<# 
   This script is used for updating assembly versions (via variable and regex) during 
   the VSTS build process and was graciously taken and modified for our purposes from GitHub here: 
   https://github.com/wulfland/ScriptRepository/blob/master/TFSBuild/TFSBuild/AssemblyVersion/Set-AssemblyVersion/Set-AssemblyVersion.ps1
#>

<#
.Synopsis
   Sets the version in all the assembly.cs, *.nuspec, servicemanfest.xml, and applicationmanifest.xml files under the source directory.

.DESCRIPTION
   A build script that can be included in TFS 2015 or Visual Studio Online (VSO) vNevt builds that update the version of all assemblies in a workspace.
   It uses the name of the build to extract the version number and updates all AssemblyInfo.cs files to use the new version.

.EXAMPLE
   Set-Version

.EXAMPLE
   Set-Version -SourceDirectory $Env:BUILD_SOURCESDIRECTORY -BuildNumber $Env:BUILD_BUILDNUMBER

.EXAMPLE
   Set-Version -SourceDirectory ".\SourceDir" -BuildNumber "CI_1.0.16300.0001" -VersionFormat "\d+\.\d+\.\d+\.\d+"
#>
[CmdletBinding(SupportsShouldProcess=$true, ConfirmImpact='Medium')]
[Alias()]
[OutputType([int])]
Param
(
    # OPTIONAL: The path to the source directory. Default $Env:BUILD_SOURCESDIRECTORY is set by TFS.
    [Parameter(Mandatory=$false, Position=0)]
    [ValidateNotNullOrEmpty()]
    [string]$SourceDirectory = $Env:BUILD_SOURCESDIRECTORY,

    # OPTIONAL: The build number. Default $Env:BUILD_BUILDNUMBER is set by TFS and must be configured according your regex.
    [Parameter(Mandatory=$false, Position=1)]
    [ValidateNotNullOrEmpty()]
    [string]$BuildNumber = $Env:BUILD_BUILDNUMBER,

    # OPTIONAL: The regex version format. Default is "\d+\.\d+\.\d+\.\d+" (e.g., major.minor.buildnumber.revision).
    [Parameter(Mandatory=$false, Position=2)]
    [ValidateNotNullOrEmpty()]
    [string]$VersionFormat = "\d+\.\d+\.\d+\.\d+"
)

<#
.Synopsis
   Sets the version in all the assembly.cs, *.nuspec, servicemanfest.xml, and applicationmanifest.xml files under the source directory.

.DESCRIPTION
   A build script that can be included in TFS 2015 or Visual Studio Online (VSO) vNevt builds that update the version 
   of all assemblies and nuspecs in a workspace. It uses the name of the build to extract the version number and 
   updates all AssemblyInfo.cs and *.nuspec files to use the new version.

.EXAMPLE
   Set-Version

.EXAMPLE
   Set-Version -SourceDirectory $Env:BUILD_SOURCESDIRECTORY -BuildNumber $Env:BUILD_BUILDNUMBER

.EXAMPLE
   Set-Version -SourceDirectory ".\SourceDir" -BuildNumber "CI_1.0.16300.0001" -VersionFormat "\d+\.\d+\.\d+\.\d+"
#>
function Set-Version
{
    [CmdletBinding(SupportsShouldProcess=$true, ConfirmImpact='Medium')]
    [Alias()]
    [OutputType([int])]
    Param
    (
        # OPTIONAL: The path to the source directory. Default $Env:BUILD_SOURCESDIRECTORY is set by TFS.
        [Parameter(Mandatory=$false, Position=0)]
        [ValidateNotNullOrEmpty()]
        [string]$SourceDirectory = $Env:BUILD_SOURCESDIRECTORY,

        # OPTIONAL: The build number. Default $Env:BUILD_BUILDNUMBER is set by TFS and must be configured according your regex.
        [Parameter(Mandatory=$false, Position=1)]
        [ValidateNotNullOrEmpty()]
        [string]$BuildNumber = $Env:BUILD_BUILDNUMBER,

        # OPTIONAL: The regex version format. Default is "\d+\.\d+\.\d+\.\d+" (e.g., major.minor.buildnumber.revision).
        [Parameter(Mandatory=$false, Position=2)]
        [ValidateNotNullOrEmpty()]
        [string]$VersionFormat = "\d+\.\d+\.\d+\.\d+"
    )

    if (-not (Test-Path $SourceDirectory)) {
        throw "The directory '$SourceDirectory' does not exist."
    }

    $Version = Get-Version -BuildNumber $BuildNumber -VersionFormat $VersionFormat
    $files = Get-Files -SourceDirectory $SourceDirectory
    Set-FileContent -Files $files -Version $Version -VersionFormat $VersionFormat
}

function Get-Version
{
    [CmdletBinding()]
    [Alias()]
    [OutputType([string])]
    Param
    (
        [Parameter(Mandatory=$true, Position=0)]
        [ValidateNotNullOrEmpty()]
        [string]$BuildNumber,
		
        [Parameter(Mandatory=$true, Position=1)]
        [ValidateNotNullOrEmpty()]
        [string]$VersionFormat
    )

    $VersionData = [regex]::matches($BuildNumber,$VersionFormat)
    if ($VersionData.Count -eq 0){
        throw "Could not find version number with format '$VersionFormat' in BUILD_BUILDNUMBER '$BuildNumber'."
    }

    return $VersionData[0]
}

function Get-Files
{
    [CmdletBinding()]
    [Alias()]
    [OutputType([System.IO.FileSystemInfo[]])]
    Param
    (
        [Parameter(Mandatory=$true, Position=0)]
        [ValidateNotNullOrEmpty()]
        [string]$SourceDirectory
    )

    # fetch all folders from the root 
    $folders = Get-ChildItem $SourceDirectory -Recurse | ?{ $_.PSIsContainer }

    # filter to only include assembly.cs, *.nuspec, servicemanfest.xml, and applicationmanifest.xml files
    return $folders | ForEach-Object { Get-ChildItem -Path $_.FullName -Recurse -Include *.nuspec,*AssemblyInfo.*,ServiceManifest.xml,ApplicationManifest.xml }
}

function Set-FileContent
{
    [CmdletBinding(SupportsShouldProcess=$true, ConfirmImpact='Medium')]
    [OutputType([int])]
    Param
    (
        [Parameter(Mandatory=$true, Position=0)]
        [System.IO.FileSystemInfo[]]$Files,
		
        [Parameter(Mandatory=$true, Position=1)]
        [string]$Version,
		
        [Parameter(Mandatory=$true, Position=2)]
        [string]$VersionFormat
    )

    foreach ($file in $Files) 
    {
        $filecontent = Get-Content $file
        if ($PSCmdlet.ShouldProcess("$file", "Set-Version"))
        {
            attrib $file -r
            $filecontent -replace $VersionFormat, $Version | Out-File $file -Encoding UTF8
            Write-Verbose -Message "Applied Version '$Version' $($file.FullName) - version applied"
        }
    }
}

if (-not ($myinvocation.line.Contains("`$here\`$sut"))) {
    Set-Version -SourceDirectory $SourceDirectory -BuildNumber $BuildNumber -VersionFormat $VersionFormat
}