﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Collections.Generic;
    using System.Fabric;
    using System.Fabric.Description;
    using System.Linq;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.BillingAPI;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ServiceHosting;
    using Microsoft.Azure.LocationServices.TimeZoneDbManager;
    using Microsoft.Azure.LocationServices.Utility;
    using Microsoft.ServiceFabric.Services.Communication.Runtime;
    using Microsoft.ServiceFabric.Services.Runtime;

    /// <summary>
    ///     The FabricRuntime creates an instance of this class for each service type instance.
    /// </summary>
    internal sealed class AzureLBS : StatelessService
    {
        /// <summary>
        /// Configuration Reader which is defaulted to a keyvault based cache solution.
        /// </summary>
        private static readonly IConfigurationReader _configurationReader = new KeyVaultConfigurationReader();

        public AzureLBS(StatelessServiceContext context)
            : base(context)
        {
        }
        
        /// <summary>
        ///     Optional override to create listeners (like tcp, http) for this service instance.
        /// </summary>
        /// <returns>The collection of listeners.</returns>
        protected override IEnumerable<ServiceInstanceListener> CreateServiceInstanceListeners()
        {
            var endpoints =
                this.Context.CodePackageActivationContext.GetEndpoints()
                     .Where(
                         endpoint =>
                             (endpoint.Protocol == EndpointProtocol.Http)
                             || (endpoint.Protocol == EndpointProtocol.Https))
                     .Select(endpoint => endpoint.Name);

            return
                endpoints.Select(
                    endpoint =>
                        new ServiceInstanceListener(
                            serviceContext =>
                                new OwinCommunicationListener(
                                    Startup.ConfigureApp,
                                    serviceContext,
                                    endpoint),
                            endpoint));
        }

        protected override async Task RunAsync(CancellationToken cancellationToken)
        {
            try
            {
                ConfigurationReader.Instance.Initialize(this.Context);
                var bootStrapConfiguration = new ServiceFabricConfigurationBootstrap(ConfigurationReader.Instance);
                await Globals.Instance.InitializeAsync(cancellationToken, bootStrapConfiguration, _configurationReader);
                await TimeZoneDb.TimeZoneDb.CreateInstance(ConfigValues.CreateForDynamicMode(bootStrapConfiguration, _configurationReader), TimeZoneLogEventCallback);
                await BillingLogWriter.CreateInstanceAsync(bootStrapConfiguration, _configurationReader);
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "proxy", "initialization");
                throw;
            }
        }

        private void TimeZoneLogEventCallback(object sender, LogEventArgs args)
        {
            var eventType = EventType.Information;
            switch (args.LogTypeId)
            {
                case LogType.Info:
                    break;
                case LogType.Error:
                    eventType = EventType.ProgramError;
                    break;
                case LogType.Verbose:
                    eventType = EventType.Verbose;
                    break;
                case LogType.Warning:
                    eventType = EventType.Warning;
                    break;
                default:
                    throw new ArgumentException($"TimeZoneLogEventCallback: unhandled type={args.LogTypeId} in switch statement");
            }

            Agent.Instance.LogEvent(eventType, args.Message, args, args.Source, timestamp: args.UtcStamp);
        }
    }
}