﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Collections.Generic;
    using System.Net.Http;
    using System.Security.Claims;
    using System.Security.Cryptography.X509Certificates;
    using System.Threading;
    using System.Threading.Tasks;
    using System.Web.Http.Filters;
    using Microsoft.Azure.LocationServices.Instrumentation;

    public class CertificateAuthenticationFilter : IAuthenticationFilter
    {
        private const string Heartbeat = "Heartbeat";

        private const string X509 = "X509";

        /// <summary>
        /// Controls whether to allow multiple instances of this filter
        /// </summary>
        public bool AllowMultiple => true;

        /// <summary>
        /// Creates the identity of the request in the context that a heart beat endpoint has been called or that 
        /// the caller as provided a client certificate for authentication.
        /// </summary>
        /// <param name="context">the current authentication context of the request.</param>
        /// <param name="cancellationToken">the cancellation token</param>
        /// <returns><see cref="Task"/></returns>
        public Task AuthenticateAsync(HttpAuthenticationContext context, CancellationToken cancellationToken)
        {
            if (context.ActionContext.ControllerContext.ControllerDescriptor.ControllerName == Heartbeat)
            {
                return Task.CompletedTask; // No authentication/authorization for heartbeat
            }

            var claimsPrincipal = context.Principal as ClaimsPrincipal;
            if (claimsPrincipal != null && claimsPrincipal.Identity.IsAuthenticated)
            {
                return Task.CompletedTask; // Already authenticated using AAD. No need for certificate fallback
            }

            Agent.Instance.LogEvent(EventType.ProgramError, "AAD authentication failed. Certificate fallback will be used.");

            try
            {
                HttpRequestMessage request = context.Request;
                var clientCert = request.GetClientCertificate();

                if (clientCert != null)
                {
                    // authenticate as a client cert
                    X509Store certStore = new X509Store(StoreName.My, StoreLocation.LocalMachine);
                    certStore.Open(OpenFlags.ReadOnly);
                    if (certStore != null)
                    {
                        X509Certificate2Collection certCollection =
                            certStore.Certificates.Find(X509FindType.FindByThumbprint, clientCert.Thumbprint, false);

                        if (certCollection != null && certCollection.Count > 0 && clientCert.Verify())
                        {
                            // authenticated!
                            var identity = new ClaimsPrincipal(
                                new ClaimsIdentity(
                                    new List<Claim>
                                    {
                                        new Claim(ClaimTypes.Name, clientCert.Thumbprint),
                                        new Claim(ClaimTypes.Authentication, bool.TrueString),
                                        new Claim(ClaimTypes.X500DistinguishedName, clientCert.Subject),
                                        new Claim(ClaimTypes.AuthenticationMethod, X509)
                                    },
                                    X509));
                            context.Principal = identity;
                        }
                    }
                }

                return Task.CompletedTask;
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "proxy", "authentication");
                throw;
            }
        }

        /// <summary>
        /// Post Processing, but this is handled at the Authorization level. Additionally there is not standard scheme for asking clients to
        /// authenticate via certificate as part of negotiation.
        /// </summary>
        /// <param name="context">authentication context</param>
        /// <param name="cancellationToken">cancellation token</param>
        /// <returns><see cref="Task"/></returns>
        public Task ChallengeAsync(HttpAuthenticationChallengeContext context, CancellationToken cancellationToken)
        {
            return Task.FromResult(false);
        }
    }
}
