﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Controllers
{
    using System;
    using System.Collections.Generic;
    using System.Globalization;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;
    using Microsoft.Azure.LocationServices.Models;
    using Microsoft.Azure.LocationServices.Utility;
    using Newtonsoft.Json;

    public class BatchController : BaseProxyController
    {
        private const int BatchSearchApiMaxQueries = 10000;
        private const int BatchRouteApiMaxQueries = 1000;
        private const string SearchApiRoot = "/search";
        private const string RouteApiRoot = "/route";

        private List<string> queries;

        #region Supported API's

        /// <summary>
        /// POST action for the Batch API.
        /// </summary>
        /// <param name="formatstring">Format for the expected response.</param>
        /// <returns>
        /// A HttpResponseMessage with [Status Code 202] - Request Accepted and a callback uri
        /// (containing the batch-id) in the Location header for fetching the batch query results later.
        /// </returns>
        [HttpPost]
        public async Task<HttpResponseMessage> PostBatch(string formatstring)
        {
            try
            {
                Versioning versionobject = new Versioning();
                BaseProxyBatch proxyObj = (BaseProxyBatch)versionobject.GetProxyObj(this.Version, BaseProxy.ProxyType.Batch, this.ProxyOperation);

                if (proxyObj != null && proxyObj is IBatchCustomSerializationProvider serializationProvider)
                {
                    this.response = await proxyObj.BatchProxyPost(
                        this.Request.RequestUri, 
                        serializationProvider.SerializeBatchQueryList(this.Request.RequestUri, this.queries), 
                        this.AccountId);
                }
                else
                {
                    this.response = this.BuildErrorResponseAndLog(new ProxyException(ProxyException.ExceptionType.ServerError, "Failed to initialize the proxy object"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        /// <summary>
        /// GET action for the Batch API.
        /// </summary>
        /// <param name="batchId">Batch id to fetch and return.</param>
        /// <returns>
        /// If batch query results are ready returns a HttpResponseMessage with [Status Code 200] and [ResponseContent] 
        /// Else returns a HttpResponseMessage with [Status Code 202] - Request Accepted and a callback uri
        /// (containing the batch-id) in the Location header for fetching the batch query results later.
        /// </returns>
        [HttpGet]
        public async Task<HttpResponseMessage> GetBatch(string batchId)
        {
            if (string.IsNullOrWhiteSpace(batchId))
            {
                return this.BuildErrorResponseAndLog(new ProxyException(ProxyException.ExceptionType.ClientParams, "Invalid BatchId - Value cannot be null/empty"));
            }

            try
            {
                Versioning versionobject = new Versioning();
                BaseProxyBatch proxyObj = (BaseProxyBatch)versionobject.GetProxyObj(this.Version, BaseProxy.ProxyType.Batch, this.ProxyOperation);

                if (proxyObj != null)
                {
                    this.response = await proxyObj.BatchProxyGet(this.Request.RequestUri, batchId);
                }
                else
                {
                    this.response = this.BuildErrorResponseAndLog(new ProxyException(ProxyException.ExceptionType.ServerError, "Failed to initialize the proxy object"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        #endregion

        #region Overrides

        internal override void ClientOverrideRequest(string query, ref Dictionary<string, List<string>> querystrings)
        {
        }

        internal override void ClientValidateParameters()
        {
            this.ValidateQueryParams();
            this.ValidateRequestBody();
        }

        /// <summary>
        /// By default all API calls (actions) will be written to billing. 
        /// Batch API behaves differently as in it ends up issuing separate requests to Azure LBS
        /// which inturn takes care of billing (on a per request basis). 
        /// </summary>
        protected override bool ShouldWriteActionToBilling()
        {
            return false;
        }

        /// <summary>
        /// By default the request body will not be parsed in BaseProxyController as part of the Initialize method. 
        /// Search Batch API behaves differently as in it needs to validate the request body before the action gets 
        /// invoked. Hence the parsing is handled in the base class.
        /// </summary>
        protected override bool ShouldParseRequestBody()
        {
            return this.Request.Method.Equals(HttpMethod.Post);
        }

        #endregion

        #region Private Helpers

        private void ValidateQueryParams()
        {
            HttpMethod requestMethodType = this.Request.Method;

            if (requestMethodType.Equals(HttpMethod.Post))
            {
                if (this.Format != "json")
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Invalid Format - Value should be json");
                }
            }
        }

        private void ValidateRequestBody()
        {
            if (this.Request.Method.Equals(HttpMethod.Post))
            {
                // First, parse the query list from the request body and validate the input format.
                try
                {
                    this.queries = this.HttpRequestContent[Constants.BatchApiQueryListKey].ToObject<List<string>>();
                }
                catch (Exception ex) when (ex is JsonSerializationException || ex is JsonReaderException)
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Invalid Request Body - Value should be a list of string queries", ex);
                }

                // Second, validate that the input batch query list has a valid count.
                int minQueryCount = 1;
                int maxQueryCOunt = this.GetMaxQueriesForRequest();

                if (this.queries.Count < minQueryCount || this.queries.Count > maxQueryCOunt)
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"Invalid Request Body - query count should be in the range: {minQueryCount}-{maxQueryCOunt}");
                }
            }
        }

        /// <summary>
        /// Calculate the max allowed queries for the current request type.
        /// </summary>
        private int GetMaxQueriesForRequest()
        {
            if (this.Request.RequestUri.AbsolutePath.StartsWith(SearchApiRoot, true, CultureInfo.InvariantCulture))
            {
                return BatchSearchApiMaxQueries;
            }

            if (this.Request.RequestUri.AbsolutePath.StartsWith(RouteApiRoot, true, CultureInfo.InvariantCulture))
            {
                return BatchRouteApiMaxQueries;
            }

            throw new ProxyException(ProxyException.ExceptionType.ServerError, "Batch is not supported for the current api.");
        }

        #endregion
    }
}