﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.Controllers
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Web.Http.Routing;
    using GeoJSON.Net;
    using Instrumentation;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Models;
    using Microsoft.Azure.LocationServices.Utility;
    using Newtonsoft.Json;
    using Newtonsoft.Json.Linq;

    public abstract class BaseProxyController : System.Web.Http.ApiController
    {
        protected HttpResponseMessage response = new HttpResponseMessage(System.Net.HttpStatusCode.BadRequest);

        protected Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>(StringComparer.OrdinalIgnoreCase);

        protected Dictionary<string, List<string>> headers = new Dictionary<string, List<string>>(StringComparer.OrdinalIgnoreCase);

        protected RequestOverride requestOverride = new RequestOverride();

        protected string AccountId;

        private readonly Stopwatch ProxyTimer;

        public BaseProxyController()
        {
            this.ProxyTimer = new Stopwatch();
            this.ProxyTimer.Start();
            this.versionObject = new Versioning();
        }

        public BaseProxyController(HttpRequestMessage request) : this()
        {
            this.Request = request;
            this.versionObject = new Versioning();
        }

        public IVersioning versionObject { get; set; }

        public IProxyOperation ProxyOperation { get; private set; }

        public string TrackingId { get; private set; }

        public string SubscriptionId { get; private set; }

        public double Version { get; private set; }

        public string Query { get; private set; }

        public string Format { get; private set; }

        public string HttpAction { get; private set; }

        public string ClientSessionId { get; private set; }

        /// <summary>
        /// Stores the content/body for POST requests as a json object.
        /// </summary>
        public JObject HttpRequestContent { get; private set; }

        /// <summary>
        /// Stores the GeoJSON geometry. Set and retrived in derived classes.
        /// </summary>
        public IGeoJSONObject GeoJsonGeometry { get; protected set; }

        public void SetProxyOperation(IProxyOperation operation)
        {
            this.ProxyOperation = operation;
        }

        public void Initialize()
        {
            // NOTE: The order of the method calls is important here.
            this.ParseRouteData();

            this.ParseRequestHeaders();

            this.ParseQueryString();

            this.ParseRequestBody();

            this.Log();

            this.ValidateParameters();

            this.OverrideRequest(this.Query, ref this.querystrings);

            this.RemoveInternalParams();
        }

        public void DeInitialize(HttpResponseMessage response = null)
        {
            if (response == null)
            {
                response = this.response;
            }

            if (response.IsSuccessStatusCode)
            {
                this.ProxyOperation.SetResultSuccess();
            }

            this.LogBilling();

            this.ProxyOperation.Log_PROXY_APIM(response.ToString(), (int)response.StatusCode);
            this.ProxyTimer.Stop();
            this.ProxyOperation.Log_Proxy_ms((int)this.ProxyTimer.ElapsedMilliseconds);
        }

        internal void SetQueryStrings(Dictionary<string, List<string>> queryStrings)
        {
            this.querystrings = queryStrings;
        }

        internal Dictionary<string, List<string>> GetQueryStrings()
        {
            return this.querystrings;
        }

        internal virtual HttpResponseMessage BuildErrorResponseAndLog(ProxyException proxyException)
        {
            Agent.Instance.LogException(proxyException, "proxy", proxyException.Type.ToString(), this.TrackingId);
            switch (proxyException.Type)
            {
                case ProxyException.ExceptionType.Unauthorized:
                    {
                        this.ProxyOperation.SetResultClientError(this.TrackingId, "invalidclientcertificate");

                        const HttpStatusCode ErrorCode = HttpStatusCode.Unauthorized;

                        var errorContract = new ErrorContract
                        {
                            Code = $"{(int)ErrorCode} {ErrorCode}",
                            Message = Constants.ErrorMessageUnauthorized
                        };

                        return this.Request.CreateResponse(ErrorCode, new ErrorContractResponse { Error = errorContract });
                    }

                case ProxyException.ExceptionType.ClientParams:
                    {
                        this.ProxyOperation.SetResultClientError(this.TrackingId, "invalidparameters");

                        const HttpStatusCode ErrorCode = HttpStatusCode.BadRequest;

                        var errorContract = new ErrorContract
                        {
                            Code = $"{(int)ErrorCode} {ErrorCode}",
                            Message = Constants.ErrorMessageBadRequest
                        };
                        
                        return this.Request.CreateResponse(ErrorCode, new ErrorContractResponse { Error = errorContract });
                    }

                case ProxyException.ExceptionType.ServerError:
                    {
                        this.ProxyOperation.SetResultFailure(this.TrackingId, "internalerror");

                        const HttpStatusCode ErrorCode = HttpStatusCode.InternalServerError;

                        var errorContract = new ErrorContract
                        {
                            Code = $"{(int)ErrorCode} {ErrorCode}",
                            Message = Constants.ErrorMessageServer
                        };

                        return this.Request.CreateResponse(ErrorCode, new ErrorContractResponse { Error = errorContract });
                    }

                case ProxyException.ExceptionType.ExtError:
                    {
                        this.ProxyOperation.SetResultFailure(this.TrackingId, "externalerror");

                        const HttpStatusCode ErrorCode = HttpStatusCode.BadGateway;

                        var errorContract = new ErrorContract
                        {
                            Code = $"{(int)ErrorCode} {ErrorCode}",
                            Message = Constants.ErrorMessageServer
                        };

                        return this.Request.CreateResponse(ErrorCode, new ErrorContractResponse { Error = errorContract });
                    }

                case ProxyException.ExceptionType.Throttled:
                    {
                        this.ProxyOperation.SetResultFailure(this.TrackingId, "requestthrottled");

                        const HttpStatusCode ErrorCode = HttpStatusCode.Forbidden;

                        var errorContract = new ErrorContract
                        {
                            Code = $"{(int)ErrorCode} {ErrorCode}",
                            Message = Constants.ErrorMessageForbidden
                        };

                        return this.Request.CreateResponse(ErrorCode, new ErrorContractResponse { Error = errorContract });
                    }

                default:
                    {
                        this.ProxyOperation.SetResultFailure(this.TrackingId, "somethingbadhappened");

                        const HttpStatusCode ErrorCode = HttpStatusCode.ServiceUnavailable;

                        var errorContract = new ErrorContract
                        {
                            Code = $"{(int)ErrorCode} {ErrorCode}",
                            Message = Constants.ErrorMessageUnknown
                        };

                        return this.Request.CreateResponse(ErrorCode, new ErrorContractResponse { Error = errorContract });
                    }
            }
        }

        internal void OverrideRequest(string query, ref Dictionary<string, List<string>> querystrings)
        {
            this.requestOverride.Override(query, ref querystrings);
            this.ClientOverrideRequest(query, ref querystrings);
        }

        internal abstract void ClientOverrideRequest(string query, ref Dictionary<string, List<string>> querystrings);
        
        internal void ValidateParameters()
        {
            this.ClientValidateParameters();
        }

        internal abstract void ClientValidateParameters();

        /// <summary>
        /// By default all API calls (actions) will be written to billing. 
        /// Child classes should override if they want a different behavior.
        /// </summary>
        protected virtual bool ShouldWriteActionToBilling()
        {
            return true;
        }

        /// <summary>
        /// By default the request body will not be parsed in BaseProxyController as part of the Initialize method. 
        /// Child classes should override if they want a different behavior. 
        /// </summary>
        protected virtual bool ShouldParseRequestBody()
        {
            return false;
        }

        protected virtual IEnumerable<Tuple<string, HttpStatusCode>> GetMethodStatus()
        {
            yield return new Tuple<string, HttpStatusCode>(this.ActionContext.ActionDescriptor.ActionName, this.response.StatusCode);
        }

        private void ParseRouteData()
        {
            this.Request.Properties.TryGetValue("MS_HttpRouteData", out var oRouteData);
            HttpRouteData routeData = (HttpRouteData)oRouteData;

            if (routeData != null)
            {
                if (routeData.Values.TryGetValue("formatstring", out var oFormatString))
                {
                    this.Format = oFormatString.ToString().ToLower();
                }

                if (routeData.Values.TryGetValue("action", out var oHttpAction))
                {
                    this.HttpAction = oHttpAction.ToString();
                }
            }
        }

        private void ParseRequestHeaders()
        {
            this.headers = this.Request.Headers.ToDictionary(x => x.Key, x => x.Value.ToList(), StringComparer.OrdinalIgnoreCase);

            string requestOriginName = string.Empty;
            if (this.headers.TryGetValue("ms-am-request-origin", out var requestOriginList))
            {
                requestOriginName = requestOriginList?.FirstOrDefault().ToLower();
            }

            this.ProxyOperation.OperationSourceName = string.IsNullOrEmpty(requestOriginName) ? "Unknown" : requestOriginName;
        }

        private void ParseQueryString()
        {
            IEnumerable<KeyValuePair<string, string>> kvpairs = this.Request.GetQueryNameValuePairs();

            for (int i = 0; i < kvpairs.Count(); i++)
            {
                KeyValuePair<string, string> kvpair = kvpairs.ElementAt(i);

                if (this.querystrings.ContainsKey(kvpair.Key))
                {
                    this.querystrings[kvpair.Key].Add(kvpair.Value);
                }
                else
                {
                    List<string> stringList = new List<string> { kvpair.Value };
                    this.querystrings.Add(kvpair.Key, stringList);
                }
            }

            this.ClientSessionId = this.Request.Headers.TryGetValues("session-id", out var clientSessionIds)
                                        ? clientSessionIds.DefaultIfEmpty("No-Session-ID").First()
                                        : "No-Session-ID";

            this.SubscriptionId = this.querystrings.TryGetValue("subscription-id", out var subscriptionId)
                                      ? subscriptionId.First()
                                      : throw new ProxyException(ProxyException.ExceptionType.ClientParams, "subscription-id invalid");

            this.TrackingId = this.querystrings.TryGetValue("tracking-id", out var trackingId)
                                  ? trackingId.First()
                                  : throw new ProxyException(ProxyException.ExceptionType.ClientParams, "tracking-id invalid");

            this.Version = this.querystrings.TryGetValue("api-version", out var version) && double.TryParse(version.First(), out var temp)
                               ? temp
                               : throw new ProxyException(ProxyException.ExceptionType.ClientParams, "api-version invalid");

            if (this.querystrings.TryGetValue("query", out var query))
            {
                this.Query = query.First();
            }
        }

        private void ParseRequestBody()
        {
            if (!this.ShouldParseRequestBody())
            {
                return;
            }

            // Content/Content-Headers/Content-Type should not be null
            if (this.Request.Content?.Headers?.ContentLength == 0 || this.Request.Content?.Headers?.ContentType == null)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Content/Content-Type should not be null.");
            }

            // Request has content and an associated type
            string requestContentType = this.Request.Content.Headers.ContentType.MediaType;

            switch (requestContentType)
            {
                case Constants.JsonMediaType:
                {
                    try
                    {
                        this.HttpRequestContent = this.Request.Content.ReadAsAsync<JObject>().GetAwaiter().GetResult();
                    }
                    catch (Exception ex) when (ex is JsonReaderException)
                    {
                        throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Request body is not a valid json", ex);
                    }

                    break;
                }

                default:
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"Content-type {requestContentType} is invalid. Available types: application/json");
            }
        }

        private void Log()
        {
            this.ProxyOperation.Log_SubscriptionId(this.SubscriptionId);
            this.ProxyOperation.Log_TrackingId(this.TrackingId);
            this.ProxyOperation.Log_APIM_PROXY(this.Request.RequestUri.ToString());
            this.ProxyOperation.Log_ResourceId(this.ParseResourceId());
            this.ProxyOperation.Log_ClientSessionId(this.ClientSessionId);
        }

        private void RemoveInternalParams()
        {
            this.querystrings.Remove(Constants.SubscriptionIdKey);
            this.querystrings.Remove(Constants.SubscriptionKey);
            this.querystrings.Remove(Constants.TrackingIdKey);
            this.querystrings.Remove(Constants.ApiVersionKey);
            this.querystrings.Remove(Constants.QueryKey);
            this.querystrings.Remove(Constants.AccountIdKey);
            this.querystrings.Remove(Constants.BatchIdKey);
            this.querystrings.Remove(Constants.BatchQueryIndexKey);
        }

        /// <summary>
        /// Parses a resource id from the account-id field.
        /// If there are any errors or anomalies, the resource id is assumed to be empty.
        /// </summary>
        /// <returns>Parsed resource id, otherwise empty string</returns>
        private string ParseResourceId()
        {
            if (!(this.querystrings.TryGetValue("account-id", out var accountIdWrapper) && accountIdWrapper.Count == 1))
            {
                return string.Empty;
            }

            var accountIdParam = accountIdWrapper.Single();

            var result = Helpers.ParseResourceIdFromAccountId(accountIdParam);

            if (!string.IsNullOrEmpty(result))
            {
                this.AccountId = accountIdParam;
            }

            return result;
        }

        /// <summary>
        /// Log Billing.
        /// If it is in hybrid mode (the header contains the hybrid signal), ignore billing for imagery,
        /// and change billing meter to hybrid for vector tiles
        /// </summary>
        private void LogBilling()
        {
            if (!string.IsNullOrEmpty(AccountId) && this.ShouldWriteActionToBilling())
            {
                List<string> values = new List<string>();
                this.headers.TryGetValue("billing", out values);

                if (values != null && values.Contains("hybrid-vector"))
                {
                    var methodStatus = new List<Tuple<string, HttpStatusCode>>() { new Tuple<string, HttpStatusCode>("GetHybridTile", this.response.StatusCode) };
                    BillingAPI.BillingLogWriter.Instance.WriteLog(this.AccountId, methodStatus, this.TrackingId);
                }
                else if (values != null && values.Contains("hybrid-imagery"))
                {
                    return;
                }
                else
                {
                    BillingAPI.BillingLogWriter.Instance.WriteLog(this.AccountId, this.GetMethodStatus().ToList(), this.TrackingId);
                }
            }
        }
    }
}