﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Controllers
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net.Http;
    using GeoJSON.Net;
    using Microsoft.Azure.LocationServices.Utility;
    using Newtonsoft.Json;
    using Newtonsoft.Json.Linq;

    public class GeometryHelper
    {
        public static IGeoJSONObject ParseAndValidateGeometry(
            HttpMethod method, 
            JObject content,
            string contentKey,
            Dictionary<string, List<string>> queryParams)
        {
            // 1. Extract the geometry json.
            string geometryJson = ExtractGeometry(method, content, contentKey, queryParams);

            // 2. Deserialize the geometry json.
            IGeoJSONObject geometry = ParseGeometry(geometryJson);

            // 3. Validate the geometry json.
            ValidateGeometry(geometry);

            return geometry;
        }

        private static string ExtractGeometry(
            HttpMethod method, 
            JObject content,
            string contentKey,
            Dictionary<string, List<string>> queryParams)
        {
            string geometryJson = null;

            if (method.Equals(HttpMethod.Get))
            {
                // For GET request, extract geometry from query params
                geometryJson = 
                    queryParams.TryGetValue(contentKey, out var geometryStringList) 
                    ? geometryStringList.First()
                    : throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"Input {contentKey} is missing");
            }

            if (method.Equals(HttpMethod.Post))
            {
                // For POST request, extract geometry from request body
                geometryJson = content.Value<JObject>(contentKey)?.ToString();
            }

            return geometryJson;
        }

        private static IGeoJSONObject ParseGeometry(string geometryJsonString)
        {
            IGeoJSONObject geoJsonGeometry;

            try
            {
                JObject geometryJson = JObject.Parse(geometryJsonString);
                geoJsonGeometry = geometryJson.Deserialize();
            }
            catch (Exception ex) when (ex is JsonSerializationException || ex is JsonReaderException || ex is ArgumentNullException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Input geometry is missing or invalid", ex);
            }

            return geoJsonGeometry;
        }

        private static void ValidateGeometry(IGeoJSONObject geoJsonGeometry)
        {
            if (!geoJsonGeometry.IsValid())
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Input geometry is not a valid GeoJSON");
            }
        }
    }
}