﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
namespace Microsoft.Azure.LocationServices.Controllers
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;
    using GeoAPI.Geometries;
    using Microsoft.Azure.LocationServices.Geography;
    using Microsoft.Azure.LocationServices.Instrumentation;

    public class GeofenceController : BaseProxyController
    {
        private const string DefaultSearchRadiusInMeters = "100";

        private static readonly HashSet<string> ValidFormatSet = new HashSet<string>
        {
            "json"
        };

        private IPoint userLocation;

        private string udid;

        private DateTime usertime;

        private double searchRadiusInMeters;

        private GeofenceResponseBody responseBody = new GeofenceResponseBody();

        #region Supported API's

        [HttpGet]
        public async Task<HttpResponseMessage> Geofence(string formatstring)
        {
            if (this.response.Content != null)
            {
                return response;
            }

            try
            {
                var proxyObj = (BaseProxyGeofence)this.versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Geofence, this.ProxyOperation);

                if (proxyObj != null)
                {
                    this.response = await proxyObj.SearchGeoFence(formatstring, this.AccountId, this.udid, this.userLocation, usertime, searchRadiusInMeters, this.responseBody);
                }
                else
                {
                    this.response = SetResponseErrorInfo(new ProxyException(ProxyException.ExceptionType.ServerError, "Failed to initialize the proxy object"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = SetResponseErrorInfo(ex);
            }
            catch (Exception ex)
            {
                this.response = SetResponseErrorInfo(new ProxyException(ProxyException.ExceptionType.ServerError, ex.Message, ex));
            }

            return this.response;
        }

        #endregion

        #region Overrides

        internal override void ClientOverrideRequest(string query, ref Dictionary<string, List<string>> querystrings)
        {
        }

        internal override void ClientValidateParameters()
        {
            try
            {
                this.ValidateRequest();
            }
            catch (ProxyException ex)
            {
                this.response = SetResponseErrorInfo(ex);
            }
            catch (Exception ex)
            {
                this.response = SetResponseErrorInfo(new ProxyException(ProxyException.ExceptionType.ServerError, ex.Message, ex));
            }
        }

        #endregion

        #region Private Helpers

        private void ValidateRequest()
        {
            // Account ID
            if (string.IsNullOrWhiteSpace(this.AccountId))
            {
                // Running locally? (We need to pass in some unique id for account id as this determines the destination container)
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Account Id should not be null.");
            }

            // Format
            if (!GeofenceController.ValidFormatSet.Contains(this.Format))
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "invalid format string");
            }

            // UDID
            var udid = string.Empty;

            if (this.querystrings.TryGetValue("udid", out var valueList))
            {
                udid = valueList.FirstOrDefault();
            }

            if (string.IsNullOrEmpty(udid))
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "udid should not be null.");
            }

            string blobExtension = udid.Split('.').Last().ToLowerInvariant();
            if (blobExtension != "geojson")
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "udid is invalid.");
            }

            this.udid = udid;

            // User Request Time
            string usertime = string.Empty;

            if (this.querystrings.TryGetValue("usertime", out valueList))
            {
                usertime = valueList.FirstOrDefault();
            }

            if (string.IsNullOrEmpty(usertime))
            {
                this.usertime = DateTime.UtcNow;
            }
            else
            {
                try
                {
                    var datetime = DateTime.Parse(usertime);

                    if (datetime.Kind == DateTimeKind.Unspecified)
                    {
                        this.usertime = DateTime.SpecifyKind(datetime, DateTimeKind.Utc);
                    }
                    else
                    {
                        this.usertime = datetime.ToUniversalTime();
                    }
                }
                catch (Exception ex)
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "usertime is invalid", ex);
                }
            }

            // User asset position 
            var latValue = string.Empty;
            var lonValue = string.Empty;

            if (this.querystrings.TryGetValue("lat", out valueList))
            {
                latValue = valueList.FirstOrDefault();
            }

            if (this.querystrings.TryGetValue("lon", out valueList))
            {
                lonValue = valueList.FirstOrDefault();
            }

            var coordinates = new Coordinates(latValue, lonValue);

            if (coordinates.IsNull)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "lat and/or lon parameters are invalid.");
            }

            this.userLocation = new NetTopologySuite.Geometries.Point(new Coordinate(coordinates.Longitude, coordinates.Latitude));

            // Search Radius
            string radius = string.Empty;

            if (this.querystrings.TryGetValue("searchradius", out valueList))
            {
                radius = valueList.FirstOrDefault();
            }

            if (string.IsNullOrEmpty(radius))
            {
                radius = DefaultSearchRadiusInMeters;
            }

            try
            {
                this.searchRadiusInMeters = Convert.ToDouble(radius);
            }
            catch (Exception ex)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "searchradius is invalid.", ex);
            }

            if (searchRadiusInMeters > MicrosoftV1GeofenceProxy.MaximumSearchRadiusInMeters)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"searchradius should be less or equal to {MicrosoftV1GeofenceProxy.MaximumSearchRadiusInMeters} meters.");
            }
        }

        private T GetGeometryFromRequestBody<T>() where T : class, IGeometry 
        {
            // Content/Content-Type
            if (this.Request.Content?.Headers?.ContentLength == 0 || this.Request.Content?.Headers?.ContentType == null)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Content/Content-Type should not be null when lat/lon is not present in query parameter.");
            }

            try
            {
                // If location geometry is a geojson in request body
                var reader = new NetTopologySuite.IO.GeoJsonReader();

                var feature = reader.Read<NetTopologySuite.Features.Feature>(this.HttpRequestContent.ToString());

                var geometry = feature.Geometry as T;

                if (geometry == null)
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"The geometry type in the request body is invalid.");
                }

                return geometry;
            }
            catch (Exception)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "The geojson in the request body is invalid.");
            }
        }

        private HttpResponseMessage SetResponseErrorInfo(ProxyException ex)
        {
            Agent.Instance.LogException(ex, "proxy", ex.Type.ToString(), this.TrackingId);

            var issue = new Issue();

            issue.type = ex.Type.ToString();

            issue.message = ex.Message;

            issue.innerException = ex.InnerException;

            responseBody.issues.Add(issue);

            switch (ex.Type)
            {
                case ProxyException.ExceptionType.Unauthorized:
                    {
                        this.ProxyOperation.SetResultClientError(this.TrackingId, "invalidclientcertificate");

                        responseBody.response_code = HttpStatusCode.Unauthorized;

                        break;
                    }

                case ProxyException.ExceptionType.ClientParams:
                    {
                        this.ProxyOperation.SetResultClientError(this.TrackingId, "invalidparameters");

                        responseBody.response_code = HttpStatusCode.BadRequest;

                        break;
                    }

                case ProxyException.ExceptionType.ServerError:
                    {
                        this.ProxyOperation.SetResultFailure(this.TrackingId, "internalerror");

                        responseBody.response_code = HttpStatusCode.InternalServerError;

                        break;
                    }

                case ProxyException.ExceptionType.ExtError:
                    {
                        this.ProxyOperation.SetResultFailure(this.TrackingId, "externalerror");

                        responseBody.response_code = HttpStatusCode.BadGateway;

                        break;
                    }

                case ProxyException.ExceptionType.Throttled:
                    {
                        this.ProxyOperation.SetResultFailure(this.TrackingId, "requestthrottled");

                        responseBody.response_code = HttpStatusCode.Forbidden;

                        break;
                    }

                default:
                    {
                        this.ProxyOperation.SetResultFailure(this.TrackingId, "somethingbadhappened");

                        responseBody.response_code = HttpStatusCode.ServiceUnavailable;

                        break;
                    }
            }

            return BaseProxyGeofence.CreateResponseMessage(responseBody);
        }

        #endregion
    }
}
