﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Controllers
{
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;
    using Microsoft.Azure.LocationServices.Utility;

    public class MapCopyrightController : BaseProxyController
    {
        private const int INCLUSIVE_MIN_ZOOM_INDEX = 0;
        private const int INCLUSIVE_MAX_ZOOM_INDEX = 18;

        [HttpGet]
        public async Task<HttpResponseMessage> GetCopyrightWorld(string formatstring)
        {
            try
            {               
                BaseProxyMap proxyObj = (BaseProxyMap)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Map, this.ProxyOperation);
                if (proxyObj != null)
                {
                    this.response =
                        await
                            proxyObj.MapCopyright(
                                    BaseProxyMap.MapType.WorldCopyright,
                                    formatstring,
                                    this.querystrings);
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        [HttpGet]
        public async Task<HttpResponseMessage> GetCopyrightTile(string formatstring)
        {
            try
            {
                BaseProxyMap proxyObj = (BaseProxyMap)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Map, this.ProxyOperation);
                if (proxyObj != null)
                {
                    List<string> zoom = new List<string>();
                    List<string> x = new List<string>();
                    List<string> y = new List<string>();

                    querystrings.TryGetValue("zoom", out zoom);
                    querystrings.TryGetValue("x", out x);
                    querystrings.TryGetValue("y", out y);

                    if ((zoom == null) || (x == null) || (y == null))
                    {
                        throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "zoom||x||y invalid"));
                    }

                    querystrings.Remove("zoom");
                    querystrings.Remove("x");
                    querystrings.Remove("y");

                    this.response =
                        await
                            proxyObj.MapCopyright(
                                BaseProxyMap.MapType.TileCopyright,
                                formatstring,
                                this.querystrings,
                                zoom.First(),
                                x.First(),
                                y.First());
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        [HttpGet]
        public async Task<HttpResponseMessage> GetCopyrightBounding(string formatstring)
        {
            try
            {
                BaseProxyMap proxyObj = (BaseProxyMap)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Map, this.ProxyOperation);
                if (proxyObj != null)
                {
                    List<string> mincoordinates = new List<string>();
                    List<string> maxcoordinates = new List<string>();

                    querystrings.TryGetValue("mincoordinates", out mincoordinates);
                    querystrings.TryGetValue("maxcoordinates", out maxcoordinates);

                    if ((mincoordinates == null) || (maxcoordinates == null))
                    {
                        throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "mincoordinates||maxcoordinates invalid"));
                    }

                    querystrings.Remove("mincoordinates");
                    querystrings.Remove("maxcoordinates");

                    this.response =
                        await
                            proxyObj.MapCopyright(
                                BaseProxyMap.MapType.BoundingCopyright,
                                formatstring,
                                this.querystrings,
                                string.Empty,
                                WebUtility.UrlDecode(mincoordinates.First()),
                                WebUtility.UrlDecode(maxcoordinates.First()));
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        public async Task<HttpResponseMessage> GetCopyrightCaption(string formatstring)
        {
            HttpResponseMessage response = new HttpResponseMessage(HttpStatusCode.BadRequest);
            try
            {
                BaseProxyMap proxyObj = (BaseProxyMap)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Map, this.ProxyOperation);
                if (proxyObj != null)
                {
                    this.response =
                        await
                            proxyObj.MapCopyright(
                                BaseProxyMap.MapType.CaptionCopyright,
                                formatstring,
                                this.querystrings);
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        internal override void ClientValidateParameters()
        {
            if (this.Format != "json" && this.Format != "xml")
            {
                throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "invalid format string"));
            }

            if (this.HttpAction == "GetCopyrightTile")
            {
                if (!querystrings.ContainsKey("zoom") || !querystrings.ContainsKey("x") || !querystrings.ContainsKey("y"))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "zoom||x||y missing"));
                }

                string xParamVal = querystrings["x"]?.FirstOrDefault();
                string yParamVal = querystrings["y"]?.FirstOrDefault();
                string zoomParamVal = querystrings["zoom"]?.FirstOrDefault();
                var copyrightTileValidator = new TileValidator(INCLUSIVE_MIN_ZOOM_INDEX, INCLUSIVE_MAX_ZOOM_INDEX);
                if (!copyrightTileValidator.IsValidTile(xParamVal, yParamVal, zoomParamVal))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, $"(x:{xParamVal}, y:{yParamVal}, z:{zoomParamVal}) is not a valid tile."));
                }
            }

            if (this.HttpAction == "GetCopyrightBounding")
            {
                if (!querystrings.ContainsKey("mincoordinates") || !querystrings.ContainsKey("maxcoordinates"))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "mincoordinates||maxcoordinates missing"));
                }
            }
        }

        internal override void ClientOverrideRequest(string query, ref Dictionary<string, List<string>> querystrings)
        {
        }
    }
}