﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Controllers
{
    using System.Collections.Generic;
    using System.Linq;
    using System.Net.Http;
    using System.Net.Http.Headers;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Utility;

    public class MapTileController : BaseProxyController
    {
        private const int INCLUSIVE_MIN_RASTER_ZOOM_INDEX = 0;
        private const int INCLUSIVE_MAX_RASTER_ZOOM_INDEX = 18;
        private const int INCLUSIVE_MIN_VECTOR_ZOOM_INDEX = 0;
        private const int INCLUSIVE_MAX_VECTOR_ZOOM_INDEX = 20;

        public async Task<HttpResponseMessage> GetMapTile(string formatstring)
        {
            try
            {
                BaseProxyMap proxyObj = (BaseProxyMap)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Map, this.ProxyOperation);
                if (proxyObj != null)
                {
                    List<string> layer = new List<string>();
                    List<string> style = new List<string>();
                    List<string> zoom = new List<string>();
                    List<string> x = new List<string>();
                    List<string> y = new List<string>();

                    querystrings.TryGetValue("layer", out layer);
                    querystrings.TryGetValue("style", out style);
                    querystrings.TryGetValue("zoom", out zoom);
                    querystrings.TryGetValue("x", out x);
                    querystrings.TryGetValue("y", out y);

                    querystrings.Remove("layer");
                    querystrings.Remove("style");
                    querystrings.Remove("zoom");
                    querystrings.Remove("x");
                    querystrings.Remove("y");

                    //Bug 1919255: Remove view param.
                    querystrings.Remove("view");

                    if (formatstring == "pbf")
                    {
                       this.response = await proxyObj.MapTileVector(
                                                    layer.First(),
                                                    style.First(),
                                                    zoom.First(),
                                                    x.First(),
                                                    y.First(),
                                                    formatstring,
                                                    this.querystrings);

                        if (this.response != null && this.response.Content != null)
                        {
                            this.response.Content.Headers.ContentType = new MediaTypeHeaderValue("application/vnd.mapbox-vector-tile");
                        }
                    }
                    else
                    {
                        this.response = await proxyObj.MapTileRaster(
                                                     layer.First(),
                                                     style.First(),
                                                     zoom.First(),
                                                     x.First(),
                                                     y.First(),
                                                     formatstring,
                                                     this.querystrings);
                    }
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        public async Task<HttpResponseMessage> GetMapTileStatic(string formatstring)
        {
            try
            {
                BaseProxyMap proxyObj = (BaseProxyMap)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Map, this.ProxyOperation);
                if (proxyObj != null)
                {
                    this.response = await proxyObj.MapStatic(formatstring, this.querystrings);
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        internal override void ClientValidateParameters()
        {
            if (this.HttpAction == "GetMapTileStatic")
            {
                if (this.Format != "png")
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "invalid format string"));
                }

                if ((querystrings.ContainsKey("center") || querystrings.ContainsKey("width") || querystrings.ContainsKey("height")) && querystrings.ContainsKey("bbox"))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "bbox may not be used in conjunction with zoom, center, height or width"));
                }
            }
            
            if (this.HttpAction == "GetMapTile")
            {
                if (this.Format != "png" && this.Format != "pbf")
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "invalid format string"));
                }

                if (!querystrings.ContainsKey("layer") || !querystrings.ContainsKey("style") || !querystrings.ContainsKey("zoom") || !querystrings.ContainsKey("x") || !querystrings.ContainsKey("y"))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "layer||style||zoom||x||y invalid"));
                }

                string xParamVal = querystrings["x"]?.FirstOrDefault();
                string yParamVal = querystrings["y"]?.FirstOrDefault();
                string zoomParamVal = querystrings["zoom"]?.FirstOrDefault();

                var mapVectorTileValidator = new TileValidator(INCLUSIVE_MIN_VECTOR_ZOOM_INDEX, INCLUSIVE_MAX_VECTOR_ZOOM_INDEX);
                if (this.Format == "pbf" && !mapVectorTileValidator.IsValidTile(xParamVal, yParamVal, zoomParamVal))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, $"(x:{xParamVal}, y:{yParamVal}, z:{zoomParamVal}) is not a valid tile."));
                }

                var mapRasterTileValidator = new TileValidator(INCLUSIVE_MIN_RASTER_ZOOM_INDEX, INCLUSIVE_MAX_RASTER_ZOOM_INDEX);
                if (this.Format == "png" && !mapRasterTileValidator.IsValidTile(xParamVal, yParamVal, zoomParamVal))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, $"(x:{xParamVal}, y:{yParamVal}, z:{zoomParamVal}) is not a valid tile."));
                }
            }
        }

        internal override void ClientOverrideRequest(string query, ref Dictionary<string, List<string>> querystrings)
        {
        }
    }
}