﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Controllers
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;
    using GeoJSON.Net;
    using GeoJSON.Net.Feature;
    using GeoJSON.Net.Geometry;
    using Microsoft.Azure.LocationServices.Geography;
    using Microsoft.Azure.LocationServices.Utility;
    using Newtonsoft.Json;
    using Newtonsoft.Json.Linq;

    public class RouteCalculateController : BaseProxyController
    {
        private List<Point> geoJSONPointList = new List<Point>();
        [HttpGet]
        public async Task<HttpResponseMessage> GetRouteCalculate(string formatstring)
        {
            try
            {
                BaseProxyRoute proxyObj = (BaseProxyRoute)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Route, this.ProxyOperation);
                if (proxyObj != null)
                {
                    this.response =
                        await
                            proxyObj.RouteProxy(
                                BaseProxyRoute.RouteType.Normal,
                                WebUtility.UrlDecode(this.Query),
                                formatstring,
                                this.querystrings);
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        [HttpPost]
        public async Task<HttpResponseMessage> PostRouteCalculate(string formatstring)
        {
            try
            {
                BaseProxyRoute proxyObj = (BaseProxyRoute)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Route, this.ProxyOperation);
                if (proxyObj != null)
                {
                    JArray cArray = new JArray();

                    //create JObject to send to TomTom
                    foreach (Point item in geoJSONPointList)
                    {
                        JObject position = new JObject(new JProperty("latitude", item.Coordinates.Latitude), new JProperty("longitude", item.Coordinates.Longitude));
                        cArray.Add(position);
                    }

                    JObject requestBody = new JObject(new JProperty("supportingPoints", cArray));
                    this.response =
                        await
                            proxyObj.RouteProxy(
                                BaseProxyRoute.RouteType.Reconstruct,
                                WebUtility.UrlDecode(this.Query),
                                formatstring,
                                this.querystrings,
                                requestBody.ToString());
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        internal override void ClientValidateParameters()
        {
            if (string.IsNullOrEmpty(this.Query))
            {
                throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "query param is empty"));
            }
            else
            {
                try
                {
                    Coordinates.CreateListOfCoordinates(this.Query);
                }
                catch
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "invalid query string"));
                }
            }

            if (this.querystrings.Count != 0 && this.querystrings.ContainsKey("departAt"))
            {
                try
                {
                    DateTime departAt = DateTime.SpecifyKind(DateTime.Parse(this.querystrings["departAt"].First()), DateTimeKind.Utc);
                    DateTime currentDateTime = TimeProvider.UtcNow();
                    if (currentDateTime > departAt)
                    {
                        throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "departAt must be a in the future."));
                    }
                }
                catch (Exception ex) when (ex.Message != "departAt must be a in the future.")
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "departAt is not in a valid datetime format"));
                }
            }

            if (this.Format != "json" && this.Format != "xml")
            {
                throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "invalid format string"));
            }

            if (this.HttpRequestContent != null)
            {
                this.ValidateRequestBody();
            }
        }

        internal override void ClientOverrideRequest(string query, ref Dictionary<string, List<string>> querystrings)
        {
        }

        /// <summary>
        /// By default the request body will not be parsed in BaseProxyController as part of the Initialize method. 
        /// Route Calculate API behaves differently as in it needs to validate the request body before the action gets 
        /// invoked. Hence the parsing is handled in the base class.
        /// </summary>
        protected override bool ShouldParseRequestBody()
        {
            return this.Request.Method.Equals(HttpMethod.Post);
        }

        private void ValidateRequestBody()
        {
            try
            {
                //Seperate and set CoordinateList for later use
                foreach (JObject item in this.HttpRequestContent.Value<JArray>("coordinateList"))
                {
                    JObject jSONPoint = item.Value<JObject>("position");
                    Feature geoJSON = (Feature)jSONPoint.Deserialize();
                    if (geoJSON.Geometry.Type != GeoJSONObjectType.Point)
                    {
                        throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Input position does not have a valid GeoJSON point");
                    }
                    else
                    {
                        this.geoJSONPointList.Add((Point)geoJSON.Geometry);
                    }
                }

                if (this.geoJSONPointList.Count < 2)
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "CoordinateList must contain at least two valid GeoJSON point");
                }
            }
            catch (Exception ex) when (ex is JsonSerializationException || ex is JsonReaderException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Input does not contain a valid CoordinateList");
            }
        }
    }
}