﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Controllers
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Text;
    using System.Threading.Tasks;
    using System.Web.Http;
    using GeoJSON.Net.Geometry;
    using Microsoft.Azure.LocationServices.Utility;
    using Newtonsoft.Json;
    
    public class RouteMatrixController : BaseProxyController
    {
        private const int CurrentMaxCellLimit = 700;
        
        private static readonly HashSet<string> ValidFormatSet = new HashSet<string>
        {
            "json"
        };

        private readonly List<Position> geoJSONOriginPositions = new List<Position>();

        private readonly List<Position> geoJSONDestinationPositions = new List<Position>();

        private List<int> responseStatusCodes;
        private string matrixRequestId;

        [HttpPost]
        public async Task<HttpResponseMessage> PostMatrixRouteCalculate(string formatstring)
        {
            try
            {
                BaseProxyRoute proxyObj = (BaseProxyRoute)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Route, this.ProxyOperation);
                if (proxyObj != null && proxyObj is IRouteCustomSerializationProvider serializationProvider)
                {
                    var forceAsync = false;
                    if (this.querystrings.TryGetValue(Constants.ForceAsyncKey, out var forcedAsyncValues))
                    {
                        if (!bool.TryParse(forcedAsyncValues.FirstOrDefault(), out forceAsync))
                        {
                            throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"{Constants.ForceAsyncKey} param couldn't be cast to boolean.");
                        }

                        this.querystrings.Remove(Constants.ForceAsyncKey);
                    }

                    this.response = await proxyObj.RouteMatrixProxyPost(
                        this.Request.RequestUri,
                        formatstring,
                        this.querystrings,
                        serializationProvider.ConvertLBSPositionsIntoTargetBody(this.geoJSONOriginPositions, this.geoJSONDestinationPositions),
                        this.AccountId,
                        this.TrackingId,
                        forceAsync);

                    var responseCode = this.response.StatusCode;
                    if (responseCode == HttpStatusCode.OK)
                    {
                        try
                        {
                            this.responseStatusCodes = serializationProvider.GetMatrixResponseCodesFromClientResponse(await this.response.Content.ReadAsStringAsync());
                        }
                        catch (FormatException fex)
                        {
                            throw new ProxyException(ProxyException.ExceptionType.Other, fex.Message);
                        }
                    }
                    else if (responseCode != HttpStatusCode.Accepted && responseCode != HttpStatusCode.SeeOther)
                    {
                        this.responseStatusCodes = new List<int>() { (int)responseCode };
                    }
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object."));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        [HttpGet]
        public async Task<HttpResponseMessage> GetMatrixRoute(string formatstring)
        {
            try
            {
                BaseProxyRoute proxyObj = (BaseProxyRoute)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Route, this.ProxyOperation);
                if (proxyObj != null)
                {
                    this.response = await proxyObj.RouteMatrixProxyGet(this.Request.RequestUri, this.matrixRequestId);
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        internal override void ClientOverrideRequest(string query, ref Dictionary<string, List<string>> querystrings)
        {
        }

        internal override void ClientValidateParameters()
        {
            this.ValidateQueryParams();
            this.ValidateRequestBody();
        }

        protected override bool ShouldWriteActionToBilling()
        {
            return this.responseStatusCodes != null;
        }

        /// <summary>
        /// By default the request body will not be parsed in BaseProxyController as part of the Initialize method. 
        /// Route Matrix API behaves differently as in it needs to validate the request body before the action gets 
        /// invoked. Hence the parsing is handled in the base class.
        /// </summary>
        protected override bool ShouldParseRequestBody()
        {
            return this.Request.Method.Equals(HttpMethod.Post);
        }

        private void ValidateQueryParams()
        {
            HttpMethod requestMethodType = this.Request.Method;

            if (requestMethodType.Equals(HttpMethod.Get))
            {
                if (this.querystrings.TryGetValue(Constants.RouteMatrixResponseUniqueIdKey, out var requestIdList))
                {
                    this.matrixRequestId = requestIdList.FirstOrDefault();
                }
                else
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "No RequestId is found in the parameters.", new ArgumentException(nameof(this.matrixRequestId)));
                }
            }

            if (!RouteMatrixController.ValidFormatSet.Contains(this.Format))
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "invalid format string");
            }
        }

        private void ValidateRequestBody()
        {
            if (this.Request.Method.Equals(HttpMethod.Get))
            {
                return;
            }

            var errorBuilder = new StringBuilder();

            try
            {
                foreach (var property in this.HttpRequestContent.Properties())
                {
                    if (property.Name == "origins")
                    {
                        this.geoJSONOriginPositions.AddRange(JsonConvert.DeserializeObject<MultiPoint>(property.Value.ToString()).Coordinates.Select(mp => mp.Coordinates as Position));
                    }
                    else if (property.Name == "destinations")
                    {
                        this.geoJSONDestinationPositions.AddRange(JsonConvert.DeserializeObject<MultiPoint>(property.Value.ToString()).Coordinates.Select(mp => mp.Coordinates as Position));
                    }
                    else
                    {
                        errorBuilder.AppendLine($"{property.Name} is not a valid property in the content JSON.");
                    }
                }
            }
            catch (Exception ex)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Input does not conform to expected format.", ex);
            }

            var numberOfCells = this.geoJSONOriginPositions.Count * this.geoJSONDestinationPositions.Count;

            if (numberOfCells < 1 || numberOfCells > CurrentMaxCellLimit)
            {
                errorBuilder.AppendLine($"Number of cells must be between 1 and {CurrentMaxCellLimit}");
            }

            if (errorBuilder.Length > 0)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, errorBuilder.ToString());
            }
        }
    }
}
