﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Controllers.Search
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;
    using Microsoft.Azure.LocationServices.Utility;
    using Newtonsoft.Json;
    using Newtonsoft.Json.Linq;

    public class SearchAlongRouteController : BaseProxyController
    {
        /// <summary>
        /// POST action for the Search Along Route API.
        /// </summary>
        /// <param name="formatstring">Format for the expected response.</param>
        /// <returns>A HttpResponseMessage with the search results.</returns>
        [HttpPost]
        public async Task<HttpResponseMessage> PostSearchAlongRoute(string formatstring)
        {
            return await this.SearchAlongRoute(formatstring);
        }

        /// <summary>
        /// GET action for the Search Along Route API.
        /// </summary>
        /// <param name="formatstring">Format for the expected response.</param>
        /// <returns>A HttpResponseMessage with the search results.</returns>
        [HttpGet]
        public async Task<HttpResponseMessage> GetSearchAlongRoute(string formatstring)
        {
            return await this.SearchAlongRoute(formatstring);
        }

        internal override void ClientOverrideRequest(string query, ref Dictionary<string, List<string>> querystrings)
        {
        }

        internal override void ClientValidateParameters()
        {
            this.ValidateRequest();
        }

        /// <summary>
        /// By default the request body will not be parsed in BaseProxyController as part of the Initialize method. 
        /// Search Along Route API behaves differently as in it needs to validate the request body before the action gets 
        /// invoked. Hence the parsing is handled in the base class.
        /// </summary>
        protected override bool ShouldParseRequestBody()
        {
            return this.Request.Method.Equals(HttpMethod.Post);
        }

        private async Task<HttpResponseMessage> SearchAlongRoute(string formatstring)
        {
            try
            {
                BaseProxySearch proxyObj = (BaseProxySearch)this.versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Search, this.ProxyOperation);

                if (proxyObj != null && proxyObj is ISearchCustomSerializationProvider serializationProvider)
                {
                    this.response = await proxyObj.SearchProxy(
                                        BaseProxySearch.SearchType.AlongRoute,
                                        WebUtility.UrlDecode(this.Query),
                                        formatstring,
                                        this.querystrings,
                                        serializationProvider.SerializeGeoJsonGeometry(BaseProxySearch.SearchType.AlongRoute, this.GeoJsonGeometry));
                }
                else
                {
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, "Failed to initialize the proxy object");
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        private void ValidateRequest()
        {
            // Query
            if (string.IsNullOrEmpty(this.Query))
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Invalid query");
            }

            // Format
            if (this.Format != "json" && this.Format != "xml")
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Invalid format string");
            }

            // Max Detour Time
            ValidationHelper.ValidateParameterWithRange(Constants.MaxDetourTimeParamKey, true, this.querystrings, 0, 3600);

            // Limit
            ValidationHelper.ValidateParameterWithRange(Constants.LimitParamKey, false, this.querystrings, 1, 20);

            // Route Geometry
            this.GeoJsonGeometry = GeometryHelper.ParseAndValidateGeometry(
                this.Request.Method, 
                this.HttpRequestContent, 
                Constants.RouteQueryParamKey, 
                this.querystrings);
        }
    }
}