﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Controllers.Search
{
    using System.Collections.Generic;
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;
    using Microsoft.Azure.LocationServices.Utility;

    public class SearchByGeometryController : BaseProxyController
    {
        /// <summary>
        /// POST action for the Geometry Search API.
        /// </summary>
        /// <param name="formatstring">Format for the expected response.</param>
        /// <returns>A HttpResponseMessage with the search results.</returns>
        [HttpPost]
        public async Task<HttpResponseMessage> PostSearchGeometry(string formatstring)
        {
            return await this.SearchGeometry(formatstring);
        }

        /// <summary>
        /// GET action for the Geometry Search API.
        /// </summary>
        /// <param name="formatstring">Format for the expected response.</param>
        /// <returns>A HttpResponseMessage with the search results.</returns>
        [HttpGet]
        public async Task<HttpResponseMessage> GetSearchGeometry(string formatstring)
        {
            return await this.SearchGeometry(formatstring);
        }

        internal override void ClientOverrideRequest(string query, ref Dictionary<string, List<string>> querystrings)
        {
        }

        internal override void ClientValidateParameters()
        {
            this.ValidateRequest();
        }

        /// <summary>
        /// By default the request body will not be parsed in BaseProxyController as part of the Initialize method. 
        /// Geometry Search API behaves differently as in it needs to validate the request body before the action gets 
        /// invoked. Hence the parsing is handled in the base class.
        /// </summary>
        protected override bool ShouldParseRequestBody()
        {
            return this.Request.Method.Equals(HttpMethod.Post);
        }

        private async Task<HttpResponseMessage> SearchGeometry(string formatstring)
        {
            try
            {
                var proxyObj = (BaseProxySearch)this.versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Search, this.ProxyOperation);

                if (proxyObj != null && proxyObj is ISearchCustomSerializationProvider serializationProvider)
                {
                    this.response = await proxyObj.SearchProxy(
                                        BaseProxySearch.SearchType.Geometry,
                                        WebUtility.UrlDecode(this.Query),
                                        formatstring,
                                        this.querystrings,
                                        serializationProvider.SerializeGeoJsonGeometry(BaseProxySearch.SearchType.Geometry, this.GeoJsonGeometry));
                }
                else
                {
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, "Failed to initialize the proxy object");
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        private void ValidateRequest()
        {
            // Query
            if (string.IsNullOrEmpty(this.Query))
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Invalid query");
            }

            // Format
            if (this.Format != "json" && this.Format != "xml")
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Invalid format string");
            }

            // Limit
            ValidationHelper.ValidateParameterWithRange(Constants.LimitParamKey, false, this.querystrings, 1, 100);

            // Geometry
            this.GeoJsonGeometry = GeometryHelper.ParseAndValidateGeometry(
                this.Request.Method, 
                this.HttpRequestContent,
                Constants.GeometryQueryParamKey,
                this.querystrings);
        }
    }
}