﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Controllers
{
    using System.Collections.Generic;
    using System.Linq;
    using System.Net.Http;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Utility;

    public class TrafficIncidentController : BaseProxyController
    {
        private const int INCLUSIVE_MIN_ZOOM_INDEX = 0;
        private const int INCLUSIVE_MAX_ZOOM_INDEX = 18;

        public async Task<HttpResponseMessage> GetTrafficIncidentTile(string formatstring)
        {
            try
            {
                List<string> style = new List<string>();
                List<string> zoom = new List<string>();
                List<string> x = new List<string>();
                List<string> y = new List<string>();

                querystrings.TryGetValue("style", out style);
                querystrings.TryGetValue("zoom", out zoom);
                querystrings.TryGetValue("x", out x);
                querystrings.TryGetValue("y", out y);

                querystrings.Remove("style");
                querystrings.Remove("zoom");
                querystrings.Remove("x");
                querystrings.Remove("y");

                BaseProxyTraffic proxyObj = (BaseProxyTraffic)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Traffic, this.ProxyOperation);
                if (proxyObj != null)
                {
                    this.response =
                        await
                            proxyObj.IncidentTile(
                                     BaseProxyTraffic.TrafficType.IncidentTile, 
                                     style.First(),
                                     zoom.First(),
                                     x.First(),
                                     y.First(),
                                     formatstring,
                                     this.querystrings);
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        public async Task<HttpResponseMessage> GetTrafficIncidentViewport(string formatstring)
        {
            try
            {
                BaseProxyTraffic proxyObj = (BaseProxyTraffic)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Traffic, this.ProxyOperation);
                if (proxyObj != null)
                {
                    List<string> boundingbox = new List<string>();
                    List<string> boundingzoom = new List<string>();
                    List<string> overviewbox = new List<string>();
                    List<string> overviewzoom = new List<string>();
                    List<string> copyright = new List<string>();

                    querystrings.TryGetValue("boundingbox", out boundingbox);
                    querystrings.TryGetValue("boundingzoom", out boundingzoom);
                    querystrings.TryGetValue("overviewbox", out overviewbox);
                    querystrings.TryGetValue("overviewzoom", out overviewzoom);
                    querystrings.TryGetValue("copyright", out copyright);

                    querystrings.Remove("boundingbox");
                    querystrings.Remove("boundingzoom");
                    querystrings.Remove("overviewbox");
                    querystrings.Remove("overviewzoom");
                    querystrings.Remove("copyright");
                    
                    this.response =
                        await
                            proxyObj.IncidentViewport(
                                BaseProxyTraffic.TrafficType.IncidentViewport,
                                boundingbox.First(),
                                boundingzoom.First(),
                                overviewbox.First(),
                                overviewzoom.First(),
                                copyright.First(),
                                formatstring,
                                this.querystrings);
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        public async Task<HttpResponseMessage> GetTrafficIncidentDetails(string formatstring)
        {
            try
            {
                BaseProxyTraffic proxyObj = (BaseProxyTraffic)versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.Traffic, this.ProxyOperation);
                if (proxyObj != null)
                {
                    List<string> style = new List<string>();
                    List<string> boundingbox = new List<string>();
                    List<string> boundingzoom = new List<string>();
                    List<string> trafficmodelid = new List<string>();

                    querystrings.TryGetValue("style", out style);
                    querystrings.TryGetValue("boundingbox", out boundingbox);
                    querystrings.TryGetValue("boundingzoom", out boundingzoom);
                    querystrings.TryGetValue("trafficmodelid", out trafficmodelid);

                    querystrings.Remove("boundingbox");
                    querystrings.Remove("boundingzoom");
                    querystrings.Remove("style");
                    querystrings.Remove("trafficmodelid");
                    
                    this.response =
                        await
                            proxyObj.IncidentDetails(
                                     BaseProxyTraffic.TrafficType.IncidentDetails,
                                     style.First(),
                                     boundingbox.First(),                                    
                                     boundingzoom.First(),
                                     trafficmodelid.First(),
                                     formatstring,
                                     this.querystrings);
                }
                else
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Maintainability", "CA1502:AvoidExcessiveComplexity")]
        internal override void ClientValidateParameters()
        {
            var incidentTileValidator = new TileValidator(INCLUSIVE_MIN_ZOOM_INDEX, INCLUSIVE_MAX_ZOOM_INDEX);

            if (this.HttpAction == "GetTrafficIncidentTile")
            {
                if (this.Format != "png")
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "invalid format string"));
                }

                if (!querystrings.ContainsKey("style") || !querystrings.ContainsKey("zoom") || !querystrings.ContainsKey("x") || !querystrings.ContainsKey("y"))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "style||zoom||x||y invalid"));
                }

                string xParamVal = querystrings["x"]?.FirstOrDefault();
                string yParamVal = querystrings["y"]?.FirstOrDefault();
                string zoomParamVal = querystrings["zoom"]?.FirstOrDefault();

                if (this.Format == "png" && !incidentTileValidator.IsValidTile(xParamVal, yParamVal, zoomParamVal))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, $"(x:{xParamVal}, y:{yParamVal}, z:{zoomParamVal}) is not a valid tile."));
                }
            }

            if (this.HttpAction == "GetTrafficIncidentViewport")
            {
                if (this.Format != "json" && this.Format != "xml")
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "invalid format string"));
                }

                if (!querystrings.ContainsKey("boundingbox") || !querystrings.ContainsKey("boundingzoom") || !querystrings.ContainsKey("overviewbox") || !querystrings.ContainsKey("overviewzoom") || !querystrings.ContainsKey("copyright"))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "boundingbox||boundingzoom||overviewbox||overviewzoom||copyright invalid"));
                }

                string boundingZoom = querystrings["boundingzoom"]?.FirstOrDefault();
                if (!incidentTileValidator.IsZoomValid(boundingZoom))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, $"Bounding Zoom ({boundingZoom}) must be between {INCLUSIVE_MIN_ZOOM_INDEX} and {INCLUSIVE_MAX_ZOOM_INDEX} inclusive."));
                }

                string overviewZoom = querystrings["overviewzoom"]?.FirstOrDefault();
                if (!incidentTileValidator.IsZoomValid(overviewZoom))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, $"Overview Zoom ({overviewZoom}) must be between {INCLUSIVE_MIN_ZOOM_INDEX} and {INCLUSIVE_MAX_ZOOM_INDEX} inclusive."));
                }
            }

            if (this.HttpAction == "GetTrafficIncidentDetails")
            {
                if (this.Format != "json" && this.Format != "xml")
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "invalid format string"));
                }

                if (!querystrings.ContainsKey("boundingbox") || !querystrings.ContainsKey("boundingzoom") || !querystrings.ContainsKey("style") || !querystrings.ContainsKey("trafficmodelid"))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, "boundingbox||boundingzoom||style||trafficmodelid invalid"));
                }

                string boundingZoom = querystrings["boundingzoom"]?.FirstOrDefault();
                if (!incidentTileValidator.IsZoomValid(boundingZoom))
                {
                    throw (new ProxyException(ProxyException.ExceptionType.ClientParams, $"Bounding Zoom ({boundingZoom}) must be between {INCLUSIVE_MIN_ZOOM_INDEX} and {INCLUSIVE_MAX_ZOOM_INDEX} inclusive."));
                }
            }
        }

        internal override void ClientOverrideRequest(string query, ref Dictionary<string, List<string>> querystrings)
        {
        }
    }
}