﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Controllers
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;
    using Microsoft.Azure.LocationServices.Models;
    using Microsoft.Azure.LocationServices.Utility;

    /// <summary>
    /// Main controller for the Upload Service API. 
    /// This api allows users to upload geographic data to azure maps in multiple formats like ShapeFile, KML, GeoJSON etc. 
    /// </summary>
    public class UploadServiceController : BaseProxyController
    {
        private BaseProxyUpload.UploadDataFormat dataFormat;

        #region Supported API's

        /// <summary>
        /// POST action for the Upload User Data API 
        /// </summary>
        /// <returns>A HttpResponseMessage with the unique data id (udid) that was uploaded.</returns>
        [HttpPost]
        public async Task<HttpResponseMessage> PostData()
        {
            try
            {
                var proxyObj = (BaseProxyUpload)this.versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.UserData, this.ProxyOperation);

                if (proxyObj != null)
                {
                    using (Stream data = await this.Request.Content.ReadAsStreamAsync())
                    {
                        this.response = await proxyObj.UploadData(this.AccountId, data, this.dataFormat, BaseProxyUpload.UploadType.Blob);
                    }
                }
                else
                {
                    this.response = this.BuildErrorResponseAndLog(new ProxyException(ProxyException.ExceptionType.ServerError, "Failed to initialize the proxy object"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        /// <summary>
        /// PUT action for the Update User Data API 
        /// </summary>
        /// <param name="udid">The unique data id of the file to update.</param>
        /// <returns>A HttpResponseMessage with the unique data id (udid) that was updated.</returns>
        [HttpPut]
        public async Task<HttpResponseMessage> PutData(string udid)
        {
            try
            {
                // Validate that we have a non-null data id to update.
                if (string.IsNullOrWhiteSpace(udid))
                {
                    return this.BuildErrorResponseAndLog(new ProxyException(ProxyException.ExceptionType.ClientParams, "udid should not be null"));
                }

                var proxyObj = (BaseProxyUpload)this.versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.UserData, this.ProxyOperation);

                if (proxyObj != null)
                {
                    using (Stream data = await this.Request.Content.ReadAsStreamAsync())
                    {
                        this.response = await proxyObj.UpdateData(this.AccountId, udid, data, BaseProxyUpload.UploadType.Blob);
                    }
                }
                else
                {
                    this.response = this.BuildErrorResponseAndLog(new ProxyException(ProxyException.ExceptionType.ServerError, "Failed to initialize the proxy object"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        /// <summary>
        /// Delete action for the User Data API 
        /// </summary>
        /// <param name="udid">The unique data id of the file to delete.</param>
        /// <returns>A HttpResponseMessage with the unique data id (udid) that was deleted.</returns>
        [HttpDelete]
        public async Task<HttpResponseMessage> DeleteData(string udid)
        {
            try
            {
                // Validate that we have a non-null data id to update.
                if (string.IsNullOrWhiteSpace(udid))
                {
                    return this.BuildErrorResponseAndLog(new ProxyException(ProxyException.ExceptionType.ClientParams, "udid should not be null"));
                }

                var proxyObj = (BaseProxyUpload)this.versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.UserData, this.ProxyOperation);

                if (proxyObj != null)
                {
                    this.response = await proxyObj.DeleteData(this.AccountId, udid, BaseProxyUpload.UploadType.Blob);
                }
                else
                {
                    this.response = this.BuildErrorResponseAndLog(new ProxyException(ProxyException.ExceptionType.ServerError, "Failed to initialize the proxy object"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        /// <summary>
        /// Delete action for the User Data API 
        /// </summary>
        /// <returns>A HttpResponseMessage with the unique data id (udid) that was deleted.</returns>
        [HttpGet]
        public async Task<HttpResponseMessage> ListData()
        {
            try
            {
                var proxyObj = (BaseProxyUpload)this.versionObject.GetProxyObj(this.Version, BaseProxy.ProxyType.UserData, this.ProxyOperation);

                if (proxyObj != null)
                {
                    this.response = await proxyObj.ListData(this.AccountId, BaseProxyUpload.UploadType.Blob);
                }
                else
                {
                    this.response = this.BuildErrorResponseAndLog(new ProxyException(ProxyException.ExceptionType.ServerError, "Failed to initialize the proxy object"));
                }
            }
            catch (ProxyException ex)
            {
                this.response = this.BuildErrorResponseAndLog(ex);
            }

            return this.response;
        }

        #endregion

        #region Overrides

        internal override void ClientOverrideRequest(string query, ref Dictionary<string, List<string>> querystrings)
        {
        }

        internal override void ClientValidateParameters()
        {
            this.ValidateRequest();
        }

        #endregion

        #region Private Helpers

        private void ValidateRequest()
        {
            // Account ID
            if (string.IsNullOrWhiteSpace(this.AccountId))
            {
                // Running locally? (We need to pass in some unique id for account id as this determines the destination container)
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Account Id should not be null.");
            }

            if (this.Request.Method.Equals(HttpMethod.Post))
            {
                // Format is available
                string formatString = this.querystrings.TryGetValue(Constants.UserDataUploadFormatKey, out var format)
                                          ? format.First()
                                          : throw new ProxyException(ProxyException.ExceptionType.ClientParams,
                                                                     $"{Constants.UserDataUploadFormatKey} should not be null.");

                // Format is valid
                if (!Enum.TryParse(formatString, true, out this.dataFormat))
                {
                    throw new ProxyException(
                        ProxyException.ExceptionType.ClientParams,
                        $"Invalid Format. Value can be one of: {{{string.Join(", ", Enum.GetNames(typeof(BaseProxyUpload.UploadDataFormat)).Skip(1)).ToLowerInvariant()}}}");
                }

                // Content/Content-Type
                if (this.Request.Content?.Headers?.ContentLength == 0 || this.Request.Content?.Headers?.ContentType == null)
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Content/Content-Type should not be null.");
                }
            }
        }

        #endregion
    }
}