﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Collections.Generic;
    using System.Net.Http;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Common.Storage;
    using Microsoft.Azure.LocationServices.GeoLocation;
    using Microsoft.Azure.LocationServices.GeoLocation.Storage;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.Properties;
    using Microsoft.Azure.LocationServices.Shared;
    using Microsoft.Azure.LocationServices.Utility;
    using Microsoft.WindowsAzure.Storage;
    using Newtonsoft.Json.Linq;
    using Validation;

    public enum ResponseFormat
    {
        UNKNOWN,
        XML,
        JSON
    }

    internal class Globals
    {
        private static readonly object _syncLock = new object();
        private static readonly Lazy<IBlobService> BlobServiceLazy = new Lazy<IBlobService>(() => new AzureBlobService(Instance.BatchServiceStorageConnectionString));
        private static readonly Lazy<IQueueService> QueueServiceLazy = new Lazy<IQueueService>(() => new AzureQueueService(Instance.BatchServiceStorageConnectionString));
        private static readonly Lazy<ProxyHostSettings> ProxyHostSettingsLazy = new Lazy<ProxyHostSettings>(() => ProxyHostSettings.CreateFromEnvironment());
        private static Globals _instance;
        private static NetworkAssignmentCache networkAssignmentCache;

        private Globals()
        {
        }

        public static Globals Instance
        {
            get
            {
                if (_instance == null)
                {
                    lock (_syncLock)
                    {
                        if (_instance == null)
                        {
                            _instance = new Globals();
                        }
                    }
                }

                return _instance;
            }
        }

        public ProxyHostSettings ProxySettings => ProxyHostSettingsLazy.Value;

        public string TomTomSearchKey { get; private set; }

        public string TomTomMapKey { get; private set; }

        public string TomTomRouteKey { get; private set; }

        public string TomTomTrafficKey { get; private set; }

        public string TomTomTrafficIncidentKey { get; private set; }

        public string TomTomReverseGeocodingKey { get; private set; }

        public string BatchServiceStorageConnectionString { get; private set; }

        public string BingMapImgUrl { get; private set; }

        public Metrics ProxyMetrics { get; private set; }

        public IIPLookup IPLookupService { get; private set; }

        public IBlobService BlobService => BlobServiceLazy.Value;
        
        public IQueueService QueueService => QueueServiceLazy.Value;

        public string ServiceHostName => ProxySettings.HostName;

        public async Task InitializeAsync(CancellationToken cancellationToken, 
                                          IBootstrapConfiguration bootstrapConfiguration, 
                                          IConfigurationReader configurationReader)
        {
            if (bootstrapConfiguration == null)
            {
                throw new ArgumentNullException(nameof(bootstrapConfiguration));
            }

            if (configurationReader == null)
            {
                throw new ArgumentNullException(nameof(configurationReader));
            }

            if (cancellationToken.IsCancellationRequested)
            {
                // If service fabric requests cancellation we should honor that.
                return;
            }

            this.TomTomSearchKey = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("mTomTomSearchKey"));
            this.TomTomMapKey = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("mTomTomMapKey"));
            this.TomTomRouteKey = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("mTomTomRouteKey"));
            this.TomTomTrafficKey = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("mTomTomTrafficKey"));
            this.TomTomTrafficIncidentKey = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("mTomTomTrafficIncidentKey"));
            this.TomTomReverseGeocodingKey = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("mTomTomReverseGeocodingKey"));
            
            this.BatchServiceStorageConnectionString = string.Format(
                await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(SharedUtilities.BatchServiceConnectionStringTemplateKey)),
                await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(SharedUtilities.BatchServiceSasTokenKey)));

            networkAssignmentCache = this.CreateNetworkAssignmentCache(
                await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("malbsipdata")),
                await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("malbsipdata-crudall")),
                cancellationToken);

            this.BingMapImgUrl = await GetBingMapImgUrlAsync(
                await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("bingMapsKey")));

            this.ProxyMetrics = new Metrics(bootstrapConfiguration);
        }

        private async Task<string> GetBingMapImgUrlAsync(string bingMapsKey)
        {
            try
            {
                using (var client = new HttpClient())
                using (var httpResponse = await client.GetAsync($"http://dev.virtualearth.net/REST/v1/Imagery/Metadata/aerial?key={bingMapsKey}"))
                {
                    string content = await httpResponse.Content.ReadAsStringAsync();
                    JObject json = JObject.Parse(content);
                    string imageUrl = json.SelectToken("resourceSets[0].resources[0].imageUrl")?.Value<string>();

                    Verify.Operation(!string.IsNullOrEmpty(imageUrl), "Bing Imagery imageUrl is null or empty");

                    return imageUrl;
                }
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "Bing Satellite imageUrl retrieval failed");
                throw;
            }
        }
       
        private NetworkAssignmentCache CreateNetworkAssignmentCache(
            string connectionStringTemplate,
            string sasKey,
            CancellationToken cancellationToken)
        {
            var account = CloudStorageAccount.Parse(string.Format(connectionStringTemplate, sasKey));
            var primaryUri = new Uri(account.BlobEndpoint, Settings.Default.IPDataContainerPath);
            var logger = new GeoLocationLogger();

            var cache = new NetworkAssignmentCache(
                new NetworkAssignmentStorage(primaryUri, account.Credentials, logger),
                TimeSpan.FromMinutes(Settings.Default.IPDataRefreshMinutes),
                logger);

            cache.GetNetworkAssignmentsWhenUpdated(
                delegate(IEnumerable<NetworkAssignment> assignments) { this.IPLookupService = new IPLookup(assignments); },
                cancellationToken).FireAndForget();

            return cache;
        }
    }
}