﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace Microsoft.Azure.LocationServices.Instrumentation
{
    using System;

    using Microsoft.Azure.LocationServices.Instrumentation.Models;
    using Microsoft.Cloud.InstrumentationFramework;

    using static System.String;

    public class ProxyOperation : OperationAgent<AzLocSvcIfx.ProxyOperation>, IProxyOperation
    {
        public ProxyOperation(string operationName)
            : base(operationName)
        {
            this.Operation.ApiType = OperationApiType.ServiceApi;

            this.Operation.PartC = new AzLocSvcIfx.ProxyOperation()
            {
                SubscriptionId = Empty,
                TrackingId = Empty,
                APIM_PROXY = Empty,
                PROXY_EXT = Empty,
                EXT_PROXY = Empty,
                PROXY_APIM = Empty,
                Proxy_ms = 0,
                EXT_ms = 0
            };
        }

        public string ResourceId => this.Operation.ResourceId;

        public BaseProxy.ProxyType ProxyType { get; set; }

        public string OperationSourceName { get; set; }

        public void Log_SubscriptionId(string SubscriptionId)
        {
            if (SubscriptionId != null)
            {
                this.Operation.PartC.SubscriptionId = SubscriptionId;
            }
        }

        public void Log_TrackingId(string TrackingId)
        {
            if (TrackingId != null)
            {
                this.Operation.PartC.TrackingId = TrackingId;
            }
        }

        public void Log_APIM_PROXY(string APIM_PROXY)
        {
            if (APIM_PROXY != null)
            {
                this.Operation.PartC.APIM_PROXY = APIM_PROXY;
            }
        }

        public void Log_PROXY_EXT(string PROXY_EXT)
        {
            if (PROXY_EXT != null)
            {
                this.Operation.PartC.PROXY_EXT = PROXY_EXT;
            }
        }

        public void Log_EXT_PROXY(string EXT_PROXY, int responseCode)
        {
            if (EXT_PROXY != null)
            {
                this.Operation.PartC.EXT_PROXY = EXT_PROXY;
            }

            this.Operation.PartC.EXT_PROXY_ResponseCode = responseCode;
        }

        public void Log_PROXY_APIM(string PROXY_APIM, int responseCode)
        {
            if (PROXY_APIM != null)
            {
                this.Operation.PartC.PROXY_APIM = PROXY_APIM;
            }

            this.Operation.PartC.PROXY_APIM_ResponseCode = responseCode;
        }

        public void Log_Proxy_ms(Int32 duration)
        {
            this.Operation.PartC.Proxy_ms = duration;
        }

        public void Log_EXT_ms(Int32 duration)
        {
            this.Operation.PartC.EXT_ms = duration;
        }

        /// <summary>
        /// Set resource id for account specific metrics and logging
        /// </summary>
        /// <param name="resourceId">The resource id for the account in Azure Resource Manager</param>
        public void Log_ResourceId(string resourceId)
        {
            if (resourceId != null)
            {
                this.Operation.ResourceId = resourceId;
            }
        }

        public void Log_ProxyOperationMetrics()
        {
            var resourceId = this.Operation.ResourceId;
            var operationName = this._operationName;
            var operationResult = this._operationResult;
            var duration = this.Operation.PartC.Proxy_ms;
            var extDuration = this.Operation.PartC.EXT_ms;

            var details = new OperationDetails
            {
                ResourceId = resourceId,
                OperationName = operationName,
                OperationResult = operationResult,
                ProxyType = this.ProxyType,
                OperationSourceName = this.OperationSourceName,
                OperationResultDetails = this._resultDetails ?? String.Empty,
                ExternalResultCode = this.Operation.PartC.EXT_PROXY_ResponseCode == 0 ? String.Empty : this.Operation.PartC.EXT_PROXY_ResponseCode.ToString(),
                OperationResultCode = this.Operation.PartC.PROXY_APIM_ResponseCode == 0 ? String.Empty : this.Operation.PartC.PROXY_APIM_ResponseCode.ToString()
            };

            // Log metrics to Geneva
            Globals.Instance.ProxyMetrics.ShoeboxLatencyLogger.LogDimensionValues(details, duration);
            Globals.Instance.ProxyMetrics.ShoeboxFailureLogger.LogDimensionValues(details, duration);
            Globals.Instance.ProxyMetrics.ExternalLatencyLogger.LogDimensionValues(details, extDuration);

            this.Operation.MarkImpactsQoS(string.Empty, resourceId);
        }

        public void Log_ClientSessionId(string clientSessionId)
        {
            if (clientSessionId != null)
            {
                this.Operation.PartC.ClientSessionId = clientSessionId;
            }
        }

#if DEBUG
        protected override void WriteEtw()
        {
            var partC = Operation.PartC;
            Agent.Instance.WriteEtwEvent(EventType.Information,
                                         new
                                         {
                                             Operation = _operationName,
                                             SubscriptionId = partC.SubscriptionId,
                                             TrackingID = partC.TrackingId,
                                             APIM_PROXY = partC.APIM_PROXY,
                                             PROXY_EXT = partC.PROXY_EXT,
                                             EXT_PROXY = partC.EXT_PROXY,
                                             PROXY_APIM = partC.PROXY_APIM,
                                             Proxy_ms = partC.Proxy_ms,
                                             EXT_ms = partC.EXT_ms,
                                             OperationResult = _operationResult.ToString(),
                                             ResultSignature = _resultSignature,
                                             ResultDetails = _resultDetails
                                         });
        }
#endif
    }
}