﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Diagnostics;
    using System.Net;
    using System.Net.Http;    
    using System.Threading.Tasks;
    using System.Web;

    using Microsoft.Azure.LocationServices.Instrumentation;

    public abstract class BaseProxy : IDisposable
    {
        protected HttpResponseMessage response = new HttpResponseMessage(HttpStatusCode.BadRequest);
        protected UriBuilder URI = new UriBuilder();
        protected ResponseOverride ResponseOverride = new ResponseOverride();

        private static readonly HttpClient httpClient = new HttpClient();

        // custom timeout, at this point, is being used by Matrix routing where TT takes longer(upto 2 mins) as they don't have true async model.
        private static readonly HttpClient httpClientWithCustomTimeout = new HttpClient();
        private static readonly TimeSpan httpClientCustomTimeout = TimeSpan.FromMinutes(3);

        public BaseProxy(IProxyOperation proxyOperation)
        {
            // Timeout can be set only once per instance.
            if (httpClientWithCustomTimeout.Timeout != httpClientCustomTimeout)
            {
                httpClientWithCustomTimeout.Timeout = httpClientCustomTimeout;
            }

            Log_EXT_PROXY = proxyOperation.Log_EXT_PROXY;
            Log_PROXY_EXT = proxyOperation.Log_PROXY_EXT;
            Log_EXT_ms = proxyOperation.Log_EXT_ms;
        }

        ~BaseProxy()
        {
            Dispose(false);
        }

        public enum ProxyType
        {
            Unknown = 0,
            Map,
            Imagery,
            Search,
            Route,
            Timezone,
            Traffic,
            Heartbeat,
            GeoLocation,
            Geofence,
            UserData,
            Batch
        }

        public Action<string> Log_PROXY_EXT { get; set; }

        public Action<string, int> Log_EXT_PROXY { get; set; }

        public Action<Int32> Log_EXT_ms { get; set; }

        public bool Disposed { get; private set; }

        public IGlobalConfigProperties globalConfigProperties { get; set; }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        internal virtual async Task<HttpResponseMessage> ExecuteRequest(HttpRequestMessage request)
        {
            return await ExecuteRequest(request, true, false);
        }

        /// <summary>
        /// Executes the request without calling the log methods. 
        /// At this point this is used by Route Matrix where the user request is executed on a background task which is waiting on TT while the user gets a response(202) back.
        /// Once the user request is completed, the log object is disposed and the background task can't use it when it is done at a later point in time. Hence this method to not use it.
        /// Created Product Backlog Item 2697387
        /// </summary>
        /// <param name="request"></param>
        /// <returns></returns>
        internal virtual async Task<HttpResponseMessage> ExecuteRequestWithCustomTimeoutNoLogging(HttpRequestMessage request)
        {
            return await ExecuteRequest(request, false, true);
        }

        internal void ValidateReponse(ref HttpResponseMessage response)
        {
            var content = "Unable to read content";
            try
            {
                content = response.Content.ReadAsStringAsync().Result;
            }
            catch
            {
                //just move on if ReadAsStringAsync failed
            }

            content = HttpUtility.HtmlEncode(content);
            switch (response.StatusCode)
            {
                case HttpStatusCode.BadRequest:
                    {
                        throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"StatusCode:{response.StatusCode} Content:{content}");
                    }

                case HttpStatusCode.BadGateway:
                case HttpStatusCode.InternalServerError:
                case HttpStatusCode.NotFound:
                case HttpStatusCode.NotImplemented:
                case HttpStatusCode.ServiceUnavailable:
                case HttpStatusCode.Unauthorized:
                case HttpStatusCode.GatewayTimeout:
                    {
                        ProxyException proxyException = new ProxyException(ProxyException.ExceptionType.ExtError, $"StatusCode:{response.StatusCode} Content:{content}");
                        throw (proxyException);
                    }

                case HttpStatusCode.Forbidden:
                    {
                        ProxyException proxyException = new ProxyException(ProxyException.ExceptionType.Throttled, $"StatusCode:{response.StatusCode} Content:{content}");
                        throw (proxyException);
                    }

                case HttpStatusCode.Accepted:
                case HttpStatusCode.SeeOther:
                    // If the response is a redirect, there is no content to be validated.
                    // Validate everything else.
                    break;
                default:
                    ResponseOverride.Override(ref response);
                    this.ClientValidateReponse(ref response);
                    break;
            }           
        }

        internal abstract void ClientValidateReponse(ref HttpResponseMessage response);

        private async Task<HttpResponseMessage> ExecuteRequest(HttpRequestMessage request, bool loggingEnabled, bool useCustomTimeoutHttpClient)
        {
            try
            {
                if (loggingEnabled)
                {
                    Log_PROXY_EXT(this.URI.Uri.ToString());
                }

                var sw = new Stopwatch();
                sw.Start();

                var currentClient = httpClient;
                if (useCustomTimeoutHttpClient)
                {
                    currentClient = httpClientWithCustomTimeout;
                }

                this.response = await currentClient.SendAsync(request);
                sw.Stop();

                if (loggingEnabled)
                {
                    Log_EXT_ms((int)sw.ElapsedMilliseconds);
                    Log_EXT_PROXY(this.response.ToString(), (int)this.response.StatusCode);
                }

                this.ValidateReponse(ref this.response);
                return this.response;
            }
            catch (ProxyException)
            {
                throw;
            }
            catch (Exception ex)
            {
                ProxyException proxyException = new ProxyException(ProxyException.ExceptionType.ServerError, "internalerror", ex);
                throw (proxyException);
            }
        }

        private void Dispose(bool disposing)
        {
            if (!Disposed)
            {
                try
                {
                    if (disposing)
                    {
                        this.response.Dispose();
                    }
                }
                finally
                {
                    Disposed = true;
                }
            }
        }
    }
}