﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Models.Common.Batch
{
    using System;
    using System.Collections.Generic;
    using System.Collections.Specialized;
    using System.IO;
    using System.IO.Compression;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Net.Http.Formatting;
    using System.Net.Http.Headers;
    using System.Security.Cryptography;
    using System.Text;
    using System.Threading.Tasks;
    using System.Web;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Common.Storage;
    using Microsoft.Azure.LocationServices.Shared;
    using Microsoft.Azure.LocationServices.Utility;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Queue;
    using Newtonsoft.Json;
    using Newtonsoft.Json.Linq;
    using ProxyBatchShared.DataModels;

    public class BatchApiProxyHelper
    {
        /// <summary>
        /// This method processes a POST "Batch (search/route) Api" request.
        /// </summary>
        /// <param name="blobService">An instance of blob service</param>
        /// <param name="queueService">An instance of queue service</param>
        /// <param name="batchRequestUri">Uri for the incoming batch api request</param>
        /// <param name="requestBody">The request body for the batch api request</param>
        /// <param name="batchId">A unique id for the current batch</param>
        /// <param name="containerName">Container name for the batch api request</param>
        /// <param name="accountId">Account Id for the batch api request</param>
        /// <returns>A Http response for the POST request</returns>
        public static async Task<HttpResponseMessage> ProcessPostBatchApiRequest(
            IBlobService blobService, 
            IQueueService queueService, 
            Uri batchRequestUri, 
            string requestBody,
            string batchId,
            string containerName, 
            string accountId)
        {
            try
            {
                // 1. Create the container if it doesn't exist already.
                await blobService.CreateContainerIfNotExistsAsync(containerName);

                // 2. Create an empty blob where blob name is the unique batchId passed-in.
                var blobReference = blobService.GetBlockBlobReference(containerName, batchId);

                // 3. Set the blob metadata to the default state.
                var metadata = new Dictionary<string, string>
                {
                    { SharedUtilities.BatchServiceResponseStatusMetadataKey, bool.FalseString },
                    { SharedUtilities.BatchServiceRequestFailedMetadataKey, bool.FalseString }
                };

                // 4. Upload the batch requests.
                await blobService.UploadDocumentAsync(containerName, batchId, requestBody, metadata);

                // 5. Create the queue if it doesn't already exist.
                CloudQueue queue = queueService.GetQueueReference(SharedUtilities.BatchServicePrimaryQueueName);
                await queueService.CreateQueueIfNotExistsAsync(queue);

                // 6. Add a new work item for this batch request to the Queue.
                var queueMessage = new BatchServiceQueueMessage
                                   {
                                       BlobUrl = blobReference.Uri.AbsoluteUri,
                                       AccountId = GenerateUniqueIdForAccount(accountId)
                                   };

                await queueService.AddMessageAsync(queue, new CloudQueueMessage(JsonConvert.SerializeObject(queueMessage)));

                // 7. Create the callback uri for the accepted request & returns the response that the request is Accepted and set the location header to the callback uri.
                Uri responseUri = 
                    new UriBuilder(
                        new Uri(
                            new UriBuilder(
                                batchRequestUri.Scheme, 
                                Globals.Instance.ServiceHostName).Uri,
                            $"/{Constants.BatchKey}/{batchId}{batchRequestUri.Query}")).Uri;

                return GetRequestAcceptedResponse(responseUri, new Dictionary<string, string>());
            }
            catch (Exception ex) when (ex is StorageException || ex is UriFormatException || ex is JsonSerializationException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "Batch Processing failed!", ex);
            }
        }

        /// <summary>
        /// This method processes a GET "Batch (search/route) Api" request.
        /// </summary>
        /// <param name="blobService">An instance of blob service</param>
        /// <param name="batchRequestUri">Uri for the incoming batch api request</param>
        /// <param name="batchId">Batch Id for the request</param>
        /// <param name="containerName">Container name for the batch api request</param>
        /// <returns>A Http response for the GET request</returns>
        public static async Task<HttpResponseMessage> ProcessGetBatchApiRequest(
            IBlobService blobService, 
            Uri batchRequestUri, 
            string batchId,
            string containerName)
        {
            try
            {
                // 1. Check if the batch-id is (still) valid.
                bool isBatchIdValid = await blobService.DoesBlockBlobExistAsync(containerName, batchId);

                if (!isBatchIdValid)
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Invalid BatchId - Either value is incorrect or has expired");
                }

                // 2. Get the blob metadata and check if response is ready.
                IDictionary<string, string> metadata = await blobService.GetBlockBlobMetadataAsync(containerName, batchId);

                if (!metadata.TryGetValue(SharedUtilities.BatchServiceResponseStatusMetadataKey, out string responseStatus) ||
                    !metadata.TryGetValue(SharedUtilities.BatchServiceRequestFailedMetadataKey, out string requestFailedStatus))
                {
                    // Just a safe check. This should not happen under any circumstances.
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, "Unexpected Error - Batch status metadata is missing in blob");
                }

                bool hasRequestFailed = bool.Parse(requestFailedStatus);
                bool isResponseAvailable = bool.Parse(responseStatus);

                // 3. Check if the request failed or not.
                if (hasRequestFailed)
                {
                    HttpStatusCode errorCode = HttpStatusCode.InternalServerError;

                    var errorContract = new ErrorContract
                    {
                        Code = $"{(int)errorCode} {errorCode}",
                        Message = "We encountered an error running your batch request. Please submit your request again."
                    };

                    var errorResponse = new ErrorContractResponse { Error = errorContract };

                    // Something Bad happened - Unexpected Error.
                    return new HttpResponseMessage(HttpStatusCode.InternalServerError)
                    {
                        Content = new ObjectContent<object>(errorResponse, new JsonMediaTypeFormatter(), Constants.JsonMediaType)
                    };
                }

                // 4. Request has not failed - Check if it's still ongoing or is a response already available.
                if (!isResponseAvailable)
                {
                    // Still processing the request. No response available yet. Return 202 Accepted with the location header set to the callback uri
                    var stillProcessingResponse = new HttpResponseMessage(HttpStatusCode.Accepted);
                    stillProcessingResponse.Headers.Location = GetUpdatedUri(batchRequestUri);
                    return stillProcessingResponse;
                }

                using (var ms = new MemoryStream())
                {
                    // 5. Request has successfully completed. Download the blob containing the batch results and deserialize it.
                    await blobService.DownloadBlockBlobToStreamAsync(containerName, batchId, ms);
                    BatchServiceBlobStorageResult batchServiceBlobStorageResult = DeserializeResponseBlob(ms);

                    // 6. Create the final batch service response from the blob result.
                    var batchServiceResponse = new BatchServiceResponse
                    {
                        Summary = batchServiceBlobStorageResult.Summary,
                        BatchResults = GenerateBatchResultsFromStorageResults(batchServiceBlobStorageResult.CompressedBatchResults)
                    };

                    // 7. Return 200 OK with the deserilized and formatted response body.
                    return new HttpResponseMessage(HttpStatusCode.OK)
                    {
                        Content = CreateHttpContent(batchServiceResponse)
                    };
                }
            }
            catch (Exception ex) when (ex is StorageException || ex is JsonSerializationException || ex is JsonReaderException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "Batch Processing failed!", ex);
            }
        }

        /// <summary>
        /// This method converts a list of relative "azure maps/lbs" queries into absolute uri's based on the uri path in the BatchRequestUri.
        /// e.g.- Relative Query     =   "?query=TIMES+SQUARE&amp;limit=5"
        ///       BatchRequestUri    =   "https://localhost/search/poi/batch/json?api-version=1.0&amp;subscription-key=12345"       
        ///       Final Absolute Uri =   "https://localhost/search/poi/json?api-version=1.0&amp;subscription-key=12345&amp;query=TIMES+SQUARE&amp;limit=5" 
        /// </summary>
        /// <param name="batchRequestUri">Uri for the incoming batch api request</param>
        /// <param name="relativeQueriesInLbsFormat">A list of relative queries in "Azure maps/lbs" format</param>
        /// <param name="batchId">A unique id for the current batch</param>
        /// <returns>A list of absolute uri's corresponding to the relative queries</returns>
        public static List<string> ConvertRelativeQueriesInBatchToAbsoluteUris(Uri batchRequestUri, IList<string> relativeQueriesInLbsFormat, string batchId)
        {
            // 1. Fetch all the "internal" key/id's from the incoming request.
            List<KeyValuePair<string, string>> internalIds = GetValidInternalIdsFromRequest(batchRequestUri);

            // 1.1 Add the unique id for the current batch to every request being executed as part of the batch to create association.
            internalIds.Add(new KeyValuePair<string, string>(Constants.BatchIdKey, batchId));

            List<string> finalQueryList = new List<string>();

            string relativePath = string.Join(string.Empty, batchRequestUri.AbsolutePath.Split(new[] { $"/{Constants.BatchKey}" }, StringSplitOptions.None));

            int queryIndex = 1;

            foreach (string lbsQuery in relativeQueriesInLbsFormat)
            {
                try
                {
                    // 2.1 Select the host based on the current env.
                    // 2.2 Create a new absolute uri using this host and the relative uri.
                    UriBuilder requestBuilder =
                        new UriBuilder(
                            new Uri(
                                new UriBuilder(
                                    batchRequestUri.Scheme,
                                    Globals.Instance.ServiceHostName).Uri,
                                $"{relativePath}{lbsQuery.Trim()}"));

                    // 3. Add back all the internal key/id's extracted from the parent to this new absolute uri (for APIM).
                    NameValueCollection finalRequestParams = HttpUtility.ParseQueryString(requestBuilder.Query);

                    foreach (KeyValuePair<string, string> internalId in internalIds)
                    {
                        finalRequestParams[internalId.Key] = internalId.Value;
                    }

                    // 3.1 Add the batch query index id to the absolute uri (for tracking)
                    finalRequestParams[Constants.BatchQueryIndexKey] = Convert.ToString(queryIndex++);

                    requestBuilder.Query = finalRequestParams.ToString();

                    // 4. Add the absolute uri to the final batch list.
                    finalQueryList.Add(requestBuilder.Uri.AbsoluteUri);
                }
                catch (Exception ex) when (ex is UriFormatException || ex is ArgumentNullException)
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Batch contains invalid queries", ex);
                }
            }

            return finalQueryList;
        }

        #region Public Helpers

        /// <summary>
        /// Builds the callback uri for the accepted request and returns the response that the request is Accepted and set the location header to the callback uri.
        /// Uses Globals.Instance.ServiceHostName as the new HostName.
        /// </summary>
        /// <param name="requestUri">The incoming URI. In non Dev environment, this points to the traffic manager.</param>
        /// <param name="newQueryParams">New Parameters to be added to the query.</param>
        /// <returns>A Http response object</returns>
        public static HttpResponseMessage GetRequestAcceptedResponse(Uri requestUri, Dictionary<string, string> newQueryParams)
        {
            // Create the callback uri for the accepted request
            var uriBuilder = new UriBuilder(requestUri);
            var query = HttpUtility.ParseQueryString(uriBuilder.Query);

            // For localHost, since APIM is not involved, leave everything as is.
            // For the rest, leave only the required params.
            if (!requestUri.IsLoopback)
            {
                var subscriptionKeyValue = query[Constants.SubscriptionKey];
                var apiVersionKeyValue = query[Constants.ApiVersionKey];

                if (string.IsNullOrEmpty(subscriptionKeyValue))
                {
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, $"{Constants.SubscriptionKey} parameter is not found in the incoming query.");
                }

                if (string.IsNullOrEmpty(apiVersionKeyValue))
                {
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, $"{Constants.ApiVersionKey} parameter is not found in the incoming query.");
                }

                query = HttpUtility.ParseQueryString(string.Empty);
                query[Constants.SubscriptionKey] = subscriptionKeyValue;
                query[Constants.ApiVersionKey] = apiVersionKeyValue;

                // The host name in the request that comes to the proxy is of the traffic manager. Replace it with the APIM end point the user sees.
                uriBuilder.Host = Globals.Instance.ServiceHostName;
            }

            foreach (var kvp in newQueryParams)
            {
                query.Add(kvp.Key, kvp.Value);
            }

            uriBuilder.Query = query.ToString();

            // Create the response that the request is Accepted and set the location header to the callback uri
            var responseObject = new HttpResponseMessage(HttpStatusCode.Accepted);
            responseObject.Headers.Location = uriBuilder.Uri;

            return responseObject;
        }

        /// <summary>
        /// Gives the Uri with the supplied hostname in non dev environment. Returns the supplied Uri otherwise.
        /// </summary>
        /// <param name="uri">The current Uri.</param>
        /// <returns>The Uri with appropriate host name.</returns>
        public static Uri GetUpdatedUri(Uri uri)
        {
            var updatedUri = uri;
            if (!uri.IsLoopback)
            {
                var uriBuilder = new UriBuilder(uri);
                uriBuilder.Host = Globals.Instance.ServiceHostName;
                updatedUri = uriBuilder.Uri;
            }

            return updatedUri;
        }

        #endregion

        #region Private Helpers

        private static List<KeyValuePair<string, string>> GetValidInternalIdsFromRequest(Uri requestUri)
        {
            // Fetch all the internal key/id's from the incoming request.
            NameValueCollection requestParams = HttpUtility.ParseQueryString(requestUri.Query);

            string subscriptionKeyValue = requestParams[Constants.SubscriptionKey];
            string subscriptionId = requestParams[Constants.SubscriptionIdKey];
            string trackingId = requestParams[Constants.TrackingIdKey];
            string accountId = requestParams[Constants.AccountIdKey];
            string apiVersion = requestParams[Constants.ApiVersionKey];

            List<KeyValuePair<string, string>> internalIds = new List<KeyValuePair<string, string>>();

            if (!string.IsNullOrEmpty(apiVersion))
            {
                internalIds.Add(new KeyValuePair<string, string>(Constants.ApiVersionKey, apiVersion));
            }

            if (!string.IsNullOrEmpty(subscriptionKeyValue))
            {
                internalIds.Add(new KeyValuePair<string, string>(Constants.SubscriptionKey, subscriptionKeyValue));
            }

            if (!string.IsNullOrEmpty(subscriptionId))
            {
                internalIds.Add(new KeyValuePair<string, string>(Constants.SubscriptionIdKey, subscriptionId));
            }

            if (!string.IsNullOrEmpty(trackingId))
            {
                internalIds.Add(new KeyValuePair<string, string>(Constants.TrackingIdKey, trackingId));
            }

            if (!string.IsNullOrEmpty(accountId))
            {
                internalIds.Add(new KeyValuePair<string, string>(Constants.AccountIdKey, accountId));
            }

            return internalIds;
        }

        private static string GenerateUniqueIdForAccount(string accountId)
        {
            // Account Id may have some special characters that are not compliant with azure storage naming guidelines.
            using (MD5 hash = MD5.Create())
            {
                // MD5 hash is fine here for this specific purpose (even though it is not cryptographically secure)
                return string.Concat("BatchServiceAccount-", string.Join("", hash.ComputeHash(Encoding.UTF8.GetBytes(accountId)).Select(ba => ba.ToString("x2")))).ToLowerInvariant();
            }
        }

        private static List<JObject> GenerateBatchResultsFromStorageResults(IEnumerable<byte[]> batchServiceBlobStorageResults)
        {
            var batchResults = new List<JObject>();

            try
            {
                batchResults.AddRange(
                    batchServiceBlobStorageResults.Select(
                        compressedBatchResult => JObject.Parse(Encoding.UTF8.GetString(Decompress(compressedBatchResult)))));
            }
            catch (Exception ex) when (ex is JsonSerializationException || ex is JsonReaderException || ex is ArgumentNullException)
            {
                // Don't rethrow - we don't want one request failure to affect the rest of the batch.
                batchResults.Add(JObject.Parse("{ \"error\": \"Request failed with an unexpected/unknown error\"}"));
            }

            return batchResults;
        }

        private static HttpContent CreateHttpContent(object content)
        {
            HttpContent httpContent = null;

            if (content != null)
            {
                var ms = new MemoryStream();
                SerializeJsonToStream(content, ms);
                ms.Seek(0, SeekOrigin.Begin);
                httpContent = new StreamContent(ms);
                httpContent.Headers.ContentType = new MediaTypeHeaderValue(Constants.JsonMediaType);
            }

            return httpContent;
        }

        private static void SerializeJsonToStream(object value, Stream stream)
        {
            using (var sw = new StreamWriter(stream, new UTF8Encoding(false), 1024, true))
            using (var jtw = new JsonTextWriter(sw) { Formatting = Formatting.Indented })
            {
                var js = new JsonSerializer();
                js.Serialize(jtw, value);
                jtw.Flush();
            }
        }

        private static BatchServiceBlobStorageResult DeserializeResponseBlob(MemoryStream ms)
        {
            using (var memoryStream = new MemoryStream())
            {
                var decompressed = Decompress(ms.ToArray());

                memoryStream.Write(decompressed, 0, decompressed.Length);
                memoryStream.Seek(0, SeekOrigin.Begin);

                return JsonConvert.DeserializeObject<BatchServiceBlobStorageResult>(Encoding.UTF8.GetString(memoryStream.ToArray()));
            }
        }

        private static byte[] Decompress(byte[] input)
        {
            if (input == null)
            {
                return new byte[0];
            }

            byte[] decompressedData;

            using (var outputStream = new MemoryStream())
            {
                using (var inputStream = new MemoryStream(input))
                {
                    using (var zip = new GZipStream(inputStream, CompressionMode.Decompress))
                    {
                        zip.CopyTo(outputStream);
                    }
                }

                decompressedData = outputStream.ToArray();
            }

            return decompressedData;
        }

        #endregion
    }
}