﻿// --------------------------------------------------------------------------------------------------------------------
// <copyright file="MicrosoftV1GeofenceProxy.cs" company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// --------------------------------------------------------------------------------------------------------------------
namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Device.Location;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Security.Cryptography;
    using System.Text;
    using System.Threading.Tasks;
    using GeoAPI.Geometries;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.Models;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using NetTopologySuite.Features;

    public class MicrosoftV1GeofenceProxy : BaseProxyGeofence
    {
        internal const double DefaultInsideDistanceValue = -999;

        internal const double MaximumSearchRadiusInMeters = 500;

        internal GeofenceResponseBody responseBody = null;

        private const string ContainerNamePrefix = "userdatacontainer-";

        public MicrosoftV1GeofenceProxy(IProxyOperation proxyOperation) : base(proxyOperation)
        {
        }

        internal override async Task<HttpResponseMessage> SearchGeoFence(string formatstring, string accountId, string udid, IPoint userLocation, DateTime usertime, double searchRadiusInMeters, GeofenceResponseBody responseBody)
        {
            this.responseBody = responseBody;

            var featureCollection = await GetGeoFenceFromStorage(accountId, udid);

            if (featureCollection == null)
            {
                CreateResponseMessage(responseBody);
            }

            foreach (var feature in featureCollection.Features)
            {
                CheckTimeValidity(feature, usertime);

                var geofence = feature.Geometry;

                var geometry = new Geometry();

                geometry.geometryId = GetPropertyValue<string>("geometryId", feature);

                if (IsCircle(feature) || feature.Geometry.OgcGeometryType == OgcGeometryType.Point)
                {
                    double radiusInMeters = GetPropertyValue<double>("radius", feature);

                    DistanceToPoint(geofence as IPoint, userLocation, radiusInMeters, searchRadiusInMeters, geometry);
                }
                else if (feature.Geometry.OgcGeometryType == OgcGeometryType.MultiPoint)
                {
                    double radiusInMeters = GetPropertyValue<double>("radius", feature);

                    DistanceToMultiPoints(userLocation, geofence as IMultiPoint, udid, geometry, searchRadiusInMeters, radiusInMeters);
                }
                else if (feature.Geometry.OgcGeometryType == OgcGeometryType.LineString)
                {
                    double radiusInMeters = GetPropertyValue<double>("radius", feature);

                    DistanceToLine(userLocation, geofence as ILineString, geometry, searchRadiusInMeters, radiusInMeters);
                }
                else if (feature.Geometry.OgcGeometryType == OgcGeometryType.MultiLineString)
                {
                    double radiusInMeters = GetPropertyValue<double>("radius", feature);

                    DistanceToPolyline(userLocation, geofence as IMultiLineString, udid, geometry, searchRadiusInMeters, radiusInMeters);
                }
                else if (feature.Geometry.OgcGeometryType == OgcGeometryType.Polygon)
                {
                    DistanceToPolygon(userLocation, geofence as IPolygon, geometry, searchRadiusInMeters);
                }
                else
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"The GeoFencing Data with id:{udid} is not supported");
                }

                if (geometry.distance != DefaultInsideDistanceValue)
                {
                    var distanceOp = new NetTopologySuite.Operation.Distance.DistanceOp(userLocation, geofence);

                    geometry.nearestLon = distanceOp.NearestPoints()[1].X;

                    geometry.nearestLat = distanceOp.NearestPoints()[1].Y;
                }
            }

            responseBody.response_code = HttpStatusCode.OK;

            return CreateResponseMessage(responseBody);
        }

        internal void CalculateResponseDistance(Geometry geometry, double distance, double searchRadius, double radius)
        {
            var distToBorder = Math.Abs(distance - radius);

            if (distance < radius)
            {
                // Inside geoFence
                if (distToBorder <= searchRadius)
                {
                    // Within search radius
                    geometry.distance = -distToBorder;
                }
                else
                {
                    // Outside search radius
                    geometry.distance = DefaultInsideDistanceValue;
                }

                responseBody.geometries.Add(geometry);
            }
            else
            {
                // Outside geoFence
                if (distToBorder <= searchRadius)
                {
                    // Within search radius
                    geometry.distance = distToBorder;

                    responseBody.geometries.Add(geometry);
                }
            }
        }

        private static double GetDistanceBetweenCoordinates(Coordinate c1, Coordinate c2)
        {
            var gc1 = new GeoCoordinate(c1.Y, c1.X);

            var gc2 = new GeoCoordinate(c2.Y, c2.X);

            return gc1.GetDistanceTo(gc2);
        }

        private static async Task<FeatureCollection> GetGeoFenceFromStorage(string accountId, string udid)
        {
            var geoJson = string.Empty;

            FeatureCollection featureCollection = null;

            try
            {
                IBlobService blobService = Globals.Instance.BlobService;

                var userDataListResponse = new UserDataListResponse { Files = new List<UserData>() };

                // Determine the container for the user/tenant.
                string containerName = GenerateContainerName(accountId);

                if (!await blobService.DoesContainerExistAsync(containerName))
                {
                    // Make sure the tenant container exists - else this is a bad request from an "invalid tenant".
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Bad request, User geofencing data not found.");
                }

                // Get the blob name using the unique data/document id (udid).
                string blobName = $"geojson/{udid}";

                if (!await blobService.DoesBlockBlobExistAsync(containerName, blobName))
                {
                    // Make sure the blob that needs to be updated exists - else the file 'udid' is invalid.
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"Bad request - Data with id:{udid} not found.");
                }

                CloudBlockBlob blobReference = blobService.GetBlockBlobReference(containerName, blobName);

                geoJson = await blobReference.DownloadTextAsync();
            }
            catch (Exception ex) when (ex is StorageException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "Listing user data failed!", ex);
            }

            try
            {
                var reader = new NetTopologySuite.IO.GeoJsonReader();

                try
                {
                    featureCollection = reader.Read<FeatureCollection>(geoJson);
                }
                catch (ArgumentException)
                {
                    var feature = reader.Read<Feature>(geoJson);
                    featureCollection = new FeatureCollection(new Collection<IFeature> { feature });
                }
            }
            catch (Exception ex)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, $"GeoFencing Data with id:{udid} can't be read as featureCollection or feature.", ex);
            }

            return featureCollection;
        }

        private static bool IsCircle(IFeature feature)
        {
            string subType = GetPropertyValue<string>("subType", feature, false);

            if (subType == null)
            {
                return false;
            }

            return subType.Equals("Circle");
        }

        private static T GetPropertyValue<T>(string key, IFeature feature, bool isRequired = true)
        {
            bool existed = feature.Attributes.Exists(key);

            object value = null;

            if (existed)
            {
                value = feature.Attributes[key];

                try
                {
                    return (T)Convert.ChangeType(value, typeof(T));
                }
                catch (Exception ex)
                {
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, $"The property with key:{key} is not valid in GeoFencing Data.", ex);
                }
            }
            else
            {
                if (isRequired)
                {
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, $"The property with key:{key} is not existing in GeoFencing Data.");
                }
                else
                {
                    return default(T);
                }
            }
        }

        private static string GenerateContainerName(string accountId)
        {
            // Account Id may have some special characters that are not compliant with azure storage container naming guidelines.
            using (MD5 hash = MD5.Create())
            {
                // MD5 hash is fine here for this specific purpose (even though it is not cryptographically secure)
                return string.Concat(ContainerNamePrefix, string.Join("", hash.ComputeHash(Encoding.UTF8.GetBytes(accountId)).Select(ba => ba.ToString("x2")))).ToLowerInvariant();
            }
        }

        private static void CheckTimeValidity(IFeature feature, DateTime usertime)
        {
            var datetime = GetPropertyValue<DateTime>("ValidityTime", feature, false);

            // Return if ValidityTime does not existing in the feature since it is not required
            if (datetime == default(DateTime))
            {
                return;
            }

            if (datetime.Kind == DateTimeKind.Unspecified)
            {
                datetime = DateTime.SpecifyKind(datetime, DateTimeKind.Utc);
            }
            else
            {
                datetime = datetime.ToUniversalTime();
            }

            if (usertime == null)
            {
                usertime = DateTime.UtcNow;
            }

            if (datetime < usertime)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"User Request Time is not in a valid time period");
            }
        }

        /// <summary>
        /// Calculates the distance between point and StringLine.
        /// This is not accurate for large polyline since it use the formula for 2 dimension
        /// This function assumes the polyline is in a reasonable size which assured by the validation service
        /// </summary>
        private static double CalculateDistanceToLine(Coordinate point, Coordinate[] linePoints)
        {
            double distance = double.MaxValue;

            for (int i = 1; i < linePoints.Length; i++)
            {
                double currentDist = 0;
                var linePoint1 = linePoints[i - 1];
                var linePoint2 = linePoints[i];

                double d0 = GetDistanceBetweenCoordinates(point, linePoint1);
                double d1 = GetDistanceBetweenCoordinates(linePoint1, linePoint2);
                double d2 = GetDistanceBetweenCoordinates(linePoint2, point);

                if (Math.Pow(d0, 2) > Math.Pow(d1, 2) + Math.Pow(d2, 2))
                {
                    currentDist = d2;
                }

                if (Math.Pow(d2, 2) > Math.Pow(d1, 2) + Math.Pow(d0, 2))
                {
                    currentDist = d0;
                }

                double halfP = (d0 + d1 + d2) * 0.5;
                double area = Math.Sqrt(halfP * (halfP - d0) * (halfP - d1) * (halfP - d2));

                currentDist = 2 * area / d1;

                if (currentDist < distance)
                {
                    distance = currentDist;
                }
            }

            return distance;
        }

        private void DistanceToMultiPoints(IPoint userLocation, IMultiPoint geofence, string udid, Geometry geometry, double searchRadius, double radius)
        {
            double minDistance = MaximumSearchRadiusInMeters + 1;

            for (int i = 0; i < geofence.Count; i++)
            {
                var point = geofence[i] as IPoint;

                if (point != null)
                {
                    var distance = GetDistanceBetweenCoordinates(point.Coordinate, userLocation.Coordinate);

                    var distToBorder = Math.Abs(radius - distance);

                    if (distToBorder <= searchRadius)
                    {
                        if (distToBorder < minDistance)
                        {
                            minDistance = distToBorder;

                            if (distance < radius)
                            {
                                geometry.distance = -distToBorder;
                            }
                            else
                            {
                                geometry.distance = distToBorder;
                            }
                        }
                    }

                    // Inside the geofence but out of the search radius to the border
                    if (distance < radius && minDistance == MaximumSearchRadiusInMeters + 1)
                    {
                        geometry.distance = DefaultInsideDistanceValue;
                    }
                }
                else
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"The GeoFencing Data with id:{udid} is not a valid multipoint");
                }
            }

            if (minDistance != MaximumSearchRadiusInMeters + 1 || geometry.distance == DefaultInsideDistanceValue)
            {
                responseBody.geometries.Add(geometry);
            }
        }

        private void DistanceToPoint(IPoint geofence, IPoint userLocation, double radius, double searchRadius, Geometry geometry)
        {
            var distanceToPoint = GetDistanceBetweenCoordinates(geofence.Coordinate, userLocation.Coordinate);

            CalculateResponseDistance(geometry, distanceToPoint, searchRadius, radius);
        }

        private void DistanceToLine(IPoint userLocation, ILineString line, Geometry geometry, double searchRadius, double radius)
        {
            var distanceToLine = CalculateDistanceToLine(userLocation.Coordinate, line.Coordinates);

            CalculateResponseDistance(geometry, distanceToLine, searchRadius, radius);
        }

        private void DistanceToPolyline(IPoint userLocation, IMultiLineString geofence, string udid, Geometry geometry, double searchRadius, double radius)
        {
            double minDistance = MaximumSearchRadiusInMeters + 1;

            for (int i = 0; i < geofence.Count; i++)
            {
                var line = geofence[i] as ILineString;

                if (line != null)
                {
                    var distance = CalculateDistanceToLine(userLocation.Coordinate, line.Coordinates);

                    var distToBorder = Math.Abs(distance - radius);

                    if (distToBorder <= searchRadius)
                    {
                        if (distToBorder < minDistance)
                        {
                            minDistance = distToBorder;

                            if (distance < radius)
                            {
                                geometry.distance = -distToBorder;
                            }
                            else
                            {
                                geometry.distance = distToBorder;
                            }
                        }
                    }

                    // inside geofence outside search radius
                    if (distance <= radius && minDistance == MaximumSearchRadiusInMeters + 1)
                    {
                        geometry.distance = DefaultInsideDistanceValue;
                    }
                }
                else
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"The GeoFencing Data with id:{udid} is not a valid polyline");
                }
            }

            if (minDistance != MaximumSearchRadiusInMeters + 1 || geometry.distance == DefaultInsideDistanceValue)
            {
                responseBody.geometries.Add(geometry);
            }
        }

        private void DistanceToPolygon(IPoint userLocation, IPolygon polygon, Geometry geometry, double searchRadius)
        {
            double distance = MaximumSearchRadiusInMeters + 1;

            var coordinates = polygon.Coordinates;

            for (int i = 1; i < coordinates.Length; i++)
            {
                var currentDist = CalculateDistanceToLine(userLocation.Coordinate, coordinates);

                if (currentDist <= searchRadius)
                {
                    if (currentDist < distance)
                    {
                        distance = currentDist;

                        geometry.distance = currentDist;
                    }
                }
            }

            if (polygon.Intersects(userLocation))
            {
                if (distance != MaximumSearchRadiusInMeters + 1)
                {
                    geometry.distance = -geometry.distance;
                }
                else
                {
                    geometry.distance = DefaultInsideDistanceValue;
                }

                responseBody.geometries.Add(geometry);
            }
            else
            {
                if (distance != MaximumSearchRadiusInMeters + 1)
                {
                    responseBody.geometries.Add(geometry);
                }
            }
        }
    }
}