﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Models.Map
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net.Http;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.Utility;

    internal class MicrosoftV2MapProxy : BaseProxyMap
    {
        ProxyHostSettings settings;

        public MicrosoftV2MapProxy(IProxyOperation proxyOperation, ProxyHostSettings settings) : base(proxyOperation)
        {
            this.settings = settings;
        }

        public override Task<HttpResponseMessage> MapCopyright(MapType type, string formatstring, Dictionary<string, List<string>> querystrings, string zoom = "", string x = "", string y = "")
        {
            throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"Request for '{type}' is not supported");
        }

        public override Task<HttpResponseMessage> MapStatic(string formatstring, Dictionary<string, List<string>> querystrings)
        {
            throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"Request for map images is not supported");
        }

        public override async Task<HttpResponseMessage> MapTileRaster(string layer, string style, string zoom, string x, string y, string formatstring, Dictionary<string, List<string>> querystrings)
        {
            this.BuildURL(ref this.URI, MapType.RasterTile, formatstring, querystrings, layer, style, zoom, x, y);
            return await this.ExecuteRequest(CreateRequest(this.URI.Uri));
        }

        public override async Task<HttpResponseMessage> MapTileVector(string layer, string style, string zoom, string x, string y, string formatstring, Dictionary<string, List<string>> querystrings)
        {
            this.BuildURL(ref this.URI, MapType.VectorTile, formatstring, querystrings, layer, style, zoom, x, y);
            return await this.ExecuteRequest(CreateRequest(this.URI.Uri));
        }

        private static HttpRequestMessage CreateRequest(Uri requestUri)
        {
            return new HttpRequestMessage(HttpMethod.Get, requestUri);
        }

        private void BuildURL(
            ref UriBuilder URI,
            MapType type,
            string formatstring,
            Dictionary<string, List<string>> querystrings,
            string layer,
            string style,
            string zoom,
            string x,
            string y)
        {
            this.URI.Scheme = this.settings.MdpApiScheme;
            this.URI.Host = this.settings.MdpApiFqdn;
                        
            this.BuildURLPath(ref this.URI, type, layer, style, zoom, x, y, formatstring);
            this.BuildURLQuery(ref this.URI, querystrings);
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Maintainability", "CA1502:AvoidExcessiveComplexity")]
        private void BuildURLPath(ref UriBuilder URI, MapType mapType, string layer, string style, string zoom, string x, string y, string format)
        {
            switch (mapType)
            {
                case (MapType.RasterTile):
                case (MapType.VectorTile):
                    {
                        if (layer == string.Empty || style == string.Empty || zoom == string.Empty || x == string.Empty || y == string.Empty)
                        {
                            break;
                        }

                        URI.Path = $"tile/{layer}/{style}/{zoom}/{x}/{y}.{format}";
                        break;
                    }

                case (MapType.StaticTile):
                case (MapType.WorldCopyright):
                case (MapType.BoundingCopyright):
                case (MapType.TileCopyright):
                case (MapType.CaptionCopyright):

                default:
                    {
                        URI.Path = $"map/notimplemented";
                        break;
                    }
            }
        }

        private void BuildURLQuery(ref UriBuilder request, Dictionary<string, List<string>> querystrings)
        {
            var query = $"api-version={Uri.EscapeDataString(this.settings.MdpApiVersion)}";
            if (querystrings.Count > 0)
            {
                query += string.Join("&", querystrings.Select(kv => $"{Uri.EscapeDataString(kv.Key)}={Uri.EscapeDataString(kv.Value.First())}"));
            }

            request.Query = query; 
        }
    }
}
