﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net.Http;
    using System.Threading.Tasks;
    using Geography;
    using Microsoft.Azure.LocationServices.Instrumentation;

    public class TomTomV1MapProxy : BaseProxyMap
    {
        private static readonly string TTKey = Globals.Instance.TomTomMapKey;
        private static readonly string TTVersion = Properties.Settings.Default.TomTomMapVersion;
        private static readonly string TTBaseURL = Properties.Settings.Default.TomTomBaseURL;
        private static readonly string TTScheme = Properties.Settings.Default.TomTomScheme;

        public TomTomV1MapProxy(IProxyOperation proxyOperation) : base(proxyOperation)
        {
        }

        public override async Task<HttpResponseMessage> MapTileVector(
            string layer,
            string style,
            string zoom,
            string x,
            string y,
            string formatstring,
            Dictionary<string, List<string>> querystrings)
        {
            this.BuildURL(ref this.URI, MapType.VectorTile, formatstring, querystrings, layer, style, zoom, x, y);
            return await this.ExecuteRequest(CreateRequest(MapType.VectorTile, this.URI.Uri));
        }

        public override async Task<HttpResponseMessage> MapTileRaster(
        string layer,
        string style,
        string zoom,
        string x,
        string y,
        string formatstring,
        Dictionary<string, List<string>> querystrings)
        {
            this.BuildURL(ref this.URI, MapType.RasterTile, formatstring, querystrings, layer, style, zoom, x, y);
            return await this.ExecuteRequest(CreateRequest(MapType.RasterTile, this.URI.Uri));
        }

        public override async Task<HttpResponseMessage> MapStatic(string formatstring, Dictionary<string, List<string>> querystrings)
        {
            /* this is the only TT URL that requires the format to be passed in as a query param.  Adding it here */
            querystrings["format"] = new List<string>() { formatstring };

            // Bug# 2494105, watermark parameter adds Microsoft watermark on the static image.
            // value true is a filler to make sure the parameter is added.
            querystrings["watermark"] = new List<string>() { "true" };

            this.BuildURL(ref this.URI, MapType.StaticTile, string.Empty, querystrings, string.Empty, string.Empty, string.Empty, string.Empty, string.Empty);
            return await this.ExecuteRequest(CreateRequest(MapType.StaticTile, this.URI.Uri));
        }

        public override async Task<HttpResponseMessage> MapCopyright(
            MapType type,
            string formatstring,
            Dictionary<string, List<string>> querystrings,
            string zoom = "",
            string x = "",
            string y = "")
      
        {
            this.BuildURL(ref this.URI, type, formatstring, querystrings, string.Empty, string.Empty, zoom, x, y);
            return await this.ExecuteRequest(CreateRequest(type, this.URI.Uri));
        }

        private static HttpRequestMessage CreateRequest(MapType mapType, Uri requestUri)
        {
            switch (mapType)
            {
                case MapType.RasterTile:
                case MapType.StaticTile:
                case MapType.VectorTile:
                case MapType.BoundingCopyright:
                case MapType.CaptionCopyright:
                case MapType.TileCopyright:
                case MapType.WorldCopyright:
                    return new HttpRequestMessage(HttpMethod.Get, requestUri);

                default:
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, $"Map type {mapType} is not supported");
            }
        }

        private void BuildURL(
            ref UriBuilder URI,
            MapType type,
            string formatstring,
            Dictionary<string, List<string>> querystrings,
            string layer,
            string style,
            string zoom,
            string x,
            string y)
        {
            this.URI.Scheme = TTScheme;
            this.URI.Host = string.Empty;

            URI.Host += TTBaseURL;

            //Override for our LBS map style.  Changing Style from "main" to "main-azure".  However, this only applies to 
            // Raster and Static Tile APIs.   This should apply to Vector tile for consistency and will be fixed later when TT makes the change on their end.
            if (type == MapType.RasterTile || type == MapType.StaticTile)
            {
                style = StyleOverride(style);

                if (type == MapType.StaticTile)
                {
                    querystrings["style"] = new List<string>() { style };
                }
            }

            this.BuildURLPath(ref this.URI, type, layer, style, zoom, x, y, formatstring);
            this.BuildURLQuery(ref this.URI, querystrings);
        }

        private string StyleOverride(string incomingStyle)
        {
            switch (incomingStyle)
            {
                case "main":
                    return "main-azure";

                default:
                    return "main-azure";
            }
        }

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Maintainability", "CA1502:AvoidExcessiveComplexity")]
        private void BuildURLPath(ref UriBuilder URI, MapType mapType, string first, string second, string third, string forth, string fifth, string formatstring)
        {
            switch (mapType)
            {
                case (MapType.RasterTile):
                    {
                        if (first == string.Empty || second == string.Empty || third == string.Empty || forth == string.Empty || fifth == string.Empty)
                        {
                            break;
                        }

                        URI.Path = $"map/" + TTVersion + $"/tile/";
                        URI.Path += first + $"/" + second + $"/" + third + $"/" + forth + $"/" + fifth + $"." + formatstring;
                        break;
                    }

                case (MapType.VectorTile):
                    {
                        if (first == string.Empty || second == string.Empty || third == string.Empty || forth == string.Empty || fifth == string.Empty)
                        {
                            break;
                        }

                        URI.Path = $"map/" + TTVersion + $"/tile/";
                        URI.Path += first + $"/" + second + $"/" + third + $"/" + forth + $"/" + fifth + $"." + formatstring;
                        break;
                    }

                case (MapType.StaticTile):
                    {
                        URI.Path = $"map/" + TTVersion + $"/staticimage";
                        break;
                    }

                case (MapType.WorldCopyright):
                    {
                        URI.Path = $"map/" + TTVersion + $"/copyrights" + $".";
                        URI.Path += formatstring;
                        break;
                    }

                case (MapType.BoundingCopyright):
                    {
                        if (forth == string.Empty || fifth == string.Empty)
                        {
                            break;
                        }

                        Coordinates minCoordinates = new Coordinates(forth);
                        Coordinates maxCoordinates = new Coordinates(fifth);
                        if (minCoordinates.IsNull || maxCoordinates.IsNull)
                        {
                            ProxyException ex = new ProxyException(ProxyException.ExceptionType.ClientParams, "failed to instantiate Coordinates object");
                            throw (ex);
                        }

                        URI.Path = $"map/" + TTVersion + $"/copyrights/" + "/" 
                            + minCoordinates.Longitude + $"/" + minCoordinates.Latitude + $"/" 
                            + maxCoordinates.Longitude + $"/" + maxCoordinates.Latitude + $".";
                        URI.Path += formatstring;
                        break;
                    }

                case (MapType.TileCopyright):
                    {
                        if (third == string.Empty || forth == string.Empty || fifth == string.Empty)
                        {
                            break;
                        }

                        URI.Path = $"map/" + TTVersion + $"/copyrights/" + third + $"/" + forth + $"/" + fifth + $".";
                        URI.Path += formatstring;
                        break;
                    }

                case (MapType.CaptionCopyright):
                    {
                        URI.Path = $"map/" + TTVersion + $"/copyrights/caption" + $".";
                        URI.Path += formatstring;
                        break;
                    }

                default:
                    {
                        URI.Path = $"map/" + TTVersion + $"/tile/";
                        break;
                    }
            }
        }

        private void BuildURLQuery(ref UriBuilder request, Dictionary<string, List<string>> querystrings)
        {
            string queryparam = $"key=" + TTKey;

            for (int index = 0; index < querystrings.Count; index++)
            {
                var item = querystrings.ElementAt(index);
                string itemKey = item.Key;

                foreach (var itemValue in item.Value)
                {
                    if (itemValue == string.Empty
                        || itemValue == "{" + itemKey + "}")
                    {
                        continue;
                    }

                    queryparam += $"&" + itemKey + "=" + itemValue;
                }
            }

            request.Query = queryparam;
        }      
    }
}