﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Net.Http.Formatting;
    using System.Text;
    using System.Threading;
    using System.Threading.Tasks;
    using System.Web;
    using GeoJSON.Net.Geometry;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Common.Storage;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.Models.Common.Batch;
    using Microsoft.Azure.LocationServices.Shared;
    using Microsoft.Azure.LocationServices.Utility;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Queue;
    using Newtonsoft.Json;
    using Newtonsoft.Json.Linq;

    public class TomTomV1RouteProxy : BaseProxyRoute, IRouteCustomSerializationProvider
    {
        private static readonly TimeSpan RouteMatrixTimeOutForAsync = TimeSpan.FromSeconds(20);
        private static readonly TimeSpan RouteMatrixTimeOutForForcedAsync = TimeSpan.FromMilliseconds(1);
        private static readonly string TTKey = Globals.Instance.TomTomRouteKey;
        private static readonly string TTVersion = Properties.Settings.Default.TomTomRouteVersion;
        private static readonly string TTBaseURL = Properties.Settings.Default.TomTomBaseURL;
        private static readonly string TTScheme = Properties.Settings.Default.TomTomScheme;

        public TomTomV1RouteProxy(IProxyOperation proxyOperation) : base(proxyOperation)
        {
        }
        
        public override async Task<HttpResponseMessage> RouteProxy(
            RouteType routeType,
            string querystring,
            string formatstring,
            Dictionary<string, List<string>> querystrings)
        {
            this.BuildURL(ref this.URI, routeType, querystring, formatstring, querystrings);
            this.Format = (ResponseFormat)Enum.Parse(typeof(ResponseFormat), formatstring, true);
            HttpRequestMessage request = CreateRequest(routeType, this.URI.Uri, null);
            return await this.ExecuteRequest(request);
        }

        public override async Task<HttpResponseMessage> RouteProxy(
            RouteType routeType,
            string querystring,
            string formatstring,
            Dictionary<string, List<string>> querystrings,
            string requestBody = null)
        {
            this.BuildURL(ref this.URI, RouteType.Normal, querystring, formatstring, querystrings);
            this.Format = (ResponseFormat)Enum.Parse(typeof(ResponseFormat), formatstring, true);
            HttpRequestMessage request = CreateRequest(routeType, this.URI.Uri, requestBody);
            return await this.ExecuteRequest(request);
        }

        public override async Task<HttpResponseMessage> RouteMatrixProxyGet(Uri requestUri, string matrixRequestId)
        {
            return await this.ProcessGetRouteMatrixRequest(Globals.Instance.BlobService, requestUri, matrixRequestId);
        }

        public override async Task<HttpResponseMessage> RouteMatrixProxyPost(Uri requestUri,
            string formatstring,
            Dictionary<string, List<string>> querystrings,
            string requestBody,
            string accountId,
            string trackingId,
            bool forceAsync)
        {
            IBlobHelper blobHelper = new BlobHelper(Globals.Instance.BlobService);
            return await this.ProcessRouteMatrixPost(Globals.Instance.BlobService, Globals.Instance.QueueService, blobHelper, requestUri, requestBody, formatstring, querystrings, accountId, trackingId, forceAsync);
        }

        public string ConvertLBSPositionsIntoTargetBody(List<Position> originList, List<Position> destinationList)
        {
            var originsObject = GetJsonObjectFromPositionsList(originList, "origins");
            var destinationsObject = GetJsonObjectFromPositionsList(destinationList, "destinations");

            return new JObject(originsObject, destinationsObject).ToString();
        }

        public List<int> GetMatrixResponseCodesFromClientResponse(string responseString)
        {
            return SharedUtilities.GetResponseCodesFromTomTomRouteMatrixResponseContent(responseString);
        }

        private static HttpRequestMessage CreateRequest(RouteType routeType, Uri requestUri, string requestBody)
        {
            switch (routeType)
            {
                case RouteType.Normal:
                case RouteType.ReachableRange:
                    return new HttpRequestMessage(HttpMethod.Get, requestUri);

                case RouteType.Reconstruct:
                case RouteType.Matrix:
                    return new HttpRequestMessage(HttpMethod.Post, requestUri) { Content = CreateJsonRequestBody(requestBody) };

                default:
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, $"Route type {routeType} is not supported");
            }
        }

        private static HttpContent CreateJsonRequestBody(string routeJson)
        {
            return new StringContent(routeJson, Encoding.UTF8, "application/json");
        }

        private static Uri GetBaseUri(Uri uri)
        {
            return new UriBuilder(uri.Scheme, uri.Host).Uri;
        }

        private static async Task WriteInfoToBlobAndQueue(
            string containerName,
            string blobName,
            string blobContent,
            bool responseAvailable,
            bool requestFailed,
            IBlobService blobService,
            IQueueService queueService,
            IBlobHelper blobHelper,
            string queueName,
            bool queueTheBlob)
        {
            // Create the container if it doesn't exist already
            await blobService.CreateContainerIfNotExistsAsync(containerName);

            // Create an empty blob where blob name is a unique id
            var blobReference = blobService.GetBlockBlobReference(containerName, blobName);

            var setBlobExpiration = !blobReference.Exists();
            var metadata = new Dictionary<string, string>
                    {
                        { SharedUtilities.BatchServiceResponseStatusMetadataKey, responseAvailable.ToString() },
                        { SharedUtilities.BatchServiceRequestFailedMetadataKey, requestFailed.ToString() }
                    };

            // Upload the matrix request
            await blobService.UploadDocumentAsync(containerName, blobName, blobContent, metadata);

            if (setBlobExpiration)
            {
                await blobHelper.SetBlobExpiryDate(blobReference, DateTime.UtcNow.AddDays(SharedUtilities.ResponseBlobExpiryTimeInDays));
            }
            
            if (queueTheBlob)
            {
                // Create the queue if it doesn't already exist
                CloudQueue queue = queueService.GetQueueReference(queueName);
                await queueService.CreateQueueIfNotExistsAsync(queue);

                // Add a new work item for this matrix request to the Queue
                await queueService.AddMessageAsync(queue, new CloudQueueMessage(blobReference.Uri.AbsoluteUri));
            }
        }

        private static JProperty GetJsonObjectFromPositionsList(List<Position> positionList, string propertyName)
        {
            JArray cArray = new JArray();
            positionList.ForEach(item => cArray.Add(new JObject(new JProperty("point", new JObject(new JProperty("latitude", item.Latitude), new JProperty("longitude", item.Longitude))))));

            return new JProperty(propertyName, cArray);
        }

        private void BuildURL(
            ref UriBuilder URI,
            RouteType routeType,
            string querystring,
            string formatstring,
            Dictionary<string, List<string>> querystrings)
        {
            URI.Scheme = TTScheme;
            URI.Host = TTBaseURL;
            this.BuildURLPath(ref this.URI, routeType, querystring, formatstring);
            this.BuildURLQuery(ref this.URI, querystrings);
        }

        private void BuildURLPath(ref UriBuilder URI, RouteType routeType, string querystring, string formatstring)
        {           
            switch (routeType)
            {
                case (RouteType.Normal):
                    {
                        URI.Path = $"routing/" + TTVersion + $"/calculateRoute/";
                        break;
                    }

                case (RouteType.ReachableRange):
                    {
                        URI.Path = $"routing/" + TTVersion + $"/calculateReachableRange/";
                        break;
                    }

                case (RouteType.Batch):
                    {
                        URI.Path = $"routing/" + TTVersion + $"/batch/";
                        break;
                    }

                case (RouteType.Matrix):
                    {
                        URI.Path = $"routing/" + TTVersion + $"/matrix";
                        break;
                    }

                default:
                    {
                        URI.Path = $"routing/" + TTVersion + $"/calculate/";
                        break;
                    }
            }

            URI.Path += querystring + $"/" + formatstring;
        }

        private void BuildURLQuery(ref UriBuilder request, Dictionary<string, List<string>> querystrings)
        {
            string queryparam = $"key=" + TTKey;

            for (int index = 0; index < querystrings.Count; index++)
            {
                var item = querystrings.ElementAt(index);
                string itemKey = item.Key;

                foreach (var itemValue in item.Value)
                {
                    if (itemValue == string.Empty
                        || itemValue == "{" + itemKey + "}")
                    {
                        continue;
                    }

                    queryparam += $"&" + itemKey + "=" + itemValue;
                }
            }

            request.Query = queryparam;
        }

        private async Task<HttpResponseMessage> ProcessRouteMatrixPost(
            IBlobService blobService, 
            IQueueService queueService,
            IBlobHelper blobHelper,
            Uri requestUri, 
            string requestBody, 
            string formatstring, 
            Dictionary<string, List<string>> queryStrings, 
            string accountId, 
            string trackingId,
            bool forceAsync)
        {
            var autoResetEvent = new AutoResetEvent(true);
            try
            {
                this.BuildURL(ref this.URI, RouteType.Matrix, string.Empty, formatstring, queryStrings);
                this.Format = (ResponseFormat)Enum.Parse(typeof(ResponseFormat), formatstring, true);
                var request = CreateRequest(RouteType.Matrix, this.URI.Uri, requestBody);

                // Create an empty blob where blob name is a unique id
                var blobName = Guid.NewGuid().ToString();
                
                var taskParamObject = new TaskParams(false, autoResetEvent, request, requestUri, blobService, queueService, blobHelper, blobName, accountId, trackingId);

                var newTask = RouteMatrixWorker(taskParamObject);
                var taskFinished = newTask.Wait(forceAsync ? RouteMatrixTimeOutForForcedAsync : RouteMatrixTimeOutForAsync);

                autoResetEvent.WaitOne();

                // re-check the status as it could've completed just at the same time.
                if (taskFinished || newTask.IsCompleted || newTask.IsFaulted || newTask.IsCanceled)
                {
                    return newTask.Result;
                }
                else
                {
                    taskParamObject.WriteToBlob = true;
                    await WriteInfoToBlobAndQueue(
                        SharedUtilities.RouteMatrixServiceResultsContainerName,
                        blobName,
                        "Temp Place holder",
                        false,
                        false,
                        blobService,
                        queueService,
                        blobHelper,
                        SharedUtilities.RouteMatrixServicePrimaryQueueName,
                        false);

                    return BatchApiProxyHelper.GetRequestAcceptedResponse(requestUri, new Dictionary<string, string> { { Constants.RouteMatrixResponseUniqueIdKey, blobName } });
                }
            }
            catch (Exception ex) when (ex is StorageException || ex is UriFormatException || ex is JsonSerializationException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "Route Matrix Processing failed!", ex);
            }
            finally
            {
                autoResetEvent.Set();
            }
        }

        private async Task<HttpResponseMessage> RouteMatrixWorker(TaskParams taskParams)
        {
            HttpResponseMessage initialResponse = null;

            try
            {
                initialResponse = await this.ExecuteRequestWithCustomTimeoutNoLogging(taskParams.Request);
            }
            catch (Exception exception)
            {
                taskParams.LockEvent.WaitOne();
                if (!taskParams.WriteToBlob)
                {
                    taskParams.LockEvent.Set();
                    throw;
                }

                var failureObj = new RouteMatrixFailureInfo() { ExceptionType = exception.GetType(), Message = exception.Message };
                await WriteInfoToBlobAndQueue(
                            SharedUtilities.RouteMatrixServiceResultsContainerName,
                            taskParams.BlobName,
                            JsonConvert.SerializeObject(failureObj),
                            true,
                            true,
                            taskParams.BlobService,
                            taskParams.QueueService,
                            taskParams.BlobHelperService,
                            SharedUtilities.RouteMatrixServicePrimaryQueueName,
                            false);
                taskParams.LockEvent.Set();
                return null;
            }

            try
            {
                taskParams.LockEvent.WaitOne();
                if (initialResponse.StatusCode != HttpStatusCode.Accepted && initialResponse.StatusCode != HttpStatusCode.SeeOther)
                {
                    if (taskParams.WriteToBlob)
                    {
                        var requestFailed = initialResponse.StatusCode != HttpStatusCode.OK;

                        var responseContent = initialResponse.Content.ReadAsStringAsync().Result;

                        var responseStatusCodes = SharedUtilities.GetResponseCodesFromTomTomRouteMatrixResponseContent(responseContent);
                        if (!string.IsNullOrEmpty(taskParams.AccountId))
                        {
                            var statusList = responseStatusCodes.Select(t => new Tuple<string, HttpStatusCode>(SharedUtilities.PostMatrixMethodName, (HttpStatusCode)t)).ToList();
                            BillingAPI.BillingLogWriter.Instance.WriteLog(taskParams.AccountId, statusList, taskParams.TrackingId);
                        }

                        // Waiting since we can't use await inside a lock 
                        await WriteInfoToBlobAndQueue(
                            SharedUtilities.RouteMatrixServiceResultsContainerName,
                            taskParams.BlobName,
                            responseContent,
                            true,
                            requestFailed,
                            taskParams.BlobService,
                            taskParams.QueueService,
                            taskParams.BlobHelperService,
                            SharedUtilities.RouteMatrixServicePrimaryQueueName,
                            false);

                        return null;
                    }
                    else
                    {
                        return initialResponse;
                    }
                }

                var retryInfo = new RouteMatrixRetryInfo(GetBaseUri(this.URI.Uri), initialResponse.Headers.Location, taskParams.AccountId, taskParams.TrackingId);
                    
                // Waiting since we can't use await inside a lock 
                await WriteInfoToBlobAndQueue(
                    SharedUtilities.RouteMatrixServiceResultsContainerName,
                    taskParams.BlobName,
                    JsonConvert.SerializeObject(retryInfo),
                    false,
                    false,
                    taskParams.BlobService,
                    taskParams.QueueService,
                    taskParams.BlobHelperService,
                    SharedUtilities.RouteMatrixServicePrimaryQueueName,
                    true);

                return BatchApiProxyHelper.GetRequestAcceptedResponse(taskParams.RequestUri, new Dictionary<string, string> { { Constants.RouteMatrixResponseUniqueIdKey, taskParams.BlobName } });
            }
            catch (Exception ex) when (ex is StorageException || ex is UriFormatException || ex is JsonSerializationException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "Route Matrix Processing failed!", ex);
            }
            finally
            {
                taskParams.LockEvent.Set();
            }
        }

        private async Task<HttpResponseMessage> ProcessGetRouteMatrixRequest(IBlobService blobService, Uri requestUri, string requestId)
        {
            try
            {
                // First check if the batch-id is (still) valid.
                bool isRequestIdValid = await blobService.DoesBlockBlobExistAsync(SharedUtilities.RouteMatrixServiceResultsContainerName, requestId);

                if (!isRequestIdValid)
                {
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Invalid RequestId - Either value is incorrect or has expired");
                }

                // Get the blob metadata and check if response is ready
                IDictionary<string, string> metadata = await blobService.GetBlockBlobMetadataAsync(SharedUtilities.RouteMatrixServiceResultsContainerName, requestId);

                if (!metadata.TryGetValue(SharedUtilities.BatchServiceResponseStatusMetadataKey, out string responseStatus) ||
                    !metadata.TryGetValue(SharedUtilities.BatchServiceRequestFailedMetadataKey, out string requestFailedStatus))
                {
                    // Just a safe check. This should not happen under any circumstances.
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, "Unexpected Error - Matrix status metadata is missing in blob");
                }

                bool hasRequestFailed = bool.Parse(requestFailedStatus);
                bool isResponseAvailable = bool.Parse(responseStatus);

                // First check if the request failed or not
                if (hasRequestFailed)
                {
                    var failureObj = await blobService.DownloadAsync<RouteMatrixFailureInfo>(SharedUtilities.RouteMatrixServiceResultsContainerName, requestId);

                    throw new ProxyException(failureObj.ExceptionType == typeof(ProxyException) ? ProxyException.ExceptionType.ExtError : ProxyException.ExceptionType.Other, failureObj.Message);
                }

                // Request has not failed - Check if it's still ongoing or is a response already available.
                if (!isResponseAvailable)
                {
                    // Still processing the request. No response available yet. Return 202 Accepted with the location header set to the callback uri
                    var stillProcessingResponse = new HttpResponseMessage(HttpStatusCode.Accepted);
                    stillProcessingResponse.Headers.Location = BatchApiProxyHelper.GetUpdatedUri(requestUri);
                    return stillProcessingResponse;
                }

                // Request has successfully completed. Download response data from blob
                var responseObj = await blobService.DownloadAsync<object>(SharedUtilities.RouteMatrixServiceResultsContainerName, requestId);

                // Return 200 OK with the deserilized and formatted response body
                return new HttpResponseMessage(HttpStatusCode.OK)
                {
                    Content = new ObjectContent<object>(responseObj, new JsonMediaTypeFormatter(), Constants.JsonMediaType)
                };
            }
            catch (Exception ex) when(ex is StorageException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "Route Matrix Processing failed!", ex);
            }
        }
        
        class TaskParams
        {
            public TaskParams(bool writeToBlob, AutoResetEvent autoResetEvent, HttpRequestMessage request, Uri requestUri, IBlobService blobService, IQueueService queueService, IBlobHelper blobHelper, string blobName, string accountId, string trackingId)
            {
                this.WriteToBlob = writeToBlob;
                this.LockEvent = autoResetEvent;
                this.Request = request;
                this.RequestUri = requestUri;
                this.BlobService = blobService;
                this.QueueService = queueService;
                this.BlobHelperService = blobHelper;
                this.BlobName = blobName;
                this.AccountId = accountId;
                this.TrackingId = trackingId;
            }

            public bool WriteToBlob { get; set; }

            public AutoResetEvent LockEvent { get; }

            public HttpRequestMessage Request { get; }

            public IBlobService BlobService { get; }

            public IQueueService QueueService { get; }

            public IBlobHelper BlobHelperService { get; }

            public string BlobName { get; }

            public string AccountId { get; }

            public string TrackingId { get; }

            public Uri RequestUri { get; }
        }
    }
}