// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Models.Search.Converters
{
    using System;
    using GeoJSON.Net.Converters;
    using GeoJSON.Net.Geometry;
    using Newtonsoft.Json;

    /// <summary>
    ///     Converter that serilaizes a <see cref="LineString"/> geometry into 
    ///     a valid JSON string for TomTom specific APIs.
    /// </summary>
    public class TomTomV2SearchAlongRouteJsonConverter : JsonConverter
    {
        /// <summary>
        /// Determines whether this instance can convert the specified object type.
        /// </summary>
        /// <param name="objectType">Type of the object.</param>
        /// <returns>
        ///     <c>true</c> if this instance can convert the specified object type; otherwise, <c>false</c>.
        /// </returns>
        public override bool CanConvert(Type objectType)
        {
            return typeof(LineString).IsAssignableFromType(objectType);
        }

        /// <summary>
        /// Reads the JSON representation of the object.
        /// </summary>
        /// <param name="reader">The <see cref="T:Newtonsoft.Json.JsonReader" /> to read from.</param>
        /// <param name="objectType">Type of the object.</param>
        /// <param name="existingValue">The existing value of object being read.</param>
        /// <param name="serializer">The calling serializer.</param>
        /// <returns>
        ///     The object value.
        /// </returns>
        public override object ReadJson(JsonReader reader, Type objectType, object existingValue, JsonSerializer serializer)
        {
            throw new NotImplementedException();
        }

        /* 
           This is a sample of what TomTom's Geometry Search API expects the input geometry to look like:
           
           {
               "route": {
                   "points": [
                       { "lat": 37.75241, "lon":-122.43576 },
                       { "lat": 37.70660, "lon":-122.43301 }, 
                       { "lat": 37.71205, "lon":-122.36434 }, 
                       { "lat": 37.75350, "lon":-122.37396 }
                   ]
               }
           }
        */

        /// <summary>
        /// Serialize <see cref="LineString"/> GeoJSON object to TomTom's route object
        /// </summary>
        /// <param name="writer">The <see cref="T:Newtonsoft.Json.JsonWriter" /> to write to.</param>
        /// <param name="value">The value.</param>
        /// <param name="serializer">The calling serializer.</param>
        public override void WriteJson(JsonWriter writer, object value, JsonSerializer serializer)
        {
            if (value is LineString lineString)
            {
                // Start
                writer.WriteStartObject();

                writer.WritePropertyName("route");
                writer.WriteStartObject();

                writer.WritePropertyName("points");
                writer.WriteStartArray();

                WriteJsonForLineStringGeometry(writer, lineString);

                writer.WriteEndArray();

                writer.WriteEndObject();

                // End
                writer.WriteEndObject();
            }
            else
            {
                throw new ArgumentException($"Error - Input should be of type: {typeof(LineString)}");
            }
        }
        
        private static void WriteJsonForLineStringGeometry(JsonWriter writer, LineString lineString)
        {
            foreach (IPosition position in lineString.Coordinates)
            {
                writer.WriteStartObject();

                writer.WritePropertyName("lat");
                writer.WriteValue(position.Latitude);

                writer.WritePropertyName("lon");
                writer.WriteValue(position.Longitude);

                writer.WriteEndObject();
            }
        }
    }
}