﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Models.Search.Converters
{
    using System;
    using GeoJSON.Net;
    using GeoJSON.Net.Converters;
    using GeoJSON.Net.Feature;
    using GeoJSON.Net.Geometry;
    using Microsoft.Azure.LocationServices.Utility;
    using Newtonsoft.Json;

    /// <summary>
    /// Converter that serilaizes a <see cref="FeatureCollection" />, 
    /// <see cref="GeometryCollection" />, <see cref="Polygon" /> 
    /// into a valid JSON string for TomTom specific APIs.
    /// </summary>
    public class TomTomV2SearchByGeometryJsonConverter : JsonConverter
    {
        /// <summary>
        /// Determines whether this instance can convert the specified object type.
        /// </summary>
        /// <param name="objectType">Type of the object.</param>
        /// <returns>
        ///     <c>true</c> if this instance can convert the specified object type; otherwise, <c>false</c>.
        /// </returns>
        public override bool CanConvert(Type objectType)
        {
            return typeof(FeatureCollection).IsAssignableFromType(objectType) || 
                typeof(GeometryCollection).IsAssignableFromType(objectType) || 
                typeof(Polygon).IsAssignableFromType(objectType);
        }

        /// <summary>
        /// Reads the JSON representation of the object.
        /// </summary>
        /// <param name="reader">The <see cref="T:Newtonsoft.Json.JsonReader" /> to read from.</param>
        /// <param name="objectType">Type of the object.</param>
        /// <param name="existingValue">The existing value of object being read.</param>
        /// <param name="serializer">The calling serializer.</param>
        /// <returns>
        ///     The object value.
        /// </returns>
        public override object ReadJson(JsonReader reader, Type objectType, object existingValue, JsonSerializer serializer)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// Writes the JSON representation of the object.
        /// </summary>
        /// <param name="writer">The <see cref="T:Newtonsoft.Json.JsonWriter" /> to write to.</param>
        /// <param name="value">The value.</param>
        /// <param name="serializer">The calling serializer.</param>
        public override void WriteJson(JsonWriter writer, object value, JsonSerializer serializer)
        {
            // Serialize GeoJSON objects to TomTom's geometry objects
            if (value is IGeoJSONObject geoJsonGeometry)
            {
                // Start
                writer.WriteStartObject();

                writer.WritePropertyName("geometryList");
                writer.WriteStartArray();

                switch (geoJsonGeometry.Type)
                {
                    case GeoJSONObjectType.FeatureCollection:
                        {
                            FeatureCollection featureCollection = (FeatureCollection)geoJsonGeometry;

                            foreach (Feature feature in featureCollection.Features)
                            {
                                switch (feature.Geometry.Type)
                                {
                                    case GeoJSONObjectType.Polygon:
                                        WriteJsonForPolygonGeometry(writer, (Polygon)feature.Geometry);
                                        break;

                                    case GeoJSONObjectType.Point:
                                        WriteJsonForCircleGeometry(writer, feature);
                                        break;

                                    default:
                                        throw new ArgumentException($"Input contains invalid type: {feature.Geometry.Type}");
                                }
                            }

                            break;
                        }

                    case GeoJSONObjectType.GeometryCollection:
                        {
                            GeometryCollection geometryCollection = (GeometryCollection)geoJsonGeometry;

                            foreach (IGeometryObject geometry in geometryCollection.Geometries)
                            {
                                switch (geometry.Type)
                                {
                                    case GeoJSONObjectType.Polygon:
                                        WriteJsonForPolygonGeometry(writer, (Polygon)geometry);
                                        break;

                                    default:
                                        throw new ArgumentException($"Input contains invalid type: {geometry.Type}");
                                }
                            }

                            break;
                        }

                    case GeoJSONObjectType.Polygon:
                        {
                            WriteJsonForPolygonGeometry(writer, (Polygon)geoJsonGeometry);
                            break;
                        }

                    default:
                        throw new ArgumentException($"Input contains invalid type: {geoJsonGeometry.Type}");
                }
                  
                // "geometryList" End
                writer.WriteEndArray();
                
                // End
                writer.WriteEndObject();
            }
            else
            {
                throw new ArgumentException($"Error - Input should be of type: {typeof(IGeoJSONObject)}");
            }
        }

        /* 
           This is a sample of what TomTom's Geometry Search API expects Polygon geometry to look like:
           
           {
               "type":"POLYGON",
               "vertices":[
                   "37.7524152343544,-122.43576049804686",
                   "37.70660472542312,-122.43301391601562",
                   "37.712059855877314,-122.36434936523438",
                   "37.75350561243041,-122.37396240234374"
               ]
           }         
        */
        private static void WriteJsonForPolygonGeometry(JsonWriter writer, Polygon polygon)
        {
            writer.WriteStartObject();

            writer.WritePropertyName("type");
            writer.WriteValue(GeoJSONObjectType.Polygon.ToString().ToUpper());
            
            writer.WritePropertyName("vertices");

            writer.WriteStartArray();

            foreach (LineString lineString in polygon.Coordinates)
            {
                foreach (IPosition position in lineString.Coordinates)
                {
                    // TomTom APIs expect the order to be "lat, long" unlike GeoJSON where 'long' comes first.
                    writer.WriteValue($"{position.Latitude},{position.Longitude}");
                }
            }

            writer.WriteEndArray();

            writer.WriteEndObject();
        }

        /* 
           This is a sample of what TomTom's Geometry Search API expects Circle geometry to look like:
           
           {
               "type":"CIRCLE",
               "position":"37.71205,-121.36434",
               "radius":6000
           }         
        */
        private static void WriteJsonForCircleGeometry(JsonWriter writer, Feature feature)
        {
            Point point = (Point)feature.Geometry;

            writer.WriteStartObject();

            writer.WritePropertyName("type");
            writer.WriteValue("CIRCLE");

            writer.WritePropertyName("position");
            writer.WriteValue($"{point.Coordinates.Latitude},{point.Coordinates.Longitude}");

            // Validate 'radius' is included in the properties section and has a valid value.
            if (feature.Properties.TryGetValue(Constants.GeoJsonCircleRadiusKey, out object radiusObject))
            {
                // Check if a valid value is provided.
                if (radiusObject == null || !double.TryParse(radiusObject.ToString(), out double radiusValue))
                {
                    throw new ArgumentException("Invalid radius value");
                }

                writer.WritePropertyName(Constants.GeoJsonCircleRadiusKey);
                writer.WriteValue(radiusValue);
            }
            else
            {
                throw new ArgumentException("Radius property for the feature is missing");
            }

            writer.WriteEndObject();
        }
    }
}