﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Text;
    using System.Threading.Tasks;
    using GeoJSON.Net;
    using Microsoft.Azure.LocationServices.Geography;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.Models.Search.Converters;
    using Microsoft.Azure.LocationServices.Utility;
    using Newtonsoft.Json;

    public class TomTomV2SearchProxy : BaseProxySearch, ISearchCustomSerializationProvider
    {
        private static readonly string TTSearchKey = Globals.Instance.TomTomSearchKey;
        private static readonly string TTReverseGeocodingKey = Globals.Instance.TomTomReverseGeocodingKey;
        private static readonly string TTVersion = Properties.Settings.Default.TomTomSearchVersion;
        private static readonly string TTBaseURL = Properties.Settings.Default.TomTomBaseURL;
        private static readonly string TTScheme = Properties.Settings.Default.TomTomScheme;

        private string TTKey;
        private Dictionary<string, List<string>> localquerystrings;

        public TomTomV2SearchProxy(IProxyOperation proxyOperation) : base(proxyOperation)
        {
        }

        public override async Task<HttpResponseMessage> SearchProxy(
            SearchType searchType,
            string querystring,
            string formatstring,
            Dictionary<string, List<string>> querystrings,
            string position,
            string heading,
            string requestBody = null)
        {
            return await this.SearchProxyInternal(searchType, querystring, formatstring, querystrings, position, heading, requestBody);
        }

        public override async Task<HttpResponseMessage> SearchProxy(
            SearchType searchType,
            string querystring,
            string formatstring,
            Dictionary<string, List<string>> querystrings,
            string requestBody = null)
        {
            return await this.SearchProxyInternal(searchType, querystring, formatstring, querystrings, null, null, requestBody);
        }

        /// <summary>
        /// Serializes the input GeoJson geometry object to a custom string object.
        /// </summary>
        /// <param name="searchType">Type of search</param>
        /// <param name="geoJsonGeometry">GeoJSON geometry object</param>
        /// <returns>A custom serialized string</returns>
        public string SerializeGeoJsonGeometry(SearchType searchType, IGeoJSONObject geoJsonGeometry)
        {
            return
                ExecuteSerializationOperation(
                    JsonConvert.SerializeObject,
                    geoJsonGeometry,
                    new[] { GetConverterBySearchType(searchType) });
        }

        private static string ExecuteSerializationOperation(
            Func<object, JsonConverter[], string> serOp, 
            object value, 
            JsonConverter[] converters)
        {
            try
            {
                return serOp(value, converters ?? new JsonConverter[0]);
            }
            catch (Exception ex) when (ex is JsonSerializationException || ex is ArgumentException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Serialization to Json failed", ex);
            }
        }

        private static JsonConverter GetConverterBySearchType(SearchType searchType)
        {
            switch (searchType)
            {
                case SearchType.AlongRoute:
                    return new TomTomV2SearchAlongRouteJsonConverter();

                case SearchType.Geometry:
                    return new TomTomV2SearchByGeometryJsonConverter();

                default:
                    throw new ArgumentException($"No converter found for search type: {searchType.ToString()}");
            }
        }

        private static HttpRequestMessage CreateRequest(SearchType searchType, Uri requestUri, string requestBody)
        {
            switch (searchType)
            {
                case SearchType.Fuzzy:
                case SearchType.Nearby:
                case SearchType.POI:
                case SearchType.POICategory:
                case SearchType.Address:
                case SearchType.AddressStructured:
                case SearchType.ReverseAddress:
                case SearchType.CrossStreet:
                case SearchType.PolygonData:
                    return new HttpRequestMessage(HttpMethod.Get, requestUri);

                case SearchType.Geometry:
                case SearchType.AlongRoute:
                    return new HttpRequestMessage(HttpMethod.Post, requestUri) { Content = CreateJsonRequestBody(requestBody) };

                default:
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, $"Search type {searchType} is not supported");
            }
        }

        /// <summary>
        /// Create HttpContent of type application/json from the serialized input object.
        /// </summary>
        /// <param name="requestBody">Input GeoJson geometry serialized to json string</param>
        private static HttpContent CreateJsonRequestBody(string requestBody)
        {
            return new StringContent(requestBody, Encoding.UTF8, Constants.JsonMediaType);
        }

        private async Task<HttpResponseMessage> SearchProxyInternal(
            SearchType searchType,
            string querystring,
            string formatstring,
            Dictionary<string, List<string>> querystrings,
            string position,
            string heading,
            string requestBody)
        {
            this.BuildURL(ref this.URI, searchType, querystring, formatstring, querystrings, position, heading);
            HttpRequestMessage request = CreateRequest(searchType, this.URI.Uri, requestBody);

            return await this.ExecuteRequest(request);
        }

        private void BuildURL(
            ref UriBuilder URI,
            SearchType searchType,
            string query,
            string format,
            Dictionary<string, List<string>> querystrings,
            string position,
            string heading)
        {
            URI.Scheme = TTScheme;
            URI.Host = TTBaseURL;

            this.localquerystrings = querystrings;

            this.BuildURLPath(ref URI, searchType, query, format, position, heading);
            this.BuildURLQuery(ref URI, this.localquerystrings);
        }

        private void BuildURLPath(
            ref UriBuilder URI,
            SearchType searchType,
            string query,
            string format,
            string position,
            string heading)
        {
            switch (searchType)
            {
                case SearchType.Fuzzy:
                    {
                        this.TTKey = TTSearchKey;
                        URI.Path = $"search/{TTVersion}/search/{WebUtility.UrlEncode(query)}.{format}";
                        break;
                    }

                case SearchType.Nearby:
                    {
                        this.TTKey = TTSearchKey;
                        URI.Path = $"search/{TTVersion}/nearbySearch/.{format}";
                        break;
                    }

                case SearchType.POI:
                    {
                        this.TTKey = TTSearchKey;
                        URI.Path = $"search/{TTVersion}/poiSearch/{WebUtility.UrlEncode(query)}.{format}";
                        break;
                    }

                case SearchType.POICategory:
                    {
                        this.TTKey = TTSearchKey;
                        URI.Path = $"search/{TTVersion}/categorySearch/{WebUtility.UrlEncode(query)}.{format}";
                        break;
                    }

                case SearchType.Address:
                    {
                        this.TTKey = TTSearchKey;
                        URI.Path = $"search/{TTVersion}/geocode/{WebUtility.UrlEncode(query)}.{format}";
                        break;
                    }

                case SearchType.AddressStructured:
                    {
                        this.TTKey = TTSearchKey;
                        URI.Path = $"search/{TTVersion}/structuredGeocode.{format}";
                        break;
                    }

                case SearchType.ReverseAddress:
                    {
                        this.TTKey = TTReverseGeocodingKey;

                        Coordinates coordinates = new Coordinates(query);
                        if (coordinates.IsNull)
                        {
                            throw new ProxyException(ProxyException.ExceptionType.ClientParams, "failed to instantiate Coordinates object");
                        }

                        List<string> tempList = new List<string>();
                        tempList.Add(coordinates.ToString().Replace(" ", ""));
                        this.localquerystrings.Add("position", tempList);

                        URI.Path = $"search/{TTVersion}/reverseGeocode/{format}";
                        break;
                    }

                case SearchType.CrossStreet:
                    {
                        this.TTKey = TTReverseGeocodingKey;
                        Coordinates coordinates = new Coordinates(query);
                        if (coordinates.IsNull)
                        {
                            throw new ProxyException(ProxyException.ExceptionType.ClientParams, "failed to instantiate Coordinates object");
                        }

                        URI.Path = $"search/{TTVersion}/reverseGeocode/crossStreet/{coordinates.Latitude},{coordinates.Longitude}.{format}";
                        break;
                    }

                case SearchType.Geometry:
                    {
                        this.TTKey = TTSearchKey;
                        URI.Path = $"search/{TTVersion}/geometrySearch/{WebUtility.UrlEncode(query)}.{format}";
                        break;
                    }

                case SearchType.AlongRoute:
                    {
                        this.TTKey = TTSearchKey;
                        URI.Path = $"search/{TTVersion}/searchAlongRoute/{WebUtility.UrlEncode(query)}.{format}";
                        break;
                    }

                case SearchType.PolygonData:
                    {
                        this.TTKey = TTSearchKey;
                        URI.Path = $"search/{TTVersion}/additionalData.{format}";
                        break;
                    }

                default:
                    {
                        this.TTKey = TTSearchKey;
                        URI.Path = $"search/{TTVersion}/search/{WebUtility.UrlEncode(query)}.{format}";
                        break;
                    }
            }
        }

        private void BuildURLQuery(ref UriBuilder request, Dictionary<string, List<string>> querystrings)
        {
            string queryparam = $"key={this.TTKey}";

            if (querystrings != null)
            {
                for (int index = 0; index < querystrings.Count; index++)
                {
                    KeyValuePair<string, List<string>> item = querystrings.ElementAt(index);
                    string itemKey = item.Key;

                    foreach (var itemValue in item.Value)
                    {
                        if (itemValue == string.Empty || itemValue == "{" + itemKey + "}")
                        {
                            continue;
                        }

                        queryparam += $"&{itemKey}={WebUtility.UrlEncode(itemValue)}";
                    }
                }

                request.Query = queryparam;
            }
        }
    }
}