﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Net;
    using System.Net.Http;
    using System.Net.Http.Headers;
    using System.Text;
    using System.Threading.Tasks;
    using Geography;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Newtonsoft.Json;
    using TimeZoneDb;

    public class MicrosoftV1TimezoneProxy : BaseProxyTimezone
    {
        public static readonly HttpResponseMessage BadRequest = new HttpResponseMessage(HttpStatusCode.BadRequest);
        private static readonly HttpResponseMessage _sBadGateway = new HttpResponseMessage(HttpStatusCode.BadGateway);
        private static readonly HttpResponseMessage _sNoContent = new HttpResponseMessage(HttpStatusCode.NoContent);
        private static readonly JsonSerializerSettings s_JsonSerializeSettings = new JsonSerializerSettings() { NullValueHandling = NullValueHandling.Ignore };

        public MicrosoftV1TimezoneProxy(IProxyOperation proxyOperation) : base(proxyOperation)
        {
        }

        public override async Task<HttpResponseMessage> TimezoneProxy(
            TimezoneType timezoneType,
            string queryParam1,
            string outputFormat,
            Dictionary<string, List<string>> headers,
            Dictionary<string, List<string>> querystrings)
        {
            if (string.IsNullOrEmpty(outputFormat) || outputFormat.ToLowerInvariant() != "json")
            {
                return BadRequest;
            }

            if (!TimeZoneDb.TimeZoneDb.IsQueriable)
            {
                return _sNoContent;
            }

            // have to manually log this as TZ doesn't call the base class methods for creating a HttpResponse
            Log_PROXY_EXT("Function: " + timezoneType.ToString() + "," + "Params: " + queryParam1.ToString() + "|" + outputFormat.ToString());
            DateTime EXTStart = DateTime.Now;
            var sw = new Stopwatch();
            sw.Start();
            var result = await Task<HttpResponseMessage>.Run(
                delegate
                {
                    switch (timezoneType)
                    {
                        case (TimezoneType.ByID):
                            return FindTimeZoneById(queryParam1, headers, querystrings);

                        case (TimezoneType.ByCoordinates):
                            return FindTimeZoneByCoordinates(queryParam1, headers, querystrings);

                        case (TimezoneType.EnumIANA):
                            return EnumIANA();

                        case (TimezoneType.EnumWindows):
                            return EnumWindows();

                        case (TimezoneType.WindowsToIANA):
                            return WindowsToIANA(queryParam1, querystrings);

                        case (TimezoneType.IANAVersion):
                            return TimeZoneIANAVersion();

                        default:
                            return BadRequest;
                    }
                });
            sw.Stop();
            Log_EXT_ms((int)sw.ElapsedMilliseconds);
            Log_EXT_PROXY(result.ToString(), (int)result.StatusCode);

            return result;
        }

        private static HttpResponseMessage CreateResponse(object tzResult)
        {
            var response = new HttpResponseMessage(HttpStatusCode.OK);
            var jString = JsonConvert.SerializeObject(tzResult, Formatting.Indented, s_JsonSerializeSettings);
            response.Content = new StringContent(jString, Encoding.UTF8, "application/json");
            return response;
        }

        private HttpResponseMessage FindTimeZoneById(string queryString, Dictionary<string, List<string>> headers, Dictionary<string, List<string>> queryStrings)
        {
            try
            {
                TimeZoneQueryParams timeZoneQueryParams;
                string errorDetails;
                if (!TimeZoneQueryParams.CreateFromRestParams(queryStrings, headers, out timeZoneQueryParams, out errorDetails))
                {
                    var response = BadRequest;
                    if (!string.IsNullOrEmpty(errorDetails))
                    {
                        response = new HttpResponseMessage(HttpStatusCode.BadRequest);
                        response.Content = new StringContent(errorDetails);
                        response.Content.Headers.ContentType = new MediaTypeHeaderValue("text/plain");
                    }

                    return response;
                }

                var tzResult = TimeZoneDb.TimeZoneDb.Instance.FindTimeZone(queryString, timeZoneQueryParams);
                return CreateResponse(tzResult);
            }
            catch (Exception ex)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "internalerror", ex);
            }
        }

        private HttpResponseMessage FindTimeZoneByCoordinates(string queryString, Dictionary<string, List<string>> headers, Dictionary<string, List<string>> queryStrings)
        {
            try
            {
                TimeZoneQueryParams timeZoneQueryParams;
                string errorDetails;
                if (!TimeZoneQueryParams.CreateFromRestParams(queryStrings, headers, out timeZoneQueryParams, out errorDetails))
                {
                    var response = BadRequest;
                    if (!string.IsNullOrEmpty(errorDetails))
                    {
                        response = new HttpResponseMessage(HttpStatusCode.BadRequest);
                        response.Content = new StringContent(errorDetails);
                        response.Content.Headers.ContentType = new MediaTypeHeaderValue("text/plain");
                    }

                    return response;
                }

                Coordinates coord = new Coordinates(queryString);
                if (coord.IsNull)
                {
                    return BadRequest;
                }

                var tzResult = TimeZoneDb.TimeZoneDb.Instance.FindTimeZone(ref coord, timeZoneQueryParams);
                return CreateResponse(tzResult);
            }
            catch (Exception ex)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "internalerror", ex);
            }
        }

        private HttpResponseMessage TimeZoneIANAVersion()
        {
            try
            {
                var tzResult = TimeZoneDb.TimeZoneDb.Instance.Version;
                return CreateResponse(tzResult);
            }
            catch (Exception ex)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "internalerror", ex);
            }
        }

        private HttpResponseMessage EnumIANA()
        {
            try
            {
                var tzResult = TimeZoneDb.TimeZoneDb.Instance.EnumerateTimeZoneIds();
                return CreateResponse(tzResult);
            }
            catch (Exception ex)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "internalerror", ex);
            }
        }

        private HttpResponseMessage EnumWindows()
        {
            try
            {
                var tzResult = TimeZoneDb.TimeZoneDb.Instance.EnumerateWindowsTimeZones();
                return CreateResponse(tzResult);
            }
            catch (Exception ex)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "internalerror", ex);
            }
        }

        private HttpResponseMessage WindowsToIANA(string querystring, Dictionary<string, List<string>> querystrings)
        {
            try
            {
                if (string.IsNullOrEmpty(querystring))
                {
                    response = new HttpResponseMessage(HttpStatusCode.BadRequest);
                    response.Content = new StringContent("No value provided for the 'windowszone' parameter");
                    response.Content.Headers.ContentType = new MediaTypeHeaderValue("text/plain");
                    return response;
                }

                string territory = null;
                if (querystrings != null && querystrings.Count > 0)
                {
                    foreach (var kvp in querystrings)
                    {
                        var key = kvp.Key.ToLowerInvariant();
                        if (key == "territory" || key == "t")
                        {
                            if (territory == null)
                            {
                                if (kvp.Value != null && kvp.Value.Count > 0 && !string.IsNullOrEmpty(kvp.Value[0]))
                                {
                                    territory = kvp.Value[0];
                                }
                            }
                            else
                            {
                                response = new HttpResponseMessage(HttpStatusCode.BadRequest);
                                response.Content = new StringContent("The 'territory' parameter was submitted more than once");
                                response.Content.Headers.ContentType = new MediaTypeHeaderValue("text/plain");
                                return response;
                            }
                        }
                    }
                }

                var tzResult = TimeZoneDb.TimeZoneDb.Instance.WindowsToIanaTimeZone(windowsZone: querystring, territory: territory);
                return CreateResponse(tzResult);
            }
            catch (Exception ex)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "internalerror", ex);
            }
        }
    }
}