﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net.Http;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Instrumentation;

    public class TomTomV4TrafficProxy : BaseProxyTraffic
    {
        private static readonly string TTTrafficKey = Globals.Instance.TomTomTrafficKey;
        private static readonly string TTIncidentKey = Globals.Instance.TomTomTrafficIncidentKey;
        private static readonly string TTVersion = Properties.Settings.Default.TomTomTrafficVersion;
        private static readonly string TTBaseURL = Properties.Settings.Default.TomTomBaseURL;
        private static readonly string TTScheme = Properties.Settings.Default.TomTomScheme;
        private string TTKey = string.Empty;

        public TomTomV4TrafficProxy(IProxyOperation proxyOperation) : base(proxyOperation)
        {
        }

        public override async Task<HttpResponseMessage> TrafficFlow(
            TrafficType type,
            string style,
            string zoom,
            string x,
            string y,
            string formatstring,
            Dictionary<string, List<string>> querystrings)
        {
            this.BuildURL(ref this.URI, type, style, zoom, x, y, formatstring, querystrings);
            return await this.ExecuteRequest(CreateRequest(type, this.URI.Uri));
        }

        public override async Task<HttpResponseMessage> IncidentTile(
            TrafficType type,
            string style,
            string zoom,
            string x,
            string y,
            string formatstring,
            Dictionary<string, List<string>> querystrings)
        {
            this.BuildURL(ref this.URI, type, style, zoom, x, y, formatstring, querystrings);
            return await this.ExecuteRequest(CreateRequest(type, this.URI.Uri));
        }

        public override async Task<HttpResponseMessage> IncidentViewport(
            TrafficType type,
            string boundingbox,
            string boundingzoom,
            string overviewbox,
            string overviewzoom,
            string copyright,
            string formatstring,
            Dictionary<string, List<string>> querystrings)
        {
            this.BuildURL(ref this.URI, type, boundingbox, boundingzoom, overviewbox, overviewzoom, copyright, formatstring, querystrings);
            return await this.ExecuteRequest(CreateRequest(type, this.URI.Uri));
        }

        public override async Task<HttpResponseMessage> IncidentDetails(
            TrafficType type,
            string style,
            string boundingbox,
            string zoom,
            string trafficmodelid,
            string formatstring,
            Dictionary<string, List<string>> querystrings)
        {
            this.BuildURL(ref this.URI, type, style, boundingbox, zoom, trafficmodelid, formatstring, querystrings);
            return await this.ExecuteRequest(CreateRequest(type, this.URI.Uri));
        }

        private static HttpRequestMessage CreateRequest(TrafficType trafficType, Uri requestUri)
        {
            switch (trafficType)
            {
                case TrafficType.FlowTile:
                case TrafficType.FlowSegment:
                case TrafficType.IncidentTile:
                case TrafficType.IncidentDetails:
                case TrafficType.IncidentViewport:
                    return new HttpRequestMessage(HttpMethod.Get, requestUri);

                default:
                    throw new ProxyException(ProxyException.ExceptionType.ServerError, $"Traffic type {trafficType} is not supported");
            }
        }

        private void BuildURL(
            ref UriBuilder URI,
            TrafficType type,
            string style,
            string zoom,
            string x,
            string y,
            string formatstring,
            Dictionary<string, List<string>> querystrings)
        {
            this.BuildURL(ref URI, type, style, zoom, x, y, string.Empty, formatstring, querystrings);
        }

        private void BuildURL(
            ref UriBuilder URI,
            TrafficType type,
            string boundingbox,
            string boundingzoom,
            string overviewbox,
            string overviewzoom,
            string copyright,
            string formatstring,
            Dictionary<string, List<string>> querystrings)
        {
            if (type == TrafficType.FlowSegment || type == TrafficType.FlowTile)
            {
                this.TTKey = TTTrafficKey;
            }
            else
            {
                this.TTKey = TTIncidentKey;
            }

            URI.Scheme = TTScheme;
            URI.Host = TTBaseURL;

            this.BuildURLPath(ref URI, type, boundingbox, boundingzoom, overviewbox, overviewzoom, copyright, formatstring);

            if (type == TrafficType.FlowSegment)
            {
                List<string> tempList = new List<string>();
                tempList.Add(overviewbox);
                querystrings.Add($"point", tempList);
            }

            this.BuildURLQuery(ref URI, querystrings);
        }

        private void BuildURLPath(ref UriBuilder URI, TrafficType trafficType, string first, string second, string third, string forth, string fifth, string formatstring)
        {
            switch (trafficType)
            {
                case (TrafficType.FlowTile):
                    {
                        if (first == string.Empty || second == string.Empty || third == string.Empty || forth == string.Empty)
                        {
                            break;
                        }

                        URI.Path = $"traffic/map/" + TTVersion + $"/tile/flow";
                        URI.Path += $"/" + first + $"/" + second + $"/" + third + $"/" + forth + $"." + formatstring;
                        break;
                    }

                case (TrafficType.FlowSegment):
                    {
                        if (first == string.Empty || second == string.Empty)
                        {
                            break;
                        }

                        URI.Path = $"traffic/services/" + TTVersion + $"/flowSegmentData";
                        URI.Path += $"/" + first + $"/" + second + $"/" + formatstring;
                        break;
                    }

                case (TrafficType.IncidentTile):
                    {
                        if (first == string.Empty || second == string.Empty || third == string.Empty || forth == string.Empty)
                        {
                            break;
                        }

                        URI.Path = $"traffic/map/" + TTVersion + $"/tile/incidents";
                        URI.Path += $"/" + first + $"/" + second + $"/" + third + $"/" + forth + $"." + formatstring;
                        break;
                    }

                case (TrafficType.IncidentDetails):
                    {
                        if (first == string.Empty || second == string.Empty || third == string.Empty || forth == string.Empty)
                        {
                            break;
                        }

                        URI.Path = $"traffic/services/" + TTVersion + $"/incidentDetails";
                        URI.Path += $"/" + first + $"/" + second + $"/" + third + $"/" + forth + $"/" + formatstring;
                        break;
                    }

                case (TrafficType.IncidentViewport):
                    {
                        if (first == string.Empty || second == string.Empty || third == string.Empty || forth == string.Empty || fifth == string.Empty)
                        {
                            break;
                        }

                        URI.Path = $"traffic/services/" + TTVersion + $"/incidentViewport";
                        URI.Path += $"/" + first + $"/" + second + $"/" + third + $"/" + forth + $"/" + fifth + $"/" + formatstring;
                        break;
                    }

                default:
                    {
                        break;
                    }
            }
        }

        private void BuildURLQuery(ref UriBuilder URI, Dictionary<string, List<string>> querystrings)
        {
            string queryparam = $"key=" + this.TTKey;

            for (int index = 0; index < querystrings.Count; index++)
            {
                var item = querystrings.ElementAt(index);
                string itemKey = item.Key;

                foreach (var itemValue in item.Value)
                {
                    if (itemValue == string.Empty
                        || itemValue == "{" + itemKey + "}")
                    {
                        continue;
                    }

                    queryparam += $"&" + itemKey + "=" + itemValue;
                }
            }

            URI.Query = queryparam;
        }
    }
}