﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Models
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Net.Http.Formatting;
    using System.Security.Cryptography;
    using System.Text;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.Utility;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;

    public class MicrosoftV1UploadProxy : BaseProxyUpload
    {
        private const string ContainerNamePrefix = "userdatacontainer-";
        private const string AccountIdMetadataKey = "AccountId";

        public MicrosoftV1UploadProxy(IProxyOperation proxyOperation)
            : base(proxyOperation)
        {
        }

        public override async Task<HttpResponseMessage> UploadData(string accountId, Stream data, UploadDataFormat format, UploadType uploadType)
        {
            try
            {
                IBlobService blobService = Globals.Instance.BlobService;

                // Set up the container for the user/tenant. The containerName below can also be thought of as a "TenantId".
                string containerName = GenerateContainerName(accountId);

                if (!await blobService.DoesContainerExistAsync(containerName))
                {
                    // Associate the AccountId with the container for reverse lookup.
                    CloudBlobContainer container = blobService.GetContainerReference(containerName);
                    container.Metadata.Add(AccountIdMetadataKey, accountId);
                    await blobService.CreateContainerIfNotExistsAsync(container);
                }

                // Create an empty blob where blob name is a unique id. The blobName below can also be thought of as a "UDID" - unique data/document id.
                string blobName = GenerateBlobName(format);
                CloudBlockBlob blobReference = blobService.GetBlockBlobReference(containerName, blobName);

                // Upload the stream data.
                await blobService.UploadBlockBlobFromStreamAsync(blobReference, data);

                return new HttpResponseMessage(HttpStatusCode.OK)
                {
                    Content = new ObjectContent<UserData>(new UserData { Udid = GetBlobName(blobReference) }, new JsonMediaTypeFormatter(), Constants.JsonMediaType)
                };
            }
            catch (Exception ex) when (ex is StorageException || ex is ArgumentException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "Uploading user data failed!", ex);
            }
        }

        public override async Task<HttpResponseMessage> UpdateData(string accountId, string udid, Stream data, UploadType uploadType)
        {
            try
            {
                IBlobService blobService = Globals.Instance.BlobService;

                // Determine the container for the user/tenant.
                string containerName = GenerateContainerName(accountId);

                if (!await blobService.DoesContainerExistAsync(containerName))
                {
                    // Make sure the tenant container exists - else the file 'udid' is invalid.
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Bad request - Tenant container not found.");
                }

                // Get the blob name using the unique data/document id (udid).
                string blobName = GenerateBlobName(GetDataFormatFromBlobName(udid), udid);

                if (!await blobService.DoesBlockBlobExistAsync(containerName, blobName))
                {
                    // Make sure the blob that needs to be updated exists - else the file 'udid' is invalid.
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"Bad request - Data with id:{udid} not found.");
                }

                // Overwrite the exisitng blob with new stream data.
                CloudBlockBlob blobReference = blobService.GetBlockBlobReference(containerName, blobName);
                await blobService.UploadBlockBlobFromStreamAsync(blobReference, data);

                return new HttpResponseMessage(HttpStatusCode.OK)
                {
                    Content = new StringContent($"{{\"udid\":\"{GetBlobName(blobReference)}\"}}", Encoding.UTF8, Constants.JsonMediaType)
                };
            }
            catch (Exception ex) when (ex is StorageException || ex is ArgumentException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "Updating user data failed!", ex);
            }
        }

        public override async Task<HttpResponseMessage> DeleteData(string accountId, string udid, UploadType uploadType)
        {
            try
            {
                IBlobService blobService = Globals.Instance.BlobService;
                
                // Determine the container for the user/tenant.
                string containerName = GenerateContainerName(accountId);

                if (!await blobService.DoesContainerExistAsync(containerName))
                {
                    // Make sure the tenant container exists - else the file 'udid' is invalid.
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Bad request - Tenant container not found.");
                }

                // Get the blob name using the unique data/document id (udid).
                string blobName = GenerateBlobName(GetDataFormatFromBlobName(udid), udid);

                if (!await blobService.DoesBlockBlobExistAsync(containerName, blobName))
                {
                    // Make sure the blob that needs to be updated exists - else the file 'udid' is invalid.
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, $"Bad request - Data with id:{udid} not found.");
                }

                // Delete the blob.
                CloudBlockBlob blobReference = blobService.GetBlockBlobReference(containerName, blobName);
                await blobService.DeleteBlobAsync(blobReference);

                return new HttpResponseMessage(HttpStatusCode.OK)
                {
                    Content = new StringContent($"{{\"udid\":\"{GetBlobName(blobReference)}\"}}", Encoding.UTF8, Constants.JsonMediaType)
                };
            }
            catch (Exception ex) when (ex is StorageException || ex is ArgumentException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "Deleting user data failed!", ex);
            }
        }

        public override async Task<HttpResponseMessage> ListData(string accountId, UploadType uploadType)
        {
            try
            {
                IBlobService blobService = Globals.Instance.BlobService;
                var userDataListResponse = new UserDataListResponse { Files = new List<UserData>() };

                // Determine the container for the user/tenant.
                string containerName = GenerateContainerName(accountId);

                if (!await blobService.DoesContainerExistAsync(containerName))
                {
                    // Make sure the tenant container exists - else this is a bad request from an "invalid tenant".
                    throw new ProxyException(ProxyException.ExceptionType.ClientParams, "Bad request - Tenant container not found.");
                }
                
                // Generate a list of all the files that this tenant has uploaded.
                var container = blobService.GetContainerReference(containerName);
                IEnumerable<IListBlobItem> userDataFiles = container.ListBlobs(useFlatBlobListing: true);

                foreach (var userDataFile in userDataFiles)
                {
                    userDataListResponse.Files.Add(new UserData { Udid = userDataFile.Uri.Segments.Last() });
                }

                return new HttpResponseMessage(HttpStatusCode.OK)
                {
                    Content = new ObjectContent<UserDataListResponse>(userDataListResponse, new JsonMediaTypeFormatter(), Constants.JsonMediaType)
                };
            }
            catch (Exception ex) when (ex is StorageException)
            {
                throw new ProxyException(ProxyException.ExceptionType.ServerError, "Listing user data failed!", ex);
            }
        }

        private static string GenerateContainerName(string accountId)
        {
            // Account Id may have some special characters that are not compliant with azure storage container naming guidelines.
            using (MD5 hash = MD5.Create())
            {
                // MD5 hash is fine here for this specific purpose (even though it is not cryptographically secure)
                return string.Concat(ContainerNamePrefix, string.Join("", hash.ComputeHash(Encoding.UTF8.GetBytes(accountId)).Select(ba => ba.ToString("x2")))).ToLowerInvariant();
            }
        }

        private static string GenerateBlobName(UploadDataFormat format, string uniqueId = null)
        {
            string directoryName = format.ToString().ToLowerInvariant();
            uniqueId = uniqueId?.Split('.').First() ?? Guid.NewGuid().ToString();

            switch (format)
            {
                case UploadDataFormat.Shapefile:
                    return $"{directoryName}/{uniqueId}.shp";

                case UploadDataFormat.Kml:
                    return $"{directoryName}/{uniqueId}.kml";

                case UploadDataFormat.Geojson:
                    return $"{directoryName}/{uniqueId}.geojson";

                default: throw new ArgumentException($"Storage for format type: {format.ToString()} not supported.");
            }
        }

        private static string GetBlobName(CloudBlockBlob blobReference)
        {
            return blobReference.Uri.Segments.Last();
        }

        private static UploadDataFormat GetDataFormatFromBlobName(string blobName)
        {
            string blobExtension = blobName.Split('.').Last().ToLowerInvariant();

            switch (blobExtension)
            {
                case "shp":
                    return UploadDataFormat.Shapefile;

                case "kml":
                    return UploadDataFormat.Kml;

                case "geojson":
                    return UploadDataFormat.Geojson;

                default: throw new ProxyException(ProxyException.ExceptionType.ClientParams, "udid is not valid", new ArgumentException($"Data format for blob:{blobExtension} not found."));
            }
        }
    }
}