﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Net;
    using System.Net.Http;
    using System.Web.Http.Controllers;
    using System.Web.Http.Filters;
    using Microsoft.Azure.LocationServices.Controllers;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.Utility;

    internal class ServiceRequestActionFilter : ActionFilterAttribute
    {
        public override void OnActionExecuting(HttpActionContext actionContext)
        {
            Controllers.BaseProxyController proxyController = null;
            Exception caughtException = null;

            try
            {
                proxyController = actionContext.ControllerContext.Controller as Controllers.BaseProxyController;
                Validation.Requires.NotNull(proxyController, "proxyController");

                string requestName = $"{actionContext.ActionDescriptor.ControllerDescriptor.ControllerName}.{actionContext.ActionDescriptor.ActionName}";
                proxyController.SetProxyOperation(new ProxyOperation(requestName));
                proxyController.Initialize();
            }
            catch (ProxyException pEx)
            {
                caughtException = pEx;
                Agent.Instance.LogException(caughtException, "proxy", "badRequest");

                const HttpStatusCode ErrorCode = HttpStatusCode.BadRequest;

                var errorContract = new ErrorContract
                {
                    Code = $"{(int)ErrorCode} {ErrorCode}",
                    Message = Constants.ErrorMessageBadRequest
                };
                
                actionContext.Response = actionContext.Request.CreateResponse(ErrorCode, new ErrorContractResponse { Error = errorContract });
            }
            catch (Exception ex)
            {
                caughtException = ex;
                Agent.Instance.LogException(caughtException, "proxy", "serverError");

                const HttpStatusCode ErrorCode = HttpStatusCode.InternalServerError;

                var errorContract = new ErrorContract
                {
                    Code = $"{(int)ErrorCode} {ErrorCode}",
                    Message = Constants.ErrorMessageServer
                };

                actionContext.Response = actionContext.Request.CreateResponse(ErrorCode, new ErrorContractResponse { Error = errorContract });
            }
            finally
            {
                // If we catch exception and construct a response, OnActionExecuted will not run.
                // We still want to log information (like we do in OnActionExecuted), rather than exit prematurely.
                if (caughtException != null && proxyController != null) // Technically proxyController nullity isn't necessary
                {
                    // Set exception information
                    var errorCode = actionContext.Response.StatusCode;
                    if (errorCode == HttpStatusCode.BadRequest)
                    {
                        proxyController.ProxyOperation.SetResultClientError(proxyController.TrackingId, caughtException.ToString());
                    }
                    else if (errorCode == HttpStatusCode.InternalServerError)
                    {
                        proxyController.ProxyOperation.SetResultFailure(proxyController.TrackingId, caughtException.ToString());
                    }

                    proxyController.DeInitialize(actionContext.Response); // Set info such as time(Response - Request)
                    proxyController.ProxyOperation.Log_ProxyOperationMetrics(); // Log to hot path
                    proxyController.ProxyOperation.Dispose(); // Log to warm path
                    proxyController.SetProxyOperation(null); // Cleanup
                }
            }
        }

        public override void OnActionExecuted(HttpActionExecutedContext actionExecutedContext)
        {
            BaseProxyController proxyController = null; 
            try
            {
                proxyController = actionExecutedContext.ActionContext.ControllerContext.Controller as Controllers.BaseProxyController;
                Validation.Requires.NotNull(proxyController, "proxyController");
                Validation.Requires.NotNull(proxyController.ProxyOperation, "proxyController.ProxyOperation");

                proxyController.DeInitialize();

                if (actionExecutedContext.Exception != null)
                {             
                    actionExecutedContext.Response = actionExecutedContext.Request.CreateResponse(HttpStatusCode.InternalServerError);
                    proxyController.ProxyOperation.SetResultFailure(proxyController.TrackingId, actionExecutedContext.Exception.ToString());
                    Agent.Instance.LogException(actionExecutedContext.Exception, "proxy", "serverError");
                }

                proxyController.ProxyOperation.Log_ProxyOperationMetrics();
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "proxy", "serverError");
            }
            finally
            {
                if (proxyController != null)
                {
                    proxyController.ProxyOperation.Dispose();
                    proxyController.SetProxyOperation(null);
                }
            }
        }
    }
}