﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Net;
    using System.Net.Http;
    using System.Security.Claims;
    using System.Web.Http;
    using System.Web.Http.Routing;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Utility;
    using Microsoft.IdentityModel.Tokens;
    using Microsoft.Owin.Security.ActiveDirectory;
    using global::Owin;

    public static class Startup
    {
        // This code configures Web API. The Startup class is specified as a type
        // parameter in the WebApp.Start method.
        public static void ConfigureApp(IAppBuilder appBuilder)
        {
            ProxyHostSettings settings = ProxyHostSettings.CreateFromEnvironment();
            if (settings.EnableAADAuthentication)
            {
                appBuilder.UseWindowsAzureActiveDirectoryBearerAuthentication(new WindowsAzureActiveDirectoryBearerAuthenticationOptions
                {
                    Tenant = settings.AADTenant,
                    TokenValidationParameters = new TokenValidationParameters
                    {
                        ValidAudience = settings.AADAudience,
                        NameClaimType = ClaimTypes.Name,
                        AuthenticationType = "AADJwt",
                        ValidateLifetime = true,
                        ValidateAudience = true,
                        ValidateIssuer = true,
                        RequireExpirationTime = true
                    }
                });
            }
            
            // Configure Web API for self-host. 
            HttpConfiguration config = new HttpConfiguration();

            //Heartbeat API
            config.Routes.MapHttpRoute(
                       "Heartbeat",
                       "heartbeat",
                       new { controller = "Heartbeat", action = "GetHeartbeat" });

            //Search APIs
            config.Routes.MapHttpRoute(
                       "SearchPolygon",
                       "search/polygon/{formatstring}",
                       new { controller = "SearchPolygon", action = "GetSearchPolygon" });

            config.Routes.MapHttpRoute(
                       "SearchFuzzy",
                       "search/fuzzy/{formatstring}",
                       new { controller = "SearchFuzzy", action = "GetSearchFuzzy" });

            config.Routes.MapHttpRoute(
                       "SearchPOI",
                       "search/poi/{formatstring}",
                       new { controller = "SearchPOI", action = "GetSearchPOI" });

            config.Routes.MapHttpRoute(
                       "SearchNearby",
                       "search/nearby/{formatstring}",
                       new { controller = "SearchNearby", action = "GetSearchNearby" });

            config.Routes.MapHttpRoute(
                       "SearchPOICategory",
                       "search/poi/category/{formatstring}",
                       new { controller = "SearchPOICategory", action = "GetSearchPOICategory" });

            config.Routes.MapHttpRoute(
                       "SearchAddress",
                       "search/address/{formatstring}",
                       new { controller = "SearchAddress", action = "GetAddress" });            

            config.Routes.MapHttpRoute(
                       "SearchAddressStructured",
                       "search/address/structured/{formatstring}",
                       new { controller = "SearchAddress", action = "GetAddressStructured" });

            config.Routes.MapHttpRoute(
                       "SearchAddressReverse",
                       "search/address/reverse/{formatstring}",
                       new { controller = "SearchReverseAddress", action = "GetAddressReverse" });

            config.Routes.MapHttpRoute(
                       "SearchAddressCrossStreet",
                       "search/address/reverse/crossStreet/{formatstring}",
                       new { controller = "SearchReverseAddress", action = "GetAddressCrossStreet" });

            config.Routes.MapHttpRoute(
                       "SearchByGeometry",
                       "search/geometry/{formatstring}",
                       new { controller = "SearchByGeometry" });

            config.Routes.MapHttpRoute(
                       "SearchAlongRoute",
                       "search/alongRoute/{formatstring}",
                       new { controller = "SearchAlongRoute" });

            //Render APIs
            config.Routes.MapHttpRoute(
                       "MapTile",
                       "map/tile/{formatstring}",
                       new { controller = "MapTile", action = "GetMapTile" });

            config.Routes.MapHttpRoute(
                       "MapStatic",
                       "map/static/{formatstring}",
                       new { controller = "MapTile", action = "GetMapTileStatic" });

            config.Routes.MapHttpRoute(
                       "MapImagery",
                       "map/imagery/{formatstring}",
                       new { controller = "MapImagery", action = "GetMapImagery" });

            config.Routes.MapHttpRoute(
                       "MapCopyrightWorld",
                       "map/copyright/world/{formatstring}",
                       new { controller = "MapCopyright", action = "GetCopyrightWorld" });

            config.Routes.MapHttpRoute(
                       "MapCopyrightCaption",
                       "map/copyright/caption/{formatstring}",
                       new { controller = "MapCopyright", action = "GetCopyrightCaption" });

            config.Routes.MapHttpRoute(
                       "MapCopyrightBounding",
                       "map/copyright/bounding/{formatstring}",
                       new { controller = "MapCopyright", action = "GetCopyrightBounding" });

            config.Routes.MapHttpRoute(
                       "MapCopyrightTile",
                       "map/copyright/tile/{formatstring}",
                       new { controller = "MapCopyright", action = "GetCopyrightTile" });

            //Timezone APIs
            config.Routes.MapHttpRoute(
                       "TimezoneIANAVersion",
                       "timezone/ianaVersion/{formatstring}",
                       new { controller = "TimezoneIANAVersion", action = "GetTimezoneIANAVersion" });

            config.Routes.MapHttpRoute(
                       "TimezoneEnumIANA",
                       "timezone/enumIana/{formatstring}",
                       new { controller = "TimezoneEnumIANA", action = "GetTimezoneIANA" });

            config.Routes.MapHttpRoute(
                       "TimezoneEnumWindows",
                       "timezone/enumWindows/{formatstring}",
                       new { controller = "TimezoneEnumWindows", action = "GetTimezoneWindows" });

            config.Routes.MapHttpRoute(
                       "TimezoneWindowsToIANA",
                       "timezone/windowsToIana/{formatstring}",
                       new { controller = "TimezoneWindowsToIANA", action = "GetTimezoneWindowsToIANA" });

            config.Routes.MapHttpRoute(
                       "TimezoneById",
                       "timezone/byId/{formatstring}",
                       new { controller = "TimezoneById", action = "GetTimezoneByID" });

            config.Routes.MapHttpRoute(
                       "TimezoneByCoordinates",
                       "timezone/byCoordinates/{formatstring}",
                       new { controller = "TimezoneByCoordinates", action = "GetTimezoneCoordinates" });

            //Route APIs
            config.Routes.MapHttpRoute(
                       "RouteDirections",
                       "route/directions/{formatstring}",
                       new { controller = "RouteCalculate" });

            config.Routes.MapHttpRoute(
                       "RouteReachableRange",
                       "route/reachablerange/{formatstring}",
                       new { controller = "RouteReachableRange" });

            config.Routes.MapHttpRoute(
                       "RouteMatrix",
                       "route/matrix/{formatstring}",
                       new { controller = "RouteMatrix" });

            //Traffic APIs
            config.Routes.MapHttpRoute(
                       "TrafficFlowTile",
                       "traffic/flow/tile/{formatstring}",
                       new { controller = "TrafficFlowTile", action = "GetTrafficFlowTile" });

            config.Routes.MapHttpRoute(
                       "TrafficFlowSegment",
                       "traffic/flow/segment/{formatstring}",
                       new { controller = "TrafficFlowSegment", action = "GetTrafficFlowSegment" });

            config.Routes.MapHttpRoute(
                       "TrafficIncidentTile",
                       "traffic/incident/tile/{formatstring}",
                       new { controller = "TrafficIncident", action = "GetTrafficIncidentTile" });

            config.Routes.MapHttpRoute(
                       "TrafficIncidentViewport",
                       "traffic/incident/viewport/{formatstring}",
                       new { controller = "TrafficIncident", action = "GetTrafficIncidentViewport" });

            config.Routes.MapHttpRoute(
                       "TrafficIncidentDetails",
                       "traffic/incident/detail/{formatstring}",
                       new { controller = "TrafficIncident", action = "GetTrafficIncidentDetails" });

            // Geo location APIs
            config.Routes.MapHttpRoute(
                       "GeoLocationIP",
                       "geolocation/ip/{formatstring}",
                       new { controller = "GeoLocationIP", action = "GetIpLocation" });

            // User Data Service APIs
            config.Routes.MapHttpRoute(
                       "UpdateOrDeleteUserData",
                       "data/{udid}",
                       new { controller = "UploadService" });

            config.Routes.MapHttpRoute(
                       "UploadUserData",
                       "data",
                       new { controller = "UploadService",  action = "PostData" }, new { httpMethod = new HttpMethodConstraint(HttpMethod.Post) });

            config.Routes.MapHttpRoute(
                       "ListUserData",
                       "data",
                       new { controller = "UploadService", action = "ListData" }, new { httpMethod = new HttpMethodConstraint(HttpMethod.Get) });

            // Batch APIs
            config.Routes.MapHttpRoute(
                       "BatchGetResultsById",
                       "batch/{batchId}",
                       new { controller = "Batch" });

            config.Routes.MapHttpRoute(
                       "SearchAddressBatch",
                       "search/address/batch/{formatstring}",
                       new { controller = "Batch" });

            config.Routes.MapHttpRoute(
                       "SearchAddressReverseBatch",
                       "search/address/reverse/batch/{formatstring}",
                       new { controller = "Batch" });

            config.Routes.MapHttpRoute(
                       "SearchFuzzyBatch",
                       "search/fuzzy/batch/{formatstring}",
                       new { controller = "Batch" });
            
            config.Routes.MapHttpRoute(
                       "RouteDirectionsBatch",
                       "route/directions/batch/{formatstring}",
                       new { controller = "Batch" });

            // Geofence APIs
            config.Routes.MapHttpRoute(
                       "Geofence",
                       "geofence/{formatstring}",
                       new { controller = "Geofence", action = "Geofence" }, new { httpMethod = new HttpMethodConstraint(HttpMethod.Get) });

            config.Filters.Add(new ServiceRequestActionFilter());            
            config.Filters.Add(new ClaimsPrincipalAuthorizationFilter());

            // We're moving away from certificate authentication to AAD. We're keeping certificate authentication as a fallback in Prod until we
            // gain more confidence in APIM->Proxy AAD authentication.
            if (string.Equals(LogicalEnvironments.Production, Environment.GetEnvironmentVariable("Environment"), StringComparison.InvariantCultureIgnoreCase))
            {
                config.Filters.Add(new CertificateAuthenticationFilter());
            }

            config.MapHttpAttributeRoutes();
            appBuilder.UseWebApi(config);

            ServicePointManager.ServerCertificateValidationCallback += (sender, certificate, chain, errors) =>
            {
                return (errors == System.Net.Security.SslPolicyErrors.None);
            };

            // The number of milliseconds after which a Domain Name Service (DNS) resolution is considered invalid.
            ServicePointManager.DnsRefreshTimeout = (int)TimeSpan.FromMinutes(1).TotalMilliseconds;

            // The number of milliseconds after which an active TomTom connection is closed. This is to prevent caching invalid/out-of-date IPs in case of a failover.
            Uri EndPoint = new UriBuilder(Properties.Settings.Default.TomTomScheme, Properties.Settings.Default.TomTomBaseURL).Uri;
            ServicePointManager.FindServicePoint(EndPoint).ConnectionLeaseTimeout = (int)TimeSpan.FromMinutes(1).TotalMilliseconds;
        }
    }
}