﻿// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Utility
{
    using System;
    using System.Linq;
    using GeoJSON.Net;
    using GeoJSON.Net.Converters;
    using GeoJSON.Net.Feature;
    using GeoJSON.Net.Geometry;
    using Newtonsoft.Json;
    using Newtonsoft.Json.Linq;

    /// <summary>
    /// An extension class that provides a bunch of utility methods for <see cref="IGeoJSONObject"/> objects.
    /// </summary>
    internal static class GeoJsonExtensions
    {
        private enum FeatureSubType
        {
            Unknown = 0,
            Circle
        }

        /// <summary>
        /// Tries to deserialize a json object into a valid <see cref="IGeoJSONObject"/> type.
        /// </summary>
        /// <param name="jsonGeometry">A valid json object representing a geometry</param>
        /// <returns>A valid <see cref="IGeoJSONObject" /> object</returns>
        /// <exception cref="JsonReaderException">Thrown when input in not a valid json.</exception>
        /// <exception cref="JsonSerializationException">Thrown when the input json cannot be deserialized into a valid GeoJSON type.</exception>
        public static IGeoJSONObject Deserialize(this JObject jsonGeometry)
        {
            return JsonConvert.DeserializeObject<IGeoJSONObject>(jsonGeometry.ToString(), converters: new GeoJsonConverter());
        }

        /// <summary>
        /// Validates the input <see cref="IGeoJSONObject" /> 
        /// </summary>
        /// <param name="geometry">Input GeoJSON geometry.</param>
        /// <returns>Returns True if input is valid. False otherwise.</returns>
        public static bool IsValid(this IGeoJSONObject geometry)
        {
            switch (geometry.Type)
            {
                case GeoJSONObjectType.FeatureCollection:
                    return IsValidFeatureCollection((FeatureCollection)geometry);

                case GeoJSONObjectType.GeometryCollection:
                    return IsValidGeometryCollection((GeometryCollection)geometry);

                case GeoJSONObjectType.Polygon:
                    return IsValidPolygon((Polygon)geometry);

                case GeoJSONObjectType.LineString:
                    return IsValidLineString((LineString)geometry);

                default:
                    return false;
            }
        }

        /// <summary>
        /// Validates the input <see cref="FeatureCollection" />.
        /// Return true if input is valid. False otherwise.
        /// </summary>
        /// <param name="featureCollection">Input featureCollection object.</param>
        private static bool IsValidFeatureCollection(FeatureCollection featureCollection)
        {
            int count = featureCollection.Features.Count;

            // The feature collection can only contain a max of 50 feature objects/geometries.
            if (count < 1 || count > 50)
            {
                return false;
            }

            // Each Polygon feature object/geometry can contain a max of 50 vertices.
            foreach (var feature in featureCollection.Features)
            {
                switch (feature.Geometry.Type)
                {
                    case GeoJSONObjectType.Polygon:
                        {
                            if (!IsValidPolygon((Polygon)feature.Geometry))
                            {
                                return false;
                            }

                            break;
                        }

                    case GeoJSONObjectType.Point:
                        {
                            // Circle geometry is implemented using a Point Geometry, so check if this a valid circle geometry.
                            if (!feature.Properties.TryGetValue(Constants.GeoJsonSubTypeKey, out object subTypeObject) ||
                                subTypeObject == null || 
                                !subTypeObject.ToString().Equals(FeatureSubType.Circle.ToString(), StringComparison.OrdinalIgnoreCase) || 
                                !IsValidCircle(feature))
                            {
                                // Either subType wasn't specified or subType value is null/incorrect or circle geometry is invalid.
                                return false;
                            }
                            
                            break;
                        }

                    default:
                        {
                            // Type not supported yet.
                            return false;
                        }
                }
            }

            return true;
        }

        /// <summary>
        /// Validates the input <see cref="GeometryCollection" />.
        /// Return true if input is valid. False otherwise.
        /// </summary>
        /// <param name="geometryCollection">Input geometryCollection object.</param>
        private static bool IsValidGeometryCollection(GeometryCollection geometryCollection)
        {
            int count = geometryCollection.Geometries.Count;

            // The geometry collection can only contain a max of 50 geometries.
            if (count < 1 || count > 50)
            {
                return false;
            }

            // Each Polygon geometry can contain a max of 50 vertices.
            return
                geometryCollection.Geometries
                .Where(geometry => geometry.Type.Equals(GeoJSONObjectType.Polygon))
                .All(geometry => IsValidPolygon((Polygon)geometry));
        }

        /// <summary>
        /// Validates the input <see cref="Polygon" />.
        /// Return true if input is valid. False otherwise.
        /// </summary>
        /// <param name="polygon">Input polygon object.</param>
        private static bool IsValidPolygon(Polygon polygon)
        {
            // Polygon is valid if it doesn't have more than 50 vertices.
            return polygon.Coordinates.SelectMany(lineString => lineString.Coordinates).Count() <= 50;
        }

        /// <summary>
        /// Validates the input <see cref="LineString" />.
        /// Return true if input is valid. False otherwise.
        /// </summary>
        /// <param name="lineString">Input LineString object.</param>
        private static bool IsValidLineString(LineString lineString)
        {
            return lineString != null && lineString.Coordinates.Count >= 2;
        }

        /// <summary>
        /// Returns true if the input <see cref="Feature"/> is a valid Circle geometry. False otherwise.
        /// </summary>
        /// <param name="feature">Input feature object.</param>
        private static bool IsValidCircle(Feature feature)
        {
            // Circle is valid if the radius value is present and it's a valid double.
            return feature.Properties.TryGetValue(Constants.GeoJsonCircleRadiusKey, out object radiusObject) 
                && radiusObject != null 
                && double.TryParse(radiusObject.ToString(), out double radiusValue);
        }
    }
}