﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Utility
{
    using System;
    using System.Globalization;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;

    /// <summary>
    /// Represents the configuration necessary to operate in an environment.
    /// </summary>
    /// <remarks>
    /// This class should only contain members that differ from environment to environment. When something doesn't 
    /// vary, it should not be present in this class, and instead should be a compile-time constant. If a property
    /// is no longer used, or no longer varies, it should be removed.
    /// </remarks>
    public abstract class ProxyHostSettings
    {
        /// <summary>
        /// AAD Tenant settings for AAD Authentication
        /// </summary>
        public abstract string AADTenant { get; }

        /// <summary>
        /// AAD Audience for the authorization of the application the token should have scope for.
        /// </summary>
        public abstract string AADAudience { get; }

        /// <summary>
        /// Gets the APIM host name
        /// </summary>
        public abstract string HostName { get; }

        /// <summary>
        /// AAD Flag to enable authentication for AAD Jwt Tokens
        /// </summary>
        public abstract bool EnableAADAuthentication { get; }

        /// <summary>
        /// Gets the Instrumentation Parameters
        /// </summary>
        public abstract InitParams MdmParameters { get; }

        /// <summary>
        /// MDP api http schema.
        /// </summary>
        public abstract string MdpApiScheme { get; }

        /// <summary>
        /// MDP api fqdn.
        /// </summary>
        public abstract string MdpApiFqdn { get; }

        /// <summary>
        /// MDP api version qualifier.
        /// </summary>
        public string MdpApiVersion { get; } = "2018-06-01-preview";

        /// <summary>
        /// MDP api calls are enabled.
        /// </summary>
        public abstract bool MdpApiEnabled { get; }

        /// <summary>
        /// Gets the current region.
        /// </summary>
        protected static string CurrentRegion
        {
            get
            {
                string currentRegion = Environment.GetEnvironmentVariable("Region");
                return currentRegion ?? "DEFAULT";
            }
        }

        /// <summary>
        /// Gets the current environment.
        /// </summary>
        /// <exception cref="ArgumentException">
        /// Throws if the current environment is not specified
        /// </exception>
        private static string CurrentEnvironment
        {
            get
            {
                string currentEnvironment = Environment.GetEnvironmentVariable("Environment");

                if (string.IsNullOrEmpty(currentEnvironment))
                {
                    throw new ArgumentException("Current Environment could not be determined");
                }

                return currentEnvironment;
            }
        }

        /// <summary>
        /// Create an environment definition from the environment environment variable.
        /// </summary>
        /// <returns>The created environment definition.</returns>
        public static ProxyHostSettings CreateFromEnvironment()
        {
            switch (CurrentEnvironment.ToUpperInvariant())
            {
                case LogicalEnvironments.Development:
                    return new ProxyHostDevSettings();
                case LogicalEnvironments.Test:
                    return new ProxyHostTestSettings();
                case LogicalEnvironments.Staging:
                    return new ProxyHostStagingSettings();
                case LogicalEnvironments.Production:
                    return new ProxyHostProdSettings();
                default:
                    throw new ArgumentException(
                        string.Format(
                            CultureInfo.InvariantCulture,
                            "Environment {0} does not have a corresponding Environment definition defined",
                            CurrentEnvironment));
            }
        }
    }
}