﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ProxyServerIntegrationTests.Common
{
    using System;
    using System.Collections.Generic;
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.ProxyServerIntegrationTests.Configuration;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using Newtonsoft.Json.Linq;

    [TestClass]
    public class ApimTestBase
    {
        private const string AadConnectionStringKeyName = "AzureServicesAuthConnectionString";
        private static readonly HashSet<string> SupportedVersions = new HashSet<string> { Constants.ApiVersion10, };
        private static bool _initialized = false;

        public static string SubscriptionKey { get; private set; }

        public static string ApiVersion { get; private set; }

        public static string BaseAddress { get; private set; }

        public MapsClient MapsClient { get; private set; }

        [AssemblyInitialize]
        public static void AssemblyInitialize(TestContext context)
        {
            if (!_initialized)
            {
                var envName = context.GetTestSetting("Env");
                ConfigurationStoreBase testConfig = null;

                // Check if this test run is targeting a CI cluster instead. If yes, the test settings should have a CI cluster name. 
                if (string.Equals(envName, ConfigurationStore.CIClusterEnvironmentName, StringComparison.OrdinalIgnoreCase))
                {
                    string clusterNameSettingKey = "CIClusterName";
                    var ciClusterName = context.GetTestSetting(clusterNameSettingKey);
                    if (string.IsNullOrWhiteSpace(ciClusterName))
                    {
                        throw new ArgumentException(
                            $"For tests targeting environment {envName}, a value for {clusterNameSettingKey} should be specified.");
                    }

                    // Update the configurations to refer to the cluster specific values.
                    testConfig = ConfigurationStore.GetCIClusterEnvironmentConfig(envName, ciClusterName);
                }
                else
                {
                    testConfig = ConfigurationStore.GetEnvironmentConfig(envName);
                }

                ApiVersion = testConfig.ApimApiVersion;
                if (!SupportedVersions.Contains(ApiVersion))
                {
                    throw new NotSupportedException($"Api version {ApiVersion} is not supported by proxy service integration tests");
                }

                var subscriptionKeyUrl = testConfig.ApimSubscriptionKey;

                var aadConnectionString = context.GetTestSetting("AadConnectionString");

                if (!string.IsNullOrWhiteSpace(aadConnectionString))
                {
                    Environment.SetEnvironmentVariable(
                        AadConnectionStringKeyName,
                        aadConnectionString,
                        EnvironmentVariableTarget.Process);
                }

                var kvReader = new KeyVaultConfigurationReader();
                var kvTask = kvReader.GetConfigurationValueAsync(subscriptionKeyUrl);
                kvTask.Wait();
                SubscriptionKey = kvTask.Result;
                BaseAddress = testConfig.ApimBaseAddress;

                _initialized = true;

                Console.WriteLine($"ApimBaseAddress: {BaseAddress}");
                Console.WriteLine($"ApimApiVersion: {ApiVersion}");
            }
        }

        [TestInitialize]
        public void TestInitialize()
        {
            this.MapsClient = new MapsClient(BaseAddress, SubscriptionKey);
        }

        #region Helper Methods

        protected async Task<HttpResponseMessage> GetAsyncWithHttpResponse(string requestUrl, bool ensureSuccess = true)
        {
            var httpResponse = await this.MapsClient.GetAsync(requestUrl).ConfigureAwait(false);

            Assert.IsNotNull(httpResponse, "Http response cannot be null");
            if (ensureSuccess)
            {
                httpResponse.EnsureSuccessStatusCode();
            }

            return httpResponse;
        }

        protected async Task<JObject> ParseJson(HttpResponseMessage httpResponse, bool ensureSuccess = true)
        {
            var content = await httpResponse.Content.ReadAsStringAsync().ConfigureAwait(false);
            Assert.IsNotNull(content, "Http response content cannot be null");

            return JObject.Parse(content);
        }

        protected async Task EnsureErrorStatusCode(
            HttpResponseMessage response,
            HttpStatusCode expected,
            string errorMessage = null)
        {
            Assert.IsNotNull(response, "Http response cannot be null");
            var content = await response.Content.ReadAsStringAsync().ConfigureAwait(false);

            if (response.IsSuccessStatusCode || response.StatusCode != expected)
            {
                var message = response.RequestMessage.RequestUri + "\n" + content;
                throw new HttpRequestException("HTTP Status Code: " + response.StatusCode + "\r\n" + message);
            }

            if (errorMessage != null)
            {
                Assert.IsNotNull(content, "Http response content cannot be null");
                Assert.IsTrue(content.Contains(errorMessage),
                              "Error message should contain \"" + errorMessage + "\", but instead, the message is:\r\n" + content);
            }
        }

        protected void EnsureTokenExists(JToken json, string tokenName)
        {
            var tokenValue = json.SelectToken(tokenName)?.Value<JToken>();
            Assert.IsNotNull(tokenValue, $"Could not find token: {tokenName} in the response");
        }

        #endregion
    }
}
