﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ProxyServerIntegrationTests.Configuration
{
    using System;
    using System.Collections.Generic;
    using Microsoft.Azure.LocationServices.Common;

    public static class ConfigurationStore
    {
        /// <summary>
        /// This value is used as an environment for running integration tests against the CI clusters.
        /// </summary>
        internal const string CIClusterEnvironmentName = "cicluster";

        private static readonly Dictionary<string, ConfigurationStoreBase> EnvToConfigurations =
            new Dictionary<string, ConfigurationStoreBase>()
            {
                { LogicalEnvironments.Development.ToLowerInvariant(), new DevConfigurationStore() },
                { LogicalEnvironments.Test.ToLowerInvariant(), new TestConfigurationStore() },
                { LogicalEnvironments.Staging.ToLowerInvariant(), new StagingConfigurationStore() },
                { LogicalEnvironments.Production.ToLowerInvariant(), new ProdConfigurationStore() }
            };

        internal static ConfigurationStoreBase GetEnvironmentConfig(string envName)
        {
            envName = envName.ToLowerInvariant();

            if (!EnvToConfigurations.ContainsKey(envName))
            {
                throw new ArgumentException(
                    $"Environment {envName} does not have a corresponding Environment definition defined");
            }

            return EnvToConfigurations[envName];
        }

        internal static ConfigurationStoreBase GetCIClusterEnvironmentConfig(string envName, string clusterName)
        {
            return new CIClusterConfigurationStore(clusterName);
        }
    }

    public abstract class ConfigurationStoreBase
    {
        public string ApimApiVersion => "1.0";

        public abstract string ApimSubscriptionKey { get; }

        public abstract string ApimBaseAddress { get; }
    }

    public class CIClusterConfigurationStore : ConfigurationStoreBase
    {
        private readonly string clusterName;

        internal CIClusterConfigurationStore(string clusterName)
        {
            this.clusterName = clusterName;
        }

        public override string ApimBaseAddress => string.Format("https://{0}-atlas.azure-api.net", clusterName);

        public override string ApimSubscriptionKey => string.Format("https://{0}-persist.vault.azure.net/secrets/ApimSubscriptionKey", clusterName);
    }

    public class DevConfigurationStore : ConfigurationStoreBase
    {
        public override string ApimBaseAddress => "https://c-atlas.azure-api.net";

        public override string ApimSubscriptionKey => "https://c-genevarunners.vault.azure.net/secrets/ApimSubscriptionKey";
    }

    public class TestConfigurationStore : ConfigurationStoreBase
    {
        public override string ApimBaseAddress => "https://t-atlas.azure-api.net";

        public override string ApimSubscriptionKey => "https://t-genevarunners.vault.azure.net/secrets/runnerSubscriptionKey";
    }

    public class StagingConfigurationStore : ConfigurationStoreBase
    {
        public override string ApimBaseAddress => "https://s-atlas.azure-api.net";

        public override string ApimSubscriptionKey => "https://s-genevarunners.vault.azure.net/secrets/runnerSubscriptionKey";
    }

    public class ProdConfigurationStore : ConfigurationStoreBase
    {
        public override string ApimBaseAddress => "https://atlas.microsoft.com";

        public override string ApimSubscriptionKey => "https://p-genevarunners.vault.azure.net/secrets/runnerSubscriptionKey";
    }
}
