﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ProxyServerIntegrationTests.Upload
{
    using System.IO;
    using System.Net;
    using System.Net.Http;
    using System.Net.Http.Headers;
    using System.Text;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices.ProxyServerIntegrationTests.Common;
    using Microsoft.VisualStudio.TestTools.UnitTesting;

    using Newtonsoft.Json.Linq;

    /// <summary>
    /// Summary description for UploadFileIntegrationTests
    /// </summary>
    [TestClass]
    public class UploadFileIntegrationTests : ApimTestBase
    {
        [Ignore("API is not finalized and not added to APIM yet")]
        [TestMethod, TestCategory("L3")]
        public async Task UploadFile_InvalidFormat_BadRequest()
        {
            // Arrange
            var url = $"/upload/file/json?api-version={ApiVersion}&subscription-key={SubscriptionKey}";

            var requestBody = new StreamContent(new MemoryStream(Encoding.UTF8.GetBytes("Test Data To Upload")));
            requestBody.Headers.ContentType = new MediaTypeHeaderValue("text/plain");

            // Act
            var httpResponse =
                await this.MapsClient.PostAsync(url, requestBody).ConfigureAwait(false);

            // Assert
            Assert.IsNotNull(httpResponse, "Response should not be null");
            Assert.AreEqual(HttpStatusCode.BadRequest, httpResponse.StatusCode, "Status code does not match");

            string content = await httpResponse.Content.ReadAsStringAsync().ConfigureAwait(false);
            Assert.IsNotNull(content, "Content should not be null");

            JObject json = JObject.Parse(content);

            string errorCode = json.SelectToken("error.code")?.Value<string>();
            Assert.AreEqual(errorCode, "400 BadRequest");

            string errorMessage = json.SelectToken("error.message")?.Value<string>();
            Assert.AreEqual(errorMessage, "Bad request: one or more parameters were incorrectly specified or are mutually exclusive.");
        }

        [Ignore("API is not finalized and not added to APIM yet")]
        [TestMethod, TestCategory("L3")]
        public async Task UploadFile_GeoJson_ValidRequest_OK()
        {
            // Arrange
            var url = $"/upload/file/geojson?api-version={ApiVersion}&subscription-key={SubscriptionKey}";

            var requestBody =
                new StreamContent(new MemoryStream(Encoding.UTF8.GetBytes("{\"type\":\"Polygon\",\"coordinates\":[[[100.0,0.0],[101.0,0.0],[100.0,0.0]]]}")));
            requestBody.Headers.ContentType = new MediaTypeHeaderValue("application/json");

            // Act
            HttpResponseMessage httpResponse = await this.MapsClient.PostAsync(url, requestBody).ConfigureAwait(false);

            // Assert
            Assert.IsNotNull(httpResponse, "Response should not be null");
            Assert.AreEqual(HttpStatusCode.OK, httpResponse.StatusCode, "Status code does not match");

            string content = await httpResponse.Content.ReadAsStringAsync().ConfigureAwait(false);
            Assert.IsNotNull(content, "Content should not be null");

            JObject json = JObject.Parse(content);

            string udid = json.SelectToken("udid")?.Value<string>();
            Assert.IsNotNull(udid);
        }

        [Ignore("API is not finalized and not added to APIM yet")]
        [TestMethod, TestCategory("L3")]
        public async Task UploadFile_Kml_ValidRequest_OK()
        {
            // Arrange
            var url = $"/upload/file/kml?api-version={ApiVersion}&subscription-key={SubscriptionKey}";

            var requestBody =
                new StreamContent(new MemoryStream(
                                      Encoding.UTF8.GetBytes(
                                          "<?xml version=\"1.0\"encoding=\"utf-8\"?><kml xmlns=\"http://www.opengis.net/kml/2.2\"><Placemark><name>My office</name><Point><coordinates>-122.0821604290386,37.42041505714916</coordinates></Point></Placemark></kml>")));
            requestBody.Headers.ContentType = new MediaTypeHeaderValue("application/json");

            // Act
            HttpResponseMessage httpResponse = await this.MapsClient.PostAsync(url, requestBody).ConfigureAwait(false);

            // Assert
            Assert.IsNotNull(httpResponse, "Response should not be null");
            Assert.AreEqual(HttpStatusCode.OK, httpResponse.StatusCode, "Status code does not match");

            string content = await httpResponse.Content.ReadAsStringAsync().ConfigureAwait(false);
            Assert.IsNotNull(content, "Content should not be null");

            JObject json = JObject.Parse(content);

            string udid = json.SelectToken("udid")?.Value<string>();
            Assert.IsNotNull(udid);
        }

        // TODO: Setup blob storage to store test data to be used in integration tests - Task 2566843
        [Ignore("API is not finalized and not added to APIM yet")]
        [TestMethod, TestCategory("L3")]
        public async Task UploadFile_Shapefile_ValidRequest_OK()
        {
            // Arrange
            var url = $"/upload/file/shapefile?api-version={ApiVersion}&subscription-key={SubscriptionKey}";

            // Download test Shapefile data from storage and upload.
        }
    }
}