﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace ProxyServerUnitTests
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http.Routing;
    using Microsoft.Azure.LocationServices;
    using Microsoft.Azure.LocationServices.Controllers;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.Models;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using Moq;

    [TestClass]
    public class Search_Fuzzy_Controller_Tests
    {
        private MockRepository mockRepository;

        private Mock<IVersioning> versionObj;

        private Mock<SearchFuzzyController> searchFuzzyController;

        [TestInitialize]
        public void TestInitialize()
        {
            this.mockRepository = new MockRepository(MockBehavior.Strict) { CallBase = true };
            this.versionObj = this.mockRepository.Create<IVersioning>();
            HttpRequestMessage Request = new HttpRequestMessage(HttpMethod.Get, "https://localhost/search/fuzzy/json?api-version=1.0&query=seattle&subscription-id=GS61xORRtw-KVNYqocNmZp4iZizSFXr2FpebcwyaXDQ&tracking-id=1");
            searchFuzzyController = new Mock<SearchFuzzyController>() { CallBase = true };
            searchFuzzyController.Object.Request = Request;
        }

        [TestMethod, TestCategory("L0")]
        public void Get_Search_Fuzzy_Test_ProxyObjNull()
        {
            //Setup vars
            ProxyException expectedException = new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed");

            //Inject dependency
            searchFuzzyController.Object.versionObject = this.versionObj.Object;

            //Setup dependency
            this.versionObj.Setup(x => x.GetProxyObj(It.IsAny<double>(), BaseProxy.ProxyType.Search, It.IsAny<IProxyOperation>())).Returns((BaseProxySearch)null);
            searchFuzzyController.Setup(x => x.BuildErrorResponseAndLog(expectedException)).Returns((HttpResponseMessage)null);

            //Call and verify
            var response = searchFuzzyController.Object.GetSearchFuzzy("");
            this.versionObj.Verify(x => x.GetProxyObj(It.IsAny<double>(), BaseProxy.ProxyType.Search, null), Times.Once);
            searchFuzzyController.Verify(x => x.BuildErrorResponseAndLog(It.Is<ProxyException>(e => e.Message == expectedException.Message)), Times.Once);
        }

        [TestMethod, TestCategory("L0")]
        public void Get_Search_Fuzzy_Test()
        {
            //Setup vars
            HttpResponseMessage response = new HttpResponseMessage();
            response.Headers.Add("WWW-Authenticate", "Basic");
            TaskCompletionSource<HttpResponseMessage> task = new TaskCompletionSource<HttpResponseMessage>();
            task.SetResult(response);
            IProxyOperation proxyOperation = new ProxyOperation("Search");
            Mock<BaseProxySearch> baseProxySearchMock = new Mock<BaseProxySearch>(proxyOperation);

            //Inject dependency
            searchFuzzyController.Object.versionObject = this.versionObj.Object;

            //Setup dependency
            this.versionObj.Setup(x => x.GetProxyObj(It.IsAny<double>(), BaseProxy.ProxyType.Search, It.IsAny<IProxyOperation>())).Returns(baseProxySearchMock.Object);
            baseProxySearchMock.Setup(x => x.SearchProxy(BaseProxySearch.SearchType.Fuzzy, It.IsAny<string>(), It.IsAny<string>(), It.IsAny<Dictionary<string, List<string>>>(), null)).Returns(task.Task);

            //Call and verify
            var output = searchFuzzyController.Object.GetSearchFuzzy("");
            this.versionObj.Verify(x => x.GetProxyObj(It.IsAny<double>(), BaseProxy.ProxyType.Search, It.IsAny<IProxyOperation>()), Times.Once);
            baseProxySearchMock.Verify(x => x.SearchProxy(BaseProxySearch.SearchType.Fuzzy, It.IsAny<string>(), It.IsAny<string>(), It.IsAny<Dictionary<string, List<string>>>(), null));
            Assert.AreEqual(output.Result, task.Task.Result);
        }

        [TestMethod, TestCategory("L0")]
        public void InitializeTest_GetSearchFuzzy()
        {
            //Setup vars
            IProxyOperation proxyOperation = new ProxyOperation("Search");
            searchFuzzyController.Object.SetProxyOperation(proxyOperation);
            string json = "json";
            IHttpRoute HttpRoute = new HttpRoute();
            HttpRouteValueDictionary routeValueDictionary = new HttpRouteValueDictionary();
            routeValueDictionary.Add("formatstring", json);
            routeValueDictionary.Add("action", "GetSearchFuzzy");
            HttpRouteData routeData = new HttpRouteData(HttpRoute, routeValueDictionary);
            searchFuzzyController.Object.Request.SetRouteData(routeData);

            //Call and verify
            searchFuzzyController.Object.Initialize();
            Assert.AreEqual(json, searchFuzzyController.Object.Format);
            Assert.AreEqual(1, searchFuzzyController.Object.Version);
            Assert.AreEqual("seattle", searchFuzzyController.Object.Query);
        }

        [TestMethod, TestCategory("L0")]
        public void InitializeTest_GetSearchFuzzy_Noquery()
        {
            //Setup vars
            HttpRequestMessage Request = new HttpRequestMessage(HttpMethod.Get, "https://localhost/search/fuzzy/json?api-version=1.0&subscription-id=GS61xORRtw-KVNYqocNmZp4iZizSFXr2FpebcwyaXDQ&tracking-id=1");
            searchFuzzyController.Object.Request = Request;
            IProxyOperation proxyOperation = new ProxyOperation("Search");
            searchFuzzyController.Object.SetProxyOperation(proxyOperation);
            string json = "json";
            IHttpRoute HttpRoute = new HttpRoute();
            HttpRouteValueDictionary routeValueDictionary = new HttpRouteValueDictionary();
            routeValueDictionary.Add("formatstring", json);
            routeValueDictionary.Add("action", "GetSearchFuzzy");
            HttpRouteData routeData = new HttpRouteData(HttpRoute, routeValueDictionary);
            searchFuzzyController.Object.Request.SetRouteData(routeData);
            ProxyException ClientParams = new ProxyException(ProxyException.ExceptionType.ClientParams, "query param is empty");
            ProxyException caughtException = null;

            //Call and verify
            try
            {
                searchFuzzyController.Object.Initialize();
            }
            catch (ProxyException e)
            {
                caughtException = e;
            }

            if (caughtException != null)
            {
                Assert.AreEqual(ClientParams.InnerException, caughtException.InnerException);
            }
            else
            {
                Assert.IsTrue(false);
            }
        }
    }
}
