﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace ProxyServerUnitTests
{
    using System;
    using System.Collections.Generic;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http.Routing;
    using Microsoft.Azure.LocationServices;
    using Microsoft.Azure.LocationServices.Controllers;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.Models;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using Moq;

    [TestClass]
    public class SearchPOICategoryControllerTests
    {
        private MockRepository mockRepository;

        private Mock<IVersioning> versionObj;

        private Mock<SearchPOICategoryController> SearchPOICategoryController;

        [TestInitialize]
        public void TestInitialize()
        {
            this.mockRepository = new MockRepository(MockBehavior.Strict) { CallBase = true };
            this.versionObj = this.mockRepository.Create<IVersioning>();
            SearchPOICategoryController = new Mock<SearchPOICategoryController>() { CallBase = true };
        }

        [TestMethod, TestCategory("L0")]
        public void GetSearchPOICategoryTest_ProxyObjNull()
        {
            //Inject dependency
            SearchPOICategoryController.Object.versionObject = this.versionObj.Object;
            ProxyException expectedException = new ProxyException(ProxyException.ExceptionType.ServerError, "failed to initialize proxy object failed");

            //Setup dependency
            this.versionObj.Setup(x => x.GetProxyObj(It.IsAny<double>(), BaseProxy.ProxyType.Search, It.IsAny<IProxyOperation>())).Returns((BaseProxySearch)null);
            SearchPOICategoryController.Setup(x => x.BuildErrorResponseAndLog(It.IsAny<ProxyException>())).Returns((HttpResponseMessage)null);

            //Call and verify
            var response = SearchPOICategoryController.Object.GetSearchPOICategory("");
            this.versionObj.Verify(x => x.GetProxyObj(It.IsAny<double>(), BaseProxy.ProxyType.Search, null), Times.Once);
            SearchPOICategoryController.Verify(x => x.BuildErrorResponseAndLog(It.Is<ProxyException>(e => e.Message == expectedException.Message)), Times.Once);
        }

        [TestMethod, TestCategory("L0")]
        public void GetSearchPOICategoryTest()
        {
            //Setup vars
            HttpResponseMessage response = new HttpResponseMessage();
            response.Headers.Add("WWW-Authenticate", "Basic");
            TaskCompletionSource<HttpResponseMessage> task = new TaskCompletionSource<HttpResponseMessage>();
            task.SetResult(response);
            IProxyOperation proxyOperation = new ProxyOperation("Map");
            Mock<BaseProxySearch> BaseProxySearchMock = new Mock<BaseProxySearch>(proxyOperation);
            Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>();
            SearchPOICategoryController.Object.SetQueryStrings(querystrings);

            //Inject dependency
            SearchPOICategoryController.Object.versionObject = this.versionObj.Object;

            //Setup dependency
            this.versionObj.Setup(x => x.GetProxyObj(It.IsAny<double>(), BaseProxy.ProxyType.Search, It.IsAny<IProxyOperation>())).Returns(BaseProxySearchMock.Object);
            BaseProxySearchMock.Setup(x => x.SearchProxy(BaseProxySearch.SearchType.POICategory, It.IsAny<string>(), It.IsAny<string>(), querystrings, null)).Returns(task.Task);

            //Call and verify
            var output = SearchPOICategoryController.Object.GetSearchPOICategory("");
            this.versionObj.Verify(x => x.GetProxyObj(It.IsAny<double>(), BaseProxy.ProxyType.Search, It.IsAny<IProxyOperation>()), Times.Once);
            BaseProxySearchMock.Verify(x => x.SearchProxy(BaseProxySearch.SearchType.POICategory, It.IsAny<string>(), It.IsAny<string>(), querystrings, null));
            Assert.AreEqual(output.Result, task.Task.Result);
        }

        [TestMethod, TestCategory("L0")]
        public void InitializeTest_GetSearchPOICategory()
        {
            //Setup vars
            HttpRequestMessage Request = new HttpRequestMessage(HttpMethod.Get, "https://localhost/search/poi/category/json?api-version=1.0&query=atm&subscription-id=GS61xORRtw-KVNYqocNmZp4iZizSFXr2FpebcwyaXDQ&tracking-id=1");
            SearchPOICategoryController.Object.Request = Request;
            IProxyOperation proxyOperation = new ProxyOperation("Search");
            SearchPOICategoryController.Object.SetProxyOperation(proxyOperation);
            string json = "json";
            IHttpRoute HttpRoute = new HttpRoute();
            HttpRouteValueDictionary routeValueDictionary = new HttpRouteValueDictionary();
            routeValueDictionary.Add("formatstring", json);
            routeValueDictionary.Add("action", "GetSearchPOICategory");
            HttpRouteData routeData = new HttpRouteData(HttpRoute, routeValueDictionary);
            SearchPOICategoryController.Object.Request.SetRouteData(routeData);

            //Call and verify
            SearchPOICategoryController.Object.Initialize();
            Assert.AreEqual(json, SearchPOICategoryController.Object.Format);
            Assert.AreEqual(1, SearchPOICategoryController.Object.Version);
            Assert.AreEqual("atm", SearchPOICategoryController.Object.Query);
        }

        [TestMethod, TestCategory("L0")]
        public void InitializeTest_GetSearchPOICategory_Noquery()
        {
            //Setup vars
            HttpRequestMessage Request = new HttpRequestMessage(HttpMethod.Get, "https://localhost/search/poi/category/json?api-version=1.0&subscription-id=GS61xORRtw-KVNYqocNmZp4iZizSFXr2FpebcwyaXDQ&tracking-id=1");
            SearchPOICategoryController.Object.Request = Request;
            IProxyOperation proxyOperation = new ProxyOperation("Search");
            SearchPOICategoryController.Object.SetProxyOperation(proxyOperation);
            string json = "json";
            IHttpRoute HttpRoute = new HttpRoute();
            HttpRouteValueDictionary routeValueDictionary = new HttpRouteValueDictionary();
            routeValueDictionary.Add("formatstring", json);
            routeValueDictionary.Add("action", "GetSearchPOICategory");
            HttpRouteData routeData = new HttpRouteData(HttpRoute, routeValueDictionary);
            SearchPOICategoryController.Object.Request.SetRouteData(routeData);
            ProxyException ClientParams = new ProxyException(ProxyException.ExceptionType.ClientParams, "query param is empty");
            ProxyException caughtException = null;

            //Call and verify
            try
            {
                SearchPOICategoryController.Object.Initialize();
            }
            catch (ProxyException e)
            {
                caughtException = e;
            }

            if (caughtException != null)
            {
                Assert.AreEqual(ClientParams.InnerException, caughtException.InnerException);
            }
            else
            {
                Assert.IsTrue(false);
            }
        }
    }
}
