﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace ProxyServerUnitTests
{
    using System;
    using System.Collections.Generic;
    using System.Net.Http;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using Moq;

    using Newtonsoft.Json.Linq;

    [TestClass]
    public class TomTomV2SearchProxyTests
    {
        private Mock<TomTomV2SearchProxy> baseProxyObj;

        [TestInitialize]
        public void TestInitialize()
        {
            IProxyOperation proxyOperation = new ProxyOperation("Search");
            baseProxyObj = new Mock<TomTomV2SearchProxy>(proxyOperation) { CallBase = true };
        }

        [TestMethod, TestCategory("L0")]
        public void SearchFuzzyTest()
        {
            //Setup vars
            HttpRequestMessage outputUri = new HttpRequestMessage(HttpMethod.Get, "https://api.azure.tomtom.com/search/2/search/seattle.png?key=");
            string query = "seattle";
            string formatString = "png";
            Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>();

            //Setup dependency
            baseProxyObj.Setup(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)));

            //Call and verify
            baseProxyObj.Object.SearchProxy(BaseProxySearch.SearchType.Fuzzy, query, formatString, querystrings);
            baseProxyObj.Verify(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)), Times.Once);
        }

        [TestMethod, TestCategory("L0")]
        public void SearchAddressTest()
        {
            //Setup vars
            HttpRequestMessage outputUri = new HttpRequestMessage(HttpMethod.Get, "https://api.azure.tomtom.com/search/2/geocode/query%3D15127+NE+24th+Street%2C+Redmond%2C+WA+98052.json?key=");
            string query = "query=15127 NE 24th Street, Redmond, WA 98052";
            string formatString = "json";
            Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>();

            //Setup dependency
            baseProxyObj.Setup(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)));

            //Call and verify
            baseProxyObj.Object.SearchProxy(BaseProxySearch.SearchType.Address, query, formatString, querystrings);
            baseProxyObj.Verify(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)), Times.Once);
        }

        [TestMethod, TestCategory("L0")]
        public void SearchAddressReverseTest()
        {
            //Setup vars
            HttpRequestMessage outputUri = new HttpRequestMessage(HttpMethod.Get, "https://api.azure.tomtom.com/search/2/reverseGeocode/json?key=&position=37.337000%2C-121.890000");
            string query = "37.337,-121.89";
            string formatString = "json";
            Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>();

            //Setup dependency
            baseProxyObj.Setup(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)));

            //Call and verify
            baseProxyObj.Object.SearchProxy(BaseProxySearch.SearchType.ReverseAddress, query, formatString, querystrings);
            baseProxyObj.Verify(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)), Times.Once);
        }

        [TestMethod, TestCategory("L0")]
        public void SearchAddressReverseCrossStreetTest()
        {
            //Setup vars
            HttpRequestMessage outputUri = new HttpRequestMessage(HttpMethod.Get, "https://api.azure.tomtom.com/search/2/reverseGeocode/crossStreet/37.337,-121.89.png?key=");
            string query = "37.337,-121.89";
            string formatString = "png";
            Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>();

            //Setup dependency
            baseProxyObj.Setup(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)));

            //Call and verify
            baseProxyObj.Object.SearchProxy(BaseProxySearch.SearchType.CrossStreet, query, formatString, querystrings);
            baseProxyObj.Verify(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)), Times.Once);
        }

        [TestMethod, TestCategory("L0")]
        public void SearchAddressStruturedTest()
        {
            //Setup vars
            HttpRequestMessage outputUri = new HttpRequestMessage(HttpMethod.Get, "https://api.azure.tomtom.com/search/2/structuredGeocode.json?key=&countryCode=US&streetNumber=15127");
            string query = "";
            string countryCode = "US";
            string formatString = "json";
            string streetNumber = "15127";
            List<string> countryCodeList = new List<string>();
            countryCodeList.Add(countryCode);
            List<string> streetNumberList = new List<string>();
            streetNumberList.Add(streetNumber);
            Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>();
            querystrings.Add("countryCode", countryCodeList);
            querystrings.Add("streetNumber", streetNumberList);

            //Setup dependency
            baseProxyObj.Setup(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)));

            //Call and verify
            baseProxyObj.Object.SearchProxy(BaseProxySearch.SearchType.AddressStructured, query, formatString, querystrings);
            baseProxyObj.Verify(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)), Times.Once);
        }

        [TestMethod, TestCategory("L0")]
        public void SearchPOITest()
        {
            //Setup vars
            HttpRequestMessage outputUri = new HttpRequestMessage(HttpMethod.Get, "https://api.azure.tomtom.com/search/2/poiSearch/statue+of+liberty.json?key=");
            string query = "statue of liberty";
            string formatString = "json";
            Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>();

            //Setup dependency
            baseProxyObj.Setup(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)));

            //Call and verify
            baseProxyObj.Object.SearchProxy(BaseProxySearch.SearchType.POI, query, formatString, querystrings);
            baseProxyObj.Verify(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)), Times.Once);
        }

        [TestMethod, TestCategory("L0")]
        public void SearchPOICategoryTest()
        {
            //Setup vars
            HttpRequestMessage outputUri = new HttpRequestMessage(HttpMethod.Get, "https://api.azure.tomtom.com/search/2/categorySearch/atm.json?key=");
            string query = "atm";
            string formatString = "json";
            Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>();

            //Setup dependency
            baseProxyObj.Setup(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)));

            //Call and verify
            baseProxyObj.Object.SearchProxy(BaseProxySearch.SearchType.POICategory, query, formatString, querystrings);
            baseProxyObj.Verify(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)), Times.Once);
        }

        [TestMethod, TestCategory("L0")]
        public void SearchNearByTest()
        {
            //Setup vars
            HttpRequestMessage outputUri = new HttpRequestMessage(HttpMethod.Get, "https://api.azure.tomtom.com/search/2/nearbySearch/.json?key=&lat=37.337&lon=-121.89");
            string query = "";
            string lat = "37.337";
            string lon = "-121.89";
            string formatString = "json";
            List<string> latList = new List<string>();
            latList.Add(lat);
            List<string> lonList = new List<string>();
            lonList.Add(lon);
            Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>();
            querystrings.Add("lat", latList);
            querystrings.Add("lon", lonList);

            //Setup dependency
            baseProxyObj.Setup(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)));

            //Call and verify
            baseProxyObj.Object.SearchProxy(BaseProxySearch.SearchType.Nearby, query, formatString, querystrings);
            baseProxyObj.Verify(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)), Times.Once);
        }

        [TestMethod, TestCategory("L0")]
        public void SearchAlongRouteTest()
        {
            //Setup vars
            HttpRequestMessage outputUri = new HttpRequestMessage(HttpMethod.Get, "https://api.azure.tomtom.com/search/2/nearbySearch/.json?key=&maxDetourTime=1000");
            string query = "burger";
            string maxDetourTime = "1000";
            string formatString = "json";
            List<string> maxDetourTimeList = new List<string>();
            maxDetourTimeList.Add(maxDetourTime);
            Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>();
            querystrings.Add("maxDetourTime", maxDetourTimeList);

            //Setup dependency
            baseProxyObj.Setup(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)));

            //Call and verify
            baseProxyObj.Object.SearchProxy(BaseProxySearch.SearchType.Nearby, query, formatString, querystrings);
            baseProxyObj.Verify(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)), Times.Once);
        }

        [TestMethod, TestCategory("L0")]
        public void SearchGeometryTest()
        {
            //Setup vars
            HttpRequestMessage outputUri = new HttpRequestMessage(HttpMethod.Get, "https://api.azure.tomtom.com/search/2/geometrySearch/pizza.json?key=&limit=2");
            string query = "pizza";
            string limit = "2";
            string formatString = "json";
            List<string> limitList = new List<string>();
            limitList.Add(limit);
            Dictionary<string, List<string>> querystrings = new Dictionary<string, List<string>>();
            querystrings.Add("limit", limitList);
            JArray jPOIArray = new JArray();
            JObject poi = new JObject();
            JObject jPosition = new JObject(new JProperty("lat", "37.337"), new JProperty("lon", "-121.89"));
            poi.Add("position", jPosition);
            jPOIArray.Add(poi);
            JObject requestBody = new JObject(new JProperty("poiList", jPOIArray));

            //Setup dependency
            baseProxyObj.Setup(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)));

            //Call and verify
            baseProxyObj.Object.SearchProxy(BaseProxySearch.SearchType.Geometry, query, formatString, querystrings, "", "0", requestBody.ToString());
            baseProxyObj.Verify(p => p.ExecuteRequest(It.Is<HttpRequestMessage>(s => s.RequestUri == outputUri.RequestUri)), Times.Once);
        }
    }
}
