﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Shared
{
    using System;
    using System.Collections.Generic;
    using System.Text;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Common.Validation;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using Newtonsoft.Json;

    public class BlobHelper : IBlobHelper
    {
        private const int MaxNumberOfRetries = 10;
        private const int IntervalBetweenRetriesInSeconds = 30;
        private const int ExpirationMapBlobAppendTextLeasePeriodInSeconds = 15;

        // The name of the blob that contains the expiration map for the container. 
        private static readonly string ExpirationMapBlobName = SharedUtilities.ExpirationMapBlobName;

        // Dependencies
        private readonly IBlobService blobService;
        
        public BlobHelper(IBlobService blobService)
        {
            Assert.ParamIsNotNull(blobService, nameof(blobService));

            this.blobService = blobService;
        }

        public async Task SetBlobExpiryDate(ICloudBlob blob, DateTime expirationDateTimeInUtc)
        {
            Assert.ParamIsNotNull(blob, nameof(blob));
            Assert.IsTrue(expirationDateTimeInUtc > DateTime.UtcNow, $"{nameof(expirationDateTimeInUtc)} should be in the future");

            // Make sure the Append Blob exists
            await this.blobService.CreateAppendBlobIfNotExistsAsync(blob.Container.Name, ExpirationMapBlobName);

            // Create and Serialize the Expiration Info
            string blobExpirationInfo = $"{JsonConvert.SerializeObject(new BlobExpirationInfo(blob.Uri, expirationDateTimeInUtc))}{Environment.NewLine}";
            CloudAppendBlob appendBlob = this.blobService.GetAppendBlobReference(blob.Container.Name, ExpirationMapBlobName);

            string leaseId = Guid.NewGuid().ToString();

            // There are multiple instances of workers processing batch requests and setting up expiry for their response blobs so 
            // we need a retry mechanism to make the whole process a bit more robust.
            await Helpers.Retry(
                async () =>
                    {
                        // Try to acquire a 15 sec lease on the Expiration map blob
                        await this.blobService.AcquireLeaseAsync(
                            appendBlob,
                            TimeSpan.FromSeconds(ExpirationMapBlobAppendTextLeasePeriodInSeconds),
                            leaseId);

                        // Append the Info to the Expiration Map Blob using the current lease
                        await this.blobService.AppendTextAsync(
                            blob.Container.Name, 
                            ExpirationMapBlobName,
                            blobExpirationInfo,
                            Encoding.UTF8,
                            AccessCondition.GenerateLeaseCondition(leaseId),
                            new BlobRequestOptions(), 
                            new OperationContext());

                        // Release the lease ASAP
                        await this.blobService.ReleaseLeaseAsync(
                            appendBlob, 
                            AccessCondition.GenerateLeaseCondition(leaseId));
                    },
                MaxNumberOfRetries,
                "SetBlobExpiryDate",
                TimeSpan.FromSeconds(IntervalBetweenRetriesInSeconds).Milliseconds,
                new List<Type> { typeof(StorageException) },
                true);
        }
    }
}