﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.AzLocApiRunner
{
    using System;
    using System.Configuration;
    using System.Diagnostics;
    using System.Net;
    using System.Net.Http;
    using System.Security.Cryptography.X509Certificates;
    using System.Threading;
    using Microsoft.Azure.LocationServices.AzLocApiRunner.Exceptions;
    using Microsoft.Azure.LocationServices.AzLocApiRunner.Utils;
    using Microsoft.Cis.Monitoring.Runners;
    using Microsoft.Cloud.InstrumentationFramework;

    public class AzLocApiRunner : RunnerAuthorBase
    {
        private readonly string assignedRequestUriShortName;
        private readonly string assignedRequestUri;
        private readonly string mdmAccountName;
        private readonly string mdmMetricNamespace;
        private readonly string keyType;

        public AzLocApiRunner(string[] args)
            : base(args)
        {
            this.assignedRequestUriShortName = this.GetConfigurationValue("UriShortName");
            this.assignedRequestUri = this.GetConfigurationValue("Uri");
            this.keyType = this.GetConfigurationValue("KeyType");
            this.mdmAccountName = this.GetConfigurationValue("MetricAccount");
            this.mdmMetricNamespace = this.GetConfigurationValue("MetricNamespace");

            Console.WriteLine($"Configured instance with Uri Short Name {this.assignedRequestUriShortName} with MdmAccount {this.mdmAccountName}");
        }

        public override void StartRunner()
        {
            string kvResult = string.Empty;
            DateTime keyVaultUpdated = DateTime.UtcNow;

            while (true)
            {
                try
                {
                    Console.WriteLine($"Runner instance started at {DateTime.UtcNow} with key type {this.keyType}");

                    // First, attempt to acquire the secret from KeyVault. 
                    // Map control does not require auth
                    if (string.IsNullOrEmpty(kvResult) && !this.keyType.Equals("NoAuth"))
                    {
                        X509Certificate2 pfxCert = null;

                        if (!string.IsNullOrWhiteSpace(this.GetConfigurationValue("AADCertSubjectName")))
                        {
                            string subjectName = this.GetConfigurationValue("AADCertSubjectName");
                            var certs = this.GetCertificatesBySubjectName(subjectName);

                            if (certs.Count <= 0)
                            {
                                Console.WriteLine($"There were no certs which matched {subjectName}");
                                throw new ArgumentOutOfRangeException(nameof(certs), $"There were no certs which matched {subjectName}");
                            }

                            pfxCert = this.GetCertificatesBySubjectName(this.GetConfigurationValue("AADCertSubjectName"))[0];
                        }

                        KeyVaultUtils kvUtils = new KeyVaultUtils(this.GetConfigurationValue("AADClientId"), pfxCert);
                        kvResult = kvUtils.AcquireKeyVaultSecret(this.GetConfigurationValue(this.keyType)).GetAwaiter().GetResult();

                        if (string.IsNullOrEmpty(kvResult))
                        {
                            // If we failed to acquire the KV secret, throw and exception to log and trigger the wait
                            // An exception is necessary to set the state of the runner to Error.
                            throw new KeyVaultSecretNotFoundException(this.keyType);
                        }
                        else
                        {
                            // Update the most recent keyvault time.
                            keyVaultUpdated = DateTime.UtcNow;
                            Console.WriteLine("Successfully bootstrapped KeyVault and read API Key secret.");
                        }
                    }

                    var requestUri = this.assignedRequestUri.Replace("{apikey}", kvResult);

                    // Remove the ampersand escaping required for the XML based configuration
                    requestUri = requestUri.Replace("amp;", string.Empty);

                    Console.WriteLine("Getting HTTP client to begin run pass.");

                    using (var client = new HttpClient())
                    {
                        Stopwatch requestStopwatch = new Stopwatch();

                        Console.WriteLine("Runner preparing to send call.");
                        requestStopwatch.Start();
                        var response = client.GetAsync(requestUri).GetAwaiter().GetResult();
                        requestStopwatch.Stop();
                        Console.WriteLine("Runner Done sending call");

                        if (response.IsSuccessStatusCode)
                        {
                            var runnerStatus = $"[Status Healthy] {this.RunnerInstance}";
                            Console.WriteLine(runnerStatus);
                            this.LogHealthy(runnerStatus);
                        }
                        else
                        {
                            if ((response.StatusCode >= HttpStatusCode.BadRequest && response.StatusCode < HttpStatusCode.InternalServerError)
                                || response.StatusCode == HttpStatusCode.GatewayTimeout)
                            {
                                // If we get a 403 result and the key is more than one hour hold, do not report unhealthy,
                                // attempt to get a new key. If that new key also fails, then report Degraded.
                                if (response.StatusCode == HttpStatusCode.Forbidden 
                                    && DateTime.UtcNow.Subtract(keyVaultUpdated) > TimeSpan.FromHours(1))
                                {
                                    kvResult = string.Empty;
                                    continue;
                                }

                                // 400 - 499 and GatewayTimeout
                                // This will show as a warning in MDM. We will trigger differently on 4xx class respone codes. 
                                var runnerStatus = $"[Status Degraded] {this.RunnerInstance} failed. Received {response.StatusCode}";
                                Console.WriteLine(runnerStatus);
                                this.LogDegraded(runnerStatus);
                            }
                            else
                            {
                                var runnerStatus = $"[Status Unhealthy] {this.RunnerInstance} failed. Received {response.StatusCode}";

                                // I hope that in the future we will log more details to console so that these calls' data payload actually diverges. 
                                Console.WriteLine(runnerStatus);
                                this.LogUnhealthy(runnerStatus);
                            }
                        }

                        // Log the MDM metric
                        this.EmitMDMMetric("APIResponseTime", requestStopwatch.ElapsedMilliseconds, $"{response.StatusCode}");

                        // Log the Runner QOS metric for the 5MinuteQOS table
                        this.LogQos(this.RunnerInstance, true, (int)response.StatusCode, (int)requestStopwatch.ElapsedMilliseconds);
                    }
                }
                catch (WebException e)
                {
                    this.LogUnhealthy($"[{this.assignedRequestUriShortName}] failed with WebException {e.Message} - Exception Status: {e.Status}");
                }
                catch (Exception e)
                {
                    // Something happened during bootstrap/creation of the runner but is un-related to TomTom.
                    Console.WriteLine(e);
                    this.LogUnhealthy($"[{this.assignedRequestUriShortName}] {e.Message}");
                }
                finally
                {
                    // One minute timeout to avoid Runner Framework imposed throttling
                    Thread.Sleep(TimeSpan.FromMinutes(1));
                }
            }
        }

        private void EmitMDMMetric(string metricName, long value, string responseMessage)
        {
            if (string.IsNullOrWhiteSpace(this.mdmMetricNamespace))
            {
                throw new ArgumentNullException(nameof(mdmMetricNamespace), "No MDM metric namespace has been defined.");
            }

            Console.WriteLine($"Sending MDM metric to {this.mdmAccountName}, {this.mdmMetricNamespace}, {metricName}");

            ErrorContext mdmError = new ErrorContext();
            MeasureMetric4D mdmMetric = MeasureMetric4D.Create(
                this.mdmAccountName,
                this.mdmMetricNamespace,
                metricName,
                "Environment",      // The Environment name as specified in the RunnerDefinition.xml.
                "RunnerName",       // The current runner name, use property RunnerName.
                "RunnerInstance",   // The current runner instance, use property RunnerInstance.
                "RunnerResponseCode",
                ref mdmError);

            if (mdmMetric == null)
            {
                Console.WriteLine("Failed to create MeasureMetric, error code is {0:X}", mdmError.ErrorCode);
                Console.WriteLine("        error message: {0}", mdmError.ErrorMessage);

                return;
            }

            Console.WriteLine($"Created MDM metric to {this.mdmAccountName}, {this.mdmMetricNamespace}, {metricName}");

            if (!mdmMetric.LogValue(value, this.Environment, this.RunnerName, this.RunnerInstance, responseMessage, ref mdmError))
            {
                Console.WriteLine("Fail to set MeasureMetric value, error code is {0:X}", mdmError.ErrorCode);
                Console.WriteLine("    error message: {0}", mdmError.ErrorMessage);
            }

            Console.WriteLine($"Sent MDM metric to {this.mdmAccountName}, {this.mdmMetricNamespace}, {metricName}");
        }
    }
}
