﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
using System.Runtime.CompilerServices;

[assembly: InternalsVisibleTo("Microsoft.Azure.LocationServices.BillingUnitTest")]

namespace Microsoft.Azure.LocationServices.BillingAPI
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Fabric;
    using System.Linq;
    using System.Threading;
    using System.Threading.Tasks;
    using System.Timers;
    using Microsoft.Azure.LocationServices.BillingAPI.Utility;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.ServiceFabric.Services.Communication.Runtime;
    using Microsoft.ServiceFabric.Services.Runtime;
    using Microsoft.WindowsAzure.Storage;

    /// <summary>
    /// An instance of this class is created for each service instance by the Service Fabric runtime.
    /// </summary>
    public sealed class BillingAPI : StatelessService
    {
        internal const int BILLING_EMIT_INCREMENTS = 100;
        internal const int BILLING_TRANSACTIONS_PER_EMIT_QUANTITY = 1000;
        internal const int BILLING_EMIT_INTERVAL_MINUTES = 20;
        private static readonly string LEASE_CONTAINER = "BillingLease";
        private static readonly string LEASE_NAME = "lease";

        /// <summary>
        /// Configuration Reader which is defaulted to a keyvault based cache solution.
        /// </summary>
        private static readonly IConfigurationReader _configurationReader = new KeyVaultConfigurationReader();
        
        /// <summary>
        /// Startup configuration wrapper
        /// </summary>
        private static IBootstrapConfiguration _bootstrapConfiguration;

        public BillingAPI(StatelessServiceContext context)
            : base(context)
        {
        }

        /// <summary>
        /// Returns partition Id to be used for logging billing transactions.
        /// </summary>
        public static string GetTransactionLogPartition()
        {
            var now = GetTransactionLogDateTime(); 
            return now.ToString("yyyyMMddHHmm");
        }

        /// <summary>
        /// Returns datetime to be used for logging billing transactions.
        /// </summary>
        internal static DateTime GetTransactionLogDateTime()
        {
            //Round up transaction logs to make sure there are no race conditions when processing
            var now = DateTime.UtcNow;
            var buffer = now.Minute % BILLING_EMIT_INTERVAL_MINUTES != 0 ? BILLING_EMIT_INTERVAL_MINUTES - (now.Minute % BILLING_EMIT_INTERVAL_MINUTES) : BILLING_EMIT_INTERVAL_MINUTES;
            now = now + TimeSpan.FromMinutes(buffer);
            return now; 
        }

        /// <summary>
        /// Returns a quentity to emit given when given a number of increments. 
        /// </summary>
        internal static double GetEmitQuantity(long increments)
        {
            if (increments < 0)
            {
                throw new ArgumentException($"Attempt to emit negative increment count: {increments}");
            }

            var x = ((double)BILLING_TRANSACTIONS_PER_EMIT_QUANTITY / (double)BILLING_EMIT_INCREMENTS);
            return increments / x;
        }

        /// <summary>
        /// Optional override to create listeners (e.g., TCP, HTTP) for this service replica to handle client or user requests.
        /// </summary>
        /// <returns>A collection of listeners.</returns>
        protected override IEnumerable<ServiceInstanceListener> CreateServiceInstanceListeners()
        {
            return new ServiceInstanceListener[0];
        }

        /// <inheritdoc/>
        protected override async Task RunAsync(CancellationToken cancellationToken)
        {
            try
            {
                Meter.Meter.Instance.Initialize();
                ConfigurationReader.Instance.Initialize(Context);
                _bootstrapConfiguration = new ServiceFabricConfigurationBootstrap(ConfigurationReader.Instance);
                await BillingHelper.InitalizeAsync(_bootstrapConfiguration, _configurationReader);
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "billing", "initialization");
                throw;
            }

            var timer = new System.Timers.Timer(TimeSpan.FromMinutes(BILLING_EMIT_INTERVAL_MINUTES).TotalMilliseconds);
            timer.Elapsed += BillingRunner;
            timer.AutoReset = true;
            timer.Start();

            while (!cancellationToken.IsCancellationRequested)
            {
                await Task.Delay(TimeSpan.FromSeconds(1), cancellationToken);
            }
        }

        private static async void BillingRunner(object source, ElapsedEventArgs e)
        {
            try
            {
                string billingSasToken = await _configurationReader.GetConfigurationValueAsync(_bootstrapConfiguration.GetValue(Globals.BillingSasTokenKey));
                string billingConnectionTemplate = await _configurationReader.GetConfigurationValueAsync(_bootstrapConfiguration.GetValue(Globals.BillingConnectionStringTemplateKey));
                string connectionString = string.Format(billingConnectionTemplate, billingSasToken);
                var account = CloudStorageAccount.Parse(connectionString);                
                var logger = new Logger("Billing");
                using (var pks = new PublisherTokenSentinel(account, logger, LEASE_CONTAINER, LEASE_NAME))
                {
                    var start = DateTime.Now;
                    while (!pks.HasLease)
                    {
                        var span = DateTime.Now - start;
                        if (span > TimeSpan.FromSeconds(PublisherTokenSentinel.LeaseSpan.TotalSeconds * 2))
                        {
                            break;
                        }

                        Thread.Sleep(100);
                    }

                    if (pks.HasLease)
                    {
                        var lastProcessed = BillingHelper.GetLastProcessedPartition();
                        var partitionsToProcess = BillingHelper.GetPartitionsToProcess(lastProcessed, BILLING_EMIT_INTERVAL_MINUTES);
                        foreach (var partition in partitionsToProcess)
                        {
                            var sw = new Stopwatch();
                            sw.Start();
                            if (partition.Day == 1 && partition.Hour == 0 && partition.Minute == 0)
                            {
                                BillingHelper.SwitchCacheTable(partition);
                            }

                            long transactionLogCount = 0; 
                            Agent.Instance.LogEvent(EventType.Information, $"Processing Partition Started {BillingHelper.DateTimeToPartition(partition)}");
                            var logs = BillingHelper.GetLogs(partition);
                            if (logs.Any())
                            {
                                ProcessLogs(logs, ref transactionLogCount, partition);
                            }

                            sw.Stop();
                            BillingHelper.InsertProcessedPartition(partition, sw.Elapsed, transactionLogCount);
                            Agent.Instance.LogEvent(EventType.Information, $"Processing Partition Completed {BillingHelper.DateTimeToPartition(partition)} {sw.Elapsed}");
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "billing", "runner"); 
            }
        }

        private static void ProcessLogs(IEnumerable<TransactionLogEntity> logs, ref long transactionLogCount, DateTime partition)
        {
            // user is equvilent to one instance of a LBS resource
            Dictionary<string, Dictionary<string, long>> userTransctionCountMap = BuildUserMethodTransactionCountMap(logs, ref transactionLogCount);

            var transactionCache = BillingHelper.GetTransactionCache(userTransctionCountMap.Keys, partition);

            UpdateUserMethodTransactionMapWithCache(userTransctionCountMap, transactionCache);

            Dictionary<string, Dictionary<string, long>> userMeterTransactionCountMap = BuildUserMeterTransactionCountMap(userTransctionCountMap);

            var usagePartition = Guid.NewGuid().ToString();

            // Go through all the meters and emit to commerce if any has hit the expected threshold 
            var usageTableEntities = new List<UsageTableEntity>();
            foreach (var userMeterCount in userMeterTransactionCountMap)
            {
                foreach (var meterCountMap in userMeterCount.Value)
                {
                    if (meterCountMap.Value >= BILLING_EMIT_INCREMENTS)
                    {
                        var increments = meterCountMap.Value / BILLING_EMIT_INCREMENTS;

                        var resourceId = Helpers.ParseResourceIdFromAccountId(userMeterCount.Key);
                        var accountId = BillingHelper.ParseAccountId(userMeterCount.Key);
                        if (string.IsNullOrEmpty(resourceId))
                        {
                            continue; 
                        }

                        var meter = meterCountMap.Key;

                        if (increments > 0)
                        {
                            double quantity = GetEmitQuantity(increments);
                            
                            var azureSubscriptionId = new Guid(accountId.Item1);
                            {
                                var meterId = Meter.Meter.Instance.SkuMeterIdMap[accountId.Item4][meter];
                                var input = new UsageTableEntity
                                {
                                    PartitionKey = usagePartition,
                                    RowKey = Guid.NewGuid().ToString(),
                                    Timestamp = DateTime.UtcNow,
                                    EventDateTime = DateTime.UtcNow,
                                    EventId = Guid.NewGuid(),
                                    Location = "eastus",
                                    MeterId = meterId,
                                    Quantity = quantity,
                                    ResourceUri = resourceId,
                                    SubscriptionId = azureSubscriptionId,
                                    Tags = ""
                                };
                                usageTableEntities.Add(input);
                                UpdateUserMeter(userMeterCount.Key, meter, increments * BILLING_EMIT_INCREMENTS, userTransctionCountMap);
                            }
                        }
                    }
                }
            }

            // Update Cache
            var transactionCacheEntities = new List<TransactionCacheEntity>();
            foreach (var userTransactionCount in userTransctionCountMap)
            {
                foreach (var transactionCount in userTransactionCount.Value)
                {
                    transactionCacheEntities.Add(new TransactionCacheEntity(userTransactionCount.Key, transactionCount.Key) { Count = transactionCount.Value });
                }
            }

            // This might be a little slow and may need to improve perf in the future. 
            Parallel.ForEach(transactionCacheEntities, new ParallelOptions() { MaxDegreeOfParallelism = 16 }, (t) =>
            {
                BillingHelper.InsertUpdateTransactionCache(t, partition);
            });

            // Write to usage
            foreach (var usage in Split(usageTableEntities, 100))
            {
                BillingHelper.WriteToUsageTable(usage);
            }
        }

        private static Dictionary<string, Dictionary<string, long>> BuildUserMeterTransactionCountMap(Dictionary<string, Dictionary<string, long>> userTransctionCountMap)
        {
            var userMeterCountMap = new Dictionary<string, Dictionary<string, long>>();
            foreach (var userTransactionCount in userTransctionCountMap)
            {
                if (!userMeterCountMap.TryGetValue(userTransactionCount.Key, out Dictionary<string, long> meterCountMap))
                {
                    meterCountMap = new Dictionary<string, long>();
                    userMeterCountMap[userTransactionCount.Key] = meterCountMap;
                }

                foreach (var transactionCount in userTransactionCount.Value)
                {
                    var meter = Meter.Meter.Instance.MethodMeterMap[transactionCount.Key];
                    if (meterCountMap.ContainsKey(meter))
                    {
                        meterCountMap[meter] += transactionCount.Value / Meter.Meter.Instance.MethodTransactionMap[transactionCount.Key];
                    }
                    else
                    {
                        meterCountMap[meter] = transactionCount.Value / Meter.Meter.Instance.MethodTransactionMap[transactionCount.Key];
                    }
                }
            }

            return userMeterCountMap;
        }

        private static void UpdateUserMethodTransactionMapWithCache(Dictionary<string, Dictionary<string, long>> userTransctionCountMap, IEnumerable<TransactionCacheEntity> transactionCache)
        {
            foreach (var tc in transactionCache)
            {
                if (!userTransctionCountMap.TryGetValue(tc.PartitionKey, out Dictionary<string, long> transactionCountMap))
                {
                    transactionCountMap = new Dictionary<string, long>();
                    userTransctionCountMap[tc.PartitionKey] = transactionCountMap;
                }

                if (transactionCountMap.ContainsKey(tc.RowKey))
                {
                    transactionCountMap[tc.RowKey] += tc.Count;
                }
                else
                {
                    transactionCountMap[tc.RowKey] = tc.Count;
                }
            }
        }

        private static Dictionary<string, Dictionary<string, long>> BuildUserMethodTransactionCountMap(IEnumerable<TransactionLogEntity> logs, ref long transactionLogCount)
        {
            var userTransctionCountMap = new Dictionary<string, Dictionary<string, long>>();

            // Build map of current transactions
            foreach (var log in logs)
            {
                transactionLogCount++;
                List<MethodStatus> methodStatuses; 
                try
                {
                    methodStatuses = log.GetMethodStatusList(); 
                }
                catch (Exception e)
                {
                    Instrumentation.Agent.Instance.LogException(e, "billing");
                    continue;
                }

                foreach (var methodStatus in methodStatuses)
                {
                    var statusCode = methodStatus.StatusCode;
                    var methodName = methodStatus.MethodName;

                    if ((statusCode >= 200 && statusCode < 300) || (statusCode >= 400 && statusCode < 500))
                    {
                        if (!userTransctionCountMap.TryGetValue(log.ResourceId, out Dictionary<string, long> transactionCountMap))
                        {
                            transactionCountMap = new Dictionary<string, long>();
                            userTransctionCountMap[log.ResourceId] = transactionCountMap;
                        }

                        if (!Meter.Meter.Instance.ServiceApiMethodMap.ContainsKey(methodName))
                        {
                            Instrumentation.Agent.Instance.LogException(new Exception($"ServiceApiMethodMap does not contain Service API {methodName}"), "billing");
                            continue;
                        }

                        var method = Meter.Meter.Instance.ServiceApiMethodMap[methodName];
                        if (transactionCountMap.ContainsKey(method))
                        {
                            transactionCountMap[method]++;
                        }
                        else
                        {
                            transactionCountMap[method] = 1;
                        }
                    }
                }
            }

            return userTransctionCountMap;
        }

        private static void UpdateUserMeter(string user, string meter, long processedCount, Dictionary<string, Dictionary<string, long>> userTransctionCountMap)
        {
            var transactionCountMap = userTransctionCountMap[user]; 
            foreach (var method in Meter.Meter.Instance.MethodMeterMap.Where(x => x.Value.Equals(meter, StringComparison.OrdinalIgnoreCase)).Select(x => x.Key))
            {
                if (!transactionCountMap.ContainsKey(method))
                {
                    continue;
                }

                var transactions = transactionCountMap[method] / Meter.Meter.Instance.MethodTransactionMap[method];
                if (processedCount >= transactions)
                {
                    userTransctionCountMap[user][method] -= transactions * Meter.Meter.Instance.MethodTransactionMap[method];
                    processedCount -= transactions;
                }
                else
                {
                    userTransctionCountMap[user][method] -= processedCount * Meter.Meter.Instance.MethodTransactionMap[method];
                    processedCount = 0;
                }

                if (processedCount <= 0)
                {
                    break;
                }
            }
        }

        private static IEnumerable<IEnumerable<T>> Split<T>(IEnumerable<T> input, int subsetSize)
        {
            //var result = new IEnumerable<IEnumerable<T>>(); 
            for (int index = 0; index < input.Count(); index += subsetSize)
            {
                yield return input.Skip(index).Take(subsetSize);
            }
        }
    }
}
