﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
namespace Microsoft.Azure.LocationServices.BillingAPI
{
    using System;
    using System.Collections.Generic;
    using System.Globalization;
    using System.Linq;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Queue;
    using Microsoft.WindowsAzure.Storage.RetryPolicies;
    using Microsoft.WindowsAzure.Storage.Table;

    internal class BillingHelper
    {
        internal const string LOG_TABLE = "UserTransactionLog";
        internal const string TRANSACTION_CACHE_TABLE_BASE = "UserTransactionCache";
        private const string LAST_PROCESSED_PARTITION_TABLE = "LastProcessedPartition";
        private const string PROCESSED_PARTITION_TABLE = "ProcessedPartitions";
        private const string BILLING_USAGE_TABLE = "UsageTable";
        private const string BILLING_ERROR_TABLE = "ErrorTable";
        private const string BILLING_USAGE_QUEUE = "UsageQueue";
        private const string BILLING_ERROR_QUEUE = "ErrorQueue";
        private static CloudStorageAccount _storageAccount;
        
        public static bool TryParsePartition(string partition, out DateTime dateTime)
        {
            return DateTime.TryParseExact(partition, "yyyyMMddHHmm", CultureInfo.InvariantCulture, DateTimeStyles.None, out dateTime);
        }

        public static bool TryParsePartition(long partition, out DateTime dateTime)
        {
            return DateTime.TryParseExact(partition.ToString(), "yyyyMMddHHmm", CultureInfo.InvariantCulture, DateTimeStyles.None, out dateTime);
        }

        public static string DateTimeToPartition(DateTime dateTime)
        {
            return dateTime.ToString("yyyyMMddHHmm");
        }

        /// <summary>
        /// At the start of each month, we will need to clear out the transaction cache table. 
        /// </summary>
        public static void SwitchCacheTable(DateTime partition)
        {
            var previousCacheTableName = GetPreviousTransactionCacheTable(partition); 
            var previousTable = GetTable(previousCacheTableName);
            var currentCacheTableName = GetCurrentTransactionCacheTable(partition);
            var currentTable = GetTable(currentCacheTableName);
            previousTable.DeleteIfExistsAsync();
            currentTable.CreateIfNotExists();
        }

        internal static string GetCurrentTransactionCacheTable(DateTime partition)
        {
            return $"{TRANSACTION_CACHE_TABLE_BASE}{partition.Month % 2}";
        }

        internal static string GetPreviousTransactionCacheTable(DateTime partition)
        {
            return $"{TRANSACTION_CACHE_TABLE_BASE}{(partition.Month % 2) ^ 1}";
        }

        internal static async Task InitalizeAsync(IBootstrapConfiguration bootstrapConfiguration, IConfigurationReader configurationReader)
        {
            if (bootstrapConfiguration == null)
            {
                throw new ArgumentNullException(nameof(bootstrapConfiguration));
            }

            if (configurationReader == null)
            {
                throw new ArgumentNullException(nameof(configurationReader));
            }

            string billingSasToken = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(Globals.BillingSasTokenKey));
            string billingConnectionTemplate = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(Globals.BillingConnectionStringTemplateKey));
            string connectionString = string.Format(billingConnectionTemplate, billingSasToken);
            _storageAccount = CloudStorageAccount.Parse(connectionString);

            var table = GetTable(LAST_PROCESSED_PARTITION_TABLE);
            var created = table.CreateIfNotExists();
            if (created)
            {
                var dateTime = BillingAPI.GetTransactionLogDateTime();
                dateTime = dateTime - TimeSpan.FromMinutes(20); 
                UpdateLastProcessedPartition(dateTime); 
            }

            table = GetTable(PROCESSED_PARTITION_TABLE);
            table.CreateIfNotExists();
            table = GetTable(BILLING_USAGE_TABLE);
            table.CreateIfNotExists();
            table = GetTable(BILLING_ERROR_TABLE);
            table.CreateIfNotExists();
            table = GetTable(LOG_TABLE);
            table.CreateIfNotExists();
            table = GetTable(GetCurrentTransactionCacheTable(DateTime.UtcNow));
            table.CreateIfNotExists();

            var queue = GetQueue(BILLING_USAGE_QUEUE);
            queue.CreateIfNotExists();
            queue = GetQueue(BILLING_ERROR_QUEUE);
            queue.CreateIfNotExists();
        }

        /// <summary>
        /// Returns transaction cache for given users
        /// </summary>
        /// <param name="userIds">Apim user Ids</param>
        /// <returns>List of TransactionCacheEntities</returns>
        internal static IEnumerable<TransactionCacheEntity> GetTransactionCache(IEnumerable<string> userIds, DateTime partition)
        {
            var currentTransactionCacheTableName = GetCurrentTransactionCacheTable(partition); 
            CloudTable table = GetTable(currentTransactionCacheTableName);
            var combined = string.Join(" or ", userIds.Select(x => $"(PartitionKey eq '{x}')"));
            TableQuery<TransactionCacheEntity> query = new TableQuery<TransactionCacheEntity>()
              .Where(combined);
            return table.ExecuteQuery(query);
        }

        /// <summary>
        /// Retrieve all logs for given partitions
        /// </summary>
        /// <param name="partitions">Partitions to get logs for</param>
        /// <returns>List of logs</returns>
        internal static IEnumerable<TransactionLogEntity> GetLogs(List<DateTime> partitions)
        {
            CloudTable table = GetTable(LOG_TABLE);
            var queryString = string.Join(" or ", partitions.Select(x => $"(PartitionKey eq '{DateTimeToPartition(x)}')"));
            TableQuery<TransactionLogEntity> query = new TableQuery<TransactionLogEntity>()
              .Where(queryString);

            return table.ExecuteQuery(query);
        }
        
        /// <summary>
        /// Retrieve all logs for given partitions
        /// </summary>
        /// <param name="partition"></param>
        /// <returns></returns>
        internal static IEnumerable<TransactionLogEntity> GetLogs(DateTime partition)
        {
            CloudTable table = GetTable(LOG_TABLE, TimeSpan.FromMinutes(30));
            var queryString = $"(PartitionKey eq '{DateTimeToPartition(partition)}')";
            TableQuery<TransactionLogEntity> query = new TableQuery<TransactionLogEntity>()
              .Where(queryString);

            return table.ExecuteQuery(query);
        }

        /// <summary>
        /// Insert/Update one transaction cache. 
        /// </summary>
        /// <param name="transaction"></param>
        internal static void InsertUpdateTransactionCache(TransactionCacheEntity transaction, DateTime partition)
        {
            var currentTransactionCacheTableName = GetCurrentTransactionCacheTable(partition);
            CloudTable table = GetTable(currentTransactionCacheTableName);
            TableOperation tableOperation = TableOperation.InsertOrReplace(transaction);
            table.Execute(tableOperation);
        }

        /// <summary>
        /// Emit usage to billing
        /// </summary>
        /// <param name="usage"></param>
        internal static void WriteToUsageTable(UsageTableEntity usage)
        {
            CloudTable table = GetTable(BILLING_USAGE_TABLE);
            TableOperation tableOperation = TableOperation.InsertOrReplace(usage);
            table.Execute(tableOperation);
        }

        /// <summary>
        /// Writes to the usage table
        /// </summary>
        /// <param name="usage"></param>
        internal static void WriteToUsageTable(IEnumerable<UsageTableEntity> usage)
        {
            CloudTable table = GetTable(BILLING_USAGE_TABLE);
            TableBatchOperation batchOperation = new TableBatchOperation();
            foreach (var u in usage)
            {
                batchOperation.Insert(u);
            }

            try
            {
                table.ExecuteBatch(batchOperation);
            }
            catch (Exception e)
            {
                throw new Exception("Failed to emit to usage table", e); 
            }

            CloudQueue queue = GetQueue(BILLING_USAGE_QUEUE);
            var message = $"{{'partitionKey':'{usage.First().PartitionKey}','batchId':'{Guid.NewGuid().ToString()}'}}";
            try
            {
                queue.AddMessage(new CloudQueueMessage(message));
            }
            catch (Exception e)
            {
                throw new Exception("Failed to emit to usage queue", e);
            }

            Instrumentation.Agent.Instance.BillingEmitLogEvent(usage.First().PartitionKey, Guid.NewGuid().ToString());
        }

        internal static void InsertIntoBillingQueue(List<string> partitions)
        {
            CloudQueue queue = GetQueue(BILLING_USAGE_QUEUE);
            partitions.ForEach(x =>
            {
                CloudQueueMessage message = new CloudQueueMessage(x);
                queue.AddMessage(message);
            });
        }

        internal static DateTime GetLastProcessedPartition()
        {
            CloudTable table = GetTable(LAST_PROCESSED_PARTITION_TABLE);
            TableQuery<LastProcessedPartitionEntity> query = new TableQuery<LastProcessedPartitionEntity>().Where(
                TableQuery.CombineFilters(
                    TableQuery.GenerateFilterCondition("PartitionKey", QueryComparisons.Equal, "a"),
                    TableOperators.And,
                    TableQuery.GenerateFilterCondition("RowKey", QueryComparisons.Equal, "a")));
            var sPartition = table.ExecuteQuery(query).ToList().Single().LastPartition;
            TryParsePartition(sPartition, out DateTime result);
            return result;
        }

        internal static void InsertProcessedPartition(DateTime partition, TimeSpan duration, long transactionCount)
        {
            CloudTable table = GetTable(PROCESSED_PARTITION_TABLE);

            var p = partition.ToString("yyyyMMdd");
            var k = DateTimeToPartition(partition);
            var entity = new ProcessedPartitionEntity(p, k)
            {
                Duration = duration.ToString(),
                TransactionCount = transactionCount
            };

            var tableOperation = TableOperation.InsertOrReplace(entity);
            table.Execute(tableOperation);

            UpdateLastProcessedPartition(partition);
        }

        /// <summary>
        /// Returns partitions to be processed when given current time and partition interval
        /// </summary>
        internal static List<DateTime> GetPartitionsToProcess(DateTime lastProcessedPartition, int partitionInternalInMinutes)
        {
            if (partitionInternalInMinutes <= 0)
            {
                throw new ArgumentException("Interval must be greater than 0");
            }

            var result = new List<DateTime>();
            var toProcess = lastProcessedPartition + TimeSpan.FromMinutes(partitionInternalInMinutes);

            var now = DateTime.UtcNow;
            now = new DateTime(now.Year, now.Month, now.Day, now.Hour, now.Minute, 0, now.Kind);

            // Logs are all rounded up, so we can process up to now
            while (toProcess <= now)
            {
                result.Add(toProcess);
                toProcess = toProcess + TimeSpan.FromMinutes(partitionInternalInMinutes);
            }

            return result;
        }

        internal static Tuple<string, string, string, string> ParseAccountId(string accountIdParam)
        {
            var tokens = accountIdParam.Split(new[] { "::" }, StringSplitOptions.RemoveEmptyEntries);
            string sku;
            if (tokens.Length == 3)
            {
                sku = SkuName.S0.ToString();
            }
            else if (tokens.Length == 4)
            {
                sku = tokens[3];
            }
            else
            {
                throw new ArgumentException($"Incorrectly formatted AccountId {accountIdParam}");
            }

            var subscriptionId = tokens[0];
            var resourceGroup = tokens[1];
            var resourceName = Uri.EscapeUriString(tokens[2]);

            return new Tuple<string, string, string, string>(subscriptionId, resourceGroup, resourceName, sku);
        }

        private static CloudTable GetTable(string tableName)
        {                        
            return GetTable(tableName, null);
        }
        
        private static CloudTable GetTable(string tableName, TimeSpan? maximumExecutionTime)
        {
            CloudTableClient tableClient = _storageAccount.CreateCloudTableClient();
            tableClient.DefaultRequestOptions = new TableRequestOptions()
            {
                RetryPolicy = new LinearRetry(TimeSpan.FromSeconds(10), 3),
                MaximumExecutionTime = maximumExecutionTime.HasValue ? maximumExecutionTime : TimeSpan.FromMinutes(1)
            };

            return tableClient.GetTableReference(tableName);
        }

        private static CloudQueue GetQueue(string queueName)
        {
            CloudQueueClient queueClient = _storageAccount.CreateCloudQueueClient();
            queueClient.DefaultRequestOptions = new QueueRequestOptions()
                                                {
                                                    RetryPolicy = new LinearRetry(TimeSpan.FromSeconds(10), 3),
                                                    MaximumExecutionTime = TimeSpan.FromSeconds(30)
                                                };
            return queueClient.GetQueueReference(queueName.ToLowerInvariant());
        }

        private static void UpdateLastProcessedPartition(DateTime partition)
        {
            CloudTable table = GetTable(LAST_PROCESSED_PARTITION_TABLE);

            var entity = new LastProcessedPartitionEntity("a", "a")
            {
                LastPartition = DateTimeToPartition(partition)
            };

            var tableOperation = TableOperation.InsertOrReplace(entity);
            table.Execute(tableOperation); 
        }
    }
}
