﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
namespace Microsoft.Azure.LocationServices.BillingAPI
{
    using System;
    using System.Collections.Concurrent;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.RetryPolicies;
    using Microsoft.WindowsAzure.Storage.Table;    

    public class BillingLogWriter
    {        
        private static BillingLogWriter _instance;
        private static object _sSyncLock = new object();
        
        private readonly ConcurrentQueue<TransactionLogEntityData> _logs;
        private readonly Thread _loggerThread;
        private readonly CloudTable _cloudTable;
        private int _writeErrorCount;

        private BillingLogWriter(CloudTable table)
        {
            _cloudTable = table;
            _logs = new ConcurrentQueue<TransactionLogEntityData>();
            _loggerThread = new Thread(FlushLogs);
            _loggerThread.IsBackground = false;
            _loggerThread.Start();
        }

        public static BillingLogWriter Instance
        {
            get
            {
                if (_instance == null)
                {
                    throw new InvalidOperationException("CreateInstanceAsync has not been called");
                }

                return _instance; 
            }
        }

        public static async Task CreateInstanceAsync(IBootstrapConfiguration bootstrapConfiguration, IConfigurationReader configurationReader)
        {
            if (bootstrapConfiguration == null)
            {
                throw new ArgumentNullException(nameof(bootstrapConfiguration));
            }

            if (configurationReader == null)
            {
                throw new ArgumentNullException(nameof(configurationReader));
            }

            string billingSasToken = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(Globals.BillingSasTokenKey));
            string billingConnectionTemplate = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(Globals.BillingConnectionStringTemplateKey));
            string connectionString = string.Format(billingConnectionTemplate, billingSasToken);
            CloudStorageAccount storageAccount = CloudStorageAccount.Parse(connectionString);
            
            lock (_sSyncLock)
            {
                if (_instance != null)
                {
                    throw new InvalidOperationException("BillingLogWriter already created");
                }

                _instance = new BillingLogWriter(CreateCloudTable(storageAccount));
            }
        }

        public void WriteLog(string resourceId, string method, HttpStatusCode statusCode, string trackingId)
        {
            var log = new TransactionLogEntityData(resourceId, method, (int)statusCode, trackingId);
            _logs.Enqueue(log); 
        }

        public void WriteLog(string resourceId, List<Tuple<string, HttpStatusCode>> methodStatusCodeTuples, string trackingId)
        {
            var log = new TransactionLogEntityData(resourceId, methodStatusCodeTuples.Select(t => new MethodStatus(t.Item1, (int)t.Item2)).ToList(), trackingId);
            _logs.Enqueue(log);
        }

        private static void WriteTransactionsToTable(List<TransactionLogEntity> logs, CloudTable table)
        {
            var operation = new TableBatchOperation();
            logs.ForEach(x => operation.Insert(x));
            table.ExecuteBatch(operation);
        }

        private static CloudTable CreateCloudTable(CloudStorageAccount storageAccount)
        {
            CloudTableClient tableClient = storageAccount.CreateCloudTableClient();
            tableClient.DefaultRequestOptions = new TableRequestOptions()
            {
                RetryPolicy = new LinearRetry(TimeSpan.FromSeconds(10), 12),
                MaximumExecutionTime = TimeSpan.FromSeconds(10)
            };
            return tableClient.GetTableReference(BillingHelper.LOG_TABLE);
        }

        private void FlushLogs()
        {
            while (true)
            {
                try
                {
                    var total = _logs.Count;
                    var processed = 0;
                    var logsToWrite = new List<TransactionLogEntityData>();
                    while (processed < total)
                    {
                        //_logs.TryPeek(out TransactionLogEntityData transaction);
                        // Write in batches, max of 100 and partition key has to be the same. 
                        var partition = BillingAPI.GetTransactionLogPartition();
                        while (processed < total && logsToWrite.Count < 100)
                        {
                            _logs.TryDequeue(out TransactionLogEntityData data);
                            processed++;
                            logsToWrite.Add(data); 
                        }

                        try
                        {
                            var logs = logsToWrite.Select(x => new TransactionLogEntity(partition, Guid.NewGuid().ToString(), x)).ToList(); 
                            WriteTransactionsToTable(logs, _cloudTable);
                            logsToWrite.Clear();
                            _writeErrorCount = 0;
                        }
                        catch (Exception e)
                        {
                            // Failed to write for some reason, push it back into the queue and try again next time
                            logsToWrite.ForEach(x => _logs.Enqueue(x));
                            logsToWrite.Clear(); 
                            _writeErrorCount++;
                            if (_writeErrorCount >= 10)
                            {
                                Instrumentation.Agent.Instance.LogException(new Exception("Failed to write billing transactions 10 times in a row", e), "billing", "writeToTable");
                                _writeErrorCount = 0;
                            }

                            break;
                        }
                    }
                }
                catch (Exception e)
                {
                    Instrumentation.Agent.Instance.LogException(e, "billing", "flushLogs");
                }

                Thread.Sleep(TimeSpan.FromMinutes(1));
            }
        }
    }
}