﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
namespace Microsoft.Azure.LocationServices.BillingAPI
{
    using System;
    using System.Collections.Generic;
    using System.Net;
    using Microsoft.WindowsAzure.Storage.Table;
    using Newtonsoft.Json;

    public class MethodStatus
    {
        public MethodStatus(string methodName, int statusCode)
        {
            this.MethodName = methodName;
            this.StatusCode = statusCode;
        }

        public String MethodName { get; private set; }

        public int StatusCode { get; private set; }
    }

    public class TransactionLogEntity : TableEntity
    {
        public TransactionLogEntity(string partitionId, string key)
        {
            PartitionKey = partitionId;
            RowKey = key;
        }

        public TransactionLogEntity(string partitionId, string key, TransactionLogEntityData data)
        {
            PartitionKey = partitionId;
            RowKey = key;
            ResourceId = data.ResourceId;
            MethodNameStatusCodesSerialized = this.Serialize(data.MethodStatuslist);
            TrackingId = data.TrackingId; 
        }

        public TransactionLogEntity()
        {
        }

        public string ResourceId { get; set; }

        public string Method { get; set; }

        public int StatusCode { get; set; }

        public string TrackingId { get; set; }

        public string MethodNameStatusCodesSerialized { get; set; }

        private List<MethodStatus> MethodStatuslist { get; set; }

        public List<MethodStatus> GetMethodStatusList()
        {
            if (this.MethodStatuslist == null)
            {
                if (this.Method == null)
                {
                    if (this.MethodNameStatusCodesSerialized == null)
                    {
                        throw new MissingFieldException("Both Method and MethodNameStatusCodesSerialized are missing");
                    }

                    try
                    {
                        this.MethodStatuslist = this.Deserialize(this.MethodNameStatusCodesSerialized);
                    }
                    catch (Exception e)
                    {
                        throw new ArgumentException($"Unable to deserialize MethodStatus: {this.MethodNameStatusCodesSerialized}", e);
                    }
                }
                else
                {
                    if (this.MethodNameStatusCodesSerialized != null)
                    {
                        throw new MissingFieldException("Both Method and MethodNameStatusCodesSerialized are available. Expecting one of them.");
                    }

                    this.MethodStatuslist = new List<MethodStatus>() { new MethodStatus(this.Method, this.StatusCode) };
                }
            }

            return this.MethodStatuslist;
        }

        private List<MethodStatus> Deserialize(string serializedString)
        {
            return JsonConvert.DeserializeObject<List<MethodStatus>>(serializedString);
        }

        private string Serialize(List<MethodStatus> statuss)
        {
            return JsonConvert.SerializeObject(statuss);
        }
    }

    public class TransactionLogEntityData
    {
        public TransactionLogEntityData(string resourceId, string method, int statusCode, string trackingId)
            : this(resourceId, new List<MethodStatus>() { new MethodStatus(method, statusCode) }, trackingId)
        {
        }

        public TransactionLogEntityData(string resourceId, List<MethodStatus> methodStatuslist, string trackingId)
        {
            ResourceId = resourceId;
            MethodStatuslist = methodStatuslist;
            TrackingId = trackingId;
        }

        public string ResourceId { get; set; }

        public List<MethodStatus> MethodStatuslist { get; set; }

        public string TrackingId { get; set; }
    }

    public class TransactionCacheEntity : TableEntity
    {
        public TransactionCacheEntity(string partitionId, string key)
        {
            PartitionKey = partitionId;
            RowKey = key;
        }

        public TransactionCacheEntity()
        {
        }

        public long Count { get; set; }

        public long TotalCount { get; set; }

        public string TrackingId { get; set; }
    }

    public class UsageTableEntity : TableEntity
    {
        public DateTime EventDateTime { get; set; }

        public Guid EventId { get; set; }

        public string Location { get; set; }

        public string MeterId { get; set; }

        public Double Quantity { get; set; }

        public string ResourceUri { get; set; }

        public Guid LocatSubscriptionIdion { get; set; }

        public Guid SubscriptionId { get; set; }

        public string Tags { get; set; }
    }

    public class LastProcessedPartitionEntity : TableEntity
    {
        public LastProcessedPartitionEntity(string partitionId, string key)
        {
            PartitionKey = partitionId;
            RowKey = key;
        }

        public LastProcessedPartitionEntity()
        {
        }

        public string LastPartition { get; set; }
    }

    public class ProcessedPartitionEntity : TableEntity
    {
        public ProcessedPartitionEntity(string partitionId, string key)
        {
            PartitionKey = partitionId;
            RowKey = key;
        }

        public ProcessedPartitionEntity()
        {
        }

        public string Duration { get; set; }

        public long TransactionCount { get; set; }
    }
}
