﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.BillingAPI.Utility
{
    using System;
    using System.Collections.Generic;
    using System.Globalization;

    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;

    /// <summary>
    /// Represents the configuration necessary to operate in an environment.
    /// </summary>
    /// <remarks>
    /// This class should only contain members that differ from environment to environment. When something doesn't 
    /// vary, it should not be present in this class, and instead should be a compile-time constant. If a property
    /// is no longer used, or no longer varies, it should be removed.
    /// </remarks>
    public abstract class BillingHostSettings
    {
        private readonly IReadOnlyDictionary<string, string> methodMeterMap = new Dictionary<string, string>(StringComparer.OrdinalIgnoreCase)
        {
            { "MapTile", "Meter1" },
            { "TrafficTile", "Meter1" },
            { "ImageryTile", "Meter5" },
            { "HybridTile", "Meter6" },
            { "Rendering", "Meter2" },
            { "Search", "Meter2" },
            { "Routing", "Meter2" },
            { "TimeZones", "Meter3" },
            { "Traffic", "Meter2" },
            { "GeoLocation", "Meter4" }
        };

        private readonly IReadOnlyDictionary<string, IReadOnlyDictionary<string, string>> skuMeterIdMap = new Dictionary<string, IReadOnlyDictionary<string, string>>(StringComparer.OrdinalIgnoreCase)
        {
            {
                SkuName.S0.ToString(),
                new Dictionary<string, string>(StringComparer.OrdinalIgnoreCase)
                {
                    { "Meter1", "5bb52c6b-cb42-4960-9de9-1567f99e966b" },   //tiles
                    { "Meter2", "098e9fa8-bb99-409c-a60a-2d7e88d7a5b5" },   //services
                    { "Meter3", "00fd30fe-b6c0-43fb-b5a7-2121c3b824fa" },   //time zones
                    { "Meter4", "ed7b0e79-13a5-4909-8326-9ed41a27d35a" },   //ipToCountry
                    { "Meter5", "ed1546b4-8dc3-4303-99b3-7f73d51361ff" },   //aerialImagery
                    { "Meter6", "ce4fbd1a-4097-499a-96c1-fc8c6ec3b555" },   //hybridAerialImagery
                }
            }
        };

        private readonly IReadOnlyDictionary<string, int> methodTransactionCountMap = new Dictionary<string, int>(StringComparer.OrdinalIgnoreCase)
        {
            { "MapTile", 15 },
            { "TrafficTile", 15 },
            { "ImageryTile", 15 },
            { "HybridTile", 15 },
            { "Rendering", 1 },
            { "Search", 1 },
            { "Routing", 1 },
            { "TimeZones", 1 },
            { "Traffic", 1 },
            { "GeoLocation", 1 }
        };

        private readonly IReadOnlyDictionary<string, string> serviceApiToMethodMap = new Dictionary<string, string>(StringComparer.OrdinalIgnoreCase)
        {
            { "GetSearchPolygon", "Search" },
            { "GetSearchFuzzy", "Search" },
            { "GetSearchPOI", "Search" },
            { "GetSearchPOICategory", "Search" },
            { "GetAddress", "Search" },
            { "GetAddressStructured", "Search" },
            { "GetAddressReverse", "Search" },
            { "GetAddressCrossStreet", "Search" },
            { "GetMapTileStatic", "Rendering" },
            { "GetMapTile", "MapTile" },
            { "GetMapImagery", "ImageryTile" },
            { "GetHybridTile", "HybridTile" },
            { "GetMapStatic", "Rendering" },
            { "GetCopyrightWorld", "Rendering" },
            { "GetCopyrightCaption", "Rendering" },
            { "GetCopyrightBounding", "Rendering" },
            { "GetCopyrightTile", "Rendering" },
            { "GetTimezoneIANAVersion", "TimeZones" },
            { "GetTimezoneIANA", "TimeZones" },
            { "GetTimezoneWindows", "TimeZones" },
            { "GetTimezoneWindowsToIANA", "TimeZones" },
            { "GetTimezoneByID", "TimeZones" },
            { "GetTimezoneCoordinates", "TimeZones" },
            { "GetRouteCalculate", "Routing" },
            { "PostRouteCalculate", "Routing" },
            { "PostMatrixRouteCalculate", "Routing" },
            { "GetRouteReachableRange", "Routing" },
            { "GetTrafficFlowTile", "TrafficTile" },
            { "GetTrafficIncidentTile", "TrafficTile" },
            { "GetTrafficFlowSegment", "Traffic" },
            { "GetTrafficIncidentViewport", "Traffic" },
            { "GetTrafficIncidentDetails", "Traffic" },
            { "GetIpLocation", "GeoLocation" },
            { "GetSearchAlongRoute", "Search" },
            { "GetSearchGeometry", "Search" },
            { "GetSearchNearby", "Search" },
            { "GetSearchRouted", "Search" },
            { "PostSearchAlongRoute", "Search" },
            { "PostSearchGeometry", "Search" },
            { "PostSearchGeometryFilter", "Search" },
            { "PostSearchRoutedFilter", "Search" }
        };

        /// <summary>
        /// Gets the Instrumentation Parameters
        /// </summary>
        public abstract InitParams MdmParameters { get; }

        public IReadOnlyDictionary<string, string> MethodMeterMap => this.methodMeterMap;

        public IReadOnlyDictionary<string, IReadOnlyDictionary<string, string>> SkuMeterIdMap => this.skuMeterIdMap;

        public IReadOnlyDictionary<string, int> MethodTransactionCountMap => this.methodTransactionCountMap;

        public IReadOnlyDictionary<string, string> ServiceApiToMethodMap => this.serviceApiToMethodMap;

        /// <summary>
        /// Gets the current region.
        /// </summary>
        protected static string CurrentRegion
        {
            get
            {
                string currentRegion = Environment.GetEnvironmentVariable("Region");
                return currentRegion ?? "DEFAULT";
            }
        }

        /// <summary>
        /// Gets the current environment.
        /// </summary>
        /// <exception cref="ArgumentException">
        /// Throws if the current environment is not specified
        /// </exception>
        private static string CurrentEnvironment
        {
            get
            {
                string currentEnvironment = Environment.GetEnvironmentVariable("Environment");

                if (string.IsNullOrEmpty(currentEnvironment))
                {
                    throw new ArgumentException("Current Environment could not be determined");
                }

                return currentEnvironment;
            }
        }

        /// <summary>
        /// Create an environment definition from the environment environment variable.
        /// </summary>
        /// <returns>The created environment definition.</returns>
        public static BillingHostSettings CreateFromEnvironment()
        {
            switch (CurrentEnvironment.ToUpperInvariant())
            {
                case LogicalEnvironments.Development:
                    return new BillingHostDevSettings();
                case LogicalEnvironments.Test:
                    return new BillingHostTestSettings();
                case LogicalEnvironments.Staging:
                    return new BillingHostStagingSettings();
                case LogicalEnvironments.Production:
                    return new BillingHostProdSettings();
                default:
                    throw new ArgumentException(
                        string.Format(
                            CultureInfo.InvariantCulture,
                            "Environment {0} does not have a corresponding Environment definition defined",
                            CurrentEnvironment));
            }
        }
    }
}