﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
namespace Microsoft.Azure.LocationServices.BillingAPI.Utility
{
    using System;
    using System.Collections.Generic;
    using System.Fabric;
    using System.Fabric.Description;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Reflection;
    using System.Security;
    using Validation;

    public class ConfigurationReader
    {
        public const string Config = "Config";

        private static readonly object _syncLock = new object();
        private static ConfigurationReader _instance;

        private ICodePackageActivationContext codePackageActivationContext;

        private ConfigurationReader()
        {
        }

        public static ConfigurationReader Instance
        {
            get
            {
                if (_instance == null)
                {
                    lock (_syncLock)
                    {
                        if (_instance == null)
                        {
                            _instance = new ConfigurationReader();
                        }
                    }
                }

                return _instance;
            }
        }

        /// <summary>
        /// Reads a tab dilimited document with two columns and returns a dictionary with column 1 as key and column 2 as the value. 
        /// </summary>
        public static void ReadTwoColumnMap(StreamReader sr, out Dictionary<string, string> result)
        {
            result = new Dictionary<string, string>();

            // Read the stream to a string, and write the string to the console.
            while (!sr.EndOfStream)
            {
                var line = sr.ReadLine();
                var arLine = line.Split('\t');
                if (arLine.Count() != 2)
                {
                    continue;
                }

                result.Add(arLine[0], arLine[1]);
            }
        }

        /// <summary>
        /// Reads a tab dilimited document with two columns and returns a dictionary with column 1 as key and column 2 as the value. 
        /// </summary>
        public static void ReadTwoColumnMap(StreamReader sr, out Dictionary<string, int> result)
        {
            result = new Dictionary<string, int>();

            // Read the stream to a string, and write the string to the console.
            while (!sr.EndOfStream)
            {
                var line = sr.ReadLine();
                var arLine = line.Split('\t');
                if (arLine.Count() != 2)
                {
                    continue;
                }

                result.Add(arLine[0], Int32.Parse(arLine[1]));
            }
        }

        public void Initialize(ServiceContext serviceContext)
        {
            Requires.NotNull(serviceContext, nameof(serviceContext));

            Validation.Verify.Operation(this.codePackageActivationContext == null, "ConfigurationReader already initialized");

            StatefulServiceContext initParams = serviceContext as StatefulServiceContext;
            if (initParams != null)
            {
                codePackageActivationContext = initParams.CodePackageActivationContext;
            }
            else
            {
                // We know this cant be null if we got here
                codePackageActivationContext = serviceContext.CodePackageActivationContext;
            }
        }

        /// <summary>
        ///     A simple overload that provides default values for everything except the parameter name.
        ///     Recommended for standard use case.
        /// </summary>
        public T GetConfigValue<T>(string parameterName)
        {
            Validation.Verify.Operation(this.codePackageActivationContext != null, "ConfigurationReader not initialized");

            Requires.NotNullOrEmpty(parameterName, nameof(parameterName));
            return GetConfigValue(Config, Config, parameterName, default(T));
        }

        public string GetDataFilePath(string dataFolder, string dataFileName)
        {
            Validation.Verify.Operation(this.codePackageActivationContext != null, "ConfigurationReader not initialized");

            Requires.NotNullOrEmpty(dataFolder, nameof(dataFolder));
            Requires.NotNullOrEmpty(dataFileName, nameof(dataFileName));

            DataPackage dataPackage = codePackageActivationContext.GetDataPackageObject(dataFolder);
            if (dataPackage?.Path != null)
            {
                return dataPackage.Path + @"\" + dataFileName;
            }

            return null;
        }

        public T ReadAndParseConfig<T>(string key)
        {
            Validation.Verify.Operation(this.codePackageActivationContext != null, "ConfigurationReader not initialized");

            Requires.NotNullOrEmpty(key, nameof(key));
            return GetConfigValue<T>(key);
        }

        private static object Parse(string value, Type type)
        {
            Requires.NotNull(type, nameof(type));

            if (string.IsNullOrEmpty(value))
            {
                return null;
            }

            try
            {
                if (type == typeof(string))
                {
                    return value;
                }

                if (type == typeof(bool))
                {
                    return bool.Parse(value);
                }

                if (type == typeof(int))
                {
                    return int.Parse(value);
                }

                if (type == typeof(double))
                {
                    return double.Parse(value);
                }

                if (type == typeof(Uri))
                {
                    return new Uri(value);
                }

                if (type.IsEnum)
                {
                    return Enum.Parse(type, value, true);
                }

                if (type.IsGenericType && (type.GetGenericTypeDefinition() == typeof(Nullable<>)))
                {
                    return value.Length == 0 ? null : Parse(value, type.GetGenericArguments()[0]);
                }

                if (type.IsArray && (type.GetElementType() == typeof(byte)))
                {
                    return Convert.FromBase64String(value);
                }

                MethodInfo parseMethod = type.GetMethod("Parse", new[] { typeof(string) });

                return parseMethod.Invoke(null, new object[] { value });
            }
            catch (ArgumentException ex)
            {
                Instrumentation.Agent.Instance.LogException(ex, "billing", "configuration");
                return null;
            }
            catch (FormatException ex)
            {
                Instrumentation.Agent.Instance.LogException(ex, "billing", "configuration");
                return null;
            }
            catch (TargetException ex)
            {
                Instrumentation.Agent.Instance.LogException(ex, "billing", "configuration");
                return null;
            }
        }

        /// <summary>
        ///     Parse the string and return a typed settingValue.
        /// </summary>
        private static bool TryParse<TValue>(string valueString, out TValue value)
        {
            Requires.NotNullOrEmpty(valueString, nameof(valueString));

            object result = Parse(valueString, typeof(TValue));
            if (result == null)
            {
                value = default(TValue);
                return false;
            }

            value = (TValue)result;
            return true;
        }

        private T GetConfigValue<T>(string configPackageName,
                                    string sectionName,
                                    string parameterName,
                                    T defaultValue,
                                    bool isOptional = false)
        {
            T value = defaultValue;

            ConfigurationPackage configPackageDesc =
                codePackageActivationContext.GetConfigurationPackageObject(configPackageName);
            if (configPackageDesc != null)
            {
                ConfigurationSettings configSettings = configPackageDesc.Settings;
                if ((null != configSettings) && (null != configSettings.Sections) &&
                    configSettings.Sections.Contains(sectionName))
                {
                    ConfigurationSection section = configSettings.Sections[sectionName];
                    if ((null != section.Parameters) && section.Parameters.Contains(parameterName))
                    {
                        ConfigurationProperty property = section.Parameters[parameterName];
                        string propertyValue = null;

                        // TODO add support for returning the secure string if the T is secure string
                        if (property.IsEncrypted)
                        {
                            try
                            {
                                SecureString decryptedValue = property.DecryptValue();

                                // Hack to get the string value for now
                                propertyValue = new NetworkCredential(string.Empty, decryptedValue).Password;
                            }
                            catch (Exception ex)
                            {
                                Instrumentation.Agent.Instance.LogException(ex, "billing", "configuration", $"{parameterName}");
                                if (!isOptional)
                                {
                                    throw;
                                }

                                return defaultValue;
                            }
                        }
                        else
                        {
                            propertyValue = property.Value;
                        }

                        if (!TryParse(propertyValue, out value))
                        {
                            value = defaultValue;
                        }
                    }
                }
            }

            return value;
        }
    }
}
