<#
 .SYNOPSIS
    Module of functions used by the CI cluster management automation scripts

 .DESCRIPTION
    Module of functions used by the CI cluster management automation scripts
#>

function New-CIResourceGroup {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ResourceGroupName,
    
        [Parameter(Mandatory = $true, Position = 2)]
        [string] $Location
    )

    Write-Host "Creating Resource Group" $ResourceGroupName
    $resourceGroup = Get-AzureRmResourceGroup -Name $ResourceGroupName -ErrorAction:SilentlyContinue 

    if (-not $resourceGroup) {
        New-AzureRmResourceGroup -Name $ResourceGroupName -Location $Location -Force
    }
    else {
        Write-Host "The Resource Group" $ResourceGroupName "Already Exists. Nothing to Create!"
    }
}

function New-CIApiManagementInstance {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ResourceGroupName,
    
        [Parameter(Mandatory = $true, Position = 2)]
        [string] $Location,
    
        [Parameter(Mandatory = $true, Position = 3)]
        [string] $SubscriptionId
    )

    $apimServiceName = Get-ApimInstanceName -InstancePrefix $ResourceGroupName

    # First check if an APIM instance already exists, bail out if it does.
    $existingApimInstance = Get-AzureRmApiManagement -ResourceGroupName $ResourceGroupName -Name $apimServiceName -ErrorAction:SilentlyContinue 

    if ($existingApimInstance) {
        Write-Host "The APIM instance" $apimServiceName "Already Exists. Nothing to Create!"
        return
    }

    $DeployAPIMScript = $PSScriptRoot + "\..\Provision\APIM\DeployAPIM.ps1"

    $deploymentName = 'new-deployment-' + (New-Guid).ToString()
    $APIMParametersTemplateJson = Get-Content -Path ($PSScriptRoot + "\..\Provision\NEW_CREATE\parameters\APIMParameters.json") | ConvertFrom-Json
    $APIMParametersTemplateJson.parameters.service_atlas_name.value = $apimServiceName
    $APIMParametersTemplateJson.parameters.adminEmail.value = 'shsi@microsoft.com'

    $jsonData = $APIMParametersTemplateJson | ConvertTo-Json -Depth 100

    $tempFile = New-TemporaryFile
    $jsonData | Set-Content -Path $tempFile

    Write-Host "Creating APIM instance..." 
    & $DeployAPIMScript -SubscriptionId $SubscriptionId -resourceGroupName $ResourceGroupName -resourceGroupLocation $Location -deploymentName $deploymentName -parametersFilePath $tempFile -isAPIMExisting $false -deployCert $false -ErrorAction $ErrorActionPreference
    Write-Host "APIM creation succeeded..." -ForegroundColor Green

	# Setting up event hub logger
	Write-Host "Setting up APIM event hub logger..." 
	$loggerId = Get-ApimEventHubLoggerId -ApimServiceName $apimServiceName
	$eventHubName = Get-ApimEventHubName -ClusterName $ResourceGroupName
	$secretKey = Get-EventHubConnectionStringKeyInVault -Type 'sending'
	$persistentVaultName = Get-PersistentKeyVaultName -ClusterName $ResourceGroupName
	$connectionString = Get-AzureKeyVaultSecret -VaultName $persistentVaultName -Name $secretKey

	$apimContext = New-AzureRmApiManagementContext -ResourceGroupName $ResourceGroupName -ServiceName $apimServiceName
	New-AzureRmApiManagementLogger -Context $apimContext -LoggerId $loggerId -Name $eventHubName -ConnectionString $connectionString.SecretValueText
	Write-Host "Event hub logger setup for APIM" -ForegroundColor Green
}

function New-CIApimEventHub {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ResourceGroupName,
    
        [Parameter(Mandatory = $true, Position = 2)]
        [string] $Location
    )
	
	$apimServiceName = Get-ApimInstanceName -InstancePrefix $ResourceGroupName
	$eventHubNamespaceName = Get-ApimEventHubNamespaceName -ClusterName $ResourceGroupName
	$eventHubName = Get-ApimEventHubName -ClusterName $ResourceGroupName

	$existingEventHubNamespace = Get-AzureRmEventHubNamespace -ResourceGroupName $ResourceGroupName -NamespaceName $eventHubNamespaceName -ErrorAction:SilentlyContinue 
	if($existingEventHubNamespace){
		Write-Host "Apim event hub namespace " $eventHubNamespaceName "already exists!" -ForegroundColor Green
	}
	else{
		New-AzureRmEventHubNamespace -ResourceGroupName $ResourceGroupName -NamespaceName $eventHubNamespaceName -Location $Location
		Write-Host "Apim event hub namespace " $eventHubNamespaceName " created!" -ForegroundColor Green
	}

	$existingEventHub = Get-AzureRmEventHub -ResourceGroupName $ResourceGroupName -Namespace $eventHubNamespaceName -Name $eventHubName -ErrorAction:SilentlyContinue 
	if($existingEventHub){
		Write-Host "Apim event hub " $eventHubName "already exists!" -ForegroundColor Green
	}
	else{
		Write-Host "Creating Apim event hub " $eventHubName 
		New-AzureRmEventHub -ResourceGroupName $ResourceGroupName -Namespace $eventHubNamespaceName -Name $eventHubName -MessageRetentionInDays 1 -PartitionCount 2
		Write-Host "Apim event hub " $eventHubName " created!" -ForegroundColor Green
	}

	# Create Sending and Receiving keys
	New-AzureRmEventHubAuthorizationRule -ResourceGroupName $ResourceGroupName -Namespace $eventHubNamespaceName -EventHub $eventHubName -AuthorizationRuleName "Sending" -Rights @("Send")
	New-AzureRmEventHubAuthorizationRule -ResourceGroupName $ResourceGroupName -Namespace $eventHubNamespaceName -EventHub $eventHubName -AuthorizationRuleName "Receiving" -Rights @("Listen")
	
	$persistentVaultName = Get-PersistentKeyVaultName -ClusterName $ResourceGroupName
	$sendingKey = Get-AzureRmEventHubKey -ResourceGroupName $ResourceGroupName -NamespaceName $eventHubNamespaceName -EventHub $eventHubName -Name "Sending"
    $sendingKeySecure = ConvertTo-SecureString $sendingKey.PrimaryConnectionString -AsPlainText -Force
	$sendingSecretKey = Get-EventHubConnectionStringKeyInVault -Type 'sending'

    Set-AzureKeyVaultSecret -VaultName $persistentVaultName -Name $sendingSecretKey -SecretValue $sendingKeySecure
    Write-Host "Apim event hub sending key added to key vault!" -ForegroundColor Green

	$receivingKey = Get-AzureRmEventHubKey -ResourceGroupName $ResourceGroupName -NamespaceName $eventHubNamespaceName -EventHub $eventHubName -Name "Receiving"
    $receivingKeySecure = ConvertTo-SecureString $receivingKey.PrimaryConnectionString -AsPlainText -Force
	$receivingSecretKey = Get-EventHubConnectionStringKeyInVault -Type 'receiving'

    Set-AzureKeyVaultSecret -VaultName $persistentVaultName -Name $receivingSecretKey -SecretValue $receivingKeySecure
    Write-Host "Apim event hub receiving key added to key vault!" -ForegroundColor Green
}

function Invoke-SyncClusterApimWithCI {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ClusterName
    )

    Add-Type -AssemblyName System.Web

    Set-GitConfig

    # Identify a working folder for cloning the APIM git repositories.
    # On local machine, it's a folder on the D drive.
    # On vsts agents, it's the system's default working directory.
    $BaseDirectoryGit = $env:SYSTEM_DEFAULTWORKINGDIRECTORY

    Write-Host "The working directory is $BaseDirectoryGit"
    if ($BaseDirectoryGit -eq $null) {
        $BaseDirectoryGit = "D:\APIMTest"
    }
    
    # Begin sync of git config.
    $ciResourceGroupName = 'Global'
    $apimInstanceNameCI = Get-ApimInstanceName -InstancePrefix 'c'
    $apimContextCI = New-AzureRmApiManagementContext -ResourceGroupName $ciResourceGroupName -ServiceName $apimInstanceNameCI
    $ciProperties = Get-AzureRmApiManagementProperty -Context $apimContextCI

    # Create a random name for a temporary remote branch for CI to sync it's current state to.
    $temporaryCIApimBranchName = -join ((65..90) + (97..122) | Get-Random -Count 8 | % {[char]$_})

    # Save existing configuration of CI APIM to a temp remote branch before cloning so that we get latest changes.
    Save-AzureRmApiManagementTenantGitConfiguration -Context $apimContextCI -Branch $temporaryCIApimBranchName -PassThru -Force

    $pw = Get-ApimRepositoryToken -ResourceGroupName $ciResourceGroupName -ApimServiceName $apimInstanceNameCI

    $apimGitRepositoryTemplate = "https://apim:{0}@{1}.scm.azure-api.net"

    $gitCloneParamCI = $apimGitRepositoryTemplate -f [System.Web.HttpUtility]::UrlEncode($pw), $apimInstanceNameCI
    $APIMGitPathCI = $BaseDirectoryGit + "\ci-apim"
    New-Item -ItemType Directory -Path $APIMGitPathCI
    Set-Location -Path $APIMGitPathCI
    git clone $gitCloneParamCI 2> $null
    Set-Location -Path ("{0}\{1}.scm.azure-api.net" -f $APIMGitPathCI, $apimInstanceNameCI)
    git checkout -b $temporaryCIApimBranchName "origin/$temporaryCIApimBranchName" 2> $null
    Write-Host "Successfully cloned CI APIM git repo..." -ForegroundColor Green

    $apimInstanceNameCluster = Get-ApimInstanceName -InstancePrefix $ClusterName
    $apimContextCluster = New-AzureRmApiManagementContext -ResourceGroupName $ClusterName -ServiceName $apimInstanceNameCluster

    # Sync all properties/named values first
    foreach ($property in $ciProperties) {
        $existingProperty = Get-AzureRmApiManagementProperty -Context $apimContextCluster -PropertyId $property.PropertyId -ErrorAction:SilentlyContinue 

        if ($existingProperty -eq $null) {
            # Create a new property with the same id, name and value
            if ($property.Secret -eq $true) {
                New-AzureRmApiManagementProperty -Context $apimContextCluster -PropertyId $property.PropertyId -Name $property.Name -Value $property.Value -Tag $property.Tags -Secret
            }
            else {
                New-AzureRmApiManagementProperty -Context $apimContextCluster -PropertyId $property.PropertyId -Name $property.Name -Value $property.Value -Tag $property.Tags
            }
        }
        else {
            # Update existing property
            Set-AzureRmApiManagementProperty -Context $apimContextCluster -PropertyId $property.PropertyId -Name $property.Name -Value $property.Value -Tag $property.Tags -Secret $property.Secret
        }
    }

    Set-Location -Path $BaseDirectoryGit

    # Save existing configuration of cluster to it's repository before cloning so that we get latest changes.
    Save-AzureRmApiManagementTenantGitConfiguration -Context $apimContextCluster -Branch 'master' -PassThru -Force

    $pw = Get-ApimRepositoryToken -ResourceGroupName $ClusterName -ApimServiceName $apimInstanceNameCluster

    $gitCloneParamCluster = $apimGitRepositoryTemplate -f [System.Web.HttpUtility]::UrlEncode($pw), $apimInstanceNameCluster
    $APIMGitPathCluster = "{0}\{1}-apim" -f $BaseDirectoryGit, $ClusterName
    New-Item -ItemType Directory -Path $APIMGitPathCluster

    Set-Location -Path $APIMGitPathCluster
    git clone $gitCloneParamCluster 2> $null
    Write-Host "Successfully cloned cluster APIM git repo into a temp branch..." -ForegroundColor Green

    $sourceLocation = "{0}\{1}.scm.azure-api.net\api-management" -f $APIMGitPathCI, $apimInstanceNameCI
    $destinationLocation = "{0}\{1}.scm.azure-api.net\api-management" -f $APIMGitPathCluster, $apimInstanceNameCluster

    # Delete all directories and files from the destination repo to replace them.
    Remove-Item $destinationLocation -Force -Recurse

    # Copy all files from CI APIM source to Cluster APIM destination
    Write-Host "Copying all files from CI cluster APIM config to cluster APIM config git repo..."
    Copy-Item -Path $sourceLocation -Destination $destinationLocation -Recurse -force
    Write-Host "Copying done..." -ForegroundColor Green

	# Update logging event hub configuration
	Set-ApimEventHubLoggerConfiguration -ClusterName $ClusterName -RepoRoot $destinationLocation

    # Commit and push changes to master for cluster.
    Write-Host "Pushing changes to cluster APIM Git..."
    Set-Location $destinationLocation
    git add * 2> $null
    git commit -a -m "Syncing changes from CI APIM c-atlas to cluster APIM." 2> $null
    git push -f 2> $null
    Write-Host "Pushed changes to Git..." -ForegroundColor Green

    # Publish changes to the cluster APIM from the repository
    Write-Host "Publishing APIM Git config..."
    Publish-AzureRmApiManagementTenantGitConfiguration -Context $apimContextCluster -Branch 'master' -Force
    Write-Host "APIM config published..." -ForegroundColor Green

    # Delete upstream CI APIM temp git branch
    Set-Location -Path ($sourceLocation + '\..\')
    git push origin --delete $temporaryCIApimBranchName 2> $null
    Write-Host "Deleted temporary git APIM remote branch for CI" -ForegroundColor Green

    # Delete local git directories
    Set-Location -Path $BaseDirectoryGit
    Remove-Item $APIMGitPathCI -Force -Recurse
    Remove-Item $APIMGitPathCluster -Force -Recurse
    Write-Host "Cleaned local git clones" -ForegroundColor Green
}

function Get-ApimRepositoryToken {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ResourceGroupName,
    
        [Parameter(Mandatory = $true, Position = 2)]
        [string] $ApimServiceName
    )

    $apimContext = New-AzureRmApiManagementContext -ResourceGroupName $ResourceGroupName -ServiceName $ApimServiceName
    $tenantGitAccess = Get-AzureRmApiManagementTenantGitAccess -Context $apimContext

    $userId = $tenantGitAccess.Id
    
    $expiry = Get-Date ((Get-Date).AddDays(1)) -Format "yyyy-MM-ddTHH:mm:ss.000Z"
    $parameters = @{ "keyType" = "primary"; "expiry" = "$expiry"; }
    
    $pw = Invoke-AzureRmResourceAction  -ResourceGroupName $ResourceGroupName -ResourceType 'Microsoft.ApiManagement/service/users' -Action 'token' -ResourceName "$ApimServiceName/$userId" -ApiVersion "2016-10-10" -Parameters $parameters -Force

    return $pw.value
}

function New-PersistentKeyVault {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ResourceGroupName,

        [Parameter(Mandatory = $true, Position = 2)]
        [string] $ResourceGroupLocation
    )

    $persistentVaultName = Get-PersistentKeyVaultName -ClusterName $ResourceGroupName
	
    $existingKeyVault = Get-AzureRmKeyVault -VaultName $persistentVaultName -ResourceGroupName $ResourceGroupName

    if ($existingKeyVault) {
        Write-Host "The key vault $persistentVaultName already exists in resource group $ResourceGroupName. Nothing to Create!"
    }
    else {
        Write-Host "Creating key vault $persistentVaultName in resource group $ResourceGroupName..."
        New-AzureRmKeyVault -VaultName $persistentVaultName -ResourceGroupName $ResourceGroupName -Location $ResourceGroupLocation

        # Give the service principal used for CI access to the key vault secrets (LBS.Web.App)
        Set-AzureRmKeyVaultAccessPolicy $persistentVaultName -ResourceGroupName $ResourceGroupName -ServicePrincipalName 'c01717a5-32dc-4012-aa09-177212753506' -PermissionsToSecrets get, list, set
        Write-Host "Key vault $persistentVaultName created!" -ForegroundColor Green

		# Give the AzLBSDevSetup app access to the key vault too until we can create a new secret for LBS.Web.App.
        Set-AzureRmKeyVaultAccessPolicy $persistentVaultName -ResourceGroupName $ResourceGroupName -ObjectId '97281a77-e3e6-4ea0-a647-0ebc72608190' -BypassObjectIdValidation -PermissionsToSecrets get, list, set
    }
}

function Set-ApimSubscriptionKeyInVault {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ClusterName
    )

    $apimServiceName = Get-ApimInstanceName -InstancePrefix $ClusterName 
    $apimContext = New-AzureRmApiManagementContext -ResourceGroupName $ClusterName -ServiceName $apimServiceName

    $productId = (Get-AzureRmApiManagementProduct -Context $apimContext | where {$_.Title -match 'Azure LBS All'}).ProductId 

    $subscriptions = Get-AzureRmApiManagementSubscription -Context $apimContext
    $lbsSubscription = $subscriptions | where {$_.ProductId -eq $productId}

    $subscriptionKey = $lbsSubscription.PrimaryKey
    $persistentVaultName = Get-PersistentKeyVaultName -ClusterName $ClusterName

    $subscriptionKeySecure = ConvertTo-SecureString $subscriptionKey -AsPlainText -Force

    Set-AzureKeyVaultSecret -VaultName $persistentVaultName -Name 'ApimSubscriptionKey' -SecretValue $subscriptionKeySecure
    Write-Host "Apim subscription key added to key vault!" -ForegroundColor Green
}

function Set-ApimEventHubLoggerConfiguration {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ClusterName,

		[Parameter(Mandatory = $true, Position = 2)]
        [string] $RepoRoot
    )

    if (-not (Test-Path $RepoRoot))
	{
		throw "The APIM git repo root '$RepoRoot' does not exist."
	}

	# Update the Apim logger name in the global policy.
	$globalPolicy = Get-ChildItem -Path $RepoRoot -Include global.xml -File -Recurse
	if($globalPolicy -eq $null){
		throw "Global policy document not found."
	}

	$ciApimServiceName = Get-ApimInstanceName -InstancePrefix 'c'
	$ciLoggerId = Get-ApimEventHubLoggerId -ApimServiceName $ciApimServiceName
	
	$apimServiceName = Get-ApimInstanceName -InstancePrefix $ClusterName
	$loggerId = Get-ApimEventHubLoggerId -ApimServiceName $apimServiceName
	
	$xml = Get-Content $globalPolicy.FullName
	$xml | % { $_.Replace($ciLoggerId, $loggerId) } | Set-Content $globalPolicy.FullName

	$AzureEventHubDirConst = "AzureEventHub_"
	# Update folder and files under /logger
	$loggersDir = Get-ChildItem -Path $RepoRoot -Include loggers -Directory -Recurse
	$eventHubDir = Get-ChildItem -Path $loggersDir -Include $AzureEventHubDirConst* -Directory -Recurse

	# Assuming there's only one such directory in there, grab the first one.
	$eventHubDirItem = $eventHubDir[0]
	$newDirectoryName = ($AzureEventHubDirConst + $loggerId)
	Rename-Item -Path $eventHubDirItem.FullName -newName $newDirectoryName

	# Process files under the directory.
	$apimEventHubName = Get-ApimEventHubName -ClusterName $ClusterName
	$ciApimEventHubName = Get-ApimEventHubName -ClusterName 'c'

	$newDirectoryPath = Join-Path $loggersDir $newDirectoryName
	$configurationFilePath = Join-Path $newDirectoryPath 'configuration.json'
	$configurationFileContent = Get-Content -Path $configurationFilePath

	$configurationFileContent = $configurationFileContent | % { $_.Replace($ciLoggerId, $loggerId) }
	$configurationFileContent = $configurationFileContent | % { $_.Replace($ciApimEventHubName, $apimEventHubName) }
	$configurationFileContent | Set-Content $configurationFilePath

	$secretKey = Get-EventHubConnectionStringKeyInVault -Type 'sending'
	$persistentVaultName = Get-PersistentKeyVaultName -ClusterName $ClusterName
	$connectionString = Get-AzureKeyVaultSecret -VaultName $persistentVaultName -Name $secretKey

	$configurationJson = Get-Content -Path $configurationFilePath | ConvertFrom-Json
	$configurationJson.credentials.connectionString = $connectionString.SecretValueText

	$jsonData = $configurationJson | ConvertTo-Json -Depth 100
    $jsonData | Set-Content -Path $configurationFilePath

    Write-Host "Updated event hub logger configuration" -ForegroundColor Green
}

function Get-ApimInstanceName {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $InstancePrefix
    )

    $name = $InstancePrefix + "-atlas";
    return $name.ToLower();
}

function Get-ApimEventHubNamespaceName {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ClusterName
    )

    $name = $ClusterName + "-apim-hub";
    return $name.ToLower();
}

function Get-ApimEventHubName {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ClusterName
    )

    $name = $ClusterName + "-hubs-log";
    return $name.ToLower();
}

function Get-ApimEventHubLoggerId {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ApimServiceName
    )

    $name = $ApimServiceName + "-log";
    return $name.ToLower();
}

function Get-PersistentKeyVaultName {
    param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $ClusterName
    )

    $name = $ClusterName + "-persist";
    return $name.ToLower();
}

function Get-EventHubConnectionStringKeyInVault {
	param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $Type
    )

	$keyName = "";
	switch ( $Type )
    {
        'receiving' { $keyName = 'ApimEventHubReceivingKey'  }
        'sending'   { $keyName = 'ApimEventHubSendingKey' }
		default     { throw "Unsupported key type." }
    }

	return $keyName
}

function Set-GitConfig {
    git config --global user.email "lbswebapp@vsts.com" 2> $null
    git config --global user.name "LBS.Web.App" 2> $null
}