﻿<#
    Sets up required dev resources and caches config values
#>

function GenerateCertificateName
{
    Param ([String] $sfCluster, [String] $certType)
    $certificateName = $sfCluster + "-" + $certType + "-" + [System.Environment]::UserName.ToLower();
    return $certificateName
}

<# --------------- Init & Imports -------------------- #>
$ErrorActionPreference = "Stop"

$modulePath = $PSScriptRoot + ".\..\Utils\lbsSetupSupport.psd1"
Import-Module -Name $modulePath -Force

$isAdmin = Get-IsRunningAsAdmin
if ($isAdmin -eq $false) { throw "Devsetup requires powershell to be running as admin" }

<# --------------- Set params -------------------- #>
$devsetupVersion = Get-lbsEnvironmentDevelopmentVersion # NOTE - bump it up whenever rp or svc cloud.config params change
$sourceEnv = 't'
$region = Get-lbsEnvironmentDevelopmentRegion # we create your KV here, but must also exist in the "t" environment so we can read from it
$testSubscriptionId = "002509a6-fb5b-4fe0-abbf-b28ea9cc115c" # Azure_LBS_Test
$devSubscriptionId = "21a9967a-e8a9-4656-a70b-96ff1c4d05a0" # IOT_AzureLBS

# set system environment variable to authenticate to key vault for
# the purpose is to bootstrap credetials on dev machines to talk to azure services
# this is because service fabric runs under NetworkService so it is the recommended option.
$externalSubscription = Select-lbsSubscriptionWithQuickLogin -subscriptionId $devSubscriptionId
$AzureServicesAuthConnectionStringSecret = Get-AzureKeyVaultSecret -VaultName "codesign-snk" -Name "DevSetupAzureServicesAuthConnectionString"
[Environment]::SetEnvironmentVariable("AzureServicesAuthConnectionString", $AzureServicesAuthConnectionStringSecret.SecretValueText, [System.EnvironmentVariableTarget]::Machine) 

# prepare configcache folder
$configCacheFolder = $PSScriptRoot + '\..\..\..\.configcache\'
Remove-Item $configCacheFolder -Recurse -ErrorAction Ignore
New-Item -ItemType directory -Path $configCacheFolder

# download the localhost managed key vault certificate and install on local machine. 
$generatedPassword = New-RandomPassword 10
$localHostCertName = 'localhost.pfx'
$localHostCertificateSecret = Get-AzureKeyVaultSecret -VaultName "codesign-snk" -Name 'localhostcertificate'
$pfxFilePath = $configCacheFolder + $localHostCertName
$passwordFile = [system.io.path]::ChangeExtension($pfxFilePath, "pwd")	
$generatedPassword | Out-File $passwordFile -Force
$kvSecretBytes = [System.Convert]::FromBase64String($localHostCertificateSecret.SecretValueText)
$certCollection = New-Object System.Security.Cryptography.X509Certificates.X509Certificate2Collection
$certCollection.Import($kvSecretBytes,$null,[System.Security.Cryptography.X509Certificates.X509KeyStorageFlags]::Exportable)
$protectedCertificateBytes = $certCollection.Export([System.Security.Cryptography.X509Certificates.X509ContentType]::Pkcs12, $generatedPassword)
[System.IO.File]::WriteAllBytes($pfxFilePath, $protectedCertificateBytes)
$pfxPasswordSecure = ConvertTo-SecureString -String $generatedPassword -Force -AsPlainText
Import-PfxCertificate -Exportable -FilePath $pfxFilePath -CertStoreLocation 'Cert:\LocalMachine\My' -Password $pfxPasswordSecure -Verbose
$localCertificate = $certCollection[0]

$devSubscription = Select-lbsSubscriptionWithQuickLogin -subscriptionId $devSubscriptionId

# Get global settings from t- KV
$targetSubcription = Select-lbsSubscriptionWithQuickLogin -subscriptionId $testSubscriptionId 
$globalVaultConfigPairs = Get-lbsKeyVaultSecretKeyValuePairs -env $sourceEnv
$devSubscription = Select-lbsSubscriptionWithQuickLogin -subscriptionId $devSubscriptionId

$sfClusters = Get-lbsServiceFabricClusters
foreach ($sfCluster in $sfClusters) 
{        
    # in-memory version of configcache
    $configCache = @{
        "kv_subscriptionid" = $devSubscriptionId;
    }

    $commonSourceVault = Get-lbsVaultName -env $sourceEnv -sfCluster $sfCluster -region $region
    # install the 'ma' certificate for Jarvis-INT.  Not referenced in config cache
    $certName = Get-lbsCertificateName -env $sourceEnv -sfCluster $sfCluster -region $region -certType 'ma'
    Get-SSLAdminCertificateState -keyVaultName $commonSourceVault -CertificateName $certName
    $configCache.Add("ma_CertName", $certName)    

	$configCache.Add("client_CertName", $localHostCertName)
    $configCache.Add("client_X509FindValue", $localCertificate.Thumbprint)	

	$configCache.Add("secrets_CertName", $localHostCertName)
	$configCache.Add("secrets_X509FindValue", $localCertificate.Thumbprint)
      	
	$globalVaultConfigPairs.Keys | foreach {
         Write-Host ' -' $_ -foregroundcolor Green
		 $curValue = $globalVaultConfigPairs[$_]
		 if(-Not ($configCache.ContainsKey($_))){
			$configCache.Add($_, $curValue)
		 }         
    }

    $fileName = "."+$sfCluster+"-configcache.json"
    Set-ConfigurationData -FileName $fileName -Data $configCache
    Write-Host "Saved Configuration: " $fileName -foregroundcolor Cyan

    $scriptVersion = @{}
    if ($sfCluster -eq 'rp')
    {
        $scriptVersion.Add("devsetupVersion" , $devsetupVersion)
    }
    elseif ($sfCluster -eq 'svc')
    {
        $scriptVersion.Add("devsetupVersion" , $devsetupVersion)
    }
    $fileName = "."+$sfCluster+"-devsetupinfo.json"
    Set-ConfigurationData -FileName $fileName -Data $scriptVersion
    Write-Host "Saved DevSetupInfo: " $fileName -foregroundcolor Cyan
}

cd $PSScriptRoot
& .\GitSetup.ps1

Write-Host 'End of devSetupEV2.ps1'
