<#
 .SYNOPSIS
    One script to deploy Azure location based service APIM

 .DESCRIPTION
    1.Deploy APIM instance to target rescource group
    2.Upload cert to APIM if the cert does not exist in target APIM
    3.Set up the Location based service policy for Global, LBS All Api(product level), Map Control APi(product level), LBS All Api(Api level)

 .PARAMETER subscriptionId
    The subscription id where the template will be deployed.

 .PARAMETER resourceGroupName
    The resource group where the template will be deployed. Can be the name of an existing or a new resource group.

 .PARAMETER resourceGroupLocation
    Optional, a resource group location. If specified, will try to create a new resource group in this location. If not specified, assumes resource group is existing.

 .PARAMETER deploymentName
    The deployment name.

 .PARAMETER templateFilePath
    Optional, path to the template file. Defaults to template.json.

 .PARAMETER parametersFilePath
    Optional, path to the parameters file. Defaults to parameters.json. If file is not found, will prompt for parameter values based on template.

 .Example
 .\DeployAPIM.ps1 `
    -subscriptionId {Target_SUBSCRIPTIONID} `
    -resourceGroupName {Target_ResourceGroup_Name}  `
    -resourceGroupLocation {Target_ResourceGroup_Location} `
    -deploymentName {Deployment_Name} `
#>

param(
 [Parameter(Mandatory=$True)]
 [string]
 $subscriptionId,

 [Parameter(Mandatory=$True)]
 [string]
 $resourceGroupName,

 [string]
 $resourceGroupLocation,

 [Parameter(Mandatory=$True)]
 [string]
 $deploymentName,

 [string]
 $templateFilePath = $PSScriptRoot + "\..\Global\APIMOneTemplate.json",

 [string]
 $parametersFilePath = $PSScriptRoot + "\..\Global\NEW_CREATE\parameters\APIMParameters.json",

 [Byte[]]
 $pfxBytes,

 [string]
 $certificatePassword = "",

 [Parameter(Mandatory=$True)]
 [bool]
 $isAPIMExisting = $false,

 [bool]
 $deployCert = $true
)

<#
.SYNOPSIS
    Registers RPs
#>
Function RegisterRP {
    Param(
        [string]$ResourceProviderNamespace
    )

    Write-Host "Registering resource provider '$ResourceProviderNamespace'";
    Register-AzureRmResourceProvider -ProviderNamespace $ResourceProviderNamespace;
}

#******************************************************************************
# Script body
# Execution begins here
#******************************************************************************
$ErrorActionPreference = "Stop"
$modulePath = $PSScriptRoot + "\..\..\Utils\lbsSetupSupport.psd1"
Import-Module -Name $modulePath -Force

# select subscription
Write-Host "Selecting subscription '$subscriptionId'";
Select-lbsSubscriptionWithQuickLogin -SubscriptionID $subscriptionId;

# Register RPs
$resourceProviders = @("microsoft.apimanagement");
if($resourceProviders.length) {
    Write-Host "Registering resource providers"
    foreach($resourceProvider in $resourceProviders) {
        RegisterRP($resourceProvider);
    }
}

#Create or check for existing resource group
$resourceGroup = Get-AzureRmResourceGroup -Name $resourceGroupName -ErrorAction SilentlyContinue
if(!$resourceGroup)
{
    Write-Host "Resource group '$resourceGroupName' does not exist. To create a new resource group, please enter a location.";
    if(!$resourceGroupLocation) {
        $resourceGroupLocation = Read-Host "resourceGroupLocation";
    }
    Write-Host "Creating resource group '$resourceGroupName' in location '$resourceGroupLocation'";
    New-AzureRmResourceGroup -Name $resourceGroupName -Location $resourceGroupLocation
}
else{
    Write-Host "Using existing resource group '$resourceGroupName'";
}

# Start the deployment
Write-Host "Starting deployment..." 
New-AzureRmResourceGroupDeployment -ResourceGroupName $resourceGroupName -TemplateFile $templateFilePath -TemplateParameterFile $parametersFilePath -ErrorAction $ErrorActionPreference -Name $deploymentName
Write-Host "Deployment succeeded..." -ForegroundColor Green

if(!$isAPIMExisting -or ($pfxBytes -ne $null -and $pfxBytes.count -gt 0))
{
    # Get the deployment parameters
    $params = @{}
    $json = Get-Content -Raw -Path $parametersFilePath;
    ((ConvertFrom-Json $json).psobject.properties | where Name -match "parameters").Value.psobject.properties | Foreach { if($_.Name -ieq "service_atlas_name") { $params[$_.Name] = $_.Value.value } }

    # Compare if cert to deploy is different from the existing one
    $context = New-AzureRmApiManagementContext -ResourceGroupName $resourceGroupName -ServiceName $params['service_atlas_name'] -ErrorAction $ErrorActionPreference

    # Get production ID for Azure LBS All
    $AzureLBSAllproductId = (Get-AzureRmApiManagementProduct -Context $context | where {$_.Title -match 'Azure LBS All'}).ProductId 
}

if(!$isAPIMExisting -and $deployCert)
{
    if(!$pfxBytes)
    {
        # Use test client cert as default new env cert  
        $pfxBytes = Get-CertBytesByEnv -env 't' -pfxPassword $nonEmptyString
        Select-lbsSubscriptionWithQuickLogin -SubscriptionID $subscriptionId;
    }
}

if(($pfxBytes -ne $null -and $pfxBytes.count -gt 0) -and $deployCert)
{  
    $X509Cert = New-Object -TypeName System.Security.Cryptography.X509Certificates.X509Certificate2
    $X509Cert.Import([byte[]]$pfxBytes, $certificatePassword,"Exportable,PersistKeySet")
    $clientCerts = Get-AzureRmApiManagementCertificate -Context $context

    if($clientCerts.Thumbprint -ne $X509Cert.Thumbprint) 
    {
        # Upload the certificate
        $cert = New-AzureRmApiManagementCertificate -Context $context -PfxBytes $pfxBytes -PfxPassword $certificatePassword -ErrorAction $ErrorActionPreference
        Write-Host "Upload certificate succeeded..." -ForegroundColor Green
    }
    else
    {
        Write-Host "The certificate is already existed" -ForegroundColor Green
    }
}