﻿Param
(
    [Parameter(Mandatory=$true)]
    [ValidateSet('p', 's', 't', 'c')]
    [String]$env,

    [bool]$deployExistingApi = $true,

    [String]$apimProduct = "Azure LBS All",

    [String]$swaggerPath = "..\..\..\ALSProxy\ALS\Swagger\Microsoft.Maps\v1\Maps.json"
)

# Import LBS Util
$modulePath = "..\..\Utils\lbsSetupSupport.psd1"
Import-Module -Name $modulePath -Force

# Install Armclient if not existed in current context
if((Get-Command Armclient | ? {$_.Name}) -eq $null)
{
    Set-ExecutionPolicy Bypass -Scope Process -Force; iex ((New-Object System.Net.WebClient).DownloadString('https://chocolatey.org/install.ps1'))
    choco install armclient
}

ARMClient login

# Get target Apim instance Info 
$subscriptionId  = Get-lbsEnvironmentSubscriptionId -env $env
Write-Host "Selecting subscription '$subscriptionId'"
Select-lbsSubscriptionWithQuickLogin -SubscriptionID $subscriptionId
$resourceGroupName =  'Global'
$serviceName = $env + "-atlas"

# Read LBS swagger file
$allApiSwagger = Get-Content $swaggerPath -raw | ConvertFrom-Json
$apiSwagger = Get-Content $swaggerPath -raw | ConvertFrom-Json

# Make parameters in swagger parameters section optional
foreach($params in $apiSwagger.parameters.PSobject.Properties)
{
    if($params.Value.required -ne $null)
    {
        # Parsing error(s): JSON is valid against more than one schema from 'oneOf'
        if($params.Value.name -ne "format" )
        {
            $params.Value.required = $false
        }
    }
    else
    {
        Add-Member -InputObject $params.Value -MemberType NoteProperty -Name "required" -Value $false
    }
}

# $pathObj corresponds to every single api, exp : { Name: /map/copyright/bounding/{format} , .... }
foreach($pathObj in $allApiSwagger.paths.PSobject.Properties)
{
    $apiSwagger.paths = $apiSwagger.paths | Select-Object -Property * -ExcludeProperty *

    # $pathObj.Value corresponds to object which has operation as field , exp: { get : object ,  put: object }
    foreach($operationObj in $pathObj.Value.PSobject.Properties)
    {
        # $operationObj.Value corresponds to object { description : ... , operationId: ... ,...., "x-albs-env": ["c","t","s","p"] } for every single operation
              
        $envProperty = $operationObj.Value.PSobject.Properties | Where-Object {$_.Name -match "x-albs-env"}

        if($envProperty)
        {
            if(!$envProperty.Value.Contains($env))
            {
                # Trim/Filter the api whose x-albs-env doesn't contain target env
                $pathObj.Value.PSobject.Properties.Remove($operationObj.Name)
            }
            else
            {
                $params = $operationObj.Value.PSobject.Properties | Where-Object {$_.Name -eq "parameters"}

                # Make api inline parameters optional
                foreach($param in $params.Value)
                {
                    if($param.required)
                    {
                        $param.required = $false
                    }
                }
            }
        }
        else
        {
            $error = "API:" + $pathObj.Name + ",Operation:" + $operationObj.Value.operationId + " does not have tag 'x-albs-env' which is required."
			throw $error
        }
    }

    # Deploy Api if it is not empty after the environemnt filter
    if(($pathObj.Value.PSobject.Properties | ? {$_.value}) -ne $null)
    {
        # /map/copyright/bounding/{format} -> map/copyright/bounding
        $apimPath = $pathObj.Name.Substring(1, $pathObj.Name.LastIndexOf('/')-1)
        
        # Get Existing Api Info
        $context = New-AzureRmApiManagementContext -ResourceGroupName $resourceGroupName -ServiceName $serviceName -ErrorAction $ErrorActionPreference
        $apimProductId = (Get-AzureRmApiManagementProduct -Context $context | where {$_.Title -match $apimProduct}).ProductId
        $api = Get-AzureRmApiManagementApi -Context $context | ? { $_.Path -ieq  $apimPath }

        if($api)
        {
            # This api already Existed in target apim instance
            $apiId = $api.ApiId
            $apiSwagger.info.description = [string]$api.Description
            $apiSwagger.info.title = [string]$api.Name.Replace(' ', '')

            if($api.ServiceUrl.GetType() -eq @().GetType())
            {
                $apiSwagger.host = [string](([System.Uri]$api.ServiceUrl[0]).Host + ":" + ([System.Uri]$api.ServiceUrl[0]).Port)
            }
            else
            {
                $apiSwagger.host = [string](([System.Uri]$api.ServiceUrl).Host + ":" + ([System.Uri]$api.ServiceUrl).Port)
            }          

            $apiInProduct = Get-AzureRmApiManagementApi -ProductId $apimProductId -Context $context | ? { $_.Path -ieq  $apimPath }
            $isNew = $false
            
            if(!$apiInProduct)
            {
                $isNew = $true
            }
        }
        else
        {
            # This Api is a new Api
            $apiId = $apimPath.Replace('/', '')
            $apiInfo = $apimPath.Replace('/', ' ')
            $apiSwagger.info.description = $apiInfo
            $apiSwagger.info.title = $apiInfo
            $serviceUrl = "{0}-svc-trafficmgr.trafficmanager.net:443" -f $env
            $apiSwagger.host = [string]$serviceUrl
            $isNew = $true
        }
        
        if($deployExistingApi -or !$api)
        {
            # Get the swagger json content for this API
            $apiSwagger.paths | Add-Member -MemberType NoteProperty -Name $pathObj.Name.Substring($pathObj.Name.LastIndexOf('/')) -Value $pathObj.Value -Force
            $jsonContent = $apiSwagger | ConvertTo-Json -Depth 20

            # Prepare put call body content
            $putContents = 'ApimBodyContent.json'
            $template = Get-Content ".\$putContents" -raw | ConvertFrom-Json
            $template.properties.path = $apimPath
            $template.properties.contentValue = [string]$jsonContent
            $template | ConvertTo-Json -Depth 5 | set-content $putContents

            # Upsert/put Api
            $armUrl = "/subscriptions/$subscriptionId/resourceGroups/$resourceGroupName/providers/Microsoft.ApiManagement/service/$serviceName/apis/$apiId" + "?api-version=2017-03-01";
            Write-Host "armclient PUT $armUrl @$putContents"
            armclient PUT $armUrl @$putContents
        }

        if($isNew)
        {
           Add-AzureRmApiManagementApiToProduct -Context $context -ProductId $apimProductId -ApiId $apiId
        }
    }
}