﻿# ##########################################################################################################################################################
# see Guide to Onboarding Key Vault for Microsoft Internal Services
# https://microsoft.sharepoint.com/teams/AzureKeyVault/_layouts/15/WopiFrame.aspx?sourcedoc=%7BC8391802-C772-4466-8635-88B8C4C302C7%7D&file=Guide%20to%20Onboarding%20Key%20Vault%20for%20Microsoft%20Internal%20Services&action=default&IsFolder=1&ListId=%7BBE71BE2C-C7D8-4D34-A997-3D950FBB049D%7D&ListItemId=7
# for details, pfx retrieval, etc.
# ##########################################################################################################################################################

Param
(
    [Parameter(Mandatory=$True)]
    [ValidateSet('p', 's', 't', 'c')]
    [string]$env
)

function Add-CertificateIfNotExists
{
    Param (
        [String] $vaultName,
        [String] $certName,     #Name of the Certificate object 
        [String] $issuerName,
        [System.Collections.Generic.List[String]] $dnsNames
    )

    if (($dnsNames -eq $null) -or ($dnsNames.Count -eq 0)) { 
        Write-Error 'Invalid dnsNames list'
        return;
    }

    $cert = Get-AzureKeyVaultCertificate -VaultName $vaultName -Name $certName
    if ($cert) {
        if ($cert.Enabled) {
            Write-Host "Skipping creation of existing certificate" $certName "in kv" $vaultName -foregroundcolor Yellow
            return;
        } else {
            Write-Host "Removing disabled certificate" $certName "in kv" $vaultName -foregroundcolor Yellow
            Remove-AzureKeyVaultCertificate -VaultName $vaultName -Name $certName -Force
            Start-Sleep -Milliseconds 1000
        }
    }

    $subjectName = "CN=" + $dnsNames[0];
    Write-Host "Adding certificate" $certName "in kv" $vaultName "for" $subjectName -foregroundcolor Cyan
    # create the certificate policy that defines the certificate details and the issuer
	# NOTE: These are compliant renewal rules: https://microsoft.sharepoint.com/teams/AzureKeyVault/_layouts/OneNote.aspx?id=%2Fteams%2FAzureKeyVault%2FDistribution%2FAzure%20Key%20Vault%20Certificates%20-%20Microsoft%20Internal%20Only%20Release%2FGuide%20to%20Onboarding%20Key%20Vault%20for%20Microsoft%20Internal%20Services&wd=target%28Cert%20Renewal.one%7CC189FB90-02C2-4CB2-90F1-383E4E08CD47%2FSet%20KV%20correctly%20for%20Renewal%7CCD4CF8C2-15FF-40A8-8D5D-947A145179EC%2F%29
    if ($dnsNames.Count -gt 1) { 
        $policy = New-AzureKeyVaultCertificatePolicy -SubjectName $subjectName -IssuerName $issuerName -DnsNames $dnsNames -RenewAtNumberOfDaysBeforeExpiry 181 -ValidityInMonths 24
    } else {
        $policy = New-AzureKeyVaultCertificatePolicy -SubjectName $subjectName -IssuerName $issuerName -RenewAtNumberOfDaysBeforeExpiry 181 -ValidityInMonths 24
    }  
    # kick off the enrollment process 
    Add-AzureKeyVaultCertificate -VaultName $vaultName -Name $certName -CertificatePolicy $policy
}

function Add-KeyvaultContactIfNotExists
{
	Param (
        [String] $vaultName,
        [String] $email
    )
	$emailContact = Get-AzureKeyVaultCertificateContact -VaultName $vaultName | where { ($_.Email -eq $email) }
	if($emailContact -eq $null)
	{
		Write-Host "Adding" $email "as an email contact for notifications." -ForegroundColor Green
		Add-AzureKeyVaultCertificateContact -VaultName $vaultName -EmailAddress $email
	}
	else
	{
		Write-Host $email "is already added as an email contact for notifications." -ForegroundColor Cyan
	}
}

import-module $env:LOCALAPPDATA\Microsoft\AzureServiceDeployClient\1.0.34-beta\Microsoft.Azure.Deployment.Common.Contracts.dll
import-module $env:LOCALAPPDATA\Microsoft\AzureServiceDeployClient\1.0.34-beta\Microsoft.Azure.Deployment.Common.dll
import-module $env:LOCALAPPDATA\Microsoft\AzureServiceDeployClient\1.0.34-beta\Microsoft.Azure.Deployment.Express.Client.dll

$modulePath = $PSScriptRoot + "\..\..\Utils\lbsSetupSupport.psd1"
Import-Module -Name $modulePath -Force
$ErrorActionPreference = "Stop"
$subscriptionId = Get-lbsEnvironmentSubscriptionId -env $env;
$subscription = Select-lbsSubscriptionWithQuickLogin -subscriptionId $subscriptionId

$ev2Folder = Get-lbsEnvironmentExpressV2Folder -env $env;
$serviceGroupRoot = $PSScriptRoot + "\..\" + $ev2Folder;

$ServiceDeployPrincipalName = "5744bc69-8a73-45f7-958d-4640d865f04d"
$Scope = "/subscriptions/"+$subscriptionId
try {
    Get-AzureRmRoleAssignment -Scope $Scope -ServicePrincipalName $ServiceDeployPrincipalName -ErrorAction Stop
} catch {
    try {
        New-AzureRmRoleAssignment -Scope $Scope -ServicePrincipalName $ServiceDeployPrincipalName -RoleDefinitionName Contributor  -ErrorAction Stop
    } catch {
        # broken in AzureRM 4.4.1: Get-AzureRmRoleAssignment always fails and can't control whehter to invoke New-AzureRoleAssignment
        if ($_.Exception.Message -ne "The role assignment already exists.") {
            write-Error "Exception Message: $($_.Exception.Message)"
        }
    }
}

$dnsNames = New-Object "System.Collections.Generic.List[String]"

try
{
    # ---------------------------------------------------------------------------------------------------------------------------------------------
    # create Secrets resource group and all its Key Vault resources
    # ---------------------------------------------------------------------------------------------------------------------------------------------
    New-AzureServiceRollout -ServiceGroupRoot $serviceGroupRoot -RolloutSpec "RolloutSpec_KeyVault.json" -RolloutInfra Prod -WaitToComplete

    # ---------------------------------------------------------------------------------------------------------------------------------------------
    # create SSL certificates and add them to their Key Vaults
    # ---------------------------------------------------------------------------------------------------------------------------------------------

    #add SSLAdmin Issuer 
    $issuerName = "SSLAdminBasedIssuer"  #Name of the Issuer object.  It is a variable. 
    $providerName = "SslAdmin"           #providerName HAS to be SslAdmin if you are enrolling from SSLAdmin 
	$certEmailAddress = 'lbsdevs@microsoft.com'

    # ---------------------------------------------------------------------------------------------------------------------------------------------
    # Cluster/Region certs
    # ---------------------------------------------------------------------------------------------------------------------------------------------
    $sfClusters = Get-lbsServiceFabricClusters
    $certTypes = Get-lbsCertificateTypes
	$tags = @{ 'updatedByProcess' = 'provisioning'}	
    foreach ($sfCluster in $sfClusters) 
    {
        $regions = Get-lbsEnvironmentRegions -env $env -sfCluster $sfCluster
    
	    foreach ($region in $regions) 
        {
		    $vaultName = Get-lbsVaultName -env $env -sfCluster $sfCluster -region $region
            Set-AzureKeyVaultCertificateIssuer -VaultName $vaultName -IssuerProvider $providerName -Name $issuerName
			Add-KeyvaultContactIfNotExists -vaultName $vaultName -email $certEmailAddress	
       	    foreach ($certType in $certTypes) 
            {
                $certName = Get-lbsCertificateName -env $env -sfCluster $sfCluster -region $region -certType $certType
                $dnsNames.Clear();

                # first name in the list is used as subject name (i.e. "CN=..."
                $dnsNames.Add(([System.String]::Format("{0}{1}{2}.{3}.cloudapp.azure.com", $env, $sfCluster, $certType, $region)));

                #for the client cert we add to the valid DNS' list the traffic manager URL and all SF clusters
                if ($certType.ToLowerInvariant() -eq 'client' -or $certType.ToLowerInvariant() -eq 'tls') {
    	            $dnsNames.Add(([System.String]::Format("{0}-{1}-trafficmgr.trafficmanager.net", $env, $sfCluster)));

                    # for APIM: include all 'cluster' endpoints to the DNS list
                    foreach ($region2 in $regions) {
                        $dnsNames.Add(([System.String]::Format("{0}{1}{2}.{3}.cloudapp.azure.com", $env, $sfCluster, "cluster", $region2)));
                    }
                }

                Add-CertificateIfNotExists -vaultName $vaultName -certName $certName -issuerName $issuerName -DnsNames $dnsNames
            }
        }
    }
}
catch
{
    Write-Error "an error has occurred: " $_
    Write-Error $_.Exception.ToString()
    Exit
}

# ---------------------------------------------------------------------------------------------------------------------------------------------
# check SSL cert creation status
# ---------------------------------------------------------------------------------------------------------------------------------------------
# Get-AzureKeyVaultCertificateOperation -VaultName $vaultName -Name srphttps-eastus

