﻿Param
(
    [Parameter(Mandatory=$True)]
    [ValidateSet('p', 's', 't', 'c')]
    [string]$env,

    [Parameter(Mandatory=$False)]
    [ValidateSet('all', 'rp', 'svc', 'jobs')]
    [string]$clusterFunction = 'all'
)

function Get-lbsGenevaRunnersVaultName
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'c')]
        [string]$env
    )

    $name = $env + "-genevarunners";
    return $name.ToLower();
}

function Get-ArmCertificateLocation
{
	Param(
		[Parameter(Mandatory=$True)]	
		[ValidateSet('p', 's', 't', 'c')]
		[string]$env
	)
    $armCertUrl = "https://api-dogfood.resources.windows-int.net:24582/metadata/authentication?api-version=2015-01-01"
	if ($env -eq "p") {
		$armCertUrl = "https://management.azure.com:24582/metadata/authentication?api-version=2015-01-01"
	}

	return $armCertUrl
}

$isTestEnvironment = $false
$lowerCaseEnv = $env.ToLowerInvariant()
if ($lowerCaseEnv -eq 't' -or $lowerCaseEnv -eq 'c') {
	$isTestEnvironment = $true
}

$modulePath = $PSScriptRoot + "\..\Utils\lbsSetupSupport.psd1"
Import-Module -Name $modulePath -Force
$ErrorActionPreference = "Stop"
$tags = @{ 'updatedByProcess' = 'provisioning'}

$sfClusters = Get-lbsServiceFabricClusters
$globalKeyVaultName = Get-lbsGlobalVaultName -env $env

$sasStorageConnectionStringTemplate = "BlobEndpoint=https://{0}.blob.core.windows.net;QueueEndpoint=https://{0}.queue.core.windows.net;TableEndpoint=https://{0}.table.core.windows.net;FileEndpoint=https://{0}.file.core.windows.net;SharedAccessSignature={1}"
$globalKeyVaultResourceGroup = 'Vaults'

# -------------------------------------------------------------------------------------------
# global svc & rp info
# -------------------------------------------------------------------------------------------
$billingStorageAcctName = $env + "albsbilling"
$billingStorageAcctNameSecure = ConvertTo-SecureString $billingStorageAcctName -AsPlainText -Force

$billingSasConnectionStringSecretName = New-lbsManagedName -name $billingStorageAcctName.Substring(1)
$billingSasConnectionStringTemplateSecure = ConvertTo-SecureString ($sasStorageConnectionStringTemplate -f $billingStorageAcctName, '{0}') -AsPlainText -Force

# -------------------------------------------------------------------------------------------
# global jobs info
# -------------------------------------------------------------------------------------------
$APIMEventHubStorageAcctName = $env + "apim"
$APIMEventHubSasConnectionStringSecretName = New-lbsManagedName -name $APIMEventHubStorageAcctName.Substring(1)
$APIMEventHubSasConnectionStringTemplateSecure = ConvertTo-SecureString ($sasStorageConnectionStringTemplate -f $APIMEventHubStorageAcctName, '{0}') -AsPlainText -Force

$IPDataStorageAcctName = $env + "albsipdata"
$IPDataSasConnectionStringSecretName = New-lbsManagedName -name $IPDataStorageAcctName.Substring(1)
$IPDataSasConnectionStringTemplateSecure = ConvertTo-SecureString ($sasStorageConnectionStringTemplate -f $IPDataStorageAcctName, '{0}') -AsPlainText -Force

$BatchServiceStorageAcctName = $env + "albsbatchservice"
$BatchServiceSasConnectionStringSecretName = New-lbsManagedName -name $BatchServiceStorageAcctName.Substring(1)
$BatchServiceSasConnectionStringTemplateSecure = ConvertTo-SecureString ($sasStorageConnectionStringTemplate -f $BatchServiceStorageAcctName, '{0}') -AsPlainText -Force

# Get the timezones storage account SAS token from the external data subscription for the staging and production environments
# and the Azure_LBS_Test subscription for the other environments.
$timezonesVaultName = 'timezones'
if($isTestEnvironment -eq $true){
	$timezonesVaultName = 'test-timezones'
}

$TimezonesStorageAccountSasTokenSecretName = 'timezoneshapes-rcelacontainer'
$TimezonesStorageAccountSasToken = Get-AzureKeyVaultSecret -Name $TimezonesStorageAccountSasTokenSecretName -VaultName $timezonesVaultName
$TimezonesStorageAccountSasTokenSecure = ConvertTo-SecureString $TimezonesStorageAccountSasToken.SecretValueText -AsPlainText -Force

# -------------------------------------------------------------------------------------------
# global osm mirror
# -------------------------------------------------------------------------------------------
$OSMMirrorStorageAcctName = $env + "lbsosmmirror"
$OSMMirrorSasConnectionStringSecretName = New-lbsManagedName -name $OSMMirrorStorageAcctName.Substring(1)
$OSMMirrorSasConnectionStringTemplateSecure = ConvertTo-SecureString ($sasStorageConnectionStringTemplate -f $OSMMirrorStorageAcctName, '{0}') -AsPlainText -Force

#--------------------------------------------------------------------------------------------
# set the secret to the global key vault.
# since global key vault will be used during runtime it is not safe to just remove secrets and re-add them	
#--------------------------------------------------------------------------------------------
Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $billingSasConnectionStringSecretName -SecretValue $billingSasConnectionStringTemplateSecure -Tag $tags
Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $APIMEventHubSasConnectionStringSecretName -SecretValue $APIMEventHubSasConnectionStringTemplateSecure -Tag $tags
Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $IPDataSasConnectionStringSecretName -SecretValue $IPDataSasConnectionStringTemplateSecure -Tag $tags
Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $OSMMirrorSasConnectionStringSecretName -SecretValue $OSMMirrorSasConnectionStringTemplateSecure -Tag $tags
Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $BatchServiceSasConnectionStringSecretName -SecretValue $BatchServiceSasConnectionStringTemplateSecure -Tag $tags
Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $TimezonesStorageAccountSasTokenSecretName -SecretValue $TimezonesStorageAccountSasTokenSecure -Tag $tags

# set the event hub connection strings in global keyvault
$eventHubNamespaces = Get-AzureRmEventHubNamespace -ResourceGroupName 'Global'
foreach($eventHubNamespace in $eventHubNamespaces)
{
	$eventHubs = Get-AzureRmEventHub -ResourceGroupName 'Global' -NamespaceName $eventHubNamespace.Name
	foreach($eventHub in $eventHubs)
	{		
		$authorizationRuleName = 'Receiving'
		$eventHubReceivingKey = Get-AzureRmEventHubKey -ResourceGroupName 'Global' -EventHubName $eventHub.Name -AuthorizationRuleName $authorizationRuleName -NamespaceName $eventHubNamespace.Name
		$eventHubSecretName = New-lbsManagedName -name ($eventHubNamespace.Name.Substring(1) + '-' + $eventHubReceivingKey.KeyName)
		$eventHubSasConnectionStringSecure = ConvertTo-SecureString $eventHubReceivingKey.PrimaryConnectionString -AsPlainText -Force
		Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $eventHubSecretName -SecretValue $eventHubSasConnectionStringSecure -Tag $tags
	}	
}

# -------------------------------------------------------------------------------------------
# global svc only info
# -------------------------------------------------------------------------------------------
if(($clusterFunction -eq 'all') -or ($clusterFunction -eq "svc"))
{
    $tzStorageAcctName = $env + "tzdata"
    $tzStorageAcctNameSecure = ConvertTo-SecureString $tzStorageAcctName -AsPlainText -Force

	$tzSasConnectionStringSecretName = New-lbsManagedName -name $tzStorageAcctName.Substring(1)
	$tzSasConnectionStringTemplateSecure = ConvertTo-SecureString ($sasStorageConnectionStringTemplate -f $tzStorageAcctName, '{0}') -AsPlainText -Force

    # Grab keys from the Azure_LBS_External_Data keyvault for Staging and Production subscriptions
	# and the Azure_LBS_Test subscription's key vault for the other subscriptions.
    $vaultName = 'developer-keys'
    if ($lowerCaseEnv -eq 'p') {
         $vaultName = 'production-keys'
    }
	elseif ($lowerCaseEnv -eq 's') {
         $vaultName = 'tt-staging-keys'
    }

    $ttCommonKeySecretName = 'ttCommonKey'
    $secret = Get-AzureKeyVaultSecret -Name $ttCommonKeySecretName -VaultName $vaultName
    $ttCommonKeySecure = ConvertTo-SecureString $secret.SecretValueText -AsPlainText -Force
	
    $ttReverseGeocodeKeySecretName = 'ttReverseGeocodeKey'
    $secret = Get-AzureKeyVaultSecret -Name $ttReverseGeocodeKeySecretName -VaultName $vaultName
    $ttReverseGeocodeKeySecure = ConvertTo-SecureString $secret.SecretValueText -AsPlainText -Force	
	
    $bingMapsKeySecretName = 'bingMapsKey'
    $secret = Get-AzureKeyVaultSecret -Name $bingMapsKeySecretName -VaultName $vaultName
    $bingMapsKeySecure = ConvertTo-SecureString $secret.SecretValueText -AsPlainText -Force	

	#---------------
	# set the tt key secret to the geneva runners key vault
	#---------------
	$genevaRunnerKeyVaultName = Get-lbsGenevaRunnersVaultName -env $env
	Set-AzureKeyVaultSecret -VaultName $genevaRunnerKeyVaultName -SecretValue $ttCommonKeySecure -Name (New-lbsManagedName -name TomTomSearchKey) -Tag $tags

	#--------------------------------------------------------------------------------------------
	# set the secret to the global key vault.
	# since global key vault will be used during runtime it is not safe to just remove secrets and re-add them	
	#--------------------------------------------------------------------------------------------
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $tzSasConnectionStringSecretName -SecretValue $tzSasConnectionStringTemplateSecure	 -Tag $tags
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $bingMapsKeySecretName -SecretValue $bingMapsKeySecure -Tag $tags
    
	# add TomTom keys to global    
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $ttCommonKeySecure -Name (New-lbsManagedName -name TomTomMapKey)  -Tag $tags
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $ttCommonKeySecure -Name (New-lbsManagedName -name TomTomRouteKey) -Tag $tags
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $ttCommonKeySecure -Name (New-lbsManagedName -name TomTomSearchKey) -Tag $tags
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $ttCommonKeySecure -Name (New-lbsManagedName -name TomTomTrafficIncidentKey) -Tag $tags
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $ttCommonKeySecure -Name (New-lbsManagedName -name TomTomTrafficKey) -Tag $tags
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $ttReverseGeocodeKeySecure -Name (New-lbsManagedName -name TomTomReverseGeocodingKey) -Tag $tags
}

# -------------------------------------------------------------------------------------------
# global rp only info
# -------------------------------------------------------------------------------------------
if(($clusterFunction -eq 'all') -or ($clusterFunction -eq "rp"))
{
    $apimName = $env + "-atlas"
    $apimGroupId = "5925c98a5996f60c3cb1083e"
    $apimGroupIdSecure = ConvertTo-SecureString $apimGroupId -AsPlainText -Force

    $apimId = "5829f7e2edee580083030003"
    $apimIdSecure = ConvertTo-SecureString $apimId -AsPlainText -Force

    # grab APIM Key
    $apimCtxt = New-AzureRmApiManagementContext -ResourceGroupName "Global" -ServiceName $apimName
    $managementAccessInformation = Get-AzureRmApiManagementTenantAccess -Context $apimCtxt
    if ((!$managementAccessInformation) -or ($managementAccessInformation.Enabled -ine "true")){
        throw "APIM management key is not found or is not enabled. Make sure you add and enable one."
    }
    $apimKeySecure = ConvertTo-SecureString $managementAccessInformation.PrimaryKey -AsPlainText -Force

    $apimProdId = "5925c92f5996f60c3cb1083d"
    $apimProdIdSecure = ConvertTo-SecureString $apimProdId -AsPlainText -Force

    $apimUri = "https://" + $apimName + ".management.azure-api.net"
    $apimUriSecure = ConvertTo-SecureString $apimUri -AsPlainText -Force

    <#[SuppressMessage("Microsoft.Security", "CS002:SecretInNextLine", Justification="Not a secret: CredScan false positive")]#>
    $rpSubscriptionTblNameSecure = ConvertTo-SecureString "LocationServicesSubscriptions" -AsPlainText -Force
    <#[SuppressMessage("Microsoft.Security", "CS002:SecretInNextLine", Justification="Not a secret: CredScan false positive")]#>
    $rpTblNameSecure = ConvertTo-SecureString "LocationServicesAccounts" -AsPlainText -Force

    $tmpName = $env + "rpaccount"
	$trpSasConnectionStringSecretName = New-lbsManagedName -name  $tmpName.Substring(1)
	$rpSasConnectionStringTemplateSecure = ConvertTo-SecureString ($sasStorageConnectionStringTemplate -f $tmpName, '{0}') -AsPlainText -Force

    #RP backup info
    $rpbackupStorageAccountName = (Get-AzureRmStorageAccount -ResourceGroupName 'Storage' | Where {$_.StorageAccountName.Contains('rpbackup')}).StorageAccountName
    $rpbackupSasConnectionStringTemplateSecure = ConvertTo-SecureString ($sasStorageConnectionStringTemplate -f $rpbackupStorageAccountName, '{0}') -AsPlainText -Force
    $rpbackupSasConnectionStringSecretName = 'mrpbackup'


	#--------------------------------------------------------------------------------------------
	# set the secret to the global key vault.
	# since global key vault will be used during runtime it is not safe to just remove secrets and re-add them	
	#--------------------------------------------------------------------------------------------
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $trpSasConnectionStringSecretName -SecretValue $rpSasConnectionStringTemplateSecure -Tag $tags
    Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $rpbackupSasConnectionStringSecretName -SecretValue $rpbackupSasConnectionStringTemplateSecure -Tag $tags
 
	# -------------------------------------------------------------------------------------------
    # add rp table storage info to global
    # -------------------------------------------------------------------------------------------
    Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $rpSubscriptionTblNameSecure  -Name (New-lbsManagedName -name ResourceProviderSubscriptionTableName) -Tag $tags
    Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $rpTblNameSecure              -Name mResourceProviderTableName -Tag $tags
    
    # add APIM to global    
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $apimGroupIdSecure -Name (New-lbsManagedName -name APIManagementGroupId) -Tag $tags
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $apimIdSecure      -Name (New-lbsManagedName -name APIManagementId) -Tag $tags
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $apimKeySecure     -Name (New-lbsManagedName -name APIManagementKey) -Tag $tags
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $apimProdIdSecure  -Name (New-lbsManagedName -name APIManagementProductId) -Tag $tags
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $apimUriSecure     -Name (New-lbsManagedName -name ApiManagementUri) -Tag $tags
	
    # add ARMClientAuthCertificateLocation    
	$armCertUrl = Get-ArmCertificateLocation $env
	$secure = ConvertTo-SecureString $armCertUrl -AsPlainText -Force
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -SecretValue $secure -Name (New-lbsManagedName -name ARMClientAuthCertificateLocation) -Tag $tags
}																 

# ---------------------------------------------------------------CosmosDB----------------------------
# Add primary RW key to global keyvault
# -------------------------------------------------------------------------------------------
$cosmosDbResourceName = "$env-geofencing"
$geofencingSecrets = Invoke-AzureRmResourceAction -Action listKeys `
    -ResourceType "Microsoft.DocumentDb/databaseAccounts" `
    -ApiVersion "2015-04-08" `
    -ResourceGroupName Global `
    -Name $cosmosDbResourceName `
    -Force
$cosmosDbPrimaryKeySecret = ConvertTo-SecureString -String $geofencingSecrets.primaryMasterKey -AsPlainText -Force
Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name "cosmosDbAuthorizationKey" -SecretValue $cosmosDbPrimaryKeySecret -Tag $tags

if($isTestEnvironment -eq $true)
{
	$mapsFeName = Get-lbsMapsFrontEndCosmosDbName -env $env
	$mapsFeCosmosDbSecrets = Invoke-AzureRmResourceAction -Action listKeys `
    -ResourceType "Microsoft.DocumentDb/databaseAccounts" `
    -ApiVersion "2015-04-08" `
    -ResourceGroupName Global `
    -Name $mapsFeName `
    -Force

	$cosmosDbPrimaryKeySecret = ConvertTo-SecureString -String $mapsFeCosmosDbSecrets.primaryMasterKey -AsPlainText -Force
	Set-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name "mapsFrontEndCosmosDbKey" -SecretValue $cosmosDbPrimaryKeySecret -Tag $tags
}
