﻿# ##########################################################################################################################################################
# This script will take all storage accounts in the resource group and upgrade them to be managed on the global key vault.
# https://docs.microsoft.com/en-us/azure/key-vault/key-vault-ovw-storage-keys
# for details, on the purpose of this script
# ##########################################################################################################################################################

Param
(
    [Parameter(Mandatory=$True)]
    [ValidateSet('p', 's', 't', 'c')]
    [string] $env,
	[Parameter(Mandatory=$False)]    
    [string] $storageResourceGroup = 'Storage',
	[Parameter(Mandatory=$False)]    
    [string] $lbsModulePath,
	[Parameter(Mandatory=$False)]    
    [string] $keyVaultResourceGroup = 'Vaults'
)

if([string]::IsNullOrEmpty($lbsModulePath))
{
	$modulePath = $PSScriptRoot + '\..\Utils\lbsSetupSupport.psd1'
}
else
{
	$modulePath = $lbsModulePath
}
Import-Module -Name $modulePath -Force
$ErrorActionPreference = "Stop"
$subscriptionId = Get-lbsEnvironmentSubscriptionId -env $env
$subscription = Select-lbsSubscriptionWithQuickLogin -subscriptionId $subscriptionId

# get the generic Key Vault Service Principal ID
$KeyVaultServicePrincipalId = (Get-AzureRmADServicePrincipal -ServicePrincipalName cfa8b339-82a2-471a-a3c9-0fc0be7a4093).Id 	
$userAccount = Get-AzureRmADUser -UserPrincipalName $subscription.Account.Id
# details on MSAK: https://microsoft.sharepoint.com/teams/AzureKeyVault/_layouts/OneNote.aspx?id=%2Fteams%2FAzureKeyVault%2FDistribution%2FAzure%20Key%20Vault%20Certificates%20-%20Microsoft%20Internal%20Only%20Release%2FGuide%20to%20Onboarding%20Key%20Vault%20for%20Microsoft%20Internal%20Services&wd=target%28KV%20ManagedStorageKeys.one%7C7423D464-40DD-4E4F-B9E7-2BC89F899423%2FOverview%7CF822BA1F-457D-41BE-B44C-795CF5345AFC%2F%29
# setting to be less than the standard renewal period for AKV's policy once we enable AutoRegenerateKey
# We will just need to make sure we deploy or restart the apps before 364 days.
$compliantValidity = [System.Timespan]::FromDays(364)
$storages = Get-AzureRmStorageAccount -ResourceGroupName $storageResourceGroup

# set the user to have access to storage on key vault
Set-lbsKeyVaultStorageAccess -env $env -resourceGroup $keyVaultResourceGroup -objectId $userAccount.Id -elevate

foreach($storage in $storages)
{
	# we are assigning a new role so Key Vault can manage the storage keys		
	$ManageRoleAssignment = Get-AzureRmRoleAssignment -ObjectId $KeyVaultServicePrincipalId `
	-ResourceGroupName $storageResourceGroup `
	-RoleDefinitionName 'Storage Account Key Operator Service Role' `
	-ResourceName $storage.StorageAccountName `
	-ResourceType 'Microsoft.Storage/storageAccounts' `
	-ErrorAction SilentlyContinue

	if($ManageRoleAssignment -eq $null)
	{
		$ManageRoleAssignment = New-AzureRmRoleAssignment -ObjectId $KeyVaultServicePrincipalId -RoleDefinitionName 'Storage Account Key Operator Service Role' -Scope $storage.Id
	}	
	else 
	{
		Write-Host $storage.StorageAccountName "was skipped for key vault role assignment as it has already been set." -foregroundcolor Cyan
	}
	
	# onboards the managed storage account in an idempotent way
	$existingManangedStorageAccount = Get-lbsManagedStorageAccount -env $env -storageAccountName $storage.StorageAccountName	
	if($existingManangedStorageAccount -eq $null)
	{				
		$ManagedStorageAccountName = Get-lbsManagedStorageAccountName -storageAccountName $storage.StorageAccountName
		$ActiveKeyName = "key2"
		$keyVaultName = Get-lbsStorageAccountKeyVaultName -env $env -storageAccountName $storage.StorageAccountName
		Write-Host "New managed storage will be created with name " $ManagedStorageAccountName " and active key: " $ActiveKeyName
	}
	else
	{
		$ManagedStorageAccountName = $existingManangedStorageAccount.AccountName
		$ActiveKeyName = $existingManangedStorageAccount.ActiveKeyName
		$keyVaultName = $existingManangedStorageAccount.VaultName
	}

	Add-AzureKeyVaultManagedStorageAccount `
			-VaultName $keyVaultName `
			-Name $ManagedStorageAccountName `
			-AccountResourceId $storage.Id `
			-ActiveKeyName $ActiveKeyName `
			-DisableAutoRegenerateKey # remove this later
			# set the rotation policy for the dev account
			# but we will only onboard that once the app code is using SAS for KV
			#-RegenerationPeriod [System.Timespan]::FromDays(90)
	Write-Host $storage.StorageAccountName "was set as a managed storage account."

	# create the sas so the service can use it at runtime
	# sas can be 2 years because the secret name is not the exact storage name will be read at runtime 
	# managed service identity should only have access to KV and DRI through JIT
	# if you want to share a managed SAS you should create a new SS
	# documentation https://docs.microsoft.com/en-us/powershell/module/azurerm.keyvault/set-azurekeyvaultmanagedstoragesasdefinition?view=azurermps-4.4.1#optional-parameters	
	# we are assigning this to billing because PAV2 takes a dependency on a read only sas token.
	Set-AzureKeyVaultManagedStorageSasDefinition -Service Blob,Table,Queue,File `
		-ResourceType Container,Service,Object `
		-VaultName $keyVaultName `
		-AccountName $ManagedStorageAccountName `
		-Name "crudall" `
		-Protocol HttpsOnly `
		-ValidityPeriod $compliantValidity `
		-Permission Add,Create,Delete,List,Process,Read,Update,Write		
		Write-Host $storage.StorageAccountName" default SAS credential has been created."

	if($keyVaultName.Contains('billing'))
	{			
		# will be used by PAV2 billing
		Set-AzureKeyVaultManagedStorageSasDefinition -Service Table,Queue `
			-ResourceType Container,Service,Object `
			-VaultName $keyVaultName `
			-AccountName $ManagedStorageAccountName `
			-Name "pav2" `
			-Protocol HttpsOnly `
			-ValidityPeriod $compliantValidity `
			-Permission Add,Process,Read,Update
			Write-Host $storage.StorageAccountName" PAV2 credential has been created."
	}	
}
# set the user to not have access to storage on key vault anymore since it is managed
Set-lbsKeyVaultStorageAccess -env $env -resourceGroup $keyVaultResourceGroup -objectId $userAccount.Id