﻿# ##########################################################################################################################################################
# Used by the build pipeline to copy service fabric application parameters & version number to EV2 parameter files.
#
# Example:
# $PATH:\CopySFSettings.ps1 -srcSFApplicationManifest $PATH:\Azure-IoT-LocationBasedServices\src\ALSProxy\SF\ApplicationPackageRoot\ApplicationManifest.xml `
# -srcSFtoDestEv2 @{"$PATH:\Azure-IoT-LocationBasedServices\src\ALSProxy\SF\ApplicationParameters\c-svc-eastus-cloud.xml" = "$PATH:\Azure-IoT-LocationBasedServices\src\DeployAndSetup\Provision\CI\Parameters\SFAPI.svc.eastus.Parameters.json"}
# ##########################################################################################################################################################
Param
(
	[Parameter(Mandatory=$True)]
	[ValidateScript({Test-Path $_ -PathType Leaf})]
	[string]$srcSFApplicationManifestFile, 
	
	[Parameter(Mandatory=$True)]
	[hashtable]$srcSFtoDestEv2 # xml application parameter file -> EV2 json parameter file
)

Write-Host "Reading $srcSFApplicationManifestFile"
[xml]$applicationManifest = Get-Content $srcSFApplicationManifestFile
$sfVersion = $applicationManifest.ApplicationManifest.ApplicationTypeVersion
Write-Host "ApplicationTypeVersion: $sfVersion"

foreach ($entry in $srcSFtoDestEv2.GetEnumerator()) {
	$srcSFParameterFile = $entry.Key
	$destEv2ParameterFile = $entry.Value

	if ((Test-Path $srcSFParameterFile -PathType Leaf) -eq 0) {
		throw "$srcSFParameterFile not found"
	}
	
	if ((Test-Path $destEv2ParameterFile -PathType Leaf) -eq 0) {
		throw "$destEv2ParameterFile not found"
	}
	
	# Read EV2 parameter file
	Write-Host "Reading $destEv2ParameterFile"
	$ev2Parameters = Get-Content $destEv2ParameterFile | ConvertFrom-Json

	# Update the application type version
	Write-Host "Updating $destEV2ParameterFile with version number '$sfVersion'"
	$ev2Parameters.parameters.applicationTypeVersion.value = $sfVersion

	# Update the application parameters
	Write-Host "Reading $srcSFParameterFile"
	[xml]$applicationParameters = Get-Content $srcSFParameterFile
	$parameters = $applicationParameters.Application.Parameters.ChildNodes
	$applicationParametersHash = @{}
	foreach ($parameter in $parameters) {
		Write-Host "Adding '$($parameter.Name)' parameter"
		$applicationParametersHash[$parameter.Name] = $parameter.Value
	}
	$ev2Parameters.parameters.applicationParameters.value = $applicationParametersHash

	# Write EV2 parameter file
	Write-Host "Writing updated EV2 parameter file to $destEv2ParameterFile"
	$jsonOutput = $ev2Parameters | ConvertTo-Json -Depth 20 | Set-Content -Encoding UTF8 -Path $destEv2ParameterFile
}