﻿# ##########################################################################################################################################################
# This script will take all storage accounts and other resources which have secrets (comos db, event hub) and rotate/renew the key which is not currently in use then set that key as active.
# applications must smart enough to retrieve the key after this script executes.
# https://docs.microsoft.com/en-us/azure/key-vault/key-vault-ovw-storage-keys 
# https://microsoft.sharepoint.com/teams/AzureSecurityCompliance/Security/SitePages/SecretsManagement.aspx
# for details, on the purpose of this script
# ##########################################################################################################################################################

Param
(
    [Parameter(Mandatory=$True)]
    [ValidateSet('p', 's', 't', 'c')]
    [string] $env,
	[Parameter(Mandatory=$False)]    
    [string] $lbsModulePath,
	[Parameter(Mandatory=$False)]    
    [string] $keyVaultResourceGroup = 'Vaults'
)

function Get-lbsNewStorageActiveKey 
{
	Param 
	(
		[string] $currentKey
	)

	if($currentKey -ieq 'key2')
	{			
		return 'key1'
	}
	else
	{
		return 'key2'
	}
}

function Rotate-lbsCosmosDbAccount
{
	Param
	(
		[string] $env,
		[string] $cosmosDbAccount,
		[string] $cosmosDbSecretName
	)

	$globalVaultName = Get-lbsGlobalVaultName -env $env
	$globalResourceGroup = 'Global'
	$resourceType = "Microsoft.DocumentDb/databaseAccounts"
	$apiVersion = "2015-04-08"
	$tags = @{ 'updatedByProcess' = 'rotation'}
	$existingCosmosDbSecret = Get-AzureKeyVaultSecret -Name $cosmosDbSecretName -VaultName $globalVaultName

	$cosmosDbSecrets = Invoke-AzureRmResourceAction -Action listKeys `
		-ResourceType $resourceType `
		-ApiVersion $apiVersion `
		-ResourceGroupName $globalResourceGroup `
		-Name $cosmosDbResourceName `
		-Force

	if($existingCosmosDbSecret.SecretValueText -eq $cosmosDbSecrets.primaryMasterKey)
	{
		$message = "Apps using " + $globalVaultName + " " + $cosmosDbSecretName + " will lose access, move them to primaryMasterKey if you think any app hasn't been switched, move it first. Continue: Y/N?"
		$reply = Read-Host $message		
		if($reply.Trim() -ieq 'Y')
		{
			$cosmosDbSecrets = Invoke-AzureRmResourceAction -Action regenerateKey `
				-ResourceType $resourceType `
				-ApiVersion $apiVersion `
				-ResourceGroupName $globalResourceGroup `
				-Name $cosmosDbResourceName `
				-Parameters @{"keyKind"="Secondary"} `
				-Force
		
			$cosmosDbKeySecretSecure = ConvertTo-SecureString -String $cosmosDbSecrets.secondaryMasterKey -AsPlainText -Force
			Set-AzureKeyVaultSecret -VaultName $globalVaultName -Name $cosmosDbSecretName -SecretValue $cosmosDbKeySecretSecure -Tag $tags
		}	
	}
	else
	{
		$message = "Apps using " + $globalVaultName + " " + $cosmosDbSecretName + " will lose access, move them to secondaryMasterKey if you think any app hasn't been switched, move it first. Continue: Y/N?"
		$reply = Read-Host $message		
		if($reply.Trim() -ieq 'Y')
		{
			$cosmosDbSecrets = Invoke-AzureRmResourceAction -Action regenerateKey `
				-ResourceType $resourceType `
				-ApiVersion $apiVersion `
				-ResourceGroupName $globalResourceGroup `
				-Name $cosmosDbResourceName `
				-Parameters @{"keyKind"="Primary"} `
				-Force

			$cosmosDbKeySecretSecure = ConvertTo-SecureString -String $cosmosDbSecrets.primaryMasterKey -AsPlainText -Force
			Set-AzureKeyVaultSecret -VaultName $globalVaultName -Name $cosmosDbSecretName -SecretValue $cosmosDbKeySecretSecure -Tag $tags
		}	
	}
}

function Update-lbsManagedStorageKey
{
	Param
	(
		[string] $env,
		[string] $storageAccountName
	)
	
	$managedStorage = Get-lbsManagedStorageAccount -env $env -storageAccountName $storageAccountName

	if($managedStorage -eq $null){
		Write-Host "Managed Storage does not exist or has an incorrect name for storage account " $storageAccountName " contact the owner of this account and onboard the account." -ForegroundColor Yellow
		return;
	}
		
	Write-Host $storageAccountName "is currently using:" $managedStorage.ActiveKeyName -ForegroundColor Cyan	
	$newKey = Get-lbsNewStorageActiveKey -currentKey $managedStorage.ActiveKeyName
	$message = "Apps using " + $storageAccountName + " " + $newKey + " will lose access, move them to " + $managedStorage.ActiveKeyName + " if you think any app hasn't been switched, move it first. Continue: Y/N?"
	$reply = Read-Host $message		
	if($reply.Trim() -ine 'Y')
	{
		Write-Host -ForegroundColor Yellow "Skipping" $storageAccountName $managedStorage.ActiveKeyName
		return;
	}
	
	# https://docs.microsoft.com/en-us/powershell/module/AzureRM.KeyVault/Update-AzureKeyVaultManagedStorageAccountKey?view=azurermps-4.4.1
	# 1. Regenerates newKey of account 'mystorageaccount' - This will invalidate any existing SasToken.
	# 2. Sets newKey as the active of the Key Vault managed Azure Storage Account.
	$managedStorage = Update-AzureKeyVaultManagedStorageAccountKey -VaultName $managedStorage.VaultName -AccountName $managedStorage.AccountName -KeyName $newKey -ErrorAction Stop

	# weird ps bug doesn't update the object with the new state.
	$managedStorage = Get-lbsManagedStorageAccount -env $env -storageAccountName $storageAccountName

	Write-Host $managedStorage.AccountName "now has" $managedStorage.ActiveKeyName "as active." -ForegroundColor Green
}

if([string]::IsNullOrEmpty($lbsModulePath))
{
	$modulePath = $PSScriptRoot + '\..\Utils\lbsSetupSupport.psd1'
}
else
{
	$modulePath = $lbsModulePath
}

Import-Module -Name $modulePath -Force
$subscriptionId = Get-lbsEnvironmentSubscriptionId -env $env
$subscription = Select-lbsSubscriptionWithQuickLogin -subscriptionId $subscriptionId
$storageResourceGroup = 'Storage'
$userAccount = Get-AzureRmADUser -UserPrincipalName $subscription.Account.Id
$storages = Get-AzureRmStorageAccount -ResourceGroupName $storageResourceGroup

# set the user to have access to storage on key vault
Set-lbsKeyVaultStorageAccess -env $env -resourceGroup $keyVaultResourceGroup -objectId $userAccount.Id -elevate
foreach($storage in $storages)
{
	Update-lbsManagedStorageKey -env $env -storageAccountName $storage.StorageAccountName
}
# set the user to not have access to storage on key vault anymore since it is managed
Set-lbsKeyVaultStorageAccess -env $env -resourceGroup $keyVaultResourceGroup -objectId $userAccount.Id

Set-lbsKeyVaultSecretAccess -env $env -resourceGroup $keyVaultResourceGroup -objectId $userAccount.Id -elevate
# rotate Event Hub Keys
$tags = @{ 'updatedByProcess' = 'rotation'}
$globalResourceGroup = 'Global'
$eventHubNamespaces = Get-AzureRmEventHubNamespace -ResourceGroupName $globalResourceGroup
$globalVaultName = Get-lbsGlobalVaultName -env $env
foreach($eventHubNamespace in $eventHubNamespaces)
{
	$eventHubs = Get-AzureRmEventHub -ResourceGroupName $globalResourceGroup -NamespaceName $eventHubNamespace.Name
	foreach($eventHub in $eventHubs)
	{		
		$authorizationRuleName = 'Receiving'
		$eventHubReceivingKey = Get-AzureRmEventHubKey -ResourceGroupName $globalResourceGroup -EventHubName $eventHub.Name -AuthorizationRuleName $authorizationRuleName -NamespaceName $eventHubNamespace.Name
		$eventHubSecretName = New-lbsManagedName -name ($eventHubNamespace.Name.Substring(1) + '-' + $eventHubReceivingKey.KeyName)

		$existingReceivingHubSasSecret = Get-AzureKeyVaultSecret -Name $eventHubSecretName -VaultName $globalVaultName
		if($existingReceivingHubSasSecret.SecretValueText -eq $eventHubReceivingKey.PrimaryConnectionString)
		{
			$message = "Apps using " + $globalVaultName + " " + $eventHubSecretName + " will lose access, move them to PrimaryConnectionString if you think any app hasn't been switched, move it first. Continue: Y/N?"
			$reply = Read-Host $message		
			if($reply.Trim() -ine 'Y')
			{
				Write-Host -ForegroundColor Yellow "Skipping" $globalVaultName $eventHubSecretName
				continue;
			}	

			$eventHubReceivingKey = New-AzureRmEventHubKey -ResourceGroupName $globalResourceGroup -Namespace $eventHubNamespace.Name -Name $authorizationRuleName -EventHub $eventHub.Name -RegenerateKey SecondaryKey			
			$eventHubSasConnectionStringSecure = ConvertTo-SecureString $eventHubReceivingKey.SecondaryConnectionString -AsPlainText -Force
		}
		else
		{
			$message = "Apps using " + $globalVaultName + " " + $eventHubSecretName + " will lose access, move them to SecondaryConnectionString if you think any app hasn't been switched, move it first. Continue: Y/N?"
			$reply = Read-Host $message		
			if($reply.Trim() -ine 'Y')
			{
				Write-Host -ForegroundColor Yellow "Skipping" $globalVaultName $eventHubSecretName
				continue;
			}

			$eventHubReceivingKey = New-AzureRmEventHubKey -ResourceGroupName $globalResourceGroup -Namespace $eventHubNamespace.Name -Name $authorizationRuleName -EventHub $eventHub.Name -RegenerateKey PrimaryKey
			$eventHubSasConnectionStringSecure = ConvertTo-SecureString $eventHubReceivingKey.PrimaryConnectionString -AsPlainText -Force			
		}
		
		Set-AzureKeyVaultSecret -VaultName $globalVaultName -Name $eventHubSecretName -SecretValue $eventHubSasConnectionStringSecure -Tag $tags
	}	
}


# rotate cosmos db account
$cosmosDbResourceName = "$env-geofencing"

Rotate-lbsCosmosDbAccount -env $env -cosmosDbAccount $cosmosDbResourceName -cosmosDbSecretName "cosmosDbAuthorizationKey"
$mapsFrontEndCosmosDbAccountName = Get-lbsMapsFrontEndCosmosDbName -env $env
Rotate-lbsCosmosDbAccount -env $env -cosmosDbAccount $mapsFrontEndCosmosDbAccountName -cosmosDbSecretName "mapsFrontEndCosmosDbKey"
Set-lbsKeyVaultSecretAccess -env $env -resourceGroup $keyVaultResourceGroup -objectId $userAccount.Id