# ##########################################################################################################################################################
# This script will create two applications (and service principals) for managing a service fabric cluster. The Web app is used as the cluster's cluster application and the native
# app is used as the cluster's client application.
# ##########################################################################################################################################################

param(
 [Parameter(Mandatory=$True)]
 [string] $TenantId,

 [Parameter(Mandatory=$True)]
 [string] $ClusterName,

 [Parameter(Mandatory=$True)]
 [string]  $ClusterDomainNamePrefix,

 [ValidateSet('AzureGermanyCloud','AzureCloud','AzureChinaCloud')]
 [string] $AzureEnvironmentName = "AzureCloud"
)

function Invoke-CallGraphAPI()
{
    Param(
        [Parameter(Mandatory = $true, Position = 1)]
        $Uri,

        [Parameter(Mandatory = $true, Position = 2)]
        $Headers,

        [Parameter(Mandatory = $true, Position = 3)]
        $Body
    )

    $json = $Body | ConvertTo-Json -Depth 4 -Compress
    $response =  (Invoke-RestMethod $Uri -Method Post -Headers $Headers -Body $json -ContentType "application/json")
    return $response
}

function Get-GraphEndpoint()
{
    Param(
        [Parameter(Mandatory = $true, Position = 1)]
        [string] $Environment
    )

    switch ($Environment)
    {
        "AzureCloud"
        {
            return "https://graph.windows.net"
        }
        "AzureChinaCloud"
        {
            return "https://graph.chinacloudapi.cn"
        }
        "AzureGermanyCloud"
        {
            return "https://graph.cloudapi.de"
        }
    }
}

function Get-GraphAuthHeader
{
    $header = @{}

    if ([Microsoft.Open.Azure.AD.CommonLibrary.AzureSession]::AccessTokens.Count -eq 0)
    {
        throw "please use Connect-AzureAD to login to Active Directory"
    }

    $token = $null
    if ([Microsoft.Open.Azure.AD.CommonLibrary.AzureSession]::AccessTokens.ContainsKey("AccessToken") -eq $true)
    {
        $token = [Microsoft.Open.Azure.AD.CommonLibrary.AzureSession]::AccessTokens["AccessToken"].AccessToken
    }
    elseif ([Microsoft.Open.Azure.AD.CommonLibrary.AzureSession]::AccessTokens.ContainsKey([Microsoft.Open.Azure.AD.CommonLibrary.AzureEnvironment+Endpoint]::AadGraphEndpointResourceId) -eq $true)
    {
        $token = [Microsoft.Open.Azure.AD.CommonLibrary.AzureSession]::AccessTokens[[Microsoft.Open.Azure.AD.CommonLibrary.AzureEnvironment+Endpoint]::AadGraphEndpointResourceId].AccessToken
    }
    elseif ([Microsoft.Open.Azure.AD.CommonLibrary.AzureSession]::AccessTokens.ContainsKey([Microsoft.Open.Azure.AD.CommonLibrary.AzureEnvironment+Endpoint]::ActiveDirectory) -eq $true)
    {
        $token = [Microsoft.Open.Azure.AD.CommonLibrary.AzureSession]::AccessTokens[[Microsoft.Open.Azure.AD.CommonLibrary.AzureEnvironment+Endpoint]::ActiveDirectory].AccessToken
    }
    elseif ([Microsoft.Open.Azure.AD.CommonLibrary.AzureSession]::AccessTokens.ContainsKey([Microsoft.Open.Azure.AD.CommonLibrary.AzureEnvironment+Endpoint]::ActiveDirectoryServiceEndpointResourceId) -eq $true)
    {
        $token = [Microsoft.Open.Azure.AD.CommonLibrary.AzureSession]::AccessTokens[[Microsoft.Open.Azure.AD.CommonLibrary.AzureEnvironment+Endpoint]::ActiveDirectoryServiceEndpointResourceId].AccessToken
    }
    elseif ([Microsoft.Open.Azure.AD.CommonLibrary.AzureSession]::AccessTokens.ContainsKey([Microsoft.Open.Azure.AD.CommonLibrary.AzureEnvironment+Endpoint]::MsGraphEndpointResourceId) -eq $true)
    {
        $token = [Microsoft.Open.Azure.AD.CommonLibrary.AzureSession]::AccessTokens[[Microsoft.Open.Azure.AD.CommonLibrary.AzureEnvironment+Endpoint]::MsGraphEndpointResourceId].AccessToken
    }

    if ($token -eq $null)    
    {
        throw "Please connect to AzureAD see Connect-AzureAD"
    }

    $header.Add("Authorization", $token)
    return $header;
}

# Example -TenantId 9e88becc-48d3-42b4-9766-543fd455ad51 -ClusterName "albs-c-jobs-eastus"  -ClusterDomainName  "cjobscluster.eastus"

Import-Module -Name AzureAD -Force

Connect-AzureAD -TenantId $TenantId -AzureEnvironmentName $AzureEnvironmentName

$WebAppName = $ClusterName + "_webapp"
$NativeAppName = $ClusterName + "_nativeapp"

$WebApplicationUri = 'https://' + $ClusterDomainNamePrefix
$HomePage = 'https://' + $ClusterDomainNamePrefix + '.cloudapp.azure.com:19080/Explorer'
$WebApplicationReplyUrl = 'https://' + $ClusterDomainNamePrefix + '.cloudapp.azure.com:19080/Explorer/index.html'

# App Roles
$appRole =
@{
    allowedMemberTypes = @("User")
    description = "ReadOnly roles have limited query access"
    displayName = "ReadOnly"
    id = [guid]::NewGuid()
    isEnabled = "true"
    value = "User"
},
@{
    allowedMemberTypes = @("User")
    description = "Admins can manage roles and perform all task actions"
    displayName = "Admin"
    id = [guid]::NewGuid()
    isEnabled = "true"
    value = "Admin"
}


$webApp = Get-AzureADApplication  -Filter ("displayName eq '" + $WebAppName + "'")

if ($webApp -eq $null)
{
    $appRoles = New-Object System.Collections.Generic.List[Microsoft.Open.AzureAD.Model.AppRole]
    $role1 = New-Object Microsoft.Open.AzureAD.Model.AppRole
    $role1.AllowedMemberTypes = @("User")
    $role1.Description = "ReadOnly roles have limited query access"
    $role1.DisplayName = "ReadOnly"
    $role1.Id = [guid]::NewGuid()
    $role1.IsEnabled = $True
    $role1.Value = "User"
    $appRoles.Add($role1)
    $role2 = New-Object Microsoft.Open.AzureAD.Model.AppRole
    $role2.AllowedMemberTypes = @("User")
    $role2.Description = "Admins can manage roles and perform all task actions"
    $role2.DisplayName = "Admin"
    $role2.Id = [guid]::NewGuid()
    $role2.IsEnabled = $True
    $role2.Value = "Admin"
    $appRoles.Add($role2)

    Write-Host "Creating Web Application " $WebAppName
    $webApp = New-AzureADApplication -AppRoles $appRoles -DisplayName $WebAppName -IdentifierUris @($WebApplicationUri) -Homepage $HomePage -ReplyUrls @($WebApplicationReplyUrl)
}
else 
{
    Write-Host "Using existing Web Application " $webApp.DisplayName
}

$webAppSp = Get-AzureADServicePrincipal -Filter ("displayName eq '" + $webApp.DisplayName + "'")

if ($webAppSp -eq $null) 
{
    Write-Host "Creating Web Application SP " $webApp.DisplayName
    $webAppSp = New-AzureADServicePrincipal -AccountEnabled $True -AppId $webApp.AppId -DisplayName $webApp.DisplayName -AppRoleAssignmentRequired $True
}
else
{
    Write-Host "Using existing Web Application SP " $webApp.DisplayName    
}


$requiredResourceAccess = New-Object System.Collections.Generic.List[Microsoft.Open.AzureAD.Model.RequiredResourceAccess]

$rrAdAccess = New-Object Microsoft.Open.AzureAD.Model.RequiredResourceAccess
$rrAdAccess.ResourceAppId = "00000002-0000-0000-c000-000000000000"
$rrAdAccess.ResourceAccess = new-object System.Collections.Generic.List[Microsoft.Open.AzureAD.Model.ResourceAccess]
$rrAdAccess.ResourceAccess.Add((New-Object -TypeName "Microsoft.Open.AzureAD.Model.ResourceAccess" -ArgumentList "311a71cc-e848-46a1-bdf8-97ff7156d8e6","Scope"))
$requiredResourceAccess.Add($rrAdAccess)

$rrWebAccess = New-Object Microsoft.Open.AzureAD.Model.RequiredResourceAccess
$rrWebAccess.ResourceAppId = $webApp.AppId
$rrWebAccess.ResourceAccess = new-object System.Collections.Generic.List[Microsoft.Open.AzureAD.Model.ResourceAccess]
$rrWebAccess.ResourceAccess.Add((New-Object -TypeName "Microsoft.Open.AzureAD.Model.ResourceAccess" -ArgumentList $webApp.Oauth2Permissions[0].Id,"Scope"))
$requiredResourceAccess.Add($rrWebAccess)

$nativeApp = Get-AzureADApplication  -Filter ("displayName eq '" + $NativeAppName + "'")

if ($nativeApp -eq $null)
{
    Write-Host "Creating Native Application " $NativeAppName
    $nativeApp = New-AzureADApplication -PublicClient $True -DisplayName $NativeAppName -ReplyUrls @("urn:ietf:wg:oauth:2.0:oob") -RequiredResourceAccess $requiredResourceAccess
}
else 
{
    Write-Host "Using existing Native Application " $nativeApp.DisplayName    
}

$nativeSp = Get-AzureADServicePrincipal -Filter ("displayName eq '" + $nativeApp.DisplayName + "'")

if ($nativeSp -eq $null)
{
    Write-Host "Creating Native Application SP " $nativeSp.DisplayName
    $nativeSp = New-AzureADServicePrincipal -AccountEnabled $True -AppId $nativeApp.AppId -DisplayName $nativeApp.DisplayName
}
else
{
    Write-Host "Using existing Native Application SP " $nativeApp.DisplayName    
}

# This is a built in SP 
$aadSp = Get-AzureADServicePrincipal -Filter "appId eq '00000002-0000-0000-c000-000000000000'"

# Get the existing oauth grants 
$grants = Get-AzureADServicePrincipalOAuth2PermissionGrant -All $true -ObjectId $nativeSp.ObjectId

# Setup to make direct graph calls 
$graphEndpoint = Get-GraphEndpoint $AzureEnvironmentName

$graphAPIFormat = $graphEndpoint + "/" + $TenantId + "/{0}?api-version=1.5"

$headers = Get-GraphAuthHeader

$userRead = $grants | Where-Object {$_.Scope -eq 'User.Read'} | Select-Object ObjectId

if ($userRead -eq $null)
{
    $uri = [string]::Format($graphAPIFormat, "oauth2PermissionGrants")
    $oauth2PermissionGrants = @{
        clientId = $nativeSp.ObjectId
        consentType = "AllPrincipals"
        resourceId = $aadSp.ObjectId
        scope = "User.Read"
        startTime = (Get-Date).ToUniversalTime().ToString("yyyy-MM-ddTHH:mm:ss.fffffff")
        expiryTime = (Get-Date).AddYears(1800).ToUniversalTime().ToString("yyyy-MM-ddTHH:mm:ss.fffffff")
    }

    Invoke-CallGraphAPI $uri $headers $oauth2PermissionGrants | Out-Null
}

$userimp = $grants | Where-Object {$_.Scope -eq 'user_impersonation'} | Select-Object ObjectId

if ($userimp -eq $null)
{
    $oauth2PermissionGrants = @{
        clientId = $nativeSp.objectId
        consentType = "AllPrincipals"
        resourceId = $webAppSp.ObjectId
        scope = "user_impersonation"
        startTime = (Get-Date).ToUniversalTime().ToString("yyyy-MM-ddTHH:mm:ss.fffffff")
        expiryTime = (Get-Date).AddYears(1800).ToUniversalTime().ToString("yyyy-MM-ddTHH:mm:ss.fffffff")
    }

    Invoke-CallGraphAPI $uri $headers $oauth2PermissionGrants | Out-Null
}

Write-Host
Write-Host '-----' $ClusterName 'ARM template-----'
Write-Host '"azureActiveDirectory": {'
Write-Host ("  `"tenantId`":`"{0}`"," -f $TenantId)
Write-Host ("  `"clusterApplication`":`"{0}`"," -f $webApp.AppId)
Write-Host ("  `"clientApplication`":`"{0}`"" -f $nativeApp.AppId)
Write-Host '},'

$result = @{"TenantId" = $TenantId; "CLUSTERAPPLICATION" = $webApp.AppId; "CLIENTAPPLICATION" = $nativeApp.AppId;}
return $result