﻿$modulePath = $PSScriptRoot + ".\IotLBSDevSetup.psd1"
Import-Module -Name $modulePath -Force

$RegKey = "HKLM:\SOFTWARE\Microsoft\Service Fabric SDK"
$modulePath = (Get-ItemProperty -Path $RegKey -Name FabricSDKPSModulePath).FabricSDKPSModulePath
Import-Module "$modulePath\ServiceFabricSDK.psm1" -Force

function Get-CertBytesByEnv
{
    Param(
        [Parameter(Mandatory = $true)] [string] $env,
        [Parameter(Mandatory = $true)] [string] $pfxPassword
    )

    $subscriptionId = Get-lbsEnvironmentSubscriptionId -env $env;
    $subscription = Select-lbsSubscriptionWithQuickLogin -subscriptionId $subscriptionId

    switch($env.ToLower())
    {
        'p' { $secretsCertName = 'psvc-apim'} # PROD
        's' { $secretsCertName = 'ssvc-apim'} # STG
        't' { $secretsCertName = 'tsvc-apim'} # TEST
        'c' { $secretsCertName = 'csvc-apim'} # TEST
    }

    $globalKeyVaultName = Get-lbsGlobalVaultName -env $env
    $kvSecret = Get-AzureKeyVaultSecret -VaultName $globalKeyVaultName -Name $secretsCertName
    $kvSecretBytes = [System.Convert]::FromBase64String($kvSecret.SecretValueText)
    $certCollection = New-Object System.Security.Cryptography.X509Certificates.X509Certificate2Collection
    $certCollection.Import($kvSecretBytes,$null,[System.Security.Cryptography.X509Certificates.X509KeyStorageFlags]::Exportable)
    $pfxBytes = $certCollection.Export([System.Security.Cryptography.X509Certificates.X509ContentType]::Pkcs12, $pfxPassword)

    return $pfxBytes
}

function Debug-lbsPromptToContinue
{
    Param(
        [Parameter(Mandatory = $false)] [string] $message
    )

    try {
        if ($message.Length -gt 0) {
            Write-Host -NoNewLine $message "Continue? (Y/N) " -foregroundcolor Cyan -backgroundcolor Black
        } else {
            Write-Host -NoNewLine "Continue? (Y/N) " -foregroundcolor Cyan -backgroundcolor Black
        }
        do {
		        $key = [Console]::ReadKey("noecho")
                if ($key.Key.ToString().ToLower() -eq 'n') {
                    Write-Host $key.Key.ToString();
                    Exit;
                }
        } while($key.Key.ToString().ToLower() -ne 'y')
        Write-Host $key.Key.ToString();
    } catch {
    }
}

function Select-lbsSubscriptionWithQuickLogin
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [string]$subscriptionId
    )

    for($i = 0; $i -le 1; $i++)
    {
        try
        {
            $s = Select-AzureRmSubscription -SubscriptionId $subscriptionId -ErrorAction Stop
            Write-Host "----------------------------------------------------------------------------------------------------------"
            Write-Host 'Account          :' $s.Account.Id
            Write-Host 'TenantId         :' $s.Subscription.TenantId
            Write-Host 'SubscriptionId   :' $s.Subscription.Id
            Write-Host 'SubscriptionName :' $s.Subscription.Name
            Write-Host "----------------------------------------------------------------------------------------------------------"
            return $s
        }
        catch
        {
            if ($i -eq 0) {
                Login-AzureRmAccount -SubscriptionId $subscriptionId -ErrorAction Stop
            }
        }
    }
    $error = "Failed to login and/or selecting subscription"
    Throw $error
}

function Get-lbsEnvironmentSubscriptionId
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'd', 'e', 'c')]
        [string]$env
    )

    switch($env.ToLower())
    {
		'p' { $subscriptionId = '387e8e8b-95ec-4229-ac0b-271e73dec988'; } # PROD
        's' { $subscriptionId = '68ff275e-741f-4cda-90bb-c6065a46d098'; } # STG
        't' { $subscriptionId = '002509a6-fb5b-4fe0-abbf-b28ea9cc115c'; } # TEST
        'd' { $subscriptionId = '21a9967a-e8a9-4656-a70b-96ff1c4d05a0'; } # DEV
		'e' { $subscriptionId = 'e4b1c29f-c726-42e0-9603-092535a49705'; } # External Data
		'c' { $subscriptionId = 'b465bdfb-8975-44c9-86ec-36b250c9b942'; } # CI
    }
    return $subscriptionId
}

function Get-lbsEnvironmentExpressV2Folder
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'c')]
        [string]$env
    )

    switch($env.ToLower())
    {
        'p' { $name = 'PROD_BN' } # PROD
        's' { $name = 'STG_BY'  } # STG
        't' { $name = 'TEST_AM' } # TEST
        'c' { $name = 'CI' } # TEST
    }
    return $name
}

function Get-lbsEnvironments
{
    $environments = @('t', 's', 'p', 'c');
    return $environments;
}

# This is the dev setup version which should be referenced in Deploy-FabricApplication and devSetup
function Get-lbsEnvironmentDevelopmentVersion
{
	# NOTE - bump it up whenever rp or svc cloud.config params change
    $devVersion = '12';
    return $devVersion
}

# This is special because we don't treat development as an environment 'd'
function Get-lbsEnvironmentDevelopmentRegion
{
    $region ='eastus';
    return $region
}

function Get-lbsEnvironmentRegions
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'c')]
        [string]$env,

		[string]$sfCluster
    )

	if($sfCluster -ieq 'jobs')
	{
		return  @('eastus')
	}

    switch($env.ToLower())
    {
        'p' { $regions = @('eastus', 'westeurope', 'westus2') } # PROD
        's' { $regions = @('eastus', 'westeurope', 'westus2') } # STG
        't' { $regions = @('eastus') } # TEST
        'c' { $regions = @('eastus') } # CI
    }
    return $regions
}

function Get-lbsServiceFabricClusters
{
    $sfClusters = @('rp', 'svc', 'jobs');
    return $sfClusters;
}

function Get-lbsCertificateTypes
{
    $certTypes = @('cluster', 'client', 'secrets', 'ma', 'tls');
    return $certTypes;
}

function Get-lbsStorageAccountKeyVaultName
{
	Param
	(
		[Parameter(Mandatory=$True)]    
		[string]$storageAccountName,
		[Parameter(Mandatory=$True)] 
		[ValidateSet('p', 's', 't', 'c')]
		[string]$env
	)
	if($storageAccountName.ToLower().Contains('billing'))
	{		
		$keyVaultName = Get-lbsBillingVaultName -env $env		
	}		
	else
	{
		$keyVaultName = Get-lbsGlobalVaultName -env $env		
	}
	return $keyVaultName
}

function Get-lbsManagedStorageAccount
{
	Param
	(
		[Parameter(Mandatory=$True)]    
		[string] $storageAccountName,
		[Parameter(Mandatory=$True)] 
		[ValidateSet('p', 's', 't', 'c')]
		[string] $env
	)

	$keyVaultName = Get-lbsStorageAccountKeyVaultName -storageAccountName $storageAccountName -env $env
	$managedStorageAccountName = Get-lbsManagedStorageAccountName -storageAccountName $storageAccountName

	$managedStorage = Get-AzureKeyVaultManagedStorageAccount -VaultName $keyVaultName -AccountName $managedStorageAccountName -ErrorAction SilentlyContinue

	if($managedStorage -eq $null)
	{		
		Write-Host "Could not find " $managedStorageAccountName " in KeyVault " $keyVaultName
		return;
	}
	else
	{
		return $managedStorage
	}	
}

function Get-lbsManagedStorageAccountName
{
	Param
    (
        [Parameter(Mandatory=$True)]        
        [string] $storageAccountName
	)
    	
	# add 'm' to indicate managed and
	#remove the env so it is different than the account name
	$ManagedStorageAccountName = New-lbsManagedName -name $storageAccountName.Substring(1)

	if($storageAccountName.ToLower().Contains('rpbackup'))
	{
		$ManagedStorageAccountName = "mrpbackup" # special because it has unique string name.
	}

    return $ManagedStorageAccountName
}

function New-lbsManagedName
{
	   Param
    (
        [Parameter(Mandatory=$True)]        
        [string]$name
	)
    
    return 'm' + $name
}

function Get-lbsMapsFrontEndCosmosDbName
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'c')]
        [string]$env
    )

    $name = $env + "-azuremaps";
    return $name.ToLower();
}

function Get-lbsGlobalVaultName
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'c')]
        [string]$env
    )

    $name = $env + "-global";
    return $name.ToLower();
}

function Get-lbsBillingVaultName
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'c')]
        [string]$env
    )

    $name = $env + "-billing";
    return $name.ToLower();
}

function Get-lbsVaultName
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'c')]
        [string]$env,

        [Parameter(Mandatory=$True)]
        [ValidateSet('rp', 'svc', 'jobs')]
        [string]$sfCluster,

        [Parameter(Mandatory=$True)]
        [ValidateSet('eastus', 'westeurope', 'westus2')]
        [string]$region
    )

    $name = $env + "-" + $sfCluster + "-" + $region;
    return $name.ToLower();
}


function Get-lbsKeyVaultSecretKeyValuePairs
{    
    Param(
        [Parameter(Mandatory = $true, Position = 1)]
		[ValidateSet('p', 's', 't', 'c')]
        [string] $env,

		[Parameter()]
		[switch] $versioned
    )
	$ConfigurationStrings = @{};
	$globalkeyVault = Get-lbsGlobalVaultName -env $env
	$billingKeyVault = Get-lbsBillingVaultName -env $env
    $keys = Get-AzureKeyVaultSecret -VaultName $globalkeyVault

    $keys | foreach {
		if($versioned)
		{
			if($_.ContentType -ieq 'application/vnd.ms-sastoken-storage')
			{
				# treat storage differently because a Sas token doesn't need to be version controlled, we just want the latest.
				$ConfigurationStrings.Add($_.Name,  $_.Id)
			}
			else
			{
				# any secrets should be versioned to maintain safe rollout and rollback of confguration
				# key vault is used at runtime to read the value and it is necessary to control which version to read.
				$secretWithVersions = Get-AzureKeyVaultSecret -VaultName $globalkeyVault -Name $_.Name -IncludeVersions
				# take the latest created version
				$sortedSecretVersions = $secretWithVersions | Sort-Object { $_.Created } -Descending
				$latestSecretVersion = $sortedSecretVersions[0]
				$ConfigurationStrings.Add($latestSecretVersion.Name,  $latestSecretVersion.Id)				
			}
		}
		else{
			$ConfigurationStrings.Add($_.Name,  $_.Id);
		}        
    }

	# intentional overwrite, when creating configuration, the keys will be unique in almost all cases
	# the reason to overwrite here is to move the billing managed storage account sas to billing key vault
	# this is because PAV2 takes a dependency on the storage account to read from table.
	$keys = Get-AzureKeyVaultSecret -VaultName $billingKeyVault
	$keys | foreach {
		if($ConfigurationStrings.ContainsKey($_.Name)){
			$ConfigurationStrings[$_.Name] = $_.Id
		}
		else{
			$ConfigurationStrings.Add($_.Name,  $_.Id);	
		}		
	}

	return $ConfigurationStrings
}

function Get-lbsGlobalCertificateName
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'c')]
        [string]$env,

        [Parameter(Mandatory=$True)]
        [ValidateSet('rp', 'svc', 'jobs')]
        [string]$sfCluster,

        [Parameter(Mandatory=$True)]
        [ValidateSet('apim')]
        [string]$certType
    )

    if ($certType.ToLower() -eq 'apim') { $name = $env + 'svc-apim'; }

    return $name.ToLower();
}

function Get-lbsCertificateName
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'c')]
        [string]$env,

        [Parameter(Mandatory=$True)]
        [ValidateSet('rp', 'svc', 'jobs')]
        [string]$sfCluster,

        [Parameter(Mandatory=$True)]
        [ValidateSet('eastus', 'westeurope', 'westus2')]
        [string]$region,

        [Parameter(Mandatory=$True)]
        [ValidateSet('cluster', 'client', 'secrets', 'ma', 'tls')]
        [string]$certType
    )

    $name = $env + $sfCluster + '-' + $certType + '-' + $region;
    return $name.ToLower();
}

# used to allow a principal id to have access to modify the key vault storage permissions.
# useful when you need to allow key vault to manage storage account keys
# read more here: https://docs.microsoft.com/en-us/powershell/module/azurerm.keyvault/set-azurermkeyvaultaccesspolicy?view=azurermps-4.4.1
function Set-lbsKeyVaultStorageAccess
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'c')]
        [string]$env,

        [Parameter(Mandatory=$True)]
        [string]$resourceGroup,

        [Parameter(Mandatory=$True)]
        [guid]$objectId,

        [Parameter()]
        [switch]$elevate
    )	

	$globalKeyVault = Get-lbsGlobalVaultName -env $env
	$billingKeyVault = $env + "-billing"
	if($elevate){
		Set-AzureRmKeyVaultAccessPolicy -ResourceGroupName $resourceGroup -VaultName $globalKeyVault -ObjectId $objectId -PermissionsToStorage all
		Set-AzureRmKeyVaultAccessPolicy -ResourceGroupName $resourceGroup -VaultName $billingKeyVault -ObjectId $objectId -PermissionsToStorage all
	}
	else{
		Remove-AzureRmKeyVaultAccessPolicy -ResourceGroupName $resourceGroup -VaultName $globalKeyVault -ObjectId $objectId
		Remove-AzureRmKeyVaultAccessPolicy -ResourceGroupName $resourceGroup -VaultName $billingKeyVault -ObjectId $objectId
	}
	
	$sfClusters = Get-lbsServiceFabricClusters
	foreach ($sfCluster in $sfClusters) 
	{
		$regions = Get-lbsEnvironmentRegions -env $env -sfCluster $sfCluster

		foreach ($region in $regions) 
		{
			$vault = Get-lbsVaultName -env $env -sfCluster $sfCluster -region $region

			if($elevate){
				Set-AzureRmKeyVaultAccessPolicy -ResourceGroupName $resourceGroup -VaultName $vault -ObjectId $objectId -PermissionsToStorage all
			}
			else{
				Remove-AzureRmKeyVaultAccessPolicy -ResourceGroupName $resourceGroup -VaultName $vault -ObjectId $objectId
			}
		}
	}
}

# used to allow a principal id to have access to modify the key vault secret permissions.
# useful when you need to allow key vault to manage secrets adhoc
# read more here: https://docs.microsoft.com/en-us/powershell/module/azurerm.keyvault/set-azurermkeyvaultaccesspolicy?view=azurermps-4.4.1
function Set-lbsKeyVaultSecretAccess
{
    Param
    (
        [Parameter(Mandatory=$True)]
        [ValidateSet('p', 's', 't', 'c')]
        [string]$env,

		[Parameter(Mandatory=$True)]
        [string]$resourceGroup,

        [Parameter(Mandatory=$True)]
        [guid]$objectId,

        [Parameter()]
        [switch]$elevate
    )	

	$globalKeyVault = Get-lbsGlobalVaultName -env $env
	if($elevate)
	{
		Set-AzureRmKeyVaultAccessPolicy -ResourceGroupName $resourceGroup  -VaultName $globalKeyVault -ObjectId $objectId -PermissionsToSecrets get,list,set
	}
	else
	{
		Remove-AzureRmKeyVaultAccessPolicy -ResourceGroupName $resourceGroup -VaultName $globalKeyVault -ObjectId $objectId
	}	
}

# ------------------------------------------------------------------------------------------------------------------------------------------------
# From Deploy-FabricApplication.ps1: reference the same function instead of maintaing 3 (or more) copies (provider, proxy and billing)
# ------------------------------------------------------------------------------------------------------------------------------------------------
function Get-ConfigDataPath(){
    [CmdletBinding()]
    Param(
        [Parameter(
            Mandatory = $true,
            Position = 1
        )] [string] $FileName
    )

    $currentPath = (get-item $PSScriptRoot)
    while($true)
    {
        $configPath = [System.IO.Path]::Combine($currentPath.FullName, ".\.configcache\")

        if ((Test-Path $configPath) -eq 1)
        {
            return [System.IO.Path]::Combine($configPath, $fileName)
        }

        $currentPath = $currentPath.Parent

        if ($curentPath -eq $currentPath.Root)
        {
            throw "Unable to locate .configCache please ensure that deestup has been run"
        }
    }
}

function Get-ConfigData()
{
    [CmdletBinding()]
    Param(
        [Parameter(Mandatory = $true, Position = 1)] [string] $FileName
    )

	$configCacheFile = Get-ConfigDataPath -FileName $FileName

	if ((Test-Path $configCacheFile) -eq 0)
	{
		return $null
	}

	$configData = Get-Content $configCacheFile | ConvertFrom-Json

	$configHashtable = @{}

    $configData | get-member -MemberType NoteProperty | Where-Object{ -not [string]::IsNullOrEmpty($configData."$($_.name)")} | ForEach-Object {$configHashtable.add($_.name,$configData."$($_.name)")} 

	return $configHashtable
} 

function Read-XmlElementAsHashtable
{
    Param (
        [System.Xml.XmlElement]
        $Element
    )

    $hashtable = @{}
    if ($Element.Attributes)
    {
        $Element.Attributes | 
            ForEach-Object {
                $boolVal = $null
                if ([bool]::TryParse($_.Value, [ref]$boolVal)) {
                    $hashtable[$_.Name] = $boolVal
                }
                else {
                    $hashtable[$_.Name] = $_.Value
                }
            }
    }

    return $hashtable
}

function Read-PublishProfile
{
    Param (
        [ValidateScript({Test-Path $_ -PathType Leaf})]
        [String]
        $PublishProfileFile
    )

    $publishProfileXml = [Xml] (Get-Content $PublishProfileFile)
    $publishProfile = @{}

    $publishProfile.ClusterConnectionParameters = Read-XmlElementAsHashtable $publishProfileXml.PublishProfile.Item("ClusterConnectionParameters")
    $publishProfile.UpgradeDeployment = Read-XmlElementAsHashtable $publishProfileXml.PublishProfile.Item("UpgradeDeployment")
    $publishProfile.CopyPackageParameters = Read-XmlElementAsHashtable $publishProfileXml.PublishProfile.Item("CopyPackageParameters")

    if ($publishProfileXml.PublishProfile.Item("UpgradeDeployment"))
    {
        $publishProfile.UpgradeDeployment.Parameters = Read-XmlElementAsHashtable $publishProfileXml.PublishProfile.Item("UpgradeDeployment").Item("Parameters")
        if ($publishProfile.UpgradeDeployment["Mode"])
        {
            $publishProfile.UpgradeDeployment.Parameters[$publishProfile.UpgradeDeployment["Mode"]] = $true
        }
    }

    $publishProfileFolder = (Split-Path $PublishProfileFile)
    $publishProfile.ApplicationParameterFile = [System.IO.Path]::Combine($PublishProfileFolder, $publishProfileXml.PublishProfile.ApplicationParameterFile.Path)

    return $publishProfile
}