﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Instrumentation
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Diagnostics.Tracing;
    using System.Threading.Tasks;
    using Microsoft.Cloud.InstrumentationFramework;

    public class Agent : EventSource
    {
        private static readonly string[] _sEventNames = BuildEventNameCollection();

        private static readonly string _processName = Process.GetCurrentProcess().ProcessName;

        private static Agent _instance = null;

#if DEBUG
        private readonly bool _etwTraceLogging;
#endif

        /// <summary>Instance constructor is private to enforce singleton semantics</summary>
        private Agent(InitParams initParams)
            : base(initParams.EventSourceName, InitParams.sourceSettings, initParams.EtwTelemetryTraits)
        {
            // A workaround for the problem where ETW activities do not get tracked until Tasks infrastructure is initialized.
            // This problem will be fixed in .NET Framework 4.6.2.
            Task.Run(() => { });

            // see http://stackoverflow.microsoft.com/a/6855/748
            // and https://jarvis-west.dc.ad.msft.net/?page=documents&section=75a7fd41-e140-46d4-9b87-516f42b2b646&id=0e680aa9-8e98-4501-8389-b4e12e0569e9#/
            IfxInitializer.IfxInitialize(initParams.IfxSession, initParams.IfxInstrumentSpec, initParams.IfxAuditSpec);
            if (!IfxInitializer.IfxEnabled)
            {
                IfxInitializer.IfxEnabled = true;
            }

#if DEBUG
            _etwTraceLogging = initParams.EmitEtwTraceloggingEvents;
#endif
        }

        private enum EventsExtended
        {
            /// <summary>Exeption event</summary>
            Exception = EventType._max,

            /// <summary>A pair of events sharing the same name prefix with a "Begin/End" suffix implicitly marks boundaries of an event tracing activity.
            /// These activities can be automatically picked up by debugging and profiling tools, which can compute their execution time, child activities, and other statistics.</summary>
            RequestBegin,

            /// <summary>A pair of events sharing the same name prefix with a "Begin/End" suffix implicitly marks boundaries of an event tracing activity.
            /// These activities can be automatically picked up by debugging and profiling tools, which can compute their execution time, child activities, and other statistics.</summary>
            RequestEndOk,

            /// <summary>A pair of events sharing the same name prefix with a "Begin/End" suffix implicitly marks boundaries of an event tracing activity.
            /// These activities can be automatically picked up by debugging and profiling tools, which can compute their execution time, child activities, and other statistics.</summary>
            RequestEndError,

            /// <summary>Service event raised by OwinCommunicationListener</summary>
            ServiceEventInfo,

            /// <summary>Service event raised by OwinCommunicationListener</summary>
            ServiceEventError,

            /// <summary>private use only: must be the last defined element</summary>
            _max
        }

        /// <summary>Get the singleton</summary>
        public static Agent Instance => _instance;

        /// <summary>Create and initialize the singleton</summary>
        public static void CreateInstance(InitParams initParams)
        {
            if (_instance != null)
            {
                throw new InvalidOperationException("ServiceEventSource already initialized");
            }

            _instance = new Agent(initParams);
        }

        /// <summary>Log an info/trace/warning event</summary>
        public void LogEvent(EventType eventType, string message)
        {
            LogEvent(eventType, message, new { Message = message });
        }

        /// <summary>
        /// Log an info/trace/warning event.
        /// </summary>
        /// <typeparam name="T">
        /// The type of data to be written to ETW (for debug mode only).
        /// This type must be an anonymous type or marked with the System.Diagnostics.Tracing.EventSourceAttribute
        /// attribute.
        /// </typeparam>
        /// <param name="eventType">The type of <see cref="EventType"/>.</param>
        /// <param name="message">The log message.</param>
        /// <param name="data">
        /// The data to be written to ETW (for debug mode only).
        /// This type must be an anonymous type or marked with the System.Diagnostics.Tracing.EventDataAttribute
        /// attribute.
        /// </param>
        /// <param name="tagId">The source of the event.</param>
        /// <param name="trackingId">The tracking ID to track the logs end to end.</param>
        /// <param name="timestamp">The timestamp to be written (in UTC).</param>
        public void LogEvent<T>(
            EventType eventType, 
            string message, 
            T data, 
            string tagId = null, 
            string trackingId = null, 
            DateTime? timestamp = null)
        {
            // TODO: Consider adding validation for attributes on the type parameter T and the parameter 'data' in lieu of 
            // the constraints defined in the parameter definitions of the docstring.
            if ((int)eventType >= (int)EventType._max)
            {
                throw new InvalidOperationException($"EventType={eventType.ToString()} is not supported by this method");
            }

            var eventName = GetEventName(eventType);
            DateTime eventTime = timestamp ?? DateTime.UtcNow;
            var partC = new AzLocIfx.AzLocTraceEventSchema()
            {
                env_time = eventTime.ToString(),
                message = message,
                eventType = eventName ?? String.Empty,
                tagId = tagId,
                processName = _processName ?? String.Empty,
                TrackingId = trackingId ?? String.Empty
            };

            IfxEvent.Log<AzLocIfx.AzLocTraceEventSchema>(partC, ToIfxTracingLevel(eventType));

#if DEBUG
            WriteEtwEvent(eventName, data);
#endif
        }

        /// <summary>
        /// Logs childern of an aggregate exception individually.
        /// </summary>
        public void LogException(AggregateException ex, string exceptionSource, string exceptionType = null, string trackingId = null)
        {
            if (ex != null)
            {
                foreach (var innerException in ex.InnerExceptions)
                {
                    Agent.Instance.LogException(innerException, exceptionSource, exceptionType, trackingId);
                }
            }
        }

        /// <summary>Log an exception</summary>
        /// <param name="ex">Puts exception class, callstack and message into the log.  Can't be null</param>
        /// <param name="tagId">A string representing what component generated the exception (for example, proxy vs billing), useful for log querying.  Can be null, but try to fill it in!</param>
        /// <param name="exceptionType">A string representing the broad categorization of the exception (for example, client input vs server error), useful for log querying and alerting. 
        /// This should be a subcategorization of exceptionSource, so it may not always be applicable if there are no interesting subcategories.  It can be null, but that can limit the usefulness of logs.</param>
        /// <param name="trackingId">A string representing a tracking id where the exception occurred (proxy has a tracking id for each operation), useful for joining log tables.  Can be null.</param>
        public void LogException(Exception ex, string tagId, string exceptionType = null, string trackingId = null)
        {
            var partC = new AzLocIfx.ExceptionSchema()
                        {
                            message = ex.Message,
                            exception = ex.GetType().ToString(),
                            call_stack = (string.IsNullOrEmpty(ex.ToString())) ? "no callstack" : ex.ToString(),
                            exceptionType = exceptionType ?? String.Empty,
                            tagId = tagId ?? String.Empty,
                            TrackingId = trackingId ?? String.Empty,
                            processName = _processName ?? String.Empty
            };

            IfxEvent.Log<AzLocIfx.ExceptionSchema>(partC, IfxTracingLevel.Error);

#if DEBUG
            WriteEtwEvent(GetEventName(EventsExtended.Exception), new ExceptionEventData(partC.exception, partC.message, partC.call_stack));
#endif
        }

        /// <summary>Custom call to record service registration</summary>
        public void LogServiceTypeRegistered(int processId, string typeName)
        {
            var partC = new AzLocIfx.ServiceRegisterSchema()
                        {
                            process_id = processId.ToString(),
                            type_name = typeName
                        };
            IfxEvent.Log<AzLocIfx.ServiceRegisterSchema>(partC, IfxTracingLevel.Informational);

#if DEBUG
            WriteEtwEvent(GetEventName(EventType.Information), new { ProcessId = processId, TypeName = typeName });
#endif
        }

        /// <summary>Custom call to record service events</summary>
        public void LogServiceEvent(ServiceEventData ed)
        {
            var extEvent = ed.IsError
                               ? EventsExtended.ServiceEventError
                               : EventsExtended.ServiceEventInfo;
            var eventName = GetEventName(extEvent);

            if (ed.IsError)
            {
                var partC = ed.ToServiceErrorPartC();
                IfxEvent.Log<AzLocIfx.ServiceErrorSchema>(partC, IfxTracingLevel.Critical);
            }
            else
            {
                var partC = ed.ToServiceInfoPartC();
                IfxEvent.Log<AzLocIfx.ServiceInfoSchema>(partC, IfxTracingLevel.Informational);
            }

#if DEBUG
            WriteEtwEvent(eventName, ed);
#endif
        }

        public void BillingEmitLogEvent(string partitionId, string batchId)
        {
            var billEmit = new AzLocBillIfx.BillEmit()
            {
                PartitionId = partitionId,
                BatchId = batchId
            };

            IfxEvent.Log<AzLocBillIfx.BillEmit>(billEmit);

#if DEBUG

            WriteEtwEvent(EventType.Information,
                                         new
                                         {
                                             Operation = "BillEmit",
                                             PartitionId = billEmit.PartitionId, 
                                             BatchId = billEmit.BatchId
                                         });
#endif
        }

#if DEBUG
        public void WriteEtwEvent<T>(string eventName, T data)
        {
            if (_etwTraceLogging && base.IsEnabled())
            {
                base.Write(eventName, data);
            }
        }

        [NonEvent]
        public void WriteEtwEvent<T>(EventType t, T data)
        {
            WriteEtwEvent(GetEventName(t), data);
        }
#endif

        /// <summary>@todo</summary>
        private static IfxTracingLevel ToIfxTracingLevel(EventType t)
        {
            switch (t)
            {
                case EventType.Information:
                    return IfxTracingLevel.Informational;
                case EventType.ProgramError:
                case EventType.UserError:
                    return IfxTracingLevel.Error;
                case EventType.Verbose:
                    return IfxTracingLevel.Verbose;
                case EventType.Warning:
                    return IfxTracingLevel.Warning;
                default:
                    throw new ArgumentException($"Unhandled EventType = '{t}' in switch statement");
            }
        }

        /// <summary>@todo</summary>
        private static string GetEventName(EventType t) => _sEventNames[(int)t];

        /// <summary>@todo</summary>
        private static string GetEventName(EventsExtended x) => _sEventNames[(int)x];

        /// <summary>@todo</summary>
        private static string[] BuildEventNameCollection()
        {
            var result = new string[(int)EventsExtended._max];

            int i = 0;
            for (; i < (int)EventType._max; i++)
            {
                result[i] = ((EventType)i).ToString();
            }

            for (; i < (int)EventsExtended._max; i++)
            {
                result[i] = ((EventsExtended)i).ToString();
            }

            //////////////////////////////////////////////////////////////////////////////////////////
            // verify compliance
            var unique = new HashSet<string>();
            for (i = 0; i < result.Length; i++)
            {
                var s = result[i];
                if (string.IsNullOrEmpty(s))
                {
                    throw new ArgumentException("Tracelogging event names cannot be null or empty");
                }

                if (!unique.Add(s.ToLowerInvariant()))
                {
                    throw new ArgumentException($"Tracelogging event names must be unique. Dup entry is '{s}'");
                }
            }

            return result;
        }
    }
}
