﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Instrumentation.Apim
{
    using System;
    using static System.String;
    using Microsoft.Cloud.InstrumentationFramework;
    using Newtonsoft.Json.Linq;

    public class ServiceApimOperation : OperationAgent<AzLocApimIfx.ApimOperation>
    {
        public ServiceApimOperation(string operationName)
            : base(operationName)
        {
            this.Operation.ApiType = OperationApiType.ClientProxy;

            this.Operation.PartC = new AzLocApimIfx.ApimOperation()
            {
                LogType = Empty,
                SubscriptionId = Empty,
                TrackingId = Empty,
                StartTime = Empty,
                Client_APIM = Empty,
                APIM_Proxy = Empty,
                Request = Empty,
                EndTime = Empty,
                APIM_ms = Empty,
                Response = Empty,
                APIMProduct = Empty,
                APIMOperation = Empty,
                APIMApi = Empty,
                APIMLastError = Empty,
                APIMInstanceName = Empty,
                DeploymentRegion = Empty,
                ResponseStatusCode = Empty
            };
        }

        public void LogRequest(RequestInfo requestInfo)
        {
            this.Operation.ResourceId = this.GetResourceId(requestInfo.User);

            this.Operation.PartC = new AzLocApimIfx.ApimOperation()
            {
                LogType = "Request",
                SubscriptionId = requestInfo.Subscription.Id ?? string.Empty,
                TrackingId = requestInfo.TrackingId,
                StartTime = requestInfo.StartTime.ToString("o"),
                Client_APIM = requestInfo.Request.OriginalUrl,
                APIM_Proxy = requestInfo.Request.Url,
                Request = requestInfo.Request.Content,
                APIMProduct = ((JObject)JToken.FromObject(requestInfo.Product)).ToString(),
                APIMOperation = ((JObject)JToken.FromObject(requestInfo.Operation)).ToString(),
                APIMApi = requestInfo.Api,
                APIMInstanceName = requestInfo.Deployment.ServiceName,
                DeploymentRegion = requestInfo.Deployment.Region
            };

            this.Operation.SetResult(OperationResult.Success);
        }

        public void LogResponse(ResponseInfo responseInfo)
        {
            this.Operation.PartC = new AzLocApimIfx.ApimOperation()
            {
                LogType = "Response",
                TrackingId = responseInfo.TrackingId,
                EndTime = responseInfo.StopTime.ToString("o"),
                APIM_ms = responseInfo.Duration.ToString(),
                APIMLastError = responseInfo.LastError,
                Response = ((JObject)JToken.FromObject(responseInfo.Response)).ToString(),
                APIMOperation = ((JObject)JToken.FromObject(responseInfo.Operation)).ToString(),
                DeploymentRegion = responseInfo.Deployment.Region,
                ResponseStatusCode = responseInfo.Response.statusCode.ToString()
            };

            var result = APIMLoggerHelper.GetResultFromStatusCode(responseInfo);

            switch (result)
            {
                case "Success":
                    this.Operation.SetResult(OperationResult.Success);
                    break;
                case "ClientError":
                case "APIMError":
                    this.Operation.SetResult(OperationResult.ClientError);
                    break;
                default:
                    this.Operation.SetResult(OperationResult.Failure);
                    break;
            }
        }

#if DEBUG
        protected override void WriteEtw()
        {
            var partC = this.Operation.PartC;

            if (Agent.Instance != null)
            {
                Agent.Instance.WriteEtwEvent(Instrumentation.EventType.Information,
                                             new
                                                 {
                                                     Source = "ETW",
                                                     LogType = partC.LogType,
                                                     SubscriptionId = partC.SubscriptionId,
                                                     TrackingId = partC.TrackingId,   
                                                     StartTime = partC.StartTime,
                                                     EndTime = partC.EndTime,
                                                     APIM_ms = partC.APIM_ms,
                                                     Client_APIM = partC.Client_APIM,
                                                     APIM_Proxy = partC.APIM_Proxy,
                                                     Request = partC.Request,
                                                     Response = partC.Response,
                                                     APIMProduct = partC.APIMProduct,
                                                     APIMOperation = partC.APIMOperation,
                                                     APIMApi = partC.APIMApi,
                                                     APIMLastError = partC.APIMLastError,
                                                     APIMInstanceName = partC.APIMInstanceName,
                                                     DeploymentRegion = partC.DeploymentRegion
                                             });
            }
        }
#endif

        private string GetResourceId(User userInfo)
        {
            var azureSubscriptionId = userInfo.AccountResourceFirstName;

            var tokens = userInfo.AccountResourceLastName?.Split(new[] { "::" }, StringSplitOptions.RemoveEmptyEntries);

            if (tokens?.Length == 2)
            {
                return $"/subscriptions/{azureSubscriptionId}/resourceGroups/{tokens[0]}/providers/Microsoft.Maps/accounts/{tokens[1]}";
            }

            return string.Empty;
        }
    }
}