﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Instrumentation
{
    using System;
    using System.Diagnostics.Tracing;
    using System.IO;
    using Microsoft.Cloud.InstrumentationFramework;

    public sealed class InitParams
    {
        public readonly string IfxSession;
        public readonly string MdmAccount;
        internal static readonly EventSourceSettings sourceSettings = EventSourceSettings.ThrowOnEventWriteErrors | EventSourceSettings.EtwSelfDescribingEventFormat;
        private string _eventSourceName;

        /// <summary>Constructor</summary>
        /// <param name="ifxSessionName">same as IfxEvents.sessionNamee in the MDS XML</param>
        /// <param name="mdmAccountName">same as AgentMetrics.mdmMonitoringAccount in the MDS XML</param>
        public InitParams(string ifxSessionName, string mdmAccountName)
        {
            if (string.IsNullOrEmpty(ifxSessionName))
            {
                throw new ArgumentException("ifxSessionName cannot be null or empty");
            }

            if (string.IsNullOrEmpty(mdmAccountName))
            {
                throw new ArgumentException("mdmAccountName cannot be null or empty");
            }

            IfxSession = ifxSessionName;
            MdmAccount = mdmAccountName;
        }

        public string IfxTraceDirectory { get; private set; }

        public string[] EtwTelemetryTraits { get; private set; }

        public bool EmitEtwTraceloggingEvents => !string.IsNullOrEmpty(_eventSourceName) && EtwTelemetryTraits != null;

        public string EventSourceName => string.IsNullOrEmpty(_eventSourceName)
            ? "LocationServicesDefault"
            : _eventSourceName;

        internal AuditSpecification IfxAuditSpec => new AuditSpecification { heartbeatSpan = TimeSpan.FromMinutes(1) };

        internal InstrumentationSpecification IfxInstrumentSpec => new InstrumentationSpecification
        {
#if DEBUG
                EmitIfxMetricsEvents = false,
#else
            EmitIfxMetricsEvents = true,
#endif
            MaxSizeCompletedTraceFilesInMb = 50,
            MaxSizeTraceFileInKb = 20,
            MonitoringAccountName = MdmAccount,
            SamplingEnabled = true,
            TraceDirectory = IfxTraceDirectory,
            WriteIfxTracerDiskLogs = !string.IsNullOrEmpty(IfxTraceDirectory)
        };

        /// <summary>Write events locally in addition to sending them to Geneva</summary>
        /// <param name="traceDirectory">full path where to log.  If empty or null, events will only be sent to Geneva</param>
        public void SetLocalTracing(string traceDirectory)
        {
            IfxTraceDirectory = traceDirectory;

            if (!string.IsNullOrEmpty(IfxTraceDirectory) && !Directory.Exists(IfxTraceDirectory))
            {
                Directory.CreateDirectory(IfxTraceDirectory);
            }
        }

        /// <summary>Optionally, use TraceLogging (EventSource v2) instead of IfxTracing</summary>
        /// <param name="eventSourceName"></param>
        /// <param name="etwProviderGuid">same as EtwProvider.guid in the MDS XML</param>
        public void SetEtwProvider(string eventSourceName, string etwProviderGuid)
        {
            if (string.IsNullOrEmpty(eventSourceName) != string.IsNullOrEmpty(etwProviderGuid))
            {
                throw new ArgumentException("Both parameters must be empty or non-empty");
            }

            _eventSourceName = eventSourceName;
            EtwTelemetryTraits = null;
            if (!string.IsNullOrEmpty(etwProviderGuid))
            {
                // see example under "ETW Provider Groups" at https://jarvis-west.dc.ad.msft.net/?page=documents&section=9c95f4eb-8689-4c9f-81bf-82d688e860fd&id=19e3200f-aa43-42e1-84dd-2bd2105151d1
                //if (!etwProviderGuid.StartsWith("{") && !etwProviderGuid.EndsWith("}"))
                //{
                //    etwProviderGuid = "{" + etwProviderGuid + "}";
                //}
                EtwTelemetryTraits = new string[] { "ETW_GROUP", etwProviderGuid };
            }
        }
    }
}
