﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace Microsoft.Azure.LocationServices.Instrumentation
{
    using System;

    using Microsoft.Cloud.InstrumentationFramework;

    public abstract class OperationAgent<T> : IOperationAgent
        where T : global::Ifx.OperationSchema, new()
    {
        protected ExtendedOperation<T> Operation;

        // sigh.  going to duplicate some fields from Operation here, so that ETW can report them.  (Operation hides them)
        protected string _operationName;
        protected OperationResult _operationResult;
        protected string _resultSignature;
        protected string _resultDetails;
        protected Exception _exception;

        // Flag: Has Dispose already been called?
        private bool disposed = false;

        protected OperationAgent(string operationName)
        {
            this.Operation = new ExtendedOperation<T>(operationName);

            // assume client error unless set otherwise
            this.Operation.SetResult(OperationResult.ClientError);

            this._operationName = operationName;
        }

        ~OperationAgent()
        {
            Dispose(false);
        }

        /// <summary>
        /// Although supported by Operation, consider using SetResultFailure instead.
        /// This call appears to just write the type and message in the resultSignature and resultDescription fields.
        /// Instead, SetResultFailure can write some or all of the callstack
        /// </summary>
        /// <param name="ex"></param>
        public void SetException(Exception ex)
        {
            this.Operation.SetResult(ex);
            this._exception = ex;
        }

        public void SetResultSuccess()
        {
            this.Operation.SetResult(OperationResult.Success);
            this._operationResult = OperationResult.Success;
        }

        public void SetResultClientError(string resultSignature, string resultDetails)
        {
            this.Operation.SetResult(OperationResult.ClientError, resultSignature, resultDetails);
            this._operationResult = OperationResult.ClientError;
            this._resultSignature = resultSignature;
            this._resultDetails = resultDetails;
        }

        public void SetResultFailure(string resultSignature, string resultDetails)
        {
            this.Operation.SetResult(OperationResult.Failure, resultSignature, resultDetails);
            this._operationResult = OperationResult.Failure;
            this._resultSignature = resultSignature;
            this._resultDetails = resultDetails;
        }

        // Public implementation of Dispose pattern callable by consumers.
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        // Protected implementation of Dispose pattern.
        protected virtual void Dispose(bool disposing)
        {
            if (disposed)
            {
                return;
            }

            if (disposing)
            {
                // Free any other managed objects here.
                //
#if DEBUG
                WriteEtw();
#endif
                this.Operation.Dispose();
                this.Operation = null;
            }

            // Free any unmanaged objects here.
            //
            disposed = true;
        }

#if DEBUG
        protected abstract void WriteEtw();
#endif
    }
}