﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Instrumentation
{
    using System;
    using System.Diagnostics.Tracing;
    using System.Fabric;

    [EventData] // [EventData] makes it possible to pass an instance of the class as an argument to EventSource.Write()
    public class ServiceEventData
    {
        public ServiceEventData(ServiceContext ctxt, string message)
        {
            if (string.IsNullOrEmpty(message))
            {
                throw new ArgumentException("ServiceEventData message param cannot be null or empty");
            }

            ServiceName = ctxt.ServiceName.ToString();
            ServiceTypeName = ctxt.ServiceTypeName;
            ReplicaOrInstanceId = GetReplicaOrInstanceId(ctxt);
            PartitionId = ctxt.PartitionId;
            ApplicationName = ctxt.CodePackageActivationContext.ApplicationName;
            ApplicationTypeName = ctxt.CodePackageActivationContext.ApplicationTypeName;
            NodeName = ctxt.NodeContext.NodeName;
            Message = message;
            ExceptionType = string.Empty;
            ExceptionCallStack = string.Empty;
        }

        public ServiceEventData(ServiceContext ctxt, Exception ex, string messagePrefix = null)
            : this(ctxt, "Error")
        {
            var exData = new ExceptionEventData(ex, messagePrefix);
            Message = exData.Message;
            ExceptionType = exData.Type;
            ExceptionCallStack = exData.CallStack;
        }

        public string ServiceName { get; private set; }

        public string ServiceTypeName { get; private set; }

        public long ReplicaOrInstanceId { get; private set; }

        public Guid PartitionId { get; private set; }

        public string ApplicationName { get; private set; }

        public string ApplicationTypeName { get; private set; }

        public string NodeName { get; private set; }

        public string Message { get; private set; }

        public string ExceptionType { get; set; }

        public string ExceptionCallStack { get; set; }

        internal bool IsError => !string.IsNullOrEmpty(ExceptionType);

        internal AzLocIfx.ServiceInfoSchema ToServiceInfoPartC()
        {
            if (IsError)
            {
                throw new InvalidOperationException("ServiceEvent: error to info conversion is disallowed");
            }

            return new AzLocIfx.ServiceInfoSchema()
                   {
                       service_name = ServiceName,
                       service_type = ServiceTypeName,
                       replica_or_instance_id = ReplicaOrInstanceId,
                       partition_id = PartitionId.ToString(),
                       application_name = ApplicationName,
                       application_type = ApplicationTypeName,
                       node_name = NodeName,
                       message = Message
                   };
        }

        internal AzLocIfx.ServiceErrorSchema ToServiceErrorPartC()
        {
            if (!IsError)
            {
                throw new InvalidOperationException("ServiceEvent: info to error conversion is disallowed");
            }

            return new AzLocIfx.ServiceErrorSchema()
                   {
                       service_name = ServiceName,
                       service_type = ServiceTypeName,
                       replica_or_instance_id = ReplicaOrInstanceId,
                       partition_id = PartitionId.ToString(),
                       application_name = ApplicationName,
                       application_type = ApplicationTypeName,
                       node_name = NodeName,
                       message = Message,
                       exception = ExceptionType,
                       call_stack = ExceptionCallStack
                   };
        }

        private static long GetReplicaOrInstanceId(ServiceContext context)
        {
            StatelessServiceContext stateless = context as StatelessServiceContext;
            if (stateless != null)
            {
                return stateless.InstanceId;
            }

            StatefulServiceContext stateful = context as StatefulServiceContext;
            if (stateful != null)
            {
                return stateful.ReplicaId;
            }

            throw new NotSupportedException("Context type not supported.");
        }
    }
}
