﻿Param
(
    [Parameter(Mandatory=$true)]
	[ValidateSet("u", "i")]
		[string]$mode = "m",
    [Parameter(Mandatory=$true)]
		[string]$profileFile = "f"
)

$modulePath = $PSScriptRoot + ".\..\..\..\..\..\..\DeployAndSetup\Utils\lbsSetupSupport.psd1"
Import-Module -Name $modulePath -Force

$ErrorActionPreference = "Stop"

if (($mode -ne "u") -and ($mode -ne "i")) {
	Write-Error "Invalid mode '$mode'. Use '-i' to deploy and '-u' to uninstall."
	Exit 
}

$localFolder = (Split-Path $MyInvocation.MyCommand.Path)
$appPackagePath = "$localFolder\..\MonitoringServiceWithMA"
$appPackagePath = Resolve-Path $appPackagePath

$appManifestXml = $appPackagePath.ToString()+"\ApplicationManifest.xml"
[xml]$serviceManifest = Get-Content $appManifestXml
[string]$appTypeName = $serviceManifest.ApplicationManifest.ApplicationTypeName
[string]$version = $serviceManifest.ApplicationManifest.ApplicationTypeVersion

$publishProfile = Read-PublishProfile $profileFile
$ClusterConnectionParameters = $publishProfile.ClusterConnectionParameters
if ($SecurityToken)
{
	$ClusterConnectionParameters["SecurityToken"] = $SecurityToken
}

[void](Connect-ServiceFabricCluster @ClusterConnectionParameters)

try
{
	Write-Output "Uninstalling (if exists) app = $appTypeName ver = $version"

	Remove-ServiceFabricService -ServiceName fabric:/Monitoring/GenevaAgent -Force
	Remove-ServiceFabricService -ServiceName fabric:/Monitoring/FabricHealth -Force
	Remove-ServiceFabricApplication -ApplicationName fabric:/Monitoring -Force
	Unregister-ServiceFabricApplicationType -ApplicationTypeName $appTypeName -ApplicationTypeVersion $version -Force
}
catch
{
	if ($mode -ne "i") {
		Write-Error $_.Exception.ToString()
	}
}

if ($mode -ne "i") {
	Exit
}

# else, DEPLOY
$gcsExactVersion = $false
$appParamsFromFile = Get-ApplicationParametersFromApplicationParameterFile $publishProfile.ApplicationParameterFile
Write-Host "AppParamsFromFile: "
foreach ($h in $appParamsFromFile.GetEnumerator()) {
	Write-Host "  - $($h.Name): $($h.Value)"  -foregroundcolor yellow -backgroundcolor black
    if ($h.Name.ToLowerInvariant() -eq "gcsconfigversion") {
        $gcsConfigVersion = $h.Value
    }
    if ($h.Name.ToLowerInvariant() -eq "gcsexactversion") {
        if ($h.Value.ToLowerInvariant() -eq 'true') {
            $gcsExactVersion = $true
        }
    }
}

$minReplicaSetSize = "error" 
$targetReplicaSetSize = "error"
$clusterManifestXml = [xml](Get-ServiceFabricClusterManifest)
$clusterManager = $clusterManifestXml.ClusterManifest.FabricSettings.Section | where { ($_.Name -eq "ClusterManager") }
foreach( $n in $clusterManager.ChildNodes) 
{
	if ($n.Name -eq "MinReplicaSetSize") { $minReplicaSetSize = $n.Value }
	if ($n.Name -eq "TargetReplicaSetSize") { $targetReplicaSetSize = $n.Value }
}

# when running on a dev box with 5 nodes, set -InstanceCount to "1" to avoid multiple agents trying to report the same Dynamic Events
$instanceCount = "-1"
if (($profileFile.ToLower().EndsWith("local.xml")) -and ($targetReplicaSetSize -ne "1")) {
	$instanceCount = "1"
}
else {
	$instanceCount = "-1"
}

Write-Host "ServiceFabricServiceParams"
Write-Host " - Stateful Service"
Write-Host "     MinReplicaSetSize = $minReplicaSetSize" -foregroundcolor yellow -backgroundcolor black
Write-Host "     TargetReplicaSetSize = $targetReplicaSetSize" -foregroundcolor yellow -backgroundcolor black
Write-Host " - Stateless Service:"
Write-Host "     InstanceCount = $instanceCount" -foregroundcolor yellow -backgroundcolor black
Write-Host "Deploying app=$appTypeName ver=$version"

if ($gcsExactVersion) {
    $msg = "Deploying to pick up a Geneva config change? Does GCSConfigVersion '" + $gcsConfigVersion + "' match the latest Geneva version in Jarvis?";
    Debug-lbsPromptToContinue -message $msg;
}

Copy-ServiceFabricApplicationPackage                                                                                  `
	-ApplicationPackagePath $appPackagePath                                                                           `
	-ImageStoreConnectionString (Get-ImageStoreConnectionStringFromClusterManifest(Get-ServiceFabricClusterManifest)) `
	-ApplicationPackagePathInImageStore MA

Register-ServiceFabricApplicationType -ApplicationPathInImageStore MA

Write-Host "-----------------------------------------------------------------------"
Write-Host "  New-ServiceFabricApplication: $appTypeName v. $version               "
Write-Host "-----------------------------------------------------------------------"
New-ServiceFabricApplication            `
	-ApplicationName fabric:/Monitoring `
	-ApplicationTypeName $appTypeName   `
	-ApplicationTypeVersion $version    `
	-ApplicationParameter $appParamsFromFile

Write-Host "-----------------------------------------------------------------------"
Write-Host "  New-ServiceFabricService (stateful) fabric:/Monitoring/FabricHealth  "
Write-Host "-----------------------------------------------------------------------"
New-ServiceFabricService                         `
	-ApplicationName fabric:/Monitoring 		 `
	-MinReplicaSetSize $minReplicaSetSize		 `
	-TargetReplicaSetSize $targetReplicaSetSize	 `
	-PartitionSchemeSingleton 					 `
	-ServiceName fabric:/Monitoring/FabricHealth `
	-ServiceTypeName FabricMonitoringServiceType `
	-Stateful 									 `
	-Correlation "fabric:/System/ClusterManagerService Affinity"

Write-Host "-----------------------------------------------------------------------"
Write-Host "  New-ServiceFabricService (stateless) fabric:/Monitoring/GenevaAgent  "
Write-Host "-----------------------------------------------------------------------"
New-ServiceFabricService                         `
	-ApplicationName fabric:/Monitoring 		 `
	-InstanceCount $instanceCount				 `
	-PartitionSchemeSingleton 					 `
	-ServiceName fabric:/Monitoring/GenevaAgent  `
	-ServiceTypeName FabricMdsAgentServiceType 	 `
	-Stateless

Write-Host "**************************************************************************" -foregroundcolor black -backgroundcolor green
Write-Host "*      Deployment completed successfully                                 *" -foregroundcolor black -backgroundcolor green
Write-Host "**************************************************************************" -foregroundcolor black -backgroundcolor green
#Get-ServiceFabricClusterManifest
