﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace Microsoft.Azure.LocationServices
{
    using System.IO;
    using System.Threading.Tasks;
    using Validation;

    /// <summary>
    /// Implements a csv reader based on a text reader.
    /// </summary>
    public class CsvFormatReader : ICsvFormatReader
    {
        private readonly TextReader reader;

        private readonly char separator;

        public CsvFormatReader(TextReader reader, char separator)
        {
            Requires.NotNull(reader, nameof(reader));
            this.reader = reader;
            this.separator = separator;
        }

        /// <summary>
        /// Returns the next set of fields.
        /// </summary>
        public async Task<string[]> ReadNextAsync()
        {
            string line;

            do
            {
                line = await this.reader.ReadLineAsync();
                if (line == null)
                {
                    break;
                }
            }
            while (ShouldSkipLine(line)); 

            return line?.Split(this.separator);
        }

        /// <summary>
        /// Returns the next set of fields.
        /// </summary>
        public string[] ReadNext()
        {
            string line;

            do
            {
                line = this.reader.ReadLine();
                if (line == null)
                {
                    break;
                }
            }
            while (ShouldSkipLine(line)); // skipping comments and empty lines

            return line?.Split(this.separator);
        }

        /// <summary>
        /// Dispose current instance.
        /// </summary>
        public void Dispose()
        {
            this.reader?.Dispose();
        }

        private static bool ShouldSkipLine(string line)
        {
            return string.IsNullOrEmpty(line) || line.TrimStart().StartsWith("#"); // skipping comments and empty lines
        }
    }
}