﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace Microsoft.Azure.LocationServices.GeoLocation
{
    using System.Collections.Generic;
    using System.Net;

    /// <summary>
    /// Implements the logic for IP address lookup
    /// </summary>
    public class IPLookup : IIPLookup
    {
        private readonly SortedList<byte[], NetworkAssignment> networks = new SortedList<byte[], NetworkAssignment>(ByteArrayComparer.Default);

        /// <summary>
        /// Creates new instance of IPLookup service.
        /// </summary>
        public IPLookup(IEnumerable<NetworkAssignment> assignments)
        {
            foreach (var item in assignments)
            {
                this.networks.Add(item.Network.StartAddress.MapToIPv6().GetAddressBytes(), item);
            }
        }
        
        /// <summary>
        /// Resolved IP address to a geo location
        /// </summary>
        /// <param name="address">IPAddress</param>
        /// <returns>Result</returns>
        public NetworkAssignmentSearchResult GetNetworkAssignment(IPAddress address)
        {
            var ipv6Mapped = address.MapToIPv6();
            var assignment = this.FindNearestNetwork(ipv6Mapped.GetAddressBytes());

            var result = new NetworkAssignmentSearchResult { IPAddress = address.ToString() };

            if (assignment != null && assignment.Network.ContainsAddress(address))
            {
                result.Country = new NetworkAssignmentSearchResultCountry { IsoCode = assignment.IsoCountryCode };
            }

            return result;
        }

        private NetworkAssignment FindNearestNetwork(byte[] key)
        {
            var index = this.networks.BinarySearch(key);

            if (index >= 0)
            {
                return this.networks.Values[index];
            }

            // get the index of the first element larger than the value
            index = ~index;

            // index now points at the first item greater than the lookup key
            // if the index value is zero, it means that our lookup key is not in the list
            // otherwise, return the previous item. if the lookup key is greater than every item in the list,
            // it may still be contained in the very last IP range
            if (index > 0)
            {
                return this.networks.Values[index - 1];
            }

            // provided key values doesn't belong to the current dataset
            return null;
        }
    }
}