﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace Microsoft.Azure.LocationServices
{
    using System;
    using System.Net;
    using System.Net.Sockets;
    using System.Numerics;

    /// <summary>
    /// Implements a structure that represent a range of IP addresses.
    /// </summary>
    public struct IPNetwork
    {
        public IPNetwork(IPAddress startAddress, ulong nodeNumber)
        {
            this.StartAddress = startAddress;
            this.NodeNumber = nodeNumber;
            this.CidrPrefix = 0;
        }

        public IPNetwork(IPAddress startAddress, byte cidrPrefix)
        {
            var addressInteger = AsBigInteger(startAddress);
            var mask = GetMask(cidrPrefix, GetAddressLengthBits(startAddress.AddressFamily));

            if (addressInteger != (mask & addressInteger))
            {
                throw new ArgumentException("Node number in start address must be equal to zero.", nameof(startAddress));
            }

            this.StartAddress = startAddress;
            this.CidrPrefix = cidrPrefix;
            this.NodeNumber = 0;
        }

        public IPAddress StartAddress { get; }

        public ulong NodeNumber { get; }

        public byte CidrPrefix { get; }

        public override string ToString()
        {
            if (this.CidrPrefix == 0)
            {
                return $"{this.StartAddress}[{this.NodeNumber}]";
            }

            var addr = this.StartAddress.IsIPv4MappedToIPv6
                           ? this.StartAddress.MapToIPv4()
                           : this.StartAddress.MapToIPv6();

            return $"{addr}/{this.CidrPrefix}";
        }

        public bool ContainsAddress(IPAddress address)
        {
            // Need to reverse address bytes because BigInteger expects the input in little-endian order
            BigInteger startAddress = AsBigInteger(this.StartAddress);
            BigInteger inputAddress = AsBigInteger(address);

            if (this.CidrPrefix > 0)
            {
                var mask = GetMask(this.CidrPrefix, GetAddressLengthBits(address.AddressFamily));
                return (inputAddress & mask) == startAddress;
            }

            return (startAddress <= inputAddress) && (inputAddress < startAddress + this.NodeNumber);
        }
        
        private static BigInteger GetMask(byte cidrPrefix, int addressLengthBits)
        {
            BigInteger m = 1;
            return ((m << cidrPrefix) - 1) << (addressLengthBits - cidrPrefix);
        }

        private static int GetAddressLengthBits(AddressFamily addressFamily)
        {
            switch (addressFamily)
            {
                case AddressFamily.InterNetwork:
                    return 32;
                case AddressFamily.InterNetworkV6:
                    return 128;
            }

            throw new ArgumentException($"Address family '{Enum.GetName(typeof(AddressFamily), addressFamily)}' not supported.", nameof(addressFamily));
        }

        private static BigInteger AsBigInteger(IPAddress address)
        {
            return new BigInteger(ReverseBytes(address.GetAddressBytes()));
        }

        private static byte[] ReverseBytes(byte[] a)
        {
            // it is important to append target with zero byte 
            // to ensure that the number is always treated as positive
            byte[] b = new byte[a.Length + 1];
            Array.Copy(a, 0, b, 1, a.Length);
            Array.Reverse(b);
            return b;
        }
    }
}