﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace Microsoft.Azure.LocationServices.Iana
{
    using System;
    using System.Globalization;

    using Validation;

    /// <summary>
    /// The regional registry exchange reader.
    /// </summary>
    public class RirDatasetReader
    {
        private ICsvFormatReader csvReader;

        public string Version { get; private set; }

        public string Registry { get; private set; }

        public string Serial { get; private set; }

        public int RecordsNumber { get; private set; }

        public DateTimeOffset StartDate { get; private set; }

        public DateTimeOffset EndDate { get; private set; }

        public int NumberOfAsnRecords { get; private set; }

        public int NumberOfIpv4Records { get; private set; }

        public int NumberOfIpv6Records { get; private set; }

        public static RirDatasetReader FromStringReader(ICsvFormatReader reader)
        {
            Requires.NotNull(reader, nameof(reader));

            var datasetReader = new RirDatasetReader();

            string[] nextLine = reader.ReadNext();
            if (nextLine == null)
            {
                throw new InvalidOperationException("Unexpected end of stream. Dataset header is expected.");
            }

            datasetReader.ParseHeader(nextLine);
            datasetReader.csvReader = reader;

            // skip summary lines
            for (int i = 0; i < 3; i++)
            {
                var components = reader.ReadNext();
                if (components == null)
                {
                    throw new InvalidOperationException("Unexpected end of stream. Summary line is expected.");
                }

                if (!string.Equals(components[5], "summary", StringComparison.OrdinalIgnoreCase))
                {
                    throw new InvalidOperationException("Unknown line type when summary is expected.");
                }

                switch (components[2])
                {
                    case "asn":
                        datasetReader.NumberOfAsnRecords = int.Parse(components[4], NumberFormatInfo.InvariantInfo);
                        break;
                    case "ipv4":
                        datasetReader.NumberOfIpv4Records = int.Parse(components[4], NumberFormatInfo.InvariantInfo);
                        break;
                    case "ipv6":
                        datasetReader.NumberOfIpv6Records = int.Parse(components[4], NumberFormatInfo.InvariantInfo);
                        break;
                }
            }

            return datasetReader;
        }

        public RirRecord GetRecord()
        {
            var components = this.csvReader.ReadNext();
            if (components == null)
            {
                return null;
            }

            Enum.TryParse(components[2], true, out RirRecordType recordType);
            Enum.TryParse(components[6], true, out RirRecordStatus recordStatus);
            return new RirRecord(components[0], components[1], recordType, components[3], components[4], components[5], recordStatus);
        }
        
        private void ParseHeader(string[] headerComponents)
        {
            Requires.NotNull(headerComponents, nameof(headerComponents));
            Requires.Range(headerComponents.Length >= 6, nameof(headerComponents));

            this.Version = headerComponents[0];
            this.Registry = headerComponents[1];
            this.Serial = headerComponents[2];
            this.RecordsNumber = int.Parse(headerComponents[3]);

            TimeSpan offset = TimeSpan.Zero;
            if (!string.IsNullOrWhiteSpace(headerComponents[6]))
            {
                offset = TimeSpan.ParseExact(headerComponents[6], new[] { @"\0hhmm", @"\+hhmm", @"\-hhmm" }, CultureInfo.InvariantCulture, headerComponents[6][0] == '-' ? TimeSpanStyles.AssumeNegative : TimeSpanStyles.None);
            }

            if (!string.IsNullOrWhiteSpace(headerComponents[4]))
            {
                if (string.Equals(headerComponents[4], "00000000"))
                {
                    this.StartDate = DateTimeOffset.MinValue;
                }
                else
                {
                    this.StartDate = new DateTimeOffset(DateTime.ParseExact(headerComponents[4], new[] { @"yyyyMMdd" }, CultureInfo.InvariantCulture, DateTimeStyles.None), offset);
                }
            }

            if (!string.IsNullOrWhiteSpace(headerComponents[5]))
            {
                this.EndDate = new DateTimeOffset(DateTime.ParseExact(headerComponents[5], "yyyyMMdd", CultureInfo.InvariantCulture), offset);
            }
        }
    }
}
