﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.GeoLocation.Storage
{
    using System;
    using System.IO;
    using System.Threading;
    using System.Threading.Tasks;

    using Microsoft.WindowsAzure.Storage.Auth;
    using Microsoft.WindowsAzure.Storage.Blob;

    using Validation;

    /// <summary>
    /// Abstract blob storage for content copy operations.
    /// </summary>
    public class BlobContentProxy : IContentTarget, IContentSource
    {
        private readonly CloudBlockBlob client;

        /// <summary>
        /// Creates new instance of BlobContentProxy.
        /// </summary>
        public BlobContentProxy(Uri blobPath)
        {
            Requires.NotNull(blobPath, nameof(blobPath));

            this.client = new CloudBlockBlob(blobPath);
        }

        /// <summary>
        /// Creates new instance of BlobContentProxy.
        /// </summary>
        public BlobContentProxy(Uri blobPath, StorageCredentials storageCredentials)
        {
            Requires.NotNull(blobPath, nameof(blobPath));
            Requires.NotNull(storageCredentials, nameof(storageCredentials));

            this.client = new CloudBlockBlob(blobPath, storageCredentials);
        }

        /// <summary>
        /// Name of this source.
        /// </summary>
        public string Name => this.client.Uri.AbsolutePath;

        /// <summary>
        /// Sets the content for the target container.
        /// </summary>
        public Task SetTargetAsync(Stream contents, CancellationToken cancellationToken)
        {
            return this.client.UploadFromStreamAsync(contents, cancellationToken);
        }
        
        /// <summary>
        /// Get the content from the source container.
        /// </summary>
        public Task<Stream> GetSourceAsync(CancellationToken cancellationToken)
        {
            return this.client.OpenReadAsync(cancellationToken);
        }

        /// <summary>
        /// Gets the timestamp of last update.
        /// </summary>
        public async Task<DateTimeOffset> GetLastModifiedDateTime(CancellationToken cancellationToken)
        {
            await this.client.FetchAttributesAsync(cancellationToken);
            return this.client.Properties.LastModified ?? DateTimeOffset.MinValue;
        }

        /// <summary>
        /// Checks if the target container exists.
        /// </summary>
        public Task<bool> ExistsAsync(CancellationToken cancellationToken)
        {
            return this.client.ExistsAsync(cancellationToken);
        }
    }
}