﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.GeoLocation.Storage
{
    using System;
    using System.Threading;
    using System.Threading.Tasks;

    /// <summary>
    /// Implements a components that conditionally copies content from a source to the target 
    /// </summary>
    public class ConditionalContentMirror
    {
        private readonly IContentSource source;
        private readonly IContentTarget target;

        /// <summary>
        /// Creates a new instance of ConditionalContentMirror.
        /// </summary>
        public ConditionalContentMirror(IContentSource source, IContentTarget target)
        {
            this.source = source;
            this.target = target;
        }
        
        /// <summary>
        /// Defines a callback for condition test.
        /// </summary>
        /// <returns></returns>
        public delegate bool ShouldCopyDelegate(DateTimeOffset sourceTimestamp, DateTimeOffset targetTimestamp);

        /// <summary>
        /// Source used by this component.
        /// </summary>
        public IContentSource Source => this.source;

        /// <summary>
        /// Target used by this component.
        /// </summary>
        public IContentTarget Target => this.target;

        /// <summary>
        /// Performs a content copy if the condition is satisfied.
        /// </summary>
        public async Task<bool> CheckAndUpdateAsync(CancellationToken token)
        {
            if (!await this.target.ExistsAsync(token)
                || await this.source.GetLastModifiedDateTime(token) > await this.target.GetLastModifiedDateTime(token))
            {
                await this.target.SetTargetAsync(await this.source.GetSourceAsync(token), token);
                return true;
            }

            return false;
        }
    }
}