﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.GeoLocation
{
    using System;
    using System.Collections.Generic;
    using System.Threading;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices.GeoLocation.Storage;

    using Validation;

    /// <summary>
    /// Implements the conditional data refresh in the background.
    /// </summary>
    public class NetworkAssignmentCache
    {
        private readonly INetworkAssignmentStorage networkAssignmentStorage;
        private readonly TimeSpan refreshPeriod;
        private readonly IGeoLocationLogger logger;
        private string currentFingerprint = null;

        /// <summary>
        /// Create new instance of the NetworkAssignmentCache.
        /// </summary>
        public NetworkAssignmentCache(INetworkAssignmentStorage networkAssignmentStorage, TimeSpan refreshPeriod, IGeoLocationLogger logger)
        {
            Requires.NotNull(networkAssignmentStorage, nameof(networkAssignmentStorage));
            Requires.NotNull(logger, nameof(logger));

            this.networkAssignmentStorage = networkAssignmentStorage;
            this.refreshPeriod = refreshPeriod;
            this.logger = logger;
        }

        /// <summary>
        /// Periodically checks and reloads network assignemnts from storage.
        /// </summary>
        public async Task GetNetworkAssignmentsWhenUpdated(Action<IEnumerable<NetworkAssignment>> callback, CancellationToken cancellationToken)
        {
            do
            {
                try
                {
                    var latestFingerprint = await this.networkAssignmentStorage.FetchFingerprintAsync();

                    if (this.currentFingerprint == null || !string.Equals(this.currentFingerprint, latestFingerprint, StringComparison.OrdinalIgnoreCase))
                    {
                        var assignments = await this.networkAssignmentStorage.LoadAsync();
                        callback(assignments);
                        this.currentFingerprint = latestFingerprint;
                        this.logger.OnNetworkAssignmentsReloaded(latestFingerprint);
                    }

                    // delay only on successful execution
                    await Task.Delay(this.refreshPeriod, cancellationToken);
                }
                catch (TaskCanceledException ex)
                {
                    // this is expected from Task.Delay on service shutdown
                    if (!cancellationToken.IsCancellationRequested)
                    {
                        this.logger.OnNetworkAssignmentRefreshError(ex);
                    }
                }
                catch (Exception ex)
                {
                    this.logger.OnNetworkAssignmentRefreshError(ex);
                    await Task.Delay(TimeSpan.FromSeconds(1), cancellationToken);
                }
            }
            while (!cancellationToken.IsCancellationRequested);
        }
    }
}