﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.GeoLocation.Storage
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.IO.Compression;
    using System.Linq;
    using System.Threading.Tasks;

    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Auth;
    using Microsoft.WindowsAzure.Storage.Blob;

    using Newtonsoft.Json;

    using Validation;

    /// <summary>
    /// Implements blob storage for IP to country assignments.
    /// </summary>
    public class NetworkAssignmentStorage : INetworkAssignmentStorage
    {
        internal const string FingerprintPropertyName = "Fingerprint";
        private const string AssignmentsBlobName = "current.gzip";

        private readonly IGeoLocationLogger logger;
        private readonly ICloudBlob blobClient;
        private readonly JsonSerializer serializer;

        /// <summary>
        /// Creates new instance of NetworkAssignmentStorage.
        /// </summary>
        public NetworkAssignmentStorage(Uri containerPath, string sasToken, IGeoLocationLogger logger)
            : this(new CloudBlockBlob(GetBlobUri(containerPath, AssignmentsBlobName), new StorageCredentials(sasToken)), logger)
        {
        }

        /// <summary>
        /// Creates new instance of NetworkAssignmentStorage.
        /// </summary>
        public NetworkAssignmentStorage(Uri containerPath, StorageCredentials credentials, IGeoLocationLogger logger)
            : this(new CloudBlockBlob(new StorageUri(GetBlobUri(containerPath, AssignmentsBlobName)), null, credentials), logger)
        {
        }

        /// <summary>
        /// Creates new instance of NetworkAssignmentStorage.
        /// </summary>
        public NetworkAssignmentStorage(ICloudBlob blobClient, IGeoLocationLogger logger)
        {
            Requires.NotNull(blobClient, nameof(blobClient));
            Requires.NotNull(logger, nameof(logger));

            this.logger = logger;
            this.blobClient = blobClient;
            this.serializer = new JsonSerializer { Formatting = Formatting.None };
        }

        /// <summary>
        /// Loads dataset from a compressed storage blob.
        /// </summary>
        public async Task<NetworkAssignment[]> LoadAsync()
        {
            using (var stream = await this.blobClient.OpenReadAsync())
            {
                using (var textReader = new JsonTextReader(new StreamReader(new GZipStream(stream, CompressionMode.Decompress))))
                {
                    var records = this.serializer.Deserialize<NetworkAssignmentRecord[]>(textReader);
                    return records.Select(r => r.ToNetworkAssignment()).ToArray();
                }
            }
        }

        /// <summary>
        /// Get a dataset fingerprint from storage.
        /// </summary>
        public Task<string> FetchFingerprintAsync()
        {
            return this.FetchFingerprintAsync(null);
        }

        /// <summary>
        /// Get a dataset fingerprint from storage.
        /// </summary>
        public async Task<string> FetchFingerprintAsync(string leaseToken)
        {
            await this.blobClient.FetchAttributesAsync(
                string.IsNullOrEmpty(leaseToken) ? AccessCondition.GenerateEmptyCondition() : AccessCondition.GenerateLeaseCondition(leaseToken),
                new BlobRequestOptions(),
                new OperationContext());

            if (this.blobClient.Metadata.ContainsKey(FingerprintPropertyName))
            {
                return this.blobClient.Metadata[FingerprintPropertyName];
            }

            return null;
        }
        
        /// <summary>
        /// Updates network assignments in storage if the fingerprint is different.
        /// </summary>
        public async Task<bool> ReplaceIfFingerprintDoesntMatchAsync(IEnumerable<NetworkAssignment> assignments, string fingerprint, TimeSpan leaseTimeout)
        {
            string leaseToken = null;
            
            if (await this.blobClient.ExistsAsync())
            {
                leaseToken = await this.blobClient.AcquireLeaseAsync(leaseTimeout);
                
                if (string.Equals(fingerprint, await this.FetchFingerprintAsync(leaseToken), StringComparison.OrdinalIgnoreCase))
                {
                    return false;
                }
            }

            try
            {
                using (var buffer = new MemoryStream())
                {
                    using (var writer = new StreamWriter(new GZipStream(buffer, CompressionMode.Compress, true)))
                    {
                        this.serializer.Serialize(writer, assignments.Select(r => r.ToNetworkAssignmentRecord()));
                    }

                    this.blobClient.Metadata[FingerprintPropertyName] = fingerprint;
                    buffer.Seek(0, SeekOrigin.Begin);

                    await this.blobClient.UploadFromStreamAsync(
                        buffer,
                        AccessCondition.GenerateLeaseCondition(leaseToken),
                        new BlobRequestOptions(),
                        new OperationContext());

                    var outputLength = buffer.Length;
                    this.logger?.OnNetworkAssignmentsContentUpdated(this.blobClient.Uri, outputLength, leaseToken, fingerprint);
                }

                return true;
            }
            catch (Exception ex)
            {
                this.logger?.OnNetworkAssignmentBlobUpdateError(ex, this.blobClient.Uri, fingerprint);
                throw;
            }
            finally
            {
                if (!string.IsNullOrEmpty(leaseToken))
                {
                    await this.blobClient.ReleaseLeaseAsync(AccessCondition.GenerateLeaseCondition(leaseToken));
                }
            }
        }

        private static Uri GetBlobUri(Uri containerUri, string blobName)
        {
            return new Uri(Path.Combine(containerUri.AbsoluteUri, blobName));
        }
    }
}