﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace Microsoft.Azure.LocationServices.GeoLocation.Tests
{
    using System;
    using System.Net;
    using System.Numerics;

    using VisualStudio.TestTools.UnitTesting;

    [TestClass]
    public class IPNetworkTests
    {
        [TestMethod]
        public void TestThrowsWhenCidrNetworkStartAddressIsIncorrect()
        {
            try
            {
                // node number is not zero
                var ipNetwork = new IPNetwork(GetAddressFromLong(0xABCDEF00), 16);
                Assert.Fail("Should validate the network start address.");
            }
            catch (System.ArgumentException)
            {
                // ignored
            }

            try
            {
                // node number is not zero
                var ipNetwork = new IPNetwork(IPAddress.Parse("192.168.1.1"), 24);
                Assert.Fail("Should validate the network start address.");
            }
            catch (System.ArgumentException)
            {
                // ignored
            }

            // start address has zero node-specifying bits 
            var ipNetwork2 = new IPNetwork(GetAddressFromLong(0xABCD0000), 16);

            // start address has zero node-specifying bits 
            ipNetwork2 = new IPNetwork(IPAddress.Parse("192.168.1.0"), 24);

            // no network nodes, so any address is allowed
            ipNetwork2 = new IPNetwork(GetAddressFromLong(0xABCDEF01), (ulong)32);

            // no network nodes, so any address is allowed
            ipNetwork2 = new IPNetwork(IPAddress.Parse("192.168.1.1"), (ulong)32);
        }

        [TestMethod]
        public void TestIp4AddressContainedInCidrNetwork()
        {
            const uint NetworkAddress = 0xABCDEF00;
            
            // define the network with 255 nodes
            var network = new IPNetwork(GetAddressFromLong(NetworkAddress), 24);
            Assert.IsTrue(network.ContainsAddress(GetAddressFromLong(NetworkAddress)));
            Assert.IsTrue(network.ContainsAddress(GetAddressFromLong(NetworkAddress + 1)));
            Assert.IsTrue(network.ContainsAddress(GetAddressFromLong(NetworkAddress + 255)));
            Assert.IsFalse(network.ContainsAddress(GetAddressFromLong(NetworkAddress + 256)));

            // define the network using string form with MSB > 128
            network = new IPNetwork(IPAddress.Parse("192.168.0.0"), 16);
            Assert.IsTrue(network.ContainsAddress(IPAddress.Parse("192.168.0.0")));
            Assert.IsTrue(network.ContainsAddress(IPAddress.Parse("192.168.255.255")));
            Assert.IsFalse(network.ContainsAddress(IPAddress.Parse("192.169.0.0")));

            // define the network using string form with MSB < 128
            network = new IPNetwork(IPAddress.Parse("10.1.0.0"), 16);
            Assert.IsTrue(network.ContainsAddress(IPAddress.Parse("10.1.0.0")));
            Assert.IsTrue(network.ContainsAddress(IPAddress.Parse("10.1.255.255")));
            Assert.IsFalse(network.ContainsAddress(IPAddress.Parse("10.2.0.0")));
        }

        [TestMethod]
        public void TestIp6AddressContainedInCidrNetwork()
        {
            BigInteger networkAddress = new BigInteger(new byte[] { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0xA, 0xB, 0xC, 0xD, 0xE, 0xF });

            // define the network with 255 nodes
            var network = new IPNetwork(GetAddressFromBigInteger(networkAddress), 120);
            Assert.IsTrue(network.ContainsAddress(GetAddressFromBigInteger(networkAddress)));
            Assert.IsTrue(network.ContainsAddress(GetAddressFromBigInteger(networkAddress + 1)));
            Assert.IsTrue(network.ContainsAddress(GetAddressFromBigInteger(networkAddress + 255)));
            Assert.IsFalse(network.ContainsAddress(GetAddressFromBigInteger(networkAddress + 256)));

            // define the network using string form with MSB > 128
            network = new IPNetwork(IPAddress.Parse("FAAA::"), 16);
            Assert.IsTrue(network.ContainsAddress(IPAddress.Parse("FAAA::")));
            Assert.IsTrue(network.ContainsAddress(IPAddress.Parse("FAAA::1")));
            Assert.IsTrue(network.ContainsAddress(IPAddress.Parse("FAAA:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF")));
            Assert.IsFalse(network.ContainsAddress(IPAddress.Parse("FAAB::")));

            // define the network using string form with MSB < 128
            network = new IPNetwork(IPAddress.Parse("0100::"), 16);
            Assert.IsTrue(network.ContainsAddress(IPAddress.Parse("0100::")));
            Assert.IsTrue(network.ContainsAddress(IPAddress.Parse("0100::1")));
            Assert.IsTrue(network.ContainsAddress(IPAddress.Parse("0100:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF")));
            Assert.IsFalse(network.ContainsAddress(IPAddress.Parse("0101::")));
        }

        private static IPAddress GetAddressFromLong(uint n)
        {
            // Long value in IPAddress constructor is expected to be in MSB order,
            // which is not the default on x86 platform, thus we need this conversion
            // IPAddress.HostToNetworkOrder seems like the right tool, 
            // but it actually returns incorrect value for some reason
            return GetAddressFromLittleEndianBytes(BitConverter.GetBytes(n));
        }

        private static IPAddress GetAddressFromBigInteger(BigInteger n)
        {
            return GetAddressFromLittleEndianBytes(n.ToByteArray());
        }

        /// <summary>
        /// Reverses byte order as IPAddress constructor expects big-endian
        /// </summary>
        private static IPAddress GetAddressFromLittleEndianBytes(byte[] bytes)
        {
            Array.Reverse(bytes);
            return new IPAddress(bytes);
        }
    }
}