﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace Microsoft.Azure.LocationServices.GeoLocation.Tests
{
    using System;
    using System.IO;
    using System.Text;

    using Microsoft.Azure.LocationServices.Iana;
    using Microsoft.VisualStudio.TestTools.UnitTesting;

    [TestClass]
    public class RirDatasetReaderTests
    {
        [TestMethod]
        public void CanParseDatasetHeader()
        {
            var reader = RirDatasetReader.FromStringReader(
                new CsvFormatReader(
                    BuildReader(
                        new[]
                            {
                                "2|lacnic|1512169199|28568|19870101|20171206|-0200",
                                "lacnic|*|ipv4|*|13941|summary",
                                "lacnic|*|ipv6|*|6736|summary",
                                "lacnic|*|asn|*|7891|summary"
                            }),
                    '|'));

            Assert.AreEqual("2", reader.Version);
            Assert.AreEqual("lacnic", reader.Registry);
            Assert.AreEqual("1512169199", reader.Serial);
            Assert.AreEqual(28568, reader.RecordsNumber);
            Assert.AreEqual(new DateTimeOffset(1987, 01, 01, 0, 0, 0, TimeSpan.FromHours(-2)), reader.StartDate);
            Assert.AreEqual(new DateTimeOffset(2017, 12, 06, 0, 0, 0, TimeSpan.FromHours(-2)), reader.EndDate);

            Assert.AreEqual(13941, reader.NumberOfIpv4Records);
            Assert.AreEqual(6736, reader.NumberOfIpv6Records);
            Assert.AreEqual(7891, reader.NumberOfAsnRecords);
        }

        [TestMethod]
        public void CanParseDatasetHeaderWithPositiveOffset()
        {
            var reader = RirDatasetReader.FromStringReader(
                             new CsvFormatReader(
                                 BuildReader(
                                     new[]
                                         {
                                             "2|lacnic|1512169199|28568|19870101|20171206|+0200",
                                             "lacnic|*|ipv4|*|13941|summary",
                                             "lacnic|*|ipv6|*|6736|summary",
                                             "lacnic|*|asn|*|7891|summary"
                                         }),
                                 '|'));

            Assert.AreEqual(new DateTimeOffset(1987, 01, 01, 0, 0, 0, TimeSpan.FromHours(2)), reader.StartDate);
            Assert.AreEqual(new DateTimeOffset(2017, 12, 06, 0, 0, 0, TimeSpan.FromHours(2)), reader.EndDate);
        }

        [TestMethod]
        public void CanParseDatasetHeaderWithZeroOffset()
        {
            var reader = RirDatasetReader.FromStringReader(
                             new CsvFormatReader(
                                 BuildReader(
                                     new[]
                                         {
                                             "2|lacnic|1512169199|28568|19870101|20171206|00000",
                                             "lacnic|*|ipv4|*|13941|summary",
                                             "lacnic|*|ipv6|*|6736|summary",
                                             "lacnic|*|asn|*|7891|summary"
                                         }),
                                 '|'));

            Assert.AreEqual(new DateTimeOffset(new DateTime(1987, 01, 01), TimeSpan.Zero), reader.StartDate);
            Assert.AreEqual(new DateTimeOffset(new DateTime(2017, 12, 06), TimeSpan.Zero), reader.EndDate);
        }

        [TestMethod]
        public void CanParseDatasetHeaderWithZeroStartDate()
        {
            var reader = RirDatasetReader.FromStringReader(
                             new CsvFormatReader(
                                 BuildReader(
                                     new[]
                                         {
                                             "2|lacnic|1512169199|28568|00000000|20171206|00000",
                                             "lacnic|*|ipv4|*|13941|summary",
                                             "lacnic|*|ipv6|*|6736|summary",
                                             "lacnic|*|asn|*|7891|summary"
                                         }),
                                 '|'));

            Assert.AreEqual(DateTimeOffset.MinValue, reader.StartDate);
        }

        [TestMethod]
        public void CanParseIpRanges()
        {
            var reader = RirDatasetReader.FromStringReader(
                             new CsvFormatReader(
                                 BuildReader(
                                     new[]
                                         {
                                             "2|lacnic|1512169199|28568|19870101|20171206|00000",
                                             "lacnic|*|ipv4|*|13941|summary",
                                             "lacnic|*|ipv6|*|6736|summary",
                                             "lacnic|*|asn|*|7891|summary",
                                             "ripencc|FR|ipv4|2.0.0.0|1048576|20100712|allocated|66fc48da-6acc-403e-b8dc-4a67f36e854a",
                                             "ripencc|FR|ipv6|2001:67c:280c::|48|20120412|assigned|1b039fde-f5b0-4913-a2d5-01aa315bc276",
                                             "ripencc||ipv4|194.33.29.0|256||available",
                                             "ripencc||ipv6|2a01:6482::|31||reserved",
                                             "ripencc|DE|asn|28|1|19930901|allocated|6351ffab-f6fd-4e9f-ae1d-85ebd0a6917e"
                                         }),
                                 '|'));

            var record = reader.GetRecord();
            Assert.AreEqual("FR", record.CountryCode);
            Assert.AreEqual(RirRecordType.Ipv4, record.Type);
            Assert.AreEqual("2.0.0.0", record.StartAddress);
            Assert.AreEqual("1048576", record.Value);
            Assert.AreEqual("20100712", record.Date);
            Assert.AreEqual(RirRecordStatus.Allocated, record.Status);

            record = reader.GetRecord();
            Assert.AreEqual("FR", record.CountryCode);
            Assert.AreEqual(RirRecordType.Ipv6, record.Type);
            Assert.AreEqual("2001:67c:280c::", record.StartAddress);
            Assert.AreEqual("48", record.Value);
            Assert.AreEqual("20120412", record.Date);
            Assert.AreEqual(RirRecordStatus.Assigned, record.Status);

            record = reader.GetRecord();
            Assert.AreEqual(string.Empty, record.CountryCode);
            Assert.AreEqual(RirRecordType.Ipv4, record.Type);
            Assert.AreEqual("194.33.29.0", record.StartAddress);
            Assert.AreEqual("256", record.Value);
            Assert.AreEqual(string.Empty, record.Date);
            Assert.AreEqual(RirRecordStatus.Available, record.Status);

            record = reader.GetRecord();
            Assert.AreEqual(string.Empty, record.CountryCode);
            Assert.AreEqual(RirRecordType.Ipv6, record.Type);
            Assert.AreEqual("2a01:6482::", record.StartAddress);
            Assert.AreEqual("31", record.Value);
            Assert.AreEqual(string.Empty, record.Date);
            Assert.AreEqual(RirRecordStatus.Reserved, record.Status);

            record = reader.GetRecord();
            Assert.AreEqual("DE", record.CountryCode);
            Assert.AreEqual(RirRecordType.Asn, record.Type);
            Assert.AreEqual("28", record.StartAddress);
            Assert.AreEqual("1", record.Value);
            Assert.AreEqual("19930901", record.Date);
            Assert.AreEqual(RirRecordStatus.Allocated, record.Status);
        }

        private static TextReader BuildReader(string[] strings)
        {
            var sb = new StringBuilder();

            foreach (var s in strings)
            {
                sb.AppendLine(s);
            }

            return new StringReader(sb.ToString());
        }
    }
}