﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.GeoLocation.Tests.Storage
{
    using System;
    using System.IO;
    using System.Threading;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices.GeoLocation.Storage;
    using Microsoft.VisualStudio.TestTools.UnitTesting;

    using Moq;

    [TestClass]
    public class ConditionalContentMirrorTests
    {
        [TestMethod]
        public async Task WillCopyWhenSourceIsNewer()
        {
            var token = new CancellationToken();
            var stream = new MemoryStream();

            var sourceTimeStamp = new DateTimeOffset(DateTime.UtcNow);
            var source = new Mock<IContentSource>();
            source.Setup(s => s.GetLastModifiedDateTime(token)).ReturnsAsync(sourceTimeStamp);
            source.Setup(s => s.GetSourceAsync(token)).ReturnsAsync(stream);

            bool setTargetCalled = false;
            var targetTimeStamp = new DateTimeOffset(DateTime.UtcNow).AddMinutes(-1);
            var target = new Mock<IContentTarget>();
            target.Setup(t => t.GetLastModifiedDateTime(token)).ReturnsAsync(targetTimeStamp);
            target.Setup(t => t.ExistsAsync(token)).ReturnsAsync(true);
            target.Setup(t => t.SetTargetAsync(stream, token)).Callback(delegate(Stream s, CancellationToken t) { setTargetCalled = true; }).Returns(Task.FromResult(true));

            var mirror = new ConditionalContentMirror(source.Object, target.Object);

            Assert.IsTrue(await mirror.CheckAndUpdateAsync(token));
            Assert.IsTrue(setTargetCalled);
        }

        [TestMethod]
        public async Task WillNotCopyWhenSourceIsNotNewer()
        {
            var token = new CancellationToken();
            var stream = new MemoryStream();

            var sameTimeStamp = new DateTimeOffset(DateTime.UtcNow);
            var source = new Mock<IContentSource>();
            source.Setup(s => s.GetLastModifiedDateTime(token)).ReturnsAsync(sameTimeStamp);
            source.Setup(s => s.GetSourceAsync(token)).ReturnsAsync(stream);

            var target = new Mock<IContentTarget>();
            target.Setup(t => t.GetLastModifiedDateTime(token)).ReturnsAsync(sameTimeStamp.AddMinutes(1));
            target.Setup(t => t.ExistsAsync(token)).ReturnsAsync(true);
            target.Setup(t => t.SetTargetAsync(stream, token)).Callback(delegate(Stream s, CancellationToken t) { Assert.Fail("Shouldn't update the target."); }).Returns(Task.FromResult(true));

            var mirror = new ConditionalContentMirror(source.Object, target.Object);

            Assert.IsFalse(await mirror.CheckAndUpdateAsync(token));
        }

        [TestMethod]
        public async Task WillNotCopyWhenSourceIsTheSame()
        {
            var token = new CancellationToken();
            var stream = new MemoryStream();

            var sameTimeStamp = new DateTimeOffset(DateTime.UtcNow);
            var source = new Mock<IContentSource>();
            source.Setup(s => s.GetLastModifiedDateTime(token)).ReturnsAsync(sameTimeStamp);
            source.Setup(s => s.GetSourceAsync(token)).ReturnsAsync(stream);

            var target = new Mock<IContentTarget>();
            target.Setup(t => t.GetLastModifiedDateTime(token)).ReturnsAsync(sameTimeStamp);
            target.Setup(t => t.ExistsAsync(token)).ReturnsAsync(true);
            target.Setup(t => t.SetTargetAsync(stream, token)).Callback(delegate(Stream s, CancellationToken t) { Assert.Fail("Shouldn't update the target."); }).Returns(Task.FromResult(true));

            var mirror = new ConditionalContentMirror(source.Object, target.Object);

            Assert.IsFalse(await mirror.CheckAndUpdateAsync(token));
        }

        [TestMethod]
        public async Task WillCopyWhenTargetDoesntExist()
        {
            var token = new CancellationToken();
            var stream = new MemoryStream();

            var source = new Mock<IContentSource>();
            source.Setup(s => s.GetSourceAsync(token)).ReturnsAsync(stream);

            bool setTargetCalled = false;
            var target = new Mock<IContentTarget>();
            target.Setup(t => t.ExistsAsync(token)).ReturnsAsync(false);
            target.Setup(t => t.SetTargetAsync(stream, token)).Callback(delegate(Stream s, CancellationToken t) { setTargetCalled = true; }).Returns(Task.FromResult(true));

            var mirror = new ConditionalContentMirror(source.Object, target.Object);

            Assert.IsTrue(await mirror.CheckAndUpdateAsync(token));
            Assert.IsTrue(setTargetCalled);
        }
    }
}