﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.GeoLocation.Tests.Storage
{
    using System;
    using System.Collections.Generic;
    using System.Threading;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices.Common.UnitTests;
    using Microsoft.Azure.LocationServices.GeoLocation.Storage;
    using Microsoft.VisualStudio.TestTools.UnitTesting;

    using Moq;

    [TestClass]
    public class NetworkAssignmentCacheTests
    {
        [TestMethod]
        public async Task LoadsImmediatelyWhenEmpty()
        {
            var logger = new Mock<IGeoLocationLogger>();
            var cancellationSource = new CancellationTokenSource();
            NetworkAssignment[] networkAssignments = new NetworkAssignment[0];
            var storage = new Mock<INetworkAssignmentStorage>();
            storage.Setup(s => s.LoadAsync()).ReturnsAsync(networkAssignments);

            var callbackTriggered = false;
            cancellationSource.Cancel();

            var cache = new NetworkAssignmentCache(storage.Object, TimeSpan.FromMilliseconds(1), logger.Object);
            var refreshTask = cache.GetNetworkAssignmentsWhenUpdated(
                delegate(IEnumerable<NetworkAssignment> newAssignments)
                    {
                        Assert.AreEqual(networkAssignments, newAssignments);
                        callbackTriggered = true;
                    },
                cancellationSource.Token);
            
            await refreshTask;
            Assert.IsTrue(callbackTriggered);
        }

        [TestMethod]
        public async Task DoesNotRefreshWhenFingerprintIsTheSame()
        {
            var logger = new Mock<IGeoLocationLogger>();
            var cancellationSource = new CancellationTokenSource();
            NetworkAssignment[] networkAssignments = new NetworkAssignment[0];
            var callbackTriggered = 0;
            var fetchTriggered = 0;

            var storage = new Mock<INetworkAssignmentStorage>();
            storage.Setup(s => s.LoadAsync()).ReturnsAsync(networkAssignments);
            storage.Setup(s => s.FetchFingerprintAsync())
                   .Callback(delegate { fetchTriggered++; })
                   .ReturnsAsync("same_fingerprint_every_time");

            var cache = new NetworkAssignmentCache(storage.Object, TimeSpan.FromMilliseconds(1), logger.Object);
            var refreshTask = cache.GetNetworkAssignmentsWhenUpdated(
                delegate(IEnumerable<NetworkAssignment> newAssignments)
                    {
                        Assert.IsTrue(callbackTriggered < 1);
                        Assert.AreEqual(networkAssignments, newAssignments);
                        callbackTriggered++;
                    }, 
                cancellationSource.Token);
            
            cancellationSource.CancelAfter(1000);
            await refreshTask;
            Assert.IsTrue(fetchTriggered > 1);
        }

        [TestMethod]
        public async Task DoesRefreshWhenFingerprintIsNew()
        {
            var logger = new Mock<IGeoLocationLogger>();
            var cancellationSource = new CancellationTokenSource();
            NetworkAssignment[] networkAssignments = new NetworkAssignment[0];
            var callbackTriggered = 0;
            var fetchTriggered = 0;

            var storage = new Mock<INetworkAssignmentStorage>();
            storage.Setup(s => s.LoadAsync()).ReturnsAsync(networkAssignments);
            storage.Setup(s => s.FetchFingerprintAsync())
                   .Callback(delegate { fetchTriggered++; })
                   .ReturnsAsync(() => TestHelpers.GetRandomString(5));

            var cache = new NetworkAssignmentCache(storage.Object, TimeSpan.FromMilliseconds(1), logger.Object);
            var refreshTask = cache.GetNetworkAssignmentsWhenUpdated(
                delegate(IEnumerable<NetworkAssignment> newAssignments)
                    {
                        Assert.AreEqual(networkAssignments, newAssignments);
                        callbackTriggered++;
                    },
                cancellationSource.Token);

            cancellationSource.CancelAfter(100);
            await refreshTask;
            Assert.AreEqual(fetchTriggered, callbackTriggered);
        }
    }
}