﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.GeoLocation.Tests.Storage
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Reflection;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices.Common.UnitTests;
    using Microsoft.Azure.LocationServices.GeoLocation.Storage;
    using Microsoft.Azure.LocationServices.GeoLocation.Tests.TestDataBuilders;
    using Microsoft.VisualStudio.TestTools.UnitTesting;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;

    using Moq;

    [TestClass]
    public class NetworkAssignmentStorageTests
    {
        [TestMethod]
        public async Task CanReadFromCompressedForm()
        {
            var assembly = Assembly.GetExecutingAssembly();

            var jsonStream = assembly.GetManifestResourceStream("Microsoft.Azure.LocationServices.GeoLocation.Tests.Storage.SampleAssignments.gzip");
            var blobClient = new Mock<ICloudBlob>();
            blobClient.Setup(c => c.OpenReadAsync()).ReturnsAsync(jsonStream);

            var storage = new NetworkAssignmentStorage(blobClient.Object, new Mock<IGeoLocationLogger>().Object);
            var networks = await storage.LoadAsync();

            var result = networks.Single(n => n.Network.ContainsAddress(IPAddress.Parse("1.33.1.1")));
            Assert.AreEqual("JP", result.IsoCountryCode);
        }

        [TestMethod]
        public async Task CanReadWhatHasBeenWritten()
        {
            string leaseToken = TestHelpers.GetRandomString(10);
            string fingerprint = TestHelpers.GetRandomString(10);
            var metadata = new Dictionary<string, string>() { { NetworkAssignmentStorage.FingerprintPropertyName, TestHelpers.GetRandomString(5) } };

            var network = new NetworkAssignmentDataBuilder().WithNetwork(new IPNetwork(IPAddress.Any, 1)).Build();
            MemoryStream dataStream = new MemoryStream();

            var blobClient = new Mock<ICloudBlob>();
            
            // assume that blob exists
            blobClient.Setup(c => c.ExistsAsync()).ReturnsAsync(true);
            blobClient.SetupGet(c => c.Uri).Returns(new Uri("http://tempuri.org"));

            // retrieve the lease Id
            blobClient.Setup(c => c.AcquireLeaseAsync(It.IsAny<TimeSpan>(), It.IsAny<string>())).ReturnsAsync(leaseToken);

            // no-op on fetch attributes
            blobClient.Setup(c => c.FetchAttributesAsync(It.IsAny<AccessCondition>(), It.IsAny<BlobRequestOptions>(), It.IsAny<OperationContext>()))
                      .Callback(
                          delegate(AccessCondition accessCondition, BlobRequestOptions requestOptions, OperationContext context)
                              {
                                  Assert.AreEqual(leaseToken, accessCondition.LeaseId);
                              })
                      .Returns(Task.FromResult(true));

            // returns metadata with the random finderprint
            blobClient.SetupGet(c => c.Metadata).Returns(metadata);
            
            // accept for the update
            blobClient.Setup(c => c.UploadFromStreamAsync(It.IsAny<Stream>(), It.IsAny<AccessCondition>(), It.IsAny<BlobRequestOptions>(), It.IsAny<OperationContext>()))
                      .Callback(delegate(Stream s, AccessCondition accessCondition, BlobRequestOptions requestOptions, OperationContext context)
                          {
                              s.CopyTo(dataStream);
                              dataStream.Seek(0, SeekOrigin.Begin);
                              Assert.AreEqual(leaseToken, accessCondition.LeaseId);
                              Assert.AreEqual(fingerprint, metadata[NetworkAssignmentStorage.FingerprintPropertyName]);
                          })
                      .Returns(Task.FromResult(true));

            // return the same stream on read
            blobClient.Setup(c => c.OpenReadAsync()).ReturnsAsync(() => dataStream);
            
            // verify that lease has been removed
            blobClient.Setup(c => c.ReleaseLeaseAsync(It.IsAny<AccessCondition>()))
                      .Callback(
                          delegate(AccessCondition accessCondition)
                              {
                                  Assert.AreEqual(leaseToken, accessCondition.LeaseId);
                              })
                      .Returns(Task.FromResult(true));

            var storage = new NetworkAssignmentStorage(blobClient.Object, new Mock<IGeoLocationLogger>().Object);
            await storage.ReplaceIfFingerprintDoesntMatchAsync(new[] { network }, fingerprint, TimeSpan.Zero);
            var loadedAssignments = await storage.LoadAsync();

            var loadedAssignment = loadedAssignments.Single();
            Assert.AreEqual(network.Network.StartAddress, loadedAssignment.Network.StartAddress);
            Assert.AreEqual(network.IsoCountryCode, loadedAssignment.IsoCountryCode);

            blobClient.VerifyAll();
        }
    }
}