﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.APIMLogToGeneva
{
    using System;
    using System.Fabric;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.ServiceBus.Messaging;
    using Microsoft.ServiceFabric.Services.Runtime;

    /// <summary>
    /// An instance of this class is created for each service instance by the Service Fabric runtime.
    /// </summary>
    internal sealed class APIMLogToGeneva : StatelessService
    {
        /// <summary>
        /// Configuration Reader which is defaulted to a keyvault based cache solution.
        /// </summary>
        private static readonly IConfigurationReader _configurationReader = new KeyVaultConfigurationReader();

        private EventProcessorHost eventProcessorHost = null;

        public APIMLogToGeneva(StatelessServiceContext context)
            : base(context)
        {
        }

        /// <summary>
        /// This is the main entry point for your service instance.
        /// </summary>
        /// <param name="cancellationToken">Canceled when Service Fabric needs to shut down this service instance.</param>
        protected override async Task RunAsync(CancellationToken cancellationToken)
        {
            FabricRuntimeHelper.ServiceFabricInstrumentInstance.LogEvent(Instrumentation.EventType.Information, "Registering EventProcessor");

            try
            {
                string StorageContainerName = FabricRuntimeHelper.GetConfigurationValue("Config", "StorageContainerName");

                string apimEventHubSasToken = await _configurationReader.GetConfigurationValueAsync(FabricRuntimeHelper.GetConfigurationValue("Config", "SasTokenSecretId"));
                string apimEventHubConnectionTemplate = await _configurationReader.GetConfigurationValueAsync(FabricRuntimeHelper.GetConfigurationValue("Config", "ConnectionTemplateSecretId"));
                string connectionString = string.Format(apimEventHubConnectionTemplate, apimEventHubSasToken);

                string EhConnectionString = await _configurationReader.GetConfigurationValueAsync(FabricRuntimeHelper.GetConfigurationValue("Config", "EhConnectionStringSecretId"));
                string EhConnStrNoEntityPath = EhConnectionString.Substring(0, EhConnectionString.LastIndexOf(';'));
                string EhEntityPath = EhConnectionString.Substring(EhConnectionString.LastIndexOf('=') + 1);

                string eventProcessorHostName = Guid.NewGuid().ToString();
                this.eventProcessorHost = new EventProcessorHost(
                    eventProcessorHostName,
                    EhEntityPath,
                    EventHubConsumerGroup.DefaultGroupName,
                    EhConnStrNoEntityPath,
                    connectionString,
                    StorageContainerName);

                // Registers the Event Processor Host and starts receiving messages
                var options = new EventProcessorOptions();
                options.ExceptionReceived += (sender, e) =>
                    {
                        // <see cref="https://stackoverflow.com/questions/41496754/what-is-causing-azure-event-hubs-receiverdisconnectedexception-leaselostexceptio?answertab=votes#tab-top"/>
                        Type exceptionType = e.Exception.GetType();
                        if (exceptionType != typeof(LeaseLostException) && exceptionType != typeof(ReceiverDisconnectedException))
                        {
                            FabricRuntimeHelper.ServiceFabricInstrumentInstance.LogException(e.Exception, "apimLogToGeneva", "logProcessing");
                        }
                    };

                eventProcessorHost.RegisterEventProcessorAsync<SimpleEventProcessor>(options).Wait();
            }
            catch (Exception e)
            {
                FabricRuntimeHelper.ServiceFabricInstrumentInstance.LogException(e, "apimLogToGeneva", "registerEventProcessor");
                throw;
            }
        }

        protected override async Task OnCloseAsync(CancellationToken cancellationToken)
        {
            // Disposes of the Event Processor Host
            if (this.eventProcessorHost != null)
            {
                FabricRuntimeHelper.ServiceFabricInstrumentInstance.LogEvent(EventType.Information, "Disposing EventProcessor");
                await this.eventProcessorHost.UnregisterEventProcessorAsync();
            }

            await base.OnCloseAsync(cancellationToken);
        }
    }
}
