﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.APIMLogToGeneva
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Text;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.APIMLogToGeneva.Configuration;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.Instrumentation.Apim;
    using Microsoft.ServiceBus.Messaging;
    using Newtonsoft.Json;
    using Newtonsoft.Json.Linq;

    class SimpleEventProcessor : IEventProcessor
    {
        private ApimOperationMetricLogger metricLogger;

        public Task CloseAsync(PartitionContext context, CloseReason reason)
        {
            FabricRuntimeHelper.ServiceFabricInstrumentInstance.LogEvent(EventType.Information, $"Processor Shutting Down. Partition '{context.Lease.PartitionId}', Reason: '{reason}'.");
            return Task.CompletedTask;
        }

        public Task OpenAsync(PartitionContext context)
        {
            FabricRuntimeHelper.ServiceFabricInstrumentInstance.LogEvent(EventType.Information, $"SimpleEventProcessor initialized. Partition: '{context.Lease.PartitionId}'");
            metricLogger = new ApimOperationMetricLogger(ApimLogHostSettings.CreateFromEnvironment().MdmParameters.MdmAccount);
            return Task.CompletedTask;
        }

        public Task ProcessErrorAsync(PartitionContext context, Exception error)
        {
            FabricRuntimeHelper.ServiceFabricInstrumentInstance.LogEvent(EventType.ProgramError, $"Error on Partition: {context.Lease.PartitionId}, Error: {error.Message}");
            return Task.CompletedTask;
        }

        public async Task ProcessEventsAsync(PartitionContext context, IEnumerable<EventData> messages)
        {
            foreach (var eventData in messages)
            {
                var data = Encoding.UTF8.GetString(eventData.GetBytes());
                FabricRuntimeHelper.ServiceFabricInstrumentInstance.LogEvent(EventType.Verbose, $"Message received. Partition: '{context.RuntimeInfo.PartitionId}', Data: {data}");

                JsonReader reader = new JsonTextReader(new StringReader(data));
                reader.DateParseHandling = DateParseHandling.None;
                dynamic dataObj = JObject.Load(reader);

                if (((string)dataObj?.Type) == "request")
                {
                    var requestInfo = JsonConvert.DeserializeObject<RequestInfo>(
                        data,
                        new JsonSerializerSettings()
                        {
                            TypeNameHandling = TypeNameHandling.Auto
                        });

                    // Using Geneva Operation API to trace APIM Request
                    using (var op = new ServiceApimOperation(nameof(SimpleEventProcessor) + "_LogRequest"))
                    {
                        op.LogRequest(requestInfo);
                    }

                    // Using Geneva MDM API to update metric from APIM Request
                    metricLogger.UpdateMetric(0, "request", requestInfo.TrackingId, requestInfo.Operation.Name, "Success", requestInfo.Deployment.Region);
                }

                if (((string)dataObj?.Type) == "response")
                {
                    var responseInfo = JsonConvert.DeserializeObject<ResponseInfo>(
                        data,
                        new JsonSerializerSettings()
                        {
                            TypeNameHandling = TypeNameHandling.Auto
                        });

                    // Using Geneva Operation API to trace APIM Response
                    using (var op = new ServiceApimOperation(nameof(SimpleEventProcessor) + "_LogResponse"))
                    {
                        op.LogResponse(responseInfo);
                    }

                    // Using Geneva MDM API to update metric from APIM Response
                    var opResult = APIMLoggerHelper.GetResultFromStatusCode(responseInfo);
                    metricLogger.UpdateMetric(responseInfo.Duration, "response", responseInfo.TrackingId, responseInfo.Operation.Name, opResult, responseInfo.Deployment.Region);
                }
            }

            await context.CheckpointAsync();
        }
    }
}
