﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.IPDataImport.Configuration
{
    using System;
    using System.Globalization;

    using Microsoft.Azure.LocationServices.Common;

    /// <summary>
    /// Represents the configuration necessary to operate in an environment.
    /// </summary>
    /// <remarks>
    /// This class should only contain members that differ from environment to environment. When something doesn't 
    /// vary, it should not be present in this class, and instead should be a compile-time constant. If a property
    /// is no longer used, or no longer varies, it should be removed.
    /// </remarks>
    public abstract class IPDataImportHostSettings
    {                               
        /// <summary>
        /// Gets the MDM session name.
        /// </summary>
        public abstract string MdmSessionName { get; }

        /// <summary>
        /// Gets the MDM account name.
        /// </summary>
        public abstract string MdmAccountName { get; }

        /// <summary>
        /// Indicates if local MDM output should be enabled.
        /// </summary>
        public virtual bool EnableLocalMdmOutput { get; } = false;

        /// <summary>
        /// Gets the current region.
        /// </summary>
        protected static string CurrentRegion
        {
            get
            {
                string currentRegion = Environment.GetEnvironmentVariable("Region");
                return currentRegion ?? "DEFAULT";
            }
        }

        /// <summary>
        /// Gets the current environment.
        /// </summary>
        /// <exception cref="ArgumentException">
        /// Throws if the current environment is not specified
        /// </exception>
        private static string CurrentEnvironment
        {
            get
            {
                string currentEnvironment = Environment.GetEnvironmentVariable("Environment");

                if (string.IsNullOrEmpty(currentEnvironment))
                {
                    throw new ArgumentException("Current Environment could not be determined");
                }

                return currentEnvironment;
            }
        }

        /// <summary>
        /// Create an environment definition from the environment environment variable.
        /// </summary>
        /// <returns>The created environment definition.</returns>
        public static IPDataImportHostSettings CreateFromEnvironment()
        {
            switch (CurrentEnvironment.ToUpperInvariant())
            {
                case LogicalEnvironments.Development:
                    return new IPDataImportHostDevSettings();
                case LogicalEnvironments.Test:
                    return new IPDataImportHostTestSettings();
                case LogicalEnvironments.Staging:
                    return new IPDataImportHostStagingSettings();
                case LogicalEnvironments.Production:
                    return new IPDataImportHostProdSettings();
                default:
                    throw new ArgumentException(
                        string.Format(
                            CultureInfo.InvariantCulture,
                            "Environment {0} does not have a corresponding Environment definition defined",
                            CurrentEnvironment));
            }
        }
    }
}