﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.IPDataImport
{
    using System;
    using System.Collections.Generic;
    using System.Fabric;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Threading;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.GeoLocation.Storage;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.IPDataImport.Properties;
    using Microsoft.Azure.LocationServices.ServiceFabricHosting;
    using Microsoft.ServiceFabric.Services.Runtime;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Auth;
    using Microsoft.WindowsAzure.Storage.Blob;

    using Validation;

    /// <summary>
    /// An instance of this class is created for each service instance by the Service Fabric runtime.
    /// </summary>
    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1001:TypesThatOwnDisposableFieldsShouldBeDisposable", Justification = "There is only one instance of this type per process.")]
    internal sealed class IPDataImport : StatelessService
    {
        private readonly ConfigurationReader serviceParameters;

        public IPDataImport(StatelessServiceContext context)
            : base(context)
        {
            this.serviceParameters = new ConfigurationReader(context);
        }

        /// <summary>
        /// This is the main entry point for your service instance.
        /// </summary>
        /// <param name="cancellationToken">Canceled when Service Fabric needs to shut down this service instance.</param>
        protected override async Task RunAsync(CancellationToken cancellationToken)
        {
            try
            {
                var secretReader = new KeyVaultConfigurationReader();
                var connectionStringTemplate = await this.GetSecretValue(secretReader, "malbsipdata");

                var sasKey = await this.GetSecretValue(secretReader, "malbsipdata-crudall");
                var account = CloudStorageAccount.Parse(string.Format(connectionStringTemplate, sasKey));
                var rirContainerPath = new Uri(account.BlobEndpoint, Settings.Default.RirMirrorContainerPath);

                var mirrors = CreateMirrorsForFtpLocations(Settings.Default.RirFtpLocations.Cast<string>(), rirContainerPath, account.Credentials, out var blobProxies);

                var ipdataContainerPath = new Uri(account.BlobEndpoint, Settings.Default.IPDataContainerPath);
                var logger = new GeoLocationLogger();

                var storage = new NetworkAssignmentStorage(ipdataContainerPath, account.Credentials, logger);
                var builder = new NetworkAssignmentBuilder(blobProxies, storage, logger);

                while (!cancellationToken.IsCancellationRequested)
                {
                    try
                    {
                        await this.EnsureContainerExists(rirContainerPath, account.Credentials);

                        await this.EnsureContainerExists(ipdataContainerPath, account.Credentials);
                    
                        await Task.WhenAll(mirrors.Select(
                            async delegate(ConditionalContentMirror mr) 
                            {
                                try
                                {
                                    await mr.CheckAndUpdateAsync(cancellationToken);
                                }
                                catch (WebException ex)
                                {
                                    Agent.Instance.LogEvent(EventType.Warning, $"[IPData] Failed to refresh the source '{mr.Source.Name}'. Reason: '{ex.Message}'");
                                }
                            }));
                    
                        await builder.BuildAndUpdateAsync(cancellationToken);
                    }
                    catch (AggregateException ex)
                    {
                        Agent.Instance.LogException(ex, "ipDataImport", "update");
                    }
                    catch (Exception ex)
                    {
                        Agent.Instance.LogException(ex, "ipDataImport", "update");
                    }

                    await Task.Delay(TimeSpan.FromMinutes(Settings.Default.IPDataRefreshMinutes), cancellationToken);
                }
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "ipDataImport", "update");
                throw;
            }
        }

        /// <summary>
        /// Returns a list of content trackers.
        /// </summary>
        private static ConditionalContentMirror[] CreateMirrorsForFtpLocations(
            IEnumerable<string> ftpLocations,
            Uri targetBlobContainer,
            StorageCredentials storageCredentials,
            out IList<BlobContentProxy> proxies)
        {
            Requires.NotNull(storageCredentials, nameof(storageCredentials));
            Requires.NotNull(targetBlobContainer, nameof(targetBlobContainer));
            
            proxies = new List<BlobContentProxy>();

            var pairs = new List<ConditionalContentMirror>();
            foreach (var ftpLocation in ftpLocations)
            {
                Requires.NotNullOrEmpty(ftpLocation, nameof(ftpLocation));
                Requires.That(
                    Uri.IsWellFormedUriString(ftpLocation, UriKind.Absolute),
                    nameof(ftpLocation),
                    $"Ftp location {ftpLocation} is not a well-formatted Uri.");

                var targetUri = new Uri(
                    Path.Combine(
                        targetBlobContainer.AbsoluteUri,
                        Path.GetFileName(ftpLocation) ?? throw new InvalidOperationException($"Ftp location {ftpLocation} doesn't specify a file name.")));

                var proxy = new BlobContentProxy(targetUri, storageCredentials);
                proxies.Add(proxy);
                pairs.Add(
                    new ConditionalContentMirror(
                        new FtpContentSource(new Uri(ftpLocation)),
                        proxy));
            }

            return pairs.ToArray();
        }

        private async Task EnsureContainerExists(Uri containerAddress, StorageCredentials storageCredentials)
        {
            var client = new CloudBlobContainer(containerAddress, storageCredentials);
            await client.CreateIfNotExistsAsync();
        }

        private Task<string> GetSecretValue(KeyVaultConfigurationReader secretReader, string keyVaultKeyName)
        {
            return secretReader.GetConfigurationValueAsync(this.serviceParameters.GetConfigValue<string>(keyVaultKeyName));
        }
    }
}
