﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.IPDataImport
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Globalization;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Security.Cryptography;
    using System.Threading;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices;
    using Microsoft.Azure.LocationServices.GeoLocation.Storage;
    using Microsoft.Azure.LocationServices.Iana;
    using Microsoft.Azure.LocationServices.IPDataImport.Properties;

    /// <summary>
    /// Extracts network assignments from multiple sources and stores them as one dataset.
    /// </summary>
    public class NetworkAssignmentBuilder
    {
        private readonly IEnumerable<IContentSource> regionalDelegations;

        private readonly INetworkAssignmentStorage storage;

        private readonly IGeoLocationLogger logger;

        /// <summary>
        /// Creates new instance of NetworkAssignmentBuilder.
        /// </summary>
        public NetworkAssignmentBuilder(IEnumerable<IContentSource> regionalDelegations, INetworkAssignmentStorage storage, IGeoLocationLogger logger)
        {
            this.regionalDelegations = regionalDelegations;
            this.storage = storage;
            this.logger = logger;
        }

        /// <summary>
        /// Performs the build operation and stores result.
        /// </summary>
        public async Task BuildAndUpdateAsync(CancellationToken cancellationToken)
        {
            List<byte[]> sourceHashes = new List<byte[]>();
            SortedList<byte[], NetworkAssignment> assignments = new SortedList<byte[], NetworkAssignment>(ByteArrayComparer.Default);

            Stopwatch stopwatch = Stopwatch.StartNew();

            foreach (var source in this.regionalDelegations)
            {
                try
                {
                    SHA256Managed sha256Managed = new SHA256Managed();
                    using (var hashStream = new CryptoStream(await source.GetSourceAsync(CancellationToken.None), sha256Managed, CryptoStreamMode.Read))
                    {
                        this.AppendFromStream(hashStream, assignments, cancellationToken);
                        sourceHashes.Add(sha256Managed.Hash);
                    }
                }
                catch (Exception e)
                {
                    this.logger?.OnFailedToParseRirDataSource(source.Name, e);
                    throw;
                }
            }

            if (!cancellationToken.IsCancellationRequested)
            {
                var fingerprint = Convert.ToBase64String(sourceHashes.SelectMany(a => a).ToArray());
                var result = await this.storage.ReplaceIfFingerprintDoesntMatchAsync(assignments.Values, fingerprint, Settings.Default.IPDataBlobLeaseTimeout);
                this.logger?.OnNetworkAssignmentRebuildCompleted(result, stopwatch.ElapsedMilliseconds, fingerprint);
            }
            else
            {
                this.logger?.OnNetworkAssignmentRebuildCompleted(false, -1, "Cancelled");
            }
        }

        private void AppendFromStream(Stream stream, IDictionary<byte[], NetworkAssignment> assignments, CancellationToken cancellationToken)
        {
            var reader = RirDatasetReader.FromStringReader(new CsvFormatReader(new StreamReader(stream), '|'));
            
            while (!cancellationToken.IsCancellationRequested)
            {
                IPNetwork network;

                RirRecord record = reader.GetRecord();

                if (record == null)
                {
                    // end of input reached
                    break;
                }

                switch (record.Type)
                {
                    case RirRecordType.Ipv4:
                        network = new IPNetwork(IPAddress.Parse(record.StartAddress), uint.Parse(record.Value));
                        break;
                    case RirRecordType.Ipv6:
                        network = new IPNetwork(IPAddress.Parse(record.StartAddress), byte.Parse(record.Value));
                        break;
                    default: continue;
                }

                if (record.Status == RirRecordStatus.Allocated || record.Status == RirRecordStatus.Assigned)
                {
                    try
                    {
                        var assignment = new NetworkAssignment(
                            network,
                            record.CountryCode,
                            DateTime.ParseExact(record.Date, "yyyyMMdd", DateTimeFormatInfo.InvariantInfo),
                            record.Registry);

                        assignments.Add(assignment.Network.StartAddress.MapToIPv6().GetAddressBytes(), assignment);
                    }
                    catch (Exception ex)
                    {
                        this.logger?.OnFailedToAddNetworkAssigment(ex, record.StartAddress, record.Registry);
                    }
                }
            }
        }
    }
}