﻿// --------------------------------------------------------------------------------------------------------------------
// <copyright file="ResourceProviderDataBackupJob.cs" company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// --------------------------------------------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.LBSBatchService
{
    using System;
    using System.Collections.Generic;
    using System.Fabric;
    using System.Net;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Common.Storage;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.LBSBatchService.Properties;
    using Microsoft.Azure.LocationServices.Shared;
    using Microsoft.ServiceFabric.Services.Runtime;

    /// <summary>
    /// An instance of this class is created for each service instance by the Service Fabric runtime.
    /// </summary>
    internal sealed class LBSBatchService : StatelessService
    {
        public LBSBatchService(StatelessServiceContext context)
            : base(context)
        {
        }

        /// <summary>
        /// This is the main entry point for your service instance.
        /// </summary>
        /// <param name="cancellationToken">Canceled when Service Fabric needs to shut down this service instance.</param>
        protected override async Task RunAsync(CancellationToken cancellationToken)
        {
            try
            {
                // Configure app/service level network settings
                ConfigureServicePointProperties();

                // Initialize the Config.
                await Config.Instance.InitializeAsync(this.Context);

                // Initialize the Logger.
                IBatchServiceLogger logger = new BatchServiceLogger();

                // Create the service dependencies for workers
                IQueueService queueService = new AzureQueueService(Config.Instance.ConnectionString);
                IBlobService blobService = new AzureBlobService(Config.Instance.ConnectionString);
                ITableService tableService = new AzureTableService(Config.Instance.ConnectionString);
                IBlobHelper blobHelper = new BlobHelper(blobService);

                var pollingConfigForBatch = 
                    new PollingQueueConfig(
                        Settings.Default.MaxQueueMessagesToProcessInParallelForBatch,
                        Settings.Default.QueueMessageVisibilityTimeoutInMinutesForBatch,
                        Settings.Default.QueueWorkerDefaultSleepTimeInSecondsForBatch,
                        Settings.Default.QueueWorkerMaxSleepTimeInSecondsForBatch);

                var pollingConfigForRouteMatrix = 
                    new PollingQueueConfig(
                        Settings.Default.MaxQueueMessagesToProcessInParallelForRouteMatrix,
                        Settings.Default.QueueMessageVisibilityTimeoutInMinutesForRouteMatrix,
                        Settings.Default.QueueWorkerDefaultSleepTimeInSecondsForRouteMatrix,
                        Settings.Default.QueueWorkerMaxSleepTimeInSecondsForRouteMatrix);
                
                while (true)
                {
                    cancellationToken.ThrowIfCancellationRequested();

                    // All these are long/continuously running workers. None of them should actually ever "finish".
                    await Task.WhenAll(GetAllWorkerTasks(queueService, blobService, tableService, blobHelper, logger, cancellationToken, pollingConfigForBatch, pollingConfigForRouteMatrix));
                }
            }
            catch (AggregateException ex)
            {
                Agent.Instance.LogException(ex, "batchService", "batchExecution");
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "batchService", "batchExecution");
                throw;
            }
        }

        private static IEnumerable<Task> GetAllWorkerTasks(
            IQueueService queueService,
            IBlobService blobService,
            ITableService tableService,
            IBlobHelper blobHelper,
            IBatchServiceLogger logger,
            CancellationToken cancellationToken,
            PollingQueueConfig pollingConfigForBatch,
            PollingQueueConfig pollingConfigForRouteMatrix)
        {
            return new List<Task>
            {
                // Worker for polling and processing Batch Api requests
                GetBatchServiceQueuePollingWorker(queueService, blobService, tableService, blobHelper, logger, cancellationToken, pollingConfigForBatch),

                // Worker for polling and processing Route Matrix Api requests
                GetRouteMatrixServiceQueuePollingWorker(queueService, blobService, blobHelper, logger, cancellationToken, pollingConfigForRouteMatrix),

                // Worker for deleting expired blobs
                GetBatchServiceBlobExpirationWorker(blobService, logger, cancellationToken)
            };
        }

        private static async Task GetBatchServiceQueuePollingWorker(
            IQueueService queueService, 
            IBlobService blobService,
            ITableService tableService,
            IBlobHelper blobHelper,
            IBatchServiceLogger logger, 
            CancellationToken cancellationToken,
            PollingQueueConfig pollingConfigForBatch)
        {
            await new BatchQueuePollingWorker(queueService, blobService, tableService, blobHelper, logger, pollingConfigForBatch)
                .Start(SharedUtilities.BatchServicePrimaryQueueName, cancellationToken);
        }

        private static async Task GetRouteMatrixServiceQueuePollingWorker(
            IQueueService queueService, 
            IBlobService blobService,
            IBlobHelper blobHelper,
            IBatchServiceLogger logger, 
            CancellationToken cancellationToken,
            PollingQueueConfig pollingConfigForRouteMatrix)
        {
            await new QueuePollingWorker(
                new MatrixQueueMessageHandlerFactory(),
                queueService,
                blobService,
                blobHelper,
                logger,
                pollingConfigForRouteMatrix).Start(SharedUtilities.RouteMatrixServicePrimaryQueueName, cancellationToken);
        }

        private static async Task GetBatchServiceBlobExpirationWorker(
            IBlobService blobService,
            IBatchServiceLogger logger,
            CancellationToken cancellationToken)
        {
            var containerList = new List<string>
            {
                SharedUtilities.BatchApiContainerName,
                SharedUtilities.RouteMatrixServiceResultsContainerName
            };

            await new BlobExpirationWorker(blobService, logger, containerList).Start(cancellationToken);
        }

        private static void ConfigureServicePointProperties()
        {
            ServicePointManager.DefaultConnectionLimit = Settings.Default.MaxConcurrentConnections;
            ServicePointManager.UseNagleAlgorithm = false;
            ServicePointManager.Expect100Continue = false;
            
            // TODO: Consider adding ServicePointManager properties for addressing ConnectionLeaseTimeout and DnsRefreshTimeout 
        }
    }
}