﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.LBSBatchService
{
    using System;
    using System.Collections.Generic;
    using System.Globalization;
    using System.IO;
    using System.Linq;
    using System.Text;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Common.Validation;
    using Microsoft.Azure.LocationServices.LBSBatchService.Properties;
    using Microsoft.Azure.LocationServices.Shared;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using Microsoft.WindowsAzure.Storage.Blob.Protocol;
    using Newtonsoft.Json;

    /// <summary>
    /// This class runs infinitely and once every day queries the blob storage for the blob "expiration map"
    /// The "expiration map" contains a list of blobs and their expiration dates.
    /// The worker filters the blobs that have expired and deletes them from the container. 
    /// </summary>
    public class BlobExpirationWorker
    {
        private const string LastProcessedDateUtcMetadataKey = "LastProcessedDate";

        // Default time in hours that the blob expiration worker sleeps for, once all expired blobs for the day have been deleted.
        private static readonly int BlobExpirationWorkerDefaultSleepTimeInHours = Settings.Default.BlobExpirationWorkerDefaultSleepTimeInHours;

        // The name of the blob that contains the expiration map for the container. 
        private static readonly string ExpirationMapBlobName = SharedUtilities.ExpirationMapBlobName;

        // Dependencies
        private readonly IBlobService blobService;
        private readonly IBatchServiceLogger logger;
        private readonly List<string> containers;

        public BlobExpirationWorker(IBlobService blobService, IBatchServiceLogger logger, List<string> containers)
        {
            Assert.ParamIsNotNull(blobService, nameof(blobService));
            Assert.ParamIsNotNull(logger, nameof(logger));
            Assert.IsTrue(containers?.Count > 0, "Containers list should not be empty");

            this.blobService = blobService;
            this.logger = logger;
            this.containers = containers;
        }

        // TODO: Only run this operation at night time? (Task 2294036)
        public async Task Start(CancellationToken cancellationToken)
        {
            try
            {
                // Run Continuously
                while (true)
                {
                    cancellationToken.ThrowIfCancellationRequested();

                    // 1. Delete expired blobs in all the containers
                    // 2. Update the expiration map in all the containers.
                    await Task.WhenAll(this.containers.Select(this.DeleteExpiredBlobsAndUpdateExpirationMapAsync));

                    // Sleep. Typically this period should be around 24 hours.
                    await Task.Delay(TimeSpan.FromHours(BlobExpirationWorkerDefaultSleepTimeInHours), cancellationToken);
                }
            }
            catch (AggregateException aggEx)
            {
                this.logger.LogAggregateError(aggEx);
            }
            catch (Exception ex)
            {
                // Catch all - Something unexpected happened.
                // Don't rethrow the exception here. Don't want to kill the application if there is an issue deleting one blob.
                // We should map an alert to this log message and manually intervene if for some reason expired blobs are not getting deleted successfully.
                if (ex is OperationCanceledException opcEx)
                {
                    this.logger.LogError(opcEx, "BlobExpirationWorker Stopped!");
                    throw;
                }

                this.logger.LogError(ex, "BlobExpirationWorker Failed!");
            }
        }

        private async Task DeleteExpiredBlobsAndUpdateExpirationMapAsync(string containerName)
        {
            // Check if both the container and the expiration-map blob in the container exist. Return if not.
            if (!await this.blobService.DoesContainerExistAsync(containerName) ||
                !await this.blobService.DoesAppendBlobExistAsync(containerName, ExpirationMapBlobName))
            {
                return;
            }

            string leaseId = null;

            try
            {
                // Try and Acquire the lease on the Expiration Map Blob
                leaseId = await this.TryAcquireLeaseAsync(containerName);

                // Coudn't acquire lease || A different worker instance already ran for the day.
                if (string.IsNullOrEmpty(leaseId) || await this.HasAlreadyRanForTheDayAsync(containerName, leaseId))
                {
                    return;
                }

                // Get the Blob expiration list for the Container
                IEnumerable<BlobExpirationInfo> blobExpirationList = await this.GetBlobExpirationListForContainerAsync(containerName, leaseId);

                // Generate the new expired and non-expired Blob lists.
                List<BlobExpirationInfo> expiredBlobs = new List<BlobExpirationInfo>();
                List<BlobExpirationInfo> nonExpiredBlobs = new List<BlobExpirationInfo>();
                this.SetExpiredAndNonExpiredBlobs(blobExpirationList, expiredBlobs, nonExpiredBlobs);

                // Delete the Expired Blobs.
                IEnumerable<Task> deleteBlobTasks = expiredBlobs.Select(this.DeleteBlobAsync);
                await Task.WhenAll(deleteBlobTasks);

                // Update the Container's Expiration Map with the non-expired Blobs.
                await this.UpdateExpirationMapForContainerAsync(containerName, nonExpiredBlobs, leaseId);
            }
            catch (Exception ex) when (ex is StorageException || ex is JsonSerializationException)
            {
                this.logger.LogError(ex, $"BlobExpirationWorker failed for container:{containerName}");
                throw;
            }
            finally
            {
                // Release the Lease on the Expiration Map Blob
                if (leaseId != null)
                {
                    await this.blobService.ReleaseLeaseAsync(
                        this.blobService.GetAppendBlobReference(containerName, ExpirationMapBlobName),
                        AccessCondition.GenerateLeaseCondition(leaseId));
                }
            }
        }

        private async Task<bool> HasAlreadyRanForTheDayAsync(string containerName, string leaseId)
        {
            IDictionary<string, string> blobMetadata = 
                await this.blobService.GetAppendBlobMetadataAsync(
                    containerName, 
                    ExpirationMapBlobName,
                    AccessCondition.GenerateLeaseCondition(leaseId),
                    new BlobRequestOptions(), 
                    new OperationContext());

            // LastProcessedDate Metadata Exists?
            if (blobMetadata.TryGetValue(LastProcessedDateUtcMetadataKey, out var lastProcessedDateUtcString))
            {
                // LastProcessedDate is a valid Date
                if (DateTime.TryParse(lastProcessedDateUtcString, CultureInfo.InvariantCulture, DateTimeStyles.RoundtripKind, out DateTime lastProcessedDateUtc))
                {
                    // Process only once every <BlobExpirationWorkerDefaultSleepTimeInHours>
                    if ((DateTime.UtcNow - lastProcessedDateUtc).TotalHours < BlobExpirationWorkerDefaultSleepTimeInHours)
                    {
                        // Worker already ran once in <BlobExpirationWorkerDefaultSleepTimeInHours>
                        return true;
                    }
                }
            }

            return false;
        }

        private async Task<string> TryAcquireLeaseAsync(string containerName)
        {
            try
            {
                return await this.blobService.AcquireLeaseAsync(
                    this.blobService.GetAppendBlobReference(containerName, ExpirationMapBlobName), 
                    null);
            }
            catch (StorageException ex)
            {
                if (ex.RequestInformation != null && ex.RequestInformation.ErrorCode == BlobErrorCodeStrings.LeaseAlreadyPresent)
                {
                    this.logger.LogInfo("Another instance of BlobExpirationWorker is already running.");
                }

                // Don't rethrow the exception here. This is expected & accetable when running in a multi-instance env.
                return null;
            }
        }

        private async Task<IEnumerable<BlobExpirationInfo>> GetBlobExpirationListForContainerAsync(string containerName, string leaseId)
        {
            try
            {
                List<BlobExpirationInfo> expirationInfoList = new List<BlobExpirationInfo>();

                using (var ms = new MemoryStream())
                {
                    // Download expiration map blob
                    await this.blobService.DownloadAppendBlobToStreamAsync(
                        containerName, 
                        ExpirationMapBlobName, 
                        ms,
                        AccessCondition.GenerateLeaseCondition(leaseId),
                        new BlobRequestOptions(), 
                        new OperationContext());

                    ms.Position = 0;

                    // Deserialize blob to get the expiration list.
                    using (StreamReader reader = new StreamReader(ms))
                    {
                        string expirationInfo = reader.ReadLine();

                        if (!string.IsNullOrEmpty(expirationInfo))
                        {
                            BlobExpirationInfo blobExpirationInfo = this.DeserializeBlobExpirationInfo(expirationInfo);

                            if (blobExpirationInfo != null)
                            {
                                // This entry has some issues. Skip this one for now and continue with others.
                                expirationInfoList.Add(blobExpirationInfo);
                            }
                        }
                    }
                }

                return expirationInfoList;
            }
            catch (Exception ex) when (ex is StorageException || ex is JsonSerializationException)
            {
                this.logger.LogError(ex, "Error downloading/deserializing append blob content");
                throw;
            }
        }

        private BlobExpirationInfo DeserializeBlobExpirationInfo(string blobExpirationInfo)
        {
            try
            {
                return JsonConvert.DeserializeObject<BlobExpirationInfo>(blobExpirationInfo);
            }
            catch (JsonSerializationException jse)
            {
                // There is some issue with this entry. Ansorb the exception and skip this entry for now.
                this.logger.LogError(jse, $"Error deserializing the blob expiration info entry. Info: {blobExpirationInfo}");
                return null;
            }
        }

        private void SetExpiredAndNonExpiredBlobs(
            IEnumerable<BlobExpirationInfo> blobExpirationList, 
            List<BlobExpirationInfo> expiredBlobs, 
            List<BlobExpirationInfo> nonExpiredBlobs)
        {
            foreach (BlobExpirationInfo blobExpirationInfo in blobExpirationList)
            {
                if (blobExpirationInfo.BlobExpirationDate < DateTime.UtcNow)
                {
                    // Blob has already expired
                    expiredBlobs.Add(blobExpirationInfo);
                }
                else
                {
                    // Blob hasn't expired yet.
                    nonExpiredBlobs.Add(blobExpirationInfo);
                }
            }
        }

        private async Task DeleteBlobAsync(BlobExpirationInfo blobExpirationInfo)
        {
            try
            {
                ICloudBlob blob = await this.blobService.GetBlobReferenceAsync(blobExpirationInfo.BlobUri);
                await this.blobService.DeleteBlobIfExistsAsync(blob.Container.Name, blob.Name);
            }
            catch (StorageException se)
            {
                this.logger.LogError(se, $"Deleting expired blob with uri:{blobExpirationInfo.BlobUri.AbsoluteUri} failed!");
                throw;
            }
        }

        private async Task UpdateExpirationMapForContainerAsync(string containerName, List<BlobExpirationInfo> expirationList, string leaseId)
        {
            try
            {
                // Serialize the expiration list.
                var serializedExpirationInfoListBuilder = new StringBuilder();

                foreach (BlobExpirationInfo blobExpirationInfo in expirationList)
                {
                    serializedExpirationInfoListBuilder.AppendLine(JsonConvert.SerializeObject(blobExpirationInfo));
                }

                // Overwrite the existing map with the new map of only non-expired blobs.
                using (var ms = new MemoryStream(Encoding.UTF8.GetBytes(serializedExpirationInfoListBuilder.ToString())))
                {
                    CloudAppendBlob appendBlob = this.blobService.GetAppendBlobReference(containerName, ExpirationMapBlobName);

                    // Set the metadata for the last processed time.
                    appendBlob.Metadata.Add(LastProcessedDateUtcMetadataKey, DateTime.UtcNow.ToString("O"));

                    await this.blobService.UploadAppendBlobFromStreamAsync(
                        appendBlob, 
                        ms,
                        AccessCondition.GenerateLeaseCondition(leaseId),
                        new BlobRequestOptions(), 
                        new OperationContext());
                }
            }
            catch (Exception ex) when (ex is StorageException || ex is JsonSerializationException)
            {
                this.logger.LogError(ex, $"Failed to update the blob expiration map in container:{containerName}");
                throw;
            }
        }
    }
}