﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.LBSBatchService
{
    using System;
    using System.Collections.Concurrent;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.IO;
    using System.IO.Compression;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Text;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Common.Storage;
    using Microsoft.Azure.LocationServices.Common.Validation;
    using Microsoft.Azure.LocationServices.LBSBatchService.Properties;
    using Microsoft.Azure.LocationServices.Shared;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using Microsoft.WindowsAzure.Storage.Queue;
    using Newtonsoft.Json;
    using ProxyBatchShared.DataModels;

    /// <summary>
    /// This class handles/processes one queue message/batch request.
    /// It fetches all the sub-requests/queries contained in the batch and executes them concurrently using partitions.
    /// Once all the requests are completed it aggregates and serializes the response and uploads it back to Blob Storage.
    /// </summary>
    public class BatchQueueMessageHandler : QueueMessageHandler
    {
        // Max number of attempts to process a queue message
        protected static readonly int QueueMessageMaxAttempts = Settings.Default.QueueMessageMaxAttempts;

        // Number of paritions = (RPS) Number of requests executed concurrently per second
        protected static readonly int QueueMessageMaxPartitions = Settings.Default.BatchRequestsPerSecond;

        // HttpClient timeout in minutes. Default value is 100 secs.
        protected static readonly int HttpClientTimeoutInMinutes = Settings.Default.HttpClientTimeoutInMinutesForBatch;

        // Max number of times we should retry a "failed" request (response status code is not present in the white list property below)
        protected static readonly int MaxNumberOfRequestRetries = Settings.Default.MaxNumberOfRequestRetries;

        // Create the HttpClient lazily.
        protected static readonly Lazy<HttpClient> HttpClientLazy =
            new Lazy<HttpClient>(() => CreateHttpClient(TimeSpan.FromMinutes(HttpClientTimeoutInMinutes)));

        // Minimum amount of time that must be spent on a single request before moving on to the next request (To enable request throttling).
        private static readonly int MinRequestIntervalInMilliseconds = 1000;

        // A request is considered to be a failure if it's response status code is not present in the white list and hence is a candidate for retry.
        private static HashSet<string> requestRetryWhiteList;

        public BatchQueueMessageHandler(
            IQueueService queueService,
            IBlobService blobService,
            CloudQueue queue,
            IBlobHelper blobHelper,
            IBatchServiceLogger logger)
        {
            Assert.ParamIsNotNull(queueService, nameof(queueService));
            Assert.ParamIsNotNull(blobService, nameof(blobService));
            Assert.ParamIsNotNull(queue, nameof(queue));
            Assert.ParamIsNotNull(logger, nameof(logger));
            Assert.ParamIsNotNull(blobHelper, nameof(blobHelper));

            this.blobHelper = blobHelper;
            this.queueService = queueService;
            this.blobService = blobService;
            this.queue = queue;
            this.logger = logger;
        }

        // A request is considered to be a failure if it's response status code is not present in the white list and hence is a candidate for retry.
        private static HashSet<string> RequestRetryWhiteList => 
            requestRetryWhiteList ?? (requestRetryWhiteList = new HashSet<string>(Settings.Default.RequestRetryWhiteListCommaSeparated.Split(',')));

        public override async Task HandleMessageAsync(CloudQueueMessage queueMessage)
        {
            Assert.ParamIsNotNull(queueMessage, nameof(queueMessage));

            this.logger.LogInfo($"Starting processing for Batch queue message:{queueMessage.Id} attempt-number:{queueMessage.DequeueCount}");

            // Get the Blob associated with the queue message.
            ICloudBlob blob = await this.GetBlobFromBatchQueueMessageAsync(queueMessage);

            try
            {
                if (blob == null)
                {
                    // TODO: Think about cases where delete may not make sense.
                    // Fetching blob from the queue message failed. Cannot proceed ahead. Delete message and return.
                    await this.DeleteMessageAsync(queueMessage);
                    return;
                }

                // Get all the batch requests contained in the blob
                List<string> requests = await this.GetAllRequestsFromBlob(blob);

                // Process all the requests and aggregate the responses as BatchServiceResponse
                BatchServiceBlobStorageResult batchServiceBlobStorageResult = await this.ExecuteAllRequestsInBatch(requests);

                // Serialize the responses into a single batch response
                byte[] serializedResponse = this.SerializeBatchServiceResult(batchServiceBlobStorageResult);

                // Set the correct metadata and upload response to blob storage
                await this.UploadBatchResponseToBlob(blob, serializedResponse);

                // Delete the Message
                await this.DeleteMessageAsync(queueMessage);

                // Finally, set the expiration for the blob. The blob will be deleted when it's expired by the <BlobExpirationWorker>
                await this.blobHelper.SetBlobExpiryDate(blob, DateTime.UtcNow.AddDays(SharedUtilities.ResponseBlobExpiryTimeInDays));

                this.logger.LogInfo($"Successfully processed queue message:{queueMessage.Id} with {requests.Count} requests at attempt-number:{queueMessage.DequeueCount}");
            }
            catch (Exception ex)
            {
                // Something bad happened. Upload an error response to the blob storage
                if (blob != null)
                {
                    await this.UploadErrorResponseToBlob(blob, ex.Message);
                }

                this.logger.LogError(ex, $"Processing queue message with id:{queueMessage.Id} failed");
                await this.DeleteMessageAsync(queueMessage);
            }
        }

        public override HttpClient GetHttpClient() => HttpClientLazy.Value;

        public override int GetQueueMessageMaxAttempts() => QueueMessageMaxAttempts;

        private async Task<ICloudBlob> GetBlobFromBatchQueueMessageAsync(CloudQueueMessage queueMessage)
        {
            ICloudBlob blob = null;
            BatchServiceQueueMessage batchQueueMessage = null;

            try
            {
                // Deserialize the batch service queue message and use the blobUri to get a reference to the request blob.
                batchQueueMessage = JsonConvert.DeserializeObject<BatchServiceQueueMessage>(queueMessage.AsString);
                Uri blobUri = new Uri(batchQueueMessage.BlobUrl);

                blob = await this.blobService.GetBlobReferenceAsync(blobUri);
            }
            catch (Exception ex) when (ex is StorageException || ex is UriFormatException || ex is JsonSerializationException)
            {
                if (ex is UriFormatException)
                {
                    this.logger.LogError(ex, $"Queue message processing failed. Message with id:{queueMessage.Id} does not contain a valid URI:{queueMessage.AsString}");
                }

                // Something went really wrong here and we don't even have access to the blob so we can't update the request's failure status.
                // Very unlikely to happen so just log and return.
                if (ex is StorageException se)
                {
                    if (se.InnerException is WebException webEx)
                    {
                        HttpWebResponse webExResponse = webEx.Response as HttpWebResponse;

                        if (webExResponse?.StatusCode == HttpStatusCode.NotFound)
                        {
                            this.logger.LogError(ex, $"Queue message processing failed. Blob uri:{batchQueueMessage?.BlobUrl} in message with id:{queueMessage.Id} not found.");
                        }
                    }
                }
            }

            return blob;
        }

        private async Task<List<string>> GetAllRequestsFromBlob(ICloudBlob blob)
        {
            bool isDownloadCompleted = false;

            try
            {
                List<string> requests;

                using (var ms = new MemoryStream())
                {
                    // Download request content from blob
                    await this.blobService.DownloadBlockBlobToStreamAsync(blob.Container.Name, blob.Name, ms);
                    isDownloadCompleted = true;

                    // Deserialize blob to get all the batch requests.
                    requests = JsonConvert.DeserializeObject<List<string>>(Encoding.UTF8.GetString(ms.ToArray()));
                }

                return requests;
            }
            catch (Exception ex) when (ex is StorageException || ex is JsonSerializationException)
            {
                string errorMessage = isDownloadCompleted ? "Error deserializing blob content" : "Error downloading blob content";
                this.logger.LogError(ex, errorMessage);
                throw;
            }
        }

        private async Task<BatchServiceBlobStorageResult> ExecuteAllRequestsInBatch(IReadOnlyCollection<string> requests)
        {
            // Initialize the result object.
            var batchServiceBlobStorageResult = 
                new BatchServiceBlobStorageResult
                {
                    Summary = new BatchServiceResultSummary { TotalRequests = requests.Count, SuccessfulRequests = 0 }
                };

            Dictionary<string, byte[]> requestResponseMap = new Dictionary<string, byte[]>();

            // Create paritions from distinct requests.
            IList<IEnumerator<string>> partitions = Partitioner.Create(requests).GetPartitions(QueueMessageMaxPartitions);

            // Here Partitioner creates the "MaxPartitions" partitions which are operated on by an equivalent number of tasks.
            // New requests are fed into the partitions/tasks when they finish processing the old ones.
            var tasks = partitions.Select(partition => Task.Run(async () =>
            {
                while (partition.MoveNext())
                {
                    string currentRequest = partition.Current;

                    // Start a new timer instance per request.
                    var timer = Stopwatch.StartNew();

                    // Execute the Request.
                    using (HttpResponseMessage httpResponse = await this.ExecuteRequestWithRetry(currentRequest, MaxNumberOfRequestRetries).ConfigureAwait(false))
                    {
                        byte[] responseContent = Compress(await this.ReadResponseContent(httpResponse).ConfigureAwait(false));
                        requestResponseMap.Add(currentRequest, responseContent);

                        if (httpResponse.IsSuccessStatusCode)
                        {
                            batchServiceBlobStorageResult.Summary.SuccessfulRequests += 1;
                        }
                    }

                    timer.Stop();

                    // If the request completed before MinRequestIntervalInMilliseconds - elapse the remaining time before moving on to the next request (To enable request throttling)
                    if (timer.ElapsedMilliseconds < MinRequestIntervalInMilliseconds)
                    {
                        await Task.Delay(MinRequestIntervalInMilliseconds - (int)timer.ElapsedMilliseconds).ConfigureAwait(false);
                    }
                }
            }));

            await Task.WhenAll(tasks);

            batchServiceBlobStorageResult.CompressedBatchResults = this.SortResponsesByRequestsOrder(requests, requestResponseMap);

            return batchServiceBlobStorageResult;
        }

        private async Task<byte[]> ReadResponseContent(HttpResponseMessage httpResponseMessage)
        {
            // Just a precaution to prevent a null reference exception. Shouldn't happen in practice.
            if (httpResponseMessage.Content != null)
            {
                return await httpResponseMessage.Content.ReadAsByteArrayAsync();
            }

            this.logger.LogVerbose($"Request {httpResponseMessage.RequestMessage?.RequestUri?.AbsoluteUri} failed with status code {httpResponseMessage.StatusCode}");
            return Encoding.UTF8.GetBytes($"Request failed with status code {httpResponseMessage.StatusCode}");
        }

        private bool ShouldRetryRequest(HttpStatusCode httpStatusCode)
        {
            // Retry if the status code is NOT present in the white list.
            bool isStatusCodeInWhiteList = RequestRetryWhiteList.Contains(((int)httpStatusCode).ToString());
            return !isStatusCodeInWhiteList;
        }

        private async Task<HttpResponseMessage> ExecuteRequestWithRetry(string currentRequest, int maxRetryCount)
        {
            try
            {
                // Total try count = "first request itself" + "max number of retries"
                int totalTryCount = 1 + maxRetryCount;
                HttpResponseMessage httpResponse;

                do
                {
                    // Execute the request.
                    httpResponse = await this.ExecuteRequest(currentRequest);

                    // Decrement the total count and continue to loop only if this request "failed"
                    // (so we should retry the request) AND there are (> 0) retries left.
                    totalTryCount--;
                }
                while (this.ShouldRetryRequest(httpResponse.StatusCode) && totalTryCount > 0);

                return httpResponse;
            }
            catch (Exception ex)
            {
                this.logger.LogError(ex, $"Request: {currentRequest} failed unexpectedly with error: {ex.Message}");

                // Don't rethrow the exception here - let the other requests in the Batch execute and use a default error response for this one.
                return new HttpResponseMessage(HttpStatusCode.InternalServerError) { Content = new StringContent("Request failed with an unexpected/unknown error") };
            }
        }

        private List<byte[]> SortResponsesByRequestsOrder(IEnumerable<string> requests, Dictionary<string, byte[]> requestResponseMap)
        {
            return requests.Select(request => requestResponseMap.TryGetValue(request, out byte[] response) ? response : null).ToList();
        }

        private byte[] SerializeBatchServiceResult(BatchServiceBlobStorageResult batchServiceBlobStorageResult)
        {
            return Compress(Encoding.UTF8.GetBytes(JsonConvert.SerializeObject(batchServiceBlobStorageResult)));
        }

        private byte[] Compress(byte[] input)
        {
            if (input == null)
            {
                return new byte[0];
            }

            byte[] compressedData;

            using (var outputStream = new MemoryStream())
            {
                using (var zip = new GZipStream(outputStream, CompressionMode.Compress))
                {
                    zip.Write(input, 0, input.Length);
                }

                compressedData = outputStream.ToArray();
            }

            return compressedData;
        }

        private async Task UploadBatchResponseToBlob(ICloudBlob blob, byte[] serializedResponse)
        {
            Assert.ParamIsNotNull(blob, nameof(blob));
            Assert.ParamIsNotNull(serializedResponse, nameof(serializedResponse));

            try
            {
                // Get blob reference
                CloudBlockBlob responseBlob = this.blobService.GetBlockBlobReference(blob.Container.Name, blob.Name);

                // Set metadata
                responseBlob.Metadata.Add(SharedUtilities.BatchServiceResponseStatusMetadataKey, true.ToString());
                responseBlob.Metadata.Add(SharedUtilities.BatchServiceRequestFailedMetadataKey, false.ToString());

                // Upload response stream to blob
                using (var ms = new MemoryStream(serializedResponse))
                {
                    await this.blobService.UploadBlockBlobFromStreamAsync(responseBlob, ms);
                }
            }
            catch (StorageException ex)
            {
                this.logger.LogError(ex, $"Upload to blob {blob.Uri.AbsoluteUri} failed");
                throw;
            }
        }
    }
}