﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.LBSBatchService
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Common.Storage;
    using Microsoft.Azure.LocationServices.Common.Validation;
    using Microsoft.Azure.LocationServices.Shared;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Queue;
    using Microsoft.WindowsAzure.Storage.Table;
    using Microsoft.WindowsAzure.Storage.Table.Protocol;

    using Newtonsoft.Json;
    using ProxyBatchShared.DataModels;

    /// <summary>
    /// This class runs infinitely and continuously polls the batch queue for new messages to process.
    /// Each queue message signifies a batch request and messages can be processed in parallel.
    /// If no messages are found the worker goes to sleep for a configured time interval, wakes up and repeats the polling/processing.
    /// </summary>
    public class BatchQueuePollingWorker
    {
        // Table that stores the list of accounts for which we are processing batch requests at any given time.
        private const string ActiveAccountsTableName = "ActiveAccountsTable";

        // If queue itself does not exist the worker will sleep for this time
        private const int QueueDoesNotExistWaitTimeInMinutes = 60;

        // Number of messages to retrieve from Batch Queue. We use the Max possible.
        private const int NumberOfMessagesToRetrieve = 32;

        // The messages retrived from the queue by this instance will be invisible to other Batch Service instances for this time period.
        private const int MessageVisibilityTimeoutInSeconds = 60;

        // Max time period for which an account id can remain valid in the active accounts list.
        private const int ActiveAccountExpiryPeriodInDays = 1;

        // Time interval that's left for the queue message to become visible before we update it.
        private const int TimeIntervalToTriggerQueueMessageUpdateInMinutes = 10;

        // Dependencies
        private readonly IQueueService queueService;
        private readonly IBlobService blobService;
        private readonly ITableService tableService;
        private readonly IBlobHelper blobHelper;
        private readonly IBatchServiceLogger logger;
        private readonly PollingQueueConfig configuration;
        private readonly List<CloudQueueMessage> currentQueueMessages = new List<CloudQueueMessage>();

        private CloudQueue queue;
        private CloudTable table;

        public BatchQueuePollingWorker(
            IQueueService queueService,
            IBlobService blobService,
            ITableService tableService,
            IBlobHelper blobHelper,
            IBatchServiceLogger logger,
            PollingQueueConfig config)
        {
            Assert.ParamIsNotNull(queueService, nameof(queueService));
            Assert.ParamIsNotNull(blobService, nameof(blobService));
            Assert.ParamIsNotNull(tableService, nameof(tableService));
            Assert.ParamIsNotNull(blobHelper, nameof(blobHelper));
            Assert.ParamIsNotNull(logger, nameof(logger));

            this.queueService = queueService;
            this.blobService = blobService;
            this.tableService = tableService;
            this.blobHelper = blobHelper;
            this.logger = logger;
            this.configuration = config;
        }

        public async Task Start(string queueName, CancellationToken cancellationToken)
        {
            try
            {
                this.table = this.tableService.GetTableReference(ActiveAccountsTableName);
                await this.tableService.CreateTableIfNotExistsAsync(this.table, cancellationToken);

                this.queue = this.queueService.GetQueueReference(queueName);
                await this.SleepIfQueueDoesNotExist(cancellationToken);

                // Polling Loop Starts
                while (true)
                {
                    cancellationToken.ThrowIfCancellationRequested();

                    // Check to see if any new messages have been added to the queue since we last checked.
                    List<CloudQueueMessage> newQueueMessages = await this.GetQueueMessagesAsync();

                    foreach (CloudQueueMessage queueMessage in newQueueMessages)
                    {
                        // Add to current active list.
                        this.currentQueueMessages.Add(queueMessage);

                        // Run the message and move on. When it completes it'll remove itself from the active list.
                        this.HandleMessageAsync(queueMessage).FireAndForget();
                    }

                    // Refresh the visibility timeout of currently active messages.
                    await this.CheckAndUpdateVisibilityTimeoutAsync();

                    this.logger.LogVerbose($"Current count of items in the Queue : {this.currentQueueMessages.Count}");

                    // TODO: Currently this uses the Linear Backoff policy - Implement a configurable Exponential/Custom Backoff Policy
                    await this.SleepFor(TimeSpan.FromSeconds(this.configuration.QueueWorkerDefaultSleepTimeInSeconds), cancellationToken);
                }
            }
            catch (Exception ex)
            {
                // Something unexpected happened.
                this.logger.LogError(ex, "Batch Queue Polling Worker Failed!");
                throw;
            }
        }

        /// <summary>
        /// Get more messages/batch requests to process. 
        /// If the account that submitted the batch-request already has another request that is currently
        /// being processed by this or any other batch service instance then skip that message. 
        /// </summary>
        private async Task<List<CloudQueueMessage>> GetQueueMessagesAsync()
        {
            // Number of new messages to fetch is (total capacity - currently active tasks)
            int messageCountToFetch = this.configuration.MaxQueueMessagesToProcessInParallel - this.currentQueueMessages.Count;

            if (messageCountToFetch <= 0)
            {
                // Only proceed if we are not at full capacity and need to fetch more messages.
                return new List<CloudQueueMessage>();
            }

            // Get all the accounts that are currently active across all instances 
            Dictionary<string, ActiveAccountEntity> activeAccountsMap = this.GetActiveAccounts();

            // Get max number of messages from queue
            List<CloudQueueMessage> queueMessages =
                (await this.queueService.GetMessagesAsync(
                    this.queue, 
                    NumberOfMessagesToRetrieve, 
                    TimeSpan.FromSeconds(MessageVisibilityTimeoutInSeconds), 
                    null, 
                    null)).ToList();

            var finalMessageList = new List<CloudQueueMessage>();

            foreach (CloudQueueMessage queueMessage in queueMessages)
            {
                BatchServiceQueueMessage batchQueueMessage = this.GetBatchQueueMessage(queueMessage);

                if (!activeAccountsMap.ContainsKey(batchQueueMessage.AccountId))
                {
                    // This account is not currently active. Try adding it to active accounts table to check if it's still available to process.
                    if (await this.AddToActiveAccountsAsync(batchQueueMessage.AccountId, activeAccountsMap))
                    {
                        if (!await this.AddMessageToFinalListAsync(finalMessageList, queueMessage))
                        {
                            // Something went wrong while adding the message to the final list. Remove it from the active list and move on to the next option.
                            await this.RemoveFromActiveAccountsAsync(batchQueueMessage.AccountId, activeAccountsMap);
                            continue;
                        }
                    }
                }
                else
                {
                    // It seems like like this account maybe active (or is present but might have expired - Instance went down before removing it)
                    ActiveAccountEntity activeAccountEntity = activeAccountsMap[batchQueueMessage.AccountId];

                    if (activeAccountEntity.ExpirationDate > DateTime.UtcNow)
                    {
                        // Entity hasn't expired yet so this account is currently being processed. Move to evaluate next item.
                        continue;
                    }

                    // Entity has expired. This might happen if the batch service instance athat added this entry went down abruptly and wasn't able to delete the entry. See if we can claim it.
                    activeAccountEntity.ExpirationDate = DateTime.UtcNow.AddDays(ActiveAccountExpiryPeriodInDays);

                    if (await this.UpdateActiveAccountsAsync(activeAccountEntity, activeAccountsMap))
                    {
                        if (!await this.AddMessageToFinalListAsync(finalMessageList, queueMessage))
                        {
                            // Something went wrong while adding the message to the final list. Remove it from the active list and move on to the next option.
                            await this.RemoveFromActiveAccountsAsync(batchQueueMessage.AccountId, activeAccountsMap);
                            continue;
                        }
                    }
                    else
                    {
                        // Looks like someone else claimed it before we could. Let's move on to the next message.
                        continue;
                    }
                }

                if (finalMessageList.Count.Equals(messageCountToFetch))
                {
                    break;
                }
            }
            
            // Return Handler Tasks for the fetched messages
            return finalMessageList;
        }

        private async Task HandleMessageAsync(CloudQueueMessage queueMessage)
        {
            // Process the request.
            await new BatchQueueMessageHandler(this.queueService, this.blobService, this.queue, this.blobHelper, this.logger).HandleMessageAsync(queueMessage);

            // Now that the request is done, remove the account from Active Accounts Table.
            await this.RemoveFromActiveAccountsAsync(this.GetBatchQueueMessage(queueMessage).AccountId);

            // Also, remove the message from the current active messages list
            this.currentQueueMessages.Remove(queueMessage);
        }

        private async Task<bool> AddMessageToFinalListAsync(
            List<CloudQueueMessage> finalMessageList, 
            CloudQueueMessage queueMessage)
        {
            try
            {
                // Update message visibility timeout
                await this.queueService.UpdateMessageAsync(
                    this.queue,
                    queueMessage,
                    TimeSpan.FromMinutes(this.configuration.QueueMessageVisibilityTimeoutInMinutes),
                    MessageUpdateFields.Visibility);

                finalMessageList.Add(queueMessage);

                return true;
            }
            catch (StorageException)
            {
                // Something went wrong while updating the visibility timeout. Can't use this message anymore. Don't expect this to happen at all. Absorb the exception.
                return false;
            }
        }

        #region ActiveAccountsTable Operations

        private Dictionary<string, ActiveAccountEntity> GetActiveAccounts()
        {
            TableQuery<ActiveAccountEntity> getActiveAccountsQuery = new TableQuery<ActiveAccountEntity>()
                .Where(TableQuery.GenerateFilterCondition("PartitionKey", QueryComparisons.Equal, ActiveAccountEntity.ActiveAccountEntityPartitionKey));

            IEnumerable<ActiveAccountEntity> activeAccounts = this.tableService.ExecuteQuery(this.table, getActiveAccountsQuery);

            return activeAccounts.ToDictionary(activeAccountEntity => activeAccountEntity.RowKey);
        }

        private async Task<bool> AddToActiveAccountsAsync(string accountId, Dictionary<string, ActiveAccountEntity> activeAccountsMap)
        {
            try
            {
                // Add the account to the remote active list 
                var activeAccountEntity = new ActiveAccountEntity(accountId, DateTime.UtcNow.AddDays(ActiveAccountExpiryPeriodInDays));
                await this.tableService.InsertAsync(this.table, activeAccountEntity);

                // Add the account to the local active list (saves remote check for remaining messages)
                activeAccountsMap.Add(accountId, activeAccountEntity);

                return true;
            }
            catch (StorageException se)
            {
                if (se.RequestInformation != null 
                    && se.RequestInformation.HttpStatusCode == (int)HttpStatusCode.Conflict 
                    && se.RequestInformation.ErrorCode == TableErrorCodeStrings.EntityAlreadyExists)
                {
                    // Entity already exists in the Table.
                    return false;
                }
            }

            // default
            return false;
        }

        private async Task RemoveFromActiveAccountsAsync(string accountId)
        {
            try
            {
                // First fetch the entity from the table.
                ActiveAccountEntity activeAccountEntity = await this.FetchActiveAccountAsync(accountId);

                if (activeAccountEntity == null)
                {
                    // The entity was not found in the table
                    return;
                }

                // Entity located - Delete it
                await this.tableService.DeleteAsync(this.table, activeAccountEntity);
            }
            catch (StorageException)
            {
                // Absorb any exceptions. Delete should just go through without any issues. Try/Catch is just for defensive purposes.
            }
        }

        private async Task RemoveFromActiveAccountsAsync(string accountId, Dictionary<string, ActiveAccountEntity> activeAccountsMap)
        {
            // Remove the account from the remote active list 
            await this.RemoveFromActiveAccountsAsync(accountId);

            // Remove the account from the local active list (saves remote check for remaining messages)
            activeAccountsMap.Remove(accountId);
        }

        private async Task<bool> UpdateActiveAccountsAsync(ActiveAccountEntity activeAccountEntity, Dictionary<string, ActiveAccountEntity> activeAccountsMap)
        {
            try
            {
                // Try updating the existing entity in the remote active list
                TableResult updateResult = await this.tableService.ReplaceAsync(this.table, activeAccountEntity);
                activeAccountEntity = updateResult.Result as ActiveAccountEntity;

                // Try updating the existing entity in the local active list (saves remote check for remaining messages)
                activeAccountsMap[activeAccountEntity.RowKey] = activeAccountEntity;

                return true;
            }
            catch (Exception ex) when (ex is StorageException || ex is NullReferenceException)
            {
                // Absorb any exceptions. Replace should just go through without any issues. Try/Catch is just for defensive purposes.
                return false;
            }
        }

        private async Task<ActiveAccountEntity> FetchActiveAccountAsync(string accountId)
        {
            TableOperation retrieveOperation = TableOperation.Retrieve<ActiveAccountEntity>(ActiveAccountEntity.ActiveAccountEntityPartitionKey, accountId);
            TableResult result = await this.tableService.ExecuteAsync(this.table, retrieveOperation);
            ActiveAccountEntity activeAccountEntity = result.Result as ActiveAccountEntity;

            return activeAccountEntity;
        }

        #endregion

        private BatchServiceQueueMessage GetBatchQueueMessage(CloudQueueMessage queueMessage)
        {
            return JsonConvert.DeserializeObject<BatchServiceQueueMessage>(queueMessage.AsString);
        }

        private async Task CheckAndUpdateVisibilityTimeoutAsync()
        {
            try
            {
                foreach (CloudQueueMessage queueMessage in this.currentQueueMessages.AsReadOnly())
                {
                    // Is the message about to become visible in the next 10 minutes
                    if (queueMessage.NextVisibleTime.HasValue && 
                        queueMessage.NextVisibleTime.Value - DateTime.UtcNow < TimeSpan.FromMinutes(TimeIntervalToTriggerQueueMessageUpdateInMinutes))
                    {
                        // Update message visibility timeout
                        await this.queueService.UpdateMessageAsync(
                            this.queue,
                            queueMessage,
                            TimeSpan.FromMinutes(this.configuration.QueueMessageVisibilityTimeoutInMinutes),
                            MessageUpdateFields.Visibility).ConfigureAwait(false);
                    }
                }
            }
            catch (Exception ex) when (ex is StorageException || ex is NullReferenceException)
            {
                // Something went wrong while updating the visibility timeout. Can't really do much at this point. 
                // Just absorb the exception for now and we will try this again soon.
            }
        }

        private async Task SleepIfQueueDoesNotExist(CancellationToken cancellationToken)
        {
            while (true)
            {
                cancellationToken.ThrowIfCancellationRequested();

                if (await this.queueService.QueueExistsAsync(this.queue, cancellationToken))
                {
                    // Success! Queue Exists - Go back and Start Polling for messages.
                    return;
                }

                await this.SleepFor(TimeSpan.FromMinutes(QueueDoesNotExistWaitTimeInMinutes), cancellationToken);
            }
        }

        private async Task SleepFor(TimeSpan time, CancellationToken cancellationToken)
        {
            await Task.Delay(time, cancellationToken);
        }

        private async Task<List<Task>> GetQueueMessageTasksWithNoConcurrencyControlAsync(int currentlyActiveMessageTasks = 0)
        {
            // Number of new messages to fetch is (total capacity - currently active tasks)
            int messageCountToFetch = this.configuration.MaxQueueMessagesToProcessInParallel - currentlyActiveMessageTasks;

            // Try and fetch messages from Azure Queue
            List<CloudQueueMessage> queueMessages =
                (await this.queueService.GetMessagesAsync(
                     this.queue,
                     messageCountToFetch,
                     TimeSpan.FromMinutes(this.configuration.QueueMessageVisibilityTimeoutInMinutes),
                     null,
                     null)).ToList();

            // Return Handler Tasks for the fetched messages
            return
                queueMessages.Count > 0
                    ? queueMessages.Select(this.HandleMessageAsync).ToList()
                    : new List<Task>();
        }
    }
}