﻿// --------------------------------------------------------------------------------------------------------------------
// <copyright file="MatrixQueueMessageHandler.cs" company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// <summary>
//   Defines the MatrixQueueMessageHandler type.
// </summary>
// --------------------------------------------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.LBSBatchService
{
    using System;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Text;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Common.Storage;
    using Microsoft.Azure.LocationServices.Common.Validation;
    using Microsoft.Azure.LocationServices.LBSBatchService.Properties;
    using Microsoft.Azure.LocationServices.Shared;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using Microsoft.WindowsAzure.Storage.Queue;
    using Newtonsoft.Json;

    /// <summary>
    /// This class handles/processes one queue message. 
    /// Every queue message has a redirect URL TomTom has given for the long running route matrix requests.
    /// Once the results are available, the blob is updated with the results and billing is notified.
    /// </summary>
    public class MatrixQueueMessageHandler : QueueMessageHandler
    {
        // Max number of attempts to process a queue message
        protected static readonly int QueueMessageMaxAttempts = Settings.Default.QueueMessageMaxAttempts;

        // HttpClient timeout in minutes. Default value is 100 secs.
        protected static readonly int HttpClientTimeoutInMinutes = Settings.Default.HttpClientTimeoutInMinutesForRouteMatrix;
        
        // Create the HttpClient lazily.
        protected static readonly Lazy<HttpClient> HttpClientLazy = new Lazy<HttpClient>(() => CreateHttpClient(TimeSpan.FromMinutes(HttpClientTimeoutInMinutes)));

        public MatrixQueueMessageHandler(
            IQueueService queueService,
            IBlobService blobService,
            CloudQueue queue,
            IBlobHelper blobHelper,
            IBatchServiceLogger logger)
        {
            Assert.ParamIsNotNull(queueService, nameof(queueService));
            Assert.ParamIsNotNull(blobService, nameof(blobService));
            Assert.ParamIsNotNull(queue, nameof(queue));
            Assert.ParamIsNotNull(blobHelper, nameof(blobHelper));
            Assert.ParamIsNotNull(logger, nameof(logger));

            this.queueService = queueService;
            this.blobService = blobService;
            this.queue = queue;
            this.blobHelper = blobHelper;
            this.logger = logger;
        }

        public override async Task HandleMessageAsync(CloudQueueMessage queueMessage)
        {
            Assert.ParamIsNotNull(queueMessage, nameof(queueMessage));

            this.logger.LogInfo($"Starting processing for Matrix queue message:{queueMessage.Id} attempt-number:{queueMessage.DequeueCount}");

            // Get the Blob associated with the queue message.
            ICloudBlob blob = await this.GetBlobFromMessageAsync(queueMessage);

            try
            {
                if (blob == null)
                {
                    // TODO: Think about cases where delete may not make sense.
                    // Fetching blob from the queue message failed. Cannot proceed ahead. Delete message and return.
                    await this.DeleteMessageAsync(queueMessage);
                    return;
                }

                // Validate queue message.
                if (!this.IsValidQueueMessage(queueMessage))
                {
                    // Message is invalid and will not be processed. Deleting message from queue and uploading error to blob.
                    await this.DeleteMessageAsync(queueMessage);
                    await this.UploadErrorResponseToBlob(blob, "Matrix request failed. Please try issuing the request again.");
                    return;
                }

                var metadata = blob.Metadata;
                if (!metadata.TryGetValue(SharedUtilities.BatchServiceResponseStatusMetadataKey, out string responseStatus) ||
                    !metadata.TryGetValue(SharedUtilities.BatchServiceRequestFailedMetadataKey, out string requestFailedStatus))
                {
                    // Just a safe check. This should not happen under any circumstances.
                    throw new InvalidOperationException("Unexpected Error - Matrix status metadata is missing in blob");
                }

                bool isResponseAvailable = bool.Parse(responseStatus);

                if (!isResponseAvailable)
                {
                    var requestInfo = await this.GetRequestFromBlob(blob);
                    var absoluteUri = new Uri(requestInfo.BaseUri, requestInfo.RelativeURi);
                    var responseMessage = await this.ExecuteRequest(absoluteUri.ToString());

                    if (responseMessage.StatusCode == HttpStatusCode.Accepted || responseMessage.StatusCode == HttpStatusCode.SeeOther)
                    {
                        if (responseMessage.Headers.Location != requestInfo.RelativeURi)
                        {
                            var newRequestInfo = new RouteMatrixRetryInfo(requestInfo.BaseUri, responseMessage.Headers.Location, requestInfo.AccountId, requestInfo.TrackingId);
                            await this.UploadToBlob(blob, JsonConvert.SerializeObject(newRequestInfo), false, false);
                        }

                        // Add a new Q entry to get back in the Q. The current one get deleted down below.
                        await this.AddMessageToQueueAsync(new CloudQueueMessage(queueMessage.AsString));
                    }
                    else
                    {
                        var requestSucceeded = responseMessage.StatusCode == HttpStatusCode.OK;
                        await this.UploadToBlob(blob, responseMessage.Content.ReadAsStringAsync().Result, true, !requestSucceeded);
                        var responseStatusCodes = SharedUtilities.GetResponseCodesFromTomTomRouteMatrixResponseContent(responseMessage.Content.ReadAsStringAsync().Result);
                        if (!string.IsNullOrEmpty(requestInfo.AccountId))
                        {
                            var statusList = responseStatusCodes.Select(t => new Tuple<string, HttpStatusCode>(SharedUtilities.PostMatrixMethodName, (HttpStatusCode)t)).ToList();
                            BillingAPI.BillingLogWriter.Instance.WriteLog(requestInfo.AccountId, statusList, requestInfo.TrackingId);
                        }
                    }
                }

                // Delete the Message
                await this.DeleteMessageAsync(queueMessage);

                this.logger.LogInfo($"Successfully processed Matrix queue message:{queueMessage.Id} at attempt-number:{queueMessage.DequeueCount}");
            }
            catch (Exception ex)
            {
                // Something bad happened. Upload an error response to the blob storage
                if (blob != null)
                {
                    await this.UploadErrorResponseToBlob(blob, ex.Message);
                }

                this.logger.LogError(ex, $"Processing Matrix queue message with id:{queueMessage.Id} failed");
                await this.DeleteMessageAsync(queueMessage);
            }
        }

        public override HttpClient GetHttpClient() => HttpClientLazy.Value;

        public override int GetQueueMessageMaxAttempts() => QueueMessageMaxAttempts;

        private async Task<RouteMatrixRetryInfo> GetRequestFromBlob(ICloudBlob blob)
        {
            bool isDownloadCompleted = false;

            try
            {
                RouteMatrixRetryInfo requestInfo;

                using (var ms = new MemoryStream())
                {
                    // Download request content from blob
                    await this.blobService.DownloadBlockBlobToStreamAsync(blob.Container.Name, blob.Name, ms);
                    isDownloadCompleted = true;

                    // Deserialize blob to get all the batch requests.
                    requestInfo = JsonConvert.DeserializeObject<RouteMatrixRetryInfo>(Encoding.UTF8.GetString(ms.ToArray()));
                }

                return requestInfo;
            }
            catch (Exception ex) when (ex is StorageException || ex is JsonSerializationException)
            {
                string errorMessage = isDownloadCompleted ? "Error deserializing blob content" : "Error downloading blob content";
                this.logger.LogError(ex, errorMessage);
                throw;
            }
        }
    }
}
