﻿// --------------------------------------------------------------------------------------------------------------------
// <copyright file="QueueMessageHandlerBase.cs" company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// <summary>
//   Defines the QueueMessageHandlerBase type.
// </summary>
// --------------------------------------------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.LBSBatchService
{
    using System;
    using System.Collections.Generic;
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Common.Storage;
    using Microsoft.Azure.LocationServices.Common.Validation;
    using Microsoft.Azure.LocationServices.Shared;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using Microsoft.WindowsAzure.Storage.Queue;

    public abstract class QueueMessageHandler
    {
        // Dependencies
        protected IBlobHelper blobHelper;
        protected IQueueService queueService;
        protected IBlobService blobService;
        protected CloudQueue queue;
        protected IBatchServiceLogger logger;
        
        public abstract Task HandleMessageAsync(CloudQueueMessage queueMessage);

        public abstract HttpClient GetHttpClient();

        public abstract int GetQueueMessageMaxAttempts();
        
        protected static HttpClient CreateHttpClient(TimeSpan clientTimeout)
        {
            return new HttpClient
                       {
                           Timeout = clientTimeout
                       };
        }

        protected async Task<ICloudBlob> GetBlobFromMessageAsync(CloudQueueMessage queueMessage)
        {
            ICloudBlob blob = null;

            try
            {
                Uri blobUri = new Uri(queueMessage.AsString);

                blob = await this.blobService.GetBlobReferenceAsync(blobUri);
            }
            catch (Exception ex) when (ex is StorageException || ex is UriFormatException)
            {
                if (ex is UriFormatException)
                {
                    this.logger.LogError(ex, $"Queue message processing failed. Message with id:{queueMessage.Id} does not contain a valid URI:{queueMessage.AsString}");
                }

                // Something went really wrong here and we don't even have access to the blob so we can't update the request's failure status.
                // Very unlikely to happen so just log and return.
                if (ex is StorageException se)
                {
                    if (se.InnerException is WebException webEx)
                    {
                        HttpWebResponse webExResponse = webEx.Response as HttpWebResponse;

                        if (webExResponse?.StatusCode == HttpStatusCode.NotFound)
                        {
                            this.logger.LogError(ex, $"Queue message processing failed. Blob uri:{queueMessage.AsString} in message with id:{queueMessage.Id} not found.");
                        }
                    }
                }
            }

            return blob;
        }

        protected async Task DeleteMessageAsync(CloudQueueMessage queueMessage)
        {
            try
            {
                await this.queueService.DeleteMessageAsync(this.queue, queueMessage);
            }
            catch (StorageException se)
            {
                this.logger.LogError(se, $"Deleting queue message with id {queueMessage.Id} failed");
                throw;
            }
        }

        protected async Task AddMessageToQueueAsync(CloudQueueMessage queueMessage)
        {
            try
            {
                await this.queueService.AddMessageAsync(this.queue, queueMessage);
            }
            catch (StorageException se)
            {
                this.logger.LogError(se, $"Adding queue message with id {queueMessage.Id} failed");
                throw;
            }
        }

        protected bool IsValidQueueMessage(CloudQueueMessage queueMessage)
        {
            // Check if max attempts reached on the message using the dequeue count
            return queueMessage.DequeueCount <= GetQueueMessageMaxAttempts();
        }

        protected async Task<HttpResponseMessage> ExecuteRequest(string query)
        {
            Assert.ParamIsNotNullOrEmpty(query, nameof(query));

            try
            {
                // Only 'GET' calls are supported.
                return await this.GetHttpClient().GetAsync(query);
            }
            catch (Exception ex) when (ex is InvalidOperationException || ex is ArgumentNullException || ex is HttpRequestException)
            {
                // Log Error and return a default error response
                this.logger.LogError(ex, $"Request to uri {query} failed");

                return new HttpResponseMessage(HttpStatusCode.InternalServerError)
                {
                    Content = new StringContent("Request failed with an unexpected/unknown error")
                };
            }
        }

        protected async Task UploadErrorResponseToBlob(ICloudBlob blob, string errorMessage)
        {
            await this.UploadToBlob(blob, errorMessage, false, true);
        }

        protected async Task UploadToBlob(ICloudBlob blob, string content, bool hasResponse, bool hasFailed)
        {
            Assert.ParamIsNotNull(blob, nameof(blob));
            Assert.ParamIsNotNullOrEmpty(content, nameof(content));

            var metadata = new Dictionary<string, string>
            {
                { SharedUtilities.BatchServiceResponseStatusMetadataKey, hasResponse.ToString() },
                { SharedUtilities.BatchServiceRequestFailedMetadataKey, hasFailed.ToString() }
            };

            try
            {
                await this.blobService.UploadDocumentAsync(blob.Container.Name, blob.Name, content, metadata);
            }
            catch (StorageException ex)
            {
                this.logger.LogError(ex, $"Upload to blob {blob.Uri.AbsoluteUri} failed");
                throw;
            }
        }
    }
}