﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.LBSBatchService
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Common.Storage;
    using Microsoft.Azure.LocationServices.Common.Validation;
    using Microsoft.Azure.LocationServices.Shared;
    using Microsoft.WindowsAzure.Storage.Queue;

    /// <summary>
    /// This class runs infinitely and continuously polls the batch queue for new messages to process.
    /// Each queue message signifies a batch request and messages can be processed in parallel.
    /// If no messages are found the worker goes to sleep for a configured time interval, wakes up and repeats the polling/processing.
    /// </summary>
    public class QueuePollingWorker
    {
        // If queue itself does not exist the worker will sleep for this time
        private const int QueueDoesNotExistWaitTimeInMinutes = 60;

        // Dependencies
        private readonly IQueueService QueueService;
        private readonly IBlobService BlobService;
        private readonly IBlobHelper BlobHelper;
        private readonly IBatchServiceLogger Logger;
        private readonly PollingQueueConfig Configuration;
        private readonly IMessageHandlerFactory MesssageHandlerFactory;

        private CloudQueue queue;

        public QueuePollingWorker(
            IMessageHandlerFactory messsageHandlerFactory,
            IQueueService queueService, 
            IBlobService blobService,
            IBlobHelper blobHelper,
            IBatchServiceLogger logger,
            PollingQueueConfig config)
        {
            Assert.ParamIsNotNull(messsageHandlerFactory, nameof(messsageHandlerFactory));
            Assert.ParamIsNotNull(queueService, nameof(queueService));
            Assert.ParamIsNotNull(blobService, nameof(blobService));
            Assert.ParamIsNotNull(blobHelper, nameof(blobHelper));
            Assert.ParamIsNotNull(logger, nameof(logger));

            this.MesssageHandlerFactory = messsageHandlerFactory;
            this.QueueService = queueService;
            this.BlobService = blobService;
            this.BlobHelper = blobHelper;
            this.Logger = logger;
            this.Configuration = config;
        }

        public async Task Start(string queueName, CancellationToken cancellationToken)
        {
            try
            {
                var typeName = MesssageHandlerFactory.GetType().Name;
                this.queue = this.QueueService.GetQueueReference(queueName);

                // Sleep till Queue doesn't exist.
                await this.SleepIfQueueDoesNotExist(cancellationToken);
                
                // Polling Loop Starts
                while (true)
                {
                    cancellationToken.ThrowIfCancellationRequested();

                    // No active queue message tasks at this time. Pre-fill list if messages are available.
                    List<Task> activeQueueTasks = await this.GetQueueMessageTasksAsync();

                    while (activeQueueTasks.Count > 0)
                    {
                        this.Logger.LogVerbose($"{typeName}: Processing long running matrix requests from the Queue. Items in the Queue : {activeQueueTasks.Count}");
                        
                        // Wait for any one of the currently active tasks to complete
                        Task completedTask = await Task.WhenAny(activeQueueTasks);

                        // Remove the completed task
                        activeQueueTasks.Remove(completedTask);

                        // Check if we have any more tasks that we can process
                        activeQueueTasks.AddRange(await this.GetQueueMessageTasksAsync(activeQueueTasks.Count));
                    }

                    this.Logger.LogVerbose($"{typeName}: no more long running matrix requests in the Queue. Going to sleep for {this.Configuration.QueueWorkerDefaultSleepTimeInSeconds} seconds");

                    // TODO: Currently this uses the Linear Backoff policy - Implement a configurable Exponential/Custom Backoff Policy
                    await this.SleepFor(TimeSpan.FromSeconds(this.Configuration.QueueWorkerDefaultSleepTimeInSeconds), cancellationToken);
                }
            }
            catch (Exception ex)
            {
                // Something unexpected happened.
                this.Logger.LogError(ex, "Queue Polling Worker Failed!");
                throw;
            }
        }

        private async Task SleepIfQueueDoesNotExist(CancellationToken cancellationToken)
        {
            while (true)
            {
                cancellationToken.ThrowIfCancellationRequested();

                if (await this.QueueService.QueueExistsAsync(this.queue, cancellationToken))
                {
                    // Success! Queue Exists - Go back and Start Polling for messages.
                    return;
                }

                await this.SleepFor(TimeSpan.FromMinutes(QueueDoesNotExistWaitTimeInMinutes), cancellationToken);
            }
        }

        private async Task<List<Task>> GetQueueMessageTasksAsync(int currentlyActiveMessageTasks = 0)
        {
            // Number of new messages to fetch is (total capacity - currently active tasks)
            int messageCountToFetch = this.Configuration.MaxQueueMessagesToProcessInParallel - currentlyActiveMessageTasks;

            // Try and fetch messages from Azure Queue
            List<CloudQueueMessage> queueMessages = 
                (await this.QueueService.GetMessagesAsync(
                    this.queue, 
                    messageCountToFetch, 
                    TimeSpan.FromMinutes(this.Configuration.QueueMessageVisibilityTimeoutInMinutes), 
                    null, 
                    null)).ToList();

            // Return Handler Tasks for the fetched messages
            return 
                queueMessages.Count > 0 
                ? queueMessages.Select(this.HandleMessageAsync).ToList() 
                : new List<Task>();
        }

        private async Task HandleMessageAsync(CloudQueueMessage queueMessage)
        {
            await this.MesssageHandlerFactory.CreateInstance(
                this.QueueService,
                this.BlobService,
                this.queue,
                this.BlobHelper,
                this.Logger).HandleMessageAsync(queueMessage);
        }

        private async Task SleepFor(TimeSpan time, CancellationToken cancellationToken)
        {
            await Task.Delay(time, cancellationToken);
        }
    }
}