﻿// --------------------------------------------------------------------------------------------------------------------
// <copyright file="OSMMirrorOnAzure.cs" company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// <summary>
//   An instance of this class is created for each service instance by the Service Fabric runtime.
// </summary>
// --------------------------------------------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.OSMMirrorOnAzure.Configuration
{
    using System;
    using System.Globalization;

    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;

    /// <summary>
    /// Represents the configuration necessary to operate in an environment.
    /// </summary>
    /// <remarks>
    /// This class should only contain members that differ from environment to environment. When something doesn't 
    /// vary, it should not be present in this class, and instead should be a compile-time constant. If a property
    /// is no longer used, or no longer varies, it should be removed.
    /// </remarks>
    public abstract class OSMMirrorHostSettings
    {
        /// <summary>
        /// Gets the Instrumentation Parameters
        /// </summary>
        public abstract InitParams MdmParameters { get; }

        /// <summary>
        /// Gets the current region.
        /// </summary>
        protected static string CurrentRegion
        {
            get
            {
                string currentRegion = Environment.GetEnvironmentVariable("Region");
                return currentRegion ?? "DEFAULT";
            }
        }

        /// <summary>
        /// Gets the current environment.
        /// </summary>
        /// <exception cref="ArgumentException">
        /// Throws if the current environment is not specified
        /// </exception>
        private static string CurrentEnvironment
        {
            get
            {
                string currentEnvironment = Environment.GetEnvironmentVariable("Environment");

                if (string.IsNullOrEmpty(currentEnvironment))
                {
                    throw new ArgumentException("Current Environment could not be determined");
                }

                return currentEnvironment;
            }
        }

        /// <summary>
        /// Create an environment definition from the environment environment variable.
        /// </summary>
        /// <returns>The created environment definition.</returns>
        public static OSMMirrorHostSettings CreateFromEnvironment()
        {
            switch (CurrentEnvironment.ToUpperInvariant())
            {
                case LogicalEnvironments.Development:
                    return new OSMMirrorHostDevSettings();
                case LogicalEnvironments.Test:
                    return new OSMMirrorHostTestSettings();
                case LogicalEnvironments.Staging:
                    return new OSMMirrorHostStagingSettings();
                case LogicalEnvironments.Production:
                    return new OSMMirrorHostProdSettings();
                default:
                    throw new ArgumentException(
                        string.Format(
                            CultureInfo.InvariantCulture,
                            "Environment {0} does not have a corresponding Environment definition defined",
                            CurrentEnvironment));
            }
        }
    }
}
