﻿// --------------------------------------------------------------------------------------------------------------------
// <copyright file="HtmlBuilder.cs" company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// <summary>
//   Defines the HtmlBuilder type.
// </summary>
// --------------------------------------------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.OSMMirrorOnAzure
{
    using System;
    using System.Collections.Generic;
    using System.Text;

    /// <summary>
    /// The html builder.
    /// </summary>
    public class HtmlBuilder
    {
        private string titleText;

        private string pageHeadingText;

        private Tuple<string, string> previousPageTextAndUrl;

        private List<string> tableHeaderList;

        private List<Tuple<string, string>> directoryNameUrlList;

        private List<List<string>> fileContentsList;

        public HtmlBuilder()
        {
            this.directoryNameUrlList = new List<Tuple<string, string>>();
            this.tableHeaderList = new List<string>();
            this.fileContentsList = new List<List<string>>();
        }

        public void SetTitle(string title)
        {
            this.titleText = title;
        }

        public void SetHeading(string heading)
        {
            this.pageHeadingText = heading;
        }

        public void SetPreviousPageLink(string previousPageName, string previousPageUrl)
        {
            this.previousPageTextAndUrl = new Tuple<string, string>(previousPageName, previousPageUrl);
        }

        public void SetTableHeaders(params string[] tableHeaderNames)
        {
            if (tableHeaderNames.Length == 0)
            {
                throw new Exception("Table should have atleast one column.");
            }

            this.tableHeaderList.AddRange(tableHeaderNames);
        }

        public void AddDirectoryToTable(string directoryName, string directoryUrl)
        {
            this.directoryNameUrlList.Add(new Tuple<string, string>(directoryName, directoryUrl));
        }

        public void AddFileContentRow(string fileUrl, params string[] rowContents)
        {
            if (this.tableHeaderList?.Count != rowContents.Length)
            {
                throw new Exception($"Number of columns passed({rowContents.Length}) doesn't match the number of columns in the header({this.tableHeaderList?.Count})");
            }

            var newRow = new List<string>() { fileUrl };
            newRow.AddRange(rowContents);
            this.fileContentsList.Add(newRow);
        }

        public string GetHtml()
        {
            var htmlContent = new StringBuilder(@"<!DOCTYPE HTML PUBLIC "" -//W3C//DTD HTML 4.01 Transitional//EN"" ""http://www.w3.org/TR/html4/loose.dtd"">");
            htmlContent.AppendLine("<html>");

            htmlContent.AppendLine($"<head> <title>{this.titleText}</title> </head>");

            htmlContent.AppendLine($"<body>");
            htmlContent.AppendLine($"<h1>{this.pageHeadingText}</h1>");
            htmlContent.AppendLine($"</body>");

            // Parent Directory link
            if (this.previousPageTextAndUrl != null)
            {
                htmlContent.AppendLine($"<table>");
                htmlContent.AppendLine(
                    $@"<tr><th valign=""top""><a></a></th></tr>");
                htmlContent.AppendLine(
                    $@"<tr><td valign=""top""><a href=""{this.previousPageTextAndUrl.Item2}"">{
                            this.previousPageTextAndUrl.Item1
                        }</a></td></tr>");
                htmlContent.AppendLine($"</table>");
            }

            // Actual data table.
            htmlContent.AppendLine($"<table>");

            // Header
            htmlContent.AppendLine($"<tr>");
            foreach (var columnName in this.tableHeaderList)
            {
                htmlContent.Append($@"<th align=""center"">{columnName}</th> ");
            }

            htmlContent.AppendLine($"</tr>");

            // folders if any
            foreach (var directoryTuple in this.directoryNameUrlList)
            {
                htmlContent.AppendLine($@"<tr><td  align=""left""><a href=""{directoryTuple.Item2}"">{directoryTuple.Item1}</a> </td></tr>");
            }

            // File Contents
            foreach (var fileContents in this.fileContentsList)
            {
                htmlContent.AppendLine($@"<tr><td align=""left""><a href=""{fileContents[0]}"">{fileContents[1]}</a> </td>");
                for (var i = 2; i < fileContents.Count; i++)
                {
                    htmlContent.Append($@"<td align=""center"">{fileContents[i]}</td> ");
                }

                htmlContent.AppendLine($"</tr>");
            }

            htmlContent.AppendLine($"</table>");

            htmlContent.AppendLine("</html>");
            return htmlContent.ToString();
        }
    }
}
