﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.OSMMirrorOnAzure
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Text;
    using System.Threading;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using Microsoft.WindowsAzure.Storage.RetryPolicies;

    public class OSMDownloadUploadManager
    {
        private const int HoursInAWeek = 24 * 7;
        private const int BufferHours = 4;
        private const string DateTimeCustomFormatString = "MM_dd_yyyy_HH_mm";
        private const string KeyWordToReplaceWithTimeStamp = "latest";
        private const string HtmlFileStatusStart = "#";
        private const char ControllerFileColumnDelimitter = '\t';
        private const string ControllerBlobName = "OsmMirrorControllerBlob.txt";
        private const string ControllerBlobContainerName = "osmmirrorcontroller";
        private const string OsmMirrorContainerName = "osmplanet";
        private const string OsmBaseUrl = "https://ftp.osuosl.org/pub/openstreetmap/";
        private static readonly List<string> DownloadableFileList = new List<string>()
         {
            "pbf/planet-latest.osm.pbf",
            "pbf/planet-latest.osm.pbf.md5",
            "planet/changesets-latest.osm.bz2",
            "planet/changesets-latest.osm.bz2.md5",
            "planet/discussions-latest.osm.bz2",
            "planet/discussions-latest.osm.bz2.md5",
            "planet/planet-latest.osm.bz2",
            "planet/planet-latest.osm.bz2.md5"
         };

        private CloudStorageAccount controllerCloudStorageAccount;
        private ILogger log;

        private ControllerContent configuredDefaultControllerContent = new ControllerContent(DownloadableFileList);
        
        public OSMDownloadUploadManager(ILogger logger, CloudStorageAccount storageAccount)
        {
            // To increase the through put to the Blob storage. Otherwise, .Net uses only 2 threads to send.
            ServicePointManager.DefaultConnectionLimit = 1000;

            this.log = logger;
            this.controllerCloudStorageAccount = storageAccount;
        }

        public async Task Initialize()
        {
            var blobClient = this.controllerCloudStorageAccount.CreateCloudBlobClient();
            var controllerContainerReference = blobClient.GetContainerReference(ControllerBlobContainerName);
            await controllerContainerReference.CreateIfNotExistsAsync();

            var mirrorContainerReference = blobClient.GetContainerReference(OsmMirrorContainerName);
            await mirrorContainerReference.CreateIfNotExistsAsync(BlobContainerPublicAccessType.Blob, null, null);
        }

        public async Task<TimeSpan> RefreshOsmFilesAsNeeded(CancellationToken cancellationToken)
        {
            var controllerContent = this.GetControllerContent();
            Dictionary<string, CloudBlobDirectory> updateFileDirectories = new Dictionary<string, CloudBlobDirectory>();
            object lockObj = new object();
            var taskList = new List<Task>();

            foreach (var kvp in controllerContent.FileDetailsMap)
            {
                taskList.Add(this.RunAsync(kvp, lockObj, updateFileDirectories, cancellationToken));
            }

            await Task.WhenAll(taskList.ToArray());

            if (cancellationToken.IsCancellationRequested)
            {
                log.Log(LogType.Error, "OSM File Refresh received cancellation. Work is abandoned. Please retry.");
                return TimeSpan.FromHours(1);
            }

            // Get the Controller Content again and check the HTML status
            controllerContent = this.GetControllerContent();
            if (!controllerContent.IsHTMLUpdated)
            {
                if (this.ReBuildOSMHtml(updateFileDirectories.Values.ToList(), cancellationToken))
                {
                    controllerContent.SetIsHtmlUpdatedStatus(true);
                    this.WriteControllerBlob(controllerContent);
                }
            }

            var currentTime = DateTimeOffset.UtcNow;
            var timeElapsedBetweenOldestAndNow = currentTime - GetOldestTimeStamp(controllerContent);
            var hoursToSleep = HoursInAWeek - (int)Math.Ceiling(timeElapsedBetweenOldestAndNow.TotalHours) - BufferHours;
            return TimeSpan.FromHours(hoursToSleep > 0 ? hoursToSleep : 1);
        }

        internal static string GetDestinationBlobName(string fileBaseName, DateTimeOffset timeStamp)
        {
            if (!fileBaseName.Contains(KeyWordToReplaceWithTimeStamp))
            {
                throw new FormatException($"File base name is expected to have the key word : {KeyWordToReplaceWithTimeStamp}");
            }

            return fileBaseName.Replace("/", $"/archives/{timeStamp.Year}/").Replace(KeyWordToReplaceWithTimeStamp, timeStamp.ToString(DateTimeCustomFormatString));
        }

        private static string GetSourceUrlFromFileBaseName(string fileBaseName) => $"{OsmBaseUrl}/{fileBaseName}";

        private static async Task<bool> CheckIfFileHasToBeUpdated(string fileBaseName, DateTimeOffset lastModifiedDateTime, long lastKnownSizeInBytes)
        {
            var sourceUrl = GetSourceUrlFromFileBaseName(fileBaseName);
            DateTimeOffset sourceLastModifieDateTime;
            long sourceSize = -1;
            using (var httpFileReader = new HttpFileBlockReader(sourceUrl, null, false))
            {
                await httpFileReader.InitializeAsync();
                sourceSize = httpFileReader.LengthOfTheStream;
                sourceLastModifieDateTime = httpFileReader.LastModifieDateTime;
            }

            return sourceSize != lastKnownSizeInBytes || sourceLastModifieDateTime > lastModifiedDateTime;
        }

        private static DateTimeOffset GetOldestTimeStamp(ControllerContent controllerContent)
        {
            return controllerContent.FileDetailsMap.Select(t => t.Value.LastModifDateTimeOffset).Min();
        }

        private static void GetFileNameAndDirectoryNameFromBlobUri(Uri uri, out string directoryName, out string fileName)
        {
            var parts = uri.ToString().TrimEnd('/').Split('/');
            if (parts[parts.Length - 1].Contains("."))
            {
                fileName = parts[parts.Length - 1];
                directoryName = parts[parts.Length - 2];
            }
            else
            {
                fileName = string.Empty;
                directoryName = parts[parts.Length - 1];
            }
        }

        private static async Task<OsmNewBlobFileInfo> BringTheNewOsmFile(string fileBaseName, int? previousBlockSize, ILogger logger, CloudStorageAccount storageAccount, CancellationToken cancellationToken, Action<int> updateBlockSizeInController)
        {
            var sourceUrl = GetSourceUrlFromFileBaseName(fileBaseName);

            logger.Log(LogType.Info, $"Starting to bring {fileBaseName}. Using custom parallel upload.");
            logger.Log(LogType.Info, $"sourceURL : {sourceUrl}");
            logger.Log(LogType.Info, $"public blob Container: {OsmMirrorContainerName}");

            var reader = new HttpFileBlockReader(sourceUrl, null, false);
            await reader.InitializeAsync();

            var lastModifiedDateTime = reader.LastModifieDateTime;
            CloudBlobDirectory cloudBlobDirectory = null;
            var lastKnownSizeInBytes = reader.LengthOfTheStream;
            var datedDestinationBlobName = GetDestinationBlobName(fileBaseName, lastModifiedDateTime);

            logger.Log(LogType.Info, $"blob file name : {datedDestinationBlobName}");

            var writer = new AzureBlockBlobWriter(
                storageAccount,
                OsmMirrorContainerName,
                datedDestinationBlobName);
            await writer.InitializeAsync();

            var controller = new BlockStreamManager(reader, writer, logger, fileBaseName, previousBlockSize);
            await controller.InitializeAsync();

            if (controller.CurrentBlockLength != previousBlockSize)
            {
                updateBlockSizeInController(controller.CurrentBlockLength);
            }

            var success = await controller.CopyAsync(cancellationToken);
            logger.Log(LogType.Info, $"Done with {fileBaseName}");

            if (success)
            {
                logger.Log(LogType.Info, $"Deleting the old Latest File({fileBaseName})...");
                Helpers.GetBlockBlob(logger, storageAccount, OsmMirrorContainerName, fileBaseName, out var currentLatestBlobRef);
                currentLatestBlobRef.DeleteIfExists();
                logger.Log(LogType.Info, $"Copying the current Latest File({datedDestinationBlobName}) into Latest({fileBaseName})...");
                await Helpers.CopyBlockBlobAsync(logger, storageAccount, OsmMirrorContainerName, datedDestinationBlobName, fileBaseName);
                Helpers.GetBlockBlob(logger, storageAccount, OsmMirrorContainerName, datedDestinationBlobName, out var currentDatedBlobRef);
                cloudBlobDirectory = currentDatedBlobRef.Parent;
            }
            else
            {
                logger.Log(LogType.Info, $"Download and upload from URL ({sourceUrl}) to blob ({datedDestinationBlobName}) failed.");
            }

            return new OsmNewBlobFileInfo(success, datedDestinationBlobName, lastModifiedDateTime, lastKnownSizeInBytes, cloudBlobDirectory);
        }

        private static async Task<Tuple<long, DateTimeOffset>> GetInputFileDetails(string sourceUrl)
        {
            using (var httpClient = new HttpClient())
            {
                long lastKnownSizeInBytes;
                DateTimeOffset lastModifiedDateTime;

                var response = await httpClient.GetAsync(sourceUrl, HttpCompletionOption.ResponseHeadersRead);
                if (response.IsSuccessStatusCode)
                {
                    if (response.Content.Headers.ContentLength.HasValue)
                    {
                        lastKnownSizeInBytes = response.Content.Headers.ContentLength.Value;
                    }
                    else
                    {
                        throw new Exception("The server did not return the ContentLength.");
                    }

                    if (response.Content.Headers.LastModified.HasValue)
                    {
                        lastModifiedDateTime = response.Content.Headers.LastModified.Value;
                    }
                    else
                    {
                        throw new Exception("The server did not return the LastModified value.");
                    }
                }
                else
                {
                    throw new HttpRequestException($"Got status code({response.IsSuccessStatusCode}) while trying to gather headers from url({sourceUrl})");
                }

                return new Tuple<long, DateTimeOffset>(lastKnownSizeInBytes, lastModifiedDateTime);
            }
        }

        private static async Task MoveFile(string sourceUrl, CloudBlockBlob targetBlob, ILogger logger, CancellationToken cancellationToken)
        {
            using (var httpClient = new HttpClient())
            {
                var inputStream = await httpClient.GetStreamAsync(sourceUrl);
                try
                {
                    var requestOptions = new BlobRequestOptions()
                    {
                        UseTransactionalMD5 = true,
                        StoreBlobContentMD5 = true,
                        RetryPolicy = new ExponentialRetry(
                                             TimeSpan.FromMinutes(10),
                                             3),
                        MaximumExecutionTime = TimeSpan.FromHours(5)
                    };

                    var operationContext = new OperationContext();
                    operationContext.Retrying += (obj, args) => logger.Log(LogType.Warning, $"UploadFromStreamAsync is retrying to upload he file ({sourceUrl}) to blob ({targetBlob.Uri}).");

                    await targetBlob.UploadFromStreamAsync(inputStream, null, requestOptions, operationContext, cancellationToken);
                }
                catch (Exception ex)
                {
                    logger.Log(LogType.Error, $"Exception while copying the file ({sourceUrl}) to blob ({targetBlob.Uri}). Exception : {ex.ToString()}");
                    throw;
                }
            }
        }

        private static async Task<OsmNewBlobFileInfo> BringTheNewOsmFileBuiltIn(string fileBaseName, ILogger logger, CloudStorageAccount storageAccount, CancellationToken cancellationToken)
        {
            var sourceUrl = GetSourceUrlFromFileBaseName(fileBaseName);
            logger.Log(LogType.Info, $"Starting to bring {fileBaseName}. Using Azure UploadStreamAsync.");
            logger.Log(LogType.Info, $"sourceURL : {sourceUrl}");
            logger.Log(LogType.Info, $"public blob Container: {OsmMirrorContainerName}");

            var inputFileDetails = await GetInputFileDetails(sourceUrl);
            long lastKnownSizeInBytes = inputFileDetails.Item1;
            DateTimeOffset lastModifiedDateTime = inputFileDetails.Item2;
            CloudBlobDirectory cloudBlobDirectory = null;
            var datedDestinationBlobName = GetDestinationBlobName(fileBaseName, lastModifiedDateTime);
            logger.Log(LogType.Info, $"blob file name : {datedDestinationBlobName}");

            Helpers.GetBlockBlob(logger, storageAccount, OsmMirrorContainerName, datedDestinationBlobName, out var targetBlob);
            await MoveFile(sourceUrl, targetBlob, logger, cancellationToken);

            logger.Log(LogType.Info, $"Deleting the old Latest File({fileBaseName})...");
            Helpers.GetBlockBlob(logger, storageAccount, OsmMirrorContainerName, fileBaseName, out var currentLatestBlobRef);
            currentLatestBlobRef.DeleteIfExists();
            logger.Log(LogType.Info, $"Copying the current Latest File({datedDestinationBlobName}) into Latest({fileBaseName})...");
            await Helpers.CopyBlockBlobAsync(logger, storageAccount, OsmMirrorContainerName, datedDestinationBlobName, fileBaseName);
            Helpers.GetBlockBlob(logger, storageAccount, OsmMirrorContainerName, datedDestinationBlobName, out var currentDatedBlobRef);
            cloudBlobDirectory = currentDatedBlobRef.Parent;

            return new OsmNewBlobFileInfo(true, datedDestinationBlobName, lastModifiedDateTime, lastKnownSizeInBytes, cloudBlobDirectory);
        }

        private void UpdateTheControllerBlobForTheCurrentFile(string fileBaseName, DateTimeOffset lastModifiedDateTime, long lastKnownSizeInBytes, int currentBlockSize)
        {
            var controllerContent = this.GetControllerContent();
            controllerContent.FileDetailsMap[fileBaseName] = new ControllerFileInfo(lastModifiedDateTime, lastKnownSizeInBytes, currentBlockSize);
            controllerContent.SetIsHtmlUpdatedStatus(false);
            this.WriteControllerBlob(controllerContent);
        }

        private async Task RunAsync(KeyValuePair<string, ControllerFileInfo> keyValuePair, object lockObj, Dictionary<string, CloudBlobDirectory> updateFileDirectories, CancellationToken cancellationToken)
        {
            var updatedNeeded = await CheckIfFileHasToBeUpdated(keyValuePair.Key, keyValuePair.Value.LastModifDateTimeOffset, keyValuePair.Value.LengthInBytes);
            if (updatedNeeded)
            {
                this.log.Log(LogType.Info, $"File({keyValuePair.Key}) is going to be downloaded.");
                var newBlobFileInfo = await BringTheNewOsmFile(keyValuePair.Key, keyValuePair.Value.PreviousBlockSize == 0 ? (int?)null : keyValuePair.Value.PreviousBlockSize, this.log, this.controllerCloudStorageAccount, cancellationToken,
                                                               newBlockSize =>
                                                                   {
                                                                       lock (lockObj)
                                                                       {
                                                                           // The moment we know the new block size for this file, put it in the controller. It'll be used in the fail-restart scenario.
                                                                           this.UpdateTheControllerBlobForTheCurrentFile(keyValuePair.Key, keyValuePair.Value.LastModifDateTimeOffset, keyValuePair.Value.LengthInBytes, newBlockSize);
                                                                       }
                                                                   });
                if (newBlobFileInfo.MoveSuccessful)
                {
                    lock (lockObj)
                    {
                        updateFileDirectories[newBlobFileInfo.NewBlobDirectory.Prefix] = newBlobFileInfo.NewBlobDirectory;
                        
                        // Set the last block size to 0 when the download is complete.
                        this.UpdateTheControllerBlobForTheCurrentFile(keyValuePair.Key, newBlobFileInfo.ModifiedDateTime, newBlobFileInfo.FileSizeInBytes, 0);
                    }
                }
            }
            else
            {
                this.log.Log(LogType.Info, $"File({keyValuePair.Key}) hasn't changed since the last download.");
            }
        }

        private bool ReBuildOSMHtml(List<CloudBlobDirectory> leafDirectoriesToBeUpdated, CancellationToken cancellationToken)
        {
            foreach (var directory in leafDirectoriesToBeUpdated)
            {
                var parentDirectory = directory;
                var currentLevel = 0;

                // Always update the current and the parent level. This leaf directory could have been created just now.
                // For all other levels, the contents are the same. So, unless the file is missing, no need to create them.
                while (parentDirectory != null)
                {
                    if (cancellationToken.IsCancellationRequested)
                    {
                        return false;
                    }

                    CreateHtmlByEnumeratingFilesInADir(parentDirectory, currentLevel < 2);
                    parentDirectory = parentDirectory.Parent;
                    currentLevel++;
                }
            }

            return true;
        }

        private void CreateHtmlByEnumeratingFilesInADir(CloudBlobDirectory blobDirectory, bool overWrite)
        {
            var currentDirRelativePath = blobDirectory.Prefix;
            Helpers.GetBlockBlob(this.log, this.controllerCloudStorageAccount, OsmMirrorContainerName, currentDirRelativePath + @"Index.html", out var htmlBlob);

            if (htmlBlob.Exists() && !overWrite)
            {
                return;
            }

            var htmlBuilder = new HtmlBuilder();
            var title = $"Index of {currentDirRelativePath}";
            htmlBuilder.SetTitle(title);
            htmlBuilder.SetHeading(title);
            htmlBuilder.SetTableHeaders("File Name", "Last modified", "Size In Bytes");
            if (blobDirectory.Parent != null)
            {
                var prevPage = blobDirectory.Parent.Prefix;
                if (string.IsNullOrEmpty(prevPage))
                {
                    prevPage = "Home";
                }

                var prevPageHtmlUrl = $"{blobDirectory.Parent.Uri.ToString().TrimEnd('/')}/Index.html";
                htmlBuilder.SetPreviousPageLink(prevPage, prevPageHtmlUrl);
            }

            foreach (var listBlobItem in blobDirectory.ListBlobs())
            {
                var uri = listBlobItem.Uri;
                if (uri.ToString().EndsWith("Index.html"))
                {
                    continue;
                }

                var directory = listBlobItem as CloudBlobDirectory;
                GetFileNameAndDirectoryNameFromBlobUri(uri, out var directoryName, out var fileName);

                if (directory == null)
                {
                    var fileBlob = listBlobItem as CloudBlockBlob;
                    htmlBuilder.AddFileContentRow(uri.ToString(), fileName, fileBlob.Properties.LastModified.ToString(), fileBlob.Properties.Length.ToString());
                }
                else
                {
                    htmlBuilder.AddDirectoryToTable(directoryName, $"{uri.ToString()}Index.html");
                }
            }
            
            var bytes = Encoding.ASCII.GetBytes(htmlBuilder.GetHtml());
            htmlBlob.UploadFromByteArray(bytes, 0, bytes.Length);
        }

        private void WriteControllerBlob(ControllerContent controllerContent)
        {
            Helpers.GetBlockBlob(this.log, this.controllerCloudStorageAccount, ControllerBlobContainerName, ControllerBlobName, out var cloudBlockBlob);

            using (StreamWriter writer = new StreamWriter(cloudBlockBlob.OpenWrite()))
            {
                foreach (var kvp in controllerContent.FileDetailsMap)
                {
                    writer.WriteLine($"{kvp.Key}\t{kvp.Value.LastModifDateTimeOffset}\t{kvp.Value.LengthInBytes}\t{kvp.Value.PreviousBlockSize}");
                }

                writer.WriteLine($"#HTMLFileUpdated\t{controllerContent.IsHTMLUpdated}");
            }
        }

        private ControllerContent GetControllerContent()
        {
            ControllerContent controllerContent = null;
            Helpers.GetBlockBlob(this.log, this.controllerCloudStorageAccount, ControllerBlobContainerName, ControllerBlobName, out var cloudBlockBlob);
            
            if (cloudBlockBlob.Exists())
            {
                controllerContent = new ControllerContent();
                using (var streamReader = new StreamReader(cloudBlockBlob.OpenRead()))
                {
                    while (!streamReader.EndOfStream)
                    {
                        var newLine = streamReader.ReadLine();
                        if (newLine.StartsWith(HtmlFileStatusStart))
                        {
                            var columns = newLine.Split(ControllerFileColumnDelimitter);
                            if (columns.Length != 2)
                            {
                                throw new FormatException($"HTML file status line is expected to have two columns seperated by ({ControllerFileColumnDelimitter}). Line encountered : {newLine}");
                            }

                            controllerContent.SetIsHtmlUpdatedStatus(Convert.ToBoolean(columns[1]));
                        }
                        else
                        {
                            var columns = newLine.Split(ControllerFileColumnDelimitter);
                            if (columns.Length != 4)
                            {
                                throw new FormatException($"File info line is expected to have three columns seperated by ({ControllerFileColumnDelimitter}). Line encountered : {newLine}");
                            }

                            if (!DateTimeOffset.TryParse(columns[1], out var dateTime))
                            {
                                throw new FormatException($"Couldn't convert the second column in the file info line to DateTime. Value encountered : {columns[1]} ");
                            }

                            if (!long.TryParse(columns[2], out var size))
                            {
                                throw new FormatException($"Couldn't convert the third column in the file info line to long. Value encountered : {columns[2]} ");
                            }

                            if (!Int32.TryParse(columns[3], out var blockSize))
                            {
                                throw new FormatException($"Couldn't convert the fourth column in the file info line to int. Value encountered : {columns[3]} ");
                            }

                            controllerContent.AddFileDetails(columns[0], dateTime, size, blockSize);
                        }
                    }
                }
            }

            if (controllerContent == null)
            {
                controllerContent = new ControllerContent(this.configuredDefaultControllerContent);
            }
            else
            {
                controllerContent.AdjustWithMasterCopy(this.configuredDefaultControllerContent);
            }

            return controllerContent;
        }
    }
}
