﻿// --------------------------------------------------------------------------------------------------------------------
// <copyright file="OSMMirrorOnAzure.cs" company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// <summary>
//   An instance of this class is created for each service instance by the Service Fabric runtime.
// </summary>
// --------------------------------------------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.OSMMirrorOnAzure
{
    using System;
    using System.Fabric;
    using System.IO;
    using System.Threading;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ServiceFabricHosting;
    using Microsoft.ServiceFabric.Services.Runtime;
    using Microsoft.WindowsAzure.Storage;

    /// <summary>
    /// An instance of this class is created for each service instance by the Service Fabric runtime.
    /// </summary>
    internal sealed class OSMMirrorOnAzure : StatelessService
    {
        private readonly ConfigurationReader serviceParameters;

        public OSMMirrorOnAzure(StatelessServiceContext context)
            : base(context)
        {
            this.serviceParameters = new ConfigurationReader(context);
        }

        /// <summary>
        /// This is the main entry point for your service instance.
        /// </summary>
        /// <param name="cancellationToken">Canceled when Service Fabric needs to shut down this service instance.</param>
        protected override async Task RunAsync(CancellationToken cancellationToken)
        {
            ILogger logger = new OsmAgentLogger();
            try
            {
                var keyVaultUrl = this.serviceParameters.GetConfigValue<string>("mlbsosmmirror");
                var sasKeyUrl = this.serviceParameters.GetConfigValue<string>("mlbsosmmirror-crudall");
                
                while (!cancellationToken.IsCancellationRequested)
                {
                    IConfigurationReader configurationReader = new KeyVaultConfigurationReader();

                    var connectionStringTemplate = await configurationReader.GetConfigurationValueAsync(keyVaultUrl);
                    var sasKey = await configurationReader.GetConfigurationValueAsync(sasKeyUrl);
                    var account = CloudStorageAccount.Parse(string.Format(connectionStringTemplate, sasKey));

                    var osmDownloadUploadManager = new OSMDownloadUploadManager(logger, account);
                    await osmDownloadUploadManager.Initialize();

                    var suggestedSleepTime = osmDownloadUploadManager.RefreshOsmFilesAsNeeded(cancellationToken).GetAwaiter().GetResult();

                    await Task.Delay(suggestedSleepTime, cancellationToken);
                }
            }
            catch (Exception ex)
            {
                logger.Log(LogType.Error, $"RunAsync got an exception. Message : {ex.Message} , ToString : {ex.ToString()}");
                Agent.Instance.LogException(ex, "osmMirror", "update");
                throw;
            }
        }
    }
}
