﻿// --------------------------------------------------------------------------------------------------------------------
// <copyright file="ResourceProviderDataBackupJob.cs" company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// --------------------------------------------------------------------------------------------------------------------
namespace Microsoft.Azure.LocationServices.ResourceProviderDataBackupJob
{
    using System;
    using System.Collections.Generic;
    using System.Fabric;
    using System.IO;
    using System.IO.Compression;
    using System.Linq;
    using System.Reflection;
    using System.Text;
    using System.Threading;
    using System.Threading.Tasks;

    using CsvHelper;

    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Utility;
    using Microsoft.ServiceFabric.Services.Communication.Runtime;
    using Microsoft.ServiceFabric.Services.Runtime;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Auth;
    using Microsoft.WindowsAzure.Storage.Blob;
    using Microsoft.WindowsAzure.Storage.Table;

    using Newtonsoft.Json;

    /// <inheritdoc />
    /// <summary>
    /// An instance of this class is created for each service instance by the Service Fabric runtime.
    /// </summary>
    internal sealed class ResourceProviderDataBackupJob : StatelessService
    {
        private readonly TimeSpan delay;

        public ResourceProviderDataBackupJob(StatelessServiceContext context)
            : base(context)
        {
            this.delay = TimeSpan.FromDays(1);
        }

        /// <inheritdoc />
        /// <summary>
        /// Optional override to create listeners (e.g., TCP, HTTP) for this service replica to handle client or user requests.
        /// </summary>
        /// <returns>A collection of listeners.</returns>
        protected override IEnumerable<ServiceInstanceListener> CreateServiceInstanceListeners()
        {
            return new ServiceInstanceListener[0];
        }

        /// <inheritdoc />
        /// <summary>
        /// This is the main entry point for your service instance.
        /// </summary>
        /// <param name="cancellationToken">Canceled when Service Fabric needs to shut down this service instance.</param>
        protected override async Task RunAsync(CancellationToken cancellationToken)
        {
            ConfigurationReader.Instance.Initialize(this.Context);
            IBootstrapConfiguration bootstrapConfiguration = new ServiceFabricConfigurationBootstrap(ConfigurationReader.Instance);

            while (!cancellationToken.IsCancellationRequested)
            {
                cancellationToken.ThrowIfCancellationRequested();

                StorageClients clients = await StorageClients.FromConfiguration(bootstrapConfiguration);

                await BackupTablesAsync(clients.SourceTableClient, clients.BackupBlobClient, cancellationToken);

                await Task.Delay(this.delay, cancellationToken);
            }

            cancellationToken.ThrowIfCancellationRequested();
        }

        private static async Task BackupTablesAsync(CloudTableClient tableClient, CloudBlobClient blobClient, CancellationToken cancellationToken)
        {
            using (var archiveStream = new MemoryStream())
            {
                using (var zip = new ZipArchive(archiveStream, ZipArchiveMode.Create, true))
                {
                    CloudTable accountsTable = tableClient.GetTableReference("LocationServicesAccounts");
                    WriteTableAsCsvToZip<AccountRecord>(accountsTable, zip);

                    CloudTable subscriptionsTable = tableClient.GetTableReference("LocationServicesSubscriptions");
                    WriteTableAsCsvToZip<SubscriptionRecord>(subscriptionsTable, zip);
                }

                await UploadSnapshotAsync(blobClient, cancellationToken, archiveStream);
            }
        }

        private static void WriteTableAsCsvToZip<T>(CloudTable table, ZipArchive zip)
            where T : TableEntity, new()
        {
            IEnumerable<T> rows = IterablyListAllRows<T>(table);
            ZipArchiveEntry entry = zip.CreateEntry($"{table.Name}.csv");
            using (Stream stream = entry.Open())
                using (var writer = new StreamWriter(stream))
                    using (var csv = new CsvWriter(writer))
                    {
                        csv.WriteRecords(rows);
                    }
        }

        private static async Task UploadSnapshotAsync(CloudBlobClient blobClient, CancellationToken cancellationToken, MemoryStream archiveStream)
        {
            CloudBlobContainer backupsContainer = blobClient.GetContainerReference("snapshots");
            await backupsContainer.CreateIfNotExistsAsync(cancellationToken);
            DateTime now = DateTime.UtcNow.Date;
            CloudBlockBlob blob = backupsContainer.GetBlockBlobReference($"{now.Year}.{now.Month}.{now.Day}.zip");
            archiveStream.Seek(0, SeekOrigin.Begin);
            await blob.UploadFromStreamAsync(archiveStream, cancellationToken);
        }

        private static IEnumerable<T> IterablyListAllRows<T>(CloudTable table)
            where T : TableEntity, new()
        {
            TableContinuationToken token = null;
            do
            {
                TableQuerySegment<T> queryResult = table.ExecuteQuerySegmented(new TableQuery<T>(), token);
                foreach (T row in queryResult.Results)
                {
                    yield return row;
                }

                token = queryResult.ContinuationToken;
            }
            while (token != null);
        }

        private class StorageClients
        {
            private readonly CloudStorageAccount sourceStorageAccount;

            private readonly CloudStorageAccount backupStorageAccount;

            private StorageClients(CloudStorageAccount sourceStorageAccount, CloudStorageAccount backupStorageAccount)
            {
                this.sourceStorageAccount = sourceStorageAccount;
                this.backupStorageAccount = backupStorageAccount;
            }

            public CloudTableClient SourceTableClient => this.sourceStorageAccount.CreateCloudTableClient();

            public CloudBlobClient BackupBlobClient => this.backupStorageAccount.CreateCloudBlobClient();

            public static async Task<StorageClients> FromConfiguration(IBootstrapConfiguration bootstrapConfiguration)
            {
                IConfigurationReader configurationReader = new KeyVaultConfigurationReader();

                string sourceConnectionStringFormat = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("SourceStorageAccountConnectionString"));
                string sourceSasToken = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("SourceStorageAccountSasTokenSecretUrl"));
                CloudStorageAccount sourceStorageAccount = CloudStorageAccount.Parse(string.Format(sourceConnectionStringFormat, sourceSasToken));

                string backupConnectionStringFormat = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("BackupStorageAccountConnectionString"));
                string backupSasToken = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue("BackupStorageAccountSasTokenSecretUrl"));
                CloudStorageAccount backupStorageAccount = CloudStorageAccount.Parse(string.Format(backupConnectionStringFormat, backupSasToken));

                return new StorageClients(sourceStorageAccount, backupStorageAccount);
            }
        }
    }
}
