﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZonesDataUpdate
{
    using System;
    using System.Collections.Generic;
    using System.Fabric;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ServiceFabricHosting;
    using Microsoft.Azure.LocationServices.TimeZoneDb;
    using Microsoft.Azure.LocationServices.TimeZoneDbManager;
    using Microsoft.ServiceFabric.Services.Communication.Runtime;
    using Microsoft.ServiceFabric.Services.Runtime;

    /// <summary>
    /// An instance of this class is created for each service instance by the Service Fabric runtime.
    /// </summary>
    internal sealed class TimeZonesDataUpdate : StatelessService
    {
        public TimeZonesDataUpdate(StatelessServiceContext context)
            : base(context)
        {
        }

        /// <summary>
        /// Optional override to create listeners (e.g., TCP, HTTP) for this service replica to handle client or user requests.
        /// </summary>
        /// <returns>A collection of listeners.</returns>
        protected override IEnumerable<ServiceInstanceListener> CreateServiceInstanceListeners()
        {
            return new ServiceInstanceListener[0];
        }

        /// <summary>
        /// This is the main entry point for your service instance.
        /// </summary>
        /// <param name="cancellationToken">Canceled when Service Fabric needs to shut down this service instance.</param>
        protected override async Task RunAsync(CancellationToken cancellationToken)
        {
            try
            {
                while (!cancellationToken.IsCancellationRequested)
                {
                    Agent.Instance.LogEvent(EventType.Information, "TimeZonesDataUpdate: Starting job");

                    var serviceFabricConfigReader = new ConfigurationReader(Context);
                    var config = new TimeZonesConfigBootStrap(serviceFabricConfigReader);

                    var cfg = ConfigValues.CreateForPublisherMode(config);
                    CronJobResult results = await TimeZoneDb.CronDataUpdateCheckAsync(cfg, LogEventCallback);

                    Agent.Instance.LogEvent(EventType.Information, $"TimeZonesDataUpdate: Finished job with {results}");

                    await Task.Delay(TimeSpan.FromHours(12), cancellationToken);
                }
            }
            catch (TaskCanceledException ex)
            {
                if (cancellationToken.IsCancellationRequested)
                {
                    Agent.Instance.LogEvent(EventType.Information, "TimeZonesDataUpdate: Service stopped because it received a cancellation request");
                }
                else
                {
                    Agent.Instance.LogException(ex, "timezone", "updateJobFailed");
                    throw;
                }
            }
            catch (Exception e)
            {
                Agent.Instance.LogException(e, "timezone", "updateJobFailed");
                throw;
            }
        }

        private static void LogEventCallback(object sender, LogEventArgs args)
        {
            switch (args.LogTypeId)
            {
                case LogType.Verbose:
                    Agent.Instance.LogEvent(EventType.Verbose, args.Message);
                    break;
                case LogType.Info:
                    Agent.Instance.LogEvent(EventType.Information, args.Message);
                    break;
                case LogType.Warning:
                    Agent.Instance.LogEvent(EventType.Warning, args.Message);
                    break;
                case LogType.Error:
                    Agent.Instance.LogEvent(EventType.ProgramError, args.Message);
                    break;
                default:
                    Agent.Instance.LogEvent(EventType.ProgramError, $"TimeZonesDataUpdate: unhandled type={args.LogTypeId} in switch statement. Source={args.Source}, Msg={args.Message}");
                    break;
            }
        }
    }
}
