﻿import ClientResources = require("ClientResources");
import Constants = require("../../../Shared/Constants");

import ExtensionDefinition = require("../../../_generated/ExtensionDefinition");

import ResourceArea = require("../../MyResourceArea");

import FxAzure = MsPortalFx.Azure;
import Arm = FxAzure.ResourceManager;
import Forms = MsPortalFx.ViewModels.Forms;
import * as SubscriptionsDropDown from "Fx/Controls/SubscriptionDropDown";
import * as LocationDropDown from "Fx/Controls/LocationDropDown";
import * as ResourceGroupDropDown from "Fx/Controls/ResourceGroupDropDown";
import * as DropDown from "Fx/Controls/DropDown";

var resourceType = Constants.resourceProvider + "/" + Constants.resourceType;


/**
 * The data model for creating resource. This data type will be used as the parameter provider's
 * TResult (the provider inputs and outputs) and TEditScope (the edit scope).
 */
export interface CreateModel {

    /**
     * The name.
     */
    name: KnockoutObservable<string>;

    /**
     * The subscription
     */
    subscription: KnockoutObservable<MsPortalFx.Azure.Subscription>;

    /**
     * The resourcegroup.
     */
    resourceGroup: KnockoutObservable<MsPortalFx.Azure.CreatorAndDropdDownValue<MsPortalFx.Azure.ResourceGroup>>;

    /**
     * The location.
     */
    location: KnockoutObservable<MsPortalFx.Azure.Location>;

    pricingSpec: KnockoutObservable<string>;
}

/**
 * Create blade view model
 */
export class CreateBladeViewModel
    extends Forms.Form.ViewModel<CreateModel>{

    public title: KnockoutObservable<string>;

    public subtitle: KnockoutObservable<string>;

    public icon: KnockoutObservable<MsPortalFx.Base.Image>;

    /**
     * View model for the create action bar.
     */
    public actionBar: MsPortalFx.ViewModels.ActionBars.CreateActionBar.ViewModel;

    /**
     * View model for the parameter provider. It must have the name 'parameterProvider' (no other
     * name will work), and it's referenced in the PDL file by setting the ParameterProvider property
     * on this part view model to 'true'.
     */
    public parameterProvider: MsPortalFx.ViewModels.ParameterProvider<CreateModel, CreateModel>;

    /**
     * View model for the ARM provisioner.
     */
    public armProvisioner: Arm.Provisioner<CreateModel>;

    /**
     * The section that hosts the form controls.
    */
    public generalSection: Forms.Section.ViewModel;

    public nameTextBox: MsPortalFx.ViewModels.Forms.TextBox.ViewModel;

    /**
     * The view model for the subscriptions drop down.
     */
    public subscriptionsDropDown: SubscriptionsDropDown.Contract;

    private _subscriptionId = ko.pureComputed(() => {
        const sub = this.subscriptionsDropDown.value();
        return sub && sub.subscriptionId;
    });

    /**
     * The view model for the resource group drop down.
     */
    public resourceGroupDropDown: ResourceGroupDropDown.Contract;

    /**
     * The view model for the locations drop down.
     */
    public locationsDropDown: LocationDropDown.Contract;

    private _legalCheckbox: Forms.CheckBox.ViewModel;

    private _galleryItem: KnockoutObservableBase<Arm.GalleryItemExtract>;

    private _initialSubscriptionid = ko.observable<string[]>();

    private _initialResourceGroupName = ko.observable<string[]>();

    private _validResourceNameRegex = new RegExp("^[a-zA-Z0-9][a-zA-Z0-9_.-]*$");

    private _pricingSpecDropDown : DropDown.Contract<string>;

    constructor(container: MsPortalFx.ViewModels.ContainerContract, initialState: any, dataContext: ResourceArea.DataContext) {
        super(container);

        //set blade titles and icons
        this.title = ko.observable(ClientResources.CreateBlade.title);
        this.icon = ko.observable(MsPortalFx.Base.Images.Polychromatic.CloudService());

        this._galleryItem = ko.observable<Arm.GalleryItemExtract>();

        // Instantiate an action bar view model. The property must be called 'actionBar' so the PDL
        // can wire the action bar control to this view model. We're instantiating a create action
        // bar because this is a create scenario.
        this.actionBar = new MsPortalFx.ViewModels.ActionBars.CreateActionBar.ViewModel(container);
        this.actionBar.disablePrimaryButton(true);
        this.actionBar.valid(false);

        // We need to hook up the action bar's valid to the form's valid (which enables/disables the
        // 'Create' button based on whether the form is valid/invalid).

        // Instantiate a parameter provider view model.
        this.parameterProvider = new MsPortalFx.ViewModels.ParameterProvider<CreateModel, CreateModel>(container, {

            // This is where we receive the initial data defined in the UI definition file uploaded
            // with the gallery package. We'll use the data to seed the edit scope. This is the time
            // do do any necessary data transformations or account for incomplete inputs.
            mapIncomingDataForEditScope: this._mapIncomingDataForEditScope.bind(this),

            // This is where we transform the edit scope data to outputs. We returning the data as
            // it is, so no work needed here.
            mapOutgoingDataForCollector: (outgoing) => {
                return outgoing;
            }
        });

        // The parameter provider takes care of instantiating and initializing an edit scope for us,
        // so all we need to do is point our form's edit scope to the parameter provider's edit scope.
        this.editScope = this.parameterProvider.editScope;

        // Instantiate a ARM provisioner view model.
        this.armProvisioner = new Arm.Provisioner<CreateModel>(container, initialState, {
            // This is where we supply the ARM provisioner with the template deployment options
            // required by the deployment operation.
            supplyTemplateDeploymentOptions: this._supplyProvisioningPromise.bind(this),

            // Supplying an action bar and a parameter provider allows for automatic provisioning.
            actionBar: this.actionBar,
            parameterProvider: this.parameterProvider,

            // TODO: Enable this before you ship to PROD. Enable this once you have a RP for your service
            createFeatures: Arm.Provisioner.CreateFeatures.EnableArmValidation,
        });

        // Initialize the form fields.
        this._initializeFormFields(container, initialState);

        // Account create form is complete when,
        this.valid = ko.pureComputed(() => {
            return this.nameTextBox.valid()
                && this.subscriptionsDropDown.valid()
                && this.resourceGroupDropDown.valid()
                && this.locationsDropDown.valid()
                && this.resourceGroupDropDown.valid()
                && this._pricingSpecDropDown.valid()
                && this._legalCheckbox.valid();
        });

        this.valid.subscribe(container, this.actionBar.valid);
    }

    private _initializeFormFields(container: any, initialState?: any): void {
        const validResourceName = ko.pureComputed<boolean>(() => {
            var resourceGroup = this.resourceGroupDropDown.value() && this.resourceGroupDropDown.value().value.name;
            var name = this.nameTextBox.value();

            if (resourceGroup && name) {
                return resourceGroup.length + name.length < 98;
            }

            return true;
        });

        const resourceId = ko.pureComputed<string>(() => {
            var subscription = this._subscriptionId();
            var resourceGroup = this.resourceGroupDropDown.value() && this.resourceGroupDropDown.value().value.name;
            var name = this.nameTextBox.value();

            if (subscription && resourceGroup && name) {
                return `/subscriptions/${subscription}/resourceGroups/${resourceGroup}/providers/${resourceType}/${name}`;
            }
        });

        // The resource name text box
        this.nameTextBox = new MsPortalFx.ViewModels.Forms.TextBox.ViewModel(
            container,
            this,
            "name",
            {
                label: ko.observable(ClientResources.resourceName),
                emptyValueText: ko.observable(ClientResources.enterTheName),
                validations: ko.observableArray([
                    // Validate empty input
                    new MsPortalFx.ViewModels.RequiredValidation(ClientResources.CreateBlade.Error.resourceNameRequired),

                    // Validate valid characters
                    new MsPortalFx.ViewModels.CustomValidation(
                        ClientResources.CreateBlade.Error.resourceNameInvalidCharacters, value => 
                        Q(<MsPortalFx.ViewModels.ValidationResult>{
                                valid: this._validResourceNameRegex.test(value)
                        })
                    ),

                    // Validate length
                    new MsPortalFx.ViewModels.CustomValidation(
                        ClientResources.CreateBlade.Error.resourceNameAndResourceGroupNameTooLong, () =>
                        Q(<MsPortalFx.ViewModels.ValidationResult>{
                                valid: validResourceName()
                        })
                    ),

                    // Validate account does not exist
                    new MsPortalFx.ViewModels.CustomValidation(
                        ClientResources.CreateBlade.Error.resourceNameUnavailable, () => {
                            if (resourceId()) {
                                return Q(MsPortalFx.Base.Net2.ajax({
                                    uri: Constants.ArmServiceEndpoint + resourceId() + "?" + Constants.ArmServiceVersion,
                                    type: "GET",
                                    dataType: "json",
                                    cache: false,
                                    traditional: true,
                                    contentType: "application/json",
                                    setAuthorizationHeader: true,
                                    invokeApi: "api/invoke"
                                }).then(
                                    (value) => <MsPortalFx.ViewModels.ValidationResult>{
                                        valid: false,
                                        message: ClientResources.CreateBlade.Error.resourceNameUnavailable
                                    },
                                    (reason) => <MsPortalFx.ViewModels.ValidationResult>{
                                        valid: true
                                    }));
                            }

                            return Q(<MsPortalFx.ViewModels.ValidationResult>{
                                valid: true
                            });
                        }
                    )
                ]),
                delayValidationTimeout: ko.observable(500),
                valueUpdateTrigger: MsPortalFx.ViewModels.Controls.ValueUpdateTrigger.Input
            });

        // The subscriptions dropdown
        const subscriptionsDropDownOptions: SubscriptionsDropDown.Options = {
            initialSubscriptionId: this._initialSubscriptionid,
            validations: ko.observableArray([
                new MsPortalFx.ViewModels.RequiredValidation(ClientResources.selectSubscription)
            ]),
            resourceProviders: [Constants.resourceProvider],
            filterByGalleryItem: this._galleryItem
        };
        this.subscriptionsDropDown = SubscriptionsDropDown.create(container, subscriptionsDropDownOptions);

        // The resource group dropdown with creator inputs
        const resourceGroupsDropDownOptions: ResourceGroupDropDown.Options = {
            subscriptionId: this._subscriptionId,
            initialResourceGroupName: this._initialResourceGroupName,
            validations: ko.observableArray([
                new MsPortalFx.ViewModels.RequiredValidation(ClientResources.selectResourceGroup)
            ]),
            value: {
                mode: ResourceGroupDropDown.Mode.UseExisting,
                value: <FxAzure.ResourceGroup>{}
            }
        };
        this.resourceGroupDropDown = ResourceGroupDropDown.create(container, resourceGroupsDropDownOptions);

        // The locations dropdown
        const locationsDropDownOptions: LocationDropDown.Options = {
            subscriptionId: this._subscriptionId,
            initialLocationName: "westus",
            validations: ko.observableArray([
                new MsPortalFx.ViewModels.RequiredValidation(ClientResources.selectLocation)
            ]),
            resourceTypes: [],
            label: ClientResources.resourceGroupLocation

        };
        this.locationsDropDown = LocationDropDown.create(container, locationsDropDownOptions);

        // Subscribe to resource group changes to update the location drop down with the resource group location.
        this.resourceGroupDropDown.value.subscribe(container, (resourceGroup) => {
            if (this.locationsDropDown) { // Make sure it's initialized.
                let resourceGroupLocation = resourceGroup && resourceGroup.value && resourceGroup.value.location;
                // Try to find the resource group location in the locations list.
                // This is a hierarchical option (i.e. not the full location object).
                let location = this.locationsDropDown.fetchedValues().first((item) => {
                    return item.name === resourceGroupLocation;
                });

                // Since the location dropdown is filtering locations based on subscription and resource type,
                // the list will only contain allowed locations. If the resource group location doesn't exist
                // in the list, then it is not an allowed location, hence shouldn’t be selected. So set the
                // new location only if it's allowed.
                if (location) {
                    this.locationsDropDown.value(location);
                }
            }
        });

        // The pricing dropdown
        this._pricingSpecDropDown = this._createSpecDropdown(container);
        this._pricingSpecDropDown.value(Constants.serviceTiers.standard.id);

        // Terms of Service checkbox
        this._legalCheckbox = this._createLegalCheckbox(container);

        ko.reactor(
            container,
            [this.subscriptionsDropDown.value, this.resourceGroupDropDown.value],
            () => {
                this.nameTextBox.triggerValidation();
            });

        // 'Create an account' form
        const children: KnockoutObservableArray<any> = ko.observableArray([
            this.nameTextBox,
            // Also add the selector fields for the pickers.
            this.subscriptionsDropDown,
            this.resourceGroupDropDown,
            this._pricingSpecDropDown,
            this._legalCheckbox
        ]);

        // Locations dropdown menu logic
        ko.reactor(container, this.resourceGroupDropDown.value, (resourceGroupValue) => {
            if (resourceGroupValue == null) return;

            const locationsDropDownIdx = children.indexOf(this.locationsDropDown);

            // If locations dropdown menu already exists
            if (locationsDropDownIdx !== -1) {
                if (resourceGroupValue.mode === ResourceGroupDropDown.Mode.UseExisting) {
                    children.remove(this.locationsDropDown);
                }
                return;
            }

            // If creating a new resource group, add a locations dropdown menu
            if (resourceGroupValue.mode === ResourceGroupDropDown.Mode.CreateNew) {
                children.splice(locationsDropDownIdx, 0, this.locationsDropDown);
            }
        });

        // The form section.
        var sectionOptions = <Forms.Section.Options>{
            children: children
        };

        this.generalSection = new Forms.Section.ViewModel(container, sectionOptions);
        this.sections.push(this.generalSection);
    }

    private _mapIncomingDataForEditScope(incoming: CreateModel): CreateModel {
        const galleryCreateOptions = this.armProvisioner.armProvisioningConfig.galleryCreateOptions;

        if (galleryCreateOptions) {
            this._galleryItem(galleryCreateOptions.galleryItem);

            const initialSubscriptionId = galleryCreateOptions.subscriptionId;
            if (initialSubscriptionId) {
                this._initialSubscriptionid([initialSubscriptionId]);
            } else {
                this._initialSubscriptionid([]);
            }

            const initialResourceGroupName = galleryCreateOptions.resourceGroupName;
            if (initialResourceGroupName) {
                this._initialResourceGroupName([initialResourceGroupName]);
            } else {
                this._initialResourceGroupName([]);
            }
        }

        const model: CreateModel = {
            name: ko.observable<string>(),
            subscription: ko.observable<FxAzure.Subscription>(),
            resourceGroup: ko.observable<FxAzure.CreatorAndDropdDownValue<FxAzure.ResourceGroup>>({
                value: {
                    name: ClientResources.selectResourceGroup,
                    location: null
                },
                createNew: false
            }),
            pricingSpec: ko.observable<string>(),
            location: ko.observable<FxAzure.Location>()
        };

        return model;
    }

    private _supplyProvisioningPromise(data: CreateModel): MsPortalFx.Base.PromiseV<Arm.TemplateDeploymentOptions> {
        // This is where we supply the ARM provisioner with the template deployment options required
        // by the deployment operation.
        var galleryCreateOptions = this.armProvisioner.armProvisioningConfig && this.armProvisioner.armProvisioningConfig.galleryCreateOptions;

        const name = data.name();
        const subscriptionId = this._subscriptionId();

        const location = this.locationsDropDown.value();
        const resourceGroupName = this.resourceGroupDropDown.value().value.name;
        const resourceGroupLocation = this.resourceGroupDropDown.value().mode === ResourceGroupDropDown.Mode.CreateNew ? location.name : this.resourceGroupDropDown.value().value.location;

        const resourceIdFormattedString = `/subscriptions/${subscriptionId}/resourcegroups/${resourceGroupName}/providers/${resourceType}/${name}`;

        const pricingSpec = this._pricingSpecDropDown.value();

        const deferred = Q.defer<Arm.TemplateDeploymentOptions>();

        if (data.name()) {
            // Construct the parameters required by the ARM template.
            const parameters: StringMap<string> = {
                name: name,
                //location: location.name,
                pricingSpec: pricingSpec
            };

            // Fill out the template deployment options.
            const templateDeploymentOptions: Arm.TemplateDeploymentOptions = {
                subscriptionId: subscriptionId,
                resourceGroupName: resourceGroupName,
                resourceGroupLocation: resourceGroupLocation,
                parameters: parameters,
                deploymentName: galleryCreateOptions.deploymentName,
                resourceProviders: [Constants.resourceProvider],
                resourceId: resourceIdFormattedString,
                templateJson: this._getResourceTemplateJson()
                // Using below instead of above results in less repetition, but is static and immutable
                // templateLinkUri: galleryCreateOptions.deploymentTemplateFileUris["CreateResource"]
            };

            deferred.resolve(templateDeploymentOptions);
        } else {
            deferred.reject();
        }

        return deferred.promise;
    }

    private _getResourceTemplateJson(): string {
        return JSON.stringify({
            $schema: "http://schema.management.azure.com/schemas/2014-04-01-preview/deploymentTemplate.json#",
            contentVersion: "1.0.0.0",
            parameters: {
                name: { type: "string" },
                //location: { type: "string" },
                pricingSpec: { type: "string" }
            },
            resources: [
                {
                    apiVersion: "2018-05-01",
                    name: "[parameters('name')]",
                    location: "global",
                    type: resourceType,
                    properties: {
                    },
                    sku: {
                        name: "[parameters('pricingSpec')]"
                    }
                }
            ]
        });
    }

    private _createSpecDropdown(container: MsPortalFx.ViewModels.ContainerContract) {
        const pricingTierResources = ClientResources.CreateBlade.PricingTier;
        const pricingDocumentationLink = `<a href=\"https://aka.ms/azuremapspricing\" target=\"_blank\">${pricingTierResources.linkText}</a>`
        const options = <DropDown.Options<string>>{
            label: pricingTierResources.labelFormat.format(pricingDocumentationLink),
            items: [
                <DropDown.Group<string>>{
                    text: pricingTierResources.Groups.all,
                    children: [
                        <DropDown.Item<string>>{
                            text: Constants.serviceTiers.standard.name,
                            value: Constants.serviceTiers.standard.id
                        }
                    ]
                }
            ],
            validations: ko.observableArray<MsPortalFx.ViewModels.Validation>([
                new MsPortalFx.ViewModels.RequiredValidation(ClientResources.CreateBlade.Validation.specMessage)
            ])
    };
        
        return DropDown.create(container, options);
    }

    private _createLegalCheckbox(container: MsPortalFx.ViewModels.ContainerContract) {
        const licenseAgreementLink =
            "<a href=\"https://azure.microsoft.com/en-us/support/legal/\" target=\"_blank\">{0}</a>".format(
                ClientResources.CreateBlade.Legal.Checkbox.Label.license);
        const privacyStatementLink =
            "<a href=\"https://go.microsoft.com/fwlink/?LinkId=521839&clcid=0x409/\" target=\"_blank\">{0}</a>".format(
                ClientResources.CreateBlade.Legal.Checkbox.Label.privacy);

        const checkbox = new Forms.CheckBox.ViewModel(container, {
            inlineLabel: true,
            label: ko.observable(ClientResources.CreateBlade.Legal.Checkbox.Label.format.format(licenseAgreementLink, privacyStatementLink)),
            validations: ko.observableArray<MsPortalFx.ViewModels.Validation>([
                new MsPortalFx.ViewModels.RequiredValidation(ClientResources.CreateBlade.Legal.acceptTerms),
                new MsPortalFx.ViewModels.CustomValidation(ClientResources.CreateBlade.Legal.acceptTerms, (value: boolean) => {
                    var deferred = Q.defer<MsPortalFx.ViewModels.ValidationResult>();
                    deferred.resolve({ valid: value, message: ClientResources.CreateBlade.Legal.acceptTerms });
                    return deferred.promise;
                })
            ])
        });

        return checkbox;
    }
}
