import * as TemplateBlade from "Fx/Composition/TemplateBlade";
import { DataContext } from "./MyResourceArea";
import * as ClientResources from "ClientResources";
import Toolbars = MsPortalFx.ViewModels.Toolbars;
import Images = MsPortalFx.Base.Images;
import CopyableLabel = MsPortalFx.ViewModels.Controls.Forms.CopyableLabel;
import Constants = require("../Shared/Constants");
import Dialogs = MsPortalFx.ViewModels.Dialogs;
import Definition = require("../_generated/ExtensionDefinition");
import FxGrid = MsPortalFx.ViewModels.Controls.Lists.Grid;

export interface KeysBladeParameters 
{
	id: string;
}

interface KeyItemControls {
	keyValue: CopyableLabel.Contract;
	regenCommand: Toolbars.Toolbar;	
}

interface KeyItem 
{
	name: string;
	controls: KnockoutObservableBase<KeyItemControls>;
}

@TemplateBlade.Decorator({htmlTemplate: 
`
    <div class='msportalfx-docking'>
        <div class='msportalfx-docking-body msportalfx-padding'>
		    <div class='msportalfx-form' data-bind='pcControl: description' style="margin-bottom: 10px"></div>
			<br/>
            <div data-bind="text: keysLabel" class="msportalfx-text-header-regular"></div>
			<div data-bind='pcControl: keysGrid'></div>
        </div>
    </div>
`})
export class KeysBlade
{
    public title = ClientResources.KeysBladeBlade.title;
    public subtitle = ClientResources.KeysBladeBlade.subtitle;
    public context: TemplateBlade.Context<KeysBladeParameters, DataContext>;

    public description: MsPortalFx.ViewModels.Controls.Markdown.ViewModel;

	public keysLabel = ko.observable<string>(ClientResources.ResourceBlade.keysCommandLabel);

	public keysGrid: MsPortalFx.ViewModels.Controls.Lists.Grid.ViewModel<KeyItem, any>;

    /**
         * View model for the copyableLabel controls
         */
    private _primaryKeyCopyableViewModel: CopyableLabel.Contract;

    /**
         * View model for the copyableLabel controls
         */
    private _secondaryKeyCopyableViewModel: CopyableLabel.Contract;

    public onInitialize()
    {
		const { container, parameters } = this.context;

	    this.description = new MsPortalFx.ViewModels.Controls.Markdown.ViewModel(container, {
			markdown: ko.observable(ClientResources.KeysBlade.description)    
		});

	    this._initializeGrid(container);

        return Q(MsPortalFx.Base.Net2.ajax({
                uri: Constants.ArmServiceEndpoint + parameters.id + "/listKeys?" + Constants.ArmServiceVersion,
                type: "POST",
                dataType: "json",
                cache: false,
                traditional: true,
                contentType: "application/json",
                setAuthorizationHeader: true,
                invokeApi: "api/invoke"
            }))
            .then((response: any) => {
                this._primaryKeyCopyableViewModel.value(response.primaryKey);
                this._secondaryKeyCopyableViewModel.value(response.secondaryKey);
            })
            .catch((reason) => {
                MsPortalFx.Base.Diagnostics.Log.writeEntry(
                    MsPortalFx.Base.Diagnostics.LogEntryLevel.Error,
                    "KeyManagementBlade",
                    "Failed to list all keys: {0}".format(ko.toJSON(reason)));
            });
    }

	private _initializePrimaryKeyItem(container: MsPortalFx.Base.LifetimeManager): KeyItem {
		this._primaryKeyCopyableViewModel = new CopyableLabel.ViewModel(container);
		this._primaryKeyCopyableViewModel.disabled(true);
		
		const primaryKeyButton = new Toolbars.DialogButton({
			label: ClientResources.KeysBlade.regeneratePrimaryKeyCommandLabel,
			icon: Images.Pending(),
			dialogOptions: new Dialogs.MessageBox(
				ClientResources.KeysBlade.regeneratePrimaryKeyDialogTitle,
				ClientResources.KeysBlade.regenerateKeyConfirmation.format(ClientResources.KeysBlade.primaryKeyLabel.toLocaleLowerCase()),
				Dialogs.MessageBoxButtons.YesNo),
			command: { 
				canExecute: ko.pureComputed(() => { return true; }),
				execute: (dialogResult: Dialogs.DialogResult) => {
					if (dialogResult === MsPortalFx.ViewModels.Dialogs.DialogResult.Yes) {
						this._primaryKeyCopyableViewModel.value(ClientResources.KeysBlade.regeneratingAction);
						const { parameters } = this.context;
						return this._updateKeys(parameters.id, true);
					} else {
						return null;
					}
				},
			}
		});

		return this._createKeyItem(container, ClientResources.KeysBlade.primaryKeyLabel, this._primaryKeyCopyableViewModel, primaryKeyButton);
	}

	private _initializeGrid(container: MsPortalFx.Base.LifetimeManager): void {
		const nameColumnWidth = "15%";
		const actionColumnWidth = "40px";
		
		const keyItems = ko.observableArray<KeyItem>([
			this._initializePrimaryKeyItem(container),
			this._initializeSecondaryKeyItem(container)
		]);

		this.keysGrid = new FxGrid.ViewModel<KeyItem, any>(container, keyItems);

		const columnResources = ClientResources.KeysBlade.Grid.Columns;

		const formattedColumns: FxGrid.Column[] = [
			{
				itemKey: "name",
				name: ko.observable<string>(columnResources.Name.label),
				width: ko.observable<string>(nameColumnWidth)
			},
			{
				itemKey: "controls",
				name: ko.observable<string>(columnResources.Key.label),
				format: FxGrid.Format.HtmlBindings,
				formatOptions: {
					htmlBindingsTemplate: "<div data-bind='pcControl: value().keyValue'></div>"
				},
				editableFormat: FxGrid.Format.Text
			},
			{
				itemKey: "controls",
				format: FxGrid.Format.HtmlBindings,
				formatOptions: {
					htmlBindingsTemplate: "<div data-bind='pcControl: value().regenCommand'></div>"
				},  
				editableFormat: FxGrid.Format.Text,
				width: ko.observable<string>(actionColumnWidth)
			}
		];

		this.keysGrid.columns = ko.observableArray(formattedColumns);
		this.keysGrid.showHeader = true;
	}

	private _initializeSecondaryKeyItem(container: MsPortalFx.Base.LifetimeManager): KeyItem {
		this._secondaryKeyCopyableViewModel = new CopyableLabel.ViewModel(container);
		this._secondaryKeyCopyableViewModel.disabled(true);

		const secondaryKeyButton = new Toolbars.DialogButton({
			label: ClientResources.KeysBlade.regenerateSecondaryKeyCommandLabel,
			icon: Images.Pending(),
			dialogOptions: new Dialogs.MessageBox(
				ClientResources.KeysBlade.regenerateSecondaryKeyDialogTitle,
				ClientResources.KeysBlade.regenerateKeyConfirmation.format(ClientResources.KeysBlade.secondaryKeyLabel.toLocaleLowerCase()),
				Dialogs.MessageBoxButtons.YesNo),
			command: {
				canExecute: ko.pureComputed(() => { return true; }),
				execute: (dialogResult: Dialogs.DialogResult) => {
					if (dialogResult === MsPortalFx.ViewModels.Dialogs.DialogResult.Yes) {
						this._secondaryKeyCopyableViewModel.value(ClientResources.KeysBlade.regeneratingAction);
						const { parameters } = this.context;
						return this._updateKeys(parameters.id, false);
					} else {
						return null;
					}
				},
			}
		});

		return this._createKeyItem(container, ClientResources.KeysBlade.secondaryKeyLabel, this._secondaryKeyCopyableViewModel, secondaryKeyButton);
	}

	private _createKeyItem(container: MsPortalFx.Base.LifetimeManager, 
		name: string,
		label: CopyableLabel.Contract, 
		button: Toolbars.DialogButton): KeyItem {
		const toolbar = new Toolbars.Toolbar(container, {
			showLabels: false
		});
		toolbar.setItems([button]);

		return <KeyItem>{
			name: name,
			controls: ko.observable<KeyItemControls>({
				keyValue: label,
				regenCommand: toolbar
			})
		};
	}

    private _updateKeys(id: string, primaryKey: boolean): MsPortalFx.Base.Promise {
		const keyType = primaryKey ? ClientResources.KeysBlade.primaryKeyLabel : ClientResources.KeysBlade.secondaryKeyLabel;
		
        var body: string = `{\"keyType\": \"${primaryKey ? "primary" : "secondary"}\"}`;

		const resourceName = MsPortalFx.ViewModels.Services.ResourceTypes.parseResourceDescriptor(id).resource;
        const notificationResources = ClientResources.Keys.Regenerate.Notification;

        const notification = new MsPortalFx.Hubs.Notifications.ClientNotification({
            title: notificationResources.InProgress.title.format(keyType, resourceName),
            description: notificationResources.InProgress.description.format(keyType, resourceName),
            status: MsPortalFx.Hubs.Notifications.NotificationStatus.InProgress,
            percentComplete: 0,
            asset: {
                extensionName: Definition.definitionName,
                assetType: Definition.AssetTypeNames.maps,
                assetId: id
            }
        });

        notification.publish();

        return MsPortalFx.Base.Net.ajax({
                uri: Constants.ArmServiceEndpoint + id + "/regenerateKey?" + Constants.ArmServiceVersion,
                type: "POST",
                dataType: "json",
                cache: false,
                traditional: true,
                contentType: "application/json",
                setAuthorizationHeader: true,
                data: body
            })
            .done((keysResponse: any) => {
                this._primaryKeyCopyableViewModel.value(keysResponse.primaryKey);
                this._secondaryKeyCopyableViewModel.value(keysResponse.secondaryKey);

				notification.title = notificationResources.Success.title.format(keyType, resourceName);
                notification.description = notificationResources.Success.description.format(keyType, resourceName);
                notification.status = MsPortalFx.Hubs.Notifications.NotificationStatus.Success;
                notification.percentComplete = 100;

                notification.publish();
            })
            .catch((reason) => {

				notification.title = notificationResources.Success.title.format(keyType, resourceName);
                notification.description = notificationResources.Success.description.format(keyType, resourceName);
                notification.status = MsPortalFx.Hubs.Notifications.NotificationStatus.Success;
                notification.percentComplete = 100;

                notification.publish();

                MsPortalFx.Base.Diagnostics.Log.writeEntry(
                    MsPortalFx.Base.Diagnostics.LogEntryLevel.Error,
                    "RegenKeyCommand",
                    "Failed to generate key for resource '{0}': {1}".format(id, ko.toJSON(reason)));
            });
    }

}