import * as TemplateBlade from "Fx/Composition/TemplateBlade";
import { DataContext } from "./MyResourceArea";
import * as ClientResources from "ClientResources";
import Toolbars = MsPortalFx.ViewModels.Toolbars;
import Images = MsPortalFx.Base.Images;
import Manager = FxImpl.AssetManager;
import Definition = require("../_generated/ExtensionDefinition");
import DataModels = LocSvcExtension.DataModels;
import MoveResourceButton = require("Fx/Controls/MoveResourceToolbarButton");
import { AccountDeletionConfirmationBladeReference, AccountDeletionConfirmationBladeOutputs } from "../_generated/BladeReferences";

import * as Essentials from "Fx/Controls/Essentials";
import * as MonitorChartV2 from "Fx/Controls/MonitorChartV2";
import * as OptionsGroup from "Fx/Controls/OptionsGroup";

export interface MyResourceBladeOverviewBladeParameters {
    id: string;
}

const enum TimespanEnums {
    OneHour,
    SixHours,
    TwelveHours,
    OneDay,
    SevenDays,
    ThirtyDays
}

@TemplateBlade.Decorator({
    htmlTemplate: "../Templates/MyResourceBladeOverviewBlade.html"    
})

export class MyResourceBladeOverviewBlade {

    title = ClientResources.MyResourceBladeOverviewBladeBlade.title;
    subtitle = ClientResources.MyResourceBladeOverviewBladeBlade.subtitle;
    context: TemplateBlade.Context<MyResourceBladeOverviewBladeParameters, DataContext>;
    essentialsViewModel: KnockoutObservable<Essentials.ResourceLayoutContract>;

    // Chart objects
    timespanOptionsGroup: OptionsGroup.Contract<TimespanEnums>;
    requestsChartContract: MonitorChartV2.Contract;
    errorsChartContract: MonitorChartV2.Contract;
    availabilityChartContract: MonitorChartV2.Contract;
    
    private _entityView: MsPortalFx.Data.EntityView<DataModels.Account, string>;

    onInitialize() {
        const { container, parameters } = this.context;

        const resourceId = parameters.id; // this is how you get parameters that were passed into your blade 

        this._entityView = this.context.model.accountEntities.createView(container);
        this._entityView.fetch(resourceId);

        this._initializeEssentials();
        this._initializeCommandBar();
        this._initializeCharts();

        return Q();   // todo - if you load data then return a loading promise here
    }

    /** Initializes the essentials control */
    private _initializeEssentials(): void {

        const { container, parameters } = this.context;
        const resourceId = parameters.id;
        const essentialsVm = Essentials.create(container,
            {
                resourceId: resourceId,
                hiddenBuiltInTypes: [Essentials.BuiltInType.Status],
                hiddenChangeLink: [],
                expanded: ko.observable(true), // todo - hook up to settings if you want to remember the state
                additionalRight: [
                    {
                        label: ClientResources.Account.pricingTier,
                        value: ko.computed(container,
                            () => {
                                const account = this._entityView.item();
                                const sku = account && account.sku && account.sku();
                                return sku && sku.name && sku.tier();
                            })
                    } as Essentials.Item
                ]
            });

        this.essentialsViewModel = ko.observable(essentialsVm);
    }

    /** Initializes the command bar */
    private _initializeCommandBar(): void {

        const { container, parameters } = this.context;

        const resourceId = parameters.id;

        const moveResourceButton = new MoveResourceButton.ViewModel(container,
            {
                resourceId: resourceId
            });

        const deleteButton = new Toolbars.CommandButton({
            label: ClientResources.OverviewBlade.deleteCommandLabel,
            icon: Images.Delete(),
            command: {
                canExecute: ko.pureComputed(() => { return true; }),
                execute: () => {
                    const deleteConfirmationBlade = new AccountDeletionConfirmationBladeReference(
                        { data: resourceId },
                        {
                            onOutputsChanged: (outputs: AccountDeletionConfirmationBladeOutputs) => {
                                const action = outputs && outputs.actionBarOutput && outputs.actionBarOutput.action;
                                if (action) {
                                    if (action === MsPortalFx.ViewModels.ActionBars.Base.Action.Complete) {
                                        container.closeChildBlade();
                                        this._delete(resourceId);
                                    } else if (action === MsPortalFx.ViewModels.ActionBars.Base.Action.Cancel) {
                                        container.closeChildBlade();
                                    }
                                }
                            }
                        });

                    return container.openBlade(deleteConfirmationBlade);
                },
            }
        });

        const commandBar = new Toolbars.Toolbar(container);
        commandBar.setItems([moveResourceButton, deleteButton]);
        container.commandBar = commandBar;
    }

    /** Initialize and Render the charts */
    private _initializeCharts(): void {
        const { container, parameters } = this.context;
        const resourceId = parameters.id;

        // Configuration for charts
        const requestsChartOptions: MonitorChartV2.Options = {
            title: "Total Requests",
            metrics: [
                {
                    name: "Usage",
                    resourceMetadata: {
                        id: resourceId
                    },
                    aggregationType: MonitorChartV2.Metric.AggregationType.Count
                }
            ]
        };

        const errorsChartOptions: MonitorChartV2.Options = {
            title: "Total Errors",
            metrics: [
                {
                    name: "Usage",
                    resourceMetadata: {
                        id: resourceId
                    },
                    aggregationType: MonitorChartV2.Metric.AggregationType.Count
                }
            ],
            filterCollection: {
                filters: [
                    {
                        key: "ResultType",
                        values: [
                            "ClientError",
                            "Failure"
                        ]
                    }
                ]
            }
        };

        const availabilityChartOptions: MonitorChartV2.Options = {
            title: "Availability",
            metrics: [
                {
                    name: "Availability",
                    resourceMetadata: {
                        id: resourceId
                    },
                    aggregationType: MonitorChartV2.Metric.AggregationType.Avg
                }
            ]
        };

        this.requestsChartContract = MonitorChartV2.create(container, requestsChartOptions);
        this.errorsChartContract = MonitorChartV2.create(container, errorsChartOptions);
        this.availabilityChartContract = MonitorChartV2.create(container, availabilityChartOptions);

        const chartContracts = [
            this.requestsChartContract,
            this.errorsChartContract,
            this.availabilityChartContract
        ];

        // Create an OptionsGroup UI object that defines timespan buttons 
        const timespanSelected: KnockoutObservable<TimespanEnums> = ko.observable(TimespanEnums.OneDay);
        this.timespanOptionsGroup = OptionsGroup.create<TimespanEnums>(container, {
            label: ClientResources.MyResourceBladeOverviewBlade.Charts.OptionsGroup.label,
            items: [
                {
                    text: ClientResources.MyResourceBladeOverviewBlade.Charts.oneHour,
                    value: TimespanEnums.OneHour
                },
                {
                    text: ClientResources.MyResourceBladeOverviewBlade.Charts.sixHours,
                    value: TimespanEnums.SixHours
                },
                {
                    text: ClientResources.MyResourceBladeOverviewBlade.Charts.twelveHours,
                    value: TimespanEnums.TwelveHours
                },
                {
                    text: ClientResources.MyResourceBladeOverviewBlade.Charts.oneDay,
                    value: TimespanEnums.OneDay
                },
                {
                    text: ClientResources.MyResourceBladeOverviewBlade.Charts.sevenDays,
                    value: TimespanEnums.SevenDays
                },
                {
                    text: ClientResources.MyResourceBladeOverviewBlade.Charts.thirtyDays,
                    value: TimespanEnums.ThirtyDays
                }
            ],
            value: timespanSelected,
            suppressDirtyBehavior: true
        });

        // Attach events for interacting with above OptionsGroup
        timespanSelected.subscribe(container, (updatedTimespan: TimespanEnums) => {
            // Update timespan for below charts
            chartContracts.forEach(chartContract => {
                chartContract.timespan(this._getTimespan(updatedTimespan));
            });
        });
    }

    private _getTimespan(timeRange: TimespanEnums): MonitorChartV2.Timespan {
        let durationMs: number;
        switch (timeRange) {
            case TimespanEnums.OneHour:
                durationMs = 3600000;
                break;
            case TimespanEnums.SixHours:
                durationMs = 21600000;
                break;
            case TimespanEnums.TwelveHours:
                durationMs = 43200000;
                break;
            case TimespanEnums.OneDay:
                durationMs = 86400000;
                break;
            case TimespanEnums.SevenDays:
                durationMs = 604800000;
                break;
            case TimespanEnums.ThirtyDays:
                durationMs = 2592000000;
                break;
        };

        return <MonitorChartV2.Timespan>{
            relative: {
                duration: durationMs
            }
        }
    }

    private _delete(id: string): MsPortalFx.Base.Promise {
        const invokeDelete = this.context.model.deleteResource(id);

        const resourceName = MsPortalFx.ViewModels.Services.ResourceTypes.parseResourceDescriptor(id).resource;
        const notificationResources = ClientResources.Overview.Delete.Notification;

        const notification = new MsPortalFx.Hubs.Notifications.ClientNotification({
            title: notificationResources.InProgress.title.format(ClientResources.MyResourceBladeAsset.lowerSingularDisplayName, resourceName),
            description: notificationResources.InProgress.description.format(ClientResources.MyResourceBladeAsset.lowerSingularDisplayName, resourceName),
            status: MsPortalFx.Hubs.Notifications.NotificationStatus.InProgress,
            percentComplete: 0,
            asset: {
                extensionName: Definition.definitionName,
                assetType: Definition.AssetTypeNames.maps,
                assetId: id
            }
        });

        notification.publish();

        return invokeDelete
            .progress(() => {
                notification.percentComplete = 50;
                notification.publish();
            })
            .then(() => {
                this.context.container.closeCurrentBlade();
                this.context.model.accountEntities.forceRemove(id);
                notification.title = notificationResources.Success.title.format(ClientResources.MyResourceBladeAsset.lowerSingularDisplayName, resourceName);
                notification.description = notificationResources.Success.description.format(ClientResources.MyResourceBladeAsset.lowerSingularDisplayName, resourceName);
                notification.status = MsPortalFx.Hubs.Notifications.NotificationStatus.Success;
                notification.percentComplete = 100;

                notification.publish();
                Manager.notifyAssetDeleted(Definition.AssetTypes.Maps.name, id, Definition.definitionName);
            })
            .catch(() => {
                notification.title = notificationResources.Error.title.format(ClientResources.MyResourceBladeAsset.lowerSingularDisplayName, resourceName);
                notification.description = notificationResources.Error.description.format(ClientResources.MyResourceBladeAsset.lowerSingularDisplayName, resourceName);
                notification.status = MsPortalFx.Hubs.Notifications.NotificationStatus.Error;

                notification.publish();
            });
    }
}