// Copyright (c) Microsoft Corporation. All rights reserved.

/**
 * Runs an operation against ARM using a CORS request.
 */
export function runOperation(uri: string, operationType?: string, data?: string, contentType?: string): MsPortalFx.Base.PromiseN<OperationProgress> {

    return runOperationWithSettings({ uri: uri, type: operationType || "POST", data: data, contentType: contentType });
}

export function runOperationWithSettings<T>(settings: MsPortalFx.Base.Net.NetAjaxSettings<T>): MsPortalFx.Base.PromiseVN<T, OperationProgress> {
    var deferred = Q.defer<T>();

    Q(MsPortalFx.Base.Net.ajax({
        uri: settings.uri,
        type: settings.type || "POST",
        setAuthorizationHeader: true,
        data: settings.data,
        dataType: settings.dataType,
        contentType: settings.contentType,
        invokeApi: "api/invoke"
    }).then((data, status, jqXHR) => { return { data: data, status: status, jqXHR: jqXHR }; }))
        .then(reportCorrelationId.bind(null, deferred))
        .then(deferred.resolve, deferred.reject);

    return deferred.promise;
}

export interface OperationProgress {
    correlationId: string;
}

function reportCorrelationId(deferred: Q.Deferred<any>, result: jQueryAjaxResult): MsPortalFx.Base.Promise {
    var correlationId = result.jqXHR.getResponseHeader("x-ms-correlation-request-id");
    if (correlationId) {
        deferred.notify({ correlationId: correlationId });
    }
    return pollOperation(result);
}

function pollOperation(result: jQueryAjaxResult): MsPortalFx.Base.Promise {
    var location = result.jqXHR.getResponseHeader("Location");
    if (location) {
        return Q.delay(pollInterval)
            .then(() => {
                return Q(MsPortalFx.Base.Net.ajax({
                    uri: location,
                    setAuthorizationHeader: true
                }).then((data, status, jqXHR) => { return { data: data, status: status, jqXHR: jqXHR }; }))
                    .then(pollOperation);
            });
    }
    return result.data;
}

var pollInterval = 10000;

interface jQueryAjaxResult {
    data: any;
    status: string;
    jqXHR: JQueryXHR<any>;
}

/**
 * A reusable data supplier for CORS requests.
 */
export function corsDataSupplier<T>(httpMethod: string, uri: string, headers: StringMap<any>, data: any): MsPortalFx.Base.PromiseV<T> {
    return MsPortalFx.Base.Net.ajax({
        uri: uri,
        type: httpMethod,
        dataType: "json",
        traditional: true,
        headers: $.extend({}, headers),
        contentType: "application/json",
        setAuthorizationHeader: true,
        data: data,
        invokeApi: "api/invoke"
    });
}