﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderPermissions.Authorization
{
    using System;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Threading.Tasks;
    using Microsoft.Azure.Documents;
    using Microsoft.Azure.Documents.Client;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ResourceProviderPermissions.Configuration;    

    public class DocumentsRoleAssignmentsRepository : IRepository<RoleAssignments>
    {
        private readonly PermissionsJobHostSettings settings;
        private readonly IDocumentClient documentClient;
        
        public DocumentsRoleAssignmentsRepository(IDocumentClient documentClient, PermissionsJobHostSettings settings)
        {
            this.documentClient = documentClient;
            this.settings = settings;
        }

        /// <summary>
        /// Factory method to construct a new instance of <see cref="DocumentsRoleAssignmentsRepository"/>
        /// </summary>
        /// <param name="settings">the settings for the environment</param>
        /// <param name="documentClient">Document Storage Client</param>
        /// <returns>new instance of <see cref="DocumentsRoleAssignmentsRepository"/></returns>
        public static async Task<IRepository<RoleAssignments>> BuildAsync(PermissionsJobHostSettings settings, DocumentClient documentClient)
        {
            try
            {
                if (settings == null)
                {
                    throw new ArgumentNullException(nameof(settings));
                }

                if (documentClient == null)
                {
                    throw new ArgumentNullException(nameof(documentClient));
                }
                
                Database db = await documentClient.CreateDatabaseIfNotExistsAsync(new Database() { Id = CosmosDbConstants.AccountsDatabaseName });
                
                await documentClient.CreateDocumentCollectionIfNotExistsAsync(db.SelfLink, new DocumentCollection()
                {
                    PartitionKey = new PartitionKeyDefinition() { Paths = new Collection<string>() { "/" + CosmosDbConstants.PartitionKeyConst } },
                    Id = settings.CosmosDbRbacCollectionName,
                    IndexingPolicy = new IndexingPolicy(
                        new RangeIndex(DataType.String, -1))
                        {
                            Automatic = true,
                            IncludedPaths = new Collection<IncludedPath>()
                            {
                                new IncludedPath()
                                {
                                        Indexes = new Collection<Index>() { new RangeIndex(DataType.String, -1) },
                                        Path = "/*"
                                }
                            }
                        },                    
                }, new RequestOptions() { OfferThroughput = settings.CosmosDbRbacCollectionRequestUnits });

                Agent.Instance.LogEvent(EventType.Information, $"Connected to document db account {settings.CosmosDbAccountsUri}");
                return new DocumentsRoleAssignmentsRepository(documentClient, settings);
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "permissionsJob", "DocumentsRoleAssignmentsRepository_Startup");
                throw;
            }
        }
        
        public async Task UpsertAsync(RoleAssignments obj)
        {
            if (obj == null)
            {
                throw new ArgumentNullException(nameof(obj));
            }

            if (string.IsNullOrWhiteSpace(obj.SubscriptionId))
            {
                throw new ArgumentNullException(nameof(obj.SubscriptionId));
            }

            try
            {
                List<CheckAccessAssignment> checkAccessAssignments = obj.ToCheckAccessAssignments();
                Uri collectionUri = UriFactory.CreateDocumentCollectionUri(CosmosDbConstants.AccountsDatabaseName, settings.CosmosDbRbacCollectionName);

                Document roleAssignments = new Document()
                {
                    Id = CosmosDbConstants.RoleAssignmentTypePropertyValueConst
                };
                roleAssignments.SetPropertyValue("value", checkAccessAssignments);
                roleAssignments.SetPropertyValue(CosmosDbConstants.PartitionKeyConst, obj.SubscriptionId);
                await documentClient.UpsertDocumentAsync(collectionUri, roleAssignments, new RequestOptions()
                {
                    IndexingDirective = IndexingDirective.Exclude
                });
            }
            catch (DocumentClientException dx)
            {
                Agent.Instance.LogException(dx, "permissionsJob", "DocumentsRoleAssignmentsRepository_UpsertAsync");
                throw;
            }
        }
    }
}
