﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------'

namespace Microsoft.Azure.LocationServices.ResourceProviderPermissions.Authorization
{
    using System;
    using System.Security.Cryptography.X509Certificates;
    using System.Threading.Tasks;
    using Microsoft.Azure.KeyVault;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.ResourceProviderPermissions.Configuration;
    using Microsoft.Azure.Services.AppAuthentication;    

    public class KeyVaultCertificateResolver : ICertificateResolver, IDisposable
    {
        private readonly AzureServiceTokenProvider azureServiceTokenProvider;
        private readonly KeyVaultClient keyVaultClient;
        private readonly PermissionsJobHostSettings settings;
        private readonly IConfigurationCache configurationCache;
        private bool disposedValue = false; // To detect redundant calls              
        
        public KeyVaultCertificateResolver(PermissionsJobHostSettings settings, IConfigurationCache configurationCache)
        {
            azureServiceTokenProvider = new AzureServiceTokenProvider();
            keyVaultClient = new KeyVaultClient(new KeyVaultClient.AuthenticationCallback(azureServiceTokenProvider.KeyVaultTokenCallback));
            this.settings = settings;
            this.configurationCache = configurationCache;
        }

        public async Task<X509Certificate2> ResolveCertificateAsync(string subjectName)
        {
            if (!settings.AAD_ARMS2S_SubjectToSecrets.TryGetValue(subjectName, out string secretId))
            {
                throw new ArgumentException(nameof(subjectName));
            }

            var keyvaultSettingFactory = new Lazy<Task<X509Certificate2>>(
                async () =>
                {
                    var secret = await keyVaultClient.GetSecretAsync(secretId).ConfigureAwait(false);
                    byte[] base64BytesCertificate = Convert.FromBase64String(secret.Value);
                    X509Certificate2Collection x509Certificate2Collection = new X509Certificate2Collection();
                    x509Certificate2Collection.Import(base64BytesCertificate, null, X509KeyStorageFlags.Exportable);
                    string tempPassword = DateTime.UtcNow.ToString("s");
                    byte[] exportedPfx = x509Certificate2Collection.Export(X509ContentType.Pfx, tempPassword);
                    return new X509Certificate2(exportedPfx, tempPassword);
                });

            return await configurationCache.ResolveSettingAsync(subjectName, keyvaultSettingFactory, TimeSpan.FromMinutes(20)).ConfigureAwait(false);
        }        

        #region IDisposable Support        
        // This code added to correctly implement the disposable pattern.
        public void Dispose()
        {
            // Do not change this code. Put cleanup code in Dispose(bool disposing) above.
            Dispose(true);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    configurationCache.Dispose();
                    keyVaultClient.Dispose();
                }

                disposedValue = true;
            }
        }
        #endregion
    }
}
