﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderPermissions.Authorization
{
    using System;
    using System.Collections.Generic;
    using System.Net.Http;
    using System.Net.Http.Headers;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.ResourceProviderPermissions.Configuration;        

    public class RoleAssignments : IRoleAssignments
    {
        private const string AuthorizationHeaderName = "Authorization";
        private const string ApplicationJson = "application/json";
        private const string ApiVersion = "2018-05-01";
        private readonly HttpClient client;
        private readonly PermissionsJobHostSettings settings;
        private readonly ITokenClient tokenClient;
        private readonly IRepository<RoleAssignments> repository;

        public RoleAssignments(HttpClient client, PermissionsJobHostSettings settings, ITokenClient tokenClient, IRepository<RoleAssignments> repository)
        {
            this.client = client;
            this.settings = settings;
            this.tokenClient = tokenClient;
            this.repository = repository;
            Assignments = new List<RoleAssignment>();
        }

        public bool IsLoaded { get; private set; }

        public string SubscriptionId { get; private set; }

        public List<RoleAssignment> Assignments { get; private set; }
        
        public async Task LoadAsync(Subscription subscription, CancellationToken cancellationToken)
        {
            if (subscription == null)
            {
                throw new ArgumentNullException(nameof(subscription));
            }

            string AuthorizationHeader = await tokenClient.AcquireAuthorizationHeaderAsync(subscription.TenantId);
            SubscriptionId = subscription.Id;
            HttpRequestMessage getRoleAssignmentsRequest = null;
            string url = $"{settings.ARM_Endpoint}subscriptions/{subscription.Id}/providers/Microsoft.Authorization/roleAssignments?api-version=2015-07-01";
            getRoleAssignmentsRequest = BuildGetRequestMessage(AuthorizationHeader, url);

            RoleAssignmentListResponse roleAssignmentListResponse = null;

            // todo: work with the nextLink
            IsLoaded = false;
            Assignments.Clear();
            bool hasNextLink = true;
            while (hasNextLink)
            {
                if (cancellationToken.IsCancellationRequested)
                {
                    return;
                }

                HttpResponseMessage getRoleAssignmentsResponse = await client.SendAsync(getRoleAssignmentsRequest);
                if (!getRoleAssignmentsResponse.IsSuccessStatusCode)
                {
                    Instrumentation.Agent.Instance.LogEvent(Instrumentation.EventType.ProgramError,
                        $"PermissionsJob: retrieving role assignments for {subscription.Id} failed with {getRoleAssignmentsResponse.StatusCode} {await getRoleAssignmentsResponse.Content.ReadAsStringAsync()}");
                    IsLoaded = false;
                    return;
                }
                else
                {
                    roleAssignmentListResponse = await getRoleAssignmentsResponse.Content.ReadAsAsync<RoleAssignmentListResponse>();
                    if (roleAssignmentListResponse == null)
                    {
                        Instrumentation.Agent.Instance.LogEvent(Instrumentation.EventType.Warning, $"Did not load Role Assignments for {url}");
                        return;
                    }

                    Assignments.AddRange(roleAssignmentListResponse.Assignments);
                    if (string.IsNullOrWhiteSpace(roleAssignmentListResponse.NextLink))
                    {
                        hasNextLink = false;
                    }
                    else
                    {
                        url = roleAssignmentListResponse.NextLink; // need to see if this works...
                        getRoleAssignmentsRequest = BuildGetRequestMessage(AuthorizationHeader, url);
                    }
                }
            }
            
            IsLoaded = true;
        }

        public async Task SaveAsync()
        {
            if (Assignments == null || Assignments.Count == 0)
            {
                return;
            }

            if (!IsLoaded)
            {
                throw new InvalidOperationException("RoleAssignments has not be been loaded.");
            }

            await repository.UpsertAsync(this).ConfigureAwait(false);
        }

        public List<CheckAccessAssignment> ToCheckAccessAssignments()
        {
            if (Assignments == null)
            {
                return new List<CheckAccessAssignment>();
            }

            List<CheckAccessAssignment> checkAccessAssignments = new List<CheckAccessAssignment>(Assignments.Count);

            if (Assignments.Count == 0)
            {
                return checkAccessAssignments;
            }

            foreach (RoleAssignment assignment in Assignments)
            {
                CheckAccessAssignment checkAccessAssignment = assignment.ToCheckAccessAssignment();
                Guid roleDefintionId = Guid.Parse(checkAccessAssignment.RoleDefinitionId);
                if (roleDefintionId == settings.AzureMapsDataReaderRoleDefinitionId)
                {
                    checkAccessAssignments.Add(checkAccessAssignment);
                }   
            }

            return checkAccessAssignments;
        }

        private HttpRequestMessage BuildGetRequestMessage(string authorizationHeader, string url)
        {
            HttpRequestMessage message = new HttpRequestMessage(HttpMethod.Get, url);
            message.Headers.Add(AuthorizationHeaderName, authorizationHeader);
            message.Headers.Accept.Add(new MediaTypeWithQualityHeaderValue(ApplicationJson));
            return message;
        }
    }
}
