﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderPermissions.Configuration
{
    using System;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Globalization;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;

    /// <summary>
    /// Represents the configuration necessary to operate in an environment.
    /// </summary>
    /// <remarks>
    /// This class should only contain members that differ from environment to environment. When something doesn't 
    /// vary, it should not be present in this class, and instead should be a compile-time constant. If a property
    /// is no longer used, or no longer varies, it should be removed.
    /// </remarks>
    public abstract class PermissionsJobHostSettings
    {        
        /// <summary>
        /// Gets the Instrumentation Parameters
        /// </summary>
        public abstract InitParams MdmParameters { get; }

        /// <summary>
        /// Preffered Locations for cosmos db read locations.
        /// </summary>
        public abstract Collection<string> CosmosDbPrefferedLocations { get; }       

        /// <summary>
        /// Document DB accounts URI
        /// </summary>
        public Uri CosmosDbAccountsUri => new Uri($"https://{MapsFrontEndCosmosDbAccountName}.documents.azure.com");

        /// <summary>
        /// Document DB account key secret id
        /// </summary>
        public string CosmosDbAccountsKeyId => $"https://{GlobalKeyVaultName}.vault.azure.net:443/secrets/mapsFrontEndCosmosDbKey";

        /// <summary>
        /// Document DB account collection request unit throughput, adjust this to scale.
        /// </summary>
        public int CosmosDbLeaseCollectionRequestUnits => 400;

        /// <summary>
        /// Flag to enable cosmos DB account feature work.
        /// </summary>
        public abstract bool EnableAccountReplication { get; }

        /// <summary>
        /// The minutes a current time to look back to query for stale subscriptions.
        /// </summary>
        public int MinuteStalenessSubscriptionRefresh => -5;

        /// <summary>
        /// The authority to request access from, the template parameter represents the customer AAD tenant id.
        /// </summary>
        public string AAD_ARMS2S_AuthorityTemplate => "https://login.windows-ppe.net/{0}/";

        /// <summary>
        /// The authorized certificate subjects which should be installed on the VM authorized to grant tokens.
        /// Note this should be configured in AAD first party apps.
        /// </summary>
        public string AAD_ARMS2S_CertificateSubject => "CN=c-rp-arm.atlas.microsoft.com";

        /// <summary>
        /// Mapping of certificate subject names to their secret ids in Key Vault. These are used to download the certificate.
        /// </summary>
        public ReadOnlyDictionary<string, string> AAD_ARMS2S_SubjectToSecrets => new ReadOnlyDictionary<string, string>(
           new Dictionary<string, string>()
           {
                { AAD_ARMS2S_CertificateSubject, $"https://{RegionalKeyVaultName}.vault.azure.net/secrets/c-rp-arm/" }
           });

        /// <summary>
        /// The AAD Application registration which represents the Resource Provider as the client requesting access from the
        /// authority. This is configured in AAD first party apps.
        /// </summary>
        public string AAD_ARMS2S_ClientId => "608f6f31-fed0-4f7b-809f-90f6c9b3de78";

        /// <summary>
        /// This is defined in the ACIS commands for PAS RP, currently this is the only data role but we can extend this later.
        /// </summary>
        public Guid AzureMapsDataReaderRoleDefinitionId => new Guid("423170ca-a8f6-4b0f-8487-9e4eb8f49bfa");

        /// <summary>
        /// The AAD resource (ARM) which the client ID will request a token for in the authority.
        /// </summary>
        public string AAD_ARMS2S_Resource => "https://management.core.windows.net/";

        /// <summary>
        /// Endpoint to call ARM for this environment.
        /// </summary>
        public string ARM_Endpoint => "https://api-dogfood.resources.windows-int.net/";

        /// <summary>
        /// Document DB account collection request unit throughput, adjust this to scale.
        /// </summary>
        public int CosmosDbRbacCollectionRequestUnits => 5500;

        /// <summary>
        /// Document Db account collection name for role definitions and role assignments.
        /// </summary>
        public string CosmosDbRbacCollectionName => "rbac";

        /// <summary>
        /// Gets the current region.
        /// </summary>
        public string CurrentRegion
        {
            get
            {
                string currentRegion = Environment.GetEnvironmentVariable("Region");
                return currentRegion ?? "DEFAULT";
            }
        }

        /// <summary>
        /// Name of the account for maps front end cosmos db instance
        /// </summary>
        protected abstract string MapsFrontEndCosmosDbAccountName { get; }

        /// <summary>
        /// Name of the key vault where all runtime secrets are referenced
        /// </summary>
        protected abstract string GlobalKeyVaultName { get; }

        /// <summary>
        /// Name of the key vault where certificates are installed and it is region specific.
        /// </summary>
        protected abstract string RegionalKeyVaultName { get; }

        /// <summary>
        /// Gets the current environment.
        /// </summary>
        /// <exception cref="ArgumentException">
        /// Throws if the current environment is not specified
        /// </exception>
        private static string CurrentEnvironment
        {
            get
            {
                string currentEnvironment = Environment.GetEnvironmentVariable("Environment");

                if (string.IsNullOrEmpty(currentEnvironment))
                {
                    throw new ArgumentException("Current Environment could not be determined");
                }

                return currentEnvironment;
            }
        }

        /// <summary>
        /// Create an environment definition from the environment environment variable.
        /// </summary>
        /// <returns>The created environment definition.</returns>
        public static PermissionsJobHostSettings CreateFromEnvironment()
        {
            switch (CurrentEnvironment.ToUpperInvariant())
            {
                case LogicalEnvironments.Development:
                    return new PermissionsJobHostDevSettings();
                case LogicalEnvironments.Test:
                    return new PermissionsJobHostTestSettings();
                case LogicalEnvironments.Staging:
                    return new PermissionsJobHostStagingSettings();
                case LogicalEnvironments.Production:
                    return new PermissionsJobHostProdSettings();
                default:
                    throw new ArgumentException(
                        string.Format(
                            CultureInfo.InvariantCulture,
                            "Environment {0} does not have a corresponding Environment definition defined",
                            CurrentEnvironment));
            }
        }
    }
}