﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderPermissions
{
    using System;
    using System.Linq;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.Documents;
    using Microsoft.Azure.Documents.Client;
    using Microsoft.Azure.Documents.Linq;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ResourceProviderPermissions.Configuration;

    public class DocumentSubscriptionsRepository : IQueueRepository<Subscriptions>
    {
        private const int takeCountConst = 32;
        private readonly IDocumentClient documentClient;
        private readonly PermissionsJobHostSettings settings;
        
        public DocumentSubscriptionsRepository(IDocumentClient documentClient, PermissionsJobHostSettings settings)
        {
            this.documentClient = documentClient;
            this.settings = settings;
        }

        public async Task DeleteAsync(string id)
        {
            if (string.IsNullOrWhiteSpace(id))
            {
                throw new ArgumentNullException(nameof(id));
            }

            // Let's delete this record in queue if it's not in use and move on.
            Uri docUri = UriFactory.CreateDocumentUri(CosmosDbConstants.AccountsDatabaseName, CosmosDbConstants.SubscriptionCollectionName, id);
            await documentClient.DeleteDocumentAsync(docUri);
        }

        public async Task<Subscriptions> PullAsync(Subscriptions subscriptions, CancellationToken cancellationToken)
        {
            if (subscriptions == null)
            {
                throw new ArgumentNullException(nameof(subscriptions));
            }

            try
            {
                Uri collectionUri = UriFactory.CreateDocumentCollectionUri(CosmosDbConstants.AccountsDatabaseName, CosmosDbConstants.SubscriptionCollectionName);
                IDocumentQuery<Subscription> query = documentClient.CreateDocumentQuery<Subscription>(collectionUri, new FeedOptions()
                {
                    MaxDegreeOfParallelism = -1,
                    MaxItemCount = takeCountConst,
                })
                    .Where(s => s.EnqueueTimeUtc <= DateTime.UtcNow.AddMinutes(settings.MinuteStalenessSubscriptionRefresh))
                    .Take(takeCountConst)
                    .AsDocumentQuery();
                FeedResponse<Subscription> feedResponse = await query.ExecuteNextAsync<Subscription>(cancellationToken);

                // TODO: We need to collect some better metrics than via trace events.
                Agent.Instance.LogEvent(EventType.Information, $"ResourceProviderPermissions: Query Metadata: RUs {feedResponse.RequestCharge}, Activity ID {feedResponse.ActivityId}, CollectionKBSize: {feedResponse.CollectionSizeUsage}, collectionSizeQuota: {feedResponse.CollectionSizeQuota}");

                // this should only appear of the query is cross partitioned.
                if (feedResponse.QueryMetrics != null)
                {
                    foreach (var queryKey in feedResponse.QueryMetrics.Keys)
                    {
                        Agent.Instance.LogEvent(EventType.Information, $"ResourceProviderPermissions: {queryKey} QueryMetrics: {feedResponse.QueryMetrics[queryKey].ToString()}");
                    }
                }

                subscriptions.Items.Clear();
                subscriptions.Items.AddRange(feedResponse);                
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "permissionsjob", "pullasync");
            }

            return subscriptions;
        }

        public async Task UpsertAsync(Subscriptions items)
        {
            if (items == null)
            {
                throw new ArgumentNullException(nameof(items));
            }

            Uri collectionUri = UriFactory.CreateDocumentCollectionUri(CosmosDbConstants.AccountsDatabaseName, CosmosDbConstants.SubscriptionCollectionName);
            foreach (Subscription subscription in items.Items)
            {
                await documentClient.UpsertDocumentAsync(collectionUri, subscription);
            }            
        }
    }
}
