﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderPermissions
{
    using System;
    using System.Fabric;
    using System.Net.Http;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.Documents.Client;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ResourceProviderPermissions.Authorization;
    using Microsoft.Azure.LocationServices.ResourceProviderPermissions.Configuration;
    using Microsoft.ServiceFabric.Services.Runtime;

    /// <summary>
    /// An instance of this class is created for each service instance by the Service Fabric runtime.
    /// </summary>
    internal sealed class PermissionsService : StatelessService
    {
        public PermissionsService(StatelessServiceContext context)
            : base(context)
        {
        }

        /// <summary>
        /// This is the main entry point for your service instance.
        /// </summary>
        /// <param name="cancellationToken">Canceled when Service Fabric needs to shut down this service instance.</param>
        protected override async Task RunAsync(CancellationToken cancellationToken)
        {
            PermissionsJobHostSettings settings = PermissionsJobHostSettings.CreateFromEnvironment();

            if (!settings.EnableAccountReplication)
            {
                return;
            }

            IConfigurationReader configurationReader = new KeyVaultConfigurationReader();
            string cosmosDbAccountsKey = await configurationReader.GetConfigurationValueAsync(settings.CosmosDbAccountsKeyId);
            DocumentClient documentClient = new DocumentClient(settings.CosmosDbAccountsUri, cosmosDbAccountsKey, new ConnectionPolicy()
            {
                ConnectionMode = ConnectionMode.Direct,
                ConnectionProtocol = Protocol.Tcp,
                EnableEndpointDiscovery = true,
                MaxConnectionLimit = 250
            });
            await documentClient.OpenAsync();
            HttpClient httpClient = new HttpClient();
            CachedTokenClient tokenClient = new CachedTokenClient(settings, new KeyVaultCertificateResolver(settings, new MemoryConfigurationCache("certificates")));
            IRepository<RoleAssignments> roleAssignmentsRepository = await DocumentsRoleAssignmentsRepository.BuildAsync(settings, documentClient);
            IRoleAssignments roleAssignments = new RoleAssignments(httpClient, settings, tokenClient, roleAssignmentsRepository);

            while (!cancellationToken.IsCancellationRequested)
            {
                try
                {
                    // create a lease based on a read and write regions. Since we can guarantee 1 process per region.
                    Agent.Instance.LogEvent(EventType.Information, $"Region {settings.CurrentRegion} currently subscribing on read {documentClient.ReadEndpoint} and {documentClient.WriteEndpoint}.");
                    if (documentClient.ReadEndpoint == documentClient.WriteEndpoint)
                    {
                        IQueueRepository<Subscriptions> queueRepository = new DocumentSubscriptionsRepository(documentClient, settings);
                        Subscriptions subscriptions = new Subscriptions(queueRepository, roleAssignments);
                        SubscriptionQueueObserver subscriptionMonitor = new SubscriptionQueueObserver(subscriptions);
                        await subscriptionMonitor.RunAsync(cancellationToken);                        
                    }

                    await Task.Delay(TimeSpan.FromSeconds(30), cancellationToken);
                }
                catch (TaskCanceledException)
                {
                    // expected when cancellation requested.
                    Agent.Instance.LogEvent(EventType.Information, "Permissions Job cancellation requested.");
                    break;
                }
                catch (Exception ex)
                {
                    Agent.Instance.LogException(ex, "permissionsjob", "runasync");
                }                
            }
        }
    }
}
