﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderPermissions
{
    using System;
    using System.Collections.Generic;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ResourceProviderPermissions.Authorization;

    public class Subscriptions : ISubscriptions
    {
        private readonly IQueueRepository<Subscriptions> repository;
        private readonly IRoleAssignments roleAssignments;

        public Subscriptions(IQueueRepository<Subscriptions> repository, IRoleAssignments roleAssignments)
        {
            this.repository = repository;
            this.roleAssignments = roleAssignments;
            Items = new List<Subscription>();
        }

        public List<Subscription> Items { get; private set; }

        public async Task LoadAsync(CancellationToken cancellationToken)
        {
            await repository.PullAsync(this, cancellationToken);
        }

        public async Task PruneStaleAsync(CancellationToken cancellationToken)
        {
            List<Subscription> PrunedItems = new List<Subscription>(Items.Count);
            foreach (Subscription subscription in Items)
            {
                if (cancellationToken.IsCancellationRequested)
                {
                    return;
                }

                if (subscription.State == SubscriptionStateType.Suspended ||
                    subscription.State == SubscriptionStateType.Unregistered ||
                    subscription.State == SubscriptionStateType.Deleted)
                {
                    Agent.Instance.LogEvent(EventType.Information, $"ResourceProviderPermissions: deleting subscription with id: {subscription.Id} state {subscription.State}");
                    await repository.DeleteAsync(subscription.Id);                    
                }
                else
                {
                    PrunedItems.Add(subscription);
                }
            }

            Items.Clear();
            Items.AddRange(PrunedItems);
        }

        public async Task UpdatePermissionsAsync(CancellationToken cancellationToken)
        {
            if (Items.Count == 0)
            {
                return;
            }

            foreach (Subscription subscription in Items)
            {
                if (cancellationToken.IsCancellationRequested)
                {
                    return;
                }

                Agent.Instance.LogEvent(EventType.Verbose, $"ResourceProviderPermissions: updating subscription queue for id: {subscription.Id}");

                //// todo: call pas-rp via ARM S2S to download the role definitions for the subscription.
                //// update the subscription and save the definitions and assignments afterwards
                await roleAssignments.LoadAsync(subscription, cancellationToken).ConfigureAwait(false);

                if (roleAssignments.IsLoaded)
                {
                    await roleAssignments.SaveAsync().ConfigureAwait(false);
                    subscription.EnqueueTimeUtc = DateTime.UtcNow;
                }
                
                Agent.Instance.LogEvent(EventType.Verbose, $"ResourceProviderPermissions: updated subscription queue for id: {subscription.Id}");
            }
            
            await repository.UpsertAsync(this).ConfigureAwait(false);            
        }
    }
}
