﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderUnitTests
{
    using System;
    using System.Collections.Generic;
    using System.Dynamic;
    using System.Linq;
    using System.Net;
    using System.Net.Http;        
    using System.Threading.Tasks;
    using System.Web.Http;
    using System.Web.Http.Routing;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ResourceProviderUnitTests.Mocks;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Controllers;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ExternalContracts;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;
    using Moq;
    using Xunit;
    using Account = ResourceProviderWebAPI.Account;
    using ExternalAccount = ResourceProviderWebAPI.ExternalContracts.Account;

    public class AccountTests : IDisposable
    {
        private const string RpApiVersion = "2018-05-01";
        private readonly HttpConfiguration configuration;
        private bool disposedValue = false; // To detect redundant calls

        public AccountTests()
        {
            configuration = new HttpConfiguration();
            configuration.Routes.Add("PutAccount", new HttpRoute(@"subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{providerName:regex(^Microsoft\.Maps$)}/accounts/{resourceName}", new HttpRouteValueDictionary(
                new { id = RouteParameter.Optional }),
                constraints: null));
            configuration.Routes.Add("MoveResources", new HttpRoute(@"subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/moveResources", new HttpRouteValueDictionary(
                new { id = RouteParameter.Optional }),
                constraints: null));
        }

        [Fact(DisplayName = "AccountController_PUT_CreateAccount")]
        public async Task AccountController_PUT_CreateAccount()
        {
            var recordStorage = new MockAccountRecordStorage();
            var apiManagementStorage = new MockApiManagementStorage();

            var apiManagementStorageFactoryMock = new Mock<IApiManagementStorageFactory>();
            apiManagementStorageFactoryMock.Setup(factory => factory.GetApiManagementStorage(It.IsAny<SkuName>())).Returns(apiManagementStorage);

            var accountController = new AccountController(recordStorage, apiManagementStorageFactoryMock.Object);
            SetupController(accountController);

            var response = await accountController.Put("1234", "rg", "microsoft.maps", "name", new ExternalAccount()
            {
                Location = "global",
                Sku = new SkuType() { Name = SkuName.S0, Tier = "Standard" },
                Name = "name"                
            }, RpApiVersion);
            Assert.NotNull(response);

            // sniff the storage and make sure the account records align with the apim records
            var acctRecords = recordStorage.GetRecords();
            var acctRecord = acctRecords.SingleOrDefault(e => e.SubscriptionId == "1234" && e.ResourceGroupName == "rg" && e.ResourceName == "name");
            Assert.NotNull(acctRecord);
            var retrievedAcct = new Account(acctRecord);

            var apimRecords = apiManagementStorage.GetRecords();
            var apimRecord = apimRecords.SingleOrDefault(e => e.ApimSubscriptionId == retrievedAcct.ApiManagementId.SubscriptionId);
            Assert.NotNull(apimRecord);
            Assert.True(apimRecord.AccountResourceGroupName == retrievedAcct.Id.ResourceGroupName);
            Assert.True(apimRecord.AccountResourceName == retrievedAcct.Id.ResourceName);
        }

        [Fact(DisplayName = "AccountController_PUTCaseInsensitive_CreateAccount")]
        public async Task AccountController_PUTCaseInsensitive_CreateAccount()
        {
            var recordStorage = new MockAccountRecordStorage();
            var apiManagementStorage = new MockApiManagementStorage();

            var apiManagementStorageFactoryMock = new Mock<IApiManagementStorageFactory>();
            apiManagementStorageFactoryMock.Setup(factory => factory.GetApiManagementStorage(It.IsAny<SkuName>())).Returns(apiManagementStorage);

            var accountController = new AccountController(recordStorage, apiManagementStorageFactoryMock.Object);
            SetupController(accountController);
            var response = await accountController.Put("1234", "resourceGroupName", "microsoft.maps", "resourceName", new ExternalAccount()
            {
                Location = "global",
                Sku = new SkuType() { Name = SkuName.S0, Tier = "Standard" },
                Name = "resourceName"
            }, RpApiVersion);

            Assert.NotNull(response);

            response = await accountController.Get("1234", "ResourceGROUPNAME", "microsoft.maps", "resourceNAME", RpApiVersion);
            Assert.NotNull(response);

            var result = await response.Content.ReadAsAsync<ExternalAccount>();

            // the result should still have the original casing
            Assert.Contains("1234", result.Id);
            Assert.Contains("resourceGroupName", result.Id);
            Assert.Contains("resourceName", result.Id);
        }

        [Fact(DisplayName = "ExternalAccount_ConvertTo_AccoundRecord")]
        public void ExternalAccount_ConvertTo_AccoundRecord()
        {
            dynamic tags = new ExpandoObject();
            tags.Foo = "bar";
            tags.SomeTag = "SomeValue";
            var externalAccount = new ExternalAccount()
                    {
                        Location = "global",
                        Sku = new SkuType() { Name = SkuName.S0, },
                        Tags = tags
                    };

            // convert to internal account
            // to simulate what really happens, create an apimanagement identity for this
            // (this would not come from external account representation, so must be created here)
            var acct = new Account(externalAccount, "1234", "resourceGroupName", "resourceName")
                           {
                               ApiManagementId =
                                   new ApiManagementIdentity("ApimSubscriptionId")
                           };

            // convert to record
            var record = new AccountRecord(acct, new ApiManagementKeys("PrimaryKey", "SecondaryKey", DateTime.UtcNow, DateTime.UtcNow));

            // convert back to internal
            var retrievedAcct = new Account(record);

            // convert to external
            var retrievedExternalAccount = new ExternalAccount(retrievedAcct, ResourceProviderWebApiService.ResourceProviderName);
        }

        [Fact(DisplayName = "MoveController_MoveAccount_NewRg")]
        public async Task MoveController_MoveAccount_NewRg()
        {
            var recordStorage = new MockAccountRecordStorage();
            var apiManagementStorage = new MockApiManagementStorage();

            var apiManagementStorageFactoryMock = new Mock<IApiManagementStorageFactory>();
            apiManagementStorageFactoryMock.Setup(factory => factory.GetApiManagementStorage(It.IsAny<SkuName>())).Returns(apiManagementStorage);

            var accountController = new AccountController(recordStorage, apiManagementStorageFactoryMock.Object);
            SetupController(accountController);

            var response = await accountController.Put("1234", "resourceGroupName", "microsoft.maps", "name1", new ExternalAccount()
            {
                Location = "global",
                Sku = new SkuType() { Name = SkuName.S0, Tier = "Standard" },
                Name = "name1"
            }, RpApiVersion);
            Assert.NotNull(response);
            var result = await response.Content.ReadAsAsync<ExternalAccount>();

            var moveController = new AccountsMoveController(recordStorage, apiManagementStorageFactoryMock.Object);
            SetupController(moveController);

            response = await moveController.Post("1234", "resourceGroupName", new MoveRequest()
                {
                    TargetResourceGroup = "/subscriptions/1234/resourcegroups/destRg",
                    Resources = new List<string>() { result.Id }
                },
                RpApiVersion);

            Assert.NotNull(response);
            Assert.Equal(HttpStatusCode.OK, response.StatusCode);
            response = await accountController.Get("1234", "destRg", "microsoft.maps", "name1", RpApiVersion);
            Assert.NotNull(response);
            Assert.Equal(HttpStatusCode.OK, response.StatusCode);
        }

        [Theory]
        [InlineData("sub", "rg", null)]
        [InlineData("sub", "rg", "")]
        [InlineData("sub", "rg", "Primary")]
        [InlineData("sub", "rg", "secondary")]
        [InlineData("sub", "rg", "invalid")]
        public async Task TestAccountsKeyRegeneration(string subscriptionId, string resourceGroupName, string keyType)
        {
            const string PrimaryKey = "primary";
            const string SecondaryKey = "secondary";

            const string ProviderName = "Microsoft.Maps";
            const string AccountName = "name1";

            // Setup
            var recordStorage = new MockAccountRecordStorage();
            var apiManagementStorage = new MockApiManagementStorage();

            var apiManagementStorageFactoryMock = new Mock<IApiManagementStorageFactory>();
            apiManagementStorageFactoryMock.Setup(factory => factory.GetApiManagementStorage(It.IsAny<SkuName>())).Returns(apiManagementStorage);

            var accountController = new AccountController(recordStorage, apiManagementStorageFactoryMock.Object);
            SetupController(accountController);

            dynamic tags = new ExpandoObject();
            tags.Foo = "bar";

            var id1 = new AccountIdentifier(subscriptionId, resourceGroupName, AccountName);
            var acct1 = new Account(id1) { Location = Account.locationInfoDefault, Sku = Account.skuInfoDefault, Tags = tags };

            var response = await accountController.Put(subscriptionId, resourceGroupName, "microsoft.maps", AccountName, new ExternalAccount()
            {
                Location = "global",
                Sku = new SkuType() { Name = SkuName.S0, Tier = "Standard" },
                Name = AccountName,
                Tags = tags
            }, RpApiVersion);
            Assert.NotNull(response);

            // Test
            var listKeysController = new ListAccountKeyController(recordStorage, apiManagementStorageFactoryMock.Object);
            listKeysController.ConfigureForUnitTests();

            var listKeysResponse1 = await listKeysController.Post(subscriptionId, resourceGroupName, ProviderName, AccountName, RpApiVersion);
            Assert.True(listKeysResponse1.IsSuccessStatusCode);
            Keys keys1 = await listKeysResponse1.Content.ReadAsAsync<Keys>();
            Assert.NotNull(keys1);

            var regenerateKeysController = new RegenerateAccountKeyController(recordStorage, apiManagementStorageFactoryMock.Object);
            regenerateKeysController.ConfigureForUnitTests();

            var requestBody = new KeySpecification() { KeyType = keyType };
            var regenerateKeysResponse = await regenerateKeysController.RegenerateKey(
                subscriptionId,
                resourceGroupName,
                ProviderName,
                AccountName,
                requestBody,
                RpApiVersion);

            if (string.IsNullOrEmpty(keyType)
                || !keyType.Equals(PrimaryKey, StringComparison.InvariantCultureIgnoreCase)
                && !keyType.Equals(SecondaryKey, StringComparison.InvariantCultureIgnoreCase))
            {
                Assert.False(regenerateKeysResponse.IsSuccessStatusCode);
                return;
            }

            Assert.True(regenerateKeysResponse.IsSuccessStatusCode);
            Keys keysRegenerated = await regenerateKeysResponse.Content.ReadAsAsync<Keys>();
            Assert.NotNull(keysRegenerated);

            Assert.Equal(keys1.Id, keysRegenerated.Id);

            if (keyType.Equals(PrimaryKey, StringComparison.InvariantCultureIgnoreCase))
            {
                Assert.NotEqual(keys1.PrimaryKey, keysRegenerated.PrimaryKey);
                Assert.Equal(keys1.SecondaryKey, keysRegenerated.SecondaryKey);
            }
            else if (keyType.Equals(SecondaryKey, StringComparison.InvariantCultureIgnoreCase))
            {
                Assert.Equal(keys1.PrimaryKey, keysRegenerated.PrimaryKey);
                Assert.NotEqual(keys1.SecondaryKey, keysRegenerated.SecondaryKey);
            }

            var listKeysResponse2 = await listKeysController.Post(subscriptionId, resourceGroupName, ProviderName, AccountName, RpApiVersion);
            Assert.True(listKeysResponse2.IsSuccessStatusCode);
            Keys keys2 = await listKeysResponse2.Content.ReadAsAsync<Keys>();
            Assert.NotNull(keys2);

            Assert.Equal(keysRegenerated.Id, keys2.Id);
            Assert.Equal(keysRegenerated.PrimaryKey, keys2.PrimaryKey);
            Assert.Equal(keysRegenerated.SecondaryKey, keys2.SecondaryKey);
        }

        #region IDisposable Support       
        // This code added to correctly implement the disposable pattern.
        public void Dispose()
        {            
            Dispose(true);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    configuration.Dispose();
                }

                disposedValue = true;
            }
        }
        #endregion
        
        private void SetupController(ResourceProviderControllerBase controller)
        {
            controller.SetAuthorized();
            controller.SetOperation(new Mock<IResourceProviderOperation>().Object);
            controller.Request = new HttpRequestMessage(HttpMethod.Put, "arbitrary");
            controller.Configuration = configuration;
        }
    }
}