﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace Microsoft.Azure.LocationServices.ResourceProviderUnitTests
{
    using System;
    using System.Dynamic;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Text;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ExternalContracts;

    using Newtonsoft.Json;

    using Xunit;

    using Account = ResourceProviderWebAPI.ExternalContracts.Account;

    public class ControllerTests : ControllerTestsBase
    {
        private const string ProviderName = "Microsoft.Maps";

        private const string ApiVersion = "2018-05-01";

        [Theory]
        [InlineData(SkuName.S0, "Standard")]
        public async Task CanListUpdateAndDeleteAccount(SkuName sku, string tier)
        {
            var subscriptionId = Guid.NewGuid().ToString();
            var resourceGroupName = "testRg";
            var resourceName = GetTestAccountName();

            var accountPath = FormatAccountPath(subscriptionId, resourceGroupName, resourceName);

            var requestBody = GetPutAccountRequestBody(sku);

            dynamic tags = new ExpandoObject();
            tags.foo = "bar";
            var update = new Account { Tags = tags };
            var updateBody = ToJsonContent(update);

            using (var client = this.CreateTestClient())
            {
                var putResponseMessage = await client.PutAsync(accountPath, requestBody);
                putResponseMessage.EnsureSuccessStatusCode();

                var getAllInSubscriptionResponseMessage = GetAccounts(client, subscriptionId);
                var accountsInSubscription = getAllInSubscriptionResponseMessage.EnsureSuccessStatusCode().GetContentValueAs<MultipleAccounts>();
                Assert.True(accountsInSubscription.Value.Count > 0);

                var getAllInResourceGroupResponseMessage = GetAccounts(client, subscriptionId, resourceGroupName);
                var accountsInResourceGroup = getAllInResourceGroupResponseMessage.EnsureSuccessStatusCode().GetContentValueAs<MultipleAccounts>();
                Assert.True(accountsInResourceGroup.Value.Count > 0);

                var getAccountResponseMessage = await client.GetAsync(accountPath);
                var retrievedAccount = getAccountResponseMessage.EnsureSuccessStatusCode().GetContentValueAs<Account>();
                Assert.Equal("global", retrievedAccount.Location);
                Assert.Equal(sku, retrievedAccount.Sku.Name);
                Assert.Equal(tier, retrievedAccount.Sku.Tier, ignoreCase: true);

                var patchAccountResponseMessage = await client.PatchAsync(accountPath, updateBody);
                patchAccountResponseMessage.EnsureSuccessStatusCode();

                var getAccountPostUpdateResponseMessage = client.Get(accountPath);
                var retrievedAccountPostUpdate = getAccountPostUpdateResponseMessage.EnsureSuccessStatusCode().GetContentValueAs<Account>();
                Assert.Equal("global", retrievedAccountPostUpdate.Location);
                Assert.Equal(retrievedAccountPostUpdate.Sku.Name, sku);
                Assert.Equal(retrievedAccountPostUpdate.Sku.Tier, tier, ignoreCase: true);
                Assert.NotEqual(retrievedAccount.Tags, retrievedAccountPostUpdate.Tags);

                var deleteResponseMessage = await client.DeleteAsync(accountPath);
                deleteResponseMessage.EnsureSuccessStatusCode();
            }
        }

        [Fact]
        public void TestBadResponse()
        {
            var subscriptionId = Guid.NewGuid().ToString();
            var resourceGroupName = "testRg";
            var resourceName = GetTestAccountName();
            var accountPath = FormatAccountPath(subscriptionId, resourceGroupName, resourceName);

            var listKeysPath =
                $"/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{ProviderName}/accounts/{resourceName}/listKeys?api-version={ApiVersion}";

            var regenerateKeyPath =
                $"/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{ProviderName}/accounts/{resourceName}/regenerateKey?api-version={ApiVersion}";

            using (var client = this.CreateTestClient())
            {
                var getMessage = client.Get(accountPath);
                Assert.False(getMessage.IsSuccessStatusCode);

                var listKeysMessage = client.Post(listKeysPath);
                Assert.False(listKeysMessage.IsSuccessStatusCode);

                var regenBadPath = client.Post(regenerateKeyPath, ToJsonContent(new KeySpecification { KeyType = "primary" }));
                Assert.False(regenBadPath.IsSuccessStatusCode);
                regenBadPath = client.Post(regenerateKeyPath, new StringContent(string.Empty));
                Assert.False(regenBadPath.IsSuccessStatusCode);

                var deleteMessage = client.Delete(accountPath);
                Assert.True(deleteMessage.StatusCode.Equals(HttpStatusCode.NoContent));
            }
        }

        [Fact]
        public void CanListAndRegenerateKeys()
        {
            var subscriptionId = Guid.NewGuid().ToString();
            var resourceGroupName = "testRg";
            var resourceName = GetTestAccountName();

            var accountPath = FormatAccountPath(subscriptionId, resourceGroupName, resourceName);

            var listKeysPath =
                $"/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{ProviderName}/accounts/{resourceName}/listKeys?api-version={ApiVersion}";

            var regenerateKeyPath =
                $"/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{ProviderName}/accounts/{resourceName}/regenerateKey?api-version={ApiVersion}";

            Func<HttpClient, string> listKeys = (client) => client.Post(listKeysPath).EnsureSuccessStatusCode().Content.ReadAsStringAsync().GetAwaiter().GetResult();

            using (var client = this.CreateTestClient())
            {
                client.Put(accountPath, GetPutAccountRequestBody()).EnsureSuccessStatusCode();

                var listKeysResponse = listKeys(client);
                var regeneratePrimaryKeyResponse =
                    client.Post(regenerateKeyPath, ToJsonContent(new KeySpecification { KeyType = "primary" })).EnsureSuccessStatusCode().Content.ReadAsStringAsync().GetAwaiter().GetResult();
                Assert.NotEqual(listKeysResponse, regeneratePrimaryKeyResponse);

                var listKeysResponse2 = listKeys(client);
                Assert.Equal(regeneratePrimaryKeyResponse, listKeysResponse2);

                var regenerateSecondaryKeyResponse = 
                    client.Post(regenerateKeyPath, ToJsonContent(new KeySpecification { KeyType = "secondary" })).EnsureSuccessStatusCode().Content.ReadAsStringAsync().GetAwaiter().GetResult();
                Assert.NotEqual(listKeysResponse2, regenerateSecondaryKeyResponse);

                var listKeysResponse3 = listKeys(client);
                Assert.Equal(regenerateSecondaryKeyResponse, listKeysResponse3);
            }
        }

        [Theory]
        [InlineData("sub1", "source", "sub2", "destination", 1)]
        [InlineData("sub", "source", "sub", "destination", 2)]
        public void CanMoveAccounts(string sourceSubscription, string sourceResourceGroup, string destinationSubscription, string destinationResourceGroup, int numberOfAccounts)
        {
            var testGuid = Guid.NewGuid().ToString();
            sourceSubscription += testGuid;
            destinationSubscription += testGuid;
            sourceResourceGroup += testGuid;
            destinationResourceGroup += testGuid;

            var movePath = $"/subscriptions/{sourceSubscription}/resourceGroups/{sourceResourceGroup}/moveResources?api-version={ApiVersion}";

            using (var client = this.CreateTestClient())
            {
                var accountIds = new string[numberOfAccounts];
                for (var i = 0; i < numberOfAccounts; i++)
                {
                    var accountPath = FormatAccountPath(sourceSubscription, sourceResourceGroup, GetTestAccountName());
                    var account = client.Put(accountPath, GetPutAccountRequestBody()).EnsureSuccessStatusCode().GetContentValueAs<Account>();
                    accountIds[i] = account.Id;
                }

                var moveBody = ToJsonContent(
                    new MoveRequest()
                        {
                            Resources = accountIds.ToList(),
                            TargetResourceGroup = $"/subscriptions/{destinationSubscription}/resourceGroups/{destinationResourceGroup}"
                        });
                client.Post(movePath, moveBody).EnsureSuccessStatusCode();

                var accountsInSourceSub = GetAccounts(client, sourceSubscription).EnsureSuccessStatusCode().GetContentValueAs<MultipleAccounts>();
                var accountsInSourceRg = GetAccounts(client, sourceSubscription, sourceResourceGroup).EnsureSuccessStatusCode().GetContentValueAs<MultipleAccounts>();

                var accountsInDestinationSub = GetAccounts(client, destinationSubscription).EnsureSuccessStatusCode().GetContentValueAs<MultipleAccounts>();
                var accountsInDestinationRg = GetAccounts(client, destinationSubscription, destinationResourceGroup).EnsureSuccessStatusCode().GetContentValueAs<MultipleAccounts>();

                Assert.Empty(accountsInSourceRg.Value);
                Assert.Equal(numberOfAccounts, accountsInDestinationRg.Value.Count);

                Assert.Equal(numberOfAccounts, accountsInDestinationSub.Value.Count);
                if (!string.Equals(sourceSubscription, destinationSubscription))
                {
                    Assert.Empty(accountsInSourceSub.Value);
                }
            }
        }

        [Fact]
        public void CanListOperations()
        {
            var operationsPath = $"/providers/{ProviderName}/operations?api-version={ApiVersion}";
            using (var client = this.CreateTestClient())
            {
                var operations = client.Get(operationsPath).EnsureSuccessStatusCode();
            }
        }

        [Fact(DisplayName = "RP_Put_Subscription")]
        public async Task CanPutSubscription()
        {
            var subscriptionsPath = $"/subscriptions/{Guid.NewGuid()}?api-version=2.0";
            var subscriptionRequestBody = ToJsonContent(new Subscription { State = "Registered", Properties = new ExpandoObject() });
            using (var client = this.CreateTestClient())
            {
                var response = await client.PutAsync(subscriptionsPath, subscriptionRequestBody);
                Assert.True(response.IsSuccessStatusCode, $"statusCode is {response.StatusCode}, {await response.Content.ReadAsStringAsync()}");
            }
        }

        private static string GetTestAccountName()
        {
            return "testAccount" + Guid.NewGuid();
        }

        private static HttpResponseMessage GetAccounts(HttpClient client, string subscriptionId)
        {
            var subscriptionPath = $"/subscriptions/{subscriptionId}/providers/{ProviderName}/accounts?api-version={ApiVersion}";
            return client.Get(subscriptionPath);
        }

        private static HttpResponseMessage GetAccounts(HttpClient client, string subscriptionId, string resourceGroupName)
        {
            var resourceGroupPath = $"/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{ProviderName}/accounts?api-version={ApiVersion}";
            return client.Get(resourceGroupPath);
        }

        private static StringContent ToJsonContent(object value)
        {
            return new StringContent(JsonConvert.SerializeObject(value), Encoding.UTF8, "application/json");
        }

        private static string FormatAccountPath(string subscriptionId, string resourceGroupName, string resourceName) => $"/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{ProviderName}/accounts/{resourceName}?api-version={ApiVersion}";

        private static StringContent GetPutAccountRequestBody(SkuName sku = SkuName.S0)
        {
            var account = new Account { Location = "global", Sku = new SkuType() { Name = sku } };
            return ToJsonContent(account);
        }
    }
}