﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 
namespace Microsoft.Azure.LocationServices.ResourceProviderUnitTests
{
    using System.Collections.Generic;
    using System.Net.Http;
    using System.Net.Http.Formatting;
    using System.Web.Http;

    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.ResourceProviderUnitTests.Mocks;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;
    using Microsoft.Owin.Testing;

    using Moq;

    using Newtonsoft.Json;
    using Newtonsoft.Json.Converters;
    using Newtonsoft.Json.Serialization;

    using global::Owin;

    using Unity;

    public class ControllerTestsBase
    {
        private static readonly TestServer TestServer;

        static ControllerTestsBase()
        {
            TestServer = TestServer.Create(
                app =>
                    {
                        var config = new HttpConfiguration();
                        config.MapHttpAttributeRoutes();

                        ConfigureJsonFormatters(config);
                        ConfigureFilters(config);
                        Register(config);

                        app.UseWebApi(config);
                    });
        }

        protected HttpClient CreateTestClient() => new HttpClient(TestServer.Handler)
                                                       {
                                                           BaseAddress = TestServer.BaseAddress
                                                       };

        private static void ConfigureFilters(HttpConfiguration config)
        {
            config.Filters.Add(new ServiceRequestActionFilter(new MockResourceProviderOperationFactory()));
            config.Filters.Add(new ValidModelsOnlyFilter());
            config.Filters.Add(new CheckClientCertificateFilter(new MockClientCertificateResolverFactory()));
        }

        private static void ConfigureJsonFormatters(HttpConfiguration config)
        {
            var defaultSettings = new JsonSerializerSettings
                                      {
                                          Formatting = Formatting.Indented,
                                          ContractResolver = new CamelCasePropertyNamesContractResolver(),
                                          DefaultValueHandling = DefaultValueHandling.Ignore,
                                          Converters = new List<JsonConverter> { new StringEnumConverter { CamelCaseText = true }, }
                                      };

            JsonConvert.DefaultSettings = () => defaultSettings;

            config.Formatters.Clear();
            config.Formatters.Add(new JsonMediaTypeFormatter());
            config.Formatters.JsonFormatter.SerializerSettings = defaultSettings;
        }

        private static void Register(HttpConfiguration config)
        {
            var container = new UnityContainer();

            var apiManagementStorage = new MockApiManagementStorage();
            var apiManagementStorageFactoryMock = new Mock<IApiManagementStorageFactory>();
            apiManagementStorageFactoryMock.Setup(f => f.GetApiManagementStorage(It.IsAny<SkuName>())).Returns(apiManagementStorage);

            container.RegisterInstance<IAccountRecordStorage>(new MockAccountRecordStorage());
            container.RegisterInstance<ISubscriptionRecordStorage>(new MockSubscriptionRecordStorage());
            container.RegisterInstance<IDocumentDbStorage>(new MockDocumentDbStorage());
            container.RegisterInstance<IApiManagementStorageFactory>(apiManagementStorageFactoryMock.Object);

            config.DependencyResolver = new UnityResolver(container);
        }
    }
}