﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderUnitTests.Mocks
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;

    internal class MockAccountRecordStorage : IAccountRecordStorage
    {
        // outer dictionary is by primary key
        private readonly Dictionary<string, Dictionary<string, AccountRecord>> dictionary =
            new Dictionary<string, Dictionary<string, AccountRecord>>();

        public async Task DeleteAccountRecord(AccountIdentifier id)
        {
            var record = await this.GetAccountRecord(id);
            if (record != null)
            {
                record.Deleted = true;
            }
        }

        public Task<AccountRecord> GetAccountRecord(AccountIdentifier id)
        {
            Validation.Requires.NotNull(id, nameof(id));

            return Task.FromResult(this.GetAccountRecordsByResourceGroup(id.SubscriptionId, id.ResourceGroupName).SingleOrDefault(
                r => string.Equals(r.ResourceName, id.ResourceName, StringComparison.InvariantCultureIgnoreCase)));
        }

        public IEnumerable<AccountRecord> GetAccountRecordsByResourceGroup(string subscriptionId,
                                                                           string resourceGroupName,
                                                                           bool includeDeleted = false)
        {
            Validation.Requires.NotNull(subscriptionId, nameof(subscriptionId));
            Validation.Requires.NotNull(resourceGroupName, nameof(resourceGroupName));

            return this.GetAccountRecordsBySubscription(subscriptionId, includeDeleted)
                        .Where(r => string.Equals(r.ResourceGroupName, resourceGroupName, StringComparison.InvariantCultureIgnoreCase));
        }

        public IEnumerable<AccountRecord> GetAccountRecordsBySubscription(string subscriptionId,
                                                                          bool includeDeleted = false)
        {
            Validation.Requires.NotNull(subscriptionId, nameof(subscriptionId));

            var subDict = this.dictionary.SingleOrDefault(p => string.Equals(p.Key, subscriptionId, StringComparison.InvariantCultureIgnoreCase)).Value;

            if (subDict == null)
            {
                return Enumerable.Empty<AccountRecord>();
            }

            return subDict.Select(p => p.Value).Where(v => !v.Deleted || includeDeleted);
        }        

        public Task PutAccountRecord(AccountRecord record)
        {
            Validation.Requires.NotNull(record, nameof(record));

            Dictionary<string, AccountRecord> partition;
            if (!this.dictionary.TryGetValue(record.PartitionKey, out partition))
            {
                partition = new Dictionary<string, AccountRecord>();
                this.dictionary[record.PartitionKey] = partition;
            }

            // replace semantics
            partition[record.RowKey] = record;
            return Task.CompletedTask;
        }

        internal IEnumerable<AccountRecord> GetRecords()
        {
            return from partition in this.dictionary from record in partition.Value select record.Value;
        }

        internal void Reset()
        {
            this.dictionary.Clear();
        }
    }
}