﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderUnitTests.Mocks
{
    using System;
    using System.Collections.Generic;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement;

    internal class MockApiManagementStorage : IApiManagementStorage
    {
        private readonly Dictionary<string, MockApiManagementRecord> subscriptions =
            new Dictionary<string, MockApiManagementRecord>();

        public ApiManagementIdentity WriteServiceKeys(AccountIdentifier accountIdentifier, string apiManagementSubscriptionId, out ApiManagementKeys keys, ApiManagementKey primaryKey, ApiManagementKey secondaryKey)
        {
            // manufacture a new identity
            ApiManagementIdentity id = new ApiManagementIdentity(apiManagementSubscriptionId);
            
            keys = new ApiManagementKeys(primaryKey, secondaryKey);
            MockApiManagementRecord record = new MockApiManagementRecord(id.SubscriptionId,
                                                                         keys,
                                                                         accountIdentifier.SubscriptionId,
                                                                         accountIdentifier.ResourceGroupName,
                                                                         accountIdentifier.ResourceName);

            this.subscriptions.Add(id.SubscriptionId, record);

            return id;
        }

        public ApiManagementIdentity AddNewServiceKeys(AccountIdentifier accountIdentifier, out ApiManagementKeys keys)
        {
            var subscriptionId = Guid.NewGuid().ToString("N");
            var primaryKey = new ApiManagementKey(Guid.NewGuid().ToString("N"), DateTime.UtcNow);
            var secondaryKey = new ApiManagementKey(Guid.NewGuid().ToString("N"), DateTime.UtcNow);

            return this.WriteServiceKeys(accountIdentifier, subscriptionId, out keys, primaryKey, secondaryKey);
        }

        public void PatchUser(ApiManagementIdentity apimIdentity, AccountIdentifier accountIdentifier)
        {
            var apimRecord = this.subscriptions[apimIdentity.SubscriptionId];
            var newRecord = new MockApiManagementRecord(apimRecord.ApimSubscriptionId,
                                                        apimRecord.Keys,
                                                        accountIdentifier.SubscriptionId, 
                                                        accountIdentifier.ResourceGroupName,
                                                        accountIdentifier.ResourceName);
            this.subscriptions[apimIdentity.SubscriptionId] = newRecord;
        }

        public void DeleteUser(ApiManagementIdentity id)
        {
            this.subscriptions.Remove(id.SubscriptionId);
        }

        public ApiManagementKeys GetServiceKeys(ApiManagementIdentity id)
        {
            return this.subscriptions.TryGetValue(id.SubscriptionId, out var record) ? record.Keys : null;
        }

        public ApiManagementKey RegeneratePrimaryKey(ApiManagementIdentity id)
        {
            if (!this.subscriptions.TryGetValue(id.SubscriptionId, out var record))
            {
                return null;
            }

            var key = Guid.NewGuid().ToString("N");
            var newKeys = new ApiManagementKeys(key, record.Keys.SecondaryKey.Key, DateTime.UtcNow, DateTime.UtcNow);
            record.Keys = newKeys;

            return record.Keys.PrimaryKey;
        }

        public ApiManagementKey RegenerateSecondaryKey(ApiManagementIdentity id)
        {
            if (!this.subscriptions.TryGetValue(id.SubscriptionId, out var record))
            {
                return null;
            }

            var key = Guid.NewGuid().ToString("N");
            var newKeys = new ApiManagementKeys(record.Keys.PrimaryKey.Key, key, DateTime.UtcNow, DateTime.UtcNow);
            record.Keys = newKeys;

            return record.Keys.SecondaryKey;
        }

        public IEnumerable<MockApiManagementRecord> GetRecords()
        {
            return this.subscriptions.Values;
        }

        public void Reset()
        {
            this.subscriptions.Clear();
        }

        public bool UserExists(ApiManagementIdentity id)
        {
            return this.subscriptions.ContainsKey(id.SubscriptionId);
        }
    }

    internal class MockApiManagementRecord
    {
        public MockApiManagementRecord(string apimSubscriptionId,
                                       ApiManagementKeys keys,
                                       string accountSubscriptionId,
                                       string accountResourceGroupName,
                                       string accountResourceName)
        {
            this.Keys = keys;
            this.AccountSubscriptionId = accountSubscriptionId;
            this.AccountResourceGroupName = accountResourceGroupName;
            this.AccountResourceName = accountResourceName;
            this.ApimSubscriptionId = apimSubscriptionId;
        }

        public string ApimSubscriptionId { get; }

        public ApiManagementKeys Keys { get; set; }

        public string AccountSubscriptionId { get; }

        public string AccountResourceGroupName { get; }

        public string AccountResourceName { get; }
    }
}