﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderUnitTests
{
    using System;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.ResourceProviderPermissions;
    using Microsoft.Azure.LocationServices.ResourceProviderPermissions.Authorization;
    using Moq;
    using Xunit;

    public class SubscriptionsUnitTest
    {
        public SubscriptionsUnitTest()
        {
            if (Instrumentation.Agent.Instance == null)
            {
                var p = new Instrumentation.InitParams("DevTestIfx", "AzLocTest");
                p.SetLocalTracing(@"c:\ifxTrace");
                p.SetEtwProvider("DevTest", "DB25DB83-F4C7-4D76-9DBA-5D08A7DEC472");
                Instrumentation.Agent.CreateInstance(p);
            }
        }

        [Fact(DisplayName = "Subscriptions_LoadAsync")]
        [Trait("Category", "L0")]
        public async Task Subscriptions_LoadAsync()
        {
            Mock<IQueueRepository<Subscriptions>> mockRepository = new Mock<IQueueRepository<Subscriptions>>();
            Mock<IRoleAssignments> mockRoleAssignments = new Mock<IRoleAssignments>();
            Subscriptions subscriptions = new Subscriptions(mockRepository.Object, mockRoleAssignments.Object);
            mockRepository
                .Setup(r => r.PullAsync(It.Is<Subscriptions>(s => s == subscriptions), It.IsAny<CancellationToken>()))                
                .ReturnsAsync(subscriptions)
                .Callback<Subscriptions, CancellationToken>((subs, token) =>
                {
                    subs.Items.Add(new Subscription() { State = SubscriptionStateType.Registered });
                })
                .Verifiable();

            await subscriptions.LoadAsync(CancellationToken.None);

            Assert.Single(subscriptions.Items);
            Assert.Equal(SubscriptionStateType.Registered, subscriptions.Items[0].State);
            mockRepository.VerifyAll();
        }

        [Fact(DisplayName = "Subscriptions_UpdatePermissionsAsync")]
        [Trait("Category", "L0")]
        public async Task Subscriptions_UpdatePermissionsAsync()
        {
            DateTime createdTime = DateTime.UtcNow.AddMinutes(-5);
            Mock<IQueueRepository<Subscriptions>> mockRepository = new Mock<IQueueRepository<Subscriptions>>();
            Mock<IRoleAssignments> mockRoleAssignments = new Mock<IRoleAssignments>();
            Subscriptions subscriptions = new Subscriptions(mockRepository.Object, mockRoleAssignments.Object);
            string subId = Guid.NewGuid().ToString();
            subscriptions.Items.Add(new Subscription()
            {
                Id = subId,
                State = SubscriptionStateType.Registered,
                EnqueueTimeUtc = createdTime
            });
        
            mockRepository
               .Setup(r => r.UpsertAsync(It.Is<Subscriptions>(s => s == subscriptions)))
               .Returns(Task.CompletedTask)
               .Verifiable();

            mockRoleAssignments.Setup(r => r.LoadAsync(It.Is<Subscription>(s => s.Id == subId), It.IsAny<CancellationToken>()))
                .Returns(Task.CompletedTask)
                .Callback<Subscription, CancellationToken>((s, c) =>
                {
                    mockRoleAssignments.Setup(r => r.IsLoaded).Returns(true).Verifiable();
                })
                .Verifiable();

            mockRoleAssignments.Setup(r => r.SaveAsync())
                .Returns(Task.CompletedTask)
                .Verifiable();

            await subscriptions.UpdatePermissionsAsync(CancellationToken.None);

            Assert.Single(subscriptions.Items);
            Assert.Equal(SubscriptionStateType.Registered, subscriptions.Items[0].State);
            Assert.InRange(subscriptions.Items[0].EnqueueTimeUtc, DateTime.UtcNow.AddMilliseconds(-300), DateTime.UtcNow.AddMilliseconds(4));
            mockRepository.VerifyAll();
            mockRoleAssignments.VerifyAll();
        }

        [Fact(DisplayName = "Subscriptions_PruneStaleAsync")]
        [Trait("Category", "L0")]
        public async Task Subscriptions_PruneStaleAsync()
        {            
            Mock<IQueueRepository<Subscriptions>> mockRepository = new Mock<IQueueRepository<Subscriptions>>();
            Mock<IRoleAssignments> mockRoleAssignments = new Mock<IRoleAssignments>();
            Subscriptions subscriptions = new Subscriptions(mockRepository.Object, mockRoleAssignments.Object);

            subscriptions.Items.Add(new Subscription()
            {
                Id = Guid.NewGuid().ToString(),
                State = SubscriptionStateType.Deleted,
                EnqueueTimeUtc = DateTime.UtcNow.AddMinutes(-5)
            });
            subscriptions.Items.Add(new Subscription()
            {
                Id = Guid.NewGuid().ToString(),
                State = SubscriptionStateType.Registered,
                EnqueueTimeUtc = DateTime.UtcNow.AddMinutes(-5)
            });
            mockRepository
               .Setup(r => r.DeleteAsync(It.Is<string>(id => id == subscriptions.Items[0].Id)))
               .Returns(Task.CompletedTask)
               .Verifiable();

            await subscriptions.PruneStaleAsync(CancellationToken.None);

            Assert.Single(subscriptions.Items);
            mockRepository.VerifyAll();
        }
    }
}
