﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderUnitTests
{
    using System.Dynamic;
    using System.Linq;
    using System.Net.Http;
    using System.Web.Http;

    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Controllers;

    using Moq;

    using Newtonsoft.Json;

    internal static class TestExtensions
    {
        /// <summary>
        /// Gets the full Azure resource id for an account identifier
        /// </summary>
        internal static string ToResourceId(this AccountIdentifier accountIdentifier)
        {
            return ResourceProviderWebAPI.ExternalContracts.Account.BuildExternalId(accountIdentifier, ResourceProviderWebApiService.ResourceProviderName);
        }

        /// <summary>
        /// Checks that the account properties are equal
        /// </summary>
        /// <param name="account1">First account to compare</param>
        /// <param name="account2">Second account to compare</param>
        /// <returns>A value indicating whether these accounts have the same property values</returns>
        internal static bool AccountPropertiesAreEqual(this Account account1, Account account2)
        {
            if (IsEitherXorNull(account1.Tags, account2.Tags))
            {
                return false;
            }

            var tagsAreEqual = true;
            if (account1.Tags != null && account2.Tags != null)
            {
                tagsAreEqual = TagsAreEqual(account1.Tags, account2.Tags);
            }

            return tagsAreEqual
                && account1.ApiManagementId.SubscriptionId.Equals(account2.ApiManagementId.SubscriptionId)
                && account1.Location.Equals(account2.Location)
                && account1.Sku.Equals(account2.Sku);
        }

        /// <summary>
        /// Configure the controller for unit testing
        /// </summary>
        internal static void ConfigureForUnitTests(this ResourceProviderControllerBase controller)
        {
            var operationMock = new Mock<IResourceProviderOperation>();
            var operation = operationMock.Object;
            controller.SetOperation(operation);

            var mockRequest = new HttpRequestMessage();
            mockRequest.SetConfiguration(new HttpConfiguration());
            controller.Request = mockRequest;

            controller.SetAuthorized();
        }

        internal static T GetContentValueAs<T>(this HttpResponseMessage httpResponseMessage)
        {
            var serializedContent = httpResponseMessage.Content.ReadAsStringAsync().GetAwaiter().GetResult();
            return JsonConvert.DeserializeObject<T>(serializedContent);
        }

        private static bool IsEitherXorNull(object obj1, object obj2)
        {
            return ((obj1 != null && obj2 == null)
                || (obj1 == null && obj2 != null));
        }

        private static bool TagsAreEqual(ExpandoObject obj1, ExpandoObject obj2)
        {
            return obj1.Intersect(obj2).Count() == obj1.Union(obj2).Count();
        }
    }
}
