﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ARM
{
    using System;
    using System.Linq;
    using System.Net;
    using System.Security.Cryptography.X509Certificates;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Newtonsoft.Json;
    using Validation;

    internal class ArmCertificateReader
    {
        private const string ArmUrlConfigKey = "mARMClientAuthCertificateLocation";
        
        private readonly IConfigurationReader _configurationReader;
        private readonly string _armUrlKey;

        public ArmCertificateReader(IBootstrapConfiguration bootstrapConfiguration, IConfigurationReader configurationReader)
        {
            Requires.NotNull(bootstrapConfiguration, nameof(bootstrapConfiguration));
            Requires.NotNull(configurationReader, nameof(configurationReader));
            _configurationReader = configurationReader;
            _armUrlKey = bootstrapConfiguration.GetValue(ArmUrlConfigKey);
        }        

        /// <summary>
        ///     Asks ARM for client certificates.
        ///     Returns null in case of any error.   Does not throw exceptions.
        /// </summary>
        /// <returns></returns>
        public async Task<X509Certificate2Collection> DownloadCertificates()
        {
            X509Certificate2Collection certificates = null;
            try
            {
                using (var client = new WebClient())
                {
                    client.Headers.Add("Content-Type", @"application/json;charset=UTF-8");
                    string armUrl = await _configurationReader.GetConfigurationValueAsync(_armUrlKey);
                    string result = await client.DownloadStringTaskAsync(armUrl);

                    ArmCertificates armCertificates = JsonConvert.DeserializeObject<ArmCertificates>(result);

                    Assumes.NotNull(armCertificates);
                    Assumes.NotNullOrEmpty(armCertificates.ClientCertificates);

                    // for each cert, convert the certdata (base64 string) into an x509 and wrap it in a Collection
                    certificates = new X509Certificate2Collection(armCertificates.ClientCertificates.Select(
                                                                                      e =>
                                                                                          new X509Certificate2(
                                                                                              Convert.FromBase64String(
                                                                                                  e.CertData)))
                                                                                                  .ToArray());
                }
            }
            catch (Exception)
            {
                // swallow everything; the caller will retry
                // TODO: log the error
            }

            return certificates;
        }
    }
}