﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI
{
    using System;
    using System.Collections.Generic;
    using System.Dynamic;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;

    /// <summary>
    ///     The internal representation of an Account.
    ///     Can be converted to the External representation and persisted in the AccountRecordStorage
    /// </summary>
    public class Account
    {
        // mapping between string (which comes externally, and is stored in account database) to sku information
        // (note:  normalize to lowercase for lookup)
        private static readonly Dictionary<string, SkuInfo> skuInfoMap =
            new Dictionary<string, SkuInfo>(StringComparer.OrdinalIgnoreCase)
            {
                { SkuName.S0.ToString(), new SkuInfo() { Name = SkuName.S0, Tier = SkuTierType.Standard } }
            };

        // mapping between string (which comes externally, and is stored in account database) and location information
        // (note:  normalize to lowercase for lookup)
        private static readonly Dictionary<string, LocationInfo> locationInfoMap =
            new Dictionary<string, LocationInfo>
            {
                {
                    "global", new LocationInfo() { LocationString = "global", Location = LocationType.Global }
                }
            };

        /// <summary>
        /// Creates an internal Account from an external representation.
        /// Throws ArgumentException on validation failures.
        /// </summary>
        /// <param name="accountExternal"></param>
        /// <param name="subscriptionId"></param>
        /// <param name="resourceGroupName"></param>
        /// <param name="resourceName"></param>
        /// <exception cref="System.ArgumentException">Throws on validation errors, such as missing values, or a bad Location or Sku.</exception>
        public Account(ExternalContracts.Account accountExternal,
                       string subscriptionId,
                       string resourceGroupName,
                       string resourceName)
        {
            Validation.Requires.NotNull(accountExternal, "account");
            Validation.Requires.NotNull(accountExternal.Location, "account.location");
            Validation.Requires.NotNull(accountExternal.Sku, "account.sku");

            Id = new AccountIdentifier(subscriptionId, resourceGroupName, resourceName);

            Location = LocationFromExternal(accountExternal);
            Sku = SkuTypeFromExternal(accountExternal);

            Tags = accountExternal.Tags;
        }

        /// <summary>
        /// Special ctor that partially constructs an account for an update.   Assumes only certain fields will be used.
        /// Throws ArgumentException on validation failures
        /// </summary>
        /// <param name="accountExternal"></param>
        /// <exception cref="System.ArgumentException">Throws on validation errors, such as missing values, or a bad Location or Sku.</exception>
        public Account(ExternalContracts.Account accountExternal)
        {
            Tags = accountExternal.Tags;

            // assume if sku is specified, then all fields are
            if (accountExternal.Sku != null)
            {
                Sku = SkuTypeFromExternal(accountExternal);
            }
        }

        public Account(AccountRecord record)
        {
            Validation.Requires.NotNull(record, nameof(record));

            Id = new AccountIdentifier(record.SubscriptionId, record.ResourceGroupName, record.ResourceName);
            ClientId = record.ClientId;
            Location = LocationFromRecord(record);
            Sku = SkuTypeFromRecord(record);

            Tags = record.GetTags();

            ApiManagementId = record.GetApiManagementIdentity();
        }

        public Account(AccountRecord record, AccountIdentifier newId)
            : this(record)
        {
            this.Id = newId;
        }

        /// <summary>
        ///     internal ctor used for tests
        /// </summary>
        internal Account(AccountIdentifier id)
        {
            this.Id = id;
        }

        /// <summary>
        /// Sku tiers (which are determined from sku name)
        /// </summary>
        public enum SkuTierType
        {
            Standard = 0,

            Premium
        }

        public enum LocationType
        {
            Global = 0
        }

        public string ClientId { get; set; }

        public AccountIdentifier Id { get; private set; }

        public SkuInfo Sku { get; set; }

        public LocationInfo Location { get; set; }

        public ExpandoObject Tags { get; set; }

        public ApiManagementIdentity ApiManagementId { get; set; }

        /// <summary>
        /// A nice default sku.   This is used in only a couple of places:
        /// - for testing
        /// - for backwards compat when retrieving an unrecognized value from storage
        /// </summary>
        internal static SkuInfo skuInfoDefault => skuInfoMap[SkuName.S0.ToString()];

        /// <summary>
        /// A nice default location.   This is used in only a couple of places:
        /// - for testing
        /// - for backwards compat when retrieving an unrecognized value from storage
        /// </summary>
        internal static LocationInfo locationInfoDefault => locationInfoMap["global"];

        /// <summary>
        /// Creates a new client id which will be associated with this account.
        /// Client ID should only be generated from the server and never changed from the client.
        /// </summary>
        public void CreateNewClientId()
        {
            ClientId = Guid.NewGuid().ToString();
        }

        private SkuInfo SkuTypeFromExternal(ExternalContracts.Account account)
        {
            if (!skuInfoMap.TryGetValue(account.Sku.Name.ToString(), out SkuInfo result))
            {
                Validation.Requires.Argument(false, "skuName", "invalid skuName");
            }

            return result;
        }

        private LocationInfo LocationFromExternal(ExternalContracts.Account account)
        {
            if (!locationInfoMap.TryGetValue(account.Location, out LocationInfo result))
            {
                Validation.Requires.Argument(false, "location", "invalid location");
            }

            return result;
        }

        private SkuInfo SkuTypeFromRecord(AccountRecord record)
        {
            if (!skuInfoMap.TryGetValue(record.SkuName, out SkuInfo result))
            {
                // set a default for backwards compat
                result = skuInfoDefault;
            }

            return result;
        }

        private LocationInfo LocationFromRecord(AccountRecord record)
        {
            if (!locationInfoMap.TryGetValue(record.Location, out LocationInfo result))
            {
                // set a default for backwards compat
                result = locationInfoDefault;
            }

            return result;
        }

        public class SkuInfo
        {
            public SkuName Name { get; set; }

            public SkuTierType Tier { get; set; }
        }

        public class LocationInfo
        {
            public string LocationString { get; set; }

            public LocationType Location { get; set; }
        }
    }
}