// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI
{
    public class AccountIdentifier
    {
        public AccountIdentifier(string subscriptionId, string resourceGroupName, string resourceName)
        {
            // more aggressive restrictions exist on these strings, but they are validated at the external interface
            Validation.Requires.NotNullOrWhiteSpace(subscriptionId, nameof(subscriptionId));
            Validation.Requires.NotNullOrWhiteSpace(resourceGroupName, nameof(resourceGroupName));
            Validation.Requires.NotNullOrWhiteSpace(resourceName, nameof(resourceName));

            SubscriptionId = subscriptionId;
            ResourceGroupName = resourceGroupName;
            ResourceName = resourceName;
        }

        /// <summary>
        ///     Build id from url form
        ///     /subscriptions/{id}/resourceGroups/{source}/providers/{namespace}/{type}/{name}
        ///     Fully parses and validates
        /// </summary>
        /// <param name="resourceId"></param>
        public AccountIdentifier(string resourceId)
        {
            Validation.Requires.NotNullOrWhiteSpace(resourceId, nameof(resourceId));

            // require beginning /, and remove it
            if (resourceId.StartsWith("/"))
            {
                resourceId = resourceId.Substring(1);
            }
            else
            {
                Validation.Requires.Fail(null, "resourceId format");
            }

            string[] tokens = resourceId.Split('/');
            Validation.Requires.That(tokens.Length == 8, "resourceId", "resourceId format");
            Validation.Requires.That(
                string.Compare(tokens[0], "subscriptions", System.StringComparison.InvariantCultureIgnoreCase) == 0,
                "resourceId",
                "resourceId format: subscriptions");
            Validation.Requires.That(
                string.Compare(tokens[2], "resourceGroups", System.StringComparison.InvariantCultureIgnoreCase) == 0,
                "resourceId",
                "resourceId format: resourceGroups");
            Validation.Requires.That(
                string.Compare(tokens[4], "providers", System.StringComparison.InvariantCultureIgnoreCase) == 0,
                "resourceId",
                "resourceId format: providers");
            Validation.Requires.That(
                (string.Compare(tokens[5],
                               "Microsoft.Maps",
                               System.StringComparison.InvariantCultureIgnoreCase) == 0)
                               || (string.Compare(tokens[5],
                                                  "Microsoft.Maps",
                                                  System.StringComparison.InvariantCultureIgnoreCase) == 0),
                "resourceId",
                "resourceId format: provider");
            Validation.Requires.That(
                string.Compare(tokens[6],
                               ResourceProviderWebApiService.ResourceTypeName,
                               System.StringComparison.InvariantCultureIgnoreCase) == 0,
                "resourceId",
                "resourceId format: resourcetype");

            ResourceProviderControllerBase.ValidateCommonParameters(tokens[1], tokens[3], tokens[7]);

            // phew if we get through all that, consider this valid
            SubscriptionId = tokens[1];
            ResourceGroupName = tokens[3];
            ResourceName = tokens[7];
        }

        public string SubscriptionId { get; }

        public string ResourceGroupName { get; }

        public string ResourceName { get; }
    }
}