﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement
{
    using System;
    using System.Collections.Specialized;
    using System.Globalization;
    using System.Net;
    using System.Security.Cryptography;
    using System.Text;
    using System.Text.RegularExpressions;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement.Contracts.ResponseContract;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement.Models;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Utility;

    using Validation;

    using RequestContract = Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement.Contracts.RequestContract;
    using ResponseContract = Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement.Contracts.ResponseContract;

    /// <summary>
    /// The APIMManager class
    /// </summary>
    public class ApiManagementClient
    {
        /// <summary>
        /// The cleanup regex
        /// </summary>
        private const string CleanupRegex = @"(\/users\/)|(\/subscriptions\/)|(\/products\/)|(\/groups\/)";

        /// <summary>
        /// The base URL
        /// </summary>
        private Uri baseUrl = null;

        /// <summary>
        /// The APIM identifier
        /// </summary>
        private string apimId = string.Empty;

        /// <summary>
        /// The APIM key
        /// </summary>
        private string apimKey = string.Empty;

        /// <summary>
        /// The token TTL
        /// </summary>
        private double tokenTtl = 10;

        /// <summary>
        /// The API version
        /// </summary>
        private string apiVersion = string.Empty;

        /// <summary>
        /// The encoded token
        /// </summary>
        private string encodedToken = string.Empty;

        /// <summary>
        /// The token expiry time
        /// </summary>
        private DateTime tokenExpiryTime = DateTime.UtcNow;

        /// <summary>
        /// The _lock
        /// </summary>
        private object lockobj = new object();

        /// <summary>
        /// Initializes a new instance of the <see cref="ApiManagementClient"/> class.
        /// </summary>
        /// <param name="uri">The URI.</param>
        /// <param name="id">The identifier.</param>
        /// <param name="key">The key.</param>
        /// <param name="apiEndpointDomain">The API endpoint domain.</param>
        /// <param name="name">API Management name</param>
        private ApiManagementClient(Uri uri, string id, string key, string apiEndpointDomain = null, string name = null)
        {
            this.baseUrl = uri;
            this.apimId = id;
            this.apimKey = key;
            this.apiVersion = "2014-02-14-preview";

            if (string.IsNullOrEmpty(apiEndpointDomain))
            {
                this.ApiEndpointDomain = uri.Host.Replace(".management.", ".");
            }
            else
            {
                this.ApiEndpointDomain = apiEndpointDomain;
            }

            if (string.IsNullOrEmpty(name))
            {
                this.Name = uri.Host.Substring(0, uri.Host.IndexOf('.'));
            }
            else
            {
                this.Name = name;
            }
        }

        /// <summary>
        /// Gets the API endpoint domain.
        /// </summary>
        /// <value>
        /// The API endpoint domain.
        /// </value>
        public string ApiEndpointDomain { get; private set; }

        public string Name { get; private set; }

        /// <summary>
        /// Factory method to construct a new instance of <see cref="ApiManagementClient"/>
        /// </summary>
        /// <param name="bootstrapConfiguration">configuration read as a bootstrap</param>
        /// <param name="configurationReader">configuration source</param>
        /// <returns>new instance of <see cref="ApiManagementClient"/></returns>
        public static async Task<ApiManagementClient> BuildAsync(IBootstrapConfiguration bootstrapConfiguration, IConfigurationReader configurationReader)
        {
            Requires.NotNull(configurationReader, nameof(configurationReader));
            Requires.NotNull(bootstrapConfiguration, nameof(bootstrapConfiguration));
            string apimClientId = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(Globals.ApimClientIdKey));
            string apimClientKey = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(Globals.ApimAuthenticationKey));
            string apimClientUri = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(Globals.ApimUriKey));
            return new ApiManagementClient(new Uri(apimClientUri), apimClientId, apimClientKey);
        }

        #region API APIs

        /// <summary>
        /// Gets the APIS.
        /// </summary>
        /// <returns>The APIS</returns>
        public ResponseContract.Apis GetApis()
        {
            string jsonSubs = this.Download(Constants.GetApisUrlTemplate);
            return Utility.DeserializeObject<ResponseContract.Apis>(jsonSubs);
        }

        /// <summary>
        /// Gets the APIS asynchronously.
        /// </summary>
        /// <returns>The APIS</returns>
        public async Task<ResponseContract.Apis> GetApisAsync()
        {
            string jsonSubs = await this.DownloadAsync(Constants.GetApisUrlTemplate);
            return Utility.DeserializeObject<ResponseContract.Apis>(jsonSubs);
        }

        /// <summary>
        /// Gets the API.
        /// </summary>
        /// <param name="id">The identifier.</param>
        /// <returns>The API</returns>
        public ResponseContract.Api GetApi(string id)
        {
            string jsonSubs = this.Download(string.Format(CultureInfo.InvariantCulture, Constants.GetApiUrlTemplate, id));
            return Utility.DeserializeObject<ResponseContract.Api>(jsonSubs);
        }

        #endregion

        #region Operation APIs
        /// <summary>
        /// Gets the operations.
        /// </summary>
        /// <param name="apiIdentifier">The API identifier.</param>
        /// <returns>The OperationSummaries</returns>
        public ResponseContract.OperationSummarys GetOperations(string apiIdentifier)
        {
            string jsonSubs = this.Download(string.Format(CultureInfo.InvariantCulture, Constants.GetOperationsUrlTemplate, apiIdentifier));
            return Utility.DeserializeObject<ResponseContract.OperationSummarys>(jsonSubs);
        }
        #endregion

        #region USER APIs

        /// <summary>
        /// Gets the user.
        /// </summary>
        /// <param name="id">The identifier.</param>
        /// <returns>The User</returns>
        public ResponseContract.User GetUser(string id)
        {
            string jsonSubs = this.Download(string.Format(CultureInfo.InvariantCulture, Constants.GetUserUrlTemplate, id));
            return Utility.DeserializeObject<ResponseContract.User>(jsonSubs);
        }

        /// <summary>
        /// Gets the user asynchronous.
        /// </summary>
        /// <param name="id">The identifier.</param>
        /// <returns>The User</returns>
        public async Task<ResponseContract.User> GetUserAsync(string id)
        {
            string jsonSubs = await this.DownloadAsync(string.Format(CultureInfo.InvariantCulture, Constants.GetUserUrlTemplate, id));
            return Utility.DeserializeObject<ResponseContract.User>(jsonSubs);
        }

        /// <summary>
        /// Gets the users.
        /// </summary>
        /// <returns>The Users</returns>
        public ResponseContract.Users GetUsers()
        {
            string jsonSubs = this.Download(Constants.GetUsersUrlTemplate);
            return Utility.DeserializeObject<ResponseContract.Users>(jsonSubs);
        }

        /// <summary>
        /// Gets the users with groups.
        /// </summary>
        /// <param name="batchSize">Size of the batch.</param>
        /// <param name="skipSize">Size of the skip.</param>
        /// <returns>The Users</returns>
        public ResponseContract.Users GetUsersWithGroups(int batchSize, int skipSize)
        {
            NameValueCollection parameters = new NameValueCollection();
            parameters.Add("expandGroups", "true");
            var top = batchSize.ToString();
            parameters.Add("interval", "P500D");
            parameters.Add("$top", batchSize.ToString());
            parameters.Add("$skip", skipSize.ToString());

            string jsonSubs = this.Download(Constants.GetUsersUrlTemplate, null, parameters);
            return Utility.DeserializeObject<ResponseContract.Users>(jsonSubs);
        }

        /// <summary>
        /// Gets the users asynchronous.
        /// </summary>
        /// <returns>The Users</returns>
        public async Task<ResponseContract.Users> GetUsersAsync()
        {
            string jsonSubs = await this.DownloadAsync(Constants.GetUsersUrlTemplate);
            return Utility.DeserializeObject<ResponseContract.Users>(jsonSubs);
        }

        /// <summary>
        /// Creates the user.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="firstName">The first name.</param>
        /// <param name="lastName">The last name.</param>
        /// <param name="email">The email.</param>
        /// <returns>The string value</returns>
        public string CreateUser(string userId, string firstName, string lastName, string email)
        {
            return this.CreateUser(userId, firstName, lastName, email, null);
        }

        /// <summary>
        /// Creates the user.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="firstName">The first name.</param>
        /// <param name="lastName">The last name.</param>
        /// <param name="email">The email.</param>
        /// <param name="userMetadata">The user metadata.</param>
        /// <returns>The string value</returns>
        public string CreateUser(string userId, string firstName, string lastName, string email, string userMetadata)
        {
            RequestContract.User user = new RequestContract.User() { FirstName = firstName, LastName = lastName, Email = email, Password = Guid.NewGuid().ToString(), State = UserState.Active, Note = string.IsNullOrEmpty(userMetadata) ? "createdNewUser" : userMetadata };
            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.CreateUserUrlTemplate, userId), HttpVerb.Put.ToString(), Utility.SerializeObject(user));
        }

        /// <summary>
        /// Creates the user asynchronous.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="firstName">The first name.</param>
        /// <param name="lastName">The last name.</param>
        /// <param name="email">The email.</param>
        /// <returns>The string value</returns>
        public Task<string> CreateUserAsync(string userId, string firstName, string lastName, string email)
        {
            return this.CreateUserAsync(userId, firstName, lastName, email, null);
        }

        /// <summary>
        /// Creates the user asynchronous.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="firstName">The first name.</param>
        /// <param name="lastName">The last name.</param>
        /// <param name="email">The email.</param>
        /// <param name="userMetadata">The user metadata.</param>
        /// <returns>The string value</returns>
        public Task<string> CreateUserAsync(string userId, string firstName, string lastName, string email, string userMetadata)
        {
            RequestContract.User user = new RequestContract.User() { FirstName = firstName, LastName = lastName, Email = email, Password = Guid.NewGuid().ToString(), State = UserState.Active, Note = string.IsNullOrEmpty(userMetadata) ? "createdNewUser" : userMetadata };
            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.CreateUserUrlTemplate, userId), HttpVerb.Put.ToString(), Utility.SerializeObject(user));
        }

        /// <summary>
        /// Updates the user.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="firstName">The first name.</param>
        /// <param name="lastName">The last name.</param>
        /// <param name="email">The email.</param>
        /// <param name="state">The state.</param>
        /// <returns>The string value</returns>
        public string UpdateUser(string userId, string firstName, string lastName, string email, UserState? state)
        {
            return this.UpdateUser(userId, firstName, lastName, email, state, null);
        }

        /// <summary>
        /// Updates the user.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="firstName">The first name.</param>
        /// <param name="lastName">The last name.</param>
        /// <param name="email">The email.</param>
        /// <param name="state">The state.</param>
        /// <param name="userMetadata">The user metadata.</param>
        /// <returns>The string value</returns>
        public string UpdateUser(string userId, string firstName, string lastName, string email, UserState? state, string userMetadata)
        {
            RequestContract.User user = new RequestContract.User()
            {
                FirstName = firstName,
                LastName = lastName,
                Email = email,
                Note = string.IsNullOrEmpty(userMetadata) ? "updatedUser" : userMetadata
            };

            if (state.HasValue)
            {
                user.State = state.Value;
            }

            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.UpdateUserUrlTemplate, userId), HttpVerb.Patch.ToString(), Utility.SerializeObject(user));
        }

        /// <summary>
        /// Updates the user asynchronous.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="firstName">The first name.</param>
        /// <param name="lastName">The last name.</param>
        /// <param name="email">The email.</param>
        /// <param name="state">The state.</param>
        /// <returns>The string value</returns>
        public Task<string> UpdateUserAsync(string userId, string firstName, string lastName, string email, UserState? state)
        {
            return this.UpdateUserAsync(userId, firstName, lastName, email, state, null);
        }

        /// <summary>
        /// Updates the user asynchronous.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="firstName">The first name.</param>
        /// <param name="lastName">The last name.</param>
        /// <param name="email">The email.</param>
        /// <param name="state">The state.</param>
        /// <param name="note">The note.</param>
        /// <returns>The string value</returns>
        public Task<string> UpdateUserAsync(string userId, string firstName, string lastName, string email, UserState? state, string note)
        {
            RequestContract.User user = new RequestContract.User()
            {
                FirstName = firstName,
                LastName = lastName,
                Email = email,
                Note = note
            };

            if (state.HasValue)
            {
                user.State = state.Value;
            }

            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.UpdateUserUrlTemplate, userId), HttpVerb.Patch.ToString(), Utility.SerializeObject(user));
        }

        /// <summary>
        /// Updates the user policy.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="note">The note.</param>
        /// <returns>The string value</returns>
        public string UpdateUserPolicy(string userId, string note)
        {
            RequestContract.User user = new RequestContract.User() { Note = note };
            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.UpdateUserUrlTemplate, userId), HttpVerb.Patch.ToString(), Utility.SerializeObject(user));
        }

        /// <summary>
        /// Updates the user policy asynchronous.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="note">The note.</param>
        /// <returns>The string value</returns>
        public Task<string> UpdateUserPolicyAsync(string userId, string note)
        {
            RequestContract.User user = new RequestContract.User() { Note = note };
            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.UpdateUserUrlTemplate, userId), HttpVerb.Patch.ToString(), Utility.SerializeObject(user));
        }

        /// <summary>
        /// Gets the SSO URL.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The string value</returns>
        public string GetSsoUrl(string userId)
        {
            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.GetSsoUrlUrlTemplate, userId), HttpVerb.Post.ToString(), string.Empty);
        }

        /// <summary>
        /// Gets the SSO URL asynchronous.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The string value</returns>
        public Task<string> GetSsoUrlAsync(string userId)
        {
            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.GetSsoUrlUrlTemplate, userId), HttpVerb.Post.ToString(), string.Empty);
        }

        /// <summary>
        /// Deletes the user.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The string value</returns>
        public string DeleteUser(string userId)
        {
            return this.DeleteUser(userId, true);
        }

        /// <summary>
        /// Deletes the user.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="deleteSubscriptions">if set to <c>true</c> [delete subscriptions].</param>
        /// <returns>The string value</returns>
        public string DeleteUser(string userId, bool deleteSubscriptions)
        {
            var queryParams = new NameValueCollection();
            if (deleteSubscriptions)
            {
                queryParams.Add("deleteSubscriptions", "true");
            }

            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.DeleteUserUrlTemplate, userId), HttpVerb.Delete.ToString(), string.Empty, null, queryParams);
        }

        /// <summary>
        /// Deletes the user subscription.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The string value</returns>
        public string DeleteUserSubscription(string userId)
        {
            return this.DeleteUser(userId, true);
        }

        /// <summary>
        /// Deletes the subscription.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The string value</returns>
        public string DeleteSubscription(string userId, string productId)
        {
            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.DeleteUserSubscriptionUrlTemplate, userId, productId), HttpVerb.Delete.ToString(), string.Empty);
        }

        /// <summary>
        /// Deletes the specific subscription.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <returns>The string value</returns>
        public string DeleteSpecificSubscription(string subscriptionId)
        {
            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.DeleteSpecificSubscriptionUrlTemplate, subscriptionId), HttpVerb.Delete.ToString(), string.Empty);
        }

        /// <summary>
        /// Updates the specific subscription.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <returns>The string value</returns>
        public string UpdateSpecificSubscription(string subscriptionId)
        {
            return this.UpdateSpecificSubscription(subscriptionId, null, null, null, null, null, null);
        }

        /// <summary>
        /// Updates the specific subscription.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <param name="state">The state.</param>
        /// <returns>The string value</returns>
        public string UpdateSpecificSubscription(string subscriptionId, SubscriptionState? state)
        {
            return this.UpdateSpecificSubscription(subscriptionId, state, null, null, null, null, null);
        }

        /// <summary>
        /// Updates the specific subscription.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <param name="state">The state.</param>
        /// <param name="primaryKey">The primary key.</param>
        /// <returns>The string value</returns>
        public string UpdateSpecificSubscription(string subscriptionId, SubscriptionState? state, string primaryKey)
        {
            return this.UpdateSpecificSubscription(subscriptionId, state, primaryKey, null, null, null, null);
        }

        /// <summary>
        /// Updates the specific subscription.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <param name="state">The state.</param>
        /// <param name="primaryKey">The primary key.</param>
        /// <param name="secondaryKey">The secondary key.</param>
        /// <returns>The string value</returns>
        public string UpdateSpecificSubscription(string subscriptionId, SubscriptionState? state, string primaryKey, string secondaryKey)
        {
            return this.UpdateSpecificSubscription(subscriptionId, state, primaryKey, secondaryKey, null, null, null);
        }

        /// <summary>
        /// Updates the specific subscription.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <param name="state">The state.</param>
        /// <param name="primaryKey">The primary key.</param>
        /// <param name="secondaryKey">The secondary key.</param>
        /// <param name="stateComment">The state comment.</param>
        /// <returns>The string value</returns>
        public string UpdateSpecificSubscription(string subscriptionId, SubscriptionState? state, string primaryKey, string secondaryKey, string stateComment)
        {
            return this.UpdateSpecificSubscription(subscriptionId, state, primaryKey, secondaryKey, stateComment, null, null);
        }

        /// <summary>
        /// Updates the specific subscription.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <param name="state">The state.</param>
        /// <param name="primaryKey">The primary key.</param>
        /// <param name="secondaryKey">The secondary key.</param>
        /// <param name="stateComment">The state comment.</param>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The string value</returns>
        public string UpdateSpecificSubscription(string subscriptionId, SubscriptionState? state, string primaryKey, string secondaryKey, string stateComment, string productId)
        {
            return this.UpdateSpecificSubscription(subscriptionId, state, primaryKey, secondaryKey, stateComment, productId, null);
        }

        /// <summary>
        /// Updates the specific subscription.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <param name="state">The state.</param>
        /// <param name="primaryKey">The primary key.</param>
        /// <param name="secondaryKey">The secondary key.</param>
        /// <param name="stateComment">The state comment.</param>
        /// <param name="productId">The product identifier.</param>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The string value</returns>
        public string UpdateSpecificSubscription(string subscriptionId, SubscriptionState? state, string primaryKey, string secondaryKey, string stateComment, string productId, string userId)
        {
            var request = GetRequestForSubscriptionUpdate(state, primaryKey, secondaryKey, stateComment, productId, userId);
            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.GetSubscriptionUrlTemplate, subscriptionId), HttpVerb.Patch.ToString(), Utility.SerializeObject(request));
        }

        /// <summary>
        /// Updates the specific subscription asynchronous.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <returns>The string value</returns>
        public Task<string> UpdateSpecificSubscriptionAsync(string subscriptionId)
        {
            var request = GetRequestForSubscriptionUpdate(null, null, null, null, null, null);
            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.GetSubscriptionUrlTemplate, subscriptionId), HttpVerb.Patch.ToString(), Utility.SerializeObject(request));
        }

        /// <summary>
        /// Updates the specific subscription asynchronous.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <param name="state">The state.</param>
        /// <param name="primaryKey">The primary key.</param>
        /// <param name="secondaryKey">The secondary key.</param>
        /// <param name="stateComment">The state comment.</param>
        /// <param name="productId">The product identifier.</param>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The string value</returns>
        public Task<string> UpdateSpecificSubscriptionAsync(string subscriptionId, SubscriptionState? state, string primaryKey, string secondaryKey, string stateComment, string productId, string userId)
        {
            var request = GetRequestForSubscriptionUpdate(state, primaryKey, secondaryKey, stateComment, productId, userId);
            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.GetSubscriptionUrlTemplate, subscriptionId), HttpVerb.Patch.ToString(), Utility.SerializeObject(request));
        }

        /// <summary>
        /// Deletes the specific subscription asynchronous.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <returns>The string value</returns>
        public Task<string> DeleteSpecificSubscriptionAsync(string subscriptionId)
        {
            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.DeleteSpecificSubscriptionUrlTemplate, subscriptionId), HttpVerb.Delete.ToString(), string.Empty);
        }

        /// <summary>
        /// Deletes the user asynchronous.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The string value</returns>
        public Task<string> DeleteUserAsync(string userId)
        {
            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.DeleteUserUrlTemplate, userId), HttpVerb.Delete.ToString(), string.Empty);
        }

        /// <summary>
        /// Adds the user to group.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="groupId">The group identifier.</param>
        /// <returns>The string value</returns>
        public string AddUserToGroup(string userId, string groupId)
        {
            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.AddUserToGroupUrlTemplate, groupId, userId), HttpVerb.Put.ToString(), string.Empty);
        }

        /// <summary>
        /// Adds the user to group asynchronous.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="groupId">The group identifier.</param>
        /// <returns>The string value</returns>
        public Task<string> AddUserToGroupAsync(string userId, string groupId)
        {
            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.AddUserToGroupUrlTemplate, groupId, userId), HttpVerb.Put.ToString(), string.Empty);
        }

        /// <summary>
        /// Deletes the user from group.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="groupId">The group identifier.</param>
        /// <returns>The string value</returns>
        public string DeleteUserFromGroup(string userId, string groupId)
        {
            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.DeleteUserFromGroupUrlTemplate, groupId, userId), HttpVerb.Delete.ToString(), string.Empty);
        }

        /// <summary>
        /// Deletes the user from group asynchronous.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="groupId">The group identifier.</param>
        /// <returns>The string value</returns>
        public Task<string> DeleteUserFromGroupAsync(string userId, string groupId)
        {
            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.DeleteUserFromGroupUrlTemplate, groupId, userId), HttpVerb.Delete.ToString(), string.Empty);
        }

        #endregion

        #region SUBSCRIPTION APIs

        /// <summary>
        /// Gets the subscriptions.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The Subscriptions</returns>
        public ResponseContract.Subscriptions GetSubscriptions(string userId)
        {
            string jsonSubs = this.Download(string.Format(CultureInfo.InvariantCulture, Constants.GetSubscriptionsForUserUrlTemplate, userId));
            return Utility.DeserializeObject<ResponseContract.Subscriptions>(jsonSubs);
        }

        /// <summary>
        /// Gets the subscriptions asynchronous.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The Subscriptions</returns>
        public async Task<ResponseContract.Subscriptions> GetSubscriptionsAsync(string userId)
        {
            string jsonSubs = await this.DownloadAsync(string.Format(CultureInfo.InvariantCulture, Constants.GetSubscriptionsForUserUrlTemplate, userId));
            return Utility.DeserializeObject<ResponseContract.Subscriptions>(jsonSubs);
        }

        /// <summary>
        /// Gets the subscriptions.
        /// </summary>
        /// <returns>The Subscriptions</returns>
        public ResponseContract.Subscriptions GetSubscriptions()
        {
            string jsonSubs = this.Download(Constants.GetSubscriptionsUrlTemplate);
            return Utility.DeserializeObject<ResponseContract.Subscriptions>(jsonSubs);
        }

        /// <summary>
        /// Gets the subscriptions asynchronous.
        /// </summary>
        /// <returns>The Subscriptions</returns>
        public async Task<ResponseContract.Subscriptions> GetSubscriptionsAsync()
        {
            string jsonSubs = await this.DownloadAsync(Constants.GetSubscriptionsUrlTemplate);
            return Utility.DeserializeObject<ResponseContract.Subscriptions>(jsonSubs);
        }

        /// <summary>
        /// Gets the subscription.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <returns>The Subscription</returns>
        public ResponseContract.Subscription GetSubscription(string subscriptionId)
        {
            string jsonSubs = this.Download(string.Format(CultureInfo.InvariantCulture, Constants.GetSubscriptionUrlTemplate, subscriptionId));
            return Utility.DeserializeObject<ResponseContract.Subscription>(jsonSubs);
        }

        /// <summary>
        /// Tries to get the subscription
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <param name="subscription">The subscription</param>
        /// <returns>A value indicating whether the subscription was retrieved.</returns>
        public bool TryGetSubscription(string subscriptionId, out ResponseContract.Subscription subscription)
        {
            subscription = default(ResponseContract.Subscription);
            try
            {
                subscription = this.GetSubscription(subscriptionId);
                return true;
            }
            catch (ApimException e)
            {
                if (e.ExceptionType == ApimExceptionType.ResourceNotFound)
                {
                    return false;
                }

                throw;
            }
        }

        /// <summary>
        /// Gets the subscription asynchronous.
        /// </summary>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <returns>The Subscription</returns>
        public async Task<ResponseContract.Subscription> GetSubscriptionAsync(string subscriptionId)
        {
            string jsonSubs = await this.DownloadAsync(string.Format(CultureInfo.InvariantCulture, Constants.GetSubscriptionUrlTemplate, subscriptionId));
            return Utility.DeserializeObject<ResponseContract.Subscription>(jsonSubs);
        }

        /// <summary>
        /// Creates the product subscription.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The Subscription</returns>
        public ResponseContract.Subscription CreateProductSubscription(string userId, string productId)
        {
            return this.CreateProductSubscription(userId, productId, null, null, null, null, null);
        }

        /// <summary>
        /// Creates the product subscription.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="productId">The product identifier.</param>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <returns>The Subscription</returns>
        public ResponseContract.Subscription CreateProductSubscription(string userId, string productId, string subscriptionId)
        {
            return this.CreateProductSubscription(userId, productId, null, subscriptionId, null, null, null);
        }

        /// <summary>
        /// Creates the product subscription.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="productId">The product identifier.</param>
        /// <param name="subscriptionName">Name of the subscription.</param>
        /// <param name="stateComment">The state comment.</param>
        /// <returns>The Subscription</returns>
        public ResponseContract.Subscription CreateProductSubscription(string userId, string productId, string subscriptionName, string stateComment)
        {
            return this.CreateProductSubscription(userId, productId, subscriptionName, stateComment, null, null, null);
        }

        /// <summary>
        /// Creates the product subscription.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="productId">The product identifier.</param>
        /// <param name="subscriptionName">Name of the subscription.</param>
        /// <param name="stateComment">The state comment.</param>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <returns>The Subscription</returns>
        public ResponseContract.Subscription CreateProductSubscription(string userId, string productId, string subscriptionName, string stateComment, string subscriptionId)
        {
            return this.CreateProductSubscription(userId, productId, subscriptionName, stateComment, subscriptionId, null, null);
        }

        /// <summary>
        /// Creates the product subscription.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="productId">The product identifier.</param>
        /// <param name="subscriptionName">Name of the subscription.</param>
        /// <param name="stateComment">The state comment.</param>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <param name="primaryKey">The primary key.</param>
        /// <param name="secondaryKey">The secondary key.</param>
        /// <returns>The Subscription</returns>
        public ResponseContract.Subscription CreateProductSubscription(string userId, string productId, string subscriptionName, string stateComment, string subscriptionId, string primaryKey, string secondaryKey)
        {
            // Check if the user is subscribed to any other product
            RequestContract.Subscription newSub = new RequestContract.Subscription()
            {
                UserId = string.Format(CultureInfo.InvariantCulture, Constants.ApimUserFormat, userId),
                ProductId = string.Format(CultureInfo.InvariantCulture, Constants.ApimProductFormat, productId),
                PrimaryKey = !string.IsNullOrWhiteSpace(primaryKey) ? primaryKey : Guid.NewGuid().ToString("N"),
                SecondaryKey = !string.IsNullOrWhiteSpace(secondaryKey) ? secondaryKey : Guid.NewGuid().ToString("N"),
                State = SubscriptionState.Active,
                Name = subscriptionName
            };

            if (!string.IsNullOrWhiteSpace(stateComment))
            {
                newSub.StateComment = stateComment;
            }

            // create a new subId if not passed            
            var newSubId = !string.IsNullOrWhiteSpace(subscriptionId) ? subscriptionId : Guid.NewGuid().ToString("N");

            // create the new subscription
            this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.CreateSpecificSubscriptionUrlTemplate, newSubId), HttpVerb.Put.ToString(), Utility.SerializeObject(newSub));

            return this.GetSubscription(newSubId);
        }

        /// <summary>
        /// Switches the product subscription.
        /// </summary>
        /// <param name="userId">The user identifier.</param>
        /// <param name="newProductId">The new product identifier.</param>
        /// <param name="oldSubscriptionId">The old subscription identifier.</param>
        /// <param name="newSubscriptionId">The new subscription identifier.</param>
        /// <returns>The Subscription</returns>
        public ResponseContract.Subscription SwitchProductSubscription(string userId, string newProductId, string oldSubscriptionId, string newSubscriptionId)
        {
            ResponseContract.Subscription result = null;

            var oldSub = this.GetSubscription(oldSubscriptionId);
            if (oldSub != null)
            {
                var primaryKey = oldSub.PrimaryKey;
                var secondaryKey = oldSub.SecondaryKey;

                this.UpdateSpecificSubscription(oldSub.Id, SubscriptionState.Cancelled, Guid.NewGuid().ToString("N"), Guid.NewGuid().ToString("N"));

                try
                {
                    var newSub = new RequestContract.Subscription()
                    {
                        UserId = string.Format(CultureInfo.InvariantCulture, Constants.ApimUserFormat, userId),
                        Name = oldSub.Name,
                        ProductId = string.Format(CultureInfo.InvariantCulture, Constants.ApimProductFormat, newProductId),
                        PrimaryKey = primaryKey,
                        SecondaryKey = secondaryKey,
                        State = SubscriptionState.Active,
                        StateComment = oldSub.StateComment
                    };

                    //// create the new subscription
                    this.Upload(
                        string.Format(CultureInfo.InvariantCulture, Constants.CreateSpecificSubscriptionUrlTemplate, newSubscriptionId),
                        HttpVerb.Put.ToString(),
                        Utility.SerializeObject(newSub));
                }
                catch (Exception)
                {
                    ////Rollback
                    try
                    {
                        this.DeleteSpecificSubscription(newSubscriptionId);
                    }
                    catch
                    {
                    }

                    this.UpdateSpecificSubscription(oldSub.Id, SubscriptionState.Active, primaryKey, secondaryKey);
                    throw;
                }

                result = this.GetSubscription(newSubscriptionId);
            }

            return result;
        }

        #endregion

        #region PRODUCT APIs

        /// <summary>
        /// Gets the products.
        /// </summary>
        /// <returns>The Products</returns>
        public ResponseContract.Products GetProducts()
        {
            string jsonSubs = this.Download(Constants.GetProductsUrlTemplate);
            return Utility.DeserializeObject<ResponseContract.Products>(jsonSubs);
        }

        /// <summary>
        /// Gets the products asynchronous.
        /// </summary>
        /// <returns>The Products</returns>
        public async Task<ResponseContract.Products> GetProductsAsync()
        {
            string jsonSubs = await this.DownloadAsync(Constants.GetProductsUrlTemplate);
            return Utility.DeserializeObject<ResponseContract.Products>(jsonSubs);
        }

        /// <summary>
        /// Gets the product.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The Product</returns>
        public ResponseContract.Product GetProduct(string productId)
        {
            string jsonSubs = this.Download(string.Format(CultureInfo.InvariantCulture, Constants.GetProductUrlTemplate, productId));
            return Utility.DeserializeObject<ResponseContract.Product>(jsonSubs);
        }

        /// <summary>
        /// Gets the policy XML.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The string value</returns>
        public string GetPolicyXml(string productId)
        {
            try
            {
                return this.Download(string.Format(CultureInfo.InvariantCulture, Constants.GetPolicyUrlTemplate, productId), null, null, false).ToString();
            }
            catch (ResponseContract.ApimException ex)
            {
                if (ex.ExceptionType == ResponseContract.ApimExceptionType.ResourceNotFound)
                {
                    return null;
                }
                else
                {
                    throw;
                }
            }
        }

        /// <summary>
        /// Sets the policy.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="data">The data.</param>
        /// <returns>The string value</returns>
        public string SetPolicy(string productId, string data)
        {
            this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.GetPolicyUrlTemplate, productId), HttpVerb.Delete.ToString(), data);

            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.GetPolicyUrlTemplate, productId), HttpVerb.Put.ToString(), data, null, null, true, "application/vnd.ms-azure-apim.policy+xml");
        }

        /// <summary>
        /// Gets the policy.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The APIM Policy</returns>
        public ApimPolicy GetPolicy(string productId)
        {
            PolicyManager pm = new PolicyManager(this.GetPolicyXml(productId));
            return pm.GetQuotaAndRatelimit();
        }

        /// <summary>
        /// Sets the rate limit policy.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="calls">The calls.</param>
        /// <param name="renewalPeriod">The renewal period.</param>
        public void SetRateLimitPolicy(string productId, int calls, int renewalPeriod)
        {
            if (calls == 0 || renewalPeriod == 0)
            {
                return;
            }

            PolicyManager pm = new PolicyManager(this.GetPolicyXml(productId));
            this.SetPolicy(productId, pm.SetPolicyXml("rate-limit", calls, renewalPeriod));
        }

        /// <summary>
        /// Sets the quota policy.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="calls">The calls.</param>
        /// <param name="renewalPeriod">The renewal period.</param>
        public void SetQuotaPolicy(string productId, int calls, int renewalPeriod)
        {
            if (calls == 0 || renewalPeriod == 0)
            {
                return;
            }

            PolicyManager pm = new PolicyManager(this.GetPolicyXml(productId));
            this.SetPolicy(productId, pm.SetPolicyXml("quota", calls, renewalPeriod));
        }

        /// <summary>
        /// Gets the product asynchronous.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The Product</returns>
        public async Task<ResponseContract.Product> GetProductAsync(string productId)
        {
            string jsonSubs = await this.DownloadAsync(string.Format(CultureInfo.InvariantCulture, Constants.GetProductUrlTemplate, productId));
            return Utility.DeserializeObject<ResponseContract.Product>(jsonSubs);
        }

        /// <summary>
        /// Gets the product APIS.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The APIS</returns>
        public ResponseContract.Apis GetProductApis(string productId)
        {
            string jsonSubs = this.Download(string.Format(CultureInfo.InvariantCulture, Constants.GetProductApisUrlTemplate, productId));
            return Utility.DeserializeObject<ResponseContract.Apis>(jsonSubs);
        }

        /// <summary>
        /// Gets the product APIS asynchronous.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The APIS</returns>
        public async Task<ResponseContract.Apis> GetProductApisAsync(string productId)
        {
            string jsonSubs = await this.DownloadAsync(string.Format(CultureInfo.InvariantCulture, Constants.GetProductApisUrlTemplate, productId));
            return Utility.DeserializeObject<ResponseContract.Apis>(jsonSubs);
        }

        /// <summary>
        /// Creates the product.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="description">The description.</param>
        /// <param name="terms">The terms.</param>
        /// <returns>The string value</returns>
        public string CreateProduct(string name, string description, string terms)
        {
            return this.CreateProduct(name, description, terms, true, true, null);
        }

        /// <summary>
        /// Creates the product.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="description">The description.</param>
        /// <param name="terms">The terms.</param>
        /// <param name="subscriptionRequired">if set to <c>true</c> [subscription required].</param>
        /// <param name="approvalRequired">if set to <c>true</c> [approval required].</param>
        /// <param name="subscriptionsLimit">The subscriptions limit.</param>
        /// <returns>The string value</returns>
        public string CreateProduct(string name, string description, string terms, bool subscriptionRequired, bool approvalRequired, int? subscriptionsLimit)
        {
            return this.CreateProduct(Guid.NewGuid().ToString("N"), name, description, terms, subscriptionRequired, approvalRequired, subscriptionsLimit);
        }

        /// <summary>
        /// Creates the product.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="name">The name.</param>
        /// <param name="description">The description.</param>
        /// <param name="terms">The terms.</param>
        /// <returns>The string value</returns>
        public string CreateProduct(string productId, string name, string description, string terms)
        {
            return this.CreateProduct(productId, name, description, terms, true, true, null);
        }

        /// <summary>
        /// Creates the product.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="name">The name.</param>
        /// <param name="description">The description.</param>
        /// <param name="terms">The terms.</param>
        /// <param name="subscriptionRequired">if set to <c>true</c> [subscription required].</param>
        /// <param name="approvalRequired">if set to <c>true</c> [approval required].</param>
        /// <param name="subscriptionsLimit">The subscriptions limit.</param>
        /// <returns>The string value</returns>
        public string CreateProduct(string productId, string name, string description, string terms, bool subscriptionRequired, bool approvalRequired, int? subscriptionsLimit)
        {
            RequestContract.Product product = new RequestContract.Product()
            {
                Name = name,
                Description = description,
                Terms = terms,
                SubscriptionRequired = subscriptionRequired,
                ApprovalRequired = approvalRequired,
                SubscriptionsLimit = subscriptionsLimit,
                State = ProductState.NotPublished
            };

            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.CreateProductUrlTemplate, productId), HttpVerb.Put.ToString(), Utility.SerializeObject(product));
        }

        /// <summary>
        /// Creates the product asynchronous.
        /// </summary>
        /// <param name="name">The name.</param>
        /// <param name="description">The description.</param>
        /// <param name="terms">The terms.</param>
        /// <returns>The string value</returns>
        public Task<string> CreateProductAsync(string name, string description, string terms)
        {
            return this.CreateProductAsync(Guid.NewGuid().ToString("N"), name, description, terms, true, true, null);
        }

        /// <summary>
        /// Creates the product asynchronous.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="name">The name.</param>
        /// <param name="description">The description.</param>
        /// <param name="terms">The terms.</param>
        /// <param name="subscriptionRequired">if set to <c>true</c> [subscription required].</param>
        /// <param name="approvalRequired">if set to <c>true</c> [approval required].</param>
        /// <param name="subscriptionsLimit">The subscriptions limit.</param>
        /// <returns>The string value</returns>
        public Task<string> CreateProductAsync(string productId, string name, string description, string terms, bool subscriptionRequired, bool approvalRequired, int? subscriptionsLimit)
        {
            RequestContract.Product product = new RequestContract.Product()
            {
                Name = name,
                Description = description,
                Terms = terms,
                SubscriptionRequired = subscriptionRequired,
                ApprovalRequired = approvalRequired,
                SubscriptionsLimit = subscriptionsLimit,
                State = ProductState.NotPublished
            };

            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.CreateProductUrlTemplate, productId), HttpVerb.Put.ToString(), Utility.SerializeObject(product));
        }

        /// <summary>
        /// Updates the product.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="name">The name.</param>
        /// <param name="description">The description.</param>
        /// <param name="terms">The terms.</param>
        /// <returns>The string value</returns>
        public string UpdateProduct(string productId, string name, string description, string terms)
        {
            return this.UpdateProduct(productId, name, description, terms, true, true, null, ProductState.NotPublished);
        }

        /// <summary>
        /// Updates the product.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="name">The name.</param>
        /// <param name="description">The description.</param>
        /// <param name="terms">The terms.</param>
        /// <param name="subscriptionRequired">if set to <c>true</c> [subscription required].</param>
        /// <param name="approvalRequired">if set to <c>true</c> [approval required].</param>
        /// <param name="subscriptionsLimit">The subscriptions limit.</param>
        /// <param name="productState">State of the product.</param>
        /// <returns>The string value</returns>
        public string UpdateProduct(string productId, string name, string description, string terms, bool subscriptionRequired, bool approvalRequired, int? subscriptionsLimit, ProductState productState)
        {
            RequestContract.Product product = new RequestContract.Product()
            {
                Name = name,
                Description = description,
                Terms = terms,
                SubscriptionRequired = subscriptionRequired,
                ApprovalRequired = approvalRequired,
                SubscriptionsLimit = subscriptionsLimit,
                State = productState
            };

            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.UpdateProductUrlTemplate, productId), HttpVerb.Patch.ToString(), Utility.SerializeObject(product));
        }

        /// <summary>
        /// Updates the product asynchronous.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="name">The name.</param>
        /// <param name="description">The description.</param>
        /// <param name="terms">The terms.</param>
        /// <returns>The string value</returns>
        public Task<string> UpdateProductAsync(string productId, string name, string description, string terms)
        {
            return this.UpdateProductAsync(productId, name, description, terms, true, true, null, ProductState.NotPublished);
        }

        /// <summary>
        /// Updates the product asynchronous.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="name">The name.</param>
        /// <param name="description">The description.</param>
        /// <param name="terms">The terms.</param>
        /// <param name="subscriptionRequired">if set to <c>true</c> [subscription required].</param>
        /// <param name="approvalRequired">if set to <c>true</c> [approval required].</param>
        /// <param name="subscriptionsLimit">The subscriptions limit.</param>
        /// <param name="productState">State of the product.</param>
        /// <returns>The string value</returns>
        public Task<string> UpdateProductAsync(string productId, string name, string description, string terms, bool subscriptionRequired, bool approvalRequired, int? subscriptionsLimit, ProductState productState)
        {
            RequestContract.Product product = new RequestContract.Product()
            {
                Name = name,
                Description = description,
                Terms = terms,
                SubscriptionRequired = subscriptionRequired,
                ApprovalRequired = approvalRequired,
                SubscriptionsLimit = subscriptionsLimit,
                State = productState
            };

            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.UpdateProductUrlTemplate, productId), HttpVerb.Patch.ToString(), Utility.SerializeObject(product));
        }

        /// <summary>
        /// Deletes the product.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The string value</returns>
        public string DeleteProduct(string productId)
        {
            return this.DeleteProduct(productId, false);
        }

        /// <summary>
        /// Deletes the product.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="deleteSubscriptions">if set to <c>true</c> [delete subscriptions].</param>
        /// <returns>The string value</returns>
        public string DeleteProduct(string productId, bool deleteSubscriptions)
        {
            var nvc = new NameValueCollection();
            nvc.Add("deleteSubscriptions", deleteSubscriptions.ToString());

            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.UpdateProductUrlTemplate, productId), HttpVerb.Delete.ToString(), string.Empty, additionalQueryParams: nvc);
        }

        /// <summary>
        /// Deletes the product asynchronous.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The string value</returns>
        public Task<string> DeleteProductAsync(string productId)
        {
            return this.DeleteProductAsync(productId, false);
        }

        /// <summary>
        /// Deletes the product asynchronous.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="deleteSubscriptions">if set to <c>true</c> [delete subscriptions].</param>
        /// <returns>The string value</returns>
        public Task<string> DeleteProductAsync(string productId, bool deleteSubscriptions)
        {
            var nvc = new NameValueCollection();
            nvc.Add("deleteSubscriptions", deleteSubscriptions.ToString());

            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.UpdateProductUrlTemplate, productId), HttpVerb.Delete.ToString(), string.Empty, additionalQueryParams: nvc);
        }

        /// <summary>
        /// Creates the product API.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="apiId">The API identifier.</param>
        /// <returns>The string value</returns>
        public string CreateProductApi(string productId, string apiId)
        {
            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.CreateProductApiUrlTemplate, productId, apiId), HttpVerb.Put.ToString(), string.Empty);
        }

        /// <summary>
        /// Creates the product API asynchronous.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="apiId">The API identifier.</param>
        /// <returns>The string value</returns>
        public Task<string> CreateProductApiAsync(string productId, string apiId)
        {
            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.CreateProductApiUrlTemplate, productId, apiId), HttpVerb.Put.ToString(), string.Empty);
        }

        /// <summary>
        /// Deletes the product API.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="apiId">The API identifier.</param>
        /// <returns>The string value</returns>
        public string DeleteProductApi(string productId, string apiId)
        {
            return this.Upload(string.Format(CultureInfo.InvariantCulture, Constants.DeleteProductApiUrlTemplate, productId, apiId), HttpVerb.Delete.ToString(), string.Empty);
        }

        /// <summary>
        /// Deletes the product API asynchronous.
        /// </summary>
        /// <param name="productId">The product identifier.</param>
        /// <param name="apiId">The API identifier.</param>
        /// <returns>The string value</returns>
        public Task<string> DeleteProductApiAsync(string productId, string apiId)
        {
            return this.UploadAsync(string.Format(CultureInfo.InvariantCulture, Constants.DeleteProductApiUrlTemplate, productId, apiId), HttpVerb.Delete.ToString(), string.Empty);
        }

        #endregion

        #region Report APIs

        /// <summary>
        /// Gets the report by operations.
        /// </summary>
        /// <param name="queryParameter">The query parameter.</param>
        /// <returns>The ReportCollection</returns>
        public ResponseContract.ReportCollection GetReportByOperations(string queryParameter)
        {
            NameValueCollection parameters = new NameValueCollection();
            parameters.Add("interval", "P500D");
            parameters.Add("$filter", queryParameter);
            string jsonSubs = this.Download(Constants.GetReportByOperationUrlTemplate, null, parameters);
            return Utility.DeserializeObject<ResponseContract.ReportCollection>(jsonSubs);
        }

        /// <summary>
        /// Gets the report by user.
        /// </summary>
        /// <param name="queryParameter">The query parameter.</param>
        /// <returns>The ReportCollection</returns>
        public ResponseContract.ReportCollection GetReportByUser(string queryParameter)
        {
            string jsonSubs = this.Download(Constants.GetReportByUserUrlTemplate + queryParameter);
            return Utility.DeserializeObject<ResponseContract.ReportCollection>(jsonSubs);
        }

        /// <summary>
        /// Gets the report by user.
        /// </summary>
        /// <param name="fromDate">From date.</param>
        /// <param name="toDate">To date.</param>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The ReportCollection</returns>
        public ResponseContract.ReportCollection GetReportByUser(DateTime fromDate, DateTime toDate, string userId)
        {
            return this.GetReportByUser(fromDate, toDate, userId, string.Empty);
        }

        /// <summary>
        /// Gets the report by user.
        /// </summary>
        /// <param name="fromDate">From date.</param>
        /// <param name="toDate">To date.</param>
        /// <param name="userId">The user identifier.</param>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The ReportCollection</returns>
        public ResponseContract.ReportCollection GetReportByUser(DateTime fromDate, DateTime toDate, string userId, string productId)
        {
            NameValueCollection parameters = new NameValueCollection();

            StringBuilder filterParameters = new StringBuilder();
            if (fromDate != DateTime.MinValue)
            {
                filterParameters.Append("timestamp ge datetime'" + fromDate.ToString("yyyy-MM-ddTHH:mm:ss.fffZ", CultureInfo.InvariantCulture) + "'");
            }

            if (toDate != DateTime.MinValue)
            {
                if (filterParameters.Length > 0)
                {
                    filterParameters.Append("and ");
                }

                filterParameters.Append("timestamp le datetime'" + toDate.ToString("yyyy-MM-ddTHH:mm:ss.fffZ", CultureInfo.InvariantCulture) + "'");
            }

            if (!string.IsNullOrWhiteSpace(productId))
            {
                if (filterParameters.Length > 0)
                {
                    filterParameters.Append("and ");
                }

                filterParameters.Append(string.Format(CultureInfo.InvariantCulture, "productId eq '{0}'", productId));
            }

            if (!string.IsNullOrWhiteSpace(userId))
            {
                if (filterParameters.Length > 0)
                {
                    filterParameters.Append("and ");
                }

                filterParameters.Append(string.Format(CultureInfo.InvariantCulture, "UserId eq '{0}'", userId));
            }

            var filter = filterParameters.ToString();

            parameters.Add("interval", "P500D");
            parameters.Add("$filter", filter);

            string jsonSubs = this.Download(Constants.GetReportByTimeUrlTemplate, null, parameters);
            return Utility.DeserializeObject<ResponseContract.ReportCollection>(jsonSubs);
        }

        /// <summary>
        /// Gets the report by user.
        /// </summary>
        /// <param name="fromDate">From date.</param>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The ReportCollection</returns>
        public ResponseContract.ReportCollection GetReportByUser(DateTime fromDate, string userId)
        {
            return this.GetReportByUser(fromDate, DateTime.MinValue, userId, string.Empty);
        }

        /// <summary>
        /// Gets the report by user.
        /// </summary>
        /// <param name="fromDate">From date.</param>
        /// <param name="userId">The user identifier.</param>
        /// <param name="productId">The product identifier.</param>
        /// <returns>The ReportCollection</returns>
        public ResponseContract.ReportCollection GetReportByUser(DateTime fromDate, string userId, string productId)
        {
            return this.GetReportByUser(fromDate, DateTime.MinValue, userId, productId);
        }

        /// <summary>
        /// Gets the report by subscription.
        /// </summary>
        /// <param name="fromDate">From date.</param>
        /// <param name="toDate">To date.</param>
        /// <param name="subscriptionId">The subscription identifier.</param>
        /// <returns>The ReportCollection</returns>
        public ResponseContract.ReportCollection GetReportBySubscription(DateTime fromDate, DateTime toDate, string subscriptionId)
        {
            NameValueCollection parameters = new NameValueCollection();

            StringBuilder filterParameters = new StringBuilder();
            if (fromDate != DateTime.MinValue)
            {
                filterParameters.Append("timestamp ge datetime'" + fromDate.ToString("yyyy-MM-ddTHH:mm:ss.fffZ", CultureInfo.InvariantCulture) + "'");
            }

            if (toDate != DateTime.MinValue)
            {
                if (filterParameters.Length > 0)
                {
                    filterParameters.Append(" and ");
                }

                filterParameters.Append("timestamp le datetime'" + toDate.ToString("yyyy-MM-ddTHH:mm:ss.fffZ", CultureInfo.InvariantCulture) + "'");
            }

            if (!string.IsNullOrWhiteSpace(subscriptionId))
            {
                if (filterParameters.Length > 0)
                {
                    filterParameters.Append(" and ");
                }

                filterParameters.Append(string.Format(CultureInfo.InvariantCulture, "subscriptionId eq '{0}'", subscriptionId));
            }

            var filter = filterParameters.ToString();

            parameters.Add("interval", "P500D");
            parameters.Add("$filter", filter);
            string jsonSubs = this.Download(Constants.GetReportBySubscriptionUriTemplate, null, parameters);
            return Utility.DeserializeObject<ResponseContract.ReportCollection>(jsonSubs);
        }

        #endregion

        /// <summary>
        /// Gets the request for subscription update.
        /// </summary>
        /// <param name="state">The state.</param>
        /// <param name="primaryKey">The primary key.</param>
        /// <param name="secondaryKey">The secondary key.</param>
        /// <param name="stateComment">The state comment.</param>
        /// <param name="productId">The product identifier.</param>
        /// <param name="userId">The user identifier.</param>
        /// <returns>The Subscription</returns>
        private static RequestContract.Subscription GetRequestForSubscriptionUpdate(
                                                                                    SubscriptionState? state,
                                                                                    string primaryKey,
                                                                                    string secondaryKey,
                                                                                    string stateComment,
                                                                                    string productId,
                                                                                    string userId)
        {
            var request = new RequestContract.Subscription();

            if (state.HasValue)
            {
                request.State = state.Value;
            }

            if (!string.IsNullOrWhiteSpace(primaryKey))
            {
                request.PrimaryKey = primaryKey;
            }

            if (!string.IsNullOrWhiteSpace(secondaryKey))
            {
                request.SecondaryKey = secondaryKey;
            }

            if (!string.IsNullOrWhiteSpace(stateComment))
            {
                request.StateComment = stateComment;
            }

            if (!string.IsNullOrWhiteSpace(productId))
            {
                request.ProductId = string.Format(CultureInfo.InvariantCulture, Constants.ApimProductFormat, productId);
            }

            if (!string.IsNullOrWhiteSpace(userId))
            {
                request.UserId = string.Format(CultureInfo.InvariantCulture, Constants.ApimUserFormat, userId);
            }

            return request;
        }

        #region Helpers

        /// <summary>
        /// Cleans up response asynchronous.
        /// </summary>
        /// <param name="respFromApim">The RESP from APIM.</param>
        /// <returns>The string value</returns>
        private static async Task<string> CleanUpResponseAsync(Task<string> respFromApim)
        {
            var result = await respFromApim;
            Regex cleanUpRegex = new Regex(CleanupRegex);
            return cleanUpRegex.Replace(result, string.Empty);
        }

        /// <summary>
        /// Cleans up response.
        /// </summary>
        /// <param name="respFromApim">The RESP from APIM.</param>
        /// <returns>The string value</returns>
        private static string CleanUpResponse(string respFromApim)
        {
            Regex cleanUpRegex = new Regex(CleanupRegex);
            return cleanUpRegex.Replace(respFromApim, string.Empty);
        }

        /// <summary>
        /// For POST/PUT REST calls
        /// </summary>
        /// <param name="apiTemplate">The API template.</param>
        /// <param name="requestType">Type of the request.</param>
        /// <param name="data">The data.</param>
        /// <param name="additionalHeaders">The additional headers.</param>
        /// <param name="additionalQueryParams">The additional query parameters.</param>
        /// <param name="ifMatchHeaderRequired">if set to <c>true</c> [if match header required].</param>
        /// <param name="headerContentType">Type of the header content.</param>
        /// <returns>The string value</returns>
        private string Upload(string apiTemplate, string requestType, string data, NameValueCollection additionalHeaders = null, NameValueCollection additionalQueryParams = null, bool ifMatchHeaderRequired = false, string headerContentType = null)
        {
            try
            {
                using (var client = new WebClient())
                {
                    client.Headers.Add("Authorization", this.GenerateTransientToken());
                    if (headerContentType == null)
                    {
                        client.Headers.Add("Content-Type", @"application/json");
                    }
                    else
                    {
                        client.Headers.Add("Content-Type", headerContentType);
                    }

                    if (requestType == HttpVerb.Put.ToString() || requestType == HttpVerb.Patch.ToString() || requestType == HttpVerb.Delete.ToString() || ifMatchHeaderRequired)
                    {
                        client.Headers.Add("If-Match", "*");
                    }

                    if (additionalHeaders != null)
                    {
                        foreach (string key in additionalHeaders.Keys)
                        {
                            client.Headers.Add(key, additionalHeaders[key]);
                        }
                    }

                    if (additionalQueryParams != null)
                    {
                        foreach (string key in additionalQueryParams.Keys)
                        {
                            client.QueryString.Add(key, additionalQueryParams[key]);
                        }
                    }

                    client.QueryString.Add("api-version", this.apiVersion);
                    return CleanUpResponse(client.UploadString(this.baseUrl + apiTemplate, requestType, data));
                }
            }
            catch (WebException ex)
            {
                throw ApimExceptionFactory.GetApimException(ex);
            }
        }

        /// <summary>
        /// Uploads the asynchronous.
        /// </summary>
        /// <param name="apiTemplate">The API template.</param>
        /// <param name="requestType">Type of the request.</param>
        /// <param name="data">The data.</param>
        /// <param name="additionalHeaders">The additional headers.</param>
        /// <param name="additionalQueryParams">The additional query parameters.</param>
        /// <returns>The string value</returns>
        private async Task<string> UploadAsync(string apiTemplate, string requestType, string data, NameValueCollection additionalHeaders = null, NameValueCollection additionalQueryParams = null)
        {
            try
            {
                using (var client = new WebClient())
                {
                    client.Headers.Add("Authorization", this.GenerateTransientToken());
                    client.Headers.Add("Content-Type", @"application/json");

                    if (requestType == HttpVerb.Patch.ToString() || requestType == HttpVerb.Delete.ToString())
                    {
                        client.Headers.Add("If-Match", "*");
                    }

                    if (additionalHeaders != null)
                    {
                        foreach (string key in additionalHeaders.Keys)
                        {
                            client.Headers.Add(key, additionalHeaders[key]);
                        }
                    }

                    if (additionalQueryParams != null)
                    {
                        foreach (string key in additionalQueryParams.Keys)
                        {
                            client.QueryString.Add(key, additionalQueryParams[key]);
                        }
                    }

                    client.QueryString.Add("api-version", this.apiVersion);
                    return await CleanUpResponseAsync(client.UploadStringTaskAsync(this.baseUrl + apiTemplate, requestType, data));
                }
            }
            catch (WebException ex)
            {
                throw ApimExceptionFactory.GetApimException(ex);
            }
        }

        /// <summary>
        /// For GET REST calls
        /// </summary>
        /// <param name="apiTemplate">The API template.</param>
        /// <param name="additionalHeaders">The additional headers.</param>
        /// <param name="additionalQueryParams">The additional query parameters.</param>
        /// <param name="isContentTypeJson">if set to <c>true</c> [is content type JSON].</param>
        /// <returns>The string value</returns>
        private string Download(string apiTemplate, NameValueCollection additionalHeaders = null, NameValueCollection additionalQueryParams = null, bool isContentTypeJson = true)
        {
            try
            {
                using (var client = new WebClient())
                {
                    client.Headers.Add("Authorization", this.GenerateTransientToken());
                    if (isContentTypeJson)
                    {
                        client.Headers.Add("Content-Type", @"application/json;charset=UTF-8");
                    }
                    else
                    {
                        client.Headers.Add("Content-Type", @"application/xml");
                    }

                    if (additionalHeaders != null)
                    {
                        foreach (string key in additionalHeaders.Keys)
                        {
                            client.Headers.Add(key, additionalHeaders[key]);
                        }
                    }

                    client.QueryString.Add("api-version", this.apiVersion);
                    if (additionalQueryParams != null)
                    {
                        foreach (string key in additionalQueryParams.Keys)
                        {
                            client.QueryString.Add(key, additionalQueryParams[key]);
                        }
                    }

                    return CleanUpResponse(client.DownloadString(this.baseUrl + apiTemplate));
                }
            }
            catch (WebException ex)
            {
                throw ApimExceptionFactory.GetApimException(ex);
            }
        }

        /// <summary>
        /// Downloads the asynchronous.
        /// </summary>
        /// <param name="apiTemplate">The API template.</param>
        /// <param name="additionalHeaders">The additional headers.</param>
        /// <param name="additionalQueryParams">The additional query parameters.</param>
        /// <returns>The string value</returns>
        private async Task<string> DownloadAsync(string apiTemplate, NameValueCollection additionalHeaders = null, NameValueCollection additionalQueryParams = null)
        {
            try
            {
                using (var client = new WebClient())
                {
                    client.Headers.Add("Authorization", this.GenerateTransientToken());
                    client.Headers.Add("Content-Type", @"application/json");
                    if (additionalHeaders != null)
                    {
                        foreach (string key in additionalHeaders.Keys)
                        {
                            client.Headers.Add(key, additionalHeaders[key]);
                        }
                    }

                    client.QueryString.Add("api-version", this.apiVersion);
                    if (additionalQueryParams != null)
                    {
                        foreach (string key in additionalQueryParams.Keys)
                        {
                            client.QueryString.Add(key, additionalQueryParams[key]);
                        }
                    }

                    return await CleanUpResponseAsync(client.DownloadStringTaskAsync(new Uri(this.baseUrl + apiTemplate)));
                }
            }
            catch (WebException ex)
            {
                throw ApimExceptionFactory.GetApimException(ex);
            }
        }

        /// <summary>
        /// Generates the transient token.
        /// </summary>
        /// <returns>The string value</returns>
        private string GenerateTransientToken()
        {
            lock (this.lockobj)
            {
                if (DateTime.UtcNow.AddSeconds(60) >= this.tokenExpiryTime)
                {
                    using (var encoder = new HMACSHA512(Encoding.UTF8.GetBytes(this.apimKey)))
                    {
                        this.tokenExpiryTime = DateTime.UtcNow.AddMinutes(this.tokenTtl);
                        var dataToSign = this.apimId + "\n" + this.tokenExpiryTime.ToString("O", CultureInfo.InvariantCulture);
                        var hash = encoder.ComputeHash(Encoding.UTF8.GetBytes(dataToSign));
                        var signature = Convert.ToBase64String(hash);
                        this.encodedToken = string.Format(CultureInfo.InvariantCulture, "SharedAccessSignature uid={0}&ex={1:o}&sn={2}", this.apimId, this.tokenExpiryTime, signature);
                        return this.encodedToken;
                    }
                }
                else
                {
                    return this.encodedToken;
                }
            }
        }

        #endregion
    }
}
