﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement
{
    using System;
    using System.Security.Cryptography;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Validation;

    internal class ApiManagementStorage : IApiManagementStorage
    {        
        private readonly ApiManagementClient apimClient;
        private readonly string groupId;
        private readonly string productId;

        public ApiManagementStorage(string groupId, string productId, ApiManagementClient apimClient)
        {
            this.groupId = groupId;
            this.productId = productId;
            this.apimClient = apimClient;
            Agent.Instance.LogEvent(EventType.Information, $"Connected to ApiM {this.apimClient.ApiEndpointDomain}");
        }

        public ApiManagementIdentity AddNewServiceKeys(AccountIdentifier accountIdentifier, out ApiManagementKeys keys)
        {
            var apiManagementSubscriptionId = CreateRandomId();
            var primaryKey = CreateKey();
            var secondaryKey = CreateKey();

            return this.WriteServiceKeys(accountIdentifier, apiManagementSubscriptionId, out keys, primaryKey, secondaryKey);
        }

        public ApiManagementIdentity WriteServiceKeys(
            AccountIdentifier accountIdentifier,
            string apiManagementSubscriptionId,
            out ApiManagementKeys keys,
            ApiManagementKey primaryKey,
            ApiManagementKey secondaryKey)
        {
            // TODO:  what if the account already has an apim user/subscription?  this will associate another and orphan the original
            // TODO:  for now, assume account.ApiManagementId is null
            using (var op = new Instrumentation.ResourceProviderApimOperation(nameof(this.WriteServiceKeys), apiManagementSubscriptionId))
            {
                var userId = CreateRandomId();

                // FirstName must be subscriptionId
                // LastName must be resourceGroupName :: resourceName
                // FirstName :: LastName is passed to proxy as account-id query-parameter
                // This is needed for account specific logging and metrics for Azure
                this.apimClient.CreateUser(userId,
                                       accountIdentifier.SubscriptionId,
                                       $"{accountIdentifier.ResourceGroupName}::{accountIdentifier.ResourceName}",
                                       $"{userId}@locationservices.microsoft.com");

                this.apimClient.AddUserToGroup(userId, this.groupId);

                keys = new ApiManagementKeys(primaryKey, secondaryKey);
                var subscription = this.apimClient.CreateProductSubscription(userId, this.productId, null, null, apiManagementSubscriptionId, primaryKey.Key, secondaryKey.Key);

                op.AddNewUser(userId, subscription.Id, this.productId, this.groupId);
                op.SetResultSuccess();

                return new ApiManagementIdentity(subscription.Id);
            }
        }

        public void PatchUser(ApiManagementIdentity apimIdentity, AccountIdentifier accountIdentifier)
        {
            using (var op = new Instrumentation.ResourceProviderApimOperation(nameof(this.PatchUser), apimIdentity.SubscriptionId))
            {
                var azureSubscriptionId = accountIdentifier.SubscriptionId;
                var resourceGroupName = accountIdentifier.ResourceGroupName;
                var resourceName = accountIdentifier.ResourceName;

                var userId = this.apimClient.GetSubscription(apimIdentity.SubscriptionId).UserId;

                this.apimClient.UpdateUser(userId, 
                                       azureSubscriptionId,
                                       $"{resourceGroupName}::{resourceName}",
                                       $"{userId}@locationservices.microsoft.com",
                                       UserState.Active);

                op.SetResultSuccess();
            }
        }

        public ApiManagementKeys GetServiceKeys(ApiManagementIdentity id)
        {
            using (var op = new Instrumentation.ResourceProviderApimOperation("GetServiceKeys", id.SubscriptionId))
            {
                var subscription = this.apimClient.GetSubscription(id.SubscriptionId);
                op.SetResultSuccess();
                return new ApiManagementKeys(subscription.PrimaryKey, subscription.SecondaryKey, DateTime.UtcNow, DateTime.UtcNow);
            }
        }

        public ApiManagementKey RegeneratePrimaryKey(ApiManagementIdentity id)
        {
            using (var op = new Instrumentation.ResourceProviderApimOperation("RegeneratePrimaryKey", id.SubscriptionId))
            {
                var newKey = CreateKey();

                // TODO: failure?
                this.apimClient.UpdateSpecificSubscription(id.SubscriptionId, null, newKey.Key);
                op.SetResultSuccess();
                return newKey;
            }
        }

        public ApiManagementKey RegenerateSecondaryKey(ApiManagementIdentity id)
        {
            using (var op = new Instrumentation.ResourceProviderApimOperation("RegenerateSecondaryKey", id.SubscriptionId))
            {
                var newKey = CreateKey();

                // TODO: failure?
                this.apimClient.UpdateSpecificSubscription(id.SubscriptionId, null, null, newKey.Key);
                op.SetResultSuccess();
                return newKey;
            }
        }

        public void DeleteUser(ApiManagementIdentity id)
        {
            using (var op = new Instrumentation.ResourceProviderApimOperation(nameof(this.DeleteUser), id.SubscriptionId))
            {
                var sub = this.apimClient.GetSubscription(id.SubscriptionId);
                if (!string.IsNullOrWhiteSpace(sub.UserId))
                {
                    this.apimClient.DeleteUser(sub.UserId);
                }

                op.SetResultSuccess();
            }
        }

        public bool UserExists(ApiManagementIdentity id)
        {
            return this.apimClient.TryGetSubscription(id.SubscriptionId, out _);
        }

        /// <summary>
        /// Create a cryptographically random key and url encode it
        /// (currently a 32 byte value)
        /// </summary>
        /// <returns>ApiManagementKey with key created at UtcNow</returns>
        private static ApiManagementKey CreateKey()
        {
            using (RandomNumberGenerator rng = new RNGCryptoServiceProvider())
            {
                var tokenData = new byte[32];
                rng.GetBytes(tokenData);

                return new ApiManagementKey(IdentityModel.Tokens.Base64UrlEncoder.Encode(tokenData), DateTime.UtcNow);
            }
        }

        /// <summary>
        /// Create a cryptographically random string (for user or subscription id)
        /// (currently 16 bytes written as hex)
        /// </summary>
        /// <returns></returns>
        private static string CreateRandomId()
        {
            using (RandomNumberGenerator rng = new RNGCryptoServiceProvider())
            {
                var tokenData = new byte[16];
                rng.GetBytes(tokenData);

                return string.Concat(Array.ConvertAll(tokenData, x => x.ToString("x2")));
            }
        }
    }
}