﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement
{
    using System.Globalization;
    using System.IO;
    using System.Xml;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement.Models;

    /// <summary>
    /// The PolicyManager class
    /// </summary>
    public class PolicyManager
    {
        /// <summary>
        /// The policy XML
        /// </summary>
        private string policyXml = string.Empty;

        /// <summary>
        /// The doc
        /// </summary>
        private XmlDocument doc = new XmlDocument();

        /// <summary>
        /// The root
        /// </summary>
        private XmlNode root;

        /// <summary>
        /// Initializes a new instance of the <see cref="PolicyManager" /> class.
        /// </summary>
        /// <param name="policyXml">The policy XML.</param>
        public PolicyManager(string policyXml)
        {
            this.policyXml = policyXml;
            if (!string.IsNullOrEmpty(this.policyXml))
            {
                this.doc.LoadXml(policyXml);
                this.root = this.doc.DocumentElement;
            }
        }

        /// <summary>
        /// Sets the policy XML.
        /// </summary>
        /// <param name="nodeName">Name of the node.</param>
        /// <param name="calls">The calls.</param>
        /// <param name="renewalPeriod">The renewal period.</param>
        /// <returns>The string value</returns>
        public string SetPolicyXml(string nodeName, int calls, int renewalPeriod)
        {
            XmlNodeType type = XmlNodeType.Element;

            if (string.IsNullOrEmpty(this.policyXml))
            {
                this.root = this.doc.CreateNode(type, "policies", string.Empty);
                this.doc.AppendChild(this.root);
                XmlNode inBaseNode = this.doc.CreateNode(type, "inbound", string.Empty);
                XmlNode baseNode = this.doc.CreateNode(type, "base", string.Empty);
                inBaseNode.AppendChild(baseNode);
                this.root.AppendChild(inBaseNode);
                XmlNode outBaseNode = this.doc.CreateNode(type, "outbound", string.Empty);
                XmlNode baseNode1 = this.doc.CreateNode(type, "base", string.Empty);
                outBaseNode.AppendChild(baseNode1);
                this.root.AppendChild(outBaseNode);
                this.root = this.doc.DocumentElement;
                this.SetCallNodeValueInXml(nodeName, calls, renewalPeriod);
            }
            else
            {
                this.doc.LoadXml(this.policyXml);
                this.root = this.doc.DocumentElement;
                this.SetCallNodeValueInXml(nodeName, calls, renewalPeriod);

                XmlNode node = this.root.SelectSingleNode("inbound/base");
                if (node == null)
                {
                    XmlNode inboundNode = this.root.SelectSingleNode("policy/inbound");
                    XmlNode baseNode = this.doc.CreateNode(type, "base", string.Empty);
                    inboundNode.AppendChild(baseNode);
                }

                node = this.root.SelectSingleNode("outbound/base");
                if (node == null)
                {
                    XmlNode inboundNode = this.root.SelectSingleNode("policy/outbound");
                    XmlNode baseNode = this.doc.CreateNode(type, "base", string.Empty);
                    inboundNode.AppendChild(baseNode);
                }
            }

            XmlWriterSettings settings = new XmlWriterSettings();
            settings.OmitXmlDeclaration = true;
            settings.ConformanceLevel = ConformanceLevel.Fragment;
            settings.CloseOutput = false;

            using (var stringWriter = new StringWriter(CultureInfo.InvariantCulture))
            {
                using (var xmlTextWriter = XmlWriter.Create(stringWriter, settings))
                {
                    this.doc.WriteTo(xmlTextWriter);
                    xmlTextWriter.Flush();
                }

                return stringWriter.GetStringBuilder().ToString();
            }
        }

        /// <summary>
        /// Gets the quota and rate limit.
        /// </summary>
        /// <returns>The APIM Policy</returns>
        public ApimPolicy GetQuotaAndRatelimit()
        {
            if (this.root == null)
            {
                return null;
            }

            ApimPolicy policy = new ApimPolicy();
            int calls = 0;
            int renewalPeriod = 0;
            XmlNode quotaNode = this.root.SelectSingleNode("inbound/quota");

            if (quotaNode != null)
            {
                int.TryParse(quotaNode.Attributes["calls"].InnerText, out calls);
                int.TryParse(quotaNode.Attributes["renewal-period"].InnerText, out renewalPeriod);
                policy.QuotaCalls = calls;
                policy.QuotaRenewalPeriod = renewalPeriod;
            }

            XmlNode ratelimitNode = this.root.SelectSingleNode("inbound/rate-limit");
            if (ratelimitNode != null)
            {
                int.TryParse(ratelimitNode.Attributes["calls"].InnerText, out calls);
                int.TryParse(ratelimitNode.Attributes["renewal-period"].InnerText, out renewalPeriod);
                policy.RateLimitCalls = calls;
                policy.RateLimitRenewalPeriod = renewalPeriod;
            }

            return policy;
        }

        /// <summary>
        /// Sets the call node value in XML.
        /// </summary>
        /// <param name="nodeName">Name of the node.</param>
        /// <param name="calls">The calls.</param>
        /// <param name="renewalPeriod">The renewal period.</param>
        public void SetCallNodeValueInXml(string nodeName, int calls, int renewalPeriod)
        {
            XmlNode callsParentNode = this.root.SelectSingleNode("inbound/" + nodeName);

            if (callsParentNode == null)
            {
                XmlElement element = this.doc.CreateElement(nodeName);
                element.Attributes.Append(this.doc.CreateAttribute("calls"));
                element.Attributes.Append(this.doc.CreateAttribute("renewal-period"));

                XmlNode inboundNode = this.root.SelectSingleNode("inbound");
                inboundNode.AppendChild(element);
                callsParentNode = this.root.SelectSingleNode("inbound/" + nodeName);
                callsParentNode.Attributes["calls"].InnerText = (calls != 0) ? calls.ToString(CultureInfo.InvariantCulture) : "0";
                callsParentNode.Attributes["renewal-period"].InnerText = renewalPeriod.ToString(CultureInfo.InvariantCulture);
            }
            else
            {
                callsParentNode.Attributes["calls"].InnerText = calls.ToString(CultureInfo.InvariantCulture);
                callsParentNode.Attributes["renewal-period"].InnerText = renewalPeriod.ToString(CultureInfo.InvariantCulture);
            }
        }
    }
}
