﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI
{
    using System;
    using System.Security.Cryptography.X509Certificates;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ARM;
    using Validation;

    internal class ClientCertificateResolver : IClientCertificateResolver
    {
        private const string ClientAuthCertificateConfigKey = "ClientAuth_Certificate_Thumbprint";
        
        private readonly X509Certificate2Collection _allowedCertificates = new X509Certificate2Collection();

        private readonly ReaderWriterLockSlim _allowedCertificatesLock = new ReaderWriterLockSlim();
        private readonly ArmCertificateReader _armCertificateReader;
        private readonly string _clientAuthCertificateThumbprint;
        private bool _disposed;
        
        public ClientCertificateResolver(IBootstrapConfiguration bootstrapConfiguration,
                                         ArmCertificateReader armCertificateReader)
        {
            Requires.NotNull(armCertificateReader, nameof(armCertificateReader));
            Requires.NotNull(bootstrapConfiguration, nameof(bootstrapConfiguration));
            _armCertificateReader = armCertificateReader;
            _clientAuthCertificateThumbprint = bootstrapConfiguration.GetValue(ClientAuthCertificateConfigKey);
        }

        ~ClientCertificateResolver()
        {
            Dispose(false);
        }

        /// <summary>
        ///     Since this is a global singleton, fully implementing the Dispose pattern isn't really necessary.
        ///     But do it anyway for good hygiene.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        public async Task FetchClientAuthCertificatesAsync(CancellationToken cancellationToken)
        {
            while (!cancellationToken.IsCancellationRequested)
            {
                // try asking ARM for client certs
                X509Certificate2Collection armCertificates = null;
                while (true)
                {
                    armCertificates = await _armCertificateReader.DownloadCertificates();
                    if (armCertificates == null)
                    {
                        // retry
                        await Task.Delay(new TimeSpan(0, 1, 0), cancellationToken);
                        if (cancellationToken.IsCancellationRequested)
                        {
                            return;
                        }

                        // TODO: count retries and throw fatal error?
                    }
                    else
                    {
                        break;
                    }
                }

                X509Certificate2Collection personalCertificates = null;
                LookforPersonalCertificates(ref personalCertificates);

                // add these certs to the allowed collection
                try
                {
                    _allowedCertificatesLock.TryEnterWriteLock(-1);
                    _allowedCertificates.Clear();
                    if (personalCertificates != null)
                    {
                        _allowedCertificates.AddRange(personalCertificates);
                        Instrumentation.Agent.Instance.LogEvent(Instrumentation.EventType.Information, $"Adding {personalCertificates.Count} dev client certificates");
                    }

                    if (armCertificates != null)
                    {
                        _allowedCertificates.AddRange(armCertificates);
                        Instrumentation.Agent.Instance.LogEvent(Instrumentation.EventType.Information, $"Adding {armCertificates.Count} client certificates");
                    }
                }
                finally
                {
                    _allowedCertificatesLock.ExitWriteLock();
                }

                await Task.Delay(new TimeSpan(1, 0, 0, 0), cancellationToken);
            }
        }

        public bool ValidateClientCertificate(X509Certificate2 clientCertificate)
        {
            if (clientCertificate != null)
            {
                try
                {
                    _allowedCertificatesLock.TryEnterReadLock(-1);

                    if (!clientCertificate.Verify())
                    {
                        return false;
                    }

                    foreach (var allowedCertificate in _allowedCertificates)
                    {
                        if (clientCertificate.Equals(allowedCertificate))
                        {
                            return true;
                        }
                    }
                }
                finally
                {
                    _allowedCertificatesLock.ExitReadLock();
                }
            }

            return false;
        }

        protected virtual void Dispose(bool disposing)
        {
            if (_disposed)
            {
                return;
            }

            if (disposing)
            {
                // Free any other managed objects here.
                _allowedCertificatesLock.Dispose();
            }

            // Free any unmanaged objects here.
            _disposed = true;
        }

        private void LookforPersonalCertificates(ref X509Certificate2Collection personalCertificates)
        {
            // look for any personal certs
            if (!string.IsNullOrEmpty(_clientAuthCertificateThumbprint))
            {
                X509Store certStore = null;
                try
                {
                    certStore = new X509Store(StoreName.My, StoreLocation.LocalMachine);
                    certStore.Open(OpenFlags.ReadOnly);
                    personalCertificates = certStore.Certificates.Find(X509FindType.FindByThumbprint,
                                                                       _clientAuthCertificateThumbprint,
                                                                       false);
                }
                finally
                {
                    if (certStore != null)
                    {
                        certStore.Close();
                    }
                }
            }
        }
    }
}