﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Controllers
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement.Contracts.ResponseContract;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;

    [Route(@"subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{providerName:regex(^Microsoft\.Maps$)}/accounts/{resourceName}")]
    public class AccountController : ResourceProviderControllerBase
    {
        private readonly IAccountRecordStorage accountRecordStorage;
        private readonly IApiManagementStorageFactory apiManagementStorageFactory;
        
        public AccountController(IAccountRecordStorage accountRecordStorage,
                                 IApiManagementStorageFactory apiManagementStorageFactory)
        {
            this.accountRecordStorage = accountRecordStorage;
            this.apiManagementStorageFactory = apiManagementStorageFactory;
        }

        [HttpGet]
        public async Task<HttpResponseMessage> Get([FromUri] string subscriptionId,
                                       [FromUri] string resourceGroupName,
                                       [FromUri] string providerName,
                                       [FromUri] string resourceName,
                                       [FromUri(Name = "api-version")] string apiVersion = "")
        {
            HttpResponseMessage response;
            try
            {
                Operation.LogCommonParameters(subscriptionId, resourceGroupName, resourceName);

                // these will throw a ResourceProviderException on failure
                ValidateAuthorized();
                ValidateApiVersion(apiVersion);
                ValidateCommonParameters(subscriptionId, resourceGroupName, resourceName);
                ValidateModelParse();

                AccountIdentifier id = new AccountIdentifier(subscriptionId, resourceGroupName, resourceName);
                var accountResult = await GetWorker(id);

                if (accountResult != null)
                {
                    ExternalContracts.Account externalResult = new ExternalContracts.Account(accountResult, providerName);
                    response = Request.CreateResponse(HttpStatusCode.OK, externalResult);
                    Operation.SetResultSuccess();
                }
                else
                {
                    Operation.SetResultClientError("notfound", null);
                    response = Request.CreateResponse(HttpStatusCode.NotFound);
                }
            }
            catch (ResourceProviderException ex)
            {
                // these are errors which get converted to nice return codes and messages
                // let unknown exceptions fall through - they will be handled by global hook
                response = this.BuildErrorResponseAndLog(ex);
            }

            return response;
        }

        [HttpPatch]
        public async Task<HttpResponseMessage> Patch([FromUri] string subscriptionId,
                                         [FromUri] string resourceGroupName,
                                         [FromUri] string providerName,
                                         [FromUri] string resourceName,
                                         [FromBody] ExternalContracts.Account body,
                                         [FromUri(Name = "api-version")] string apiVersion = "")
        {
            HttpResponseMessage response;
            try
            {
                Operation.LogCommonParameters(subscriptionId, resourceGroupName, resourceName);

                // these will throw a ResourceProviderException on failure
                ValidateAuthorized();
                ValidateApiVersion(apiVersion);
                ValidateCommonParameters(subscriptionId, resourceGroupName, resourceName);
                ValidateModelParse();

                // validate there at least is a body.  Other errors are caught below when building the internal Account object
                if (body == null)
                {
                    throw new ResourceProviderException(ResourceProviderException.ExceptionType.ClientParams, "LocationServicesAccount");
                }

                // get the specified account
                AccountIdentifier id = new AccountIdentifier(subscriptionId, resourceGroupName, resourceName);

                // the creation of an internal Account does many validations that are client actionable,
                // so wrap them all and catch any ArgumentException and rethrow as ResourceProviderException to get the common handling
                Account accountResult;
                try
                {
                    // transform the input account info into internal representation (note the identity here is not used; the Account object is just used to instantiate the other parts of an account)
                    var accountChanges = new Account(body);
                    accountResult = await PatchWorker(id, accountChanges);
                }
                catch (ArgumentException ex)
                {
                    throw new ResourceProviderException(ex);
                }

                if (accountResult != null)
                {
                    ExternalContracts.Account externalResult = new ExternalContracts.Account(accountResult, providerName);
                    response = Request.CreateResponse(HttpStatusCode.OK, externalResult);
                    Operation.SetResultSuccess();
                }
                else
                {
                    Operation.SetResultClientError("notfound", null);
                    response = Request.CreateResponse(HttpStatusCode.NotFound);
                }
            }
            catch (ResourceProviderException ex)
            {
                // these are errors which get converted to nice return codes and messages
                // let unknown exceptions fall through - they will be handled by global hook
                response = this.BuildErrorResponseAndLog(ex);
            }

            return response;
        }

        [HttpDelete]
        public async Task<HttpResponseMessage> Delete([FromUri] string subscriptionId,
                                          [FromUri] string resourceGroupName,
                                          [FromUri] string resourceName,
                                          [FromUri(Name = "api-version")] string apiVersion = "")
        {
            HttpResponseMessage response;
            try
            {
                Operation.LogCommonParameters(subscriptionId, resourceGroupName, resourceName);

                // these will throw a ResourceProviderException on failure
                ValidateAuthorized();
                ValidateApiVersion(apiVersion);
                ValidateCommonParameters(subscriptionId, resourceGroupName, resourceName);
                ValidateModelParse();

                AccountIdentifier id = new AccountIdentifier(subscriptionId, resourceGroupName, resourceName);

                if (await DeleteWorker(id))
                {
                    response = Request.CreateResponse(HttpStatusCode.OK);
                    Operation.SetResultSuccess();
                }
                else
                {
                    Operation.SetResultClientError("notfound", null);
                    response = Request.CreateResponse(HttpStatusCode.NoContent);
                }
            }
            catch (ResourceProviderException ex)
            {
                // these are errors which get converted to nice return codes and messages
                // let unknown exceptions fall through - they will be handled by global hook
                response = this.BuildErrorResponseAndLog(ex);
            }

            return response;
        }

        [HttpPut]
        public async Task<HttpResponseMessage> Put([FromUri] string subscriptionId,
                                       [FromUri] string resourceGroupName,
                                       [FromUri] string providerName,
                                       [FromUri] string resourceName,
                                       [FromBody] ExternalContracts.Account body,
                                       [FromUri(Name = "api-version")] string apiVersion = "")
        {
            HttpResponseMessage response;
            try
            {
                Operation.LogCommonParameters(subscriptionId, resourceGroupName, resourceName);

                // these will throw a ResourceProviderException on failure
                ValidateAuthorized();
                ValidateApiVersion(apiVersion);
                ValidateCommonParameters(subscriptionId, resourceGroupName, resourceName);
                ValidateModelParse();

                // validate there at least is a body.  Other errors are caught below when building the internal Account object
                if (body == null)
                {
                    throw new ResourceProviderException(ResourceProviderException.ExceptionType.ClientParams, "LocationServicesAccount");
                }

                // the creation of an internal Account does many validations that are client actionable,
                // so wrap them all and catch any ArgumentException and rethrow as ResourceProviderException to get the common handling
                Account account;
                bool isUpdate;
                Account accountResult;
                try
                {
                    // transform the account info into internal representation
                    account = new Account(body, subscriptionId, resourceGroupName, resourceName);
                    AccountRecord existingAccount = await accountRecordStorage.GetAccountRecord(account.Id);

                    isUpdate = existingAccount != null ? true : false;
                    accountResult = await PutWorker(account, isUpdate, existingAccount);
                }
                catch (ArgumentException ex)
                {
                    throw new ResourceProviderException(ex);
                }

                ExternalContracts.Account externalResult = new ExternalContracts.Account(accountResult, providerName);                
                response = Request.CreateResponse(isUpdate ? HttpStatusCode.OK : HttpStatusCode.Created, externalResult);

                Operation.SetResultSuccess();
            }
            catch (ResourceProviderException ex)
            {
                // these are errors which get converted to nice return codes and messages
                // let unknown exceptions fall through - they will be handled by global hook
                response = this.BuildErrorResponseAndLog(ex);
            }

            return response;
        }

        /// <summary>
        /// Deletes all records matching the given account identifier
        /// </summary>
        /// <param name="id">Identifier for LBS account</param>
        /// <returns>A value indicating whether deletion occurred</returns>
        private async Task<bool> DeleteWorker(AccountIdentifier id)
        {
            IEnumerable<AccountRecord> accountRecords = this.accountRecordStorage.GetAccountRecordsByResourceGroup(id.SubscriptionId, id.ResourceGroupName)
                .Where(r => r.ResourceName.Equals(id.ResourceName, StringComparison.OrdinalIgnoreCase));

            var hadRecord = false;

            foreach (var accountRecord in accountRecords)
            {
                // convert to internal account
                var account = new Account(accountRecord);

                try
                {
                    // delete the keys
                    IApiManagementStorage apiManagementStorage = this.apiManagementStorageFactory.GetApiManagementStorage(account.Sku.Name);
                    apiManagementStorage.DeleteUser(account.ApiManagementId);
                }
                catch (ApimException e)
                {
                    if (e.ExceptionType != ApimExceptionType.ResourceNotFound)
                    {
                        throw;
                    }
                }

                // delete the account record
                await accountRecordStorage.DeleteAccountRecord(id);

                hadRecord = true;
            }
            
            return hadRecord;
        }

        private async Task<Account> PutWorker(Account account, bool update, AccountRecord existingAccount)
        {
            // see if there is already a record for this account
            AccountRecord updatedRecord = null;
            if (!update)
            {
                IApiManagementStorage apiManagementStorage = this.apiManagementStorageFactory.GetApiManagementStorage(account.Sku.Name);

                ApiManagementKeys keys = null;

                // generate ApiManagement subscription and attach to account
                account.ApiManagementId = apiManagementStorage.AddNewServiceKeys(account.Id, out keys);                
                account.CreateNewClientId();

                // build the persistent record
                updatedRecord = new AccountRecord(account, keys);
            }
            else
            {
                var oldAccount = new Account(existingAccount);
                if (oldAccount.Sku.Name != account.Sku.Name)
                {
                    throw new ArgumentException($"Changing SKUs from {oldAccount.Sku.Name} to {account.Sku.Name} is not supported", nameof(account.Sku.Name));
                }

                if (oldAccount.Location.Location != account.Location.Location)
                {
                    throw new ArgumentException($"Changing Location from {oldAccount.Location.LocationString} to {oldAccount.Location.LocationString} is not supported", nameof(account.Location.LocationString));
                }

                if (string.IsNullOrWhiteSpace(oldAccount.ClientId))
                {
                    // backfill existing accounts
                    oldAccount.CreateNewClientId();
                    existingAccount.ClientId = oldAccount.ClientId;
                }

                // create a new account record based on the incoming data (with the previous key and previous identity)
                updatedRecord = new AccountRecord(account, existingAccount);               
                updatedRecord.ClientId = existingAccount.ClientId;                
            }

            // send persistent record to storage (this will rewrite the other properties if the account already existed, which is what we want)
            // TODO: check for error!
            await accountRecordStorage.PutAccountRecord(updatedRecord);

            // convert back to external format
            return new Account(updatedRecord);
        }

        private async Task<Account> GetWorker(AccountIdentifier id)
        {
            AccountRecord resultRecord = await accountRecordStorage.GetAccountRecord(id);

            if (resultRecord != null && string.IsNullOrWhiteSpace(resultRecord.ClientId))
            {
                // backfill the client id record.
                Account account = new Account(resultRecord);
                account.CreateNewClientId();
                resultRecord.ClientId = account.ClientId;
                await accountRecordStorage.PutAccountRecord(resultRecord);
                return new Account(resultRecord);
            }
            else if (resultRecord != null)
            {
                // convert to internal account
                return new Account(resultRecord);
            }

            return null;
        }

        private async Task<Account> PatchWorker(AccountIdentifier id, Account accountChanges)
        {
            AccountRecord originalRecord = await accountRecordStorage.GetAccountRecord(id);

            if (originalRecord != null)
            {
                Account accountOriginal = new Account(originalRecord);

                // TODO: validate!

                // copy fields that are updatable
                if (accountChanges.Tags != null)
                {
                    accountOriginal.Tags = accountChanges.Tags;
                }

                if (accountChanges.Sku != null)
                {
                    if (accountOriginal.Sku.Name != accountChanges.Sku.Name)
                    {
                        throw new ArgumentException($"Changing SKUs from {accountOriginal.Sku.Name} to {accountChanges.Sku.Name} is not supported", nameof(accountChanges.Sku.Name));
                    }
                }

                if (string.IsNullOrWhiteSpace(accountOriginal.ClientId))
                {
                    // backfill existing accounts
                    accountOriginal.CreateNewClientId();
                    originalRecord.ClientId = accountOriginal.ClientId;
                }

                // TODO: validate!
                // create the new record and save it
                // TODO: this is an IN PLACE write of new data.  does it have to be journaled?
                var changedRecord = new AccountRecord(accountChanges, originalRecord);
                await accountRecordStorage.PutAccountRecord(changedRecord);

                // convert back to external format
                return new Account(changedRecord);
            }
            else
            {
                return null;
            }
        }
    }
}