﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Controllers
{
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Web.Http;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;

    [Route(@"subscriptions/{subscriptionId}/providers/{providerName:regex(^Microsoft\.Maps$)}/accounts")]
    public class AccountsBySubscriptionController : ResourceProviderControllerBase
    {
        private readonly IAccountRecordStorage accountRecordStorage;

        public AccountsBySubscriptionController(IAccountRecordStorage accountRecordStorage)
        {
            this.accountRecordStorage = accountRecordStorage;
        }

        [HttpGet]
        public HttpResponseMessage Get([FromUri] string subscriptionId,
                                       [FromUri] string providerName,
                                       [FromUri(Name = "api-version")] string apiVersion = "")
        {
            HttpResponseMessage response;
            try
            {
                Operation.LogCommonParameters(subscriptionId);

                // these will throw a ResourceProviderException on failure
                ValidateAuthorized();
                ValidateApiVersion(apiVersion);
                ValidateCommonParameters(subscriptionId);
                ValidateModelParse();

                ExternalContracts.MultipleAccounts accounts = new ExternalContracts.MultipleAccounts()
                                                              {
                                                                  Value = new List<ExternalContracts.Account>()
                                                              };
                foreach (var account in GetWorker(subscriptionId))
                {
                    var externalAccountResult = new ExternalContracts.Account(account, providerName);

                    accounts.Value.Add(externalAccountResult);
                }

                response = Request.CreateResponse(HttpStatusCode.OK, accounts);
                Operation.SetResultSuccess();
            }
            catch (ResourceProviderException ex)
            {
                // these are errors which get converted to nice return codes and messages
                // let unknown exceptions fall through - they will be handled by global hook
                response = this.BuildErrorResponseAndLog(ex);
            }

            return response;
        }

        internal IEnumerable<Account> GetWorker(string subscriptionId)
        {
            return this.accountRecordStorage
                .GetAccountRecordsBySubscription(subscriptionId)
                .DistinctBy(record => $"{record.ResourceGroupName}/{record.ResourceName}")
                .Select(record => new Account(record));
        }
    }
}
