﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Controllers
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;

    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;

    [Route("subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/moveResources")]
    public class AccountsMoveController : ResourceProviderControllerBase
    {
        private readonly IAccountRecordStorage accountRecordStorage;

        private readonly IApiManagementStorageFactory apiManagementStorageFactory;

        public AccountsMoveController(IAccountRecordStorage accountRecordStorage, IApiManagementStorageFactory apiManagementStorageFactory)
        {
            this.accountRecordStorage = accountRecordStorage;
            this.apiManagementStorageFactory = apiManagementStorageFactory;
        }

        [HttpPost]
        public async Task<HttpResponseMessage> Post([FromUri] string subscriptionId,
                                        [FromUri] string resourceGroupName,
                                        [FromBody] ExternalContracts.MoveRequest body,
                                        [FromUri(Name = "api-version")] string apiVersion = "")
        {
            HttpResponseMessage response;
            try
            {
                Operation.LogCommonParameters(subscriptionId, resourceGroupName);

                // these will throw a ResourceProviderException on failure
                ValidateAuthorized();
                ValidateApiVersion(apiVersion);
                ValidateCommonParameters(subscriptionId, resourceGroupName);
                ValidateModelParse();

                if (body == null)
                {
                    throw new ResourceProviderException(ResourceProviderException.ExceptionType.ClientParams, "LocationServicesAccountsMoveRequest");
                }

                // since there are so many validations here, wrap them all and catch any ArgumentException and rethrow as ResourceProviderException
                // to get the common handling
                List<AccountIdentifier> resourceList;
                string destinationSubscriptionId;
                string destinationResourceGroup;
                try
                {
                    Validation.Requires.NotNullOrWhiteSpace(body.TargetResourceGroup, "LocationServicesAccountsMoveRequest.targetResourceGroup");

                    // parse the destination - it has a very specific format:
                    // "/subscriptions/{targetId}/resourceGroups/{targetName}"

                    // strip the beginning '/' if it exists
                    string targetResourceGroupString = body.TargetResourceGroup;
                    Validation.Requires.Argument(targetResourceGroupString.StartsWith("/"), nameof(body.TargetResourceGroup), $"{nameof(body.TargetResourceGroup)} is not correctly formatted");
                    targetResourceGroupString = targetResourceGroupString.Substring(1);

                    string[] tokens = targetResourceGroupString.Split('/');
                    Validation.Requires.That(tokens.Length == 4, "LocationServicesAccountsMoveRequest.targetResourceGroup", "targetResourceGroup format");
                    Validation.Requires.That(
                        string.Compare(tokens[0], "subscriptions", StringComparison.InvariantCultureIgnoreCase) == 0,
                        "LocationServicesAccountsMoveRequest.targetResourceGroup",
                        "targetResourceGroup format");
                    Validation.Requires.That(
                        string.Compare(tokens[2], "resourceGroups", StringComparison.InvariantCultureIgnoreCase) == 0,
                        "LocationServicesAccountsMoveRequest.targetResourceGroup",
                        "targetResourceGroup format");
                    Validation.Requires.NotNullOrWhiteSpace(tokens[1], "LocationServicesAccountsMoveRequest.targetResourceGroup");
                    Validation.Requires.NotNullOrWhiteSpace(tokens[3], "LocationServicesAccountsMoveRequest.targetResourceGroup");

                    destinationSubscriptionId = tokens[1];
                    destinationResourceGroup = tokens[3];

                    // TODO: these will fail with wrong parameters names in the exception
                    ValidateCommonParameters(destinationSubscriptionId, destinationResourceGroup);

                    Validation.Requires.NotNullEmptyOrNullElements(body.Resources, "resources");

                    // do this here because creating AccountIdentifiers will validate and throw
                    resourceList = body.Resources.Select(s => new AccountIdentifier(s)).ToList();
                }
                catch (ArgumentException ex)
                {
                    throw new ResourceProviderException(ex);
                }

                await PostWorker(destinationSubscriptionId, destinationResourceGroup, resourceList);

                response = Request.CreateResponse(HttpStatusCode.OK);
                Operation.SetResultSuccess();
            }
            catch (ResourceProviderException ex)
            {
                // these are errors which get converted to nice return codes and messages
                // let unknown exceptions fall through - they will be handled by global hook
                response = this.BuildErrorResponseAndLog(ex);
            }

            return response;
        }

        internal async Task PostWorker(string destinationSubscriptionId, string destinationResourceGroup, List<AccountIdentifier> resourceList)
        {
            foreach (AccountIdentifier id in resourceList)
            {
                var record = await accountRecordStorage.GetAccountRecord(id);

                if (record == null)
                {
                    continue;
                }

                // TODO:  can't wrap this in a transaction since the partition key is subscription,
                // TODO:  and movement can happen between subscriptions
                // turn into an account and create a new Record
                var newId = new AccountIdentifier(destinationSubscriptionId,
                                                  destinationResourceGroup,
                                                  record.ResourceName);

                var existingRecord = await this.accountRecordStorage.GetAccountRecord(newId);
                if (existingRecord != null && !existingRecord.Deleted)
                {
                    throw new ResourceProviderException(
                        new ArgumentException(
                            $"Resource: {ExternalContracts.Account.BuildExternalId(newId)} already exists."));
                }

                var account = new Account(record, newId);
                var newRecord = new AccountRecord(account, record.GetKeys());

                // (soft) delete the old
                // Order matters here because cosmos db only uses 1 record.
                record.Deleted = true;
                await accountRecordStorage.PutAccountRecord(record);

                await accountRecordStorage.PutAccountRecord(newRecord);
                var apiManagementStorage = this.apiManagementStorageFactory.GetApiManagementStorage(account.Sku.Name);
                apiManagementStorage.PatchUser(newRecord.GetApiManagementIdentity(), newId);
            }
        }
    }
}
