﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Controllers.Deprecated
{
    using System.Collections.Generic;
    using System.Net;
    using System.Net.Http;
    using System.Text;
    using System.Web.Http;

    [Route(@"providers/{providerName:regex(^Microsoft\.Location(Based)?Services$)}/operations")]
    public class DeprecatedOperationsController : ResourceProviderControllerBase
    {
        private static IDictionary<string, string> displayProviderDict;
        private static IDictionary<string, string> displayResourceDict;

        public DeprecatedOperationsController()
        {
            displayProviderDict = new Dictionary<string, string>()
            {
                { "Microsoft.LocationBasedServices", "Microsoft Location Based Services" },
                { "Microsoft.LocationServices", "Microsoft Location Services" }
            };
            displayResourceDict = new Dictionary<string, string>()
            {
                { "Microsoft.LocationBasedServices", "Location Based Services Account" },
                { "Microsoft.LocationServices", "Location Services Account" }
            };
        }

        [HttpGet]
        public HttpResponseMessage Get([FromUri] string providerName, [FromUri(Name = "api-version")] string apiVersion = "")
        {
            HttpResponseMessage response;
            try
            {
                // these will throw a ResourceProviderException on failure
                ValidateAuthorized();
                ValidateApiVersion(apiVersion);
                ValidateModelParse();

                ValidateApiVersion(apiVersion);

                response = Request.CreateResponse(HttpStatusCode.OK);
                response.Content = new StringContent(_operationsJson(providerName), Encoding.UTF8, "application/json");
                Operation.SetResultSuccess();
            }
            catch (ResourceProviderException ex)
            {
                // these are errors which get converted to nice return codes and messages
                // let unknown exceptions fall through - they will be handled by global hook
                response = this.BuildErrorResponseAndLog(ex);
            }

            return response;
        }

        // quick and dirty json formatted output.  So far, this doesn't need to be data driven or localized (?)
        private static string _operationsJson(string providerName)
        {
            var FriendlyDisplayProvider = displayProviderDict.TryGetValue(providerName, out string displayProvider) ? displayProvider : "Generic Provider";
            var FriendlyResourceName = displayResourceDict.TryGetValue(providerName, out string displayResource) ? displayResource : "Generic Resource";
            return
$@"{{
  ""value"": [
    {{
      ""name"": ""{providerName}/register/action"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""(Deprecated: Please use /providers/Microsoft.Maps) Register the provider"",
        ""description"": ""(Deprecated: Please use /providers/Microsoft.Maps) Register the provider""
      }}
    }},
    {{
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/write"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""(Deprecated: Please use /providers/Microsoft.Maps) Create or update a {FriendlyResourceName}."",
        ""description"": ""(Deprecated: Please use /providers/Microsoft.Maps) Create or update a {FriendlyResourceName}.""
      }}
    }},
    {{
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/read"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""(Deprecated: Please use /providers/Microsoft.Maps) Get a {FriendlyResourceName}."",
        ""description"": ""(Deprecated: Please use /providers/Microsoft.Maps) Get a {FriendlyResourceName}.""
      }}
    }},
    {{
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/delete"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""(Deprecated: Please use /providers/Microsoft.Maps) Delete a {FriendlyResourceName}."",
        ""description"": ""(Deprecated: Please use /providers/Microsoft.Maps) Delete a {FriendlyResourceName}.""
      }}
    }},
    {{
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/listKeys/action"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""(Deprecated: Please use /providers/Microsoft.Maps) List keys"",
        ""description"": ""(Deprecated: Please use /providers/Microsoft.Maps)List Location Based Services Account keys""
      }}
    }},
    {{
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/regenerateKey/action"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""(Deprecated: Please use /providers/Microsoft.Maps) Generate new primary or secondary key"",
        ""description"": ""(Deprecated: Please use /providers/Microsoft.Maps) Generate new Location Based Services Account primary or secondary key""
      }}
    }},
    {{
      ""origin"": ""system"",
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/providers/Microsoft.Insights/metricDefinitions/read"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""The metric definition of Location Based Services Account"",
        ""operation"": ""(Deprecated: Please use /providers/Microsoft.Maps) Read Location Based Services Account metric definitions"",
        ""description"": ""(Deprecated: Please use /providers/Microsoft.Maps) Gets the available metrics for Location Based Services Accounts""
      }},
      ""properties"": {{
        ""serviceSpecification"": {{
          ""metricSpecifications"": [
            {{
              ""name"": ""Usage"",
              ""displayName"": ""{Operations.UsageMetric_DisplayName}"",
              ""displayDescription"": ""{Operations.UsageMetric_DisplayDescription}"",
              ""internalMetricName"": ""Latency"",
              ""unit"": ""Count"",
              ""aggregationType"": ""Count"",
              ""supportedAggregationTypes"": [
                ""Count""
              ],
              ""sourceMdmAccount"": ""MicrosoftLocationBasedServicesShoebox"",
              ""sourceMdmNamespace"": ""ServiceOperations"",
              ""fillGapWithZero"": false,
              ""dimensions"": [
                {{
                  ""name"": ""ApiCategory"",
                  ""displayName"": ""{Operations.Metric_Dimensions_DisplayName_ApiCategory}"",
                  ""internalName"": ""Service Category"",
                  ""toBeExportedToShoebox"": true
                }},
                {{
                  ""name"": ""ApiName"",
                  ""displayName"": ""{Operations.Metric_Dimensions_DisplayName_ApiName}"",
                  ""internalName"": ""OperationName"",
                  ""toBeExportedToShoebox"": true
                }}
              ]
            }}
          ]
        }}
      }}
    }},
    {{
      ""origin"": ""system"",
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/providers/Microsoft.Insights/diagnosticSettings/read"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""(Deprecated: Please use /providers/Microsoft.Maps) {Operations.Diagnostics_Read_Operation}"",
        ""description"": ""(Deprecated: Please use /providers/Microsoft.Maps) {Operations.Diagnostics_Read_Description}""
      }}
    }},
    {{
      ""origin"": ""system"",
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/providers/Microsoft.Insights/diagnosticSettings/write"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""(Deprecated: Please use /providers/Microsoft.Maps) {Operations.Diagnostics_Write_Operation}"",
        ""description"": ""(Deprecated: Please use /providers/Microsoft.Maps) {Operations.Diagnostics_Write_Description}""
      }}
    }}
  ]
}}";
        }
    }
}