﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Controllers
{
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ExternalContracts;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;

    [Route(@"subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{providerName:regex(^Microsoft\.Maps$)}/accounts/{resourceName}/listKeys")]
    public class ListAccountKeyController : ResourceProviderControllerBase
    {
        private readonly IAccountRecordStorage accountRecordStorage;
        private readonly IApiManagementStorageFactory apiManagementStorageFactory;
        
        public ListAccountKeyController(IAccountRecordStorage accountRecordStorage,
                                        IApiManagementStorageFactory apiManagementStorageFactory)
        {
            this.accountRecordStorage = accountRecordStorage;
            this.apiManagementStorageFactory = apiManagementStorageFactory;
        }

        [HttpPost]
        public async Task<HttpResponseMessage> Post([FromUri] string subscriptionId,
                                        [FromUri] string resourceGroupName,
                                        [FromUri] string providerName,
                                        [FromUri] string resourceName,
                                        [FromUri(Name = "api-version")] string apiVersion = "")
        {
            HttpResponseMessage response;
            try
            {
                Operation.LogCommonParameters(subscriptionId, resourceGroupName, resourceName);

                // these will throw a ResourceProviderException on failure
                ValidateAuthorized();
                ValidateApiVersion(apiVersion);
                ValidateCommonParameters(subscriptionId, resourceGroupName, resourceName);
                ValidateModelParse();

                var accountIdentifier = new AccountIdentifier(subscriptionId,
                                                              resourceGroupName,
                                                              resourceName);
                var keys = await PostWorker(accountIdentifier);
                if (keys != null)
                {
                    var keyResult = new Keys(accountIdentifier, keys, providerName);
                    response = Request.CreateResponse(HttpStatusCode.OK, keyResult);
                    Operation.SetResultSuccess();
                }
                else
                {
                    response = Request.CreateResponse(HttpStatusCode.NotFound);
                    Operation.SetResultClientError("notfound", null);
                }
            }
            catch (ResourceProviderException ex)
            {
                // these are errors which get converted to nice return codes and messages
                // let unknown exceptions fall through - they will be handled by global hook
                response = this.BuildErrorResponseAndLog(ex);
            }

            return response;
        }

        private async Task<ApiManagementKeys> PostWorker(AccountIdentifier accountIdentifier)
        {
            var record = await this.accountRecordStorage.GetAccountRecord(accountIdentifier);

            if (record == null)
            {
                return null;
            }

            var account = new ResourceProviderWebAPI.Account(record);
            ApiManagementKeys keys = record.GetKeys();
            var apiManagementStorage = this.apiManagementStorageFactory.GetApiManagementStorage(account.Sku.Name);

            // Check if account is legacy -- doesn't include keys in record
            if (keys == null)
            {
                // Key exists in APIM and not account record. Retrieve
                if (apiManagementStorage.UserExists(account.ApiManagementId))
                {
                    keys = apiManagementStorage.GetServiceKeys(account.ApiManagementId);
                }

                // No keys exist. Regenerate both
                else
                {
                    account.ApiManagementId = apiManagementStorage.AddNewServiceKeys(account.Id, out keys);
                }

                record.ApiManagementId = AccountRecord.ApiManagmentIdentityToString(account.ApiManagementId);
                record.UpdateKeys(keys);
                await accountRecordStorage.PutAccountRecord(record);
            }

            // APIM does not exist, so write keys to enable
            else if (!apiManagementStorage.UserExists(account.ApiManagementId))
            {
                apiManagementStorage.WriteServiceKeys(account.Id, account.ApiManagementId.SubscriptionId, out _, keys.PrimaryKey, keys.SecondaryKey);
            }

            return keys;
        }
    }
}