﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Controllers
{
    using System.Net;
    using System.Net.Http;
    using System.Text;
    using System.Web.Http;

    [Route(@"providers/{providerName:regex(^Microsoft\.Maps$)}/operations")]
    public class OperationsController : ResourceProviderControllerBase
    {
        private static string FriendlyDisplayProvider => Operations.Provider;

        private static string FriendlyResourceName => Operations.ResourceName;

        [HttpGet]
        public HttpResponseMessage Get([FromUri] string providerName, [FromUri(Name = "api-version")] string apiVersion = "")
        {
            HttpResponseMessage response;
            try
            {
                // these will throw a ResourceProviderException on failure
                ValidateAuthorized();
                ValidateApiVersion(apiVersion);
                ValidateModelParse();

                ValidateApiVersion(apiVersion);

                response = Request.CreateResponse(HttpStatusCode.OK);
                response.Content = new StringContent(_operationsJson(providerName), Encoding.UTF8, "application/json");
                Operation.SetResultSuccess();
            }
            catch (ResourceProviderException ex)
            {
                // these are errors which get converted to nice return codes and messages
                // let unknown exceptions fall through - they will be handled by global hook
                response = this.BuildErrorResponseAndLog(ex);
            }

            return response;
        }

        // quick and dirty json formatted output.  So far, this doesn't need to be data driven or localized (?)
        private static string _operationsJson(string providerName)
        {
            return
$@"{{
  ""value"": [
    {{
      ""name"": ""{providerName}/register/action"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""Register the provider"",
        ""description"": ""Register the provider""
      }}
    }},
    {{
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/write"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""Create or update a {FriendlyResourceName}."",
        ""description"": ""Create or update a {FriendlyResourceName}.""
      }}
    }},
    {{
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/read"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""Get a {FriendlyResourceName}."",
        ""description"": ""Get a {FriendlyResourceName}.""
      }}
    }},
    {{
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/delete"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""Delete a {FriendlyResourceName}."",
        ""description"": ""Delete a {FriendlyResourceName}.""
      }}
    }},
    {{
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/listKeys/action"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""List keys"",
        ""description"": ""List Maps Account keys""
      }}
    }},
    {{
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/regenerateKey/action"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""Generate new primary or secondary key"",
        ""description"": ""Generate new Maps Account primary or secondary key""
      }}
    }},
    {{
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/*/read"",
      ""isDataAction"" : true,
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}"",
        ""operation"": ""{Operations.RBAC_Data_Read_OperationName}"",
        ""description"": ""{Operations.RBAC_Data_Read_Description}""
      }}
    }},
    {{
      ""origin"": ""system"",
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/providers/Microsoft.Insights/metricDefinitions/read"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{Operations.Metrics_Display_Resource}"",
        ""operation"": ""{Operations.Metrics_Display_Operation}"",
        ""description"": ""{Operations.Metrics_Display_Description}""
      }},
      ""properties"": {{
        ""serviceSpecification"": {{
          ""metricSpecifications"": [
            {{
              ""name"": ""Usage"",
              ""displayName"": ""{Operations.UsageMetric_DisplayName}"",
              ""displayDescription"": ""{Operations.UsageMetric_DisplayDescription}"",
              ""internalMetricName"": ""Latency"",
              ""unit"": ""Count"",
              ""aggregationType"": ""Count"",
              ""supportedAggregationTypes"": [
                ""Count""
              ],
              ""sourceMdmAccount"": ""MicrosoftLocationBasedServicesShoebox"",
              ""sourceMdmNamespace"": ""ServiceOperations"",
              ""fillGapWithZero"": false,
              ""dimensions"": [
                {{
                  ""name"": ""ApiCategory"",
                  ""displayName"": ""{Operations.Metric_Dimensions_DisplayName_ApiCategory}"",
                  ""internalName"": ""Service Category"",
                  ""toBeExportedToShoebox"": true
                }},
                {{
                  ""name"": ""ApiName"",
                  ""displayName"": ""{Operations.Metric_Dimensions_DisplayName_ApiName}"",
                  ""internalName"": ""OperationName"",
                  ""toBeExportedToShoebox"": true
                }},
                {{
                  ""name"": ""ResultType"",
                  ""displayName"": ""{Operations.Metric_Dimensions_DisplayName_ResultType}"",
                  ""internalName"": ""OperationResult"",
                  ""toBeExportedToShoebox"": true
                }},
                {{
                  ""name"": ""ResponseCode"",
                  ""displayName"": ""{Operations.Metric_Dimensions_DisplayName_ResponseCode}"",
                  ""internalName"": ""OperationResultCode"",
                  ""toBeExportedToShoebox"": true
                }}
              ]
            }},
            {{
              ""name"": ""Availability"",
              ""displayName"": ""{Operations.AvailabilityMetric_DisplayName}"",
              ""displayDescription"": ""{Operations.AvailabilityMetric_DisplayDescription}"",
              ""internalMetricName"": ""Availability"",
              ""unit"": ""Percent"",
              ""aggregationType"": ""Average"",
              ""supportedAggregationTypes"": [
                ""Average""
              ],
              ""sourceMdmAccount"": ""MicrosoftLocationBasedServicesShoebox"",
              ""sourceMdmNamespace"": ""ServiceOperations"",
              ""fillGapWithZero"": false,
              ""dimensions"": [
                {{
                  ""name"": ""ApiCategory"",
                  ""displayName"": ""{Operations.Metric_Dimensions_DisplayName_ApiCategory}"",
                  ""internalName"": ""Service Category"",
                  ""toBeExportedToShoebox"": true
                }},
                {{
                  ""name"": ""ApiName"",
                  ""displayName"": ""{Operations.Metric_Dimensions_DisplayName_ApiName}"",
                  ""internalName"": ""OperationName"",
                  ""toBeExportedToShoebox"": true
                }}
              ]
            }}
          ]
        }}
      }}
    }},
    {{
      ""origin"": ""system"",
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/providers/Microsoft.Insights/diagnosticSettings/read"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""{Operations.Diagnostics_Read_Operation}"",
        ""description"": ""{Operations.Diagnostics_Read_Description}""
      }}
    }},
    {{
      ""origin"": ""system"",
      ""name"": ""{providerName}/{ResourceProviderWebApiService.ResourceTypeName}/providers/Microsoft.Insights/diagnosticSettings/write"",
      ""display"": {{ 
        ""provider"": ""{FriendlyDisplayProvider}"",
        ""resource"": ""{FriendlyResourceName}"",
        ""operation"": ""{Operations.Diagnostics_Write_Operation}"",
        ""description"": ""{Operations.Diagnostics_Write_Description}""
      }}
    }}
  ]
}}";
        }
    }
}