﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Controllers
{
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ExternalContracts;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;

    using Account = Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Account;

    [Route(@"subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/{providerName:regex(^Microsoft\.Maps$)}/accounts/{resourceName}/regenerateKey")]
    public class RegenerateAccountKeyController : ResourceProviderControllerBase
    {
        private const string primaryKey = "primary";
        private const string secondaryKey = "secondary";

        private readonly IAccountRecordStorage accountRecordStorage;
        private readonly IApiManagementStorageFactory apiManagementStorageFactory;

        public RegenerateAccountKeyController(IAccountRecordStorage accountRecordStorage,
                                              IApiManagementStorageFactory apiManagementStorageFactory)
        {
            this.accountRecordStorage = accountRecordStorage;
            this.apiManagementStorageFactory = apiManagementStorageFactory;
        }

        internal enum KeyType
        {
            Primary,
            Secondary
        }

        [HttpPost]
        public async Task<HttpResponseMessage> RegenerateKey([FromUri] string subscriptionId,
                                                 [FromUri] string resourceGroupName,
                                                 [FromUri] string providerName,
                                                 [FromUri] string resourceName,
                                                 [FromBody] KeySpecification keySpecification,
                                                 [FromUri(Name = "api-version")] string apiVersion = "")
        {
            HttpResponseMessage response;
            try
            {
                Operation.LogCommonParameters(subscriptionId, resourceGroupName, resourceName);

                // these will throw a ResourceProviderException on failure
                ValidateAuthorized();
                ValidateApiVersion(apiVersion);
                ValidateCommonParameters(subscriptionId, resourceGroupName, resourceName);
                ValidateModelParse();
                if (keySpecification == null)
                {
                    throw new ResourceProviderException(ResourceProviderException.ExceptionType.ClientParams, "LocationServicesKeySpecification");
                }

                var accountIdentifier = new AccountIdentifier(subscriptionId,
                                                              resourceGroupName,
                                                              resourceName);

                KeyType keyType = KeyType.Primary;
                if (string.Compare(keySpecification.KeyType, primaryKey, System.StringComparison.InvariantCultureIgnoreCase) == 0)
                {
                    keyType = KeyType.Primary;
                }
                else if (string.Compare(keySpecification.KeyType, secondaryKey, System.StringComparison.InvariantCultureIgnoreCase) == 0)
                {
                    keyType = KeyType.Secondary;
                }
                else
                {
                    throw new ResourceProviderException(ResourceProviderException.ExceptionType.ClientParams, "LocationServicesKeySpecification.KeyType");
                }

                var keys = await RegenerateKeyWorker(accountIdentifier, keyType);

                if (keys != null)
                {
                    response = Request.CreateResponse(HttpStatusCode.OK, new Keys(accountIdentifier, keys, providerName));
                    Operation.SetResultSuccess();
                }
                else
                {
                    response = Request.CreateResponse(HttpStatusCode.NotFound);
                    Operation.SetResultClientError("notfound", null);
                }
            }
            catch (ResourceProviderException ex)
            {
                // these are errors which get converted to nice return codes and messages
                // let unknown exceptions fall through - they will be handled by global hook
                response = this.BuildErrorResponseAndLog(ex);
            }

            return response;
        }

        internal async Task<ApiManagementKeys> RegenerateKeyWorker(AccountIdentifier accountIdentifier, KeyType keyType)
        {
            var record = await accountRecordStorage.GetAccountRecord(accountIdentifier);

            if (record == null)
            {
                return null;
            }

            var account = new Account(record);
            var keys = record.GetKeys();
            var apiManagementStorage = this.apiManagementStorageFactory.GetApiManagementStorage(account.Sku.Name);

            // Check if account is legacy -- doesn't include keys in record
            if (keys == null)
            {
                // Key exists in APIM and not account record. Retrieve
                if (apiManagementStorage.UserExists(account.ApiManagementId))
                {
                    keys = apiManagementStorage.GetServiceKeys(account.ApiManagementId);
                    record.UpdateKeys(keys);
                }

                // No keys exist. Regenerate both
                else
                {
                    account.ApiManagementId = apiManagementStorage.AddNewServiceKeys(account.Id, out keys);
                    record.ApiManagementId = AccountRecord.ApiManagmentIdentityToString(account.ApiManagementId);
                    record.UpdateKeys(keys);
                }

                await accountRecordStorage.PutAccountRecord(record);

                account = new Account(record);
            }

            // APIM does not exist, so write keys to enable
            else if (!apiManagementStorage.UserExists(account.ApiManagementId))
            {
                apiManagementStorage.WriteServiceKeys(account.Id, account.ApiManagementId.SubscriptionId, out _, keys.PrimaryKey, keys.SecondaryKey);
            }

            keys = this.RegenerateKey(keyType, keys, account);

            var changedRecord = new AccountRecord(record);

            // save record which is identical to the old, except with new keys
            changedRecord.UpdateKeys(keys);
            await accountRecordStorage.PutAccountRecord(changedRecord);

            return keys;
        }

        private ApiManagementKeys RegenerateKey(KeyType keyType, ApiManagementKeys keys, Account account)
        {
            var apiManagementStorage = this.apiManagementStorageFactory.GetApiManagementStorage(account.Sku.Name);
            switch (keyType)
            {
                case KeyType.Primary:
                    return new ApiManagementKeys(apiManagementStorage.RegeneratePrimaryKey(account.ApiManagementId), keys.SecondaryKey);
                case KeyType.Secondary:
                default:
                    return new ApiManagementKeys(keys.PrimaryKey, apiManagementStorage.RegenerateSecondaryKey(account.ApiManagementId));
            }
        }
    }
}