﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Controllers
{
    using System.Net;
    using System.Net.Http;
    using System.Threading.Tasks;
    using System.Web.Http;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;

    [Route("subscriptions/{subscriptionId}")]
    public class SubscriptionLifecycleController : ResourceProviderControllerBase
    {
        private readonly ISubscriptionRecordStorage _subscriptionRecordStorage;
        private readonly IDocumentDbStorage _documentDbStorage;

        public SubscriptionLifecycleController(ISubscriptionRecordStorage subscriptionRecordStorage,
            IDocumentDbStorage documentDbStorage)
        {
            _subscriptionRecordStorage = subscriptionRecordStorage;
            _documentDbStorage = documentDbStorage;
        }

        [HttpPut]
        public async Task<HttpResponseMessage> Put([FromUri] string subscriptionId,
                                       [FromBody] ExternalContracts.Subscription body,
                                       [FromUri(Name = "api-version")] string apiVersion = "")
        {
            HttpResponseMessage response;
            Subscription subscription;
            try
            {
                Operation.LogCommonParameters(subscriptionId);

                // these will throw a ResourceProviderException on failure
                ValidateAuthorized();
                ValidateCommonParameters(subscriptionId);
                ValidateModelParse();

                // special case --- for subscription lifecycle, the api version is "2.0"
                if (string.Compare(apiVersion, "2.0", System.StringComparison.InvariantCultureIgnoreCase) != 0)
                {
                    throw new ResourceProviderException(ResourceProviderException.ExceptionType.ClientParams, "api-version");
                }

                // convert from external format to internal (which does some validation)
                try
                {
                    subscription = new Subscription(subscriptionId, body);
                }
                catch (System.ArgumentException ex)
                {
                    throw new ResourceProviderException(ResourceProviderException.ExceptionType.ClientParams, ex.ParamName);
                }

                // convert to storage format and save
                SubscriptionRecord record = new SubscriptionRecord(subscription);
                _subscriptionRecordStorage.PutSubscriptionRecord(record);

                await _documentDbStorage.UpsertSubscriptionAsync(subscription);

                response = Request.CreateResponse(HttpStatusCode.OK);
                Operation.SetResultSuccess();
            }
            catch (ResourceProviderException ex)
            {
                // these are errors which get converted to nice return codes and messages
                // let unknown exceptions fall through - they will be handled by global hook
                response = this.BuildErrorResponseAndLog(ex);
            }

            return response;
        }
    }
}