﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI
{
    using System;
    using System.Linq;
    using System.Net.Http;

    using Microsoft.Azure.LocationServices.Instrumentation;

    public class ResourceProviderControllerBase : System.Web.Http.ApiController
    {
        private static readonly string[] AllowedVersions = { "2017-01-01-preview", "2018-05-01" };

        public bool IsAuthorized { get; private set; }

        public Exception ModelParsingException { get; private set; }

        public IResourceProviderOperation Operation { get; private set; }

        /// <summary>
        ///     Validates external strings
        /// </summary>
        /// <param name="subscriptionId"></param>
        /// <param name="resourceGroupName"></param>
        /// <param name="resourceName"></param>
        /// <exception cref="ResourceProviderException">Throws if any field fails validation</exception>
        public static void ValidateCommonParameters(string subscriptionId, string resourceGroupName, string resourceName)
        {
            try
            {
                ValidateCommonParameters(subscriptionId, resourceGroupName);
                Validation.Requires.NotNullOrWhiteSpace(resourceName, nameof(resourceName));
                Validation.Requires.Argument(resourceGroupName.Length + resourceName.Length < 98, "resourceGroupName and resourceName", "The ResourceName and ResourceGroupName together must be less than 98 characters");
                Validation.Requires.Argument(System.Text.RegularExpressions.Regex.IsMatch(resourceName, @"^[^\p{C}%&:\\/\?]+$"),
                                             nameof(resourceName),
                                             "Invalid characters");
            }
            catch (ArgumentException e)
            {
                throw new ResourceProviderException(ResourceProviderException.ExceptionType.ClientParams, e.ParamName);
            }
        }

        /// <summary>
        ///     Validates external strings
        /// </summary>
        /// <param name="subscriptionId"></param>
        /// <param name="resourceGroupName"></param>
        /// <exception cref="ResourceProviderException">Throws if any field fails validation</exception>
        public static void ValidateCommonParameters(string subscriptionId, string resourceGroupName)
        {
            try
            {
                ValidateCommonParameters(subscriptionId);
                Validation.Requires.NotNullOrWhiteSpace(resourceGroupName, nameof(resourceGroupName));
                Validation.Requires.Argument(System.Text.RegularExpressions.Regex.IsMatch(resourceGroupName, @"^[\w-_\(\)\.]+$"),
                                             nameof(resourceGroupName),
                                             "Invalid characters");
            }
            catch (ArgumentException e)
            {
                throw new ResourceProviderException(ResourceProviderException.ExceptionType.ClientParams, e.ParamName);
            }
        }

        /// <summary>
        ///     Validates external subscriptionId
        /// </summary>
        /// <param name="subscriptionId"></param>
        /// <exception cref="ResourceProviderException">Throws if any field fails validation</exception>
        public static void ValidateCommonParameters(string subscriptionId)
        {
            try
            {
                Validation.Requires.NotNullOrWhiteSpace(subscriptionId, nameof(subscriptionId));
            }
            catch (ArgumentException e)
            {
                throw new ResourceProviderException(ResourceProviderException.ExceptionType.ClientParams, e.ParamName);
            }
        }

        public void SetAuthorized()
        {
            this.IsAuthorized = true;
        }

        public void SetModelParsingException(Exception ex)
        {
            this.ModelParsingException = ex;
        }

        public void SetOperation(IResourceProviderOperation operation)
        {
            this.Operation = operation;
        }

        /// <summary>
        ///     Validates the incoming api version string is allowed.
        ///     This is the default implementation; individual entry point could supply their own
        /// </summary>
        /// <param name="version"></param>
        /// <exception cref="ResourceProviderException">Throws if api is unexpected version</exception>
        protected virtual void ValidateApiVersion(string version)
        {
            if (!AllowedVersions.Any(v => string.Equals(version, v, StringComparison.OrdinalIgnoreCase)))
            {
                throw new ResourceProviderException(ResourceProviderException.ExceptionType.ClientParams, "api-version");
            }
        }

        /// <summary>
        ///     Validates the call is authorized.
        ///     Note that the actual determination happens earlier (in the global filter) and the results cached here.
        ///     It's done this way so that our code can still log and trap the error appropriately.
        ///     See SetAuthorized
        /// </summary>
        /// <exception cref="ResourceProviderException">Throws if service call is not authorized</exception>
        protected void ValidateAuthorized()
        {
            if (!this.IsAuthorized)
            {
                throw new ResourceProviderException(ResourceProviderException.ExceptionType.Unauthorized, null);
            }
        }

        /// <summary>
        ///     Validates the body was correctly parsed.
        ///     Note that the actual determination happens earlier (in the global filter) and the results cached here.
        ///     It's done this way so that our code can still log and trap the error appropriately.
        ///     See SetModelParsingException
        /// </summary>
        /// <exception cref="ResourceProviderException">Throws if the body was not correctly parsed</exception>
        protected void ValidateModelParse()
        {
            if (this.ModelParsingException != null)
            {
                throw new ResourceProviderException(ResourceProviderException.ExceptionType.JscriptParsing, "Invalid body", this.ModelParsingException);
            }
        }

        /// <summary>
        ///     Constructs an appropriate *client-visible* response for the error, and logs it to the Operation
        /// </summary>
        /// <param name="resourceProviderException"></param>
        /// <returns></returns>
        protected HttpResponseMessage BuildErrorResponseAndLog(ResourceProviderException resourceProviderException)
        {
            switch (resourceProviderException.Type)
            {
                case ResourceProviderException.ExceptionType.Unauthorized:
                {
                    this.Operation.SetResultClientError("clientcertificate", null);
                    return this.Request.CreateResponse(System.Net.HttpStatusCode.Forbidden);
                }

                case ResourceProviderException.ExceptionType.ClientParams:
                {
                    this.Operation.SetResultClientError("invalidparameter", resourceProviderException.Message);
                    return this.Request.CreateResponse(System.Net.HttpStatusCode.BadRequest, new ExternalContracts.ErrorResponse(resourceProviderException));
                }

                case ResourceProviderException.ExceptionType.JscriptParsing:
                {
                    this.Operation.SetResultClientError("invalidjscript", resourceProviderException.ToString().Truncate(256));
                    return this.Request.CreateResponse(System.Net.HttpStatusCode.BadRequest, new ExternalContracts.ErrorResponse(resourceProviderException));
                }

                case ResourceProviderException.ExceptionType.Other:
                default:
                {
                    // create a unique guid, return that to the user, and log it for cross reference
                    Guid guid = Guid.NewGuid();

                    // log a specific entry for exception, so various tools can spot it
                    Agent.Instance.LogException(resourceProviderException, "resourceProvider", "unknown client error", guid.ToString());

                    this.Operation.SetResultFailure($"error: {guid}", resourceProviderException.ToString().Truncate(1024));
                    return this.Request.CreateResponse(System.Net.HttpStatusCode.InternalServerError, new ExternalContracts.ErrorResponse(guid.ToString()));
                }
            }
        }
    }
}