﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI
{
    using System.Net.Http;
    using System.Web.Http.Controllers;
    using System.Web.Http.Filters;

    using Microsoft.Azure.LocationServices.Instrumentation;

    internal class ServiceRequestActionFilter : ActionFilterAttribute
    {
        private readonly IResourceProviderOperationFactory resourceProviderOperationFactory;

        public ServiceRequestActionFilter(IResourceProviderOperationFactory resourceProviderOperationFactory)
        {
            this.resourceProviderOperationFactory = resourceProviderOperationFactory;
        }

        public override void OnActionExecuting(HttpActionContext actionContext)
        {
            if (actionContext.ControllerContext.ControllerDescriptor.ControllerType == typeof(Controllers.HeartbeatController))
            {
                return;
            }

            // this hook sets up the logging operation
            var rpController = actionContext.ControllerContext.Controller as ResourceProviderControllerBase;
            Validation.Requires.NotNull(rpController, "rpController");

            string operationName = $"{actionContext.ActionDescriptor.ControllerDescriptor.ControllerName}.{actionContext.ActionDescriptor.ActionName}";
            rpController.SetOperation(this.resourceProviderOperationFactory.CreateOperation(operationName));
        }

        public override void OnActionExecuted(HttpActionExecutedContext actionExecutedContext)
        {
            if (actionExecutedContext.ActionContext.ControllerContext.ControllerDescriptor.ControllerType == typeof(Controllers.HeartbeatController))
            {
                return;
            }

            try
            {
                var rpController = actionExecutedContext.ActionContext.ControllerContext.Controller as ResourceProviderControllerBase;
                Validation.Requires.NotNull(rpController, "rpController");
                Validation.Requires.NotNull(rpController.Operation, "rpController.Operation");

                if (actionExecutedContext.Exception != null)
                {
                    // this is the last line of defense.   Known exception cases (bad parameters etc) should be handled gracefully by now.
                    // things like apim or storage failures will land here

                    // create a unique guid, return that to the user, and log it for cross reference
                    System.Guid guid = System.Guid.NewGuid();

                    // log a specific entry for exception, so various tools can spot it
                    Instrumentation.Agent.Instance.LogException(actionExecutedContext.Exception, "resourceProvider", "serverError", guid.ToString());

                    actionExecutedContext.Response = actionExecutedContext.Request.CreateResponse(System.Net.HttpStatusCode.InternalServerError,
                                                                                                  new ExternalContracts.ErrorResponse(guid.ToString()));

                    rpController.Operation.SetResultFailure($"exception: {guid.ToString()}", actionExecutedContext.Exception.ToString().Truncate(1024));
                }

                // this hook closes out the logging operation
                rpController.Operation.Dispose();
                rpController.SetOperation(null);
            }
            catch (System.Exception)
            {
                // somehow got here, probably without a ResourceProviderBaseController, or already disposed Operation
                // TODO: how to handle?
            }
        }
    }
}
