﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI
{
    using System.Collections.Generic;
    using System.Net.Http.Formatting;
    using System.Threading.Tasks;
    using System.Web.Http;    
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ARM;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Utility;

    using Newtonsoft.Json;
    using Newtonsoft.Json.Converters;
    using Newtonsoft.Json.Serialization;
    using global::Owin;    
    using Swashbuckle.Application;

    using Unity;
    using Unity.Lifetime;

    public static class Startup
    {
        /// <summary>
        /// Single container for dependencies, will be disposed and managed by <see cref="UnityResolver"/>
        /// </summary>
        private static readonly IUnityContainer Container;

        static Startup()
        {
            Container = new UnityContainer();
        }
        
        // This code configures Web API. The Startup class is specified as a type
        // parameter in the WebApp.Start method.
        public static void ConfigureApp(IAppBuilder appBuilder)
        {            
            // Configure Web API for self-host. 
            HttpConfiguration config = new HttpConfiguration();

            var defaultSettings = new JsonSerializerSettings
                                  {
                                      Formatting = Formatting.Indented,
                                      ContractResolver = new CamelCasePropertyNamesContractResolver(),
                                      DefaultValueHandling = DefaultValueHandling.Ignore,
                                      Converters = new List<JsonConverter>
                                                   {
                                                       new StringEnumConverter { CamelCaseText = true },
                                                   }
                                  };

            JsonConvert.DefaultSettings = () => defaultSettings;

            config.Formatters.Clear();
            config.Formatters.Add(new JsonMediaTypeFormatter());
            config.Formatters.JsonFormatter.SerializerSettings = defaultSettings;

            config.EnableSwagger(c => c.SingleApiVersion("v1", "Location Services Resource Provider"))
                   .EnableSwaggerUi();

            config.MapHttpAttributeRoutes();

            // the order of these is significant (I believe they are executed in reverse order)
            // this means that exceptions thrown by ValidModels are correctly handled in ServiceRequestActionFilter, etc
            config.DependencyResolver = new UnityResolver(Container);
            config.Filters.Add(new ServiceRequestActionFilter(new ResourceProviderOperationFactory()));
            config.Filters.Add(new ValidModelsOnlyFilter());
            config.Filters.Add(new CheckClientCertificateFilter(new ClientCertificateResolverFactory(config.DependencyResolver)));
            
            appBuilder.UseWebApi(config);
            
            System.Net.ServicePointManager.ServerCertificateValidationCallback += (sender, certificate, chain, errors) => (errors == System.Net.Security.SslPolicyErrors.None);
        }
        
        /// <summary>
        /// Registers dependencies for the application.
        /// </summary>
        /// <returns>awaitable <see cref="Task"/></returns>
        public static async Task RegisterDependenciesAsync()
        {
            Container.RegisterInstance(ConfigurationReader.Instance);
            Container.RegisterType<IBootstrapConfiguration, ServiceFabricConfigurationBootstrap>(new ContainerControlledLifetimeManager());
            Container.RegisterType<IConfigurationReader, KeyVaultConfigurationReader>(new ContainerControlledLifetimeManager());
            Container.RegisterType<IClientCertificateResolver, ClientCertificateResolver>(new ContainerControlledLifetimeManager());
            Container.RegisterType<ArmCertificateReader>(new ContainerControlledLifetimeManager());

            // register environment host settings
            var hostSettings = ResourceProviderHostSettings.CreateFromEnvironment();
            Container.RegisterInstance(hostSettings);

            IBootstrapConfiguration bootstrapConfiguration = Container.Resolve<IBootstrapConfiguration>();
            IConfigurationReader configurationReader = Container.Resolve<IConfigurationReader>();

            // register subscription storage
            ISubscriptionRecordStorage subscriptionRecordStorage = await SubscriptionRecordStorage.BuildAsync(bootstrapConfiguration, configurationReader);
            Container.RegisterInstance(subscriptionRecordStorage);

            IDocumentDbStorage documentDbAccountStorage = new MockDocumentDbStorage(); 
            if (hostSettings.EnableAccountReplication)
            {
                 documentDbAccountStorage = await DocumentDbStorage.BuildAsync(configurationReader, hostSettings, subscriptionRecordStorage);
            }

            Container.RegisterInstance(documentDbAccountStorage);

            // register account storage
            IAccountRecordStorage accountRecordStorage = await AccountRecordStorage.BuildAsync(bootstrapConfiguration, configurationReader, 
                hostSettings, documentDbAccountStorage);
            Container.RegisterInstance(accountRecordStorage);
            
            // register apim client
            Container.RegisterInstance(await ApiManagementClient.BuildAsync(bootstrapConfiguration, configurationReader));

            IApiManagementStorageFactory apiManagementStorageFactory = new ApiManagementStorageFactory(Container.Resolve<ApiManagementClient>(), hostSettings.SkuApimMap);
            Container.RegisterInstance(apiManagementStorageFactory);
        }

        /// <summary>
        /// Resolves an instance of <see cref="IClientCertificateResolver"/>, this is required for any calls that require service locator.
        /// This is initially not generic to discourage service locator as a pattern.
        /// </summary>  
        /// <returns>instance of <see cref="IClientCertificateResolver"/></returns>
        internal static IClientCertificateResolver ResolveClientCertificateResolver()
        {
            return Container.Resolve<IClientCertificateResolver>();
        }
    }
}
