﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage
{
    using System;
    using System.Dynamic;
    using Microsoft.WindowsAzure.Storage.Table;
    using Newtonsoft.Json;

    public class AccountRecord : TableEntity
    {
        /// <summary>
        /// ctor required by storage deserialization
        /// </summary>
        public AccountRecord()
        {
        }

        /// <summary>
        /// Use this constructor when creating new records.
        /// </summary>
        /// <param name="account">New account information</param>
        /// <param name="keys">Keys which belong to this account</param>
        public AccountRecord(Account account, ApiManagement.ApiManagementKeys keys)
            : base()
        {
            Validation.Requires.NotNull(account, nameof(account));

            this.PartitionKey = AccountRecordStorage.PartitionKey(account.Id);
            this.ClientId = account.ClientId;
            this.RowKey = AccountRecordStorage.RowKey(account.Id);
            this.Created = DateTime.UtcNow;
            this.Deleted = false;
            this.SubscriptionId = account.Id.SubscriptionId;
            this.ResourceGroupName = account.Id.ResourceGroupName;
            this.ResourceName = account.Id.ResourceName;
            this.SkuName = account.Sku.Name.ToString();
            this.Location = account.Location.LocationString;
            this.Tags = TagsToString(account.Tags);
            this.ApiManagementId = ApiManagmentIdentityToString(account.ApiManagementId);

            // these are saved here, so the RP is the source of Truth (they should always match what ApiM has) but otherwise don't do anything
            // (can be null for backwards compatible; older accounts do not have these saved)
            UpdateKeys(keys);
        }

        /// <summary>
        /// Applies changes to the account record based on a changed account, consider usage for patch operations
        /// </summary>
        /// <param name="changedAccount">changed account provided from external callers</param>
        /// <param name="originalRecord">original record prior to modifications</param>
        public AccountRecord(Account changedAccount, AccountRecord originalRecord)
            : this(originalRecord)
        {            
            Validation.Requires.NotNull(changedAccount, nameof(changedAccount));
            this.Tags = TagsToString(changedAccount.Tags);
        }

        /// <summary>
        /// Copy constructor.
        /// </summary>
        /// <param name="originalRecord"></param>
        public AccountRecord(AccountRecord originalRecord)            
        {
            Validation.Requires.NotNull(originalRecord, nameof(originalRecord));
            this.RowKey = originalRecord.RowKey;
            this.PartitionKey = originalRecord.PartitionKey;
            this.ClientId = originalRecord.ClientId;
            this.Created = originalRecord.Created;
            this.Deleted = originalRecord.Deleted;
            this.Location = originalRecord.Location;
            this.PrimaryKey = originalRecord.PrimaryKey;
            this.PrimaryKeyLastUpdated = originalRecord.PrimaryKeyLastUpdated;
            this.ResourceGroupName = originalRecord.ResourceGroupName;
            this.ResourceName = originalRecord.ResourceName;
            this.SecondaryKey = originalRecord.SecondaryKey;
            this.SecondaryKeyLastUpdated = originalRecord.SecondaryKeyLastUpdated;
            this.SkuName = originalRecord.SkuName;
            this.SubscriptionId = originalRecord.SubscriptionId;
            this.Tags = originalRecord.Tags;
            this.ApiManagementId = originalRecord.ApiManagementId;            
        }

        public string SubscriptionId { get; set; }

        public string ResourceGroupName { get; set; }

        public string ResourceName { get; set; }

        public string SkuName { get; set; }

        public string Location { get; set; }

        public string Tags { get; set; }

        public DateTime Created { get; set; }

        public bool Deleted { get; set; }

        public string ApiManagementId { get; set; }

        public string PrimaryKey { get; set; }

        public string SecondaryKey { get; set; }

        public DateTime? PrimaryKeyLastUpdated { get; set; }

        public DateTime? SecondaryKeyLastUpdated { get; set; }

        public string ClientId { get; set; }

        internal static string ApiManagmentIdentityToString(ApiManagement.ApiManagementIdentity id)
        {
            return JsonConvert.SerializeObject(id);
        }

        internal static string TagsToString(System.Dynamic.ExpandoObject tags)
        {
            return JsonConvert.SerializeObject(tags);
        }

        internal ApiManagement.ApiManagementIdentity GetApiManagementIdentity()
        {
            return JsonConvert.DeserializeObject<ApiManagement.ApiManagementIdentity>(ApiManagementId);
        }

        internal ExpandoObject GetTags()
        {
            return JsonConvert.DeserializeObject<ExpandoObject>(Tags);
        }

        /// <summary>
        /// Gets any keys in this record.  For backwards compatibility, this can be null as older records don't have this.   But newer records should
        /// </summary>
        /// <returns></returns>
        internal ApiManagement.ApiManagementKeys GetKeys()
        {
            if (this.PrimaryKey != null && this.SecondaryKey != null)
            {
                return new ApiManagement.ApiManagementKeys(this.PrimaryKey, this.SecondaryKey, this.PrimaryKeyLastUpdated.GetValueOrDefault(this.Created), this.SecondaryKeyLastUpdated.GetValueOrDefault(this.Created));
            }

            return null;
        }

        internal void UpdateKeys(ApiManagement.ApiManagementKeys keys)
        {
            // these are saved here, so the RP is the source of Truth (they should always match what ApiM has) but otherwise don't do anything
            // (can be null for backwards compatible; older accounts do not have these saved)
            if (keys != null)
            {
                this.PrimaryKey = keys.PrimaryKey.Key;
                this.SecondaryKey = keys.SecondaryKey.Key;
                this.PrimaryKeyLastUpdated = keys.PrimaryKey.LastUpdated;
                this.SecondaryKeyLastUpdated = keys.SecondaryKey.LastUpdated;
            }
        }
    }
}
