﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using System.Text.RegularExpressions;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Utility;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.RetryPolicies;
    using Microsoft.WindowsAzure.Storage.Table;
    using Validation;

    internal class AccountRecordStorage : IAccountRecordStorage
    {
        // Primary and Row keys in tables can't include these characters.   In most cases, we don't even have to check since 
        // strings like subscription ids are already checked against a more restrictive set.
        public static readonly Regex DisallowedCharsInTableKeys = new Regex(@"[\\\\#%+/?\u0000-\u001F\u007F-\u009F]");

        private const string ResourceProviderTableNameKey = "mResourceProviderTableName";

        /// <summary>
        /// An integer specifying the maximum number of retry attempts.
        /// </summary>
        private const int MaxRetries = 3;

        /// <summary>
        /// The back-off interval between retries.
        /// </summary>
        private static readonly TimeSpan DeltaBackoff = TimeSpan.FromSeconds(2);

        /// <summary>
        /// The maximum execution time across all potential retries.
        /// </summary>
        private static readonly TimeSpan MaximumExecutionTime = TimeSpan.FromSeconds(30);

        /// <summary>
        /// The table request options.
        /// </summary>
        private static readonly TableRequestOptions TableRequestOptions = new TableRequestOptions
        {
            RetryPolicy = new ExponentialRetry(DeltaBackoff, MaxRetries),
            MaximumExecutionTime = MaximumExecutionTime
        };

        // this is the main fact table
        private readonly CloudTable table;

        /// <summary>
        /// The document client used for account look and validation.
        /// </summary>
        private readonly IDocumentDbStorage accountDbStorage;

        private readonly ResourceProviderHostSettings settings;

        private AccountRecordStorage(CloudTable table, IDocumentDbStorage accountDbStorage,
            ResourceProviderHostSettings settings)
        {            
            this.table = table;
            this.accountDbStorage = accountDbStorage;
            this.settings = settings;
        }

        /// <summary>
        /// Factory method to construct a new instance of <see cref="AccountRecordStorage"/>
        /// </summary>
        /// <param name="bootstrapConfiguration">configuration read as a bootstrap</param>
        /// <param name="configurationReader">configuration source</param>
        /// <param name="settings">Environment configuration settings</param>
        /// <param name="accountDbStorage">Document database storage for replication</param>
        /// <returns>new instance of <see cref="AccountRecordStorage"/></returns>
        public static async Task<IAccountRecordStorage> BuildAsync(IBootstrapConfiguration bootstrapConfiguration, IConfigurationReader configurationReader, 
            ResourceProviderHostSettings settings, IDocumentDbStorage accountDbStorage)
        {
            try
            {
                Requires.NotNull(configurationReader, nameof(configurationReader));
                Requires.NotNull(bootstrapConfiguration, nameof(bootstrapConfiguration));
                Requires.NotNull(accountDbStorage, nameof(accountDbStorage));
                Requires.NotNull(settings, nameof(settings));
                string connectionStringTemplate = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(Globals.StorageAccountConnectionTemplateKey));
                string storageAccountSasToken = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(Globals.StorageAccountSasTokenKey));
                string storageConnectionString = string.Format(connectionStringTemplate, storageAccountSasToken);
                CloudStorageAccount account = CloudStorageAccount.Parse(storageConnectionString);
                CloudTableClient client = account.CreateCloudTableClient();
                client.DefaultRequestOptions = TableRequestOptions;
                string tableName = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(ResourceProviderTableNameKey));
                CloudTable table = client.GetTableReference(tableName);
                table.CreateIfNotExists();

                Agent.Instance.LogEvent(EventType.Information, $"Connected to table {table.StorageUri.PrimaryUri}");
                return new AccountRecordStorage(table, accountDbStorage, settings);
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "resourceProvider", "accountRecordStorage");
                throw;
            }
        }

        /// <summary>
        /// <para>
        /// Get AccountRecords from storage
        /// </para>
        /// Success (as logged in geneva) is defined as not getting an error from storage.  Not finding the record is still success.
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        public async Task<AccountRecord> GetAccountRecord(AccountIdentifier id)
        {
            using (var op = new Instrumentation.ResourceProviderStorageOperation(nameof(GetAccountRecord)))
            {
                Validation.Requires.NotNull(id, nameof(id));

                op.LogCommonParameters(id.SubscriptionId, id.ResourceGroupName, id.ResourceName);

                var queryString = string.Format("({0}) {1} ({2})",
                                                TableQuery.GenerateFilterCondition("PartitionKey",
                                                                                   QueryComparisons.Equal,
                                                                                   PartitionKey(id)),
                                                TableOperators.And,
                                                TableQuery.GenerateFilterConditionForBool("Deleted",
                                                                                          QueryComparisons.Equal,
                                                                                          false));

                var query = new TableQuery<AccountRecord>().Where(queryString);

                // this results in null if no match
                // there should be at most one record which matches all criteria (and is not deleted)

                // server filtering is case sensitive.   So, get all non-deleted records for a subscription and client filter (there
                // shouldn't be too many anyway)

                // FirstOrDefault() is used to be resilient for instances when multiple records share a resource id
                // It would still be exceptional for multiple accounts to share a resource id
                var record = this.table.ExecuteQuery(query)
                                    .Where(e => string.Equals(e.ResourceGroupName,
                                                             id.ResourceGroupName,
                                                             StringComparison.InvariantCultureIgnoreCase) &&
                                               string.Equals(e.ResourceName,
                                                             id.ResourceName,
                                                             StringComparison.InvariantCultureIgnoreCase))
                                    .OrderByDescending(e => e.Timestamp)
                                    .FirstOrDefault();

                op.SetNumberOfRecordsAffected(record != null ? 1 : 0);
                op.SetResultSuccess();

                if (!settings.EnableAccountReplication && record != null)
                {
                    // set this null so we don't actually create IDs until we enable the feature.
                    record.ClientId = null;
                }

                return await Task.FromResult(record);
            }
        }

        /// <summary>
        ///     Adds or updates account information.  Throws on failure
        /// </summary>
        /// <param name="account"></param>
        public async Task PutAccountRecord(AccountRecord account)
        {
            using (var op = new Instrumentation.ResourceProviderStorageOperation(nameof(PutAccountRecord)))
            {
                op.LogCommonParameters(account.SubscriptionId, account.ResourceGroupName, account.ResourceName);

                Requires.NotNull(account, nameof(account));

                if (!settings.EnableAccountReplication)
                {
                    // set this null so we don't actually create IDs until we enable the feature.
                    account.ClientId = null;
                }

                // last one wins semantics
                var insertOperation = TableOperation.InsertOrReplace(account);
                Task docDbAccountTask = null;
                if (settings.EnableAccountReplication)
                {
                    docDbAccountTask = accountDbStorage.UpsertAccountAsync(account);
                }

                this.table.Execute(insertOperation);

                if (docDbAccountTask != null)
                {
                    await docDbAccountTask;
                }
                          
                op.SetNumberOfRecordsAffected(2);
                op.SetResultSuccess();
            }
        }

        /// <summary>
        /// <para>
        /// (Soft) deletes an account.
        /// </para>
        /// Success is logged even if the account is not found.   (NumberOfRecordsAffected will be 1 if the account is found, 0 if not)
        /// </summary>
        /// <param name="id"></param>
        public async Task DeleteAccountRecord(AccountIdentifier id)
        {
            using (var op = new Instrumentation.ResourceProviderStorageOperation(nameof(DeleteAccountRecord)))
            {
                Validation.Requires.NotNull(id, nameof(id));

                op.LogCommonParameters(id.SubscriptionId, id.ResourceGroupName, id.ResourceName);

                var record = await GetAccountRecord(id);

                if (record != null)
                {
                    // soft delete
                    record.Deleted = true;
                    await PutAccountRecord(record);                    
                    op.SetNumberOfRecordsAffected(1);
                }
                else
                {
                    op.SetNumberOfRecordsAffected(0);
                }

                op.SetResultSuccess();
            }
        }

        /// <summary>
        /// <para>
        /// Gets all records by subscription id.
        /// </para>
        /// Logs Success even if no records found, as long as there are no exceptions from storage.
        /// </summary>
        /// <param name="subscriptionId"></param>
        /// <param name="includeDeleted"></param>
        /// <returns></returns>
        public IEnumerable<AccountRecord> GetAccountRecordsBySubscription(string subscriptionId,
                                                                          bool includeDeleted = false)
        {
            using (var op = new Instrumentation.ResourceProviderStorageOperation(nameof(GetAccountRecordsBySubscription)))
            {
                Validation.Requires.NotNullOrWhiteSpace(subscriptionId, nameof(subscriptionId));

                op.LogCommonParameters(subscriptionId);

                var query = new TableQuery<AccountRecord>()
                    .Where(TableQuery.GenerateFilterCondition("PartitionKey",
                                                              QueryComparisons.Equal,
                                                              subscriptionId.ToLowerInvariant()));

                int recordCount = 0;
                foreach (var record in this.table.ExecuteQuery(query).Where(r => !r.Deleted || includeDeleted))
                {
                    recordCount++;

                    if (!settings.EnableAccountReplication)
                    {
                        // set this null so we don't actually create IDs until we enable the feature.
                        record.ClientId = null;
                    }

                    yield return record;
                }

                op.SetNumberOfRecordsAffected(recordCount);
                op.SetResultSuccess();
            }
        }

        /// <summary>
        /// <para>
        /// Gets all records by resource group
        /// </para>
        /// Logs Success even if no records found, as long as there are no exceptions from storage.
        /// </summary>
        /// <param name="subscriptionId"></param>
        /// <param name="resourceGroupName"></param>
        /// <param name="includeDeleted"></param>
        /// <returns></returns>
        public IEnumerable<AccountRecord> GetAccountRecordsByResourceGroup(string subscriptionId,
                                                                           string resourceGroupName,
                                                                           bool includeDeleted = false)
        {
            using (var op = new Instrumentation.ResourceProviderStorageOperation(nameof(GetAccountRecordsBySubscription)))
            {
                Validation.Requires.NotNullOrWhiteSpace(subscriptionId, nameof(subscriptionId));
                Validation.Requires.NotNullOrWhiteSpace(resourceGroupName, nameof(resourceGroupName));

                op.LogCommonParameters(subscriptionId, resourceGroupName);

                // fetch all records in the subscription and include only those with the same resourceGroupName (case-insenstive compare)
                int recordCount = 0;
                foreach (
                    var record in
                    GetAccountRecordsBySubscription(subscriptionId, includeDeleted)
                        .Where(r => string.Equals(r.ResourceGroupName,
                                                 resourceGroupName,
                                                 StringComparison.InvariantCultureIgnoreCase)))
                {
                    recordCount++;

                    if (!settings.EnableAccountReplication)
                    {
                        // set this null so we don't actually create IDs until we enable the feature.
                        record.ClientId = null;
                    }

                    yield return record;
                }

                op.SetNumberOfRecordsAffected(recordCount);
                op.SetResultSuccess();
            }
        }

        internal static string PartitionKey(AccountIdentifier id)
        {
            return id.SubscriptionId.ToLowerInvariant();
        }

        /// <summary>
        /// the rowkey is prepended with the subscription id; this way, records with the same sub id are grouped together in a partition
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        internal static string RowKey(AccountIdentifier id)
        {
            return id.SubscriptionId.ToLowerInvariant() + ":" + Guid.NewGuid().ToString();
        }
    }
}