﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage
{
    using System;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Net;
    using System.Threading.Tasks;
    using Microsoft.Azure.Documents;
    using Microsoft.Azure.Documents.Client;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Utility;
    using Newtonsoft.Json.Linq;
    using Validation;

    public class DocumentDbStorage : IDocumentDbStorage
    {
        private const string partitionKeyConst = "pk";
        private const string trustedTenantIdConst = "tids";
        private const string resourceNameConst = "nm";
        private const string resourceGroupConst = "rg";
        private const string subscriptionIdConst = "subid";
        private const string primaryKeyConst = "key1";
        private const string secondaryKeyConst = "key2";
        private const string skuConst = "sku";
        private const string deletedConst = "deleted";
        private const string throttleConfigConst = "throttle";

        /// <summary>
        /// The document client used for account look and validation.
        /// </summary>
        private readonly IDocumentClient documentClient;

        /// <summary>
        /// Subscription Record store to retrieve the aad tenant ID for the subscription
        /// </summary>
        private readonly ISubscriptionRecordStorage subscriptionRecordStorage;

        /// <summary>
        /// Environment settings for the app.
        /// </summary>
        private readonly ResourceProviderHostSettings settings;

        private DocumentDbStorage(IDocumentClient documentClient, ISubscriptionRecordStorage subscriptionRecordStorage, ResourceProviderHostSettings settings)
        {
            this.documentClient = documentClient;
            this.subscriptionRecordStorage = subscriptionRecordStorage;
            this.settings = settings;
        }

        /// <summary>
        /// Factory method to construct a new instance of <see cref="AccountRecordStorage"/>
        /// </summary>
        /// <param name="configurationReader">configuration source</param>
        /// <param name="settings">the settings for the environment</param>
        /// <param name="subscriptionRecordStorage">Azure Subscription Storage repository</param>
        /// <returns>new instance of <see cref="DocumentDbStorage"/></returns>
        public static async Task<IDocumentDbStorage> BuildAsync(IConfigurationReader configurationReader,
            ResourceProviderHostSettings settings, ISubscriptionRecordStorage subscriptionRecordStorage)
        {
            try
            {
                Requires.NotNull(configurationReader, nameof(configurationReader));
                Requires.NotNull(settings, nameof(settings));
                Requires.NotNull(subscriptionRecordStorage, nameof(subscriptionRecordStorage));
                string cosmosDbAccountsKey = await configurationReader.GetConfigurationValueAsync(settings.CosmosDbAccountsKeyId);
                DocumentClient documentClient = new DocumentClient(settings.CosmosDbAccountsUri, cosmosDbAccountsKey, new ConnectionPolicy()
                {
                    ConnectionMode = ConnectionMode.Direct,
                    ConnectionProtocol = Protocol.Tcp,
                    EnableEndpointDiscovery = true,
                    MaxConnectionLimit = 250
                });
                await documentClient.OpenAsync();
                Database db = await documentClient.CreateDatabaseIfNotExistsAsync(new Database() { Id = CosmosDbConstants.AccountsDatabaseName });
                await documentClient.CreateDocumentCollectionIfNotExistsAsync(db.SelfLink, new DocumentCollection()
                {
                    PartitionKey = new PartitionKeyDefinition() { Paths = new Collection<string>() { "/" + partitionKeyConst } },
                    Id = CosmosDbConstants.AccountDocumentCollectionName,
                    IndexingPolicy = new IndexingPolicy()
                    {
                        Automatic = false,
                        IndexingMode = IndexingMode.None
                    }
                }, new RequestOptions() { OfferThroughput = settings.CosmosDbAccountsCollectionRequestUnits });
                
                await documentClient.CreateDocumentCollectionIfNotExistsAsync(db.SelfLink, new DocumentCollection()
                {
                    Id = CosmosDbConstants.SubscriptionCollectionName,
                    IndexingPolicy = new IndexingPolicy(new RangeIndex(DataType.String) { Precision = -1 })
                }, new RequestOptions() { OfferThroughput = settings.CosmosDbAccountsSubscriptionQueueCollectionRequestUnits });

                Agent.Instance.LogEvent(EventType.Information, $"Connected to document db account {settings.CosmosDbAccountsUri}");
                return new DocumentDbStorage(documentClient, subscriptionRecordStorage, settings);
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "resourceProvider", "DocumentDbAccountRecordStorage");
                throw;
            }
        }

        public async Task UpsertAccountAsync(AccountRecord accountRecord)
        {
            Requires.NotNull(accountRecord, nameof(accountRecord));

            if (accountRecord.Deleted)
            {
                await DeleteAccountAsync(accountRecord.ClientId);
                return;
            }

            // upsert the reverse lookup account for account validation.                
            SubscriptionRecord subRecord = subscriptionRecordStorage.GetSubscriptionRecord(accountRecord.SubscriptionId);
            if (subRecord == null)
            {
                Agent.Instance.LogEvent(EventType.Warning, $"subscription {accountRecord.SubscriptionId} doesn't exist in storage, will not create account record.");
                return;
            }

            Document existingRecord = null;
            try
            {
                var existingDocId = UriFactory.CreateDocumentUri(CosmosDbConstants.AccountsDatabaseName, CosmosDbConstants.AccountDocumentCollectionName, accountRecord.ClientId);
                existingRecord = await documentClient.ReadDocumentAsync(existingDocId, new RequestOptions()
                {
                    PartitionKey = new PartitionKey(accountRecord.ClientId)
                });                
            }
            catch (DocumentClientException dx)
            {
                if (dx.StatusCode != HttpStatusCode.NotFound)
                {
                    Agent.Instance.LogException(dx, "resourceProvider", "DocumentDbAccountRecordStorage");
                    throw;
                }       
            }

            JObject properties = JObject.Parse(subRecord.Properties);

            List<string> trustedTenantIds = new List<string>(5);
            string tenantId = properties["tenantId"]?.Value<string>();
            trustedTenantIds.Add(tenantId);
            Document record = existingRecord ?? new Document()
            {
                Id = accountRecord.ClientId
            };
                        
            record.SetPropertyValue(partitionKeyConst, accountRecord.ClientId);
            record.SetPropertyValue(resourceNameConst, accountRecord.ResourceName);
            record.SetPropertyValue(resourceGroupConst, accountRecord.ResourceGroupName);
            record.SetPropertyValue(subscriptionIdConst, accountRecord.SubscriptionId);
            record.SetPropertyValue(skuConst, accountRecord.SkuName);
            record.SetPropertyValue(primaryKeyConst, accountRecord.PrimaryKey);
            record.SetPropertyValue(secondaryKeyConst, accountRecord.SecondaryKey);
            record.SetPropertyValue(deletedConst, false);
            record.SetPropertyValue(trustedTenantIdConst, trustedTenantIds);

            ThrottlingConfiguration throttlingConfiguration = record.GetPropertyValue<ThrottlingConfiguration>(throttleConfigConst);

            if (throttlingConfiguration == null)
            {
                SkuName sku = (SkuName)Enum.Parse(typeof(SkuName), accountRecord.SkuName);
                ThrottlingConfiguration configuration = settings.DefaultThrottlingPerAccountMap[sku];
                record.SetPropertyValue(throttleConfigConst, configuration);
            }
            
            Uri docUri = UriFactory.CreateDocumentCollectionUri(CosmosDbConstants.AccountsDatabaseName, CosmosDbConstants.AccountDocumentCollectionName);
            try
            {
                await documentClient.UpsertDocumentAsync(docUri, record);
            }
            catch (DocumentClientException dx)
            {
                Agent.Instance.LogException(dx, "resourceProvider", "DocumentDbAccountRecordStorage");
                throw;
            }
        }

        public async Task DeleteAccountAsync(string id)
        {
            Requires.NotNullOrWhiteSpace(id, id);

            try
            {
                var existingDocId = UriFactory.CreateDocumentUri(CosmosDbConstants.AccountsDatabaseName, CosmosDbConstants.AccountDocumentCollectionName, id);
                Document existingDocument = await documentClient.ReadDocumentAsync(existingDocId, new RequestOptions()
                {
                    PartitionKey = new PartitionKey(id)
                });
                existingDocument.SetPropertyValue(deletedConst, true);
                Uri docUri = UriFactory.CreateDocumentCollectionUri(CosmosDbConstants.AccountsDatabaseName, CosmosDbConstants.AccountDocumentCollectionName);                                

                await documentClient.UpsertDocumentAsync(docUri, existingDocument);
            }
            catch (DocumentClientException dx)
            {
                if (dx.StatusCode == HttpStatusCode.NotFound)
                {                    
                    return;
                }
                
                Agent.Instance.LogException(dx, "resourceProvider", "DocumentDbAccountRecordStorage");
                throw;
            }
        }

        public async Task UpsertSubscriptionAsync(ResourceProviderWebAPI.Subscription subscription)
        {
            Requires.NotNull(subscription, nameof(subscription));
            Requires.NotNullOrWhiteSpace(subscription.Id, nameof(subscription.Id));
            try
            {
                Document document = new Document()
                {
                    Id = subscription.Id
                };
                document.SetPropertyValue(CosmosDbConstants.EnqueueTimePropertyConst, DateTime.UtcNow);
                document.SetPropertyValue(CosmosDbConstants.SubscriptionStatePropertyConst, subscription.State.ToString());

                var properties = (dynamic)subscription.Properties;
                document.SetPropertyValue(CosmosDbConstants.SubscriptionTenantIdPropertyConst, properties.tenantId);
                Uri docUri = UriFactory.CreateDocumentCollectionUri(CosmosDbConstants.AccountsDatabaseName, CosmosDbConstants.SubscriptionCollectionName);                                

                await documentClient.UpsertDocumentAsync(docUri, document);
            }
            catch (DocumentClientException dx)
            {
                Agent.Instance.LogException(dx, "resourceProvider", "DocumentDbAccountRecordStorage");
                throw;
            }
        }
    }
}
