﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage
{
    using System;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Utility;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.RetryPolicies;
    using Microsoft.WindowsAzure.Storage.Table;
    using Validation;

    internal class SubscriptionRecordStorage : ISubscriptionRecordStorage
    {
        private const string ResourceProviderSubscriptionTableNameKey = "mResourceProviderSubscriptionTableName";

        /// <summary>
        /// An integer specifying the maximum number of retry attempts.
        /// </summary>
        private const int MaxRetries = 3;

        /// <summary>
        /// The back-off interval between retries.
        /// </summary>
        private static readonly TimeSpan DeltaBackoff = TimeSpan.FromSeconds(2);

        /// <summary>
        /// The maximum execution time across all potential retries.
        /// </summary>
        private static readonly TimeSpan MaximumExecutionTime = TimeSpan.FromSeconds(30);

        /// <summary>
        /// The table request options.
        /// </summary>
        private static readonly TableRequestOptions TableRequestOptions = new TableRequestOptions
        {
            RetryPolicy = new ExponentialRetry(DeltaBackoff, MaxRetries),
            MaximumExecutionTime = MaximumExecutionTime
        };

        private readonly CloudTable table;

        private SubscriptionRecordStorage(CloudTable table)
        {
            this.table = table;
        }

        /// <summary>
        /// Factory method to construct a new instance of <see cref="SubscriptionRecordStorage"/>
        /// </summary>
        /// <param name="bootstrapConfiguration">configuration read as a bootstrap</param>
        /// <param name="configurationReader">configuration source</param>
        /// <returns>new instance of <see cref="SubscriptionRecordStorage"/></returns>
        public static async Task<SubscriptionRecordStorage> BuildAsync(IBootstrapConfiguration bootstrapConfiguration, IConfigurationReader configurationReader)
        {
            try
            {
                Requires.NotNull(configurationReader, nameof(configurationReader));
                Requires.NotNull(bootstrapConfiguration, nameof(bootstrapConfiguration));
                string connectionStringTemplate = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(Globals.StorageAccountConnectionTemplateKey));
                string storageAccountSasToken = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(Globals.StorageAccountSasTokenKey));
                string storageConnectionString = string.Format(connectionStringTemplate, storageAccountSasToken);
                CloudStorageAccount account = CloudStorageAccount.Parse(storageConnectionString);
                CloudTableClient client = account.CreateCloudTableClient();
                client.DefaultRequestOptions = TableRequestOptions;
                string tableName = await configurationReader.GetConfigurationValueAsync(bootstrapConfiguration.GetValue(ResourceProviderSubscriptionTableNameKey));
                CloudTable table = client.GetTableReference(tableName);
                table.CreateIfNotExists();

                Agent.Instance.LogEvent(EventType.Information, $"Connected to table {table.StorageUri.PrimaryUri}");

                return new SubscriptionRecordStorage(table);
            }
            catch (Exception ex)
            {
                Agent.Instance.LogException(ex, "resourceProvider", "subscriptionRecordStorage");
                throw;
            }
        }

        public void PutSubscriptionRecord(SubscriptionRecord record)
        {
            using (var op = new Instrumentation.ResourceProviderSubscriptionStorageOperation("PutSubscriptionRecord"))
            {
                Validation.Requires.NotNull(record, nameof(record));
                op.LogCommonParameters(record.PartitionKey);

                var insertOperation = TableOperation.InsertOrReplace(record);
                this.table.Execute(insertOperation);

                op.SetResultSuccess();
            }
        }

        public SubscriptionRecord GetSubscriptionRecord(string subscriptionId)
        {
            using (var op = new Instrumentation.ResourceProviderSubscriptionStorageOperation("PutSubscriptionRecord"))
            {
                Validation.Requires.NotNull(subscriptionId, nameof(subscriptionId));
                var partitionKey = PartitionKey(subscriptionId);
                var rowKey = RowKey(subscriptionId);
                op.LogCommonParameters(partitionKey);

                var operation = TableOperation.Retrieve<SubscriptionRecord>(partitionKey, rowKey);
                var result = this.table.Execute(operation);

                op.SetResultSuccess();

                return result?.Result as SubscriptionRecord;
            }
        }

        internal static string PartitionKey(string subscriptionId)
        {
            return subscriptionId.ToLowerInvariant();
        }

        internal static string RowKey(string subscriptionId)
        {
            return "s";   // one record per subscription, so it does not matter what this is
        }
    }
}