﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Utility
{
    using System;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Globalization;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Instrumentation;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.ApiManagement;
    using Microsoft.Azure.LocationServices.ResourceProviderWebAPI.Storage;

    /// <summary>
    /// Represents the configuration necessary to operate in an environment.
    /// </summary>
    /// <remarks>
    /// This class should only contain members that differ from environment to environment. When something doesn't 
    /// vary, it should not be present in this class, and instead should be a compile-time constant. If a property
    /// is no longer used, or no longer varies, it should be removed.
    /// </remarks>
    public abstract class ResourceProviderHostSettings
    {
        private static readonly Dictionary<SkuName, ThrottlingConfiguration> throttlingPerAccountMap = new Dictionary<SkuName, ThrottlingConfiguration>()
        {
            { SkuName.S0, new ThrottlingConfiguration() { NumberOfCalls = 300, RenewalPeriodSeconds = 5 } }
        };

        /// <summary>
        /// Gets the Instrumentation Parameters
        /// </summary>
        public abstract InitParams MdmParameters { get; }

        /// <summary>
        /// Gets the mapping from Sku name to Azure API Management group id and product id
        /// </summary>
        public abstract IReadOnlyDictionary<SkuName, ApiManagementStorageOptions> SkuApimMap { get; }

        /// <summary>
        /// Preferred Locations for cosmos db read locations.
        /// </summary>
        public abstract Collection<string> CosmosDbPreferredLocations { get; }
       
        /// <summary>
        /// Document DB accounts URI
        /// </summary>
        public Uri CosmosDbAccountsUri => new Uri($"https://{MapsFrontEndCosmosDbAccountName}.documents.azure.com");

        /// <summary>
        /// Document DB account key secret id
        /// </summary>
        public string CosmosDbAccountsKeyId => $"https://{GlobalKeyVaultName}.vault.azure.net:443/secrets/mapsFrontEndCosmosDbKey";

        /// <summary>
        /// Document DB account collection request unit throughput, adjust this to scale.
        /// </summary>
        public int CosmosDbAccountsCollectionRequestUnits => 5500;

        /// <summary>
        /// Document DB subscription queue collection request unit throughput, adjust to scale.
        /// </summary>
        public int CosmosDbAccountsSubscriptionQueueCollectionRequestUnits => 5500;

        /// <summary>
        /// Flag to enable cosmos DB account feature work.
        /// </summary>
        public abstract bool EnableAccountReplication { get; }

        /// <summary>
        /// The Default throttling per account on a particular sku
        /// </summary>
        public Dictionary<SkuName, ThrottlingConfiguration> DefaultThrottlingPerAccountMap
        {
            get
            {
                return throttlingPerAccountMap;
            }
        }

        /// <summary>
        /// Gets the current region.
        /// </summary>
        protected static string CurrentRegion
        {
            get
            {
                string currentRegion = Environment.GetEnvironmentVariable("Region");
                return currentRegion ?? "DEFAULT";
            }
        }

        /// <summary>
        /// Name of the account for maps front end cosmos db instance
        /// </summary>
        protected abstract string MapsFrontEndCosmosDbAccountName { get; }

        /// <summary>
        /// Name of the key vault where all runtime secrets are referenced
        /// </summary>
        protected abstract string GlobalKeyVaultName { get; }

        /// <summary>
        /// Gets the current environment.
        /// </summary>
        /// <exception cref="ArgumentException">
        /// Throws if the current environment is not specified
        /// </exception>
        private static string CurrentEnvironment
        {
            get
            {
                string currentEnvironment = Environment.GetEnvironmentVariable("Environment");

                if (string.IsNullOrEmpty(currentEnvironment))
                {
                    throw new ArgumentException("Current Environment could not be determined");
                }

                return currentEnvironment;
            }
        }

        /// <summary>
        /// Create an environment definition from the environment environment variable.
        /// </summary>
        /// <returns>The created environment definition.</returns>
        public static ResourceProviderHostSettings CreateFromEnvironment()
        {
            switch (CurrentEnvironment.ToUpperInvariant())
            {
                case LogicalEnvironments.Development:
                    return new ResourceProviderHostDevSettings();
                case LogicalEnvironments.Test:
                    return new ResourceProviderHostTestSettings();
                case LogicalEnvironments.Staging:
                    return new ResourceProviderHostStagingSettings();
                case LogicalEnvironments.Production:
                    return new ResourceProviderHostProdSettings();
                default:
                    throw new ArgumentException(
                        string.Format(
                            CultureInfo.InvariantCulture,
                            "Environment {0} does not have a corresponding Environment definition defined",
                            CurrentEnvironment));
            }
        }
    }
}