import { Map } from "../../map/Map";
import { ControlOptions, ControlPosition } from "../../map/options/ControlOptions";
import { Control } from "../Control";
import { CompassControlOptions } from "./CompassControlOptions";

/**
 * A control for changing the rotation of the map.
 * @module atlas.controls
 */
export class CompassControl implements Control {
    private static ROTATION_DURATION_MS = 100;
    private static DEFAULT_ROTATION = 0;

    private container: HTMLElement;
    private options: CompassControlOptions;
    private map: Map;

    /**
     * Constructs a CompassControl.
     * @param options The options for the control.
     */
    constructor(options?: CompassControlOptions) {
        this.options = new CompassControlOptions().merge(options);
    }

    public onAdd(map: Map, options: ControlOptions): HTMLElement {
        this.map = map;

        this.container = document.createElement("div");
        this.container.classList.add("azure-maps-control-container");
        this.container.setAttribute("aria-label", "Rotation Control");

        const rotationButton = this.constructRotationButton();
        const rotationRightButton = this.constructRightRotationButton();
        const rotationLeftButton = this.constructLeftRotationButton();

        this.container.addEventListener("mouseover", (event) => {
            rotationRightButton.classList.remove("hidden-accessible-element");
            rotationLeftButton.classList.remove("hidden-accessible-element");
        });

        this.container.addEventListener("mouseout", (event) => {
            rotationRightButton.classList.add("hidden-accessible-element");
            rotationLeftButton.classList.add("hidden-accessible-element");
        });

        if (options.position === ControlPosition.TopRight ||
            options.position === ControlPosition.BottomRight) {
            this.container.style.flexDirection = "row-reverse";
            this.container.appendChild(rotationButton);
            this.container.appendChild(rotationLeftButton);
            this.container.appendChild(rotationRightButton);
        } else {
            this.container.style.flexDirection = "row";
            this.container.appendChild(rotationButton);
            this.container.appendChild(rotationRightButton);
            this.container.appendChild(rotationLeftButton);
        }

        return this.container;
    }

    private constructRotationButton(): HTMLElement {
        const rotationButton = document.createElement("button");
        rotationButton.classList.add("azure-maps-control-button");
        rotationButton.classList.add("rotation");
        rotationButton.classList.add(this.options.style);
        rotationButton.setAttribute("title", "Reset to Default Rotation");

        rotationButton.addEventListener("click", (event) => {
            this.map.setCamera({
                bearing: CompassControl.DEFAULT_ROTATION
            });
        });

        return rotationButton;
    }

    private constructRightRotationButton(): HTMLElement {
        const rotationRightButton = document.createElement("button");
        rotationRightButton.classList.add("azure-maps-control-button");
        rotationRightButton.classList.add("rotation-right");
        rotationRightButton.classList.add("hidden-accessible-element");
        rotationRightButton.classList.add(this.options.style);
        rotationRightButton.setAttribute("title", "Rotate Right");

        rotationRightButton.addEventListener("click", (event) => {
            this.map.setCamera({
                bearing: this.map.getCamera().bearing + this.options.rotationDegreesDelta,
                type: "ease",
                duration: CompassControl.ROTATION_DURATION_MS
            });
        });

        return rotationRightButton;
    }

    private constructLeftRotationButton(): HTMLElement {
        const rotationLeftButton = document.createElement("button");
        rotationLeftButton.classList.add("azure-maps-control-button");
        rotationLeftButton.classList.add("rotation-left");
        rotationLeftButton.classList.add("hidden-accessible-element");
        rotationLeftButton.classList.add(this.options.style);
        rotationLeftButton.setAttribute("title", "Rotate Left");

        rotationLeftButton.addEventListener("click", (event) => {
            this.map.setCamera({
                bearing: this.map.getCamera().bearing - this.options.rotationDegreesDelta,
                type: "ease",
                duration: CompassControl.ROTATION_DURATION_MS
            });
        });

        return rotationLeftButton;
    }

    public onRemove() {
        this.container.remove();
    }

}
