import { Map } from "../../map/Map";
import { ControlOptions, ControlPosition } from "../../map/options/ControlOptions";
import { Control } from "../Control";
import { PitchControlOptions } from "./PitchControlOptions";

/**
 * A control for changing the pitch of the map.
 * @module atlas.controls
 */
export class PitchControl implements Control {
    private static PITCH_DURATION_MS = 100;
    private static DEFAULT_PITCH = 0;

    private container: HTMLElement;
    private options: PitchControlOptions;
    private map: Map;

    /**
     * Constructs a PitchControl.
     * @param options The options for the control.
     */
    constructor(options?: PitchControlOptions) {
        this.options = new PitchControlOptions().merge(options);
    }

    public onAdd(map: Map, options: ControlOptions): HTMLElement {
        this.map = map;

        this.container = document.createElement("div");
        this.container.classList.add("azure-maps-control-container");
        this.container.setAttribute("aria-label", "Pitch Control");

        this.container.style.flexDirection = (options.position === ControlPosition.TopRight ||
            options.position === ControlPosition.BottomRight) ? "row-reverse" : "row";

        const pitchButton = this.constructPitchButton();
        const pitchIncrementButton = this.constructPitchIncrementButton();
        const pitchDecrementButton = this.constructPitchDecrementButton();

        this.container.addEventListener("mouseover", (event) => {
            pitchIncrementButton.classList.remove("hidden-accessible-element");
            pitchDecrementButton.classList.remove("hidden-accessible-element");
        });

        this.container.addEventListener("mouseout", (event) => {
            pitchIncrementButton.classList.add("hidden-accessible-element");
            pitchDecrementButton.classList.add("hidden-accessible-element");
        });

        this.container.appendChild(pitchButton);
        this.container.appendChild(pitchIncrementButton);
        this.container.appendChild(pitchDecrementButton);
        return this.container;
    }

    private constructPitchButton() {
        const pitchButton = document.createElement("button");
        pitchButton.classList.add("azure-maps-control-button");
        pitchButton.classList.add("pitch");
        pitchButton.classList.add(this.options.style);
        pitchButton.setAttribute("title", "Reset to Default Pitch");

        pitchButton.addEventListener("click", (event) => {
            this.map.setCamera({
                pitch: PitchControl.DEFAULT_PITCH
            });
        });

        return pitchButton;
    }

    private constructPitchDecrementButton() {
        const pitchDecrementButton = document.createElement("button");
        pitchDecrementButton.classList.add("azure-maps-control-button");
        pitchDecrementButton.classList.add("pitch-down");
        pitchDecrementButton.classList.add("hidden-accessible-element");
        pitchDecrementButton.classList.add(this.options.style);
        pitchDecrementButton.setAttribute("title", "Decrease Pitch");

        pitchDecrementButton.addEventListener("click", (event) => {
            this.map.setCamera({
                pitch: this.map.getCamera().pitch - this.options.pitchDegreesDelta,
                type: "ease",
                duration: PitchControl.PITCH_DURATION_MS
            });
        });

        return pitchDecrementButton;
    }

    private constructPitchIncrementButton() {
        const pitchIncrementButton = document.createElement("button");
        pitchIncrementButton.classList.add("azure-maps-control-button");
        pitchIncrementButton.classList.add("pitch-up");
        pitchIncrementButton.classList.add("hidden-accessible-element");
        pitchIncrementButton.classList.add(this.options.style);
        pitchIncrementButton.setAttribute("title", "Increase Pitch");

        pitchIncrementButton.addEventListener("click", (event) => {
            this.map.setCamera({
                pitch: this.map.getCamera().pitch + this.options.pitchDegreesDelta,
                type: "ease",
                duration: PitchControl.PITCH_DURATION_MS
            });
        });

        return pitchIncrementButton;
    }

    public onRemove() {
        this.container.remove();
    }

}
