import { env } from "../../../variables.json";
import { Url, UrlOptions } from "../../helpers/Url";
import { Map } from "../../map/Map";
import { ControlOptions, ControlPosition } from "../../map/options/ControlOptions";
import { Style } from "../../map/style/Style";
import { Control } from "../Control";
import { StyleControlOptions } from "./StyleControlOptions";

interface StyleDef { name: string; iconUrl: string; }

/**
 * A control for changing the style of the map.
 * @module atlas.controls
 */
export class StyleControl implements Control {
    private container: HTMLElement;
    private options: StyleControlOptions;
    private styleOptionsGrid: HTMLElement;
    private currStyleImage: HTMLImageElement;
    private map: Map;
    private availableStyles: { [name: string]: StyleDef } = {};

    /**
     * Constructs a StyleControl.
     * @param options The options for the control.
     */
    constructor(options?: StyleControlOptions) {
        this.options = new StyleControlOptions().merge(options);

        this.container = document.createElement("div");
        this.container.classList.add("azure-maps-control-container");
        this.container.setAttribute("aria-label", "Map Style Control");

        this.styleOptionsGrid = this.constructStyleOptionsGrid();
        this.container.addEventListener("mouseover", (event) => {
            this.styleOptionsGrid.classList.remove("hidden-accessible-element");
        });
        this.container.addEventListener("mouseout", (event) => {
            this.styleOptionsGrid.classList.add("hidden-accessible-element");
        });

        const currStyleButton = this.constructCurrStyleButton(this.options.style);
        this.container.appendChild(currStyleButton);
        this.container.appendChild(this.styleOptionsGrid);

        const styleDefinitionsUrl = new Url(new UrlOptions().merge({
            domain: env.staticAssetsDomain,
            path: env.styleDefinitionsPath
        }));

        styleDefinitionsUrl.get().then((response) => {
            response["styles"].forEach((s) => {
                const iconUrl = new Url(new UrlOptions().merge({
                    domain: response["domain"],
                    path: s["iconPath"]
                }));

                const style: StyleDef = {
                    name: s["name"],
                    iconUrl: iconUrl.toString()
                };

                this.availableStyles[s.name] = style;

                const styleOptionButton = this.constructSelectStyleButton(style);
                this.styleOptionsGrid.appendChild(styleOptionButton);

                const currMapStyle = this.map.getStyle().style;
                if (style.name === currMapStyle) {
                    this.currStyleImage.src = style.iconUrl;
                }
            });
        });
    }

    private constructSelectStyleButton(style: StyleDef) {
        const styleOptionButton = document.createElement("button");
        styleOptionButton.classList.add("azure-maps-control-button");
        styleOptionButton.classList.add(this.options.style);
        styleOptionButton.setAttribute("title", style.name);
        styleOptionButton.dataset.style = style.name;

        const styleIconImage = new Image(28, 28);
        styleIconImage.src = style.iconUrl;
        styleIconImage.alt = style.name;
        styleOptionButton.appendChild(styleIconImage);

        styleOptionButton.addEventListener("click", (event) => {
            if (event.currentTarget instanceof HTMLElement) {
                const clickedButton: HTMLElement = event.currentTarget;
                this.map.setStyle({ style: clickedButton.dataset.style });
            }
        });

        return styleOptionButton;
    }

    private constructCurrStyleButton(styleName: string): HTMLElement {
        const selectCurrButton = document.createElement("button");
        selectCurrButton.classList.add("azure-maps-control-button");
        selectCurrButton.classList.add("curr-style");
        selectCurrButton.classList.add(styleName);
        selectCurrButton.setAttribute("title", "Select Style");

        this.currStyleImage = new Image(28, 28);
        selectCurrButton.appendChild(this.currStyleImage);

        const selectCurrButtonIcon = document.createElement("div");
        selectCurrButtonIcon.classList.add("icon");
        selectCurrButton.appendChild(selectCurrButtonIcon);

        return selectCurrButton;
    }

    private constructStyleOptionsGrid(): HTMLElement {
        const styleOptionsGrid = document.createElement("div");
        styleOptionsGrid.classList.add("style-options");
        styleOptionsGrid.classList.add("hidden-accessible-element");
        styleOptionsGrid.setAttribute("aria-label", "Style Options");

        return styleOptionsGrid;
    }

    public onAdd(map: Map, options?: ControlOptions): HTMLElement {
        this.map = map;

        this.container.style.flexDirection = (options.position === ControlPosition.TopRight ||
            options.position === ControlPosition.BottomRight) ? "row-reverse" : "row";

        this.map.addEventListener("styledata", () => {
            const currMapStyle = this.map.getStyle().style;

            if (this.availableStyles.hasOwnProperty(currMapStyle)) {
                const style: StyleDef = this.availableStyles[currMapStyle];
                this.currStyleImage.src = style.iconUrl;
            }
        });

        return this.container;
    }

    public onRemove() {
        this.container.remove();
    }

}
