import { Map } from "../../map/Map";
import { Control } from "../Control";
import { ZoomControlOptions } from "./ZoomControlOptions";

/**
 * A control for changing the zoom of the map.
 * @module atlas.controls
 */
export class ZoomControl implements Control {
    private static ZOOM_DURATION_MS = 200;

    private container: HTMLElement;
    private options: ZoomControlOptions;

    /**
     * Constructs a ZoomControl.
     * @param options The options for the control.
     */
    constructor(options?: ZoomControlOptions) {
        this.options = new ZoomControlOptions().merge(options);
    }

    public onAdd(map: Map): HTMLElement {
        this.container = document.createElement("div");
        this.container.classList.add("azure-maps-control-container");
        this.container.setAttribute("aria-label", "Zoom Control");
        this.container.style.flexDirection = "column";

        const zoomInButton = this.constructZoomInButton(map);
        const zoomOutButton = this.constructZoomOutButton(map);

        this.container.appendChild(zoomInButton);
        this.container.appendChild(zoomOutButton);
        return this.container;
    }

    private constructZoomInButton(map: Map) {
        const zoomInButton = document.createElement("button");
        zoomInButton.classList.add("azure-maps-control-button");
        zoomInButton.classList.add("zoom-in");
        zoomInButton.classList.add(this.options.style);
        zoomInButton.setAttribute("title", "Zoom In");
        zoomInButton.addEventListener("click", (event) => {
            map.setCamera({
                zoom: map.getCamera().zoom + this.options.zoomDelta,
                type: "ease",
                duration: ZoomControl.ZOOM_DURATION_MS
            });
        });

        return zoomInButton;
    }

    private constructZoomOutButton(map: Map) {
        const zoomOutButton = document.createElement("button");
        zoomOutButton.classList.add("azure-maps-control-button");
        zoomOutButton.classList.add("zoom-out");
        zoomOutButton.classList.add(this.options.style);
        zoomOutButton.setAttribute("title", "Zoom Out");
        zoomOutButton.addEventListener("click", (event) => {
            map.setCamera({
                zoom: map.getCamera().zoom - this.options.zoomDelta,
                type: "ease",
                duration: ZoomControl.ZOOM_DURATION_MS
            });
        });

        return zoomOutButton;
    }

    public onRemove() {
        this.container.remove();
    }

}
