﻿import * as _ from "lodash";
import * as mapboxgl from "mapbox-gl";
import { env } from "../../variables.json";
import { AzureLogoControl } from "../controls/AzureLogoControl";
import { Control } from "../controls/Control";
import { CopyrightControl } from "../controls/CopyrightControl";
import {
    BoundingBox, Feature, LineString, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon,
    Position
} from "../data";
import { Insights } from "../instrumentation/Insights";
import { ServiceOptions } from "../services/options/ServiceOptions";
import { CircleProperties } from "./data_properties/CircleProperties";
import { LinestringProperties } from "./data_properties/LinestringProperties";
import { PinProperties } from "./data_properties/PinProperties";
import { PolygonProperties } from "./data_properties/PolygonProperties";
import { AccessibleMapDelegate } from "./delegates/accessibility/AccessibleMapDelegate";
import { FlowServiceDelegate } from "./delegates/flow/FlowServiceDelegate";
import { IncidentServiceDelegate } from "./delegates/incidents/IncidentServiceDelegate";
import { MapCallbackHandler } from "./events/MapCallbackHandler";
import { AnimationOptions } from "./options/AnimationOptions";
import { CameraBoundsOptions } from "./options/CameraBoundsOptions";
import { CameraOptions } from "./options/CameraOptions";
import { CircleLayerOptions } from "./options/CircleLayerOptions";
import { ControlOptions, ControlPosition } from "./options/ControlOptions";
import { LinestringLayerOptions } from "./options/LinestringLayerOptions";
import { PinLayerOptions } from "./options/PinLayerOptions";
import { PolygonLayerOptions } from "./options/PolygonLayerOptions";
import { RasterLayerOptions } from "./options/RasterLayerOptions";
import { StyleOptions } from "./options/StyleOptions";
import { TrafficOptions } from "./options/TrafficOptions";
import { UserInteractionOptions } from "./options/UserInteractionOptions";
import { CircleLayer } from "./style/layers/basic/CircleLayer";
import { LinestringLayer } from "./style/layers/basic/LinestringLayer";
import { PinLayer } from "./style/layers/basic/PinLayer";
import { PolygonLayer } from "./style/layers/basic/PolygonLayer";
import { RasterLayer } from "./style/layers/basic/RasterLayer";
import { Layer } from "./style/layers/Layer";
import { StyleBuilder } from "./style/StyleBuilder";
import { StyleDefinitions } from "./style/StyleDefinitions";

/**
 * The control for a visual and interactive web map.
 * @module atlas
 */
export class Map {
    private readonly map: mapboxgl.Map;
    private readonly styleBuilder: StyleBuilder;
    private readonly htmlElements: { [id: string]: mapboxgl.Marker } = {};
    private htmlElementId: number = 0;
    private mapCallbackHandler: MapCallbackHandler;
    private styleDefinitions: StyleDefinitions;
    private insights: Insights;

    // Control Containers
    private controlContainer: HTMLElement;
    private copyrightControl: CopyrightControl;

    // Options
    private serviceOptions: ServiceOptions;
    private styleOptions: StyleOptions;
    private trafficOptions: TrafficOptions;
    private userInteractionOptions: UserInteractionOptions;

    // Delegates
    private readonly incidentDelegate: IncidentServiceDelegate;
    private readonly flowDelegate: FlowServiceDelegate;
    private readonly accessibleMapDelegate: AccessibleMapDelegate;

    /**
     * Displays a map in the specified container.
     * @param container The id of the element where the map should be displayed.
     * @param options Options for the initial display and interactability with the map.
     */
    constructor(container: string, options: ServiceOptions & CameraOptions & StyleOptions & UserInteractionOptions) {
        let successfulLoad = true;
        this.serviceOptions = new ServiceOptions().merge(options);
        this.styleOptions = new StyleOptions().merge(options);
        this.userInteractionOptions = new UserInteractionOptions().merge(options);
        const cameraOptions: CameraOptions = new CameraOptions().merge(options);
        this.trafficOptions = new TrafficOptions();

        this.styleBuilder = new StyleBuilder();
        this.styleDefinitions = new StyleDefinitions();
        this.insights = new Insights(this.serviceOptions);

        const mapboxMapOptions: mapboxgl.MapboxOptions = {
            container: container,
            attributionControl: false,
            transformRequest: (url: string, resourceType: string) => {
                if (resourceType === "Tile" && _.startsWith(url, `https://${env.domain}`)) {
                    this.insights.trackEvent("TileLoad");
                    const billingInfo = this.styleDefinitions.getBillingHeaderAndCleanUpURL(url);
                    const billingHeader = billingInfo["billing"];
                    url = billingInfo["url"];
                    return {
                        url: url,
                        headers: {
                            "Session-Id": this.serviceOptions["session-id"],
                            "Ms-Am-Request-Origin": "MapControl",
                            ...billingHeader
                        }
                    };
                }
            }
        } as any;

        if (this.serviceOptions["subscription-key"]) {
            // AuthUserId (session id) / Account Id (sub key)
            this.insights.setAuthenticatedUserContext(this.serviceOptions["session-id"], this.serviceOptions["subscription-key"]);
        }

        try {
            const mapboxMap: mapboxgl.Map = new mapboxgl.Map(mapboxMapOptions);
            this.map = mapboxMap;

            this.map.getCanvasContainer().id = "atlas-map";

            // Add containers for map controls
            this.controlContainer = document.createElement("div");
            this.controlContainer.classList.add("atlas-control-container");

            for (const controlPosition in ControlPosition) {
                if (ControlPosition.hasOwnProperty(controlPosition)) {
                    const subcontrolContainer = document.createElement("div");
                    subcontrolContainer.classList.add(ControlPosition[controlPosition]);
                    subcontrolContainer.classList.add("subcontrol-container");
                    this.controlContainer.appendChild(subcontrolContainer);
                }
            }

            this.getMapContainer().appendChild(this.controlContainer);

            // Add controls for the Azure logo and copyrights
            this.addControl(new AzureLogoControl(), {
                position: ControlPosition.NonFixed
            });
            this.copyrightControl = new CopyrightControl();
            this.addControl(this.copyrightControl, {
                position: ControlPosition.NonFixed
            });

            // Initialize state of map

            // --> Initialize map to default style
            this.map.setStyle(this.styleBuilder.build());

            // --> Download style definitions and restyle map with data
            this.styleDefinitions.initialize().then(() => {
                this.setStyle(this.styleOptions);
            });

            // --> Set initial camera state of map
            this.setCamera({ ...cameraOptions, type: "jump", duration: 0 });

            // --> Set initial user interactions of map
            this.setUserInteraction(this.userInteractionOptions);

            // Initialize the callback handler
            this.mapCallbackHandler = new MapCallbackHandler();

            // Add delegates to map
            this.incidentDelegate = new IncidentServiceDelegate(this);
            this.flowDelegate = new FlowServiceDelegate(this);
            this.accessibleMapDelegate = new AccessibleMapDelegate(this);
            this.accessibleMapDelegate.addToMap();
        } catch (e) {
            successfulLoad = false;
            this.insights.trackException(e);
            throw e;
        } finally {
            // Track the successful session load
            this.insights.trackEvent("MapLoad", { success: String(successfulLoad) });
            this.insights.flush();
        }
    }

    /**
     * Returns the HTMLCanvasElement that the map is drawn to.
     */
    public getCanvas(): HTMLCanvasElement {
        return this.map.getCanvas();
    }

    /**
     * Returns the HTMLElement that contains the map's HTMLCanvasElement. The map's events (e.g. panning and zooming)
     * are attached to this element.
     */
    public getCanvasContainer(): HTMLElement {
        return this.map.getCanvasContainer();
    }

    /**
     * Returns the HTMLElement that contains the map.
     */
    public getMapContainer(): HTMLElement {
        return this.map.getContainer();
    }

    /**
     * Returns the service options with which the map control was initialized.
     */
    public getServiceOptions(): ServiceOptions {
        return this.serviceOptions;
    }

    /**
     * Set the camera of the map control with an animated transition. Any options not specified will default to their
     * current values.
     * @param options The options for setting the map's camera and for the animation of any view change.
     */
    public setCamera(options?: CameraOptions & AnimationOptions) {
        const cameraOptions: CameraOptions = new CameraOptions().merge(this.getCamera(), options);
        const animOptions: AnimationOptions = new AnimationOptions().merge(options);

        const mapboxCameraOptions: mapboxgl.CameraOptions = {
            center: cameraOptions.center,
            zoom: cameraOptions.zoom,
            bearing: cameraOptions.bearing,
            pitch: cameraOptions.pitch,
            around: undefined
        };

        const mapboxAnimationOptions: mapboxgl.AnimationOptions = {
            duration: animOptions.duration,
            easing: (t: number) => t,
            offset: [0, 0],
            animate: true
        };

        this.map.setMinZoom(cameraOptions.minZoom);
        this.map.setMaxZoom(cameraOptions.maxZoom);

        if (animOptions.type === "jump") {
            this.map.jumpTo(mapboxCameraOptions);
        } else if (animOptions.type === "ease") {
            this.map.easeTo({ ...mapboxCameraOptions, ...mapboxAnimationOptions });
        } else if (animOptions.type === "fly") {
            this.map.flyTo({ ...mapboxCameraOptions, ...mapboxAnimationOptions });
        }
    }

    /**
     * Set the camera bounds of the map control.
     * @param options The options for setting the map's camera bounds.
     */
    public setCameraBounds(options?: CameraBoundsOptions) {
        const cameraBoundsOptions: CameraBoundsOptions = new CameraBoundsOptions().merge(options);

        const sourthwestPosition: Position = new Position(cameraBoundsOptions.bounds[0], cameraBoundsOptions.bounds[1]);
        const northeastPosition: Position = new Position(cameraBoundsOptions.bounds[2], cameraBoundsOptions.bounds[3]);
        const padding: any = (Array.isArray(cameraBoundsOptions.padding)) ? {
            top: cameraBoundsOptions.padding[0],
            right: cameraBoundsOptions.padding[1],
            bottom: cameraBoundsOptions.padding[2],
            left: cameraBoundsOptions.padding[3]
        } : cameraBoundsOptions.padding;

        this.map.fitBounds([sourthwestPosition, northeastPosition], {
            duration: 0,
            padding: padding
        } as any);
    }

    /**
     * Returns the camera's current properties.
     */
    public getCamera(): CameraOptions & CameraBoundsOptions {
        const southwestLngLat: mapboxgl.LngLat = this.map.getBounds().getSouthWest().wrap();
        const northeastLngLat: mapboxgl.LngLat = this.map.getBounds().getNorthEast().wrap();

        const cameraOptions: CameraOptions & CameraBoundsOptions = {
            zoom: this.map.getZoom(),
            center: this.map.getCenter().toArray(),
            bearing: this.map.getBearing(),
            pitch: this.map.getPitch(),
            bounds: new BoundingBox(southwestLngLat.toArray(), northeastLngLat.toArray()),
            minZoom: this.map.getMinZoom(),
            maxZoom: this.map.getMaxZoom()
        };

        return cameraOptions;
    }

    /**
     * Set the map control's style options. Any options not specified will default to their current values.
     * @param options The options for setting the style of the map control.
     */
    public setStyle(options?: StyleOptions) {
        this.styleOptions = new StyleOptions().merge(this.styleOptions, options);

        const styleLayerOptions: StyleOptions & ServiceOptions = _.merge({}, this.styleOptions, this.serviceOptions);

        this.styleDefinitions.getLayers(styleLayerOptions).then((layers) => {
            this.styleBuilder.upsertAll(layers);
            this.styleBuilder.setGlyphs(this.styleDefinitions.getGlyphs(options));
            this.styleBuilder.setSprite(this.styleDefinitions.getSprite(options));
            this.map.setStyle(this.styleBuilder.build());
            this.copyrightControl.update(this.styleDefinitions.getCopyrightCaption(options));
        });
    }

    /**
     * Returns the map control's current style settings.
     */
    public getStyle(): StyleOptions {
        return this.styleOptions;
    }

    /**
     * Add an icon to the map to use for pins. Map and image must be fully loaded before the icon can be added. Use the
     * addEventListener method with event type 'load'.
     * @param id The identifier of the icon.
     * @param icon The icon image.
     */
    public addIcon(id: string, icon: HTMLImageElement) {
        this.map.addImage(id, icon);
    }

    /**
     * Add a collection of points to a layer of the map as pins. The layer and its options can be specified through a
     * [PinLayerOptions]{@link azure-maps-javascript.PinLayerOptions} object. Options for the layer can
     * only be specified upon the layer's initial creation. Map must be fully loaded before the pins can be added. Use
     * the addEventListener method with event type 'load'.
     * @param pins The points to add.
     * @param options The layer options for the pins.
     */
    public addPins(pins: Array<Feature<Point, PinProperties>>, options?: PinLayerOptions) {
        const layerOptions: PinLayerOptions = new PinLayerOptions().merge(options);
        const layer: Layer = this.styleBuilder.get(layerOptions.name) || new PinLayer(layerOptions);

        if (layerOptions.overwrite) {
            layer.clear();
        }
        layer.push(pins);

        this.styleBuilder.upsert(layer, layerOptions.before);
        if (!layerOptions.defer) {
            const geoJsonSource = this.map.getSource(layer.sourceName) as mapboxgl.GeoJSONSource;
            if (geoJsonSource) {
                geoJsonSource.setData(layer.sources[layer.sourceName].data);
            } else {
                this.map.setStyle(this.styleBuilder.build());
            }
        }
    }

    /**
     * Add a collection of points to a layer of the map as circles. The layer and its options can be specified through a
     * [CircleLayerOptions]{@link azure-maps-javascript.CircleLayerOptions} object. Options for the layer
     * can only be specified upon the layer's initial creation. Map must be fully loaded before the circles can be
     * added. Use the addEventListener method with event type 'load'.
     * @param circles The circles to add.
     * @param options The layer options for the circles.
     */
    public addCircles(circles: Array<Feature<Point, CircleProperties>>, options?: CircleLayerOptions) {
        const layerOptions: CircleLayerOptions = new CircleLayerOptions().merge(options);
        const layer: Layer = this.styleBuilder.get(layerOptions.name) || new CircleLayer(layerOptions);

        if (layerOptions.overwrite) {
            layer.clear();
        }
        layer.push(circles);

        this.styleBuilder.upsert(layer, layerOptions.before);
        if (!layerOptions.defer) {
            const geoJsonSource = this.map.getSource(layer.sourceName) as mapboxgl.GeoJSONSource;
            if (geoJsonSource) {
                geoJsonSource.setData(layer.sources[layer.sourceName].data);
            } else {
                this.map.setStyle(this.styleBuilder.build());
            }
        }
    }

    /**
     * Add a collection of linestrings to a layer of the map. The layer and its options can be specified through a
     * [LinestringLayerOptions]{@link azure-maps-javascript.LinestringLayerOptions} object. Options for
     * the layer can only be specified upon the layer's initial creation. Map must be fully loaded before the
     * linestrings can be added. Use the addEventListener method with event type 'load'.
     * @param linestrings The linestrings to add.
     * @param options The layer options for the linestrings.
     */
    public addLinestrings(linestrings: Array<Feature<LineString | MultiLineString, LinestringProperties>>,
                          options?: LinestringLayerOptions) {
        const layerOptions: LinestringLayerOptions = new LinestringLayerOptions().merge(options);
        const layer: Layer = this.styleBuilder.get(layerOptions.name) || new LinestringLayer(layerOptions);

        if (layerOptions.overwrite) {
            layer.clear();
        }
        layer.push(linestrings);

        this.styleBuilder.upsert(layer, layerOptions.before);
        if (!layerOptions.defer) {
            const geoJsonSource = this.map.getSource(layer.sourceName) as mapboxgl.GeoJSONSource;
            if (geoJsonSource) {
                geoJsonSource.setData(layer.sources[layer.sourceName].data);
            } else {
                this.map.setStyle(this.styleBuilder.build());
            }
        }
    }

    /**
     * Add a collection of polygons to a layer of the map. The layer and its options can be specified through a
     * [PolygonLayerOptions]{@link azure-maps-javascript.PolygonLayerOptions} object. Options for
     * the layer can only be specified upon the layer's initial creation. Map must be fully loaded before the polygons
     * can be added. Use the addEventListener method with event type 'load'.
     * @param polygons The polygons to add.
     * @param options The layer options for the polygons.
     */
    public addPolygons(polygons: Array<Feature<Polygon | MultiPolygon, PolygonProperties>>,
                       options?: PolygonLayerOptions) {
        const layerOptions: PolygonLayerOptions = new PolygonLayerOptions().merge(options);
        const layer: Layer = this.styleBuilder.get(layerOptions.name) || new PolygonLayer(layerOptions);

        if (layerOptions.overwrite) {
            layer.clear();
        }
        layer.push(polygons);

        this.styleBuilder.upsert(layer, layerOptions.before);
        if (!layerOptions.defer) {
            const geoJsonSource = this.map.getSource(layer.sourceName) as mapboxgl.GeoJSONSource;
            if (geoJsonSource) {
                geoJsonSource.setData(layer.sources[layer.sourceName].data);
            } else {
                this.map.setStyle(this.styleBuilder.build());
            }
        }
    }

    /**
     * Adds a raster layer to the map. The layer and its options can be specified through a
     * [RasterLayerOptions]{@link azure-maps-javascript.RasterLayerOptions} object. Options for
     * the layer can only be specified upon the layer's initial creation.
     * @param tileSources  A list of endpoints specified as strings from which raster images can be requested. The
     * endpoints can be parameterized with the tags '{z}', '{x}' and '{y}' to specify the zoom, x-index, and y-index of
     * the needed tile respectively. The map control will request and place the tiles that are contained in the map's
     * viewport.
     * @param options The options for the raster layer.
     */
    public addRaster(tileSources: string[], options?: RasterLayerOptions) {
        const layerOptions: RasterLayerOptions = new RasterLayerOptions().merge(options);
        const layer: Layer = this.styleBuilder.get(layerOptions.name) || new RasterLayer(layerOptions);

        if (layerOptions.overwrite) {
            layer.clear();
        }
        layer.push(tileSources);

        this.styleBuilder.upsert(layer, layerOptions.before);
        if (!layerOptions.defer) {
            this.map.setStyle(this.styleBuilder.build());
        }
    }

    /**
     * Returns a list of the map's layers from bottom to top.
     */
    public getLayers(): string[] {
        return this.styleBuilder.getLayers().map((layer) => layer.name);
    }

    /**
     * Removes a collection of layers from the map.
     * @param layerNames An array of layer names to remove from the map.
     */
    public removeLayers(layerNames: string[]) {
        for (const layerName of layerNames) {
            this.styleBuilder.remove(layerName);
        }
        this.map.setStyle(this.styleBuilder.build());
    }

    /**
     * Adds a custom HTMLElement to the map at a specified position.
     * @param element The HTMLElement to add.
     * @param position The position to place the element.
     */
    public addHtml(element: HTMLElement, position: Position): string {
        const marker: mapboxgl.Marker = new mapboxgl.Marker(element)
            .setLngLat(position)
            .addTo(this.map);

        const htmlElementId: number = this.htmlElementId++;
        this.htmlElements[htmlElementId] = marker;

        return `${htmlElementId}`;
    }

    /**
     * Removes a custom HTMLElement from the map.
     * @param elementId The id of element to remove.
     */
    public removeHtml(elementId: string) {
        if (this.htmlElements.hasOwnProperty(elementId)) {
            this.htmlElements[elementId].remove();
        }
    }

    /**
     * Add an event listener to the map.
     * @param type The event listener types:
     * <p>'click':  A pointing device is pressed and released at the same point on the map.</p>
     * <p>'dblclick':  A pointing device is clicked twice on the map.</p>
     * <p>'mousemove': A pointing device is moved within the map. </p>
     * <p>'mouseup': A pointing device is released within the map. </p>
     * <p>'mousedown': A pointing device is pressed within the map. </p>
     * <p>'mouseout': A pointing device leaves the map's canvas.</p>
     * <p>'touchstart': A touch point is placed on the touch surface. </p>
     * <p>'touchend': A touch point is removed from the touch surface. </p>
     * <p>'touchmove': A touch point is moved along the touch surface. </p>
     * <p>'drag': An element is dragged. </p>
     * <p>'dragstart': The user starts dragging an element.</p>
     * <p>'dragend': A drag operation is ended.</p>
     * <p>'zoomstart': Just before the map transitions from one zoom level to another.</p>
     * <p>'zoom': During the map transitions from one zoom level to another.</p>
     * <p>'zoomend': After the map completes a transition from one zoom level to another.</p>
     * <p>'rotatestart': When a "drag to rotate" starts.</p>
     * <p>'rotate': During a "drag to rotate" interaction.</p>
     * <p>'rotateend': After When a "drag to rotate" ends.</p>
     * <p>'pitchstart': When the map's pitch (tilt) starts to change.</p>
     * <p>'pitch': Whenever the map's pitch (tilt) changes.</p>
     * <p>'pitchend': After the map's pitch (tilt) completes the change.</p>
     * <p>'resize': The document view has been resized.<p>
     * <p>'load': Progression has been successful.</p>
     * @param callback The callback to fire when the event occurs.
     */
    public addEventListener(type: string, callback: any): void;

    /**
     * Add an event listener to a layer of the map. Event listeners cannot be added to the default "base", "transit" and
     * "labels" layers of the map.
     * @param type The event listener types:
     * <p>'click':  A pointing device is pressed and released at the same point on the map.</p>
     * <p>'dblclick':  A pointing device is clicked twice on the map.</p>
     * <p>'mousemove': A pointing device is moved within the map. </p>
     * <p>'mouseup': A pointing device is released within the map. </p>
     * <p>'mousedown': A pointing device is pressed within the map. </p>
     * <p>'mouseenter': A pointing device is moved onto a specific layer that has the listener attached. </p>
     * <p>'mouseleave': A pointing device is moved off a specific layer that has the listener attached.</p>
     * <p>'mouseover': A pointing device is moved within the map.</p>
     * <p>'touchstart': A touch point is placed on the touch surface. </p>
     * <p>'touchend': A touch point is removed from the touch surface. </p>
     * @param layer The layer of the map.
     * @param callback The callback to fire when the event occurs.
     */
    public addEventListener(type: string, layer: string, callback: any): void;

    public addEventListener(type: string, layerOrCallback: string | any, callback?: any) {
        if (typeof layerOrCallback === "function") {
            this.mapCallbackHandler.addCallback(type, "", layerOrCallback);
            const modifiedCallback = this.mapCallbackHandler.getModifiedCallback(type, "", layerOrCallback);
            this.map.off(type, modifiedCallback);
            this.map.on(type, modifiedCallback);
        } else {
            this.mapCallbackHandler.addCallback(type, layerOrCallback, callback);
            const modifiedCallback = this.mapCallbackHandler.getModifiedCallback(type, layerOrCallback, callback);
            this.map.off(type, layerOrCallback, modifiedCallback);
            this.map.on(type, layerOrCallback, modifiedCallback);
        }
    }

    /**
     * Remove an event listener from the map.
     * @param type The event type.
     * @param callback The callback of the event listener.
     */
    public removeEventListener(type: string, callback: any): void;

    /**
     * Remove an event listener from a layer of the map.
     * @param type The event type.
     * @param layer The layer of the map.
     * @param callback The callback of the event listener.
     */
    public removeEventListener(type: string, layer: string, callback: any): void;

    public removeEventListener(type: string, layerOrCallback: string | any, callback?: any) {
        if (typeof layerOrCallback === "function") {
            const modifiedCallback = this.mapCallbackHandler.getModifiedCallback(type, "", layerOrCallback);
            this.map.off(type, modifiedCallback);
        } else {
            const modifiedCallback = this.mapCallbackHandler.getModifiedCallback(type, layerOrCallback, callback);
            this.map.off(type, layerOrCallback, modifiedCallback);
        }
    }

    /**
     * Set the map control's user interaction handlers. Any options not specified will default to their current values.
     * @param options The options for enabling/disabling the user interaction handlers.
     */
    public setUserInteraction(options?: UserInteractionOptions) {
        this.userInteractionOptions = new UserInteractionOptions().merge(this.userInteractionOptions, options);

        if (!this.userInteractionOptions.interactive) {
            this.map.boxZoom.disable();
            this.map.doubleClickZoom.disable();
            this.map.dragPan.disable();
            this.map.dragRotate.disable();
            this.map.keyboard.disable();
            this.map.scrollZoom.disable();
            this.map.touchZoomRotate.disable();
        } else {
            (this.userInteractionOptions.boxZoomInteraction) ? this.map.boxZoom.enable() : this.map.boxZoom.disable();
            (this.userInteractionOptions.dblClickZoomInteraction) ? this.map.doubleClickZoom.enable() : this.map.doubleClickZoom.disable();
            (this.userInteractionOptions.dragPanInteraction) ? this.map.dragPan.enable() : this.map.dragPan.disable();
            (this.userInteractionOptions.dragRotateInteraction) ? this.map.dragRotate.enable() : this.map.dragRotate.disable();
            (this.userInteractionOptions.keyboardInteraction) ? this.map.keyboard.enable() : this.map.keyboard.disable();
            (this.userInteractionOptions.scrollZoomInteraction) ? this.map.scrollZoom.enable() : this.map.scrollZoom.disable();
            (this.userInteractionOptions.touchInteraction) ? this.map.touchZoomRotate.enable() : this.map.touchZoomRotate.disable();
        }
    }

    /**
     * Return the map control's current user interaction handler settings.
     */
    public getUserInteraction(): UserInteractionOptions {
        return this.userInteractionOptions;
    }

    /**
     * Add a control to the map.
     * @param control The control to add.
     * @param options The options for the added control.
     */
    public addControl(control: Control, options?: ControlOptions) {
        const controlOptions = new ControlOptions().merge(options);

        const controlHtml = control.onAdd(this, options);

        const subcontrolContainers = this.controlContainer.getElementsByClassName(controlOptions.position);
        if (subcontrolContainers.length > 0) {
            subcontrolContainers.item(0).appendChild(controlHtml);
        } else {
            throw new Error(`Control position ${controlOptions.position} does not exist.`);
        }
    }

    /**
     * Remove a control from the map.
     * @param control The control to remove.
     */
    public removeControl(control: Control) {
        control.onRemove();
    }

    /**
     * Set the traffic options for the map. Any options not specified will default to their current values.
     * @param options The options for defining the map's traffic display.
     */
    public setTraffic(options?: TrafficOptions) {
        const previousIncidentsOption = this.trafficOptions.incidents;
        const previousFlowOption = this.trafficOptions.flow;
        this.trafficOptions = new TrafficOptions().merge(this.trafficOptions, options);

        if (this.trafficOptions.incidents) {
            if (!previousIncidentsOption) {
                this.incidentDelegate.addToMap();
            }
        } else {
            this.incidentDelegate.removeFromMap();
        }

        if (this.trafficOptions.flow && this.trafficOptions.flow !== "none") {
            if (this.trafficOptions.flow !== previousFlowOption) {
                this.flowDelegate.addToMap();
            }
        } else {
            this.flowDelegate.removeFromMap();
        }
    }

    /**
     * Return the map control's current traffic settings.
     */
    public getTraffic(): TrafficOptions {
        return this.trafficOptions;
    }

    /**
     * Clean up the map's resources. Map will not function correctly after calling this method.
     */
    public remove() {
        this.map.remove();
        while (this.getMapContainer().firstChild) {
            const currChild = this.getMapContainer().firstChild;
            this.getMapContainer().removeChild(currChild);
        }
    }

    /**
     * Resize the map according to the dimensions of its container element.
     */
    public resize() {
        this.map.resize();
    }

}
