import * as _ from "lodash";
import { Position } from "../data/Position";
import { Map } from "./Map";
import { PopupOptions } from "./options/PopupOptions";

/**
 * An information window anchored at a specified position on a map.
 * @module atlas
 */
export class Popup {
    private contentContainerHtml: HTMLElement;
    private closeButton: HTMLElement;
    private arrowHtml: HTMLElement;
    private containerHtml: HTMLElement;
    private id: string;
    private popupOptions: PopupOptions;
    private map: Map;

    /**
     * Constructs a Popup object and initializes it with the specified options.
     * @param options The options for the popup.
     */
    constructor(options?: PopupOptions) {
        this.popupOptions = new PopupOptions().merge(options);

        // Create popup HTMLElement parts
        this.contentContainerHtml = document.createElement("div");
        this.contentContainerHtml.classList.add("popup-content-container");
        this.contentContainerHtml.setAttribute("aria-role", "note");
        this.contentContainerHtml.setAttribute("aria-label", "Map Information Window");
        this.contentContainerHtml.setAttribute("tabindex", "0");

        this.arrowHtml = document.createElement("div");
        this.arrowHtml.classList.add("popup-arrow");

        this.closeButton = document.createElement("div");
        this.closeButton.classList.add("popup-close");
        this.closeButton.setAttribute("aria-hidden", "true");
        this.closeButton.innerHTML = "x";

        this.containerHtml = document.createElement("div");
        this.containerHtml.classList.add("popup-container");
        this.containerHtml.classList.add("hidden-accessible-element");

        // Assemble popup
        this.containerHtml.appendChild(this.contentContainerHtml);
        this.containerHtml.appendChild(this.arrowHtml);
        this.containerHtml.appendChild(this.closeButton);
    }

    /**
     * Sets the options for the popup.
     * @param options The options for the popup.
     */
    public setPopupOptions(options?: PopupOptions) {
        this.popupOptions = new PopupOptions().merge(this.popupOptions, options);
    }

    /**
     * Returns the options for the popup.
     */
    public getPopupOptions(): PopupOptions {
        return this.popupOptions;
    }

    /**
     * Attaches the popup to the HTML document in a hidden state.
     * @param map The map.
     */
    public attach = (map: Map) => {
        // If attaching to a different map, close popup on current map
        if (this.map && this.map !== map) {
            this.close();
        }

        this.map = map;

        // If not already created, create container for holding all popups
        const popupContainerClassName = "popup-collection-container";
        let popupContainer = this.map.getMapContainer().querySelector(`.${popupContainerClassName}`);
        if (!popupContainer) {
            popupContainer = document.createElement("div");
            popupContainer.classList.add(popupContainerClassName);
            popupContainer.setAttribute("aria-role", "region");
            popupContainer.setAttribute("aria-label", "Map Information");

            this.map.getMapContainer().appendChild(popupContainer);
        }

        // Set popup content. Attach popup to the HTML document. Initial state of popup container is hidden.
        this.contentContainerHtml.innerHTML = "";
        this.contentContainerHtml.appendChild(this.popupOptions.content);
        this.id = this.map.addHtml(this.containerHtml, this.popupOptions.position);

        popupContainer.appendChild(this.containerHtml);

        this.closeButton.addEventListener("click", this.close);
    }

    /**
     * Opens the popup on the passed in map.
     * @param map The map.
     */
    public open = (map: Map) => {
        this.attach(map);

        this.containerHtml.classList.remove("hidden-accessible-element");

        this.containerHtml.style.height = `${this.contentContainerHtml.clientHeight +
            this.arrowHtml.style.borderTopWidth}px`;
        this.containerHtml.style.top = `${this.popupOptions.pixelOffset[1] -
            (this.contentContainerHtml.clientHeight / 2 + 10)}px`;
        this.containerHtml.style.left = `${this.popupOptions.pixelOffset[0]}px`;
    }

    /**
     * Closes the popup on the map. The popup remains attached to the HTML document.
     */
    public close = () => {
        this.containerHtml.classList.add("hidden-accessible-element");
    }

    /**
     * Closes the popup on the map and removes it from the HTML document.
     */
    public remove = () => {
        if (this.map) {
            this.map.removeHtml(this.id);
        }
    }

}
