import { NearbyGeographySearchService } from "../../../services/NearbyGeographySearchService";
import { Map } from "../../Map";
import { MapDelegate } from "../MapDelegate";

/**
 * @private
 */
export class AccessibleMapDelegate implements MapDelegate {
    private static readonly COORDINATE_FRACTION_PRECISION: number = 5;
    private static readonly ZOOM_FRACTION_PRECISION: number = 2;
    private static readonly UNKNOWN_LOCATION_STRING: string = "unknown";

    private map: Map;
    private atlasMapLiveStateInfo: HTMLElement;
    private nearbyGeographySearch: NearbyGeographySearchService;

    constructor(map: Map) {
        this.map = map;
    }

    public addToMap = () => {
        this.nearbyGeographySearch = new NearbyGeographySearchService(this.map.getServiceOptions());

        this.createMapKeyBindingInfo();
        this.initializeMapLiveStateInfo();

        this.map.getMapContainer().setAttribute("role", "application");
        this.map.getMapContainer().setAttribute("aria-label", "Map Application");

        this.map.getCanvasContainer().setAttribute("role", "widget");
        this.map.getCanvasContainer().setAttribute("aria-label", "Interactive Map");
        this.map.getCanvasContainer().setAttribute("aria-details", "atlas-map-details");
        this.map.getCanvasContainer().setAttribute("tabindex", "0");
        this.map.getCanvasContainer().setAttribute("aria-describedby", "atlas-map-state");

        this.map.getCanvas().setAttribute("tabindex", "-1");
        this.map.getCanvas().setAttribute("aria-hidden", "true");
    }

    public removeFromMap = () => {
        // Never called
    }

    private createMapKeyBindingInfo = () => {
        const atlasMapKeyBindings = document.createElement("div");
        atlasMapKeyBindings.setAttribute("tabindex", "0");
        atlasMapKeyBindings.setAttribute("aria-label", "Interactive Map Key Bindings");
        atlasMapKeyBindings.id = "atlas-map-details";
        atlasMapKeyBindings.classList.add("hidden-accessible-element");
        this.map.getCanvasContainer().appendChild(atlasMapKeyBindings);

        const zoomOut = document.createElement("p");
        zoomOut.innerText = "Zoom out: hyphen";
        atlasMapKeyBindings.appendChild(zoomOut);

        const zoomIn = document.createElement("p");
        zoomIn.innerText = "Zoom in: plus";
        atlasMapKeyBindings.appendChild(zoomIn);

        const panRight = document.createElement("p");
        panRight.innerText = "Pan right 100 pixels: right arrow";
        atlasMapKeyBindings.appendChild(panRight);

        const panLeft = document.createElement("p");
        panLeft.innerText = "Pan left 100 pixels: left arrow";
        atlasMapKeyBindings.appendChild(panLeft);

        const panUp = document.createElement("p");
        panUp.innerText = "Pan up 100 pixels: up arrow";
        atlasMapKeyBindings.appendChild(panUp);

        const panDown = document.createElement("p");
        panDown.innerText = "Pan down 100 pixels: down arrow";
        atlasMapKeyBindings.appendChild(panDown);

        const rotateClockwise = document.createElement("p");
        rotateClockwise.innerText = "Rotate 15 degrees clockwise: shift + right arrow";
        atlasMapKeyBindings.appendChild(rotateClockwise);

        const rotateCounterClockwise = document.createElement("p");
        rotateCounterClockwise.innerText = "Rotate 15 degrees counter clockwise: shift + left arrow";
        atlasMapKeyBindings.appendChild(rotateCounterClockwise);

        const increasePitch = document.createElement("p");
        increasePitch.innerText = "Increase pitch 10 degrees: shift + up arrow";
        atlasMapKeyBindings.appendChild(increasePitch);

        const decreasePitch = document.createElement("p");
        decreasePitch.innerText = "Decrease pitch 10 degrees: shift + down arrow";
        atlasMapKeyBindings.appendChild(decreasePitch);
    }

    private initializeMapLiveStateInfo = () => {
        this.atlasMapLiveStateInfo = document.createElement("div");
        this.atlasMapLiveStateInfo.setAttribute("tabindex", "0");
        this.atlasMapLiveStateInfo.setAttribute("aria-label", "Interactive Map State");
        this.atlasMapLiveStateInfo.setAttribute("aria-live", "polite");
        this.atlasMapLiveStateInfo.setAttribute("aria-atomic", "true");
        this.atlasMapLiveStateInfo.id = "atlas-map-state";
        this.atlasMapLiveStateInfo.classList.add("hidden-accessible-element");
        this.map.getCanvasContainer().appendChild(this.atlasMapLiveStateInfo);

        this.updateMapState();
        this.map.addEventListener("moveend", this.updateMapState);
    }

    private updateMapState = () => {
        const cameraState = this.map.getCamera();

        this.nearbyGeographySearch.request(cameraState.center).then((response) => {
            this.atlasMapLiveStateInfo.innerHTML = "";

            const zoom = document.createElement("p");
            zoom.innerText = `Zoom: ${cameraState.zoom.toFixed(AccessibleMapDelegate.ZOOM_FRACTION_PRECISION)}`;
            this.atlasMapLiveStateInfo.appendChild(zoom);

            const location = document.createElement("p");
            const locationAddress = response["results"].length > 0 ?
                response["results"][0]["address"]["freeformAddress"]
                : AccessibleMapDelegate.UNKNOWN_LOCATION_STRING;
            location.innerText = `Location: ${locationAddress}`;
            this.atlasMapLiveStateInfo.appendChild(location);

            const lat = document.createElement("p");
            lat.innerText = `Latitude: ${cameraState.center[1].toFixed(AccessibleMapDelegate.COORDINATE_FRACTION_PRECISION)}`;
            this.atlasMapLiveStateInfo.appendChild(lat);

            const lon = document.createElement("p");
            lon.innerText = `Longitude: ${cameraState.center[0].toFixed(AccessibleMapDelegate.COORDINATE_FRACTION_PRECISION)}`;
            this.atlasMapLiveStateInfo.appendChild(lon);
        }).catch((e) => {
            this.atlasMapLiveStateInfo.innerHTML = "";

            const zoom = document.createElement("p");
            zoom.innerText = `Zoom: ${cameraState.zoom.toFixed(AccessibleMapDelegate.ZOOM_FRACTION_PRECISION)}`;
            this.atlasMapLiveStateInfo.appendChild(zoom);

            const location = document.createElement("p");
            location.innerText = `Location: ${AccessibleMapDelegate.UNKNOWN_LOCATION_STRING}`;
            this.atlasMapLiveStateInfo.appendChild(location);

            const lat = document.createElement("p");
            lat.innerText = `Latitude: ${cameraState.center[1].toFixed(AccessibleMapDelegate.COORDINATE_FRACTION_PRECISION)}`;
            this.atlasMapLiveStateInfo.appendChild(lat);

            const lon = document.createElement("p");
            lon.innerText = `Longitude: ${cameraState.center[0].toFixed(AccessibleMapDelegate.COORDINATE_FRACTION_PRECISION)}`;
            this.atlasMapLiveStateInfo.appendChild(lon);
        });
    }

}
