import { LengthConverter } from "../../../helpers/LengthConverter";
import { Popup } from "../../Popup";
import { Incident } from "./Incident";

/**
 * @private
 */
enum IncidentPopupIconName {
    DELAY = "delay",
    LENGTH = "length",
    FROM = "from",
    TO = "to"
}

/**
 * @private
 */
export class IncidentPopupFactory {

    public static build(incident: Incident, languageCode: string = "en-US"): Popup {
        const titleEl: HTMLElement = document.createElement("div");
        titleEl.classList.add("incident-popup-title");
        titleEl.classList.add("font-segoeui-b");
        titleEl.setAttribute("aria-label", "Incident Type");
        titleEl.setAttribute("tabindex", "0");
        titleEl.innerText = incident.properties.incidentType;

        const subtitleEl: HTMLElement = document.createElement("div");
        subtitleEl.classList.add("incident-popup-subtitle");
        subtitleEl.classList.add("font-segoeui-b");
        subtitleEl.setAttribute("aria-label", "Incident Description");
        subtitleEl.setAttribute("tabindex", "0");
        subtitleEl.innerText = incident.properties.description;

        const headerEl: HTMLElement = document.createElement("div");
        headerEl.classList.add("incident-popup-header");
        headerEl.style.backgroundColor = incident.properties.color;
        headerEl.appendChild(titleEl);
        headerEl.appendChild(subtitleEl);

        const bodyEl: HTMLElement = document.createElement("div");
        bodyEl.classList.add("incident-popup-body");

        if (incident.properties.delay) {
            const delay = `${Math.floor(incident.properties.delay / 60)} min`;
            const delayInfoEl: HTMLElement = this.createInfoSection(delay, IncidentPopupIconName.DELAY);
            bodyEl.appendChild(delayInfoEl);
        }

        if (incident.properties.length) {
            const i18nNumberFormat = new Intl.NumberFormat([languageCode, "en-US"], {
                maximumFractionDigits: 2
            });

            let i18nLength: string;
            if (languageCode === "en-US") {
                const lengthInMiles: number = LengthConverter.metersToMiles(incident.properties.length);
                i18nLength = `${i18nNumberFormat.format(lengthInMiles)} miles`;
            } else {
                const lengthInKm = LengthConverter.metersToKilometers(incident.properties.length);
                i18nLength = `${i18nNumberFormat.format(lengthInKm)} km`;
            }

            const lengthInfoEl: HTMLElement = this.createInfoSection(i18nLength, IncidentPopupIconName.LENGTH);
            bodyEl.appendChild(lengthInfoEl);
        }

        if (incident.properties.from) {
            const fromInfoEl: HTMLElement = this.createInfoSection(incident.properties.from, IncidentPopupIconName.FROM);
            bodyEl.appendChild(fromInfoEl);
        }

        if (incident.properties.to) {
            const toInfoEl: HTMLElement = this.createInfoSection(incident.properties.to, IncidentPopupIconName.TO);
            bodyEl.appendChild(toInfoEl);
        }

        const incidentPopupContentEl: HTMLElement = document.createElement("div");
        incidentPopupContentEl.classList.add("incident-popup-content");
        incidentPopupContentEl.appendChild(headerEl);
        incidentPopupContentEl.appendChild(bodyEl);

        const incidentPopup: Popup = new Popup({
            content: incidentPopupContentEl,
            position: incident.geometry.coordinates
        });

        return incidentPopup;
    }

    private static createInfoSection(message: string, iconName: IncidentPopupIconName): HTMLElement {
        const infoSectionEl = document.createElement("div");
        infoSectionEl.classList.add("incident-popup-info");

        const iconEl = document.createElement("div");
        iconEl.classList.add("icon");
        iconEl.classList.add(iconName);

        const messageEl = document.createElement("span");
        messageEl.classList.add("message");
        messageEl.classList.add("font-segoeui");
        messageEl.setAttribute("aria-label", `Incident ${iconName}`);
        messageEl.setAttribute("tabindex", "0");
        messageEl.innerText = message;

        infoSectionEl.appendChild(iconEl);
        infoSectionEl.appendChild(messageEl);

        return infoSectionEl;
    }

}
