import { BoundingBox, Feature, Point, Position } from "../../../data";
import { IncidentService } from "../../../services/IncidentService";
import { ServiceOptions } from "../../../services/options/ServiceOptions";
import { Map } from "../../Map";
import { Popup } from "../../Popup";
import { MapDelegate } from "../MapDelegate";
import { Incident } from "./Incident";
import { IncidentPopupFactory } from "./IncidentPopupFactory";

/**
 * @private
 */
export class IncidentServiceDelegate implements MapDelegate {
    private readonly incidentLayerName: string = "incidents";
    private readonly incidentService: IncidentService;
    private popups: { [popupId: string]: Popup };
    private openPopupId: string;
    private readonly map: Map;

    constructor(map: Map) {
        this.map = map;
        this.incidentService = new IncidentService(map.getServiceOptions());
    }

    public addToMap = () => {
        this.map.addPins([], {
            name: this.incidentLayerName,
            cluster: false
        });

        this.map.addEventListener("mouseenter", this.incidentLayerName, this.cursorToPointer);
        this.map.addEventListener("mouseleave", this.incidentLayerName, this.cursorToDefault);
        this.map.addEventListener("click", this.incidentLayerName, this.openIncidentPopup);

        this.map.addEventListener("moveend", this.updateIncidents);

        this.updateIncidents();
    }

    public removeFromMap = () => {
        this.map.removeLayers([this.incidentLayerName]);

        this.map.removeEventListener("click", this.incidentLayerName, this.openIncidentPopup);
        this.map.removeEventListener("mouseenter", this.incidentLayerName, this.cursorToPointer);
        this.map.removeEventListener("mouseleave", this.incidentLayerName, this.cursorToDefault);

        this.map.removeEventListener("moveend", this.updateIncidents);

        this.cleanUpPopups();
    }

    private updateIncidents = () => {
        const integerZoom: number = Math.ceil(this.map.getCamera().zoom);
        const boundingBox: BoundingBox = this.map.getCamera().bounds;
        const mapLanguage: string = this.map.getStyle().language;

        this.incidentService.request({
            bbox: boundingBox,
            zoom: integerZoom,
            params: {
                expandCluster: true,
                originalPosition: true,
                language: mapLanguage
            }
        }).then((incidentResponse) => {
            if (!this.map.getTraffic().incidents) {
                return;
            }

            const prevOpenPopupId = this.openPopupId;
            this.cleanUpPopups();

            const incidents: Incident[] = [];
            for (const poi of incidentResponse["tm"]["poi"] || []) {
                for (const cpoi of poi.cpoi || [poi]) {
                    const incident: Incident = new Incident(cpoi);
                    incidents.push(incident);

                    this.popups[incident.properties.id] = IncidentPopupFactory.build(incident, mapLanguage);

                    if (incident.properties.id === prevOpenPopupId) {
                        this.popups[incident.properties.id].open(this.map);
                        this.openPopupId = incident.properties.id;
                    } else {
                        this.popups[incident.properties.id].attach(this.map);
                    }
                }
            }

            this.map.addPins(incidents, {
                overwrite: incidents.length > 0,
                name: this.incidentLayerName
            });
        });
    }

    private cursorToPointer = (data: any) => {
        this.map.getCanvas().style.cursor = "pointer";
    }

    private cursorToDefault = (data: any) => {
        this.map.getCanvas().style.cursor = "";
    }

    private openIncidentPopup = (data: any) => {
        if (this.popups.hasOwnProperty(this.openPopupId)) {
            this.popups[this.openPopupId].close();
        }

        const incident: Incident = data.features[0];
        const nextPopupId = incident.properties.id;
        if (this.popups.hasOwnProperty(nextPopupId)) {
            this.popups[nextPopupId].open(this.map);
            this.openPopupId = nextPopupId;
        }
    }

    private cleanUpPopups = () => {
        for (const popupId in this.popups) {
            if (this.popups.hasOwnProperty(popupId)) {
                this.popups[popupId].remove();
            }
        }

        this.popups = {};
        this.openPopupId = null;
    }

}
