import { Feature, Point } from "../../../../data";
import { PinProperties } from "../../../data_properties/PinProperties";
import { LayerOptions } from "../../../options/LayerOptions";
import { PinLayerOptions } from "../../../options/PinLayerOptions";
import { Layer } from "../Layer";

/**
 * @private
 */
export class PinLayer implements Layer {
    public name: string;
    public sources: any;
    public layers: any;
    public sourceName: string;

    constructor(pinLayerOptions: PinLayerOptions & PinProperties) {
        this.name = pinLayerOptions.name;
        this.sourceName = `${this.name}-source`;

        this.sources = {};
        this.sources[this.sourceName] = {
            type: "geojson",
            data: {
                type: "FeatureCollection",
                features: []
            },
            cluster: pinLayerOptions.cluster
        };

        this.layers = [
            {
                id: this.name,
                type: "symbol",
                source: this.sourceName,
                layout: {
                    "icon-size": pinLayerOptions.iconSize,
                    "icon-image": { type: "identity", property: "icon", default: pinLayerOptions.icon },
                    "text-field": { type: "identity", property: "title", default: pinLayerOptions.title },
                    "text-font": [pinLayerOptions.textFont],
                    "text-size": pinLayerOptions.fontSize,
                    "text-anchor": "center",
                    "icon-ignore-placement": !pinLayerOptions.cluster
                },
                paint: {
                    "icon-opacity": pinLayerOptions.opacity,
                    "text-opacity": pinLayerOptions.opacity,
                    "text-translate": pinLayerOptions.textOffset,
                    "text-color": pinLayerOptions.fontColor
                },
                minzoom: pinLayerOptions.minZoom,
                maxzoom: pinLayerOptions.maxZoom,
                filter: ["!has", "point_count"]
            }
        ];

        if (pinLayerOptions.cluster) {
            this.layers.push({
                id: `${this.name}-cluster`,
                type: "symbol",
                source: this.sourceName,
                layout: {
                    "icon-size": pinLayerOptions.iconSize,
                    "icon-image": pinLayerOptions.clusterIcon || {
                        type: "identity",
                        property: "icon",
                        default: pinLayerOptions.icon
                    },
                    "icon-allow-overlap": true
                },
                paint: {
                    "icon-opacity": pinLayerOptions.opacity,
                    "text-opacity": pinLayerOptions.opacity
                },
                minzoom: pinLayerOptions.minZoom,
                maxzoom: pinLayerOptions.maxZoom,
                filter: ["has", "point_count"]
            });

            this.layers.push({
                id: `${this.name}-cluster-count`,
                type: "symbol",
                source: this.sourceName,
                layout: {
                    "icon-image": "cluster-red-circle",
                    "text-field": "{point_count_abbreviated}",
                    "text-font": ["StandardFontCondensed-Bold"],
                    "text-size": 10,
                    "text-justify": "center",
                    "text-anchor": "top",
                    "text-line-height": 1
                },
                paint: {
                    "icon-opacity": pinLayerOptions.opacity,
                    "text-opacity": pinLayerOptions.opacity,
                    "icon-translate": [12, -12],
                    "text-translate": [12, -13],
                    "text-color": "#fff"
                },
                minzoom: pinLayerOptions.minZoom,
                maxzoom: pinLayerOptions.maxZoom,
                filter: ["has", "point_count"]
            });
        }

    }

    public push(pins: Array<Feature<Point, PinProperties>>) {
        for (const pin of pins) {
            this.sources[this.sourceName].data.features.push(pin);
        }
    }

    public clear() {
        this.sources[this.sourceName].data.features = [];
    }

}
