﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Text;
    using System.Threading;
    using System.Threading.Tasks;

    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using Microsoft.WindowsAzure.Storage.RetryPolicies;

    public class AzureBlockBlobWriter : IRandomBlockWriter
    {
        private const string ContentTypeKey = "ContentType";
        private const string BlockIdPrefix = "BId:";
        private const int Mb = 1024 * 1024;
        private const int MaxNumberOfRetries = 3;
        private const int DeltaBackOffSecs = 60;
        private const int MaxExecutionTimeInSecs = 5 * 60 * 60;
        private const int MaxExecutionTimeInSecsForBatchCommit = 10 * 60 * 60;

        private readonly CloudStorageAccount storageAccount;
        private readonly string containerName;
        private readonly string fileName;
        private CloudBlockBlob cloudBlockBlob;
        private IDictionary<string, string> sourceMetadataMap;
        
        /// <summary>
        /// Initializes a new instance of the <see cref="AzureBlockBlobWriter"/> class.
        /// </summary>
        /// <param name="cloudStorageAccount">
        /// The cloud Storage Account.
        /// </param>
        /// <param name="blobContainerName">
        /// The blob container name.
        /// </param>
        /// <param name="blobFileName">
        /// The blob file name.
        /// </param>
        /// <exception cref="Exception">
        /// Throws Exception if couldn't connect to the storage account.
        /// </exception>
        public AzureBlockBlobWriter(CloudStorageAccount cloudStorageAccount, string blobContainerName, string blobFileName)
        {
            // To increase the through put to the Blob storage. Otherwise, .Net uses only 2 threads to send.
            ServicePointManager.DefaultConnectionLimit = 1000;
            this.storageAccount = cloudStorageAccount;
            this.containerName = blobContainerName;
            this.fileName = blobFileName;
        }

        /// <summary>
        /// Gets the md 5 hash of target.
        /// </summary>
        public string Md5HashOfTarget { get; private set; }

        /// <summary>
        /// Gets the max length of block allowed.
        /// </summary>
        public int MaxLengthOfBlockAllowed => 2 * Mb;

        /// <summary>
        /// Gets the preferred length of block.
        /// </summary>
        public int? PreferredLengthOfBlock => null;

        /// <summary>
        /// Gets the max number of blocks allowed.
        /// </summary>
        public long MaxNumberOfBlocksAllowed => 50000;

        /// <summary>
        /// Gets the max degree of parallelism allowed.
        /// </summary>
        public int MaxDegreeOfParallelismAllowed => 200;

        /// <summary>
        /// Initializes the required objects.
        /// </summary>
        /// <returns></returns>
        public async Task InitializeAsync()
        {
            var cloudBlobClient = this.storageAccount.CreateCloudBlobClient();
            cloudBlobClient.DefaultRequestOptions = new BlobRequestOptions()
                                                        {
                                                            UseTransactionalMD5 = true,
                                                            StoreBlobContentMD5 = true,
                                                            RetryPolicy = new ExponentialRetry(
                                                                                                            TimeSpan.FromSeconds(DeltaBackOffSecs),
                                                                                                            MaxNumberOfRetries),
                                                            MaximumExecutionTime = TimeSpan.FromSeconds(MaxExecutionTimeInSecs)
            };

            var cloudBlobContainer = cloudBlobClient.GetContainerReference(this.containerName);
            await cloudBlobContainer.CreateIfNotExistsAsync(BlobContainerPublicAccessType.Container, null, null);
            this.cloudBlockBlob = cloudBlobContainer.GetBlockBlobReference(this.fileName);
        }

        /// <summary>
        /// Returns the previously saved blockIds. This allows the controller to skip these.
        /// </summary>
        /// <returns></returns>
        public async Task<HashSet<long>> GetSavedBlockIds()
        {
            if (await this.cloudBlockBlob.ExistsAsync())
            {
                return new HashSet<long>();
            }

            var blockList = await GetUncommittedBlockIds(this.cloudBlockBlob);
            if (blockList == null)
            {
                return new HashSet<long>();
            }

            return new HashSet<long>(blockList.Select(this.GetLongFromBase64EncodedString));
        }

        public async Task ClearPriorUncommittedBlocks()
        {
            if (await this.cloudBlockBlob.ExistsAsync())
            {
                return;
            }

            var blockList = await GetUncommittedBlockIds(this.cloudBlockBlob); // .DownloadBlockListAsync(BlockListingFilter.Uncommitted, null, null, null);
            if (blockList == null)
            {
                return;
            }

            await this.cloudBlockBlob.PutBlockListAsync(blockList);
            await this.cloudBlockBlob.DeleteAsync();
            await this.InitializeAsync();
        }

        /// <summary>
        /// The set metadata map.
        /// </summary>
        /// <param name="metadataMap">
        /// The metadata Map.
        /// </param>
        public void SetMetadataMap(IDictionary<string, string> metadataMap)
        {
            this.sourceMetadataMap = metadataMap;
            if (this.sourceMetadataMap.TryGetValue(ContentTypeKey, out var value))
            {
                this.cloudBlockBlob.Properties.ContentType = value;
                this.sourceMetadataMap.Remove(ContentTypeKey);
            }

            if (this.sourceMetadataMap.TryGetValue("ContentMD5", out value))
            {
                this.cloudBlockBlob.Properties.ContentMD5 = value;
                this.sourceMetadataMap.Remove("ContentMD5");
            }

            foreach (var kvp in this.sourceMetadataMap)
            {
                if (this.cloudBlockBlob.Metadata.TryGetValue(kvp.Key, out var v))
                {
                    if (v != kvp.Value)
                    {
                        throw new Exception($"Container({this.containerName}) file Name ({this.fileName}) has conflicting metadata. Key : {kvp.Key}, Existing value : {this.cloudBlockBlob.Metadata[kvp.Key]}, Incoming value : {kvp.Value}");
                    }
                }
                else
                {
                    this.cloudBlockBlob.Metadata.Add(kvp.Key, kvp.Value);
                }
            }

            this.cloudBlockBlob.SetProperties();
            this.cloudBlockBlob.SetMetadata();
        }

        /// <summary>
        /// The write block async.
        /// </summary>
        /// <param name="blockId">
        /// The block id.
        /// </param>
        /// <param name="blockContent">
        /// The block content.
        /// </param>
        /// <returns>
        /// The <see cref="Task"/>.
        /// </returns>
        public async Task<bool> WriteBlockAsync(long blockId, byte[] blockContent)
        {
            using (var memoryStream = new MemoryStream(blockContent))
            {
                await this.cloudBlockBlob.PutBlockAsync(GetBase64EncodedString(blockId), memoryStream, null);
            }

            return true;
        }

        /// <summary>
        /// The commit blocks async.
        /// </summary>
        /// <param name="orderedListOfBlockIds">
        /// The ordered list of block ids.
        /// </param>
        /// <param name="cancellationToken"></param>
        /// <returns>
        /// The <see cref="Task"/>.
        /// </returns>
        public async Task CommitBlocksAsync(List<long> orderedListOfBlockIds, CancellationToken cancellationToken)
        {
            var requestOptions = new BlobRequestOptions()
                                     {
                                         UseTransactionalMD5 = true,
                                         StoreBlobContentMD5 = true,
                                         RetryPolicy = new ExponentialRetry(
                                             TimeSpan.FromSeconds(DeltaBackOffSecs * 10),
                                             MaxNumberOfRetries),
                                         MaximumExecutionTime = TimeSpan.FromSeconds(MaxExecutionTimeInSecsForBatchCommit)
                                     };

            var blockIds = orderedListOfBlockIds.Select(GetBase64EncodedString).ToList();
            await this.cloudBlockBlob.PutBlockListAsync(blockIds, null, requestOptions, null, cancellationToken);
            this.GetMd5Hash();
        }

        private static async Task<IEnumerable<string>> GetUncommittedBlockIds(CloudBlockBlob blockBlob)
        {
            try
            {
                var blockList = await blockBlob.DownloadBlockListAsync(BlockListingFilter.Uncommitted, null, null, null);
                return blockList.Select(t => t.Name);
            }
            catch (Exception)
            {
                return null;
            }
        }

        /// <summary>
        /// The get base 64 encoded string.
        /// </summary>
        /// <param name="id">
        /// The id.
        /// </param>
        /// <returns>
        /// The <see cref="string"/>.
        /// </returns>
        private string GetBase64EncodedString(long id)
        {
            return Convert.ToBase64String(ASCIIEncoding.ASCII.GetBytes($"{BlockIdPrefix}{id.ToString("d12")}"));
        }

        private long GetLongFromBase64EncodedString(string idString)
        {
            return Convert.ToInt64(ASCIIEncoding.ASCII.GetString(Convert.FromBase64String(idString)).Replace(BlockIdPrefix, ""));
        }

        private void GetMd5Hash()
        {
            //this.cloudBlockBlob.FetchAttributes();
            this.Md5HashOfTarget = this.cloudBlockBlob.Properties.ContentMD5;
        }
    }
}
