﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common
{
    using System;
    using System.Collections.Concurrent;
    using System.Collections.Generic;
    using System.Threading;
    using System.Threading.Tasks;

    using SharpCompress.Crypto;

    /// <summary>
    /// Moves the blocks of stream from the source to the destination.
    /// </summary>
    public class BlockStreamManager : IDisposable
    {
        /// <summary>
        /// The kb.
        /// </summary>
        private const int Kb = 1024;

        /// <summary>
        /// The mb.
        /// </summary>
        private const int Mb = 1024 * Kb;

        /// <summary>
        /// The gb.
        /// </summary>
        private const int Gb = 1024 * Mb;

        /// <summary>
        /// The reader.
        /// </summary>
        private readonly IRandomBlockReader reader;

        /// <summary>
        /// The writer.
        /// </summary>
        private readonly IRandomBlockWriter writer;

        /// <summary>
        /// The logger.
        /// </summary>
        private readonly ILogger logger;

        /// <summary>
        /// The total length of source content.
        /// </summary>
        private long totalLengthOfSourceContent;

        /// <summary>
        /// The max degree of parallelism allowed on reader.
        /// </summary>
        private int maxDegreeOfParallelismAllowedOnReader;

        /// <summary>
        /// The max block length allowed on reader.
        /// </summary>
        private long maxBlockLengthAllowedOnReader;

        /// <summary>
        /// The max degree of parallelism allowed on writer.
        /// </summary>
        private int maxDegreeOfParallelismAllowedOnWriter;

        /// <summary>
        /// The max block length allowed on writer.
        /// </summary>
        private long maxBlockLengthAllowedOnWriter;

        /// <summary>
        /// The max number of blocks allowed on writer.
        /// </summary>
        private long maxNumberOfBlocksAllowedOnWriter;

        /// <summary>
        /// The max block length supported.
        /// </summary>
        private long maxBlockLengthSupported;

        /// <summary>
        /// The current block length.
        /// </summary>
        private int defaultBlockLength = 100 * Kb;

        /// <summary>
        /// The max blocks to be cached.
        /// </summary>
        private int maxBlocksToBeCachedForWriter;

        /// <summary>
        /// The block info list for reader.
        /// </summary>
        private ConcurrentQueue<BlockInfo> blockInfosForReader;

        /// <summary>
        /// The block info list for writer.
        /// </summary>
        private BlockingCollection<BlockInfo> blockInfosForWriter;

        /// <summary>
        /// The ordered block ids.
        /// </summary>
        private List<long> orderedBlockIds;

        private IDictionary<string, string> sourceMetadataMap;

        private int? blockLengthOverrideFromClient;

        private string jobName;

        /// <summary>
        /// Initializes a new instance of the <see cref="BlockStreamManager"/> class.
        /// </summary>
        /// <param name="blockReader">
        /// The block reader.
        /// </param>
        /// <param name="blockWriter">
        /// The block writer.
        /// </param>
        /// <param name="loggerInstance">logger instance</param>
        public BlockStreamManager(IRandomBlockReader blockReader, IRandomBlockWriter blockWriter, ILogger loggerInstance, string jobIdentifier, int? blockLengthOverride = null)
        {
            this.reader = blockReader;
            this.writer = blockWriter;
            this.logger = loggerInstance;
            this.jobName = jobIdentifier;
            this.blockLengthOverrideFromClient = blockLengthOverride;
        }

        /// <summary>
        /// The current block length.
        /// </summary>
        public int CurrentBlockLength { get; private set; }

        /// <summary>
        /// Initializes the internal variables based on the reader and writer thresholds.
        /// </summary>
        public async Task InitializeAsync()
        {
            this.maxBlockLengthAllowedOnReader = this.reader.MaxLengthOfBlockAllowed;
            this.maxDegreeOfParallelismAllowedOnReader = this.reader.MaxDegreeOfParallelismAllowed;
            this.totalLengthOfSourceContent = this.reader.LengthOfTheStream;
            this.sourceMetadataMap = this.reader.MetadataMap;

            this.maxBlockLengthAllowedOnWriter = this.writer.MaxLengthOfBlockAllowed;
            this.maxDegreeOfParallelismAllowedOnWriter = this.writer.MaxDegreeOfParallelismAllowed;
            this.maxNumberOfBlocksAllowedOnWriter = this.writer.MaxNumberOfBlocksAllowed;
            this.maxBlocksToBeCachedForWriter = this.maxDegreeOfParallelismAllowedOnWriter * 2;

            if (!this.blockLengthOverrideFromClient.HasValue)
            {
                this.maxBlockLengthSupported = this.maxBlockLengthAllowedOnReader > this.maxBlockLengthAllowedOnWriter
                                        ? this.maxBlockLengthAllowedOnWriter
                                        : this.maxBlockLengthAllowedOnReader;

                var tempBlockLength = this.reader.PreferredLengthOfBlock;

                // Pick the max preferred block length out of reader and writer.
                tempBlockLength = this.writer.PreferredLengthOfBlock.HasValue ?
                                      tempBlockLength.HasValue ? tempBlockLength.Value > this.writer.PreferredLengthOfBlock.Value ?
                                                tempBlockLength.Value : this.writer.PreferredLengthOfBlock : this.writer.PreferredLengthOfBlock : tempBlockLength;

                // Pick the max preferred block length out of predefined Vs supplied.
                this.CurrentBlockLength = tempBlockLength.HasValue ? tempBlockLength.Value > this.defaultBlockLength ?
                                                                        tempBlockLength.Value : this.defaultBlockLength : this.defaultBlockLength;

                this.CalculateBlockSizeWithinConstraints();
            }
            else
            {
                this.CurrentBlockLength = this.blockLengthOverrideFromClient.Value;
            }

            await this.ComputeBlockLists();

            this.logger.Log(LogType.Info, $"{this.jobName} : Reader Threads: {this.maxDegreeOfParallelismAllowedOnReader}");
            this.logger.Log(LogType.Info, $"{this.jobName} : Writer threads : {this.maxDegreeOfParallelismAllowedOnWriter}");
            this.logger.Log(LogType.Info, $"{this.jobName} : Source Content Length: {this.totalLengthOfSourceContent}");
            this.logger.Log(LogType.Info, $"{this.jobName} : Block Length: {this.CurrentBlockLength}");
            this.logger.Log(LogType.Info, $"{this.jobName} : Number of Blocks to be cached for writer : {this.maxBlocksToBeCachedForWriter}");
            this.logger.Log(LogType.Info, $"{this.jobName} : Total number of Blocks : {this.orderedBlockIds.Count}");
        }

        /// <summary>
        /// Moves the data from the source to the destination in chunks in parallel.
        /// </summary>
        /// <returns>
        /// The <see cref="Task"/>.
        /// </returns>
        public async Task<bool> CopyAsync(CancellationToken cancellationToken)
        {
            this.logger.Log(LogType.Info, $"{this.jobName} : Starting CopyAsync");
            var readerTask = Task.Factory.StartNew(this.ReaderTask, (object)cancellationToken);
            var writerTask = Task.Factory.StartNew(this.WriterTask, (object)cancellationToken);

            await writerTask;
            await readerTask;

            this.logger.Log(LogType.Verbose, $"{this.jobName} : Committing the blocks..");
            try
            {
                await this.writer.CommitBlocksAsync(this.orderedBlockIds, cancellationToken);
                this.writer.SetMetadataMap(this.sourceMetadataMap);
                this.logger.Log(LogType.Info, $"{this.jobName} : Done with copying.");
            }
            catch (OperationCanceledException)
            {
                this.logger.Log(LogType.Info, $"{this.jobName} : CopyAsync received Cancellation.");
                return false;
            }
            catch (Exception ex)
            {
                this.logger.Log(LogType.Info, $"{this.jobName} : CopyAsync ran into an exception. Message : {ex.Message} . ToString : {ex}");
                throw;
            }

            return true; // this.reader.Md5HashOfSource == this.writer.Md5HashOfTarget; Not using the hash comparison.
        }

        public void Dispose()
        {
            this.blockInfosForWriter.Dispose();
        }

        /// <summary>
        /// The reader task.
        /// </summary>
        private void ReaderTask(object cancellationTokenObj)
        {
            var cancellationToken = (CancellationToken)cancellationTokenObj;
            var tasksList = new Task[this.maxDegreeOfParallelismAllowedOnReader];
            BlockInfo blockInfo = null;
            var index = 0;
            var blocksProcessed = 0;
            for (; index < this.maxDegreeOfParallelismAllowedOnReader; index++)
            {
                if (this.blockInfosForReader.TryDequeue(out blockInfo))
                {
                    tasksList[index] = this.ReadBlockAsync(blockInfo);
                    blocksProcessed++;
                }
            }

            try
            {
                while (!this.blockInfosForReader.IsEmpty)
                {
                    var completedTaskIndex = Task.WaitAny(tasksList, cancellationToken);

                    // Conversion to eliminate Resharper warning : https://www.jetbrains.com/help/resharper/2017.3/CoVariantArrayConversion.html
                    this.blockInfosForWriter.Add(((Task<BlockInfo>)tasksList[completedTaskIndex]).Result, cancellationToken);

                    if (this.blockInfosForReader.TryDequeue(out blockInfo))
                    {
                        tasksList[completedTaskIndex] = this.ReadBlockAsync(blockInfo);
                        blocksProcessed++;
                    }

                    if (blocksProcessed % 100 == 0)
                    {
                        this.logger.Log(LogType.Verbose, $"{this.jobName} : Reader Thread: Blocks processed so far : {blocksProcessed}");
                    }
                }

                this.logger.Log(LogType.Verbose, $"{this.jobName} : Reader Thread: Waiting on remaining threads to be done. Blocks processed so far : {blocksProcessed}");

                // Now wait on them in sequence.
                for (index = 0; index < this.maxDegreeOfParallelismAllowedOnReader; index++)
                {
                    if (tasksList[index] != null)
                    {
                        tasksList[index].Wait(cancellationToken);

                        // Conversion to eliminate Resharper warning : https://www.jetbrains.com/help/resharper/2017.3/CoVariantArrayConversion.html
                        this.blockInfosForWriter.Add(((Task<BlockInfo>)tasksList[index]).Result, cancellationToken);
                    }
                }

                this.logger.Log(LogType.Info, $"{this.jobName} : Reader Thread: Done with all the items in the incoming Q");
            }
            catch (OperationCanceledException)
            {
                this.logger.Log(LogType.Info, $"{this.jobName} : Reader Thread: Operation was cancelled by the caller.");
            }
            catch (Exception ex)
            {
                this.logger.Log(LogType.Error, $"{this.jobName} : Reader Thread exception Message : {ex.Message} ; StackTrace : {ex.StackTrace}");
                throw;
            }
            finally
            {
                this.blockInfosForWriter.CompleteAdding();
            }            
        }

        /// <summary>
        /// The writer task.
        /// </summary>
        private void WriterTask(object cancellationTokenObj)
        {
            var cancellationToken = (CancellationToken)cancellationTokenObj;
            var blocksProcessed = 0;
            var tasksList = new Task[this.maxDegreeOfParallelismAllowedOnWriter];
            BlockInfo blockInfo = null;
            var index = 0;
            for (; index < this.maxDegreeOfParallelismAllowedOnWriter; index++)
            {
                if (this.blockInfosForWriter.TryTake(out blockInfo, -1))
                {
                    tasksList[index] = this.WriteBlockAsync(blockInfo);
                }

                blocksProcessed++;
            }

            try
            {
                while (!this.blockInfosForWriter.IsCompleted)
                {
                    var completedTaskIndex = Task.WaitAny(tasksList, cancellationToken);

                    var completed = ((Task<bool>)tasksList[completedTaskIndex]).Result;

                    if (this.blockInfosForWriter.TryTake(out blockInfo, -1))
                    {
                        tasksList[completedTaskIndex] = this.WriteBlockAsync(blockInfo);
                    }

                    blocksProcessed++;

                    if (blocksProcessed % 100 == 0)
                    {
                        this.logger.Log(LogType.Verbose, $"{this.jobName} : Writer Thread: Blocks processed by writer : {blocksProcessed}, Items waiting in the writer Q : {this.blockInfosForWriter.Count}");
                    }
                }

                this.logger.Log(LogType.Info, $"{this.jobName} : Writer Thread: Done with all the blocks. Waiting on the threads to be done. Blocks processed by writer : {blocksProcessed}, Items waiting in the writer Q : {this.blockInfosForWriter.Count}");

                // Now wait on them in sequence.
                for (index = 0; index < this.maxDegreeOfParallelismAllowedOnReader; index++)
                {
                    if (tasksList[index] != null)
                    {
                        tasksList[index].Wait(cancellationToken);
                        var completed = ((Task<bool>)tasksList[index]).Result;
                    }
                }

                this.logger.Log(LogType.Info, $"{this.jobName} : Writer Thread: All threads are done.");
            }
            catch (OperationCanceledException)
            {
                this.logger.Log(LogType.Info, $"{this.jobName} : Writer Thread: Operation was cancelled by the caller.");
            }
            catch (Exception ex)
            {
                this.logger.Log(LogType.Error, $"{this.jobName} : Writer Thread exception Message : {ex.Message} ; StackTrace : {ex.StackTrace}");
                throw;
            }
        }

        /// <summary>
        /// The write block async.
        /// </summary>
        /// <param name="blockInfo">
        /// The block info.
        /// </param>
        /// <returns>
        /// The <see cref="Task"/>.
        /// </returns>
        private async Task<bool> WriteBlockAsync(BlockInfo blockInfo)
        {
            return await this.writer.WriteBlockAsync(blockInfo.BlockId, blockInfo.BlockBuffer);
        }

        /// <summary>
        /// The read block async.
        /// </summary>
        /// <param name="blockInfo">
        /// The block info.
        /// </param>
        /// <returns>
        /// The <see cref="Task"/>.
        /// </returns>
        private async Task<BlockInfo> ReadBlockAsync(BlockInfo blockInfo)
        {
            blockInfo.BlockBuffer = await this.reader.GetBufferFromOffset(blockInfo.StartOffset, blockInfo.Length);
            if (blockInfo.BlockBuffer.Length != blockInfo.Length)
            {
                throw new DataLengthException($"{this.jobName} : Requested length for the reader({blockInfo.Length}) doesn't match the returned length of the stream({blockInfo.BlockBuffer.Length})");
            }

            return blockInfo;
        }

        /// <summary>
        /// The compute block lists.
        /// </summary>
        private async Task ComputeBlockLists()
        {
            this.blockInfosForReader = new ConcurrentQueue<BlockInfo>();
            this.blockInfosForWriter = new BlockingCollection<BlockInfo>(this.maxBlocksToBeCachedForWriter);
            this.orderedBlockIds = new List<long>();
            var processedBlockIds = await this.writer.GetSavedBlockIds() ?? new HashSet<long>();

            var blockIndex = 0;
            for (; blockIndex < (int)Math.Floor((double)this.totalLengthOfSourceContent / this.CurrentBlockLength); blockIndex++)
            {
                if (!processedBlockIds.Contains(blockIndex))
                {
                    this.blockInfosForReader.Enqueue(new BlockInfo(blockIndex, blockIndex * (long)this.CurrentBlockLength, this.CurrentBlockLength));
                }

                this.orderedBlockIds.Add(blockIndex);
            }

            var remainingBytes = (int)(this.totalLengthOfSourceContent % this.CurrentBlockLength);
            if (remainingBytes > 0)
            {
                if (!processedBlockIds.Contains(blockIndex))
                {
                    this.blockInfosForReader.Enqueue(new BlockInfo(blockIndex, blockIndex * (long)this.CurrentBlockLength, remainingBytes));
                }

                this.orderedBlockIds.Add(blockIndex);
            }
        }

        /// <summary>
        /// The calculate block size within constraints.
        /// </summary>
        /// <exception cref="Exception">
        /// Throws an Exception when it can't figure out block size within the constraints.
        /// </exception>
        private void CalculateBlockSizeWithinConstraints()
        {
            var numberOfChunks = (long)Math.Floor((double)this.totalLengthOfSourceContent / this.CurrentBlockLength);
            if (this.CurrentBlockLength > this.maxBlockLengthSupported && numberOfChunks > this.maxNumberOfBlocksAllowedOnWriter)
            {
                throw new Exception($@"{this.jobName} : Block size({this.CurrentBlockLength}) & number of blocks allowed ({numberOfChunks}) 
                        combination is exceeding the max length({this.maxBlockLengthSupported}) & max number of blocks({this.maxNumberOfBlocksAllowedOnWriter}) supported by both reader and writer.");
            }

            if (numberOfChunks > this.maxNumberOfBlocksAllowedOnWriter)
            {
                this.CurrentBlockLength += 100 * Kb;
                this.CalculateBlockSizeWithinConstraints();
            }
        }
    }
}
