﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Threading.Tasks;

    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using SharpCompress.Common;
    using SharpCompress.Readers;
    using SharpCompress.Writers;

    public static class Helpers
    {
        public static async Task RenameBlockBlobAsync(ILogger logger, CloudStorageAccount acct, string containerName, string oldBlobName, string newBlobName)
        {
            await MoveBlockBlobAsync(logger, acct, containerName, oldBlobName, newBlobName, true);
        }

        public static async Task CopyBlockBlobAsync(ILogger logger, CloudStorageAccount acct, string containerName, string oldBlobName, string newBlobName)
        {
            await MoveBlockBlobAsync(logger, acct, containerName, oldBlobName, newBlobName, false);
        }

        public static bool GetBlockBlob(ILogger logger, CloudStorageAccount acct, string containerName, string blobName, out CloudBlockBlob blob)
        {
            return GetBlob(logger, acct, containerName, blobName, out blob, false, false);
        }

        public static bool TryGetBlockBlob(ILogger logger, CloudStorageAccount acct, string containerName, string blobName, out CloudBlockBlob blob)
        {
            return GetBlob(logger, acct, containerName, blobName, out blob, false, true);
        }

        public static bool GetBlockBlobWithAttributes(ILogger logger, CloudStorageAccount acct, string containerName, string blobName, out CloudBlockBlob blob)
        {
            return GetBlob(logger, acct, containerName, blobName, out blob, true, false);
        }

        public static bool TryGetBlockBlobWithAttributes(ILogger logger, CloudStorageAccount acct, string containerName, string blobName, out CloudBlockBlob blob)
        {
            return GetBlob(logger, acct, containerName, blobName, out blob, true, true);
        }

        public static int TryExtractArchiveFile(ILogger logger, string src, string destPath)
        {
            int count = 0;
            try
            {
                if (!Directory.Exists(destPath))
                {
                    Directory.CreateDirectory(destPath);
                }

                using (var stream = File.OpenRead(src))
                {
                    var r = ReaderFactory.Open(stream);
                    r.WriteAllToDirectory(destPath);

                    stream.Position = 0;
                    r = ReaderFactory.Open(stream);
                    while (r.MoveToNextEntry())
                    {
                        ++count;
                    }
                }
            }
            catch (Exception ex)
            {
                logger.Log(LogType.Error, $"TryExtract error: {ex.Message}");
                return int.MinValue;
            }

            return count;
        }

        public static void ExtractArchiveStream(Stream stream, string destPath)
        {
            if (!Directory.Exists(destPath))
            {
                Directory.CreateDirectory(destPath);
            }

            var r = ReaderFactory.Open(stream);
            r.WriteAllToDirectory(destPath);
        }

        public static int TryCreateArchive(ILogger logger, ArchiveType type, string archivePathAndName, IEnumerable<string> srcFiles)
        {
            int count = 0;

            switch (type)
            {
                case ArchiveType.Zip:
                    try
                    {
                        var opts = new WriterOptions(CompressionType.BZip2);
                        using (var zip = File.OpenWrite(archivePathAndName))
                        using (var zipWriter = WriterFactory.Open(zip, type, opts))
                        {
                            foreach (var filePath in srcFiles)
                            {
                                zipWriter.Write(Path.GetFileName(filePath), filePath);
                                ++count;
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                        logger.Log(LogType.Error, $"TryCreateArchive error: {ex.Message}");
                        count = int.MinValue;
                    }

                    break;

                default:
                    throw new NotImplementedException($"TryCreateArchive: {type} is not implemented");
            }

            return count;
        }

        /// <summary>
        /// The retry Utility.
        /// </summary>
        /// <param name="action">
        /// The action.
        /// </param>
        /// <param name="numberOfRetries">
        /// The number of retries.
        /// </param>
        /// <param name="actionDescription">
        /// The action description.
        /// </param>
        /// <param name="milliSecsToWaitBetweenRetries">
        /// Number of milliseconds to wait between retries.
        /// </param>
        /// <param name="listOfExceptionTypesToRetryFor">
        /// The list of exception types to retry for.
        /// </param>
        /// <param name="listIsInclusive">
        /// The list provided is inclusive if True. All the given exception types will be retried.
        /// The list provided is to exclusive if False. All the given exception types will NOT be retried.
        /// </param>
        /// <returns>
        /// The <see cref="Task"/>.
        /// </returns>
        /// <exception cref="AggregateException">
        /// Throws AggregateException if all the retries fail.
        /// </exception>
        public static async Task Retry(Func<Task> action, int numberOfRetries, string actionDescription, long milliSecsToWaitBetweenRetries, List<Type> listOfExceptionTypesToRetryFor = null, bool listIsInclusive = true)
        {
            var exceptionsOccurred = new List<Exception>();
            var currentRetryCounter = numberOfRetries;
            while (true)
            {
                try
                {
                    await action();
                    break;
                }
                catch (Exception ex) 
                    when (listOfExceptionTypesToRetryFor == null 
                        || !listOfExceptionTypesToRetryFor.Any() 
                        || (listOfExceptionTypesToRetryFor.Any(e => e.IsAssignableFrom(ex.GetType())) == listIsInclusive))
                {
                    exceptionsOccurred.Add(ex);
                    if (currentRetryCounter == 0)
                    {
                        throw new AggregateException($"All retry attempts({numberOfRetries}) failed for {actionDescription}", exceptionsOccurred);
                    }

                    await Task.Delay(TimeSpan.FromMilliseconds(milliSecsToWaitBetweenRetries * (numberOfRetries - currentRetryCounter)));
                    currentRetryCounter--;
                }
            }
        }

        public static void EnsureCleanFolder(string folder)
        {
            if (Directory.Exists(folder))
            {
                foreach (var file in Directory.EnumerateFiles(folder))
                {
                    File.Delete(file);
                }

                foreach (var dir in Directory.EnumerateDirectories(folder))
                {
                    try
                    {
                        Directory.Delete(dir, true);
                    }
                    catch
                    {
                    }
                }
            }
            else
            {
                Directory.CreateDirectory(folder);
            }
        }

        /// <summary>
        /// Parse a resource id from APIM account-id field. 
        /// </summary>
        /// <param name="accountIdParam">APIM account-id</param>
        /// <returns>Parsed resource id, otherwise empty string</returns>
        public static string ParseResourceIdFromAccountId(string accountIdParam)
        {
            var tokens = accountIdParam.Split(new[] { "::" }, StringSplitOptions.RemoveEmptyEntries);
            if (tokens.Length != 3 && tokens.Length != 4)
            {
                return string.Empty; 
            }

            var subscriptionId = tokens[0];
            var resourceGroup = tokens[1];
            var resourceName = Uri.EscapeUriString(tokens[2]);
            return $"/subscriptions/{subscriptionId}/resourceGroups/{resourceGroup}/providers/Microsoft.Maps/accounts/{resourceName}";
        }

        private static async Task MoveBlockBlobAsync(ILogger logger, CloudStorageAccount acct, string containerName, string oldBlobName, string newBlobName, bool deleteOldBlob)
        {
            GetBlob(logger, acct, containerName, oldBlobName, out CloudBlockBlob oldBlobRef, true, false);
            GetBlob(logger, acct, containerName, newBlobName, out CloudBlockBlob newBlobRef, true, false);

            if (!await oldBlobRef.ExistsAsync())
            {
                throw new FileNotFoundException($"Old BlockBlob({oldBlobName}) doesn't exist.");
            }

            if (await newBlobRef.ExistsAsync())
            {
                throw new IOException($"New BlockBlob({newBlobName}) already exist.");
            }

            await newBlobRef.StartCopyAsync(oldBlobRef);
            if (deleteOldBlob)
            {
                await oldBlobRef.DeleteAsync();
            }
        }

        private static bool GetBlob(ILogger logger, CloudStorageAccount acct, string containerName, string blobName, out CloudBlockBlob blob, bool getAttributes, bool isTry)
        {
            try
            {
                var blobClient = acct.CreateCloudBlobClient();
                var container = blobClient.GetContainerReference(containerName);
                container.CreateIfNotExists();

                blob = container.GetBlockBlobReference(blobName);

                if (getAttributes && blob.Exists())
                {
                    blob.FetchAttributes();
                }

                return true;
            }
            catch (Exception ex)
            {
                logger.Log(LogType.Error, $"TryGetBlockBlobReference error getting {containerName}.{blobName }blob reference: {ex.Message}");
                blob = null;

                if (isTry)
                {
                    return false;
                }
                else
                {
                    throw;
                }
            }
        }
    }
}
