﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Net.Http;
    using System.Net.Http.Headers;
    using System.Threading.Tasks;

    /// <summary>
    /// Reads the Http file in blocks
    /// </summary>
    public class HttpFileBlockReader : IRandomBlockReader, IDisposable
    {
        /// <summary>
        /// ContentType Key
        /// </summary>
        private const string ContentTypeKey = "ContentType";

        /// <summary>
        /// Last modified key
        /// </summary>
        private const string LastModifiedKey = "LastModified";

        /// <summary>
        /// Mb
        /// </summary>
        private const int Mb = 1024 * 1024;

        /// <summary>
        /// The number of retries allowed.
        /// </summary>
        private const int NumberOfRetriesAllowed = 3;

        /// <summary>
        /// The source url.
        /// </summary>
        private string sourceUrl;

        /// <summary>
        /// The md 5 hash file url.
        /// </summary>
        private string md5HashFileUrl;

        /// <summary>
        /// The compute md 5 locally.
        /// </summary>
        private bool computeMd5Locally;

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpFileBlockReader"/> class.
        /// </summary>
        /// <param name="sourceFileUrl">
        /// The source File Url.
        /// </param>
        /// <param name="md5HashFileUrl">
        /// The md 5 Hash File Url.
        /// </param>
        /// <param name="computeMd5Locally">
        /// The compute Md 5 Locally.
        /// </param>
        public HttpFileBlockReader(string sourceFileUrl, string md5HashFileUrl, bool computeMd5Locally)
        {
            this.sourceUrl = sourceFileUrl;
            this.md5HashFileUrl = md5HashFileUrl;
            this.computeMd5Locally = computeMd5Locally;
        }

        /// <summary>
        /// Gets the max length of block allowed.
        /// </summary>
        public int MaxLengthOfBlockAllowed => 2 * Mb;

        /// <summary>
        /// Gets the preferred length of block.
        /// </summary>
        public int? PreferredLengthOfBlock => null;

        /// <summary>
        /// Gets the length of the stream.
        /// </summary>
        public long LengthOfTheStream { get; private set; }

        /// <summary>
        /// Last Modified DateTime.
        /// </summary>
        public DateTimeOffset LastModifieDateTime { get; private set; }

        /// <summary>
        /// Gets the md 5 hash of source.
        /// </summary>
        public string Md5HashOfSource { get; private set; }

        public Dictionary<string, string> MetadataMap { get; private set; }

        /// <summary>
        /// Gets the max degree of parallelism allowed.
        /// </summary>
        public int MaxDegreeOfParallelismAllowed => 10;

        /// <summary>
        /// The initializer.
        /// </summary>
        public async Task InitializeAsync()
        {
            using (var httpClient = new HttpClient { Timeout = TimeSpan.FromMinutes(30) })
            {
                var response = await httpClient.GetAsync(this.sourceUrl, HttpCompletionOption.ResponseHeadersRead);
                if (response.IsSuccessStatusCode)
                {
                    if (response.Content.Headers.ContentLength.HasValue)
                    {
                        this.LengthOfTheStream = response.Content.Headers.ContentLength.Value;
                    }
                    else
                    {
                        throw new Exception("The server did not return the ContentLength.");
                    }

                    if (response.Content.Headers.LastModified.HasValue)
                    {
                        this.LastModifieDateTime = response.Content.Headers.LastModified.Value;
                    }
                    else
                    {
                        throw new Exception("The server did not return the LastModified value.");
                    }

                    this.MetadataMap = GetMetadataMapFromContentHeaders(response.Content.Headers);
                }
                else
                {
                    throw new Exception($"Http Call failed. Reason returned by the server : {response.ReasonPhrase}");
                }

                // Now take care of the hash.
                if (String.IsNullOrEmpty(this.md5HashFileUrl))
                {
                    this.Md5HashOfSource = null;
                }
                else
                {
                    await Helpers.Retry(
                        async () =>
                            {
                                var request = new HttpRequestMessage() { RequestUri = new Uri(this.md5HashFileUrl) };
                                using (var md5Response = await httpClient.SendAsync(request))
                                {
                                    this.Md5HashOfSource = await md5Response.Content.ReadAsStringAsync();
                                }
                            },
                        NumberOfRetriesAllowed,
                        "GetMd5HashValue From source file URL",
                        1000);
                }
            }
        }

        /// <summary>
        /// The get bytes from offset.
        /// </summary>
        /// <param name="offset">
        /// The offset.
        /// </param>
        /// <param name="numberOfBytes">
        /// The number of bytes.
        /// </param>
        /// <returns>
        /// The <see cref="Task"/>.
        /// </returns>
        public async Task<byte[]> GetBufferFromOffset(long offset, int numberOfBytes)
        {
            byte[] byteArray = null;
            await Helpers.Retry(async () =>
                {
                    var request = new HttpRequestMessage() { RequestUri = new Uri(this.sourceUrl) };
                    request.Headers.Range = new RangeHeaderValue(offset, offset + numberOfBytes - 1);
                    using (var httpClient = new HttpClient { Timeout = TimeSpan.FromMinutes(30) })
                    {
                        var response = await httpClient.SendAsync(request);
                        if (!response.IsSuccessStatusCode)
                        {
                            throw new Exception($"Http block read didn't get success code. Got : {response.StatusCode}");
                        }

                        using (var httpStream = await response.Content.ReadAsStreamAsync())
                        {
                            byteArray = new byte[numberOfBytes];
                            var bytesRead = await httpStream.ReadAsync(byteArray, 0, numberOfBytes);
                            if (bytesRead != numberOfBytes)
                            {
                                throw new Exception($"HttpFileBlockReader at offset {offset}, got only {bytesRead} bytes. Expected bytes: {numberOfBytes}");
                            }
                        }
                    }
                },
                NumberOfRetriesAllowed,
                "Source-GetBufferFromOffset",
                10000);

            return byteArray;
        }
        
        public void Dispose()
        {
        }

        /// <summary>
        /// The get metadata map from content headers.
        /// </summary>
        /// <param name="headers">
        /// The headers.
        /// </param>
        private static Dictionary<string, string> GetMetadataMapFromContentHeaders(HttpContentHeaders headers)
        {
            var map = new Dictionary<string, string>();

            if (headers.ContentMD5 != null)
            {
                map.Add("ContentMD5", Convert.ToBase64String(headers.ContentMD5));
            }

            if (headers.ContentType != null)
            {
                map.Add(ContentTypeKey, headers.ContentType.ToString());
            }

            if (headers.LastModified != null)
            {
                map.Add(LastModifiedKey, headers.LastModified.ToString());
            }

            if (headers.ContentRange != null)
            {
                map.Add("ContentRange", headers.ContentRange.ToString());
            }

            if (headers.Expires != null)
            {
                map.Add("Expires", headers.Expires.ToString());
            }

            return map;
        }
    }
}
