﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common
{
    using System.Collections.Generic;
    using System.IO;
    using System.Threading;
    using System.Threading.Tasks;

    /// <summary>
    /// Supports writing blocks of data to the destination stream/file in random order and then commit all of them in the order of the block ids given.
    /// </summary>
    public interface IRandomBlockWriter
    {
        /// <summary>
        /// Gets the max length of block allowed.
        /// </summary>
        int MaxLengthOfBlockAllowed { get; }

        /// <summary>
        /// Gets the preferred length of block.
        /// </summary>
        int? PreferredLengthOfBlock { get; }

        /// <summary>
        /// Gets the max number of blocks allowed.
        /// </summary>
        long MaxNumberOfBlocksAllowed { get; }

        /// <summary>
        /// Gets the max degree of parallelism allowed.
        /// </summary>
        int MaxDegreeOfParallelismAllowed { get; }

        /// <summary>
        /// Gets the md 5 hash of Target.
        /// </summary>
        string Md5HashOfTarget { get; }

        /// <summary>
        /// Sets the metadata map.
        /// </summary>
        /// <param name="metadataMap">
        /// The metadata Map.
        /// </param>
        void SetMetadataMap(IDictionary<string, string> metadataMap);

        /// <summary>
        /// The write block async.
        /// </summary>
        /// <param name="blockId">
        /// The block id.
        /// </param>
        /// <param name="blockBuffer">
        /// The block content.
        /// </param>
        /// <returns>
        /// The <see cref="Task"/>.
        /// </returns>
        Task<bool> WriteBlockAsync(long blockId, byte[] blockBuffer);

        /// <summary>
        /// Returns the previously saved blockIds. This allows the controller to skip these.
        /// </summary>
        /// <returns></returns>
        Task<HashSet<long>> GetSavedBlockIds();

        /// <summary>
        /// Initializes the required objects.
        /// </summary>
        /// <returns></returns>
        Task InitializeAsync();

        /// <summary>
        /// The commit blocks async.
        /// </summary>
        /// <param name="orderedListOfBlockIds">
        /// The ordered list of block ids.
        /// </param>
        /// <param name="cancellationToken">cancellation token</param>
        /// <returns>
        /// The <see cref="Task"/>.
        /// </returns>
        Task CommitBlocksAsync(List<long> orderedListOfBlockIds, CancellationToken cancellationToken);
    }
}