﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common
{
    using System;
    using System.Threading.Tasks;
    using Microsoft.Azure.KeyVault;
    using Microsoft.Azure.Services.AppAuthentication;

    public class KeyVaultConfigurationReader : IConfigurationReader
    {
        private readonly AzureServiceTokenProvider _azureServiceTokenProvider;
        private readonly IConfigurationCache _cache;
        private readonly KeyVaultClient _keyVaultClient;
        private bool disposedValue = false; // To detect redundant calls        
        
        public KeyVaultConfigurationReader()
        {
            _azureServiceTokenProvider = new AzureServiceTokenProvider();
            _keyVaultClient = new KeyVaultClient(new KeyVaultClient.AuthenticationCallback(_azureServiceTokenProvider.KeyVaultTokenCallback));            
            _cache = new MemoryConfigurationCache(Guid.NewGuid().ToString(("n")));
        }

        /// <summary>
        /// Get the the configuration setting/secret from key vault and returns back the value.
        /// Uses an underlying caching mechanism to avoid trashing key vault.
        /// </summary>
        /// <param name="key">the url which is the pointer in key vault to download the configuration value</param>
        /// <returns>the configuration value</returns>
        public async Task<string> GetConfigurationValueAsync(string key)
        {
            var keyvaultSettingFactory = new Lazy<Task<string>>(
                async () =>
                {
                    var secret = await _keyVaultClient.GetSecretAsync(key).ConfigureAwait(false);
                    return secret.Value;
                });

            return await _cache.ResolveSettingAsync(key, keyvaultSettingFactory, null).ConfigureAwait(false);
        }

        #region IDisposable Support
        // This code added to correctly implement the disposable pattern.
        public void Dispose()
        {
            // Do not change this code. Put cleanup code in Dispose(bool disposing) above.
            Dispose(true);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    // TODO: dispose managed state (managed objects).
                    _cache.Dispose();
                    _keyVaultClient.Dispose();
                }

                disposedValue = true;
            }
        }
        #endregion
    }
}
