﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common
{
    using System;
    using System.Runtime.Caching;
    using System.Threading.Tasks;

    public class MemoryConfigurationCache : IConfigurationCache
    {        
        private const string _cacheName = "ALSMemoryCache";
        private static DateTimeOffset _defaultCacheExpiration;
        private readonly MemoryCache _memoryCache;
        private bool _disposedValue = false; // To detect redundant calls                                     

        public MemoryConfigurationCache(string cacheName)
        {
            if (string.IsNullOrWhiteSpace(cacheName))
            {
                cacheName = _cacheName;
            }

            _memoryCache = new MemoryCache(cacheName);
            _defaultCacheExpiration = ObjectCache.InfiniteAbsoluteExpiration;
        }

        /// <summary>
        /// Resolves secret setting by calling Key Vault
        /// </summary>
        /// <param name="settingName"> The setting name </param>
        /// <param name="lazyFactoryMethod">The factory method to resovle the secret's configuration</param>
        /// <param name="cachedExpirationTimeSpan"> the time span to keep the secret value in the cache </param>
        /// <returns> the secret value </returns>
        public async Task<T> ResolveSettingAsync<T>(string settingName, Lazy<Task<T>> lazyFactoryMethod, TimeSpan? cachedExpirationTimeSpan)
        {
            var absoluteExpiration = _defaultCacheExpiration;
            if (cachedExpirationTimeSpan.HasValue)
            {
                absoluteExpiration = DateTimeOffset.Now + cachedExpirationTimeSpan.Value;
            }
            
            return await CacheAddOrGet(settingName, lazyFactoryMethod, absoluteExpiration).ConfigureAwait(false);
        }

        #region IDisposable Support        
        // This code added to correctly implement the disposable pattern.
        public void Dispose()
        {
            // Do not change this code. Put cleanup code in Dispose(bool disposing) above.
            Dispose(true);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposedValue)
            {
                if (disposing)
                {
                    _memoryCache.Dispose();
                }

                _disposedValue = true;
            }
        }
        #endregion

        /// <summary>
        /// Gets a cache value. If the value does not exists add that to the cache and then return
        /// </summary>
        /// <typeparam name="T"> The type of the cached value </typeparam>
        /// <param name="settingName"> The setting name </param>
        /// <param name="secretFactory"> A factory method to resolve the setting name to its value </param>
        /// <param name="absoluteExpiration">expiration of the cache item</param>
        /// <returns> Cached value corresponds to the setting </returns>
        private T CacheAddOrGet<T>(string settingName, Lazy<T> secretFactory, DateTimeOffset absoluteExpiration)
        {
            // Get the existing cache member or if not available add new value. 
            // AddOrGetExisting is a thread-safe atomic operation which handles the locking mechanism for thread safty. 
            // To use the operation, the value is passes in as Lazy initialization to only be calculated when the key is not in the cache and to be atomic and thread-safe.
            Lazy<T> cachedValue = _memoryCache.AddOrGetExisting(
                               settingName,
                               secretFactory,
                               absoluteExpiration) as Lazy<T>;
            try
            {
                // For the first time adding the cache entry, cachedValue is set to null so the lazy object will be initialized
                return (cachedValue ?? secretFactory).Value;
            }
            catch
            {                
                _memoryCache.Remove(settingName);
                throw;
            }
        }
    }
}
