﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
namespace Microsoft.Azure.LocationServices.Common
{
    using System;
    using System.Text;
    using System.Threading;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using Microsoft.WindowsAzure.Storage.RetryPolicies;

    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.OrderingRules", "SA1201:ElementsMustAppearInTheCorrectOrder")]
    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.OrderingRules", "SA1202:ElementsMustBeOrderedByAccess")]
    public class PublisherTokenSentinel : IDisposable
    {
        public static readonly TimeSpan LeaseSpan = TimeSpan.FromSeconds(60); //see http://msdn.microsoft.com/en-us/library/ee691972.aspx
        
        private static readonly TimeSpan _sLeaseRequestTimeout = TimeSpan.FromSeconds(5);
        private static readonly BlobRequestOptions _sLeaseBlobRequestOptions = new BlobRequestOptions()
        {
            MaximumExecutionTime = TimeSpan.FromSeconds(30),
            ServerTimeout = TimeSpan.FromSeconds(5),
            RetryPolicy = new LinearRetry(TimeSpan.FromMilliseconds(500), 10)
        };

        private readonly string _leaseBlobName;
        private readonly string _leaseBlobContainer;
        private readonly Logger _loggerRef;
        private ManualResetEvent _disposeEvent;
        private Thread _leaseThread;
        private string _leaseId;
        private CloudStorageAccount _account;
        private CloudBlockBlob _leaseBlob;

        public PublisherTokenSentinel(CloudStorageAccount acct, Logger logger, string leaseBlobContainer, string leaseBlobName)
        {
            _leaseBlobName = leaseBlobName; 
            _leaseBlobContainer = leaseBlobContainer.ToLowerInvariant(); 
            _loggerRef = logger;
            _account = acct;
            _disposeEvent = new ManualResetEvent(false);
            _leaseThread = new Thread(new ThreadStart(LeaseThread));
            _leaseThread.Name = "PublisherTokenSentinel";
            _leaseThread.Priority = ThreadPriority.BelowNormal;
            _leaseThread.IsBackground = true;
            _leaseThread.Start();
        }

        /// <summary>If true, this instance is responsible for compiling and publishing data updates</summary>
        public bool HasLease => !string.IsNullOrEmpty(_leaseId);

        /// <summary>for Lease Blob, see http://msdn.microsoft.com/en-us/library/ee691972.aspx </summary>
        private void LeaseThread()
        {
            _leaseId = null;
            _leaseBlob = null;

            var buffer = TimeSpan.FromSeconds(1);
            var sleep = LeaseSpan;

            do
            {
                var start = DateTime.Now;
                try
                {
                    if (TryInitLeaseBlob())
                    {
                        System.Diagnostics.Debug.Assert(_leaseBlob != null, "_leaseBlob shouldn't be null at this point");
                        if (HasLease) // renew
                        {
                            var acc = new AccessCondition() { LeaseId = _leaseId };
                            _leaseBlob.RenewLease(acc, _sLeaseBlobRequestOptions, null);
                        }
                        else // try acquiring
                        {
                            _leaseBlob.FetchAttributes();
                            if (_leaseBlob.Properties.LeaseStatus != LeaseStatus.Locked || _leaseBlob.Properties.LeaseState != LeaseState.Leased)
                            {
                                _leaseId = _leaseBlob.AcquireLease(LeaseSpan, null, null, _sLeaseBlobRequestOptions, null);
                                _loggerRef.Log(LogType.Info, "lease acquired");
                                TryStampLeaseBlobAsync(Environment.MachineName);
                            }
                        }
                    }

                    if (HasLease)
                    {
                        sleep = LeaseSpan - (DateTime.Now - start).Add(buffer);

                        // if trying the lease takes more than 59 seconds there is no point chasing it
                        if (sleep.TotalSeconds < 1)
                        {
                            sleep = LeaseSpan;
                        }
                    }
                    else
                    {
                        sleep = LeaseSpan;
                    }
                }
                catch (Exception ex)
                {
                    _loggerRef.Log(LogType.Error, $"LeaseThread error: {ex.Message}");
                    _leaseId = null;
                    sleep = LeaseSpan;
                }

                _loggerRef.Log(LogType.Info, $"LeaseThread sleep for {sleep.TotalSeconds.ToString("0.00")}. HasLease={HasLease}.");
            }
            while (!_disposeEvent.WaitOne(sleep, false));

            _loggerRef.Log(LogType.Verbose, "LeaseThread exited while loop and cleaning up");
            ReleaseLease();
            _loggerRef.Log(LogType.Verbose, "LeaseThread exiting");
        }

        /// <summary>Init the lease blob container if it doesn't exist</summary>
        private bool TryInitLeaseBlob()
        {
            if (_leaseBlob != null && _leaseBlob.Exists())
            {
                return true;
            }

            if (!Helpers.TryGetBlockBlob(_loggerRef, _account, _leaseBlobContainer, _leaseBlobName, out _leaseBlob))
            {
                return false;
            }

            if (!_leaseBlob.Exists())
            {
                _loggerRef.Log(LogType.Verbose, "TryInitLeaseBlob creating and locking _leaseBlob");
                try
                {
                    _leaseBlob.UploadText($"Created by {Environment.MachineName} on {DateTime.UtcNow}", Encoding.UTF8, null, _sLeaseBlobRequestOptions, null);
                }
                catch (Exception ex)
                {
                    _loggerRef.Log(LogType.Error, $"TryInitLeaseBlob error creating blob: {ex.Message}");
                    _leaseBlob = null;
                    return false;
                }

                // locking the blob now is gravy:  no need to return false as long as the blob has been created by writing some text to it
                try
                {
                    _leaseId = _leaseBlob.AcquireLease(LeaseSpan, null, null, _sLeaseBlobRequestOptions, null);
                }
                catch (Exception ex)
                {
                    _loggerRef.Log(LogType.Error, $"TryInitLeaseBlob error locking newly created blob: {ex.Message}");
                }

                System.Diagnostics.Debug.Assert(_leaseBlob.Exists(), "_leaseBlob has to exist at this point");
            }

            return _leaseBlob != null && _leaseBlob.Exists();
        }

        /// <summary>Release the lease - if it holds it</summary>
        private void ReleaseLease()
        {
            if (!HasLease)
            {
                return;
            }

            _loggerRef.Log(LogType.Verbose, "ReleaseLease");
            TryStampLeaseBlobAsync();
            try
            {
                var acc = new AccessCondition() { LeaseId = _leaseId };
                _leaseBlob.ReleaseLease(acc, _sLeaseBlobRequestOptions, null);
            }
            catch (Exception ex)
            {
                _loggerRef.Log(LogType.Error, $"ReleaseLease error: {ex.Message}");
            }

            _leaseId = null;
        }

        /// <summary>Set the owner name in the lease blob</summary>
        private void TryStampLeaseBlobAsync(string owner = null)
        {
            System.Diagnostics.Debug.Assert(HasLease, "StampLeaseBlob invoked without a lease");
            try
            {
                var blobText = string.IsNullOrEmpty(owner)
                    ? $"Unlocked since {DateTime.UtcNow}"
                    : $"Claimed by {owner} on {DateTime.UtcNow}";

                var acc = new AccessCondition() { LeaseId = _leaseId };
                _leaseBlob.UploadTextAsync(blobText, Encoding.UTF8, acc, null, null);
            }
            catch (Exception ex)
            {
                _loggerRef.Log(LogType.Error, $"StampLeaseBlob error: {ex.Message}");
            }
        }

        #region ----- IDisposable -----
        /// <summary>Track whether Dispose has been called or not</summary>
        public bool Disposed { get; private set; }

        /// <summary>IDispose pattern, straight out of MSDN</summary>
        public void Dispose()
        {
            Dispose(true);

            // This object will be cleaned up by the Dispose method.
            // Therefore, you should call GC.SupressFinalize to take this object off the finalization queue 
            // and prevent finalization code for this object from executing a second time.
            GC.SuppressFinalize(this);
        }

        /// <summary>Dispose(bool disposing) executes in two distinct scenarios.
        /// If disposing equals true, the method has been called directly or indirectly by a user's code. 
        /// Managed and unmanaged resources can be disposed.
        /// If disposing equals false, the method has been called by the runtime from inside the finalizer
        /// and you should not reference other objects. Only unmanaged resources can be disposed.
        /// </summary>
        /// <param name="disposing"></param>
        private void Dispose(bool disposing)
        {
            // Check to see if Dispose has already been called.
            if (!Disposed)
            {
                Disposed = true;

                // If disposing equals true, dispose all managed and unmanaged resources.
                if (disposing)
                {
                    //signal threads to exit and give them a chance to exit gracefully
                    _loggerRef.Log(LogType.Verbose, "PublisherTokenSentinel signaling _disposeEvent");
                    _disposeEvent.Set();
                    _loggerRef.Log(LogType.Verbose, "PublisherTokenSentinel signaled _disposeEvent and waiting on _leaseThread to exit");
                    _leaseThread.Join(2000);
                    _loggerRef.Log(LogType.Verbose, "PublisherTokenSentinel done waiting on _coordinatorThread.Join");
                    while (_leaseThread.ThreadState == ThreadState.Running || _leaseThread.ThreadState == ThreadState.Background)
                    {
                        _loggerRef.Log(LogType.Warning, "PublisherTokenSentinel _leaseThread still running!!");
                        try
                        {
                            _leaseThread.Abort();
                        }
                        catch (ThreadAbortException)
                        {
                             /*nothing*/
                        }
                    }

                    _loggerRef.LogIf(HasLease, LogType.Warning, "PublisherTokenSentinel still has publisher token!");
                    ReleaseLease();

                    if (_disposeEvent != null)
                    {
                        _disposeEvent.Dispose();
                        _disposeEvent = null;
                    }

                    _loggerRef.Log(LogType.Verbose, "PublisherTokenSentinel done");
                }
            }
        }

        /// <summary>Use C# destructor syntax for finalization code.
        /// This destructor will run only if the Dispose method does not get called.
        /// It gives your base class the opportunity to finalize.
        /// Do not provide destructors in types derived from this class.
        /// </summary>
        ~PublisherTokenSentinel()
        {
            Dispose(false);
        }
        #endregion // IDisposable
    }
}
