﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
namespace Microsoft.Azure.LocationServices.Common
{
    using System;
    using System.Collections.Generic;
    using System.Data;
    using System.Data.SqlClient;
    using System.Threading;

    public class SqlUtility
    {
        private const int _maxRetries = 5;

        private readonly string _connectionString;
        private readonly TimeSpan _sqlDefaultTimeoutInSecs;
        private readonly TimeSpan _sqlRetryWaitTime;
        private readonly List<ConnectionState> _inActiveConnectionStates;

        public SqlUtility(string connectionString, TimeSpan sqlDefaultTimeout, TimeSpan sqlRetryWaitTime)
        {
            _connectionString = connectionString;
            _sqlDefaultTimeoutInSecs = sqlDefaultTimeout;
            _sqlRetryWaitTime = sqlRetryWaitTime;
            _inActiveConnectionStates = new List<ConnectionState>() { ConnectionState.Closed, ConnectionState.Broken };
        }

        /// <summary>
        /// Performs the given SQL action with retry mechanism.
        /// </summary>
        /// <param name="action"></param>
        public void ExecuteSqlActionWithRetries(Action action)
        {
            var retries = 0;
            while (true)
            {
                try
                {
                    action();
                    break;
                }
                catch (SqlException)
                {
                    retries++;
                    if (retries >= _maxRetries)
                    {
                        throw;
                    }

                    Thread.Sleep(_sqlRetryWaitTime);
                }
            }
        }

        /// <summary>
        /// Executes the SQL read command with retry mechanism.
        /// </summary>
        /// <param name="action"></param>
        public void ExecuteSqlReadCommandWithRetries(Action<SqlCommand> action)
        {
            var retries = 0;
            while (true)
            {
                using (SqlConnection connection = GetActiveSqlConnectionObject())
                {
                    using (SqlCommand command = new SqlCommand())
                    {
                        command.Connection = connection;
                        command.CommandTimeout = (int)_sqlDefaultTimeoutInSecs.TotalSeconds;
                        try
                        {
                            action(command);
                            break;
                        }
                        catch (SqlException)
                        {
                            retries++;
                            if (retries >= _maxRetries)
                            {
                                throw;
                            }

                            Thread.Sleep(_sqlRetryWaitTime);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Performs the SQL action using the given connection instance, with retry logic.
        /// This method is to support temp tables, where we have to perform multiple operations using the same connection.
        /// </summary>
        /// <param name="connection"></param>
        /// <param name="action"></param>
        public void ExecuteSqlCommandWithRetries(SqlConnection connection, Action<SqlCommand> action)
        {
            var retries = 0;
            while (true)
            {
                // If the given connection was broken, throw the exception 
                if (!CheckIfConnectionIsStillActive(connection))
                {
                    throw new Exception("Cannot proceed further as the connection was not active.");
                }

                using (SqlTransaction transaction = connection.BeginTransaction())
                {
                    using (SqlCommand command = new SqlCommand())
                    {
                        command.Transaction = transaction;
                        command.Connection = connection;
                        command.CommandTimeout = (int)_sqlDefaultTimeoutInSecs.TotalSeconds;
                        try
                        {
                            action(command);
                            transaction.Commit();
                            break;
                        }
                        catch (SqlException)
                        {
                            // Attempt to roll back the transaction.
                            if (!CheckIfConnectionIsStillActive(transaction.Connection))
                            {
                                //Do not proceed further if we can not rollback the current transaction.
                                throw;
                            }

                            RollbackSQLTransaction(transaction);
                            retries++;
                            if (retries >= _maxRetries)
                            {
                                throw;
                            }

                            Thread.Sleep(_sqlRetryWaitTime);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Perform the given Sql action using the new SQL connection instance, with retry logic.
        /// </summary>
        public void ExecuteSqlCommandWithRetries(Action<SqlCommand> action)
        {
            var retries = 0;
            while (true)
            {
                using (SqlConnection connection = GetActiveSqlConnectionObject())
                {
                    using (SqlTransaction transaction = connection.BeginTransaction())
                    {
                        using (SqlCommand command = new SqlCommand())
                        {
                            command.Transaction = transaction;
                            command.Connection = connection;
                            command.CommandTimeout = (int)_sqlDefaultTimeoutInSecs.TotalSeconds;
                            try
                            {
                                action(command);
                                transaction.Commit();
                                break;
                            }
                            catch (SqlException)
                            {
                                // Attempt to roll back the transaction.
                                if (!CheckIfConnectionIsStillActive(transaction.Connection))
                                {
                                    //Do not proceed further if we can not rollback the current transaction.
                                    throw;
                                }

                                RollbackSQLTransaction(transaction);
                                retries++;
                                if (retries >= _maxRetries)
                                {
                                    throw;
                                }

                                Thread.Sleep(_sqlRetryWaitTime);
                            }
                        }
                    }
                }
            }
        }

        ///// <summary>
        ///// Returns an active Sql Connection object.
        ///// </summary>
        ///// <returns></returns>
        public SqlConnection GetActiveSqlConnectionObject()
        {
            var retries = 0;
            SqlConnection conn = new SqlConnection(_connectionString);
            while (true)
            {
                try
                {
                    conn.Open();
                    break;
                }
                catch (SqlException)
                {
                    retries++;
                    if (retries >= _maxRetries)
                    {
                        throw;
                    }

                    Thread.Sleep(_sqlRetryWaitTime);
                }
            }

            return conn;
        }

        /// <summary>
        /// Method to rollback the SQL transaction.
        /// </summary>
        /// <param name="transaction"></param>
        public void RollbackSQLTransaction(SqlTransaction transaction)
        {
            // Attempt to roll back the transaction.
            try
            {
                //Transaction will not be rolled back, if the associated connection was closed / null.
                if (!CheckIfConnectionIsStillActive(transaction.Connection))
                {
                    throw new Exception("Can not roll back the transaction, as the connection was not active.");
                }

                transaction.Rollback();
            }
            catch (Exception ex)
            {
                // This catch block will handle any errors that may have occurred
                // on the server that would cause the rollback to fail, such as a closed connection.                                    
                //Terminate the process if Rollback fails.
                throw new Exception("Exception occured during the Transaction Rollback.", ex);
            }
        }

        /// <summary>
        /// Method to check  whether the given connection was in active state or not.
        /// </summary>
        /// <param name="connection"></param>
        /// <returns></returns>
        public bool CheckIfConnectionIsStillActive(SqlConnection connection)
        {
            if (connection == null || _inActiveConnectionStates.Contains(connection.State))
            {
                return false;
            }

            return true;
        }
    }
}