﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Text;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common.Validation;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using Newtonsoft.Json;

    public class AzureBlobService : IBlobService
    {
        private readonly CloudBlobClient client;

        #region Constructors

        public AzureBlobService(string connectionString)
        {
            Assert.ParamIsNotNullOrEmpty(connectionString, nameof(connectionString));

            CloudStorageAccount storageAccount = CloudStorageAccount.Parse(connectionString);
            this.client = storageAccount.CreateCloudBlobClient();
        }

        public AzureBlobService(CloudStorageAccount storageAccount)
        {
            Assert.ParamIsNotNull(storageAccount, nameof(storageAccount));

            this.client = storageAccount.CreateCloudBlobClient();
        }

        #endregion

        #region IBlobService

        public CloudBlockBlob GetBlockBlobReference(string containerName, string blobName)
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);

            return container.GetBlockBlobReference(blobName);
        }

        public CloudAppendBlob GetAppendBlobReference(string containerName, string blobName)
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);

            return container.GetAppendBlobReference(blobName);
        }

        public async Task<ICloudBlob> GetBlobReferenceAsync(Uri blobUri)
        {
            Assert.ParamIsNotNull(blobUri, nameof(blobUri));
            
            return await this.client.GetBlobReferenceFromServerAsync(blobUri);
        }

        public CloudBlobContainer GetContainerReference(string containerName)
        {
            Assert.ParamIsNotNull(containerName, nameof(containerName));

            return this.client.GetContainerReference(containerName);
        }

        public async Task<bool> CreateContainerIfNotExistsAsync(
            string containerName, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            return await container.CreateIfNotExistsAsync(cancellationToken);
        }

        public async Task<bool> CreateContainerIfNotExistsAsync(
            CloudBlobContainer container,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNull(container, nameof(container));
            
            return await container.CreateIfNotExistsAsync(cancellationToken);
        }

        public async Task CreateAppendBlobIfNotExistsAsync(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudAppendBlob appendBlob = container.GetAppendBlobReference(blobName);

            if (!await this.DoesAppendBlobExistAsync(containerName, blobName, cancellationToken))
            {
                await appendBlob.CreateOrReplaceAsync(cancellationToken);
            }
        }

        public async Task<bool> DeleteContainerIfExistsAsync(
            string containerName, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            return await container.DeleteIfExistsAsync(cancellationToken);
        }

        public async Task<bool> DeleteBlobIfExistsAsync(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudBlockBlob blockBlob = container.GetBlockBlobReference(blobName);

            return await blockBlob.DeleteIfExistsAsync(cancellationToken);
        }

        public async Task DeleteBlobAsync(
            CloudBlockBlob blockBlob,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNull(blockBlob, nameof(blockBlob));

            await blockBlob.DeleteAsync(cancellationToken);
        }

        public async Task UploadBlockBlobFromStreamAsync(
            CloudBlockBlob blockBlob,
            Stream stream,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNull(blockBlob, nameof(blockBlob));
            Assert.ParamIsNotNull(stream, nameof(stream));

            await blockBlob.UploadFromStreamAsync(stream, cancellationToken);
        }

        public async Task UploadBlockBlobFromStreamAsync(
            CloudBlockBlob blockBlob,
            MemoryStream ms,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNull(blockBlob, nameof(blockBlob));
            Assert.ParamIsNotNull(ms, nameof(ms));

            await blockBlob.UploadFromStreamAsync(ms, cancellationToken);
        }

        public async Task UploadBlockBlobFromStreamAsync(
            string containerName, 
            string blobName, 
            MemoryStream ms, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));
            Assert.ParamIsNotNull(ms, nameof(ms));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudBlockBlob blockBlob = container.GetBlockBlobReference(blobName);

            await blockBlob.UploadFromStreamAsync(ms, cancellationToken);
        }

        public async Task UploadAppendBlobFromStreamAsync(
            CloudAppendBlob appendBlob,
            MemoryStream ms,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNull(appendBlob, nameof(appendBlob));
            Assert.ParamIsNotNull(ms, nameof(ms));
            Assert.ParamIsNotNull(accessCondition, nameof(accessCondition));
            Assert.ParamIsNotNull(blobRequestOptions, nameof(blobRequestOptions));
            Assert.ParamIsNotNull(operationContext, nameof(operationContext));

            await appendBlob.UploadFromStreamAsync(ms, accessCondition, blobRequestOptions, operationContext, cancellationToken);
        }

        public async Task UploadFromByteArrayAsync(
            string containerName, 
            string blobName, 
            byte[] buffer, 
            int offset, 
            int count, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));
            Assert.ParamIsNotNull(buffer, nameof(buffer));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudBlockBlob blockBlob = container.GetBlockBlobReference(blobName);

            await blockBlob.UploadFromByteArrayAsync(buffer, offset, count, cancellationToken);
        }

        public async Task UploadObjectAsJsonAsync<T>(
            string containerName, 
            string blobName, 
            T data, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));
            Assert.ParamIsNotNull(data, nameof(data));

            using (var ms = new MemoryStream(Encoding.UTF8.GetBytes(JsonConvert.SerializeObject(data))))
            {
                await this.UploadBlockBlobFromStreamAsync(containerName, blobName, ms, cancellationToken);
            }
        }

        public async Task UploadDocumentAsync(
            string containerName, 
            string blobName, 
            string data, 
            IDictionary<string, string> metadata = null,
            string contentType = "application/json",
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));
            Assert.ParamIsNotNullOrEmpty(contentType, nameof(contentType));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudBlockBlob blockBlob = container.GetBlockBlobReference(blobName);

            blockBlob.Properties.ContentType = contentType;

            if (metadata != null)
            {
                foreach (KeyValuePair<string, string> kv in metadata)
                {
                    blockBlob.Metadata.Add(kv.Key, kv.Value);
                }
            }

            using (var ms = new MemoryStream(Encoding.UTF8.GetBytes(data)))
            {
                await blockBlob.UploadFromStreamAsync(ms, cancellationToken);
            }
        }

        public async Task AppendTextAsync(
            string containerName,
            string blobName,
            string text,
            Encoding encoding,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));
            Assert.ParamIsNotNullOrEmpty(text, nameof(text));
            Assert.ParamIsNotNull(accessCondition, nameof(accessCondition));
            Assert.ParamIsNotNull(blobRequestOptions, nameof(blobRequestOptions));
            Assert.ParamIsNotNull(operationContext, nameof(operationContext));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudAppendBlob appendBlob = container.GetAppendBlobReference(blobName);

            await appendBlob.AppendTextAsync(text, encoding, accessCondition, blobRequestOptions, operationContext, cancellationToken);
        }

        public async Task DownloadBlockBlobToStreamAsync(
            string containerName, 
            string blobName, 
            MemoryStream target, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudBlockBlob blockBlob = container.GetBlockBlobReference(blobName);

            await blockBlob.DownloadToStreamAsync(target, cancellationToken);
        }

        public async Task DownloadAppendBlobToStreamAsync(
            string containerName,
            string blobName,
            MemoryStream target,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));
            Assert.ParamIsNotNull(accessCondition, nameof(accessCondition));
            Assert.ParamIsNotNull(blobRequestOptions, nameof(blobRequestOptions));
            Assert.ParamIsNotNull(operationContext, nameof(operationContext));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudAppendBlob appendBlob = container.GetAppendBlobReference(blobName);

            await appendBlob.DownloadToStreamAsync(target, accessCondition, blobRequestOptions, operationContext, cancellationToken);
        }

        public async Task<T> DownloadAsync<T>(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            using (var ms = new MemoryStream())
            {
                await this.DownloadBlockBlobToStreamAsync(containerName, blobName, ms, cancellationToken);
                return JsonConvert.DeserializeObject<T>(Encoding.UTF8.GetString(ms.ToArray()));
            }
        }

        public async Task<bool> DoesBlockBlobExistAsync(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudBlockBlob blockBlob = container.GetBlockBlobReference(blobName);

            return await blockBlob.ExistsAsync(cancellationToken);
        }

        public async Task<bool> DoesAppendBlobExistAsync(
            string containerName,
            string blobName,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudAppendBlob appendBlob = container.GetAppendBlobReference(blobName);

            return await appendBlob.ExistsAsync(cancellationToken);
        }

        public async Task<bool> DoesContainerExistAsync(
            string containerName, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);

            return await container.ExistsAsync(cancellationToken);
        }

        public async Task<CloudBlockBlob> GetBlockBlobAttributesAsync(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudBlockBlob blockBlob = container.GetBlockBlobReference(blobName);

            await blockBlob.FetchAttributesAsync(cancellationToken);

            return blockBlob;
        }

        public async Task<CloudAppendBlob> GetAppendBlobAttributesAsync(
            string containerName,
            string blobName,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNullOrEmpty(containerName, nameof(containerName));
            Assert.ParamIsNotNullOrEmpty(blobName, nameof(blobName));
            Assert.ParamIsNotNull(accessCondition, nameof(accessCondition));
            Assert.ParamIsNotNull(blobRequestOptions, nameof(blobRequestOptions));
            Assert.ParamIsNotNull(operationContext, nameof(operationContext));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);
            CloudAppendBlob appendBlob = container.GetAppendBlobReference(blobName);

            await appendBlob.FetchAttributesAsync(accessCondition, blobRequestOptions, operationContext, cancellationToken);

            return appendBlob;
        }

        public async Task<BlobProperties> GetBlockBlobPropertiesAsync(
            string containerName,
            string blobName,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            CloudBlockBlob blob = await this.GetBlockBlobAttributesAsync(containerName, blobName, cancellationToken);

            return blob.Properties;
        }

        public async Task<BlobProperties> GetAppendBlobPropertiesAsync(
            string containerName,
            string blobName,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            CloudAppendBlob blob = 
                await this.GetAppendBlobAttributesAsync(
                    containerName, 
                    blobName,
                    accessCondition,
                    blobRequestOptions,
                    operationContext,
                    cancellationToken);

            return blob.Properties;
        }

        public async Task<IDictionary<string, string>> GetBlockBlobMetadataAsync(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken))
        {
            CloudBlockBlob blob = await this.GetBlockBlobAttributesAsync(containerName, blobName, cancellationToken);

            return blob.Metadata;
        }

        public async Task<IDictionary<string, string>> GetAppendBlobMetadataAsync(
            string containerName,
            string blobName,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            CloudAppendBlob blob = 
                await this.GetAppendBlobAttributesAsync(
                    containerName, 
                    blobName, 
                    accessCondition, 
                    blobRequestOptions, 
                    operationContext, 
                    cancellationToken);

            return blob.Metadata;
        }

        public async Task SetAppendBlobMetadataAsync(
            CloudAppendBlob appendBlob,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNull(appendBlob, nameof(appendBlob));

            await appendBlob.SetMetadataAsync(cancellationToken);
        }

        public async Task SetContainerMetadataAsync(
            string containerName, 
            IDictionary<string, string> metadata,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNull(containerName, nameof(containerName));
            Assert.ParamIsNotNull(metadata, nameof(metadata));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);

            // Fetch container attributes in order to populate the container's properties and metadata.
            await container.FetchAttributesAsync(cancellationToken);

            foreach (KeyValuePair<string, string> kv in metadata)
            {
                container.Metadata.Add(kv.Key, kv.Value);
            }
            
            // Set the container's metadata.
            await container.SetMetadataAsync(cancellationToken);
        }

        public async Task<IDictionary<string, string>> GetContainerMetadataAsync(
            string containerName,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNull(containerName, nameof(containerName));

            CloudBlobContainer container = this.client.GetContainerReference(containerName);

            // Fetch container attributes in order to populate the container's properties and metadata.
            await container.FetchAttributesAsync(cancellationToken);

            // Return the container's metadata.
            return container.Metadata;
        }

        public async Task<string> AcquireLeaseAsync(
            CloudAppendBlob appendBlob,
            TimeSpan? leaseTime,
            string proposedLeaseId = null,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNull(appendBlob, nameof(appendBlob));

            return await appendBlob.AcquireLeaseAsync(leaseTime, proposedLeaseId, cancellationToken);
        }

        public async Task ReleaseLeaseAsync(
            CloudAppendBlob appendBlob,
            AccessCondition accessCondition,
            CancellationToken cancellationToken = default(CancellationToken))
        {
            Assert.ParamIsNotNull(appendBlob, nameof(appendBlob));
            Assert.ParamIsNotNull(accessCondition, nameof(accessCondition));

            await appendBlob.ReleaseLeaseAsync(accessCondition, cancellationToken);
        }

        #endregion

        #region Private Helpers

        #endregion
    }
}
