﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common.Storage
{
    using System;
    using System.Collections.Generic;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common.Validation;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Queue;

    /// <summary>
    /// A simplified interface to the Azure Storage Queue Service 
    /// </summary>
    public class AzureQueueService : IQueueService
    {
        private readonly CloudQueueClient client;

        #region Constructors

        public AzureQueueService(string connectionString)
        {
            Assert.ParamIsNotNullOrEmpty(connectionString, nameof(connectionString));

            CloudStorageAccount storageAccount = CloudStorageAccount.Parse(connectionString);
            this.client = storageAccount.CreateCloudQueueClient();
        }

        public AzureQueueService(CloudStorageAccount storageAccount)
        {
            Assert.ParamIsNotNull(storageAccount, nameof(storageAccount));

            this.client = storageAccount.CreateCloudQueueClient();
        }

        #endregion

        #region IQueueService

        public CloudQueue GetQueueReference(string queueName)
        {
            Assert.ParamIsNotNullOrEmpty(queueName, nameof(queueName));

            return this.client.GetQueueReference(queueName);
        }

        public async Task<bool> QueueExistsAsync(CloudQueue queue, CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(queue, nameof(queue));

            return await queue.ExistsAsync(token);
        }

        public async Task<bool> CreateQueueIfNotExistsAsync(CloudQueue queue, CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(queue, nameof(queue));

            return await queue.CreateIfNotExistsAsync(token);
        }

        public async Task<int?> GetApproximateMessageCountAsync(CloudQueue queue, CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(queue, nameof(queue));

            // First refresh queue attributes
            await queue.FetchAttributesAsync(token);

            // Check the current number of messages in the queue
            return queue.ApproximateMessageCount;
        }

        public async Task ClearQueue(CloudQueue queue, CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(queue, nameof(queue));

            await queue.ClearAsync(token);
        }

        public async Task AddMessageAsync(CloudQueue queue, CloudQueueMessage message, CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(queue, nameof(queue));
            Assert.ParamIsNotNull(message, nameof(message));

            await queue.AddMessageAsync(message, token);
        }

        public async Task AddMessageAsync(
            CloudQueue queue, 
            CloudQueueMessage message, 
            TimeSpan? timeToLive, 
            TimeSpan? initialVisibilityDelay, 
            QueueRequestOptions queueRequestOptions,
            OperationContext operationContext,
            CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(queue, nameof(queue));
            Assert.ParamIsNotNull(message, nameof(message));

            await queue.AddMessageAsync(
                message, 
                timeToLive, 
                initialVisibilityDelay, 
                queueRequestOptions ?? new QueueRequestOptions(), 
                operationContext ?? new OperationContext(), 
                token);
        }

        public async Task<CloudQueueMessage> GetMessageAsync(
            CloudQueue queue, 
            TimeSpan? visibilityTimeout,
            QueueRequestOptions queueRequestOptions,
            OperationContext operationContext,
            CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(queue, nameof(queue));

            return await queue.GetMessageAsync(
                visibilityTimeout,
                queueRequestOptions ?? new QueueRequestOptions(),
                operationContext ?? new OperationContext(),
                token);
        }

        public async Task<IEnumerable<CloudQueueMessage>> GetMessagesAsync(
            CloudQueue queue, 
            int messageCount, 
            TimeSpan? visibilityTimeout,
            QueueRequestOptions queueRequestOptions,
            OperationContext operationContext,
            CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(queue, nameof(queue));

            return await queue.GetMessagesAsync(
                messageCount, 
                visibilityTimeout,
                queueRequestOptions ?? new QueueRequestOptions(),
                operationContext ?? new OperationContext(),
                token);
        }

        public async Task UpdateMessageAsync(
            CloudQueue queue,
            CloudQueueMessage message,
            TimeSpan visibilityTimeout,
            MessageUpdateFields updateFields,
            CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(queue, nameof(queue));
            Assert.ParamIsNotNull(message, nameof(message));
            Assert.ParamIsNotNull(visibilityTimeout, nameof(visibilityTimeout));
            Assert.ParamIsNotNull(updateFields, nameof(updateFields));

            await queue.UpdateMessageAsync(message, visibilityTimeout, updateFields, token);
        }

        public async Task DeleteMessageAsync(
            CloudQueue queue, 
            CloudQueueMessage message, 
            CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(queue, nameof(queue));
            Assert.ParamIsNotNull(message, nameof(message));

            await queue.DeleteMessageAsync(message, token);
        }

        #endregion
    }
}
