﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common.Storage
{
    using System.Collections.Generic;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common.Validation;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Table;

    public class AzureTableService : ITableService
    {
        private readonly CloudTableClient client;

        #region Constructors

        public AzureTableService(string connectionString)
        {
            Assert.ParamIsNotNullOrEmpty(connectionString, nameof(connectionString));

            CloudStorageAccount storageAccount = CloudStorageAccount.Parse(connectionString);
            this.client = storageAccount.CreateCloudTableClient();
        }

        public AzureTableService(CloudStorageAccount storageAccount)
        {
            Assert.ParamIsNotNull(storageAccount, nameof(storageAccount));

            this.client = storageAccount.CreateCloudTableClient();
        }

        #endregion

        #region ITableService
        
        public CloudTable GetTableReference(string tableName)
        {
            Assert.ParamIsNotNullOrEmpty(tableName, nameof(tableName));

            return this.client.GetTableReference(tableName);
        }

        public async Task<bool> CreateTableIfNotExistsAsync(CloudTable table, CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(table, nameof(table));

            return await table.CreateIfNotExistsAsync(token);
        }

        public async Task<TableResult> InsertAsync<TEntity>(CloudTable table, TEntity entity, CancellationToken token = default(CancellationToken)) where TEntity : ITableEntity, new()
        {
            Assert.ParamIsNotNull(table, nameof(table));
            Assert.ParamIsNotNull(entity, nameof(entity));
            
            return await table.ExecuteAsync(TableOperation.Insert(entity), token);
        }

        public async Task<TableResult> ReplaceAsync<TEntity>(CloudTable table, TEntity entity, CancellationToken token = default(CancellationToken)) where TEntity : ITableEntity, new()
        {
            Assert.ParamIsNotNull(table, nameof(table));
            Assert.ParamIsNotNull(entity, nameof(entity));

            return await table.ExecuteAsync(TableOperation.Replace(entity), token);
        }

        public async Task<TableResult> DeleteAsync<TEntity>(CloudTable table, TEntity entity, CancellationToken token = default(CancellationToken)) where TEntity : ITableEntity, new()
        {
            Assert.ParamIsNotNull(table, nameof(table));
            Assert.ParamIsNotNull(entity, nameof(entity));

            return await table.ExecuteAsync(TableOperation.Delete(entity), token);
        }

        public IEnumerable<TEntity> ExecuteQuery<TEntity>(CloudTable table, TableQuery<TEntity> query) where TEntity : class, ITableEntity, new()
        {
            Assert.ParamIsNotNull(table, nameof(table));
            Assert.ParamIsNotNull(query, nameof(query));

            return table.ExecuteQuery(query);
        }

        public async Task<TableResult> ExecuteAsync(CloudTable table, TableOperation tableOperation, CancellationToken token = default(CancellationToken))
        {
            Assert.ParamIsNotNull(table, nameof(table));
            Assert.ParamIsNotNull(tableOperation, nameof(tableOperation));

            return await table.ExecuteAsync(tableOperation, token);
        }

        #endregion
    }
}