﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Text;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;

    /// <summary>
    /// A simplified interface to the Azure Storage Blob Service 
    /// </summary>
    public interface IBlobService
    {
        #region Get
        
        /// <summary>
        /// Returns a reference to the block blob with the passed in container and blob name
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <returns>A reference to the Blob</returns>
        CloudBlockBlob GetBlockBlobReference(string containerName, string blobName);

        /// <summary>
        /// Returns a reference to the append blob with the passed in container and blob name
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <returns>A reference to the Blob</returns>
        CloudAppendBlob GetAppendBlobReference(string containerName, string blobName);

        /// <summary>
        /// Initiates an asynchronous operation to fetch the blob reference for the given Uri
        /// </summary>
        /// <param name="blobUri">Uri for the blob</param>
        /// <returns>A reference to the Blob</returns>
        Task<ICloudBlob> GetBlobReferenceAsync(Uri blobUri);

        /// <summary>
        /// Initiates an operation to fetch the container reference for the given container name
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <returns>A reference to the Container</returns>
        CloudBlobContainer GetContainerReference(string containerName);

        #endregion

        #region Create

        /// <summary>
        /// Initiates an asynchronous operation to create a container with the given name if one doesn't exist
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>
        /// A bool value indicating the create container result:
        ///     true   -  Container was successfully created
        ///     false  -  Container was not created
        ///     false  -  Container already exists 
        /// </returns>
        Task<bool> CreateContainerIfNotExistsAsync(
            string containerName, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to create a container if one doesn't exist
        /// </summary>
        /// <param name="container">A valid container reference</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>
        /// A bool value indicating the create container result:
        ///     true   -  Container was successfully created
        ///     false  -  Container was not created
        ///     false  -  Container already exists 
        /// </returns>
        Task<bool> CreateContainerIfNotExistsAsync(
            CloudBlobContainer container, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to create an append blob if one doesn't exist
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="cancellationToken">A cancellation token</param>
        Task CreateAppendBlobIfNotExistsAsync(
            string containerName,
            string blobName,
            CancellationToken cancellationToken = default(CancellationToken));

        #endregion

        #region Delete

        /// <summary>
        /// Initiates an asynchronous operation to deletes a container if it already exists
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="cancellationToken">A cancellation token</param>
        Task<bool> DeleteContainerIfExistsAsync(
            string containerName, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to delete the blob if it already exists
        /// </summary>
        /// <param name="containerName">The name of the container containing the blob</param>
        /// <param name="blobName">The name of the blob to delete</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>Returns True if the blob existed and was successfully deleted. False otherwise.</returns>
        Task<bool> DeleteBlobIfExistsAsync(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to delete the passed-in blob
        /// </summary>
        /// <param name="blockBlob">The blob to be deleted</param>
        /// <param name="cancellationToken">A cancellation token</param>
        Task DeleteBlobAsync(
            CloudBlockBlob blockBlob,
            CancellationToken cancellationToken = default(CancellationToken));

        #endregion

        #region Upload

        /// <summary>
        /// Initiates an asynchronous operation to upload a <see cref="Stream"/> to a block blob. If the blob already exists, it will be overwritten.
        /// </summary>
        /// <param name="blockBlob">A valid blob reference</param>
        /// <param name="stream">The memory stream to upload</param>
        /// <param name="cancellationToken">A cancellation token</param>
        Task UploadBlockBlobFromStreamAsync(
            CloudBlockBlob blockBlob,
            Stream stream,
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to upload a <see cref="MemoryStream"/> to a block blob. If the blob already exists, it will be overwritten.
        /// </summary>
        /// <param name="blockBlob">A valid blob reference</param>
        /// <param name="ms">The memory stream to upload</param>
        /// <param name="cancellationToken">A cancellation token</param>
        Task UploadBlockBlobFromStreamAsync(
            CloudBlockBlob blockBlob,
            MemoryStream ms,
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to upload a <see cref="MemoryStream"/> to a block blob. If the blob already exists, it will be overwritten.
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="ms">The memory stream to upload</param>
        /// <param name="cancellationToken">A cancellation token</param>
        Task UploadBlockBlobFromStreamAsync(
            string containerName, 
            string blobName, 
            MemoryStream ms, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to upload a <see cref="MemoryStream"/> to an append blob. If the blob already exists, it will be overwritten.
        /// </summary>
        /// <param name="appendBlob">Append Blob to which the data should be uploaded to</param>
        /// <param name="ms">The memory stream to upload</param>
        /// <param name="accessCondition">An AccessCondition object that represents the condition that must be met in order for the request to proceed.</param>
        /// <param name="blobRequestOptions">A BlobRequestOptions object that specifies additional options for the request.</param>
        /// <param name="operationContext">An OperationContext object that represents the context for the current operation.</param>
        /// <param name="cancellationToken">A cancellation token</param>
        Task UploadAppendBlobFromStreamAsync(
            CloudAppendBlob appendBlob,
            MemoryStream ms,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to upload the contents of a byte array to a blob. If the blob already exists, it will be overwritten.
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="buffer">An array of bytes</param>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin</param>
        /// <param name="count">The number of bytes to be written</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns></returns>
        Task UploadFromByteArrayAsync(
            string containerName, 
            string blobName, 
            byte[] buffer, 
            int offset, 
            int count, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to upload a typed data model as json to a blob. If the blob already exists, it will be overwritten. 
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="data">The object to upload</param>
        /// <param name="cancellationToken">A cancellation token</param>
        Task UploadObjectAsJsonAsync<T>(
            string containerName, 
            string blobName, 
            T data, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to upload document data to a blob. If the blob already exists, it will be overwritten.
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="data">Document data to upload</param>
        /// <param name="metadata">Metadata for the blob</param>
        /// <param name="contentType">The content type for the document. Default is application/json.</param>
        /// <param name="cancellationToken">A cancellation cancellationToken</param>
        Task UploadDocumentAsync(
            string containerName,
            string blobName,
            string data,
            IDictionary<string, string> metadata = null,
            string contentType = "application/json",
            CancellationToken cancellationToken = default(CancellationToken));

        #endregion

        #region Append

        /// <summary>
        /// Initiates an asynchronous operation to append a string of text to an append blob.
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="text">A string containing the text to upload.</param>
        /// <param name="encoding">A Encoding object that indicates the text encoding to use. If null, UTF-8 will be used.</param>
        /// <param name="accessCondition">An AccessCondition object that represents the condition that must be met in order for the request to proceed.</param>
        /// <param name="blobRequestOptions">A BlobRequestOptions object that specifies additional options for the request.</param>
        /// <param name="operationContext">An OperationContext object that represents the context for the current operation.</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns></returns>
        Task AppendTextAsync(
            string containerName,
            string blobName,
            string text,
            Encoding encoding,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken));

        #endregion

        #region Download

        /// <summary>
        /// Initiates an asynchronous operation to download the contents of a block blob to a stream.
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="target">Stream to download blob contents to.</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns></returns>
        Task DownloadBlockBlobToStreamAsync(
            string containerName, 
            string blobName, 
            MemoryStream target, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to download the contents of an append blob to a stream.
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="target">Stream to download blob contents to.</param>
        /// <param name="accessCondition">An AccessCondition object that represents the condition that must be met in order for the request to proceed.</param>
        /// <param name="blobRequestOptions">A BlobRequestOptions object that specifies additional options for the request.</param>
        /// <param name="operationContext">An OperationContext object that represents the context for the current operation.</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns></returns>
        Task DownloadAppendBlobToStreamAsync(
            string containerName,
            string blobName,
            MemoryStream target,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to download the contents of a blob to a typed object.
        /// </summary>
        /// <typeparam name="T">The generic type to deserialize the object to</typeparam>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="cancellationToken">A cancellation token</param>
        Task<T> DownloadAsync<T>(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken));

        #endregion

        #region Check Existence

        /// <summary>
        /// Initiates an asynchronous operation to check existence of the block blob
        /// </summary>
        /// <param name="containerName">The container for the blob</param>
        /// <param name="blobName">The blob name</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>True if blob exists. False otherwise.</returns>
        Task<bool> DoesBlockBlobExistAsync(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to check existence of the append blob
        /// </summary>
        /// <param name="containerName">The container for the blob</param>
        /// <param name="blobName">The blob name</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>True if blob exists. False otherwise.</returns>
        Task<bool> DoesAppendBlobExistAsync(
            string containerName,
            string blobName,
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation that checks whether the container exists
        /// </summary>
        /// <param name="containerName">The container name to check</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>True if the container exists. False otherwise.</returns>
        Task<bool> DoesContainerExistAsync(
            string containerName, 
            CancellationToken cancellationToken = default(CancellationToken));

        #endregion

        #region Properties and Metadata

        /// <summary>
        /// Initiates an asynchronous operation to fetch the blob's properties and metadata
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>A <see cref="CloudBlockBlob"/> object with it's properties and metadata filled in</returns>
        Task<CloudBlockBlob> GetBlockBlobAttributesAsync(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to fetch the blob's properties and metadata
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="accessCondition">An AccessCondition object that represents the condition that must be met in order for the request to proceed.</param>
        /// <param name="blobRequestOptions">A BlobRequestOptions object that specifies additional options for the request.</param>
        /// <param name="operationContext">An OperationContext object that represents the context for the current operation.</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>A <see cref="CloudAppendBlob"/> object with it's properties and metadata filled in</returns>
        Task<CloudAppendBlob> GetAppendBlobAttributesAsync(
            string containerName,
            string blobName,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to fetch the blob's properties
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>A <see cref="BlobProperties"/> object containing the blob's properties</returns>
        Task<BlobProperties> GetBlockBlobPropertiesAsync(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to fetch the blob's properties
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="accessCondition">An AccessCondition object that represents the condition that must be met in order for the request to proceed.</param>
        /// <param name="blobRequestOptions">A BlobRequestOptions object that specifies additional options for the request.</param>
        /// <param name="operationContext">An OperationContext object that represents the context for the current operation.</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>A <see cref="BlobProperties"/> object containing the blob's properties</returns>
        Task<BlobProperties> GetAppendBlobPropertiesAsync(
            string containerName,
            string blobName,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to fetch the blob's metadata
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>A dictionary containing the blob's metadata</returns>
        Task<IDictionary<string, string>> GetBlockBlobMetadataAsync(
            string containerName, 
            string blobName, 
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to fetch the blob's metadata
        /// </summary>
        /// <param name="containerName">A valid name for the container</param>
        /// <param name="blobName">A valid name for the blob</param>
        /// <param name="accessCondition">An AccessCondition object that represents the condition that must be met in order for the request to proceed.</param>
        /// <param name="blobRequestOptions">A BlobRequestOptions object that specifies additional options for the request.</param>
        /// <param name="operationContext">An OperationContext object that represents the context for the current operation.</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>A dictionary containing the blob's metadata</returns>
        Task<IDictionary<string, string>> GetAppendBlobMetadataAsync(
            string containerName,
            string blobName,
            AccessCondition accessCondition,
            BlobRequestOptions blobRequestOptions,
            OperationContext operationContext,
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to set the blob's metadata
        /// </summary>
        /// <param name="appendBlob">Blob's whose metadata needs to be updated</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>A dictionary containing the blob's metadata</returns>
        Task SetAppendBlobMetadataAsync(
            CloudAppendBlob appendBlob,
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to set the containers's metadata.
        /// </summary>
        /// <param name="containerName">A valid name for an exisitng container</param>
        /// <param name="metadata">Metadata for the container</param>
        /// <param name="cancellationToken">A cancellation token</param>
        Task SetContainerMetadataAsync(
            string containerName,
            IDictionary<string, string> metadata,
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to fetch the container's metadata
        /// </summary>
        /// <param name="containerName">A valid name for an exisitng container</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>A dictionary containing the container's metadata</returns>
        Task<IDictionary<string, string>> GetContainerMetadataAsync(
            string containerName,
            CancellationToken cancellationToken = default(CancellationToken));

        #endregion

        #region Leasing

        /// <summary>
        /// Initiates an asynchronous operation to acquire a lease on this blob.
        /// </summary>
        /// <param name="appendBlob">Blob on which the lease has to be acquired</param>
        /// <param name="leaseTime">A span of time for which to acquire the lease. If null, an infinite lease will be acquired. If not null, this must be 15 to 60 seconds.</param>
        /// <param name="proposedLeaseId">A string representing the proposed lease ID for the new lease, or null if no lease ID is proposed.</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>A dictionary containing the blob's metadata</returns>
        Task<string> AcquireLeaseAsync(
            CloudAppendBlob appendBlob,
            TimeSpan? leaseTime,
            string proposedLeaseId = null,
            CancellationToken cancellationToken = default(CancellationToken));

        /// <summary>
        /// Initiates an asynchronous operation to release the lease on this blob.
        /// </summary>
        /// <param name="appendBlob">Blob on which the lease has to be released</param>
        /// <param name="accessCondition">An object representing the condition that must be met in order for the request to proceed, including a required lease ID</param>
        /// <param name="cancellationToken">A cancellation token</param>
        /// <returns>A dictionary containing the blob's metadata</returns>
        Task ReleaseLeaseAsync(
            CloudAppendBlob appendBlob,
            AccessCondition accessCondition,
            CancellationToken cancellationToken = default(CancellationToken));

        #endregion
    }
}