﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common.Storage
{
    using System;
    using System.Collections.Generic;
    using System.Threading;
    using System.Threading.Tasks;

    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Queue;

    public interface IQueueService
    {
        #region Queue Operations

        /// <summary>
        /// Get a reference to the queue.
        /// </summary>
        /// <param name="queueName">The name of the queue to refer.</param>
        CloudQueue GetQueueReference(string queueName);

        /// <summary>
        /// Check if a queue exists.
        /// </summary>
        /// <param name="queue">The queue to check.</param>
        /// <param name="token">Cancellation token</param>
        Task<bool> QueueExistsAsync(CloudQueue queue, CancellationToken token = default(CancellationToken));

        /// <summary>
        /// Create a queue if it doesn't exist already.
        /// </summary>
        /// <param name="queue">The queue to check.</param>
        /// <param name="token">Cancellation token.</param>
        Task<bool> CreateQueueIfNotExistsAsync(CloudQueue queue, CancellationToken token = default(CancellationToken));

        /// <summary>
        /// Get an approximate count of all the messages in the Queue (both visible and invisible).
        /// </summary>
        /// <param name="queue">The queue to get the message count for.</param>
        /// <param name="token">Cancellation token.</param>
        Task<int?> GetApproximateMessageCountAsync(CloudQueue queue, CancellationToken token = default(CancellationToken));

        /// <summary>
        /// Clear all the message in the Queue.
        /// </summary>
        /// <param name="queue">The queue to clear.</param>
        /// <param name="token">Cancellation token.</param>
        Task ClearQueue(CloudQueue queue, CancellationToken token = default(CancellationToken));

        #endregion

        #region Add Messages to Queue

        /// <summary>
        /// Start adding a message async.
        /// </summary>
        /// <param name="queue">The queue to add the message to.</param>
        /// <param name="message">Message to add</param>
        /// <param name="token">Cancellation token</param>
        /// <returns>Task to add message</returns>
        Task AddMessageAsync(CloudQueue queue, CloudQueueMessage message, CancellationToken token = default(CancellationToken));

        /// <summary>
        /// Start adding a message async with a visibilityTimeout
        /// </summary>
        /// <param name="queue">The queue to add the message to.</param>
        /// <param name="message">Message to add</param>
        /// <param name="timeToLive">Time for which the message will live in the queue</param>
        /// <param name="initialVisibilityDelay">Timeout after which message will be visible</param>
        /// <param name="queueRequestOptions">Queue Request Options</param>
        /// <param name="operationContext">Operation context</param>
        /// <param name="token">Cancellation token</param> 
        /// <returns>Task to add message</returns>
        Task AddMessageAsync(
            CloudQueue queue,
            CloudQueueMessage message,
            TimeSpan? timeToLive,
            TimeSpan? initialVisibilityDelay,
            QueueRequestOptions queueRequestOptions,
            OperationContext operationContext,
            CancellationToken token = default(CancellationToken));

        #endregion

        #region Get Messages from Queue

        /// <summary>
        /// Start getting a message async
        /// </summary>
        /// <param name="queue">The queue to get the message from.</param>
        /// <param name="visibilityTimeout">Visibiltiy timeout to set.</param>
        /// <param name="queueRequestOptions">Queue Request Options</param>
        /// <param name="operationContext">Operation context</param>
        /// <param name="token">Cancellation token</param>
        /// <returns>New message if available</returns>
        Task<CloudQueueMessage> GetMessageAsync(
            CloudQueue queue,
            TimeSpan? visibilityTimeout,
            QueueRequestOptions queueRequestOptions,
            OperationContext operationContext,
            CancellationToken token = default(CancellationToken));

        /// <summary>
        /// Start getting multiple messages async
        /// </summary>
        /// <param name="queue">The queue to get the message from.</param>
        /// <param name="messageCount">Maximum number of messages.</param>
        /// <param name="visibilityTimeout">Visibiltiy timeout to set.</param>
        /// <param name="queueRequestOptions">Queue Request Options</param>
        /// <param name="operationContext">Operation context</param>
        /// <param name="token">Cancellation token</param>
        /// <returns>New message(s) if available</returns>
        Task<IEnumerable<CloudQueueMessage>> GetMessagesAsync(
            CloudQueue queue,
            int messageCount,
            TimeSpan? visibilityTimeout,
            QueueRequestOptions queueRequestOptions,
            OperationContext operationContext,
            CancellationToken token = default(CancellationToken));

        #endregion

        #region Update Message in Queue

        /// <summary>
        /// Start Updating a message async.
        /// </summary>
        /// <param name="queue">The queue to delete the message from.</param>
        /// <param name="message">Delete m</param>
        /// <param name="visibilityTimeout">Visibility timeout for message</param>
        /// <param name="updateFields">Message fields to update</param>
        /// <param name="token">Cancellation token</param>
        /// <returns>Task to await on.</returns>
        Task UpdateMessageAsync(
            CloudQueue queue,
            CloudQueueMessage message,
            TimeSpan visibilityTimeout,
            MessageUpdateFields updateFields,
            CancellationToken token = default(CancellationToken));

        #endregion

        #region Delete Messages from Queue

        /// <summary>
        /// Start Deleting a message async.
        /// </summary>
        /// <param name="queue">The queue to delete the message from.</param>
        /// <param name="message">Delete message async</param>
        /// <param name="token">Cancellation token</param>
        /// <returns>Task to await on.</returns>
        Task DeleteMessageAsync(
            CloudQueue queue,
            CloudQueueMessage message,
            CancellationToken token = default(CancellationToken));

        #endregion
    }
}