﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common.Storage
{
    using System.Collections.Generic;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.WindowsAzure.Storage.Table;

    public interface ITableService
    {
        /// <summary>
        /// Get a reference to the table.
        /// </summary>
        /// <param name="tableName">The name of the table to refer.</param>
        CloudTable GetTableReference(string tableName);

        /// <summary>
        /// Create a table if it doesn't exist already.
        /// </summary>
        /// <param name="table">The table to create.</param>
        /// <param name="token">Cancellation token.</param>
        Task<bool> CreateTableIfNotExistsAsync(CloudTable table, CancellationToken token = default(CancellationToken));

        /// <summary>
        /// Insert an item, exception will be throw if the item exist.
        /// </summary>
        /// <typeparam name="TEntity">Type Entity to add</typeparam>
        /// <param name="table">The table to add the entity to.</param>
        /// <param name="entity">Entity to add.</param>
        /// <param name="token">Cancellation token</param>
        /// <returns>Results of operation</returns>
        Task<TableResult> InsertAsync<TEntity>(CloudTable table, TEntity entity, CancellationToken token = default(CancellationToken)) where TEntity : ITableEntity, new();

        /// <summary>
        /// Replaces an entity in the table using Etag support. The ETag check does not work if the Etag in the new entity is null.
        /// </summary>
        /// <typeparam name="TEntity">Type Entity to replace</typeparam>
        /// <param name="table">The table to replace the entity in.</param>
        /// <param name="entity">The new entity that will replace the existing entity</param>
        /// <param name="token">Cancellation token</param>
        /// <returns>The operation result</returns>
        Task<TableResult> ReplaceAsync<TEntity>(CloudTable table, TEntity entity, CancellationToken token = default(CancellationToken)) where TEntity : ITableEntity, new();

        /// <summary>
        /// Deletes an entity in the table using Etag support. The ETag check does not work if the Etag in the new entity is null.
        /// </summary>
        /// <typeparam name="TEntity">Type Entity to delete</typeparam>
        /// <param name="table">The table to delete the entity in.</param>
        /// <param name="entity">The entity to be deleted</param>
        /// <param name="token">Cancellation token</param>
        /// <returns>The operation result</returns>
        Task<TableResult> DeleteAsync<TEntity>(CloudTable table, TEntity entity, CancellationToken token = default(CancellationToken)) where TEntity : ITableEntity, new();

        /// <summary>
        /// Execute the query on the table and return all the results.
        /// </summary>
        /// <typeparam name="TEntity">Type Entity to query</typeparam>
        /// <param name="table">The table to query.</param>
        /// <param name="query">The query to execute.</param>
        /// <returns>The operation result</returns>
        IEnumerable<TEntity> ExecuteQuery<TEntity>(CloudTable table, TableQuery<TEntity> query) where TEntity : class, ITableEntity, new();

        /// <summary>
        /// Execute the query on the table and return all the results.
        /// </summary> 
        /// <param name="table">The table to query.</param>
        /// <param name="tableOperation">The table operation to execute.</param>
        /// <param name="token">Cancellation token</param>
        /// <returns>The operation result</returns>
        Task<TableResult> ExecuteAsync(CloudTable table, TableOperation tableOperation, CancellationToken token = default(CancellationToken));
    }
}