﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Common.Validation
{
    using System;
    using System.Globalization;

    /// <summary>
    /// Assertion utilities.
    /// </summary>
    public static class Assert
    {
        /// <summary>
        /// Asserts that a string parameter is not null or empty.
        /// </summary>
        /// <param name="value">The parameter to validate.</param>
        /// <param name="parameterName">The name of the parameter.</param>
        public static void ParamIsNotNullOrEmpty(string value, string parameterName)
        {
            LocalParameterIsNotNullOrWhiteSpace(parameterName, nameof(parameterName));

            string message = null;

            if (value == null)
            {
                message = $"{parameterName} should not be null.";
            }
            else if (string.IsNullOrWhiteSpace(value))
            {
                message = $"{parameterName} should not be empty or whitespace.";
            }

            if (message != null)
            {
                FailAndThrow(message);
            }
        }

        /// <summary>
        /// Asserts that a parameter is not null.
        /// </summary>
        /// <param name="value">The parameter to validate.</param>
        /// <param name="parameterName">The name of the parameter.</param>
        public static void ParamIsNotNull(object value, string parameterName)
        {
            LocalParameterIsNotNullOrWhiteSpace(parameterName, nameof(parameterName));

            if (value == null)
            {
                FailAndThrow($"{parameterName} should not be null.");
            }
        }

        /// <summary>
        /// Asserts that an type value is defined and valid.
        /// </summary>
        /// <param name="value">The value to validate.</param>
        /// <param name="parameterName">The name of the parameter.</param>
        public static void EnumIsDefined<TEnum>(TEnum value, string parameterName)
        {
            LocalParameterIsNotNullOrWhiteSpace(parameterName, nameof(parameterName));
            ParamIsNotNull(value, nameof(value));

            Type enumType = typeof(TEnum);

            if (!Enum.IsDefined(enumType, value))
            {
                FailAndThrow($"{parameterName} has an invalid value '{value}' for type '{enumType.Name}'.");
            }
        }

        /// <summary>
        /// Asserts that the <paramref name="condition" /> is <c>true</c> for the <paramref name="parameterName" />.
        /// </summary>
        /// <param name="condition">The condition to check.</param>
        /// <param name="parameterName">The parameter name.</param>
        /// <param name="errorMessage">Message to show if condition is false</param>
        /// <exception cref="System.ArgumentException"><paramref name="condition" /> evaluates to <c>false</c>.</exception>
        public static void ParameterConditionIsTrue(bool condition, string parameterName, string errorMessage)
        {
            LocalParameterIsNotNullOrWhiteSpace(parameterName, nameof(parameterName));
            LocalParameterIsNotNullOrWhiteSpace(errorMessage, nameof(errorMessage));

            if (!condition)
            {
                FailAndThrow(errorMessage);
            }
        }

        /// <summary>
        /// Asserts that the <paramref name="condition" /> is <c>true</c>.
        /// </summary>
        /// <param name="condition">The condition to check.</param>
        /// <param name="errorMessage">Message to show if condition is false</param>
        /// <exception cref="System.InvalidOperationException"><paramref name="condition" /> evaluates to <c>false</c>.</exception>
        public static void IsTrue(bool condition, string errorMessage)
        {
            LocalParameterIsNotNullOrWhiteSpace(errorMessage, nameof(errorMessage));

            if (!condition)
            {
                Fail(errorMessage);
            }
        }

        /// <summary>
        /// Asserts that a guid parameter is not empty.
        /// </summary>
        /// <param name="value">The guid parameter to validate.</param>
        /// <param name="parameterName">The name of the parameter.</param>
        public static void GuidIsNotEmpty(Guid value, string parameterName)
        {
            LocalParameterIsNotNullOrWhiteSpace(parameterName, nameof(parameterName));

            if (value == Guid.Empty)
            {
                FailAndThrow($"{parameterName} should not be an empty Guid.");
            }
        }

        /// <summary>
        /// Fails with the specified message.
        /// </summary>
        /// <param name="message">The message.</param>
        public static void Fail(string message)
        {
            LocalParameterIsNotNullOrWhiteSpace(message, nameof(message));

            DebugUtilities.Fail(message);
            throw new InvalidOperationException(message);
        }

        /// <summary>
        /// Asserts that a parameter is not null or whitespace.
        /// </summary>
        /// <param name="value">Value to validate.</param>
        /// <param name="parameterName">Name of the parameter.</param>
        private static void LocalParameterIsNotNullOrWhiteSpace(string value, string parameterName)
        {
            if (string.IsNullOrWhiteSpace(value))
            {
                FailAndThrow($"{parameterName} should not be null or empty.");
            }
        }

        /// <summary>
        /// Breaks the Debugger and throws an <see cref="ArgumentException"/> 
        /// </summary>
        /// <param name="message">Error Message</param>
        private static void FailAndThrow(string message)
        {
            DebugUtilities.Fail(message);
            throw new InvalidOperationException(message);
        }
    }
}