﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.Common.UnitTests
{
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Diagnostics.CodeAnalysis;
    using System.Globalization;
    using System.Linq;
    using System.Reflection;

    using Microsoft.VisualStudio.TestTools.UnitTesting;

    /// <summary>
    /// Minor test helper utilities.
    /// </summary>
    [ExcludeFromCodeCoverage]
    public static class TestHelpers
    {
        private static readonly char[] AlphaNumericCharacters;

        private static readonly Random Random;

        static TestHelpers()
        {
            AlphaNumericCharacters = Enumerable.Range('a', 26).Concat(Enumerable.Range('A', 26).Concat(Enumerable.Range('0', 10))).Select(c => (char)c).ToArray();
            Random = new Random();
        }

        /// <summary>
        /// Generates a random alphanumeric string.
        /// </summary>
        /// <param name="length">Desired string length.</param>
        public static string GetRandomString(int length)
        {
            byte[] buffer = new byte[length];
            lock (TestHelpers.Random)
            {
                TestHelpers.Random.NextBytes(buffer);
            }

            return new string(buffer.Select(b => TestHelpers.AlphaNumericCharacters[b % TestHelpers.AlphaNumericCharacters.Length]).ToArray());
        }

        /// <summary>
        /// Generates a random number.
        /// </summary>
        /// <returns>Cid</returns>
        public static long GetRandomLong()
        {
            byte[] bytes = new byte[8];
            lock (TestHelpers.Random)
            {
                TestHelpers.Random.NextBytes(bytes);
            }

            return BitConverter.ToInt64(bytes, 0);
        }

        /// <summary>
        /// Runs a test step and validates that the expected exception type occurs
        /// </summary>
        /// <typeparam name="T">Expected exception type</typeparam>
        /// <param name="test">Test step to run</param>
        [SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        [SuppressMessage("Microsoft.Design", "CA1004:GenericMethodsShouldProvideTypeParameter")]
        public static void AssertThrows<T>(Action test)
            where T : Exception
        {
            try
            {
                test();
                Assert.Fail("Did not throw the expected exception type");
            }
            catch (T)
            {
                // expected
            }
        }

        /// <summary>
        /// Compares all public instance properties of an object
        /// </summary>
        /// <param name="obj1">Object to compare 1</param>
        /// <param name="obj2">Object to compare 2</param>
        /// <param name="skipProperties">Properties to ignore</param>
        public static void PropertyComparer(object obj1, object obj2, List<string> skipProperties = null)
        {
            var type = obj1.GetType();
            Assert.AreEqual(type.FullName, obj2.GetType().FullName, "Objects must be of the same type");

            var properties = type.GetProperties(BindingFlags.Public | BindingFlags.Instance);
            foreach (var property in properties)
            {
                if (skipProperties?.Contains(property.Name) == true)
                {
                    continue;
                }

                var left = property.GetValue(obj1);
                var right = property.GetValue(obj2);

                if (left is IEnumerable leftEnum && right is IEnumerable rightEnum)
                {
                    CollectionAssert.AreEqual(
                        leftEnum.Cast<object>().ToList(),
                        rightEnum.Cast<object>().ToList(),
                        string.Format(CultureInfo.InvariantCulture, "property '{0}' inconsistent", property.Name));
                }
                else
                {
                    Assert.AreEqual(left, right, string.Format(CultureInfo.InvariantCulture, "property '{0}' inconsistent", property.Name));
                }
            }
        }
    }
}