﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace Microsoft.Azure.LocationServices.ServiceFabricHosting
{
    using System;
    using System.ComponentModel;
    using System.Fabric;
    using System.Fabric.Description;
    using System.Net;
    using System.Reflection;
    using System.Security;

    using Validation;

    /// <summary>
    /// Implements the functionality to read configuration from service fabric parameters.
    /// </summary>
    public class ConfigurationReader
    {
        private const string Config = "Config";

        private readonly ICodePackageActivationContext codePackageActivationContext;

        public ConfigurationReader(ServiceContext serviceContext) 
            : this(serviceContext.CodePackageActivationContext)
        {
        }

        public ConfigurationReader(ICodePackageActivationContext activationContext)
        {
            Requires.NotNull(activationContext, nameof(activationContext));
            this.codePackageActivationContext = activationContext;
        }

        /// <summary>
        /// Returns parameter value or default value if not found.
        /// </summary>
        public T GetConfigValue<T>(string parameterName)
        {
            return this.GetConfigValue<T>(Config, Config, parameterName);
        }
        
        /// <summary>
        /// Returns parameter value or default value if not found.
        /// </summary>
        public T GetConfigValue<T>(string sectionName, string parameterName)
        {
            return this.GetConfigValue<T>(Config, sectionName, parameterName);
        }

        /// <summary>
        /// Returns parameter value or default value if not found.
        /// </summary>
        public T GetConfigValue<T>(string configPackageName, string sectionName, string parameterName)
        {
            Validation.Verify.Operation(this.codePackageActivationContext != null, "ConfigurationReader not initialized");

            Requires.NotNullOrEmpty(configPackageName, nameof(configPackageName));
            Requires.NotNullOrEmpty(sectionName, nameof(sectionName));
            Requires.NotNullOrEmpty(parameterName, nameof(parameterName));

            return this.GetConfigValue(Config, Config, parameterName, default(T));
        }

        /// <summary>
        /// Resolve the parameter value if found.
        /// </summary>
        public bool TryGetConfigValue<T>(string sectionName, string parameterName, out T value)
        {
            return this.TryGetConfigValue(Config, sectionName, parameterName, out value);
        }

        /// <summary>
        /// Resolve the parameter value if found.
        /// </summary>
        public bool TryGetConfigValue<T>(string parameterName, out T value)
        {
            return this.TryGetConfigValue(Config, Config, parameterName, out value);
        }

        public string GetDataFilePath(string dataFolder, string dataFileName)
        {
            Validation.Verify.Operation(this.codePackageActivationContext != null, "ConfigurationReader not initialized");

            Requires.NotNullOrEmpty(dataFolder, nameof(dataFolder));
            Requires.NotNullOrEmpty(dataFileName, nameof(dataFileName));

            DataPackage dataPackage = this.codePackageActivationContext?.GetDataPackageObject(dataFolder);
            if (dataPackage?.Path != null)
            {
                return dataPackage.Path + @"\" + dataFileName;
            }

            return null;
        }

        internal static object Parse(string value, Type type)
        {
            Requires.NotNull(type, nameof(type));

            if (string.IsNullOrEmpty(value))
            {
                return null;
            }

            try
            {
                var foo = TypeDescriptor.GetConverter(type);
                if (foo != null && foo.CanConvertFrom(typeof(string)))
                {
                    return foo.ConvertFromInvariantString(value);
                }
            }
            catch (ArgumentException ex)
            {
                Instrumentation.Agent.Instance.LogException(ex, "common", "configuration");
            }
            catch (FormatException ex)
            {
                Instrumentation.Agent.Instance.LogException(ex, "common", "configuration");
            }
            catch (NotSupportedException ex)
            {
                Instrumentation.Agent.Instance.LogException(ex, "common", "configuration");
            }
            catch (Exception ex)
            {
                if (ex.InnerException != null && ex.InnerException is FormatException)
                {
                    Instrumentation.Agent.Instance.LogException(ex.InnerException, "Common", "configuration");
                }
                else
                {
                    throw;
                }
            }

            return null;
        }

        /// <summary>
        ///     Parse the string and return a typed settingValue.
        /// </summary>
        private static bool TryParse<TValue>(string valueString, out TValue value)
        {
            Requires.NotNullOrEmpty(valueString, nameof(valueString));

            object result = Parse(valueString, typeof(TValue));
            if (result == null)
            {
                value = default(TValue);
                return false;
            }

            value = (TValue)result;
            return true;
        }

        private T GetConfigValue<T>(string configPackageName,
                                    string sectionName,
                                    string parameterName,
                                    T defaultValue,
                                    bool isOptional = false)
        {
            if (this.TryGetConfigValue<T>(configPackageName, sectionName, parameterName, out var value))
            {
                return value;
            }

            if (isOptional)
            {
                return defaultValue;
            }

            throw new ArgumentException($"Failed to read the value of parameter {parameterName}");
        }

        private bool TryGetConfigValue<T>(string configPackageName,
                                    string sectionName,
                                    string parameterName,
                                    out T value)
        {
            value = default(T);

            ConfigurationPackage configPackageDesc = this.codePackageActivationContext.GetConfigurationPackageObject(configPackageName);
            ConfigurationSettings configSettings = configPackageDesc?.Settings;
            if (configSettings?.Sections != null && configSettings.Sections.Contains(sectionName))
            {
                ConfigurationSection section = configSettings.Sections[sectionName];
                if (section.Parameters?.Contains(parameterName) ?? false)
                {
                    ConfigurationProperty property = section.Parameters[parameterName];
                    string propertyValue;

                    // TODO add support for returning the secure string if the T is secure string
                    if (property.IsEncrypted)
                    {
                        try
                        {
                            SecureString decryptedValue = property.DecryptValue();
                            
                            // Hack to get the string value for now
                            propertyValue = new NetworkCredential(string.Empty, decryptedValue).Password;
                        }
                        catch (Exception ex)
                        {
                            Instrumentation.Agent.Instance.LogException(ex, "common", "configuration", $"{parameterName}");
                            return false;
                        }
                    }
                    else
                    {
                        propertyValue = property.Value;
                    }

                    return TryParse(propertyValue, out value);
                }
            }

            return false;
        }        
    }
}