const RefParser = require("json-schema-ref-parser");
const FileSystem = require("fs");
const Path = require("path");
const Oav = require("oav");
const Resolvers = require("./lib/resolvers")
const mkdirp = require("mkdirp");
const OperationFilter = require("./lib/operationFilter");
const renameKeys = require('deep-rename-keys');
const ChildProcess = require('child_process');


const argv = require("yargs")
    .usage("Usage: $0 [options]")
    .example('$0 -i ./data-plane/render/render.yml', 'Build the Render Service spec with all operations.')
    .example('$0 -i ./data-plane/render/render.yml -p', 'Build the Render Service spec with only published operations.')
    .options({
        "input": {
            alias: "i",
            describe: 'The input schema to build.',
            denandOption: true,
            string: true
        },
        "publish": {
            alias: "p",
            describe: "Whether to only include published operations.",
            boolean: true,
            default: false
        }
    })
    .help('h')
    .argv;

const inputDir = Path.dirname(argv.input);
const tmpOutputPath = Path.resolve(inputDir + "/" + Path.basename(argv.input, ".yml") + ".tmp.json");
const outputPath = Path.resolve(inputDir + "/" + Path.basename(argv.input, ".yml") + ".json");


const buildSpec = new Promise((resolve, reject) => { resolve(); });

buildSpec

    .then(() => {
        console.log("\nBundling Spec")
    }).then(() => {
        console.log("-> Resolving Spec's References");
        return RefParser.bundle(argv.input, {
            resolve: {
                examples: Resolvers.exampleResolver
            }
        });
    }).then((spec) => {
        console.log("-> Replacing Temporary '%ref-placeholder' keys with '$ref'");
        return renameKeys(spec, (key, value) => {
            return key === "$ref-placeholder" ? "$ref" : key;
        });
    })

    .then((spec) => {
        console.log("\nFiltering Spec");
        return spec;
    }).then((spec) => {
        if (!argv.publish) {
            return spec;
        }

        console.log("-> Removing Non Published Operations");
        const operationFilter = new OperationFilter("x-publish", true);
        return operationFilter.filter(spec);
    })

    .then((spec) => {
        console.log("\nCreating Output Directory and Output File");
        mkdirp(inputDir, (err) => { console.error(err) });
        FileSystem.writeFileSync(tmpOutputPath, JSON.stringify(spec, null, 2));
    })

    .then(() => {
        console.log("\nValidating Output");
    }).then(() => {
        console.log("-> OAV Validating Spec");
        return Oav.validateSpec(tmpOutputPath);
    }).then(() => {
        console.log("-> OAV Validating Spec Examples");
        return Oav.validateExamples(tmpOutputPath);
    }).then(() => {
        console.log("-> Autorest Validating Spec and Examples");
        ChildProcess.execSync("autorest --input-file=" + tmpOutputPath + " --model-validator --semantic-validator");
    })

    .then(() => {
        FileSystem.copyFileSync(tmpOutputPath, outputPath);
        FileSystem.unlinkSync(tmpOutputPath);
        process.exit(0);
    }).catch((error) => {
        console.error("\n" + error);
        FileSystem.unlinkSync(tmpOutputPath);
        process.exit(1);
    });
