﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.Geography
{
    using System;
    using System.Collections.Generic;
    using System.Runtime.Serialization;

    /// <summary>Earth WGS84 coordinates</summary>
    [DataContract]
    public struct Coordinates : IComparable<Coordinates>, IEquatable<Coordinates>
    {
        private static readonly char[] _sSeparators = { ',', ' ', '|' };
        private static readonly string _sSeparator = ":";
        private byte _notNull;
        private double _lat;
        private double _lon;

        /// <summary>Constructor</summary>
        /// <param name="lat">latitude (or Y)</param>
        /// <param name="lon">longitude (or X)</param>
        public Coordinates(double lat, double lon)
        {
            ValidateLatitude(lat);
            ValidateLongitude(lon);

            _lat = lat;
            _lon = lon;
            _notNull = 1;
        }

        /// <summary>Construct by parsing a lat/lon string</summary>
        /// <remarks>If the string can't be parsed, the constructed Coordinates will be IsNull=TRUE</remarks>
        public Coordinates(string lat, string lon)
        {
            _lat = _lon = 0;
            _notNull = 0;

            if (Double.TryParse(lat, out _lat) && Double.TryParse(lon, out _lon) && AreValidCoordinates(_lat, _lon))
            {
                _notNull = 1;
            }
        }

        /// <summary>Construct by parsing a lat/lon string</summary>
        /// <param name="s">string to be parsed</param>
        /// <param name="separators">if null, default separators {',', ' ', '|'} are supplied</param>
        /// <remarks>If the string can't be parsed, the constructed Coordinates will be IsNull=TRUE</remarks>
        public Coordinates(string s, char[] separators = null)
        {
            _lat = _lon = 0;
            _notNull = 0;

            if (!string.IsNullOrEmpty(s))
            {
                if (separators == null || separators.Length == 0)
                {
                    separators = _sSeparators;
                }

                var toks = s.Split(separators, StringSplitOptions.RemoveEmptyEntries);
                if (toks.Length != 2)
                {
                    return;
                }

                if (Double.TryParse(toks[0], out _lat) && Double.TryParse(toks[1], out _lon) && AreValidCoordinates(_lat, _lon))
                {
                    _notNull = 1;
                }
            }
        }

        public Coordinates(Coordinates copy)
        {
            if (copy.IsNull)
            {
                throw new InvalidOperationException("Coordinates copy ctor: source not initialized");
            }

            _lat = copy._lat;
            _lon = copy._lon;
            _notNull = copy._notNull;
        }

        /// <summary>Latitude (or Y)</summary>
        public double Y
        {
            get { return CheckAndGetValue(_lat); }
            set { SetLatitude(value); }
        }

        /// <summary>Longitude (or X)</summary>
        public double X
        {
            get { return CheckAndGetValue(_lon); }
            set { SetLongitude(value); }
        }

        /// <summary>Test whether the "get" accessors can be invoked or not</summary>
        public bool IsNull => (_notNull == 0);

        /// <summary>Latitude (or Y)</summary>
        [DataMember]
        public double Latitude
        {
            get { return CheckAndGetValue(_lat); }
            set { SetLatitude(value); }
        }

        /// <summary>Longitude (or X)</summary>
        [DataMember]
        public double Longitude
        {
            get { return CheckAndGetValue(_lon); }
            set { SetLongitude(value); }
        }

        /// <summary>Overload operator==</summary>
        public static bool operator ==(Coordinates a, Coordinates b) => a.Equals(b);

        /// <summary>Overload operator!=</summary>
        public static bool operator !=(Coordinates a, Coordinates b) => !a.Equals(b);

        /// <summary>Test and throw if out of range</summary>
        public static void ValidateLatitude(double lat)
        {
            if ((lat < -90.0 || lat > 90.0))
            {
                throw new ArgumentOutOfRangeException($"Invalid latitude: {lat}");
            }
        }

        public static List<Coordinates> CreateListOfCoordinates(string cString, string separator = null)
        {
            List<Coordinates> cList = new List<Coordinates>();
            if (!string.IsNullOrEmpty(cString))
            {
                if (string.IsNullOrEmpty(separator))
                {
                    separator = _sSeparator;
                }

                string[] cArray = cString.Split(separator.ToCharArray(), StringSplitOptions.RemoveEmptyEntries);
                for (int i = 0; i < cArray.Length; ++i)
                {
                    Coordinates coordinates = new Coordinates(cArray[i]);
                    if (coordinates.IsNull)
                    {
                        throw new InvalidOperationException("Coordinates could not be initialized");
                    }
                    else
                    {
                        cList.Add(coordinates);
                    }
                }
            }
            else
            {
                throw new InvalidOperationException("Coordinates could not be initialized");
            }

            return cList;
        }

        /// <summary>Test and throw if out of range</summary>
        public static void ValidateLongitude(double lon)
        {
            if ((lon < -180.0 || lon > 180.0))
            {
                throw new ArgumentOutOfRangeException($"Invalid longitude: {lon}");
            }
        }

        /// <summary>Test and throw if out of range</summary>
        public static void ValidateCoordinates(double lat, double lon)
        {
            ValidateLatitude(lat);
            ValidateLongitude(lon);
        }

        /// <summary>Test if within range</summary>
        public static bool IsValidLatitude(double lat)
        {
            return lat >= -90.0 && lat <= 90.0;
        }

        /// <summary>Test if within range</summary>
        public static bool IsValidLongitude(double lon)
        {
            return lon >= -180.0 && lon <= 180.0;
        }

        /// <summary>Test if within range</summary>
        public static bool AreValidCoordinates(double lat, double lon)
        {
            return IsValidLatitude(lat) && IsValidLongitude(lon);
        }

        /// <summary>Used to override default values (i.e. Null == true) or set new values</summary>
        /// <param name="lat"></param>
        /// <param name="lon"></param>
        public void Set(double lat, double lon)
        {
            ValidateLatitude(lat);
            ValidateLongitude(lon);

            _lat = lat;
            _lon = lon;
            _notNull = 1;
        }

        /// <summary>get a spatially-aware hash code for these coordinates</summary>
        public string GetGeoHashCode(int precision = 12)
        {
            if (IsNull)
            {
                throw new InvalidOperationException("Cannot read for a null (uninitialized) object");
            }

            return ThirdPartyIP_GeoHash.Get(_lat, _lon, precision);
        }

        /// <summary>Override</summary>
        public override int GetHashCode() => _lat.GetHashCode() ^ _lon.GetHashCode();

        /// <summary>Override</summary>
        public override string ToString() => IsNull ? "NULL" : $"{_lat.ToString("00.000000")}, {_lon.ToString("000.000000")}";

        /// <summary>Override. *** DO NOT USE ***</summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1065:DoNotRaiseExceptionsInUnexpectedLocations")]
        public override bool Equals(object obj)
        {
            throw new InvalidOperationException("disallowed comparer");
        }

        /// <summary>IComparable</summary>
        public int CompareTo(Coordinates other)
        {
            return GetGeoHashCode().CompareTo(other.GetGeoHashCode());
        }

        /// <summary>IEquatable</summary>
        public bool Equals(Coordinates other) => (this._lat == other._lat && this._lon == other._lon);

        /// <summary>Check and set value</summary>
        private void SetLatitude(double lat)
        {
            ValidateLatitude(lat);
            _lat = lat;
        }

        /// <summary>Get accessor check for non-null</summary>
        private double CheckAndGetValue(double v)
        {
            if (IsNull)
            {
                throw new InvalidOperationException("Cannot read for a null (uninitialized) object");
            }

            return v;
        }

        /// <summary>Check and set value</summary>
        private void SetLongitude(double lon)
        {
            ValidateLongitude(lon);
            _lon = lon;
        }
    }
}
