﻿// Copyright 2013 The Noda Time Authors. All rights reserved.
// Use of this source code is governed by the Apache License 2.0,
// as found in the LICENSE.txt file.

using System;
using JetBrains.Annotations;
using NodaTime.Utility;

namespace NodaTime.TimeZones
{
    /// <summary>
    /// A country represented within an entry in the "zone1970.tab" file, with the English name
    /// mapped from the "iso3166.tab" file.
    /// </summary>
    public sealed class Country : IEquatable<Country>, Microsoft.Azure.LocationServices.TimeZoneDb.ICountry
    {
        /// <summary>
        /// Gets the English name of the country.
        /// </summary>
        /// <value>The English name of the country.</value>
        public string Name { get; set; }

        /// <summary>
        /// Gets the ISO-3166 2-letter country code for the country.
        /// </summary>
        /// <value>The ISO-3166 2-letter country code for the country.</value>
        public string Code { get; set; }

        /// <summary>
        /// Constructs a new country from its name and ISO-3166 2-letter code.
        /// </summary>
        /// <param name="name">Country name; must not be empty.</param>
        /// <param name="code">2-letter code</param>
        internal Country([NotNull] string name, [NotNull] string code)
        {
            Name = Preconditions.CheckNotNull(name, nameof(name));
            Code = Preconditions.CheckNotNull(code, nameof(code));
            Preconditions.CheckArgument(Name.Length > 0, nameof(name), "Country name cannot be empty");
            Preconditions.CheckArgument(Code.Length == 2, nameof(code), "Country code must be two characters");
        }

        /// <summary>
        /// Compares countries for equality, by name and code.
        /// </summary>
        /// <param name="other">The country to compare with this one.</param>
        /// <returns><c>true</c> if the given country has the same name and code as this one; <c>false</c> otherwise.</returns>
        public bool Equals(Country other) => other != null && other.Code == Code && other.Name == Name;

        /// <summary>
        /// Compares countries for equality, by name and code.
        /// </summary>
        /// <param name="obj">The object to compare this one with.</param>
        /// <returns><c>true</c> if the given object is a country with the same name and code as this one; <c>false</c> otherwise.</returns>
        public override bool Equals(object obj) => Equals(obj as Country);

        /// <summary>
        /// Returns a hash code for this country.
        /// </summary>
        /// <returns>A hash code for this country.</returns>
        public override int GetHashCode() => HashCodeHelper.Hash(Name, Code);

        /// <summary>
        /// Returns a string representation of this country, including the code and name.
        /// </summary>
        /// <returns>A string representation of this country.</returns>
        public override string ToString() => $"{Code} ({Name})";
    }
}
