// Copyright 2012 The Noda Time Authors. All rights reserved.
// Use of this source code is governed by the Apache License 2.0,
// as found in the LICENSE.txt file.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using JetBrains.Annotations;
using NodaTime.TimeZones.Cldr;
using NodaTime.Utility;

namespace NodaTime.TimeZones.IO
{
    /// <summary>
    /// Provides the raw data exposed by <see cref="TzdbDateTimeZoneSource"/>.
    /// </summary>
    public sealed class TzdbStreamData
    {
        private static readonly Dictionary<TzdbStreamFieldId, Action<Builder, TzdbStreamField>> FieldHanders =
            new Dictionary<TzdbStreamFieldId, Action<Builder, TzdbStreamField>>
        {
            [TzdbStreamFieldId.StringPool] = (builder, field) => builder.HandleStringPoolField(field),
            [TzdbStreamFieldId.TimeZone] = (builder, field) => builder.HandleZoneField(field),
            [TzdbStreamFieldId.TzdbIdMap] = (builder, field) => builder.HandleTzdbIdMapField(field),
            [TzdbStreamFieldId.TzdbVersion] = (builder, field) => builder.HandleTzdbVersionField(field),
            [TzdbStreamFieldId.CldrSupplementalWindowsZones] = (builder, field) => builder.HandleSupplementalWindowsZonesField(field),
            [TzdbStreamFieldId.WindowsAdditionalStandardNameToIdMapping] = (builder, field) => builder.HandleWindowsAdditionalStandardNameToIdMappingField(field),
            [TzdbStreamFieldId.Zone1970Locations] = (builder, field) => builder.HandleZone1970LocationsField(field)
        };

        private const int AcceptedVersion = 0;

        private readonly IList<string> stringPool;
        private readonly IDictionary<string, TzdbStreamField> zoneFields;

        /// <summary>
        /// Returns the TZDB version string.
        /// </summary>
        [NotNull] public string TzdbVersion { get; }

        /// <summary>
        /// Returns the TZDB ID dictionary (alias to canonical ID). This needn't be read-only; it won't be exposed directly.
        /// For example, the key "Europe/Jersey" would be associated with the value "Europe/London
        /// </summary>
        [NotNull] public IDictionary<string, string> TzdbIdMap { get; }

        /// <summary>
        /// Gets a lookup from canonical time zone ID (e.g. "Europe/London") to a group of aliases for that time zone
        /// (e.g. {"Europe/Belfast", "Europe/Guernsey", "Europe/Jersey", "Europe/Isle_of_Man", "GB", "GB-Eire"}).
        /// </summary>
        /// <remarks>
        /// The group of values for a key never contains the canonical ID, only aliases. Any time zone
        /// ID which is itself an alias or has no aliases linking to it will not be present in the lookup.
        /// The aliases within a group are returned in alphabetical (ordinal) order.
        /// </remarks>
        /// <value>A lookup from canonical ID to the aliases of that ID.</value>
        public ILookup<string, string> Aliases { get; }

        /// <summary>
        /// Returns the Windows mapping dictionary. (As the type is immutable, it can be exposed directly to callers.)
        /// </summary>
        [NotNull] public WindowsZones WindowsMapping { get; }

        /// <summary>
        /// Returns the "zone 1970" locations for the source, or null if no such location data is available.
        /// This needn't be a read-only collection; it won't be exposed directly.
        /// </summary>
        public IList<TzdbZone1970Location> Zone1970Locations { get; }

        /// <summary>Additional mappings from Windows standard name to TZDB ID. This never returns null</summary>
        public IDictionary<string, string> WindowsAdditionalStandardNameToIdMapping { get; }

        /// <summary>
        /// Creates the <see cref="DateTimeZone"/> for the given canonical ID, which will definitely
        /// be one of the values of the TzdbAliases dictionary.
        /// </summary>
        /// <param name="id">ID for the returned zone, which may be an alias.</param>
        /// <param name="canonicalId">Canonical ID for zone data</param>
        public DateTimeZone CreateZone([NotNull] string id, [NotNull] string canonicalId)
        {
            Preconditions.CheckNotNull(id, nameof(id));
            Preconditions.CheckNotNull(canonicalId, nameof(canonicalId));
            using (var stream = zoneFields[canonicalId].CreateStream())
            {
                var reader = new DateTimeZoneReader(stream, stringPool);
                // Skip over the ID before the zone data itself
                reader.ReadString();
                var type = (DateTimeZoneWriter.DateTimeZoneType)reader.ReadByte();
                switch (type)
                {
                    case DateTimeZoneWriter.DateTimeZoneType.Fixed:
                        return FixedDateTimeZone.Read(reader, id);
                    case DateTimeZoneWriter.DateTimeZoneType.Precalculated:
                        return CachedDateTimeZone.ForZone(PrecalculatedDateTimeZone.Read(reader, id));
                    default:
                        throw new InvalidNodaDataException("Unknown time zone type " + type);
                }
            }
        }

        private TzdbStreamData(Builder builder)
        {
            stringPool = CheckNotNull(builder.stringPool, "string pool");
            TzdbIdMap = CheckNotNull(builder.tzdbIdMap, "TZDB alias map");
            TzdbVersion = CheckNotNull(builder.tzdbVersion, "TZDB version");
            WindowsMapping = CheckNotNull(builder.windowsMapping, "CLDR Supplemental Windows Zones");
            zoneFields = builder.zoneFields;
#if DEPRECATED
            ZoneLocations = builder.zoneLocations;
#endif
            Zone1970Locations = builder.zone1970Locations;

            // Add in the canonical IDs as mappings to themselves.
            foreach (var id in zoneFields.Keys)
            {
                TzdbIdMap[id] = id;
            }

            WindowsAdditionalStandardNameToIdMapping = CheckNotNull(builder.windowsAdditionalStandardNameToIdMapping,
                "Windows additional standard name to ID mapping");

            Aliases = TzdbIdMap
                .Where(pair => pair.Key != pair.Value)
                .OrderBy(pair => pair.Key, StringComparer.Ordinal)
                .ToLookup(pair => pair.Value, pair => pair.Key);
        }

        // Like Preconditions.CheckNotNull, but specifically for incomplete data.
        private static T CheckNotNull<T>(T input, string name) where T : class
        {
            if (input == null)
            {
                throw new InvalidNodaDataException("Incomplete TZDB data. Missing field: " + name);
            }
            return input;
        }

        public static TzdbStreamData OpenFile([NotNull] string nzdFile)
        {
            if(!File.Exists(nzdFile))
            {
                throw new ArgumentException($"TzdbStreamData.Open(). File not found: {nzdFile}");
            }

            TzdbStreamData result = null;
            using (var stream = new FileStream(nzdFile, FileMode.Open))
            {
                Preconditions.CheckNotNull(stream, nameof(stream));
                int version = new BinaryReader(stream).ReadInt32();
                if (version != AcceptedVersion)
                {
                    throw new InvalidNodaDataException("Unable to read stream with version " + version);
                }
                Builder builder = new Builder();
                foreach (var field in TzdbStreamField.ReadFields(stream))
                {
                    // Only handle fields we know about
                    Action<Builder, TzdbStreamField> handler;
                    if (FieldHanders.TryGetValue(field.Id, out handler))
                    {
                        handler(builder, field);
                    }
                }
                result = new TzdbStreamData(builder);
            }
            return result;
        }

        /// <summary>
        /// Validates that the data within this source is consistent with itself.
        /// </summary>
        /// <remarks>
        /// Source data is not validated automatically when it's loaded, but any source
        /// loaded from data produced by <c>NodaTime.TzdbCompiler</c> (including the data shipped with Noda Time)
        /// will already have been validated via this method when it was originally produced. This method should
        /// only normally be called explicitly if you have data from a source you're unsure of.
        /// </remarks>
        /// <exception cref="InvalidNodaDataException">The source data is invalid. The source may not function
        /// correctly.</exception>
        internal IEnumerable<InvalidNodaDataException> Validate()
        {
            // Check that each entry has a canonical value. (Every mapping x to y
            // should be such that y maps to itself.)
            foreach (var entry in TzdbIdMap)
            {
                string canonical;
                if (!TzdbIdMap.TryGetValue(entry.Value, out canonical))
                {
                    yield return new InvalidNodaDataException("Mapping for entry {entry.Key} ({entry.Value}) is missing");
                }
                if (entry.Value != canonical)
                {
                    yield return new InvalidNodaDataException("Mapping for entry {entry.Key} ({entry.Value}) is not canonical ({entry.Value} maps to {canonical}");
                }
            }

            // Check that every Windows mapping has a primary territory
            foreach (var mapZone in WindowsMapping.MapZones)
            {
                // Simplest way of checking is to find the primary mapping...
                if (!WindowsMapping.PrimaryMapping.ContainsKey(mapZone.WindowsId))
                {
                    yield return new InvalidNodaDataException($"Windows mapping for standard ID {mapZone.WindowsId} has no primary territory");
                }
            }

            // Check that each Windows mapping has a known canonical ID.
            foreach (var mapZone in WindowsMapping.MapZones)
            {
                foreach (var id in mapZone.IanaIds)
                {
                    if (!TzdbIdMap.ContainsKey(id))
                    {
                        yield return new InvalidNodaDataException($"Windows mapping uses canonical ID {id} which is missing");
                    }
                }
            }

            // Check that each additional Windows standard name mapping has a known canonical ID.
            var additionalMappings = WindowsAdditionalStandardNameToIdMapping;
            if (additionalMappings != null)
            {
                foreach (var id in additionalMappings.Values)
                {
                    if (!TzdbIdMap.ContainsKey(id))
                    {
                        yield return new InvalidNodaDataException($"Windows additional standard name mapping uses canonical ID {id} which is missing");
                    }
                }
            }

            // Check that each zone location has a valid zone ID
            if (Zone1970Locations != null)
            {
                foreach (var location in Zone1970Locations)
                {
                    if (!TzdbIdMap.ContainsKey(location.ZoneId))
                    {
                        yield return new InvalidNodaDataException($"Zone 1970 location {location.Countries[0].Name} uses zone ID {location.ZoneId} which is missing");
                    }
                }
            }
        }

        /// <summary>Mutable builder class used during parsing</summary>
        private class Builder
        {
            internal IList<string> stringPool;
            internal string tzdbVersion;
            internal IDictionary<string, string> tzdbIdMap;
            internal IList<TzdbZoneLocation> zoneLocations = null;
            internal IList<TzdbZone1970Location> zone1970Locations = null;
            internal WindowsZones windowsMapping;
            internal readonly IDictionary<string, TzdbStreamField> zoneFields = new Dictionary<string, TzdbStreamField>();
            internal IDictionary<string, string> windowsAdditionalStandardNameToIdMapping;

            internal void HandleStringPoolField(TzdbStreamField field)
            {
                CheckSingleField(field, stringPool);
                using (var stream = field.CreateStream())
                {
                    var reader = new DateTimeZoneReader(stream, null);
                    int count = reader.ReadCount();
                    stringPool = new string[count];
                    for (int i = 0; i < count; i++)
                    {
                        stringPool[i] = reader.ReadString();
                    }
                }
            }

            internal void HandleZoneField(TzdbStreamField field)
            {
                CheckStringPoolPresence(field);
                // Just read the ID from the zone - we don't parse the data yet.
                // (We could, but we might as well be lazy.)
                using (var stream = field.CreateStream())
                {
                    var reader = new DateTimeZoneReader(stream, stringPool);
                    string id = reader.ReadString();
                    if (zoneFields.ContainsKey(id))
                    {
                        throw new InvalidNodaDataException("Multiple definitions for zone " + id);
                    }
                    zoneFields[id] = field;
                }
            }

            internal void HandleTzdbVersionField(TzdbStreamField field)
            {
                CheckSingleField(field, tzdbVersion);
                tzdbVersion = field.ExtractSingleValue(reader => reader.ReadString(), null);
            }

            internal void HandleTzdbIdMapField(TzdbStreamField field)
            {
                CheckSingleField(field, tzdbIdMap);
                tzdbIdMap = field.ExtractSingleValue(reader => reader.ReadDictionary(), stringPool);
            }

            internal void HandleSupplementalWindowsZonesField(TzdbStreamField field)
            {
                CheckSingleField(field, windowsMapping);
                windowsMapping = field.ExtractSingleValue(WindowsZones.Read, stringPool);
            }

            internal void HandleWindowsAdditionalStandardNameToIdMappingField(TzdbStreamField field)
            {
                // Even on the non-portable build, we still read the data: the cost is minimal, and it makes it much simpler to validate.
                if (windowsMapping == null)
                {
                    throw new InvalidNodaDataException("Field " + field.Id + " without earlier Windows mapping field");
                }
                windowsAdditionalStandardNameToIdMapping = field.ExtractSingleValue(reader => reader.ReadDictionary(), stringPool);
            }

            internal void HandleZoneLocationsField(TzdbStreamField field)
            {
                CheckSingleField(field, zoneLocations);
                CheckStringPoolPresence(field);
                using (var stream = field.CreateStream())
                {
                    var reader = new DateTimeZoneReader(stream, stringPool);
                    var count = reader.ReadCount();
                    var array = new TzdbZoneLocation[count];
                    for (int i = 0; i < count; i++)
                    {
                        array[i] = TzdbZoneLocation.Read(reader);
                    }
                    zoneLocations = array;
                }
            }

            internal void HandleZone1970LocationsField(TzdbStreamField field)
            {
                CheckSingleField(field, zone1970Locations);
                CheckStringPoolPresence(field);
                using (var stream = field.CreateStream())
                {
                    var reader = new DateTimeZoneReader(stream, stringPool);
                    var count = reader.ReadCount();
                    var array = new TzdbZone1970Location[count];
                    for (int i = 0; i < count; i++)
                    {
                        array[i] = TzdbZone1970Location.Read(reader);
                    }
                    zone1970Locations = array;
                }
            }

            private void CheckSingleField(TzdbStreamField field, object expectedNullField)
            {
                if (expectedNullField != null)
                {
                    throw new InvalidNodaDataException("Multiple fields of ID " + field.Id);
                }
            }

            private void CheckStringPoolPresence(TzdbStreamField field)
            {
                if (stringPool == null)
                {
                    throw new InvalidNodaDataException("String pool must be present before field " + field.Id);
                }
            }
        }
    }
}
