// Copyright 2009 The Noda Time Authors. All rights reserved.
// Use of this source code is governed by the Apache License 2.0,
// as found in the LICENSE.txt file.

using System;
using System.IO;
using System.Linq;

using NodaTime.TimeZones.IO;
using NodaTime.TimeZones.Cldr;
using NodaTime.TzdbCompiler.Tzdb;

using Microsoft.Azure.LocationServices.Common;

namespace NodaTime.TzdbCompiler
{
    /// <summary>
    /// Main entry point for the time zone information compiler. In theory we could support
    /// multiple sources and formats but currently we only support one:
    /// http://www.twinsun.com/tz/tz-link.htm. This system refers to it as TZDB.
    /// This also requires a windowsZone.xml file from the Unicode CLDR repository, to
    /// map Windows time zone names to TZDB IDs.
    /// </summary>
    public static class Compiler
    {
        public static readonly string AzureNzdFileName = "AzureNodaTime.nzd";

        /// <summary>Run the compiler</summary>
        /// <param name="tzdbSourceFolder"></param>
        /// <param name="cldrPathOrFile"></param>
        /// <param name="outFilePath"></param>
        /// <returns>nzd file path and name</returns>
        public static string Run(Logger logger, string tzdbSourceFolder, string cldrPathOrFile, string outFilePath, out NodaTime.Utility.InvalidNodaDataException[] errors)
        {
            var tzdbCompiler = new TzdbZoneInfoCompiler(logger);
            var tzdb = tzdbCompiler.Compile(tzdbSourceFolder);
            tzdb.LogCounts(logger);

            // @todo davsta: this may be a better way to get intervals...
            //if (options.ZoneId != null)
            //{
            //    tzdb.GenerateDateTimeZone(options.ZoneId);
            //    return 0;
            //}

            var windowsZones = LoadWindowsZones(logger, cldrPathOrFile, tzdb.Version);
            //if (options.WindowsOverride != null)
            //{
            //    var overrideFile = CldrWindowsZonesParser.Parse(options.WindowsOverride);
            //    windowsZones = MergeWindowsZones(windowsZones, overrideFile);
            //}
            LogWindowsZonesSummary(logger, windowsZones);

            if (!Directory.Exists(outFilePath))
                Directory.CreateDirectory(outFilePath);

            var outFile = Path.Combine(outFilePath, AzureNzdFileName);
            using (var stream = File.Create(outFile))
            {
                var writer = new TzdbStreamWriter();
                writer.Write(tzdb, windowsZones, PclSupport.StandardNameToIdMap, stream);
            }

            var db = TzdbStreamData.OpenFile(outFile);
            errors = db.Validate().ToArray();

            return outFile;
        }

        /// <summary>
        /// Loads the best windows zones file based on the options. If the WindowsMapping option is
        /// just a straight file, that's used. If it's a directory, this method loads all the XML files
        /// in the directory (expecting them all to be mapping files) and then picks the best one based
        /// on the version of TZDB we're targeting - basically, the most recent one before or equal to the
        /// target version.
        /// </summary>
        private static WindowsZones LoadWindowsZones(Logger logger, string cldrPathOrFile, string targetTzdbVersion)
        {
            if (File.Exists(cldrPathOrFile))
            {
                return CldrWindowsZonesParser.Parse(cldrPathOrFile);
            }
            if (!Directory.Exists(cldrPathOrFile))
            {
                throw new Exception($"{cldrPathOrFile} does not exist as either a file or a directory");
            }
            var xmlFiles = Directory.GetFiles(cldrPathOrFile, "*.xml");
            if (xmlFiles.Length == 0)
            {
                throw new Exception($"{cldrPathOrFile} does not contain any XML files");
            }
            var allFiles = xmlFiles
                .Select(file => CldrWindowsZonesParser.Parse(file))
                .OrderByDescending(zones => zones.TzdbVersion)
                .ToList();

            var versions = string.Join(", ", allFiles.Select(z => z.TzdbVersion).ToArray());

            var bestFile = allFiles
                .Where(zones => StringComparer.Ordinal.Compare(zones.TzdbVersion, targetTzdbVersion) <= 0)
                .FirstOrDefault();

            if (bestFile == null)
            {
                throw new Exception($"No zones files suitable for version {targetTzdbVersion}. Found versions targeting: [{versions}]");
            }
            logger.Log(LogType.Info, $"Picked Windows Zones with TZDB version {bestFile.TzdbVersion} out of [{versions}] as best match for {targetTzdbVersion}");
            return bestFile;
        }

        private static void LogWindowsZonesSummary(Logger logger, WindowsZones windowsZones)
        {
            logger.Log(LogType.Info, $"Windows Zones: Version = {windowsZones.Version}");
            logger.Log(LogType.Info, $"Windows Zones: TZDB version = {windowsZones.TzdbVersion}");
            logger.Log(LogType.Info, $"Windows Zones: Windows version = {windowsZones.WindowsVersion}");
            logger.Log(LogType.Info, $"Windows Zones: MapZones = {windowsZones.MapZones.Count}");
            logger.Log(LogType.Info, $"Windows Zones: PrimaryMappings = {windowsZones.PrimaryMapping.Count}");
        }

        /// <summary>
        /// Merge two WindowsZones objects together. The result has versions present in override,
        /// but falling back to the original for versions absent in the override. The set of MapZones
        /// in the result is the union of those in the original and override, but any ID/Territory
        /// pair present in both results in the override taking priority, unless the override has an
        /// empty "type" entry, in which case the entry is removed entirely.
        /// 
        /// While this method could reasonably be in WindowsZones class, it's only needed in
        /// TzdbCompiler - and here is as good a place as any.
        /// 
        /// The resulting MapZones will be ordered by Windows ID followed by territory.
        /// </summary>
        /// <param name="windowsZones">The original WindowsZones</param>
        /// <param name="overrideFile">The WindowsZones to override entries in the original</param>
        /// <returns>A merged zones object.</returns>
        internal static WindowsZones MergeWindowsZones(WindowsZones originalZones, WindowsZones overrideZones)
        {
            var version = overrideZones.Version == "" ? originalZones.Version : overrideZones.Version;
            var tzdbVersion = overrideZones.TzdbVersion == "" ? originalZones.TzdbVersion : overrideZones.TzdbVersion;
            var windowsVersion = overrideZones.WindowsVersion == "" ? originalZones.WindowsVersion : overrideZones.WindowsVersion;

            // Work everything out using dictionaries, and then sort.
            var mapZones = originalZones.MapZones.ToDictionary(mz => new { mz.WindowsId, mz.Territory });
            foreach (var overrideMapZone in overrideZones.MapZones)
            {
                var key = new { overrideMapZone.WindowsId, overrideMapZone.Territory };
                if (overrideMapZone.IanaIds.Count == 0)
                {
                    mapZones.Remove(key);
                }
                else
                {
                    mapZones[key] = overrideMapZone;
                }
            }
            var mapZoneList = mapZones
                .OrderBy(pair => pair.Key.WindowsId)
                .ThenBy(pair => pair.Key.Territory)
                .Select(pair => pair.Value)
                .ToList();
            return new WindowsZones(version, tzdbVersion, windowsVersion, mapZoneList);
        }
    }
}
