﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDb
{
    using System.Collections.Generic;
    using System.IO;
    using GeoAPI.Geometries;
    using NetTopologySuite.Geometries;
    using NetTopologySuite.Index.Quadtree;
    using NetTopologySuite.IO;

    internal class ShapeIndex
    {
        private static readonly string tzIdColName = "ianaid"; // CELA controls the header names

        private Dictionary<int, TimeZoneFeature> map;
        private Quadtree<int> index;
        private GeometryFactory factory;

        /// <summary>construct from CELA provided shape (*.shp) file(s)</summary>
        internal ShapeIndex(string dataPath)
        {
            map = new Dictionary<int, TimeZoneFeature>(500);
            index = new Quadtree<int>();
            factory = new GeometryFactory();

            int rowId = 0;
            foreach (var file in Directory.EnumerateFiles(dataPath, "*.shp"))
            {
                using (var reader = new ShapefileDataReader(file, factory))
                {
                    int tzIdColIdx = int.MinValue;
                    var header = reader.DbaseHeader;
                    for (int i = 0; i < header.NumFields; i++)
                    {
                        if (header.Fields[i].Name.ToLowerInvariant().Equals(tzIdColName))
                        {
                            tzIdColIdx = i;
                            break;
                        }
                    }

                    if (tzIdColIdx < 0)
                    {
                        throw new InvalidDataException($"Failed to find {tzIdColName} header in {file}. Check with CELA for format changes");
                    }

                    while (reader.Read())
                    {
                        var tzId = (string)reader.GetValue(tzIdColIdx);
                        if (tzId.ToLowerInvariant().Equals("uninhabited"))
                        {
                            continue;
                        }

                        ++rowId;

                        var geom = reader.Geometry;
                        var feature = new TimeZoneFeature() { Shape = geom, TzId = tzId };
                        map.Add(rowId, feature);
                        index.Insert(geom.Simplify().EnvelopeInternal, rowId);
                    }
                }
            }
        }

        internal IEnumerable<string> Find(double lat, double lon)
        {
            var point = factory.CreatePoint(new Coordinate(lon, lat));
            var indexResult = index.Query(point.EnvelopeInternal);
            foreach (var i in indexResult)
            {
                var feature = map[i];
                if (feature.Shape.Contains(point))
                {
                    yield return feature.TzId;

                    // improves perf, but need more testing to verify there are no overlapping regions
                    //break;
                }
            }
        }

        internal class TimeZoneFeature
        {
            public string TzId { get; set; }

            public IGeometry Shape { get; set; }
        }
    }
}
