﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDb
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Geography;
    using Microsoft.Azure.LocationServices.TimeZoneDbManager;
    using NodaTime.TimeZones.IO;

    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.OrderingRules", "SA1201:ElementsMustAppearInTheCorrectOrder")]
    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.OrderingRules", "SA1202:ElementsMustBeOrderedByAccess")]
    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.OrderingRules", "SA1204:StaticElementsMustAppearBeforeInstanceElements")]
    public sealed class TimeZoneDb : IDisposable, ITimeZoneDbManaged
    {
        private readonly ConfigValues _configValues;

        private bool _initialized;
        private Logger _logger;
        private DataManagementPool _dataManagementPool;

        private ShapeIndex _shapeIndex;
        private ShapeIndex _shapeIndexUpdate;

        private TzdbStreamData _nzd;
        private TzdbStreamData _nzdUpdate;

        /// <summary>CreateInstance constructor</summary>
        private TimeZoneDb(ConfigValues config, EventHandler<LogEventArgs> logEventCallback)
        {
            _configValues = config;

            _logger = new Logger("TimeZoneDb");
            _logger.LogEvent += LogEventCallback;

            this.LogEvent += logEventCallback;
        }

        /// <summary>Test data constructor without DataManagementPool and Logging</summary>
        private TimeZoneDb(string nzdPath, string shapesPath)
        {
            _nzd = TzdbStreamData.OpenFile(nzdPath);
            _shapeIndex = new ShapeIndex(shapesPath);
            _initialized = true;
        }

        public event EventHandler<LogEventArgs> LogEvent;

        #region ----- Singleton -----
        private static object _sSyncLock = new object();
        private static SemaphoreSlim _syncronizationSemaphoreSlim = new SemaphoreSlim(1);
        private static TimeZoneDb _sInstance;

        /// <summary>Create and get the singleton instance. The actual initialization takes several seconds to complete</summary>
        public static async Task<TimeZoneDb> CreateInstance(ConfigValues config, EventHandler<LogEventArgs> logEventCallback, int timeoutSeconds = 60)
        {
            if (config == null)
            {
                throw new ArgumentNullException(nameof(config));
            }

            lock (_sSyncLock)
            {
                if (_sInstance != null)
                {
                    throw new InvalidOperationException("An instance of TimeZoneDb already exists");
                }

                _sInstance = new TimeZoneDb(config, logEventCallback);
            }

            await _sInstance.InitializeAsync(timeoutSeconds);
            return _sInstance;
        }

        /// <summary>Get the instance (must have been created first)</summary>
        public static TimeZoneDb Instance
        {
            get
            {
                if (_sInstance == null)
                {
                    throw new InvalidOperationException("Instance does not exist");
                }

                if (_sInstance.Disposed)
                {
                    throw new InvalidOperationException("Instance is disposed");
                }

                return _sInstance;
            }
        }

        /// <summary>Check if the instance exists and not disposed</summary>
        public static bool IsQueriable => (_sInstance != null && !_sInstance.Disposed && _sInstance._initialized);
        #endregion // Singleton

        /// <summary>Retrofit function to support data update and compilation synchronously from a Cron job</summary>
        /// <returns>True if a new tzdb was published, false otherwise.  Logged errors must be handled by the callback function</returns>
        public static async Task<CronJobResult> CronDataUpdateCheckAsync(ConfigValues config, EventHandler<LogEventArgs> logEventCallback)
        {
            if (config == null)
            {
                throw new ArgumentNullException(nameof(config));
            }

            await _syncronizationSemaphoreSlim.WaitAsync();
            try
            {
                if (_sInstance != null)
                {
                    throw new InvalidOperationException("An instance of TimeZoneDb already exists");
                }

                _sInstance = new TimeZoneDb(config, logEventCallback);
                return await DataManagementPool.CronDataUpdateCheckAsync(config, _sInstance);
            }
            finally
            {
                if (_sInstance != null)
                {
                    _sInstance.Dispose();
                    _sInstance = null;
                }

                _syncronizationSemaphoreSlim.Release();
            }
        }

        /// <summary>Return the underlying IANA tzdb version</summary>
        public TimeZoneIanaVersion Version => new TimeZoneIanaVersion { Version = (_nzd == null) ? string.Empty : _nzd.TzdbVersion };

        /// <summary>Create a private instance of self using given input data and run a suite of public API's to verify publish suitability</summary>
        public static bool TestDatabase(string nzdPath, string shapesPath, out string ianaVersion, out IList<string> errors)
        {
            TimeZoneDb tzdb = null;
            try
            {
                tzdb = new TimeZoneDb(nzdPath, shapesPath);
            }
            catch (Exception ex)
            {
                errors = new List<string>();
                errors.Add($"Exception: {ex.Message}");
                ianaVersion = null;
                return false;
            }

            return tzdb.TestPublicApis(out ianaVersion, out errors);
        }

        /// <summary>Enumerate all IANA timezone id's</summary>
        /// <returns>IANA key and, if an alias, the primary key it is an alias of in the value field.
        /// If the value field is null, the key is primary and not an alias.</returns>
        public IEnumerable<IanaId> EnumerateTimeZoneIds()
        {
            ValidateNotDisposed();
            foreach (var kvp in _nzd.TzdbIdMap)
            {
                bool hit = _nzd.Zone1970Locations.FirstOrDefault(g => g.ZoneId.Equals(kvp.Value, StringComparison.InvariantCultureIgnoreCase)) != null;
                yield return new IanaId(kvp, hit);
            }
        }

        /// <summary>Enumerate all WIndows timezone id's</summary>
        public IList<NodaTime.TimeZones.Cldr.MapZone> EnumerateWindowsTimeZones()
        {
            ValidateNotDisposed();
            return _nzd.WindowsMapping.MapZones;
        }

        /// <summary>Given a Windows system time zone ID, return the correspondingTZDB zone ID</summary>
        public IEnumerable<IanaId> WindowsToIanaTimeZone(string windowsZone, string territory = null)
        {
            ValidateNotDisposed();

            var zones = _nzd.WindowsMapping.MapZones.Where(g =>
                            g.WindowsId.Equals(windowsZone, StringComparison.InvariantCultureIgnoreCase)
                            && (string.IsNullOrEmpty(territory) || g.Territory.Equals(territory, StringComparison.InvariantCultureIgnoreCase)));

            if (zones != null)
            {
                foreach (var zone in zones)
                {
                    foreach (var key in zone.IanaIds)
                    {
                        var value = _nzd.TzdbIdMap[key];
                        bool hit = _nzd.Zone1970Locations.FirstOrDefault(g => g.ZoneId.Equals(value, StringComparison.InvariantCultureIgnoreCase)) != null;
                        yield return new IanaId(new KeyValuePair<string, string>(key, value), hit);
                    }
                }
            }
        }

        /// <summary>Find the time zone covering the supplied location</summary>
        public TimeZoneQueryResults FindTimeZone(string timezoneId, TimeZoneQueryParams queryParams)
        {
            ValidateNotDisposed();
            try
            {
                var results = new List<TimeZoneQueryResult>(1);
                string tzId;
                if (_nzd.TzdbIdMap.TryGetValue(timezoneId, out tzId))
                {
                    results.Add(new TimeZoneQueryResult(tzId, queryParams, _nzd));
                }

                return new TimeZoneQueryResults(_nzd.TzdbVersion, queryParams, results);
            }
            catch (TimeZoneDbException tzdbEx)
            {
                return new TimeZoneQueryResults(_nzd.TzdbVersion, queryParams, tzdbEx);
            }
            catch (Exception ex)
            {
                var tzdbEx = new TimeZoneDbException("Unhandled error in FindTimeZone by id", ex);
                return new TimeZoneQueryResults(_nzd.TzdbVersion, queryParams, tzdbEx);
            }
        }

        /// <summary>Find the time zone covering the supplied location</summary>
        public TimeZoneQueryResults FindTimeZone(ref Coordinates coord, TimeZoneQueryParams queryParams)
        {
            ValidateNotDisposed();
            try
            {
                var results = new List<TimeZoneQueryResult>(1);
                foreach (var tzId in _shapeIndex.Find(coord.Latitude, coord.Longitude))
                {
                    results.Add(new TimeZoneQueryResult(tzId, queryParams, _nzd));
                }

                if (results.Count == 0)
                {
                    results.Add(new TimeZoneQueryResult(ref coord, queryParams, _nzd));
                }

                return new TimeZoneQueryResults(_nzd.TzdbVersion, queryParams, results);
            }
            catch (TimeZoneDbException tzdbEx)
            {
                return new TimeZoneQueryResults(_nzd.TzdbVersion, queryParams, tzdbEx);
            }
            catch (Exception ex)
            {
                var tzdbEx = new TimeZoneDbException("Unhandled error in FindTimeZone by coord", ex);
                return new TimeZoneQueryResults(_nzd.TzdbVersion, queryParams, tzdbEx);
            }
        }

        /// <summary>Iinitializer async CreateInstance</summary>
        private async Task InitializeAsync(int timeoutSeconds)
        {
            var start = DateTime.Now;
            if (_shapeIndex != null || _nzd != null)
            {
                throw new TimeZoneDbException("Initialize() invoked from the wrong code path", null);
            }

            if (0 != (_configValues.Mode & ExecutionMode.Static))
            {
                CurrentContainerName = "static init";
                _shapeIndex = new ShapeIndex(_configValues.TimeZoneDbFolder);
                _nzd = TzdbStreamData.OpenFile(Path.Combine(_configValues.TimeZoneDbFolder, NodaTime.TzdbCompiler.Compiler.AzureNzdFileName));
            }
            else
            {
                _dataManagementPool = await DataManagementPool.RegisterAsync(_configValues, this);

                var args = _dataManagementPool.GetCurrentDb(timeoutSeconds);
                CurrentContainerName = args.ContainerName;

                // This ensures that the database is initialized before we start reading from the disk.
                CancellationTokenSource source = new CancellationTokenSource(TimeSpan.FromSeconds(timeoutSeconds));
                while (!_dataManagementPool.DatabaseInitialized && !source.Token.IsCancellationRequested)
                {
                    await Task.Delay(100);
                }

                if (_dataManagementPool.DatabaseInitialized)
                {
                    _shapeIndex = new ShapeIndex(args.DbFolderPath);
                    _nzd = TzdbStreamData.OpenFile(Path.Combine(args.DbFolderPath, NodaTime.TzdbCompiler.Compiler.AzureNzdFileName));
                    _initialized = true;
                    _logger.Log(LogType.Info, $"TimeZoneDb initialized in {(int)((DateTime.Now - start).TotalMilliseconds)} ms");
                }
                else
                {
                    _logger.Log(LogType.Error, $"TimeZoneDb was NOT initialized within the timeout window.");
                }
            }
        }

        /// <summary>Logger LogEvent passthrough</summary>
        private void LogEventCallback(object sender, LogEventArgs e)
        {
            LogEvent?.Invoke(this, e);
        }

        /// <summary>tag</summary>
        private void ValidateNotDisposed()
        {
            if (Disposed)
            {
                throw new InvalidOperationException("TimeZoneDb: cannot invoke methods on a disposed object");
            }
        }

        #region ----- ITimeZoneDbManaged -----
        /// <summary>Return the logger instance</summary>
        public Logger GetLogger => _logger;

        /// <summary>Get the current container name.  Null if nothing loaded</summary>
        public string CurrentContainerName { get; private set; }

        /// <summary>Get the awaiting container name.  Null if the awaiting container is not loaded</summary>
        public string AwaitingContainerName { get; private set; }

        /// <summary>Notification to load a new container to a private cache</summary>
        public void LoadAwaitingContainer(DbUpdateEventArgs args)
        {
            if (IsReadyToSwapContainers(args))
            {
                _logger.Log(LogType.Warning, "awaiting containers already loaded!  Overlapping updates?");
                return;
            }

            _logger.Log(LogType.Info, $"LoadAwaitingContainer: {args.ContainerName}. Effective={args.EffectiveFrom} due in {args.DueTimeFromNow}");

            AwaitingContainerName = args.ContainerName;
            _shapeIndexUpdate = new ShapeIndex(args.DbFolderPath);
            _nzdUpdate = TzdbStreamData.OpenFile(Path.Combine(args.DbFolderPath, NodaTime.TzdbCompiler.Compiler.AzureNzdFileName));

            _logger.Log(LogType.Info, $"LoadAwaitingContainer: awaiting container {AwaitingContainerName} is loaded and ready to swap");
        }

        /// <summary>Check if load has completed and subscriber is ready to swap</summary>
        public bool IsReadyToSwapContainers(DbUpdateEventArgs args)
        {
            if (args == null)
            {
                return false;
            }

            return args.IsSwitchoverTime
                && _shapeIndexUpdate != null
                && _nzdUpdate != null
                && !string.IsNullOrEmpty(AwaitingContainerName)
                && AwaitingContainerName.Equals(args.ContainerName);
        }

        /// <summary>Notification to serve requests from the newer (awaiting) container</summary>
        public void PublishAwaitingContainer()
        {
            if (string.IsNullOrEmpty(AwaitingContainerName) || _shapeIndexUpdate == null || _nzdUpdate == null)
            {
                throw new InvalidOperationException("PublishAwaitingContainer: awaitng containers not loaded");
            }

            // @todo davsta: locking to consistently swap both sources doesn't seem necessary, but give it more thought.
            _logger.Log(LogType.Info, $"PublishAwaitingContainer: switching from {CurrentContainerName} to {AwaitingContainerName}");

            CurrentContainerName = AwaitingContainerName;
            AwaitingContainerName = null;

            _shapeIndex = _shapeIndexUpdate;
            _nzd = _nzdUpdate;

            _shapeIndexUpdate = null;
            _nzdUpdate = null;

            _logger.Log(LogType.Info, "PublishAwaitingContainer: tzdb swap completed");
        }

        /// <summary>Run a suite of public API's to verify its viability</summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Maintainability", "CA1502:AvoidExcessiveComplexity")]
        public bool TestPublicApis(out string ianaVersion, out IList<string> errors)
        {
            errors = new List<string>();
            try
            {
                ianaVersion = Version.Version;
                if (string.IsNullOrEmpty(ianaVersion))
                {
                    errors.Add("IsNullOrEmpty: ianaVersion");
                }

                int count = 0;
                foreach (var rec in EnumerateTimeZoneIds())
                {
                    if (!string.IsNullOrEmpty(rec.Id) && rec.HasZone1970Location)
                    {
                        ++count;
                    }
                }

                if (count < (350 / 2))
                {
                    errors.Add($"EnumerateTimeZoneIds: count of Zone1970Locations = {count} does not meet 50% count bar");
                }

                count = 0;
                foreach (var rec in EnumerateWindowsTimeZones())
                {
                    if (rec.IanaIds.Count > 0)
                    {
                        ++count;
                    }
                }

                if (count < (495 / 2))
                {
                    errors.Add($"EnumerateTimeZoneIds: count of Zone1970Locations = {count} does not meet 50% count bar");
                }

                count = 0;
                double dCount = 0;
                foreach (var wz in EnumerateWindowsTimeZones())
                {
                    ++dCount;
                    foreach (var tz in WindowsToIanaTimeZone(wz.WindowsId, wz.Territory))
                    {
                        ++count;
                    }
                }

                var ratio = (dCount == 0) ? 0.0 : count / dCount;
                if (ratio < 0.75)
                {
                    errors.Add($"WindowsToIanaTimeZone: mapping ratio to iana of {ratio} is below the 0.75 bar");
                }

                var queryParams = new TimeZoneQueryParams(TimeZoneQueryParams.AdditionalData.All, null, null, 5);

                //////////////////////////////////////////////////////////////////
                // search by Id
                bool aliases = false;
                bool transitions = false;
                foreach (var rec in EnumerateTimeZoneIds())
                {
                    var results = FindTimeZone(rec.Id, queryParams);
                    if (results.Error == null)
                    {
                        if (results.Count == 0)
                        {
                            errors.Add($"FindTimeZone by Id: no results searching for {rec.Id}");
                        }
                        else if (results.Count > 1)
                        {
                            errors.Add($"FindTimeZone by Id: multiple results searching for {rec.Id}");
                        }
                        else
                        {
                            if (results.TimeZones[0].Aliases != null && results.TimeZones[0].Aliases.Count > 0)
                            {
                                aliases = true;
                            }

                            if (results.TimeZones[0].TimeTransitions != null && results.TimeZones[0].TimeTransitions.Count > 0)
                            {
                                transitions = true;
                            }
                        }
                    }
                    else
                    {
                        errors.Add($"FindTimeZone by Id exception: {results.Error.Message}");
                    }
                }

                if (!aliases)
                {
                    errors.Add("FindTimeZone by Id: no results containing Aliases");
                }

                if (!transitions)
                {
                    errors.Add("FindTimeZone by Id: no results containing Transitions");
                }

                //////////////////////////////////////////////////////////////////
                // spatial search
                aliases = false;
                transitions = false;
                double increment = 5.0;
                var coords = new Coordinates(-90.0, -180.0);
                while (true)
                {
                    while (true)
                    {
                        var results = FindTimeZone(ref coords, queryParams);
                        if (results.Error == null)
                        {
                            if (results.Count == 0)
                            {
                                errors.Add($"FindTimeZone by Coordinates: no results searching at {coords}");
                            }
                            else if (results.Count > 1)
                            {
                                errors.Add($"FindTimeZone by Coordinates: multiple results searching at {coords}");
                            }
                            else
                            {
                                if (results.TimeZones[0].Aliases != null && results.TimeZones[0].Aliases.Count > 0)
                                {
                                    aliases = true;
                                }

                                if (results.TimeZones[0].TimeTransitions != null && results.TimeZones[0].TimeTransitions.Count > 0)
                                {
                                    transitions = true;
                                }
                            }
                        }
                        else
                        {
                            errors.Add($"FindTimeZone by Coordinates exception: {results.Error.Message}");
                        }

                        if (coords.Latitude + increment > 90.0)
                        {
                            break;
                        }

                        coords.Latitude += increment;
                    }

                    if (coords.Longitude + increment > 180.0)
                    {
                        break;
                    }

                    coords.Longitude += increment;
                    coords.Latitude = -90.0;
                }

                if (!aliases)
                {
                    errors.Add("FindTimeZone by Coordinates: no results containing Aliases");
                }

                if (!transitions)
                {
                    errors.Add("FindTimeZone by Coordinates: no results containing Transitions");
                }
            }
            catch (Exception ex)
            {
                errors.Add($"Exception: {ex.Message}");
                ianaVersion = null;
            }

            return errors.Count == 0;
        }

        /// <summary>Create a private instance of self using given input data and run a suite of public API's to verify publish suitability</summary>
        public bool TestNewDatabase(string nzdPath, string shapesPath, out string ianaVersion, out IList<string> errors)
        {
            return TestDatabase(nzdPath, shapesPath, out ianaVersion, out errors);
        }
        #endregion // ITimeZoneDbManaged

        #region ----- IDisposable -----
        /// <summary>Track whether Dispose has been called or not</summary>
        public bool Disposed { get; private set; }

        /// <summary>IDispose pattern, straight out of MSDN</summary>
        public void Dispose()
        {
            Dispose(true);

            // This object will be cleaned up by the Dispose method.
            // Therefore, you should call GC.SupressFinalize to take this object off the finalization queue 
            // and prevent finalization code for this object from executing a second time.
            GC.SuppressFinalize(this);
        }

        /// <summary>Dispose(bool disposing) executes in two distinct scenarios.
        /// If disposing equals true, the method has been called directly or indirectly by a user's code. 
        /// Managed and unmanaged resources can be disposed.
        /// If disposing equals false, the method has been called by the runtime from inside the finalizer
        /// and you should not reference other objects. Only unmanaged resources can be disposed.
        /// </summary>
        /// <param name="disposing"></param>
        private void Dispose(bool disposing)
        {
            // Check to see if Dispose has already been called.
            if (!Disposed)
            {
                try
                {
                    // If disposing equals true, dispose all managed and unmanaged resources.
                    if (disposing)
                    {
                        if (_dataManagementPool != null)
                        {
                            DataManagementPool.Unregister(_dataManagementPool);
                        }

                        _dataManagementPool = null;
                        _logger = null;
                    }
                }
                finally
                {
                    Disposed = true;
                }
            }
        }

        /// <summary>Use C# destructor syntax for finalization code.
        /// This destructor will run only if the Dispose method does not get called.
        /// It gives your base class the opportunity to finalize.
        /// Do not provide destructors in types derived from this class.
        /// </summary>
        ~TimeZoneDb()
        {
            Dispose(false);
        }
        #endregion // IDisposable
    }
}
