﻿// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDb
{
    using System;
    using System.Collections.Generic;
    using System.Globalization;
    using System.Linq;

    public class TimeZoneQueryParams
    {
        private const int _cTransitionsInResultsLowerBound = 100;
        private const int _cTransitionsInResultsUpperBound = 2000;
        private const int _cDefaultYears = 1;
        private const string languageCodesHeaderName = "Accept-Language";

        private static Dictionary<ParamList, HashSet<string>> _sParamTags = InitParamTags();
        private static int _sCurrentMaxAllowedTransitionsInResults = 250;

        private List<string> _iso6391LanguageCodeOptions;
        private AdditionalData _options;
        private DateTime _utcTimestamp;
        private DateTime _utcTransitionFrom;
        private int _transitionYears;

        public TimeZoneQueryParams(DateTime? stamp, List<string> iso6391LanguageCodes = null)
        {
            _options = AdditionalData.None;
            _utcTimestamp = (stamp == null) ? DateTime.UtcNow : stamp.Value;

            _utcTransitionFrom = _utcTimestamp;
            _transitionYears = _cDefaultYears;

            _iso6391LanguageCodeOptions = iso6391LanguageCodes;

            string error;
            if (!Validate(out error))
            {
                throw new ArgumentException(error);
            }
        }

        public TimeZoneQueryParams(AdditionalData opts, DateTime? stamp, DateTime? transitionFrom, int? transitionYears)
        {
            _options = opts;
            _utcTimestamp = (stamp == null) ? DateTime.UtcNow : stamp.Value;
            _utcTransitionFrom = (transitionFrom == null) ? _utcTimestamp : transitionFrom.Value;
            _transitionYears = (transitionYears == null) ? _cDefaultYears : transitionYears.Value;

            string error;
            if (!Validate(out error))
            {
                throw new ArgumentException(error);
            }
        }

        [Flags]
        public enum AdditionalData : byte
        {
            None = 0x00,
            ZoneInfo = 0x01,
            Transitions = 0x02,
            All = ZoneInfo | Transitions
        }

        private enum ParamList : byte
        {
            ResultOptions = 0,
            Timestamp,
            TransitionsFrom,
            TransitionsYearsCount,
            _max
        }

        public static int MaxAllowedTransitionsInResults
        {
            get { return _sCurrentMaxAllowedTransitionsInResults; }
            set
            {
                if (value < _cTransitionsInResultsLowerBound || value > _cTransitionsInResultsUpperBound)
                {
                    throw new ArgumentException($"MaxAllowedTransitionsInResults range is {_cTransitionsInResultsLowerBound} to {_cTransitionsInResultsUpperBound}");
                }

                _sCurrentMaxAllowedTransitionsInResults = value;
            }
        }

        public List<string> ISO6391LanguageCodeOptions => _iso6391LanguageCodeOptions;

        public AdditionalData Options => _options;

        public DateTime UtcTimestamp => _utcTimestamp;

        public DateTime UtcTransitionFrom => _utcTransitionFrom;

        public DateTime UtcTransitionTo => _utcTransitionFrom.AddYears(_transitionYears);

        public int TransitionYearsCount => _transitionYears;

        public int MaxTransitionsInResults => _sCurrentMaxAllowedTransitionsInResults;

        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Maintainability", "CA1502:AvoidExcessiveComplexity")]
        public static bool CreateFromRestParams(Dictionary<string, List<string>> requestParams, Dictionary<string, List<string>> headerParams, out TimeZoneQueryParams tzdbParams, out string errorDetails)
        {
            tzdbParams = new TimeZoneQueryParams(null);
            if (headerParams != null && headerParams.Count != 0)
            {
                var lowerCaseHeaderParams = headerParams.ToDictionary(h => h.Key.ToLower(), h => h.Value);
                if (lowerCaseHeaderParams.ContainsKey(languageCodesHeaderName.ToLower()))
                {
                    tzdbParams._iso6391LanguageCodeOptions = headerParams[languageCodesHeaderName];
                }
            }

            // If Accept-Language is not provided, default to English
            if (tzdbParams._iso6391LanguageCodeOptions == null)
            {
                tzdbParams._iso6391LanguageCodeOptions = new List<string>() { "en" };
            }

            if (requestParams == null || requestParams.Count == 0)
            {
                errorDetails = null;
                return true;
            }

            var checkedParams = new bool[(int)ParamList._max];
            foreach (var kvpRest in requestParams)
            {
                foreach (var kvpLkp in _sParamTags)
                {
                    if (!kvpLkp.Value.Contains(kvpRest.Key.ToLowerInvariant()))
                    {
                        continue; //quietly ignore parm keys we do not understand
                    }

                    var checkedParamsIdx = (int)kvpLkp.Key;
                    if (checkedParams[checkedParamsIdx])
                    {
                        errorDetails = $"'{kvpLkp.Key}' parameter submitted more than once";
                        return false;
                    }

                    switch (kvpLkp.Key)
                    {
                        case ParamList.ResultOptions:
                            switch (kvpRest.Value.First().ToLowerInvariant())
                            {
                                case "none":
                                case "n":
                                case "0x00":
                                case "0":
                                    tzdbParams._options = AdditionalData.None;
                                    break;

                                case "zoneinfo":
                                case "zi":
                                case "0x01":
                                case "1":
                                    tzdbParams._options = AdditionalData.ZoneInfo;
                                    break;

                                case "transitions":
                                case "tr":
                                case "0x02":
                                case "2":
                                    tzdbParams._options = AdditionalData.Transitions;
                                    break;

                                case "all":
                                case "a":
                                    tzdbParams._options = AdditionalData.All;
                                    break;

                                default:
                                    errorDetails = $"'{kvpRest.Value}' is not a valid '{kvpRest.Key}' value";
                                    return false;
                            }

                            break;

                        case ParamList.Timestamp:
                            if (!ParseUtcTime(kvpRest.Value.First(), out tzdbParams._utcTimestamp))
                            {
                                errorDetails = $"'{kvpRest.Value.First()}' is not a valid UTC date time for '{kvpRest.Key}'";
                                return false;
                            }

                            break;

                        case ParamList.TransitionsFrom:
                            if (!ParseUtcTime(kvpRest.Value.First(), out tzdbParams._utcTransitionFrom))
                            {
                                errorDetails = $"'{kvpRest.Value.First()}' is not a valid UTC date time for '{kvpRest.Key}'";
                                return false;
                            }

                            break;

                        case ParamList.TransitionsYearsCount:
                            if (!int.TryParse(kvpRest.Value.First(), out tzdbParams._transitionYears)
                                || tzdbParams._transitionYears < 1)
                            {
                                errorDetails = $"'{kvpRest.Value.First()}' is not a valid value for '{kvpRest.Key}'";
                                return false;
                            }

                            break;

                        default:
                            throw new InvalidOperationException("Internal error: unhandled ParamList case in switch statement.");
                    }

                    checkedParams[checkedParamsIdx] = true;
                    break;
                }
            }

            return tzdbParams.Validate(out errorDetails);
        }

        public bool Validate(out string errorDetails)
        {
            errorDetails = null;

            if (UtcTimestamp.Kind != DateTimeKind.Utc)
            {
                errorDetails = "UtcTimestamp is not DateTimeKind.Utc";
                return false;
            }

            if (UtcTransitionFrom.Kind != DateTimeKind.Utc)
            {
                errorDetails = "UtcTransitionFrom is not DateTimeKind.Utc";
                return false;
            }

            if (UtcTransitionTo.Kind != DateTimeKind.Utc)
            {
                errorDetails = "UtcTransitionTo is not DateTimeKind.Utc";
                return false;
            }

            if (0 != (_options & AdditionalData.Transitions)
                && TransitionYearsCount < 1)
            {
                errorDetails = "Count of transition years must be greater than zero";
                return false;
            }

            return true;
        }

        private static Dictionary<ParamList, HashSet<string>> InitParamTags()
        {
            var dic = new Dictionary<ParamList, HashSet<string>>();

            for (int i = 0; i < (int)ParamList._max; i++)
            {
                dic.Add((ParamList)i, new HashSet<string>());
            }

            dic[ParamList.ResultOptions].Add("options".ToLowerInvariant());
            dic[ParamList.ResultOptions].Add("o".ToLowerInvariant());

            dic[ParamList.Timestamp].Add("timestamp".ToLowerInvariant());
            dic[ParamList.Timestamp].Add("stamp".ToLowerInvariant());
            dic[ParamList.Timestamp].Add("s".ToLowerInvariant());

            dic[ParamList.TransitionsFrom].Add("transitionsFrom".ToLowerInvariant());
            dic[ParamList.TransitionsFrom].Add("tf".ToLowerInvariant());

            dic[ParamList.TransitionsYearsCount].Add("transitionsYears".ToLowerInvariant());
            dic[ParamList.TransitionsYearsCount].Add("ty".ToLowerInvariant());

            return dic;
        }

        private static bool ParseUtcTime(string val, out DateTime dt)
        {
            if (!string.IsNullOrEmpty(val)
                && DateTime.TryParse(val, CultureInfo.InvariantCulture, DateTimeStyles.AdjustToUniversal, out dt)
                && dt.Kind == DateTimeKind.Utc)
            {
                return true;
            }

            dt = DateTime.MinValue;
            return false;
        }
    }
}
