﻿// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDb
{
    using System;
    using System.Collections.Generic;
    using System.Linq;
    using Microsoft.Azure.LocationServices.Geography;
    using NodaTime;
    using NodaTime.TimeZones.IO;
    using TimeZoneNames;

    public class TimeZoneQueryResult
    {
        /// <summary>Serialization ctor</summary>
        public TimeZoneQueryResult()
        {
        }

        /// <summary>Construct an unnamed time zone</summary>
        internal TimeZoneQueryResult(ref Coordinates coord, TimeZoneQueryParams queryParams, TzdbStreamData nzd)
        {
            var offsetHours = Utilities.CalculateOffsetHoursFromLongitude(coord.Longitude);
            var tzId = Utilities.GetTimeZoneId(offsetHours);
            Id = nzd.TzdbIdMap[tzId];
            RepresentativePoint = coord;

            ReferenceTime = new ZoneTime(queryParams.UtcTimestamp, tzId, offsetHours);
            SetTimezoneNames(queryParams);
        }

        /// <summary>Construct a Noda/IANA time zone</summary>
        internal TimeZoneQueryResult(string tzId, TimeZoneQueryParams queryParams, TzdbStreamData nzd)
        {
            Id = nzd.TzdbIdMap[tzId];

            if (0 != (queryParams.Options & TimeZoneQueryParams.AdditionalData.ZoneInfo))
            {
                ///////////////////////////////////////////////////////////////////////////////////////
                // Alternate names
                if (nzd.Aliases.Contains(Id))
                {
                    var list = new List<string>();
                    foreach (var alias in nzd.Aliases[Id])
                    {
                        list.Add(alias);
                    }

                    list.TrimExcess();
                    Aliases = list;
                }

                ///////////////////////////////////////////////////////////////////////////////////////
                // Location
                var zone1970 = nzd.Zone1970Locations.FirstOrDefault(g => g.ZoneId == Id);
                if (zone1970 != null)
                {
                    RepresentativePoint = new Coordinates(zone1970.Latitude, zone1970.Longitude);
                    var list = new List<ICountry>();
                    foreach (var country in zone1970.Countries)
                    {
                        list.Add(country);
                    }

                    list.TrimExcess();
                    Countries = list;
                }
            }

            ///////////////////////////////////////////////////////////////////////////////////////
            // cache the corresponding Noda's DateTimeZone and ZoneInterval
            try
            {
                var dateTimeZone = nzd.CreateZone(tzId, Id);
                ReferenceTime = new ZoneTime(dateTimeZone, queryParams.UtcTimestamp);

                if (0 != (queryParams.Options & TimeZoneQueryParams.AdditionalData.Transitions))
                {
                    var list = new List<TimeTransition>();
                    if (dateTimeZone != null)
                    {
                        foreach (var zi in dateTimeZone.GetZoneIntervals(Instant.FromDateTimeUtc(queryParams.UtcTransitionFrom), Instant.FromDateTimeUtc(queryParams.UtcTransitionTo)))
                        {
                            list.Add(new TimeTransition(zi));
                            if (list.Count == queryParams.MaxTransitionsInResults)
                            {
                                break;
                            }
                        }
                    }

                    TimeTransitions = list;
                }
            }
            catch (Exception ex)
            {
                //ReferenceWallTime = new ReferenceTime(utcTimeStamp, TimeZoneId, offsetHours);
                throw new TimeZoneDbException($"TimeZoneQueryResult.CreateZone({tzId}, {Id})", ex);
            }

            if (TimeTransitions != null && TimeTransitions.Count > queryParams.MaxTransitionsInResults)
            {
                throw new TimeZoneDbException("Too many time transitions.  Try reducing the range.", null);
            }

            SetTimezoneNames(queryParams);
        }

        /// <summary>The IANA id or UTC offset for maritime zones</summary>
        public string Id { get; set; }

        /// <summary>tzId aliases.  Can be null.</summary>
        public IList<string> Aliases { get; set; }

        /// <summary>Countries in this time zone.  Can be null.</summary>
        public IList<ICountry> Countries { get; set; }

        /// <summary>The localized names of the time zone</summary>
        public LocalizedTimeZoneNames Names { get; set; }

        /// <summary>Time details in effect @ReferenceTimestamp</summary>
        public ZoneTime ReferenceTime { get; set; }

        /// <summary>representative coordinates</summary>
        public Coordinates? RepresentativePoint { get; set; }

        /// <summary>Time zone transitions (tz rules in tzdb) from start of time until timestamp + 1 year</summary>
        public IList<TimeTransition> TimeTransitions { get; set; }

        /// <summary>Override</summary>
        public override string ToString() => $"{Id}: {ReferenceTime}";

        public void SetTimezoneNames(TimeZoneQueryParams queryParams)
        {
            Names = new LocalizedTimeZoneNames();

            if (queryParams.ISO6391LanguageCodeOptions == null)
            {
                return;
            }

            if (queryParams.ISO6391LanguageCodeOptions.Count == 1 && queryParams.ISO6391LanguageCodeOptions.First() == "*")
            {
                queryParams.ISO6391LanguageCodeOptions[0] = "en";
            }

            foreach (var languageCode in queryParams.ISO6391LanguageCodeOptions)
            {
                var code = languageCode.Split(';')[0]; // Some default browser Accept-Language values will have weights, for example: fr;q=0.9

                if (!string.IsNullOrWhiteSpace(code) && SupportedLocalizationLanguages.IsSupported(code))
                {
                    try
                    {
                        var tzNames = TZNames.GetNamesForTimeZone(Id, code);
                        var iso6391LanguageCode = code.Split('-').First();
                        Names = new LocalizedTimeZoneNames(iso6391LanguageCode, tzNames);
                        break;
                    }
                    catch
                    {
                        // Continue processing the next language code option
                        continue;
                    }
                }
            }
        }
    }
}
