﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDb
{
    using System;

    public static class Utilities
    {
        public static string FromDateTime(DateTime dt)
        {
            return $"{FromDate(dt)} {FromTime(dt)}";
        }

        public static string FromDate(DateTime dt)
        {
            return $"{dt.Year.ToString("0000")}-{dt.Month.ToString("00")}-{dt.Day.ToString("00")}";
        }

        public static string FromTime(DateTime dt)
        {
            return $"{dt.Hour.ToString("00")}:{dt.Minute.ToString("00")}:{dt.Second.ToString("00")}";
        }

        public static string To00v00(TimeSpan ts)
        {
            return $"{ts.TotalHours.ToString("+00.00;-00.00")}";
        }

        internal static int CalculateOffsetHoursFromLongitude(double lon)
        {
            var dir = lon < 0 ? -1 : 1;
            var posNo = Math.Sqrt(Math.Pow(lon, 2));
            if (posNo <= 7.5)
            {
                return 0;
            }

            posNo -= 7.5;
            var offset = posNo / 15;
            if (posNo % 15 > 0)
            {
                offset++;
            }

            return dir * (int)Math.Floor(offset);
        }

        internal static string GetTimeZoneId(int offsetHours)
        {
            if (offsetHours == 0)
            {
                return NodaTime.DateTimeZone.UtcId;
            }

            // see https://opensource.apple.com/source/system_cmds/system_cmds-230/zic.tproj/datfiles/etcetera
            // We use POSIX-style signs in the Zone names and the output abbreviations, even though this is the opposite of what many people expect.
            // POSIX has positive signs west of Greenwich, but many people expect positive signs east of Greenwich.  For example, TZ='Etc/GMT+4' uses
            // the abbreviation "GMT+4" and corresponds to 4 hours behind UTC (i.e. west of Greenwich) even though many people would expect it to
            // mean 4 hours ahead of UTC (i.e. east of Greenwich).
            // 
            // In the draft 5 of POSIX 1003.1-200x, the angle bracket notation (which is not yet supported by the tz code) allows for
            // TZ='<GMT-4>+4'; if you want time zone abbreviations conforming to ISO 8601 you can use TZ='<-0400>+4'.  Thus the commonly-expected
            // offset is kept within the angle bracket (and is used for display) while the POSIX sign is kept outside the angle bracket (and is used
            // for calculation).
            var reversed = (offsetHours >= 0 ? "-" : "+") + Math.Abs(offsetHours);
            return "Etc/GMT" + reversed;
        }
    }
}
