﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDb
{
    using System;
    using System.Text;

    using NodaTime;
    using NodaTime.TimeZones;

    /// <summary>Ground truth time in reference to the user provided timestamp and location</summary>
    public class ZoneTime
    {
        /// <summary>Constructor</summary>
        internal ZoneTime(DateTime utcStamp, string tag, int offsetHours)
        {
            if (string.IsNullOrEmpty(tag))
            {
                throw new ArgumentException("ReferenceTime ctor: Tag can't be null or empty");
            }

            Tag = tag;
            StandardOffset = TimeSpan.FromHours(offsetHours);
            DaylightSavings = TimeSpan.Zero;
            WallTime = new DateTime(utcStamp.Add(StandardOffset).Ticks, DateTimeKind.Local);
            PosixTzValidYear = utcStamp.Year;
            PosixTz = GetPosixString(offsetHours);
        }

        /// <summary>Constructor</summary>
        internal ZoneTime(DateTimeZone dtz, DateTime utcStamp)
        {
            var zi = dtz.GetZoneInterval(Instant.FromDateTimeUtc(utcStamp));
            if (string.IsNullOrEmpty(zi.Name))
            {
                throw new ArgumentException("ReferenceTime ctor: Tag can't be null or empty");
            }

            Tag = zi.Name;
            StandardOffset = zi.StandardOffset.ToTimeSpan();
            DaylightSavings = zi.Savings.ToTimeSpan();
            var tsWall = zi.WallOffset.ToTimeSpan();
            WallTime = new DateTimeOffset(utcStamp.Add(tsWall).Ticks, tsWall);
            PosixTzValidYear = utcStamp.Year;
            PosixTz = GetPosixString(dtz, PosixTzValidYear);

#if DEBUG
            var dt = new DateTime(utcStamp.Add(tsWall).Ticks, DateTimeKind.Local);
            System.Diagnostics.Debug.Assert(WallTime.DateTime == dt);
            System.Diagnostics.Debug.Assert(WallTime.UtcDateTime == utcStamp);
            System.Diagnostics.Debug.Assert(WallTime.Offset == tsWall);
#endif
        }

        /// <summary>Time zone name in effect @ReferenceTimestamp (i.e. PST or PDT depending whether DST is in effect or not)</summary>
        public string Tag { get; set; }

        /// <summary>UTC-only (no DST) offset in effect @ReferenceTimestamp</summary>
        public TimeSpan StandardOffset { get; set; }

        /// <summary>DST-only time saving in effect @ReferenceTimestamp</summary>
        public TimeSpan DaylightSavings { get; set; }

        /// <summary>Time @ReferenceTimestamp in the given time zone</summary>
        public DateTimeOffset WallTime { get; set; }

        /// <summary>The one and only year this POSIX string is valid for</summary>
        public int PosixTzValidYear { get; set; }

        /// <summary>POSIX string used to set the TZ environment variable</summary>
        public string PosixTz { get; set; }

        /// <summary>tag</summary>
        public static string GetPosixString(int offsetHours)
        {
            // see https://opensource.apple.com/source/system_cmds/system_cmds-230/zic.tproj/datfiles/etcetera
            // We use POSIX-style signs in the Zone names and the output abbreviations, even though this is the opposite of what many people expect.
            // POSIX has positive signs west of Greenwich, but many people expect positive signs east of Greenwich.  For example, TZ='Etc/GMT+4' uses
            // the abbreviation "GMT+4" and corresponds to 4 hours behind UTC (i.e. west of Greenwich) even though many people would expect it to
            // mean 4 hours ahead of UTC (i.e. east of Greenwich).
            // 
            // In the draft 5 of POSIX 1003.1-200x, the angle bracket notation (which is not yet supported by the tz code) allows for
            // TZ='<GMT-4>+4'; if you want time zone abbreviations conforming to ISO 8601 you can use TZ='<-0400>+4'.  Thus the commonly-expected
            // offset is kept within the angle bracket (and is used for display) while the POSIX sign is kept outside the angle bracket (and is used
            // for calculation).
            if (offsetHours == 0)
            {
                return "GMT+0";
            }
            else
            {
                var hours = Math.Abs(offsetHours);
                return (offsetHours > 0)
                    ? $"<-{hours}>+{hours}"
                    : $"<+{hours}>-{hours}";
            }
        }

        /// <summary>tag</summary>
        public static string GetPosixString(DateTimeZone dtz, int year)
        {
            var jan = dtz.GetZoneInterval(Instant.FromUtc(year, 1, 1, 0, 0));
            var jun = dtz.GetZoneInterval(Instant.FromUtc(year, 6, 1, 0, 0));

            var std = jan;
            var dlt = jun;
            if (jan.Savings.Ticks > dlt.Savings.Ticks)
            {
                std = jun;
                dlt = jan;
            }

            var sb = new StringBuilder();
            bool bracket = std.Name[0] == '+' || std.Name[0] == '-';
            if (!bracket)
            {
                for (int i = 1; i < std.Name.Length; i++)
                {
                    if (char.IsNumber(std.Name[i]))
                    {
                        bracket = true;
                        break;
                    }
                }
            }

            sb.Append(bracket ? $"<{std.Name}>" : std.Name);

            var offset = std.StandardOffset.ToTimeSpan();
            sb.AppendFormat("{0:+#;-#;+0}", -offset.Hours);
            if (offset.Minutes != 0)
            {
                sb.Append($":{offset.Minutes.ToString("00")}");
            }

            if (std.WallOffset.Ticks != dlt.WallOffset.Ticks)
            {
                sb.Append(bracket ? $"<{dlt.Name}>" : dlt.Name);

                // Explicit savings if not the standard 1 hour in 4 digit format, i.e. 0030 for 0.5 hours 
                if (Math.Abs(dlt.WallOffset.Seconds - std.WallOffset.Seconds) != 3600)
                {
                    offset = dlt.WallOffset.ToTimeSpan();
                    sb.Append(offset.Hours.ToString("00"));
                    if (offset.Minutes != 0)
                    {
                        sb.Append($"{offset.Minutes.ToString("00")}");
                    }
                }

                if (AppendTransitionForPosix(sb, std))
                {
                    AppendTransitionForPosix(sb, dlt);
                }
            }

            return sb.ToString();
        }

        /// <summary>tag</summary>
        public override string ToString() => 
            $"{Utilities.FromDateTime(WallTime.DateTime)} {Tag} ({Utilities.To00v00(StandardOffset)} DST {Utilities.To00v00(DaylightSavings)})";

        /// <summary>tag</summary>
        private static bool AppendTransitionForPosix(StringBuilder sb, ZoneInterval zi)
        {
            sb.Append(",");

            if (!zi.HasEnd)
            {
                sb.Append("J365/25");
                return false;
            }

            // TZ = local_timezone,date/time,date/time.
            // The date part is in Mm.n.d format, where:
            //   Mm(1 - 12) for 12 months
            //   n(1 - 5) 1 for the first week and 5 for the last week in the month
            //   d(0 - 6) 0 for Sunday and 6 for Saturday
            // The time component can be omitted when it's 02:00:00 (the standard transition time)
            // For instance, the two strings beloe are equivalent:
            //   TZ = CST6CDT,M3.2.0/2:00:00,M11.1.0/2:00:00
            //   TZ = CST6CDT,M3.2.0,M11.1.0
            var dt = zi.IsoLocalEnd.ToDateTimeUnspecified();
            sb.Append($"M{dt.Month}");
            sb.Append($".{((dt.Day - 1) / 7) + 1}");
            sb.Append($".{(int)dt.DayOfWeek}");
            if (dt.TimeOfDay != TimeSpan.FromHours(2))
            {
                if (dt.Minute == 0)
                {
                    sb.Append($"/{dt.Hour}");
                }
                else
                {
                    sb.Append($"/{dt.Hour.ToString("00")}:{dt.Minute.ToString("00")}");
                }
            }

            return true;
        }
    }
}
