﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDbManager
{
    using System;
    using System.IO;
    using System.Linq;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Blob;
    using Microsoft.WindowsAzure.Storage.RetryPolicies;
    using Microsoft.WindowsAzure.Storage.Table;

    /// <summary>tag</summary>
    internal class DbCloudStore
    {
        internal const int ExpectedZipPolyFileCount = 4; // 4 polygon files
        internal const int ExpectedZipDbFileCount = ExpectedZipPolyFileCount + 1; // nzd + 4 polygon files

        private const int CUR = 0;
        private const int NXT = 1;
        private const int _cTableCount = 2;
        private static readonly string[] _sTableNames = new string[_cTableCount] { "tzdbCurrent", "tzdbNext" };

        private readonly CloudStorageAccount _account;
        private readonly CloudTable[] _tables;
        private readonly Logger _loggerRef;

        /// <summary>CloudTable is not thread safe and any execute call needs to be serialized</summary>
        private object _cloudTableLock = new object();
        
        internal DbCloudStore(CloudStorageAccount acct, Logger logger)
        {
            _loggerRef = logger;
            _account = acct;
            _tables = new CloudTable[_cTableCount];            
        }

        internal enum SourceBlobType : byte
        {
            db = 0,
            cela,
            cldr,
            iana,
            _max
        }

        /// <summary>
        /// Creates the tables used by the DB. Should be done on initialization.
        /// </summary>
        /// <returns>awaitable <see cref="Task"/></returns>
        internal async Task CreateTablesAsync()
        {
            var client = _account.CreateCloudTableClient();
            var tasks = new Task[_cTableCount];
            for (int i = 0; i < _cTableCount; i++)
            {
                _tables[i] = client.GetTableReference(_sTableNames[i]);
                tasks[i] = _tables[i].CreateIfNotExistsAsync();
            }

            await Task.WhenAll(tasks);
            foreach (Task task in tasks)
            {
                if (task.IsFaulted)
                {
                    throw task.Exception;
                }
            }
        }

        /// <summary>Get info for the db currently in use</summary>
        /// <returns>An empty TimeZoneDbTableEntity if the store is empty and waiting for bootstrap to complete,
        /// a populated TimeZoneDbTableEntity object with the version in use otherwise</returns>
        internal DbTableEntity GetCurrentDbInfo()
        {
            DbTableEntity cur, nxt;
            GetDbInfo(out cur, out nxt, false);

            if (cur == null) // bootstrap
            {
                return new DbTableEntity();
            }
            else
            {
                return (nxt != null && DateTime.UtcNow.AddMinutes(2) >= nxt.Effective) ? nxt : cur;
            }
        }

        /// <summary>If exists, return the info of the db about to go live.</summary>
        /// <returns>null if 
        ///     (a) db currently in use is the most recent, or
        ///     (b) the store is empty and waiting for bootstrap to complete
        /// otherwise the db info to be prep'd for use by the Effective data</returns>
        internal DbTableEntity GetNextToBeDbInfo()
        {
            DbTableEntity cur, nxt;
            GetDbInfo(out cur, out nxt, true);
            return nxt;
        }

        internal string DownloadArchiveFile(string containerName, SourceBlobType type, string destFolder)
        {
            var blobClient = _account.CreateCloudBlobClient();
            var container = blobClient.GetContainerReference(containerName);
            var blob = container.GetBlockBlobReference(type.ToString());

            var destFilePath = Path.Combine(destFolder, $"{type}.archive");
            blob.DownloadToFile(destFilePath, FileMode.Create);
            return destFilePath;
        }

        internal void DownloadArchiveContents(string containerName, SourceBlobType type, string destFolder)
        {
            var blobClient = _account.CreateCloudBlobClient();
            var container = blobClient.GetContainerReference(containerName);
            var blob = container.GetBlockBlobReference(type.ToString());
            using (var stream = blob.OpenRead())
            {
                Helpers.ExtractArchiveStream(stream, destFolder);
            }
        }

        /// <summary>Upload a new version</summary>
        internal void UploadNewVersion(DbTableEntity current, DbUpdateInfo nextInfo)
        {
            nextInfo.Validate();

            if (current != null)
            {
                if (current.Equals(nextInfo))
                {
                    System.Diagnostics.Debug.Fail("how is this possible?!");
                    return;
                }
            }
            else
            {
                System.Diagnostics.Debug.Assert(nextInfo._isBootstrap, "can only bootstrap when current is null");
            }

            try
            {
                var nxt = new DbTableEntity(nextInfo);

                ////////////////////////////////////////////////////////////////
                // first, upload the blobs
                var blobClient = _account.CreateCloudBlobClient();
                var container = blobClient.GetContainerReference(nxt.Container);
                container.Create();

                var tasks = new Task[(int)SourceBlobType._max];
                var blobs = new CloudBlockBlob[tasks.Length];
                for (int i = 0; i < tasks.Length; i++)
                {
                    var type = (SourceBlobType)i;
                    _loggerRef.Log(LogType.Info, $"TimeZoneDbCloudStore.UploadNewVersion: uploading {type.ToString()} blob");
                    blobs[i] = container.GetBlockBlobReference(type.ToString());
                    tasks[i] = blobs[i].UploadFromFileAsync(nextInfo[type]);
                }

                Task.WaitAll(tasks);

                for (int i = 0; i < tasks.Length; i++)
                {
                    if (tasks[i].IsFaulted)
                    {
                        throw tasks[i].Exception;
                    }
                }

                ////////////////////////////////////////////////////////////////
                // then, insert the record in the table to seal the transaction
                // if this fails, the blobs will be orphaned and invisible for downloading
                var op = TableOperation.Insert(nxt);
                lock (_cloudTableLock)
                {
                    if (nextInfo._isBootstrap)
                    {
                        _tables[CUR].Execute(op);
                        _loggerRef.Log(LogType.Info, $"TimeZoneDbCloudStore.UploadNewVersion: bootstrap {nxt} uploaded to {_tables[CUR].Name} table");
                    }
                    else
                    {
                        _tables[NXT].Execute(op);
                        _loggerRef.Log(LogType.Info, $"TimeZoneDbCloudStore.UploadNewVersion: uploaded {nxt} to {_tables[CUR].Name} table");
                    }
                }
            }
            catch (StorageException se)
            {
                _loggerRef.Log(LogType.Error, $"TimeZoneDbCloudStore.UploadNewVersion: {se.Message}");
                throw;
            }
        }

        /// <summary>tag</summary>
        internal void CommitNewVersion(DbTableEntity current, DbUpdateEventArgs nextInfo)
        {
            var tblQuery = new TableQuery<DbTableEntity>();
            lock (_cloudTableLock)
            {
                var cur = _tables[CUR].ExecuteQuery(tblQuery).OrderByDescending(a => a.Effective).SingleOrDefault();
                if (!cur.Equals(current))
                {
                    throw new InvalidOperationException($"CommitNewVersion: recs for current do not match");
                }

                var nxt = _tables[NXT].ExecuteQuery(tblQuery).OrderByDescending(a => a.Effective).SingleOrDefault();
                if (!nxt.Equals(nextInfo))
                {
                    throw new InvalidOperationException($"CommitNewVersion: recs for next do not match");
                }

                var opts = new TableRequestOptions()
                {
                    MaximumExecutionTime = TimeSpan.FromMinutes(2),
                    RetryPolicy = new LinearRetry(TimeSpan.FromSeconds(5), 12),
                    ServerTimeout = TimeSpan.FromSeconds(10)
                };

                var op = TableOperation.Insert(new DbTableEntity(nxt));
                _tables[CUR].Execute(op, opts);

                op = TableOperation.Delete(nxt);
                _tables[NXT].Execute(op, opts);

                op = TableOperation.Delete(cur);
                _tables[CUR].Execute(op, opts);

                var blobClient = _account.CreateCloudBlobClient();
                var container = blobClient.GetContainerReference(current.Container);
                container.Delete(null);
            }
        }

        private void GetDbInfo(out DbTableEntity cur, out DbTableEntity nxt, bool nxtOnly)
        {
            var qryString = string.Format("({0})", TableQuery.GenerateFilterCondition("Effective", QueryComparisons.LessThanOrEqual, DateTime.UtcNow.ToString()));
            var tblQuery = new TableQuery<DbTableEntity>().Where(qryString);

            lock (_cloudTableLock)
            {
                if (nxtOnly)
                {
                    cur = null;
                }
                else
                {
                    cur = _tables[CUR].ExecuteQuery(tblQuery).OrderByDescending(a => a.Effective).SingleOrDefault();
                }

                nxt = _tables[NXT].ExecuteQuery(tblQuery).OrderByDescending(a => a.Effective).SingleOrDefault();
            }
        }
    }
}
