﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDbManager
{
    using System;

    using Microsoft.WindowsAzure.Storage.Table;

    /// <summary>tzDb table record with version, effective date and a pointer to the blob container</summary>
    public class DbTableEntity : TableEntity, IEquatable<DbTableEntity>
    {
        public readonly DateTime _snapshotTime;

        /// <summary>default ctor (required)</summary>
        public DbTableEntity()
        {
            _snapshotTime = DateTime.UtcNow;
        }

        /// <summary>ctor</summary>
        internal DbTableEntity(DbUpdateInfo rec)
             : base(rec._pk, rec._rk)
        {
            System.Diagnostics.Debug.Assert(rec._zipDbFileCount == DbCloudStore.ExpectedZipDbFileCount, "TimeZoneDbUpdateInfo.Validate() missed ExpectedZipDbFileCount check");

            _snapshotTime = DateTime.MaxValue.ToUniversalTime();

            Container = rec.ContainerName;
            Effective = rec._dbEffectiveDate.UtcDateTime;
            ZipDbFileCount = rec._zipDbFileCount;
            ver_IANA = rec._ianaVersion;
            ver_CLDR_type = rec._cldrTypeVersion;
            ver_CLDR_other = rec._cldrOtherVersion;
            utc_IANA = rec._ianaStamp.UtcDateTime;
            utc_CELA = rec._celaStamp.UtcDateTime;
            Publisher = rec._dbPublisher;
        }

        /// <summary>Copy constructor</summary>
        internal DbTableEntity(DbTableEntity o)
             : base(o.PartitionKey, o.RowKey)
        {
            _snapshotTime = DateTime.MaxValue.ToUniversalTime();

            Container = o.Container;
            Effective = o.Effective;
            ZipDbFileCount = o.ZipDbFileCount;
            ver_IANA = o.ver_IANA;
            ver_CLDR_type = o.ver_CLDR_type;
            ver_CLDR_other = o.ver_CLDR_other;
            utc_IANA = o.utc_IANA;
            utc_CELA = o.utc_CELA;
            Publisher = o.Publisher;
        }

        public TimeSpan SnapshotAge => DateTime.UtcNow - _snapshotTime;

        public bool IsEmpty => string.IsNullOrEmpty(PartitionKey);

        #region ----- Table Fields -----
        public string Container { get; set; }

        public DateTime Effective { get; set; }

        public int ZipDbFileCount { get; set; }

        public string ver_IANA { get; set; }

        public string ver_CLDR_type { get; set; }

        public string ver_CLDR_other { get; set; }

        public DateTime utc_IANA { get; set; }

        public DateTime utc_CELA { get; set; }

        public string Publisher { get; set; }
        #endregion

        public bool Equals(DbTableEntity o)
        {
            if (o == null)
            {
                return false;
            }

            return (this.PartitionKey == o.PartitionKey && this.RowKey == o.RowKey) || this.Container == o.Container;
        }

        public bool Equals(DbUpdateInfo o)
        {
            if (o == null)
            {
                return false;
            }

            return (this.PartitionKey == o._pk && this.RowKey == o._rk) || this.Container == o.ContainerName;
        }

        public bool Equals(DbUpdateEventArgs o)
        {
            if (o == null)
            {
                return false;
            }

            return this.Container == o.ContainerName;
        }

        public override string ToString() => (IsEmpty) ? "Empty" : $"{PartitionKey} {RowKey} from {Effective}, files={ZipDbFileCount}";
    }
}
