﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDbManager
{
    using System;
    using System.IO;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;

    [Flags]
    public enum ExecutionMode : byte
    {
        /// <summary>TimeZoneDb library initializes using a local and static tz db</summary>
        Static = 0,

        /// <summary>TimeZoneDb library uses the most current tz db available in cloud storage</summary>
        Dynamic = 1,

        /// <summary>TimeZoneDb library monitors external sources updates, builds and publishes new tz db's</summary>
        Publisher = 2
    }

    public class ConfigValues
    {
        /// <summary>The intended exec mode (loca, cluster or db publisher)</summary>
        public readonly ExecutionMode Mode;

        /// <summary>
        /// the app setting key to key vault which enables CRUD on the time zone storage account
        /// </summary>
        private const string _timeZoneAccountSasAppSettingKey = "mtzdata-crudall";

        /// <summary>
        /// The app setting key to key vault which is used to format a Sas Connection String
        /// https://docs.microsoft.com/en-us/azure/storage/common/storage-dotnet-shared-access-signature-part-1#authenticating-from-a-client-application-with-a-sas
        /// </summary>
        private const string _timeZoneConnectionStringTemplateAppSettingKey = "mtzdata";

        /// <summary>
        /// The app setting key to key vault which is used to do read access to the cela container.
        /// </summary>
        private const string _celaTimeZoneAppSettingKey = "timezoneshapes-rcelacontainer";

        /// <summary>Local folder used for application data</summary>
        private readonly string _appDataPath;

        /// <summary>
        /// configuration reader
        /// </summary>
        private IConfigurationReader _configurationReader;

        /// <summary>
        /// The key vault id for the Sas Token to CELA
        /// </summary>
        private string _celaAccountSasId;

        /// <summary>
        /// The key vault id for the Sas Token to the Timezone storage
        /// </summary>
        private string _timezoneAccountSasId;

        /// <summary>
        /// The key vault id for the Connection string template for time zone storage
        /// </summary>
        private string _connectionStringTemplateId;
       
        private ConfigValues(ExecutionMode mode, IBootstrapConfiguration bootstrap, IConfigurationReader configurationReader, string localDataPath)
        {
            Mode = mode;

            _connectionStringTemplateId = bootstrap.GetValue(_timeZoneConnectionStringTemplateAppSettingKey);
            if (string.IsNullOrEmpty(_connectionStringTemplateId))
            {
                throw new ArgumentException("connectionStringTemplate cannot be null or empty.", nameof(_connectionStringTemplateId));
            }

            _timezoneAccountSasId = bootstrap.GetValue(_timeZoneAccountSasAppSettingKey);
            if (string.IsNullOrEmpty(_timezoneAccountSasId))
            {
                throw new ArgumentException("timezoneAccountSasId cannot be null or empty.", nameof(_timezoneAccountSasId));
            }
            
            _celaAccountSasId = bootstrap.GetValue(_celaTimeZoneAppSettingKey);
            _configurationReader = configurationReader;

            if (Mode == ExecutionMode.Static)
            {
                if (string.IsNullOrEmpty(localDataPath) || !Directory.Exists(localDataPath))
                {
                    throw new ArgumentException("localDataPath null, empty or folder does not exist");
                }

                _appDataPath = localDataPath;
            }
            else if (0 != (Mode & ExecutionMode.Dynamic) || 0 != (Mode & ExecutionMode.Publisher))
            {
                if (0 != (Mode & ExecutionMode.Publisher))
                {
                    if (string.IsNullOrEmpty(_celaAccountSasId))
                    {
                        throw new ArgumentException("celaAccountKeyId cannot be null or empty", nameof(_celaAccountSasId));
                    }
                }

                if (string.IsNullOrEmpty(localDataPath))
                {
                    localDataPath = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData);
                }

                _appDataPath = Path.Combine(localDataPath, "TimeZoneService");

                if (!Directory.Exists(_appDataPath))
                {
                    Directory.CreateDirectory(_appDataPath);
                }

                if (!Directory.Exists(TimeZoneDbFolder))
                {
                    Directory.CreateDirectory(TimeZoneDbFolder);
                }

                Helpers.EnsureCleanFolder(ExternalSourcesFolder);
                Helpers.EnsureCleanFolder(CompileFolder);
            }
            else
            {
                throw new InvalidOperationException($"Unhandled case for ExecutionMode.{Mode}");
            }
        }
        
        /// <summary>Local folder with runtime data</summary>
        public string TimeZoneDbFolder => (Mode == ExecutionMode.Static) ? _appDataPath : Path.Combine(_appDataPath, "TimeZoneDb");

        /// <summary>External sources download folder</summary>
        public string ExternalSourcesFolder => Path.Combine(_appDataPath, "ExternalDownloads");

        /// <summary>Data compiler work folder</summary>
        public string CompileFolder => Path.Combine(_appDataPath, "Compile");
        
        /// <summary>Construct for static/standalone mode</summary>
        public static ConfigValues CreateForStaticMode(string localDataPath)
        {
            return new ConfigValues(ExecutionMode.Static, new AppSettingBootstapConfiguration(), new KeyVaultConfigurationReader(), localDataPath);
        }

        /// <summary>Construct for static/standalone mode</summary>
        public static ConfigValues CreateForDynamicMode(IBootstrapConfiguration configuration, IConfigurationReader configurationReader)
        {
            return new ConfigValues(ExecutionMode.Dynamic, configuration, configurationReader, null);
        }

        /// <summary>Construct for static/standalone mode, with <see cref="AppSettingBootstapConfiguration"/></summary>
        public static ConfigValues CreateForDynamicMode()
        {
            return new ConfigValues(ExecutionMode.Dynamic, new AppSettingBootstapConfiguration(), new KeyVaultConfigurationReader(), null);
        }

        /// <summary>Construct for static/standalone mode</summary>
        public static ConfigValues CreateForPublisherMode(IBootstrapConfiguration configuration)
        {
            return new ConfigValues(ExecutionMode.Publisher, configuration, new KeyVaultConfigurationReader(), null);
        }
        
        /// <summary>Construct for static/standalone mode, with <see cref="AppSettingBootstapConfiguration"/></summary>
        public static ConfigValues CreateForPublisherMode()
        {
            return new ConfigValues(ExecutionMode.Publisher, new AppSettingBootstapConfiguration(), new KeyVaultConfigurationReader(), null);
        }

        /// <summary>Construct for static/standalone mode</summary>
        public static ConfigValues CreateForUnsafeMode(IBootstrapConfiguration configuration, string localDataPath = null)
        {
            var mode = ExecutionMode.Dynamic | ExecutionMode.Publisher;
            return new ConfigValues(mode, configuration, new KeyVaultConfigurationReader(), localDataPath);
        }

        /// <summary>Construct for static/standalone mode, with <see cref="AppSettingBootstapConfiguration"/></summary>
        public static ConfigValues CreateForUnsafeMode(string localDataPath = null)
        {
            var mode = ExecutionMode.Dynamic | ExecutionMode.Publisher;
            return new ConfigValues(mode, new AppSettingBootstapConfiguration(), new KeyVaultConfigurationReader(), localDataPath);
        }

        /// <summary>
        /// Loads the CELA storage account account Sas Token
        /// </summary>
        /// <returns>the Sas Token</returns>
        public async Task<string> LoadCelaAccountSasTokenAsync()
        {
            return await _configurationReader.GetConfigurationValueAsync(_celaAccountSasId);
        }

        /// <summary>
        /// Loads the connection string based on the Sas Token for the time zone storage account
        /// </summary>
        /// <returns>the connection string</returns>
        public async Task<string> LoadTimeZoneAccountConnectionStringAsync()
        {
            string template = await _configurationReader.GetConfigurationValueAsync(_connectionStringTemplateId);
            string sasToken = await _configurationReader.GetConfigurationValueAsync(_timezoneAccountSasId);
            string connectionString = string.Format(template, sasToken);
            return connectionString;
        }
    }
}
