﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDbManager
{
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Linq;
    using System.Threading;
    using System.Threading.Tasks;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.WindowsAzure.Storage;
    using Microsoft.WindowsAzure.Storage.Auth;
    using NodaTime.TzdbCompiler;
    using NodaTime.Utility;

    /// <summary>Originally a singleton to handle data updates and orchestrated publication.  Now it also sports a 
    /// CronDataUpdateCheck API for invocation from a Cron Azure Function.  SIngleton and CronDataUpdateCheck are mutually exclusive</summary>
    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.OrderingRules", "SA1201:ElementsMustAppearInTheCorrectOrder")]
    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.OrderingRules", "SA1202:ElementsMustBeOrderedByAccess")]
    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.OrderingRules", "SA1204:StaticElementsMustAppearBeforeInstanceElements")]
    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1001:TypesThatOwnDisposableFieldsShouldBeDisposable")]
    public class DataManagementPool
    {
        private const string LEASE_CONTAINER = "tzLease";
        private const string LEASE_NAME = "lease";

        #region ---------- Singleton ----------
        private static object _sSyncLock = new object();
        private static DataManagementPool _sInstance;
        private static SemaphoreSlim _syncronizationSemaphore = new SemaphoreSlim(1);

        /// <summary>
        /// Using int to allow for interlocked exchange of downloaded state of the database.
        /// </summary>
        private static int databaseInitialized = 0;

        /// <summary>
        /// Determines whether the timezone database has been initialed before accessing the files.
        /// </summary>
        public bool DatabaseInitialized => databaseInitialized > 0;

        public static async Task<DataManagementPool> RegisterAsync(ConfigValues configValues, ITimeZoneDbManaged target)
        {                        
            await _syncronizationSemaphore.WaitAsync();
            try
            {
                if (_sInstance != null)
                {
                    throw new InvalidOperationException("An instance of DataManagementPool already exists");
                }

                _sInstance = new DataManagementPool(configValues, target, forCronJob: false);
                await _sInstance.StartAsync();
            }
            finally
            {
                _syncronizationSemaphore.Release();
            }
            
            return _sInstance;
        }

        public static void Unregister(DataManagementPool instance)
        {
            lock (_sSyncLock)
            {
                if (!object.ReferenceEquals(_sInstance, instance))
                {
                    throw new InvalidOperationException("unmatched instance");
                }

                if (_sInstance != null)
                {
                    _sInstance.PrivateDispose();
                    _sInstance = null;
                }
            }
        }
        #endregion // Singleton

        /// <summary>Retrofit function to support data update and compilation synchronously from a Cron job</summary>
        /// <returns>Enumerator result code. Logged errors must be handled by the callback function</returns>
        public static async Task<CronJobResult> CronDataUpdateCheckAsync(ConfigValues configValues, ITimeZoneDbManaged target)
        {
            CronJobResult result = CronJobResult.Error;
            await _syncronizationSemaphore.WaitAsync();
            if (_sInstance != null)
            {
                _syncronizationSemaphore.Release();
                throw new InvalidOperationException("An instance of DataManagementPool already exists");
            }

            try
            {
                _sInstance = new DataManagementPool(configValues, target, forCronJob: true);
                await _sInstance.StartAsync();
                
                //wait until it completes or the Azure Function times out
                result = _sInstance._cronJob.WaitForCompletion();
            }
            catch (Exception ex)
            {
                target.GetLogger.Log(LogType.Error, msgFmt: $"SingleUpdateDataCheck error: {ex.Message}");
                result = CronJobResult.Error;
            }
            finally
            {
                if (_sInstance != null)
                {
                    _sInstance.PrivateDispose();
                    _sInstance = null;
                }

                _syncronizationSemaphore.Release();
            }

            return result;
        }

        private readonly int _dbUpdateAndServerRegistrationFrequencySeconds;
        private readonly int _dataPublishAllowanceSeconds;
        private readonly ConfigValues _config;        
        private readonly ITimeZoneDbManaged _managedDb;

        private CloudStorageAccount _account;
        private DataUpdateCronJob _cronJob;
        private ManualResetEvent _disposeEvent;
        private Thread _mainThread;
        private PublisherTokenSentinel _publisherTokenSentinel;
        private DbCloudStore _tzdbCloudStore;
        private DbTableEntity _currentDbInfo;

        /// <summary>Private constructor</summary>
        private DataManagementPool(ConfigValues configValues, ITimeZoneDbManaged target, bool forCronJob)
        {
            if (configValues == null || target == null)
            {
                throw new ArgumentException("DataManagementPool: invalid constructor arguments");
            }

            if (0 != (configValues.Mode & ExecutionMode.Static))
            {
                throw new InvalidOperationException("DataManagementPool shouldn't be initialized when ExecutionMode is Static");
            }

            _config = configValues;
            _managedDb = target;
            _disposeEvent = new ManualResetEvent(false);
            _cronJob = new DataUpdateCronJob(forCronJob ? _disposeEvent : null);

            // **NOTE**: The default timeout for functions on a Consumption plan is 5 minutes. We ovverride the default to 
            //           the allowed max of 10 minutes.  The Cron job needs to download, build and commit within that 10 minute window.
            //           Concurrently, this setting should be mindful of the instances running on the Service servers and limit
            //           pinging the Azure Tables to the minimum required
            _dbUpdateAndServerRegistrationFrequencySeconds = 60 * 5;
            _dataPublishAllowanceSeconds = _dbUpdateAndServerRegistrationFrequencySeconds + 60;
#if DEBUG
            if (forCronJob)
            {
                _dbUpdateAndServerRegistrationFrequencySeconds = 60;
                _dataPublishAllowanceSeconds = _dbUpdateAndServerRegistrationFrequencySeconds + 30;
            }
#endif
        }

        /// <summary>
        /// Starts the Data Management Pool, will retrieve any keys and start any threads.
        /// </summary>
        /// <returns>the continuation task</returns>
        public async Task StartAsync()
        {
            _managedDb.GetLogger.Log(LogType.Info, $"tz DataManagementPool: initialized with ExecutionMode={_config.Mode}");
            string connectionString = await _config.LoadTimeZoneAccountConnectionStringAsync();
            _account = CloudStorageAccount.Parse(connectionString);

            if (0 != (_config.Mode & ExecutionMode.Publisher))
            {
                _publisherTokenSentinel = new PublisherTokenSentinel(_account, _managedDb.GetLogger, LEASE_CONTAINER, LEASE_NAME);
            }

            _tzdbCloudStore = new DbCloudStore(_account, _managedDb.GetLogger);
            await _tzdbCloudStore.CreateTablesAsync();
            _mainThread = new Thread(new ThreadStart(CoordinatorThread));
            _mainThread.Name = "tz DataManagementPool Coordinator";
            _mainThread.Priority = _cronJob.IsEnabled ? ThreadPriority.Highest : ThreadPriority.BelowNormal;
            _mainThread.IsBackground = true;
            _mainThread.Start();
        }

        /// <summary>Ensure the current db is present on a local disk</summary>
        public DbUpdateEventArgs GetCurrentDb(int timeoutseconds)
        {
            var result = GetDbToDisk(true, timeoutseconds);
            Interlocked.Exchange(ref databaseInitialized, 1);
            return result;
        } 

        /// <summary>tag</summary>
        private void CoordinatorThread()
        {
            Task commitDbUpdateTask = null;
            int consecutiveFailuresCommitDbUpdate = 0;

            ServerTableEntity serverRegistrationRecord = null;
            Task dbUpdateAndServerRegistrationTask = null;
            DbUpdateEventArgs dbUpdateArgs = null;
            int consecutiveFailuresDbUpdateAndServerRegistration = 0;

            DateTimeOffset lastSourcesCheck = DateTimeOffset.MinValue;
            Task checkForExternalSourcesUpdatesTask = null;
            int consecutiveFailuresCheckingExternalSources = 0;            
            while (!_disposeEvent.WaitOne(1000, false))
            {
                try
                {
                    // -------------------------------------------------------------------------------------
                    // ------------------------------- DB UPDATE SWITCHOVER  -------------------------------
                    // -------------------------------------------------------------------------------------
                    if (_managedDb.IsReadyToSwapContainers(dbUpdateArgs))
                    {
                        _managedDb.PublishAwaitingContainer();

                        if (_publisherTokenSentinel != null && _publisherTokenSentinel.HasLease)
                        {
                            commitDbUpdateTask = Task.Run(() =>
                                {
                                    _tzdbCloudStore.CommitNewVersion(_currentDbInfo, dbUpdateArgs);
                                    dbUpdateArgs = null;
                                    _cronJob.SetSubtaskResult(DataUpdateCronJob.Status.NewDataCommitted);
                                });
                        }
                    }

                    MonitorTask(ref commitDbUpdateTask, "commitDbUpdate", ref consecutiveFailuresCommitDbUpdate);

                    // -------------------------------------------------------------------------------------
                    // --------------------- DB UPDATE AND SERVER REGISTRATION CHECKS ----------------------
                    // -------------------------------------------------------------------------------------
                    if (DbUpdateAndServerRegistrationOverdue(serverRegistrationRecord))
                    {
                        if (dbUpdateAndServerRegistrationTask == null)
                        {
                            dbUpdateAndServerRegistrationTask = Task.Run(
                                delegate
                                {
                                    // these calls may block the task for a little while. Upon completion, the call to
                                    // ServerTableEntity.UpdateRegistration writes an entry signaling that the new
                                    // awaiting container is loaded
                                    var args = GetDbToDisk(false, 300);
                                    if (args != null && !args.Equals(dbUpdateArgs))
                                    {
                                        dbUpdateArgs = args;
                                        _managedDb.LoadAwaitingContainer(dbUpdateArgs);
                                        _cronJob.SetSubtaskResult(DataUpdateCronJob.Status.NewDataLoaded);
                                    }

                                    serverRegistrationRecord = ServerTableEntity.UpdateRegistration(_account, _managedDb, _managedDb.GetLogger);
                                });
                        }
                    }

                    MonitorTask(ref dbUpdateAndServerRegistrationTask, "dbUpdateAndServerRegistration", ref consecutiveFailuresDbUpdateAndServerRegistration);

                    // -------------------------------------------------------------------------------------
                    // ------------------------------ EXTERNAL SOURCES CHECK  ------------------------------
                    // -------------------------------------------------------------------------------------
                    if (_publisherTokenSentinel != null && _publisherTokenSentinel.HasLease)
                    {
                        if ((DateTimeOffset.UtcNow - lastSourcesCheck) >= TimeSpan.FromMinutes(30))
                        {
                            if (checkForExternalSourcesUpdatesTask == null)
                            {
                                checkForExternalSourcesUpdatesTask = Task.Run(
                                    async delegate
                                    {
                                        if (RefreshCache())
                                        {
                                            Helpers.EnsureCleanFolder(_config.ExternalSourcesFolder);
                                            Helpers.EnsureCleanFolder(_config.CompileFolder);

                                            try
                                            {
                                                bool sourceDataCheckResult = await CheckForSourceDataUpdatesAsync();
                                                var cronResult = sourceDataCheckResult
                                                    ? DataUpdateCronJob.Status.NewDataPosted
                                                    : DataUpdateCronJob.Status.NoNewData;
                                                _cronJob.SetSubtaskResult(cronResult);                                                
                                            }
                                            catch (Exception ex)
                                            {
                                                _managedDb.GetLogger.Log(LogType.Error, $"CheckForSourceDataUpdates error: {ex.Message}");
                                            }

                                            Helpers.EnsureCleanFolder(_config.ExternalSourcesFolder);
                                            Helpers.EnsureCleanFolder(_config.CompileFolder);
                                        }

                                        lastSourcesCheck = DateTimeOffset.UtcNow;                                        
                                    });
                            }
                        }
                    }
                    else
                    {
                        FlushCache();
                    }

                    MonitorTask(ref checkForExternalSourcesUpdatesTask, "checkForExternalSourcesUpdates", ref consecutiveFailuresCheckingExternalSources);
                }
                catch (Exception ex)
                {
                    _managedDb.GetLogger.Log(LogType.Error, $"DataManagementPool._mainThread error: {ex.Message}");
                }
            } // while !disposed

            _managedDb.GetLogger.Log(LogType.Verbose, "DataManagementPool._mainThread: exited while loop and cleaning up");
            if (_publisherTokenSentinel != null)
            {
                _publisherTokenSentinel.Dispose();
                _publisherTokenSentinel = null;
            }

            _managedDb.GetLogger.Log(LogType.Verbose, "DataManagementPool._mainThread: exiting");
        }

        /// <summary>tag</summary>
        internal DateTimeOffset GetPublicationTime(bool isBootstrap) => 
            isBootstrap 
                ? new DateTime(1900, 1, 1, 0, 0, 0, DateTimeKind.Utc)
                : DateTimeOffset.UtcNow.AddSeconds(_dataPublishAllowanceSeconds);

        /// <summary>tag</summary>
        private bool DbUpdateAndServerRegistrationOverdue(ServerTableEntity e) => 
            (e == null || (DateTimeOffset.UtcNow - e.Timestamp) >= TimeSpan.FromSeconds(_dbUpdateAndServerRegistrationFrequencySeconds));

        /// <summary>tag</summary>
        private void MonitorTask(ref Task task, string tag, ref int failCount)
        {
            if (task == null)
            {
                return;
            }

            _managedDb.GetLogger.Log(LogType.Verbose, $"DataManagementPool {tag} task status = {task.Status}");
            switch (task.Status)
            {
                case TaskStatus.RanToCompletion:
                    break; 

                case TaskStatus.Canceled: // impossible
                    break;

                case TaskStatus.Faulted:
                    if (++failCount > 2)
                    {
                        _managedDb.GetLogger.Log(LogType.Error, $"DataManagementPool {tag} task error: {task.Exception}");

                        //@todo davsta: sound alarm
                    }

                    break;

                case TaskStatus.Created:
                case TaskStatus.WaitingToRun:
                    // @todo davsta: should we worry about scheduler taking too long to run the task?
                    return;

                case TaskStatus.Running:
                case TaskStatus.WaitingForActivation:
                case TaskStatus.WaitingForChildrenToComplete:
                default:
                    return;
            }

            task = null;
            failCount = 0;
        }

        /// <summary>tag</summary>
        private DbUpdateEventArgs GetDbToDisk(bool currentDb, int timeoutSeconds)
        {
            var start = DateTime.Now;
            DbTableEntity info = null;

            do
            {
                try
                {
                    if (info == null || info.IsEmpty)
                    {
                        info = currentDb ? _tzdbCloudStore.GetCurrentDbInfo() : _tzdbCloudStore.GetNextToBeDbInfo();
                    }

                    if (info == null || info.IsEmpty)
                    {
                        if (currentDb)
                        {
                            continue; // bootstrapping: a server is building and posting the first db
                        }
                        else
                        {
                            return null; // no db updates waiting
                        }
                    }

                    System.Diagnostics.Debug.Assert(info != null && !info.IsEmpty, "review GetDbToDisk logic");

                    var result = new DbUpdateEventArgs(_config.TimeZoneDbFolder, info);

                    if (Directory.Exists(result.DbFolderPath))
                    {
                        var count = Directory.EnumerateFiles(result.DbFolderPath).Count();
                        if (count == info.ZipDbFileCount)
                        {
                            return result; // data already cached to local disk
                        }
                        else if (count > 0)
                        {
                            Helpers.EnsureCleanFolder(result.DbFolderPath);
                        }
                    }
                    else
                    {
                        Directory.CreateDirectory(result.DbFolderPath);
                    }

                    _tzdbCloudStore.DownloadArchiveContents(info.Container, DbCloudStore.SourceBlobType.db, result.DbFolderPath);
                    return result;
                }
                catch
                {
                }

                Thread.Sleep(250);
            }
            while (timeoutSeconds < 1 || (DateTime.Now - start) < TimeSpan.FromSeconds(timeoutSeconds));

            throw new TimeoutException($"DataSourceHelper.GetCurrentDb timed out after {(DateTime.Now - start).TotalSeconds} seconds");
        }

        /// <summary>tag</summary>
        private bool IsCurrentDbInfoCached => (_currentDbInfo != null && _currentDbInfo.SnapshotAge.TotalHours < 1);

        /// <summary>tag</summary>
        private void FlushCache()
        {
            _currentDbInfo = null;
        }

        /// <summary>tag</summary>
        private bool RefreshCache()
        {
            if (IsCurrentDbInfoCached)
            {
                return true;
            }

            try
            {
                _currentDbInfo = _tzdbCloudStore.GetCurrentDbInfo();
            }
            catch
            {
                // @todo davsta: logging & alarms
                FlushCache();
            }

            return IsCurrentDbInfoCached;
        }

        /// <summary>tag</summary>
        private async Task<bool> CheckForSourceDataUpdatesAsync()
        {
            var cldrTypeVersion = _currentDbInfo.ver_CLDR_type;
            var cldrOtherVersion = _currentDbInfo.ver_CLDR_other;

            DateTimeOffset celaStamp = _currentDbInfo.utc_CELA;
            DateTimeOffset ianaStamp = _currentDbInfo.utc_IANA;

            string cldrZipFilePath, ianaTarGzFilePath;

            var celaResult = await ExternalSources.CheckUpdateCelaAsync(_managedDb.GetLogger, _config, celaStamp);
            if (celaResult.Result == ExternalSources.Result.NewDataReady)
            {
                celaStamp = celaResult.TimeStamp;
            }

            string celaZipFilePath = celaResult.ZipFilePath;
            
            var cldrResult = ExternalSources.CheckUpdateCLDR(_managedDb.GetLogger, _config, ref cldrTypeVersion, ref cldrOtherVersion, out cldrZipFilePath);
            var ianaResult = ExternalSources.CheckUpdateIANA(_managedDb.GetLogger, _config, ref ianaStamp, out ianaTarGzFilePath);

            //// ********************************************************************
            //// @todo davsta: log & alarms if we can't get to external resources
            //// ********************************************************************

            _managedDb.GetLogger.Log(LogType.Verbose, $"external source changes for CELA = {celaResult.Result}");
            _managedDb.GetLogger.Log(LogType.Verbose, $"external source changes for CLDR = {cldrResult}");
            _managedDb.GetLogger.Log(LogType.Verbose, $"external source changes for IANA = {ianaResult}");

            if (0 == ((celaResult.Result | cldrResult | ianaResult) & ExternalSources.Result.NewDataReady))
            {
                return false;
            }

            int extractCount = int.MinValue;

            //////////////////////////////////////////////////////////////////////////////////////////////
            // stage IANA
            if (ianaResult != ExternalSources.Result.NewDataReady)
            {
                if (_currentDbInfo.IsEmpty)
                {
                    _managedDb.GetLogger.Log(LogType.Error, $"CheckForSourceDataUpdates cannot proceed: IANA download result = '{ianaResult}' and no current db sources to download");
                    return false;
                }

                ianaTarGzFilePath = _tzdbCloudStore.DownloadArchiveFile(_currentDbInfo.Container, DbCloudStore.SourceBlobType.iana, _config.ExternalSourcesFolder);
            }

            var ianaCompileDir = Path.Combine(_config.CompileFolder, "iana");
            extractCount = Helpers.TryExtractArchiveFile(_managedDb.GetLogger, ianaTarGzFilePath, ianaCompileDir);
            if (extractCount < 1)
            {
                throw new ArgumentException($"IANA tar.gz failed to extract: extract count = {extractCount}");
            }

            // bug 1476060: in one instance, IANA touched the gz files and affected the file timestamp without acutally changing 
            // the contents and the tzdb version. In addition to detecting a file create change, we now look into the files and 
            // check for an actual verison change UNLESS we are in bootstrap mode
            if (ianaResult == ExternalSources.Result.NewDataReady && !_currentDbInfo.IsEmpty)
            {
                var newVer = NodaTime.TzdbCompiler.Tzdb.TzdbZoneInfoCompiler.InferTimeZoneDbVersion(ianaCompileDir);
                if (_currentDbInfo.ver_IANA.Equals(newVer))
                {
                    _managedDb.GetLogger.Log(LogType.Warning, "IANA touched gz timestamp but tzdb version didn't change");
                    return false;
                }
            }

            //////////////////////////////////////////////////////////////////////////////////////////////
            // stage CELA
            if (celaResult.Result != ExternalSources.Result.NewDataReady)
            {
                if (_currentDbInfo.IsEmpty)
                {
                    _managedDb.GetLogger.Log(LogType.Error, $"CheckForSourceDataUpdates cannot proceed: CELA download result = '{celaResult}' and no current db sources to download");
                    return false;
                }

                celaZipFilePath = _tzdbCloudStore.DownloadArchiveFile(_currentDbInfo.Container, DbCloudStore.SourceBlobType.cela, _config.ExternalSourcesFolder);
            }

            extractCount = Helpers.TryExtractArchiveFile(_managedDb.GetLogger, celaZipFilePath, _config.CompileFolder);
            if (extractCount != DbCloudStore.ExpectedZipPolyFileCount)
            {
                throw new ArgumentException($"CELA zip failed to extract: extract count = {extractCount} instead of {DbCloudStore.ExpectedZipPolyFileCount}");
            }

            //////////////////////////////////////////////////////////////////////////////////////////////
            // stage CLDR
            if (cldrResult != ExternalSources.Result.NewDataReady)
            {
                if (_currentDbInfo.IsEmpty)
                {
                    _managedDb.GetLogger.Log(LogType.Error, $"CheckForSourceDataUpdates cannot proceed: CLDR download result = '{cldrResult}' and no current db sources to download");
                    return false;
                }

                cldrZipFilePath = _tzdbCloudStore.DownloadArchiveFile(_currentDbInfo.Container, DbCloudStore.SourceBlobType.cldr, _config.ExternalSourcesFolder);
            }

            var cldrCompileDir = Path.Combine(_config.CompileFolder, "cldr");
            extractCount = Helpers.TryExtractArchiveFile(_managedDb.GetLogger, cldrZipFilePath, cldrCompileDir);
            if (extractCount != 1)
            {
                throw new ArgumentException($"CLDR zip failed to extract: extract count = {extractCount} instead of 1");
            }

            //////////////////////////////////////////////////////////////////////////////////////////////
            // compile
            InvalidNodaDataException[] compileErrors;
            var nzdFilePath = Compiler.Run(_managedDb.GetLogger, ianaCompileDir, cldrCompileDir, _config.CompileFolder, out compileErrors);
            if (compileErrors != null && compileErrors.Length > 0)
            {
                foreach (var err in compileErrors)
                {
                    _managedDb.GetLogger.Log(LogType.Error, $"NodaCompiler error: {err}");
                }

                return false;
            }

            //////////////////////////////////////////////////////////////////////////////////////////////
            // basic tests
            string ianaVersion;
            IList<string> testErrors;
            if (!_managedDb.TestNewDatabase(nzdFilePath, _config.CompileFolder, out ianaVersion, out testErrors))
            {
                _managedDb.GetLogger.Log(LogType.Error, $"TestNewData failed with {testErrors.Count} errors");
                for (int i = 0; i < testErrors.Count; i++)
                {
                    _managedDb.GetLogger.Log(LogType.Error, $"TestNewData error #{i+1}: {testErrors[i]}");
                }

                return false;
            }

            //////////////////////////////////////////////////////////////////////////////////////////////
            // set publish time, zip and upload
            var zipDbFiles = Directory.EnumerateFiles(_config.CompileFolder).ToArray();
            var isBootstrap = _currentDbInfo.IsEmpty;
            var info = new DbUpdateInfo(zipDbFiles.Length, isBootstrap, GetPublicationTime(isBootstrap))
            {
                _dbPublisher = Environment.MachineName,
                _celaStamp = celaStamp,
                _cldrTypeVersion = cldrTypeVersion,
                _cldrOtherVersion = cldrOtherVersion,
                _ianaStamp = ianaStamp,
                _ianaVersion = ianaVersion,
            };

            info[DbCloudStore.SourceBlobType.db] = Path.Combine(_config.CompileFolder, "TimeZoneDb.zip");
            info[DbCloudStore.SourceBlobType.cela] = celaZipFilePath;
            info[DbCloudStore.SourceBlobType.cldr] = cldrZipFilePath;
            info[DbCloudStore.SourceBlobType.iana] = ianaTarGzFilePath;

            if (zipDbFiles.Length != Helpers.TryCreateArchive(_managedDb.GetLogger, SharpCompress.Common.ArchiveType.Zip, info[DbCloudStore.SourceBlobType.db], zipDbFiles))
            {
                return false;
            }

            _tzdbCloudStore.UploadNewVersion(_currentDbInfo, info);
            return true;
        }

        /// <summary>tag</summary>
        private void PrivateDispose()
        {
            _managedDb.GetLogger.Log(LogType.Verbose, "DataManagementPool.PrivateDispose() start");

            //signal threads to exit and give them a chance to exit gracefully
            _disposeEvent.Set();
            if (_mainThread != null)
            {
                _managedDb.GetLogger.Log(LogType.Verbose, "DataManagementPool signaled _disposeEvent and waiting on _coordinatorThread to exit");
                _mainThread.Join(5000);
                _managedDb.GetLogger.Log(LogType.Verbose, "DataManagementPool done waiting on _coordinatorThread.Join");
                if (_mainThread.ThreadState == ThreadState.Running || _mainThread.ThreadState == ThreadState.Background)
                {
                    _managedDb.GetLogger.Log(LogType.Warning, "DataManagementPool _coordinatorThread still running!!");
                    try
                    {
                        _mainThread.Abort();
                    }
                    catch (ThreadAbortException)
                    {
                        /*nothing*/
                    }
                }
            }

            if (_publisherTokenSentinel != null)
            {
                _managedDb.GetLogger.Log(LogType.Warning, "DataManagementPool _publisherSentinel not null!!");
                _publisherTokenSentinel.Dispose();
                _publisherTokenSentinel = null;
            }

            if (_disposeEvent != null)
            {
                _disposeEvent.Dispose();
                _disposeEvent = null;
            }

            _managedDb.GetLogger.Log(LogType.Verbose, "DataManagementPool.PrivateDispose() end");
        }
    }
}
