﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDbManager
{
    using System;
    using System.Diagnostics;
    using System.Threading;

    public enum CronJobResult
    {
        Error = 0,
        NoNewData = 1,
        NewDataPublished = 2
    }

    internal class DataUpdateCronJob
    {
        private readonly ManualResetEvent _completionEventRef;
        private Status _status;

        internal DataUpdateCronJob(ManualResetEvent doneSignal)
        {
            _completionEventRef = doneSignal;

            //this._status = Status.None;
        }

        [Flags]
        internal enum Status : byte
        {
            None = 0x00,
            NoNewData = 0x01,
            NewDataPosted = 0x02,
            NewDataLoaded = 0x04,
            NewDataCommitted = 0x08
        }

        internal bool IsEnabled => (_completionEventRef != null);

        internal CronJobResult WaitForCompletion()
        {
            if (!IsEnabled)
            {
                throw new InvalidOperationException("CronJob is not enabled");
            }

            Debug.Assert(_completionEventRef != null, "_completionEventRef cannot be null if the cron job is enabled");
            _completionEventRef.WaitOne();

            if (_status == Status.NoNewData)
            {
                return CronJobResult.NoNewData;
            }

            if (0 != (_status & Status.NewDataCommitted))
            {
                return CronJobResult.NewDataPublished;
            }

            throw new InvalidProgramException($"Unexpected DataUpdateCronJob state '{_status}' upon exit from DataManagementPool.CoordinatorThread()");
        }

        /// <summary>Set the result/completion of each sub task</summary>
        internal void SetSubtaskResult(Status taskResult)
        {
            if (!IsEnabled)
            {
                return;
            }

            switch (taskResult) //cannot be masked...
            {
                case Status.NoNewData:
                    if (_status != Status.None)
                    {
                        throw new InvalidProgramException($"SetSubtaskResult(): cannot set '{taskResult}' on top of '{_status}'");
                    }

                    _status = taskResult;
                    break;

                case Status.NewDataPosted:
                    if (_status != Status.None)
                    {
                        throw new InvalidProgramException($"SetSubtaskResult(): cannot set '{taskResult}' on top of '{_status}'");
                    }

                    _status = taskResult;
                    break;

                case Status.NewDataLoaded:  // new data is pre-loaded and waits for publish time: usually it happens before publishing is committed
                    if (0 == (_status & Status.NewDataPosted))
                    {
                        throw new InvalidProgramException($"SetSubtaskResult(): cannot set '{taskResult}' if not '{Status.NewDataCommitted}'");
                    }

                    _status |= taskResult;
                    break;

                case Status.NewDataCommitted:
                    if (0 == (_status & Status.NewDataPosted))
                    {
                        throw new InvalidProgramException($"SetSubtaskResult(): cannot set '{taskResult}' if not '{Status.NewDataPosted}'");
                    }

                    _status |= taskResult;
                    break;

                default:
                    throw new InvalidProgramException($"SetSubtaskResult(): unhandled taskResult '{taskResult}'");
            }

            // we are done if no new data, or the new data has been published AND load verified by this instance
            if (_status == Status.NoNewData
            || _status == (Status.NewDataPosted | Status.NewDataLoaded | Status.NewDataCommitted))
            {
                _completionEventRef.Set();
            }
        }
    }
}
