﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZoneDbManager
{
    using System;
    using System.IO;
    using System.Linq;
    using System.Threading.Tasks;
    using System.Xml.Linq;
    using FluentFTP;
    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.WindowsAzure.Storage.Blob;

    /// <summary>static methods to download external sources if a version change has been detected</summary>
    internal static class ExternalSources
    {
        internal static readonly string _celaAccountName = "timezoneshapes";
        internal static readonly string _celaContainer = "celaupload".ToLowerInvariant();
        internal static readonly string _celaBlob = "TimeZonePoly.zip";
        internal static readonly string _cldrFileNameNoExt = "windowsZones";
        internal static readonly string _cldrXmlFile = $"{_cldrFileNameNoExt}.xml";
        internal static readonly string _cldrXmlUrl = "http://unicode.org/repos/cldr/trunk/common/supplemental/" + _cldrXmlFile;
        internal static readonly string _cldrZipFile = $"{_cldrFileNameNoExt}.zip";
        internal static readonly string _ianaFtp = "ftp.iana.org";
        internal static readonly string _ianaDirectory = "tz";
        internal static readonly string _ianaTarGzLink = "tzdata-latest.tar.gz";
        
        [Flags]
        internal enum Result : byte
        {
            Undefined = 0,

            /// <summary>An error occurred trying to access the resource. Check the err out param</summary>
            ConnectError = 0x01,

            /// <summary>External resource version is different but data download failed. Check the err out param</summary>
            DownloadError = 0x02,

            /// <summary>Downloaded data couldn't be parsed. Check the err out param</summary>
            ParseError = 0x04,

            /// <summary>External resource does not exist</summary>
            NotFound = 0x08,

            /// <summary>Failed to zip file</summary>
            ZipError = 0x10,

            /// <summary>External resource version matches input param</summary>
            NoChanges = 0x20,

            /// <summary>External resource version is different and new data downloaded successfully</summary>
            NewDataReady = 0x40,
        }
        
        internal static async Task<UpdateResponse> CheckUpdateCelaAsync(Logger logger, ConfigValues config, DateTimeOffset stamp)
        {
            string sasToken = await config.LoadCelaAccountSasTokenAsync();
            var blob = new CloudBlockBlob(new Uri($"https://{_celaAccountName}.blob.core.windows.net/{_celaContainer}/{_celaBlob}{sasToken}"));
            UpdateResponse response = new UpdateResponse();
            try
            {                
                if (!blob.Exists())
                {
                    response.Result = Result.NotFound;
                    return response;
                }

                blob.FetchAttributes();
                if (blob.Properties.LastModified.Equals(stamp))
                {
                    response.Result = Result.NoChanges;
                    return response;
                }
            }
            catch (Exception ex)
            {
                logger.Log(LogType.Error, $"CheckUpdateCELA exception: {ex.Message}");
                response.Result = Result.ConnectError;
                return response;
            }

            /////////////////////////////////////////////////////////////////
            // version changed: download
            try
            {
                string zipFilePath = Path.Combine(config.ExternalSourcesFolder, _celaBlob);
                blob.DownloadToFile(zipFilePath, FileMode.Create);
                stamp = blob.Properties.LastModified.Value;
                response.Result = Result.NewDataReady;
                response.TimeStamp = blob.Properties.LastModified.Value;
                response.ZipFilePath = zipFilePath;
                return response;
            }
            catch (Exception ex)
            {
                logger.Log(LogType.Error, $"CheckUpdateCELA exception: {ex.Message}");
                response.Result = Result.DownloadError;
                return response;
            }
        }

        internal static Result CheckUpdateCLDR(Logger logger, ConfigValues config, ref string typeVersion, ref string otherVersion, out string zipFilePath)
        {
            zipFilePath = null;

            XDocument xdoc = null;
            try
            {
                xdoc = XDocument.Load(_cldrXmlUrl);
            }
            catch (Exception ex)
            {
                logger.Log(LogType.Error, $"CheckUpdateCLDR exception: {ex.Message}");
                return Result.DownloadError;
            }

            try
            {
                var tzdbVersion = xdoc.Root.Element("windowsZones")?.Element("mapTimezones")?.Attribute("typeVersion")?.Value ?? "";
                var windowsVersion = xdoc.Root.Element("windowsZones")?.Element("mapTimezones")?.Attribute("otherVersion")?.Value ?? "";

                if (string.IsNullOrEmpty(typeVersion)
                    || string.IsNullOrEmpty(otherVersion)
                    || !typeVersion.Equals(tzdbVersion, StringComparison.InvariantCultureIgnoreCase)
                    || !otherVersion.Equals(windowsVersion, StringComparison.InvariantCultureIgnoreCase))
                {
                    typeVersion = tzdbVersion;
                    otherVersion = windowsVersion;

                    string[] xmlFilePath = new string[] { Path.Combine(config.ExternalSourcesFolder, _cldrXmlFile) };
                    xdoc.Save(xmlFilePath[0]);
                    zipFilePath = Path.Combine(config.ExternalSourcesFolder, _cldrZipFile);

                    if (1 != Helpers.TryCreateArchive(logger, SharpCompress.Common.ArchiveType.Zip, zipFilePath, xmlFilePath))
                    {
                        return Result.ZipError;
                    }

                    File.Delete(xmlFilePath[0]);
                    return Result.NewDataReady;
                }
            }
            catch (Exception ex)
            {
                logger.Log(LogType.Error, $"CheckUpdateCLDR exception: {ex.Message}");
                return Result.ParseError;
            }

            return Result.NoChanges;
        }

        internal static Result CheckUpdateIANA(Logger logger, ConfigValues config, ref DateTimeOffset stamp, out string tarGzFilePath)
        {
            tarGzFilePath = null;

            using (var ftpClient = new FtpClient())
            {
                try
                {
                    ftpClient.Credentials = new System.Net.NetworkCredential("anonymous", "anonymous@microsoft.com");
                    ftpClient.Host = _ianaFtp;
                    ftpClient.Connect();
                }
                catch (Exception ex)
                {
                    logger.Log(LogType.Error, $"CheckUpdateIANA exception: {ex.Message}");
                    return Result.ConnectError;
                }

                try
                {
                    if (!ftpClient.DirectoryExists(_ianaDirectory))
                    {
                        return Result.NotFound;
                    }

                    ftpClient.SetWorkingDirectory(_ianaDirectory);
                    var ftpListItems = ftpClient.GetListing();
                    var linkItem = ftpListItems
                        .Where(t => t.Type == FtpFileSystemObjectType.Link && t.Name.Equals(_ianaTarGzLink, StringComparison.InvariantCultureIgnoreCase))
                        .FirstOrDefault();

                    if (linkItem == null)
                    {
                        return Result.NotFound;
                    }

                    // *** NOTE **** don't use the linkItem.Modified attribute because it isn't set
                    var modTimeUtc = ftpClient.GetModifiedTime(linkItem.FullName).ToUniversalTime();
                    if (modTimeUtc != stamp.UtcDateTime)
                    {
                        tarGzFilePath = Path.Combine(config.ExternalSourcesFolder, _ianaTarGzLink);
                        if (!ftpClient.DownloadFile(tarGzFilePath, linkItem.FullName, true))
                        {
                            throw new Exception($"FluentFTP.DownloadFile({tarGzFilePath}, {linkItem.FullName}, true) failed");
                        }

                        stamp = modTimeUtc;
                        return Result.NewDataReady;
                    }
                }
                catch (Exception ex)
                {
                    tarGzFilePath = null;
                    logger.Log(LogType.Error, $"CheckUpdateIANA exception: {ex.Message}");
                    return Result.DownloadError;
                }
                finally
                {
                    ftpClient.Disconnect();
                }
            }

            return Result.NoChanges;
        }

        /// <summary>
        /// Response Object class  which is used to enable async/await 
        /// pattern to call dependent libraries. This can replace ref and out when
        /// using async.
        /// </summary>
        internal class UpdateResponse
        {
            public Result Result { get; set; }

            public DateTimeOffset TimeStamp { get; set; }

            public string ZipFilePath { get; set; }
        }
    }
}
