﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------

namespace Microsoft.Azure.LocationServices.TimeZonesUnitTests.TimeZoneDb
{
    using System.Collections.Generic;
    using FluentAssertions;
    using Microsoft.Azure.LocationServices.TimeZoneDb;
    using Microsoft.VisualStudio.TestTools.UnitTesting;

    [TestClass]
    public class TimeZoneQueryResultUnitTests
    {
        private const string sampleTzId = "Asia%252FBahrain";
        private const string sampleIanaId = "Asia/Qatar";
        private const string unsupportedLanguageCode = "No supported language code supplied";

        [TestMethod, TestCategory("L0")]
        public void NamesShouldBeEmptyIfLanguageCodeOptionsAreNull()
        {
            var tzQueryParams = new TimeZoneQueryParams(null, null);
            var tzResult = new TimeZoneQueryResult();

            tzResult.Id = sampleIanaId;
            tzResult.SetTimezoneNames(tzQueryParams);

            ValidateNames(tzResult, unsupportedLanguageCode, string.Empty, string.Empty, string.Empty);
        }

        [TestMethod, TestCategory("L0")]
        public void NamesShouldBeEmptyIfLanguageCodeOptionsAreEmpty()
        {
            var tzQueryParams = new TimeZoneQueryParams(null, new List<string>());
            var tzResult = new TimeZoneQueryResult();

            tzResult.Id = sampleIanaId;
            tzResult.SetTimezoneNames(tzQueryParams);

            ValidateNames(tzResult, unsupportedLanguageCode, string.Empty, string.Empty, string.Empty);
        }

        [TestMethod, TestCategory("L0")]
        public void NamesShouldBeEmptyIfLanguageCodeOptionsAreInvalid()
        {
            var tzQueryParams = new TimeZoneQueryParams(null, new List<string> { "enus" });
            var tzResult = new TimeZoneQueryResult();

            tzResult.Id = sampleIanaId;
            tzResult.SetTimezoneNames(tzQueryParams);

            ValidateNames(tzResult, unsupportedLanguageCode, string.Empty, string.Empty, string.Empty);
        }

        [TestMethod, TestCategory("L0")]
        public void NamesShouldBeEmptyIfLanguageCodeOptionsAreNotSupported()
        {
            var tzQueryParams = new TimeZoneQueryParams(null, new List<string> { "co" });
            var tzResult = new TimeZoneQueryResult();

            tzResult.Id = sampleIanaId;
            tzResult.SetTimezoneNames(tzQueryParams);

            ValidateNames(tzResult, unsupportedLanguageCode, string.Empty, string.Empty, string.Empty);
        }

        [TestMethod, TestCategory("L0")]
        public void NamesShouldBeEmptyIfSupportedLanguageCodeOptionsHaveWeights()
        {
            var tzQueryParams = new TimeZoneQueryParams(null, new List<string> { "fr;q=0.9" });
            var tzResult = new TimeZoneQueryResult();

            tzResult.Id = sampleIanaId;
            tzResult.SetTimezoneNames(tzQueryParams);

            ValidateNames(tzResult, unsupportedLanguageCode, string.Empty, string.Empty, string.Empty);
        }

        [TestMethod, TestCategory("L0")]
        public void NamesShouldBeNonEmptyIfAtleastOneOfTheLanguageCodeOptionsIsSupported()
        {
            var tzQueryParams = new TimeZoneQueryParams(null, new List<string> { "co", "en" });
            var tzResult = new TimeZoneQueryResult();

            tzResult.Id = sampleIanaId;
            tzResult.SetTimezoneNames(tzQueryParams);

            ValidateNames(tzResult, "en", "Arabian Time", "Arabian Standard Time", "Arabian Daylight Time");
        }

        [TestMethod, TestCategory("L0")]
        public void NamesShouldBeNonEmptyIfTheLanguageCodeOptionAlsoContainsTheRegion()
        {
            var tzQueryParams = new TimeZoneQueryParams(null, new List<string> { "en-us" });
            var tzResult = new TimeZoneQueryResult();

            tzResult.Id = sampleIanaId;
            tzResult.SetTimezoneNames(tzQueryParams);

            ValidateNames(tzResult, "en", "Arabian Time", "Arabian Standard Time", "Arabian Daylight Time");
        }

        [TestMethod, TestCategory("L0")]
        public void NamesShouldCorrespondToTheFirstSupportedLanguageCodeOption()
        {
            var tzQueryParams = new TimeZoneQueryParams(null, new List<string> { "co", "en", "fr" });
            var tzResult = new TimeZoneQueryResult();

            tzResult.Id = sampleIanaId;
            tzResult.SetTimezoneNames(tzQueryParams);

            ValidateNames(tzResult, "en", "Arabian Time", "Arabian Standard Time", "Arabian Daylight Time");
        }

        [TestMethod, TestCategory("L0")]
        public void NamesShouldDefaultToEnIfWildcardIsSpecified()
        {
            var tzQueryParams = new TimeZoneQueryParams(null, new List<string> { "*" });
            var tzResult = new TimeZoneQueryResult();

            tzResult.Id = sampleIanaId;
            tzResult.SetTimezoneNames(tzQueryParams);

            ValidateNames(tzResult, "en", "Arabian Time", "Arabian Standard Time", "Arabian Daylight Time");
        }

        private void ValidateNames(TimeZoneQueryResult tzResult, string languageCode, string genericName, string standardName, string daylightName)
        {
            tzResult.Names.ISO6391LanguageCode.Should().Be(languageCode);
            tzResult.Names.Generic.Should().Be(genericName);
            tzResult.Names.Standard.Should().Be(standardName);
            tzResult.Names.Daylight.Should().Be(daylightName);
        }
    }
}
