﻿// ---------------------------------------------------------------------------------
// <copyright company="Microsoft">
//   Copyright (c) 2017 Microsoft Corporation. All rights reserved.
// </copyright>
// ---------------------------------------------------------------------------------
// 

namespace tzDemo
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Globalization;
    using System.IO;
    using System.Linq;
    using System.Threading;
    using System.Threading.Tasks;

    using Microsoft.Azure.LocationServices.Common;
    using Microsoft.Azure.LocationServices.Geography;
    using Microsoft.Azure.LocationServices.TimeZoneDb;
    using Microsoft.Azure.LocationServices.TimeZoneDbManager;

    using Newtonsoft.Json;

    public enum Platform
    {
        si = 0,
        sql = 1,
        post = 2,
        netop = 3,
        _max
    }

    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.OrderingRules", "SA1201:ElementsMustAppearInTheCorrectOrder")]
    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.OrderingRules", "SA1202:ElementsMustBeOrderedByAccess")]
    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.OrderingRules", "SA1204:StaticElementsMustAppearBeforeInstanceElements")]
    [System.Diagnostics.CodeAnalysis.SuppressMessage("StyleCop.CSharp.LayoutRules", "SA1503:CurlyBracketsMustNotBeOmitted")]
    [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1001:TypesThatOwnDisposableFieldsShouldBeDisposable")]
    class Program
    {
        const int maxPlatforms = (int)Platform._max;
        static readonly JsonSerializerSettings s_JsonSerializeSettings = new JsonSerializerSettings() { NullValueHandling = NullValueHandling.Ignore };
        static readonly string dataFolder = @"D:\";

        static TimeZoneDb timeZoneDb;

        static KeyValuePair<string, TestDelegate>[] testDelegates = new KeyValuePair<string, TestDelegate>[]
          {
                new KeyValuePair<string, TestDelegate>("Exit", null), //0 = Exit
                new KeyValuePair<string, TestDelegate>("EnumerateTimeZoneIds", new TestDelegate(EnumerateTimeZoneIds)),
                new KeyValuePair<string, TestDelegate>("EnumerateWondowsZones", new TestDelegate(EnumerateWondowsZones)),
                new KeyValuePair<string, TestDelegate>("WindowsToIanaTimeZone", new TestDelegate(WindowsToIanaTimeZone)),
                new KeyValuePair<string, TestDelegate>("FindTimeZoneById", new TestDelegate(FindTimeZoneById)),
                new KeyValuePair<string, TestDelegate>("FindBySpatialScan", new TestDelegate(TestAzureTzDbSpatial)),
                new KeyValuePair<string, TestDelegate>("TimeConversion", new TestDelegate(TimeConversion)),
                new KeyValuePair<string, TestDelegate>("PosixTest", new TestDelegate(PosixTest)),
                new KeyValuePair<string, TestDelegate>("CompileNzd", new TestDelegate(CompileNzd)),
                new KeyValuePair<string, TestDelegate>("RunTimeZoneDbApiSelfTest", new TestDelegate(RunTimeZoneDbApiSelfTest)),
                new KeyValuePair<string, TestDelegate>("TestLocalizedNamesForAllTimeZones", new TestDelegate(TestLocalizedNamesForAllTimeZones)),
#if RESTORE_ME
                new KeyValuePair<string, TestDelegate>("PointInPolyTest", new TestDelegate(PointInPolyTest)),
#endif
          };

        delegate void TestDelegate(string fnName);
#if RESTORE_ME
        delegate TimeZoneDetails FindDelegate(double y, double x);
#endif
        static void Main(string[] args)
        {
            try
            {
                var cfg = ConfigValues.CreateForUnsafeMode(dataFolder);

                Console.WriteLine("Initializing TimeZoneDb...");
                DateTime start = DateTime.Now;
                Task<TimeZoneDb> tzdbTask = TimeZoneDb.CreateInstance(cfg, TimeZoneLogEventCallback, 300);
                var ts = DateTime.Now - start;
                Console.WriteLine($" - CreateInstance call returned in {ts.TotalSeconds.ToString("0.00")} sec.");
                Console.WriteLine("\tTimeZoneDb.IsQueriable = {0}", TimeZoneDb.IsQueriable);

                while (!tzdbTask.IsCompleted)
                    Thread.Sleep(50);
                timeZoneDb = tzdbTask.Result;

                ts = DateTime.Now - start;
                Console.WriteLine($" - CreateInstance task completed in in {ts.TotalSeconds.ToString("0.00")} sec.");
                Console.WriteLine("\tTimeZoneDb.IsQueriable = {0}", TimeZoneDb.IsQueriable);
            }
            catch (Exception e)
            {
                Console.WriteLine($"Error: {e.Message}");
                return;
            }

            DisplayTestOptions();

            string input = string.Empty;
            while (true)
            {
                input = Console.ReadLine();
                int test = 0;
                if (!int.TryParse(input, out test))
                    continue;

                try
                {
                    if (test >= testDelegates.Length)
                    {
                        DisplayTestOptions();
                    }
                    else if (testDelegates[test].Value == null)
                    {
                        timeZoneDb.Dispose();
                        return;
                    }
                    else
                    {
                        testDelegates[test].Value(testDelegates[test].Key);
                    }
                }
                catch (Exception ex)
                {
                    Console.WriteLine("Error: {0}. Inner: {1}", ex.Message, ex.InnerException == null ? "none" : ex.InnerException.Message);
                }

                DisplayTestOptions();
            }
        }

        static void TimeZoneLogEventCallback(object sender, LogEventArgs e)
        {
            Console.WriteLine($" - LogEvent: {e}");
        }

        static void DisplayTestOptions()
        {
            Console.WriteLine();
            Console.WriteLine("=================================================================");
            for (int i = 0; i < testDelegates.Length; i++)
                Console.WriteLine($" {i}. {testDelegates[i].Key}");
            Console.WriteLine();
            Console.Write("Run test? ");
        }

        static void PrintTestHeader(string fnName)
        {
            Console.WriteLine("----------------------------------------------------------------------------------------------");
            Console.WriteLine($"\tTest: {fnName}, tzdb version: {timeZoneDb.Version.Version}");
            Console.WriteLine();
        }

        static bool GetTimeStamp(out DateTime stamp)
        {
            var defaultStamp = DateTime.UtcNow;
            stamp = defaultStamp;
            Console.Write($"Timestamp [{Utilities.FromDateTime(defaultStamp)} UTC]? ");
            var tmp = Console.ReadLine();
            if (!string.IsNullOrEmpty(tmp))
            {
                if (!DateTime.TryParse(tmp, CultureInfo.InvariantCulture, DateTimeStyles.AdjustToUniversal | DateTimeStyles.AssumeUniversal, out stamp))
                {
                    Console.WriteLine("Error!  Cannot parse Timestamp to a valid DateTime");
                    return false;
                }
            }

            return true;
        }

        static void EnumerateTimeZoneIds(string fnName)
        {
            PrintTestHeader(fnName);

            int cnt = 0;
            foreach (var tz in timeZoneDb.EnumerateTimeZoneIds())
            {
                ++cnt;
                Console.WriteLine($" {cnt.ToString("000")}: {tz}");
            }
        }

        static void TestLocalizedNamesForAllTimeZones(string fnName)
        {
            PrintTestHeader(fnName);

            int cnt = 0;
            var timeZoneIds = timeZoneDb.EnumerateTimeZoneIds();
            foreach (var tz in timeZoneIds)
            {
                ++cnt;
                var timeZoneData = timeZoneDb.FindTimeZone(tz.Id, new TimeZoneQueryParams(null, "en-us"));
                var names = timeZoneData.TimeZones.First().Names;
                if (string.IsNullOrEmpty(names.Generic) || string.IsNullOrEmpty(names.Standard) || string.IsNullOrEmpty(names.Daylight))
                {
                    throw new Exception($"Expected valid values for the localized timezone name for the timezoneId: {tz.Id}");
                }

                var title = $" {cnt.ToString("000")}: {tz}";
                Console.Write(title);
                for (var i = 0; i < Math.Max(48 - title.Length, 1); i++)
                    Console.Write(" ");
                Console.Write($"{names.Generic}");
                for (var i = 0; i < Math.Max(48 - names.Generic.Length, 1); i++)
                    Console.Write(" ");
                Console.Write($"{names.Standard}");
                for (var i = 0; i < Math.Max(48 - names.Standard.Length, 1); i++)
                    Console.Write(" ");
                Console.Write($"{names.Daylight}");
                Console.WriteLine();
            }
        }

        static void EnumerateWondowsZones(string fnName)
        {
            PrintTestHeader(fnName);

            var list = timeZoneDb.EnumerateWindowsTimeZones();
            var jString = JsonConvert.SerializeObject(list, Formatting.Indented, s_JsonSerializeSettings);
            Console.WriteLine(jString);
            Console.WriteLine();
            Console.WriteLine($"Windows zones count: {list.Count}");
        }

        static void WindowsToIanaTimeZone(string fnName)
        {
            PrintTestHeader(fnName);

            int cnt = 0;
            foreach (var wz in timeZoneDb.EnumerateWindowsTimeZones())
            {
                ++cnt;
                var s = $"{cnt.ToString("000")}. {wz.WindowsId}, {wz.Territory}";
                Console.Write(s);
                int spaces = 45 - s.Length;

                bool first = true;
                foreach (var tz in timeZoneDb.WindowsToIanaTimeZone(wz.WindowsId, wz.Territory))
                {
                    for (int i = 0; i < spaces; i++)
                        Console.Write(" ");
                    if (first)
                    {
                        Console.Write(" => ");
                        spaces = 49;
                        first = false;
                    }

                    Console.WriteLine(tz);
                }
            }
        }

        static void FindTimeZoneById(string fnName)
        {
            PrintTestHeader(fnName);

            DateTime utcStamp;
            if (!GetTimeStamp(out utcStamp))
            {
                return;
            }

            Console.WriteLine($"Timestamp: {Utilities.FromDateTime(utcStamp)} UTC");

            var resultCache = new Dictionary<IanaId, TimeZoneQueryResults>();
            var sw = new Stopwatch();
            int cnt = 0;
            long msTotal = 0;
            foreach (var tz in timeZoneDb.EnumerateTimeZoneIds())
            {
                sw.Restart();
                var results = timeZoneDb.FindTimeZone(tz.Id, new TimeZoneQueryParams(utcStamp, "en-us"));
                sw.Stop();
                msTotal += sw.ElapsedMilliseconds;
                ++cnt;
                resultCache.Add(tz, results);
            }

            Console.WriteLine();
            Console.WriteLine($"Queries: {cnt}, average exec time: {msTotal / cnt} ms");

            do
            {
                Console.WriteLine();
                Console.Write("Output result? By Id [I], Time [T], Offset [O] or [X] to exit: ");
                switch (Console.ReadLine().ToUpperInvariant())
                {
                    case "I":
                        foreach (var kvp in resultCache.OrderBy(g => g.Key.Id))
                            OutputResults(kvp.Value, kvp.Key);
                        break;
                    case "O":
                        foreach (var kvp in resultCache.OrderBy(g => (g.Value.TimeZones == null) ? TimeSpan.FromDays(2) : g.Value.TimeZones[0].ReferenceTime.StandardOffset))
                            OutputResults(kvp.Value, kvp.Key);
                        break;
                    case "T":
                        foreach (var kvp in resultCache.OrderBy(g => (g.Value.TimeZones == null) ? long.MaxValue : g.Value.TimeZones[0].ReferenceTime.WallTime.Ticks))
                            OutputResults(kvp.Value, kvp.Key);
                        break;
                    case "X":
                        return;
                }
            }
            while (true);
        }

        static void TestAzureTzDbSpatial(string fnName)
        {
            PrintTestHeader(fnName);

#if DEBUG
            double increment = 0.750;
#else
            double increment = 0.050;
#endif

            Console.Write($"Spatial increment scan [{increment}]: ");
            var input = Console.ReadLine();
            if (!string.IsNullOrEmpty(input))
            {
                if (!double.TryParse(input, out increment))
                {
                    Console.WriteLine($"Error: {input} cannot be parsed to a double");
                    return;
                }
            }

            Console.WriteLine($"Using {increment.ToString("0.000")} as spatial increment");

            HashSet<string> visited = new HashSet<string>();
            Dictionary<int, List<TimeZoneQueryResult>> resultsByOffset = new Dictionary<int, List<TimeZoneQueryResult>>();

            var tzdbParams = new TimeZoneQueryParams(null);

            int msTotal = 0;
            var taskList = new List<Task>();
            var syncLock = new object();
            var coords = new Coordinates(-90, -180);
            var coordList = new List<Coordinates>();
            while (true)
            {
                if (coords.Y + increment > 90.0)
                    break;
                coords.Y += increment;
                coords.X = -180;

                while (true)
                {
                    if (coords.X + increment > 180.0)
                        break;
                    coords.X += increment;
                    coordList.Add(coords);
                }
            }

            Console.WriteLine($"Running {coordList.Count} concurrent spatial queries");
            for (int i = 0; i < coordList.Count; i++)
            {
                var pt = coordList[i];
                var task = Task.Run(
                    delegate
                    {
                        try
                        {
                            var start = DateTime.Now;
                            var results = timeZoneDb.FindTimeZone(ref pt, tzdbParams);
                            var ts = DateTime.Now - start;

                            if (results.Count == 0)
                                Console.WriteLine($"At {pt} *** no results ***");
                            else
                            {
                                if (results.Count > 1)
                                    Console.WriteLine($"*** {results.Count} results *** at {pt}");

                                for (int r = 0; r < results.Count; r++)
                                {
                                    var rec = results.TimeZones[r];
                                    if (!visited.Contains(rec.Id))
                                    {
                                        lock (syncLock)
                                        {
                                            if (!visited.Contains(rec.Id))
                                            {
                                                Console.WriteLine($"At {pt} - {rec}");
                                                visited.Add(results.TimeZones[0].Id);
                                                Interlocked.Add(ref msTotal, (int)ts.TotalMilliseconds);

                                                List<TimeZoneQueryResult> list;
                                                if (!resultsByOffset.TryGetValue((int)rec.ReferenceTime.StandardOffset.TotalSeconds, out list))
                                                {
                                                    list = new List<TimeZoneQueryResult>();
                                                    resultsByOffset.Add((int)rec.ReferenceTime.StandardOffset.TotalSeconds, list);
                                                }

                                                list.Add(rec);
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        catch (Exception ex)
                        {
                            Console.Write($"*** Error *** {ex.Message}");
                            if (ex.InnerException != null)
                            {
                                Console.WriteLine($" - inner ex: {ex.InnerException.Message}");
                            }
                            else
                            {
                                Console.WriteLine();
                            }
                        }
                    });
                taskList.Add(task);
            }

            Console.WriteLine($"Waiting for {taskList.Count} tasks to complete...");
            Task.WaitAll(taskList.ToArray());

            Console.WriteLine();
            Console.WriteLine($"Queries = {coordList.Count}   Unique hits = {visited.Count}   Average response = {msTotal / (float)visited.Count} ms");
            Console.WriteLine();

            Directory.CreateDirectory(@"c:\Temp");
            string fileName = @"c:\Temp\tzSpatialTest.tab";
            using (var w = new StreamWriter(fileName))
            {
                w.WriteLine("StandardOffset\tSavings\tWallDate\tWallTime\tPosixTz\tRefPoint\tTimeZone");
                foreach (var kvp in resultsByOffset.OrderBy(g => g.Key))
                {
                    Console.WriteLine($"Offset = {kvp.Key / 3600}");
                    foreach (var r in kvp.Value.OrderBy(g => g.RepresentativePoint.HasValue ? g.RepresentativePoint.Value.X : int.MinValue))
                    {
                        var zt = r.ReferenceTime;
                        string s = $"{zt.StandardOffset.TotalHours}"
                            + $"\t{zt.DaylightSavings.TotalHours}"
                            + $"\t{zt.WallTime.Date.ToShortDateString()}"
                            + $"\t{zt.WallTime.TimeOfDay.Hours}:{zt.WallTime.TimeOfDay.Minutes}"
                            + $"\t{zt.PosixTz}"
                            + $"\t{r.RepresentativePoint}"
                            + $"\t{r.Id}";
                        Console.WriteLine(s);
                        w.WriteLine(s);
                    }
                }
            }

            Console.WriteLine();
            Console.WriteLine($"----- Contents written to: {fileName} -----");
        }

        static void TimeConversion(string fnName)
        {
            PrintTestHeader(fnName);

            var defaultTzId = "America/Los_Angeles";
            var defaulCoords = new Coordinates(47.0, -121.0);
            var additionalInfo = 0;

            Console.Write($"Additonal info None=[0], ZoneInfo=[1], Transitions=[2], All=[3]? ");
            var tmp = Console.ReadLine();
            if (!int.TryParse(tmp, out additionalInfo))
            {
                switch (additionalInfo)
                {
                    case 0:
                    case 1:
                    case 2:
                    case 3:
                        break;
                    default:
                        Console.WriteLine("Error: invalid option");
                        return;
                }
            }

            DateTime utcStamp;
            if (!GetTimeStamp(out utcStamp))
                return;

            string findMode = string.Empty;
            while (findMode != "0" && findMode != "1")
            {
                Console.Write("Find by [0]=ID, [1]=Location, [X]=Exit ? ");
                findMode = Console.ReadLine();
                if (findMode.ToLowerInvariant() == "x")
                    return;
            }

            var options = new TimeZoneQueryParams((TimeZoneQueryParams.AdditionalData)additionalInfo, utcStamp, null, null);
            TimeZoneQueryResults results = null;
            switch (findMode)
            {
                case "0":
                    Console.Write($"Timezone ID [{defaultTzId}]? ");
                    var tzId = Console.ReadLine();
                    if (string.IsNullOrEmpty(tzId))
                        tzId = defaultTzId;
                    results = timeZoneDb.FindTimeZone(tzId, options);
                    break;

                case "1":
                    var coords = defaulCoords;
                    Console.Write($"Coordinates [{defaulCoords}]: ");
                    tmp = Console.ReadLine();
                    if (!string.IsNullOrEmpty(tmp))
                    {
                        coords = new Coordinates(tmp);
                        if (coords.IsNull)
                        {
                            Console.WriteLine("Invalid coordinates");
                            return;
                        }
                    }

                    results = timeZoneDb.FindTimeZone(ref coords, options);
                    break;

                default:
                    return;
            }

            OutputResults(results);
        }

        static void PosixTest(string fnName)
        {
            PrintTestHeader(fnName);

            var tests = new KeyValuePair<string, string>[]
                {
                new KeyValuePair<string, string>("America/New_York", "EST5EDT,M3.1.0,M11.1.0"),
                new KeyValuePair<string, string>("Australia/Sydney", "AEST-10AEDT,M10.1.0,M4.1.0/3"),
                new KeyValuePair<string, string>("America/Sao_Paulo", "BRT3BRST,M10.3.0/0,M2.3.0/0"),
                new KeyValuePair<string, string>("Europe/London", "GMT0BST,M3.3.0/1,M10.5.0"),
                new KeyValuePair<string, string>("Australia/Lord_Howe", "LHST-10:30LHDT-11,M10.1.0,M4.1.0"),
                new KeyValuePair<string, string>("Pacific/Chatham", "CHAST-12:45CHADT,M9.3.0/2:45,M4.1.0/3:45"),
                new KeyValuePair<string, string>("Europe/Astrakhan", "<+03>-3<+04>,M3.3.0,J365/25"),
                };

            var utcStamp = DateTime.Parse("2016-03-01T00:00:00Z", CultureInfo.InvariantCulture, DateTimeStyles.AssumeUniversal | DateTimeStyles.AdjustToUniversal);
            Debug.Assert(utcStamp.Kind == DateTimeKind.Utc);

            foreach (var kvp in tests)
            {
                var results = timeZoneDb.FindTimeZone(kvp.Key, new TimeZoneQueryParams(utcStamp));
                if (results.Count == 0)
                {
                    if (results.Error != null)
                        Console.WriteLine($"Error: {results.Error.Message}");
                    else
                        Console.WriteLine("no results");
                }
                else if (results.Count == 1)
                {
                    var tzResult = results.TimeZones[0];
                    if (string.IsNullOrEmpty(tzResult.ReferenceTime.PosixTz) || tzResult.ReferenceTime.PosixTz != kvp.Value)
                    {
                        string s = $"Fail: {kvp.Key}";
                        Console.Write(s);
                        int i = s.Length;
                        for (; i < 35; i++) Console.Write(" ");
                        Console.WriteLine($"{tzResult.ReferenceTime.PosixTz}");

                        s = $"          instead of";
                        Console.Write(s);
                        i -= s.Length;
                        for (; i > 0; i--) Console.Write(" ");
                        Console.WriteLine($"{kvp.Value}");
                    }
                    else
                    {
                        Console.Write($"Pass: {kvp.Key}");
                        for (int i = 0; i < 30 - kvp.Key.Length; i++) Console.Write(" ");
                        Console.WriteLine($"{tzResult.ReferenceTime.PosixTz}");
                    }
                }
                else
                {
                    Console.WriteLine($"Multiple results for {kvp.Key}");
                }
            }
        }

        static void CompileNzd(string fnName)
        {
            PrintTestHeader(fnName);

            var srcIana = @"D:\Data\TimeZone\tzdata\2017b";
            var srcCldr = @"D:\Data\TimeZone\CLDR";
            var outFolder = @"D:\tzCompilerTest";

            Console.Write($"IANA source [{srcIana}]: ");
            var s = Console.ReadLine();
            if (!string.IsNullOrEmpty(s))
                srcIana = s;

            Console.Write($"CLDR source [{srcCldr}]: ");
            s = Console.ReadLine();
            if (!string.IsNullOrEmpty(s))
                srcCldr = s;

            Console.Write($"Output folder [{outFolder}]: ");
            s = Console.ReadLine();
            if (!string.IsNullOrEmpty(s))
                outFolder = s;

            NodaTime.Utility.InvalidNodaDataException[] errors = null;
            var logger = new Logger("tzDemo");
            logger.LogEvent += TimeZoneLogEventCallback;
            var sw = new Stopwatch();
            sw.Start();
            var file = NodaTime.TzdbCompiler.Compiler.Run(logger, srcIana, srcCldr, outFolder, out errors);
            sw.Stop();
            Console.WriteLine($"Created {file} in {sw.ElapsedMilliseconds / 1000} sec.");
            if (errors != null && errors.Length > 0)
            {
                Console.WriteLine($" *** {errors.Length} validaiton errors ***");
                foreach (var err in errors)
                    Console.WriteLine($"{err.Message}");
            }
        }

        static void RunTimeZoneDbApiSelfTest(string fnName)
        {
            PrintTestHeader(fnName);

            string ianaVersion;
            IList<string> errors;

            var sw = new Stopwatch();
            sw.Start();
            bool result = timeZoneDb.TestPublicApis(out ianaVersion, out errors);
            sw.Stop();

            Console.WriteLine();
            Console.WriteLine($"TimeZoneDb TestPublicApis ran in {sw.ElapsedMilliseconds} ms.  Success = {result}");
        }

#if RESTORE_ME
       static void PointInPolyTest(string fnName)
        {
            PrintTestHeader(fnName);

            TimeZones.Initialize(@"D:\Data\TimeZone\DemoData");
            double lat = 0;
            double lon = 0;
            double count = 0;
            double[] totals = new double[maxPlatforms];
            TimeSpan[] ts = new TimeSpan[maxPlatforms];
            TimeZoneDetails[] tzd = new TimeZoneDetails[maxPlatforms];

            // warm up
            TimeZones.GetTimeZoneWithSpatialIndex(lat, lon); 
            TimeZones.GetTimeZoneWithSql(lat, lon);
            TimeZones.GetTimeZoneWithPostgreSql(lat, lon);
            TimeZones.GetTimeZoneWithNetTopologyIndex(lat, lon);

            FindDelegate[] findDelegates = new FindDelegate[maxPlatforms];
            findDelegates[(int)Platform.si] = new FindDelegate(TimeZones.GetTimeZoneWithSpatialIndex);
            findDelegates[(int)Platform.sql] = new FindDelegate(TimeZones.GetTimeZoneWithSql);
            findDelegates[(int)Platform.post] = new FindDelegate(TimeZones.GetTimeZoneWithPostgreSql);
            findDelegates[(int)Platform.netop] = new FindDelegate(TimeZones.GetTimeZoneWithNetTopologyIndex);

            DateTime start = DateTime.Now;
            for (int x = 0; x < 2; x++)
            {
                lat = 0;
                lon = 0;

                if (x > 0)
                    Console.WriteLine("\r\n================ reset ================\r\n");

                for (int y = 0; y < 15; y++)
                {
                    ++count;
                    for (int p = 0; p < findDelegates.Length; p++)
                    {
                        start = DateTime.Now;
                        tzd[p] = findDelegates[p](lat, lon);
                        ts[p] = DateTime.Now - start;
                    }

                    OutputPointInPolyResult(lat, lon, tzd, ts, totals);

                    lat += 3;
                    if (x == 0)
                        lon -= 8;
                    else if (x == 1)
                        lon += 8;
                    else
                        throw new NotImplementedException("unhandled");
                }
            }

            Console.WriteLine("Average responses:");
            for (int p = 0; p < maxPlatforms; p++)
            {
                Platform plat = (Platform)p;
                Console.WriteLine($"\t{plat}={totals[p] / count}");
            }
        }

        private static void OutputPointInPolyResult(double lat, double lon, TimeZoneDetails[] tzd, TimeSpan[] ts, double[] totals)
        {
            Console.WriteLine("----------------------------------------------------------------------");
            Console.Write($"At {lat}, {lon} the time zone is {tzd[0].IanaId}.");
            for (int p = 0; p < maxPlatforms; p++)
            {
                Console.Write($" {(Platform)p}={ts[p].TotalMilliseconds}");
                totals[p] += ts[p].TotalMilliseconds;
            }
            Console.WriteLine();

            for (int p = 0; p < maxPlatforms; p++)
            {
                if (p > 0 && tzd[p].IanaId != tzd[0].IanaId)
                    Console.WriteLine($"  *** {(Platform)p} result is {tzd[p].IanaId}");
            }

            for (int p = 0; p < maxPlatforms; p++)
            {
                if (tzd[p].AlternativeResults != null && tzd[p].AlternativeResults.Count > 0)
                {
                    Console.WriteLine($"\talt names returned by {(Platform)p} query:");
                    for (int i = 0; i < tzd[p].AlternativeResults.Count; i++)
                        Console.WriteLine($"\t\t{i}: {tzd[p].AlternativeResults[i]}");
                }
            }
            Console.WriteLine();
        }
#endif
        static void OutputResults(TimeZoneQueryResults results, IanaId tz = null)
        {
            if (results.Error != null)
            {
                Console.WriteLine($"*** Error: {results.Error} ***");
                return;
            }

            if (results.Count == 0)
            {
                Console.WriteLine("*** No Results ***");
                return;
            }

            if (results.Count > 1)
            {
                Console.Write($"*** {results.Count} results ***");
                Console.Write($" {results.TimeZones.Select(r => r.Id).Aggregate((i, j) => $"{i}, {j}")}");
                Console.WriteLine();
                return;
            }

            if (tz == null)
            {
                var jString = JsonConvert.SerializeObject(results, Formatting.Indented, s_JsonSerializeSettings);
                Console.WriteLine(jString);
            }
            else
            {
                var res = results.TimeZones[0];
                Console.Write(res.ReferenceTime);
                for (int i = 0; i < Math.Max(48 - res.ReferenceTime.ToString().Length, 1); i++)
                    Console.Write(" ");
                Console.Write(res.Names.Generic);
                for (int i = 0; i < Math.Max(48 - res.Names.Generic.ToString().Length, 1); i++)
                    Console.Write(" ");
                Console.Write(res.ReferenceTime.PosixTz);
                for (int i = 0; i < Math.Max(48 - res.ReferenceTime.PosixTz.Length, 1); i++)
                    Console.Write(" ");
                Console.Write(res.Id);
                if (tz.IsAlias)
                    Console.Write($" for alias {tz.Id}");

                if ((tz.IsAlias && tz.AliasOf != res.Id)
                || (!tz.IsAlias && tz.Id != res.Id))
                {
                    Debug.Fail("unexpected id");
                    Console.Write($" ??? unexected id for {tz.Id}");
                }

                Console.WriteLine();
            }
        }

        private static string ToTimeString(TimeSpan ts)
        {
            return $"{ts.Hours.ToString("00")}:{ts.Minutes.ToString("00")}:{ts.Seconds.ToString("00")}";
        }
    }
}
