"use strict";
// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License. See License.txt in the project root for license information.
Object.defineProperty(exports, "__esModule", { value: true });
const uuid = require("uuid/v4");
const rhea_promise_1 = require("./rhea-promise");
const amqp_common_1 = require("./amqp-common");
/**
 * Describes the methods on the EventData interface.
 * @module EventData
 */
var EventData;
(function (EventData) {
    /**
     * Converts the AMQP message to an EventData.
     * @param {AmqpMessage} msg The AMQP message that needs to be converted to EventData.
     */
    function fromAmqpMessage(msg) {
        const data = {
            body: msg.body,
            _raw_amqp_mesage: msg
        };
        if (msg.message_annotations) {
            data.annotations = msg.message_annotations;
            if (msg.message_annotations[amqp_common_1.Constants.partitionKey])
                data.partitionKey = msg.message_annotations[amqp_common_1.Constants.partitionKey];
            if (msg.message_annotations[amqp_common_1.Constants.sequenceNumber])
                data.sequenceNumber = msg.message_annotations[amqp_common_1.Constants.sequenceNumber];
            if (msg.message_annotations[amqp_common_1.Constants.enqueuedTime])
                data.enqueuedTimeUtc = new Date(msg.message_annotations[amqp_common_1.Constants.enqueuedTime]);
            if (msg.message_annotations[amqp_common_1.Constants.offset])
                data.offset = msg.message_annotations[amqp_common_1.Constants.offset];
        }
        // Since rhea expects message properties as top level properties we will look for them and unflatten them inside properties.
        for (const prop of rhea_promise_1.messageProperties) {
            if (msg[prop] !== undefined) {
                if (!data.properties) {
                    data.properties = {};
                }
                data.properties[prop] = msg[prop];
            }
        }
        // Since rhea expects message headers as top level properties we will look for them and unflatten them inside header.
        for (const prop of rhea_promise_1.messageHeader) {
            if (msg[prop] !== undefined) {
                if (!data.header) {
                    data.header = {};
                }
                data.header[prop] = msg[prop];
            }
        }
        if (msg.application_properties) {
            data.applicationProperties = msg.application_properties;
        }
        if (msg.delivery_annotations) {
            data.lastEnqueuedOffset = msg.delivery_annotations.last_enqueued_offset;
            data.lastSequenceNumber = msg.delivery_annotations.last_enqueued_sequence_number;
            data.lastEnqueuedTime = new Date(msg.delivery_annotations.last_enqueued_time_utc);
            data.retrievalTime = new Date(msg.delivery_annotations.runtime_info_retrieval_time_utc);
        }
        return data;
    }
    EventData.fromAmqpMessage = fromAmqpMessage;
    /**
     * Converts an EventData object to an AMQP message.
     * @param {EventData} data The EventData object that needs to be converted to an AMQP message.
     */
    function toAmqpMessage(data) {
        const msg = {
            body: data.body,
        };
        // As per the AMQP 1.0 spec If the message-annotations or delivery-annotations section is omitted,
        // it is equivalent to a message-annotations section containing anempty map of annotations.
        msg.message_annotations = {};
        msg.delivery_annotations = {};
        if (data.annotations) {
            msg.message_annotations = data.annotations;
        }
        if (data.properties) {
            // Set amqp message properties as top level properties, since rhea sends them as top level properties.
            for (const prop in data.properties) {
                msg[prop] = data.properties[prop];
            }
        }
        if (!msg.message_id) {
            msg.message_id = uuid();
        }
        if (data.applicationProperties) {
            msg.application_properties = data.applicationProperties;
        }
        if (data.partitionKey) {
            msg.message_annotations[amqp_common_1.Constants.partitionKey] = data.partitionKey;
            // Event Hub service cannot route messages to a specific partition based on the partition key
            // if AMQP message header is an empty object. Hence we make sure that header is always present
            // with atleast one property. Setting durable to true, helps us achieve that.
            msg.durable = true;
        }
        if (data.sequenceNumber != undefined) {
            msg.message_annotations[amqp_common_1.Constants.sequenceNumber] = data.sequenceNumber;
        }
        if (data.enqueuedTimeUtc) {
            msg.message_annotations[amqp_common_1.Constants.enqueuedTime] = data.enqueuedTimeUtc.getTime();
        }
        if (data.offset != undefined) {
            msg.message_annotations[amqp_common_1.Constants.offset] = data.offset;
        }
        if (data.lastEnqueuedOffset != undefined) {
            msg.delivery_annotations.last_enqueued_offset = data.lastEnqueuedOffset;
        }
        if (data.lastSequenceNumber != undefined) {
            msg.delivery_annotations.last_enqueued_sequence_number = data.lastSequenceNumber;
        }
        if (data.lastEnqueuedTime) {
            msg.delivery_annotations.last_enqueued_time_utc = data.lastEnqueuedTime.getTime();
        }
        if (data.retrievalTime) {
            msg.delivery_annotations.runtime_info_retrieval_time_utc = data.retrievalTime.getTime();
        }
        if (data.header) {
            // Set amqp message header as top level properties, since rhea expects them as top level properties.
            for (const prop in data.header) {
                msg[prop] = data.header[prop];
            }
        }
        return msg;
    }
    EventData.toAmqpMessage = toAmqpMessage;
})(EventData = exports.EventData || (exports.EventData = {}));
//# sourceMappingURL=eventData.js.map