import { EventHubReceiver } from "./eventHubReceiver";
import { EventHubSender } from "./eventHubSender";
import { TokenProvider, CbsClient, DataTransformer, ConnectionConfig } from "./amqp-common";
import { ManagementClient } from "./managementClient";
import { ClientOptions } from "./eventHubClient";
import { Connection, Dictionary } from "./rhea-promise";
/**
 * @interface ConnectionContext
 * Provides contextual information like the underlying amqp connection, cbs session, management session,
 * tokenProvider, senders, receivers, etc. about the EventHub client.
 */
export interface ConnectionContext {
    /**
     * @property {ConnectionConfig} config The EventHub connection config that is created after parsing the connection string.
     */
    readonly config: ConnectionConfig;
    /**
     * @property {Connection} [connection] The underlying AMQP connection.
     */
    connection: Connection;
    /**
     * @property {string} [connectionId] The amqp connection id that uniquely identifies the connection within a process.
     */
    connectionId: string;
    /**
     * @property {DataTransformer} dataTransformer A DataTransformer object that has methods named
     * - encode Responsible for encoding the AMQP message before sending it on the wire.
     * - decode Responsible for decoding the received AMQP message before passing it to the customer.
     */
    dataTransformer: DataTransformer;
    /**
     * @property {boolean} wasConnectionCloseCalled Indicates whether the close() method was
     * called on theconnection object.
     */
    wasConnectionCloseCalled: boolean;
    /**
     * @property {TokenProvider} tokenProvider The TokenProvider to be used for getting tokens for authentication for the EventHub client.
     */
    readonly tokenProvider: TokenProvider;
    /**
     * @property {Dictionary<EventHubReceiver>} receivers A dictionary of the EventHub Receivers associated with this client.
     */
    receivers: Dictionary<EventHubReceiver>;
    /**
     * @property {Dictionary<EventHubSender>} senders A dictionary of the EventHub Senders associated with this client.
     */
    senders: Dictionary<EventHubSender>;
    /**
     * @property {ManagementClient} managementSession A reference to the management session ($management endpoint) on
     * the underlying amqp connection for the EventHub Client.
     */
    managementSession?: ManagementClient;
    /**
     * @property {CbsClient} cbsSession A reference to the cbs session ($cbs endpoint) on the underlying
     * the amqp connection for the EventHub Client.
     */
    cbsSession: CbsClient;
    /**
     * @property {string} connectionLock The unqiue lock name per connection that is used to acquire the lock
     * for establishing an aqmp connection per client if one does not exist.
     */
    readonly connectionLock: string;
    /**
     * @property {string} negotiateClaimLock The unqiue lock name per connection that is used to acquire the lock
     * for negotiating cbs claim by an entity on that connection.
     */
    readonly negotiateClaimLock: string;
}
export interface ConnectionContextOptions extends ClientOptions {
    managementSessionAddress?: string;
    managementSessionAudience?: string;
}
export declare namespace ConnectionContext {
    /**
     * @property {string} userAgent The user agent string for the event hub client. Constant value: "/js-event-hubs".
     */
    const userAgent: string;
    function create(config: ConnectionConfig, options?: ConnectionContextOptions): ConnectionContext;
}
