import { Message, MessageProperties, MessageHeader, Dictionary } from "./rhea-promise";
import { EventHubMessageAnnotations } from "./amqp-common";
/**
 * Describes the structure of an event to be sent or received from the EventHub.
 * @interface EventData
 */
export interface EventData {
    /**
     * @property {MessageHeader} [header] - The message headers.
     */
    header?: MessageHeader;
    /**
     * @property {any} body - The message body that needs to be sent or is received.
     */
    body: any;
    /**
     * @property {Date} [enqueuedTimeUtc] The enqueued time of the event.
     */
    enqueuedTimeUtc?: Date;
    /**
     * @property {string | null} [partitionKey] If specified EventHub will hash this to a partitionId.
     * It guarantees that messages end up in a specific partition on the event hub.
     */
    partitionKey?: string | null;
    /**
     * @property {string} [offset] The offset of the event.
     */
    offset?: string;
    /**
     * @property {number} [sequenceNumber] The sequence number of the event.
     */
    sequenceNumber?: number;
    /**
     * @property {AmqpMessageAnnotations} [annotations] The amqp message attributes.
     */
    annotations?: EventHubMessageAnnotations;
    /**
     * @property {AmqpMessageProperties} [properties] The predefined AMQP properties like message_id, correlation_id, reply_to, etc.
     */
    properties?: MessageProperties;
    /**
     * @property {Dictionary<any>} [applicationProperties] The application specific properties.
     */
    applicationProperties?: Dictionary<any>;
    /**
     * @property {number} [lastSequenceNumber] The last sequence number of the event within the partition stream of the Event Hub.
     */
    lastSequenceNumber?: number;
    /**
     * @property {string} [lastEnqueuedOffset] The offset of the last enqueued event.
     */
    lastEnqueuedOffset?: string;
    /**
     * @property {Date} [lastEnqueuedTime] The enqueued UTC time of the last event.
     */
    lastEnqueuedTime?: Date;
    /**
     * @property {Date} [retrievalTime] The time when the runtime info was retrieved
     */
    retrievalTime?: Date;
    /**
     * @property {AmqpMessage} _raw_amqp_mesage The underlying raw amqp message.
     */
    _raw_amqp_mesage?: Message;
}
/**
 * Describes the methods on the EventData interface.
 * @module EventData
 */
export declare namespace EventData {
    /**
     * Converts the AMQP message to an EventData.
     * @param {AmqpMessage} msg The AMQP message that needs to be converted to EventData.
     */
    function fromAmqpMessage(msg: Message): EventData;
    /**
     * Converts an EventData object to an AMQP message.
     * @param {EventData} data The EventData object that needs to be converted to an AMQP message.
     */
    function toAmqpMessage(data: EventData): Message;
}
