// Compiled with TypeScript Version 2.3.3

// FILE: Fx\Ajax.d.ts
declare module "Fx/Ajax" {
    export = Main;
    module Main {
        import FxBase = MsPortalFx.Base;
        import Net2 = FxBase.Net2;
        /**
         * The contract of settings input into ajax calls.
         */
        type NetAjaxSettings<T> = Net2.NetAjaxSettings<T>;
        /**
         * The contract for the batch settings.
         */
        interface BatchMultipleSettings {
            /**
             * The endpoint to make the request to.
             */
            readonly endpoint: string;
            /**
             * The list of batch requests. All URIs have to be relative URIs in the request.
             */
            readonly batchRequests: ReadonlyArray<BatchRequest>;
            /**
             * Determines whether the ajax request is part of a background task.
             * If true the batch request will be pushed on to the background queue.
             */
            readonly isBackgroundTask?: boolean;
            /**
             * Determines whether to append a telemetry header for the ARM calls.
             *
             * Set to a non-empty string to append the header. The value should be 60 characters or less and will be trimmed
             * if longer.
             */
            readonly telemetryHeader?: string;
        }
        /**
         * Response for a request within a batch.
         */
        interface BatchResponseItem<T> {
            /**
             * The response content. Can be success or failure.
             */
            readonly content: T;
            /**
             * The response headers.
             */
            readonly headers: ReadonlyStringMap<string>;
            /**
             * The response status code.
             */
            readonly httpStatusCode: number;
        }
        /**
         * Batch response.
         */
        interface BatchResponse {
            /**
             * The success response from ARM.
             */
            readonly responses: ReadonlyArray<BatchResponseItem<any>>;
        }
        /**
         * Individual batch request.
         */
        interface BatchRequest {
            /**
             * The URI to call.
             */
            readonly uri: string;
            /**
             * The http method for the call. Defaults to GET
             */
            readonly httpMethod?: string;
            /**
             * Optional request details.
             */
            readonly requestHeaderDetails?: {
                /**
                 * The command name.
                 */
                readonly commandName?: string;
            };
        }
        /**
         * Calls the API by batching multiple requests together.
         * Use this API if you have a single AJAX request but there is potential for batching this
         * with other requests.
         *
         * There are a few limitations when using this:
         *   - All requests must have absolute URIs
         *   - The max number of requests in a batch call is 20
         *   - Supports tenant level operations (/subscriptions, /resources, /locations, /providers, /tenants)
         *   - Allowed methods are - GET, POST, DELETE, HEAD without a request body
         *
         * @param settings The settings to use to call batch.
         * @return A promise for the batch call.
         */
        function batch<T>(settings: NetAjaxSettings<T>, serverEndpoint?: string): Q.Promise<BatchResponseItem<T>>;
        /**
         * Performs a batch ajax request using the given set of URIs.
         * This API is recommended if you have a set of URIs that can be called concurrently using batch.
         * There are a few limitations when using this:
         *   - All requests must have absolute URIs
         *   - The max number of requests in a batch call is 20
         *   - Supports tenant level operations (/subscriptions, /resources, /locations, /providers, /tenants)
         *   - Allowed methods are - GET, POST, DELETE, HEAD without a request body
         *
         * @param settings The settings that are to be passed to the batch call.
         * @return A promise for the batch call.
         */
        function batchMultiple(settings: BatchMultipleSettings): Q.Promise<BatchResponse>;
        /**
         * Performs an ajax request with the appropriate headers set.
         *
         * @param settings The settings that are to be passed to the ajax call.
         * @return A promise for the ajax call.
         */
        function postJSON<T>(settings: NetAjaxSettings<T>, data: any): Q.Promise<T>;
        /**
         * Performs an ajax request using caching with the appropriate headers set.
         *
         * @param settings The settings that are to be passed to the ajax call.
         * @return A promise for the ajax call.
         */
        function ajax<T>(settings: NetAjaxSettings<T>): Q.Promise<T>;
    }
}

// FILE: Fx\Assets\Assets.d.ts
declare module "Fx/Assets/Assets" {
    export = Main;
    module Main {
        import FxAssets = MsPortalFx.Assets;
        /**
         * Finds the asset type information for the given resource type from the collection of asset types.
         * NOTE: The asset types must come from the API MsPortalFx.Assets.getResourceTypeAssetTypeInformation().
         *
         * @param {string} resourceType The resource type to find.
         * @param {FxAssets.ResourceTypeAssetTypeInformation[]} assetTypes The array of asset type information from the getResourceTypeAssetTypeInformation API.
         * @returns {FxAssets.ResourceTypeAssetTypeInformation} The result asset type information if found, otherwise undefined.
         */
        function findAssetTypeInformation(resourceType: string, assetTypes: FxAssets.ResourceTypeAssetTypeInformation[]): FxAssets.ResourceTypeAssetTypeInformation;
        /**
         * Finds the asset type information for the given resource ID from the collection of asset types.
         * NOTE: The asset types must come from the API MsPortalFx.Assets.getResourceTypeAssetTypeInformation().
         *
         * @param {string} resourceID The resource ID for the resource type to find.
         * @param {FxAssets.ResourceTypeAssetTypeInformation[]} assetTypes The array of asset type information from the getResourceTypeAssetTypeInformation API.
         * @returns {FxAssets.ResourceTypeAssetTypeInformation} The result asset type information if found, otherwise undefined.
         */
        function findAssetTypeInformationFromResourceId(resourceId: string, assetTypes: FxAssets.ResourceTypeAssetTypeInformation[]): FxAssets.ResourceTypeAssetTypeInformation;
    }
}

// FILE: Fx\Base\Base.Net.Batch.d.ts
declare module "Fx/Base/Base.Net.Batch" {
    module Main {
        /**
         * This module is obsolete. Please use "Fx/Ajax"
         */
        type BatchMultipleSettings = Obsolete;
        /**
         * This module is obsolete. Please use "Fx/Ajax"
         */
        type BatchRequest = Obsolete;
        /**
         * This module is obsolete. Please use "Fx/Ajax"
         */
        type BatchResponse = Obsolete;
        /**
         * This module is obsolete. Please use "Fx/Ajax"
         */
        type BatchResponseItem<T> = Obsolete;
        /**
         * This module is obsolete. Please use "Fx/Ajax"
         */
        let batch: Obsolete;
        /**
         * This module is obsolete. Please use "Fx/Ajax"
         */
        let batchMultiple: Obsolete;
    }
    export = Main;
}

// FILE: Fx\ClientIp.d.ts
declare module "Fx/ClientIp" {
    const _default: () => Q.Promise<string>;
    export = _default;
}

// FILE: Fx\Composition\Blade.d.ts
declare module "Fx/Composition/Blade" {
    import FxComposition = require("Fx/Composition");
    import { DialogManagement } from "Fx/Composition/Dialog";
    import FxBladeBase = require("Fx/Composition/BladeBase");
    import FxViewModelBase = require("Fx/Composition/ViewModelBase");
    import * as Provisioning from "Fx/ResourceManagement/Provisioning";
    export = MsPortalFxBlade;
    module MsPortalFxBlade {
        import PartReference = MsPortalFx.Composition.PartReference;
        /**
         * The @Blade decorator.  Identifies a class within the extension as a Blade view model class.
         *
         * @options Metadata describing the Blade and how it will be treated by the FX.
         */
        function Decorator(options?: Options): (bladeClass: BladeClass) => void;
        /**
         * Constrains the @Blade decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface BladeClass extends FxBladeBase.BladeClass {
            new (): Contract<any, any>;
        }
        /**
         * Constrains the @Blade decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface Contract<TParameters, TModel> extends FxBladeBase.Contract<TParameters, TModel> {
            /**
             * The view model that the FX will bind into the view for this Blade.  This property must be set before 'container.revealContent()'
             * is called (if it is) and before the promise returned from 'onInitialize' is resolved.
             */
            viewModel: any;
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' method.
             */
            context: Context<TParameters, TModel>;
        }
        /**
         * Status bar for a Blade.
         */
        interface StatusBar extends FxBladeBase.StatusBar {
            /**
             * The state of the Blade.
             */
            state: ContentState;
        }
        /**
         * Specifies which content state decoration should be applied to a Blade.
         */
        export import ContentState = FxBladeBase.ContentState;
        /**
         * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' method.
         */
        interface Context<TParameters, TModel> extends FxBladeBase.Context<TParameters, TModel> {
            /**
             * A container object that can be used to control the chrome of the Blade.
             */
            container: Container;
        }
        /**
         * Describes if/when to display an alert when the blade closes.
         */
        export import AlertLevel = FxBladeBase.AlertLevel;
        /**
         * Options supplied to the @Blade decorator.  Includes metadata describing the Blade and how it will be treated by
         * the FX.
         */
        interface Options extends FxBladeBase.Options, FxViewModelBase.AcceptsStyleSheetsOptions, FxBladeBase.ShrinkOnOpenChildBladeOptions {
        }
        /**
         * Represents a container object that can be used to control the chrome of the Blade.
         */
        interface Container extends FxBladeBase.Container, FxBladeBase.CanHaveCommandBarContainer, DialogManagement {
            /**
             * The status bar of the Blade.
             */
            statusBar: KnockoutObservable<StatusBar>;
            /**
             * The path to the help content file.
             */
            helpContentUri: KnockoutObservable<string>;
            /**
             * Moves the container into an 'unauthorized' display mode, allowing for an optional custom error message or notice options.
             *
             * @param message A custom error message or notice options in place of the default.
             */
            unauthorized(message?: string | NoticeOptions): void;
        }
        /**
         * The type of shield to show when there is one or more operations in progress.
         */
        export import ShieldType = FxComposition.ShieldType;
        /**
         * The options for the notice blade.
         */
        export import NoticeOptions = MsPortalFx.ViewModels.NoticeOptions;
        module Rebindable {
            /**
             * The @Rebindable decorator.  This expands the Blade view model such that it implements an 'onRebind' method that will be
             * called when the Blade's parameters are changed.  Parameter changes are caused by, for example, the parent Blade/Part making
             * a call to a 'container.openBlade()'-related API or by the user updating browser's address bar.
             *
             * When this decorator is not used, whenever parameters change value, the Blade view model will be disposed and
             * a new one instantiated and initialized.
             */
            function Decorator(): (rebindableClass: RebindableClass) => void;
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface RebindableClass {
                new (): Contract;
                _fx?: {
                    rebindable?: boolean;
                };
            }
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface Contract {
                /**
                 * A method called by the FX whenever the view model's parameters change value.  When called, the view model should
                 * be updated to reflect the new parameter/settings values.
                 *
                 * @param changes A description of the parameters that have changed.
                 * @return A promise that should be resolved once the view model has been entirely updated to reflect the new values for
                 * parameters/settings.
                 */
                onRebind(changes: Changes): Q.Promise<any>;
            }
            /**
             * Passed to the view model's 'onRebind' method, reflecting the changes to the view model's parameters.
             */
            type Changes = FxBladeBase.Rebindable.Changes;
        }
        module Configurable {
            /**
             * The @Configurable decorator.  View models that use this decorator will make use of 'context.configuration' to
             * obtain the view model's configuration API, with which the view model can update its persisted 'settings' values.  The 'settings'
             * values are made available in the view model's 'context', for use in the view model's 'onInitialize' and optional 'onRebind'
             * methods.
             *
             * @param options Options used to configure the @Configurable decorator, supplying additional metadata for the view model's
             * settings.
             */
            function Decorator(options?: Options): (configurableClass: ConfigurableClass) => void;
            /**
             * Options supplied to the @Configurable decorator.  Used to supply metadata describing the Blade's settings.
             */
            type Options = FxBladeBase.Configurable.Options;
            /**
             * Metadata describing one setting of the Blade's persisted settings.
             */
            type SettingMetadata = FxBladeBase.Configurable.SettingMetadata;
            /**
             * The scope at which the Blade's settings are persisted.
             */
            export import SettingsScope = FxBladeBase.Configurable.SettingsScope;
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the Blade's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface ConfigurableClass {
                new (): Contract<any>;
                _fx?: {
                    configurableOptions?: Options;
                };
            }
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the Blade's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface Contract<TSettings> {
                /**
                 * Context injected by the FX into the view model.  This context will include 'configuration' (the Blade's configuration
                 * API) as well as 'settings' (the Blade's persisted settings values).
                 */
                context: Context<TSettings>;
            }
            /**
             * Context injected by the FX into the view model.  This context will include 'configuration' (the Blade's configuration API),
             * the 'parameters' of the Blade, as well as its persisted 'settings' values.
             */
            interface Context<TSettings> extends FxBladeBase.Configurable.Context<TSettings> {
                /**
                 * The view model's configuration API, with which the view model can update its settings.
                 */
                configuration: Configuration<TSettings>;
            }
            /**
             * The Blade's configuration API returned from 'context.configuration'.
             */
            type Configuration<TSettings> = FxBladeBase.Configurable.Configuration<TSettings>;
        }
        module ReturnsData {
            /**
             * The @ReturnsData decorator.  Adds a strongly-typed 'context.container.closeCurrentBlade(data: TData)' to the
             * Blade.
             */
            function Decorator(): (returnsDataClass: ReturnsDataClass) => void;
            /**
             * Constrains the @ReturnsData decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface ReturnsDataClass {
                new (): Contract<any>;
            }
            /**
             * Constrains the @ReturnsData decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface Contract<TData> {
                /**
                 * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and
                 * optional 'onRebind' methods.
                 */
                context: Context<TData>;
            }
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
             * 'onRebind' methods.
             */
            interface Context<TData> {
                /**
                 * A container object that can be used to control the chrome of the Blade.
                 */
                container: FxBladeBase.ReturnsData.Container<TData>;
            }
        }
        module Pinnable {
            /**
             * The @Pinnable decorator.  Requires that the Blade implement an 'onPin' method that determines what Part
             * to pin when the user pins the Blade.
             */
            function Decorator(): (pinnableClass: PinnableClass) => void;
            /**
             * Constrains the @Pinnable decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface PinnableClass {
                new (): Contract;
            }
            /**
             * Constrains the @Pinnable decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface Contract {
                /**
                 * A callback that will be called when the user pins the Blade.  Returns a PartReference to the Part that will be pinned to the
                 * user's Dashboard.
                 */
                onPin(): PartReference<any>;
            }
        }
        module DoesProvisioning {
            /**
             * The @DoesProvisioning decorator. Adds a strongly-typed 'context.provisioning' to the TemplateBlade.
             * This provisioning API allows the view model to perform provisioning operations, including support
             * for telemetry around provisioning.
             *
             * @param options Options used to configure the @DoesProvisioning decorator.
             */
            function Decorator(options?: Options): (doesProvisioningClass: DoesProvisioningClass) => void;
            /**
             * Constrains the @DoesProvisioning decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface DoesProvisioningClass {
                new (): Contract;
                _fx?: {
                    doesProvisioning?: boolean;
                };
            }
            /**
             * Options supplied to the @DoesProvisioning decorator.
             */
            interface Options {
                /**
                 * A flag indicating whether the caller is required to supply a Marketplace item id
                 * when instantiating a BladeReference to this TemplateBlade. Defaults to true.
                 */
                requiresMarketplaceId?: boolean;
            }
            /**
             * Constrains the @DoesProvisioning decorator so that it can be applied only to classes
             * implementing 'Contract'.
             */
            interface Contract {
                /**
                 * Context injected by the FX into the view model. This context should be used within
                 * the view model's 'onInitialize' and optional 'onRebind' methods.
                 */
                context: Context<any>;
            }
            /**
             * Context injected by the FX into the view model. This context should be used within the
             * view model's 'onInitialize' and optional 'onRebind' methods.
             */
            interface Context<TUIMetadata = void> {
                /**
                 * The view model's provisioning API, with which the view model can
                 * perform provisioning operations, including support for telemetry
                 * around provisioning.
                 */
                provisioning: FxBladeBase.DoesProvisioning.Provisioning<TUIMetadata>;
            }
            /**
             * Options for the DeployTemplate method on provisioning context
             */
            export import DeployTemplateOptions = Provisioning.DeployTemplateOptions;
            /**
             * Options for the DeployCustom method on provisioning context
             */
            export import DeployCustomOptions = Provisioning.DeployCustomOptions;
            /**
             * The template deployment operation mode. Defaults to 'RequestDeploymentOnly'.
             */
            export import TemplateDeploymentMode = Provisioning.TemplateDeploymentMode;
        }
        module LegacyFeatures {
            /**
             * The @LegacyFeatures decorator.  Used to access legacy features for Blades, like defining an explicit Blade width.
             *
             * @param options Options used to configure the @LegacyFeatures decorator.
             */
            function Decorator(options: Options): (bladeClass: BladeClass) => void;
            /**
             * Options supplied to the @LegacyFeatures decorator.
             */
            interface Options {
                /**
                 * The width for this Blade.
                 */
                width?: Width;
            }
            /**
             * The available widths for the Blade.
             */
            export import Width = MsPortalFx.Blades.BladeWidth;
        }
        module ForContextPane {
            /**
             * The @ForContextPane decorator.  Used to configure the behavior of the Blade when it is opened in the ContextPane.
             *
             * @param options Options used to configure the @ForContextPane decorator.
             */
            function Decorator(options: Options): (bladeClass: BladeClass) => void;
            /**
             * Options supplied to the @ForContextPane decorator.
             */
            interface Options {
                /**
                 * The width of the Blade when it is opened in the ContextPane.
                 */
                width?: Width;
            }
            /**
             * The available widths for the ContextPane.
             */
            export import Width = MsPortalFx.Blades.BladeWidth;
        }
        /**
         * The @ProxiedMember decorator.  Identifies a public member of a Blade class that is to be proxied for use in the view.
         * This decorator is not necessary in most scenarios.
         * This decorator suppresses any compile-time errors that state "The '<MEMBER>' member is public and is not bound to the Blade's view."
         */
        export import ProxiedMember = FxViewModelBase.ProxiedMember;
    }
}

// FILE: Fx\Composition\BladeBase.d.ts
declare module "Fx/Composition/BladeBase" {
    import { ClickableLink } from "Fx/Composition";
    import { ItemType } from "Fx/Composition/ItemType";
    import { Container as PdlBladeContainer } from "Fx/Composition/Pdl/Blade";
    import { Context as MarketplaceContext } from "Fx/ResourceManagement/Marketplace";
    import { InitialValues, DeployTemplateOptions, DeployCustomOptions } from "Fx/ResourceManagement/Provisioning";
    import FxViewModelBase = require("Fx/Composition/ViewModelBase");
    export = Main;
    module Main {
        import FxConfiguration = MsPortalFx.Composition.Configuration;
        import FxToolbars = MsPortalFx.ViewModels.Toolbars;
        import FxResourceManager = MsPortalFx.Azure.ResourceManager;
        /**
         * The class type to which the various Blade decorators can be applied.
         */
        interface BladeClass extends FxViewModelBase.ViewModelClass<any> {
            _fx?: {
                itemType?: ItemType;
                options?: Options;
                ownsEditScope?: boolean;
                rebindable?: boolean;
                parameterProvider?: boolean;
                doesProvisioning?: boolean;
                configurableOptions?: FxViewModelBase.Configurable.Options;
            };
        }
        /**
         * A common interface type for various Blade view model classes.
         */
        interface Contract<TParameters, TModel> extends FxViewModelBase.Contract<TParameters, TModel> {
            /**
             * The displayed title of the Blade.
             * Supply 'null' if no title is desired for the Blade.
             */
            title: string | KnockoutObservableBase<string>;
            /**
             * The displayed subtitle of the Blade.
             * Supply 'null' if no subtitle is desired for the Blade.
             */
            subtitle: string | KnockoutObservableBase<string>;
        }
        /**
         * Status bar for a Blade.
         */
        interface StatusBar {
            /**
             * The text to display in the Blade's status bar.
             */
            text: string;
            /**
             * Optional callback called by the FX when the status bar is clicked.
             */
            onClick?: (() => void) | ClickableLink;
        }
        /**
         * Specifies which content state decoration should be applied to a Blade.
         */
        export import ContentState = MsPortalFx.ViewModels.ContentState;
        /**
         * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
         * 'onRebind' methods.
         */
        interface Context<TParameters, TModel> extends FxViewModelBase.Context<TParameters, TModel> {
            /**
             * An API a Blade can use to interact with its host MenuBlade (if applicable).
             */
            menu: MenuBladeManagement;
            /**
             * An API a Blade can use to interact with the Form Fields on the Blade.
             */
            form: FormManagement;
        }
        /**
         * Represents a container object that can be used to control the chrome of the Blade.
         */
        interface Container extends FxViewModelBase.Container {
            /**
             * Supplies the Blade's icon.
             * Any non-null value supplied here will override the image inferred from this Blade's 'forAsset' option (if supplied).
             */
            icon: KnockoutObservable<MsPortalFx.Base.Image>;
        }
        /**
         * Represents a container object that can be used to specify the CommandBar of the Blade.
         */
        interface CanHaveCommandBarContainer {
            /**
             * The command bar of the Blade.  If required, must be set before 'container.revealContent' is called and before the
             * Promise returned from 'onInitialized' is resolved.
             */
            commandBar: FxToolbars.Toolbar;
        }
        /**
         * Options passed to the various Blade decorators.
         */
        interface Options extends FxViewModelBase.Options {
            /**
             * Specifies whether the Blade is pinnable.  Defaults to 'true'.
             */
            isPinnable?: boolean;
        }
        /**
         * Shared by decorator options of @Blade/@TemplateBlade/@FrameBlade
         */
        interface ShrinkOnOpenChildBladeOptions {
            /**
             * Specifies whether the Blade shrinks to the small width on open child blade.
             */
            shrinkOnOpenChildBlade?: boolean;
        }
        module Rebindable {
            /**
             * Passed to the view model's 'onRebind' method, reflecting the changes to the view model's parameters.
             */
            type Changes = FxViewModelBase.Rebindable.Changes;
        }
        module Configurable {
            /**
             * Options supplied to the @Configurable decorator.  Used to supply metadata describing the Blade's settings.
             */
            type Options = FxViewModelBase.Configurable.Options;
            /**
             * Metadata describing one setting of the Blade's persisted settings.
             */
            type SettingMetadata = FxViewModelBase.Configurable.SettingMetadata;
            /**
             * The scope at which the Blade's settings are persisted.
             */
            export import SettingsScope = FxViewModelBase.Configurable.SettingsScope;
            /**
             * Context injected by the FX into the view model.  This context will include 'configuration' (the Blades's
             * configuration API) as well as 'settings' (the Blade's persisted settings values).
             */
            type Context<TSettings> = FxViewModelBase.Configurable.Context<TSettings>;
            /**
             * The Blade's configuration API returned from 'context.configuration'.
             */
            type Configuration<TSettings> = FxConfiguration.Blade.Contract<TSettings>;
        }
        module ReturnsData {
            /**
             * Represents a container object that can be used to control the chrome of the Blade.
             */
            interface Container<TData> {
                /**
                 * Closes the Blade, returning 'data' to the parent Blade/Part.
                 *
                 * @param data The data to return to the parent Blade/Part.
                 * @return A promise indicating if the Blade was successfully closed.
                 */
                closeCurrentBlade(data: TData): Q.Promise<boolean>;
            }
        }
        module DoesProvisioning {
            /**
             * The view model's provisioning API, with which the view model can provision async
             * operations, such as creates and deployments, (and logs the necessary telemetry).
             */
            interface Provisioning<TUIMetadata> extends MarketplaceContext<TUIMetadata> {
                /**
                 * Initial values for form initialization. Use those values to initialize the subscription,
                 * resource group, and location drop down controls.
                 */
                readonly initialValues: InitialValues;
                /**
                 * Deploy a template to ARM and recieve a promise for a deployment result
                 */
                deployTemplate(options: DeployTemplateOptions): Q.Promise<FxResourceManager.TemplateDeploymentResult>;
                /**
                 * Deploy a template to ARM and recieve a promise for a deployment result
                 */
                deployCustom<TResult>(options: DeployCustomOptions<TResult>): Q.Promise<FxResourceManager.TemplateDeploymentResult>;
            }
        }
        /**
         * An API available as 'context.form' that a Blade can use to interact with the Form Fields on the Blade.
         */
        type FormManagement = MsPortalFx.ViewModels.FormProperties;
        /**
         * Describes if/when to display an alert when the blade closes.
         */
        export import AlertLevel = MsPortalFx.ViewModels.AlertLevel;
        /**
         * An API available as 'context.menu' that a Blade can use to interact with its host MenuBlade (if applicable).
         */
        interface MenuBladeManagement extends MsPortalFx.ViewModels.MenuBladeManagement {
            /**
             * A suffix applied to the title of the MenuBlade.
             */
            titleSuffix: KnockoutObservable<string>;
            /**
             * A suffix applied to the subtitle of the MenuBlade.
             */
            subtitleSuffix: KnockoutObservable<string>;
        }
        /**.
         * The blade 'Container' API for all variations of blades. It's through this API that extensions
         * can interact with the blade's chrome UI.
         */
        type AnyBladeContainer = Container | PdlBladeContainer;
    }
}

// FILE: Fx\Composition\ButtonPart.d.ts
declare module "Fx/Composition/ButtonPart" {
    import FxComposition = require("Fx/Composition");
    import FxViewModelBase = require("Fx/Composition/ViewModelBase");
    import FxPartBase = require("Fx/Composition/PartBase");
    export = MsPortalFxButtonPart;
    module MsPortalFxButtonPart {
        /**
         * The @ButtonPart decorator.  Identifies a class within the extension as a ButtonPart view model class.
         *
         * @options Metadata describing the ButtonPart and how it will be treated by the FX.
         */
        function Decorator(options?: Options): (partClass: ButtonPartClass) => void;
        /**
         * Constrains the @ButtonPart decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface ButtonPartClass extends FxPartBase.PartClass {
            new (): Contract<any, any>;
        }
        /**
         * Constrains the @ButtonPart decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface Contract<TParameters, TModel> {
            /**
             * The title of the part.
             */
            title: string | KnockoutObservableBase<string>;
            /**
             * A shorter title of the part (for size: mini).
             */
            shortTitle: string | KnockoutObservableBase<string>;
            /**
             * The displayed subtitle for this Part, typically the Asset name for the asset/resource associated with this Part.
             * Supply 'null' if no subtitle is desired for the Part.
             */
            subtitle: string | KnockoutObservableBase<string>;
            /**
             * A description for the part.
             */
            description: string | KnockoutObservableBase<string>;
            /**
             * An icon for the the part.
             */
            icon: MsPortalFx.Base.Image | KnockoutObservableBase<MsPortalFx.Base.Image>;
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and
             * optional 'onRebind' methods.
             */
            context: Context<TParameters, TModel>;
            /**
             * Called by the FX to initialize this ButtonPart.
             *
             * @return A promise that should be resolved by the ButtonPart when: (1) all the data is loaded for this view model and (2) the
             * ButtonPart state is updated to reflect the loaded data.
             */
            onInitialize(): Q.Promise<any>;
            /**
             * Describes the behavior of the ButtonPart when it is clicked by the user.
             * Supply 'null' if the Part is not clickable.
             */
            onClick: (() => void) | FxComposition.ClickableLink;
        }
        /**
         * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
         * 'onRebind' methods.
         */
        interface Context<TParameters, TModel> extends FxPartBase.Context<TParameters, TModel> {
            /**
             * A container object that can be used to control the chrome of the ButtonPart.
             */
            container: Container;
        }
        /**
         * Represents a container object that can be used to control the chrome of the ButtonPart.
         */
        interface Container extends FxPartBase.Container {
            /**
             * Detailed ButtonPart size information.
             * The ButtonPart class should subscribe to this observable to be notified of size changes.
             */
            size: KnockoutReadOnlyObservableBase<SizeInfo>;
            /**
             * Indicates the location of this ButtonPart (on a Dashboard, on a Blade, etc.).
             */
            location: Location;
        }
        /**
         * The type of shield to show when there is one or more operations in progress.
         */
        export import ShieldType = FxComposition.ShieldType;
        /**
         * Detailed ButtonPart size information.
         */
        interface SizeInfo extends FxPartBase.SizeInfo {
            /**
             * One of the defined ButtonPart sizes, including 'Custom' if the ButtonPart is configured to support arbitrary sizing.
             */
            partSize: Size;
        }
        /**
         * Indicates the location of this ButtonPart (on a Dashboard, on a Blade, etc.).
         */
        export import Location = FxPartBase.Location;
        /**
         * Options supplied to the @ButtonPart decorator.  Includes metadata describing the ButtonPart and how it will be treated by
         * the FX.
         */
        interface Options extends FxViewModelBase.Options {
            /**
             * The initial size with which the ButtonPart will be rendered.
             */
            initialSize?: Size.Mini | Size.Small | Size.Normal | Size.Wide;
            /**
             * Metadata describing the parameters passed to this Part.
             */
            parameterMetadata?: {
                [key: string]: ParameterMetadata;
            };
            /**
             * Metadata that describes this Part in the Part Gallery.
             */
            galleryMetadata?: FxPartBase.GalleryMetadata;
        }
        /**
         * The available sizes for the ButtonPart.
         */
        export import Size = FxPartBase.Size;
        /**
         * Metadata describing a parameter passed to this ButtonPart.
         */
        type ParameterMetadata = FxPartBase.ParameterMetadata;
        /**
         * Describes how the image should be stretched to fill allocated space in the Part Gallery entry.
         */
        export import PartGalleryThumbnailStretch = FxPartBase.PartGalleryThumbnailStretch;
        module Rebindable {
            /**
             * The @Rebindable decorator.  This expands the ButtonPart view model such that it implements an 'onRebind' method that will be
             * called when the ButtonPart's parameters are changed.  Parameter changes are caused by, for example, the user modifying a Dashboard
             * value ('timeRange', for instance) to which the ButtonPart is bound.
             *
             * When this decorator is not used, whenever parameters change value, the ButtonPart view model will be disposed and
             * a new one instantiated and initialized.
             */
            function Decorator(): (rebindableClass: RebindableClass) => void;
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface RebindableClass {
                new (): Contract;
                _fx?: {
                    rebindable?: boolean;
                };
            }
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface Contract {
                /**
                 * A method called by the FX whenever the view model's parameters change value.  When called, the view model should
                 * be updated to reflect the new parameter/settings values.
                 *
                 * @param changes A description of the parameters that have changed.
                 * @return A promise that should be resolved once the view model has been entirely updated to reflect the new values for
                 * parameters/settings.
                 */
                onRebind(changes: Changes): Q.Promise<any>;
            }
            /**
             * Passed to the view model's 'onRebind' method, reflecting the changes to the view model's parameters.
             */
            type Changes = FxPartBase.Rebindable.Changes;
        }
        module ReplacesOldPart {
            /**
             * Options that are supplied to the @ReplacesOldPart decorator
             */
            type Options = FxPartBase.ReplacesOldPart.Options;
            /**
             * The @ReplacesOldPart decorator.  This decorator is used to identify a old part that this
             * ButtonPart replaces.
             *
             * @param options options used to configure the @ReplacesOldPart decorator.
             */
            function Decorator(options: Options): (partClass: ButtonPartClass) => void;
        }
    }
}

// FILE: Fx\Composition\DashboardFilters.d.ts
declare module "Fx/Composition/DashboardFilters" {
    export = MsPortalFxDashboardFilters;
    module MsPortalFxDashboardFilters {
        namespace TimeRangeFilter {
            /**
             * The filter model supplied to a Part's 'onInitialize' and 'onRebind' methods.
             * A Part will load/reload its data by constructing an AJAX query based on this filter model.
             */
            interface FilterModel {
                /**
                 * An absolute time range.
                 */
                absolute?: {
                    /**
                     * The start of the time range.
                     */
                    fromDate: Date;
                    /**
                     * The end of the time range.
                     */
                    toDate: Date;
                };
                /**
                 * A relative time range.
                 */
                relative?: {
                    /**
                     * The duration for the relative time range.
                     */
                    duration: number;
                    /**
                     * The time unit for the duration of the relative time range.
                     */
                    timeUnit: TimeUnit;
                };
                /**
                 * The time standard for the provided time range.
                 */
                format?: DateTimeKind;
                /**
                 * The time granularity for the provided time range.
                 */
                granularity?: TimeGranularity;
            }
            /**
             * Values represent units of time.  Used in FilterModel above.
             */
            const enum TimeUnit {
                Minute = 0,
                Hour = 1,
                Day = 2,
                Week = 3,
                Month = 4,
                Year = 5,
            }
            /**
             * Values represent time zone formats.  Used in FilterModel above.
             */
            const enum DateTimeKind {
                UTC = 0,
                Local = 1,
            }
            /**
             * Values represent time granularity units.  Used in FilterModel above.
             */
            const enum TimeGranularity {
                Day = 0,
                Week = 1,
                Month = 2,
                Year = 3,
            }
            /**
             * The filter id of the TimeRange filter.
             */
            const filterId = "MsPortalFx_TimeRange";
            /**
             * Used to compose a Part's 'TFilterModels' type (see the 'Filterable.Context<TFilterModels>'-typing for its 'context' member).
             */
            interface FilterIdAndModel {
                MsPortalFx_TimeRange: FilterModel;
            }
        }
        /**
         * An abstract class representing an extension-defined filter, one supported by a particular Part.
         * Each supported FX filter variant will have a 'createDefinition' factory function that returns a FilterDefinition.
         * A Part returns FilterDefinitions covering its supported filters from its 'getFilterDefinitions' and 'getDynamicFilterDefinitions'
         * methods.
         */
        class FilterDefinition {
        }
        /**
         * Identifies a specific localizable filter value.
         */
        interface LocalizableFilterValue {
            /**
             * Identifies a specific localizable filter value.  Part's return localized display text for such a value by implementing
             * a 'getLocalizedFilterValues' method.
             */
            valueId: string;
        }
        /**
         * A discrete value the user can pick in the filter UI.
         */
        type FilterValue = string | LocalizableFilterValue;
        namespace EqualityFilter {
            /**
             * Operators supported by the equality filter.
             */
            const enum Operator {
                Equals = 0,
                NotEquals = 1,
            }
            /**
             * A discrete value the user can pick in the equality filter UI.
             */
            type FilterValue = MsPortalFxDashboardFilters.FilterValue;
            /**
             * The filter model supplied to a Part's 'onInitialize' and 'onRebind' methods.
             * A Part will load/reload its data by constructing an AJAX query based on this filter model.
             *
             * Re: "Select all":
             * - "User clicks 'select all'" is normalized to { operator: NotEquals, values: [] }
             * - "User clicks 'select all' then unchecks specific values" is normalized to { operator: NotEquals, value: [ <UNCHECKED_VALUES> ]}
             */
            interface FilterModel {
                /**
                 * The active filter operator.
                 */
                operator: Operator;
                /**
                 * A set of filter values.
                 */
                values: FilterValue[];
            }
            /**
             * Options supplied to 'createDefinition' (below), identifying and configuring an extension-declared equality filter.
             */
            interface Options {
                /**
                 * The id of the extension-declared filter.  For extensions to jointly participate in a filter (for instance, to each supply
                 * distinct filter values), this id must be the same across the extensions.
                 */
                filterId: string;
                /**
                 * Display text for this filter.
                 */
                displayText: string;
                /**
                 * A callback called when the user changes the filter text in the equality filter UI.  In response, the Part should
                 * return a set of filter values, where each value matches 'filterText'.
                 *
                 * @param filterText User input from the equality filter text box.
                 * @return An array or Promise resolving with an array of filter values, where each value matches 'filterText'.
                 */
                getFilterValues(filterText: string): FilterValue[] | Q.Promise<FilterValue[]>;
            }
            /**
             * Creates an equality filter definition. Returned from a Part's 'getFilterDefinitions' or 'getDynamicFilterDefinitions'.
             *
             * @param options Options from which to create the filter definition.
             * @return The filter definition.
             */
            function createDefinition(options: Options): MsPortalFxDashboardFilters.FilterDefinition;
        }
        /**
         * A type used to describe dynamic filters (that is, those filters not known at compile-time).  In a Part's 'onInitialize' and 'onRebind'
         * methods, for dynamic filters, the Part will downcast filter models statically-typed as 'AnyFilterModel'.
         */
        type AnyFilterModel = EqualityFilter.FilterModel | TimeRangeFilter.FilterModel | SubscriptionFilter.FilterModel | LocationFilter.FilterModel | ResourceGroupFilter.FilterModel | ResourceTypeFilter.FilterModel;
        /**
         * Part 'TFilterModels' interface types will inherit this interface type when they support dynamic filters (those filters not known at
         * compile-time).  In the Part's 'onInitialize' and 'onRebind' methods, it will reason over filter models for dynamic filters
         * according to this interface type.
         */
        interface DynamicFilterModels {
            /**
             * The filter {id, model}-pairs for dynamic filters.
             */
            [dynamicFilterId: string]: AnyFilterModel;
        }
        /**
         * A localized filter value returned from a Part's 'getLocalizedFilterValues' method.
         */
        interface LocalizedFilterValue {
            /**
             * The value id.
             */
            valueId: string;
            /**
             * The localized filter value.
             */
            value: string;
        }
        /**
         * A dynamic filter definition returned from a Part's 'getDynamicFilterDefinitions'.
         */
        interface DynamicFilterDefinition {
            /**
             * The filter definition.
             */
            definition: FilterDefinition;
            /**
             * Specifies whether telemetry should be logged for this filter.  Defaults to 'false' if not supplied.
             * Telemetry should only be logged if the filter (its filter id, for instance) doesn't include any sensitive user information.
             */
            logTelemetry?: boolean;
        }
        namespace SubscriptionFilter {
            /**
             * Operators supported by the subscription filter.
             */
            export import Operator = EqualityFilter.Operator;
            /**
             * The filter model supplied to a Part's 'onInitialize' and 'onRebind' methods.
             * A Part will load/reload its data by constructing an AJAX query based on this filter model.
             *
             * Re: "Select all":
             * - "User clicks 'select all'" is normalized to { operator: NotEquals, values: [] }
             * - "User clicks 'select all' then unchecks specific values" is normalized to { operator: NotEquals, value: [ <UNCHECKED_VALUES> ]}
             */
            interface FilterModel {
                /**
                 * The active filter operator.
                 */
                operator: Operator;
                /**
                 * A set of subscription ids.
                 */
                subscriptionIds: string[];
            }
            /**
             * Options supplied to 'createDefinition' (below), configuring the Part's support for the Dashboard subscription filter.
             */
            interface Options {
                /**
                 * A callback called to populate the list of subscriptions from which the user can pick, reflecting those subscriptions supported
                 * by the associated Part.
                 *
                 * @return An array or Promise resolving with an array of supported subscription ids.
                 */
                getSubscriptionIds(): string[] | Q.Promise<string[]>;
            }
            /**
             * Creates a subscription filter definition. Returned from a Part's 'getFilterDefinitions' or 'getDynamicFilterDefinitions'.
             *
             * @param options Options from which to create the filter definition.
             * @return The filter definition.
             */
            function createDefinition(options: Options): MsPortalFxDashboardFilters.FilterDefinition;
            /**
             * The filter id of the subscription filter.
             */
            const filterId = "MsPortalFx_Subscription";
            /**
             * Used to compose a Part's 'TFilterModels' type (see the 'Filterable.Context<TFilterModels>'-typing for its 'context' member).
             */
            interface FilterIdAndModel {
                MsPortalFx_Subscription: FilterModel;
            }
        }
        namespace LocationFilter {
            /**
             * Operators supported by the location filter.
             */
            export import Operator = EqualityFilter.Operator;
            /**
             * A location known to the Portal/ARM for which locale-specific text can be mapped to by the FX.
             */
            interface KnownLocation {
                /**
                 * The non-localized location name.
                 */
                name: string;
            }
            /**
             * A discrete location value the user can pick in the location filter UI.
             */
            type Location = KnownLocation | MsPortalFxDashboardFilters.FilterValue;
            /**
             * The filter model supplied to a Part's 'onInitialize' and 'onRebind' methods.
             * A Part will load/reload its data by constructing an AJAX query based on this filter model.
             *
             * Re: "Select all":
             * - "User clicks 'select all'" is normalized to { operator: NotEquals, values: [] }
             * - "User clicks 'select all' then unchecks specific values" is normalized to { operator: NotEquals, value: [ <UNCHECKED_VALUES> ]}
             */
            interface FilterModel {
                /**
                 * The active filter operator.
                 */
                operator: Operator;
                /**
                 * A set of locations.
                 */
                locations: Location[];
            }
            /**
             * Options supplied to 'createDefinition' (below), configuring the Part's support for the Dashboard location filter.
             */
            interface Options {
                /**
                 * A callback called to populate the list of locations from which the user can pick, reflecting those locations supported
                 * by the associated Part.
                 *
                 * @param filterText User input from the location filter text box.
                 * @return An array or Promise resolving with an array of supported locations ids.
                 */
                getLocations(filterText: string): Location[] | Q.Promise<Location[]>;
            }
            /**
             * Creates a location filter definition. Returned from a Part's 'getFilterDefinitions' or 'getDynamicFilterDefinitions'.
             *
             * @param options Options from which to create the filter definition.
             * @return The filter definition.
             */
            function createDefinition(options: Options): MsPortalFxDashboardFilters.FilterDefinition;
            /**
             * The filter id of the location filter.
             */
            const filterId = "MsPortalFx_Location";
            /**
             * Used to compose a Part's 'TFilterModels' type (see the 'Filterable.Context<TFilterModels>'-typing for its 'context' member).
             */
            interface FilterIdAndModel {
                MsPortalFx_Location: FilterModel;
            }
        }
        namespace ResourceGroupFilter {
            /**
             * Operators supported by the resource groups filter.
             */
            export import Operator = EqualityFilter.Operator;
            /**
             * The filter model supplied to a Part's 'onInitialize' and 'onRebind' methods.
             * A Part will load/reload its data by constructing an AJAX query based on this filter model.
             *
             * Re: "Select all":
             * - "User clicks 'select all'" is normalized to { operator: NotEquals, values: [] }
             * - "User clicks 'select all' then unchecks specific values" is normalized to { operator: NotEquals, value: [ <UNCHECKED_VALUES> ]}
             */
            interface FilterModel {
                /**
                 * The active filter operator.
                 */
                operator: Operator;
                /**
                 * A set of resource group ids.
                 */
                resourceGroupIds: string[];
            }
            /**
             * Options supplied to 'createDefinition' (below), configuring the Part's support for the Dashboard resource groups filter.
             */
            interface Options {
                /**
                 * A callback called to populate the list of resource groups from which the user can pick, reflecting those resource groups supported
                 * by the associated Part.
                 *
                 * @return An array or Promise resolving with an array of supported resource group ids.
                 */
                getResourceGroupIds(): string[] | Q.Promise<string[]>;
            }
            /**
             * Creates a resource groups filter definition. Returned from a Part's 'getFilterDefinitions' or 'getDynamicFilterDefinitions'.
             *
             * @param options Options from which to create the filter definition.
             * @return The filter definition.
             */
            function createDefinition(options: Options): MsPortalFxDashboardFilters.FilterDefinition;
            /**
             * The filter id of the resource groups filter.
             */
            const filterId = "MsPortalFx_ResourceGroup";
            /**
             * Used to compose a Part's 'TFilterModels' type (see the 'Filterable.Context<TFilterModels>'-typing for its 'context' member).
             */
            interface FilterIdAndModel {
                MsPortalFx_ResourceGroup: FilterModel;
            }
        }
        namespace ResourceTypeFilter {
            /**
             * Operators supported by the resource types filter.
             */
            export import Operator = EqualityFilter.Operator;
            /**
             * A discrete resource type value the user can pick in the resource types filter UI.
             */
            interface ResourceType {
                /**
                 * The name of the resource type, used to locate <ResourceType> metadata defined on some <AssetType>.
                 */
                resourceType: string;
                /**
                 * An optional resource kind, used to locate <ResourceKind> metadata around the <ResourceType> named by 'resourceType'.
                 */
                resourceKind?: string;
            }
            /**
             * The filter model supplied to a Part's 'onInitialize' and 'onRebind' methods.
             * A Part will load/reload its data by constructing an AJAX query based on this filter model.
             *
             * Re: "Select all":
             * - "User clicks 'select all'" is normalized to { operator: NotEquals, values: [] }
             * - "User clicks 'select all' then unchecks specific values" is normalized to { operator: NotEquals, value: [ <UNCHECKED_VALUES> ]}
             */
            interface FilterModel {
                /**
                 * The active filter operator.
                 */
                operator: Operator;
                /**
                 * A set of resource types.
                 */
                resourceTypes: ResourceType[];
            }
            /**
             * Options supplied to 'createDefinition' (below), configuring the Part's support for the Dashboard resource types filter.
             */
            interface Options {
                /**
                 * A callback called to populate the list of resource types from which the user can pick, reflecting those resource types supported
                 * by the associated Part.
                 *
                 * @return An array or Promise resolving with an array of supported resource types.
                 */
                getResourceTypes(): ResourceType[] | Q.Promise<ResourceType[]>;
            }
            /**
             * Creates a resource types filter definition. Returned from a Part's 'getFilterDefinitions' or 'getDynamicFilterDefinitions'.
             *
             * @param options Options from which to create the filter definition.
             * @return The filter definition.
             */
            function createDefinition(options: Options): MsPortalFxDashboardFilters.FilterDefinition;
            /**
             * The filter id of the resource types filter.
             */
            const filterId = "MsPortalFx_ResourceType";
            /**
             * Used to compose a Part's 'TFilterModels' type (see the 'Filterable.Context<TFilterModels>'-typing for its 'context' member).
             */
            interface FilterIdAndModel {
                MsPortalFx_ResourceType: FilterModel;
            }
        }
    }
}

// FILE: Fx\Composition\DataContext.d.ts
declare module "Fx/Composition/DataContext" {
    export = MsPortalFxDataContext;
    module MsPortalFxDataContext {
        /**
         * The @DataContext decorator.  Identifies a class within the extension as the DataContext class for a certain extension area.
         * The DataContext class loads and caches data for the area's contained Blades and Parts.
         *
         * @options Options supplied to the decorator, including the area for which the DataContext loads/caches data.
         */
        function Decorator(options?: Options): (dataContextClass: Function) => void;
        /**
         * Options supplied to the @DataContext decorator.
         */
        interface Options {
            /**
             * The extension area for which the DataContext loads and caches data for the area's contained Blades and Parts.
             */
            area: string;
        }
    }
}

// FILE: Fx\Composition\Dialog.d.ts
declare module "Fx/Composition/Dialog" {
    import { HtmlContent } from "Fx/Controls/ControlsBase";
    import { FxElement } from "Fx/Events";
    /**
     * Standard buttons that can be displayed in a dialog.
     */
    export const enum DialogButtons {
        /**
         * The message box contains Abort, Retry and Ignore buttons.
         */
        AbortRetryIgnore = 0,
        /**
         * The message box contains an OK button.
         */
        OK = 1,
        Ok = 1,
        /**
         * The message box contains OK and Cancel buttons.
         */
        OKCancel = 2,
        /**
         * The message box contains Retry and Cancel buttons.
         */
        RetryCancel = 3,
        /**
         * The message box contains Yes and No buttons.
         */
        YesNo = 4,
        /**
         * The message box contains Yes, No and Cancel buttons.
         */
        YesNoCancel = 5,
    }
    /**
     * Button identifiers
     */
    export const enum DialogButton {
        /**
         * The abort button.
         */
        Abort = 0,
        /**
         * The cancel button.
         */
        Cancel = 1,
        /**
         * The ignore button.
         */
        Ignore = 2,
        /**
         * The no button.
         */
        No = 3,
        /**
         * The ok button.
         */
        Ok = 4,
        /**
         * The retry button.
         */
        Retry = 5,
        /**
         * The yes button.
         */
        Yes = 6,
    }
    /**
     * Describes a custom button in a dialog
     */
    export interface CustomDialogButton {
        /**
         * Display text of the button
         */
        displayText: string;
        /**
         * Button identifier that is programatically passed back
         */
        button: DialogButton;
    }
    /**
     * The result returned after a dialog was closed
     */
    export interface DialogResult {
        /**
         * The result returned from the dialog
         */
        button: DialogButton;
    }
    /**
     * option hint for requested dialog position.  Defaults to BottomLeftEdge.
     */
    export const enum DialogPosition {
        TopLeftEdge = 1,
        RightTopEdge = 2,
        BottomLeftEdge = 3,
        LeftTopEdge = 4,
    }
    /**
     * Options supplied when opening a dialog
     */
    export interface BaseDialogOptions {
        /**
         * Name used in telemetry
         */
        telemetryName: string;
        /**
         * Dialog title
         */
        title: string;
        /**
         * Dialog content
         *
         * The dialog content can be a string message to display or an HTML template / view model pair.
         */
        content: string | HtmlContent;
        /**
         * Buttons can be either standard buttons provided by the framework or
         * this can be custom buttons with labels provided by the extension
         */
        buttons: DialogButtons | CustomDialogButton[];
        /**
         * Default button is the button which has focus if custom buttons was specified.
         * This property is not applicable when standard buttons are specified.
         */
        defaultButton?: DialogButton;
        /**
         * Optional callback that is invoked when the dialog is closed.
         * Dialogs are closed when the user clicks on one of the buttons or
         * clicks anywhere on the screen outside of the dialog.   If the user clicks
         * somewhere else in the portal the cancel button or equivalent is passed to this callback.
         */
        onClosed?: (result: DialogResult) => void;
    }
    /**
     * Options supplied when opening a dialog that supports targeting
     */
    export interface DialogOptions extends BaseDialogOptions {
        /**
         * The target which is used to position the dialog.
         * This is a class name or a reference to a DOM element.
         */
        target?: string | FxElement;
        /**
         * Hint where to postion the dialog relative to the target.
         */
        positionHint?: DialogPosition;
    }
    /**
     * Methods for managing dialogs which don't support targeting
     */
    export interface BaseDialogManagement {
        /**
         * Opens a new dialog.
         *
         * @param options Describes the dialog that is too be opened
         * @returns Promise that is resolved the dialog is successfully opened
         */
        openDialog(options: BaseDialogOptions): Q.Promise<boolean>;
    }
    /**
     * Methods for managing dialogs
     */
    export interface DialogManagement {
        /**
         * Opens a new dialog.
         *
         * @param options Describes the dialog that is too be opened
         * @returns Promise that is resolved the dialog is successfully opened
         */
        openDialog(options: DialogOptions): Q.Promise<boolean>;
    }
}

// FILE: Fx\Composition\FrameBlade.d.ts
declare module "Fx/Composition/FrameBlade" {
    import FxComposition = require("Fx/Composition");
    import FxBladeBase = require("Fx/Composition/BladeBase");
    import { BaseDialogManagement } from "Fx/Composition/Dialog";
    import FxFramePart = require("Fx/Composition/FramePart");
    import * as Provisioning from "Fx/ResourceManagement/Provisioning";
    export = MsPortalFxFrameBlade;
    module MsPortalFxFrameBlade {
        import PartReference = MsPortalFx.Composition.PartReference;
        /**
         * The @FrameBlade decorator.  Identifies a class within the extension as a FrameBlade view model class.
         *
         * @options Metadata describing the FrameBlade and how it will be treated by the FX.
         */
        function Decorator(options?: Options): (frameBladeClass: FrameBladeClass) => void;
        /**
         * Constrains the @FrameBlade decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface FrameBladeClass extends FxBladeBase.BladeClass {
            new (): Contract<any, any>;
        }
        /**
         * Constrains the @FrameBlade decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface Contract<TParameters, TModel> extends FxBladeBase.Contract<TParameters, TModel> {
            /**
             * The FrameBlade view model.  This property must be set before 'container.revealContent()'
             * is called (if it is) and before the promise returned from 'onInitialize' is resolved.
             */
            viewModel: ViewModel;
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and
             * optional 'onRebind' methods.
             */
            context: Context<TParameters, TModel>;
        }
        /**
         * Status bar for a FrameBlade.
         */
        interface StatusBar extends FxBladeBase.StatusBar {
            /**
             * The state of the FrameBlade.
             */
            state: ContentState;
        }
        /**
         * Specifies which content state decoration should be applied to a FrameBlade.
         */
        export import ContentState = FxBladeBase.ContentState;
        /**
         * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
         * 'onRebind' methods.
         */
        interface Context<TParameters, TModel> extends FxBladeBase.Context<TParameters, TModel> {
            /**
             * A container object that can be used to control the chrome of the FrameBlade.
             */
            container: Container;
        }
        /**
         * Options supplied to the @FrameBlade decorator.  Includes metadata describing the FrameBlade and how it will be treated by
         * the FX.
         */
        interface Options extends FxBladeBase.Options, FxBladeBase.ShrinkOnOpenChildBladeOptions {
        }
        /**
         * Represents a container object that can be used to control the chrome of the FrameBlade.
         */
        interface Container extends FxBladeBase.Container, FxBladeBase.CanHaveCommandBarContainer, BaseDialogManagement {
            /**
             * The status bar of the FrameBlade.
             */
            statusBar: KnockoutObservable<StatusBar>;
            /**
             * Moves the container into an 'unauthorized' display mode, allowing for an optional custom error message or notice options.
             *
             * @param message A custom error message or notice options in place of the default.
             */
            unauthorized(message?: string | NoticeOptions): void;
        }
        /**
         * The type of shield to show when there is one or more operations in progress.
         */
        export import ShieldType = FxComposition.ShieldType;
        /**
         * The options for the notice blade.
         */
        export import NoticeOptions = MsPortalFx.ViewModels.NoticeOptions;
        module Rebindable {
            /**
             * The @Rebindable decorator.  This expands the FrameBlade view model such that it implements an 'onRebind' method that will be
             * called when the FrameBlade's parameters are changed.  Parameter changes are caused by, for example, the parent Blade/Part making
             * a call to a 'container.openBlade()'-related API or by the user updating browser's address bar.
             *
             * When this decorator is not used, whenever parameters change value, the FrameBlade view model will be disposed and
             * a new one instantiated and initialized.
             */
            function Decorator(): (rebindableClass: RebindableClass) => void;
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface RebindableClass {
                new (): Contract;
                _fx?: {
                    rebindable?: boolean;
                };
            }
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface Contract {
                /**
                 * A method called by the FX whenever the view model's parameters change value.  When called, the view model should
                 * be updated to reflect the new parameter/settings values.
                 *
                 * @param changes A description of the parameters that have changed.
                 * @return A promise that should be resolved once the view model has been entirely updated to reflect the new values for
                 * parameters/settings.
                 */
                onRebind(changes: Changes): Q.Promise<any>;
            }
            /**
             * Passed to the view model's 'onRebind' method, reflecting the changes to the view model's parameters.
             */
            type Changes = FxBladeBase.Rebindable.Changes;
        }
        module Configurable {
            /**
             * The @Configurable decorator.  View models that use this decorator will make use of 'context.configuration' to
             * obtain the view model's configuration API, with which the view model can update its persisted 'settings' values.  The 'settings'
             * values are made available in the view model's 'context', for use in the view model's 'onInitialize' and optional 'onRebind'
             * methods.
             *
             * @param options Options used to configure the @Configurable decorator, supplying additional metadata for the view model's
             * settings.
             */
            function Decorator(options?: Options): (configurableClass: ConfigurableClass) => void;
            /**
             * Options supplied to the @Configurable decorator.  Used to supply metadata describing the FrameBlade's settings.
             */
            type Options = FxBladeBase.Configurable.Options;
            /**
             * Metadata describing one setting of the FrameBlade's persisted settings.
             */
            type SettingMetadata = FxBladeBase.Configurable.SettingMetadata;
            /**
             * The scope at which the FrameBlade's settings are persisted.
             */
            export import SettingsScope = FxBladeBase.Configurable.SettingsScope;
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the FrameBlade's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface ConfigurableClass {
                new (): Contract<any>;
                _fx?: {
                    configurableOptions?: Options;
                };
            }
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the FrameBlade's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface Contract<TSettings> {
                /**
                 * Context injected by the FX into the view model.  This context will include 'configuration' (the FrameBlade's configuration
                 * API) as well as 'settings' (the FrameBlade's persisted settings values).
                 */
                context: Context<TSettings>;
            }
            /**
             * Context injected by the FX into the view model.  This context will include 'configuration' (the FrameBlade's configuration API),
             * the 'parameters' of the FrameBlade, as well as its persisted 'settings' values.
             */
            interface Context<TSettings> extends FxBladeBase.Configurable.Context<TSettings> {
                /**
                 * The view model's configuration API, with which the view model can update its settings.
                 */
                configuration: Configuration<TSettings>;
            }
            /**
             * The FrameBlade's configuration API returned from 'context.configuration'.
             */
            type Configuration<TSettings> = FxBladeBase.Configurable.Configuration<TSettings>;
        }
        module ReturnsData {
            /**
             * The @ReturnsData decorator.  Adds a strongly-typed 'context.container.closeCurrentBlade(data: TData)' to the
             * FrameBlade.
             */
            function Decorator(): (returnsDataClass: ReturnsDataClass) => void;
            /**
             * Constrains the @ReturnsData decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface ReturnsDataClass {
                new (): Contract<any>;
            }
            /**
             * Constrains the @ReturnsData decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface Contract<TData> {
                /**
                 * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and
                 * optional 'onRebind' methods.
                 */
                context: Context<TData>;
            }
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
             * 'onRebind' methods.
             */
            interface Context<TData> {
                /**
                 * A container object that can be used to control the chrome of the FrameBlade.
                 */
                container: FxBladeBase.ReturnsData.Container<TData>;
            }
        }
        module Pinnable {
            /**
             * The @Pinnable decorator.  Requires that the FrameBlade implement an 'onPin' method that determines what Part
             * to pin when the user pins the FrameBlade.
             */
            function Decorator(): (pinnableClass: PinnableClass) => void;
            /**
             * Constrains the @Pinnable decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface PinnableClass {
                new (): Contract;
            }
            /**
             * Constrains the @Pinnable decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface Contract {
                /**
                 * A callback that will be called when the user pins the FrameBlade.  Returns a PartReference to the Part that will be pinned to the
                 * user's Dashboard.
                 */
                onPin(): PartReference<any>;
            }
        }
        module DoesProvisioning {
            /**
             * The @DoesProvisioning decorator. Adds a strongly-typed 'context.provisioning' to the TemplateBlade.
             * This provisioning API allows the view model to perform provisioning operations, including support
             * for telemetry around provisioning.
             *
             * @param options Options used to configure the @DoesProvisioning decorator.
             */
            function Decorator(options?: Options): (doesProvisioningClass: DoesProvisioningClass) => void;
            /**
             * Constrains the @DoesProvisioning decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface DoesProvisioningClass {
                new (): Contract;
                _fx?: {
                    doesProvisioning?: boolean;
                };
            }
            /**
             * Options supplied to the @DoesProvisioning decorator.
             */
            interface Options {
                /**
                 * A flag indicating whether the caller is required to supply a Marketplace item id
                 * when instantiating a BladeReference to this TemplateBlade. Defaults to true.
                 */
                requiresMarketplaceId?: boolean;
            }
            /**
             * Constrains the @DoesProvisioning decorator so that it can be applied only to classes
             * implementing 'Contract'.
             */
            interface Contract {
                /**
                 * Context injected by the FX into the view model. This context should be used within
                 * the view model's 'onInitialize' and optional 'onRebind' methods.
                 */
                context: Context<any>;
            }
            /**
             * Context injected by the FX into the view model. This context should be used within the
             * view model's 'onInitialize' and optional 'onRebind' methods.
             */
            interface Context<TUIMetadata = void> {
                /**
                 * The view model's provisioning API, with which the view model can
                 * perform provisioning operations, including support for telemetry
                 * around provisioning.
                 */
                provisioning: FxBladeBase.DoesProvisioning.Provisioning<TUIMetadata>;
            }
            /**
             * Options for the DeployTemplate method on provisioning context
             */
            export import DeployTemplateOptions = Provisioning.DeployTemplateOptions;
            /**
             * Options for the DeployCustom method on provisioning context
             */
            export import DeployCustomOptions = Provisioning.DeployCustomOptions;
            /**
             * The template deployment operation mode. Defaults to 'RequestDeploymentOnly'.
             */
            export import TemplateDeploymentMode = Provisioning.TemplateDeploymentMode;
        }
        module LegacyFeatures {
            /**
             * The @LegacyFeatures decorator.  Used to access legacy features for FrameBlades, like defining an explicit FrameBlade width.
             *
             * @param options Options used to configure the @LegacyFeatures decorator.
             */
            function Decorator(options: Options): (bladeClass: FrameBladeClass) => void;
            /**
             * Options supplied to the @LegacyFeatures decorator.
             */
            interface Options {
                /**
                 * The width for this FrameBlade.
                 */
                width?: Width;
            }
            /**
             * The available widths for the FrameBlade.
             */
            export import Width = MsPortalFx.Blades.BladeWidth;
        }
        module ForContextPane {
            /**
             * The @ForContextPane decorator.  Used to configure the behavior of the FrameBlade when it is opened in the ContextPane.
             *
             * @param options Options used to configure the @ForContextPane decorator.
             */
            function Decorator(options: Options): (bladeClass: FrameBladeClass) => void;
            /**
             * Options supplied to the @ForContextPane decorator.
             */
            interface Options {
                /**
                 * The width of the FrameBlade when it is opened in the ContextPane.
                 */
                width?: Width;
            }
            /**
             * The available widths for the ContextPane.
             */
            export import Width = MsPortalFx.Blades.BladeWidth;
        }
        /**
         * Options used to configure the FrameBlade view model.
         */
        export import ViewModelOptions = FxFramePart.ViewModelOptions;
        /**
         * Defines the signature for a message handler.
         */
        export import MessageHandler = FxFramePart.MessageHandler;
        /**
         * The view model of the FrameBlade.
         */
        class ViewModel extends FxFramePart.ViewModel {
            /**
             * Creates the view model for a FrameBlade
             *
             * @param container The container for this FrameBlade.
             * @param options Options for the FrameBlade view model.
             */
            constructor(container: Container, options: ViewModelOptions);
        }
    }
}

// FILE: Fx\Composition\FramePart.d.ts
declare module "Fx/Composition/FramePart" {
    import FxComposition = require("Fx/Composition");
    import FxFilters = require("Fx/Composition/DashboardFilters");
    import FxPartBase = require("Fx/Composition/PartBase");
    export = MsPortalFxFramePart;
    module MsPortalFxFramePart {
        /**
         * The @FramePart decorator.  Identifies a class within the extension as a FramePart view model class.
         *
         * @options Metadata describing the FramePart and how it will be treated by the FX.
         */
        function Decorator(options?: Options): (framePartClass: FramePartClass) => void;
        /**
         * Constrains the @FramePart decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface FramePartClass extends FxPartBase.PartClass {
            new (): Contract<any, any>;
        }
        /**
         * Constrains the @FramePart decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface Contract<TParameters, TModel> extends FxPartBase.Contract<TParameters, TModel> {
            /**
             * The view model that the FX will bind into the view for this FramePart.  This property must be set before 'container.revealContent()'
             * is called (if it is) and before the promise returned from 'onInitialize' is resolved.
             */
            viewModel: ViewModel;
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and
             * optional 'onRebind' methods.
             */
            context: Context<TParameters, TModel>;
        }
        /**
         * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
         * 'onRebind' methods.
         */
        interface Context<TParameters, TModel> extends FxPartBase.Context<TParameters, TModel> {
            /**
             * A container object that can be used to control the chrome of the FramePart.
             */
            container: Container;
        }
        /**
         * Represents a container object that can be used to control the chrome of the FramePart.
         */
        interface Container extends FxPartBase.Container {
            /**
             * Detailed Part size information.
             * The FramePart class should subscribe to this observable to be notified of size changes.
             */
            size: KnockoutReadOnlyObservableBase<SizeInfo>;
            /**
             * Indicates the location of this FramePart (on a Dashboard, on a Blade, etc.).
             */
            location: Location;
        }
        /**
         * The type of shield to show when there is one or more operations in progress.
         */
        export import ShieldType = FxComposition.ShieldType;
        /**
         * Detailed Part size information.
         */
        interface SizeInfo extends FxPartBase.SizeInfo {
            /**
             * One of the defined Part sizes, including 'Custom' if the FramePart is configured to support arbitrary sizing.
             */
            partSize: Size;
        }
        /**
         * Indicates the location of this FramePart (on a Dashboard, on a Blade, etc.).
         */
        export import Location = FxPartBase.Location;
        /**
         * Options supplied to the @FramePart decorator.  Includes metadata describing the FramePart and how it will be treated by
         * the FX.
         */
        interface Options extends FxPartBase.Options {
            /**
             * The sizes supported by this FramePart.
             */
            supportedSizes?: Size[];
            /**
             * The initial size with which the FramePart will be rendered.  Must be one of the 'supportedSizes'.
             */
            initialSize?: Size;
            /**
             * The resize mode for the FramePart.  If not supplied, the default value of 'Fixed' will be used.
             */
            resizeMode?: ResizeMode;
        }
        /**
         * The available sizes for the FramePart.
         */
        export import Size = FxPartBase.Size;
        /**
         * The available resize modes for the FramePart.
         */
        export import ResizeMode = FxPartBase.ResizeMode;
        /**
         * Metadata describing a parameter passed to this FramePart.
         */
        type ParameterMetadata = FxPartBase.ParameterMetadata;
        /**
         * Describes how the image should be stretched to fill allocated space in the Part Gallery entry.
         */
        export import PartGalleryThumbnailStretch = FxPartBase.PartGalleryThumbnailStretch;
        module Rebindable {
            /**
             * The @Rebindable decorator.  This expands the FramePart view model such that it implements an 'onRebind' method that will be
             * called when the FramePart's parameters are changed.  Parameter changes are caused by, for example, the user modifying a Dashboard
             * value ('timeRange', for instance) to which the FramePart is bound.
             *
             * When this decorator is not used, whenever parameters change value, the FramePart view model will be disposed and
             * a new one instantiated and initialized.
             */
            function Decorator(): (rebindableClass: RebindableClass) => void;
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface RebindableClass {
                new (): Contract;
                _fx?: {
                    rebindable?: boolean;
                };
            }
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface Contract {
                /**
                 * A method called by the FX whenever the view model's parameters change value.  When called, the view model should
                 * be updated to reflect the new parameter/settings values.
                 *
                 * @param changes A description of the parameters that have changed.
                 * @return A promise that should be resolved once the view model has been entirely updated to reflect the new values for
                 * parameters/settings.
                 */
                onRebind(changes: Changes): Q.Promise<any>;
            }
            /**
             * Passed to the view model's 'onRebind' method, reflecting the changes to the view model's parameters.
             */
            type Changes = FxPartBase.Rebindable.Changes;
        }
        module Configurable {
            /**
             * The @Configurable decorator.  View models that use this decorator will make use of 'context.configuration' to
             * obtain the view model's configuration API, with which the view model can update its 'parameters' and persisted 'settings'
             * values.  A new 'settings' property is added to the view model's 'context' which -- along with 'parameters' -- are used in the
             * view model's 'onInitialize' and optional 'onRebind' methods.
             *
             * @param options Options used to configure the @Configurable decorator, supplying additional metadata for the view model's
             * settings.
             */
            function Decorator(options?: Options): (configurableClass: ConfigurableClass) => void;
            /**
             * Options supplied to the @Configurable decorator.  Used to supply metadata describing the FramePart's settings.
             */
            type Options = FxPartBase.Configurable.Options;
            /**
             * Metadata describing one setting of the FramePart's persisted settings.
             */
            type SettingMetadata = FxPartBase.Configurable.SettingMetadata;
            /**
             * The scope at which the FramePart's settings are persisted.
             */
            export import SettingsScope = FxPartBase.Configurable.SettingsScope;
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the FramePart's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface ConfigurableClass {
                new (): Contract<any, any>;
                _fx?: {
                    configurableOptions?: Options;
                };
            }
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the FramePart's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface Contract<TParameters, TSettings> {
                /**
                 * Context injected by the FX into the view model.  This context will include a 'configuration' (the FramePart's
                 * configuration API), as well as 'parameters' and 'settings' (the FramePart's persisted settings values).
                 */
                context: Context<TParameters, TSettings>;
            }
            /**
             * Context injected by the FX into the view model.  This context will include 'configuration' (the FramePart's configuration
             * API), the 'parameters' of the FramePart, as well as its persisted 'settings' values.
             */
            interface Context<TParameters, TSettings> extends FxPartBase.Configurable.Context<TSettings> {
                /**
                 * The view model's configuration API, with which the view model can update its parameters and settings.
                 */
                configuration: Configuration<TParameters, TSettings>;
            }
            /**
             * The FramePart's configuration API returned from 'context.configuration'.
             */
            type Configuration<TParameters, TSettings> = FxPartBase.Configurable.Configuration<TParameters, TSettings>;
        }
        module ReplacesOldPart {
            /**
             * Options that are supplied to the @ReplacesOldPart decorator
             */
            type Options = FxPartBase.ReplacesOldPart.Options;
            /**
             * The @ReplacesOldPart decorator.  This decorator is used to identify a old part that this
             * FramePart replaces.
             *
             * @param options options used to configure the @ReplacesOldPart decorator.
             */
            function Decorator(options: Options): (partClass: FramePartClass) => void;
        }
        /**
         * The signature of messages used by the FramePart.
         */
        const signature = "FxFramePart";
        /**
         * The kind of message indicating that content can be revealed.
         */
        const revealContentMessageKind = "revealcontent";
        /**
         * The kind of message indicating that initialization has completed.
         */
        const initializationCompleteMessageKind = "initializationcomplete";
        module Internal {
            interface Data<T> {
                /**
                 * Data passed from the shell to the control.
                 */
                postMessage: KnockoutObservable<Message<T>>;
                /**
                 * Data passed from the control to the shell.
                 */
                receiveMessage: KnockoutObservable<Message<T>>;
                /**
                 * Indicates whether the iframe control has initialized.
                 */
                isFrameReady: KnockoutObservable<boolean>;
                /**
                 * The signature to be used for messages routed to/from the IFrame.
                 */
                signature: string;
            }
            interface ViewModelOptions extends MsPortalFxFramePart.ViewModelOptions {
                /**
                 * The signature of messages used by the FramePart.
                 */
                signature: string;
            }
        }
        /**
         * Defines the signature for a message handler.
         */
        type MessageHandler = (data: any) => void;
        /**
         * Defines a message sent between a frame part's view model and the frame.
         */
        interface Message<T> {
            /**
             * The message signature.
             */
            signature: string;
            /**
             * The message kind.
             */
            kind: string;
            /**
             * The message data.
             */
            data: T;
        }
        interface ViewModelOptions {
            /**
             * The URL of the page to load in the frame.
             */
            src: string;
        }
        class ViewModel {
            private _handlers;
            private _readyPromise;
            /**
             * The URL of the page to embed.
             */
            src: string;
            /**
             * Data passed to/from the shell and the control.
             */
            private _msPortalFxData;
            /**
             * Creates the view model for a frame part.
             *
             * @param container Container of the part.
             * @param options Options for this view model.
             */
            constructor(container: MsPortalFx.ViewModels.PartContainerContract | Container, options: ViewModelOptions);
            waitForFrameReady(): Q.Promise<void>;
            /**
             * Posts a message to the frame.
             *
             * @param kind The message kind.
             * @param data The data associated to the message.
             */
            postMessage: <T>(kind: string, data?: T) => void;
            /**
             * Attaches an event handler function for message events of the specified kind.
             *
             * @param kind The kind of message to listen to.
             * @param handler A function to execute when a message of the specified kind is received.
             */
            on(kind: string, handler: MessageHandler): void;
            /**
             * Removes an event handler.
             *
             * @param kind The kind of message.
             * @param handler A function previously attached to message events of the specified kind.
             */
            off(kind: string, handler: MessageHandler): void;
        }
        namespace Filterable {
            /**
             * The @Filterable decorator.  FrameParts that carry this decorator will receive the state of the those filters currently
             * applied to the containing Dashboard.  They'll receive a "filter model" for each applied filter that the FramePart supports.
             * (see the 'FramePart.Filterable.Context' type and its 'filterModels' property, injected into the FramePart as its 'context' member).
             * FrameParts will make use of the current "filter models":
             * - in their 'onInitialize' method, to render the FramePart
             * - in their 'onRebind' method, to update the rendered FramePart to reflect user changes to Dashboard filters
             */
            function Decorator(): (filterableClass: FilterableClass) => void;
            interface FilterableClass {
                new (): Contract<any>;
                _fx?: {
                    filterable?: boolean;
                };
            }
            /**
             * The interface type implemented by any class that carries the '@FramePart.Filterable.Decorator' decorator.
             */
            interface Contract<TFilterModels> {
                /**
                 * Context injected by the FX into the FramePart class.  This context should be used within the FramePart's 'onInitialize'
                 * and 'onRebind' methods.
                 */
                context: Context<TFilterModels>;
                /**
                 * As with the 'onRebind' method required when using the @FramePart.Rebindable.Decorator decorator,
                 * this method will be called by the FX whenever the filters for the current Dashboard are updated by the user.
                 *
                 * @param changes An indication of the filter ids whose corresponding filters have been added/removed/updated.
                 * @return A Promise reflecting the time spent loading updated FramePart data and making corresponding
                 * changes to the FramePart's view.
                 */
                onRebind(changes: Changes): Q.Promise<any>;
                /**
                 * Optional.  Returns filter definitions for extension-defined filters known statically.  These static filters are those known
                 * at development time (and not those derived from mutable backend schema, for instance).  Such extension-defined
                 * filters are ones like EqualityFilter.FilterDefinition (and not FX-intrinsic filters like TimeRangeFilter).
                 * This method should return a filter definition for each extension-defined filters declared on the FramePart's
                 * 'TFilterModels' type included in 'public context: FramePart.Filterable.Context<TFilterModels>'.
                 *
                 * This method will be called by the FX some time after the FX calls this FramePart's 'onInitialize' (to not compete
                 * with the initiation of FramePart data-loading).  When called, this method should not assume that the FramePart is
                 * fully initialized.
                 *
                 * @return An array of filter definitions for each extension-defined filter declared on the FramePart's 'TFilterModels' type.
                 */
                getFilterDefinitions?(): FxFilters.FilterDefinition[];
                /**
                 * Optional.  Returns filter definitions for extension-defined filters known only dynamically (for instance, derived from
                 * mutable backend schema).  Such extension-defined filters are ones like EqualityFilter.FilterDefinition (and not FX-intrinsic
                 * filters like TimeRangeFilter).
                 *
                 * This method will be called by the FX some time after the FX calls this FramePart's 'onInitialize' (to not compete
                 * with the initiation of FramePart data-loading).  When called, this method should not assume that the FramePart is
                 * fully initialized.
                 *
                 * @return An array of filter definitions for each extension-defined filter only known dynamically.
                 */
                getDynamicFilterDefinitions?(): Q.Promise<FxFilters.DynamicFilterDefinition[]>;
                /**
                 * Required only if any filter definition 'getValues' callbacks returns a localizable filter value.
                 * In such cases, this method will be called to map one or more such localizable filter values to display text in the Portal
                 * user's locale.
                 *
                 * @param filterId The filter id for the filter definition whose 'getValues' callback returned a localizable value.
                 * @param valueIds One or more ids of localizable values.
                 * @return A Promise resolving with corresponding display text for each localizable value, according to the user's locale.
                 */
                getLocalizedFilterValues?(filterId: string, valueIds: string[]): Q.Promise<FxFilters.LocalizedFilterValue[]>;
            }
            /**
             * A description of filter changes (adds/removes/updates) that triggered a call to the FramePart's 'onRebind' method,
             * to reflect user modification to Dashboard filters.
             */
            interface Changes {
                /**
                 * An array of filter ids covering the added/removed/updated Dashboard filters.
                 */
                filterIds: string[];
            }
            /**
             * Context injected by the FX into the FramePart class, including extra APIs supporting the
             * '@FramePart.Filterable.Decorator' decorator.
             */
            interface Context<TFilterModels> {
                /**
                 * FX APIs related to Dashboards.
                 */
                dashboard: {
                    /**
                     * FX APIs related to Dashboard filtering.
                     */
                    filtering: {
                        /**
                         * For those filters supported by this FramePart, these are the filter models for those filters currently applied
                         * to the active Dashboard and the Part-level filter overrides.  The supported filters are those static filters declared in this FramePart's
                         * 'TFilterModels' type and those dynamic filters returned from 'getDynamicFilterDefinitions'.
                         */
                        filterModels: TFilterModels;
                        /**
                         * If this FramePart includes UX to update filters that apply to the Part, the FramePart can include an option for
                         * the user to update filters at the Dashboard level, using this method.
                         *
                         * The filters accepted here are a subset of those declared to be supported statically via the 'TFilterModels' type and
                         * those filters supported dynamically via 'getDynamicFilterDefinitions'.
                         *
                         * When the returned Promise resolves, the extension can assume that 'context.filtering.filterModels' is consistent
                         * with the updated 'dashboardFilters' here passed here.
                         *
                         * @param dashboardFilters Updated dashboard filter models
                         * @return A Promise reflecting the updating of the Dashboard filter state as well as 'context.dashboard.filtering.filterModels'.
                         */
                        updateFilters(dashboardFilters: Partial<TFilterModels>): Q.Promise<void>;
                        /**
                         * If this FramePart includes UX to update filters that apply to the Part, the FramePart can include an option for
                         * the user to update filters at the Dashboard level and/or part level, using this method.
                         *
                         * The filters accepted here are a subset of those declared to be supported statically via the 'TFilterModels' type and
                         * those filters supported dynamically via 'getDynamicFilterDefinitions'.
                         *
                         * When the returned Promise resolves, the extension can assume that 'context.filtering.filterModels' is consistent
                         * with the updated 'filters' here passed here.
                         *
                         * @param filters Updated dashboard and/or part-level filter models
                         * @return A Promise reflecting the updating of the Dashboard and/or Part-level filter state as well as 'context.dashboard.filtering.filterModels'.
                         */
                        updateFilters(filters: {
                            dashboard?: Partial<TFilterModels>;
                            part?: Partial<TFilterModels>;
                        }): Q.Promise<void>;
                    };
                };
            }
        }
    }
}

// FILE: Fx\Composition\ItemType.d.ts
declare module "Fx/Composition/ItemType" {
    /**
     * These are the composition items that have TypeScript decorator support.
     */
    export const enum ItemType {
        Blade = 0,
        FrameBlade = 1,
        MenuBlade = 2,
        TemplateBlade = 3,
        ButtonPart = 4,
        FramePart = 5,
        Part = 6,
        TemplatePart = 7,
    }
}

// FILE: Fx\Composition\MenuBlade.d.ts
declare module "Fx/Composition/MenuBlade" {
    import FxComposition = require("Fx/Composition");
    import FxBladeBase = require("Fx/Composition/BladeBase");
    import * as Menu from "Fx/Controls/Menu";
    export = MsPortalFxMenuBlade;
    module MsPortalFxMenuBlade {
        /**
         * The @MenuBlade decorator.  Identifies a class within the extension as a MenuBlade view model class.
         *
         * @options Metadata describing the MenuBlade and how it will be treated by the FX.
         */
        function Decorator(options?: Options): (menuBladeClass: MenuBladeClass) => void;
        /**
         * Constrains the @MenuBlade decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface MenuBladeClass extends FxBladeBase.BladeClass {
            new (): Contract<any, any>;
        }
        /**
         * Constrains the @MenuBlade decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface Contract<TParameters, TModel> extends FxBladeBase.Contract<TParameters, TModel> {
            /**
             * The view model that the FX will bind into the view for this MenuBlade.  This property must be set before 'container.revealContent()'
             * is called (if it is) and before the promise returned from 'onInitialize' is resolved.
             */
            viewModel: ViewModel2;
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and
             * optional 'onRebind' methods.
             */
            context: Context<TParameters, TModel>;
        }
        /**
         * Status bar for a MenuBlade.
         */
        interface StatusBar extends FxBladeBase.StatusBar {
            /**
             * The state of the MenuBlade.
             */
            state: ContentState;
        }
        /**
         * Specifies which content state decoration should be applied to a MenuBlade.
         */
        export import ContentState = FxBladeBase.ContentState;
        /**
         * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
         * 'onRebind' methods.
         */
        interface Context<TParameters, TModel> extends FxBladeBase.Context<TParameters, TModel> {
            /**
             * A container object that can be used to control the chrome of the MenuBlade.
             */
            container: Container;
        }
        /**
         * Options supplied to the @MenuBlade decorator.  Includes metadata describing the MenuBlade and how it will be treated by
         * the FX.
         */
        type Options = FxBladeBase.Options;
        /**
         * Represents a container object that can be used to control the chrome of the MenuBlade.
         */
        interface Container extends FxBladeBase.Container {
            /**
             * The status bar of the MenuBlade.
             */
            statusBar: KnockoutObservable<StatusBar>;
            /**
             * Moves the container into an 'unauthorized' display mode, allowing for an optional custom error message or notice options.
             *
             * @param message A custom error message or notice options in place of the default.
             */
            unauthorized(message?: string | NoticeOptions): void;
        }
        /**
         * The type of shield to show when there is one or more operations in progress.
         */
        export import ShieldType = FxComposition.ShieldType;
        /**
         * The options for the notice blade.
         */
        export import NoticeOptions = MsPortalFx.ViewModels.NoticeOptions;
        module Rebindable {
            /**
             * The @Rebindable decorator.  This expands the MenuBlade view model such that it implements an 'onRebind' method that will be
             * called when the MenuBlade's parameters are changed.  Parameter changes are caused by, for example, the parent Blade/Part making
             * a call to a 'container.openBlade()'-related API or by the user updating browser's address bar.
             *
             * When this decorator is not used, whenever parameters change value, the MenuBlade view model will be disposed and
             * a new one instantiated and initialized.
             */
            function Decorator(): (rebindableClass: RebindableClass) => void;
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface RebindableClass {
                new (): Contract;
                _fx?: {
                    rebindable?: boolean;
                };
            }
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface Contract {
                /**
                 * A method called by the FX whenever the view model's parameters change value.  When called, the view model should
                 * be updated to reflect the new parameter/settings values.
                 *
                 * @param changes A description of the parameters that have changed.
                 * @return A promise that should be resolved once the view model has been entirely updated to reflect the new values for
                 * parameters/settings.
                 */
                onRebind(changes: Changes): Q.Promise<any>;
            }
            /**
             * Passed to the view model's 'onRebind' method, reflecting the changes to the view model's parameters.
             */
            type Changes = FxBladeBase.Rebindable.Changes;
        }
        module ReturnsData {
            /**
             * The @ReturnsData decorator.  Adds a strongly-typed 'context.container.closeCurrentBlade(data: TData)' to the
             * MenuBlade.
             */
            function Decorator(): (returnsDataClass: ReturnsDataClass) => void;
            /**
             * Constrains the @ReturnsData decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface ReturnsDataClass {
                new (): Contract<any>;
            }
            /**
             * Constrains the @ReturnsData decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface Contract<TData> {
                /**
                 * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and
                 * optional 'onRebind' methods.
                 */
                context: Context<TData>;
            }
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
             * 'onRebind' methods.
             */
            interface Context<TData> {
                /**
                 * A container object that can be used to control the chrome of the MenuBlade.
                 */
                container: FxBladeBase.ReturnsData.Container<TData>;
            }
        }
        /**
         * The sort order used for groups in the menu.
         */
        export import SortOrder = Menu.SortOrder;
        /**
         * The sort function used to order groups in the menu.
         */
        export import SortFunction = Menu.SortFunction;
        namespace Internal {
            /**
             * Options supplied to the ViewModel constructor.
             */
            interface ViewModelOptions extends Menu.Options {
                /**
                 * The groups of menu items.
                 */
                groups: MenuGroup[] | KnockoutReadOnlyObservableBase<MenuGroup[]>;
                /**
                 * Specifies the sort order used for groups in the menu.
                 */
                groupSortOrder?: SortOrder;
                /**
                 * Specifies the sort function used to order groups in the menu.
                 */
                groupSortFunction?: SortFunction;
                /**
                 * A value indicating whether or not to show the search box.
                 */
                showSearch?: boolean;
            }
        }
        /**
         * Defines the view model for a menu blade.
         */
        class ViewModel2 {
            private _msPortalFxMenu;
            /**
             * Creates the view model for a menu blade.
             *
             * @param container The container for the view model.
             * @param options Options for the view model.
             * @return The view model.
             */
            static create<T extends Internal.ViewModelOptions>(container: Container, options: T): ViewModel2 & T;
            private _initialize(container, options);
        }
        /**
         * Defines an item in a group of the menu.
         */
        interface MenuItem extends Menu.MenuItem {
        }
        /**
         * Defines a group in the menu.
         */
        interface MenuGroup extends Menu.MenuGroup {
        }
        /**
         * This class has been relocated.
         * It is now in the Fx/Composition/Pdl/MenuBlade module.
         */
        let ViewModel: Obsolete;
    }
}

// FILE: Fx\Composition\Part.d.ts
declare module "Fx/Composition/Part" {
    import FxComposition = require("Fx/Composition");
    import FxViewModelBase = require("Fx/Composition/ViewModelBase");
    import FxPartBase = require("Fx/Composition/PartBase");
    import FxFilters = require("Fx/Composition/DashboardFilters");
    export = MsPortalFxPart;
    module MsPortalFxPart {
        /**
         * The @Part decorator.  Identifies a class within the extension as a Part view model class.
         *
         * @options Metadata describing the Part and how it will be treated by the FX.
         */
        function Decorator(options?: Options): (partClass: PartClass) => void;
        /**
         * Constrains the @Part decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface PartClass extends FxPartBase.PartClass {
            new (): Contract<any, any>;
        }
        /**
         * Constrains the @Part decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface Contract<TParameters, TModel> extends FxPartBase.Contract<TParameters, TModel> {
            /**
             * The view model that the FX will bind into the view for this Part.  This property must be set before 'container.revealContent()'
             * is called (if it is) and before the promise returned from 'onInitialize' is resolved.
             */
            viewModel: any;
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and
             * optional 'onRebind' methods.
             */
            context: Context<TParameters, TModel>;
        }
        /**
         * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
         * 'onRebind' methods.
         */
        interface Context<TParameters, TModel> extends FxPartBase.Context<TParameters, TModel> {
            /**
             * A container object that can be used to control the chrome of the Part.
             */
            container: Container;
        }
        /**
         * Represents a container object that can be used to control the chrome of the Part.
         */
        interface Container extends FxPartBase.Container {
            /**
             * Detailed Part size information.
             * The Part class should subscribe to this observable to be notified of size changes.
             */
            size: KnockoutReadOnlyObservableBase<SizeInfo>;
            /**
             * Indicates the location of this Part (on a Dashboard, on a Blade, etc.).
             */
            location: Location;
        }
        /**
         * The type of shield to show when there is one or more operations in progress.
         */
        export import ShieldType = FxComposition.ShieldType;
        /**
         * Detailed Part size information.
         */
        interface SizeInfo extends FxPartBase.SizeInfo {
            /**
             * One of the defined Part sizes, including 'Custom' if the Part is configured to support arbitrary sizing.
             */
            partSize: Size;
        }
        /**
         * Indicates the location of this Part (on a Dashboard, on a Blade, etc.).
         */
        export import Location = FxPartBase.Location;
        /**
         * Options supplied to the @Part decorator.  Includes metadata describing the Part and how it will be treated by
         * the FX.
         */
        interface Options extends FxPartBase.Options, FxViewModelBase.AcceptsStyleSheetsOptions {
            /**
             * The sizes supported by this Part.
             */
            supportedSizes?: Size[];
            /**
             * The initial size with which the Part will be rendered.  Must be one of the 'supportedSizes'.
             */
            initialSize?: Size;
            /**
             * The resize mode for the Part.  If not supplied, the default value of 'Fixed' will be used.
             */
            resizeMode?: ResizeMode;
        }
        /**
         * The available sizes for the Part.
         */
        export import Size = FxPartBase.Size;
        /**
         * The available resize modes for the Part.
         */
        export import ResizeMode = FxPartBase.ResizeMode;
        /**
         * Metadata describing a parameter passed to this Part.
         */
        type ParameterMetadata = FxPartBase.ParameterMetadata;
        /**
         * Describes how the image should be stretched to fill allocated space in the Part Gallery entry.
         */
        export import PartGalleryThumbnailStretch = FxPartBase.PartGalleryThumbnailStretch;
        module Rebindable {
            /**
             * The @Rebindable decorator.  This expands the Part view model such that it implements an 'onRebind' method that will be
             * called when the Part's parameters are changed.  Parameter changes are caused by, for example, the user modifying a Dashboard
             * value ('timeRange', for instance) to which the Part is bound.
             *
             * When this decorator is not used, whenever parameters change value, the Part view model will be disposed and
             * a new one instantiated and initialized.
             */
            function Decorator(): (rebindableClass: RebindableClass) => void;
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface RebindableClass {
                new (): Contract;
                _fx?: {
                    rebindable?: boolean;
                };
            }
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface Contract {
                /**
                 * A method called by the FX whenever the view model's parameters change value.  When called, the view model should
                 * be updated to reflect the new parameter/settings values.
                 *
                 * @param changes A description of the parameters that have changed.
                 * @return A promise that should be resolved once the view model has been entirely updated to reflect the new values for
                 * parameters/settings.
                 */
                onRebind(changes: Changes): Q.Promise<any>;
            }
            /**
             * Passed to the view model's 'onRebind' method, reflecting the changes to the view model's parameters.
             */
            type Changes = FxPartBase.Rebindable.Changes;
        }
        module Configurable {
            /**
             * The @Configurable decorator.  View models that use this decorator will make use of 'context.configuration' to
             * obtain the view model's configuration API, with which the view model can update its 'parameters' and persisted 'settings'
             * values.  A new 'settings' property is added to the view model's 'context' which -- along with 'parameters' -- are used in the
             * view model's 'onInitialize' and optional 'onRebind' methods.
             *
             * @param options Options used to configure the @Configurable decorator, supplying additional metadata for the view model's
             * settings.
             */
            function Decorator(options?: Options): (configurableClass: ConfigurableClass) => void;
            /**
             * Options supplied to the @Configurable decorator.  Used to supply metadata describing the Part's settings.
             */
            type Options = FxPartBase.Configurable.Options;
            /**
             * Metadata describing one setting of the Part's persisted settings.
             */
            type SettingMetadata = FxPartBase.Configurable.SettingMetadata;
            /**
             * The scope at which the Part's settings are persisted.
             */
            export import SettingsScope = FxPartBase.Configurable.SettingsScope;
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the Part's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface ConfigurableClass {
                new (): Contract<any, any>;
                _fx?: {
                    configurableOptions?: Options;
                };
            }
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the Part's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface Contract<TParameters, TSettings> {
                /**
                 * Context injected by the FX into the view model.  This context will include a 'configuration' (the Part's
                 * configuration API), as well as 'parameters' and 'settings' (the Part's persisted settings values).
                 */
                context: Context<TParameters, TSettings>;
            }
            /**
             * Context injected by the FX into the view model.  This context will include 'configuration' (the Part's configuration
             * API), the 'parameters' of the Part, as well as its persisted 'settings' values.
             */
            interface Context<TParameters, TSettings> extends FxPartBase.Configurable.Context<TSettings> {
                /**
                 * The view model's configuration API, with which the view model can update its parameters and settings.
                 */
                configuration: Configuration<TParameters, TSettings>;
            }
            /**
             * The Part's configuration API returned from 'context.configuration'.
             */
            type Configuration<TParameters, TSettings> = FxPartBase.Configurable.Configuration<TParameters, TSettings>;
        }
        module ReplacesOldPart {
            /**
             * Options that are supplied to the @ReplacesOldPart decorator
             */
            type Options = FxPartBase.ReplacesOldPart.Options;
            /**
             * The @ReplacesOldPart decorator.  This decorator is used to identify a old part that this
             * Part replaces.
             *
             * @param options options used to configure the @ReplacesOldPart decorator.
             */
            function Decorator(options: Options): (partClass: PartClass) => void;
        }
        /**
         * The @ProxiedMember decorator.  Identifies a public member of a Part class that is to be proxied for use in the view.
         * This decorator is not necessary in most scenarios.
         * This decorator suppresses any compile-time errors that state "The '<MEMBER>' member is public and is not bound to the Part's view."
         */
        export import ProxiedMember = FxViewModelBase.ProxiedMember;
        namespace Filterable {
            /**
             * The @Filterable decorator.  Parts that carry this decorator will receive the state of the those filters currently
             * applied to the containing Dashboard.  They'll receive a "filter model" for each applied filter that the Part supports.
             * (see the 'Part.Filterable.Context' type and its 'filterModels' property, injected into the Part as its 'context' member).
             * Parts will make use of the current "filter models":
             * - in their 'onInitialize' method, to render the Part
             * - in their 'onRebind' method, to update the rendered Part to reflect user changes to Dashboard filters
             */
            function Decorator(): (filterableClass: FilterableClass) => void;
            interface FilterableClass {
                new (): Contract<any>;
                _fx?: {
                    filterable?: boolean;
                };
            }
            /**
             * The interface type implemented by any class that carries the '@Part.Filterable.Decorator' decorator.
             */
            interface Contract<TFilterModels> {
                /**
                 * Context injected by the FX into the Part class.  This context should be used within the Part's 'onInitialize'
                 * and 'onRebind' methods.
                 */
                context: Context<TFilterModels>;
                /**
                 * As with the 'onRebind' method required when using the @Part.Rebindable.Decorator decorator,
                 * this method will be called by the FX whenever the filters for the current Dashboard are updated by the user.
                 *
                 * @param changes An indication of the filter ids whose corresponding filters have been added/removed/updated.
                 * @return A Promise reflecting the time spent loading updated Part data and making corresponding
                 * changes to the Part's view.
                 */
                onRebind(changes: Changes): Q.Promise<any>;
                /**
                 * Optional.  Returns filter definitions for extension-defined filters known statically.  These static filters are those known
                 * at development time (and not those derived from mutable backend schema, for instance).  Such extension-defined
                 * filters are ones like EqualityFilter.FilterDefinition (and not FX-intrinsic filters like TimeRangeFilter).
                 * This method should return a filter definition for each extension-defined filters declared on the Part's
                 * 'TFilterModels' type included in 'public context: Part.Filterable.Context<TFilterModels>'.
                 *
                 * This method will be called by the FX some time after the FX calls this Part's 'onInitialize' (to not compete
                 * with the initiation of Part data-loading).  When called, this method should not assume that the Part is
                 * fully initialized.
                 *
                 * @return An array of filter definitions for each extension-defined filter declared on the Part's 'TFilterModels' type.
                 */
                getFilterDefinitions?(): FxFilters.FilterDefinition[];
                /**
                 * Optional.  Returns filter definitions for extension-defined filters known only dynamically (for instance, derived from
                 * mutable backend schema).  Such extension-defined filters are ones like EqualityFilter.FilterDefinition (and not FX-intrinsic
                 * filters like TimeRangeFilter).
                 *
                 * This method will be called by the FX some time after the FX calls this Part's 'onInitialize' (to not compete
                 * with the initiation of Part data-loading).  When called, this method should not assume that the Part is
                 * fully initialized.
                 *
                 * @return An array of filter definitions for each extension-defined filter only known dynamically.
                 */
                getDynamicFilterDefinitions?(): Q.Promise<FxFilters.DynamicFilterDefinition[]>;
                /**
                 * Required only if any filter definition 'getValues' callbacks returns a localizable filter value.
                 * In such cases, this method will be called to map one or more such localizable filter values to display text in the Portal
                 * user's locale.
                 *
                 * @param filterId The filter id for the filter definition whose 'getValues' callback returned a localizable value.
                 * @param valueIds One or more ids of localizable values.
                 * @return A Promise resolving with corresponding display text for each localizable value, according to the user's locale.
                 */
                getLocalizedFilterValues?(filterId: string, valueIds: string[]): Q.Promise<FxFilters.LocalizedFilterValue[]>;
            }
            /**
             * A description of filter changes (adds/removes/updates) that triggered a call to the Part's 'onRebind' method,
             * to reflect user modification to Dashboard filters.
             */
            interface Changes {
                /**
                 * An array of filter ids covering the added/removed/updated Dashboard filters.
                 */
                filterIds: string[];
            }
            /**
             * Context injected by the FX into the Part class, including extra APIs supporting the
             * '@Part.Filterable.Decorator' decorator.
             */
            interface Context<TFilterModels> {
                /**
                 * FX APIs related to Dashboards.
                 */
                dashboard: {
                    /**
                     * FX APIs related to Dashboard filtering.
                     */
                    filtering: {
                        /**
                         * For those filters supported by this Part, these are the filter models for those filters currently applied
                         * to the active Dashboard and the Part-level filter overrides.  The supported filters are those static filters declared in this Part's
                         * 'TFilterModels' type and those dynamic filters returned from 'getDynamicFilterDefinitions'.
                         */
                        filterModels: TFilterModels;
                        /**
                         * If this Part includes UX to update filters that apply to the Part, the Part can include an option for
                         * the user to update filters at the Dashboard level, using this method.
                         *
                         * The filters accepted here are a subset of those declared to be supported statically via the 'TFilterModels' type and
                         * those filters supported dynamically via 'getDynamicFilterDefinitions'.
                         *
                         * When the returned Promise resolves, the extension can assume that 'context.filtering.filterModels' is consistent
                         * with the updated 'dashboardFilters' here passed here.
                         *
                         * @param dashboardFilters Updated dashboard filter models
                         * @return A Promise reflecting the updating of the Dashboard filter state as well as 'context.dashboard.filtering.filterModels'.
                         */
                        updateFilters(dashboardFilters: Partial<TFilterModels>): Q.Promise<void>;
                        /**
                         * If this Part includes UX to update filters that apply to the Part, the Part can include an option for
                         * the user to update filters at the Dashboard level and/or part level, using this method.
                         *
                         * The filters accepted here are a subset of those declared to be supported statically via the 'TFilterModels' type and
                         * those filters supported dynamically via 'getDynamicFilterDefinitions'.
                         *
                         * When the returned Promise resolves, the extension can assume that 'context.filtering.filterModels' is consistent
                         * with the updated 'filters' here passed here.
                         *
                         * @param filters Updated dashboard and/or part-level filter models
                         * @return A Promise reflecting the updating of the Dashboard and/or Part-level filter state as well as 'context.dashboard.filtering.filterModels'.
                         */
                        updateFilters(filters: {
                            dashboard?: Partial<TFilterModels>;
                            part?: Partial<TFilterModels>;
                        }): Q.Promise<void>;
                    };
                };
            }
        }
    }
}

// FILE: Fx\Composition\PartBase.d.ts
declare module "Fx/Composition/PartBase" {
    import FxComposition = require("Fx/Composition");
    import FxViewModelBase = require("Fx/Composition/ViewModelBase");
    import { ItemType } from "Fx/Composition/ItemType";
    import { DialogManagement } from "Fx/Composition/Dialog";
    export = Main;
    module Main {
        import FxConfiguration = MsPortalFx.Composition.Configuration;
        import FxParts = MsPortalFx.Parts;
        /**
         * The class type to which the various Part decorators can be applied.
         */
        interface PartClass extends FxViewModelBase.ViewModelClass<any> {
            _fx?: {
                itemType?: ItemType;
                options?: Options;
                ownsEditScope?: boolean;
                rebindable?: boolean;
                parameterProvider?: void;
                configurableOptions?: FxViewModelBase.Configurable.Options;
                filterable?: boolean;
            };
        }
        /**
         * A common interface type for the various Part view model classes.
         */
        interface Contract<TParameters, TModel> extends FxViewModelBase.Contract<TParameters, TModel> {
            /**
             * The title text displayed for this part. Typically this specifies what kind of information
             * appears inside the part, e.g., "Deployment history", or if that is not applicable then the
             * type of the asset that the part relates to, e.g., "Website".
             * Supply 'null' if no title is desired for the Part.
             */
            title: string | KnockoutObservableBase<string>;
            /**
             * The displayed subtitle for this Part, typically the Asset name for the asset/resource associated with this Part.
             * Supply 'null' if no subtitle is desired for the Part.
             */
            subtitle: string | KnockoutObservableBase<string>;
            /**
             * Describes the behavior of the Part when it is clicked by the user.
             * Supply 'null' if the Part is not clickable.
             */
            onClick: (() => void) | FxComposition.ClickableLink;
        }
        /**
         * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
         * 'onRebind' methods.
         */
        type Context<TParameters, TModel> = FxViewModelBase.Context<TParameters, TModel>;
        /**
         * Represents a container object that can be used to control the chrome of the Part.
         */
        interface Container extends FxViewModelBase.Container, DialogManagement {
            /**
             * If allowed by ResizeMode, will adjust the size of the container to the size of the content, clamped to the grid constraints.
             *
             * The content must be enclosed in a single container div in the content area of the part.
             * If more than one div is found in the content area, the autoSize will be ignored, and a warning logged in Dev Mode.
             */
            autoSize(): void;
            /**
             * If allowed by ResizeMode, will adjust the size of the part to the size specified, clamped to the grid constraints.
             *
             * @param width The desired width, in grid steps, valid in range from 1 to 25.
             * @param height The desired height, in grid steps, valid in range from 1 to 25.
             */
            resizeTo(width: number, height: number): void;
        }
        /**
         * Detailed Part size information.
         */
        interface SizeInfo {
            /**
             * The width/height of the Part in terms of pixels.
             */
            dimensions: {
                width: number;
                height: number;
            };
        }
        /**
         * The location of a Part.
         */
        export import Location = MsPortalFx.ViewModels.PartLocation;
        /**
         * Options supplied to the specific Part decorator.  Includes metadata describing the Part and how it will be
         * treated by the FX.
         */
        interface Options extends FxViewModelBase.Options {
            /**
             * The initial height with which the Part will be rendered.
             * Is applicable only when the 'initialSize' doesn't include a fixed height.
             */
            initialHeight?: number;
            /**
             * The initial width with which the Part will be rendered.
             * Is applicable only when the 'initialSize' doesn't include a fixed width.
             */
            initialWidth?: number;
            /**
             * Metadata describing the parameters passed to this Part.
             */
            parameterMetadata?: {
                [key: string]: ParameterMetadata;
            };
            /**
             * Metadata that describes this Part in the Part Gallery.
             */
            galleryMetadata?: GalleryMetadata;
        }
        /**
         * The available sizes for the Part.
         */
        export import Size = FxParts.PartSize;
        /**
         * The available resize modes for the Part.
         */
        export import ResizeMode = FxParts.ResizeMode;
        /**
         * Metadata describing a parameter passed to this Part.
         */
        interface ParameterMetadata {
            /**
             * When a Part is added to a Dashboard, this type is used to determine compatible Dashboard values to which the
             * associated parameter can be bound.
             */
            valueType: FxConfiguration.ValueType;
        }
        /**
         * Metadata describing the parameters passed to this Part.
         */
        interface GalleryMetadata {
            /**
             * The title of the Part shown in the Part Gallery.
             */
            title: string;
            /**
             * The description of the Part shown in the Part Gallery.
             */
            description?: string;
            /**
             * The category to which the Part belongs.
             */
            category?: string;
            /**
             * A description of the thumbnail to be shown for the Part's entry in the Part Gallery.
             */
            thumbnail?: {
                /**
                 * The image to be shown for the Part's entry in the Part Gallery.
                 */
                image: MsPortalFx.Base.Image;
                /**
                 * Describes how the image should be stretched to fill allocated space in the Part Gallery entry.
                 */
                stretch?: PartGalleryThumbnailStretch;
            };
            /**
             * The feature name used to hide this Part in the Part Gallery.
             */
            feature?: string;
        }
        /**
         * Describes how the image should be stretched to fill allocated space in the Part Gallery entry.
         */
        export import PartGalleryThumbnailStretch = MsPortalFx.Widgets.PartGallery.ThumbnailStretch;
        module Rebindable {
            /**
             * Passed to the view model's 'onRebind' method, reflecting the changes to the view model's parameters.
             */
            type Changes = FxViewModelBase.Rebindable.Changes;
        }
        module Configurable {
            /**
             * Options supplied to the @Configurable decorator.  Used to supply metadata describing the Part's settings.
             */
            type Options = FxViewModelBase.Configurable.Options;
            /**
             * Metadata describing one setting of the Part's persisted settings.
             */
            type SettingMetadata = FxViewModelBase.Configurable.SettingMetadata;
            /**
             * The scope at which the Part's settings are persisted.
             */
            export import SettingsScope = FxViewModelBase.Configurable.SettingsScope;
            /**
             * Context injected by the FX into the view model.  This context will include 'configuration' (the Part's configuration
             * API) as well as 'parameters' and 'settings' (the Part's persisted settings values).
             */
            type Context<TSettings> = FxViewModelBase.Configurable.Context<TSettings>;
            /**
             * The Part's configuration API returned from 'context.configuration'.
             */
            type Configuration<TParameters, TSettings> = FxConfiguration.Part.Contract<TParameters, TSettings>;
        }
        module ReplacesOldPart {
            /**
             * Identifies a part on a blade that is to have its pinned instances replaced
             */
            interface ReplacedPartIdentifer {
                /**
                 * Name of the part that is to be replaced
                 */
                part: string;
                /**
                 * The lens which contains the part that is to be replaced
                 */
                lens: string;
                /**
                 * The blade which contains the part that is to be replaced
                 */
                blade: string;
            }
            /**
             * Options supplied to the ReplacesOldPart decorator.
             */
            interface Options {
                /**
                 * The old part that is to be replaced
                 */
                replacedPart: string | ReplacedPartIdentifer;
                /**
                 * The extension which the old part belongs to
                 */
                extension?: string;
            }
        }
    }
}

// FILE: Fx\Composition\PartReferences.d.ts
declare module "Fx/Composition/PartReferences" {
    import FxFilters = require("Fx/Composition/DashboardFilters");
    export = PartReferences;
    module PartReferences {
        interface PartReferenceOptions extends MsPortalFx.Composition.PartReferenceOptions {
            /**
             * The part-level filters to use when displaying data, if applicable.
             */
            filters?: StringMap<FxFilters.AnyFilterModel>;
        }
        /**
         * Identifies the part that is referenced. This type is not intended to be used directly, but instead
         * code generation defines sub-classes of this base type to create a strongly typed reference per Part.
         */
        class PartReference<TParameters> extends MsPortalFx.Composition.PartReference<TParameters> {
            /**
             * Options to create the target part
             */
            options: PartReferenceOptions;
            /**
             * PartReference constructor
             *
             * @param partName partName of the referenced part
             * @param parameters parameters that are passed to the target part
             * @param options options that are used to create the target part
             */
            constructor(partName: string, parameters: TParameters, options?: PartReferenceOptions);
        }
    }
}

// FILE: Fx\Composition\Pdl\Blade.d.ts
declare module "Fx/Composition/Pdl/Blade" {
    import FxComposition = require("Fx/Composition");
    import FxConfiguration = MsPortalFx.Composition.Configuration;
    export interface Container extends MsPortalFx.ViewModels.ContainerContract, MsPortalFx.ViewModels.BladeManagement, MsPortalFx.ViewModels.DialogManagement {
        /**
         * Activates this Blade's 'Configuration', which is an API that enables the Blade to manage parameter and settings overrides/inheritance via
         * configuration UI (typically via a Context Blade).  See the MsPortalFx.Composition.Configuration.Contract interface for details.
         *
         * @param options Options used to initialize the Configuration API for this Part
         * @return The Configuration API for this Part
         */
        activateConfiguration<TSettings>(options?: FxConfiguration.Blade.Options<TSettings>): FxConfiguration.Blade.Contract<TSettings>;
        /**
         * Tracks the list of asynchronous operations being performed by the Blade.
         */
        operations: FxComposition.OperationList;
        /**
         * Moves the blade into a 'not found' display mode, allowing for an optional custom error message.
         *
         * @param message An optional custom error message.
         */
        notFound(message?: string): void;
        /**
         * Moves the blade into an 'unauthorized' display mode, allowing for an an optional custom error message or notice options.
         *
         * @param message A custom error message or notice options in place of the default.
         */
        unauthorized(message?: string | MsPortalFx.ViewModels.NoticeOptions): void;
        /**
         * Shows an error message.
         *
         * @param message A message that will be displayed to the user when the blade fails.
         */
        fail(message: string): void;
        /**
         * Removes any 'failed' error message and brings the blade back to 'ready' state.
         * If the blade was not already in a 'failed' state, this method does nothing.
         */
        recover(): void;
        /**
         * Causes the blades's content to be revealed immediately, regardless of any onInputsSet promise that
         * might not yet have completed. This removes the opaque 'loading' spinner and makes the blae interactive,
         * so you must also disable or hide any UI elements that the user should not be able to see or interact with
         * until data is fully loaded.
         */
        revealContent(): void;
        /**
         * The message displayed by the Part when it has no data to display.
         */
        noDataMessage: KnockoutObservable<string>;
        /**
         * Closes this blade
         *
         * @param data Optional value to return back to the parent blade
         * @return a promise that resolves to true if the this blade is successfully closed.
         */
        closeCurrentBlade(data?: any): Q.Promise<boolean>;
        /**
         * Closes the child blade that is currently open
         *
         * @return a promise that resolves to true if the child blade is successfully closed.
         */
        closeChildBlade(): Q.Promise<boolean>;
        /**
         * Closes the context blade that was opened by this part
         *
         * @returns a promise that resolves to true if the context blade is sucessfully closed.
         */
        closeContextBlade(): Q.Promise<boolean>;
        /**
         * Properties to track form validity, dirty state of forms on this blade and methods to trigger validation and modify alert behavior
         */
        form: MsPortalFx.ViewModels.FormProperties;
        /**
         * Interface for managing menu blade.   This property is null if the blade is not a opened by a menu blade.
         */
        menu: MsPortalFx.ViewModels.MenuBladeManagement;
        /**
         * The path to the help content file.
         */
        helpContentUri: KnockoutObservable<string>;
        /**
         * API collection that facilitates dashboard interactions.
         */
        dashboard: MsPortalFx.ViewModels.DashboardManagement;
    }
}

// FILE: Fx\Composition\Pdl\MenuBlade.d.ts
declare module "Fx/Composition/Pdl/MenuBlade" {
    import * as Menu from "Fx/Controls/Menu";
    export = Main;
    module Main {
        /**
         * The options for the menu.
         */
        interface Options extends Menu.Options {
        }
        /**
         * The sort functions used to sort groups in the menu.
         */
        const SortFunction: typeof Menu.SortFunction;
        /**
         * Defines the view model for a menu blade.
         */
        class ViewModel extends MsPortalFx.ViewModels.Blade {
            menu: Menu.ViewModel;
            /**
             * Creates the view model for a menu blade.
             *
             * @param container The container for the view model.
             */
            constructor(container: MsPortalFx.ViewModels.ContainerContract);
        }
        /**
         * The sort order used for groups in the menu.
         */
        export import SortOrder = Menu.SortOrder;
        /**
         * Defines an item in a group of the menu.
         */
        interface MenuItem extends Menu.MenuItem {
        }
        /**
         * Defines a group in the menu.
         */
        interface MenuGroup extends Menu.MenuGroup {
        }
    }
}

// FILE: Fx\Composition\Permissions.d.ts
declare module "Fx/Composition/Permissions" {
    /**
     * Permissions that will be evaluated before rendering a Blade/Part.
     */
    export type Permissions = (Permission | PermissionSet)[];
    /**
     * A permission that will be evaluated for a resource before the associated Blade/Part UI will be shown/enabled.
     */
    export interface Permission {
        /**
         * The 'action' name for this permission ('read', 'write', etc.).
         */
        action: string;
        /**
         * An optional target of the permissions check.  If this is not supplied, then the target resource of this
         * permissions check is the asset associated with this Blade/Part.  To associate a Blade/Part with an asset,
         * use the 'Options.forAsset' property supplied to the Blade/Part decorator.
         */
        targetResource?: ResourceIdSource;
    }
    /**
     * A set of permissions that will be evaluated for a resource before the associated Blade/Part UI will be
     * shown/enabled.
     */
    export interface PermissionSet {
        /**
         * The set of permissions and/or permission references.
         */
        permissions: Permission[];
        /**
         * Indicates how the result from each of the N permission checks will be considered before allowing the
         * associated Blade/Part UI to be shown/enabled.
         */
        requirement: SetRequirement;
    }
    /**
     * Specifies the target asset for a permission or permission reference.
     */
    export interface ResourceIdSource {
        /**
         * A reference to an asset type definition that defines the resource that is the target of some permission or
         * permission reference.
         * If not supplied, then the asset type definition used here is that which is associated with the Blade/Part
         * (using the 'Options.forAsset' property supplied to the Blade/Part decorator).
         */
        assetType?: {
            /**
             * The name of the asset type definition.
             */
            assetType: string;
            /**
             * The name of the extension in which the asset type is defined.  When this is not supplied, this defaults
             * to the current extension.
             */
            extensionName?: string;
        };
        /**
         * If supplied, this indicates that a "property provider service" associated with the asset type definition is
         * to be used to determine the resource id that is the target of a permission / permission reference.
         * When supplied, this value contains the 'key' and 'property' parameters passed to the "property provider
         * service".
         *
         * If this is not supplied, then it is assumed that the associated permission or permission reference will be
         * evaluated against the resource associated with the Blade/Part via the 'Options.forAsset' property passed to
         * the Blade/Part decorator.
         */
        providerParameters?: {
            /**
             * The 'key' parameter passed to the "property provider service".
             */
            key: string;
            /**
             * The 'property' parameter passed to the "property provider service".  This is the view model property
             * where the resource id is to be published.
             */
            property: string;
        };
    }
    /**
     * Indicates how the result from each of the N permission checks will be considered before allowing the associated
     * Blade/Part UI to be shown/enabled.
     */
    export type SetRequirement = MsPortalFx.Extension.SetRequirement;
}

// FILE: Fx\Composition\TemplateBlade.d.ts
declare module "Fx/Composition/TemplateBlade" {
    import FxComposition = require("Fx/Composition");
    import FxBladeBase = require("Fx/Composition/BladeBase");
    import { DialogManagement } from "Fx/Composition/Dialog";
    import FxViewModelBase = require("Fx/Composition/ViewModelBase");
    import * as Provisioning from "Fx/ResourceManagement/Provisioning";
    export = MsPortalFxTemplateBlade;
    module MsPortalFxTemplateBlade {
        import PartReference = MsPortalFx.Composition.PartReference;
        /**
         * The @TemplateBlade decorator.  Identifies a class within the extension as a TemplateBlade view model class.
         *
         * @options Metadata describing the TemplateBlade and how it will be treated by the FX.
         */
        function Decorator(options: Options): (templateBladeClass: TemplateBladeClass) => void;
        /**
         * Constrains the @TemplateBlade decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface TemplateBladeClass extends FxBladeBase.BladeClass {
            new (): Contract<any, any>;
        }
        /**
         * Constrains the @TemplateBlade decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface Contract<TParameters, TModel> extends FxBladeBase.Contract<TParameters, TModel> {
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and
             * optional 'onRebind' methods.
             */
            context: Context<TParameters, TModel>;
        }
        /**
         * Status bar for a TemplateBlade.
         */
        interface StatusBar extends FxBladeBase.StatusBar {
            /**
             * The state of the TemplateBlade.
             */
            state: ContentState;
        }
        /**
         * Specifies which content state decoration should be applied to a TemplateBlade.
         */
        export import ContentState = FxBladeBase.ContentState;
        /**
         * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
         * 'onRebind' methods.
         */
        interface Context<TParameters, TModel> extends FxBladeBase.Context<TParameters, TModel> {
            /**
             * A container object that can be used to control the chrome of the TemplateBlade.
             */
            container: Container;
            /**
             * An API a TemplateBlade can use to interact with the Form Fields on the Blade.
             */
            form: FxBladeBase.FormManagement;
        }
        /**
         * Describes if/when to display an alert when the blade closes.
         */
        export import AlertLevel = FxBladeBase.AlertLevel;
        /**
         * Options supplied to the @TemplateBlade decorator.  Includes metadata describing the TemplateBlade and how it will be treated by
         * the FX.
         */
        interface Options extends FxBladeBase.Options, FxViewModelBase.RequiresHtmlTemplateOptions, FxViewModelBase.AcceptsStyleSheetsOptions, FxBladeBase.ShrinkOnOpenChildBladeOptions {
        }
        /**
         * Represents a container object that can be used to control the chrome of the TemplateBlade.
         */
        interface Container extends FxBladeBase.Container, FxBladeBase.CanHaveCommandBarContainer, DialogManagement {
            /**
             * The status bar of the TemplateBlade.
             */
            statusBar: KnockoutObservable<StatusBar>;
            /**
             * Moves the container into an 'unauthorized' display mode, allowing for an optional custom error message or notice options.
             *
             * @param message A custom error message or notice options in place of the default.
             */
            unauthorized(message?: string | NoticeOptions): void;
        }
        /**
         * The type of shield to show when there is one or more operations in progress.
         */
        export import ShieldType = FxComposition.ShieldType;
        /**
         * The options for the notice blade.
         */
        export import NoticeOptions = MsPortalFx.ViewModels.NoticeOptions;
        module Rebindable {
            /**
             * The @Rebindable decorator.  This expands the TemplateBlade view model such that it implements an 'onRebind' method that will be
             * called when the TemplateBlade's parameters are changed.  Parameter changes are caused by, for example, the parent Blade/Part making
             * a call to a 'container.openBlade()'-related API or by the user updating browser's address bar.
             *
             * When this decorator is not used, whenever parameters change value, the TemplateBlade view model will be disposed and
             * a new one instantiated and initialized.
             */
            function Decorator(): (rebindableClass: RebindableClass) => void;
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface RebindableClass {
                new (): Contract;
                _fx?: {
                    rebindable?: boolean;
                };
            }
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface Contract {
                /**
                 * A method called by the FX whenever the view model's parameters change value.  When called, the view model should
                 * be updated to reflect the new parameter/settings values.
                 *
                 * @param changes A description of the parameters that have changed.
                 * @return A promise that should be resolved once the view model has been entirely updated to reflect the new values for
                 * parameters/settings.
                 */
                onRebind(changes: Changes): Q.Promise<any>;
            }
            /**
             * Passed to the view model's 'onRebind' method, reflecting the changes to the view model's parameters.
             */
            type Changes = FxBladeBase.Rebindable.Changes;
        }
        module Configurable {
            /**
             * The @Configurable decorator.  View models that use this decorator will make use of 'context.configuration' to
             * obtain the view model's configuration API, with which the view model can update its persisted 'settings' values.  The 'settings'
             * values are made available in the view model's 'context', for use in the view model's 'onInitialize' and optional 'onRebind'
             * methods.
             *
             * @param options Options used to configure the @Configurable decorator, supplying additional metadata for the view model's
             * settings.
             */
            function Decorator(options?: Options): (configurableClass: ConfigurableClass) => void;
            /**
             * Options supplied to the @Configurable decorator.  Used to supply metadata describing the TemplateBlade's settings.
             */
            type Options = FxBladeBase.Configurable.Options;
            /**
             * Metadata describing one setting of the TemplateBlade's persisted settings.
             */
            type SettingMetadata = FxBladeBase.Configurable.SettingMetadata;
            /**
             * The scope at which the TemplateBlade's settings are persisted.
             */
            export import SettingsScope = FxBladeBase.Configurable.SettingsScope;
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the TemplateBlade's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface ConfigurableClass {
                new (): Contract<any>;
                _fx?: {
                    configurableOptions?: Options;
                };
            }
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the TemplateBlade's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface Contract<TSettings> {
                /**
                 * Context injected by the FX into the view model.  This context will include 'configuration' (the TemplateBlade's configuration
                 * API) as well as 'settings' (the TemplateBlade's persisted settings values).
                 */
                context: Context<TSettings>;
            }
            /**
             * Context injected by the FX into the view model.  This context will include 'configuration' (the TemplateBlade's configuration API),
             * the 'parameters' of the TemplateBlade, as well as its persisted 'settings' values.
             */
            interface Context<TSettings> extends FxBladeBase.Configurable.Context<TSettings> {
                /**
                 * The view model's configuration API, with which the view model can update its settings.
                 */
                configuration: Configuration<TSettings>;
            }
            /**
             * The TemplateBlade's configuration API returned from 'context.configuration'.
             */
            type Configuration<TSettings> = FxBladeBase.Configurable.Configuration<TSettings>;
        }
        module ReturnsData {
            /**
             * The @ReturnsData decorator.  Adds a strongly-typed 'context.container.closeCurrentBlade(data: TData)' to the
             * TemplateBlade.
             */
            function Decorator(): (returnsDataClass: ReturnsDataClass) => void;
            /**
             * Constrains the @ReturnsData decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface ReturnsDataClass {
                new (): Contract<any>;
            }
            /**
             * Constrains the @ReturnsData decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface Contract<TData> {
                /**
                 * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and
                 * optional 'onRebind' methods.
                 */
                context: Context<TData>;
            }
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
             * 'onRebind' methods.
             */
            interface Context<TData> {
                /**
                 * A container object that can be used to control the chrome of the TemplateBlade.
                 */
                container: FxBladeBase.ReturnsData.Container<TData>;
            }
        }
        module Pinnable {
            /**
             * The @Pinnable decorator.  Requires that the TemplateBlade implement an 'onPin' method that determines what Part
             * to pin when the user pins the TemplateBlade.
             */
            function Decorator(): (pinnableClass: PinnableClass) => void;
            /**
             * Constrains the @Pinnable decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface PinnableClass {
                new (): Contract;
            }
            /**
             * Constrains the @Pinnable decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface Contract {
                /**
                 * A callback that will be called when the user pins the TemplateBlade.  Returns a PartReference to the Part that will be pinned to the
                 * user's Dashboard.
                 */
                onPin(): PartReference<any>;
            }
        }
        module DoesProvisioning {
            /**
             * The @DoesProvisioning decorator. Adds a strongly-typed 'context.provisioning' to the TemplateBlade.
             * This provisioning API allows the view model to perform provisioning operations, including support
             * for telemetry around provisioning.
             *
             * @param options Options used to configure the @DoesProvisioning decorator.
             */
            function Decorator(options?: Options): (doesProvisioningClass: DoesProvisioningClass) => void;
            /**
             * Constrains the @DoesProvisioning decorator so that it can be applied only to classes implementing 'Contract'.
             */
            interface DoesProvisioningClass {
                new (): Contract;
                _fx?: {
                    doesProvisioning?: boolean;
                };
            }
            /**
             * Options supplied to the @DoesProvisioning decorator.
             */
            interface Options {
                /**
                 * A flag indicating whether the caller is required to supply a Marketplace item id
                 * when instantiating a BladeReference to this TemplateBlade. Defaults to true.
                 */
                requiresMarketplaceId?: boolean;
            }
            /**
             * Constrains the @DoesProvisioning decorator so that it can be applied only to classes
             * implementing 'Contract'.
             */
            interface Contract {
                /**
                 * Context injected by the FX into the view model. This context should be used within
                 * the view model's 'onInitialize' and optional 'onRebind' methods.
                 */
                context: Context<any>;
            }
            /**
             * Context injected by the FX into the view model. This context should be used within the
             * view model's 'onInitialize' and optional 'onRebind' methods.
             */
            interface Context<TUIMetadata = void> {
                /**
                 * The view model's provisioning API, with which the view model can
                 * perform provisioning operations, including support for telemetry
                 * around provisioning.
                 */
                provisioning: FxBladeBase.DoesProvisioning.Provisioning<TUIMetadata>;
            }
            /**
             * Options for the DeployTemplate method on provisioning context
             */
            export import DeployTemplateOptions = Provisioning.DeployTemplateOptions;
            /**
             * Options for the DeployCustom method on provisioning context
             */
            export import DeployCustomOptions = Provisioning.DeployCustomOptions;
            /**
             * The template deployment operation mode. Defaults to 'RequestDeploymentOnly'.
             */
            export import TemplateDeploymentMode = Provisioning.TemplateDeploymentMode;
        }
        module LegacyFeatures {
            /**
             * The @LegacyFeatures decorator.  Used to access legacy features for TemplateBlades, like defining an explicit TemplateBlade width.
             *
             * @param options Options used to configure the @LegacyFeatures decorator.
             */
            function Decorator(options: Options): (bladeClass: TemplateBladeClass) => void;
            /**
             * Options supplied to the @LegacyFeatures decorator.
             */
            interface Options {
                /**
                 * The width for this TemplateBlade.
                 */
                width?: Width;
            }
            /**
             * The available widths for the TemplateBlade.
             */
            export import Width = MsPortalFx.Blades.BladeWidth;
        }
        module ForContextPane {
            /**
             * The @ForContextPane decorator.  Used to configure the behavior of the TemplateBlade when it is opened in the ContextPane.
             *
             * @param options Options used to configure the @ForContextPane decorator.
             */
            function Decorator(options: Options): (bladeClass: TemplateBladeClass) => void;
            /**
             * Options supplied to the @ForContextPane decorator.
             */
            interface Options {
                /**
                 * The width of the TemplateBlade when it is opened in the ContextPane.
                 */
                width?: Width;
            }
            /**
             * The available widths for the ContextPane.
             */
            export import Width = MsPortalFx.Blades.BladeWidth;
        }
        /**
         * The @ProxiedMember decorator.  Identifies a public member of a TemplateBlade class that is to be proxied for use in the view.
         * This decorator is not necessary in most scenarios.
         * This decorator suppresses any compile-time errors that state "The '<MEMBER>' member is public and is not bound to the TemplateBlade's HTML template."
         */
        export import ProxiedMember = FxViewModelBase.ProxiedMember;
    }
}

// FILE: Fx\Composition\TemplatePart.d.ts
declare module "Fx/Composition/TemplatePart" {
    import FxComposition = require("Fx/Composition");
    import FxViewModelBase = require("Fx/Composition/ViewModelBase");
    import FxPartBase = require("Fx/Composition/PartBase");
    import FxFilters = require("Fx/Composition/DashboardFilters");
    export = MsPortalFxTemplatePart;
    module MsPortalFxTemplatePart {
        /**
         * The @TemplatePart decorator.  Identifies a class within the extension as a TemplatePart view model class.
         *
         * @options Metadata describing the TemplatePart and how it will be treated by the FX.
         */
        function Decorator(options: Options): (partClass: TemplatePartClass) => void;
        /**
         * Constrains the @TemplatePart decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface TemplatePartClass extends FxPartBase.PartClass {
            new (): Contract<any, any>;
        }
        /**
         * Constrains the @TemplatePart decorator so that it can be applied only to classes implementing 'Contract'.
         */
        interface Contract<TParameters, TModel> extends FxPartBase.Contract<TParameters, TModel> {
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and
             * optional 'onRebind' methods.
             */
            context: Context<TParameters, TModel>;
        }
        /**
         * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
         * 'onRebind' methods.
         */
        interface Context<TParameters, TModel> extends FxPartBase.Context<TParameters, TModel> {
            /**
             * A container object that can be used to control the chrome of the TemplatePart.
             */
            container: Container;
        }
        /**
         * Represents a container object that can be used to control the chrome of the TemplatePart.
         */
        interface Container extends FxPartBase.Container {
            /**
             * Detailed TemplatePart size information.
             * The TemplatePart class should subscribe to this observable to be notified of size changes.
             */
            size: KnockoutReadOnlyObservableBase<SizeInfo>;
            /**
             * Indicates the location of this TemplatePart (on a Dashboard, on a Blade, etc.).
             */
            location: Location;
        }
        /**
         * The type of shield to show when there is one or more operations in progress.
         */
        export import ShieldType = FxComposition.ShieldType;
        /**
         * Detailed TemplatePart size information.
         */
        interface SizeInfo extends FxPartBase.SizeInfo {
            /**
             * One of the defined TemplatePart sizes, including 'Custom' if the TemplatePart is configured to support arbitrary sizing.
             */
            partSize: Size;
        }
        /**
         * Indicates the location of this TemplatePart (on a Dashboard, on a Blade, etc.).
         */
        export import Location = FxPartBase.Location;
        /**
         * Options supplied to the @TemplatePart decorator.  Includes metadata describing the TemplatePart and how it will be treated by
         * the FX.
         */
        interface Options extends FxPartBase.Options, FxViewModelBase.RequiresHtmlTemplateOptions, FxViewModelBase.AcceptsStyleSheetsOptions {
            /**
             * The sizes supported by this TemplatePart.
             */
            supportedSizes?: Size[];
            /**
             * The initial size with which the TemplatePart will be rendered.  Must be one of the 'supportedSizes'.
             */
            initialSize?: Size;
            /**
             * The resize mode for the TemplatePart.  If not supplied, the default value of 'Fixed' will be used.
             */
            resizeMode?: ResizeMode;
        }
        /**
         * The available sizes for the TemplatePart.
         */
        export import Size = FxPartBase.Size;
        /**
         * The available resize modes for the TemplatePart.
         */
        export import ResizeMode = FxPartBase.ResizeMode;
        /**
         * Metadata describing a parameter passed to this TemplatePart.
         */
        type ParameterMetadata = FxPartBase.ParameterMetadata;
        /**
         * Describes how the image should be stretched to fill allocated space in the Part Gallery entry.
         */
        export import PartGalleryThumbnailStretch = FxPartBase.PartGalleryThumbnailStretch;
        module Rebindable {
            /**
             * The @Rebindable decorator.  This expands the TemplatePart view model such that it implements an 'onRebind' method that will be
             * called when the TemplatePart's parameters are changed.  Parameter changes are caused by, for example, the user modifying a Dashboard
             * value ('timeRange', for instance) to which the TemplatePart is bound.
             *
             * When this decorator is not used, whenever parameters change value, the TemplatePart view model will be disposed and
             * a new one instantiated and initialized.
             */
            function Decorator(): (rebindableClass: RebindableClass) => void;
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface RebindableClass {
                new (): Contract;
                _fx?: {
                    rebindable?: boolean;
                };
            }
            /**
             * Constrains the @Rebindable decorator so that it can be applied only to classes implementing 'Contract', requiring that the class
             * implement an 'onRebind' method.
             */
            interface Contract {
                /**
                 * A method called by the FX whenever the view model's parameters change value.  When called, the view model should
                 * be updated to reflect the new parameter/settings values.
                 *
                 * @param changes A description of the parameters that have changed.
                 * @return A promise that should be resolved once the view model has been entirely updated to reflect the new values for
                 * parameters/settings.
                 */
                onRebind(changes: Changes): Q.Promise<any>;
            }
            /**
             * Passed to the view model's 'onRebind' method, reflecting the changes to the view model's parameters.
             */
            type Changes = FxPartBase.Rebindable.Changes;
        }
        module Configurable {
            /**
             * The @Configurable decorator.  View models that use this decorator will make use of 'context.configuration' to
             * obtain the view model's configuration API, with which the view model can update its 'parameters' and persisted 'settings'
             * values.  A new 'settings' property is added to the view model's 'context' which -- along with 'parameters' -- are used in the
             * view model's 'onInitialize' and optional 'onRebind' methods.
             *
             * @param options Options used to configure the @Configurable decorator, supplying additional metadata for the view model's
             * settings.
             */
            function Decorator(options?: Options): (configurableClass: ConfigurableClass) => void;
            /**
             * Options supplied to the @Configurable decorator.  Used to supply metadata describing the TemplatePart's settings.
             */
            type Options = FxPartBase.Configurable.Options;
            /**
             * Metadata describing one setting of the TemplatePart's persisted settings.
             */
            type SettingMetadata = FxPartBase.Configurable.SettingMetadata;
            /**
             * The scope at which the TemplatePart's settings are persisted.
             */
            export import SettingsScope = FxPartBase.Configurable.SettingsScope;
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the TemplatePart's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface ConfigurableClass {
                new (): Contract<any, any>;
                _fx?: {
                    configurableOptions?: Options;
                };
            }
            /**
             * Constrains the @Configurable decorator so that it can only be applied to a class that implements the 'Contract' interface.
             * This adds to the view model's 'context':
             * - 'configuration', the TemplatePart's configuration API
             * - 'parameters' and persisted 'settings' values, used in the 'onInitialize' and optional 'onRebind' methods to render the view.
             */
            interface Contract<TParameters, TSettings> {
                /**
                 * Context injected by the FX into the view model.  This context will include a 'configuration' (the TemplatePart's
                 * configuration API), as well as 'parameters' and 'settings' (the TemplatePart's persisted settings values).
                 */
                context: Context<TParameters, TSettings>;
            }
            /**
             * Context injected by the FX into the view model.  This context will include 'configuration' (the TemplatePart's configuration
             * API), the 'parameters' of the TemplatePart, as well as its persisted 'settings' values.
             */
            interface Context<TParameters, TSettings> extends FxPartBase.Configurable.Context<TSettings> {
                /**
                 * The view model's configuration API, with which the view model can update its parameters and settings.
                 */
                configuration: Configuration<TParameters, TSettings>;
            }
            /**
             * The TemplatePart's configuration API returned from 'context.configuration'.
             */
            type Configuration<TParameters, TSettings> = FxPartBase.Configurable.Configuration<TParameters, TSettings>;
        }
        module ReplacesOldPart {
            /**
             * Options that are supplied to the @ReplacesOldPart decorator
             */
            type Options = FxPartBase.ReplacesOldPart.Options;
            /**
             * The @ReplacesOldPart decorator.  This decorator is used to identify a old part that this
             * TemplatePart replaces.
             *
             * @param options options used to configure the @ReplacesOldPart decorator.
             */
            function Decorator(options: Options): (partClass: TemplatePartClass) => void;
        }
        /**
         * The @ProxiedMember decorator.  Identifies a public member of a TemplatePart class that is to be proxied for use in the view.
         * This decorator is not necessary in most scenarios.
         * This decorator suppresses any compile-time errors that state "The '<MEMBER>' member is public and is not bound to the TemplatePart's HTML template."
         */
        export import ProxiedMember = FxViewModelBase.ProxiedMember;
        namespace Filterable {
            /**
             * The @Filterable decorator.  TemplateParts that carry this decorator will receive the state of the those filters currently
             * applied to the containing Dashboard.  They'll receive a "filter model" for each applied filter that the TemplatePart supports.
             * (see the 'TemplatePart.Filterable.Context' type and its 'filterModels' property, injected into the TemplatePart as its 'context' member).
             * TemplateParts will make use of the current "filter models":
             * - in their 'onInitialize' method, to render the TemplatePart
             * - in their 'onRebind' method, to update the rendered TemplatePart to reflect user changes to Dashboard filters
             */
            function Decorator(): (filterableClass: FilterableClass) => void;
            interface FilterableClass {
                new (): Contract<any>;
                _fx?: {
                    filterable?: boolean;
                };
            }
            /**
             * The interface type implemented by any class that carries the '@TemplatePart.Filterable.Decorator' decorator.
             */
            interface Contract<TFilterModels> {
                /**
                 * Context injected by the FX into the TemplatePart class.  This context should be used within the TemplatePart's 'onInitialize'
                 * and 'onRebind' methods.
                 */
                context: Context<TFilterModels>;
                /**
                 * As with the 'onRebind' method required when using the @TemplatePart.Rebindable.Decorator decorator,
                 * this method will be called by the FX whenever the filters for the current Dashboard are updated by the user.
                 *
                 * @param changes An indication of the filter ids whose corresponding filters have been added/removed/updated.
                 * @return A Promise reflecting the time spent loading updated TemplatePart data and making corresponding
                 * changes to the TemplatePart's view.
                 */
                onRebind(changes: Changes): Q.Promise<any>;
                /**
                 * Optional.  Returns filter definitions for extension-defined filters known statically.  These static filters are those known
                 * at development time (and not those derived from mutable backend schema, for instance).  Such extension-defined
                 * filters are ones like EqualityFilter.FilterDefinition (and not FX-intrinsic filters like TimeRangeFilter).
                 * This method should return a filter definition for each extension-defined filters declared on the TemplatePart's
                 * 'TFilterModels' type included in 'public context: TemplatePart.Filterable.Context<TFilterModels>'.
                 *
                 * This method will be called by the FX some time after the FX calls this TemplatePart's 'onInitialize' (to not compete
                 * with the initiation of TemplatePart data-loading).  When called, this method should not assume that the TemplatePart is
                 * fully initialized.
                 *
                 * @return An array of filter definitions for each extension-defined filter declared on the TemplatePart's 'TFilterModels' type.
                 */
                getFilterDefinitions?(): FxFilters.FilterDefinition[];
                /**
                 * Optional.  Returns filter definitions for extension-defined filters known only dynamically (for instance, derived from
                 * mutable backend schema).  Such extension-defined filters are ones like EqualityFilter.FilterDefinition (and not FX-intrinsic
                 * filters like TimeRangeFilter).
                 *
                 * This method will be called by the FX some time after the FX calls this TemplatePart's 'onInitialize' (to not compete
                 * with the initiation of TemplatePart data-loading).  When called, this method should not assume that the TemplatePart is
                 * fully initialized.
                 *
                 * @return An array of filter definitions for each extension-defined filter only known dynamically.
                 */
                getDynamicFilterDefinitions?(): Q.Promise<FxFilters.DynamicFilterDefinition[]>;
                /**
                 * Required only if any filter definition 'getValues' callbacks returns a localizable filter value.
                 * In such cases, this method will be called to map one or more such localizable filter values to display text in the Portal
                 * user's locale.
                 *
                 * @param filterId The filter id for the filter definition whose 'getValues' callback returned a localizable value.
                 * @param valueIds One or more ids of localizable values.
                 * @return A Promise resolving with corresponding display text for each localizable value, according to the user's locale.
                 */
                getLocalizedFilterValues?(filterId: string, valueIds: string[]): Q.Promise<FxFilters.LocalizedFilterValue[]>;
            }
            /**
             * A description of filter changes (adds/removes/updates) that triggered a call to the TemplatePart's 'onRebind' method,
             * to reflect user modification to Dashboard filters.
             */
            interface Changes {
                /**
                 * An array of filter ids covering the added/removed/updated Dashboard filters.
                 */
                filterIds: string[];
            }
            /**
             * Context injected by the FX into the TemplatePart class, including extra APIs supporting the
             * '@TemplatePart.Filterable.Decorator' decorator.
             */
            interface Context<TFilterModels> {
                /**
                 * FX APIs related to Dashboards.
                 */
                dashboard: {
                    /**
                     * FX APIs related to Dashboard filtering.
                     */
                    filtering: {
                        /**
                         * For those filters supported by this TemplatePart, these are the filter models for those filters currently applied
                         * to the active Dashboard and the Part-level filter overrides.  The supported filters are those static filters declared in this TemplatePart's
                         * 'TFilterModels' type and those dynamic filters returned from 'getDynamicFilterDefinitions'.
                         */
                        filterModels: TFilterModels;
                        /**
                         * If this TemplatePart includes UX to update filters that apply to the Part, the TemplatePart can include an option for
                         * the user to update filters at the Dashboard level, using this method.
                         *
                         * The filters accepted here are a subset of those declared to be supported statically via the 'TFilterModels' type and
                         * those filters supported dynamically via 'getDynamicFilterDefinitions'.
                         *
                         * When the returned Promise resolves, the extension can assume that 'context.filtering.filterModels' is consistent
                         * with the updated 'dashboardFilters' here passed here.
                         *
                         * @param dashboardFilters Updated dashboard filter models
                         * @return A Promise reflecting the updating of the Dashboard filter state as well as 'context.dashboard.filtering.filterModels'.
                         */
                        updateFilters(dashboardFilters: Partial<TFilterModels>): Q.Promise<void>;
                        /**
                         * If this TemplatePart includes UX to update filters that apply to the Part, the TemplatePart can include an option for
                         * the user to update filters at the Dashboard level and/or part level, using this method.
                         *
                         * The filters accepted here are a subset of those declared to be supported statically via the 'TFilterModels' type and
                         * those filters supported dynamically via 'getDynamicFilterDefinitions'.
                         *
                         * When the returned Promise resolves, the extension can assume that 'context.filtering.filterModels' is consistent
                         * with the updated 'filters' here passed here.
                         *
                         * @param filters Updated dashboard and/or part-level filter models
                         * @return A Promise reflecting the updating of the Dashboard and/or Part-level filter state as well as 'context.dashboard.filtering.filterModels'.
                         */
                        updateFilters(filters: {
                            dashboard?: Partial<TFilterModels>;
                            part?: Partial<TFilterModels>;
                        }): Q.Promise<void>;
                    };
                };
            }
        }
    }
}

// FILE: Fx\Composition\ViewModelAdapters.d.ts
declare module "Fx/Composition/ViewModelAdapters" {
    import Blade = require("Fx/Composition/Blade");
    import ButtonPart = require("Fx/Composition/ButtonPart");
    import FrameBlade = require("Fx/Composition/FrameBlade");
    import FramePart = require("Fx/Composition/FramePart");
    import MenuBlade = require("Fx/Composition/MenuBlade");
    import Part = require("Fx/Composition/Part");
    import TemplateBlade = require("Fx/Composition/TemplateBlade");
    import TemplatePart = require("Fx/Composition/TemplatePart");
    export = Main;
    module Main {
        class ViewModelAdapter<TParameters, TModel> implements MsPortalFx.ViewModels.BladeContract, MsPortalFx.ViewModels.HostedInMenuBlade, MsPortalFx.ViewModels.ButtonPartContract {
            title: KnockoutObservableBase<string>;
            subtitle: KnockoutObservableBase<string>;
            icon: KnockoutObservableBase<MsPortalFx.Base.Image>;
            actionBar: MsPortalFx.ViewModels.ActionBars.Base.Contract;
            parameterProvider: MsPortalFx.ViewModels.ParameterProvider<any, any>;
            commandBar: MsPortalFx.ViewModels.Toolbars.Toolbar;
            statusBar: KnockoutObservable<MsPortalFx.ViewModels.BladeStatusBar>;
            helpContentUri?: KnockoutObservableBase<string>;
            menuContent: {
                title: KnockoutObservable<string>;
                subtitle: KnockoutObservable<string>;
            };
            shortTitle: KnockoutObservableBase<string>;
            description: KnockoutObservableBase<string>;
            size: KnockoutObservable<MsPortalFx.Parts.PartSize>;
            constructor(container: MsPortalFx.ViewModels.ContainerContract | MsPortalFx.ViewModels.PartContainerContract, initialState: any, dataContext: TModel, _viewModelClass: Blade.BladeClass | FrameBlade.FrameBladeClass | MenuBlade.MenuBladeClass | TemplateBlade.TemplateBladeClass | ButtonPart.ButtonPartClass | FramePart.FramePartClass | Part.PartClass | TemplatePart.TemplatePartClass);
            onInputsSet(inputs: any, settings: any, dontCallOnRebind?: boolean, filters?: StringMap<any>): MsPortalFx.Base.Promise;
        }
    }
}

// FILE: Fx\Composition\ViewModelBase.d.ts
declare module "Fx/Composition/ViewModelBase" {
    import FxComposition = require("Fx/Composition");
    import FxPermissions = require("Fx/Composition/Permissions");
    export = Main;
    module Main {
        import FxBase = MsPortalFx.Base;
        import FxViewModels = MsPortalFx.ViewModels;
        /**
         * Options that are common to various Blade/Part decorators.  Includes metadata describing the Blade/Part and how it will be treated by
         * the FX.
         */
        interface Options {
            /**
             * Specifies whether this Blade/Part should exported from this extension,
             * so it can be imported/reused by other extensions.  The default is 'false'.
             */
            forExport?: boolean;
            /**
             * Used to associate this Blade/Part with a defined AssetType.  This allows the FX to:
             * - When an asset is deleted, give special UX treatment to associated Parts/Blades (close Blades, disable Parts)
             * - Store per-asset Part/Blade settings
             */
            forAsset?: {
                /**
                 * Names a single Blade/Part parameter that the FX will treat as the 'id' of the associated asset.
                 */
                assetIdParameter: string;
                /**
                 * Names an AssetType definition.  Values here should be one of the string constant generated by the PDL compiler from
                 * <AssetType>.
                 */
                assetType: string;
            };
            /**
             * Permissions that will be evaluated against the resource associated with this Blade/Part (see 'forAsset' above) before the
             * Blade/Part UI is shown/enabled.
             */
            permissions?: FxPermissions.Permissions;
        }
        /**
         * A common class type for Blade/Part view model classes.
         * For unit-testing purposes, this class type specifies an optional 'createModel' method that can be used to narrow the view model's
         * TModel type down from DataContext to a more-specific type that is easier to mock.
         */
        interface ViewModelClass<TModel> {
            new (): Contract<any, TModel>;
        }
        /**
         * A common interface type for Blade/Part view model classes.
         */
        interface Contract<TParameters, TModel> {
            /**
             * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
             * 'onRebind' methods.
             */
            context: Context<TParameters, TModel>;
            /**
             * Called by the FX to initialize this view model.
             *
             * @return A promise that should be resolved by the view model when: (1) all the data is loaded for this view model and (2) the
             * view model is updated to reflect the loaded data.
             */
            onInitialize(): Q.Promise<any>;
        }
        /**
         * Context injected by the FX into the view model.  This context should be used within the view model's 'onInitialize' and optional
         * 'onRebind' methods.
         */
        interface Context<TParameters, TModel> {
            /**
             * The container used by this view model.
             */
            container: Container;
            /**
             * The model used by this view model.
             */
            model: TModel;
            /**
             * Parameters supplied by the caller of this Blade/Part.
             */
            parameters: TParameters;
            /**
             * API collection that facilitates dashboard interactions.
             */
            dashboard: MsPortalFx.ViewModels.DashboardManagement;
        }
        /**
         * Represents a container object that can be used to control the chrome of the Blade/Part.
         */
        interface Container extends FxBase.LifetimeManager, FxViewModels.BladeManagement {
            /**
             * Removes the blocking loading indicator from the Blade/Part.
             */
            revealContent(): void;
            /**
             * Closes this blade.
             *
             * @param data Use the @ReturnsData<TData> decorator to supply TData-typed 'data' here. This 'data' will be returned to the
             * BladeReference with which this Blade was opened, via the BladeReference's 'onClosed' callback.
             * @return a promise that resolves to true if the this blade is successfully closed.
             */
            closeCurrentBlade(data?: void): Q.Promise<boolean>;
            /**
             * Closes the child blade that is currently open.
             *
             * @return a promise that resolves to true if the child blade is successfully closed.
             */
            closeChildBlade(): Q.Promise<boolean>;
            /**
             * Closes the context blade that was opened by this part
             *
             * @returns a promise that resolves to true if the context blade is sucessfully closed.
             */
            closeContextBlade(): Q.Promise<boolean>;
            /**
             * Tracks the list of asynchronous operations being performed by the Part.
             */
            operations: FxComposition.OperationList;
            /**
             * An observable whose value determines the message displayed when there is no data to be displayed in the
             * Blade/Part.
             * This message will be displayed if it is non-null, non-empty.
             */
            noDataMessage: KnockoutObservable<string>;
            /**
             * Shows an error message in place of the Part.
             *
             * @param message A message that will be displayed to the user when the part fails.
             */
            fail(message: string): void;
            /**
             * Used to put a failed Blade/Part back into an operational state.
             */
            recover(): void;
            /**
             * Moves the container into a 'not found' display mode, allowing for an optional custom error message.
             *
             * @param message An optional custom error message.
             */
            notFound(message?: string): void;
            /**
             * Moves the container into an 'unauthorized' display mode, allowing for an optional custom error message.
             *
             * @param message A custom error message in place of the default.
             */
            unauthorized(message?: string): void;
        }
        /**
         * Shared by the decorator options of @TemplateBlade/@TemplatePart.  Requires that the extension supply an HTML template to which the
         * view model will be bound.
         */
        interface RequiresHtmlTemplateOptions {
            /**
             * The HTML template to which the associated view model will be bound.  This can be either an HTML string or an HTML filename.
             */
            htmlTemplate: string;
        }
        /**
         * Shared by decorator options that accept a 'styleSheets' option.
         */
        interface AcceptsStyleSheetsOptions {
            /**
             * The CSS style sheets containing styles used by HTML templates of this Blade/Part.
             */
            styleSheets?: string[];
        }
        module Rebindable {
            /**
             * Passed to the view model's 'onRebind' method, reflecting the changes to the view model's parameters.
             */
            interface Changes {
                /**
                 * The keys of those view model parameters whose values have changed.
                 */
                parameterKeys: string[];
            }
        }
        module Configurable {
            /**
             * Options supplied to the @Configurable decorator.  Used to supply metadata describing the Blade/Part's settings.
             */
            interface Options {
                /**
                 * Metadata describing the settings that will be persisted for this Blade/Part.
                 */
                settings?: {
                    /**
                     * The scope at which the settings will be persisted.  If not supplied, a default value of 'PerPartOrBladeType' will be
                     * used.
                     */
                    scope?: SettingsScope;
                    /**
                     * A StringMap of metadata describing individual settings that will be persisted.
                     */
                    metadata?: {
                        [key: string]: SettingMetadata;
                    };
                };
            }
            /**
             * Metadata describing one setting of the Blade/Part's persisted settings.
             */
            interface SettingMetadata {
                /**
                 * Specifies whether this setting is one that can be shared across different Part and Blade types (that is, isn't used by a
                 * single Part or Blade type).  If not supplied, a default value of 'false' will be used.
                 */
                isSharedAcrossPartAndBladeTypes?: boolean;
                /**
                 * When 'isSharedAcrossPartAndBladeTypes' is 'true', this is the key under which the setting will be saved.  This key will
                 * be used in the different Parts and Blades that share the setting value.
                 */
                sharedKey?: string;
            }
            /**
             * The scope at which the Blade/Part's settings are persisted.
             */
            const enum SettingsScope {
                /**
                 * The settings are saved under a single key for all instances of the associated Blade/Part.
                 */
                PerPartOrBladeType = 0,
                /**
                 * The settings are saved under a discrete key for each 'id' with which the associated Blade/Part
                 * is rendered.  For instance, the 'Freezing Fog' website would have different settings values than those of the
                 * 'Jasmine Dancer' website on the 'WebsiteDetails' Blade.
                 */
                PerId = 1,
            }
            /**
             * Context injected by the FX into the view model.  This context will include a 'configuration' (the
             * Blade/Part's configuration API) as well as 'parameters' and 'settings' (the
             * Blade/Part's persisted settings values).
             */
            interface Context<TSettings> {
                /**
                 * The persisted settings values.
                 */
                settings: TSettings;
            }
        }
        /**
         * The @ProxiedMember decorator.  Identifies a public member of a Blade/Part class that is to be proxied for use in the view.
         * This decorator is not necessary in most scenarios.
         * This decorator suppresses any compile-time errors that state "The '<MEMBER>' member is public and is not bound to the Blade/Part's view."
         */
        function ProxiedMember(target: Object, key: string): void;
    }
}

// FILE: Fx\Controls\Accordion.d.ts
declare module "Fx/Controls/Accordion" {
    import { HtmlContent } from "Fx/Controls/ControlsBase";
    import Control = MsPortalFx.Control;
    /**
     * Options for configuring an accordion control
     */
    export interface Options {
        /**
         * Specifies whether the user can open multiple sections or not.
         * Set to false by default.
         * If canExpandMultipleSections = false and > 1 sections are initialized with `expanded = true`, only the 1st section will be expanded
         */
        canExpandMultipleSections?: boolean;
        /**
         * Specifies whether the user can make all the sections be closed at once. Allows collapsing the active section.
         * Set to false by default.
         * If canCollapseAllSections = false and no section is initialized with `expanded = true`, will expand the 1st section.
         *
         * 'canExpandMultipleSections' and 'canCollapseAllSections' used together exhibit the following behavior:
         * canExpandMultipleSections = true ; canCollapseAllSections = true ;  =>  Expand / collapse with no limitations.
         * canExpandMultipleSections = true ; canCollapseAllSections = false;  =>  At least one section expanded, can have more than one sections expanded.
         * canExpandMultipleSections = false; canCollapseAllSections = true ;  =>  At most one section expanded, can have all sections collapsed.
         * canExpandMultipleSections = false; canCollapseAllSections = false;  =>  Always have exact one section expanded.
         */
        canCollapseAllSections?: boolean;
        disabled?: boolean | KnockoutObservableBase<boolean>;
        visible?: boolean | KnockoutObservableBase<boolean>;
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Aria label for control.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * Sections for your accordion control.
         */
        sections: Section[] | KnockoutObservableArray<Section>;
    }
    /**
     * The interface that defines the viewmodel that respresents an Accordion.
     */
    export interface Contract extends Control {
        /**
         * Whether user can open multiple sections or not. Set to false by default.
         */
        readonly canExpandMultipleSections: boolean;
        /**
         * Whether all the sections can be closed at once. Allows collapsing the active section.
         * Set to false by default.
         */
        readonly canCollapseAllSections: boolean;
        readonly disabled: KnockoutObservableBase<boolean>;
        readonly visible: KnockoutObservableBase<boolean>;
        readonly cssClass: KnockoutObservableBase<string>;
        /**
         * Aria label for control.
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * Sections for your accordion control.
         */
        readonly sections: KnockoutObservableArray<Section>;
    }
    /**
     * Creates an Accordion control viewmodel.
     *
     * @param lifetime The lifetime for the created view model.
     * @param options The set of options to configure the accordion.
     *
     * @returns an Accordion viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options: Options): Contract;
    /**
     * Options for a Section in Accordion control
     */
    export interface SectionOptions {
        /**
         * The open / collapse status of the section, Default is false.
         */
        expanded?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Header title for your section.
         * If supplied as a string, the string is rendered as markdown https://www.markdownguide.org/basic-syntax.
         * Can be a HTML template.
         */
        title: string | Control | HtmlContent | KnockoutObservableBase<string | Control | HtmlContent>;
        /**
         * Content for your section.
         * If supplied as a string, the string is rendered as markdown https://www.markdownguide.org/basic-syntax.
         * Can also be a HTML template.
         * Can be a Q.Promise.
         */
        content: string | Control | HtmlContent | KnockoutObservableBase<string | Control | HtmlContent> | Q.Promise<string | HtmlContent>;
    }
    /**
     * The interface that defines the viewmodel that respresents a Section.
     */
    export interface Section {
        /**
         * The open / collapse status of the section
         */
        readonly expanded: KnockoutObservableBase<boolean>;
        /**
         * Header title for your section.
         */
        readonly title: KnockoutObservableBase<string | Control | HtmlContent>;
        /**
         * Content for your section.
         */
        readonly content: KnockoutObservableBase<string | Control | HtmlContent>;
    }
    /**
     * Creates a Section viewmodel.
     *
     * @param lifetime The lifetime for the created view model.
     * @param options The set of options to configure the Section.
     *
     * @returns a Section viewmodel
     */
    export function createSection(lifetime: MsPortalFx.Base.LifetimeManager, options: SectionOptions): Section;
    /**
     * Describes an HTML view, defined in terms of an HTML template bound to an accompanying view model object.
     */
    export { HtmlContent };
}

// FILE: Fx\Controls\BaseResourceDropDown.d.ts
declare module "Fx/Controls/BaseResourceDropDown" {
    import { CustomV, Required, ValidationResult } from "Fx/Controls/Validations";
    import { Item as MarketplaceItem } from "Fx/ResourceManagement/Marketplace";
    export = Main;
    module Main {
        import FxAzure = MsPortalFx.Azure;
        /**
         * The validations accepted by the resource dropdowns
         */
        type Validation<TValue> = CustomV<TValue> | Required;
        /**
         * The options accepted by all resource dropdowns
         */
        interface Options<TValue> {
            /**
             * The label of the control
             */
            label?: string | KnockoutObservableBase<string>;
            /**
             * Aria label for group.
             */
            ariaLabel?: string | KnockoutObservableBase<string>;
            filter?: boolean;
            /**
             * Sets the filter TextBox’s placeholder text.
             * Default is blank.
             */
            filterPlaceholder?: string;
            /**
             * Sets the DropDown’s placeholder text.
             * This is the placeholder if `filter` is set to false.
             * Default is blank.
             */
            placeholder?: string;
            /**
             * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
             */
            suppressDirtyBehavior?: boolean;
            /**
             * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
             */
            value?: TValue | KnockoutObservableBase<TValue>;
            /**
             * Validations on the control
             */
            validations?: Validation<TValue>[] | KnockoutObservableBase<Validation<TValue>[]>;
            /**
             * sanitized html string shown in the info balloon popup
             */
            infoBalloonContent?: string | KnockoutObservableBase<string>;
            /**
             * An object which will map values to groups and sort the groups.
             */
            grouping?: {
                /**
                 * A function which takes an item and returns the display name of the group it will belong to
                 * @param item An object of the type for this dropdown
                 * @returns The display name for the group it belongs to. If empty, item will be on top level of the dropdown
                 */
                map(item: TValue): string;
                /**
                 * A comparator to display the groups in the order desired (does not sort values within groups)
                 */
                sort?: Comparator<string>;
            };
            /**
             * An object which will hide values
             */
            hiding?: {
                /**
                 * A callback for to determine if an item is hidden
                 * @param item An object of the type for this dropdown
                 * @returns A flag for if the item should be hidden
                 */
                hide(item: TValue): boolean;
                /**
                 * A message explaining why values are hidden
                 */
                reason: string;
            };
            /**
             * A callback which will disable an item if a reason is returned to disable it.
             * Values will be grouped by the reason the item is disabled
             * @param item An object of the type for this dropdown
             * @return The reason this item is disabled. If empty the item will be enabled.
             */
            disable?(item: TValue): string;
            /**
             * A comparator to display the values in the order desired (does not sort groups if you have any)
             */
            sort?: Comparator<TValue>;
        }
        /**
         * The contract all resource dropdowns fulfill
         */
        interface Contract<TValue> extends MsPortalFx.Control {
            /**
             * The dirty state of the control.  This will only change when the user performs an action to change the value, or the dirty state is set explicitly by the extension.
             */
            readonly dirty: KnockoutObservableBase<boolean>;
            /**
             * The validation state of the control.
             */
            readonly valid: KnockoutReadOnlyObservableBase<boolean>;
            /**
             * The label of the control
             */
            readonly label: KnockoutObservableBase<string>;
            /**
             * validations on the control
             */
            readonly validations: KnockoutObservableArray<Validation<TValue>>;
            /**
             * Latest validation results.
             */
            readonly validationResults: KnockoutReadOnlyObservableArray<ValidationResult>;
            /**
             * The disabled state of the control.
             */
            readonly disabled: KnockoutObservableBase<boolean>;
            /**
             * Triggers validation on this control when called.  Returns a promise that resolves once validation on this control has completed.
             */
            readonly triggerValidation: () => Q.Promise<boolean>;
            /**
             * the value of the control
             */
            readonly value: KnockoutObservableBase<TValue>;
            /**
             * The available values for the control
             */
            readonly fetchedValues: KnockoutReadOnlyObservableArray<TValue>;
            /**
             * Puts the control into a loading state.  When in the loading state, the control will be disabled and have the text "Loading..." shown.
             */
            readonly loading: KnockoutReadOnlyObservable<boolean>;
        }
        /**
         * Subscription dropdown interfaces
         */
        module Subscription {
            import FxGalleryItem = MsPortalFx.Services.Gallery.GalleryItem;
            import ResourceManager = FxAzure.ResourceManager;
            import RequiredPermissions = FxAzure.RequiredPermissions;
            import Subscription = FxAzure.Subscription;
            /**
             * Options for the legacy dropdown which doesn't have the property initialSubscriptionId
             */
            interface BaseOptions extends Main.Options<Subscription> {
                /**
                 * The gallery item to validate against.
                 */
                filterByGalleryItem?: FxGalleryItem | ResourceManager.GalleryItemExtract | MarketplaceItem<any> | KnockoutObservableBase<FxGalleryItem | ResourceManager.GalleryItemExtract | MarketplaceItem<any>>;
                /**
                 * The list of required permissions the subscription must have. (If you need to check
                 * for resource provider registration permissions, use the 'resourceProviders' option
                 * instead.)
                 */
                requiredPermissions?: RequiredPermissions | KnockoutObservableBase<RequiredPermissions>;
                /**
                 * A list of resource providers that the
                 * subscription needs to be registered with. Providing the list will NOT register
                 * the subscription with the resource provider, but will only check if the user has
                 * the necessary permissions to register them (only if they aren't registered already).
                 */
                resourceProviders?: string[] | KnockoutObservableBase<string[]>;
            }
            /**
             * Options for the subscirption dropdown.
             */
            interface Options extends BaseOptions {
                /**
                 * The initial subscription id to be set as a default value for the dropdown if it's a member of the values fetched from ARM.
                 * Create Blades will receive this value as a Blade parameter.
                 * If an array is passed here, the first item in the array which is is a member of the values fetched from ARM will be used.
                 */
                initialSubscriptionId: string | string[] | KnockoutObservableBase<string | string[]>;
            }
            /**
             * This is the contract for the subscription dropdown
             */
            interface Contract extends Main.Contract<Subscription> {
            }
        }
        /**
         * Location dropdown interfaces module
         */
        module Location {
            import Location = MsPortalFx.Azure.Location;
            /**
             * Options for the legacy dropdown which doesn't have properties initialLocationName and subscriptionId
             */
            type BaseOptions = Main.Options<Location>;
            /**
             * Options for the location dropdown.
             */
            interface Options extends BaseOptions {
                /**
                 * The initial location name to be set as a default value for the dropdown if it's a member of the values fetched from ARM.
                 * Create blades will receive this as a parameter, so supplying this as initialLocationName is required
                 * If an array is passed here, the first item in the array which is is a member of the values fetched from ARM will be used.
                 */
                initialLocationName: string | string[] | KnockoutObservableBase<string | string[]>;
                /**
                 * The subscription used to look up available locations
                 */
                subscriptionId: KnockoutObservableBase<string>;
                /**
                 * Optional. The observable that holds the list of resource types used to filter locations.
                 */
                resourceTypes?: string[] | KnockoutObservableBase<string[]>;
            }
            /**
             * Options for the location dropdown with additional locations
             */
            interface CustomOptions extends Options {
                /**
                 * A callback to customize which locations are in the dropdown
                 * @param locs The locations retrieved from ARM
                 * @return A list of locations to be displayed in the dropdown
                 */
                customizeItems?: (locs: Location[]) => Location[];
            }
            /**
             * This is the contract for the location dropdown
             */
            interface Contract extends Main.Contract<Location> {
            }
        }
        /**
         * ResourceGroup dropdown interfaces module
         */
        module ResourceGroup {
            import RequiredPermissions = FxAzure.RequiredPermissions;
            /**
             * The modes possible for the dropdown
             */
            const enum Mode {
                UseExisting = 0,
                CreateNew = 1,
                Both = 2,
            }
            /**
             * The mode of the value returned by the control
             */
            type SelectedMode = Mode.CreateNew | Mode.UseExisting;
            /**
             * The ARM resource group value
             */
            type ResourceGroup = FxAzure.ResourceGroup;
            /**
             * The value of the resource group dropdown
             */
            interface Value {
                /**
                 * The mode of the dropdown, create new or use existing
                 */
                mode: SelectedMode;
                /**
                 * The ARM value for the resource group
                 */
                value: ResourceGroup;
            }
            /**
             * Options for the resource group dropdown.
             */
            interface Options extends BaseOptions {
                /**
                 * The inital resourceGroupId to be set as a default value for the dropdown if it's an existing resource group in the dropdown.
                 * If an array is passed, will select the first available in the list
                 */
                initialResourceGroupName?: string | string[] | KnockoutObservableBase<string | string[]>;
                /**
                 * The subscription used to look up available resource groups
                 */
                subscriptionId: KnockoutObservableBase<string>;
            }
            /**
             * Options for the legacy dropdown (which doesn't have 'initialResourceGroupName' and 'subscriptionId' properties)
             */
            interface BaseOptions extends Main.Options<Value> {
                /**
                 * The default value for a new resource group name. If set, the +New text box will
                 * default to that value. If a resource group already exists with the same value, this
                 * string will be incremented ("val", "val_1", "val_2", etc.). If not set, the text
                 * box will be left blank (default).
                 */
                defaultNewValue?: string;
                /**
                 * The list of required permissions the subscription must have.
                 */
                requiredPermissions?: RequiredPermissions | KnockoutObservableBase<RequiredPermissions>;
                /**
                 * The mode allowed for the resource group dropdown
                 */
                allowedMode?: Mode | KnockoutObservableBase<Mode>;
                /**
                 * Sets the placeholder.
                 * Default is blank.
                 */
                createNewPlaceholder?: string | KnockoutObservableBase<string>;
            }
            /**
             * This is the contract for the resource group dropdown
             */
            interface Contract extends Main.Contract<Value> {
                /**
                 * Set which mode the dropdown is in, create or use existing
                 */
                switchMode(mode?: SelectedMode): void;
            }
        }
    }
}

// FILE: Fx\Controls\Batch\HeatMap.d.ts
declare module "Fx/Controls/Batch/HeatMap" {
    export = Main;
    module Main {
        /**
         * Represent an item in the heatMap
         */
        interface Item {
            /**
             * Id of an item in the heatmap. This NEED to be unique.
             */
            id: string;
            /**
             * State of the item
             */
            state: string;
            /**
             * Optional setting to make the item less opaque on the heatmap.
             * Default: 1
             */
            opacity?: number;
        }
        /**
         * Possible position of the legend relative to the heatmap.
         * i.e. LegendLocation.Left means the legend will be on the left of the heatmap.
         *
         * LegendLocation.None will not show the legend
         */
        enum LegendLocation {
            None = 0,
            Top = 1,
            Right = 2,
            Bottom = 3,
            Left = 4,
        }
        /**
         * Represent a possible state an item can have
         */
        interface State {
            /**
             * State id, Item added to the Heatmap should have their state match this id
             */
            id: string;
            /**
             * Label for the legend
             */
            label: KnockoutObservableBase<string>;
            /**
             * Color the tile with this state will have
             */
            color: MsPortalFx.ColorUtil.ColorCode;
            /**
             * Color opacity
             * @default 1
             */
            opacity?: number;
        }
        interface Options {
            /**
             * Message to be displayed in the heatmap if the item list is empty
             */
            noItemsMessage?: KnockoutObservableBase<string>;
            /**
             * List of available states. Every state that might be used in an item later on should be here.
             */
            states?: State[];
            /**
             * When one of the item has a state not defined in the lis_t above it will use this color
             */
            unknownStateColor?: MsPortalFx.ColorUtil.ColorCode;
            /**
             * Enable selection on the heatmap
             * @default true
             */
            isSelectable?: boolean;
            /**
             * If the heatmap control should expand automatically if the number of items don't fit in the initial window
             * @default false
             */
            isExpandable?: boolean;
            /**
             * Minimum number of pixel a tile side can be in the heatmap
             * Cannot be less that 1
             * @default 10
             */
            minTileSize?: number;
            /**
             * Maximum number of pixel a tile side can be in the heatmap
             * @default 100
             */
            maxTileSize?: number;
            /**
             * Legend location relative to the heatmap.
             * None will not show the legend.
             * @default Right
             */
            legendLocation?: KnockoutObservableBase<LegendLocation>;
        }
        interface ItemListDiff {
            added: Item[];
            updated: Item[];
            removed: Item[];
        }
        class ViewModel extends MsPortalFx.ViewModels.Controls.Base.ViewModel {
            states: State[];
            unknownStateColor: MsPortalFx.ColorUtil.ColorCode;
            noItemsMessage: KnockoutObservableBase<string>;
            errorMessage: string;
            minTileSize: number;
            maxTileSize: number;
            /**
             * Enable selection on the heatmap
             * @default true
             */
            isSelectable: boolean;
            /**
             * If the heatmap control should expand automatically if the number of items don't fit in the initial window
             * @default false
             */
            isExpandable: boolean;
            /**
             * Cache of the items used to compute the edits
             */
            private itemMap;
            /**
             * Legend location relative to the heatmap.
             * None will not show the legend.
             * @default Right
             * @see LegendOptions
             */
            legendLocation: KnockoutObservableBase<LegendLocation>;
            /**
             * Internal mapping of state by ID.
             * Used in the widget.
             */
            stateMap: StringMap<State>;
            /**
             * Internal callback that will pass the difference of items everytime setItems is called
             */
            _msPortalFxUpdateCallback: KnockoutObservable<(diffs: ItemListDiff) => void>;
            /**
             * Internal update to ask for a canvas redraw
             */
            _msPortalFxUpdateCanvas: KnockoutObservableBase<number>;
            /**
             * Observable containing the currently selected item.
             * Seleletable must be set to true in the options for this to be updated
             */
            selectedItem: KnockoutObservable<Item>;
            selectable: MsPortalFx.ViewModels.Selectable<Item>;
            /**
             * Buffer for to keep the last setItems args until the implementation set the callback
             */
            private _lastItemsBuffer;
            private defaultOptions;
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, options?: Options);
            setItems(items: Item[]): void;
            /**
             * Force a render of the heatmap.
             * Use this if you have enableAutoRendering to false.
             */
            render(): void;
            private _createStateMap();
            private _createItemMap(items);
            private _computeDifference(oldMap, newMap);
        }
    }
}

// FILE: Fx\Controls\Button.d.ts
declare module "Fx/Controls/Button" {
    export import Image = MsPortalFx.Base.Image;
    /**
     * Styling of the button
     */
    export const enum Style {
        /**
         * primary button style (default)
         */
        Primary = 0,
        /**
         * secondary button style
         */
        Secondary = 1,
    }
    export const enum IconPosition {
        Left = 0,
        Right = 1,
    }
    export interface IconAndPosition {
        /**
         * The image to use as an icon.
         */
        image: Image;
        /**
         * The position of the icon in the button.
         */
        position: IconPosition;
    }
    /**
     * Options used to dynamically configure the button.
     */
    export interface Options {
        /**
         * Callback invoked when the button is clicked.
         */
        onClick: Function;
        /**
         * The text shown on the button
         */
        text: string | KnockoutObservableBase<string>;
        /**
         * Style of the button
         */
        style?: Style | KnockoutObservableBase<Style>;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Whether the form element is visible.
         */
        visible?: KnockoutObservableBase<boolean> | boolean;
        /**
         * The alternate text for screen-reader
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * loading status
         */
        loading?: boolean | KnockoutObservableBase<boolean>;
        /**
         * icon shown in the button
         */
        icon?: IconAndPosition | KnockoutObservableBase<IconAndPosition>;
    }
    /**
     * Button control
     */
    export interface Contract extends MsPortalFx.Control {
        /**
         * The disabled state of the control.
         */
        readonly disabled: KnockoutObservableBase<boolean>;
        /**
         * The text shown on the button
         */
        readonly text: KnockoutObservableBase<string>;
        /**
         * Style of the button (Primary / Secondary)
         */
        readonly style: KnockoutObservableBase<Style>;
        /**
         * Whether the form element is visible.
         */
        readonly visible: KnockoutObservableBase<boolean>;
        /**
         * The alternate text for screen-reader
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * loading status
         */
        readonly loading: KnockoutObservableBase<boolean>;
        /**
         * icon shown in the button
         */
        readonly icon: KnockoutObservableBase<IconAndPosition>;
    }
    /**
     * Creates a button control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a button viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options: Options): Contract;
}

// FILE: Fx\Controls\CheckBox.d.ts
declare module "Fx/Controls/CheckBox" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    /**
     * Options for configuring a checkbox control.
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * The aria-label on the control.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * if set to true, the label will be placed to the right of the control
         */
        labelOnRight?: boolean;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
    }
    /**
     * CheckBox control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<boolean>;
    }
    /**
     * Creates a checkbox control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a checkbox control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\Console.d.ts
declare module "Fx/Controls/Console" {
    export const enum LogItemType {
        /**
         * Indicates general information.
         */
        Text = 0,
        /**
         * Indicates important information.
         */
        Information = 1,
        /**
         * Indicates success information
         */
        Success = 2,
        /**
         * Indicates warning information.
         */
        Warning = 3,
        /**
         * Indicates error information.
         */
        Error = 4,
    }
    export interface LogItem {
        /**
         * The text to log.
         */
        text: string;
        /**
         * The type of text to log.
         */
        type?: LogItemType;
    }
    export interface Options {
        /**
         * Alternate text for screen reader
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * Provides the text used for the CLI prompt.
         */
        prompt?: string | KnockoutObservableBase<string>;
        /**
         * callback called when the user enters input
         * console is the instance of the console created with these options, userInput is the string entered by the user
         */
        processCommand: (console: Contract, userInput: string) => void;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
    }
    export interface Contract extends MsPortalFx.Control {
        /**
         * Alternate text for screen reader, label for this log-stream
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * Provides the text used for the CLI prompt.
         */
        readonly prompt: KnockoutObservableBase<string>;
        /**
         * Clears the console.
         */
        reset(): void;
        /**
         * Writes an entry to the console
         * @param entry The text to show in the control.  Passing in a LogItem or array of LogItems allows you to control the level of entry (error, warning, etc.)
         */
        log(entry: string | string[] | LogItem | LogItem[]): void;
        readonly disabled: KnockoutObservableBase<boolean>;
    }
    /**
     * Creates a Console control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a Console control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options: Options): Contract;
}

// FILE: Fx\Controls\ControlsBase.d.ts
declare module "Fx/Controls/ControlsBase" {
    /**
     * Describes an HTML view, defined in terms of an HTML template bound to an accompanying view model object.
     */
    export interface HtmlContent {
        /**
         * The HTML template to display
         */
        htmlTemplate: string;
        /**
         * The object to bind to the template.
         */
        viewModel: any;
    }
}

// FILE: Fx\Controls\CopyableLabel.d.ts
declare module "Fx/Controls/CopyableLabel" {
    export interface Options {
        /**
         * Aria label for the textbox.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Hides the textbox.  Used in scenarios where the copied text is not shown to the user.
         */
        hideTextbox?: boolean;
        /**
         * sanitized html string shown in the info balloon popup.
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * The label of the control.
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The number of lines in the textbox.
         */
        rows?: number;
        /**
         * The sub label of the control.
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * The value of the TextBox.
         */
        value?: string | KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
    }
    export interface Contract extends MsPortalFx.Control {
        /**
         * A custom css class to apply to the control.
         */
        readonly cssClass: KnockoutObservableBase<string>;
        /**
         * The disabled state of the control.
         */
        readonly disabled: KnockoutObservableBase<boolean>;
        /**
         * sanitized html string shown in the info balloon popup.
         */
        readonly infoBalloonContent: KnockoutObservableBase<string>;
        /**
         * The label of the control.
         */
        readonly label: KnockoutObservableBase<string>;
        /**
         * The sub label of the control.
         */
        readonly subLabel: KnockoutObservableBase<string>;
        /**
         * The value of the TextBox.
         */
        readonly value: KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        readonly visible: KnockoutObservableBase<boolean>;
    }
    /**
     * Creates a copyableLabel control viewmodel.
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options: Options): Contract;
}

// FILE: Fx\Controls\CustomHtml.d.ts
declare module "Fx/Controls/CustomHtml" {
    /**
     * Options for creating a custom html form control
     */
    export interface Options {
        /**
         * The html template to display as the body of the form.
         */
        htmlTemplate: string;
        /**
         * The object to bind to the template.
         */
        innerViewModel?: any;
        /**
         * Uses a new binding context for the template such that $root is innerViewModel instead of the blade/part view model.
         * Defaults to true.
         */
        isolated?: boolean;
        /**
         * The label of the control.
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sub label of the control.
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * sanitized html string shown in the info balloon popup.
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
    }
    /**
     * Custom html control viewmodel
     */
    export interface Contract extends MsPortalFx.Control {
        /**
         * The label of the control.
         */
        readonly label: KnockoutObservableBase<string>;
        /**
         * The sub label of the control.
         */
        readonly subLabel: KnockoutObservableBase<string>;
        /**
         * sanitized html string shown in the info balloon popup.
         */
        readonly infoBalloonContent: KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        readonly visible: KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        readonly cssClass: KnockoutObservableBase<string>;
    }
    /**
     * Creates a custom html control viewmodel.
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options: Options): Contract;
}

// FILE: Fx\Controls\DatePicker.d.ts
declare module "Fx/Controls/DatePicker" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export import DateTimeRange = MsPortalFx.DateUtil.DateTimeRange;
    /**
     * Options for configuring a date picker control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * Aria label for the DatePicker.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * Date/time range in which user is able to select date/time.
         */
        enabledDateTimeRange?: DateTimeRange | KnockoutObservableBase<DateTimeRange> | DateTimeRange[] | KnockoutObservableArray<DateTimeRange>;
        /**
         * Whether the control should allow empty(null) value as valid.
         */
        allowEmpty?: boolean;
        /**
         * The text shown when no value is set in the control.
         */
        placeHolderText?: string | KnockoutObservableBase<string>;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: Date | KnockoutObservableBase<Date>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
    }
    /**
     * DatePicker control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<Date>;
        /**
         * Date/time range in which user is able to select date/time.
         */
        readonly enabledDateTimeRange: KnockoutObservableBase<DateTimeRange> | KnockoutObservableArray<DateTimeRange>;
        /**
         * The text shown when no value is set in the control.
         */
        readonly placeHolderText: KnockoutObservableBase<string>;
    }
    /**
     * Creates a date picker control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a date picker control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\DateTimePicker.d.ts
declare module "Fx/Controls/DateTimePicker" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export import DateTimeRange = MsPortalFx.DateUtil.DateTimeRange;
    /**
     * Options for configuring a date time picker control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * Aria label for the DateTimePicker.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * Date/time range in which user is able to select date/time.
         */
        enabledDateTimeRange?: DateTimeRange | KnockoutObservableBase<DateTimeRange> | DateTimeRange[] | KnockoutObservableArray<DateTimeRange>;
        /**
         * Show time zone dropdown.
         */
        showTimezoneDropdown?: boolean | KnockoutObservableBase<boolean>;
        /**
         * ID of the timezone to show in the dropdown.
         * Refer to "Fx/Controls/TimeZones" for the possible values.
         */
        timezoneDropDownValue?: string | KnockoutObservableBase<string>;
        /**
         * Obsolete. This is no longer available and the datetime value is always in the local timezone.
         */
        timezoneOffset?: Obsolete;
        /**
         * Whether the control should allow empty(null) value as valid.
         */
        allowEmpty?: boolean;
        /**
         * Empty value text for date.
         */
        placeHolderTextForDate?: string | KnockoutObservableBase<string>;
        /**
         * Empty value text for time.
         */
        placeHolderTextForTime?: string | KnockoutObservableBase<string>;
        /**
         * Obsolete. Use 'placeHolderTextForTime' instead.
         */
        placeHolderText?: Obsolete;
        /**
         * If true displays hours, minutes, and seconds. If false displays only hours and minutes.
         */
        showSeconds?: boolean;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: Date | KnockoutObservableBase<Date>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
    }
    /**
     * DateTimePicker control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<Date>;
        /**
         * Date/time range in which user is able to select date/time.
         */
        readonly enabledDateTimeRange: KnockoutObservableBase<DateTimeRange> | KnockoutObservableArray<DateTimeRange>;
        /**
         * Show time zone dropdown.
         */
        readonly showTimezoneDropdown: KnockoutObservableBase<boolean>;
        /**
         * ID of the timezone to show in the dropdown.
         * Refer to "Fx/Controls/TimeZones" for the possible values.
         */
        readonly timezoneDropDownValue: KnockoutObservableBase<string>;
        /**
         * Obsolete. This is no longer available and the datetime value is always in the local timezone.
         */
        readonly timezoneOffset: Obsolete;
        /**
         * Empty value text for date.
         */
        readonly placeHolderTextForDate: KnockoutObservableBase<string>;
        /**
         * Empty value text for time.
         */
        readonly placeHolderTextForTime: KnockoutObservableBase<string>;
        /**
         * Obsolete. Use 'placeHolderTextForTime' instead.
         */
        readonly placeHolderText: Obsolete;
    }
    /**
     * Creates a date/time picker control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a date time picker control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\DateTimeRangePicker.d.ts
declare module "Fx/Controls/DateTimeRangePicker" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export import DateTimeRange = Base.DateTimeRange;
    /**
     * Options for configuring a date time range picker control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * Aria label for the DateTimeRangePicker.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * Enabled range for start date/time
         */
        startDateTimeEnabledRange?: DateTimeRange | KnockoutObservableBase<DateTimeRange>;
        /**
         * Enabled range for end date/time
         */
        endDateTimeEnabledRange?: DateTimeRange | KnockoutObservableBase<DateTimeRange>;
        /**
         * Show time zone dropdown.
         */
        showTimezoneDropdown?: boolean | KnockoutObservableBase<boolean>;
        /**
         * ID of the timezone to show in the dropdown.
         * Refer to "Fx/Controls/TimeZones" for the possible values.
         */
        timezoneDropDownValue?: string | KnockoutObservableBase<string>;
        /**
         * Obsolete. This is no longer available and the datetime value is always in the local timezone.
         */
        timezoneOffset?: Obsolete;
        /**
         * Empty value text for date.
         */
        placeHolderTextForDate?: string | KnockoutObservableBase<string>;
        /**
         * Empty value text for time.
         */
        placeHolderTextForTime?: string | KnockoutObservableBase<string>;
        /**
         * Obsolete. Use 'placeHolderTextForTime' instead.
         */
        placeHolderText?: Obsolete;
        /**
         * Display start/end date/time fields inline (false by default).
         */
        displayFieldsInline?: boolean;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: DateTimeRange | KnockoutObservableBase<DateTimeRange>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
    }
    /**
     * DateTimeRangePicker control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<DateTimeRange>;
        /**
         * Enabled range for start date/time
         */
        readonly startDateTimeEnabledRange: KnockoutObservableBase<DateTimeRange>;
        /**
         * Enabled range for end date/time
         */
        readonly endDateTimeEnabledRange: KnockoutObservableBase<DateTimeRange>;
        /**
         * Show time zone dropdown.
         */
        readonly showTimezoneDropdown: KnockoutObservableBase<boolean>;
        /**
         * ID of the timezone to show in the dropdown.
         * Refer to "Fx/Controls/TimeZones" for the possible values.
         */
        readonly timezoneDropDownValue: KnockoutObservableBase<string>;
        /**
         * Obsolete. This is no longer available and the datetime value is always in the local timezone.
         */
        readonly timezoneOffset: Obsolete;
        /**
         * Empty value text for date.
         */
        readonly placeHolderTextForDate: KnockoutObservableBase<string>;
        /**
         * Empty value text for time.
         */
        readonly placeHolderTextForTime: KnockoutObservableBase<string>;
        /**
         * Obsolete. Use 'placeHolderTextForTime' instead.
         */
        readonly placeHolderText: Obsolete;
    }
    /**
     * Creates a date/time range picker control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a date time range picker control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\DayPicker.d.ts
declare module "Fx/Controls/DayPicker" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    /**
     * Options for configuring a day picker control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * Aria label for the DayPicker.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * Validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * Sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: number[] | KnockoutObservableBase<number[]>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
    }
    /**
     * DayPicker control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<number[]>;
    }
    /**
     * Creates a day picker control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a day picker control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\DropDown.d.ts
declare module "Fx/Controls/DropDown" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export = Main;
    module Main {
        interface Item<TValue> {
            /**
             * Content to render for the item, supports html.
             * If you use html, please add an ariaLabel for screen reader support.
             */
            text: string | KnockoutObservableBase<string>;
            /**
             * Overrides what is shown in the selected state. Allows you to display different content then what is shown in the the dropdown.
             * This is especially useful when using filtering and items with html.
             */
            selectedItemOverride?: string | KnockoutObservableBase<string>;
            /**
             * Value of your item.
             */
            value: TValue;
            /**
             * Disabled state
             */
            disabled?: boolean | KnockoutObservableBase<boolean>;
            /**
             * Aria label for item.
             */
            ariaLabel?: string | KnockoutObservableBase<string>;
        }
        interface Group<TValue> {
            /**
             * Content to render for the item, supports html.
             */
            text: string | KnockoutObservableBase<string>;
            /**
             * Items to render under this group.
             */
            children?: Array<Item<TValue> | Group<TValue>> | KnockoutObservableBase<Array<Item<TValue> | Group<TValue>>>;
            /**
             * Disabled state
             */
            disabled?: boolean | KnockoutObservableBase<boolean>;
            /**
             * Aria label for group.
             */
            ariaLabel?: string | KnockoutObservableBase<string>;
        }
        interface Options<TValue> {
            /**
             * The label of the control
             */
            label?: string | KnockoutObservableBase<string>;
            /**
             * The sublabel of the control
             */
            subLabel?: string | KnockoutObservableBase<string>;
            /**
             * Aria label for group.
             */
            ariaLabel?: string | KnockoutObservableBase<string>;
            /**
             * validations on the control
             */
            validations?: Validation[] | KnockoutObservableArray<Validation>;
            /**
             * sanitized html string shown in the info balloon popup
             */
            infoBalloonContent?: string | KnockoutObservableBase<string>;
            /**
             * Items to render in your dropdown.
             */
            items?: Array<Item<TValue> | Group<TValue>> | KnockoutObservableArray<Item<TValue> | Group<TValue>>;
            /**
             * Sets the placeholder.
             * Default is blank.
             *
             * Multiselect doesn't support placeholder, use multiItemDisplayText.
             * filterPlaceholder will default to this value if it's not present.
             */
            placeholder?: string | KnockoutObservableBase<string>;
            /**
             * Turn filtering on/off.
             */
            filter?: boolean;
            /**
             * Sets the filter textbox's placeholder.
             * Default is blank.
             */
            filterPlaceholder?: string | KnockoutObservableBase<string>;
            /**
             * Callback which will override our simple substring filtering.
             * Return list of items (by item value) you wish to show.
             * Only works when filter is set to true.
             */
            customFilter?: (value: string) => MsPortalFx.Base.PromiseV<TValue[]>;
            /**
             * Toggle multiselect on.
             */
            multiselect?: boolean;
            /**
             * Multiple items text format (multiselect only).
             * By default the format string is "{0} selected"
             */
            multiItemDisplayText?: string | KnockoutObservableBase<string>;
            /**
             * Enables select all for multiselect.
             */
            selectAll?: boolean;
            /**
             * Callback which will override our selectAll functionality.
             * Allows you to select any items you wish, we don't do any selecting when you provide this override.
             */
            selectAllOverride?: () => MsPortalFx.Base.PromiseV<any>;
            /**
             * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
             */
            suppressDirtyBehavior?: boolean;
            /**
             * The loading state of the control. Defaults to false.
             */
            loading?: boolean | KnockoutObservableBase<boolean>;
            /**
             * The disabled state of the control.
             */
            disabled?: boolean | KnockoutObservableBase<boolean>;
            /**
             * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
             */
            value?: TValue | Array<TValue> | KnockoutObservableBase<TValue | Array<TValue>>;
            /**
             * The visible state of the control.
             */
            visible?: boolean | KnockoutObservableBase<boolean>;
            /**
             * A custom css class to apply to the control.
             */
            cssClass?: string | KnockoutObservableBase<string>;
        }
        /**
         * DropDown control view model
         */
        interface Contract<TValue> extends Base.Contract {
            /**
             * Popup state of of the dropdown.
             */
            readonly isPopUpOpen: KnockoutObservableBase<boolean>;
            /**
             * Current value of the filter textbox.
             */
            readonly filterString: KnockoutObservableBase<string>;
            /**
             * Sets the placeholder.
             * Default is blank.
             *
             * Multiselect doesn't support placeholder, use multiItemDisplayText.
             * filterPlaceholder will default to this value if it's not present.
             */
            readonly placeholder: KnockoutObservableBase<string>;
            /**
             * Items to render in your dropdown.
             */
            readonly items: KnockoutObservableArray<Item<TValue> | Group<TValue>>;
            /**
             * Puts the control into a loading state.  When in the loading state, the control will be disabled and have the text "Loading..." shown.
             */
            readonly loading: KnockoutObservableBase<boolean>;
            /**
             * Multiple items text format (multiselect only).
             * By default the format string is "{0} selected"
             */
            readonly multiItemDisplayText: KnockoutObservableBase<string>;
            /**
             * Stops the selected from updating, if items are changed.
             */
            readonly suppressSelectedUpdate: KnockoutObservableBase<boolean>;
            /**
             * the value of the control
             */
            readonly value: KnockoutObservableBase<TValue> & KnockoutObservableBase<Array<TValue>>;
        }
        /**
         * Creates a dropdown control viewmodel.
         *
         * @param lifetime The lifetime of the control
         * @param items The items shown in the dropdown
         * @param options Options that modify the appearance and behavior of the control.
         *
         * @returns a dropdown control viewmodel
         */
        function create<TValue>(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options<TValue>): Contract<TValue>;
        /**
         * Options used in old construction pattern.  For new controls, please use the new create pattern.
         */
        interface ObsoleteOptions<T> {
            /**
             * The label of the control
             */
            label?: string | KnockoutObservableBase<string>;
            /**
             * The sublabel of the control
             */
            subLabel?: string | KnockoutObservableBase<string>;
            /**
             * validations on the control
             */
            validations?: Validation[] | KnockoutObservableArray<Validation>;
            /**
             * sanitized html string shown in the info balloon popup
             */
            infoBalloonContent?: string | KnockoutObservableBase<string>;
            /**
             * Items to render in your dropdown.
             */
            items?: KnockoutObservableArray<Item<T> | Group<T>>;
            /**
             * Sets the placeholder.
             * Default is blank.
             *
             * Multiselect doesn't support placeholder, use multiItemDisplayText.
             * filterPlaceholder will default to this value if it's not present.
             */
            placeholder?: string | KnockoutObservableBase<string>;
            /**
             * Turn filtering on/off.
             */
            filter?: boolean | KnockoutObservableBase<boolean>;
            /**
             * Sets the filter textbox's placeholder.
             * Default is blank.
             */
            filterPlaceholder?: string | KnockoutObservableBase<string>;
            /**
             * Callback which will override our simple substring filtering.
             * Return list of items (by item value) you wish to show.
             * Only works when filter is set to true.
             */
            customFilter?: (value: string) => MsPortalFx.Base.PromiseV<T[]>;
            /**
             * Toggle multiselect on.
             */
            multiselect?: boolean | KnockoutObservableBase<boolean>;
            /**
             * Enables select all for multiselect.
             */
            selectAll?: boolean | KnockoutObservableBase<boolean>;
            /**
             * Callback which will override our selectAll functionality.
             * Allows you to select any items you wish, we don't do any selecting when you provide this override.
             */
            selectAllOverride?: () => MsPortalFx.Base.PromiseV<any>;
            /**
             * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
             */
            suppressDirtyBehavior?: boolean;
            /**
             * The disabled state of the control.
             */
            disabled?: KnockoutObservableBase<boolean>;
            /**
             * Show validation error message below the control.
             */
            showValidationMessagesBelowControl?: boolean;
            /**
             * Aria label for dropdown.
             */
            ariaLabel?: string | KnockoutObservableBase<string>;
            /**
             * Display text format for multiselect.
             * By default the format string is "{0} selected"
             */
            multiItemDisplayText?: string | KnockoutObservableBase<string>;
            /**
             * The visible state of the control.
             */
            visible?: boolean | KnockoutObservableBase<boolean>;
            /**
             * A custom css class to apply to the control.
             */
            cssClass?: string | KnockoutObservableBase<string>;
        }
        /**
         * Old construction pattern.  Instead of creating instances of this class, please use the create methods in this module instead.
         */
        class ViewModel<TValue> extends MsPortalFx.ViewModels.Forms.Base.Input.ViewModel<TValue> {
            /**
             * Items to render in your dropdown.
             */
            items: KnockoutObservableArray<Item<TValue> | Group<TValue>>;
            /**
             * Turn filtering on/off.
             */
            filter: boolean | KnockoutObservableBase<boolean>;
            /**
             * Sets the filter textbox's placeholder.
             * Default is blank.
             */
            filterPlaceholder: string | KnockoutObservableBase<string>;
            /**
             * Sets the placeholder.
             * Default is blank.
             *
             * Multiselect doesn't support placeholder, use multiItemDisplayText.
             * filterPlaceholder will default to this value if it's not present.
             */
            placeholder: string | KnockoutObservableBase<string>;
            /**
             * Current value of the filter textbox.
             */
            readonly filterString: KnockoutObservableBase<string>;
            /**
             * Callback which will override our simple substring filtering.
             * Return list of items (by item value) you wish to show.
             * Only works when filter is set to true.
             */
            customFilter: (value: string) => MsPortalFx.Base.PromiseV<TValue[]>;
            /**
             * Turn multiselect on/off.
             */
            multiselect: boolean | KnockoutObservableBase<boolean>;
            /**
             * Multiple items text format (multiselect only).
             * By default the format string is "{0} selected"
             */
            multiItemDisplayText: string | KnockoutObservableBase<string>;
            /**
             * Turn multiselect on/off.
             */
            selectAll: boolean | KnockoutObservableBase<boolean>;
            /**
             * Callback which will override our selectAll functionality.
             */
            selectAllOverride: () => MsPortalFx.Base.PromiseV<any>;
            /**
             * Stops the selected from updating, if items are changed.
             */
            suppressSelectedUpdate: KnockoutObservableBase<boolean>;
            /**
             * Value is of type T or Array<TValue> depending on whether multiselect is set;
             */
            value: KnockoutObservableBase<TValue> & KnockoutObservableBase<Array<TValue>>;
            /**
             * Aria label for dropdown.
             */
            readonly ariaLabel: KnockoutObservableBase<string>;
            /**
             * Popup state of of the dropdown.
             */
            readonly isPopUpOpen: KnockoutObservableBase<boolean>;
            /**
             * This is an old construction pattern.  Please use the create method in this module instead.
             *
             * @param lifetimeManager A LifetimeManager object that will notify when the data is no longer being used by the caller.
             * @param options Optional The set of options to configure the Dropdown control.
             */
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, options?: ObsoleteOptions<TValue> & MsPortalFx.ViewModels.Forms.Base.Input.Options<TValue>);
            /**
             * This is an old construction pattern.  Please use the create method in this module instead.
             *
             * @param lifetimeManager A LifetimeManager object that will notify when the data is no longer being used by the caller.
             * @param form The form element within which the Dropdown is contained.
             * @param observablePath The path to the value on the EditScope to be bound to this field.
             * @param options Optional The set of options to configure the Dropdown control.
             */
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, form: MsPortalFx.ViewModels.Forms.Form.ViewModel<any>, observablePath: string, options?: ObsoleteOptions<TValue> & MsPortalFx.ViewModels.Forms.Base.Input.Options<TValue>);
            /**
             * This is an old construction pattern.  Please use the create method in this module instead.
             *
             * @param lifetimeManager A LifetimeManager object that will notify when the data is no longer being used by the caller.
             * @param form The form element within which the Dropdown is contained.
             * @param accessor Used to read and write values to the edit scope.  Use Form.createEditScopeAccessor methods to create this object.
             * @param options Optional The set of options to configure the Dropdown control.
             */
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, form: MsPortalFx.ViewModels.Forms.Form.ViewModel<any>, accessor: MsPortalFx.ViewModels.Forms.EditScopeAccessors<any>, options?: ObsoleteOptions<TValue> & MsPortalFx.ViewModels.Forms.Base.Input.Options<TValue>);
        }
    }
}

// FILE: Fx\Controls\DurationPicker.d.ts
declare module "Fx/Controls/DurationPicker" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    /**
     * Options for configuring a duration picker control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: number | KnockoutObservableBase<number>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
    }
    /**
     * DurationPicker control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<number>;
    }
    /**
     * Creates a duration picker control viewmodel.
     * Note: The only options that you will be able to dynamically update are the ones that are passed in as observables.
     * ie: if you pass in { label: ko.observable("myLabel") }, the label property on your viewmodel will be observable, but the sublabel property will not be.
     * similarly, if you pass in { label: "myLabel" } as your options, and you will not be able to update your label dynamically.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a duration picker control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\EditableGrid.d.ts
declare module "Fx/Controls/EditableGrid" {
    import { SelectionMode } from "Fx/Controls/GridBase";
    /**
     * EditableGrid selection mode.
     */
    export { SelectionMode };
    /**
     * EditableGrid column header definition.
     * The ColumnHeaderDefinition contains properties that are used to create a ColumnHeader
     * when constructing the grid or adding dynamic columns via grid.columnDefinitions.
     */
    export interface ColumnHeaderDefinition {
        /**
         * Aria label for the header.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * Content for the header.
         * The content can be a string, a control, or an observable that reutrns a string or control.
         */
        content?: string | MsPortalFx.Control | KnockoutObservableBase<string | MsPortalFx.Control>;
    }
    /**
     * EditableGrid column header.
     * Accessed by grid.columns()[index].header or grid.columns.get(id).header.
     */
    export interface ColumnHeader {
        /**
         * Aria label for the header.
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * Content for the header.
         */
        readonly content: KnockoutObservableBase<string | MsPortalFx.Control>;
    }
    /**
     * EditableGrid column definition.
     * The ColumnDefinition contains properties that are used to create a Column.
     */
    export interface ColumnDefinition<T> {
        /**
         * Indicates if the column width is resizable or fixed.
         * Defaults to true.
         */
        canResize?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Creates an input control such as TextBox, DropDown, or NumericTextBox that
         * enables the user to edit the cell value.
         *
         *     createCellEditor: (editorLifetime, item) => {
         *         return TextBox.create(editorLifetime, {
         *             value: item.valueProperty // Bind the item property to the text box for editing
         *         });
         *     })
         *
         * @param editorLifetime The lifetime manager for the cell editor control.
         * @param item The data item for the row.
         * @param row The containing row.
         * @returns The input control used for editing the cell value.
         */
        createCellEditor?: (editorLifetime: MsPortalFx.Base.LifetimeManager, item: T, row: Row<T>) => MsPortalFx.Control | KnockoutReadOnlyObservableBase<MsPortalFx.Control>;
        /**
         * Callback to define cell properties.
         *
         * @param cellLifetime The cell lifetime manager.
         * @param item The item to create the cell for.
         * @param row The containing row.
         * @returns The cell definition or the cell display content as a string or control.
         */
        defineCell: (cellLifetime: MsPortalFx.Base.LifetimeManager, item: T, row: Row<T>) => CellDefinition | string | MsPortalFx.Control | KnockoutReadOnlyObservableBase<string | MsPortalFx.Control>;
        /**
         * The column header definition or text content.
         * For simple cases this can be a string.
         * If need to set additional column header properties provide a column header definition.
         */
        header?: ColumnHeaderDefinition | string | MsPortalFx.Control | KnockoutReadOnlyObservableBase<string | MsPortalFx.Control>;
        /**
         * The column identifier.
         * Each column requires a unique identifier.
         * If not provided a default one is created.
         * The id can also be used to access a cell row.cells.get(column.id).
         */
        id?: string;
        /**
         * The column width.
         * Percentage '%', pixel 'px', and fractional 'fr' units are supported.
         * Percengage widths are relative to the available space in the container.
         * Fractional widths are used divide up the remaining space proportionally.
         * When undefined the width is treated as '1fr' which equally divides the space between all visible columns of undefined width.
         */
        width?: string | KnockoutObservableBase<string>;
    }
    /**
     * EditableGrid column.
     * Part of the EditableGrid object model.
     * Accessed by grid.columns.
     */
    export interface Column<T> {
        "--noUnusedLocals"?: any | T;
        /**
         * Indicates if the column width is resizable or fixed.
         */
        readonly canResize: KnockoutObservableBase<boolean>;
        /**
         * The column header.
         */
        readonly header: ColumnHeader;
        /**
         * The column identifier.
         * Each column requires a unique identifier.
         * If not provided a default one is created.
         * The id can also be used to access a cell row.cells.get(column.id).
         */
        readonly id: string;
        /**
         * The column width.
         * Percentage '%', pixel 'px', and fractional 'fr' units are supported.
         * Percengage widths are relative to the available space in the container.
         * Fractional widths are used divide up the remaining space proportionally.
         * When undefined the width is treated as '1fr' which equally divides the space between all visible columns of undefined width.
         */
        readonly width: KnockoutObservableBase<string>;
    }
    /**
     * Editable grid column collection.
     */
    export interface ColumnCollection<T> extends KnockoutReadOnlyObservableArray<Column<T>> {
        /**
         * Adds columns to the grid.
         * The columns are created from the supplied column definitions and inserted.
         *
         * @param columnDefinitions Array of column definitions create columns from.
         * @param index The index to insert the new columns in the columns collection.  If not specified the columns are added to the end.
         */
        add(columsDefinitions: ColumnDefinition<T>[], index?: number): void;
        /**
         * Gets a column by column id.
         * @param columnId The id of the column.
         * @returns The indicated column.
         */
        get(columnId: string): Column<T>;
        /**
         * Removes columns from the grid.
         *
         * @param ids The column ids of the columns to remove from the grid.
         * @returns The original column definitions used to create the columns.
         */
        remove(ids: string[]): ColumnDefinition<T>[];
        /**
         * Removes all columns from the grid.
         *
         * @returns The original column definitions used to create the columns.
         */
        removeAll(): ColumnDefinition<T>[];
    }
    /**
     * EditableGrid cell definition.
     * The CellDefinition contains properties that are used to initialize a new Cell.
     */
    export interface CellDefinition {
        /**
         * The accesibility label for the cell.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The content to display in the cell.
         * The content will usually be a string representation of a data item value.
         * For numbers and dates this may require NumberFormat and DateTimeFormat from "MsPortalFx/Globalization".
         * If the value can change and you want the cell to update use an observable or computed for the content.
         * To display images or other UI in the cell provide CustomHtml or other control for the content.
         */
        content?: string | MsPortalFx.Control | KnockoutReadOnlyObservableBase<string | MsPortalFx.Control>;
    }
    /**
     * EditableGrid cell.
     * Part of the EditableGrid object model.
     * Accessed by row.cells()[index] or row.cells.get(columnId).
     */
    export interface Cell<T> {
        /**
         * The aria label for the cell.
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * The cell column.
         */
        readonly column: Column<T>;
        /**
         * The content to display in the cell.
         */
        readonly content: KnockoutReadOnlyObservableBase<string | MsPortalFx.Control>;
        /**
         * The cell editor control for editing cell content.
         * Will be null if cell has not been edited or is not editable.
         */
        readonly editor: KnockoutReadOnlyObservableBase<MsPortalFx.Control>;
        /**
         * Indicates if the cell editor control is dirty because the user made a change.
         */
        readonly dirty: KnockoutReadOnlyObservableBase<boolean>;
        /**
         * Triggers validation of the cell.
         */
        triggerValidation(): Q.Promise<boolean>;
        /**
         * Extension-defined cell validation errors.
         * Extensions can add custom errors to this array to flag a cell as invalid in the UI.
         * When the error condition is cleared the extension should remove the errors from this array to remove the error UI.
         */
        readonly customValidationErrors: KnockoutObservableArray<string>;
        /**
         * Validation results from the cell editor control.
         * This is a read-only collection of the validation errors extracted from the cell editor control.
         */
        readonly editorValidationErrors: KnockoutReadOnlyObservableBase<string[]>;
    }
    /**
     * EditableGrid row definition.
     * The RowDefinition contains properties that are used to initialize a new Row.
     */
    export interface RowDefinition {
        /**
         * The row accesibility label.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The row disabled state.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
    }
    /**
     * EditableGrid cell collection.
     */
    export interface CellCollection<T> extends KnockoutReadOnlyObservableArray<Cell<T>> {
        /**
         * Gets a cell by column id.
         * @param columnId The column id.
         * @returns The requested cell.
         */
        get(columnId: string): Cell<T>;
    }
    /**
     * EditableGrid row.
     * Part of the EditableGrid object model.
     * Accessed by grid.rows()[index].
     */
    export interface Row<T> {
        /**
         * The row accesibility label.
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * The row cells.
         * A particular cell can be retrieved by column id cells.get(columnId).
         */
        readonly cells: CellCollection<T>;
        /**
         * The row disabled state.
         */
        readonly disabled: KnockoutObservableBase<boolean>;
        /**
         * Indicates if the row is currently focused.
         */
        readonly focused: KnockoutReadOnlyObservable<boolean>;
        /**
         * The data item.
         */
        readonly item: T;
        /**
         * Indicates if the row is selected.
         */
        readonly selected: KnockoutReadOnlyObservable<boolean>;
        /**
         * Indicates if the row is being edited.
         */
        readonly editing: KnockoutReadOnlyObservable<boolean>;
        /**
         * Indicates if the row is dirty because the user edited the value of one or more of the cells.
         */
        readonly dirty: KnockoutReadOnlyObservableBase<boolean>;
        /**
         * Triggers validation of the row.
         */
        triggerValidation(): Q.Promise<boolean>;
        /**
         * All cell and cell editor control validation errors for the row.
         * This is a read-only collection of all the cell validation errors within the row.
         */
        readonly validationErrors: KnockoutReadOnlyObservableBase<string[]>;
    }
    /**
     * Grid header definition.
     */
    export interface HeaderDefinition {
        /**
         * Indicates if the header should be visible.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
    }
    /**
     * Grid header.
     * Part of the EditableGrid object model.
     * Accessed by grid.header.
     */
    export interface Header {
        /**
         * Determines if the header is visible.
         */
        readonly visible: KnockoutObservableBase<boolean>;
    }
    /**
     * EditableGrid focus options.
     */
    export interface FocusOptions<T> {
        /**
         * Event handler for RowEnter event.
         *
         * @param row The row being entered.
         * @param grid The grid.
         */
        onEnterRow?: (row: Row<T>, grid: Contract<T>) => void;
        /**
         * Event handler for RowExit event.
         *
         * @param row The row being exited.
         * @param grid The grid.
         */
        onExitRow?: (row: Row<T>, grid: Contract<T>) => void;
        /**
         * Event handler for CellEnter event.
         *
         * @param cell The cell being entered.
         * @param row The row containing the cell.
         * @param grid The grid.
         */
        onEnterCell?: (cell: Cell<T>, row: Row<T>, grid: Contract<T>) => void;
        /**
         * Event handler for CellExit event.
         *
         * @param cell The cell being exited.
         * @param row The row containing the cell.
         * @param grid The grid.
         */
        onExitCell?: (cell: Cell<T>, row: Row<T>, grid: Contract<T>) => void;
    }
    /**
     * EditableGrid focus contract.
     */
    export interface FocusContract<T> {
        /**
         * The current row.
         * This row is the focused row when the grid has focus.
         * When focus leaves the grid the current row remains.
         */
        readonly currentRow: KnockoutReadOnlyObservable<Row<T>>;
        /**
         * The current cell.
         * This cell is the focused cell when the grid has focus.
         * When focus leaves the grid the current cell remains.
         * The current cell will be null when no cell contains the focus.
         */
        readonly currentCell: KnockoutReadOnlyObservable<Cell<T>>;
        /**
         * Moves the focus from the current row to the specified row.
         * Makes the specified row the current row.
         * Focus will not change unless the grid already has focus.
         *
         * @param row The row to move to.
         * @returns Indicates if the focus moved.
         */
        moveToRow(row: Row<T>): boolean;
        /**
         * Moves the focus from the current row to the row below it if possible.
         *
         * @returns Indicates if the focus moved down.
         */
        moveDown(): boolean;
        /**
         * Moves the focus from the current row to the row above it if possible.
         *
         * @returns Indicates if the focus moved up.
         */
        moveUp(): boolean;
        /**
         * Moves the focus from the current row to the first row if possible.
         *
         * @returns Indicates if the focus moved to the first row.
         */
        moveFirst(): boolean;
        /**
         * Moves the focus from the current row to the last row if possible.
         *
         * @returns Indicates if the focus moved to the last row.
         */
        moveLast(): boolean;
    }
    /**
     * EditableGrid selection options.
     */
    export interface SelectionOptions<T> {
        /**
         * Selection mode (defaults to Single).
         */
        selectionMode?: SelectionMode;
        /**
         * Disables Ctrl+A or the select all checkbox from selecting all rows at once in multiple selection mode.
         */
        selectAllDisabled?: boolean;
        /**
         * Determines if a row can be selected.
         *
         * @param row The row to select.
         * @param grid The grid.
         * @returns Indicates if the row can be selected.
         */
        canSelectRow?: (row: Row<T>, grid: Contract<T>) => boolean;
        /**
         * Determines if a row can be unselected.
         *
         * @param row The row to unselect.
         * @param grid The grid.
         * @returns Indicates if the row can be unselected.
         */
        canUnselectRow?: (row: Row<T>, grid: Contract<T>) => boolean;
        /**
         * Determines if all rows can be selected at once.
         *
         * @param grid The grid.
         * @returns Indicates if all rows can be selected at once.
         */
        canSelectAllRows?: (grid: Contract<T>) => boolean;
        /**
         * Determines if all rows can be unselected at once.
         *
         * @param grid The grid.
         * @returns Indicates if all rows can be unselected at once.
         */
        canUnselectAllRows?: (grid: Contract<T>) => boolean;
    }
    /**
     * EditableGrid selection state.
     */
    export interface SelectionState<T> {
        /**
         * Indicates if all rows are selected.
         */
        readonly allRowsSelected: KnockoutReadOnlyObservableBase<boolean>;
        /**
         * The currently selected rows.
         * Commands should use the selectedOrCurrentRows property in most scenarios.
         */
        readonly selectedRows: KnockoutReadOnlyObservableArray<Row<T>>;
        /**
         * The selected rows or the current row if there are no selected rows.
         * Commands should use this property instead of selectedRows if they
         * must work on the current row when the user has not made a selection yet.
         */
        readonly selectedOrCurrentRows: KnockoutReadOnlyObservableArray<Row<T>>;
    }
    /**
     * EditableGrid selection contract.
     */
    export interface SelectionContract<T> extends SelectionState<T> {
        /**
         * The selection mode.
         */
        readonly selectionMode: SelectionMode;
        /**
         * Indicates if all rows are selected.
         */
        readonly allRowsSelected: KnockoutReadOnlyObservableBase<boolean>;
        /**
         * The currently selected rows.
         * Commands should use the selectedOrCurrentRows property in most scenarios.
         */
        readonly selectedRows: KnockoutReadOnlyObservableArray<Row<T>>;
        /**
         * The selected rows or the current row if there are no selected rows.
         * Commands should use this property instead of selectedRows if they
         * must work on the current row when the user has not made a selection yet.
         */
        readonly selectedOrCurrentRows: KnockoutReadOnlyObservableArray<Row<T>>;
        /**
         * Determines if a row can be selected.
         *
         * @row The row to select.
         * @return Indicates if a row can be selected.
         */
        canSelectRow(row: Row<T>): boolean;
        /**
         * Selects a row.
         *
         * @row The row to select.
         * @return Indicates if the row was selected.
         */
        selectRow(row: Row<T>): boolean;
        /**
         * Determines if a row can be unselected.
         *
         * @row The row to unselect.
         * @return Indicates if a row can be unselected.
         */
        canUnselectRow(row: Row<T>): boolean;
        /**
         * Unselects a row.
         *
         * @row The row to unselect.
         * @return Indicates if the row was unselected.
         */
        unselectRow(row: Row<T>): boolean;
        /**
         * Indicates if all rows can be selected.
         */
        readonly canSelectAllRows: KnockoutReadOnlyObservableBase<boolean>;
        /**
         * Selects all rows.
         *
         * @return Indicates if all rows were selected.
         */
        selectAllRows(): boolean;
        /**
         * Indicates if all rows can be unselected.
         */
        readonly canUnselectAllRows: KnockoutReadOnlyObservableBase<boolean>;
        /**
         * Unselects all rows.
         *
         * @return Indicates if all rows were unselected.
         */
        unselectAllRows(): boolean;
    }
    /**
     * EditableGrid toolbar options.
     */
    export interface ToolbarOptions<T> {
        /**
         * Hides or shows the toolbar.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Callback to create a customized toolbar.
         * If no callback is supplied the EditableGrid will create a default toolbar.
         *
         * @param toolbarLifetime The lifetime manager for the toolbar.
         * @param grid The grid for creating built-in toolbar items like grid.deletion.createDeleteRowsCommand(toolbarLifetime, grid.selection).
         *             (BREAKING: This parameter used to be a ToolbarItems interface and is now a grid contract to be used as described above.)
         * @returns The toolbar view model that will be used within the grid header.
         */
        createToolbar?: (toolbarLifetime: MsPortalFx.Base.LifetimeManager, grid: Contract<T>) => MsPortalFx.ViewModels.Toolbars.Toolbar;
    }
    /**
     * EditableGrid toolbar contract.
     */
    export interface ToolbarContract<T> {
        "--noUnusedLocals"?: any | T;
        /**
         * The toolbar view model.
         */
        readonly viewModel: MsPortalFx.ViewModels.Toolbars.Toolbar;
        /**
         * Controls the toolbar visiblity in the grid header.
         */
        readonly visible: KnockoutObservableBase<boolean>;
        /**
         * OBSOLETE: Use factory methods such as the following instead:
         * grid.deletion.createDeleteRowsCommand(toolbarLifetime, grid.selection)
         * grid.insertion.createInsertRowCommand(toolbarLifetime, grid.selection)
         * grid.reordering.createMoveRowsUpCommand(toolbarLifetime, grid.selection)
         * grid.reordering.createMoveRowsDownCommand(toolbarLifetime, grid.selection)
         * grid.reordering.createMoveRowsTopCommand(toolbarLifetime, grid.selection)
         * grid.reordering.createMoveRowsBottomCommand(toolbarLifetime, grid.selection)
         */
        readonly supportedItems: Obsolete;
    }
    /**
     * EditableGrid addition options.
     */
    export interface AdditionOptions<T> {
        /**
         * Callback to create a new item.
         * The item will be held in the placeholder row until commited.
         *
         * @returns A newly created item.
         */
        createItem: () => T;
        /**
         * Determines if the user can add additional rows.
         *
         * @param grid The grid.
         * @returns Indicates if the user can add additional rows
         */
        canAddRow?: (grid: Contract<T>) => boolean;
        /**
         * Callback when row is added.
         *
         * @param row The added row.
         * @param grid The grid.
         */
        onAddRow?: (row: Row<T>, grid: Contract<T>) => void;
    }
    /**
     * EditableGrid addition contract.
     */
    export interface AdditionContract<T> {
        "--noUnusedLocals"?: any | T;
        /**
         * Determines if rows can be added
         */
        canAddRow: KnockoutReadOnlyObservableBase<boolean>;
    }
    /**
     * EditableGrid deletion options.
     */
    export interface DeletionOptions<T> {
        /**
         * Determines if rows can be deleted.
         *
         * @param rows The rows to delete.
         * @param grid The grid.
         * @returns Indicates if the rows can be deleted.
         */
        canDeleteRows?: (rows: Row<T>[], grid: Contract<T>) => boolean;
        /**
         * Notifies when rows have been deleted.
         *
         * @param rows The rows that were deleted.
         * @param grid The grid.
         */
        onDeleteRows?: (rows: Row<T>[], grid: Contract<T>) => void;
    }
    /**
     * EditableGrid deletion contract.
     */
    export interface DeletionContract<T> {
        /**
         * Determines if rows can be deleted.
         *
         * @param rows The rows to delete.
         * @returns Indicates if the rows can be deleted.
         */
        canDeleteRows(rows: Row<T>[]): boolean;
        /**
         * Deletes the specified rows.
         *
         * @param rows The rows to delete.
         * @returns Indicates if the rows were deleted.
         */
        deleteRows(rows: Row<T>[]): boolean;
        /**
         * Creates a toolbar button for deletion using the supplied rows as the command context.
         *
         * @param lifetime The lifetime manager for the command.
         * @param context The row or rows to be deleted.
         * @param options Toolbar button options for overriding defaults.
         */
        createDeleteRowsCommand(lifetime: MsPortalFx.Base.LifetimeManager, context: Row<T> | KnockoutObservableBase<Row<T>> | Row<T>[] | KnockoutObservableBase<Row<T>[]> | SelectionState<T>, options?: MsPortalFx.ViewModels.Toolbars.ToolbarButtonOptions): MsPortalFx.ViewModels.Toolbars.ToolbarButton;
    }
    /**
     * EditableGrid editing options.
     */
    export interface EditingOptions<T> {
        /**
         * Determines if the specified row can be edited.
         *
         * @param row The row to edit.
         * @param grid The grid.
         * @returns Indicates if the row can be edited.
         */
        canEditRow?: (row: Row<T>, grid: Contract<T>) => boolean;
        /**
         * Determines if the specified row can stop editing.
         *
         * @param row The row to stop editing.
         * @param grid The grid.
         * @returns Promise that indicates if the row can stop editing.
         */
        canStopEditingRow?: (row: Row<T>, grid: Contract<T>) => Q.Promise<boolean>;
    }
    /**
     * EditableGrid editing contract.
     */
    export interface EditingContract<T> {
        /**
         * Determines if a row can be edited.
         *
         * @param row The row to edit.
         * @returns Indicates if the row can be edited.
         */
        canEditRow(row: Row<T>): boolean;
        /**
         * Puts the specified row into editing mode for the user to edit it.
         *
         * @param row The row to edit.
         * @returns Indicates if the row was put into edit mode.
         */
        editRow(row: Row<T>): boolean;
        /**
         * Determine if the row can close all cell editors.
         *
         * @param row The row to stop editing.
         * @returns Promise that indicates if the row can stop editing.
         */
        canStopEditingRow(row: Row<T>): Q.Promise<boolean>;
        /**
         * Stops editing the row by closing all the cell editors of the row.
         *
         * @param row The row to stop editing.
         * @returns Promise that indicates if the row succesfully stopped editing.
         */
        stopEditingRow(row: Row<T>): Q.Promise<boolean>;
    }
    /**
     * EditableGrid insertion options.
     */
    export interface InsertionOptions<T> {
        /**
         * Callback to create an new item for insertion.
         *
         * @returns A newly created item.
         */
        createItem: () => T;
        /**
         * Determines if a new row can be inserted before the specified row.
         *
         * @param rowBefore The row to insert a new row before.
         * @param grid The grid.
         * @returns Indicates if a new row can be inserted before the specified row.
         */
        canInsertRow?: (rowBefore: Row<T>, grid: Contract<T>) => boolean;
        /**
         * Notifies when a row has been inserted.
         *
         * @param row The row that was inserted.
         * @param grid The grid.
         */
        onInsertRow?: (row: Row<T>, grid: Contract<T>) => void;
    }
    /**
     * EditableGrid insertion contract.
     */
    export interface InsertionContract<T> {
        /**
         * Determine if a new row can be inserted before the specified row.
         *
         * @param rowBefore The row to insert a new row before.
         * @returns Indicates if a new row can be inserted before the specified row.
         */
        canInsertRow(rowBefore: Row<T>): boolean;
        /**
         * Inserts a new row before the specified row.
         * The grid calls the createItem callback to generate the new item that will be inserted.
         *
         * @param rowBefore The row to insert a new row before.
         * @returns Indicates if a new row was inserted.
         */
        insertRow(rowBefore: Row<T>): boolean;
        /**
         * Creates a toolbar button for inserting a row before the supplied row context.
         *
         * @param lifetime The lifetime manager for the command.
         * @param context The current row or rows to apply the command status to.  Will be disabled if more than one row is present.
         * @param options Toolbar button options for overriding defaults.
         */
        createInsertRowCommand(lifetime: MsPortalFx.Base.LifetimeManager, context: Row<T> | KnockoutObservableBase<Row<T>> | Row<T>[] | KnockoutObservableBase<Row<T>[]> | SelectionState<T>, options?: MsPortalFx.ViewModels.Toolbars.ToolbarButtonOptions): MsPortalFx.ViewModels.Toolbars.ToolbarButton;
    }
    /**
     * EditableGrid reordering options.
     */
    export interface ReorderingOptions<T> {
        /**
         * Determines if the rows can be moved to the specified row position.
         *
         * @param rows The rows to move.
         * @param to The row in the position where the rows are to be moved.
         * @param grid The grid.
         * @returns Indicates if the rows can be moved to the specified row position.
         */
        canMoveRowsToRow?: (rows: Row<T>[], to: Row<T>, grid: Contract<T>) => boolean;
        /**
         * Determines if the rows can be moved up.
         *
         * @param rows The rows to move.
         * @param grid The grid.
         * @returns Indicates if the rows can be moved up.
         */
        canMoveRowsUp?: (rows: Row<T>[], grid: Contract<T>) => boolean;
        /**
         * Determines if the rows can be moved down.
         *
         * @param rows The rows to move.
         * @param grid The grid.
         * @returns Indicates if the rows can be moved down.
         */
        canMoveRowsDown?: (rows: Row<T>[], grid: Contract<T>) => boolean;
        /**
         * Determines if the rows can be moved to top.
         *
         * @param rows The rows to move.
         * @param grid The grid.
         * @returns Indicates if the rows can be moved to the top.
         */
        canMoveRowsTop?: (rows: Row<T>[], grid: Contract<T>) => boolean;
        /**
         * Determines if the rows can be moved to bottom.
         *
         * @param rows The rows to move.
         * @param grid The grid.
         * @returns Indicates if the rows can be moved to the bottom.
         */
        canMoveRowsBottom?: (rows: Row<T>[], grid: Contract<T>) => boolean;
    }
    /**
     * EditableGrid reordering contract.
     */
    export interface ReorderingContract<T> {
        /**
         * Determines if rows can be moved to the specified row position.
         *
         * @param rows The rows to move.
         * @param to The row in the position where the rows are to be moved.
         * @returns Indicates if the rows can be moved to the specified row position.
         */
        canMoveRowsToRow(rows: Row<T>[], to: Row<T>): boolean;
        /**
         * Moves rows to the specified row position.
         *
         * @param rows The rows to move.
         * @param to The row in the position where the rows are to be moved.
         * @returns Indicates if the rows were moved.
         */
        moveRowsToRow(rows: Row<T>[], to: Row<T>): boolean;
        /**
         * Determines if rows can be moved up.
         *
         * @param rows The rows to move.
         * @returns Indicates if the rows can be moved up.
         */
        canMoveRowsUp(rows: Row<T>[]): boolean;
        /**
         * Moves rows up.
         *
         * @param rows The rows to move.
         * @returns Indicates if the rows were moved.
         */
        moveRowsUp(rows: Row<T>[]): boolean;
        /**
         * Determines if rows can be moved down.
         *
         * @param rows The rows to move.
         * @returns Indicates if the rows can be moved down.
         */
        canMoveRowsDown(rows: Row<T>[]): boolean;
        /**
         * Moves rows down.
         *
         * @param rows The rows to move.
         * @returns Indicates if the rows were moved.
         */
        moveRowsDown(rows: Row<T>[]): boolean;
        /**
         * Determines if rows can be moved to the top.
         *
         * @param rows The rows to move.
         * @returns Indicates if the rows can be moved to the top.
         */
        canMoveRowsTop(rows: Row<T>[]): boolean;
        /**
         * Moves rows to the top.
         *
         * @param rows The rows to move.
         * @returns Indicates if the rows were moved.
         */
        moveRowsTop(rows: Row<T>[]): boolean;
        /**
         * Determines if rows can be moved to the bottom.
         *
         * @param rows The rows to move.
         * @returns Indicates if the rows can be moved to the bottom.
         */
        canMoveRowsBottom(rows: Row<T>[]): boolean;
        /**
         * Moves rows to the bottom.
         *
         * @param rows The rows to move.
         * @returns Indicates if the rows were moved.
         */
        moveRowsBottom(rows: Row<T>[]): boolean;
        /**
         * Creates a toolbar button for moving the supplied context rows up.
         *
         * @param lifetime The lifetime manager for the command.
         * @param context The row or rows to move.
         * @param options Toolbar button options for overriding defaults.
         * @return The command button.
         */
        createMoveRowsUpCommand(lifetime: MsPortalFx.Base.LifetimeManager, context: Row<T> | KnockoutObservableBase<Row<T>> | Row<T>[] | KnockoutObservableBase<Row<T>[]> | SelectionState<T>, options?: MsPortalFx.ViewModels.Toolbars.ToolbarButtonOptions): MsPortalFx.ViewModels.Toolbars.ToolbarButton;
        /**
         * Creates a toolbar button for moving the supplied context rows down.
         *
         * @param lifetime The lifetime manager for the command.
         * @param context The row or rows to move.
         * @param options Toolbar button options for overriding defaults.
         * @return The command button.
         */
        createMoveRowsDownCommand(lifetime: MsPortalFx.Base.LifetimeManager, context: Row<T> | KnockoutObservableBase<Row<T>> | Row<T>[] | KnockoutObservableBase<Row<T>[]> | SelectionState<T>, options?: MsPortalFx.ViewModels.Toolbars.ToolbarButtonOptions): MsPortalFx.ViewModels.Toolbars.ToolbarButton;
        /**
         * Creates a toolbar button for moving the supplied context rows to the top.
         *
         * @param lifetime The lifetime manager for the command.
         * @param context The row or rows to move.
         * @param options Toolbar button options for overriding defaults.
         * @return The command button.
         */
        createMoveRowsTopCommand(lifetime: MsPortalFx.Base.LifetimeManager, context: Row<T> | KnockoutObservableBase<Row<T>> | Row<T>[] | KnockoutObservableBase<Row<T>[]> | SelectionState<T>, options?: MsPortalFx.ViewModels.Toolbars.ToolbarButtonOptions): MsPortalFx.ViewModels.Toolbars.ToolbarButton;
        /**
         * Creates a toolbar button for moving the supplied context rows to the bottom.
         *
         * @param lifetime The lifetime manager for the command.
         * @param context The row or rows to move.
         * @param options Toolbar button options for overriding defaults.
         * @return The command button.
         */
        createMoveRowsBottomCommand(lifetime: MsPortalFx.Base.LifetimeManager, context: Row<T> | KnockoutObservableBase<Row<T>> | Row<T>[] | KnockoutObservableBase<Row<T>[]> | SelectionState<T>, options?: MsPortalFx.ViewModels.Toolbars.ToolbarButtonOptions): MsPortalFx.ViewModels.Toolbars.ToolbarButton;
    }
    /**
     * EditableGrid resizing options.
     */
    export interface ResizingOptions<T> {
        /**
         * Notification when columns have been resized by the user.
         *
         * @columns The columns that changed widths.
         * @grid The grid.
         */
        onColumnsResized?: (columns: Column<T>[], grid: Contract<T>) => void;
    }
    /**
     * EditableGrid validation options.
     */
    export interface ValidationOptions<T> {
        /**
         * Custom validation call back.
         * Called to validate the grid when the form is validated.
         * Errors can be added and removed from particualar cells using the cell customValidationErrors array.
         *
         * @param grid The grid.
         * @returns Indicates if the grid is valid.
         */
        validate?: (grid: Contract<T>) => boolean;
        /**
         * Custom async validation call back.
         * Called to validate the grid when the form is validated.
         * Errors can be added and removed from particualar cells using the cell customValidationErrors array.
         *
         * @param grid The grid.
         * @returns Promise indicating if the grid is valid.
         */
        validateAsync?: (grid: Contract<T>) => Q.Promise<boolean>;
        /**
         * Creates a customized display of validation errors.
         *
         * @param lifetime The lifetime manager for the validation footer.
         * @param row The row to contain the footer. The validationErrors property of the row will have the errors to display.
         * @returns The control to present as the row validation footer.
         */
        createRowValidationFooter?: (lifetime: MsPortalFx.Base.LifetimeManager, row: Row<T>) => MsPortalFx.Control | KnockoutObservableBase<MsPortalFx.Control>;
    }
    /**
     * EditableGrid validation contract.
     */
    export interface ValidationContract<T> {
        "--noUnusedLocals"?: any | T;
        /**
         * Triggers validation of the grid.
         *
         * @returns Promise that indicates if the grid is valid.
         */
        triggerValidation(): Q.Promise<boolean>;
        /**
         * Indicates if the grid is valid.
         * This is the result of the latest validation.
         */
        readonly valid: KnockoutReadOnlyObservable<boolean>;
    }
    /**
     * EditableGrid context menu options.
     */
    export interface ContextMenuOptions<T> {
        /**
         * Determines if the row supports context menu.
         * This is true by default and only needs to be implemented to opt specific rows out.
         * This controls if the context menu shortcut is displayed in the row.
         * It also controls if right-click/menu-key call supplyMenuCommands/supplyMenuBulkCommands.
         * supplyMenuCommands/supplyMenuBulkCommands can still return nothing if they do not want a menu shown.
         *
         * @param row The row to show menus for.
         * @param grid The grid.
         * @returns Indicates if the row can show a context menu.
         */
        canShowMenu?: (row: Row<T>, grid: Contract<T>) => boolean;
        /**
         * Supplies commands to be displayed in the row context menu.
         * This is not called when multiple rows are selected.
         *
         * @param lifetime The lifetime manager for any command related allocations.
         * @param row The row to supply commands for.
         * @param grid The grid.
         * @returns The commands for the context menu or a promise for the commands.
         */
        supplyMenuCommands?: (lifetime: MsPortalFx.Base.LifetimeManager, row: Row<T>, grid: Contract<T>) => MsPortalFx.ViewModels.Toolbars.ToolbarItemContract[] | Q.Promise<MsPortalFx.ViewModels.Toolbars.ToolbarItemContract[]>;
        /**
         * Supplies bulk commands to be displayed for the rows in a context menu.
         * This is called only when multiple rows are selected.
         *
         * @param lifetime The lifetime manager for any command related allocations.
         * @param selection The selection to supply commands for.
         * @param grid The grid.
         * @returns The commands for the context menu or a promise for the commands.
         */
        supplyMenuBulkCommands?: (lifetime: MsPortalFx.Base.LifetimeManager, selection: SelectionState<T>, grid: Contract<T>) => MsPortalFx.ViewModels.Toolbars.ToolbarItemContract[] | Q.Promise<MsPortalFx.ViewModels.Toolbars.ToolbarItemContract[]>;
        /**
         * Determines the maximum buttons commands to display directly in the row.
         * The default is one.
         * Command buttons will be displayed in the row until the count of commands exceeds this max.
         * When the max is exceeded the ellipsis will be displayed indicating more commands and providing a way to open the menu.
         * If zero only the ellipsis will ever be shown.
         */
        maxButtonCommands?: number;
        /**
         * Supplies command buttons to be displayed in the row margin.
         *
         * @param lifetime The lifetime manager for any command related allocations.
         * @param row The row to supply commands for.
         * @param grid The grid.
         * @returns The commands for the context menu or a promise for the commands.
         */
        supplyButtonCommands?: (lifetime: MsPortalFx.Base.LifetimeManager, row: Row<T>, grid: Contract<T>) => MsPortalFx.ViewModels.Toolbars.ToolbarItemContract[] | Q.Promise<MsPortalFx.ViewModels.Toolbars.ToolbarItemContract[]>;
    }
    /**
     * EditableGrid context menu contract.
     */
    export interface ContextMenuContract<T> {
        "--noUnusedLocals"?: any | T;
    }
    /**
     * EditableGrid options.
     */
    export interface Options<T> {
        /**
         * The accesibility label for the grid.
         */
        ariaLabel: string | KnockoutObservableBase<string>;
        /**
         * The column definitions.
         * This is required because it is expected that the existing columns will be known or obtained before creating the grid.
         */
        columns: ColumnDefinition<T>[] | KnockoutObservableArray<ColumnDefinition<T>>;
        /**
         * Context menu options.
         */
        contextMenu?: ContextMenuOptions<T>;
        /**
         * Callback to define properties used to initialize a new row.
         * If not supplied the row properties are initialized with default values.
         *
         * @param rowLifetime The lifetime manager for the row.
         * @param item The data item for the row.
         * @returns The row definition that is used to initialize a new row.
         */
        defineRow?: (rowLifetime: MsPortalFx.Base.LifetimeManager, item: T) => RowDefinition;
        /**
         * Disabled state.
         */
        disabled?: boolean;
        /**
         * Header definition.
         * Initializes the grid header.
         */
        header?: HeaderDefinition;
        /**
         * Focus options.
         * Focus is enabled by default.
         * Focus can not be disabled by omitting the focus options or setting to null.
         */
        focus?: FocusOptions<T>;
        /**
         * Selection options.
         * Selection is enabled by default.
         * Selection can not be disabled by omitting selection options or setting to null.
         */
        selection?: SelectionOptions<T>;
        /**
         * Toolbar options.
         * The toolbar is disabled by default.
         * The toolbar can be enabled by providing toolbar options.
         */
        toolbar?: ToolbarOptions<T>;
        /**
         * The array of items to be edited.
         * Changes will be made directly to this array as the user makes edits.
         * This is required because it is expected that the existing data is obtained before creating the grid.
         * If you do not have existing data you can specify an empty array.
         */
        items: T[] | KnockoutObservableArray<T>;
        /**
         * Addition options for adding new rows.
         * Addition is disabled by default.
         * Addition can be enabled by specifying addition options.
         */
        addition?: AdditionOptions<T>;
        /**
         * Deletion options for deleting rows.
         * Deletion is disabled by default.
         * Deletion can be enabled by specifying deletion options.
         */
        deletion?: DeletionOptions<T>;
        /**
         * Editing options for editing existing rows.
         * Editing is enabled by default.
         * Editing can not be disabled by omitting editing options or setting to null.
         * Providing a cell editor control on a column will enable editing of cells withing the column.
         * The canEditRow callback can be used to disable editing of particular rows.
         */
        editing?: EditingOptions<T>;
        /**
         * Insertion options for inserting a new row.
         * Insertion is disabled by default.
         * Insertion can be enabled by specifying insertion options.
         */
        insertion?: InsertionOptions<T>;
        /**
         * Reordering options for reordering rows.
         * Reordering is disabled by default.
         * Reordering can be enabled by specifying reordering options.
         */
        reordering?: ReorderingOptions<T>;
        /**
         * Resizing options.
         * Resizing is enabled by default.
         * Resizing can be disabled by setting to null.
         */
        resizing?: ResizingOptions<T>;
        /**
         * Validation options.
         * Vaidation is enabled by default.
         * Validation can not be disabled by omitting validation options or setting to null.
         */
        validation?: ValidationOptions<T>;
    }
    /**
     * EditableGrid contract.
     */
    export interface Contract<T> extends MsPortalFx.Control {
        /**
         * The accesibility label for the grid.
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * The columns of the grid.
         */
        readonly columns: ColumnCollection<T>;
        /**
         * Context menu properties and methods.
         */
        readonly contextMenu: ContextMenuContract<T>;
        /**
         * Gets or sets the EditableGrid dirty state.
         */
        readonly dirty: KnockoutObservableBase<boolean>;
        /**
         * Disabled state.
         */
        readonly disabled: KnockoutObservableBase<boolean>;
        /**
         * The grid rows.
         */
        readonly rows: KnockoutObservableArray<Row<T>>;
        /**
         * The grid header.
         */
        readonly header: Header;
        /**
         * EditableGrid focus properties and methods.
         */
        readonly focus: FocusContract<T>;
        /**
         * EditableGrid selection properties and methods.
         */
        readonly selection: SelectionContract<T>;
        /**
         * EditableGrid toolbar properties and methods.
         */
        readonly toolbar: ToolbarContract<T>;
        /**
         * The array of items to be edited.
         * Changes will be made directly to this array as the user makes edits.
         */
        readonly items: KnockoutObservableArray<T>;
        /**
         * EditableGrid grid row addition properties and methods.
         */
        readonly addition: AdditionContract<T>;
        /**
         * EditableGrid row deletion properties and methods.
         */
        readonly deletion: DeletionContract<T>;
        /**
         * EditableGrid row editing properties and methods.
         */
        readonly editing: EditingContract<T>;
        /**
         * Editable grid row insertion properties and methods.
         */
        readonly insertion: InsertionContract<T>;
        /**
         * EditableGrid row reordering properties and methods.
         */
        readonly reordering: ReorderingContract<T>;
        /**
         * EditableGrid row validation properties and methods.
         */
        readonly validation: ValidationContract<T>;
    }
    /**
     * Creates an EditableGrid control.
     *
     * @param lifetime The lifetime manager for the grid.
     * @param options The options for creating the grid.
     * @returns The created grid.
     */
    export function create<T>(lifetime: MsPortalFx.Base.LifetimeManager, options: Options<T>): Contract<T>;
}

// FILE: Fx\Controls\Essentials.d.ts
declare module "Fx/Controls/Essentials" {
    import { ClickableLink } from "Fx/Composition";
    import * as FxCompositionBlade from "Fx/Composition/Blade";
    import * as FxCompositionPdlBlade from "Fx/Composition/Pdl/Blade";
    import * as FxCompositionTemplateBlade from "Fx/Composition/TemplateBlade";
    import FxBase = MsPortalFx.Base;
    export type Container = FxCompositionBlade.Container | FxCompositionPdlBlade.Container | FxCompositionTemplateBlade.Container;
    /**
     * Built-In properties those can be obtained from resource data.
     */
    export const enum BuiltInType {
        /**
         * Built-in resource group property.  Allowed on left or right.
         */
        ResourceGroup = 1,
        /**
         * Built-in status property.  Allowed on left or right.
         */
        Status = 2,
        /**
         * Built-in location property.  Allowed on left or right.
         */
        Location = 3,
        /**
         * Built-in subscription name property.  Allowed on left or right.
         */
        SubscriptionName = 4,
        /**
         * Built-in subscription id property.  Allowed on left or right.
         */
        SubscriptionId = 5,
        /**
         * Built-in tags property.  Allowed on bottom.
         */
        Tags = 6,
    }
    /**
     * Icon position.
     * Determine which side of the text value is used for placing the icon.
     */
    export const enum IconPosition {
        /**
         * Place the icon to the left side of the text value.
         */
        Left = 1,
        /**
         * Place the icon to the right side of the text value.
         */
        Right = 2,
    }
    /**
     * The icon image and the position for the icon
     */
    export interface ImageAndPosition {
        /**
         * Show an icon on the left side of the pill.
         */
        image: FxBase.Image;
        /**
         * Icon position.
         * Determine which side of the text value is used for placing the icon.
         */
        position: IconPosition;
    }
    /**
     * The base blade close results interface provides the base for blade close results. The BuiltInType sent to the
     * onBladeClose will determine the actual shape of these results.
     */
    export interface BladeCloseResults {
    }
    /**
     * Base options.
     */
    export interface Options {
        /**
         * Expanded state of the essentials control.
         */
        expanded?: KnockoutObservableBase<boolean>;
        /**
         * This function is called when the expander is clicked.
         */
        onExpanderClick?: (isExpanded: boolean) => void;
        /**
         * Show all items without blocking them with "See More/See Less" button.
         */
        showAllItems?: boolean;
        /**
         * A flag to use responsive multi-column support based on the blade size.
         */
        responsiveColumns?: boolean;
    }
    /**
     * Layout options for a resource.
     */
    export interface ResourceLayoutOptions extends Options {
        /**
         * Resource id.
         */
        resourceId: string;
        /**
         * Optional resource instance for the resource id.
         */
        resource?: HubsExtension.Azure.Resource | HubsExtension.Azure.ResourceGroup | KnockoutObservableBase<HubsExtension.Azure.Resource | HubsExtension.Azure.ResourceGroup>;
        /**
         * Flag to control whether the framework should fetch the resource info automatically if the resource instance is not provided.
         */
        skipResourceFetching?: boolean;
        /**
         * Flag to include the tags for the resource.
         */
        includeTags?: boolean;
        /**
         * List of built-in types to hide the "change" button, move resource blade opening button, right next to the built-in type label.
         */
        hiddenChangeLink?: BuiltInType[];
        /**
         * This callback function is called when the resource blade is opened.
         */
        onBladeOpen?: (type: BuiltInType) => void;
        /**
         * This callback function is called when the resource blade is closed.
         */
        onBladeClose?: (type: BuiltInType, bladeCloseResults?: BladeCloseResults) => void;
    }
    /**
     * Default layout options for a resource.
     */
    export interface DefaultResourceLayoutOptions extends ResourceLayoutOptions {
        /**
         * List of built-in types to hide.
         */
        hiddenBuiltInTypes?: BuiltInType[];
        /**
         * Additional static properties on the left besides the usual 5 items.
         */
        additionalLeft?: (Item | MultiLineItem)[];
        /**
         * Additional static properties on the right.
         */
        additionalRight: (Item | MultiLineItem)[];
    }
    /**
     * Custom layout options for a resource.
     */
    export interface CustomResourceLayoutOptions extends ResourceLayoutOptions {
        /**
         * List of items to render on the left pane.
         */
        left: (Item | MultiLineItem | BuiltInType)[];
        /**
         * List of items to render on the right pane.
         */
        right: (Item | MultiLineItem | BuiltInType)[];
    }
    /**
     * Layout options for an arbitrary object type (not a resource).
     */
    export interface NonResourceLayoutOptions extends Options {
        /**
         * List of items to render on the left pane.
         */
        left: (Item | MultiLineItem)[];
        /**
         * List of items to render on the right pane.
         */
        right: (Item | MultiLineItem)[];
    }
    /**
     * Properties of an item.
     */
    export interface ItemProperty {
        /**
         * Item value to show.
         */
        value: string | KnockoutObservableBase<string>;
        /**
         * OnClick handler for the item.
         */
        onClick?: (() => void) | ClickableLink;
        /**
         * Displays an icon for the item.
         */
        icon?: ImageAndPosition | KnockoutObservableBase<ImageAndPosition>;
    }
    /**
     * Single-line item.
     */
    export interface Item extends ItemProperty {
        /**
         * Label for the item.
         */
        label: string;
    }
    /**
     * Multi-line item.
     */
    export interface MultiLineItem {
        /**
         * Label for the multi-line item.
         */
        label: string;
        /**
         * Arrays of item properties.
         * Multi-line item contains one label and multiple item properties.
         */
        lines: ItemProperty[];
    }
    /**
     * Essentials base contract.
     */
    export interface Contract extends MsPortalFx.Control {
        /**
         * Expanded state.
         */
        expanded: KnockoutObservableBase<boolean>;
        /**
         * Essentials disabled state.
         */
        disabled: KnockoutObservableBase<boolean>;
        /**
         * Add items dynamically to the essentials control.
         *
         * @param left An array of item to add.
         * @param right An array of item to add.
         */
        addDynamicProperties(left: (Item | MultiLineItem)[], right: (Item | MultiLineItem)[]): void;
    }
    /**
     * Essentials contract for a resource layout.
     */
    export interface ResourceLayoutContract extends Contract {
        /**
         * Resource instance.
         */
        resource: KnockoutObservableBase<HubsExtension.Azure.Resource | HubsExtension.Azure.ResourceGroup>;
        /**
         * Set "Status" property.
         *
         * @param status A string value to set the "Status" property value.
         */
        modifyStatus(status: string): void;
    }
    /**
     * Essentials contract for an arbitrary object layout.
     */
    export interface NonResourceLayoutContract extends Contract {
    }
    /**
     * Creates an Essentials control viewmodel.
     *
     * @param container A template blade container.
     * @param options The set of options to configure the essentials.
     *
     * @returns an essentials viewmodel
     */
    export function create(container: Container, options: DefaultResourceLayoutOptions): ResourceLayoutContract;
    /**
     * Creates an Essentials control viewmodel.
     *
     * @param container A template blade container.
     * @param options The set of options to configure the essentials.
     *
     * @returns an essentials viewmodel
     */
    export function create(container: Container, options: CustomResourceLayoutOptions): ResourceLayoutContract;
    /**
     * Creates an Essentials control viewmodel.
     *
     * @param container A template blade container.
     * @param options The set of options to configure the essentials.
     *
     * @returns an essentials viewmodel
     */
    export function create(container: Container, options: NonResourceLayoutOptions): NonResourceLayoutContract;
    /**
     * This API has been deprecated.
     * Please use the create Method to instantiate the control.
     */
    export const ViewModel: Obsolete;
}

// FILE: Fx\Controls\FileUpload.d.ts
declare module "Fx/Controls/FileUpload" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export import UploadContext = MsPortalFx.ViewModels.FileUpload.UploadContext;
    export import UploadType = MsPortalFx.ViewModels.FileUpload.UploadType;
    export import ContentType = MsPortalFx.ViewModels.FileUpload.ContentType;
    export import SelectedFileContract = MsPortalFx.ViewModels.FileUpload.SelectedFileContract;
    export import AsyncUploadTaskContract = MsPortalFx.ViewModels.FileUpload.AsyncUploadTaskContract;
    /**
     * Options for configuring a file upload control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * The text shown when no value is set in the control.
         */
        placeHolderText?: string | KnockoutObservableBase<string>;
        /**
         * A comma-separated list of allowed file mime-types, excluding extensions.
         * This maps directly to the HTML accept attribute for file input controls and filters the file based on specified mime-types in the file browse dialog.
         * The file filtering  behavior depends on native browser support.
         */
        accept?: string | KnockoutObservableBase<string>;
        /**
         * A comma-separated list of allowed file extensions.
         * This is upload control specific validations that will validate if the selected file is in the allowed file extension list.
         * If the selected file is not in the extension list, a validation error will be flagged and the selected file will be marked as invalid.
         * Eg: "pfx,cer"
         */
        allowedFileExtensions?: string | KnockoutObservableBase<string>;
        /**
         * The maximum number of files allowed to be uploaded at once.
         * This limit is applied post-selection.
         */
        maxFiles?: number;
        /**
         * The upload context options around how/where to upload and size limits.
         */
        uploadContext: UploadContext | KnockoutObservableBase<UploadContext>;
        /**
         * The user selected files via browse dialog is surfaced as SelectedFile in the selectedFiles array.
         * For scenarios where selected files should automatically be associated with an UploadTask, set this value to true.
         * If this option is set to false, the extension author is responsible for explicitly adding the selectedFile to uploadTasks by invoking
         * the addUploadTasks() method on view model to create an UploadTask representation and have it available in uploadTasks array.
         * By default, this option is set to true.
         */
        autoAddUploadTasks?: boolean;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Aria label for the file upload control.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
    }
    /**
     * FileUpload control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<string>;
        /**
         * Each time user clicks on the browse dialog, user can choose one or multiple files and each selected file is represented as SelectedFile on the extension iframe.
         * The selected files are surfaced in the selectedFiles array.
         * The array will be cleared if user chooses to click browse dialog to pick a different set of files.
         * Setting the value property null or calling resetUploadTasks() will clear the array.
         */
        readonly selectedFiles: KnockoutReadOnlyObservableArray<SelectedFileContract>;
        /**
         * Observable array holds all the current file upload tasks. Each task represent the file details, status, upload progress etc.
         * UploadTask is used to initiate the file upload, pause/resume uploads, cancel and dispose the file upload.
         */
        readonly uploadTasks: KnockoutReadOnlyObservableArray<AsyncUploadTaskContract>;
        /**
         * A comma-separated list of allowed file mime-types, excluding extensions.
         * This maps directly to the HTML accept attribute for file input controls and filters the file based on specified mime-types in the file browse dialog.
         * The file filtering  behavior depends on native browser support.
         */
        readonly accept: KnockoutObservableBase<string>;
        /**
         * A comma-separated list of allowed file extensions.
         * This is upload control specific validations that will validate if the selected file is in the allowed file extension list.
         * If the selected file is not in the extension list, a validation error will be flagged and the selected file will be marked as invalid.
         * Eg: "pfx,cer"
         */
        readonly allowedFileExtensions: KnockoutObservableBase<string>;
        /**
         * The upload context options around how/where to upload and size limits.
         */
        readonly uploadContext: KnockoutObservableBase<UploadContext>;
        /**
         * The text shown when no value is set in the control.
         */
        readonly placeHolderText: KnockoutObservableBase<string>;
    }
    /**
     * Creates a FileUpload control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a file upload control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options: Options): Contract;
}

// FILE: Fx\Controls\FormBase.d.ts
declare module "Fx/Controls/FormBase" {
    import { Validation, ValidationResult } from "Fx/Controls/Validations";
    /**
     * Base Properties on all form controls
     */
    export interface Contract extends MsPortalFx.Control {
        /**
         * The dirty state of the control.  This will only change when the user performs an action to change the value, or the dirty state is set explicitly by the extension.
         */
        readonly dirty: KnockoutObservableBase<boolean>;
        /**
         * The validation state of the control.
         */
        readonly valid: KnockoutReadOnlyObservableBase<boolean>;
        /**
         * The visible state of the control.
         */
        readonly visible: KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        readonly cssClass: KnockoutObservableBase<string>;
        /**
         * The label of the control
         */
        readonly label: KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        readonly subLabel: KnockoutObservableBase<string>;
        /**
         * The aria-label on the control.
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        readonly validations: KnockoutObservableArray<Validation>;
        /**
         * Latest validation results.
         */
        readonly validationResults: KnockoutReadOnlyObservableArray<ValidationResult>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        readonly infoBalloonContent: KnockoutObservableBase<string>;
        /**
         * The disabled state of the control.
         */
        readonly disabled: KnockoutObservableBase<boolean>;
        /**
         * Triggers validation on this control when called.  Returns a promise that resolves once validation on this control has completed.
         * Promise return a value of true (valid) or false (invalid)
         */
        readonly triggerValidation: (focusOnFirstInvalid?: boolean, validateHidden?: boolean) => Q.Promise<boolean>;
    }
    export const enum ValueUpdateTrigger {
        /**
         * (default) Trigger value update when the control widget value changes
         */
        Input = 4,
        /**
         * Trigger value update on blur.
         */
        Blur = 3,
    }
    export class DateTimeRange {
        /**
         * Start date/time in local timezone.
         */
        startDateTime: KnockoutObservable<Date>;
        /**
         * End date/time in local timezone.
         */
        endDateTime: KnockoutObservable<Date>;
        /**
         * Obsolete. This is no longer available and the datetime values are always in the local timezone.
         */
        timezoneOffset: Obsolete;
        /**
         * Creates a DateTimeRange object
         *
         * @param startDateTime The start date/time in local timezone
         * @param endDateTime The end date/time in local timezone
         */
        constructor(startDateTime?: Date, endDateTime?: Date);
    }
}

// FILE: Fx\Controls\GridBase.d.ts
declare module "Fx/Controls/GridBase" {
    /**
     * Grid selection mode.
     */
    export const enum SelectionMode {
        /**
         * Select a single row at a time.
         */
        Single = 0,
        /**
         * Select a multiple rows at a time.
         */
        Multiple = 1,
    }
}

// FILE: Fx\Controls\HotSpot.d.ts
declare module "Fx/Controls/HotSpot" {
    import Selectable = require("Fx/Composition/Selectable");
    export = Main;
    module Main {
        import BaseViewModel = MsPortalFx.ViewModels.Controls.HotSpot.BaseViewModel;
        /**
         * Hotspot options.
         */
        interface ViewModelOptions {
            /**
             * A function to be called when the hotspot is clicked.
             */
            onClick?: () => void;
            /**
             * Whether to activate the hotspot during customize or not.
             */
            clickableDuringCustomize?: boolean;
            /**
             * Disabled state of the hotspot.
             */
            disabled?: KnockoutObservableBase<boolean>;
        }
        /**
         * View model for hotspots.
         */
        class ViewModel extends BaseViewModel {
            private _msPortalFxOnClick;
            /**
             * Creates a HotSpot ViewModel.
             */
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, options?: ViewModelOptions);
        }
        class ViewModel2 extends FxImpl.Controls.HotSpot.ViewModel2 {
            /**
             * Creates a HotSpot ViewModel.
             */
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, selectableOptions?: Selectable.Selectable2Options<any>, clickableDuringCustomize?: boolean);
        }
    }
}

// FILE: Fx\Controls\InfoBalloon.d.ts
declare module "Fx/Controls/InfoBalloon" {
    export interface Options {
        /**
         * Html to be shown in the balloon.
         */
        content?: KnockoutObservableBase<string> | string;
        /**
         * Desired aria-label value of the anchor. If 'content' member field is too long, please
         * make use of this field; make it a shorter summary of the full content.
         */
        ariaLabel?: KnockoutObservableBase<string> | string;
    }
    /**
     * InfoBalloon control view model
     */
    export interface Contract extends MsPortalFx.Control {
        /**
         * Html to be shown in the balloon.
         */
        readonly content: KnockoutObservableBase<string>;
        /**
         * Desired aria-label value of the anchor. If 'content' member field is too long, please
         * make use of this field; make it a shorter summary of the full content.
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
    }
    /**
     * Creates an InfoBalloon control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns an InfoBalloon viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\InfoBox.d.ts
declare module "Fx/Controls/InfoBox" {
    /**
     * The size options for info box
     */
    export const enum Size {
        /**
         * Small sized info box
         */
        Small = 0,
        /**
         * Normal sized info box
         */
        Normal = 1,
    }
    export interface Options {
        /**
         * The disabled state of the control
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The image to display in the control.
         */
        image?: MsPortalFx.Base.Image | KnockoutObservableBase<MsPortalFx.Base.Image>;
        /**
         * defines click behavior.
         * If a clickable link is provided, then clicking the info box will open that link.
         * If a function is specified, clicking the infobox will trigger a callback to the function.
         */
        onClick?: MsPortalFx.ViewModels.ClickableLink | Function;
        /**
         * Size of the infobox
         */
        size?: Size | KnockoutObservableBase<Size>;
        /**
         * The text or html to display in the control.
         */
        text?: string | KnockoutObservableBase<string>;
        /**
         * Flag indicating if the info box can be closed.
         */
        canClose?: boolean;
        /**
         * Defines the action performed once the infobox is dismissed.
         */
        onCloseClickCallback?: Function;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
    }
    export interface Contract extends MsPortalFx.Control {
        /**
         * The disabled state of the control
         */
        readonly disabled: KnockoutObservableBase<boolean>;
        /**
         * The image to display in the control.
         */
        readonly image: KnockoutObservableBase<MsPortalFx.Base.Image>;
        /**
         * Size of the infobox
         */
        readonly size: KnockoutObservableBase<Size>;
        /**
         * The text or html to display in the control.
         */
        readonly text: KnockoutObservableBase<string>;
    }
    /**
     * Creates an InfoBox control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns an InfoBox viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\Legend.d.ts
declare module "Fx/Controls/Legend" {
    export = Main;
    module Main {
        /**
         * Defines hatching patterns.
         */
        enum Pattern {
            /**
             * The area is solid.
             */
            Solid = 0,
            /**
             * The area is cross hatched.
             */
            CrossHatching = 1,
            /**
             * The area is diagonal hatched.
             */
            DiagonalHatching = 2,
            /**
             * The area is hatched horizontally.
             */
            DottedHatching = 3,
        }
        /**
         * Options for Legend ViewModel.
         */
        interface Options {
            /**
             * Data being displayed.
             */
            data?: KnockoutObservableArray<any>;
            /**
             * Data key used to identify the row index. Leave unset to use the index within data.
             */
            indexKey?: string;
            /**
             * Data key used to identify the color.
             */
            colorKey: string;
            /**
             * Data key of the hatching pattern.
             */
            hatchingKey?: string;
            /**
             * Data key used to identify the boolean column indicating if the row is selected.
             */
            selectedKey?: string;
            /**
             * Data key used to label the row for display purpose. Data has to be string, we use to show center caption for quick indication of the item.
             */
            labelKey: string;
            /**
             * Index of current hovered row in the legend.
             */
            hoveredIndex?: KnockoutObservableBase<string>;
            /**
             * Aria label for the legend.
             */
            ariaLabel?: string | KnockoutObservableBase<string>;
        }
        class ViewModel extends MsPortalFx.ViewModels.Controls.Base.ViewModel {
            /**
             * Data being displayed.
             */
            data: KnockoutObservableArray<any>;
            /**
             * Data key used to identify the row index. Leave unset to use the index within data.
             */
            indexKey: string;
            /**
             * Data key used to identify the color.
             */
            colorKey: string;
            /**
             * Data key of the hatching pattern.
             */
            hatchingKey: string;
            /**
             * Data key used to identify the boolean column indicating if the row is selected.
             */
            selectedKey: string;
            /**
             * Data key used to label the row for display purpose. Data has to be string, we use to show center caption for quick indication of the item.
             */
            labelKey: string;
            /**
             * Index of current hovered row in the legend.
             */
            hoveredIndex: KnockoutObservableBase<string>;
            /**
             * Aria label for the legend.
             */
            readonly ariaLabel: KnockoutObservableBase<string>;
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, options?: Options);
        }
    }
}

// FILE: Fx\Controls\LocationDropDown.d.ts
declare module "Fx/Controls/LocationDropDown" {
    import { AnyBladeContainer } from "Fx/Composition/BladeBase";
    import { Location, Validation as BaseValidation } from "Fx/Controls/BaseResourceDropDown";
    export = Main;
    module Main {
        /**
         * The contract for the values returned by the location dropdown
         */
        export import Location = MsPortalFx.Azure.Location;
        /**
         * The validation type accepted by the dropdown
         */
        type Validation = BaseValidation<Location>;
        /**
         * The contract for options to create the location drop down
         */
        type Options = Location.Options;
        /**
         * The contract for options to create the location drop down
         */
        type CustomOptions = Location.CustomOptions;
        /**
         * The contract for the location dropdown
         */
        interface Contract extends Location.Contract {
        }
        /**
         * This creates a location dropdown control.
         * @param container The container associated with the part or blade.
         * @param options The location dropdown options. See interface
         */
        function create(container: AnyBladeContainer, options?: Options): Contract;
        /**
         * This creates a location dropdown control with the ability to add additional locations.
         * @param container The container associated with the part or blade.
         * @param options The location dropdown options. See interface
         */
        function createCustom(container: AnyBladeContainer, options?: CustomOptions): Contract;
    }
}

// FILE: Fx\Controls\LogStream.d.ts
declare module "Fx/Controls/LogStream" {
    export const enum LogItemType {
        /**
         * Indicates general information.
         */
        Text = 0,
        /**
         * Indicates important information.
         */
        Information = 1,
        /**
         * Indicates success information
         */
        Success = 2,
        /**
         * Indicates warning information.
         */
        Warning = 3,
        /**
         * Indicates error information.
         */
        Error = 4,
    }
    export interface LogItem {
        /**
         * The text to log.
         */
        text: string;
        /**
         * The type of text to log.
         */
        type?: LogItemType;
    }
    export interface Options {
        /**
         * Alternate text for screen reader, label for this log-stream
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * Controls display updating.
         */
        paused?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Controls filtering of messages in the log stream.
         */
        filters?: LogItemType[] | KnockoutObservableArray<LogItemType>;
    }
    export interface Contract extends MsPortalFx.Control {
        /**
         * Alternate text for screen reader, label for this log-stream
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * Controls display updating.  When set to true, the log stream will buffer messages and will show them when unpaused.
         */
        readonly paused: KnockoutObservableBase<boolean>;
        /**
         * Controls filtering of messages in the log stream.
         */
        readonly filters: KnockoutObservableArray<LogItemType>;
        /**
         * Clears the log stream control.
         */
        reset(): void;
        /**
         * Writes an entry to the log
         * @param entry The text to show in the control.  Passing in a LogItem or array of LogItems allows you to control the level of entry (error, warning, etc.)
         */
        log(entry: string | string[] | LogItem | LogItem[]): void;
    }
    /**
     * Creates an LogStream control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns an LogStream viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\Markdown.d.ts
declare module "Fx/Controls/Markdown" {
    export interface Options {
        /**
         * markdown Input string containing markdown to be converted to html and displayed.
         */
        markdown?: string | KnockoutObservableBase<string>;
        /**
         * Alternate text for screen reader to describe this article shortly
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
    }
    export interface Contract extends MsPortalFx.Control {
        /**
         * markdown Input string containing markdown to be converted to html and displayed.
         */
        readonly markdown: KnockoutObservableBase<string>;
        /**
         * Alternate text for screen reader to describe this article shortly
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
    }
    /**
     * Creates an Markdown control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a Markdown viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\Menu.d.ts
declare module "Fx/Controls/Menu" {
    export = Main;
    module Main {
        import FxViewModels = MsPortalFx.ViewModels;
        import FxControls = FxViewModels.Controls;
        /**
         * Attributes common to all items and groups in the menu.
         */
        interface MenuItemBase extends FxImpl.Controls.Menu.MenuItemBase {
        }
        /**
         * Defines an item in a group of the menu.
         */
        interface MenuItem extends FxImpl.Controls.Menu.MenuItem {
        }
        /**
         * Defines a group in the menu.
         */
        interface MenuGroup extends FxImpl.Controls.Menu.MenuGroup {
        }
        /**
         * The options for the menu.
         */
        interface Options extends FxImpl.Controls.Menu.Options {
        }
        /**
         * The sort order used for groups in the menu.
         */
        export import SortOrder = MsPortalFx.ViewModels.Controls.Lists.Grid.SortOrder;
        /**
         * The sort functions used to sort groups in the menu.
         */
        enum SortFunction {
            /**
             * Uses the built-in Array.sort() function.
             */
            Default = 0,
            /**
             * Uses the getTime() value of Date object to sort.
             */
            DateTime = 1,
            /**
             * Uses custom sorting, that ensures that the support group is prioritized over others.
             */
            SupportFirst = 9999,
        }
        /**
         * Defines the view model for the menu control.
         */
        class ViewModel extends FxControls.Base.ViewModel {
            /**
             * The groups of menu items.
             */
            groups: KnockoutObservableArray<MenuGroup>;
            /**
             * Specifies the sort order used for groups in the menu.
             */
            groupSortOrder: SortOrder;
            /**
             * Specifies the sort function used to order groups in the menu.
             */
            groupSortFunction: SortFunction;
            /**
             * A value indicating whether or not to show the search box.
             */
            showSearch: boolean;
            private _msPortalFxData;
            constructor(lifetime: MsPortalFx.Base.LifetimeManager, groupSortFunction?: SortFunction);
            /**
             * Sets the options for the menu.
             *
             * @param options The options to configure the menu.
             */
            setOptions: (options: Options) => void;
        }
    }
}

// FILE: Fx\Controls\MonitorChart.d.ts
declare module "Fx/Controls/MonitorChart" {
    import * as FxCompositionBlade from "Fx/Composition/Blade";
    import * as FxCompositionPart from "Fx/Composition/Part";
    import * as FxCompositionPdlBlade from "Fx/Composition/Pdl/Blade";
    import * as FxCompositionTemplate from "Fx/Composition/TemplateBlade";
    import BladeContainer = FxCompositionBlade.Container;
    import PdlBladeContainer = FxCompositionPdlBlade.Container;
    import PartContainer = FxCompositionPart.Container;
    import TemplateContainer = FxCompositionTemplate.Container;
    /**
     * Defines how data points are aggregated for a metric.
     */
    export const enum AggregationType {
        /**
         * No aggregation is done.
         */
        None = 0,
        /**
         * Data points are aggregated by taking the average of their values.
         */
        Average = 1,
        /**
         * Data points are aggregated by taking the min of their values.
         */
        Minimum = 2,
        /**
         * Data points are aggregated by taking the max of their values.
         */
        Maximum = 3,
        /**
         * Data points are aggregated by taking the total of their values.
         */
        Total = 4,
    }
    /**
     * Defines what visualization to use when rendering the chart.
     */
    export const enum ChartType {
        /**
         * Line chart.
         */
        Line = 0,
        /**
         * Bar chart.
         */
        Bar = 1,
    }
    /**
     * Defines the unit of a metric.
     */
    export const enum Unit {
        /**
         * Count unit.
         */
        Count = 0,
        /**
         * Bytes unit.
         */
        Bytes = 1,
        /**
         * Seconds unit.
         */
        Seconds = 2,
        /**
         * CountPerSecond unit.
         */
        CountPerSecond = 3,
        /**
         * BytesPerSecond unit.
         */
        BytesPerSecond = 4,
        /**
         * Percent unit.
         */
        Percent = 5,
        /**
         * MilliSeconds unit.
         */
        MilliSeconds = 6,
        /**
         * ByteSeconds unit.
         */
        ByteSeconds = 7,
    }
    /**
     * Defines the timespan over which data points are fetched and plotted.
     */
    export interface Timespan {
        /**
         * A relative timespan indicates that data points are plotted for a moving timespan, whose
         * end time is always now().
         */
        relative?: {
            /**
             * Length of time over which metrics are plotted.
             */
            durationMs: number;
        };
        /**
         * An absolute timespan indicates that data points are plotted for a fixed timespan.
         */
        absolute?: {
            /**
             * The start time of the timespan.
             */
            start: Date;
            /**
             * The end time of the timespan.
             */
            end: Date;
        };
        /**
         * How frequently the data points shown on the chart should be updated.
         */
        refreshIntervalMs?: number;
    }
    export interface ResourceMetadata {
        /**
         * The resource id of the resource.
         */
        resourceId: string;
        /**
         * The kind of the resource.
         */
        kind?: string;
        /**
         * The sku of the resource.
         */
        sku?: {
            /**
             * The sku name.
             */
            name?: string;
            /**
             * The sku tier.
             */
            tier?: string;
            /**
             * The sku size.
             */
            size?: string;
            /**
             * The sku family.
             */
            family?: string;
            /**
             * The sku model.
             */
            model?: string;
            /**
             * The sku capacity.
             */
            capacity?: string;
        };
        /**
         * Additional properties for the resource.
         */
        properties?: any;
    }
    export interface Metric {
        /**
         * Resource information about the resource to which this metric belongs.
         */
        resourceMetadata: ResourceMetadata;
        /**
         * The non-localized metric name.
         */
        name: string;
        /**
         * The aggregation type to use for this metric.
         */
        aggregationType?: AggregationType;
        /**
         * The time grain to use for this metric.
         */
        timeGrainMs?: number;
        /**
         * The unit of this metric.
         */
        unit?: Unit;
        /**
         * The type of this metric.
         *
         * This propery should be used when dealing with resource types that have subcontainers (e.g. Storage Accounts or Virtual Machines).
         */
        type?: string;
        /**
         * Css class to apply to this metric series.
         *
         * This css class should be one of the built in msportalfx-bgcolor-* colors.
         *
         * @see https://github.com/Azure/portaldocs/blob/334634c8a7601f729c4ee6fa673eb00d25045c60/portal-sdk/generated/portalfx-style-guide-color-palette.md#coloring-to-differentiate-data
         * @see https://df.onecloud.azure-test.net/?SamplesExtension=true#blade/SamplesExtension/SDKMenuBlade/styleguidecolorpalettetitle
         */
        cssClass?: string;
    }
    export interface ChartDefinition {
        /**
         * The metrics to plot on this chart.
         */
        metrics: Metric[];
        /**
         * The visualization to use for this chart.
         *
         * Defaults to line chart.
         */
        chartType?: ChartType;
        /**
         * The timespan to use for this chart.
         *
         * Note: This overrides the top-level timespan provided in the Options object.
         */
        timespan?: Timespan;
        /**
         * The title of this chart.
         *
         * Defaults to a string generated from the metrics plotted on this chart.
         */
        title?: string;
        /**
         * The subtitle of this chart.
         *
         * Defaults to a string generated from the resources plotted on this chart.
         */
        subtitle?: string;
        /**
         * Message to display if no data can be loaded for this chart.
         */
        noDataMessage?: string;
        /**
         * Disables pinning for this chart.
         */
        disablePinning?: boolean;
        /**
         * Callback that's executed when the user clicks this chart while
         * the chart is functioning normally.
         *
         * Defaults to opening the Metrics blade.
         */
        onClick?: () => void;
        /**
         * Css class to apply to this chart.
         *
         * Multiple css classes can be added by separating them with a space.
         */
        cssClass?: string;
        /**
         * Aria label for the chart.
         *
         * Defaults to "Resource metrics" if not set.
         */
        ariaLabel?: string;
    }
    /**
     * MonitorChart options.
     */
    export interface Options {
        /**
         * The charts to render.
         */
        charts?: ChartDefinition[] | KnockoutObservableArray<ChartDefinition>;
        /**
         * The timespan used for all charts, unless overriden in an individual chart.
         *
         * Defaults to relative duration of 24 hours.
         */
        timespan?: Timespan | KnockoutObservableBase<Timespan>;
    }
    /**
     * MonitorChart control.
     */
    export interface Contract {
        /**
         * The charts rendered.
         */
        readonly charts: KnockoutObservableArray<ChartDefinition>;
        /**
         * The timespan used for all charts, unless overriden in an individual chart.
         */
        readonly timespan: KnockoutObservableBase<Timespan>;
    }
    /**
     * Creates a MonitorChart viewmodel.
     *
     * @param lifetime The lifetime of the control. This must be a container object.
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a MonitorChart viewmodel.
     */
    export function create(container: BladeContainer | PartContainer | PdlBladeContainer | TemplateContainer, options?: Options): Contract;
    export module Internal {
        interface BladeManagementApiContainer extends MsPortalFx.ViewModels.BladeManagement, MsPortalFx.Base.LifetimeManager {
        }
        /**
         * Internal viewmodel used for monitor charts.
         */
        class ViewModel extends MsPortalFx.ViewModels.Controls.Loadable.ViewModel implements Contract {
            /**
             * Charts property that stores the charts array provided by the consumer of this control and
             * is **not** proxied across to the widget.
             *
             * This is to prevent the Javascript array provided by the consumer from being frozen
             * by the PO layer.
             */
            charts: KnockoutObservableArray<ChartDefinition>;
            /**
             * Charts property that mirrors the charts property and **is proxied** to the widget.
             */
            _msPortalFxCharts: KnockoutObservableBase<ChartDefinition[]>;
            /**
             * Timespan property that stores the timespan provided by the consumer of this control and
             * is **not** proxied across to the widget.
             *
             * This is to prevent the Javascript array provided by the consumer from being frozen
             * by the PO layer.
             */
            timespan: KnockoutObservableBase<Timespan>;
            /**
             * Timespan property that mirrors the timespan property and **is proxied** to the widget.
             */
            _msPortalFxTimespan: KnockoutObservableBase<Timespan>;
            private _container;
            private _traceId;
            private _useMonitorChartV2;
            constructor(container: BladeManagementApiContainer, options?: Options);
            widgetLoaded(bladeOrPartName: string): void;
            /**
             * Allows the MonitorChart widget to open a blade when it's clicked.
             *
             * This openBlade api exists in the ViewModel because the widget does not have the container context on
             * which this control resides (which is important so the shell knows whether to open a blade in a journey
             * or to start a new journey).
             */
            openBlade(bladeReference: FxImpl.Composition.Selectable.BladeReference<any>, options?: MsPortalFx.ViewModels.OpenBladeOptions): Q.Promise<boolean>;
        }
    }
}

// FILE: Fx\Controls\MonitorChartV2.d.ts
declare module "Fx/Controls/MonitorChartV2" {
    /**
     *  Describes the ARM resource whose metric is being plotted on the chart.
     */
    export interface ResourceMetadata {
        /**
         * The ARM resource id of the resource.
         */
        id: string;
        /**
         * The resource kind of the resource.
         * This is optional and makes fetching the resource faster.
         * When this is not specified it is fetched from ARM and populated.
         */
        kind?: string;
        /**
         * The sku of the resource.
         * These optional features makes fetching the resource faster.
         * When this is not specified it is fetched from ARM and populated.
         */
        sku?: {
            /**
             * The sku name.
             */
            name?: string;
            /**
             * The sku tier.
             */
            tier?: string;
            /**
             * The sku size.
             */
            size?: string;
            /**
             * The sku family.
             */
            family?: string;
            /**
             * The sku model.
             */
            model?: string;
            /**
             * The sku capacity.
             */
            capacity?: string;
        };
    }
    /**
     * This namespace contains types related to a threshold line to be plotted on the metric on chart.
     */
    export namespace Threshold {
        /**
         * Describes the properties of a threshold line on the metric plotted on chart.
         */
        interface Options {
            /**
             * The identifier for the threshold.
             * Each threshold on a metric has a unique identifier.
             * The id can also be used to identify and access a threshold from `Metric.Threshold` array.
             * If not provided a default one is created.
             */
            id?: string;
            /**
             * The upper bound of the threshold line. Defaults to 0.
             */
            upperThreshold?: number | KnockoutObservableBase<number>;
            /**
             * The lower bound of the threshold. Defaults to 0.
             */
            lowerThreshold?: number | KnockoutObservableBase<number>;
        }
        /**
         * Returned from `MonitorChartV2.Threshold.create`. Represents a threshold line for a metric plotted on the chart.
         */
        interface Contract {
            /**
             * The identifier for the threshold.
             * Each threshold on a metric has a unique identifier.
             * The id can also be used to identify and access a threshold from `Metric.Threshold` array.
             * If not provided a default one is created.
             */
            readonly id: string;
            /**
             * The upper bound of the threshold line.
             */
            readonly upperThreshold: KnockoutObservableBase<number>;
            /**
             * The lower bound of the threshold line.
             */
            readonly lowerThreshold: KnockoutObservableBase<number>;
        }
        /**
         * Creates a new `Threshold.Contract` from the provided `Threshold.Options`.
         * @param lifetime The lifetime object.
         * @param options The `Threshold.Options` object.
         * @returns A `Threshold.Contract` object.
         */
        function create(lifetime: MsPortalFx.Base.LifetimeManager, options: Options): Contract;
    }
    /**
     * This namespace contains types related to a metric being plotted on the chart
     */
    export namespace Metric {
        /**
         * Aggregation to use for the metric values
         */
        const enum AggregationType {
            /**
             * No aggregation. Invalid value.
             */
            None = 0,
            /**
             * Sum of metric values per bucket.
             */
            Sum = 1,
            /**
             * Minimum of metric values per bucket.
             */
            Min = 2,
            /**
             * Maximum of metric values per bucket.
             */
            Max = 3,
            /**
             * Average of metric values per bucket.
             */
            Avg = 4,
            /**
             * Unique count of metric values per bucket.
             */
            Unique = 5,
            /**
             * 90th Percentile of metric values per bucket.
             */
            Percentile = 6,
            /**
             * Count of metric values per bucket.
             */
            Count = 7,
        }
        /**
         * Visualization options related to a metric on the chart
         */
        interface Visualization {
            /**
             * The localized metric display name to be displayed on the chart.
             * When not specified the metric id is used to display on the chart.
             */
            displayName?: string;
            /**
             * The localized resource display name to be displayed on the chart
             * When not specified the ARM resource name is displayed on the chart.
             */
            resourceDisplayName?: string;
            /**
             * Color of the metric when plotted on the chart, in hexadecimal format (e.g. #c7f1c7).
             * NOTE: Make sure the colors work in dark and light themes.
             * When unspecified, chart chooses a random color.
             */
            color?: string;
        }
        /**
         * Options for creating a metric.
         */
        interface Options {
            /**
             * Information that identifies the resource to which the metric belongs.
             */
            resourceMetadata: ResourceMetadata;
            /**
             * The name of the metric.
             */
            name: string;
            /**
             * The aggregation type to use for this metric.
             */
            aggregationType: AggregationType | KnockoutObservableBase<AggregationType>;
            /**
             * Additional information to use when determining the correct metrics provider.
             * For example, VM providers need to use this, because they can have multiple providers for
             * one VM resource (i.e. Host and Guest metrics).
             * This is only needed for certain providers and not all
             * When not specified, chart uses a fixed default for each metric type.
             */
            namespace?: string;
            /**
             * Visualization options for the metric.
             * Defaults to a fixed visualization options on all metrics.
             */
            metricVisualization?: Visualization | KnockoutObservableBase<Visualization>;
            /**
             * Threshold options for the metric.
             * This is used to place threshold lines on the chart for the metric.
             * When not specified, there are no threshold lines placed on the chart.
             */
            thresholds?: Threshold.Options[];
        }
        /**
         * Returned from `MonitorChartV2.Metric.create` call. Contract object for the metric.
         */
        interface Contract {
            /**
             * Name of the metric. This can be used to uniquely identify the metric, inside a chart's Metric array.
             */
            readonly name: string;
            /**
             * Visualization options for the metric.
             */
            readonly metricVisualization: KnockoutObservableBase<Visualization>;
            /**
             * The aggregation type to use for this metric.
             */
            readonly aggregationType: KnockoutObservableBase<AggregationType>;
            /**
             * Threshold options for the metric. This is used to place threshold lines on the chart
             * for the metric
             */
            readonly thresholds: KnockoutObservableArray<Threshold.Contract>;
        }
        /**
         * Creates a `Metric.Contract` for the `Metric.Options` provided by callers.
         * Use this `Metric.Contract` object to later manipulate the `MonitorChartV2.charts().metrics` array.
         *
         * @param lifetime The lifetime object. Use the lifetime associated with MonitorChart control here.
         * @param options The `Metric.Options` object for the metric.
         * @returns A `Metric.Contract` object.
         */
        function create(lifetime: MsPortalFx.Base.LifetimeManager, options: Options): Contract;
    }
    /**
     * Describes the grouping/segmentation of the metric on the chart.
     */
    export interface Grouping {
        /**
         * The dimension to group against.
         */
        dimension: string;
    }
    /**
     * A collection of all the filters on the chart.
     */
    export interface FilterCollection {
        /**
         * Array containing all the Filters applied to the chart.
         */
        filters: Filter[];
    }
    /**
     * Describes an individual filter which can be added to the FilterCollection of the chart.
     */
    export interface Filter {
        /**
         * The dimension key for the filter.
         */
        key: string;
        /**
         * The set of values for the key we want to filter against.
         */
        values: string[];
    }
    /**
     * Time range over which data points are fetched and plotted.
     */
    export interface TimeRange {
        /**
         * The start of the time range.
         */
        startTime: Date;
        /**
         * The end of the time range.
         */
        endTime: Date;
    }
    /**
     * Supported grains that can be applied on a timespan for the chart.
     */
    export const enum Grain {
        /**
         * Chart chooses an automatic grain to display.
         */
        Automatic = 1,
        /**
         * Grain to set to per minute over the given time range.
         */
        Minutely = 2,
        /**
         * Grain to set to per hour over the given time range.
         */
        Hourly = 3,
        /**
         * Grain to set to per day over the given time range.
         */
        Daily = 4,
        /**
         * Grain to set to per week over the given time range.
         */
        Weekly = 5,
        /**
         * Grain to set to per month over the given time range.
         */
        Monthly = 6,
        /**
         * Grain to set to every 5 minutes over the given time range.
         */
        Every_5_Minutes = 7,
        /**
         * Grain to set to every 15 minutes over the given time range.
         */
        Every_15_Minutes = 8,
        /**
         * Grain to set to every 30 minutes over the given time range.
         */
        Every_30_Minutes = 9,
        /**
         * Grain to set to every 6 hours over the given time range.
         */
        Every_6_Hours = 10,
        /**
         * Grain to set to every 12 hours over the given time range.
         */
        Every_12_Hours = 11,
    }
    /**
     * Timespan signifies the absolute or relative time range for which the chart displays the data.
     */
    export interface Timespan {
        /**
         * The absolute time range of the chart's timespan.
         * If this is not present relative duration is used.
         * If neither of absolute or relative is specified, chart defaults to past 24 hours time range.
         */
        absolute?: TimeRange;
        /**
         * The related duration for the chart's timespan.
         * If this is not present absolute duration is used.
         * If neither of absolute or relative is specified, chart defaults to past 24 hours time range.
         */
        relative?: {
            duration: number;
        };
        /**
         * The grain of the time context.
         * Defaults to a fixed grain with respect to the timespan.
         */
        grain?: Grain;
        /**
         * Set to true to display time as UTC on the chart
         * Defaults to using local time.
         */
        showUTCTime?: boolean;
    }
    /**
     * Represents the visual elements of a chart.
     */
    export interface Visualization {
        /**
         * The type of chart to be used. Defaults to a line chart.
         */
        chartType?: ChartType | KnockoutObservableBase<ChartType>;
        /**
         * Options related to how the legend should be displayed.
         * Defaults to a fixed legend visualization with legends on bottom.
         */
        legendVisualization?: LegendVisualization | KnockoutObservableBase<LegendVisualization>;
        /**
         * Options related to how the axis should be displayed.
         * Defaults to a fixed axis visualization.
         */
        axisVisualization?: AxisVisualization | KnockoutObservableBase<AxisVisualization>;
        /**
         * Indicates whether users can pin this chart to their dashboard.
         * Defaults to false and pin button is always shown.
         */
        disablePinning?: boolean;
    }
    /**
     * List of supported chart types.
     */
    export const enum ChartType {
        /**
         * Bar Chart.
         */
        Bar = 1,
        /**
         * Line Chart.
         */
        Line = 2,
        /**
         * Area Chart.
         */
        Area = 3,
        /**
         * Scatter plot chart.
         */
        Scatter = 4,
    }
    /**
     * Represents how to display a chart's axis.
     */
    export interface AxisVisualization {
        /**
         * x-axis visualization options.
         */
        x: IndividualAxisVisualization;
        /**
         * y-axis visualization options.
         */
        y: IndividualAxisVisualization;
    }
    /**
     * Visualization options for a particular axis.
     */
    export interface IndividualAxisVisualization {
        /**
         * Determines if the axis is visible on the chart or hidden.
         */
        isVisible: boolean;
        /**
         * Defines the smallest value shown for the axis.
         * Defaults to '0' for numeric axis and start timespan for date axis.
         */
        min?: number;
        /**
         * Defines the largest value shown for the axis.
         * Defaults to maximum data value of the plotted data for numeric axis and end timespan for date axis.
         */
        max?: number;
    }
    /**
     * List of relative positions to place legend.
     */
    export const enum LegendPosition {
        /**
         * Legends placed on top of chart.
         */
        Top = 1,
        /**
         * Legends placed on bottom of chart.
         */
        Bottom = 2,
        /**
         * Legends placed on left of chart.
         */
        Left = 3,
        /**
         * Legends placed on right of chart.
         */
        Right = 4,
    }
    /**
     * Represents how to style a chart's legend.
     */
    export interface LegendVisualization {
        /**
         * Controls if the legend is visible or hidden on the chart.
         */
        isVisible: boolean;
        /**
         * Controls the position of legend with respect to the chart.
         * Default is to position on bottom of chart.
         */
        position?: LegendPosition;
    }
    /**
     * MonitorChartV2 options.
     */
    export interface Options {
        /**
         * The list of metrics to plot on this chart.
         */
        metrics: Metric.Options[];
        /**
         * The filters to apply on the chart for all metrics.
         * If not specified, data isn't filtered.
         */
        filterCollection?: FilterCollection | KnockoutObservableBase<FilterCollection>;
        /**
         * Specify the grouping to segment the chart against.
         * If not specified chart is not segmented.
         */
        grouping?: Grouping | KnockoutObservableBase<Grouping>;
        /**
         * Controls the display properties of chart.
         */
        visualization?: Visualization | KnockoutObservableBase<Visualization>;
        /**
         * Title for the chart. Defaults to no title.
         */
        title?: string | KnockoutObservableBase<string>;
        /**
         * The timespan used for the MonitorChartV2
         * Defaults to past 24 hours.
         */
        timespan?: Timespan | KnockoutObservableBase<Timespan>;
        /**
         * Aria label for the MonitorChartV2. Defaults to "MonitorChartV2"
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
    }
    /**
     * MonitorChartV2 control.
     */
    export interface Contract {
        /**
         * The list of metrics to plot on this chart.
         */
        readonly metrics: KnockoutObservableArray<Metric.Contract>;
        /**
         * The filters applied on all metrics on this chart.
         */
        readonly filterCollection: KnockoutObservableBase<FilterCollection>;
        /**
         * Specifies the grouping the chart is segmented against.
         */
        readonly grouping: KnockoutObservableBase<Grouping>;
        /**
         * Controls the display properties of chart.
         */
        readonly visualization: KnockoutObservableBase<Visualization>;
        /**
         * Title for the chart.
         */
        readonly title: KnockoutObservableBase<string>;
        /**
         * The timespan used for the chart
         */
        readonly timespan: KnockoutObservableBase<Timespan>;
        /**
         * Aria label for the MonitorChartV2.
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
    }
    /**
     * Creates a MonitorChartV2 viewModel.
     *
     * @param lifetime The lifetime of the control.
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a MonitorChartV2 viewModel.
     */
    export function create(container: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\MoveResourceToolbarButton.d.ts
declare module "Fx/Controls/MoveResourceToolbarButton" {
    import { AnyBladeContainer } from "Fx/Composition/BladeBase";
    export = Main;
    module Main {
        import FxViewModels = MsPortalFx.ViewModels;
        import Toolbars = FxViewModels.Toolbars;
        interface Options {
            /**
             * Resource id.
             */
            resourceId: string;
        }
        /**
         * Creates the move resource toolbar button.
         */
        class ViewModel extends Toolbars.DialogButton implements Toolbars.ToolbarButtonContract {
            /**
             * Creates the move resource toolbar button for moving resources across resource groups/subscriptions.
             *
             * @param container container for the toolbar.
             * @param resourceId resource Id of the resource to be moved.
             *
             * @return A toolbar button which allows moving resources across resource groups/subscriptions.
             */
            constructor(container: AnyBladeContainer, options: Options);
        }
    }
}

// FILE: Fx\Controls\MultiLineTextBox.d.ts
declare module "Fx/Controls/MultiLineTextBox" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export import ValueUpdateTrigger = Base.ValueUpdateTrigger;
    /**
     * Options for configuring a multiline textbox control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * The text shown when no value is set in the control.
         */
        placeHolderText?: string | KnockoutObservableBase<string>;
        /**
         * the number of lines in the control.
         */
        rows?: number | KnockoutObservableBase<number>;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * Controls when the control will update it's value in response to the user typing into it.
         */
        valueUpdateTrigger?: ValueUpdateTrigger;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: string | KnockoutObservableBase<string>;
        /**
         * Aria label for the textbox.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Enable the control to be spellcheck. Default as false
         * (Warning. Perf hit. Do not enable this for a huge text).
         */
        spellcheck?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Event is triggered when user presses the Enter + Shift key.
         */
        onEnterPressed?: (value: string) => void;
    }
    /**
     * Multiline textbox control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<string>;
        /**
         * The text shown when no value is set in the control.
         */
        readonly placeHolderText: KnockoutObservableBase<string>;
        /**
         * the number of lines in the control.
         */
        readonly rows: KnockoutObservableBase<number>;
    }
    /**
     * Creates a mulitline textbox control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a multiline textbox control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\NumericTextBox.d.ts
declare module "Fx/Controls/NumericTextBox" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export import ValueUpdateTrigger = Base.ValueUpdateTrigger;
    /**
     * Options for configuring a numeric textbox control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * The text shown when no value is set in the control.
         */
        placeHolderText?: string | KnockoutObservableBase<string>;
        /**
         * Minimum number allowed.
         */
        min?: number | KnockoutObservableBase<number>;
        /**
         * Maximum number allowed.
         */
        max?: number | KnockoutObservableBase<number>;
        /**
         *  Maximum decimal points allowed for the number. No more than 20.
         */
        decimalPoint?: number | KnockoutObservableBase<number>;
        /**
         * Text to display when entered text is not numeric.
         */
        notANumberErrorMessage?: string;
        /**
         * Specifies whether an empty value should be allowed or if it should be replaced with a 0 value.  Defaults to false.
         */
        allowEmpty?: boolean;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * Controls when the control will update it's value in response to the user typing into it.
         */
        valueUpdateTrigger?: ValueUpdateTrigger;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: number | KnockoutObservableBase<number>;
        /**
         * Aria label for the textbox.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Event is triggered when user presses the enter key.
         */
        onEnterPressed?: (value: number) => void;
    }
    /**
     * Numeric TextBox control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<number>;
        /**
         * Minimum number allowed.
         */
        readonly min: KnockoutObservableBase<number>;
        /**
         * Maximum number allowed.
         */
        readonly max: KnockoutObservableBase<number>;
        /**
         * The text shown when no value is set in the control.
         */
        readonly placeHolderText: KnockoutObservableBase<string>;
        /**
         *  Maximum decimal points allowed for the number. No more than 20.
         */
        readonly decimalPoint: KnockoutObservableBase<number>;
    }
    /**
     * Creates a Numeric TextBox control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a numeric textbox control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\OAuthButton.d.ts
declare module "Fx/Controls/OAuthButton" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export interface Options {
        /**
         * Aria label for the OAuthButton.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * Button text used to populate the text content of the OAuthButton
         */
        buttonText?: string | KnockoutObservableBase<string>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The height of the popup. Set to half the height of the screen if not specified.
         */
        popupHeight?: number;
        /**
         * The width of the popup. Set to half the height of the screen if not specified.
         */
        popupWidth?: number;
        /**
         * Request url used to pop up the OAuth window.  Note: When registering app with provider (e.g., github) you need to specify Authorization callback URL: https://portal.azure.com/TokenAuthorize
         * e.g. for github it is https://github.com/login/oauth/authorize?client_id=<CLIENT_ID>&client_secret=<CLIENT_SECRET>&state=<RANDOM_STRING>
         */
        requestUrl?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * Called when authentication is successful.
         * value is a string in the form of "https://portal.azure.com/TokenAuthorize?code=abcd12345678abcdef12"
         */
        onAuthenticationSucceeded: (value: string) => void;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
    }
    export interface Contract extends Base.Contract {
        /**
         * Request url used to pop up the OAuth window.  Note: When registering app with provider (e.g., github) you need to specify Authorization callback URL: https://portal.azure.com/TokenAuthorize
         * e.g. for github it is https://github.com/login/oauth/authorize?client_id=<CLIENT_ID>&client_secret=<CLIENT_SECRET>&state=<RANDOM_STRING>
         */
        requestUrl: KnockoutObservableBase<string>;
        /**
         * Button text used to populate the text content of the OAuthButton
         */
        buttonText: KnockoutObservableBase<string>;
    }
    /**
     * Creates an OAuth button control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns an OAuth button control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options: Options): Contract;
}

// FILE: Fx\Controls\OptionsGroup.d.ts
declare module "Fx/Controls/OptionsGroup" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export interface Item<TValue> {
        text: string;
        value: TValue;
        disabled?: KnockoutObservableBase<boolean>;
        /**
         * Optional alternate text for screen readers if the text is not explicit enough.
         */
        ariaLabel?: KnockoutObservableBase<string>;
    }
    /**
     * Options for configuring a options group control
     */
    export interface Options<TValue> {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * If true, sets each item to equal width and the total size of the control to the size of the element.
         * Otherwise, sets the width of each item in the options group to the size of the content. Defaults to false.
         */
        uniformItemWidth?: boolean;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The list of items of the options group
         */
        items?: Array<Item<TValue>> | KnockoutObservableArray<Item<TValue>>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: TValue | KnockoutObservableBase<TValue>;
        /**
         * Optional alternate text for screen readers.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
    }
    /**
     * ItemPicker control view model
     */
    export interface Contract<TValue> extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<TValue>;
        /**
         * The list of items of the options group
         */
        readonly items: KnockoutObservableArray<Item<TValue>>;
    }
    /**
     * Creates a options group control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns an options group control viewmodel
     */
    export function create<TValue>(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options<TValue>): Contract<TValue>;
}

// FILE: Fx\Controls\PairedTimeline.d.ts
declare module "Fx/Controls/PairedTimeline" {
    export = Main;
    import * as PairedTimelineBadges from "Fx/Controls/PairedTimelineBadges";
    import * as RangeSelection from "Fx/Controls/RangeSelection";
    module Main {
        import Controls = MsPortalFx.ViewModels.Controls;
        import Visualization = Controls.Visualization;
        import Chart = Visualization.Chart;
        import Metrics = Visualization.Metrics;
        import ListView = Controls.Lists.ListView;
        import Loadable = Controls.Loadable;
        /**
         * Contract container for the Chart and Metrics controls
         */
        interface PairedTimelineChartContract<TX, TY> extends Loadable.Contract {
            /**
             * Identifier of chart
             */
            id: string;
            /**
             * Chart control contract
             */
            chart: Chart.Contract<TX, TY>;
            /**
             * Chart height in number of units(integer). Each unit is 45 px.
             */
            chartHeight: KnockoutObservableBase<number>;
            /**
             * The message to display when chart is un-configured or there is no chart data.
             */
            noDataMessage: KnockoutObservableBase<string>;
            /**
             * CSS class to be applied to the List View item
             */
            cssClass: KnockoutObservableBase<string>;
            /**
             * Selectable for the buttons in the PT
             */
            buttonSelectables: MsPortalFx.ViewModels.Selectable<string>[];
            /**
             * The title for the chart.
             */
            title: KnockoutObservableBase<string>;
            /**
             * Whether or not the title is shown.
             */
            showTitle: KnockoutObservableBase<boolean>;
        }
        /**
         * Container for the Chart and Metrics controls
         */
        class PairedTimelineChart<TX, TY> extends Loadable.ViewModel implements PairedTimelineChartContract<TX, TY> {
            /**
             * Identifier of chart
             */
            id: string;
            /**
             * Chart control
             */
            chart: Chart.ViewModel<TX, TY>;
            /**
             * Chart height in number of units(integer). Each unit is 45 px.
             */
            chartHeight: KnockoutObservableBase<number>;
            /**
             * The message to display when the chart is un-configured or there is no chart data.
             */
            noDataMessage: KnockoutObservableBase<string>;
            /**
             * CSS class to be applied to the List View item
             */
            cssClass: KnockoutObservableBase<string>;
            /**
             * Selectable for the buttons in the PT
             */
            buttonSelectables: MsPortalFx.ViewModels.Selectable<string>[];
            /**
             * The title for the chart.
             */
            title: KnockoutObservableBase<string>;
            /**
             * Whether or not the title is shown.
             * Title will only be shown for >1 chart height charts.
             */
            showTitle: KnockoutObservableBase<boolean>;
            /**
             * Creates a paired timeline.
             *
             * @param lifetimeManager Lifetime manager for this view model.
             * @param id Id of this view model.
             */
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, id: string);
            dispose(): void;
        }
        const enum PairedTimelineButtonVisibilityOptions {
            ShowOnHover = 1,
            AlwaysShow = 2,
        }
        interface PairedTimelineButtonOptions {
            /**
             * Whether the Edit buttons should be shown.
             */
            enabled: KnockoutObservableBase<boolean>;
            /**
             * Text to use for the edit button, default is "Edit"
             */
            text: KnockoutObservableBase<string>;
            /**
             * Controls the visibility of the buttons. (always vs onhover)
             */
            visibility: KnockoutObservableBase<PairedTimelineButtonVisibilityOptions>;
        }
        interface Contract<TX, TY> extends Loadable.Contract {
            /**
             * The collection of Paired Timeline charts.
             */
            ptCharts: KnockoutObservableArray<PairedTimelineChart<TX, TY>>;
            /**
             * Options used to configure the loaded plugins for selection.
             * Updating this property after the widget is initialized will have no effect.
             * A null or undefined value disables selection.
             */
            extensionOptions: ListView.ExtensionOptions;
            /**
             * Enable metrics rules.
             */
            enableMetricsRules: KnockoutObservableBase<boolean>;
            /**
             * Metrics options
             */
            metricsOptions: Metrics.VisualContract;
            /**
             * Enables sliders to select a range on the x axis.
             */
            enableRangeSelection: KnockoutObservableBase<boolean>;
            /**
             * Range selection view model.
             */
            rangeSelectionViewModel: RangeSelection.ViewModel<TX>;
            /**
             * Options for the buttons defined in the PT.
             */
            buttonOptions: PairedTimelineButtonOptions[];
            /**
             * Badges View model
             */
            badgesViewModel: PairedTimelineBadges.ViewModel<TX>;
        }
        class ViewModel<TX, TY> extends Loadable.ViewModel implements Contract<TX, TY> {
            /**
             * The collection of Paired Timeline charts.
             */
            ptCharts: KnockoutObservableArray<PairedTimelineChart<TX, TY>>;
            /**
             * Options used to configure the loaded plugins for selection.
             * Updating this property after the widget is initialized will have no effect.
             */
            extensionOptions: ListView.ExtensionOptions;
            /**
             * Enable Metrics Rules
             */
            enableMetricsRules: KnockoutObservableBase<boolean>;
            /**
             * MetricsOptions
             */
            metricsOptions: Metrics.VisualContract;
            /**
             * Enables sliders to select a range on the x axis.
             */
            enableRangeSelection: KnockoutObservableBase<boolean>;
            /**
             * Range selection view model.
             */
            rangeSelectionViewModel: RangeSelection.ViewModel<TX>;
            /**
             * Badges View model
             */
            badgesViewModel: PairedTimelineBadges.ViewModel<TX>;
            /**
             * Options for the buttons defined in the PT.
             */
            buttonOptions: PairedTimelineButtonOptions[];
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, ptCharts: KnockoutObservableArray<PairedTimelineChart<TX, TY>>, extensionOptions: ListView.ExtensionOptions, rangeSelectionOptions?: RangeSelection.Options<TX>, pairedTimelineBadgesOptions?: PairedTimelineBadges.Options<TX>);
        }
    }
}

// FILE: Fx\Controls\PairedTimelineBadges.d.ts
declare module "Fx/Controls/PairedTimelineBadges" {
    export = Main;
    module Main {
        import DockedBalloon = MsPortalFx.ViewModels.Controls.DockedBalloon;
        import FxViewModelsControlsBase = MsPortalFx.ViewModels.Controls.Base;
        /**
            * Contract for an individual displayable badge.
            */
        interface BadgeContract<TX> {
            /**
                * Badge Identifier
                */
            id: string;
            /**
                * Unique class name for this badge.
                */
            className: KnockoutObservableBase<string>;
            /**
                * The badge text
                */
            text: string;
            /**
                * The x value corresponding to the event this badge is reporting.
                */
            xValue: TX;
            /**
                * The coordinate where the badge is rendered.
                */
            coordinate?: KnockoutObservableBase<number>;
            /**
                * The icon to display for the badge.
                */
            icon: KnockoutObservableBase<MsPortalFx.Base.Image>;
            /**
                * Whether the badge is displayed.
                */
            visible?: KnockoutObservableBase<boolean>;
        }
        /**
            * Class for an individual Badge.
            */
        class Badge<TX> extends FxViewModelsControlsBase.ViewModel implements BadgeContract<TX> {
            /**
                * Badge Identifier
                */
            id: string;
            /**
                * Unique class name for this badge.
                */
            className: KnockoutObservableBase<string>;
            /**
                * The badge text
                */
            text: string;
            /**
                * The x value corresponding to the event this badge is reporting.
                */
            xValue: TX;
            /**
                * The icon to display for the badge.
                */
            icon: KnockoutObservableBase<MsPortalFx.Base.Image>;
            /**
                * Whether the badge is displayed.
                */
            visible: KnockoutObservableBase<boolean>;
            /**
                * The coordinate where the badge is rendered.
                */
            coordinate: KnockoutObservableBase<number>;
            /**
                * Creates a paired timeline.
                *
                * @param lifetimeManager Lifetime manager for this view model.
                * @param id Id of this view model.
                */
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, options?: BadgeContract<TX>);
        }
        interface Options<TX> {
            /**
                * The collection of badges
                */
            badges: KnockoutObservableArray<BadgeContract<TX>>;
            /**
                * Height of the badges when displayed on the paired timeline
                */
            height: KnockoutObservableBase<number>;
            /**
                * Balloon for displaying badge information.
                */
            balloon: KnockoutObservableBase<DockedBalloon.ViewModel>;
            /**
                * Enables badges that will be displayed at top of paired timeline
                */
            visible: KnockoutObservableBase<boolean>;
        }
        class ViewModel<TX> extends FxViewModelsControlsBase.ViewModel {
            /**
                * The collection of badges
                */
            badges: KnockoutObservableArray<BadgeContract<TX>>;
            /**
                * Height of the badges when displayed on the paired timeline
                */
            height: KnockoutObservableBase<number>;
            /**
                * Balloon for displaying badge information.
                */
            balloon: KnockoutObservableBase<DockedBalloon.ViewModel>;
            /**
                * Enables badges that will be displayed at top of paired timeline
                */
            visible: KnockoutObservableBase<boolean>;
            /**
                * An optional selectableSet to make the badges selectable
            */
            selectableSet: KnockoutObservableBase<MsPortalFx.ViewModels.SelectableSet<any, any>>;
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, options?: Options<TX>);
            /**
                * Set the viewmodel's array of badges to the incoming badges. Optionally set up a selectableSet on these badges with the given callbacks.
                */
            pushAllBadges(newBadges: BadgeContract<TX>[], useSelectableSet?: boolean, itemMatchesSelection?: (item: BadgeContract<TX>, selection: any) => boolean, createSelection?: (item: BadgeContract<TX>) => any): void;
        }
    }
}

// FILE: Fx\Controls\PasswordBox.d.ts
declare module "Fx/Controls/PasswordBox" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    /**
     * Options for configuring a password textbox control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * The text shown when no value is set in the control.
         */
        placeHolderText?: string | KnockoutObservableBase<string>;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: string | KnockoutObservableBase<string>;
        /**
         * Shows validation as a popup.
         */
        showValidationsAsPopup?: boolean;
        /**
         * Aria label for the control.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Event is triggered when user presses the enter key.
         */
        onEnterPressed?: (value: string) => void;
    }
    /**
     * Password box control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<string>;
        /**
         * The text shown when no value is set in the control.
         */
        readonly placeHolderText: KnockoutObservableBase<string>;
    }
    /**
     * Creates a password box control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a password box control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\Pill.d.ts
declare module "Fx/Controls/Pill" {
    import FxBase = MsPortalFx.Base;
    import Image = FxBase.Image;
    import DockedBalloon = MsPortalFx.ViewModels.Controls.DockedBalloon;
    import DialogResult = MsPortalFx.ViewModels.Dialogs.DialogResult;
    import CustomDialogButton = MsPortalFx.ViewModels.CustomDialogButton;
    import { HtmlContent } from "Fx/Controls/ControlsBase";
    /**
     * Html content.
     */
    export { HtmlContent };
    /**
     * Common edit contract properties.
     */
    export interface Editor {
        /**
         * The editor will render this content in edit mode.
         */
        content: MsPortalFx.Control | HtmlContent;
    }
    /**
     * Inline editor.
     */
    export interface InlineEditor<T> extends Editor {
        /**
         * Callback function called when the editMode is ended.
         *
         * @returns new value for the pill.
         */
        onEditModeExited: () => T;
    }
    /**
     * Dialog editor.
     */
    export interface DialogEditor<T> extends Editor {
        /**
         * Dialog title.
         */
        title: string | KnockoutObservableBase<string>;
        /**
         * Callback function called when the editMode is ended.
         * @param result Obtained dialog result by clicking a dialog button or dismiss the dialog. When the user clicks outside of the dialog, the dialog will be dismissed and will return DialogResult.Cancel.
         *
         * @returns new value for the pill.
         */
        onEditModeExited: (result: DialogResult) => T;
        /**
         * Buttons used for the dialog.
         */
        buttons: CustomDialogButton[];
    }
    /**
     * Custom editor.
     */
    export interface CustomEditor {
        /**
         * onClick handler fired when the pill is clicked.
         * The author has responsibility to manage the pill value to make changes.
         */
        onClick: () => void;
    }
    /**
     * Options for configuring a pill control.
     */
    export interface Options<T> {
        /**
         * A function that generates the displaying text for key.
         */
        getKeyText: (value: T) => string;
        /**
         * A function that generates the displaying text for operator.
         */
        getOperatorText?: (value: T) => string;
        /**
         * A function that generates the displaying text for value.
         */
        getValueText: (value: T) => string;
        /**
         * The initial value of the pill.
         */
        value: T | KnockoutObservableBase<T>;
        /**
         * The editor of the pill.
         */
        createEditor: () => InlineEditor<T> | DialogEditor<T> | CustomEditor | Q.Promise<InlineEditor<T>> | Q.Promise<DialogEditor<T>> | Q.Promise<CustomEditor>;
        /**
         * Callback that fires when the pill is removed.
         */
        onRemoved?: (value: T) => void;
        /**
         * Show a remove button to allow users to remove the pill.
         */
        showRemoveButton?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Show an icon on the left side of the pill.
         */
        icon?: Image | KnockoutObservableBase<Image>;
        /**
         * Enable "readonly" mode that removes editable behavior from the pill.
         */
        readonly?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Indicates that whether the pill is in edit mode or not.
         */
        editMode?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Tooltip to show on hover in readonly mode.
         */
        tooltip?: DockedBalloon.ViewModel | KnockoutObservableBase<DockedBalloon.ViewModel>;
        /**
         * If true then this pill will not trigger the unsaved changes dialog, even after the user makes changes.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * Aria label for pill.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * Pill's disabled state.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
    }
    /**
     * Pill control contract.
     */
    export interface Contract<T> extends MsPortalFx.Control {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<T>;
        /**
         * Show a remove button to allow users to remove the pill.
         */
        readonly showRemoveButton: KnockoutObservableBase<boolean>;
        /**
         * Show an icon on the left side of the pill.
         */
        readonly icon: KnockoutObservableBase<Image>;
        /**
         * Enable "readonly" mode that removes editable behavior from the pill.
         */
        readonly readonly: KnockoutObservableBase<boolean>;
        /**
         * Indicates that whether the pill is in edit mode or not.
         */
        readonly editMode: KnockoutObservableBase<boolean>;
        /**
         * Tooltip to show on hover in readonly mode.
         */
        readonly tooltip: KnockoutObservableBase<DockedBalloon.ViewModel>;
        /**
         * Aria label for pill.
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * Pill's disabled state.
         */
        readonly disabled: KnockoutObservableBase<boolean>;
    }
    /**
     * Creates a pill control with inline edit mode contract.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a Pill contract
     */
    export function create<T>(lifetime: FxBase.LifetimeManager, options: Options<T>): Contract<T>;
}

// FILE: Fx\Controls\PillCollection.d.ts
declare module "Fx/Controls/PillCollection" {
    import * as Pill from "Fx/Controls/Pill";
    import FxBase = MsPortalFx.Base;
    import Image = FxBase.Image;
    export const enum Theme {
        /**
         * Theme for filters (default).
         */
        Filter = 1,
        /**
         * Theme for tags.
         */
        Tag = 2,
    }
    /**
     * Options for configuring a pill collection control.
     */
    export interface Options {
        /**
         * The initial set of pills in the collection.
         */
        items: Pill.Contract<any>[] | KnockoutObservableArray<Pill.Contract<any>>;
        /**
         * Options for "add pill" button.
         */
        addition?: AdditionOptions;
        /**
         * Aria label for pill collection.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * Pill collection's disabled state.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Pill collection's theme.
         */
        theme?: Theme;
        /**
         * Max number of items of pill to show.
         */
        maxVisibleItems?: number;
        /**
         * Show pills in multiple rows or truncate them when they exceeds a single row.
         */
        multipleRows?: boolean;
    }
    /**
     * Options for "add pill" button.
     */
    export interface AdditionOptions {
        /**
         * Enable the "add pill" button.
         */
        canAddPill?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Set this true to always show label or label will be hidden, if there is any items.
         */
        alwaysShowLabel?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Icon for "add pill" button. If not given, default icon will be used.
         */
        icon?: Image | KnockoutObservableBase<Image>;
        /**
         * Label for "add pill" button.
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * A function called when the user clicks "add pill" button.
         * Returns a pill viewModel that will be added to "items".
         * If nothing is returned, this callback becomes like simple onClick handler and the author has responsibility to manage the pill collection items to make changes.
         */
        addPill?: () => Pill.Contract<any> | void;
    }
    export interface Addition {
        /**
         * Enable the "add pill" button.
         */
        readonly canAddPill: KnockoutObservableBase<boolean>;
        /**
         * Set this true to always show label or label will be hidden, if there is any items.
         */
        readonly alwaysShowLabel: KnockoutObservableBase<boolean>;
        /**
         * Icon for "add pill" button. If not given, default icon will be used.
         */
        readonly icon: KnockoutObservableBase<Image>;
        /**
         * Label for "add pill" button.
         */
        readonly label: KnockoutObservableBase<string>;
    }
    export interface Contract extends MsPortalFx.Control {
        /**
         * The set of pills in the collection.
         */
        readonly items: KnockoutObservableArray<Pill.Contract<any>>;
        /**
         * Options for "add pill" button.
         */
        readonly addition: Addition;
        /**
         * Aria label for pill collection.
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * Pill collection's disabled state.
         */
        readonly disabled: KnockoutObservableBase<boolean>;
    }
    /**
     * Creates an Pill collection control contract.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a Pill collection contract
     */
    export function create(lifetime: FxBase.LifetimeManager, options: Options): Contract;
}

// FILE: Fx\Controls\ProgressBar.d.ts
declare module "Fx/Controls/ProgressBar" {
    /**
     * Status type of progress indicators
     */
    export const enum StatusType {
        /**
         * In progress style.
         */
        InProgress = 0,
        /**
         * Error style.
         */
        Error = 1,
        /**
         * Warning style.
         */
        Warning = 2,
        /**
         * Success style.
         */
        Success = 3,
        /**
         * Indeterminate style.
         */
        Indeterminate = 4,
    }
    export interface Options {
        /**
         * Value percentage of the current progress.
         */
        valuePercentage?: number | KnockoutObservableBase<number>;
        /**
         * Status of the progress indicator.
         */
        status?: StatusType | KnockoutObservableBase<StatusType>;
        /**
         * Text for screen reader to describe this progressbar
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
    }
    export interface Contract extends MsPortalFx.Control {
        /**
         * Value percentage of the current progress.
         */
        readonly valuePercentage: KnockoutObservableBase<number>;
        /**
         * Status of the progress indicator.
         */
        readonly status: KnockoutObservableBase<StatusType>;
        /**
         * Text for screen reader to describe this progressbar
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * The disabled state of the control.
         */
        readonly disabled: KnockoutObservableBase<boolean>;
    }
    /**
     * Creates a ProgressBar control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a ProgressBar viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\RadioButtons.d.ts
declare module "Fx/Controls/RadioButtons" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export interface Item<TValue> {
        text: string;
        value: TValue;
        disabled?: KnockoutObservableBase<boolean>;
        /**
         * Optional alternate text for screen readers if the text is not explicit enough.
         */
        ariaLabel?: KnockoutObservableBase<string>;
    }
    /**
     * Options for configuring a radio button control
     */
    export interface Options<TValue> {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * If set, each button will each appear on their own line.  Defaults to false.
         */
        singleItemPerLine?: boolean;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The list of items of the radio buttons
         */
        items?: Array<Item<TValue>> | KnockoutObservableArray<Item<TValue>>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: TValue | KnockoutObservableBase<TValue>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Optional alternate text for screen readers.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
    }
    /**
     * RadioButtons control view model
     */
    export interface Contract<TValue> extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<TValue>;
        /**
         * The list of items of the radio button
         */
        readonly items: KnockoutObservableArray<Item<TValue>>;
    }
    /**
     * Creates a radio buttons control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a radio buttons control viewmodel
     */
    export function create<TValue>(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options<TValue>): Contract<TValue>;
}

// FILE: Fx\Controls\RangeSelection.d.ts
declare module "Fx/Controls/RangeSelection" {
    export = Main;
    module Main {
        import FxControls = MsPortalFx.ViewModels.Controls;
        const enum LabelPosition {
            /**
                * Range selection labels will be oriented on the inside of the selection range
                */
            Inside = 0,
            /**
                * Range selection labels will be oriented on the outside of the selection range
                */
            Outside = 1,
        }
        interface Range<T> {
            /**
                * The beginning of the range selection.  May be null to indicate open beginning of range.
                */
            start: T;
            /**
                * The end of the range selection.  May be null to indicate open ended range.
                */
            end: T;
        }
        interface Options<T> {
            /**
                * Specifies position of range selection labels.
                */
            labelPosition?: KnockoutObservableBase<LabelPosition>;
            /**
                * The range selected if enableRangeSelection was enabled.
                */
            rangeSelection?: KnockoutObservableBase<Range<T>>;
        }
        class ViewModel<T> extends FxControls.Base.ViewModel {
            /**
                * Specifies position of range selection labels.
                */
            labelPosition: KnockoutObservableBase<LabelPosition>;
            /**
                * The range selected if enableRangeSelection was enabled.
                */
            rangeSelection: KnockoutObservableBase<Range<T>>;
            constructor(lifetimeManager: MsPortalFx.Base.LifetimeManager, options?: Options<T>);
        }
    }
}

// FILE: Fx\Controls\RangeSlider.d.ts
declare module "Fx/Controls/RangeSlider" {
    import { SliderLegends } from "Fx/Controls/Slider";
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    /**
     * Options for configuring a range slider control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The aria label of the control, for screen reader
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * Validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * Sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * Minimum value of the range slider.
         */
        min?: number | KnockoutObservableBase<number>;
        /**
         * Maximum value of the range slider.
         */
        max?: number | KnockoutObservableBase<number>;
        /**
         * Determines the size or amount of each interval or step the range slider takes between min and max.
         */
        step?: number | KnockoutObservableBase<number>;
        /**
         * Displays a marker for each step.
         */
        showStepMarkers?: boolean | KnockoutObservableBase<boolean>;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: string | KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Legends for the slider.
         */
        legends?: SliderLegends;
        /**
         * Hide the textbox of the control.
         */
        hideTextbox?: boolean | KnockoutObservableBase<boolean>;
    }
    /**
     * RangeSlider control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<string>;
        /**
         * Start value of the range.
         */
        readonly start: KnockoutObservableBase<number>;
        /**
         * End value of the range.
         */
        readonly end: KnockoutObservableBase<number>;
        /**
         * Minimum value of the range slider.
         */
        readonly min: KnockoutObservableBase<number>;
        /**
         * Maximum value of the range slider.
         */
        readonly max: KnockoutObservableBase<number>;
        /**
         * Determines the size or amount of each interval or step the range slider takes between min and max.
         */
        readonly step: KnockoutObservableBase<number>;
        /**
         * Displays a marker for each step.
         */
        readonly showStepMarkers: KnockoutObservableBase<boolean>;
        /**
         * Legends for the slider.
         */
        readonly legends: SliderLegends;
        /**
         * Hide the textbox of the control.
         */
        readonly hideTextbox: KnockoutObservableBase<boolean>;
    }
    /**
     * Options for configuring the range slider control
     */
    export interface CustomValueOptions {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The aria label of the control, for screen reader
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * Validations on the control.
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * Sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * Displays a marker for each step.
         */
        showStepMarkers?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Custom range slider clickstop values.
         */
        customValues: Array<number> | KnockoutObservableBase<Array<number>>;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: string | KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Legends for the slider.
         */
        legends?: SliderLegends;
        /**
         * Hide the textbox of the control.
         */
        hideTextbox?: boolean | KnockoutObservableBase<boolean>;
    }
    /**
     * CustomRangeSlider control view model
     */
    export interface CustomValueContract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<string>;
        /**
         * Displays a marker for each step.
         */
        readonly showStepMarkers: KnockoutObservableBase<boolean>;
        /**
         * Custom slider clickstop values.
         */
        readonly customValues: KnockoutObservableBase<number[]>;
        /**
         * Legends for the slider.
         */
        readonly legends: SliderLegends;
        /**
         * Hide the textbox of the control.
         */
        readonly hideTextbox: KnockoutObservableBase<boolean>;
    }
    /**
     * Creates a range slider control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a range slider control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
    /**
     * Creates a range slider control viewmodel with custom values.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a range slider control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options: CustomValueOptions): CustomValueContract;
}

// FILE: Fx\Controls\ResourceGroupDropDown.d.ts
declare module "Fx/Controls/ResourceGroupDropDown" {
    import { AnyBladeContainer } from "Fx/Composition/BladeBase";
    import { ResourceGroup, Validation as BaseValidation } from "Fx/Controls/BaseResourceDropDown";
    export = Main;
    module Main {
        /**
         * The modes possible for the dropdown
         */
        export import Mode = ResourceGroup.Mode;
        /**
         * The mode of the value returned by the control
         */
        export import SelectedMode = ResourceGroup.SelectedMode;
        /**
         * The contract for the values returned by the resource group dropdown
         * It contains the ARM Value and a flag for if you are in create mode
         */
        export import Value = ResourceGroup.Value;
        /**
         * The validation type accepted by the dropdown
         */
        type Validation = BaseValidation<Value>;
        /**
         * The contract for options to create the resource group drop down
         */
        type Options = ResourceGroup.Options;
        /**
         * The contract for the resource group dropdown
         */
        interface Contract extends ResourceGroup.Contract {
        }
        /**
         * This creates a resource group dropdown control.
         * @param container The container associated with the part or blade.
         * @param options The resource group dropdown options. See interface
         */
        function create(container: AnyBladeContainer, options?: Options): Contract;
    }
}

// FILE: Fx\Controls\SearchBox.d.ts
declare module "Fx/Controls/SearchBox" {
    /**
     * Options for configuring a SearchBox control
     */
    export interface Options {
        /**
         * The default placeholder text.
         */
        placeHolderText?: string | KnockoutObservableBase<string>;
        /**
         * The delay after you finish typing to trigger the update callback in milliseconds.  Defaults to 400 ms
         */
        valueUpdateTimeout?: number;
        /**
         * The update callback.  Called when the user types in entries (after they stop typing for the duration of valueUpdateTimeout)
         * value is the value typed in by the user
         * returns the values shown in the dropdown
         */
        update?: (value: string) => Q.Promise<string[]>;
        /**
         * The submit callback.  Called when the user selects an item, or hits enter.
         * value is the value of the selected item
         * typedValue is the value that the user typed in (this will match value when the user hits enter)
         * returns a promise that when resolved will take the control out of the busy state
         */
        submit?: (value: string, typedValue: string) => Q.Promise<boolean>;
        /**
         * Disabled state of the search box.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Hide suggestion box.
         */
        hideSuggestions?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Aria label for search box.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * Maximum number of results shown.  Defaults to 5.
         */
        resultLimit?: number | KnockoutObservableBase<number>;
    }
    /**
     * SearchBox control view model
     */
    export interface Contract extends MsPortalFx.Control {
        /**
         * The placeholder text that is shown when the user has not entered a value.
         */
        readonly placeHolderText: KnockoutObservableBase<string>;
        /**
         * Disabled state of the search box.
         */
        readonly disabled: KnockoutObservableBase<boolean>;
        /**
         * Hide suggestion box.
         */
        readonly hideSuggestions: KnockoutObservableBase<boolean>;
        /**
         * Aria label for search box.
         */
        readonly ariaLabel: KnockoutObservableBase<string>;
        /**
         * Maximum number of results shown.
         */
        readonly resultLimit: KnockoutObservableBase<number>;
        /**
         * A helper function for that removes all items, closes dropdown & clears textbox.
         * @return A promise that is resolved when the reset has been completed.
         */
        reset(): Q.Promise<void>;
    }
    /**
     * Creates a SearchBox control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a SearchBox viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\Section.d.ts
declare module "Fx/Controls/Section" {
    import { HtmlContent } from "Fx/Controls/ControlsBase";
    /**
     * Options for configuring a form section
     */
    export interface Options {
        /**
         * The name of the section.
         * Can be an HTML template and optional view model. See HtmlContent interface in Fx/Controls/ControlsBase
         * This will be displayed as a tab header if the section is inside a tab control.
         */
        name?: string | HtmlContent | KnockoutObservableBase<string | HtmlContent>;
        /**
         * The sections & controls to nest within the section.
         */
        children?: any[] | KnockoutObservableArray<any>;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The minimum height of the section in pixels.  Use this to reduce jumpiness in forms.
         */
        minHeight?: number;
        /**
         * Labels will appear to the left of child controls when this is set to true.
         * Labels are set to the size of the largest label in a consecutive set (unless leftLabelWidth is set).
         */
        leftLabelPosition?: boolean;
        /**
         * The width of the left label in px.
         * Setting this property implies leftLabelPosition = true.
         */
        leftLabelWidth?: number;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
    }
    /**
     * Section view model
     */
    export interface Contract extends MsPortalFx.Control {
        /**
         * The sections & controls to nest within the section.
         */
        readonly children: KnockoutObservableArray<any>;
        /**
         * Changes the section state to disabled.
         */
        readonly disabled: KnockoutObservableBase<boolean>;
        /**
         * The name of the section.
         * Can be an HTML template and optional view model. See HtmlContent interface in Fx/Controls/ControlsBase
         * This will be displayed as a tab header if the section is inside a tab control.
         */
        readonly name: KnockoutObservableBase<string | HtmlContent>;
    }
    /**
     * Creates a section view model
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
    /**
     * Describes an HTML view, defined in terms of an HTML template bound to an accompanying view model object.
     */
    export { HtmlContent };
}

// FILE: Fx\Controls\Slider.d.ts
declare module "Fx/Controls/Slider" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    /**
     * Options for configuring a slider control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The aria label of the control, for screen reader
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * Minimum value of the slider.
         */
        min?: number | KnockoutObservableBase<number>;
        /**
         * Maximum value of the slider.
         */
        max?: number | KnockoutObservableBase<number>;
        /**
         * Determines the size or amount of each interval or step the slider takes between min and max.
         */
        step?: number | KnockoutObservableBase<number>;
        /**
         * Displays a marker for each step.
         */
        showStepMarkers?: boolean | KnockoutObservableBase<boolean>;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: number | KnockoutObservableBase<number>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Legends for the slider.
         */
        legends?: SliderLegends;
        /**
         * Hide the textbox of the control.
         */
        hideTextbox?: boolean | KnockoutObservableBase<boolean>;
    }
    /**
     * Legend for Slider
     */
    export interface SliderLegends {
        /**
         * The labels for the slider.
         */
        labels: SliderLabel[];
        /**
         * Label to show when the value is active.
         */
        activeValueLabel: (value: number) => string;
        /**
         * Sublabel to show when the value is active.
         */
        activeValueSublabel?: (value: number) => string;
    }
    /**
     * Label for Slider
     */
    export interface SliderLabel {
        /**
         * The value of the slider to map.
         */
        value: number;
        /**
         * The label for the corresponding slider value.
         */
        label: string;
    }
    /**
     * Slider control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<number>;
        /**
         * Minimum value of the range slider.
         */
        readonly min: KnockoutObservableBase<number>;
        /**
         * Maximum value of the range slider.
         */
        readonly max: KnockoutObservableBase<number>;
        /**
         * Determines the size or amount of each interval or step the range slider takes between min and max.
         */
        readonly step: KnockoutObservableBase<number>;
        /**
         * Displays a marker for each step.
         */
        readonly showStepMarkers: KnockoutObservableBase<boolean>;
        /**
         * Legends for the slider.
         */
        readonly legends: SliderLegends;
        /**
         * Hide the textbox of the control.
         */
        readonly hideTextbox: KnockoutObservableBase<boolean>;
    }
    /**
     * Options for configuring the slider control
     */
    export interface CustomValueOptions {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The aria label of the control, for screen reader
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * Displays a marker for each step.
         */
        showStepMarkers?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Custom slider clickstop values.
         */
        customValues: number[] | KnockoutObservableBase<number[]>;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: number | KnockoutObservableBase<number>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Legends for the slider.
         */
        legends?: SliderLegends;
        /**
         * Hide the textbox of the control.
         */
        hideTextbox?: boolean | KnockoutObservableBase<boolean>;
    }
    export interface CustomValueContract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<number>;
        /**
         * Displays a marker for each step.
         */
        readonly showStepMarkers: KnockoutObservableBase<boolean>;
        /**
         * Custom slider clickstop values.
         */
        readonly customValues: KnockoutObservableBase<number[]>;
        /**
         * Legends for the slider.
         */
        readonly legends: SliderLegends;
        /**
         * Hide the textbox of the control.
         */
        readonly hideTextbox: KnockoutObservableBase<boolean>;
    }
    /**
     * Creates a slider control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a slider control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
    /**
     * Creates a slider control viewmodel with custom values.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a slider control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options: CustomValueOptions): CustomValueContract;
}

// FILE: Fx\Controls\Splitter.d.ts
declare module "Fx/Controls/Splitter" {
    export const enum Panes {
        Primary = 0,
        Secondary = 1,
        Both = 2,
    }
    export interface Options {
        /**
         * The initial size of the primary pane.
         */
        primaryPaneSize?: number;
        /**
         * Controls which panes are visible.  Defaults to Both
         */
        showPanes?: Panes | KnockoutObservableBase<Panes>;
        /**
         * Set to true to change splitter orientation to vertical.  Defaults to false.
         */
        vertical?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Reverses the splitter's flow.  This is mostly used to put the primary pane on the right in order to set a fixed initial width via primaryPaneSize.
         */
        reversed?: boolean;
        /**
         * Minimum pane sizes
         */
        minPaneSizes?: {
            primaryPaneSize: number;
            secondaryPaneSize: number;
        };
    }
    export interface Contract extends MsPortalFx.Control {
        /**
         * Controls which panes are visible.  Defaults to Both
         */
        readonly showPanes: KnockoutObservableBase<Panes>;
        /**
         * Set to true to change splitter orientation to vertical.  Defaults to false.
         */
        readonly vertical: KnockoutObservableBase<boolean>;
    }
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\Storage\FileShareDropDown.d.ts
declare module "Fx/Controls/Storage/FileShareDropDown" {
    import { Contract as BaseContract } from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    /**
     * The contract for the values returned by the file share dropdown
     */
    export interface FileShare {
        /**
         * The name of the select File Share.
         */
        name: string;
        /**
         * The full URL path to the selected File Share.
         */
        url: string;
    }
    /**
     * Options for configuring the file share drop down.
     */
    export interface Options {
        /**
         * Required. The ID of the Storage Account for which to pull file shares
         */
        storageAccountId: string | KnockoutObservableBase<string>;
        /**
         * The aria-label for the control.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * Change the default filter string.
         */
        filterPlaceholder?: string;
        /**
         * Toggle multiselect on.
         */
        multiSelect?: boolean;
        /**
         * Multiple items text format (multiselect only).
         * By default the format string is "{0} selected"
         */
        multiItemDisplayText?: string | KnockoutObservableBase<string>;
        selectAll?: boolean;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Control or view the visible state of this control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
    }
    /**
     * File share drop down Viewmodel
     */
    export interface Contract extends BaseContract {
        /**
         * The value of the dropdown
         */
        readonly value: KnockoutObservableBase<FileShare> & KnockoutObservableBase<FileShare[]>;
        /**
         * The Storage Account ID for which to pull file shares.
         * Updating this value will cause the set of items in the drop down to get updated as well.
         */
        readonly storageAccountId: KnockoutObservableBase<string>;
        /**
         * Current value of the filter textbox.
         */
        readonly filterString: KnockoutObservableBase<string>;
        /**
         * Puts the control into a loading state.
         */
        readonly loading: KnockoutReadOnlyObservableBase<boolean>;
        /**
         * Multiple items text format (multiselect only).
         * By default the format string is "{0} selected"
         */
        readonly multiItemDisplayText: KnockoutObservableBase<string>;
    }
    /**
     * Creates a file share drop down viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a file share drop down viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options: Options): Contract;
}

// FILE: Fx\Controls\SubscriptionDropDown.d.ts
declare module "Fx/Controls/SubscriptionDropDown" {
    import { AnyBladeContainer } from "Fx/Composition/BladeBase";
    import { Subscription, Validation as BaseValidation } from "Fx/Controls/BaseResourceDropDown";
    export = Main;
    module Main {
        /**
         * The contract for the values returned by the subscription dropdown
         */
        export import Subscription = MsPortalFx.Azure.Subscription;
        /**
         * The contract for options to create the subscription drop down
         */
        type Options = Subscription.Options;
        /**
         * The validation type accepted by the dropdown
         */
        type Validation = BaseValidation<Subscription>;
        /**
         * The contract for the subscription dropdown
         */
        interface Contract extends Subscription.Contract {
        }
        /**
         * This creates a subscription dropdown control.
         * @param container The container associated with the part or blade.
         * @param options The subscription dropdown options. See interface
         */
        function create(container: AnyBladeContainer, options?: Options): Contract;
    }
}

// FILE: Fx\Controls\TabControl.d.ts
declare module "Fx/Controls/TabControl" {
    import { Contract as SectionContract } from "Fx/Controls/Section";
    /**
     * Options for configuring a tab control
     */
    export interface Options {
        /**
         * The section view models containing the contents of each tab.
         */
        tabs?: SectionContract[];
        /**
         * A custom css class to apply to the tab control.
         */
        cssClass?: string;
        /**
         * The minimum height of the tab control in pixels.  Use this to reduce jumpiness in forms.
         */
        minHeight?: number;
        /**
         * Show required asterisks on tab.
         * If there is any required control or section within the tab we show a red '*' on the left of the tab header.
         */
        showRequiredStatusOnTabs?: boolean;
        /**
         * Show validation failure on tab.
         * If there is any invalid control or section within the tab we show a red 'o' on the right of the tab header.
         */
        showValidationStatusOnTabs?: boolean;
    }
    /**
     * Tab control view model
     */
    export interface Contract extends MsPortalFx.Control {
        /**
         * The section view models containing the contents of each tab.
         */
        readonly tabs: KnockoutObservableArray<SectionContract>;
        /**
         * Gets and sets the currently active tab.
         */
        readonly activeTab: KnockoutObservableBase<SectionContract>;
        /**
         * Show required asterisks on tab.
         * If there is any required control or section within the tab we show a red '*' on the left of the tab header.
         */
        readonly showRequiredStatusOnTabs: KnockoutObservableBase<boolean>;
        /**
         * Show validation failure on tab.
         * If there is any invalid control or section within the tab we show a red 'o' on the right of the tab header.
         */
        readonly showValidationStatusOnTabs: KnockoutObservableBase<boolean>;
    }
    /**
     * Creates a tab control view model
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\TextBlock.d.ts
declare module "Fx/Controls/TextBlock" {
    /**
     * Options for configuring a TextBlock control
     */
    export interface Options {
        /**
         * Text shown in the TextBlock
         */
        text?: KnockoutObservableBase<string> | string;
    }
    /**
     * TextBlock control view model
     */
    export interface Contract extends MsPortalFx.Control {
        /**
         * Text shown in the TextBlock
         */
        readonly text: KnockoutObservableBase<string>;
    }
    /**
     * Creates a TextBlock control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a TextBlock viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\TextBox.d.ts
declare module "Fx/Controls/TextBox" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export import ValueUpdateTrigger = Base.ValueUpdateTrigger;
    /**
     * Options for configuring a textbox control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * The text shown when no value is set in the control.
         */
        placeHolderText?: string | KnockoutObservableBase<string>;
        /**
         * When set, the textbox will be read only.
         */
        readonly?: boolean | KnockoutObservableBase<boolean>;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * Controls when the control will update it's value in response to the user typing into it.
         */
        valueUpdateTrigger?: ValueUpdateTrigger;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: string | KnockoutObservableBase<string>;
        /**
         * Aria label for the textbox.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
        /**
         * Shows validation as a popup.
         */
        showValidationsAsPopup?: boolean;
        /**
         * Enable the textbox to be spellcheck. Default as false
         * (Warning. Perf hit. Do not enable this for a huge text).
         */
        spellcheck?: boolean | KnockoutObservableBase<boolean>;
        /**
         * Event is triggered when user presses the enter key.
         */
        onEnterPressed?: (value: string) => void;
    }
    /**
     * TextBox control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<string>;
        /**
         * The text shown when no value is set in the control.
         */
        readonly placeHolderText: KnockoutObservableBase<string>;
        /**
         * When set, the textbox will be read only.
         */
        readonly readonly: KnockoutObservableBase<boolean>;
        /**
         * Event is triggerred when user presses the enter key.
         */
        readonly onEnterPressed: (value: string) => void;
        /**
         * Enable the textbox to be spellcheck. Default as false
         * (Warning. Perf hit. Do not enable this for a huge text).
         */
        readonly spellcheck: KnockoutObservableBase<boolean>;
    }
    /**
     * Creates a TextBox control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a textbox viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\TimePicker.d.ts
declare module "Fx/Controls/TimePicker" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    export import DateTimeRange = Base.DateTimeRange;
    /**
     * Options for configuring a time picker control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * Aria label for the DatePicker.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * Date/time range in which user is able to select date/time.
         */
        enabledDateTimeRange?: DateTimeRange | KnockoutObservableBase<DateTimeRange> | DateTimeRange[] | KnockoutObservableArray<DateTimeRange>;
        /**
         * Whether the control should allow empty(null) value as valid.
         */
        allowEmpty?: boolean;
        /**
         * The text shown when no value is set in the control.
         */
        placeHolderText?: string | KnockoutObservableBase<string>;
        /**
         * If true displays hours, minutes, and seconds. If false displays only hours and minutes.
         */
        showSeconds?: boolean;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: Date | KnockoutObservableBase<Date>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
    }
    /**
     * TimePicker control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<Date>;
        /**
         * Date/time range in which user is able to select date/time.
         */
        readonly enabledDateTimeRange: KnockoutObservableBase<DateTimeRange> | KnockoutObservableArray<DateTimeRange>;
        /**
         * The text shown when no value is set in the control.
         */
        readonly placeHolderText: KnockoutObservableBase<string>;
    }
    /**
     * Creates a time picker control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a time picker control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\TimeZones.d.ts
declare module "Fx/Controls/TimeZones" {
    /**
     * The timezone ID (from Windows OS) for 'Dateline Standard Time'.
     */
    export const DatelineStandardTime = "Dateline Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'UTC-11'.
     */
    export const UTCMinus11 = "UTC-11";
    /**
     * The timezone ID (from Windows OS) for 'Aleutian Standard Time'.
     */
    export const AleutianStandardTime = "Aleutian Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Hawaiian Standard Time'.
     */
    export const HawaiianStandardTime = "Hawaiian Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Marquesas Standard Time'.
     */
    export const MarquesasStandardTime = "Marquesas Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Alaskan Standard Time'.
     */
    export const AlaskanStandardTime = "Alaskan Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'UTC-09'.
     */
    export const UTCMinus09 = "UTC-09";
    /**
     * The timezone ID (from Windows OS) for 'Pacific Standard Time (Mexico)'.
     */
    export const PacificStandardTimeMexico = "Pacific Standard Time (Mexico)";
    /**
     * The timezone ID (from Windows OS) for 'UTC-08'.
     */
    export const UTCMinus08 = "UTC-08";
    /**
     * The timezone ID (from Windows OS) for 'Pacific Standard Time'.
     */
    export const PacificStandardTime = "Pacific Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'US Mountain Standard Time'.
     */
    export const USMountainStandardTime = "US Mountain Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Mountain Standard Time (Mexico)'.
     */
    export const MountainStandardTimeMexico = "Mountain Standard Time (Mexico)";
    /**
     * The timezone ID (from Windows OS) for 'Mountain Standard Time'.
     */
    export const MountainStandardTime = "Mountain Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Central America Standard Time'.
     */
    export const CentralAmericaStandardTime = "Central America Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Central Standard Time'.
     */
    export const CentralStandardTime = "Central Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Easter Island Standard Time'.
     */
    export const EasterIslandStandardTime = "Easter Island Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Central Standard Time (Mexico)'.
     */
    export const CentralStandardTimeMexico = "Central Standard Time (Mexico)";
    /**
     * The timezone ID (from Windows OS) for 'Canada Central Standard Time'.
     */
    export const CanadaCentralStandardTime = "Canada Central Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'SA Pacific Standard Time'.
     */
    export const SAPacificStandardTime = "SA Pacific Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Eastern Standard Time (Mexico)'.
     */
    export const EasternStandardTimeMexico = "Eastern Standard Time (Mexico)";
    /**
     * The timezone ID (from Windows OS) for 'Eastern Standard Time'.
     */
    export const EasternStandardTime = "Eastern Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Haiti Standard Time'.
     */
    export const HaitiStandardTime = "Haiti Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Cuba Standard Time'.
     */
    export const CubaStandardTime = "Cuba Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'US Eastern Standard Time'.
     */
    export const USEasternStandardTime = "US Eastern Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Turks And Caicos Standard Time'.
     */
    export const TurksAndCaicosStandardTime = "Turks And Caicos Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Paraguay Standard Time'.
     */
    export const ParaguayStandardTime = "Paraguay Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Atlantic Standard Time'.
     */
    export const AtlanticStandardTime = "Atlantic Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Venezuela Standard Time'.
     */
    export const VenezuelaStandardTime = "Venezuela Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Central Brazilian Standard Time'.
     */
    export const CentralBrazilianStandardTime = "Central Brazilian Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'SA Western Standard Time'.
     */
    export const SAWesternStandardTime = "SA Western Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Pacific SA Standard Time'.
     */
    export const PacificSAStandardTime = "Pacific SA Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Newfoundland Standard Time'.
     */
    export const NewfoundlandStandardTime = "Newfoundland Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Tocantins Standard Time'.
     */
    export const TocantinsStandardTime = "Tocantins Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'E. South America Standard Time'.
     */
    export const ESouthAmericaStandardTime = "E. South America Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'SA Eastern Standard Time'.
     */
    export const SAEasternStandardTime = "SA Eastern Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Argentina Standard Time'.
     */
    export const ArgentinaStandardTime = "Argentina Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Greenland Standard Time'.
     */
    export const GreenlandStandardTime = "Greenland Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Montevideo Standard Time'.
     */
    export const MontevideoStandardTime = "Montevideo Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Magallanes Standard Time'.
     */
    export const MagallanesStandardTime = "Magallanes Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Saint Pierre Standard Time'.
     */
    export const SaintPierreStandardTime = "Saint Pierre Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Bahia Standard Time'.
     */
    export const BahiaStandardTime = "Bahia Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'UTC-02'.
     */
    export const UTCMinus02 = "UTC-02";
    /**
     * The timezone ID (from Windows OS) for 'Mid-Atlantic Standard Time'.
     */
    export const MidMinusAtlanticStandardTime = "Mid-Atlantic Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Azores Standard Time'.
     */
    export const AzoresStandardTime = "Azores Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Cape Verde Standard Time'.
     */
    export const CapeVerdeStandardTime = "Cape Verde Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'UTC'.
     */
    export const UTC = "UTC";
    /**
     * The timezone ID (from Windows OS) for 'Morocco Standard Time'.
     */
    export const MoroccoStandardTime = "Morocco Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'GMT Standard Time'.
     */
    export const GMTStandardTime = "GMT Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Greenwich Standard Time'.
     */
    export const GreenwichStandardTime = "Greenwich Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'W. Europe Standard Time'.
     */
    export const WEuropeStandardTime = "W. Europe Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Central Europe Standard Time'.
     */
    export const CentralEuropeStandardTime = "Central Europe Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Romance Standard Time'.
     */
    export const RomanceStandardTime = "Romance Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Central European Standard Time'.
     */
    export const CentralEuropeanStandardTime = "Central European Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'W. Central Africa Standard Time'.
     */
    export const WCentralAfricaStandardTime = "W. Central Africa Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Jordan Standard Time'.
     */
    export const JordanStandardTime = "Jordan Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'GTB Standard Time'.
     */
    export const GTBStandardTime = "GTB Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Middle East Standard Time'.
     */
    export const MiddleEastStandardTime = "Middle East Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Egypt Standard Time'.
     */
    export const EgyptStandardTime = "Egypt Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'E. Europe Standard Time'.
     */
    export const EEuropeStandardTime = "E. Europe Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Syria Standard Time'.
     */
    export const SyriaStandardTime = "Syria Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'West Bank Standard Time'.
     */
    export const WestBankStandardTime = "West Bank Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'South Africa Standard Time'.
     */
    export const SouthAfricaStandardTime = "South Africa Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'FLE Standard Time'.
     */
    export const FLEStandardTime = "FLE Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Israel Standard Time'.
     */
    export const IsraelStandardTime = "Israel Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Kaliningrad Standard Time'.
     */
    export const KaliningradStandardTime = "Kaliningrad Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Sudan Standard Time'.
     */
    export const SudanStandardTime = "Sudan Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Libya Standard Time'.
     */
    export const LibyaStandardTime = "Libya Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Namibia Standard Time'.
     */
    export const NamibiaStandardTime = "Namibia Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Arabic Standard Time'.
     */
    export const ArabicStandardTime = "Arabic Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Turkey Standard Time'.
     */
    export const TurkeyStandardTime = "Turkey Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Arab Standard Time'.
     */
    export const ArabStandardTime = "Arab Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Belarus Standard Time'.
     */
    export const BelarusStandardTime = "Belarus Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Russian Standard Time'.
     */
    export const RussianStandardTime = "Russian Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'E. Africa Standard Time'.
     */
    export const EAfricaStandardTime = "E. Africa Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Iran Standard Time'.
     */
    export const IranStandardTime = "Iran Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Arabian Standard Time'.
     */
    export const ArabianStandardTime = "Arabian Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Astrakhan Standard Time'.
     */
    export const AstrakhanStandardTime = "Astrakhan Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Azerbaijan Standard Time'.
     */
    export const AzerbaijanStandardTime = "Azerbaijan Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Russia Time Zone 3'.
     */
    export const RussiaTimeZone3 = "Russia Time Zone 3";
    /**
     * The timezone ID (from Windows OS) for 'Mauritius Standard Time'.
     */
    export const MauritiusStandardTime = "Mauritius Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Saratov Standard Time'.
     */
    export const SaratovStandardTime = "Saratov Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Georgian Standard Time'.
     */
    export const GeorgianStandardTime = "Georgian Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Caucasus Standard Time'.
     */
    export const CaucasusStandardTime = "Caucasus Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Afghanistan Standard Time'.
     */
    export const AfghanistanStandardTime = "Afghanistan Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'West Asia Standard Time'.
     */
    export const WestAsiaStandardTime = "West Asia Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Ekaterinburg Standard Time'.
     */
    export const EkaterinburgStandardTime = "Ekaterinburg Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Pakistan Standard Time'.
     */
    export const PakistanStandardTime = "Pakistan Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'India Standard Time'.
     */
    export const IndiaStandardTime = "India Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Sri Lanka Standard Time'.
     */
    export const SriLankaStandardTime = "Sri Lanka Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Nepal Standard Time'.
     */
    export const NepalStandardTime = "Nepal Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Central Asia Standard Time'.
     */
    export const CentralAsiaStandardTime = "Central Asia Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Bangladesh Standard Time'.
     */
    export const BangladeshStandardTime = "Bangladesh Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Omsk Standard Time'.
     */
    export const OmskStandardTime = "Omsk Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Myanmar Standard Time'.
     */
    export const MyanmarStandardTime = "Myanmar Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'SE Asia Standard Time'.
     */
    export const SEAsiaStandardTime = "SE Asia Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Altai Standard Time'.
     */
    export const AltaiStandardTime = "Altai Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'W. Mongolia Standard Time'.
     */
    export const WMongoliaStandardTime = "W. Mongolia Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'North Asia Standard Time'.
     */
    export const NorthAsiaStandardTime = "North Asia Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'N. Central Asia Standard Time'.
     */
    export const NCentralAsiaStandardTime = "N. Central Asia Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Tomsk Standard Time'.
     */
    export const TomskStandardTime = "Tomsk Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'China Standard Time'.
     */
    export const ChinaStandardTime = "China Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'North Asia East Standard Time'.
     */
    export const NorthAsiaEastStandardTime = "North Asia East Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Singapore Standard Time'.
     */
    export const SingaporeStandardTime = "Singapore Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'W. Australia Standard Time'.
     */
    export const WAustraliaStandardTime = "W. Australia Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Taipei Standard Time'.
     */
    export const TaipeiStandardTime = "Taipei Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Ulaanbaatar Standard Time'.
     */
    export const UlaanbaatarStandardTime = "Ulaanbaatar Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'North Korea Standard Time'.
     */
    export const NorthKoreaStandardTime = "North Korea Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Aus Central W. Standard Time'.
     */
    export const AusCentralWStandardTime = "Aus Central W. Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Transbaikal Standard Time'.
     */
    export const TransbaikalStandardTime = "Transbaikal Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Tokyo Standard Time'.
     */
    export const TokyoStandardTime = "Tokyo Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Korea Standard Time'.
     */
    export const KoreaStandardTime = "Korea Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Yakutsk Standard Time'.
     */
    export const YakutskStandardTime = "Yakutsk Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Cen. Australia Standard Time'.
     */
    export const CenAustraliaStandardTime = "Cen. Australia Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'AUS Central Standard Time'.
     */
    export const AUSCentralStandardTime = "AUS Central Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'E. Australia Standard Time'.
     */
    export const EAustraliaStandardTime = "E. Australia Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'AUS Eastern Standard Time'.
     */
    export const AUSEasternStandardTime = "AUS Eastern Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'West Pacific Standard Time'.
     */
    export const WestPacificStandardTime = "West Pacific Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Tasmania Standard Time'.
     */
    export const TasmaniaStandardTime = "Tasmania Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Vladivostok Standard Time'.
     */
    export const VladivostokStandardTime = "Vladivostok Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Lord Howe Standard Time'.
     */
    export const LordHoweStandardTime = "Lord Howe Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Bougainville Standard Time'.
     */
    export const BougainvilleStandardTime = "Bougainville Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Russia Time Zone 10'.
     */
    export const RussiaTimeZone10 = "Russia Time Zone 10";
    /**
     * The timezone ID (from Windows OS) for 'Magadan Standard Time'.
     */
    export const MagadanStandardTime = "Magadan Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Norfolk Standard Time'.
     */
    export const NorfolkStandardTime = "Norfolk Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Sakhalin Standard Time'.
     */
    export const SakhalinStandardTime = "Sakhalin Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Central Pacific Standard Time'.
     */
    export const CentralPacificStandardTime = "Central Pacific Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Russia Time Zone 11'.
     */
    export const RussiaTimeZone11 = "Russia Time Zone 11";
    /**
     * The timezone ID (from Windows OS) for 'New Zealand Standard Time'.
     */
    export const NewZealandStandardTime = "New Zealand Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'UTC+12'.
     */
    export const UTCPlus12 = "UTC+12";
    /**
     * The timezone ID (from Windows OS) for 'Fiji Standard Time'.
     */
    export const FijiStandardTime = "Fiji Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Kamchatka Standard Time'.
     */
    export const KamchatkaStandardTime = "Kamchatka Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Chatham Islands Standard Time'.
     */
    export const ChathamIslandsStandardTime = "Chatham Islands Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'UTC+13'.
     */
    export const UTCPlus13 = "UTC+13";
    /**
     * The timezone ID (from Windows OS) for 'Tonga Standard Time'.
     */
    export const TongaStandardTime = "Tonga Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Samoa Standard Time'.
     */
    export const SamoaStandardTime = "Samoa Standard Time";
    /**
     * The timezone ID (from Windows OS) for 'Line Islands Standard Time'.
     */
    export const LineIslandsStandardTime = "Line Islands Standard Time";
    /**
     * The fake timezone ID for the current (pseudo) timezone.
     */
    export const Current = "Current Time";
}

// FILE: Fx\Controls\Toolbar\MoveResourceToolbarButton.d.ts
declare module "Fx/Controls/Toolbar/MoveResourceToolbarButton" {
    export = Main;
    module Main {
        /**
         * This module is obsolete. Please use "Fx/Controls/MoveResourceToolbarButton"
         */
        type Options = Obsolete;
        /**
         * This module is obsolete. Please use "Fx/Controls/MoveResourceToolbarButton"
         */
        const ViewModel: Obsolete;
    }
}

// FILE: Fx\Controls\TriStateCheckBox.d.ts
declare module "Fx/Controls/TriStateCheckBox" {
    import * as Base from "Fx/Controls/FormBase";
    import { Validation } from "Fx/Controls/Validations";
    /**
     * CheckBoxValue value states.
     */
    export const enum Value {
        /**
         * TriStateCheckBox state representing unchecked state.
         */
        Unchecked = 0,
        /**
         * TriStateCheckBox state representing checked state.
         */
        Checked = 1,
        /**
         * TriStateCheckBox state representing indeterminate state.
         */
        Indeterminate = 2,
    }
    /**
     * Options for configuring a tri-state textbox control
     */
    export interface Options {
        /**
         * The label of the control
         */
        label?: string | KnockoutObservableBase<string>;
        /**
         * The sublabel of the control
         */
        subLabel?: string | KnockoutObservableBase<string>;
        /**
         * The aria-label on the control.
         */
        ariaLabel?: string | KnockoutObservableBase<string>;
        /**
         * validations on the control
         */
        validations?: Validation[] | KnockoutObservableArray<Validation>;
        /**
         * sanitized html string shown in the info balloon popup
         */
        infoBalloonContent?: string | KnockoutObservableBase<string>;
        /**
         * if set to true, the label will be placed to the right of the control
         */
        labelOnRight?: boolean;
        /**
         * if set to true, the user can set the indeterminate state of the control
         */
        userCanSetIndeterminate?: boolean;
        /**
         * If set, the control will never be in a dirty state.  This means it will not display dirty highlighting, nor will it trigger an alert when blades are closed.  Validation is not affected.
         */
        suppressDirtyBehavior?: boolean;
        /**
         * The disabled state of the control.
         */
        disabled?: boolean | KnockoutObservableBase<boolean>;
        /**
         * The initial value of the control.  If an observable is provided, there will be a 2 way binding set up between the observable and the value of the control.
         */
        value?: Value | KnockoutObservableBase<Value>;
        /**
         * The visible state of the control.
         */
        visible?: boolean | KnockoutObservableBase<boolean>;
        /**
         * A custom css class to apply to the control.
         */
        cssClass?: string | KnockoutObservableBase<string>;
    }
    /**
     * TriStateCheckBox control view model
     */
    export interface Contract extends Base.Contract {
        /**
         * The value of the control
         */
        readonly value: KnockoutObservableBase<Value>;
    }
    /**
     * Creates a tristate checkbox control viewmodel.
     *
     * @param lifetime The lifetime of the control
     * @param options Options that modify the appearance and behavior of the control.
     *
     * @returns a tristate checkbox control viewmodel
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
}

// FILE: Fx\Controls\Validations.d.ts
declare module "Fx/Controls/Validations" {
    export import CaseInsensitiveComparison = MsPortalFx.ViewModels.CaseInsensitiveComparisonValidation;
    export import Contains = MsPortalFx.ViewModels.ContainsValidation;
    export import ContainsCharacters = MsPortalFx.ViewModels.ContainsCharactersValidation;
    export import Custom = MsPortalFx.ViewModels.CustomValidation;
    export import CustomV = MsPortalFx.ViewModels.CustomValidationV;
    export import Equals = MsPortalFx.ViewModels.EqualsValidation;
    export import HasDigit = MsPortalFx.ViewModels.HasDigitValidation;
    export import HasLetter = MsPortalFx.ViewModels.HasLetterValidation;
    export import HasLowerCaseLetter = MsPortalFx.ViewModels.HasLowerCaseLetterValidation;
    export import HasPunctuation = MsPortalFx.ViewModels.HasPunctuationValidation;
    export import HasUpperCaseLetter = MsPortalFx.ViewModels.HasUpperCaseLetterValidation;
    export import Invalid = MsPortalFx.ViewModels.InvalidValidation;
    export import LengthRange = MsPortalFx.ViewModels.LengthRangeValidation;
    export import LocaleAwareCaseInsensitiveComparison = MsPortalFx.ViewModels.LocaleAwareCaseInsensitiveComparisonValidation;
    export import MaxLength = MsPortalFx.ViewModels.MaxLengthValidation;
    export import MaxValue = MsPortalFx.ViewModels.MaxValueValidation;
    export import MinLength = MsPortalFx.ViewModels.MinLengthValidation;
    export import MinValue = MsPortalFx.ViewModels.MinValueValidation;
    export import NotContains = MsPortalFx.ViewModels.NotContainsValidation;
    export import NotContainsCharacters = MsPortalFx.ViewModels.NotContainsCharactersValidation;
    export import NotRegExMatch = MsPortalFx.ViewModels.NotRegExMatchValidation;
    export import Numeric = MsPortalFx.ViewModels.NumericValidation;
    export import Range = MsPortalFx.ViewModels.RangeValidation;
    export import RegExMatch = MsPortalFx.ViewModels.RegExMatchValidation;
    export import Required = MsPortalFx.ViewModels.RequiredValidation;
    export import StaticMessage = MsPortalFx.ViewModels.StaticMessageValidation;
    export import Uri = MsPortalFx.ViewModels.UriValidation;
    export import Validation = MsPortalFx.ViewModels.FormValidation;
    export import ValidationResult = MsPortalFx.ViewModels.ValidationResult;
    export import ValidationState = MsPortalFx.ViewModels.Controls.Validators.ValidationState;
    export import ValidationType = MsPortalFx.ViewModels.FormValidationType;
}

// FILE: Fx\Controls\Video.d.ts
declare module "Fx/Controls/Video" {
    /**
     * Options that can be set on the Video control
     */
    export interface Options {
        /**
         * Source of the video
         */
        src?: KnockoutObservableBase<string> | string;
        /**
         * Play video immediatly after load
         */
        autoplay?: KnockoutObservableBase<boolean> | boolean;
        /**
         * True if current video type supports autoplay
         */
        canAutoplay?: KnockoutReadOnlyObservable<boolean> | boolean;
        /**
         * On first play callback function
         */
        onFirstPlay?: KnockoutObservableBase<Callback> | Callback;
    }
    /**
     * Video control view model
     */
    export interface Contract extends MsPortalFx.Control {
        /**
         * URL of the desired video (Supports HTML5 video, YouTube, Channel9, TedxTalks)
         */
        src: KnockoutObservableBase<string>;
        /**
         * Play video immediatly after load (only works for HTML5 and YouTube)
         */
        autoplay: KnockoutObservableBase<boolean>;
        /**
         * True if current video type supports autoplay
         */
        canAutoplay: KnockoutReadOnlyObservable<boolean>;
        /**
         * Called when the video is played in first time (only works for HTML5, YouTube and Channel9)
         */
        onFirstPlay: KnockoutObservableBase<Callback>;
    }
    /**
     * Callback function type for events
     */
    export type Callback = () => void;
    /**
     * Creates a Video control viewmodel.
     *
     * @param lifetime The lifetime for the created view model.
     * @param options The set of options to configure the Video view model.
     *
     * @returns a Video view model
     */
    export function create(lifetime: MsPortalFx.Base.LifetimeManager, options?: Options): Contract;
    /**
     * Obsolete.  Use the create method to create viewmodels instead.
     */
    export let ViewModel: Obsolete;
}

// FILE: Fx\Cryptography.d.ts
declare module "Fx/Cryptography" {
    import FxBase = MsPortalFx.Base;
    import FxRpc = FxImpl.Rpc;
    import FxPromiseV = FxBase.PromiseV;
    export module Internal {
        const getHmacSha256EndPoint: FxRpc.FuncEndPointDefinition<GetHmacSha256Options, string>;
        /**
         * Returns an hmac sha256 token generated by hashing the provided string with the specified key.
         *
         * @param options Parameters required by the function.
         * @param client The rpc client from which this call originates. Defaults to the current iFrame.
         *  Note: the default behavior should only be overridden for testing.
         * @return A promise that resolves to a hmac sha256 hashed string.
         */
        function getHmacSha256(options: GetHmacSha256Options, client?: FxRpc.Client): FxPromiseV<string>;
    }
    /**
     * Options that must be passed when calling the getHmacSha256 function.
     */
    export interface GetHmacSha256Options {
        /**
         * The string to hash, can be any length.
         */
        stringToHash: string;
        /**
         * The key, encoded in base 64.
         */
        key: string;
    }
    /**
     * HmacSha256 RPC endpoint response interface
     */
    export interface HmacSha256ResponseModel {
        data: string;
    }
    /**
     * Returns an hmac sha256 token generated by hashing the provided string with the specified key.
     *
     * @param options Parameters required by the function.
     * @return A promise that resolves to a hmac sha256 hashed string.
     */
    export function getHmacSha256(options: GetHmacSha256Options): FxBase.PromiseV<string>;
}

// FILE: Fx\Events.d.ts
declare module "Fx/Events" {
    /**
     * A reference to a DOM element in the portal
     */
    export import FxElement = MsPortalFx.ViewModels.FxElement;
    /**
     * Information describing a event generated by a mouse action
     */
    export import FxMouseEvent = MsPortalFx.ViewModels.FxMouseEvent;
}

// FILE: Fx\Extension.d.ts
declare module "Fx/Extension" {
    export = MsPortalFxExtension;
    module MsPortalFxExtension {
        /**
         * The @Extension decorator.  Identifies a class that defines the name of the extension and optionally defines a constructor that
         * initializes the extension.
         * This decorator can only be defined once in the extension package.
         */
        function Decorator(): (constructor: Function) => void;
    }
}

// FILE: Fx\Feedback.d.ts
declare module "Fx/Feedback" {
    export = Main;
    module Main {
        import FxBase = MsPortalFx.Base;
        import Rpc = FxImpl.Rpc;
        interface NpsDetails {
            /**
             * The display name of the product to be shown in the NPS survey.
             * The question will be of the form:
             *     How likely are you to recommend {your product name} to a friend or colleague?
             */
            productDisplayName: string;
            /**
             * The product id that will be logged in the portal telemetry table.
             */
            productId: string;
        }
        module Internal {
            /**
             * RPC endpoint to show survey.
             */
            const showNpsEndpoint: Rpc.FuncEndPointDefinition<NpsDetails, void>;
            /**
             * RPC endpoint to check if survey can be shown.
             */
            const canShowNpsEndpoint: Rpc.FuncEndPointDefinition<void, boolean>;
        }
        /**
         * Requests an NPS toast to be displayed. The response to the feedback will be recorded in the portal telemetry tables.
         *
         * Note that the NPS survey will only be shown once per session. Calling this API does not guarantee that the survey will be shown.
         *
         * @param params NPS parameters.
         */
        function showNps(params: NpsDetails): void;
        /**
         * Checks if an NPS survey can be shown. If this returns true, extensions can call the showNPSSurvey to show NPS.
         * If this returns false, an NPS survey has already been shown and calling the showNPSSurvey will result in a no-op.
         *
         */
        function canShowNps(): FxBase.PromiseV<boolean>;
        /**
         * DEPRECATED - no longer opens a feedback pane
         */
        var openResourceDeletedFeedbackPane: Obsolete;
    }
}

// FILE: Fx\NotifyResourcesCreated.d.ts
declare module "Fx/NotifyResourcesCreated" {
    export = Main;
    module Main {
        /**
         * The resource type whose set of resources for the user has created.
         */
        const enum CreatedResourcesType {
            /**
             * Subscriptions.
             */
            Subscriptions = 1,
        }
        /**
         * Notify FX that the set of resources of some type have created.
         *
         * @param createdResourcesType  The resource type whose set of resources for the user has created.
         */
        function notifyResourcesCreated(createdResourcesType: CreatedResourcesType): void;
    }
}

// FILE: Fx\Pinner.d.ts
declare module "Fx/Pinner" {
    export = Main;
    module Main {
        import FxBase = MsPortalFx.Base;
        import FxRpc = FxImpl.Rpc;
        import FxComposition = MsPortalFx.Composition;
        /**
         * Options that can be passed when calling the pin function.
         */
        interface Options {
            /**
             * The value to indicate whether to show a notification or not.
             */
            notify: boolean;
        }
        module Internal {
            /**
             * RPC endpoint to pin provided parts.
             */
            const pinEndPoint: FxRpc.FuncEndPointDefinition<PinParams, string>;
            /**
             * Options that must be passed when calling the pin function.
             */
            interface PinParams {
                /**
                 * Parts to pin to the dashboard.
                 */
                parts: FxComposition.PartReference<any>[];
                /**
                 * The value to indicate whether to show a notification or not.
                 */
                options?: Options;
            }
        }
        /**
         * Pins the specified parts to the currently opened dashboard.
         *
         * @param parts Parts to pin to the dashboard.
         * @param options Optional arguments used by the function.
         * @return A promise that resolves to completion of pinning.
         */
        function pin(parts: FxComposition.PartReference<any>[], options?: Options): FxBase.Promise;
    }
}

// FILE: Fx\ResourceManagement\Marketplace.d.ts
declare module "Fx/ResourceManagement/Marketplace" {
    /**
     * Marketplace offer plan.
     */
    export interface OfferPlan {
        /**
         * The plan id.
         */
        planId: string;
        /**
         * The plan display name.
         */
        displayName: string;
        /**
         * The summary text for the plan.
         */
        summary: string;
        /**
         * The description HTML for the plan.
         */
        description: string;
    }
    /**
     * Marketplace offer pricing details model.
     * Used to retrieve the pricing information for a Marketplace offer.
     */
    export interface OfferPricingDetails {
        /**
         * The offer id.
         */
        offerId: string;
        /**
         * The publisher id.
         */
        publisherId: string;
        /**
         * The offer plans provided by the publisher.
         */
        plans: OfferPlan[];
    }
    /**
     * Marketplace product (offer).
     */
    export interface Product {
        /**
         * The product display name.
         */
        displayName: string;
        /**
         * The publisher display name.
         */
        publisherDisplayName: string;
        /**
         * The URI to the legal terms HTML.
         */
        legalTermsUri: string;
        /**
         * The URI to the privacy policy HTML.
         */
        privacyPolicyUri: string;
        /**
         * The other pricing details URI.
         */
        pricingDetailsUri: string;
        /**
         * The offer pricing details.
         */
        offerDetails?: OfferPricingDetails;
    }
    /**
     * Marketplace artifact.
     */
    export interface Artifact {
        /**
         * The artifact name.
         */
        name: string;
        /**
         * The URI to the artifact file.
         */
        uri: string;
        /**
         * The artifact type.
         */
        type: string;
    }
    /**
     * The context from which a marketplace create is kicked off.
     */
    export interface LaunchingContext extends StringMap<any> {
        /**
         * The gallery item id.
         */
        galleryItemId: string;
        /**
         * The source entity launching the create flow (blade name, control, etc.). Used for telemetry logging.
         */
        source: string[];
        /**
         * The marketplace menu item id.
         */
        menuItemId?: string;
        /**
         * The marketplace sub menu item id.
         */
        subMenuItemId?: string;
        /**
         * The blade instance id.
         */
        bladeInstanceId?: string;
    }
    /**
     * Marketplace item.
     */
    export interface Item<TUIMetaData> {
        /**
         * The Marketplace item id.
         */
        id: string;
        /**
         * The item display name.
         */
        itemDisplayName: string;
        /**
         * The publisher display name.
         */
        publisherDisplayName: string;
        /**
         * The Marketplace item version.
         */
        version: string;
        /**
         * The list of category ids the Marketplace item belongs to.
         */
        categoryIds: string[];
        /**
         * The products associated with the Marketplace item.
         */
        products: Product[];
        /**
         * Marketplace item products with no pricing information.
         */
        productsWithNoPricing: Product[];
        /**
         * The artifacts associated with the Marketplace item.
         */
        specialArtifacts?: Artifact[];
        /**
         * The dictionary of metadata properties to be used by the extension.
         */
        metadata?: StringMap<string>;
        /**
         * The deployment name.
         */
        deploymentName: string;
        /**
         * The list of URIs for the CSM template files.
         */
        deploymentTemplateFileUris: StringMap<string>;
        /**
         * The list of URIs for the deployment fragments.
         */
        deploymentFragmentFileUris?: StringMap<string>;
        /**
         * The context from which a marketplace create is kicked off.
         */
        launchingContext: LaunchingContext;
        /**
         *
         */
        uiMetadata: TUIMetaData;
    }
    /**
     * The interface of context supplied by marketplace
     */
    export interface Context<TUIMetaData> {
        /**
         * The telemetry id; a GUID unique to each instance of the provisioning flow initiated by
         * the user (i.e. unique to each instance when the blade is launched). The same id is used
         * when the 'CreateFlowLaunched, 'ProvisioningStart/Ended' and 'CreateDeploymentStart/End'
         * events are logged. Adding this telemetry id to the telemetry logged on the blade will
         * help you connect all the data points for a given provisioning instance.
         */
        readonly telemetryId: string;
        /**
         * The Marketplace item invoking the blade. Will be undefined if 'requiresMarketplaceId'
         * is set to false on the @DoesProvisioning decorator options.
         */
        readonly marketplaceItem?: Item<TUIMetaData>;
    }
}

// FILE: Fx\ResourceManagement\Provisioning.d.ts
declare module "Fx/ResourceManagement/Provisioning" {
    import { LaunchingContext } from "Fx/ResourceManagement/Marketplace";
    import PartReference = MsPortalFx.Composition.PartReference;
    import FxResourceManager = MsPortalFx.Azure.ResourceManager;
    /**
     * The template deployment operation mode. Defaults to 'RequestDeploymentOnly'.
     */
    export import TemplateDeploymentMode = FxResourceManager.TemplateDeploymentMode;
    /**
     * Initial values for form initialization. Use those values to initialize the subscription,
     * resource group, and location drop down controls.
     */
    export interface InitialValues {
        /**
         * The list of subscription ids last used by the user.
         */
        readonly subscriptionIds?: string[];
        /**
         * The list of location names last used by the user.
         */
        readonly locationNames?: string[];
    }
    /**
     * Options for the DeployTemplate method on provisioning context
     */
    export interface DeployTemplateOptions {
        /**
         * The subscription id.
         */
        subscriptionId: string;
        /**
         * The deployment name.
         */
        deploymentName: string;
        /**
         * The resource group name.
         */
        resourceGroupName: string;
        /**
         * The resource id. Supply this to link the notifications to the asset or if the deployment
         * results in a startboard part.
         */
        resourceId?: string;
        /**
         * The context from which a gallery create is kicked off. Used for telemetry logging.
         */
        launchingContext?: LaunchingContext;
        /**
         * Debug info.
         */
        debug?: string;
        /**
         * The location/region.
         */
        resourceGroupLocation: string;
        /**
         * An array of the resource providers to be registered for the subscription.
         */
        resourceProviders: string[];
        /**
         * The parameters for the template deployment (name and value pairs).
         */
        parameters?: StringMapPrimitive;
        /**
         * The reference parameters for the template deployment.
         */
        referenceParameters?: StringMap<StringMapPrimitive>;
        /**
         * The URI for the parameters file. Use this to link to an existing parameters file. Specify
         * this or the parameters and/or referenceParameters properties, but not both.
         */
        parametersLinkUri?: string;
        /**
         * The URI for the ARM template. Specify this or the templateJson property, but not both.
         */
        templateLinkUri?: string;
        /**
         * The inline deployment template JSON. Specify this or the templateLinkUri property, but not both.
         */
        templateJson?: string;
        /**
         * The template deployment operation mode. Defaults to 'RequestDeploymentOnly'.
         */
        deploymentMode?: TemplateDeploymentMode;
        /**
         * Flag indicating that we should run ARM's preflight validation before submitting the template
         * deployment request to ARM. Defaults to false.
         */
        validateTemplate?: boolean;
        /**
         * Function to provide a part reference based on the resourceId of a deployment.
         * Defaults to the part reference provided by the marketplace UI.Definition file
         * or null if no marketplace item was provieded to this provisioning blade.
         *
         * @param resourceId The resourceId of the resource created
         */
        supplyPartReference?(resourceId: string): PartReference<any>;
    }
    /**
     * Options for the DeployCustom method on provisioning context
     */
    export interface DeployCustomOptions<TResult> {
        /**
         * A promise for when provisioning has finished
         */
        provisioningPromise: Q.Promise<TResult>;
        /**
         * Function to provide a part reference based on the resourceId of a deployment.
         * Defaults to the part reference provided by the marketplace UI.Definition file
         * or null if no marketplace item was provieded to this provisioning blade.
         * @param resource The result of the provisioning promise
         * @returns a part reference to pin to dashboard
         */
        supplyPartReference?(result: TResult): PartReference<any>;
    }
}

// FILE: Fx\Specs\DropDown.d.ts
declare module "Fx/Specs/DropDown" {
    import { AnyBladeContainer } from "Fx/Composition/BladeBase";
    import { ParameterProviderBladeReference, ParameterProviderOptions } from "Fx/Composition/Selectable";
    export = Main;
    module Main {
        import FxViewModels = MsPortalFx.ViewModels;
        import Specs = MsPortalFx.Azure.ResourceManager.Pickers.Specs;
        import FxForms = FxViewModels.Forms;
        import SpecPicker = HubsExtension.Azure.SpecPicker;
        /**
         * A spec picker drop down which will show promotoed features of the specs available.
         */
        module DropDown {
            /**
             * Inteface to select the pricing blade on a spec drop down
             */
            interface PricingBladeSelection extends FxViewModels.DynamicBladeSelection {
                /**
                 * An identifying string for the hotspot. Must be unique from all other hotspots
                 * on the part so that the parameter provider can hook into the collector correctly
                 * advise name spacing as well e.g. "Spec.DropDown.Foo1"
                 */
                hotspot: string;
            }
            /**
             * Interface to a pricing blade reference class
             */
            interface PricingBladeReferenceClass {
                new (options: ParameterProviderOptions<any, any, any>): ParameterProviderBladeReference<any, any, any>;
            }
            /**
             * Base Options for the spec picker dropdown
             */
            interface BaseOptions extends FxForms.Base.Input.Options<string> {
                /**
                 * The spec picker extender viewmodel which will be used to display specs to be picked
                 */
                specPickerExtender: SpecPicker.SpecPickerExtender;
                /**
                 * A callback that supplies initial data for the parameter provider
                 * in the child blade each time it opens.
                 *
                 * Note that the object received by the parameter provider will be a
                 * deep clone of the value you give, rather than the original instance,
                 * because it is passed (and sometimes stored) in a serialized form.
                 *
                 * @return Initial data for the child blade.
                 */
                initialData: KnockoutObservableBase<Specs.InitialData>;
                /**
                 * The Form ViewModel for the selector control.
                 */
                form?: FxForms.Form.ViewModel<any>;
                /**
                 * The path to the property on the Form data model being set
                 * Used instead of accessor
                 */
                accessor?: FxForms.EditScopeAccessors<Specs.Result>;
                /**
                 * This will map the spec to the name in option for the dropdown
                 * defaults to spec code
                 */
                specNameMap?: (spec: SpecPicker.Spec) => string;
            }
            /**
             * Options for the spec picker dropdown
             */
            interface Options extends BaseOptions {
                /**
                 * When authoring your Blade using no-PDL (that is, TypeScript decorators),
                 * supply a code-genetrated BladeReference class.
                 * When authoring your Blade using PDL, supply 'PricingBladeSelection'.
                 */
                pricingBlade?: PricingBladeSelection;
            }
            /**
             * Options for the spec picker dropdown
             */
            interface OptionsWithBladeReference extends BaseOptions {
                /**
                 * When authoring your Blade using no-PDL (that is, TypeScript decorators),
                 * supply a code-genetrated BladeReference class.
                 * When authoring your Blade using PDL, supply 'PricingBladeSelection'.
                 */
                pricingBladeReference?: PricingBladeReferenceClass;
            }
            /**
             * This is a section which has on it a selectable on it.
             * This is to keep a consistent api with the spec selector
             * and infobox with the control having a selectable
             */
            interface SelectableSection extends FxForms.Section.ViewModel {
                /**
                 * The selectable on the section control to be bound to
                 * in the pdl
                 */
                selectable?: FxViewModels.Selectable<any>;
            }
        }
        /**
         * DropDown which creates a collector and selector that launches a spec picker blade
         */
        class DropDown {
            /**
             * ParameterCollector created to interface with the spec picker blade ParameterProvider
             */
            collector: FxViewModels.ParameterCollector<SpecPicker.ParameterCollectionV3.SpecPickerProviderCollectorParameter>;
            /**
             * The info box control.
             */
            control: DropDown.SelectableSection;
            /**
             * The selected spec(s).
             */
            value: KnockoutObservableBase<Specs.Result>;
            /**
             * Constructs a drop down for specs that can launch a specpicker blade
             *
             * @param container The container associated with the part or other composition item hosting this parameter collector.
             * @param DropDown options that extends both the colector and the BaseDropDownOptions
             */
            constructor(container: FxViewModels.ContainerContract | AnyBladeContainer, options: DropDown.Options | DropDown.OptionsWithBladeReference);
        }
    }
}

// FILE: Fx\Svg.d.ts
declare module "*.svg" {
    const _default: MsPortalFx.Base.Image;
    export = _default;
}
